/* s7, a Scheme interpreter
 *
 *    derived from:
 */

/* T I N Y S C H E M E    1 . 3 9
 *   Dimitrios Souflis (dsouflis@acm.org)
 *   Based on MiniScheme (original credits follow)
 * (MINISCM)               coded by Atsushi Moriwaki (11/5/1989)
 * (MINISCM)           E-MAIL :  moriwaki@kurims.kurims.kyoto-u.ac.jp
 * (MINISCM) This version has been modified by R.C. Secrist.
 * (MINISCM)
 * (MINISCM) Mini-Scheme is now maintained by Akira KIDA.
 * (MINISCM)
 * (MINISCM) This is a revised and modified version by Akira KIDA.
 * (MINISCM)	current version is 0.85k4 (15 May 1994)
 */


/* s7, Bill Schottstaedt, Aug-08
 *
 *   changes from tinyScheme:
 *        just two files: s7.c and s7.h, source-level embeddable (no library, no run-time init files)
 *        full continuations, call-with-exit for goto or return, dynamic-wind
 *        ratios and complex numbers (and ints are 64-bit by default)
 *          optional multiprecision arithmetic for all numeric types and functions
 *        generalized set!, procedure-with-setter, applicable objects
 *        defmacro and define-macro
 *        keywords, hash tables, block comments, define*
 *        format
 *        error handling using error and catch
 *        in sndlib, the run macro works giving s7 a (somewhat limited) byte compiler
 *        no invidious distinction between built-in and "foreign"
 *          (this makes it easy to extend built-in operators like "+" -- see s7.html for a simple example)
 *        lists, strings, vectors, hash-tables, and environments are (set-)applicable objects
 *        true multiple-values
 *        multidimensional vectors
 *        hooks (conditions)
 *
 *   differences from r5rs:
 *        no syntax-rules or any of its friends
 *        no force or delay
 *        no inexact integer or ratio (so, for example, truncate returns an exact integer), no exact complex or exact real
 *           (exact? has no obvious meaning in regard to complex numbers anyway -- are we referring to the polar or
 *            the rectangular form, and are both real and imaginary parts included? -- why can't they be separate?)
 *           In s7, exact? is a synonym for rational?, inexact->exact is a synonym for rationalize.
 *        '#' does not stand for an unknown digit, and the '@' complex number notation is ignored
 *           I also choose not to include numbers such as +i (= 0+i) -- include the real part!
 *        modulo, remainder, and quotient take integer, ratio, or real args 
 *        lcm and gcd can take integer or ratio args
 *        continuation? function to distinguish a continuation from a procedure
 *        log takes an optional 2nd arg (base)
 *        '.' and an exponent can occur in a number in any base -- they do not mean "base 10"!  
 *           However, the exponent itself is always in base 10 (this follows gmp usage).
 *
 *   other additions: 
 *        random for any numeric type and any numeric argument, including 0 ferchrissake!
 *        sinh, cosh, tanh, asinh, acosh, atanh
 *        read-line, read-byte, write-byte, *stdin*, *stdout*, and *stderr*
 *        logior, logxor, logand, lognot, logbit?, ash, integer-length, integer-decode-float, nan?, infinite?
 *        procedure-source, procedure-arity, procedure-documentation, procedure-name, help
 *          if the initial expression in a function body is a string constant, it is assumed to be a documentation string
 *        symbol-table, symbol->value, global-environment, current-environment, procedure-environment, initial-environment, environment?
 *        provide, provided?, defined?
 *        port-line-number, port-filename
 *        object->string, eval-string
 *        reverse!, list-set!, sort!, make-list
 *        gc, *load-hook*, *error-hook*, *error-info*, *unbound-variable-hook*
 *        *features*, *load-path*, *vector-print-length*, *#readers*
 *        define-constant, pi, most-positive-fixnum, most-negative-fixnum, constant?
 *        stacktrace, trace and untrace, *trace-hook*, __func__, macroexpand
 *        length, copy, fill!, reverse, map, for-each are generic
 *        make-type creates a new scheme type
 *        symbol-access modifies symbol value lookup
 *        member and assoc accept an optional 3rd argument, the comparison function
 *
 *
 * Mike Scholz provided the FreeBSD support (complex trig funcs, etc)
 * Rick Taube and Andrew Burnson provided the MS Visual C++ support
 *
 *
 * Documentation is in s7.h and s7.html.  s7test.scm is a regression test.
 *
 *
 * ---------------- compile time switches ---------------- 
 */

#include <mus-config.h>

/* 
 * Your config file goes here, or just replace that #include line with the defines you need.
 * The compile-time switches involve booleans, complex numbers, and multiprecision arithmetic.
 * Currently we assume we have setjmp.h (used by the error handlers).
 *
 * s7.h includes stdbool.h if HAVE_STDBOOL_H is 1 and we're not in C++.
 * 
 * The *gc-stats* output includes timing info if HAVE_GETTIMEOFDAY.  In this
 *   case we also assume we can include <time.h> and <sys/time.h>.
 *
 *
 * Complex number support (which is problematic in C++, Solaris, and netBSD)
 *   is on the WITH_COMPLEX switch. On a Mac, or in Linux, if you're not using C++,
 *   you can use:
 *
 *   #define WITH_COMPLEX 1
 *   #define HAVE_COMPLEX_TRIG 1
 *
 *   Define the first if your compiler has any support for complex numbers.
 *   Define the second if functions like csin are defined in the math library.
 *
 *   In C++ use:
 *
 *   #define WITH_COMPLEX 1
 *   #define HAVE_COMPLEX_TRIG 0
 *
 *   Some systems (freeBSD) have complex.h, but not the trig funcs, so
 *   WITH_COMPLEX means we can find
 *
 *      cimag creal cabs csqrt carg conj
 *
 *   and HAVE_COMPLEX_TRIG means we have
 *
 *      cacos cacosh casin casinh catan catanh ccos ccosh cexp clog cpow csin csinh ctan ctanh
 *
 * When WITH_COMPLEX is 0 or undefined, the complex functions are stubs that simply return their
 *   argument -- this will be very confusing for the s7 user because, for example, (sqrt -2)
 *   will return something bogus (it will not signal an error).
 *
 * Snd's configure.ac has m4 code to handle WITH_COMPLEX and HAVE_COMPLEX_TRIG.
 *
 *
 * To get multiprecision arithmetic, set WITH_GMP to 1.
 *   You'll also need libgmp, libmpfr, and libmpc (version 0.8.0 or later)
 */


/* ---------------- initial sizes ---------------- */

#define INITIAL_HEAP_SIZE 128000
/* the heap grows as needed, this is its initial size. 
 *
 *    this size is not very important (it can be 32 or maybe smaller, but it needs to be bigger than GC_TEMPS_SIZE below):
 *      8k: 2432, 32k: 2419, 128k: 2401, 512k: 2394, 8192k: 2417
 *    (valgrind timings from 23-Feb-10 running s7test.scm)
 *
 * If the initial heap is small, s7 can run in less than 1 Mbyte of memory.
 * As of 5-May-2011, the heap size must be a multiple of 32.
 */

#define SYMBOL_TABLE_SIZE 19259
/* names are hashed into the symbol table (a vector) and collisions are chained as lists. 
 */

#define INITIAL_STACK_SIZE 512
/* the stack grows as needed, each frame takes 4 entries, this is its initial size.
 *   this needs to be big enough to handle the eval_c_string's at startup (ca 100) 
 *   In s7test.scm, the maximum stack size is ca 440.  In snd-test.scm, it's ca 200.
 *   This number matters only because call/cc copies the stack, which requires filling
 *   the unused portion of the new stack, which requires memcpy of #<unspecified>'s.
 */

#define INITIAL_PROTECTED_OBJECTS_SIZE 16  
/* a vector of objects that are (semi-permanently) protected from the GC, grows as needed */

#define GC_TEMPS_SIZE 256
/* the number of recent objects that are temporarily gc-protected; 8 works for s7test and snd-test. 
 *    For the FFI, this sets the lag between a call on s7_cons and the first moment when its result
 *    might be vulnerable to the GC. 
 */

#define INITIAL_TRACE_LIST_SIZE 2
/* a list of currently-traced functions */



/* ---------------- scheme choices ---------------- */

#ifndef WITH_GMP
  #define WITH_GMP 0
  /* this includes multiprecision arithmetic for all numeric types and functions, using gmp, mpfr, and mpc
   * WITH_GMP adds the following functions: bignum, bignum?, bignum-precision
   * using gmp with precision=128 is about 3 times slower than using C doubles and long long ints
   */
#endif

#if WITH_GMP
  #define DEFAULT_BIGNUM_PRECISION 128
#endif

#ifndef WITH_EXTRA_EXPONENT_MARKERS
  #define WITH_EXTRA_EXPONENT_MARKERS 1
  /* if 1, s7 recognizes "d", "f", "l", and "s" as exponent markers, in addition to "e" (also "D", "F", "L", "S")
   */
#endif


#ifndef WITH_OPTIMIZATION
#define WITH_OPTIMIZATION 1
  /* this currently speeds s7 up by about 236 to 122 in callgrind terms.
   *    a lot of the current optimization choices are just experiments (I'll clean up this mess some day).
   *    the completely unrealistic goal, of course, is to replace the run macro.
   */
#endif



#define DEBUGGING 0
#define PRINTING 0

#define BOLD_TEXT "\033[1m"
#define UNBOLD_TEXT "\033[22m"


#define WITH_GCC defined(__GNUC__)




/* -------------------------------------------------------------------------------- */

/* s7.c is organized as follows:
 *
 *    structs and type flags
 *    constants
 *    GC
 *    stacks
 *    symbols
 *    environments
 *    keywords
 *    continuations
 *    numbers
 *    characters
 *    strings
 *    ports
 *    lists
 *    vectors and hash-tables
 *    objects and functions
 *    hooks
 *    eq?
 *    generic length, copy, fill!
 *    format
 *    error handlers, stacktrace, trace
 *    sundry leftovers
 *    multiple-values, quasiquote
 *    eval
 *    multiprecision arithmetic
 *    initialization
 *
 * naming conventions: s7_* usually are C accessible (s7.h), g_* are scheme accessible (FFI), H_* are documentation strings,
 *   *_1 are auxilliary functions, big_* refer to gmp and friends, scheme "?" corresponds to C "_is_", scheme "->" to C "_to_".
 */

#if __cplusplus
  #ifndef WITH_COMPLEX
    #define WITH_COMPLEX 1
  #endif
  #ifndef HAVE_COMPLEX_TRIG
    #define HAVE_COMPLEX_TRIG 0
  #endif
#endif


#ifndef _MSC_VER
  #include <unistd.h>
#endif
#include <limits.h>
#include <ctype.h>
#ifndef _MSC_VER
  #include <strings.h>
  #include <errno.h>
#endif
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <time.h>
#include <stdarg.h>

#if __cplusplus
  #include <cmath>
#else
  #include <math.h>
#endif

#if WITH_COMPLEX
  #if __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif

#include <setjmp.h>

#include "s7.h"

#ifndef M_PI
  #define M_PI 3.1415926535897932384626433832795029L
#endif

#ifndef INFINITY
  #define INFINITY (-log(0.0))
#endif

#ifndef NAN
  #define NAN (INFINITY / INFINITY)
#endif


#define DISPLAY(Obj) s7_object_to_c_string(sc, Obj)
#define DISPLAY_80(Obj) object_to_truncated_string(sc, Obj, 80)


enum {OP_NO_OP, 
      OP_READ_INTERNAL, OP_EVAL, 
      OP_EVAL_ARGS, OP_EVAL_ARGS1, OP_EVAL_ARGS2, OP_EVAL_ARGS3, OP_EVAL_ARGS4, OP_EVAL_ARGS5,
      OP_APPLY, OP_EVAL_MACRO, OP_LAMBDA, OP_QUOTE, 
      OP_DEFINE, OP_DEFINE1, OP_BEGIN, OP_BEGIN1, OP_IF, OP_IF1, OP_SET, OP_SET1, OP_SET2, 
      OP_LET, OP_LET1, OP_LET_STAR, OP_LET_STAR1, 
      OP_LETREC, OP_LETREC1, OP_COND, OP_COND1, OP_COND_SIMPLE, OP_COND1_SIMPLE,
      OP_AND, OP_AND1, OP_OR, OP_OR1, OP_DEFMACRO, OP_DEFMACRO_STAR,
      OP_DEFINE_MACRO, OP_DEFINE_MACRO_STAR, OP_DEFINE_EXPANSION,
      OP_CASE, OP_CASE1, OP_READ_LIST, OP_READ_DOT, OP_READ_QUOTE, 
      OP_READ_QUASIQUOTE, OP_READ_QUASIQUOTE_VECTOR, OP_READ_UNQUOTE, OP_READ_APPLY_VALUES,
      OP_READ_VECTOR, OP_READ_DONE, 
      OP_LOAD_RETURN_IF_EOF, OP_LOAD_CLOSE_AND_POP_IF_EOF, OP_EVAL_STRING, OP_EVAL_DONE,
      OP_CATCH, OP_DYNAMIC_WIND, OP_DEFINE_CONSTANT, OP_DEFINE_CONSTANT1, 
      OP_DO, OP_DO_END, OP_DO_END1, OP_DO_STEP, OP_DO_STEP2, OP_DO_INIT,
      OP_DEFINE_STAR, OP_LAMBDA_STAR, OP_LAMBDA_STAR_DEFAULT, OP_ERROR_QUIT, OP_UNWIND_INPUT, OP_UNWIND_OUTPUT, 
      OP_TRACE_RETURN, OP_ERROR_HOOK_QUIT, OP_TRACE_HOOK_QUIT, OP_WITH_ENV, OP_WITH_ENV1, OP_EXPANSION,
      OP_FOR_EACH, OP_FOR_EACH_SIMPLE, OP_FOR_EACH_SIMPLER, OP_MAP, OP_MAP_SIMPLE, OP_BARRIER, OP_DEACTIVATE_GOTO,

      OP_DEFINE_BACRO, OP_DEFINE_BACRO_STAR, 
      OP_GET_OUTPUT_STRING, OP_SORT, OP_SORT1, OP_SORT2, OP_SORT3, OP_SORT4, OP_SORT_TWO, 
      OP_EVAL_STRING_1, OP_EVAL_STRING_2, OP_HOOK_APPLY, 
      OP_MEMBER_IF, OP_ASSOC_IF, OP_MEMBER_IF1, OP_ASSOC_IF1,
      
      OP_QUOTE_UNCHECKED, OP_LAMBDA_UNCHECKED, OP_LET_UNCHECKED, OP_CASE_UNCHECKED, 
      OP_SET_UNCHECKED, OP_SET_SYMBOL_C, OP_SET_SYMBOL_S, OP_SET_SYMBOL_Q, OP_SET_SYMBOL_P, OP_SET_SYMBOL_R, OP_SET_NORMAL, OP_SET_PAIR, OP_SET_WITH_ACCESSOR,
      OP_LET_STAR_UNCHECKED, OP_LETREC_UNCHECKED, OP_COND_UNCHECKED,
      OP_LAMBDA_STAR_UNCHECKED, OP_DO_UNCHECKED, OP_DEFINE_UNCHECKED, OP_DEFINE_STAR_UNCHECKED, 
      OP_DEFINE_WITH_ACCESSOR, OP_DEFMACRO_WITH_ACCESSOR, OP_DEFINE_MACRO_WITH_ACCESSOR,
      OP_LET_NO_VARS, OP_NAMED_LET, OP_NAMED_LET_NO_VARS, OP_CASE_SIMPLE, OP_CASE_SIMPLER, OP_CASE_SIMPLEST, OP_CASE_INT,
      OP_LET_C, OP_LET_S, OP_LET_Q, OP_LET_ALL_C, OP_LET_ALL_S, OP_LET_ALL_Q, OP_LET_ALL_G, 
      OP_LET_C_P, OP_LET_S_P, 
      OP_IF_P_P_P, OP_IF_P_P, OP_IF_P_P_X, OP_IF_P_X_P, OP_IF_P_X, OP_IF_P_X_X, 
      OP_IF_B_P, OP_IF_B_P_P, OP_IF_ANDP_P, OP_IF_ANDP_P_P, OP_IF_ORP_P, OP_IF_ORP_P_P, 
      OP_IF_PPP, OP_IF_PP, OP_IF_PPX, OP_IF_PXP, OP_IF_PX, OP_IF_PXX, 
      OP_IF_X_P_P, OP_IF_X_P, OP_IF_X_P_X, OP_IF_X_X_P, OP_IF_X_X, OP_IF_X_X_X,
      OP_AND_UNCHECKED, OP_AND_P, OP_AND_P1, OP_OR_UNCHECKED, OP_OR_P, OP_OR_P1,
      
#if WITH_OPTIMIZATION
      OP_SAFE_IF1, OP_SAFE_IF2,
      OP_SAFE_OR_S, OP_SAFER_OR_S, OP_SAFER_OR_C, OP_SAFER_OR_CEQ,
      OP_SAFE_AND_S, OP_SAFE_CASE_S, OP_SAFE_FOR_EACH, OP_SAFER_OR_X, OP_SAFER_AND_X, OP_COND_SIMPLER,
      OP_SIMPLE_DO, OP_SIMPLE_DO_STEP, OP_DOTIMES, OP_DOTIMES_STEP, OP_SIMPLE_DOTIMES, OP_SAFE_DO, OP_SAFE_DO_STEP, OP_DOTIMES_C_C,
      OP_SIMPLE_DO_P, OP_SIMPLE_DO_STEP_P, OP_SAFE_SIMPLE_DO, OP_DOX, OP_DOX_STEP, OP_SIMPLE_DO_OPT, 
      OP_SAFE_IF1_1, OP_SAFE_IF2_1, OP_SAFE_IF_CC_X_P, OP_SAFE_IF_CC_P_P, OP_SAFE_IF_CEQ_P_P, 
      OP_SAFE_IF_CS_P_P, OP_SAFE_IF_CS_X_P, OP_SAFE_IF_CC_P, OP_SAFE_IF_CS_P, OP_SAFE_IF_CS_P_X, OP_SAFE_IF_CS_Q_P,
      OP_SAFE_IF_CSS_X_P, OP_SAFE_IF_CSC_X_P, OP_SAFE_IF_CSC_X_P_A,
      OP_SAFE_IF_CSQ_P, OP_SAFE_IF_CSQ_P_P, OP_SAFE_IF_CSS_P, OP_SAFE_IF_CSS_P_P, OP_SAFE_IF_CSC_P, OP_SAFE_IF_CSC_P_P, 
      OP_SAFE_IF_IS_PAIR_P, OP_SAFE_IF_IS_PAIR_P_X, OP_SAFE_IF_IS_PAIR_P_P, OP_SAFE_IF_C_SS_P,
      OP_SAFE_IF_IS_SYMBOL_P, OP_SAFE_IF_IS_SYMBOL_P_X, OP_SAFE_IF_IS_SYMBOL_P_P, 
      OP_SAFE_IF_IS_EOF_P_P, OP_IF_O_P, OP_IF_O_P_P, OP_SAFE_IF_ANDX_P, OP_SAFE_IF_ORX_P, OP_IF_ORCEQ_P, OP_IF_ORCEQ_P_P,
      OP_SAFE_C_P_1, OP_SAFE_C_PP_1, OP_SAFE_C_PP_2, OP_EVAL_ARGS_P_1, OP_EVAL_ARGS_P_2, OP_EVAL_ARGS_P_3, OP_EVAL_ARGS_P_4,
      OP_INCREMENT_1, OP_DECREMENT_1, OP_SET_CDR, OP_SET_CONS,
      OP_LET_O, OP_LET_O1, OP_LET_R, OP_LET_ALL_R, OP_LET_C_D, OP_LET_O_P, OP_LET_O2, OP_LET_R_P, OP_LET_OX_P, OP_LET_UNKNOWN_S_P, 
      OP_LET_READ_CHAR_P, OP_LET_CAR_P,
      OP_SAFE_C_ZZ_1, OP_SAFE_C_ZZ_2, OP_SAFE_C_SZ_1, OP_SAFE_C_ZS_1,
      OP_SAFE_C_ZXX_1, OP_SAFE_C_XZX_1, OP_SAFE_C_XXZ_1, 
      OP_SAFE_C_ZZX_1, OP_SAFE_C_ZZX_2, OP_SAFE_C_ZXZ_1, OP_SAFE_C_ZXZ_2, OP_SAFE_C_XZZ_1, OP_SAFE_C_XZZ_2, 
      OP_SAFE_C_ZZZ_1, OP_SAFE_C_ZZZ_2, OP_SAFE_C_ZZZ_3, OP_SAFE_C_opSq_P_1, OP_C_P_1,
#endif
      OP_MAX_DEFINED_1};

#define OP_MAX_DEFINED (OP_MAX_DEFINED_1 + 1)


#if ((defined(SIZEOF_VOID_P)) && (SIZEOF_VOID_P == 4))
  #define opcode_t unsigned int
#else
  #define opcode_t unsigned long long int
#endif

/* these names can leak into the repl if (for example) user types set!
 */
static const char *op_names[OP_MAX_DEFINED + 1] = 
  {"no-op", 
   "read-internal", "eval", 
   "eval-args", "eval-args1", "eval-args2", "eval-args3", "eval-args4", "eval-args5",
   "apply", "eval-macro", "lambda", "quote", 
   "define", "define", "begin", "begin", "if", "if", "set!", "set!", "set!", 
   "let", "let", "let*", "let*", 
   "letrec", "letrec", "cond", "cond", "cond", "cond",
   "and", "and", "or", "or", "defmacro", "defmacro*",
   "define-macro", "define-macro*", "define-expansion",
   "case", "case", "read-list", "read-dot", "read-quote", 
   "read-quasiquote", "read-quasiquote-vector", "read-unquote", "read-apply-values",
   "read-vector", "read-done", 
   "load-return-if-eof", "load-close-and-pop-if-eof", "eval-string", "eval-done",
   "catch", "dynamic-wind", "define-constant", "define-constant", 
   "do", "do", "do", "do", "do", "do",
   "define*", "lambda*", "lambda*", "error-quit", "unwind-input", "unwind-output", 
   "trace-return", "error-hook-quit", "trace-hook-quit", "with-environment", "with-environment", "define-expansion",
   "for-each", "for-each", "for-each", "map", "map", "barrier", "deactivate-goto",
   "define-bacro", "define-bacro*", 
   "get-output-string", "sort!", "sort!", "sort!", "sort!", "sort!", "sort!", 
   "eval-string", "eval-string", "hook-apply", 
   "member", "assoc", "member", "assoc",
   
   "quote", "lambda", "let", "case", 
   "set!", "set!", "set!", "set!", "set!", "set!", "set!", "set!", "set!",
   "let*", "letrec", "cond",
   "lambda*", "do", "define", "define*", 
   "define", "defmacro", "define-macro",
   "let", "let", "let", "case", "case", "case", "case",
   "let", "let", "let", "let", "let", "let", "let",
   "let", "let",
   "if", "if", "if", "if", "if", "if", "if", "if",
   "if", "if", "if", "if", "if", "if", 
   "if", "if", "if", "if", "if", "if",
   "and", "and", "and", "or", "or", "or", 

#if WITH_OPTIMIZATION
   "safe_if1", "safe_if2",
   "safe_or_s", "safer_or_s", "safer_or_c", "safer_or_ceq",
   "safe_and_s", "safe_case_s", "safe_for_each", "safer_or_x", "safer_and_x", "cond_simpler",
   "simple_do", "simple_do_step", "dotimes", "dotimes_step", "simple_dotimes", "safe_do", "safe_do_step", "dotimes_c_c",
   "simple_do_p", "simple_do_step_p", "safe_simple_do", "dox", "dox_step", "simple_do_opt", 
   "safe_if1_1", "safe_if2_1", "safe_if_cc_x_p", "safe_if_cc_p_p", "safe_if_ceq_p_p", 
   "safe_if_cs_p_p", "safe_if_cs_x_p", "safe_if_cc_p", "safe_if_cs_p", "safe_if_cs_p_x", "safe_if_cs_q_p",
   "safe_if_css_x_p", "safe_if_csc_x_p", "safe_if_csc_x_p_a",
   "safe_if_csq_p", "safe_if_csq_p_p", "safe_if_css_p", "safe_if_css_p_p", "safe_if_csc_p", "safe_if_csc_p_p", 
   "safe_if_is_pair_p", "safe_if_is_pair_p_x", "safe_if_is_pair_p_p", "safe_if_c_ss_p",
   "safe_if_is_symbol_p", "safe_if_is_symbol_p_x", "safe_if_is_symbol_p_p", 
   "safe_if_is_eof_p_p", "if_o_p", "if_o_p_p", "safe_if_andx_p", "safe_if_orx_p", "if_orceq_p", "if_orceq_p_p",
   "safe_c_p_1", "safe_c_pp_1", "safe_c_pp_2", "eval_args_p_1", "eval_args_p_2", "eval_args_p_3", "eval_args_p_4",
   "increment_1", "decrement_1", "set_cdr", "set_cons",
   "let_o", "let_o1", "let_r", "let_all_r", "let_c_d", "let_o_p", "let_o2", "let_r_p", "let_ox_p", "let_unknown_s_p", 
   "let_read_char_p", "let_car_p",
   "safe_c_zz_1", "safe_c_zz_2", "safe_c_sz_1", "safe_c_zs_1",
   "safe_c_zxx_1", "safe_c_xzx_1", "safe_c_xxz_1", 
   "safe_c_zzx_1", "safe_c_zzx_2", "safe_c_zxz_1", "safe_c_zxz_2", "safe_c_xzz_1", "safe_c_xzz_2", 
   "safe_c_zzz_1", "safe_c_zzz_2", "safe_c_zzz_3", "safe_c_opsq_p_1", "c_p_1",
#endif
   
   "op-max"
  };


#if DEBUGGING
static const char *real_op_names[OP_MAX_DEFINED + 1] = {
  "OP_NO_OP", 
  "OP_READ_INTERNAL", "OP_EVAL", 
  "OP_EVAL_ARGS", "OP_EVAL_ARGS1", "OP_EVAL_ARGS2", "OP_EVAL_ARGS3", "OP_EVAL_ARGS4", "OP_EVAL_ARGS5",
  "OP_APPLY", "OP_EVAL_MACRO", "OP_LAMBDA", "OP_QUOTE", 
  "OP_DEFINE", "OP_DEFINE1", "OP_BEGIN", "OP_BEGIN1", "OP_IF", "OP_IF1", "OP_SET", "OP_SET1", "OP_SET2", 
  "OP_LET", "OP_LET1", "OP_LET_STAR", "OP_LET_STAR1", 
  "OP_LETREC", "OP_LETREC1", "OP_COND", "OP_COND1", "OP_COND_SIMPLE", "OP_COND1_SIMPLE",
  "OP_AND", "OP_AND1", "OP_OR", "OP_OR1", "OP_DEFMACRO", "OP_DEFMACRO_STAR",
  "OP_DEFINE_MACRO", "OP_DEFINE_MACRO_STAR", "OP_DEFINE_EXPANSION",
  "OP_CASE", "OP_CASE1", "OP_READ_LIST", "OP_READ_DOT", "OP_READ_QUOTE", 
  "OP_READ_QUASIQUOTE", "OP_READ_QUASIQUOTE_VECTOR", "OP_READ_UNQUOTE", "OP_READ_APPLY_VALUES",
  "OP_READ_VECTOR", "OP_READ_DONE", 
  "OP_LOAD_RETURN_IF_EOF", "OP_LOAD_CLOSE_AND_POP_IF_EOF", "OP_EVAL_STRING", "OP_EVAL_DONE",
  "OP_CATCH", "OP_DYNAMIC_WIND", "OP_DEFINE_CONSTANT", "OP_DEFINE_CONSTANT1", 
  "OP_DO", "OP_DO_END", "OP_DO_END1", "OP_DO_STEP", "OP_DO_STEP2", "OP_DO_INIT",
  "OP_DEFINE_STAR", "OP_LAMBDA_STAR", "OP_LAMBDA_STAR_DEFAULT", "OP_ERROR_QUIT", "OP_UNWIND_INPUT", "OP_UNWIND_OUTPUT", 
  "OP_TRACE_RETURN", "OP_ERROR_HOOK_QUIT", "OP_TRACE_HOOK_QUIT", "OP_WITH_ENV", "OP_WITH_ENV1", "OP_EXPANSION",
  "OP_FOR_EACH", "OP_FOR_EACH_SIMPLE", "OP_FOR_EACH_SIMPLER", "OP_MAP", "OP_MAP_SIMPLE", "OP_BARRIER", "OP_DEACTIVATE_GOTO",
  
  "OP_DEFINE_BACRO", "OP_DEFINE_BACRO_STAR", 
  "OP_GET_OUTPUT_STRING", "OP_SORT", "OP_SORT1", "OP_SORT2", "OP_SORT3", "OP_SORT4", "OP_SORT_TWO", 
  "OP_EVAL_STRING_1", "OP_EVAL_STRING_2", "OP_HOOK_APPLY", 
  "OP_MEMBER_IF", "OP_ASSOC_IF", "OP_MEMBER_IF1", "OP_ASSOC_IF1",
  
  "OP_QUOTE_UNCHECKED", "OP_LAMBDA_UNCHECKED", "OP_LET_UNCHECKED", "OP_CASE_UNCHECKED", 
  "OP_SET_UNCHECKED", "OP_SET_SYMBOL_C", "OP_SET_SYMBOL_S", "OP_SET_SYMBOL_Q", "OP_SET_SYMBOL_P", "OP_SET_SYMBOL_R", "OP_SET_NORMAL", "OP_SET_PAIR", "OP_SET_WITH_ACCESSOR",
  "OP_LET_STAR_UNCHECKED", "OP_LETREC_UNCHECKED", "OP_COND_UNCHECKED",
  "OP_LAMBDA_STAR_UNCHECKED", "OP_DO_UNCHECKED", "OP_DEFINE_UNCHECKED", "OP_DEFINE_STAR_UNCHECKED", 
  "OP_DEFINE_WITH_ACCESSOR", "OP_DEFMACRO_WITH_ACCESSOR", "OP_DEFINE_MACRO_WITH_ACCESSOR",
  "OP_LET_NO_VARS", "OP_NAMED_LET", "OP_NAMED_LET_NO_VARS", "OP_CASE_SIMPLE", "OP_CASE_SIMPLER", "OP_CASE_SIMPLEST", "OP_CASE_INT",
  "OP_LET_C", "OP_LET_S", "OP_LET_Q", "OP_LET_ALL_C", "OP_LET_ALL_S", "OP_LET_ALL_Q", "OP_LET_ALL_G", 
  "OP_LET_C_P", "OP_LET_S_P", 
  "OP_IF_P_P_P", "OP_IF_P_P", "OP_IF_P_P_X", "OP_IF_P_X_P", "OP_IF_P_X", "OP_IF_P_X_X", 
  "OP_IF_B_P", "OP_IF_B_P_P", "OP_IF_ANDP_P", "OP_IF_ANDP_P_P", "OP_IF_ORP_P", "OP_IF_ORP_P_P", 
  "OP_IF_PPP", "OP_IF_PP", "OP_IF_PPX", "OP_IF_PXP", "OP_IF_PX", "OP_IF_PXX", 
  "OP_IF_X_P_P", "OP_IF_X_P", "OP_IF_X_P_X", "OP_IF_X_X_P", "OP_IF_X_X", "OP_IF_X_X_X",
  "OP_AND_UNCHECKED", "OP_AND_P", "OP_AND_P1", "OP_OR_UNCHECKED", "OP_OR_P", "OP_OR_P1",
  
#if WITH_OPTIMIZATION
  "OP_SAFE_IF1", "OP_SAFE_IF2",
  "OP_SAFE_OR_S", "OP_SAFER_OR_S", "OP_SAFER_OR_C", "OP_SAFER_OR_CEQ",
  "OP_SAFE_AND_S", "OP_SAFE_CASE_S", "OP_SAFE_FOR_EACH", "OP_SAFER_OR_X", "OP_SAFER_AND_X", "OP_COND_SIMPLER",
  "OP_SIMPLE_DO", "OP_SIMPLE_DO_STEP", "OP_DOTIMES", "OP_DOTIMES_STEP", "OP_SIMPLE_DOTIMES", "OP_SAFE_DO", "OP_SAFE_DO_STEP", "OP_DOTIMES_C_C",
  "OP_SIMPLE_DO_P", "OP_SIMPLE_DO_STEP_P", "OP_SAFE_SIMPLE_DO", "OP_DOX", "OP_DOX_STEP", "OP_SIMPLE_DO_OPT",
  "OP_SAFE_IF1_1", "OP_SAFE_IF2_1", "OP_SAFE_IF_CC_X_P", "OP_SAFE_IF_CC_P_P", "OP_SAFE_IF_CEQ_P_P", 
  "OP_SAFE_IF_CS_P_P", "OP_SAFE_IF_CS_X_P", "OP_SAFE_IF_CC_P", "OP_SAFE_IF_CS_P", "OP_SAFE_IF_CS_P_X", "OP_SAFE_IF_CS_Q_P",
  "OP_SAFE_IF_CSS_X_P", "OP_SAFE_IF_CSC_X_P", "OP_SAFE_IF_CSC_X_P_A",
  "OP_SAFE_IF_CSQ_P", "OP_SAFE_IF_CSQ_P_P", "OP_SAFE_IF_CSS_P", "OP_SAFE_IF_CSS_P_P", "OP_SAFE_IF_CSC_P", "OP_SAFE_IF_CSC_P_P", 
  "OP_SAFE_IF_IS_PAIR_P", "OP_SAFE_IF_IS_PAIR_P_X", "OP_SAFE_IF_IS_PAIR_P_P", "OP_SAFE_IF_C_SS_P",
  "OP_SAFE_IF_IS_SYMBOL_P", "OP_SAFE_IF_IS_SYMBOL_P_X", "OP_SAFE_IF_IS_SYMBOL_P_P", 
  "OP_SAFE_IF_IS_EOF_P_P", "OP_IF_O_P", "OP_IF_O_P_P", "SAFE_IF_ANDX_P", "SAFE_IF_ORX_P", "OP_IF_ORCEQ_P", "OP_IF_ORCEQ_P_P",
  "OP_SAFE_C_P_1", "OP_SAFE_C_PP_1", "OP_SAFE_C_PP_2", "OP_EVAL_ARGS_P_1", "OP_EVAL_ARGS_P_2", "OP_EVAL_ARGS_P_3", "OP_EVAL_ARGS_P_4",
  "OP_INCREMENT_1", "OP_DECREMENT_1", "OP_SET_CDR", "OP_SET_CONS",
  "OP_LET_O", "OP_LET_O1", "OP_LET_R", "OP_LET_ALL_R", "OP_LET_C_D", "OP_LET_O_P", "OP_LET_O2", "OP_LET_R_P", "OP_LET_OX_P", "OP_LET_UNKNOWN_S_P", 
  "OP_LET_READ_CHAR_P", "OP_LET_CAR_P",
  "OP_SAFE_C_ZZ_1", "OP_SAFE_C_ZZ_2", "OP_SAFE_C_SZ_1", "OP_SAFE_C_ZS_1",
  "OP_SAFE_C_ZXX_1", "OP_SAFE_C_XZX_1", "OP_SAFE_C_XXZ_1", 
  "OP_SAFE_C_ZZX_1", "OP_SAFE_C_ZZX_2", "OP_SAFE_C_ZXZ_1", "OP_SAFE_C_ZXZ_2", "OP_SAFE_C_XZZ_1", "OP_SAFE_C_XZZ_2", 
  "OP_SAFE_C_ZZZ_1", "OP_SAFE_C_ZZZ_2", "OP_SAFE_C_ZZZ_3", "OP_SAFE_C_opSq_P_1", "OP_C_P_1",
#endif
  "OP_MAX_DEFINED_1"};
#endif   



#if WITH_OPTIMIZATION
enum{SAFE_C_P, SAFE_C_PP, SAFE_C_CP, SAFE_C_SP, SAFE_C_PC, SAFE_C_PS, SAFE_C_PQ, SAFE_C_QP};

enum {OP_NOT_AN_OP, HOP_NOT_AN_OP,
      OP_THUNK, HOP_THUNK, 
      OP_CLOSURE_S, HOP_CLOSURE_S, OP_CLOSURE_C, HOP_CLOSURE_C, OP_CLOSURE_Q, HOP_CLOSURE_Q, 
      OP_CLOSURE_SS, HOP_CLOSURE_SS, OP_CLOSURE_SC, HOP_CLOSURE_SC, OP_CLOSURE_CS, HOP_CLOSURE_CS, OP_CLOSURE_CC, HOP_CLOSURE_CC, 
      OP_CLOSURE_opSq, HOP_CLOSURE_opSq, OP_CLOSURE_opCq, HOP_CLOSURE_opCq, 
      OP_CLOSURE_opSq_S, HOP_CLOSURE_opSq_S, OP_CLOSURE_opSq_opSq, HOP_CLOSURE_opSq_opSq, 
      OP_CLOSURE_S_opSq, HOP_CLOSURE_S_opSq, OP_CLOSURE_SSS, HOP_CLOSURE_SSS,
      OP_CLOSURE_ALL_S, HOP_CLOSURE_ALL_S, OP_CLOSURE_ALL_C, HOP_CLOSURE_ALL_C, OP_CLOSURE_ALL_G, HOP_CLOSURE_ALL_G, 
      OP_CLOSURE_CDR_S, HOP_CLOSURE_CDR_S, 

      OP_TCLOSURE_S, HOP_TCLOSURE_S, 
      OP_TCLOSURE_SS, HOP_TCLOSURE_SS, 
      OP_TCLOSURE_opSq, HOP_TCLOSURE_opSq, OP_TCLOSURE_opCq, HOP_TCLOSURE_opCq, 
      OP_TCLOSURE_opSq_S, HOP_TCLOSURE_opSq_S,
      OP_TCLOSURE_S_opSq, HOP_TCLOSURE_S_opSq, 
      OP_TCLOSURE_CDR_S, HOP_TCLOSURE_CDR_S, 
      
      OP_SAFE_THUNK, HOP_SAFE_THUNK,
      OP_SAFE_CLOSURE_S, HOP_SAFE_CLOSURE_S, OP_SAFE_CLOSURE_C, HOP_SAFE_CLOSURE_C, OP_SAFE_CLOSURE_Q, HOP_SAFE_CLOSURE_Q, 
      OP_SAFE_CLOSURE_SS, HOP_SAFE_CLOSURE_SS, OP_SAFE_CLOSURE_SC, HOP_SAFE_CLOSURE_SC, OP_SAFE_CLOSURE_CS, HOP_SAFE_CLOSURE_CS, OP_SAFE_CLOSURE_CC, HOP_SAFE_CLOSURE_CC, 
      OP_SAFE_CLOSURE_opSq, HOP_SAFE_CLOSURE_opSq, OP_SAFE_CLOSURE_opSq_S, HOP_SAFE_CLOSURE_opSq_S, OP_SAFE_CLOSURE_opSq_opSq, HOP_SAFE_CLOSURE_opSq_opSq, 
      OP_SAFE_CLOSURE_S_opSq, HOP_SAFE_CLOSURE_S_opSq, OP_SAFE_CLOSURE_SSS, HOP_SAFE_CLOSURE_SSS,
      OP_SAFE_CLOSURE_ALL_S, HOP_SAFE_CLOSURE_ALL_S, OP_SAFE_CLOSURE_ALL_C, HOP_SAFE_CLOSURE_ALL_C, OP_SAFE_CLOSURE_ALL_G, HOP_SAFE_CLOSURE_ALL_G, 
      
      OP_SAFE_C_opSAFE_CLOSURE_SSq, HOP_SAFE_C_opSAFE_CLOSURE_SSq, OP_SAFE_C_opSAFE_CLOSURE_opSq_Sq, HOP_SAFE_C_opSAFE_CLOSURE_opSq_Sq,
      OP_SAFE_CLOSURE_opCq, HOP_SAFE_CLOSURE_opCq, 
      OP_SAFE_CLOSURE_S_one, HOP_SAFE_CLOSURE_S_one, OP_SAFE_CLOSURE_S_vref, HOP_SAFE_CLOSURE_S_vref, 
      
      OP_SAFE_C_C, HOP_SAFE_C_C, OP_SAFE_C_S, HOP_SAFE_C_S,
      OP_SAFE_C_SS, HOP_SAFE_C_SS, OP_SAFE_C_SS_2, HOP_SAFE_C_SS_2, OP_SAFE_C_SC, HOP_SAFE_C_SC, OP_SAFE_C_CS, HOP_SAFE_C_CS, 
      OP_SAFE_C_Q, HOP_SAFE_C_Q, OP_SAFE_C_SQ, HOP_SAFE_C_SQ, OP_SAFE_C_QS, HOP_SAFE_C_QS, OP_SAFE_C_QQ, HOP_SAFE_C_QQ, 
      OP_SAFE_C_XXX, HOP_SAFE_C_XXX, OP_SAFE_C_SSS, HOP_SAFE_C_SSS,
      OP_SAFE_C_ALL_S, HOP_SAFE_C_ALL_S, OP_SAFE_C_ALL_G, HOP_SAFE_C_ALL_G, OP_SAFE_C_ALL_X, HOP_SAFE_C_ALL_X,
      
      OP_SAFE_C_opCq, HOP_SAFE_C_opCq, OP_SAFE_C_opQq, HOP_SAFE_C_opQq, OP_SAFE_C_opSq, HOP_SAFE_C_opSq, 
      OP_SAFE_C_opSSq, HOP_SAFE_C_opSSq, OP_SAFE_C_opSCq, HOP_SAFE_C_opSCq, OP_SAFE_C_opSQq, HOP_SAFE_C_opSQq, 
      OP_SAFE_C_opCSq, HOP_SAFE_C_opCSq, OP_SAFE_C_S_opSq, HOP_SAFE_C_S_opSq,
      OP_SAFE_C_C_opSCq, HOP_SAFE_C_C_opSCq, 
      OP_SAFE_C_S_opSCq, HOP_SAFE_C_S_opSCq, OP_SAFE_C_S_opCSq, HOP_SAFE_C_S_opCSq, 
      OP_SAFE_C_opSq_S, HOP_SAFE_C_opSq_S, OP_SAFE_C_opSq_C, HOP_SAFE_C_opSq_C, 
      OP_SAFE_C_opSq_opSq, HOP_SAFE_C_opSq_opSq, OP_SAFE_C_S_opSSq, HOP_SAFE_C_S_opSSq, OP_SAFE_C_C_opSq, HOP_SAFE_C_C_opSq, 
      OP_SAFE_C_C_opCSq, HOP_SAFE_C_C_opCSq, OP_SAFE_C_opCSq_C, HOP_SAFE_C_opCSq_C, 
      OP_SAFE_C_S_opCq, HOP_SAFE_C_S_opCq, OP_SAFE_C_opSSq_C, HOP_SAFE_C_opSSq_C, OP_SAFE_C_C_opSSq, HOP_SAFE_C_C_opSSq, 
      OP_SAFE_C_C_opCq, HOP_SAFE_C_C_opCq, OP_SAFE_C_opCq_S, HOP_SAFE_C_opCq_S, 
      OP_SAFE_C_opCq_opCq, HOP_SAFE_C_opCq_opCq, OP_SAFE_C_opCq_C, HOP_SAFE_C_opCq_C, 
      OP_SAFE_C_opSCq_opSCq, HOP_SAFE_C_opSCq_opSCq, OP_SAFE_C_opSSq_opSSq, HOP_SAFE_C_opSSq_opSSq,
      OP_SAFE_C_opSSq_S, HOP_SAFE_C_opSSq_S, OP_SAFE_C_opSCq_S, HOP_SAFE_C_opSCq_S, OP_SAFE_C_opCSq_S, HOP_SAFE_C_opCSq_S,
      OP_SAFE_C_opSCq_C, HOP_SAFE_C_opSCq_C, OP_SAFE_C_opQSq, HOP_SAFE_C_opQSq, 
      OP_SAFE_C_opXXXq, HOP_SAFE_C_opXXXq, 
      
      OP_SAFE_IS_PAIR_S, HOP_SAFE_IS_PAIR_S, OP_SAFE_IS_SYMBOL_S, HOP_SAFE_IS_SYMBOL_S, OP_SAFE_NOT_S, HOP_SAFE_NOT_S,
      
      /* these can't be embedded, and have to be the last thing called */
      OP_C_LS, HOP_C_LS, OP_C_L_opSq, HOP_C_L_opSq, OP_C_L, HOP_C_L, OP_C_LL, HOP_C_LL, OP_C_CLL, HOP_C_CLL,
      OP_C_ALL_G, HOP_C_ALL_G, OP_CALL_WITH_EXIT, HOP_CALL_WITH_EXIT, OP_C_CATCH, HOP_C_CATCH,
      OP_C_S_opSq, HOP_C_S_opSq, OP_C_opSq_CC, HOP_C_opSq_CC, 
      OP_C_FOR_EACH_LS, HOP_C_FOR_EACH_LS, OP_C_FOR_EACH_LS_2, HOP_C_FOR_EACH_LS_2,
      OP_C_FOR_EACH_L_opSq, HOP_C_FOR_EACH_L_opSq,
      OP_C_S, HOP_C_S, OP_READ_S, HOP_READ_S, OP_C_P, HOP_C_P,
      
      OP_VECTOR_C, HOP_VECTOR_C, OP_VECTOR_S, HOP_VECTOR_S, OP_VECTOR_opCq, HOP_VECTOR_opCq, 
      OP_STRING_C, HOP_STRING_C, OP_STRING_S, HOP_STRING_S, OP_STRING_opCq, HOP_STRING_opCq, 
      OP_C_OBJECT_C, HOP_C_OBJECT_C, OP_C_OBJECT_S, HOP_C_OBJECT_S, OP_C_OBJECT_opCq, HOP_C_OBJECT_opCq, 
      OP_PAIR_C, HOP_PAIR_C, OP_PAIR_S, HOP_PAIR_S, OP_PAIR_opCq, HOP_PAIR_opCq, 
      OP_HASH_TABLE_C, HOP_HASH_TABLE_C, OP_HASH_TABLE_S, HOP_HASH_TABLE_S, OP_HASH_TABLE_opCq, HOP_HASH_TABLE_opCq, 
      OP_ENVIRONMENT_S, HOP_ENVIRONMENT_S, 
      
      OP_UNKNOWN_C, HOP_UNKNOWN_C, OP_UNKNOWN_S, HOP_UNKNOWN_S,
      OP_UNKNOWN_SS, HOP_UNKNOWN_SS, OP_UNKNOWN_SSS, HOP_UNKNOWN_SSS, 
      OP_UNKNOWN_SC, HOP_UNKNOWN_SC, OP_UNKNOWN_CS, HOP_UNKNOWN_CS, OP_UNKNOWN_CC, HOP_UNKNOWN_CC, 
      OP_UNKNOWN_opSq, HOP_UNKNOWN_opSq, OP_UNKNOWN_opSq_S, HOP_UNKNOWN_opSq_S, OP_UNKNOWN_opSq_opSq, HOP_UNKNOWN_opSq_opSq,      
      OP_UNKNOWN_S_opSq, HOP_UNKNOWN_S_opSq, OP_UNKNOWN_opCq, HOP_UNKNOWN_opCq, 
      
      OP_SAFE_C_P, HOP_SAFE_C_P, OP_SAFE_C_PP, HOP_SAFE_C_PP,
      OP_SAFE_C_opSq_P, HOP_SAFE_C_opSq_P, 
      OP_SAFE_C_SP, HOP_SAFE_C_SP, OP_SAFE_C_CP, HOP_SAFE_C_CP, 
      OP_SAFE_C_PS, HOP_SAFE_C_PS, OP_SAFE_C_PC, HOP_SAFE_C_PC, 
      
      OP_SAFE_C_Z, HOP_SAFE_C_Z, OP_SAFE_C_ZZ, HOP_SAFE_C_ZZ, OP_SAFE_C_SZ, HOP_SAFE_C_SZ, OP_SAFE_C_ZS, HOP_SAFE_C_ZS, 
      OP_SAFE_C_CZ, HOP_SAFE_C_CZ, OP_SAFE_C_ZC, HOP_SAFE_C_ZC, OP_SAFE_C_QZ, HOP_SAFE_C_QZ, OP_SAFE_C_ZQ, HOP_SAFE_C_ZQ, 
      
      OP_SAFE_C_ZXX, HOP_SAFE_C_ZXX, OP_SAFE_C_XZX, HOP_SAFE_C_XZX, OP_SAFE_C_XXZ, HOP_SAFE_C_XXZ,
      OP_SAFE_C_ZZX, HOP_SAFE_C_ZZX, OP_SAFE_C_ZXZ, HOP_SAFE_C_ZXZ, OP_SAFE_C_XZZ, HOP_SAFE_C_XZZ, 
      OP_SAFE_C_ZZZ, HOP_SAFE_C_ZZZ,
      
      OP_SAFE_C_DD, HOP_SAFE_C_DD,
      OP_SAFE_C_DXX, HOP_SAFE_C_DXX, OP_SAFE_C_XDX, HOP_SAFE_C_XDX, OP_SAFE_C_XXD, HOP_SAFE_C_XXD,
      OP_SAFE_C_DDX, HOP_SAFE_C_DDX, OP_SAFE_C_DXD, HOP_SAFE_C_DXD, OP_SAFE_C_XDD, HOP_SAFE_C_XDD, 
      OP_SAFE_C_DDD, HOP_SAFE_C_DDD,
      
      OPT_MAX_DEFINED
};


#if DEBUGGING
static const char *opt_names[OPT_MAX_DEFINED + 1] =
  {  
     "not_an_op", "h_not_an_op",
     "thunk", "h_thunk", 
     "closure_s", "h_closure_s", "closure_c", "h_closure_c", "closure_q", "h_closure_q", 
     "closure_ss", "h_closure_ss", "closure_sc", "h_closure_sc", "closure_cs", "h_closure_cs", "closure_cc", "h_closure_cc", 
     "closure_opsq", "h_closure_opsq", "closure_opcq", "h_closure_opcq", 
     "closure_opsq_s", "h_closure_opsq_s", "closure_opsq_opsq", "h_closure_opsq_opsq", 
     "closure_s_opsq", "h_closure_s_opsq", "closure_sss", "h_closure_sss",
     "closure_all_s", "h_closure_all_s", "closure_all_c", "h_closure_all_c", "closure_all_g", "h_closure_all_g",
     "closure_cdr_s", "h_closure_cdr_s",

     "tclosure_s", "h_tclosure_s", 
     "tclosure_ss", "h_tclosure_ss", 
     "tclosure_opsq", "h_tclosure_opsq", "closure_opcq", "h_closure_opcq", 
     "tclosure_opsq_s", "h_tclosure_opsq_s", 
     "tclosure_s_opsq", "h_tclosure_s_opsq", 
     "tclosure_cdr_s", "h_tclosure_cdr_s",

     "safe_thunk", "h_safe_thunk",
     "safe_closure_s", "h_safe_closure_s", "safe_closure_c", "h_safe_closure_c", "safe_closure_q", "h_safe_closure_q", 
     "safe_closure_ss", "h_safe_closure_ss", "safe_closure_sc", "h_safe_closure_sc", "safe_closure_cs", "h_safe_closure_cs", "safe_closure_cc", "h_safe_closure_cc", 
     "safe_closure_opsq", "h_safe_closure_opsq", "safe_closure_opsq_s", "h_safe_closure_opsq_s", "safe_closure_opsq_opsq", "h_safe_closure_opsq_opsq", 
     "safe_closure_s_opsq", "h_safe_closure_s_opsq", "safe_closure_sss", "h_safe_closure_sss",
     "safe_closure_all_s", "h_safe_closure_all_s", "safe_closure_all_c", "h_safe_closure_all_c", "safe_closure_all_g", "h_safe_closure_all_g", 

     "safe_c_opsafe_closure_ssq", "h_safe_c_opsafe_closure_ssq", "safe_c_opsafe_closure_opsq_sq", "h_safe_c_opsafe_closure_opsq_sq",
     "safe_closure_opcq", "h_safe_closure_opcq", 
     "safe_closure_s_one", "h_safe_closure_s_one", "safe_closure_s_vref", "h_safe_closure_s_vref",

     "safe_c_c", "h_safe_c_c", "safe_c_s", "h_safe_c_s",
     "safe_c_ss", "h_safe_c_ss", "safe_c_ss_2", "h_safe_c_ss_2", "safe_c_sc", "h_safe_c_sc", "safe_c_cs", "h_safe_c_cs", 
     "safe_c_q", "h_safe_c_q", "safe_c_sq", "h_safe_c_sq", "safe_c_qs", "h_safe_c_qs", "safe_c_qq", "h_safe_c_qq", 
     "safe_c_xxx", "h_safe_c_xxx", "safe_c_sss", "h_safe_c_sss",
     "safe_c_all_s", "h_safe_c_all_s", "safe_c_all_g", "h_safe_c_all_g", "safe_c_all_x", "h_safe_c_all_x",

     "safe_c_opcq", "h_safe_c_opcq", "safe_c_opqq", "h_safe_c_opqq", "safe_c_opsq", "h_safe_c_opsq", 
     "safe_c_opssq", "h_safe_c_opssq", "safe_c_opscq", "h_safe_c_opscq", "safe_c_opsqq", "h_safe_c_opsqq", 
     "safe_c_opcsq", "h_safe_c_opcsq", "safe_c_s_opsq", "h_safe_c_s_opsq", 
     "safe_c_c_opscq", "h_safe_c_c_opscq", 
     "safe_c_s_opscq", "h_safe_c_s_opscq", "safe_c_s_opcsq", "h_safe_c_s_opcsq", 
     "safe_c_opsq_s", "h_safe_c_opsq_s", "safe_c_opsq_c", "h_safe_c_opsq_c", 
     "safe_c_opsq_opsq", "h_safe_c_opsq_opsq", "safe_c_s_opssq", "h_safe_c_s_opssq", "safe_c_c_opsq", "h_safe_c_c_opsq", 
     "safe_c_c_opcsq", "h_safe_c_c_opcsq", "safe_c_opcsq_c", "h_safe_c_opcsq_c", 
     "safe_c_s_opcq", "h_safe_c_s_opcq", "safe_c_opssq_c", "h_safe_c_opssq_c", "safe_c_c_opssq", "h_safe_c_c_opssq", 
     "safe_c_c_opcq", "h_safe_c_c_opcq", "safe_c_opcq_s", "h_safe_c_opcq_s", 
     "safe_c_opcq_opcq", "h_safe_c_opcq_opcq", "safe_c_opcq_c", "h_safe_c_opcq_c", 
     "safe_c_opscq_opscq", "h_safe_c_opscq_opscq", "safe_c_opssq_opssq", "h_safe_c_opssq_opssq",
     "safe_c_opssq_s", "h_safe_c_opssq_s", "safe_c_opscq_s", "h_safe_c_opscq_s", "safe_c_opcsq_s", "h_safe_c_opcsq_s",
     "safe_c_opscq_c", "h_safe_c_opscq_c", "safe_c_opqsq", "h_safe_c_opqsq", 
     "safe_c_opxxxq", "h_safe_c_opxxxq", 

     "safe_is_pair_s", "h_safe_is_pair_s", "safe_is_symbol_s", "h_safe_is_symbol_s", "safe_not_s", "h_safe_not_s",

     "c_ls", "h_c_ls", "c_l_opsq", "h_c_l_opsq", "c_l", "h_c_l", "c_ll", "h_c_ll", "c_cll", "h_c_cll",
     "c_all_g", "h_c_all_g", "call-with-exit", "h-call-with-exit", "c_catch", "h_c_catch",
     "c_s_opsq", "h_c_s_opsq", "c_opsq_cc", "h_c_opsq_cc",
     "c_for_each_ls", "h_c_for_each_ls", "c_for_each_ls_2", "h_c_for_each_ls_2",
     "c_for_each_l_opsq", "h_c_for_each_l_opsq",
     "c_s", "h_c_s", "read_s", "h_read_s", "list_p", "h_list_p",

     "vector_c", "h_vector_c", "vector_s", "h_vector_s", "vector_opcq", "h_vector_opcq", 
     "string_c", "h_string_c", "string_s", "h_string_s", "string_opcq", "h_string_opcq", 
     "c_object_c", "h_c_object_c", "c_object_s", "h_c_object_s", "c_object_opcq", "h_c_object_opcq", 
     "pair_c", "h_pair_c", "pair_s", "h_pair_s", "pair_opcq", "h_pair_opcq", 
     "hash_table_c", "h_hash_table_c", "hash_table_s", "h_hash_table_s", "hash_table_opcq", "h_hash_table_opcq", 
     "environment_s", "h_environment_s", 
     
     "unknown_c", "h_unknown_c", "unknown_s", "h_unknown_s",
     "unknown_ss", "h_unknown_ss", "unknown_sss", "h_unknown_sss", 
     "unknown_sc", "h_unknown_sc", "unknown_cs", "h_unknown_cs", "unknown_cc", "h_unknown_cc", 
     "unknown_opsq", "h_unknown_opsq", "unknown_opsq_s", "h_unknown_opsq_s", "unknown_opsq_opsq", "h_unknown_opsq_opsq",      
     "unknown_s_opsq", "h_unknown_s_opsq", "unknown_opcq", "h_unknown_opcq", 

     "safe_c_p", "h_safe_c_p", "safe_c_pp", "h_safe_c_pp",
     "safe_c_opsq_p", "h_safe_c_opsq_p", 
     "safe_c_sp", "h_safe_c_sp", "safe_c_cp", "h_safe_c_cp", 
     "safe_c_ps", "h_safe_c_ps", "safe_c_pc", "h_safe_c_pc", 

     "safe_c_z", "h_safe_c_z", "safe_c_zz", "h_safe_c_zz", "safe_c_sz", "h_safe_c_sz", "safe_c_zs", "h_safe_c_zs", 
     "safe_c_cz", "h_safe_c_cz", "safe_c_zc", "h_safe_c_zc", "safe_c_qz", "h_safe_c_qz", "safe_c_zq", "h_safe_c_zq", 

     "safe_c_zxx", "h_safe_c_zxx", "safe_c_xzx", "h_safe_c_xzx", "safe_c_xxz", "h_safe_c_xxz",
     "safe_c_zzx", "h_safe_c_zzx", "safe_c_zxz", "h_safe_c_zxz", "safe_c_xzz", "h_safe_c_xzz", 
     "safe_c_zzz", "h_safe_c_zzz",
     
     "safe_c_dd", "h_safe_c_dd",
     "safe_c_dxx", "h_safe_c_dxx", "safe_c_xdx", "h_safe_c_xdx", "safe_c_xxd", "h_safe_c_xxd",
     "safe_c_ddx", "h_safe_c_ddx", "safe_c_dxd", "h_safe_c_dxd", "safe_c_xdd", "h_safe_c_xdd", 
     "safe_c_ddd", "h_safe_c_ddd",

     "opt-max",
  };
#endif

#endif

#define WITH_COUNTS 0
#if WITH_COUNTS
static int counts[65000];
static void init_counts(void) {int i;for (i=0;i<65000;i++) counts[i]=0;}
static void report_counts(void) 
{
  int i;
  for (i=0;i<65000;i++)
    if (counts[i]>0)
      fprintf(stderr, "%d: %d\n", i, counts[i]);
}
#endif



#define NUM_SMALL_INTS 2048
#define NUM_CHARS 256

typedef enum {TOKEN_EOF, TOKEN_LEFT_PAREN, TOKEN_RIGHT_PAREN, TOKEN_DOT, TOKEN_ATOM, TOKEN_QUOTE, TOKEN_DOUBLE_QUOTE, 
	      TOKEN_BACK_QUOTE, TOKEN_COMMA, TOKEN_AT_MARK, TOKEN_SHARP_CONST, TOKEN_VECTOR} token_t;


typedef enum {FILE_PORT, STRING_PORT, FUNCTION_PORT} port_type_t;

typedef struct s7_port_t {
  bool is_closed;
  port_type_t type;
  bool needs_free;
  FILE *file;
  unsigned int line_number;
  unsigned int file_number;
  char *filename;
  char *value;
  unsigned int size, point;        /* these limit the in-core portion of a string-port to 2^31 bytes */
  s7_pointer (*input_function)(s7_scheme *sc, s7_read_t read_choice, s7_pointer port);
  void (*output_function)(s7_scheme *sc, unsigned char c, s7_pointer port);
  unsigned char *data;
  /* a version of string ports using a pointer to the current location and a pointer to the end
   *   (rather than an integer for both, indexing from the base string) was not faster.
   */
  s7_pointer orig_str;                                                   /* GC protection for string port string */
  int (*read_character)(s7_scheme *sc, s7_pointer port);                 /* function to read a character */
  void (*write_character)(s7_scheme *sc, int c, s7_pointer port);        /* function to write a character */
  token_t (*read_semicolon)(s7_scheme *sc, s7_pointer port);             /* internal skip-to-semicolon reader */
  int (*read_white_space)(s7_scheme *sc, s7_pointer port);               /* internal skip white space reader */
  s7_pointer (*read_name)(s7_scheme *sc, s7_pointer pt, bool atom_case); /* internal get-next-name reader */
  s7_pointer (*read_line)(s7_scheme *sc, s7_pointer pt, bool eol_case);  /* function to display or write a string */
  void (*display)(s7_scheme *sc, const char *s, s7_pointer pt);
} s7_port_t;


typedef struct s7_func_t {
  s7_function ff;
  const char *name;
  char *doc;
  unsigned int required_args, optional_args, all_args;
  bool rest_arg;
  s7_pointer setter;
#if WITH_OPTIMIZATION
  unsigned int id;
  s7_pointer (*chooser)(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr);
  void *chooser_data;
#endif
} s7_func_t;


typedef struct {               /* call/cc */
  s7_pointer stack;
  s7_pointer *stack_start, *stack_end, *op_stack;
  unsigned int stack_size, op_stack_loc, op_stack_size;
} s7_continuation_t;


typedef struct s7_vdims_t {
  unsigned int ndims;
  s7_Int *dims, *offsets;
  s7_pointer original;
} s7_vdims_t;


/* cell structure */
typedef struct s7_cell {
  union {
    unsigned int flag;
    unsigned char type_field;
  } tf;
  int hloc;
  union {
    
    struct { 
      unsigned int length;
      unsigned int hash;  
      char *svalue;
      bool needs_free;
    } string;
    
    union {
      s7_Int integer_value;
      s7_Double real_value;

      struct {
	s7_Int numerator;
	s7_Int denominator;
      } fraction_value;

      struct {
	s7_Double rl;
	s7_Double im;
      } complex_value;

      unsigned long ul_value; /* these two are not used by s7 in any way */
      unsigned long long ull_value;

#if WITH_GMP
      mpz_t big_integer;
      mpq_t big_ratio;
      mpfr_t big_real;
      mpc_t big_complex;
#endif
    } number;

    s7_port_t *port;
    
    unsigned char chr;

    void *c_pointer;
    
    struct {
      s7_Int length;
      s7_pointer *elements;
      union {
	s7_vdims_t *dim_info;
	unsigned int entries;          
      } vextra;
      s7_pointer (*hash_func)(s7_scheme *sc, s7_pointer table, s7_pointer key);
      unsigned int fill_ptr;
    } vector;
    
    s7_func_t *ffptr;      /* C functions, macros */
    
    struct {               /* pairs */ 
      s7_pointer car, cdr, ecdr, fcdr;
      unsigned int line;
      unsigned short data;
      unsigned short data_index;
    } cons;

    struct {               /* symbols */
      s7_pointer name, global_slot, local_slot;
      long long int id;
    } sym;

    struct {               /* slots (bindings) */
      s7_pointer sym, val, nxt, pending_value, expr;
    } slt;

    struct {               /* environments (frames) */
      s7_pointer slots, nxt;
      long long int id;
    } envr;

    struct {               /* counter (internal) */
      s7_pointer result, list;
      long long int val, len;
    } ctr;

    struct {               /* additional object types (C and Scheme) */
      int type;
      void *value;
      s7_pointer (*apply)(s7_scheme *sc, s7_pointer obj, s7_pointer args);
      s7_pointer (*set)(s7_scheme *sc, s7_pointer obj, s7_pointer args);
    } fobj;
    
    s7_continuation_t *continuation;
    
    struct {               /* call-with-exit */
      unsigned int goto_loc, op_stack_loc;
      bool active;
    } rexit;
    
    struct {               /* catch */
      unsigned int goto_loc, op_stack_loc;
      s7_pointer tag;
      s7_pointer handler;
    } rcatch; /* C++ reserves "catch" I guess */
    
    struct {               /* dynamic-wind */
      s7_pointer in, out, body;
      unsigned int state;
    } winder;

    struct {               /* hook */
      s7_pointer functions, arity, documentation;
    } hook;

  } object;
} s7_cell;


typedef struct {
  s7_cell obj;
  int accessor;
  void *accessor_data;     /* FFI-accessible */
  void *op_data;           /*   same purpose, but only s7-accessible */
} s7_extended_cell;


typedef struct {
  s7_pointer *objs;
  int size, top, ref;
  int *refs;
} shared_info;

typedef struct {
  char *str;
  int len, loc;
  s7_pointer args;
} format_data;

static s7_pointer *small_ints, *chars;
static s7_pointer real_zero, real_NaN, real_pi, real_one; 

struct s7_scheme {  
  opcode_t op;                        /* making this global is much slower! */
  s7_pointer value;
  s7_pointer args;                    /* arguments of current function */
  s7_pointer code, cur_code;          /* current code */
  s7_pointer envir;                   /* current environment */
  token_t tok;

  s7_pointer stack;                   /* stack is a vector */
  unsigned int stack_size;
  s7_pointer *stack_start, *stack_end, *stack_resize_trigger;
  
  s7_pointer *op_stack, *op_stack_now, *op_stack_end;
  unsigned int op_stack_size;

  s7_cell **heap, **free_heap, **free_heap_top, **free_heap_trigger, **previous_free_heap_top;
  unsigned int heap_size;

  /* "int" or "unsigned int" seems safe here:
   *      sizeof(s7_cell) = 28 in 32-bit machines, 32 in 64
   *      so to get more than 2^32 actual objects would require ca 140 GBytes RAM
   *      vectors might be full of the same object (sc->NIL for example), so there
   *      we need ca 38 GBytes RAM (8 bytes per pointer).  
   */
  
  s7_pointer protected_objects;       /* a vector of gc-protected objects */
  unsigned int protected_objects_size, protected_objects_loc;

  struct s7_cell _NIL;
  s7_pointer NIL;                     /* empty list */
  
  struct s7_cell _T;
  s7_pointer T;                       /* #t */
  
  struct s7_cell _F;
  s7_pointer F;                       /* #f */
  
  struct s7_cell _EOF_OBJECT;         /* can't use EOF here because it's taken by C or glibc? */
  s7_pointer EOF_OBJECT;              /* #<eof> */
  
  struct s7_cell _UNDEFINED;  
  s7_pointer UNDEFINED;               /* #<undefined> */
  
  struct s7_cell _UNSPECIFIED;
  s7_pointer UNSPECIFIED;             /* #<unspecified> */
  
  struct s7_cell _NO_VALUE;
  s7_pointer NO_VALUE;                /* the (values) value */

  struct s7_cell _ELSE;
  s7_pointer ELSE;                    /* else */  

  struct s7_cell _GC_NIL;             /* marker for empty slot in gc-protected vector */
  s7_pointer GC_NIL;
  
  s7_pointer symbol_table;            /* symbol table */
  s7_pointer global_env;              /* global environment */
  s7_pointer initial_env;             /* original bindings of predefined functions */
  
  s7_pointer LAMBDA, LAMBDA_STAR, QUOTE, UNQUOTE, MACROEXPAND;
  s7_pointer APPLY, VECTOR, CDR, SET, QQ_VALUES, QQ_LIST, QQ_APPLY_VALUES, QQ_APPEND, MULTIVECTOR;
  s7_pointer ERROR, WRONG_TYPE_ARG, WRONG_TYPE_ARG_INFO, OUT_OF_RANGE, OUT_OF_RANGE_INFO;
  s7_pointer SIMPLE_WRONG_TYPE_ARG_INFO, SIMPLE_OUT_OF_RANGE_INFO;
  s7_pointer FORMAT_ERROR, WRONG_NUMBER_OF_ARGS, READ_ERROR, SYNTAX_ERROR, TOO_MANY_ARGUMENTS, NOT_ENOUGH_ARGUMENTS;
  s7_pointer KEY_KEY, KEY_OPTIONAL, KEY_REST, KEY_ALLOW_OTHER_KEYS;
  s7_pointer __FUNC__;
  s7_pointer OBJECT_SET;              /* applicable object set method */
  s7_pointer FEED_TO;                 /* => */
  s7_pointer VECTOR_SET, STRING_SET, LIST_SET, HASH_TABLE_SET, HASH_TABLE_ITERATE, ENVIRONMENT_SET;
  s7_pointer S_IS_TYPE, S_TYPE_MAKE, S_TYPE_REF, S_TYPE_ARG;
  s7_pointer s_function_args;
  s7_pointer QUOTE_UNCHECKED, CASE_UNCHECKED, SET_UNCHECKED, LAMBDA_UNCHECKED, LET_UNCHECKED;
  s7_pointer LET_STAR_UNCHECKED, LETREC_UNCHECKED, COND_UNCHECKED, COND_SIMPLE;
  s7_pointer SET_SYMBOL_C, SET_SYMBOL_S, SET_SYMBOL_Q, SET_SYMBOL_P, SET_SYMBOL_R, SET_NORMAL, SET_PAIR;
  s7_pointer LAMBDA_STAR_UNCHECKED, DO_UNCHECKED, DEFINE_UNCHECKED, DEFINE_STAR_UNCHECKED, CASE_SIMPLE, CASE_SIMPLER, CASE_SIMPLEST, CASE_INT;
  s7_pointer LET_C, LET_S, LET_Q, LET_ALL_C, LET_ALL_S, LET_ALL_Q, LET_ALL_G, LET_C_P, LET_S_P;
  s7_pointer LET_NO_VARS, NAMED_LET, NAMED_LET_NO_VARS, AND_UNCHECKED, AND_P, OR_UNCHECKED, OR_P;
  s7_pointer IF_P_P_P, IF_P_P, IF_B_P, IF_B_P_P, IF_P_P_X, IF_P_X_P, IF_P_X, IF_P_X_X, IF_X_P_P, IF_X_P, IF_X_P_X, IF_X_X_P, IF_X_X, IF_X_X_X;
  s7_pointer IF_O_P, IF_O_P_P, SAFE_IF_ANDX_P, SAFE_IF_ORX_P, IF_ANDP_P, IF_ANDP_P_P, IF_ORP_P, IF_ORP_P_P;

#if WITH_OPTIMIZATION
  s7_pointer SAFE_OR_S, SAFER_OR_S, SAFER_OR_C, SAFER_OR_CEQ, SAFE_AND_S, SAFE_CASE_S, SAFER_OR_X, SAFER_AND_X, COND_SIMPLER;
  s7_pointer SAFE_IF1, SAFE_IF2, SAFE_IF_CC_X_P, SAFE_IF_CC_P_P, SAFE_IF_CEQ_P_P;
  s7_pointer SAFE_IF_CS_P_P, SAFE_IF_CS_X_P, SAFE_IF_CC_P, SAFE_IF_CS_P, SAFE_IF_CS_P_X, SAFE_IF_CS_Q_P;
  s7_pointer SAFE_IF_CSS_X_P, SAFE_IF_CSC_X_P, SAFE_IF_CSC_X_P_A;
  s7_pointer SAFE_IF_CSQ_P, SAFE_IF_CSQ_P_P, SAFE_IF_CSS_P, SAFE_IF_CSS_P_P, SAFE_IF_CSC_P, SAFE_IF_CSC_P_P;
  s7_pointer SAFE_IF_IS_PAIR_P, SAFE_IF_IS_PAIR_P_X, SAFE_IF_IS_PAIR_P_P, SAFE_IF_C_SS_P;
  s7_pointer SAFE_IF_IS_SYMBOL_P, SAFE_IF_IS_SYMBOL_P_X, SAFE_IF_IS_SYMBOL_P_P, SAFE_IF_IS_EOF_P_P;
  s7_pointer INCREMENT_1, DECREMENT_1, SET_CDR, SET_CONS;
  s7_pointer LET_R, LET_O, LET_ALL_R, LET_C_D, LET_O_P, LET_R_P, LET_OX_P, LET_UNKNOWN_S_P, LET_READ_CHAR_P, LET_CAR_P;
  s7_pointer SIMPLE_DO, DOTIMES, SIMPLE_DOTIMES, DOTIMES_C_C, SAFE_DO, SIMPLE_DO_P, SAFE_SIMPLE_DO, SIMPLE_DO_OPT;
  s7_pointer DOX, IF_ORCEQ_P, IF_ORCEQ_P_P;
  int safe_do_level, safe_do_ids_size;
  long long int *safe_do_ids;
#endif
  
  s7_pointer input_port;              /* current-input-port */
  s7_pointer input_port_stack;        /*   input port stack (load and read internally) */
  s7_pointer output_port;             /* current-output-port */
  s7_pointer error_port;              /* current-error-port */
  s7_pointer error_info;              /* the vector bound to *error-info* */
  s7_pointer standard_input, standard_output, standard_error;

  s7_pointer sharp_readers;           /* the binding pair for the global *#readers* list */
  s7_pointer vector_print_length;     /* same for *vector-print-length* */
  s7_pointer trace_hook;              /* *trace-hook* hook object */
  s7_pointer load_hook;               /* *load-hook* hook object */
  s7_pointer unbound_variable_hook;   /* *unbound-variable-hook* hook object */
  s7_pointer error_hook;              /* *error-hook* hook object */

  bool gc_off, gc_stats;              /* gc_off: if true, the GC won't run, gc_stats: if true, print stats during GC */
  unsigned int gensym_counter;
  bool symbol_table_is_locked;  

  #define INITIAL_STRBUF_SIZE 1024
  unsigned int strbuf_size;
  char *strbuf;
  int print_width;
  
  char *read_line_buf;
  unsigned int read_line_buf_size;

  s7_pointer w, x, y, z;         /* evaluator local vars */
  s7_pointer temp1, temp2, temp3, temp4;

  struct s7_cell _TEMP_CELL, _TEMP_CELL_1, _TEMP_CELL_2, _TEMP_CELL_3;
  s7_pointer TEMP_CELL, TEMP_CELL_1, TEMP_CELL_2, TEMP_CELL_3;
  struct s7_cell _T1_1, _T2_1, _T2_2, _T3_1, _T3_2, _T3_3;
  s7_pointer T1_1, T2_1, T2_2, T3_1, T3_2, T3_3;
  s7_pointer *all_g_args;

  jmp_buf goto_start;
  bool longjmp_ok;
  void (*error_exiter)(void);
  bool (*begin_hook)(s7_scheme *sc);
  
  s7_pointer *trace_list;
  int trace_list_size, trace_top, trace_depth;
  int no_values, current_line, s7_call_line, safety;
  const char *current_file, *s7_call_file, *s7_call_name;

  shared_info *circle_info;
  format_data **fdats;
  int num_fdats;

  s7_pointer *strings, *vectors, *input_ports, *output_ports, *continuations, *c_objects, *hash_tables;
  int strings_size, vectors_size, input_ports_size, output_ports_size, continuations_size, c_objects_size, hash_tables_size;
  int strings_loc, vectors_loc, input_ports_loc, output_ports_loc, continuations_loc, c_objects_loc, hash_tables_loc;

  void *default_rng;
#if WITH_GMP
  void *default_big_rng;
#endif
};



/* is_c_function can be array lookup, as can T_PROCEDURE(?), T_ANY_MACRO(?)
 *  what about T_EXPANSION? T_KEYWORD? T_COPY_ARGS? or checks like typeflag against T_PROCEDURE|T_ANY_MACRO
 *  the array lookup has exactly the same computational cost as the bit check
 */

#define T_UNTYPED              0
#define T_NIL                  1
#define T_BOOLEAN              2
#define T_CHARACTER            3
#define T_SYMBOL               4
#define T_SYNTAX               5

#define T_INTEGER              6
#define T_RATIO                7
#define T_REAL                 8
#define T_COMPLEX              9

#define T_BIG_INTEGER         10
#define T_BIG_RATIO           11
#define T_BIG_REAL            12
#define T_BIG_COMPLEX         13
/* only used in WITH_GMP case -- order matters */

#define T_STRING              14
#define T_PAIR                15
#define T_CLOSURE             16
#define T_CLOSURE_STAR        17
#define T_CONTINUATION        18
#define T_VECTOR              19
#define T_MACRO               20
#define T_BACRO               21
#define T_C_OBJECT            22
#define T_GOTO                23
#define T_CATCH               24
#define T_DYNAMIC_WIND        25
#define T_HASH_TABLE          26
#define T_HOOK                27
#define T_ENVIRONMENT         28
#define T_STACK               29
#define T_COUNTER             30
#define T_SLOT                31
#define T_C_POINTER           32
#define T_C_MACRO             33
#define T_OUTPUT_PORT         34
#define T_INPUT_PORT          35

#define T_C_FUNCTION          36
#define T_C_ANY_ARGS_FUNCTION 37
#define T_C_OPT_ARGS_FUNCTION 38
#define T_C_RST_ARGS_FUNCTION 39
#define T_C_LST_ARGS_FUNCTION 40

#define BUILT_IN_TYPES        41


static bool t_number_p[BUILT_IN_TYPES], t_real_p[BUILT_IN_TYPES], t_rational_p[BUILT_IN_TYPES];
static bool t_simple_p[BUILT_IN_TYPES];

#if WITH_GMP
  static bool t_big_number_p[BUILT_IN_TYPES];
#endif

static void init_types(void)
{
  int i;
  for (i = 0; i < BUILT_IN_TYPES; i++) 
    {
      t_number_p[i] = false;
      t_real_p[i] = false;
      t_rational_p[i] = false;

      t_simple_p[i] = false;
    }
  t_number_p[T_INTEGER] = true;
  t_number_p[T_RATIO] = true;
  t_number_p[T_REAL] = true;
  t_number_p[T_COMPLEX] = true;
  
  t_rational_p[T_INTEGER] = true;
  t_rational_p[T_RATIO] = true;
  
  t_real_p[T_INTEGER] = true;
  t_real_p[T_RATIO] = true;
  t_real_p[T_REAL] = true;

#if WITH_GMP
  t_big_number_p[T_BIG_INTEGER] = true;
  t_big_number_p[T_BIG_RATIO] = true;
  t_big_number_p[T_BIG_REAL] = true;
  t_big_number_p[T_BIG_COMPLEX] = true;
#endif

  t_simple_p[T_NIL] = true;
  t_simple_p[T_UNTYPED] = true;
  t_simple_p[T_BOOLEAN] = true;
  t_simple_p[T_CHARACTER] = true;
  t_simple_p[T_SYMBOL] = true;
  t_simple_p[T_SYNTAX] = true;
  t_simple_p[T_C_MACRO] = true;
  t_simple_p[T_C_FUNCTION] = true;
  t_simple_p[T_C_ANY_ARGS_FUNCTION] = true;
  t_simple_p[T_C_OPT_ARGS_FUNCTION] = true;
  t_simple_p[T_C_RST_ARGS_FUNCTION] = true;
  t_simple_p[T_C_LST_ARGS_FUNCTION] = true;
  /* not completely sure about the next ones */
  t_simple_p[T_ENVIRONMENT] = true;
  t_simple_p[T_INPUT_PORT] = true;
  t_simple_p[T_OUTPUT_PORT] = true;
}

/* T_STACK, T_SLOT, and T_COUNTER are internal (stacks, bindings, map circular list checks)
 */

#define TYPE_BITS                     8
#define typeflag(p)                   ((p)->tf.flag)
#define type(p)                       ((p)->tf.type_field)

#define is_number(P)                  t_number_p[type(P)]
#define is_integer(P)                 (type(P) == T_INTEGER)
#define is_real(P)                    t_real_p[type(P)]
#define is_rational(P)                t_rational_p[type(P)]

#define is_simple(P)                  t_simple_p[type(P)]


#define NOT_IN_HEAP -1
#define T_IMMUTABLE                   (1 << (TYPE_BITS + 0))
#define is_immutable(p)               ((typeflag(p) & T_IMMUTABLE) != 0)
#define set_immutable(p)              typeflag(p) |= T_IMMUTABLE
/* immutable means the value can't be changed via set! or bind -- this is separate from the symbol access stuff
 */
#define dont_copy(p)                  ((!is_pair(p)) || (is_immutable(p)) || (p->hloc == NOT_IN_HEAP))
/* dont_copy means the object is not copied when saved in a continuation */


#define T_PROCEDURE                   (1 << (TYPE_BITS + 2))
#define is_procedure(p)               ((typeflag(p) & T_PROCEDURE) != 0)
/* closure, c_function, procedure-with-setter, settable object, goto or continuation 
 *   perhaps this should be T_APPLICABLE?
 */

#define T_SAFE_PROCEDURE              (1 << (TYPE_BITS + 3))
#define is_safe_procedure(p)          ((typeflag(p) & T_SAFE_PROCEDURE) != 0)
/* c_functions that do not return or modify the arg list directly (no :rest arg in particular),
 *    and that can't call apply themselves either directly or via s7_call.
 *    I think the latter would be safe if they gc_protect the called function.  
 * This also includes applicable objects that are known safe (lists, vectors, etc -- equivalent to vector-ref).
 */

#define T_ANY_MACRO                   (1 << (TYPE_BITS + 4))
#define is_any_macro(p)               ((typeflag(p) & T_ANY_MACRO) != 0)
/* this marks scheme and C-defined macros 
 */

#define T_DONT_EVAL_ARGS              (1 << (TYPE_BITS + 5))
#define dont_eval_args(p)             ((typeflag(p) & T_DONT_EVAL_ARGS) != 0)
/* this marks things that don't evaluate their arguments
 */

#define T_EXPANSION                   (1 << (TYPE_BITS + 6))
#define EXPANSION_TYPE                (T_EXPANSION | T_SYMBOL)
#define is_expansion(p)               (typeflag(p) == EXPANSION_TYPE)
/* this marks the symbol associated with a run-time macro and distinguishes the value from an ordinary macro
 */

#define T_GLOBAL                      (1 << (TYPE_BITS + 7))
#define is_global(p)                  ((typeflag(p) & T_GLOBAL) != 0)
/* this marks something defined (bound) at the top-level, and never defined locally */
#define set_global(p)                 typeflag(p) |= T_GLOBAL
#define set_local(p)                  typeflag(p) = (typeflag(p) & ~(T_DONT_EVAL_ARGS | T_GLOBAL | T_SYNTACTIC))


#define T_MULTIPLE_VALUE              (1 << (TYPE_BITS + 8))
#define is_multiple_value(p)          ((typeflag(p) & T_MULTIPLE_VALUE) != 0)
#define set_multiple_value(p)         typeflag(p) |= T_MULTIPLE_VALUE
#define multiple_value(p)             p
/* this bit marks a list (from "values") that is waiting for a
 *    chance to be spliced into its caller's argument list.  It is normally
 *    on only for a very short time.
 */

#define T_TAIL_CALL                   (1 << (TYPE_BITS + 9))
#define is_tail_call(p)               ((typeflag(p) & T_TAIL_CALL) != 0)
#define set_tail_call(p)              typeflag(p) |= T_TAIL_CALL
/* this bit marks tail calls for the optimizer
 */

#define T_KEYWORD                     (1 << (TYPE_BITS + 10))
#define is_keyword(p)                 ((typeflag(p) & T_KEYWORD) != 0)
/* this bit distinguishes a symbol from a symbol that is also a keyword
 */

#define T_SYNTACTIC                   (1 << (TYPE_BITS + 11))
#define is_syntactic(p)               ((typeflag(p) & T_SYNTACTIC) != 0)
#define SYNTACTIC_TYPE                (T_SYMBOL | T_DONT_EVAL_ARGS | T_SYNTACTIC)
/* this marks symbols that represent syntax objects 
 */

#define T_OVERLAY                     (1 << (TYPE_BITS + 12))
#define set_overlay(p)                typeflag(p) |= T_OVERLAY
#define is_overlaid(p)                ((typeflag(p) & T_OVERLAY) != 0)
/* optimizer flag that marks a cell whose ecdr points to the previous cell in a list
 */

#define T_OPTIMIZED                   (1 << (TYPE_BITS + 13))
#define set_optimized(p)              typeflag(p) |= T_OPTIMIZED
#define is_optimized(p)               ((typeflag(p) & T_OPTIMIZED) != 0)
#define clear_optimized(p)            typeflag(p) &= ~(T_OPTIMIZED)
/* optimizer flag for an expression that has optimization info
 */

#define T_CHECKED                     (1 << (TYPE_BITS + 14))
#define set_checked(p)                typeflag(p) |= T_CHECKED
#define is_checked(p)                 ((typeflag(p) & T_CHECKED) != 0)
#define is_not_checked(p)             ((typeflag(p) & T_CHECKED) == 0)

#define T_CHECKED_1                   (1 << (TYPE_BITS + 1))
#define set_checked_1(p)              typeflag(p) |= T_CHECKED_1
#define is_checked_1(p)               ((typeflag(p) & T_CHECKED_1) != 0)
/* optimizer flags that an expression has been checked for possible optimization 
 *    (trying to avoid circular lists during the optimization scan)
 */

#define T_UNSAFE                      (1 << (TYPE_BITS + 15))
#define set_unsafe(p)                 typeflag(p) |= T_UNSAFE
#define is_unsafe(p)                  ((typeflag(p) & T_UNSAFE) != 0)
#define clear_unsafe(p)               typeflag(p) &= (~T_UNSAFE)
/* optimizer flag saying "this expression is not completely self-contained.  It might involve the stack, etc"
 */

#define T_SAFE_CLOSURE                (1 << (TYPE_BITS + 16))
#define is_safe_closure(p)            ((typeflag(p) & T_SAFE_CLOSURE) != 0)
#define set_safe_closure(p)           typeflag(p) = (T_SAFE_CLOSURE | (typeflag(p) & (~T_COPY_ARGS)))
#define clear_safe_closure(p)         typeflag(p) = (T_COPY_ARGS | (typeflag(p) & (~T_SAFE_CLOSURE)))
/* optimizer flag for a closure body that is completely simple (every expression is safe)
 */

#define T_SETTER                      (1 << (TYPE_BITS + 17))
#define set_setter(p)                 typeflag(p) |= T_SETTER
#define is_setter(p)                  ((typeflag(p) & T_SETTER) != 0)
/* optimizer flag for a procedure that sets some variable (set-car! for example).
 */

#define T_HAS_TABLE                   (1 << (TYPE_BITS + 18))
#define set_has_table(p)              typeflag(p) |= T_HAS_TABLE
#define has_table(p)                  ((typeflag(p) & T_HAS_TABLE) != 0)
/* optimizer flag (external function using a function table for variable access)
 */

#define T_ONE_LINER                   (1 << (TYPE_BITS + 19))
#define set_one_liner(p)              typeflag(p) |= T_ONE_LINER
#define is_one_liner(p)               ((typeflag(p) & T_ONE_LINER) != 0)
/* implicit begin optimization -- I need to do this differently
 */

#define T_COPY_ARGS                   (1 << (TYPE_BITS + 20))
#define set_copy_args(p)              typeflag(p) |= T_COPY_ARGS
#define needs_copied_args(p)          ((typeflag(p) & T_COPY_ARGS) != 0)
#define clear_copy_args(p)            typeflag(p) &= (~T_COPY_ARGS)
/* this marks something that might mess with its argument list
 */

#define T_GC_MARK                     (1 << (TYPE_BITS + 23))
#define is_marked(p)                  ((typeflag(p) &  T_GC_MARK) != 0)
/* #define is_marked(p)               (typeflag(p) >= T_GC_MARK) */ /* exactly the same speed? */ 
#define set_mark(p)                   typeflag(p) |= T_GC_MARK
#define clear_mark(p)                 typeflag(p) &= (~T_GC_MARK)
/* using bit 23 for this makes a big difference in the GC
 */

#define UNUSED_BITS                   0x60000000

#define set_type(p, f)                typeflag(p) = f

#define is_true(Sc, p)                ((p) != Sc->F)
#define is_false(Sc, p)               ((p) == Sc->F)
#ifdef _MSC_VER
  #define make_boolean(sc, Val)       (((Val) & 0xff) ? sc->T : sc->F)
#else
  #define make_boolean(sc, Val)       ((Val) ? sc->T : sc->F)
#endif

#define is_unspecified(p)             (p == sc->UNSPECIFIED)

#define is_pair(p)                    (type(p) == T_PAIR)
#define is_null(p)                    (p == sc->NIL)
#define is_not_null(p)                (p != sc->NIL)

#define car(p)                        ((p)->object.cons.car)
#define cdr(p)                        ((p)->object.cons.cdr)
#define ecdr(p)                       ((p)->object.cons.ecdr)
#define fcdr(p)                       ((p)->object.cons.fcdr)

#define caar(p)                       car(car(p))
#define cadr(p)                       car(cdr(p))
#define cdar(p)                       cdr(car(p))
#define cddr(p)                       cdr(cdr(p))

#define caaar(p)                      car(car(car(p)))
#define cadar(p)                      car(cdr(car(p)))
#define cdadr(p)                      cdr(car(cdr(p)))
#define caddr(p)                      car(cdr(cdr(p)))
#define caadr(p)                      car(car(cdr(p)))
#define cdaar(p)                      cdr(car(car(p)))
#define cdddr(p)                      cdr(cdr(cdr(p)))
#define cddar(p)                      cdr(cdr(car(p)))

#define caaadr(p)                     car(car(car(cdr(p))))
#define caadar(p)                     car(car(cdr(car(p))))
#define cadaar(p)                     car(cdr(car(car(p))))
#define cadddr(p)                     car(cdr(cdr(cdr(p))))
#define caaddr(p)                     car(car(cdr(cdr(p))))
#define cddddr(p)                     cdr(cdr(cdr(cdr(p))))
#define caddar(p)                     car(cdr(cdr(car(p))))
#define cdadar(p)                     cdr(car(cdr(car(p))))
#define cdaddr(p)                     cdr(car(cdr(cdr(p))))
#define caaaar(p)                     car(car(car(car(p))))
#define cadadr(p)                     car(cdr(car(cdr(p))))
#define cdaadr(p)                     cdr(car(car(cdr(p))))
#define cdaaar(p)                     cdr(car(car(car(p))))
#define cdddar(p)                     cdr(cdr(cdr(car(p))))
#define cddadr(p)                     cdr(cdr(car(cdr(p))))
#define cddaar(p)                     cdr(cdr(car(car(p))))

#define cadddar(p)                    car(cdr(cdr(cdr(car(p)))))



#if WITH_GCC
  /* slightly tricky because cons can be called recursively */
  #define cons(Sc, A, B)             ({s7_pointer _X_; NEW_CELL(sc, _X_); car(_X_) = A; cdr(_X_) = B; set_type(_X_, T_PAIR | T_SAFE_PROCEDURE); _X_;})
#else
  #define cons(Sc, A, B)              s7_cons(Sc, A, B)
#endif

#define list_1(Sc, A)                 cons(Sc, A, sc->NIL)
#define list_2(Sc, A, B)              cons_unchecked(Sc, A, cons(Sc, B, sc->NIL))
#define list_3(Sc, A, B, C)           cons_unchecked(Sc, A, cons_unchecked(Sc, B, cons(Sc, C, sc->NIL)))
#define list_4(Sc, A, B, C, D)        cons_unchecked(Sc, A, cons_unchecked(Sc, B, cons_unchecked(Sc, C, cons(Sc, D, sc->NIL))))

#define pair_line_number(p)           (p)->object.cons.line
#define port_file_number(p)           (p)->object.port->file_number

#define optimize_data(p)              (p)->object.cons.data
#define set_optimize_data(P, X)       (P)->object.cons.data = ((X) + hop) /* this unchecking saves about 40 currently (2%) */
#define clear_optimize_data(P)        optimize_data(P) = 0
#define optimize_data_match(P, Q)     ((optimize_data(P) & 0xfffe) == Q)
#define optimize_data_index(p)        (p)->object.cons.data_index
#define is_hopping(P)                 ((optimize_data(P) & 0x1) == 1)

#define symbol_id(p)                  (p)->object.sym.id
/* we need 64-bits here, I think, since we don't want this thing to wrap around, and frames are created at a great rate 
 *    callgrind says this is faster than an unsigned int!
 */
#define string_value(p)               (p)->object.string.svalue
#define string_length(p)              (p)->object.string.length
#define string_hash(p)                (p)->object.string.hash
#define string_needs_free(p)          (p)->object.string.needs_free
#define character(p)                  (p)->object.chr

#define is_symbol(p)                  (type(p) == T_SYMBOL)
#define symbol_name_cell(p)           (p)->object.sym.name
#define symbol_name(p)                string_value(symbol_name_cell(p))
#define symbol_name_length(p)         string_length(symbol_name_cell(p))
#define symbol_hash(p)                (symbol_name_cell(p))->object.string.hash
#define symbol_accessor(p)            ((s7_extended_cell *)p)->accessor
#define symbol_has_accessor(p)        (symbol_accessor(p) != -1)
#define symbol_accessor_data(p)       ((s7_extended_cell *)p)->accessor_data
#define symbol_op_data(p)             ((s7_extended_cell *)p)->op_data

#define global_slot(p)                (p)->object.sym.global_slot
#define local_slot(p)                 (p)->object.sym.local_slot

#define is_slot(p)                    (type(p) == T_SLOT)
#define slot_value(p)                 (p)->object.slt.val
#define slot_set_value(p, Val)        (p)->object.slt.val = Val
#define slot_symbol(p)                (p)->object.slt.sym
#define next_slot(p)                  (p)->object.slt.nxt
#define slot_pending_value(p)         (p)->object.slt.pending_value
#define slot_expression(p)            (p)->object.slt.expr

#define is_syntax(p)                  (type(p) == T_SYNTAX)
#define syntax_opcode(p)              ((p)->hloc)

#define frame_id(p)                   (p)->object.envr.id
#define is_environment(p)             (type(p) == T_ENVIRONMENT)
#define environment_slots(p)          (p)->object.envr.slots
#define next_environment(p)           (p)->object.envr.nxt

#define vector_length(p)              ((p)->object.vector.length)
#define vector_element(p, i)          ((p)->object.vector.elements[i])
#define vector_elements(p)            (p)->object.vector.elements
#define vector_fill_pointer(p)        ((p)->object.vector.fill_ptr)

#define vector_dimension(p, i)        ((p)->object.vector.vextra.dim_info->dims[i])
#define vector_ndims(p)               ((p)->object.vector.vextra.dim_info->ndims)
#define vector_offset(p, i)           ((p)->object.vector.vextra.dim_info->offsets[i])
#define vector_is_multidimensional(p) ((p)->object.vector.vextra.dim_info)
#define shared_vector(p)              ((p)->object.vector.vextra.dim_info->original)

#define hash_table_length(p)          (p)->object.vector.length
#define hash_table_elements(p)        (p)->object.vector.elements
#define hash_table_entries(p)         (p)->object.vector.vextra.entries
#define hash_table_function(p)        (p)->object.vector.hash_func

#define is_input_port(p)              (type(p) == T_INPUT_PORT) 
#define is_output_port(p)             (type(p) == T_OUTPUT_PORT)
#define is_string_port(p)             ((p)->object.port->type == STRING_PORT)
#define is_file_port(p)               ((p)->object.port->type == FILE_PORT)
#define is_function_port(p)           ((p)->object.port->type == FUNCTION_PORT)
#define port_type(p)                  (p)->object.port->type
#define port_line_number(p)           (p)->object.port->line_number
#define port_filename(p)              (p)->object.port->filename
#define port_file(p)                  (p)->object.port->file
#define port_is_closed(p)             (p)->object.port->is_closed
#define port_string(p)                (p)->object.port->value
#define port_string_length(p)         (p)->object.port->size
#define port_string_point(p)          (p)->object.port->point
#define port_needs_free(p)            (p)->object.port->needs_free
#define port_output_function(p)       (p)->object.port->output_function
#define port_input_function(p)        (p)->object.port->input_function
#define port_data(p)                  (p)->object.port->data
#define port_position(p)              (p)->object.port->point
#define port_original_input_string(p) (p)->object.port->orig_str
#define port_read_character(p)        (p)->object.port->read_character
#define port_read_line(p)             (p)->object.port->read_line
#define port_display(p)               (p)->object.port->display
#define port_write_character(p)       (p)->object.port->write_character
#define port_read_semicolon(p)        (p)->object.port->read_semicolon
#define port_read_white_space(p)      (p)->object.port->read_white_space
#define port_read_name(p)             (p)->object.port->read_name

#define is_c_function(f)              (type(f) >= T_C_FUNCTION)
#define c_function(f)                 (f)->object.ffptr
#define c_function_call(f)            (f)->object.ffptr->ff
#define c_function_name(f)            (f)->object.ffptr->name
#define c_function_documentation(f)   (f)->object.ffptr->doc
#define c_function_required_args(f)   (f)->object.ffptr->required_args
#define c_function_optional_args(f)   (f)->object.ffptr->optional_args
#define c_function_has_rest_arg(f)    (f)->object.ffptr->rest_arg
#define c_function_all_args(f)        (f)->object.ffptr->all_args
#define c_function_setter(f)          (f)->object.ffptr->setter
#if WITH_OPTIMIZATION
#define c_function_class(f)           (f)->object.ffptr->id
#define c_function_chooser(f)         (f)->object.ffptr->chooser
#define c_function_chooser_data(f)    (f)->object.ffptr->chooser_data
#define c_function_choice(Sc, F, Args, X) (c_function_chooser(F))(Sc, F, Args, X)
#endif

#define c_call(f)                     ((s7_function)(fcdr(f)))
#define set_c_function(f, X)          do {ecdr(f) = X; fcdr(f) = (s7_pointer)(c_function_call(ecdr(f)));} while (0)

#define is_c_macro(p)                 (type(p) == T_C_MACRO)
#define c_macro_call(f)               (f)->object.ffptr->ff
#define c_macro_name(f)               (f)->object.ffptr->name
#define c_macro_required_args(f)      (f)->object.ffptr->required_args
#define c_macro_all_args(f)           (f)->object.ffptr->all_args

#define continuation(p)               (p)->object.continuation
#define continuation_stack(p)         (p)->object.continuation->stack
#define continuation_stack_end(p)     (p)->object.continuation->stack_end
#define continuation_stack_start(p)   (p)->object.continuation->stack_start
#define continuation_stack_size(p)    (p)->object.continuation->stack_size
#define continuation_stack_top(p)     (continuation_stack_end(p) - continuation_stack_start(p))
#define continuation_op_stack(p)      (p)->object.continuation->op_stack
#define continuation_op_loc(p)        (p)->object.continuation->op_stack_loc
#define continuation_op_size(p)       (p)->object.continuation->op_stack_size

#define call_exit_goto_loc(p)         (p)->object.rexit.goto_loc
#define call_exit_op_loc(p)           (p)->object.rexit.op_stack_loc
#define call_exit_active(p)           (p)->object.rexit.active

#define s7_stack_top(Sc)              ((Sc)->stack_end - (Sc)->stack_start)

#define is_continuation(p)            (type(p) == T_CONTINUATION)
#define is_goto(p)                    (type(p) == T_GOTO)
#define is_macro(p)                   (type(p) == T_MACRO)
#define is_bacro(p)                   (type(p) == T_BACRO)

#define is_closure(p)                 (type(p) == T_CLOSURE)
#define is_closure_star(p)            (type(p) == T_CLOSURE_STAR)
#define closure_args(Obj)             car(Obj)
#define closure_body(Obj)             cdr(Obj)
#define closure_environment(Obj)      ecdr(Obj)

#define catch_tag(p)                  (p)->object.rcatch.tag
#define catch_goto_loc(p)             (p)->object.rcatch.goto_loc
#define catch_op_loc(p)               (p)->object.rcatch.op_stack_loc
#define catch_handler(p)              (p)->object.rcatch.handler

enum {DWIND_INIT, DWIND_BODY, DWIND_FINISH};
#define dynamic_wind_state(p)         (p)->object.winder.state
#define dynamic_wind_in(p)            (p)->object.winder.in
#define dynamic_wind_out(p)           (p)->object.winder.out
#define dynamic_wind_body(p)          (p)->object.winder.body

#define is_c_object(p)                (type(p) == T_C_OBJECT)
#define object_type(p)                (p)->object.fobj.type
#define object_value(p)               (p)->object.fobj.value
#define object_ref(p)                 (p)->object.fobj.apply
#define object_set(p)                 (p)->object.fobj.set

#define is_hook(p)                    (type(p) == T_HOOK)
#define hook_arity(p)                 (p)->object.hook.arity
#define hook_functions(p)             (p)->object.hook.functions
#define hook_documentation(p)         (p)->object.hook.documentation

#define raw_pointer(p)                (p)->object.c_pointer

#define is_counter(p)                 (type(p) == T_COUNTER)
#define counter_result(p)             (p)->object.ctr.result
#define counter_list(p)               (p)->object.ctr.list
#define counter_count(p)              (p)->object.ctr.val
#define counter_length(p)             (p)->object.ctr.len


#if __cplusplus
  using namespace std;
  typedef complex<s7_Double> s7_Complex;
  static s7_Double Real(complex<s7_Double> x) {return(real(x));} /* protect the C++ name */
  static s7_Double Imag(complex<s7_Double> x) {return(imag(x));}
#endif


#define small_int(Val)                small_ints[Val]

#define integer(p)                    p->object.number.integer_value
#define real(p)                       p->object.number.real_value
#define numerator(p)                  p->object.number.fraction_value.numerator
#define denominator(p)                p->object.number.fraction_value.denominator
#define fraction(p)                   (((long double)numerator(p)) / ((long double)denominator(p)))
#define inverted_fraction(p)          (((long double)denominator(p)) / ((long double)numerator(p)))
#define real_part(p)                  p->object.number.complex_value.rl
#define imag_part(p)                  p->object.number.complex_value.im
#define as_c_complex(p)               (real_part(p) + (imag_part(p) * _Complex_I))

#if WITH_GMP
  #define big_integer(p)             (p->object.number.big_integer)
  #define big_ratio(p)               (p->object.number.big_ratio)
  #define big_real(p)                (p->object.number.big_real)
  #define big_complex(p)             (p->object.number.big_complex)
#endif


#define S7_LLONG_MAX 9223372036854775807LL
#define S7_LLONG_MIN (-S7_LLONG_MAX - 1LL)

#define S7_LONG_MAX 2147483647LL
#define S7_LONG_MIN (-S7_LONG_MAX - 1LL)

#define S7_SHORT_MAX 32767
#define S7_SHORT_MIN -32768

static s7_Int s7_Int_max = 0, s7_Int_min = 0;

/* 9007199254740991LL is where a truncated double starts to skip integers (expt 2 53) = ca 1e16 
 *   :(ceiling (+ 1e16 1))
 *   10000000000000000
 *   :(> 9007199254740993.0 9007199254740992.0)
 *   #f ; in non-gmp 64-bit doubles
 *
 * but we can't fix this except in the gmp case because:
 *   :(integer-decode-float (+ (expt 2.0 62) 100))
 *   (4503599627370496 10 1)
 *   :(integer-decode-float (+ (expt 2.0 62) 500))
 *   (4503599627370496 10 1)
 *   :(> (+ (expt 2.0 62) 500) (+ (expt 2.0 62) 100))
 *   #f ; non-gmp again
 *
 * i.e. the bits are identical.  We can't even detect when it has happened, so should
 *   we just give an error for any floor (or whatever) of an arg>1e16?  (sin has a similar problem)?
 *   I think in the non-gmp case I'll throw an error in these cases because the results are
 *   bogus:
 *   :(floor (+ (expt 2.0 62) 512))
 *   4611686018427387904
 *   :(floor (+ (expt 2.0 62) 513))
 *   4611686018427388928
 *
 * another case at the edge: (round 9007199254740992.51) -> 9007199254740992
 *
 * This spells trouble for normal arithmetic in this range.  If no gmp,
 *    (- (+ (expt 2.0 62) 512) (+ (expt 2.0 62) 513)) = -1024.0 (should be -1.0)
 *    but we don't currently give an error in this case -- not sure what the right thing is.
 */



static int safe_strlen(const char *str)
{
  /* this is safer than strlen, and slightly faster */
  char *tmp = (char *)str;
  if (!tmp) return(0);
  while (*tmp++) {};
  return(tmp - str - 1);
}


static char *copy_string_with_len(const char *str, int len)
{
  char *newstr;
  newstr = (char *)malloc((len + 1) * sizeof(char));
  if (len != 0)
    memcpy((void *)newstr, (void *)str, len + 1);
  else newstr[0] = 0;
  return(newstr);
}


static char *copy_string(const char *str)
{
  return(copy_string_with_len(str, safe_strlen(str)));
}


#define strings_are_equal(Str1, Str2) (strcmp(Str1, Str2) == 0)
/* newlib code here was slower -- this should only be used for internal strings -- scheme
 *   strings can have embedded nulls.
 */

static int safe_strcmp(const char *s1, const char *s2)
{
  int val;
  if (s1 == NULL)
    {
      if (s2 == NULL)
	return(0);
      return(-1);
    }
  if (s2 == NULL)
    return(1);

  val = strcmp(s1, s2); /* strcmp can return stuff like -97, but we want -1, 0, or 1 */

  if (val <= -1)
    return(-1);
  if (val >= 1)
    return(1);
  return(val);
}


static bool is_proper_list(s7_scheme *sc, s7_pointer lst);
static void mark_embedded_objects(s7_pointer a); /* called by gc, calls fobj's mark func */
static s7_pointer eval(s7_scheme *sc, opcode_t first_op);
static s7_pointer division_by_zero_error(s7_scheme *sc, const char *caller, s7_pointer arg);
static s7_pointer file_error(s7_scheme *sc, const char *caller, const char *descr, const char *name);
static s7_pointer safe_reverse_in_place(s7_scheme *sc, s7_pointer list);
static s7_pointer cons_unchecked(s7_scheme *sc, s7_pointer a, s7_pointer b);
static s7_pointer permanent_cons(s7_pointer a, s7_pointer b, int type);
static void free_object(s7_pointer a);
static char *object_print(s7_scheme *sc, s7_pointer a);
static s7_pointer make_atom(s7_scheme *sc, char *q, int radix, bool want_symbol);
static bool object_is_applicable(s7_pointer x);
static s7_pointer eval_symbol(s7_scheme *sc, s7_pointer sym);
static s7_pointer eval_error(s7_scheme *sc, const char *errmsg, s7_pointer obj);
static s7_pointer apply_error(s7_scheme *sc, s7_pointer obj, s7_pointer args);
static bool is_thunk(s7_scheme *sc, s7_pointer x);
static int remember_file_name(s7_scheme *sc, const char *file);
static const char *type_name(s7_pointer arg);
static s7_pointer make_string_uncopied(s7_scheme *sc, char *str);
static s7_pointer make_string_uncopied_with_length(s7_scheme *sc, char *str, int len);
static s7_pointer make_protected_string(s7_scheme *sc, const char *str);
static s7_pointer s7_copy(s7_scheme *sc, s7_pointer obj);
static s7_pointer splice_in_values(s7_scheme *sc, s7_pointer args);
static s7_pointer vector_copy(s7_scheme *sc, s7_pointer old_vect);
static void pop_input_port(s7_scheme *sc);
static s7_pointer apply_list_star(s7_scheme *sc, s7_pointer d);
static bool args_match(s7_scheme *sc, s7_pointer x, int args);
static char *object_to_truncated_string(s7_scheme *sc, s7_pointer p, int len);
static bool is_pws(s7_pointer obj);
static token_t token(s7_scheme *sc);

static bool tracing, trace_all;

static s7_pointer find_symbol_or_bust(s7_scheme *sc, s7_pointer hdl);

#if WITH_OPTIMIZATION
static bool body_is_safe(s7_scheme *sc, s7_pointer func, s7_pointer body, bool at_end, bool *bad_set);
static s7_pointer (*finder)(s7_scheme *sc, s7_pointer symbol);
#else
#define finder(Sc, Sym) find_symbol_or_bust(Sc, Sym)
#endif

static int position_of(s7_pointer p, s7_pointer args)
{
  int i;
  for (i = 1; p != args; i++, args = cdr(args));
  return(i);
}





/* -------------------------------- constants -------------------------------- */

s7_pointer s7_f(s7_scheme *sc) 
{
  return(sc->F);
}


s7_pointer s7_t(s7_scheme *sc) 
{
  return(sc->T);
}


s7_pointer s7_nil(s7_scheme *sc) 
{
  return(sc->NIL);
}


bool s7_is_null(s7_scheme *sc, s7_pointer p) 
{
  return(is_null(p));
}


s7_pointer s7_undefined(s7_scheme *sc) 
{
  return(sc->UNDEFINED);
}


s7_pointer s7_unspecified(s7_scheme *sc) 
{
  return(sc->UNSPECIFIED);
}


bool s7_is_unspecified(s7_scheme *sc, s7_pointer val)
{
  return((is_unspecified(val)) || (val == sc->NO_VALUE));
}


s7_pointer s7_eof_object(s7_scheme *sc)          /* returns #<eof> -- not equivalent to "eof-object?" */
{
  return(sc->EOF_OBJECT);
}


static s7_pointer g_not(s7_scheme *sc, s7_pointer args)
{
  #define H_not "(not obj) returns #t if obj is #f, otherwise #t: (not '()) -> #f"
  return(make_boolean(sc, is_false(sc, car(args))));
}


bool s7_boolean(s7_scheme *sc, s7_pointer x)
{
  return(x != sc->F);
}


bool s7_is_boolean(s7_pointer x)
{
  return(type(x) == T_BOOLEAN);
}


s7_pointer s7_make_boolean(s7_scheme *sc, bool x)
{
  return(make_boolean(sc, x));
}


static s7_pointer g_is_boolean(s7_scheme *sc, s7_pointer args)
{
  #define H_is_boolean "(boolean? obj) returns #t if obj is #f or #t: (boolean? '()) -> #f"
  return(s7_make_boolean(sc, s7_is_boolean(car(args))));
}


bool s7_is_constant(s7_pointer p) 
{ 
  /* this means "always evaluates to the same thing", sort of
   *   not "evaluates to itself":
   *   (let ((x 'x)) (and (not (constant? x)) (equal? x (eval x))))
   *   (and (constant? (list + 1)) (not (equal? (list + 1) (eval (list + 1)))))
   */
  return((type(p) != T_SYMBOL) ||
	 ((typeflag(p) & (T_KEYWORD | T_IMMUTABLE)) != 0));
}


static s7_pointer g_is_constant(s7_scheme *sc, s7_pointer args)
{
  #define H_is_constant "(constant? obj) returns #t if obj is a constant (unsettable): (constant? pi) -> #t"
  return(s7_make_boolean(sc, s7_is_constant(car(args))));
}





/* -------------------------------- GC -------------------------------- */

#define is_gc_nil(p) ((p) == sc->GC_NIL)

int s7_gc_protect(s7_scheme *sc, s7_pointer x)
{
  unsigned int i, loc, size, new_size;

  loc = sc->protected_objects_loc++;
  size = sc->protected_objects_size;

  if (sc->protected_objects_loc >= size)
    sc->protected_objects_loc = 0;

  if (is_gc_nil(vector_element(sc->protected_objects, loc)))
    {
      vector_element(sc->protected_objects, loc) = x;
      return(loc);
    }
  
  for (i = 0; i < size; i++)
    {
      if (is_gc_nil(vector_element(sc->protected_objects, i)))
	{
	  vector_element(sc->protected_objects, i) = x;
	  return(i);
	}
    }
  
  new_size = 2 * size;
  vector_elements(sc->protected_objects) = (s7_pointer *)realloc(vector_elements(sc->protected_objects), new_size * sizeof(s7_pointer));
  for (i = size; i < new_size; i++)
    vector_element(sc->protected_objects, i) = sc->GC_NIL;
  vector_length(sc->protected_objects) = new_size;
  sc->protected_objects_size = new_size;
  sc->protected_objects_loc = size + 1;

  vector_element(sc->protected_objects, size) = x;
  return(size);
}


void s7_gc_unprotect(s7_scheme *sc, s7_pointer x)
{
  unsigned int i;

  for (i = 0; i < sc->protected_objects_size; i++)
    if (vector_element(sc->protected_objects, i) == x)
      {
	vector_element(sc->protected_objects, i) = sc->GC_NIL;
	sc->protected_objects_loc = i;
	return;
      }
}


void s7_gc_unprotect_at(s7_scheme *sc, int loc)
{
  if ((loc >= 0) &&
      (loc < (int)sc->protected_objects_size))
    {
      vector_element(sc->protected_objects, loc) = sc->GC_NIL;
      sc->protected_objects_loc = loc;
    }
}


s7_pointer s7_gc_protected_at(s7_scheme *sc, int loc)
{
  s7_pointer obj;

  obj = sc->UNSPECIFIED;
  if ((loc >= 0) &&
      (loc < (int)sc->protected_objects_size))
    obj = vector_element(sc->protected_objects, loc);

  if (obj == sc->GC_NIL)
    return(sc->UNSPECIFIED);

  return(obj);
}


static void sweep(s7_scheme *sc)
{
  int i, j;
  if (sc->strings_loc > 0)
    {
      for (i = 0, j = 0; i < sc->strings_loc; i++)
	{
	  s7_pointer s1;
	  s1 = sc->strings[i];
	  if (type(s1) == 0)
	    {
	      if (string_needs_free(s1))
		free(string_value(s1));
	    }
	  else sc->strings[j++] = s1;
	}
      sc->strings_loc = j;
    }

  if (sc->c_objects_loc > 0)
    {
      for (i = 0, j = 0; i < sc->c_objects_loc; i++)
	{
	  if (type(sc->c_objects[i]) == 0)
	    free_object(sc->c_objects[i]);
	  else sc->c_objects[j++] = sc->c_objects[i];
	}
      sc->c_objects_loc = j;
    }

  if (sc->vectors_loc > 0)
    {
      for (i = 0, j = 0; i < sc->vectors_loc; i++)
	{
	  if (type(sc->vectors[i]) == 0)
	    {
	      s7_pointer a;
	      a = sc->vectors[i];
	      if (vector_length(a) > 0)
		{
		  if (vector_is_multidimensional(a))
		    {
		      if (shared_vector(a) == sc->F)
			{
			  free(a->object.vector.vextra.dim_info->dims);
			  free(a->object.vector.vextra.dim_info->offsets);
			  free(vector_elements(a));
			}
		      free(a->object.vector.vextra.dim_info);
		    }
		  else 
		    {
		      free(vector_elements(a));
		    }
		}
	    }
	  else sc->vectors[j++] = sc->vectors[i];
	}
      sc->vectors_loc = j;
    }

  if (sc->hash_tables_loc > 0)
    {
      for (i = 0, j = 0; i < sc->hash_tables_loc; i++)
	{
	  if (type(sc->hash_tables[i]) == 0)
	    {
	      if (hash_table_length(sc->hash_tables[i]) > 0)
		free(hash_table_elements(sc->hash_tables[i]));
	    }
	  else sc->hash_tables[j++] = sc->hash_tables[i];
	}
      sc->hash_tables_loc = j;
    }

  if (sc->input_ports_loc > 0)
    {
      for (i = 0, j = 0; i < sc->input_ports_loc; i++)
	{
	  if (type(sc->input_ports[i]) == 0)
	    {
	      s7_pointer a;
	      a = sc->input_ports[i];
	      if (port_needs_free(a))
		{
		  if (port_string(a))
		    {
		      free(port_string(a));
		      port_string(a) = NULL;
		    }
		  port_needs_free(a) = false;
		}
	      
	      if (port_filename(a))
		{
		  free(port_filename(a));
		  port_filename(a) = NULL;
		}
	      
	      free(a->object.port);
	      
	    }
	  else sc->input_ports[j++] = sc->input_ports[i];
	}
      sc->input_ports_loc = j;
    }

  if (sc->output_ports_loc > 0)
    {
      for (i = 0, j = 0; i < sc->output_ports_loc; i++)
	{
	  if (type(sc->output_ports[i]) == 0)
	    {
	      s7_close_output_port(sc, sc->output_ports[i]);
	      free(sc->output_ports[i]->object.port);
	    }
	  else sc->output_ports[j++] = sc->output_ports[i];
	}
      sc->output_ports_loc = j;
    }

  if (sc->continuations_loc > 0)
    {
      for (i = 0, j = 0; i < sc->continuations_loc; i++)
	{
	  if (type(sc->continuations[i]) == 0)
	    {
	      s7_pointer c;
	      c = sc->continuations[i];
	      if (continuation_op_stack(c))
		{
		  free(continuation_op_stack(c));
		  continuation_op_stack(c) = NULL;
		}
	      free(continuation(c));
	    }
	  else sc->continuations[j++] = sc->continuations[i];
	}
      sc->continuations_loc = j;
    }
}


static void add_string(s7_scheme *sc, s7_pointer p)
{
  if (sc->strings_loc == sc->strings_size)
    {
      sc->strings_size *= 2;
      sc->strings = (s7_pointer *)realloc(sc->strings, sc->strings_size * sizeof(s7_pointer));
    }
  sc->strings[sc->strings_loc++] = p;
}


static void add_c_object(s7_scheme *sc, s7_pointer p)
{
  if (sc->c_objects_loc == sc->c_objects_size)
    {
      sc->c_objects_size *= 2;
      sc->c_objects = (s7_pointer *)realloc(sc->c_objects, sc->c_objects_size * sizeof(s7_pointer));
    }
  sc->c_objects[sc->c_objects_loc++] = p;
}


static void add_hash_table(s7_scheme *sc, s7_pointer p)
{
  if (sc->hash_tables_loc == sc->hash_tables_size)
    {
      sc->hash_tables_size *= 2;
      sc->hash_tables = (s7_pointer *)realloc(sc->hash_tables, sc->hash_tables_size * sizeof(s7_pointer));
    }
  sc->hash_tables[sc->hash_tables_loc++] = p;
}


static void add_vector(s7_scheme *sc, s7_pointer p)
{
  if (sc->vectors_loc == sc->vectors_size)
    {
      sc->vectors_size *= 2;
      sc->vectors = (s7_pointer *)realloc(sc->vectors, sc->vectors_size * sizeof(s7_pointer));
    }
  sc->vectors[sc->vectors_loc++] = p;
}


static void add_input_port(s7_scheme *sc, s7_pointer p)
{
  if (sc->input_ports_loc == sc->input_ports_size)
    {
      sc->input_ports_size *= 2;
      sc->input_ports = (s7_pointer *)realloc(sc->input_ports, sc->input_ports_size * sizeof(s7_pointer));
    }
  sc->input_ports[sc->input_ports_loc++] = p;
}


static void add_output_port(s7_scheme *sc, s7_pointer p)
{
  if (sc->output_ports_loc == sc->output_ports_size)
    {
      sc->output_ports_size *= 2;
      sc->output_ports = (s7_pointer *)realloc(sc->output_ports, sc->output_ports_size * sizeof(s7_pointer));
    }
  sc->output_ports[sc->output_ports_loc++] = p;
}


static void add_continuation(s7_scheme *sc, s7_pointer p)
{
  if (sc->continuations_loc == sc->continuations_size)
    {
      sc->continuations_size *= 2;
      sc->continuations = (s7_pointer *)realloc(sc->continuations, sc->continuations_size * sizeof(s7_pointer));
    }
  sc->continuations[sc->continuations_loc++] = p;
}


#define INIT_GC_CACHE_SIZE 512
static void init_gc_caches(s7_scheme *sc)
{
  sc->strings_size = INIT_GC_CACHE_SIZE;
  sc->strings_loc = 0;
  sc->strings = (s7_pointer *)malloc(sc->strings_size * sizeof(s7_pointer));
  sc->vectors_size = INIT_GC_CACHE_SIZE;
  sc->vectors_loc = 0;
  sc->vectors = (s7_pointer *)malloc(sc->vectors_size * sizeof(s7_pointer));
  sc->hash_tables_size = INIT_GC_CACHE_SIZE;
  sc->hash_tables_loc = 0;
  sc->hash_tables = (s7_pointer *)malloc(sc->hash_tables_size * sizeof(s7_pointer));
  sc->input_ports_size = INIT_GC_CACHE_SIZE;
  sc->input_ports_loc = 0;
  sc->input_ports = (s7_pointer *)malloc(sc->input_ports_size * sizeof(s7_pointer));
  sc->output_ports_size = INIT_GC_CACHE_SIZE;
  sc->output_ports_loc = 0;
  sc->output_ports = (s7_pointer *)malloc(sc->output_ports_size * sizeof(s7_pointer));
  sc->continuations_size = INIT_GC_CACHE_SIZE;
  sc->continuations_loc = 0;
  sc->continuations = (s7_pointer *)malloc(sc->continuations_size * sizeof(s7_pointer));
  sc->c_objects_size = INIT_GC_CACHE_SIZE;
  sc->c_objects_loc = 0;
  sc->c_objects = (s7_pointer *)malloc(sc->c_objects_size * sizeof(s7_pointer));
}


static void (*mark_function[BUILT_IN_TYPES])(s7_pointer p);
#define S7_MARK(Obj) do {s7_pointer _p_; _p_ = Obj; (*mark_function[type(_p_)])(_p_);} while (0)

static void mark_vector_1(s7_pointer p, s7_Int top)
{
  s7_pointer *tp, *tend;

  set_mark(p);

  tp = (s7_pointer *)(vector_elements(p));
  if (!tp) return;
  tend = (s7_pointer *)(tp + top);
  
  if ((top % 8) == 0)
    {
      while (tp < tend) 
	{
	  S7_MARK(*tp++);
	  S7_MARK(*tp++);
	  S7_MARK(*tp++);
	  S7_MARK(*tp++);
	  S7_MARK(*tp++);
	  S7_MARK(*tp++);
	  S7_MARK(*tp++);
	  S7_MARK(*tp++);
	}
    }
  else
    {
      while (tp < tend) 
	S7_MARK(*tp++);
    }
}


static void mark_environment(s7_pointer env)
{
  s7_pointer x;
  for (x = env; is_environment(x) && (!is_marked(x)); x = next_environment(x)) 
    { 
      s7_pointer y;
      for (y = environment_slots(x); is_slot(y); y = next_slot(y))
	if (!is_marked(y))
	  {
	    set_mark(y);
	    S7_MARK(slot_value(y));
	  }
      set_mark(x); /* we handle car above, and cdr in the loop */
    }
}


static void just_mark(s7_pointer p)
{
  set_mark(p);
}


static void mark_noop(s7_pointer p)
{
}


static void mark_pair(s7_pointer p)
{
  if (!is_marked(p)) 
    {
      set_mark(p);
      S7_MARK(car(p));
      S7_MARK(cdr(p));
    }
}

static void mark_slot(s7_pointer p)
{
  if (!is_marked(p)) 
    {
      set_mark(p);
      S7_MARK(slot_value(p));
    }
}

static void mark_counter(s7_pointer p)
{
  if (!is_marked(p)) 
    {
      set_mark(p);
      S7_MARK(counter_result(p));
      S7_MARK(counter_list(p));
    }
}


static void mark_closure(s7_pointer p)
{
  if (!is_marked(p)) 
    {
      set_mark(p);
      S7_MARK(closure_args(p));
      S7_MARK(closure_body(p));
      mark_environment(closure_environment(p));
    }
}

static void mark_stack(s7_pointer p)
{
}

static void mark_stack_1(s7_pointer p, s7_Int top)
{
  s7_pointer *tp, *tend;

  set_mark(p);

  tp = (s7_pointer *)(vector_elements(p));
  if (!tp) return;
  tend = (s7_pointer *)(tp + top);
  
  while (tp < tend) 
    {
      S7_MARK(*tp++);
      S7_MARK(*tp++);
      S7_MARK(*tp++);
      tp++;
    }
}


static void mark_continuation(s7_pointer p)
{
  if (!is_marked(p)) 
    {
      unsigned int i;
      set_mark(p);
      mark_stack_1(continuation_stack(p), continuation_stack_top(p));
      for (i = 0; i < continuation_op_loc(p); i++)
	S7_MARK(continuation_op_stack(p)[i]);
    }
}

static void mark_vector(s7_pointer p)
{
  if (!is_marked(p)) 
    {
      /* If a subvector (an inner dimension) of a vector is the only remaining reference
       *    to the main vector, we want to make sure the main vector is not GC'd until
       *    the subvector is also GC-able.  The shared_vector field either points to the
       *    parent vector, or it is sc->F, so we need to check for a vector parent if
       *    the current is multidimensional (this will include 1-dim slices).  We need
       *    to keep the parent case separate (i.e. sc->F means the current is the original)
       *    so that we only free once (or remove_from_heap once).
       */

      if ((vector_is_multidimensional(p)) &&
	  (s7_is_vector(shared_vector(p))))
	{
	  /* set_mark(shared_vector(p)); */
	  S7_MARK(shared_vector(p));
	}
      mark_vector_1(p, vector_length(p));
    }
}

static void mark_c_object(s7_pointer p)
{
  if (!is_marked(p)) 
    {
      set_mark(p);
      mark_embedded_objects(p);
    }
}

static void mark_catch(s7_pointer p)
{
  if (!is_marked(p))
    {
      set_mark(p);
      S7_MARK(catch_tag(p));
      S7_MARK(catch_handler(p));
    }
}

static void mark_dynamic_wind(s7_pointer p)
{
  if (!is_marked(p)) 
    {
      set_mark(p);
      S7_MARK(dynamic_wind_in(p));
      S7_MARK(dynamic_wind_out(p));
      S7_MARK(dynamic_wind_body(p));
    }
}

static void mark_hash_table(s7_pointer p)
{
  if (!is_marked(p))
    {
      if (hash_table_entries(p) == 0)
	set_mark(p);
      else mark_vector_1(p, hash_table_length(p));
    }
}

static void mark_hook(s7_pointer p)
{
  if (!is_marked(p)) 
    {
      set_mark(p);
      S7_MARK(hook_functions(p));
      S7_MARK(hook_arity(p));
      S7_MARK(hook_documentation(p));
    }
}

static void mark_input_port(s7_pointer p)
{
  if (!is_marked(p)) 
    {
      set_mark(p);
      set_mark(port_original_input_string(p));
    }
}


static void init_mark_functions(void)
{
  mark_function[T_UNTYPED]             = just_mark;
  mark_function[T_NIL]                 = mark_noop;
  mark_function[T_STRING]              = just_mark;
  mark_function[T_INTEGER]             = just_mark;
  mark_function[T_RATIO]               = just_mark;
  mark_function[T_REAL]                = just_mark;
  mark_function[T_COMPLEX]             = just_mark;
  mark_function[T_BIG_INTEGER]         = just_mark;
  mark_function[T_BIG_RATIO]           = just_mark;
  mark_function[T_BIG_REAL]            = just_mark;
  mark_function[T_BIG_COMPLEX]         = just_mark;
  mark_function[T_SYMBOL]              = mark_noop;
  mark_function[T_PAIR]                = mark_pair;
  mark_function[T_CLOSURE]             = mark_closure;
  mark_function[T_CLOSURE_STAR]        = mark_closure;
  mark_function[T_CONTINUATION]        = mark_continuation;
  mark_function[T_CHARACTER]           = mark_noop;
  mark_function[T_INPUT_PORT]          = mark_input_port;
  mark_function[T_VECTOR]              = mark_vector;
  mark_function[T_MACRO]               = mark_closure;
  mark_function[T_BACRO]               = mark_closure;
  mark_function[T_C_OBJECT]            = mark_c_object;
  mark_function[T_GOTO]                = just_mark;
  mark_function[T_OUTPUT_PORT]         = just_mark;
  mark_function[T_CATCH]               = mark_catch;
  mark_function[T_DYNAMIC_WIND]        = mark_dynamic_wind;
  mark_function[T_HASH_TABLE]          = mark_hash_table;
  mark_function[T_BOOLEAN]             = mark_noop;
  mark_function[T_SYNTAX]              = mark_noop;
  mark_function[T_HOOK]                = mark_hook;
  mark_function[T_ENVIRONMENT]         = mark_environment;
  mark_function[T_STACK]               = mark_stack;
  mark_function[T_COUNTER]             = mark_counter;
  mark_function[T_SLOT]                = mark_slot;
  mark_function[T_C_MACRO]             = just_mark;
  mark_function[T_C_POINTER]           = just_mark;
  mark_function[T_C_FUNCTION]          = just_mark;
  mark_function[T_C_ANY_ARGS_FUNCTION] = just_mark;
  mark_function[T_C_OPT_ARGS_FUNCTION] = just_mark;
  mark_function[T_C_RST_ARGS_FUNCTION] = just_mark;
  mark_function[T_C_LST_ARGS_FUNCTION] = just_mark;
}


static void mark_op_stack(s7_scheme *sc)
{
  s7_pointer *p, *tp;
  tp = sc->op_stack_now;
  p = sc->op_stack;
  while (p < tp)
    S7_MARK(*p++);
}


static void mark_global_env(s7_scheme *sc)
{
  s7_pointer ge;
  s7_pointer *tmp, *top;

  ge = sc->global_env;
  tmp = vector_elements(ge);
  top = (s7_pointer *)(tmp + vector_fill_pointer(ge));

  set_mark(ge);

  while (tmp < top)
    S7_MARK(slot_value(*tmp++));
}


void s7_mark_object(s7_pointer p)
{
  S7_MARK(p);
}

#define GC_TRIGGER_SIZE 64


#if HAVE_GETTIMEOFDAY && (!_MSC_VER)
  #include <time.h>
  #include <sys/time.h>
  static struct timeval start_time;
  static struct timezone z0;
#endif

static int gc(s7_scheme *sc)
{
  s7_cell **old_free_heap_top;
  /* mark all live objects (the symbol table is in permanent memory, not the heap) */
  #define GC_CALL(P, Tp) p = (*tp++); if (is_marked(p)) clear_mark(p); else {if (typeflag(p) != 0) {typeflag(p) = 0; (*fp++) = p;}}

  if (sc->gc_stats)
    {
      fprintf(stdout, "gc ");
#if HAVE_GETTIMEOFDAY && (!_MSC_VER)
      gettimeofday(&start_time, &z0);
#endif
    }

  mark_global_env(sc);
  S7_MARK(sc->args);
  mark_environment(sc->envir);
  S7_MARK(sc->code);
  S7_MARK(sc->cur_code);
  mark_stack_1(sc->stack, s7_stack_top(sc));
  S7_MARK(sc->w);
  S7_MARK(sc->x);
  S7_MARK(sc->y);
  S7_MARK(sc->z);
  S7_MARK(sc->value);  

  S7_MARK(sc->temp1);
  S7_MARK(sc->temp2);
  S7_MARK(sc->temp3);
  S7_MARK(sc->temp4);

  set_mark(sc->input_port);
  S7_MARK(sc->input_port_stack);
  set_mark(sc->output_port);
  set_mark(sc->error_port);

  mark_pair(sc->TEMP_CELL_1);
  mark_pair(sc->TEMP_CELL_2);
  mark_pair(sc->TEMP_CELL_3);
  S7_MARK(car(sc->T1_1));
  S7_MARK(car(sc->T2_1));
  S7_MARK(car(sc->T2_2));
  S7_MARK(car(sc->T3_1));
  S7_MARK(car(sc->T3_2));
  S7_MARK(car(sc->T3_3));

  {
    s7_pointer p;
    for (p = sc->all_g_args[0]; is_pair(p); p = cdr(p))
      S7_MARK(p);
  }

  S7_MARK(sc->protected_objects);
  
  /* now protect recent allocations using the free_heap cells above the current free_heap_top (if any).
   *
   * cells above sc->free_heap_top might be malloc'd garbage (after heap reallocation), so we keep track of 
   *   where the last actually freed cells were after the previous GC call.  We're trying to
   *   GC protect the previous GC_TEMPS_SIZE allocated pointers so that the caller doesn't have
   *   to gc-protect every temporary cell.
   *
   * There's one remaining possible problem.  s7_remove_from_heap frees cells outside
   *   the GC and might push free_heap_top beyond its previous_free_heap_top, then
   *   an immediate explicit gc call might not see those temp cells.
   */
  {
    s7_pointer *tmps, *tmps_top;

    tmps = sc->free_heap_top;
    tmps_top = tmps + GC_TEMPS_SIZE;
#if 0
    if ((sc->heap_size > INITIAL_HEAP_SIZE) && 
	(tmps_top > (sc->free_heap + INITIAL_HEAP_SIZE))) /* this was the old (buggy, pre-23-Sep-11) choice */
      fprintf(stderr, "%ld of %d: protect (min %ld %ld), not %ld\n", 
	      sc->previous_free_heap_top - sc->free_heap, 
	      sc->heap_size, 
	      sc->previous_free_heap_top - tmps,
	      tmps_top - tmps,
	      sc->free_heap + INITIAL_HEAP_SIZE - tmps_top);

    /* snd-test.scm: 297114 of 512000: protect (min 155818 256), not -13552
     *   which once in a blue-moon (1/1000 tries or less) would manifest itself as confusion in 
     *   one of the tinyCL functions in s7test when run at the end of snd-test!
     */
#endif
    if (tmps_top > sc->previous_free_heap_top)
      tmps_top = sc->previous_free_heap_top;

    while (tmps < tmps_top)
      S7_MARK(*tmps++);
  }
  mark_op_stack(sc);

  /* free up all unmarked objects */
  old_free_heap_top = sc->free_heap_top;

  {
    s7_pointer *fp, *tp, *heap_top;
    fp = sc->free_heap_top;

    tp = sc->heap;
    heap_top = (s7_pointer *)(sc->heap + sc->heap_size);

    while (tp < heap_top)          /* != here or ^ makes no difference */
      {
	s7_pointer p;
	/* from here down is GC_CALL, but I wanted one case explicit for readability
	 */
	p = (*tp++);

	if (is_marked(p))          /* this order is faster than checking typeflag(p) != 0 first */
	  clear_mark(p);
	else 
	  {
	    if (typeflag(p) != 0) /* if 0, it's an already-free object -- the free_heap is usually not empty when we call the GC */
	      {
		typeflag(p) = 0;  /* (this is needed -- otherwise we try to free some objects twice) */
		(*fp++) = p;
	      }
	  }

	/* this looks crazy, but it speeds up the entire GC process by 25%!
	 *   going from 16 to 32 saves .2% so it may not matter.
	 */
	GC_CALL(p, tp);
	GC_CALL(p, tp);
	GC_CALL(p, tp);

	GC_CALL(p, tp);
	GC_CALL(p, tp);
	GC_CALL(p, tp);
	GC_CALL(p, tp);

	GC_CALL(p, tp);
	GC_CALL(p, tp);
	GC_CALL(p, tp);
	GC_CALL(p, tp);

	GC_CALL(p, tp);
	GC_CALL(p, tp);
	GC_CALL(p, tp);
	GC_CALL(p, tp);

	GC_CALL(p, tp);
	GC_CALL(p, tp);
	GC_CALL(p, tp);
	GC_CALL(p, tp);

	GC_CALL(p, tp);
	GC_CALL(p, tp);
	GC_CALL(p, tp);
	GC_CALL(p, tp);

	GC_CALL(p, tp);
	GC_CALL(p, tp);
	GC_CALL(p, tp);
	GC_CALL(p, tp);

	GC_CALL(p, tp);
	GC_CALL(p, tp);
	GC_CALL(p, tp);
	GC_CALL(p, tp);
      }
    
    sc->free_heap_top = fp;
    sweep(sc);
  }

  if (sc->gc_stats)
    {
#if HAVE_GETTIMEOFDAY && (!_MSC_VER)
      struct timeval t0;
      double secs;
      gettimeofday(&t0, &z0);
      secs = (t0.tv_sec - start_time.tv_sec) +  0.000001 * (t0.tv_usec - start_time.tv_usec);
      fprintf(stdout, "freed %d/%d, time: %f\n", (int)(sc->free_heap_top - old_free_heap_top), sc->heap_size, secs);
#else
      fprintf(stdout, "freed %d/%d\n", (int)(sc->free_heap_top - old_free_heap_top), sc->heap_size);
#endif
    }
  
  sc->previous_free_heap_top = sc->free_heap_top;
  return(sc->free_heap_top - old_free_heap_top); /* needed by cell allocator to decide when to increase heap size */
}


#define WITH_DUMP_HEAP 0
#if WITH_DUMP_HEAP
static s7_pointer g_dump_heap(s7_scheme *sc, s7_pointer args)
{
  FILE *fd;
  s7_pointer *tp;
  int i;

  gc(sc);

  fd = fopen("heap.data", "w");
  for (tp = sc->heap; (*tp); tp++)
    {
      s7_pointer p;
      p = (*tp);
      if (typeflag(p) != 0)
	fprintf(fd, "%s\n", s7_object_to_c_string(sc, p));
    }

  fflush(fd);
  fclose(fd);
  return(sc->NIL);
}
#endif


#define NEW_CELL(Sc, Obj) \
  do {						\
    if (Sc->free_heap_top <= Sc->free_heap_trigger) try_to_call_gc(Sc); \
    Obj = (*(--(Sc->free_heap_top))); \
    } while (0)

#define NEW_CELL_NO_CHECK(Sc, Obj) do {Obj = (*(--(Sc->free_heap_top)));} while (0)
  /* since sc->free_heap_trigger is GC_TRIGGER_SIZE above the free heap base, we don't need
   *   to check it repeatedly after the 1st such check.
   */


static void try_to_call_gc(s7_scheme *sc)
{
  /* called only from NEW_CELL and cons */
  unsigned int freed_heap = 0;
  
  if (!(sc->gc_off))
    freed_heap = gc(sc);
  
  if (freed_heap < sc->heap_size / 4) /* was 1000, setting it to 2 made no difference in run time */
    {
      /* alloc more heap */
      unsigned int old_size;
      old_size = sc->heap_size;
      
      if (sc->heap_size < 512000)
	sc->heap_size *= 2;
      else sc->heap_size += 512000;
      
      sc->heap = (s7_cell **)realloc(sc->heap, sc->heap_size * sizeof(s7_cell *));
      if (!(sc->heap))
	fprintf(stderr, "heap reallocation failed! tried to get %lu bytes\n", (unsigned long)(sc->heap_size * sizeof(s7_cell *)));
      
      sc->free_heap = (s7_cell **)realloc(sc->free_heap, sc->heap_size * sizeof(s7_cell *));
      if (!(sc->free_heap))
	fprintf(stderr, "free heap reallocation failed! tried to get %lu bytes\n", (unsigned long)(sc->heap_size * sizeof(s7_cell *)));	  
      
      sc->free_heap_trigger = (s7_cell **)(sc->free_heap + GC_TRIGGER_SIZE);
      sc->free_heap_top = sc->free_heap;
      
      { 
	/* optimization suggested by K Matheussen */
	unsigned int k;
	s7_cell *cells;
	cells = (s7_cell *)calloc(sc->heap_size - old_size, sizeof(s7_cell));
	for (k = old_size; k < sc->heap_size; k++)
	  {
	    sc->heap[k] = &cells[k - old_size];
	    (*sc->free_heap_top++) = sc->heap[k];
	    sc->heap[k]->hloc = k;
	  }
	sc->previous_free_heap_top = sc->free_heap_top;
      }
    }
}

  /* originally I tried to mark each temporary value until I was done with it, but
   *   that way madness lies... By delaying GC of _every_ %$^#%@ pointer, I can dispense
   *   with hundreds of individual protections.  So the free_heap's last GC_TEMPS_SIZE
   *   allocated pointers are protected during the mark sweep.
   */


static s7_pointer g_gc(s7_scheme *sc, s7_pointer args)
{
  #define H_gc "(gc (on #t)) runs the garbage collector.  If 'on' is supplied, it turns the GC on or off. \
Evaluation produces a surprising amount of garbage, so don't leave the GC off for very long!"

  if (is_not_null(args))
    {
      if (!s7_is_boolean(car(args)))
	return(s7_wrong_type_arg_error(sc, "gc", 0, car(args), "#f (turn GC off) or #t (turn it on)"));	

      sc->gc_off = (car(args) == sc->F);
      if (sc->gc_off) return(sc->F);
    }

  gc(sc);
  return(sc->UNSPECIFIED);
}


s7_pointer s7_gc_on(s7_scheme *sc, bool on)
{
  sc->gc_off = !on;
  return(s7_make_boolean(sc, on));
}


static s7_pointer g_gc_stats_set(s7_scheme *sc, s7_pointer args)
{
  if (s7_is_boolean(cadr(args)))
    {
      sc->gc_stats = (cadr(args) != sc->F);      
      return(cadr(args));
    }
  return(sc->ERROR);
}


void s7_gc_stats(s7_scheme *sc, bool on)
{
  sc->gc_stats = on;
  s7_symbol_set_value(sc, s7_make_symbol(sc, "*gc-stats*"), make_boolean(sc, on));
}


static s7_pointer g_safety_set(s7_scheme *sc, s7_pointer args)
{
  if (s7_is_integer(cadr(args)))
    {
      sc->safety = s7_integer(cadr(args));
      return(cadr(args));
    }
  return(sc->ERROR);
}


static s7_pointer g_safety_bind(s7_scheme *sc, s7_pointer args)
{
  return(sc->ERROR);
}


void s7_remove_from_heap(s7_scheme *sc, s7_pointer x)
{
  int loc;
  /* global functions are very rarely redefined, so we can remove the function body from
   *   the heap when it is defined.  If redefined, we currently lose the memory held by the
   *   old definition.  (It is not trivial to recover this memory because it is allocated
   *   in blocks, not by the pointer, I think, but s7_define is the point to try).
   * 
   * There is at least one problem with this: if, for example, a function has
   *    a quoted (constant) list, then uses list-set! to change an element of it,
   *    then a GC happens, and the new element is GC'd because no one in the heap
   *    points to it, then we call the function again, and it tries to access
   *    that element.  I wonder if removal should be on a switch, so the user can
   *    choose a "safety" level.
   *
   *    (define (bad-idea)
   *      (let ((lst '(1 2 3))) ; or #(1 2 3) and vector-ref|set
   *        (let ((result (list-ref lst 1)))
   *          (list-set! lst 1 (* 2.0 16.6))
   *          (gc)
   *          result)))
   * 
   *     put that in a file, load it (to force removal), than call bad-idea a few times.
   * so... if *safety* is not 0, remove-from-heap is disabled.
   */

  /* (catch #t (lambda () (set! *safety* "hi")) (lambda args args)) */

  loc = x->hloc; /* punned to syntax_opcode... is this necessary? */
  if (loc == NOT_IN_HEAP) return;
  
  switch (type(x))
    {
    case T_PAIR:
      x->hloc = NOT_IN_HEAP;
      sc->heap[loc] = (s7_cell *)calloc(1, sizeof(s7_cell));
      (*sc->free_heap_top++) = sc->heap[loc];
      sc->heap[loc]->hloc = loc;

      s7_remove_from_heap(sc, car(x));
      s7_remove_from_heap(sc, cdr(x));
      return;


    case T_VECTOR:
      {
	s7_Int i;
	/* this can happen if a global function has an explicit vector constant.
	 */
	x->hloc = NOT_IN_HEAP;
	sc->heap[loc] = (s7_cell *)calloc(1, sizeof(s7_cell));
	(*sc->free_heap_top++) = sc->heap[loc];
	sc->heap[loc]->hloc = loc;

	if ((!vector_is_multidimensional(x)) ||
	    (shared_vector(x) == sc->F))
	  {
	    for (i = 0; i < vector_length(x); i++)
	      s7_remove_from_heap(sc, vector_element(x, i));
	  }
      }
      return;

    case T_HASH_TABLE:
    case T_ENVIRONMENT:
      return;

    case T_SYMBOL:
    case T_SYNTAX:
      /* here hloc is usually NOT_IN_HEAP, but in the syntax case can be the syntax op code.
       */
      return;

    case T_CLOSURE:
    case T_CLOSURE_STAR:
    case T_MACRO:
    case T_BACRO:
      x->hloc = NOT_IN_HEAP;
      sc->heap[loc] = (s7_cell *)calloc(1, sizeof(s7_cell));
      (*sc->free_heap_top++) = sc->heap[loc];
      sc->heap[loc]->hloc = loc;

      s7_remove_from_heap(sc, closure_args(x));
      s7_remove_from_heap(sc, closure_body(x));
      return;

    default:
      break;
    }

  x->hloc = NOT_IN_HEAP;
  sc->heap[loc] = (s7_cell *)calloc(1, sizeof(s7_cell));
  (*sc->free_heap_top++) = sc->heap[loc];
  sc->heap[loc]->hloc = loc;
}


/* permanent memory for objects that we know will not (normally) be deallocated */

#define PERMANENT_HEAP_SIZE 65536
static unsigned char *permanent_heap = NULL, *permanent_heap_top = NULL;

static unsigned char *permanent_calloc(int bytes)
{
  unsigned char *cur;
  if (bytes >= PERMANENT_HEAP_SIZE)
    {
      /* this actually can happen!  I wrote a file unwittingly that had gmp's output from (ash 1 92233720360)
       *   or some big number like that -- 16 million digits.  During a subsequent load, s7 decided it was a 
       *   symbol name(!) and tried to store it permanently in the symbol table.  segfault.
       */
      return((unsigned char *)calloc(bytes, sizeof(unsigned char)));
    }
  if ((permanent_heap + bytes) >= permanent_heap_top)
    {
      permanent_heap = (unsigned char *)calloc(PERMANENT_HEAP_SIZE, sizeof(unsigned char));
      permanent_heap_top = (unsigned char *)(permanent_heap + PERMANENT_HEAP_SIZE);
    }
  cur = permanent_heap;
  permanent_heap += bytes;
  return(cur);
}




/* -------------------------------- stacks -------------------------------- */

#define OP_STACK_INITIAL_SIZE 128
#define push_op_stack(Sc, Op) (*Sc->op_stack_now++) = Op
#define pop_op_stack(Sc)      (*(--(Sc->op_stack_now)))

static void initialize_op_stack(s7_scheme *sc)
{
  int i;
  sc->op_stack = (s7_pointer *)malloc(OP_STACK_INITIAL_SIZE * sizeof(s7_pointer));
  sc->op_stack_size = OP_STACK_INITIAL_SIZE;
  sc->op_stack_now = sc->op_stack;
  sc->op_stack_end = (s7_pointer *)(sc->op_stack + sc->op_stack_size);
  for (i = 0; i < OP_STACK_INITIAL_SIZE; i++)
    sc->op_stack[i] = sc->NIL;
}


static void resize_op_stack(s7_scheme *sc)
{
  int i, loc, new_size;
  loc = (int)(sc->op_stack_now - sc->op_stack);
  new_size = sc->op_stack_size * 2;
  sc->op_stack = (s7_pointer *)realloc((void *)(sc->op_stack), new_size * sizeof(s7_pointer));
  for (i = sc->op_stack_size; i < new_size; i++)
    sc->op_stack[i] = sc->NIL;
  sc->op_stack_size = new_size;
  sc->op_stack_now = (s7_pointer *)(sc->op_stack + loc);
  sc->op_stack_end = (s7_pointer *)(sc->op_stack + sc->op_stack_size);
}


static void stack_reset(s7_scheme *sc) 
{ 
  sc->stack_end = sc->stack_start;
} 


#define stack_code(Stack, Loc)        vector_element(Stack, Loc - 3)
#define stack_environment(Stack, Loc) vector_element(Stack, Loc - 2)
#define stack_args(Stack, Loc)        vector_element(Stack, Loc - 1)
#define stack_op(Stack, Loc)          ((opcode_t)(vector_element(Stack, Loc)))

/* these macros are faster than the equivalent simple function calls.  If the s7_scheme struct is set up to reflect the
 *    stack order [code envir args op], we can use memcpy here: 
 *      #define pop_stack(Sc) do {Sc->stack_end -= 4; memcpy((void *)Sc, (void *)(Sc->stack_end), 4 * sizeof(s7_pointer));} while (0)
 *    but it is only slightly faster (.2% at best)!
 */

#define pop_stack(Sc) \
  do { \
  Sc->stack_end -= 4; \
  Sc->code =  Sc->stack_end[0]; \
  Sc->envir = Sc->stack_end[1]; \
  Sc->args =  Sc->stack_end[2]; \
  Sc->op =    (opcode_t)(Sc->stack_end[3]); \
  } while (0)

#define pop_stack_no_op(Sc) \
  do { \
  Sc->stack_end -= 4; \
  Sc->code =  Sc->stack_end[0]; \
  Sc->envir = Sc->stack_end[1]; \
  Sc->args =  Sc->stack_end[2]; \
  } while (0)

#define push_stack(Sc, Op, Args, Code) \
  do { \
  Sc->stack_end[0] = Code; \
  Sc->stack_end[1] = Sc->envir; \
  Sc->stack_end[2] = Args; \
  Sc->stack_end[3] = (s7_pointer)Op; \
  Sc->stack_end += 4; \
  } while (0)

#define push_stack_no_args(Sc, Op, Code) \
  do { \
  Sc->stack_end[0] = Code; \
  Sc->stack_end[1] = Sc->envir; \
  Sc->stack_end[3] = (s7_pointer)Op; \
  Sc->stack_end += 4; \
  } while (0)



static void increase_stack_size(s7_scheme *sc)
{
  int i, new_size, loc;

  loc = s7_stack_top(sc);
  new_size = sc->stack_size * 2;

  vector_elements(sc->stack) = (s7_pointer *)realloc(vector_elements(sc->stack), new_size * sizeof(s7_pointer));
  for (i = sc->stack_size; i < new_size; i++)
    vector_element(sc->stack, i) = sc->NIL;
  vector_length(sc->stack) = new_size;
  sc->stack_size = new_size;

  sc->stack_start = vector_elements(sc->stack);
  sc->stack_end = (s7_pointer *)(sc->stack_start + loc);
  sc->stack_resize_trigger = (s7_pointer *)(sc->stack_start + sc->stack_size / 2);
} 


static s7_pointer g_stack_size(s7_scheme *sc, s7_pointer args)
{
  return(s7_make_integer(sc, (sc->stack_end - sc->stack_start) / 4));
}




/* -------------------------------- symbols -------------------------------- */

#define HASH_MULT 4

static int symbol_table_hash(const char *key, unsigned int *loc) 
{ 
  unsigned int hashed = 0;
  const char *c; 
  for (c = key; *c; c++) 
    hashed = *c + hashed * HASH_MULT;
  if (loc) (*loc) = hashed;
  return(hashed % SYMBOL_TABLE_SIZE); 

  /* using ints here is much faster, and the symbol table will not be enormous, so it's worth splitting out this case */

  /* chain lengths (after s7test using mult=37): 
   *   all chars: 43 159 329 441 395 349 217 152 69 35 12 4 1 0 0 1          max: 15
   *    4 chars: 572 528 307 183 128 90 50 48 41 35 34 23 21 ...             max: 182!
   *    8 chars: 114 307 404 411 301 197 146 98 77 35 28 18 11 16 ...        max: 79
   *    16 chars: 44 160 344 400 435 348 206 143 72 31 16 4 0 1 0 0 0 2 1... max: 18
   *
   * currently the hash calculation is ca 8 (s7test) and the find_by_name process 3,
   *   if we use 4 chars, this calc goes to 6/7 but the find calc to 8/9
   *
   * the multiplier (4 currently) doesn't matter:
   *   mult  1: 1744
   *         2: 1743
   *         3: 1743
   *         7: 1744
   *         8: 1743
   *        16: 1745
   *        17: 1746
   *        32: 1746
   *        37: 1744
   *        39: 1744
   */
} 


static s7_pointer new_symbol(s7_scheme *sc, const char *name, int location) 
{ 
  s7_pointer x, str; 
  
  str = s7_make_permanent_string(name);
  x = (s7_cell *)permanent_calloc(sizeof(s7_extended_cell));
  x->hloc = NOT_IN_HEAP;
  symbol_name_cell(x) = str;
  set_type(x, T_SYMBOL);
  global_slot(x) = sc->NIL;
  local_slot(x) = sc->NIL;
  symbol_id(x) = 0;
  symbol_accessor(x) = -1;

  if ((symbol_name_length(x) > 1) &&                           /* not 0, otherwise : is a keyword */
      ((name[0] == ':') ||
       (name[symbol_name_length(x) - 1] == ':')))
    typeflag(x) |= (T_IMMUTABLE | T_KEYWORD); 

  vector_element(sc->symbol_table, location) = permanent_cons(x, vector_element(sc->symbol_table, location), 
							      T_PAIR | T_IMMUTABLE);
  return(x); 
} 


static s7_pointer symbol_table_find_by_name(s7_scheme *sc, const char *name, int location) 
{ 
  s7_pointer x; 
  for (x = vector_element(sc->symbol_table, location); is_not_null(x); x = cdr(x)) 
    { 
      const char *s; 
      s = symbol_name(car(x)); 
      if ((s) && (*s == *name) && (strings_are_equal(name, s)))
	return(car(x)); 
    }
  return(sc->NIL); 
} 


static s7_pointer g_symbol_table(s7_scheme *sc, s7_pointer args)
{
  #define H_symbol_table "(symbol-table) returns the s7 symbol table (a vector)"
  return(vector_copy(sc, sc->symbol_table));
}


void s7_for_each_symbol_name(s7_scheme *sc, bool (*symbol_func)(const char *symbol_name, void *data), void *data)
{
  int i; 
  s7_pointer x; 

  for (i = 0; i < vector_length(sc->symbol_table); i++) 
    for (x  = vector_element(sc->symbol_table, i); is_not_null(x); x = cdr(x)) 
      if (symbol_func(symbol_name(car(x)), data))
	return;
}


void s7_for_each_symbol(s7_scheme *sc, bool (*symbol_func)(const char *symbol_name, s7_pointer value, void *data), void *data)
{
  int i; 
  s7_pointer x; 

  for (i = 0; i < vector_length(sc->symbol_table); i++) 
    for (x  = vector_element(sc->symbol_table, i); is_not_null(x); x = cdr(x)) 
      if (symbol_func(symbol_name(car(x)), cdr(x), data))
	return;
}

/* (for-each
   (lambda (lst)
      (for-each
         (lambda (sym)
            (format #t "~A ~S~%" sym (symbol->value sym)))
         lst))
   (symbol-table))
   
   at normal motif-snd startup there are 5699 globals (2583 of them constant), and 411 other undefined symbols
*/


static s7_pointer g_symbol_table_is_locked(s7_scheme *sc, s7_pointer args)
{
  #define H_symbol_table_is_locked "set (-s7-symbol-table-locked?) to #t to prohibit the creation of any new symbols"
  return(make_boolean(sc, sc->symbol_table_is_locked));
}


static s7_pointer g_set_symbol_table_is_locked(s7_scheme *sc, s7_pointer args)
{
  sc->symbol_table_is_locked = (car(args) != sc->F);
  return(car(args));
}


static s7_pointer make_symbol(s7_scheme *sc, const char *name) 
{
  s7_pointer x; 
  int location;
  unsigned int loc = 0;

  location = symbol_table_hash(name, &loc); 
  x = symbol_table_find_by_name(sc, name, location); 
  if (is_not_null(x)) 
    return(x); 

  if (sc->symbol_table_is_locked)
    return(s7_error(sc, sc->ERROR, sc->NIL));

  x = new_symbol(sc, name, location); 
  symbol_hash(x) = loc;
  return(x);
} 


s7_pointer s7_make_symbol(s7_scheme *sc, const char *name) 
{ 
  if (!name) return(sc->F);
  return(make_symbol(sc, name));
}


s7_pointer s7_gensym(s7_scheme *sc, const char *prefix)
{ 
  char *name;
  int len, location;
  s7_pointer x;
  unsigned int loc = 0;
  
  len = safe_strlen(prefix) + 32;
  name = (char *)calloc(len, sizeof(char));
  
  for (; sc->gensym_counter < S7_LONG_MAX; ) 
    { 
      snprintf(name, len, "{%s}-%d", prefix, sc->gensym_counter++); 
      location = symbol_table_hash(name, &loc); 
      x = symbol_table_find_by_name(sc, name, location); 
      if (is_not_null(x))
	{
	  if (s7_symbol_value(sc, x) != sc->UNDEFINED)
	    continue; 
	  free(name);
	  return(x); 
	}
      
      x = new_symbol(sc, name, location); 
      symbol_hash(x) = loc;
      free(name);
      return(x); 
    } 
  free(name);
  return(sc->NIL); 
} 


static s7_pointer g_gensym(s7_scheme *sc, s7_pointer args) 
{
  #define H_gensym "(gensym (prefix \"gensym\")) returns a new, unused symbol"
  if (is_not_null(args))
    {
      if (!s7_is_string(car(args)))
	return(s7_wrong_type_arg_error(sc, "gensym prefix,", 0, car(args), "a string"));
      return(s7_gensym(sc, string_value(car(args))));
    }
  return(s7_gensym(sc, "gensym"));
}


s7_pointer s7_name_to_value(s7_scheme *sc, const char *name)
{
  return(s7_symbol_value(sc, make_symbol(sc, name)));
}


bool s7_is_symbol(s7_pointer p)   
{ 
  return(is_symbol(p));
}


static s7_pointer g_is_symbol(s7_scheme *sc, s7_pointer args)
{
  #define H_is_symbol "(symbol? obj) returns #t if obj is a symbol"
  return(make_boolean(sc, is_symbol(car(args))));
}


const char *s7_symbol_name(s7_pointer p)   
{ 
  return(symbol_name(p));
}


static s7_pointer g_symbol_to_string(s7_scheme *sc, s7_pointer args)
{
  #define H_symbol_to_string "(symbol->string sym) returns the symbol sym converted to a string"
  s7_pointer sym;

  sym = car(args);
  if (!is_symbol(sym))
    return(s7_wrong_type_arg_error(sc, "symbol->string", 0, sym, "a symbol"));
  
  /* s7_make_string uses strlen which stops at an embedded null
   */
  return(s7_make_string_with_length(sc, symbol_name(sym), symbol_name_length(sym)));    /* return a copy */
}

#if WITH_OPTIMIZATION
static s7_pointer symbol_to_string_uncopied;
static s7_pointer g_symbol_to_string_uncopied(s7_scheme *sc, s7_pointer args)
{
  s7_pointer sym, x;

  sym = car(args);
  if (!is_symbol(sym))
    return(s7_wrong_type_arg_error(sc, "symbol->string", 0, sym, "a symbol"));

  x = make_string_uncopied_with_length(sc, symbol_name(sym), symbol_name_length(sym)); 
  string_needs_free(x) = false;
  return(x);
}
#endif



static s7_pointer g_string_to_symbol_1(s7_scheme *sc, s7_pointer args, const char *caller)
{
  s7_pointer str;
  str = car(args);

  if (!s7_is_string(str))
    return(s7_wrong_type_arg_error(sc, caller, 0, str, "a string"));

  /* currently if the string has an embedded null, it marks the end of the new symbol name.
   *   I wonder if this is a bug...
   */
  return(make_symbol(sc, string_value(str)));
}


static s7_pointer g_string_to_symbol(s7_scheme *sc, s7_pointer args)
{
  #define H_string_to_symbol "(string->symbol str) returns the string str converted to a symbol"
  return(g_string_to_symbol_1(sc, args, "string->symbol"));
}


static s7_pointer g_symbol(s7_scheme *sc, s7_pointer args)
{
  #define H_symbol "(symbol str) returns the string str converted to a symbol"
  return(g_string_to_symbol_1(sc, args, "symbol"));
}





/* -------------------------------- environments -------------------------------- */

static unsigned long long int frame_number = 0;

#define NEW_FRAME(Sc, Old_Env, New_Env)  \
  do {                                   \
      s7_pointer _x_;                      \
      NEW_CELL(Sc, _x_);                   \
      frame_id(_x_) = ++frame_number; \
      environment_slots(_x_) = Sc->NIL;                  \
      next_environment(_x_) = Old_Env;		         \
      set_type(_x_, T_ENVIRONMENT | T_IMMUTABLE); \
      New_Env = _x_;		   \
     } while (0)


static s7_pointer new_frame_in_env(s7_scheme *sc, s7_pointer old_env) 
{ 
  /* return(cons(sc, sc->NIL, old_env)); */
  s7_pointer x;
  NEW_CELL(sc, x);
  frame_id(x) = ++frame_number;
  environment_slots(x) = sc->NIL;
  next_environment(x) = old_env;
  set_type(x, T_ENVIRONMENT | T_IMMUTABLE);
  return(x);
} 


#define ADD_SLOT(Frame, Symbol, Value) \
  do {\
    s7_pointer _slot_;\
    NEW_CELL_NO_CHECK(sc, _slot_);\
    slot_symbol(_slot_) = Symbol;\
    symbol_id(slot_symbol(_slot_)) = frame_id(Frame);	\
    local_slot(slot_symbol(_slot_)) = _slot_;\
    slot_set_value(_slot_, Value);	\
    set_type(_slot_, T_SLOT | T_IMMUTABLE);\
    next_slot(_slot_) = environment_slots(Frame);\
    environment_slots(Frame) = _slot_;\
  } while (0)

/* no set_local here -- presumably done earlier in check_* 
 */


#define NEW_FRAME_WITH_SLOT(Sc, Old_Env, New_Env, Symbol, Value) \
  do {                                   \
      s7_pointer _x_, _slot_;			 \
      NEW_CELL(Sc, _x_);                   \
      frame_id(_x_) = ++frame_number; \
      next_environment(_x_) = Old_Env;		         \
      set_type(_x_, T_ENVIRONMENT | T_IMMUTABLE); \
      New_Env = _x_;		   \
      NEW_CELL_NO_CHECK(Sc, _slot_);\
      slot_symbol(_slot_) = Symbol;\
      symbol_id(slot_symbol(_slot_)) = frame_number;	\
      local_slot(slot_symbol(_slot_)) = _slot_;\
      slot_set_value(_slot_, Value);	\
      set_type(_slot_, T_SLOT | T_IMMUTABLE);\
      next_slot(_slot_) = sc->NIL;\
      environment_slots(_x_) = _slot_;\
     } while (0)


static s7_pointer permanent_slot(s7_pointer symbol, s7_pointer value)
{
  s7_pointer x;
  x = (s7_cell *)permanent_calloc(sizeof(s7_cell));
  x->hloc = NOT_IN_HEAP;
  slot_symbol(x) = symbol;
  slot_set_value(x, value);
  set_type(x, T_SLOT | T_IMMUTABLE);
  return(x);
}


bool s7_is_environment(s7_pointer e)
{
  return(is_environment(e));
}


static s7_pointer g_is_environment(s7_scheme *sc, s7_pointer args)
{
  #define H_is_environment "(environment? obj) returns #t if obj is an environment."
  return(make_boolean(sc, is_environment(car(args))));
}


static int environment_length(s7_scheme *sc, s7_pointer e)
{
  int i;
  s7_pointer p;
  if (e == sc->global_env)
    return(vector_fill_pointer(e));
  for (i = 0, p = environment_slots(e); is_slot(p); i++, p = next_slot(p));
  return(i);
}


s7_pointer s7_make_slot(s7_scheme *sc, s7_pointer env, s7_pointer symbol, s7_pointer value) 
{ 
  s7_pointer slot;

  if ((!is_environment(env)) ||
      (env == sc->global_env))
    {
      s7_pointer ge;

      if ((sc->safety == 0) &&
	  ((is_closure(value)) ||
	   (is_closure_star(value)) ||
	   (is_macro(value)) ||
	   (is_bacro(value))))
	{
	  s7_remove_from_heap(sc, closure_args(value));
	  s7_remove_from_heap(sc, closure_body(value));
	}

      ge = sc->global_env;
      slot = permanent_slot(symbol, value);
      vector_element(ge, vector_fill_pointer(ge)++) = slot;
      if (vector_fill_pointer(ge) >= vector_length(ge))
	{
	  int i;
	  vector_length(ge) *= 2;
	  vector_elements(ge) = (s7_pointer *)realloc(vector_elements(ge), vector_length(ge) * sizeof(s7_pointer));
	  for (i = vector_fill_pointer(ge); i < vector_length(ge); i++)
	    vector_element(ge, i) = sc->NIL;
	}
      global_slot(symbol) = slot;
      if (symbol_id(symbol) == 0) /* never defined locally? */
	{
	  local_slot(symbol) = slot;
	  set_global(symbol);
	}
      /* symbol_id(symbol) = 0; */
    }
  else 
    {
      set_local(symbol);

      NEW_CELL(sc, slot);
      slot_symbol(slot) = symbol;
      slot_set_value(slot, value);
      set_type(slot, T_SLOT | T_IMMUTABLE);
      next_slot(slot) = environment_slots(env);
      environment_slots(env) = slot;
      symbol_id(symbol) = frame_id(env);
      local_slot(symbol) = slot;
    }

  /* there are about the same number of frames as local variables -- this
   *   strikes me as surprising, but it holds up across a lot of code.
   */

  return(slot);
} 


static s7_pointer add_slot(s7_scheme *sc, s7_pointer variable, s7_pointer value) 
{ 
  /* this is called when it is guaranteed that there is a local environment */
  s7_pointer y;

  NEW_CELL(sc, y);
  slot_symbol(y) = variable;
  slot_set_value(y, value);
  set_type(y, T_SLOT | T_IMMUTABLE);
  next_slot(y) = environment_slots(sc->envir);
  environment_slots(sc->envir) = y;
  set_local(variable);
  symbol_id(variable) = frame_id(sc->envir);
  local_slot(variable) = y;
  return(y);
} 


static void save_initial_environment(s7_scheme *sc)
{
  /* there are ca 270 predefined functions (and another 30 or so other things): 302 at last count
   */
  #define INITIAL_ENV_ENTRIES 400
  int i, k = 0, len;
  s7_pointer ge;
  s7_pointer *lsts, *inits;

  sc->initial_env = (s7_pointer)calloc(1, sizeof(s7_cell));
  set_type(sc->initial_env, T_VECTOR);
  vector_length(sc->initial_env) = INITIAL_ENV_ENTRIES;
  vector_elements(sc->initial_env) = (s7_pointer *)malloc(INITIAL_ENV_ENTRIES * sizeof(s7_pointer));
  inits = vector_elements(sc->initial_env);
  s7_vector_fill(sc, sc->initial_env, sc->NIL);
  sc->initial_env->hloc = NOT_IN_HEAP;

  ge = sc->global_env;
  len = vector_fill_pointer(ge);
  lsts = vector_elements(ge);
   
  for (i = 0; i < len; i++)
    {
      s7_pointer slot;
      slot = lsts[i];
      if (is_procedure(slot_value(slot)))
	{
	  inits[k++] = permanent_slot(slot_symbol(slot), slot_value(slot));
	  if (k >= INITIAL_ENV_ENTRIES)
	    return;
	}
    }
}


static s7_pointer g_initial_environment(s7_scheme *sc, s7_pointer args)
{
  /* add sc->initial_env bindings to the current environment */
  #define H_initial_environment "(initial-environment) establishes the original bindings of all the predefined functions"

  /* maybe this should be named with-initial-environment or something -- it currently looks
   *   like it simply returns the initial env, but it actually shadows the global env entries
   *   that have changed.  But that doesn't behave the way one would expect:
   *
   *     (let () (define (x) 1) (defined? 'x (initial-environment)) -> #t!
   *   
   *   I can't see how to fix this short of copying the thing.
   */
  int i;
  s7_pointer *inits;
  s7_pointer x;

  sc->w = new_frame_in_env(sc, sc->envir);
  inits = vector_elements(sc->initial_env);

  for (i = 0; (i < INITIAL_ENV_ENTRIES) && (is_slot(inits[i])); i++)
    if ((!is_global(slot_symbol(inits[i]))) ||                                      /* it's shadowed locally */
	(slot_value(inits[i]) != slot_value(global_slot(slot_symbol(inits[i])))))   /* it's not shadowed, but has been changed globally */
      s7_make_slot(sc, sc->w, slot_symbol(inits[i]), slot_value(inits[i]));
                         
  /* if (set! + -) then + needs to be overridden, but the local bit isn't set,
   *   so we have to check the actual values in the non-local case.
   *   (define (f x) (with-environment (initial-environment) (+ x 1))) 
   */

  x = sc->w;
  sc->w = sc->NIL;
  return(x);
}

/* should these two augment-envs check for symbol accessors?
 */

static s7_pointer g_augment_environment_direct(s7_scheme *sc, s7_pointer args)
{
  #define H_augment_environment_direct "(augment-environment! env ...) adds its \
arguments (each a cons: symbol . value) directly to the environment env, and returns the \
environment."

  s7_pointer x, e;
  int i;

  e = car(args);
  if (is_null(e))
    e = sc->global_env;
  else
    {
      if (!is_environment(e))
	return(s7_wrong_type_arg_error(sc, "augment-environment!", 1, e, "an environment"));
    }
  for (i = 2, x = cdr(args); is_not_null(x); x = cdr(x), i++)
    {
      s7_pointer sym, val;
      if ((!is_pair(car(x))) ||
	  (!is_symbol(caar(x))))
	return(s7_wrong_type_arg_error(sc, "augment-environment!", i, car(x), "a pair: '(symbol . value)"));

      sym = caar(x);
      val = cdar(x);
      if ((is_immutable(sym)) &&                            /* check for (eval 'pi (augment-environment! () '(pi . 1))) */
	  (!s7_is_equal(sc, val, s7_symbol_value(sc, sym))))
	return(s7_wrong_type_arg_error(sc, "augment-environment!", i, sym, "mutable symbol"));
    }
  
  if (e == sc->global_env)
    {
      for (x = cdr(args); is_not_null(x); x = cdr(x))
	{
	  if (is_slot(global_slot(caar(x))))
	    slot_set_value(global_slot(caar(x)), cdar(x));
	  else s7_make_slot(sc, e, caar(x), cdar(x));
	}
    }
  else
    {
      for (x = cdr(args); is_not_null(x); x = cdr(x))
	s7_make_slot(sc, e, caar(x), cdar(x));
    }
  return(e);
}


s7_pointer s7_augment_environment(s7_scheme *sc, s7_pointer e, s7_pointer bindings)
{
  s7_pointer new_e;
  /* fprintf(stderr, "e: %p, slots: %p, next: %p\n", e, environment_slots(e), next_environment(e)); */

  if (e == sc->global_env)
    new_e = new_frame_in_env(sc, sc->NIL);
  else new_e = new_frame_in_env(sc, e);

  if (!is_null(bindings))
    {
      s7_pointer x;
      int gc_loc;
      gc_loc = s7_gc_protect(sc, new_e);

      for (x = bindings; is_not_null(x); x = cdr(x))
	s7_make_slot(sc, new_e, caar(x), cdar(x));

      s7_gc_unprotect_at(sc, gc_loc);
    }

  /* fprintf(stderr, "new_e: %p, slots: %p, next: %p\n", new_e, environment_slots(new_e), next_environment(new_e)); */
  return(new_e);
}


#if 0
typedef struct c_example_t {
  s7_Int integer_value;
  s7_Double real_value;
} c_example_t;

static s7_pointer c_struct_to_s7_object(s7_scheme *sc, c_example_t *p)
{
  return(s7_augment_environment(sc, s7_nil(sc), 
				s7_cons(sc, s7_make_symbol(sc, "integer_value"), s7_make_integer(sc, p->integer_value)),
				s7_cons(sc, s7_make_symbol(sc, "real_value"), s7_make_real(sc, p->real_value)),

  /* now in scheme, (p 'integer_value) gives that field, but how to automate? how to set the C side from scheme?
   *    pws seems redundant, ((p 'integer_value) p) is redundant
   *    local (to the object) symbol_accessor: integer_value -> s7_make_integer(sc, unwrap_p->integer_value)
   * so we need env-local accessors, at least from C
   *    all these implicit refs (anything outside with-environment) go through environment_ref|set
   *    and how to tell (e sym) wants to evaluate the pws as if it were (sym e)?  maybe an extension of env: T_C_ENVIRONMENT?
   *
   * ultimately we want to replace hooks with objects where all hook functions take no args and return nothing,
   *   but handle all exchange through the object.  Currently hooks are bad because the caller (on both sides)
   *   has to remember how many arguments each takes, and what they mean and what the return value means, and
   *   when there are lots of hooks (Snd has nearly 100), what all the names are.  Similarly, all the C struct
   *   accesses have to go through the pws mechanism which requires endless documentation and naming and C-side boilplate.
   *   And ideally C and Scheme would share access paths, and names (xg.c for example) -- only one name to remember.
   *
   *   snd_state->sounds[selected_sound]->chans[0]->edits[current_edit]->time_graph->axis_x0 or some such path to
   *   (axis-x0 (graphs (edits (channels (sounds *snd* (selected-sound *snd*)) 0) current-edit) time-graph))
   *
   * 1st step: define a constant *snd*
   * snd-xen.c has g_snd_sound_pointer which is sort of like sounds above
   */
}
#endif


/* (define (environment . args) (apply augment-environment '() args))
 * 
 * currently (eval 'a (augment-environment () '(a . 1) '(a . 2))) -> 2,
 *   but should it be an error?
 */

static s7_pointer g_augment_environment(s7_scheme *sc, s7_pointer args)
{
  #define H_augment_environment "(augment-environment env ...) adds its \
arguments (each a cons: symbol . value) to the environment env, and returns the \
new environment."

  s7_pointer e, x;
  int i;

  e = car(args);
  if (is_null(e))
    e = sc->global_env;
  else
    {
      if (!is_environment(e))
	return(s7_wrong_type_arg_error(sc, "augment-environment", 1, e, "an environment"));
    }

  for (i = 2, x = cdr(args); is_not_null(x); x = cdr(x), i++)
    {
      s7_pointer sym, val;
      if ((!is_pair(car(x))) ||
	  (!is_symbol(caar(x))))
	return(s7_wrong_type_arg_error(sc, "augment-environment", i, car(x), "a pair: '(symbol . value)"));

      sym = caar(x);
      val = cdar(x);
      if ((is_immutable(sym)) &&                            /* check for (eval 'pi (augment-environment () '(pi . 1))) */
	  (!s7_is_equal(sc, val, s7_symbol_value(sc, sym))))
	return(s7_wrong_type_arg_error(sc, "augment-environment", i, sym, "mutable symbol"));
    }

  return(s7_augment_environment(sc, e, cdr(args)));
}


s7_pointer s7_environment_to_list(s7_scheme *sc, s7_pointer env)
{
  s7_pointer x;

  sc->w = sc->NIL;
  if (env == sc->global_env)
    {
      unsigned int i;
      for (i = 0; i < vector_fill_pointer(env); i++)
	if (is_slot(vector_element(env, i)))
	  sc->w = cons(sc, cons(sc, slot_symbol(vector_element(env, i)), slot_value(vector_element(env, i))), sc->w);
    }
  else
    {
      for (x = environment_slots(env); is_slot(x); x = next_slot(x))
	sc->w = cons(sc, cons(sc, slot_symbol(x), slot_value(x)), sc->w);
    }
  x = sc->w;
  sc->w = sc->NIL;
  return(x);
}


static s7_pointer g_environment_to_list(s7_scheme *sc, s7_pointer args)
{
  #define H_environment_to_list "(environment->list env) returns env's bindings as a list of cons's: '(symbol . value)."

  s7_pointer env;
  env = car(args);
  if (!is_environment(env))
    return(s7_wrong_type_arg_error(sc, "environment->list", 0, env, "an environment"));

  return(s7_environment_to_list(sc, env));
}


static s7_pointer environment_ref(s7_scheme *sc, s7_pointer env, s7_pointer symbol)
{
  /* (let ((a 1)) ((current-environment) 'a)) */
  s7_pointer x, y;

  for (x = env; is_environment(x); x = next_environment(x))
    for (y = environment_slots(x); is_slot(y); y = next_slot(y))
      if (slot_symbol(y) == symbol)
	return(cdr(y));

  return(sc->UNDEFINED);
}


static s7_pointer environment_set(s7_scheme *sc, s7_pointer env, s7_pointer symbol, s7_pointer value)
{
  s7_pointer x, y;
  for (x = env; is_environment(x); x = next_environment(x))
    for (y = environment_slots(x); is_slot(y); y = next_slot(y))
      if (slot_symbol(y) == symbol)
	{
	  cdr(y) = value;
	  return(value);
	}
  return(sc->UNDEFINED);
}


static s7_pointer g_environment_set(s7_scheme *sc, s7_pointer args)
{
  /* (let ((a 1)) (set! ((current-environment) 'a) 32) a) */
  s7_pointer sym;

  sym = cadr(args);
  if (!is_symbol(sym))
    return(s7_wrong_type_arg_error(sc, "environment set!", 2, sym, "a symbol"));

  return(environment_set(sc, car(args), sym, caddr(args)));
}


static s7_pointer reverse_slots(s7_scheme *sc, s7_pointer list)
{
  s7_pointer p = list, result, q;
  result = sc->NIL;

  while (is_slot(p))
    {
      q = next_slot(p);
      next_slot(p) = result;
      result = p;
      p = q;
    }
  return(result);
}


static s7_pointer environment_copy(s7_scheme *sc, s7_pointer env)
{
  if (is_environment(env))
    {
      s7_pointer x, new_e;
      int gc_loc;

      new_e = new_frame_in_env(sc, next_environment(env));
      gc_loc = s7_gc_protect(sc, new_e);

      for (x = environment_slots(env); is_slot(x); x = next_slot(x))
	ADD_SLOT(new_e, slot_symbol(x), slot_value(x));
      environment_slots(new_e) = reverse_slots(sc, environment_slots(new_e));
      
      s7_gc_unprotect_at(sc, gc_loc);      
      return(new_e);
    }
  return(sc->NIL);
}



static s7_pointer g_global_environment(s7_scheme *sc, s7_pointer ignore)
{
  #define H_global_environment "(global-environment) returns the current top-level definitions (symbol bindings). \
It is a hash-table."
  return(sc->global_env);
}

/* as with the symbol-table, this function can lead to disaster -- user could
 *   clobber the environment etc.  But we want it to be editable and augmentable,
 *   so I guess I'll leave it alone.  (See current|procedure-environment as well).
 */


s7_pointer s7_global_environment(s7_scheme *sc) 
{
  return(sc->global_env);
}


static s7_pointer g_current_environment(s7_scheme *sc, s7_pointer args)
{
  #define H_current_environment "(current-environment) returns the current definitions (symbol bindings)"

  if (is_environment(sc->envir))
    return(sc->envir);
  return(sc->global_env);
}


s7_pointer s7_current_environment(s7_scheme *sc) 
{
  return(sc->envir);
}


static s7_pointer g_outer_environment(s7_scheme *sc, s7_pointer args)
{
  #define H_outer_environment "(outer-environment env) returns the environment that contains env."
  s7_pointer env;

  env = car(args);
  if (!is_environment(env))
    return(s7_wrong_type_arg_error(sc, "outer-environment", 0, env, "an environment"));  

  if ((env == sc->global_env) ||
      (is_null(next_environment(env))))
    return(sc->global_env);
  return(next_environment(env));
}


#if 0
static char *display_locals(s7_scheme *sc)
{
  s7_pointer e;
  for (e = sc->envir, spaces = 0; is_environment(e); e = next_environment(e), spaces += 2)
    fprintf(stderr, "%s\n", str = s7_object_to_c_string(sc, e));
  return((char *)"");
}
#endif


static s7_pointer find_symbol(s7_scheme *sc, s7_pointer hdl)
{ 
  s7_pointer x;	

  for (x = sc->envir; symbol_id(hdl) < frame_id(x); x = next_environment(x));

  if (frame_id(x) == symbol_id(hdl))
    return(local_slot(hdl));	

  for (; is_environment(x); x = next_environment(x))
    {
      s7_pointer y; 
      for (y = environment_slots(x); is_slot(y); y = next_slot(y))	
	if (slot_symbol(y) == hdl)
	  return(y); 
    }

  return(global_slot(hdl));
} 


s7_pointer s7_slot(s7_scheme *sc, s7_pointer symbol)
{
  return(find_symbol(sc, symbol));
}


s7_pointer s7_slot_value(s7_scheme *sc, s7_pointer slot)
{
  return(slot_value(slot));
}


s7_pointer s7_slot_set_value(s7_scheme *sc, s7_pointer slot, s7_pointer value)
{
  slot_set_value(slot, value);
  return(value);
}


static s7_pointer find_local_symbol(s7_scheme *sc, s7_pointer env, s7_pointer hdl) 
{ 
  s7_pointer y;

  if (!is_environment(env))
    return(global_slot(hdl));

  for (y = environment_slots(env); is_slot(y); y = next_slot(y))
    if (slot_symbol(y) == hdl)
      return(y);

  return(sc->NIL);
} 


s7_pointer s7_symbol_value(s7_scheme *sc, s7_pointer sym) /* was searching just the global environment? */
{
  s7_pointer x;

  x = find_symbol(sc, sym);
  if (is_slot(x))
    return(slot_value(x));

  if (is_keyword(sym))
    return(sym);
  return(sc->UNDEFINED);
}


s7_pointer s7_symbol_local_value(s7_scheme *sc, s7_pointer sym, s7_pointer local_env)
{
  s7_pointer x;
  if (is_environment(local_env))
    {
      for (x = local_env; is_environment(x); x = next_environment(x))
	{
	  s7_pointer y;
	  for (y = environment_slots(x); is_slot(y); y = next_slot(y))
	    if (slot_symbol(y) == sym)
	      return(slot_value(y));
	}
    }
  return(s7_symbol_value(sc, sym)); 
}


static s7_pointer g_symbol_to_value(s7_scheme *sc, s7_pointer args)
{
  #define H_symbol_to_value "(symbol->value sym (env (current-environment))) returns the binding of (the value associated with) the \
symbol sym in the given environment: (let ((x 32)) (symbol->value 'x)) -> 32"

  s7_pointer sym;
  sym = car(args);

  if (!is_symbol(sym))
    return(s7_wrong_type_arg_error(sc, "symbol->value", (is_null(cdr(args))) ? 0 : 1, sym, "a symbol"));

  if (is_not_null(cdr(args)))
    {
      s7_pointer x, local_env;
      if (!is_environment(cadr(args)))
	return(s7_wrong_type_arg_error(sc, "symbol->value", 2, cadr(args), "an environment"));

      local_env = cadr(args);
      if (local_env == sc->global_env)
	{
	  x = global_slot(sym);
	  if (is_not_null(x))
	    return(slot_value(x));
	  return(sc->UNDEFINED);
	}
      return(s7_symbol_local_value(sc, sym, local_env));
    }
  
  if (is_global(sym))
    return(slot_value(global_slot(sym)));

  return(s7_symbol_value(sc, sym));
}


s7_pointer s7_symbol_set_value(s7_scheme *sc, s7_pointer sym, s7_pointer val)
{
  s7_pointer x;
  /* if immutable should this return an error? */
  x = find_symbol(sc, sym);
  if (is_slot(x))
    slot_set_value(x, val);
  return(val);
}


static s7_pointer make_closure(s7_scheme *sc, s7_pointer args, s7_pointer code, int type) 
{
  /* this is called every time a lambda form is evaluated, or during letrec, etc */

  s7_pointer x;
  NEW_CELL(sc, x);
  closure_args(x) = args;
  closure_body(x) = code;
  closure_environment(x) = sc->envir;
  if (type == T_CLOSURE)
    set_type(x, type | T_PROCEDURE | T_COPY_ARGS);
  else set_type(x, type | T_PROCEDURE);
  return(x);
}


s7_pointer s7_make_closure(s7_scheme *sc, s7_pointer a, s7_pointer c, s7_pointer e)
{
  /* c is a list: args code, so 
   *   (define (proc a b) (+ a b)) becomes
   *   make_closure ((a b) (+ a b)) e
   */
  s7_pointer p;
  p = make_closure(sc, a, c, T_CLOSURE);
  closure_environment(p) = e;
  return(p);
}


#define MAKE_CLOSURE(Sc, X, Args, Code, Env)	\
  do { \
       NEW_CELL(Sc, X); \
       closure_args(X) = Args; \
       closure_body(X) = Code; \
       closure_environment(X) = Env; \
       set_type(X, T_CLOSURE | T_PROCEDURE | T_COPY_ARGS); \
      } while (0)



static s7_pointer g_is_defined(s7_scheme *sc, s7_pointer args)
{
  #define H_is_defined "(defined? obj (env (current-environment))) returns #t if obj has a binding (a value) in the environment env"
  s7_pointer sym, x;

  sym = car(args);
  if (!is_symbol(sym))
    return(s7_wrong_type_arg_error(sc, "defined?", (is_null(cdr(args))) ? 0 : 1, sym, "a symbol"));
  
  if (is_not_null(cdr(args)))
    {
      s7_pointer e;
      if (!is_environment(cadr(args)))
	return(s7_wrong_type_arg_error(sc, "defined?", 2, cadr(args), "an environment"));
      e = cadr(args);
      if (e == sc->global_env)
	{
	  x = global_slot(sym);
	  return(make_boolean(sc, (is_not_null(x)) && (x != sc->UNDEFINED)));
	}
      x = find_local_symbol(sc, e, sym); 
      if (is_slot(x))
	return(sc->T);
    }
  else 
    {
      if (is_global(sym))
	return(sc->T);
    }
  
  x = find_symbol(sc, sym); 
  return(make_boolean(sc, is_slot(x)));
}


bool s7_is_defined(s7_scheme *sc, const char *name)
{
  s7_pointer x;
  x = find_symbol(sc, make_symbol(sc, name));
  return(is_slot(x));
}


void s7_define(s7_scheme *sc, s7_pointer envir, s7_pointer symbol, s7_pointer value) 
{
  s7_pointer x;

  if (envir == sc->global_env) envir = sc->NIL; /* for C-side backwards compatibility */
  x = find_local_symbol(sc, envir, symbol);
  if (is_slot(x)) 
    slot_set_value(x, value); 
  else s7_make_slot(sc, envir, symbol, value); /* I think this means C code can override "constant" defs */
}


void s7_define_variable(s7_scheme *sc, const char *name, s7_pointer value)
{
  s7_pointer sym;

  sym = make_symbol(sc, name);
  s7_define(sc, sc->NIL, sym, value);
}


void s7_define_constant(s7_scheme *sc, const char *name, s7_pointer value)
{
  s7_pointer x, sym;
  
  sym = make_symbol(sc, name);
  s7_define(sc, sc->NIL, sym, value);
  x = global_slot(sym);
  set_immutable(slot_symbol(x));
}

/*        (define (func a) (let ((cvar (+ a 1))) cvar))
 *        (define-constant cvar 23)
 *        (func 1)
 *        ;can't bind an immutable object: cvar
 */




/* -------- keywords -------- */

bool s7_is_keyword(s7_pointer obj)
{
  return(is_keyword(obj));
}


static s7_pointer g_is_keyword(s7_scheme *sc, s7_pointer args)
{
  #define H_is_keyword "(keyword? obj) returns #t if obj is a keyword, (keyword? :key) -> #t"
  return(make_boolean(sc, is_keyword(car(args))));
}


s7_pointer s7_make_keyword(s7_scheme *sc, const char *key)
{
  s7_pointer sym;
  char *name;
  
  name = (char *)malloc((safe_strlen(key) + 2) * sizeof(char));
  sprintf(name, ":%s", key);                     /* prepend ":" */
  sym = make_symbol(sc, name);
  typeflag(sym) |= (T_IMMUTABLE); 
  free(name);
  
  s7_make_slot(sc, sc->NIL, sym, sym); /* make it global, not in the local env! */

  return(sym);
}


static s7_pointer g_make_keyword(s7_scheme *sc, s7_pointer args)
{
  #define H_make_keyword "(make-keyword str) prepends ':' to str and defines that as a keyword"
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "make-keyword", 0, car(args), "a string"));
  return(s7_make_keyword(sc, string_value(car(args))));
}


static s7_pointer g_keyword_to_symbol(s7_scheme *sc, s7_pointer args)
{
  #define H_keyword_to_symbol "(keyword->symbol key) returns a symbol with the same name as key but no prepended colon"
  const char *name;

  if (!is_keyword(car(args)))
    return(s7_wrong_type_arg_error(sc, "keyword->symbol", 0, car(args), "a keyword"));

  name = symbol_name(car(args));
  if (name[0] == ':')
    return(make_symbol(sc, (const char *)(name + 1)));

  /* else it ends in ":", (keyword->symbol foo:) */
  {
    char *temp;
    s7_pointer res;
    temp = copy_string(name);
    temp[strlen(temp) - 1] = '\0';
    res = make_symbol(sc, (const char *)temp);
    free(temp);
    return(res);
  }
}


static s7_pointer g_symbol_to_keyword(s7_scheme *sc, s7_pointer args)
{
  #define H_symbol_to_keyword "(symbol->keyword sym) returns a keyword with the same name as sym, but with a colon prepended"
  if (!is_symbol(car(args)))
    return(s7_wrong_type_arg_error(sc, "symbol->keyword", 0, car(args), "a symbol"));
  return(s7_make_keyword(sc, symbol_name(car(args))));
}



/* ---------------- uninterpreted pointers ---------------- */

bool s7_is_c_pointer(s7_pointer arg) 
{
  return(type(arg) == T_C_POINTER);
}


void *s7_c_pointer(s7_pointer p) 
{
  if ((is_number(p)) && 
      (s7_integer(p) == 0))
    return(NULL); /* special case where the null pointer has been cons'd up by hand */

  if (type(p) != T_C_POINTER)
    {
      fprintf(stderr, "s7_c_pointer argument is not a c pointer?");
      return(NULL);
    }

  return(raw_pointer(p));
}


s7_pointer s7_make_c_pointer(s7_scheme *sc, void *ptr)
{
  s7_pointer x;
  NEW_CELL(sc, x);
  set_type(x, T_C_POINTER);
  raw_pointer(x) = ptr;
  return(x);
}




/* -------------------------------- continuations and gotos -------------------------------- */

bool s7_is_continuation(s7_pointer p)    
{ 
  return(is_continuation(p));
}


static s7_pointer g_is_continuation(s7_scheme *sc, s7_pointer args)
{
  #define H_is_continuation "(continuation? obj) returns #t if obj is a continuation"
  return(make_boolean(sc, is_continuation(car(args))));
}

/* is this the right thing?  It returns #f for call-with-exit ("goto") because
 *   that form of continuation can't continue (via a jump back to its context).
 */


static s7_pointer copy_arg_list(s7_scheme *sc, s7_pointer lst)
{
  /* lst may be dotted, but not circular */
  if (is_null(lst))
    return(sc->NIL);
  if (is_pair(cdr(lst)))
    return(cons(sc, car(lst), copy_arg_list(sc, cdr(lst))));
  return(cons(sc, car(lst), cdr(lst)));
}


static s7_pointer copy_counter(s7_scheme *sc, s7_pointer obj)
{
  s7_pointer nobj;
  NEW_CELL(sc, nobj);
  counter_result(nobj) = counter_result(obj);
  counter_list(nobj) = counter_list(obj);
  counter_count(nobj) = counter_count(obj);
  counter_length(nobj) = counter_length(obj);
  return(nobj);
}


static s7_pointer copy_object(s7_scheme *sc, s7_pointer obj)
{
  s7_pointer nobj;
  int nloc;

  NEW_CELL(sc, nobj);
  nloc = nobj->hloc;
  memcpy((void *)nobj, (void *)obj, sizeof(s7_cell));
  nobj->hloc = nloc;
  
  /* nobj is safe here because the gc is off */
  if (dont_copy(car(obj)))
    car(nobj) = car(obj);
  else car(nobj) = copy_object(sc, car(obj));
  
  if ((dont_copy(cdr(obj))) ||
      ((typeflag(obj) & (T_PROCEDURE | T_ANY_MACRO)) != 0))
    cdr(nobj) = cdr(obj); /* closure_environment in func cases */
  else cdr(nobj) = copy_object(sc, cdr(obj));
  
  return(nobj);
}


static s7_pointer copy_stack(s7_scheme *sc, s7_pointer old_v, int top)
{
  int i, len;
  s7_pointer new_v;
  s7_pointer *nv, *ov;

  /* stacks can grow temporarily, so sc->stack_size grows, but we don't normally need all that
   *   leftover space here, so choose the original stack size if it's smaller.
   */

  len = vector_length(old_v);
  if ((top < INITIAL_STACK_SIZE / 2) &&
      (len > INITIAL_STACK_SIZE * 2))
    len = INITIAL_STACK_SIZE;

  new_v = s7_make_vector(sc, len);
  set_type(new_v, T_STACK);
  /* we can't leave the upper stuff simply malloc-garbage because we're sure to call the GC.
   *   We also can't just copy the vector since that seems to confuse the gc mark process.
   */

  nv = vector_elements(new_v);
  ov = vector_elements(old_v);
  
  s7_gc_on(sc, false);

  for (i = 0; i < top; i += 4)
    {
      s7_pointer p;
      if (dont_copy(ov[i]))
	nv[i] = ov[i];
      else nv[i] = copy_object(sc, ov[i]);    /* code */

      nv[i + 1] = ov[i + 1];                  /* environment pointer */
      nv[i + 3] = ov[i + 3];                  /* op (constant int) */

      p = ov[i + 2];
      if (is_pair(p))                         /* args need not be a list (it can be a port or #f, etc) */
	nv[i + 2] = copy_arg_list(sc, p);     /* args (copy is needed -- see s7test.scm) */
      else
	{
	  if (is_counter(p))                  /* these can only occur in this context */
	    nv[i + 2] = copy_counter(sc, p);
	  else nv[i + 2] = p;                 /* is this a safe assumption? */
	}
    }
  
  s7_gc_on(sc, true);
  return(new_v);
}


static s7_pointer make_goto(s7_scheme *sc) 
{
  s7_pointer x;
  NEW_CELL(sc, x);
  call_exit_goto_loc(x) = s7_stack_top(sc);
  call_exit_op_loc(x) = (int)(sc->op_stack_now - sc->op_stack);
  call_exit_active(x) = true;
  set_type(x, T_GOTO | T_PROCEDURE);
  return(x);
}


static s7_pointer *copy_op_stack(s7_scheme *sc)
{
  int len;
  s7_pointer *ops;
  ops = (s7_pointer *)malloc(sc->op_stack_size * sizeof(s7_pointer));
  /* memcpy((void *)ops, (void *)(sc->op_stack), sc->op_stack_size * sizeof(s7_pointer)); */
  len = (int)(sc->op_stack_now - sc->op_stack);
  if (len > 0)
    memcpy((void *)ops, (void *)(sc->op_stack), len * sizeof(s7_pointer));
  return(ops);
}


s7_pointer s7_make_continuation(s7_scheme *sc) 
{
  s7_pointer x;
  int loc;
  if ((int)(sc->free_heap_top - sc->free_heap) < (int)(sc->heap_size / 4))
    gc(sc);

  /* this gc call is needed if there are lots of call/cc's -- by pure bad luck
   *   we can end up hitting the end of the gc free list time after time while
   *   in successive copy_stack's below, causing s7 to core up until it runs out of memory.
   */

  loc = s7_stack_top(sc);

  NEW_CELL(sc, x);
  continuation(x) = (s7_continuation_t *)calloc(1, sizeof(s7_continuation_t));
  continuation_stack(x) = copy_stack(sc, sc->stack, s7_stack_top(sc));
  continuation_stack_size(x) = vector_length(continuation_stack(x));   /* copy_stack can return a smaller stack than the current one */
  continuation_stack_start(x) = vector_elements(continuation_stack(x));
  continuation_stack_end(x) = (s7_pointer *)(continuation_stack_start(x) + loc);

  continuation_op_stack(x) = copy_op_stack(sc);
  continuation_op_loc(x) = (int)(sc->op_stack_now - sc->op_stack);
  continuation_op_size(x) = sc->op_stack_size;

  set_type(x, T_CONTINUATION | T_PROCEDURE);
  add_continuation(sc, x);
  return(x);
}


static bool check_for_dynamic_winds(s7_scheme *sc, s7_pointer c)
{
  int i, s_base = 0, c_base = -1;
  opcode_t op;

  for (i = s7_stack_top(sc) - 1; i > 0; i -= 4)
    {
      s7_pointer x;

      op = stack_op(sc->stack, i);
      switch (op)
	{
	case OP_DYNAMIC_WIND:
	  {
	    int j;
	    x = stack_code(sc->stack, i);
	    for (j = 3; j < continuation_stack_top(c); j += 4)
	      if ((stack_op(continuation_stack(c), j) == OP_DYNAMIC_WIND) &&
		  (x == stack_code(continuation_stack(c), j)))
		{
		  s_base = i;
		  c_base = j;
		  break;
		}
	  
	    if (s_base != 0)
	      break;	  
	  
	    if (dynamic_wind_state(x) == DWIND_BODY)
	      {
		dynamic_wind_state(x) = DWIND_FINISH;
		push_stack(sc, OP_EVAL_DONE, sc->args, sc->code); 
		sc->args = sc->NIL;
		sc->code = dynamic_wind_out(x);
		eval(sc, OP_APPLY);
	      }
	  }
	  break;

	case OP_BARRIER:
	  if (i > continuation_stack_top(c))  /* otherwise it's some unproblematic outer eval-string? */
	    return(false);
	  break;

	case OP_DEACTIVATE_GOTO:              /* here we're jumping out of an unrelated call-with-exit block */
	  if (i > continuation_stack_top(c))
	    call_exit_active(stack_args(sc->stack, i)) = false;
	  break;

	case OP_TRACE_RETURN:
	  if (i > continuation_stack_top(c))
	    {
	      sc->trace_depth--;
	      if (sc->trace_depth < 0) sc->trace_depth = 0;
	    }
	  break;
	  
	default:
	  break;
	}
    }
  
  for (i = c_base + 4; i < continuation_stack_top(c); i += 4)
    {
      op = stack_op(continuation_stack(c), i);

      if (op == OP_DYNAMIC_WIND)
	{
	  s7_pointer x;
	  x = stack_code(continuation_stack(c), i);
	  push_stack(sc, OP_EVAL_DONE, sc->args, sc->code); 
	  sc->args = sc->NIL;
	  sc->code = dynamic_wind_in(x);
	  eval(sc, OP_APPLY);
	  dynamic_wind_state(x) = DWIND_BODY;
	}
      else
	{
	  if (op == OP_DEACTIVATE_GOTO)
	    call_exit_active(stack_args(continuation_stack(c), i)) = true;
	}
    }
  return(true);
}


static void call_with_current_continuation(s7_scheme *sc)
{
  if (!check_for_dynamic_winds(sc, sc->code))
    return;

  sc->stack = copy_stack(sc, continuation_stack(sc->code), continuation_stack_top(sc->code));
  sc->stack_size = continuation_stack_size(sc->code);
  sc->stack_start = vector_elements(sc->stack);
  sc->stack_end = (s7_pointer *)(sc->stack_start + continuation_stack_top(sc->code));
  sc->stack_resize_trigger = (s7_pointer *)(sc->stack_start + sc->stack_size / 2);
  
  {
    int i, top;
    top = continuation_op_loc(sc->code);
    sc->op_stack_now = (s7_pointer *)(sc->op_stack + top);
    sc->op_stack_size = continuation_op_size(sc->code);
    sc->op_stack_end = (s7_pointer *)(sc->op_stack + sc->op_stack_size);
    for (i = 0; i < top; i++)
      sc->op_stack[i] = continuation_op_stack(sc->code)[i];
  }

  if (is_null(sc->args))
    sc->value = sc->NIL;
  else
    {
      if (is_null(cdr(sc->args)))
	sc->value = car(sc->args);
      else sc->value = splice_in_values(sc, sc->args);
    }
}


static void call_with_exit(s7_scheme *sc)
{
  int i, new_stack_top;
  
  if (!call_exit_active(sc->code))
    s7_error(sc, make_symbol(sc, "invalid-escape-function"),
	     list_1(sc, make_protected_string(sc, "call-with-exit escape procedure called outside its block")));
  call_exit_active(sc->code) = false;
  new_stack_top = call_exit_goto_loc(sc->code);
  sc->op_stack_now = (s7_pointer *)(sc->op_stack + call_exit_op_loc(sc->code));

  /* look for dynamic-wind in the stack section that we are jumping out of */
  for (i = s7_stack_top(sc) - 1; i > new_stack_top; i -= 4)
    {
      opcode_t op;

      op = stack_op(sc->stack, i);
      switch (op)
	{
	case OP_DYNAMIC_WIND:
	  {
	    sc->z = stack_code(sc->stack, i);
	    if (dynamic_wind_state(sc->z) == DWIND_BODY)
	      {
		dynamic_wind_state(sc->z) = DWIND_FINISH;
		push_stack(sc, OP_EVAL_DONE, sc->args, sc->code); 
		sc->args = sc->NIL;
		sc->code = dynamic_wind_out(sc->z);
		sc->z = sc->NIL;
		eval(sc, OP_APPLY);
	      }
	  }
	  break;

	case OP_EVAL_STRING_2:
	  s7_close_input_port(sc, sc->input_port);
	  pop_input_port(sc);
	  break;

	case OP_BARRIER:                /* oops -- we almost certainly went too far */
	  return; 

	case OP_DEACTIVATE_GOTO:        /* here we're jumping into an unrelated call-with-exit block */
	  call_exit_active(stack_args(sc->stack, i)) = false;
	  break;

	case OP_TRACE_RETURN:
	  sc->trace_depth--;
	  if (sc->trace_depth < 0) sc->trace_depth = 0;
	  break;

	  /* call/cc does not close files, but I think call-with-exit should */
	case OP_UNWIND_OUTPUT:
	  {
	    s7_pointer x;
	    x = stack_code(sc->stack, i);                /* "code" = port that we opened */
	    s7_close_output_port(sc, x);
	    x = stack_args(sc->stack, i);                /* "args" = port that we shadowed, if not #f */
	    if (x != sc->F)
	      sc->output_port = x;
	  }
	  break;

	case OP_UNWIND_INPUT:
	  s7_close_input_port(sc, stack_code(sc->stack, i)); /* "code" = port that we opened */
	  sc->input_port = stack_args(sc->stack, i);         /* "args" = port that we shadowed */
	  break;

	default:
	  break;
	}
    }
	    
  sc->stack_end = (s7_pointer *)(sc->stack_start + new_stack_top);

  /* the return value should have an implicit values call, just as in call/cc */
  if (is_null(sc->args))
    sc->value = sc->NIL;
  else
    {
      if (is_null(cdr(sc->args)))
	sc->value = car(sc->args);
      else sc->value = splice_in_values(sc, sc->args);
    }
}


static s7_pointer g_call_cc(s7_scheme *sc, s7_pointer args)
{
  #define H_call_cc "(call-with-current-continuation func) needs more than a one sentence explanation"
  s7_pointer proc_args;

  /* car(args) is the procedure passed to call/cc */
  
  if (!is_procedure(car(args)))                      /* this includes continuations */
    return(s7_wrong_type_arg_error(sc, "call/cc", 0, car(args), "a procedure"));

  proc_args = s7_procedure_arity(sc, car(args));
  if ((s7_integer(car(proc_args)) > 1) ||
      ((s7_integer(car(proc_args)) == 0) &&
       (s7_integer(cadr(proc_args)) == 0) &&
       (caddr(proc_args) == sc->F)))
    return(s7_error(sc, sc->WRONG_TYPE_ARG, 
		    list_2(sc, make_protected_string(sc, "call/cc procedure, ~A, should take one argument"), car(args))));

  sc->w = s7_make_continuation(sc);
  push_stack(sc, OP_APPLY, list_1(sc, sc->w), car(args));
  sc->w = sc->NIL;
  return(sc->NIL);
}


static s7_pointer g_call_with_exit(s7_scheme *sc, s7_pointer args)
{
  #define H_call_with_exit "(call-with-exit func) is call/cc without the ability to jump back into a previous computation."
  s7_pointer x;
  
  /* (call-with-exit (lambda (return) ...)) */
  /* perhaps "call/exit"? */
  
  if (!is_procedure(car(args)))                              /* this includes continuations */
    return(s7_wrong_type_arg_error(sc, "call-with-exit", 0, car(args), "a procedure"));

  x = make_goto(sc);
  push_stack(sc, OP_DEACTIVATE_GOTO, x, sc->NIL); /* this means call-with-exit is not tail-recursive */
  push_stack(sc, OP_APPLY, cons_unchecked(sc, x, sc->NIL), car(args));

  /* fprintf(stderr, "call-with-exit %s %s\n", DISPLAY_80(sc->code), DISPLAY_80(sc->args)); */
  
  /* if the lambda body calls the argument as a function, 
   *   it is applied to its arguments, apply notices that it is a goto, and...
   *   
   *      (conceptually...) sc->stack_top = call_exit_goto_loc(sc->code);      
   *      s_pop(sc, (is_not_null(sc->args)) ? car(sc->args) : sc->NIL);
   * 
   *   which jumps to the point of the goto returning car(args).
   *
   * There is one gotcha: we can't jump back in from outside, so if the caller saves the goto
   *   and tries to invoke it outside the call-with-exit block, we have to
   *   make sure it triggers an error.  So, if the escape is called, it then
   *   deactivates itself.  Otherwise the block returns, we pop to OP_DEACTIVATE_GOTO,
   *   and it finds the goto in sc->args.
   * Even worse:
   *
       (let ((cc #f))
         (call-with-exit
           (lambda (c3)
             (call/cc (lambda (ret) (set! cc ret)))
             (c3)))
         (cc))
   *
   * where we jump back into a call-with-exit body via call/cc, the goto has to be
   * re-established.
   */
  
  return(sc->NIL);
}



/* -------------------------------- numbers -------------------------------- */

#if WITH_GMP
  static char *big_number_to_string_with_radix(s7_pointer p, int radix, int width);
  static bool big_numbers_are_eqv(s7_pointer a, s7_pointer b);
  static s7_pointer string_to_either_integer(s7_scheme *sc, const char *str, int radix);
  static s7_pointer string_to_either_ratio(s7_scheme *sc, const char *nstr, const char *dstr, int radix);
  static s7_pointer string_to_either_real(s7_scheme *sc, const char *str, int radix);
  static s7_pointer string_to_either_complex(s7_scheme *sc,
					     char *q, char *slash1, char *ex1, bool has_dec_point1, 
					     char *plus, char *slash2, char *ex2, bool has_dec_point2, 
					     int radix, int has_plus_or_minus);
  static s7_pointer big_add(s7_scheme *sc, s7_pointer args);
  static s7_pointer big_subtract(s7_scheme *sc, s7_pointer args);
  static s7_pointer big_multiply(s7_scheme *sc, s7_pointer args);
  static s7_pointer big_divide(s7_scheme *sc, s7_pointer args);
  static s7_pointer big_random(s7_scheme *sc, s7_pointer args);
  static s7_pointer s7_Int_to_big_integer(s7_scheme *sc, s7_Int val);
  static s7_pointer s7_ratio_to_big_ratio(s7_scheme *sc, s7_Int num, s7_Int den);
  static s7_pointer s7_number_to_big_real(s7_scheme *sc, s7_pointer p);
  static s7_pointer promote_number(s7_scheme *sc, int type, s7_pointer x);
  static s7_pointer big_negate(s7_scheme *sc, s7_pointer args);
  static s7_pointer big_invert(s7_scheme *sc, s7_pointer args);
  static s7_pointer big_inexact_to_exact(s7_scheme *sc, s7_pointer args);
  static s7_pointer big_exact_to_inexact(s7_scheme *sc, s7_pointer args);
  static s7_pointer big_is_zero_1(s7_scheme *sc, s7_pointer x);
  static s7_pointer mpz_to_big_integer(s7_scheme *sc, mpz_t val);
  static s7_pointer mpq_to_big_ratio(s7_scheme *sc, mpq_t val);
  static s7_pointer mpfr_to_big_real(s7_scheme *sc, mpfr_t val);
  static s7_pointer mpc_to_big_complex(s7_scheme *sc, mpc_t val);
#endif


#define s7_Int_abs(x) ((x) >= 0 ? (x) : -(x))
/* can't use abs even in gcc -- it doesn't work with long long ints! */
#define s7_Double_abs(x) fabs(x)
#define s7_fabsl(x) (((x) < 0.0) ? -(x) : (x))
/* fabsl doesn't exist in netBSD! */

static bool is_NaN(s7_Double x) {return(x != x);}
/* callgrind says this is faster than isnan, I think (very confusing data...) */

#ifdef _MSC_VER
/* need to provide inverse hyperbolic trig funcs and cbrt */

double asinh(double x);
double asinh(double x) 
{ 
  return(log(x + sqrt(1.0 + x * x))); 
} 


double acosh(double x);
double acosh(double x)
{ 
  return(log(x + sqrt(x * x - 1.0))); 
  /* perhaps less prone to numerical troubles (untested):
   *   2.0 * log(sqrt(0.5 * (x + 1.0)) + sqrt(0.5 * (x - 1.0)))
   */
} 


double atanh(double x);
double atanh(double x)
{ 
  return(log((1.0 + x) / (1.0 - x)) / 2.0); 
} 


double cbrt(double x);
double cbrt(double x)
{
  if (x >= 0.0)
    return(pow(x, 1.0 / 3.0));
  return(-pow(-x, 1.0 / 3.0));
}

static bool isinf(s7_Double x) {return((x == x) && (is_NaN(x - x)));}

#endif

#if WITH_COMPLEX

#if __cplusplus
  #define _Complex_I (complex<s7_Double>(0.0, 1.0))
  #define creal(x) Real(x)
  #define cimag(x) Imag(x)
  #define carg(x) arg(x)
  #define cabs(x) abs(x)
  #define csqrt(x) sqrt(x)
  #define cpow(x, y) pow(x, y)
  #define clog(x) log(x)
  #define cexp(x) exp(x)
  #define csin(x) sin(x)
  #define ccos(x) cos(x)
  #define csinh(x) sinh(x)
  #define ccosh(x) cosh(x)
#else
  typedef double complex s7_Complex;
#endif

/* Trigonometric functions. 
 *    FreeBSD's math library does not include the complex form of the trig funcs.
 *    FreeBSD supplies cabs carg cimag creal conj csqrt, so can we assume those exist if complex.h exists?
 */

#if 0
static s7_Double carg(s7_Complex z)
{ 
  return(atan2(cimag(z), creal(z))); 
} 


static s7_Double cabs(s7_Complex z) 
{ 
  return(hypot(creal(z), cimag(z))); 
} 


static s7_Complex conj(s7_Complex z) 
{ 
  return(~z); 
} 


static s7_Complex csqrt(s7_Complex z) 
{ 
  if (cimag(z) < 0.0) 
    return(conj(csqrt(conj(z)))); 
  else 
    { 
      s7_Double r = cabs(z); 
      s7_Double x = creal(z); 
      
      return(sqrt((r + x) / 2.0) + sqrt((r - x) / 2.0) * _Complex_I); 
    } 
} 
#endif


#if (!HAVE_COMPLEX_TRIG)

#if (!__cplusplus)
static s7_Complex csin(s7_Complex z) 
{ 
  return(sin(creal(z)) * cosh(cimag(z)) + (cos(creal(z)) * sinh(cimag(z))) * _Complex_I); 
} 


static s7_Complex ccos(s7_Complex z) 
{ 
  return(cos(creal(z)) * cosh(cimag(z)) + (-sin(creal(z)) * sinh(cimag(z))) * _Complex_I); 
} 


static s7_Complex csinh(s7_Complex z) 
{ 
  return(sinh(creal(z)) * cos(cimag(z)) + (cosh(creal(z)) * sin(cimag(z))) * _Complex_I); 
} 


static s7_Complex ccosh(s7_Complex z) 
{ 
  return(cosh(creal(z)) * cos(cimag(z)) + (sinh(creal(z)) * sin(cimag(z))) * _Complex_I); 
} 
#endif


static s7_Complex ctan(s7_Complex z) 
{ 
  return(csin(z) / ccos(z)); 
} 


static s7_Complex ctanh(s7_Complex z) 
{ 
  return(csinh(z) / ccosh(z)); 
} 


#if (!__cplusplus)
static s7_Complex cexp(s7_Complex z) 
{ 
  return(exp(creal(z)) * cos(cimag(z)) + (exp(creal(z)) * sin(cimag(z))) * _Complex_I); 
} 


static s7_Complex clog(s7_Complex z) 
{ 
  return(log(s7_Double_abs(cabs(z))) + carg(z) * _Complex_I); 
} 


static s7_Complex cpow(s7_Complex x, s7_Complex y) 
{ 
  s7_Double r = cabs(x); 
  s7_Double theta = carg(x); 
  s7_Double yre = creal(y); 
  s7_Double yim = cimag(y); 
  s7_Double nr = exp(yre * log(r) - yim * theta); 
  s7_Double ntheta = yre * theta + yim * log(r); 
  
  return(nr * cos(ntheta) + (nr * sin(ntheta)) * _Complex_I); /* make-polar */ 
} 
#endif


static s7_Complex casin(s7_Complex z) 
{ 
  return(-_Complex_I * clog(_Complex_I * z + csqrt(1.0 - z * z))); 
} 


static s7_Complex cacos(s7_Complex z) 
{ 
  return(-_Complex_I * clog(z + _Complex_I * csqrt(1.0 - z * z))); 
} 


static s7_Complex catan(s7_Complex z) 
{ 
  return(_Complex_I * clog((_Complex_I + z) / (_Complex_I - z)) / 2.0); 
} 


static s7_Complex casinh(s7_Complex z) 
{ 
  return(clog(z + csqrt(1.0 + z * z))); 
} 


static s7_Complex cacosh(s7_Complex z) 
{ 
  return(clog(z + csqrt(z * z - 1.0))); 
  /* perhaps less prone to numerical troubles (untested):
   *   2.0 * clog(csqrt(0.5 * (z + 1.0)) + csqrt(0.5 * (z - 1.0)))
   */
} 


static s7_Complex catanh(s7_Complex z) 
{ 
  return(clog((1.0 + z) / (1.0 - z)) / 2.0); 
} 
#endif

#else
/* not WITH_COMPLEX */
  typedef double s7_Complex;
  #define _Complex_I 1
  #define creal(x) x
  #define cimag(x) x
  #define csin(x) sin(x)
  #define casin(x) x
  #define ccos(x) cos(x)
  #define cacos(x) x
  #define ctan(x) x
  #define catan(x) x
  #define csinh(x) x
  #define casinh(x) x
  #define ccosh(x) x
  #define cacosh(x) x
  #define ctanh(x) x
  #define catanh(x) x
  #define cexp(x) exp(x)
  #define cpow(x, y) pow(x, y)
  #define clog(x) log(x)
  #define csqrt(x) sqrt(x)
  #define conj(x) x
#endif


#if (!WITH_GMP)

bool s7_is_number(s7_pointer p)
{
  return(is_number(p));
}


bool s7_is_integer(s7_pointer p) 
{ 
  return(is_integer(p));
}


bool s7_is_real(s7_pointer p) 
{ 
  return(is_real(p));
}


bool s7_is_rational(s7_pointer p)
{
  return(is_rational(p));
}


bool s7_is_ratio(s7_pointer p)
{
  return(type(p) == T_RATIO);
}


bool s7_is_complex(s7_pointer p)
{
  return(is_number(p));
}

#endif 
/* !WITH_GMP */


bool s7_is_exact(s7_pointer p)
{
  return(s7_is_rational(p));
}


bool s7_is_inexact(s7_pointer p)
{
  return(!s7_is_rational(p));
}


static s7_Int c_gcd(s7_Int u, s7_Int v)
{
  s7_Int a, b, temp;
  
  a = s7_Int_abs(u);  /* trouble if either is most-negative-fixnum... */
  b = s7_Int_abs(v);
  while (b != 0)
    {
      temp = a % b;
      a = b;
      b = temp;
    }
  if (a < 0)
    return(-a);
  return(a);
}


static bool c_rationalize(s7_Double ux, s7_Double error, s7_Int *numer, s7_Int *denom)
{
  /*
    (define* (rat ux (err 0.0000001))
      ;; translated from CL code in Canny, Donald, Ressler, "A Rational Rotation Method for Robust Geometric Algorithms"
      (let ((x0 (- ux error))
	    (x1 (+ ux error)))
        (let ((i (ceiling x0))
	      (i0 (floor x0))
	      (i1 (ceiling x1))
	      (r 0))
          (if (>= x1 i)
	      i
	      (do ((p0 i0 (+ p1 (* r p0)))
	           (q0 1 (+ q1 (* r q0)))
	           (p1 i1 p0)
	           (q1 1 q0)
	           (e0 (- i1 x0) e1p)
	           (e1 (- x0 i0) (- e0p (* r e1p)))
	           (e0p (- i1 x1) e1)
	           (e1p (- x1 i0) (- e0 (* r e1))))
	          ((<= x0 (/ p0 q0) x1)
	           (/ p0 q0))
	        (set! r (min (floor (/ e0 e1))
			     (ceiling (/ e0p e1p)))))))))
  */
  
  double x0, x1, val;
  s7_Int i, i0, i1, r, r1, p0, q0, p1, q1;
  double e0, e1, e0p, e1p;
  s7_Int old_p1, old_q1;
  double old_e0, old_e1, old_e0p;
  int tries = 0;
  /* don't use s7_Double here;  if it is "long double", the loop below will hang */

  /* #e1e19 is a killer -- it's bigger than most-positive-fixnum, but if we ceil(ux) below
   *   it turns into most-negative-fixnum.  1e19 is trouble in many places.
   */
  if ((ux > s7_Int_max) || (ux < s7_Int_min))
    {
      /* can't return false here because that confuses some of the callers!
       */
      if (ux > s7_Int_min) (*numer) = s7_Int_max; else (*numer) = s7_Int_min;
      (*denom) = 1;
      return(true);
    }

  if (error < 0.0) error = -error;
  x0 = ux - error;
  x1 = ux + error;
  i = (s7_Int)ceil(x0);
  
  if (error >= 1.0) /* aw good grief! */
    {
      if (x0 < 0)
	{
	  if (x1 < 0)
	    (*numer) = (s7_Int)floor(x1);
	  else (*numer) = 0;
	}
      else
	{
	  (*numer) = i;
	}
      (*denom) = 1;
      return(true);
    }
  
  if (x1 >= i)
    {
      if (i >= 0)
	(*numer) = i;
      else (*numer) = (s7_Int)floor(x1);
      (*denom) = 1;
      return(true);
    }

  i0 = (s7_Int)floor(x0);
  i1 = (s7_Int)ceil(x1);

  p0 = i0; 
  q0 = 1;
  p1 = i1; 
  q1 = 1; 
  e0 = i1 - x0;
  e1 = x0 - i0;
  e0p = i1 - x1;
  e1p = x1 - i0;

  while (true)
    {
      val = (double)p0 / (double)q0;
      
      if (((x0 <= val) && (val <= x1)) ||
	  (e1 == 0) ||
	  (e1p == 0) ||
	  (tries > 100))
	{
	  (*numer) = p0;
	  (*denom) = q0;
	  return(true);
	}
      tries++;

      r = (s7_Int)floor(e0 / e1);
      r1 = (s7_Int)ceil(e0p / e1p);
      if (r1 < r) r = r1;

      /* Scheme "do" handles all step vars in parallel */
      old_p1 = p1;
      p1 = p0;
      old_q1 = q1;
      q1 = q0;
      old_e0 = e0;
      e0 = e1p;
      old_e0p = e0p;
      e0p = e1;
      old_e1 = e1;

      p0 = old_p1 + r * p0;
      q0 = old_q1 + r * q0;
      e1 = old_e0p - r * e1p;
      /* if the error is set too low, we can get e1 = 0 here: (rationalize (/ pi) 1e-17) */
      e1p = old_e0 - r * old_e1;
    }
  return(false);
}

#if 0
/* there is another way to rationalize.  Here is a scheme version of
 *   Bill Gosper's farint:

(define* (farint x (err 1/1000000))
  
  (define* (farint-1 x nhi dhi (ln 0) (ld 1) (hn 1) (hd 0))
    (if (> (+ ln hn) (* (+ ld hd) x))
	(let* ((m (min (if (= 0 ln) 
			   nhi 
			 (floor (/ (- nhi hn) ln)))
		       (floor (/ (- dhi hd) ld))))
	       (d (- (* x ld) ln))
	       (k (if (= 0 d) 
		      m 
		    (ceiling (/ (- hn (* x hd)) d)))))
	  (if (< k m)
	      (let ((hn1 (+ (* k ln) hn))
		    (hd1 (+ (* k ld) hd)))
		(farint-1 x nhi dhi hn1 hd1 (- hn1 ln) (- hd1 ld)))

	    (let* ((n (+ (* m ln) hn)) (d (+ (* m ld) hd)))
	      (if (< (* 2 d ld x) (+ (* ld n) (* ln d)))
		  (/ ln ld) 
		(/ n d)))))

      (let* ((m (min (floor (/ (- nhi ln) hn))
		     (if (= 0 hd) 
			 dhi 
		       (floor (/ (- dhi ld) hd)))))
	     (d (- hn (* x hd)))
	     (k (if (= 0 d) 
		    m 
		  (ceiling (/ (- (* x ld) ln) d)))))
	(if (< k m)
	    (let ((ln1 (+ (* k hn) ln))
		  (ld1 (+ (* k hd) ld)))
	    (farint-1 x nhi dhi (- ln1 hn) (- ld1 hd) ln1 ld1))
	  (let* ((n (+ (* m hn) ln)) (d (+ (* m hd) ld)))
	    (if (< (* 2 d hd x) (+ (* hd n) (* hn d)))
		(/ n d) 
	      (/ hn hd)))))))

  (farint-1 x (/ err) (/ err)))
*/
#endif


s7_pointer s7_rationalize(s7_scheme *sc, s7_Double x, s7_Double error)
{
  s7_Int numer = 0, denom = 1;
  if (c_rationalize(x, error, &numer, &denom))
    return(s7_make_ratio(sc, numer, denom));
  return(s7_make_real(sc, x));
}



#if (!WITH_GCC)
static s7_Double number_to_real(s7_pointer n)
{
  if (type(n) >= T_REAL)
    return(real(n));
  if (type(n) == T_INTEGER)
    return((s7_Double)integer(n));
  return(fraction(n));
}
#else
#define number_to_real(_n_) ({ s7_pointer N; N = _n_; ((type(N) >= T_REAL) ? real(N) : ((type(N) == T_INTEGER) ? ((s7_Double)integer(N)) : (s7_Double)(fraction(N)))); })
#endif


#if WITH_OPTIMIZATION
static s7_Int number_to_numerator(s7_pointer n)
{
  if (type(n) == T_RATIO)
    return(numerator(n));
  return(integer(n));
}


static s7_Int number_to_denominator(s7_pointer n)
{
  if (type(n) == T_RATIO)
    return(denominator(n));
  return(1);
}
#endif


#if (!WITH_GMP)
static s7_Double number_to_real_part(s7_pointer n)
{
  /* no bignum parallel */
  switch (type(n))
    {
    case T_INTEGER: return((s7_Double)integer(n));
    case T_RATIO:   return(fraction(n));
    case T_REAL:    return(real(n));
    default:        return(real_part(n));
    }
}


static s7_Double number_to_imag_part(s7_pointer n)
{
  if (type(n) >= T_COMPLEX)
    return(imag_part(n));
  return(0.0);
}
#endif


s7_pointer s7_make_integer(s7_scheme *sc, s7_Int n) 
{
  s7_pointer x;

  if ((n & ~(NUM_SMALL_INTS - 1)) == 0)            /* ((n >= 0) && (n < NUM_SMALL_INTS)) is slower */
    return(small_int(n));

  NEW_CELL(sc, x);
  set_type(x, T_INTEGER);
  integer(x) = n;

  return(x);
}


#if WITH_GCC
static s7_pointer make_integer_1(s7_scheme *sc, s7_Int n)
{
  s7_pointer x;
  NEW_CELL(sc, x);
  set_type(x, T_INTEGER);
  integer(x) = n;
  return(x);
}

#define make_integer(Sc, N) ({ s7_Int _N_; _N_ = N; (((_N_ & ~(NUM_SMALL_INTS - 1)) == 0) ? small_int(_N_) : make_integer_1(Sc, _N_)); })
#else
#define make_integer(Sc, N) s7_make_integer(Sc, N)
#endif


static s7_pointer make_mutable_integer(s7_scheme *sc, s7_Int n)
{
  s7_pointer x;
  NEW_CELL(sc, x);
  set_type(x, T_INTEGER);
  integer(x) = n;
  return(x);
}


static s7_pointer make_permanent_integer(s7_Int i)
{
  s7_pointer p;
  p = (s7_pointer)calloc(1, sizeof(s7_cell));
  typeflag(p) = T_IMMUTABLE | T_INTEGER;
  p->hloc = NOT_IN_HEAP;
  integer(p) = i;
  return(p);
}


s7_pointer s7_make_real(s7_scheme *sc, s7_Double n) 
{
  s7_pointer x;

  if (n == 0.0)
    return(real_zero);

  NEW_CELL(sc, x);
  set_type(x, T_REAL);
  real(x) = n;
  
  return(x);
}


static s7_pointer make_permanent_real(s7_Double n) 
{
  s7_pointer x;
  x = (s7_pointer)calloc(1, sizeof(s7_cell));
  set_type(x, T_IMMUTABLE | T_REAL);
  x->hloc = NOT_IN_HEAP;
  real(x) = n;
  return(x);
}


s7_pointer s7_remake_real(s7_scheme *sc, s7_pointer rl, s7_Double n) 
{
  if ((rl != real_zero) &&
      (type(rl) > T_RATIO))
    real(rl) = n;
  else rl = s7_make_real(sc, n);
  return(rl);
}


s7_pointer s7_make_complex(s7_scheme *sc, s7_Double a, s7_Double b)
{
  s7_pointer x;
  NEW_CELL(sc, x);
  if (b == 0.0)
    {
      set_type(x, T_REAL);
      real(x) = a;
    }
  else
    {
      set_type(x, T_COMPLEX);
      real_part(x) = a;
      imag_part(x) = b;
    }
  return(x);
}


s7_pointer s7_make_ratio(s7_scheme *sc, s7_Int a, s7_Int b)
{
  s7_pointer x;
  s7_Int divisor;

  if (b == 0)
    return(division_by_zero_error(sc, "make-ratio", list_2(sc, make_integer(sc, a), small_int(0))));
  if (a == 0)
    return(small_int(0));
  if (b == 1)
    return(make_integer(sc, a));

#if (!WITH_GMP)
  if (b == S7_LLONG_MIN)
    {
      if (a == b)
	return(small_int(1));

      /* we've got a problem... This should not trigger an error during reading -- we might have the
       *   ratio on a switch with-bignums or whatever, so its mere occurrence is just an annoyance.
       *   We'll try to do something...
       */
      if (a & 1)
	{
	  if (a == 1)
	    return(real_NaN);
	  /* not an error here because 1/9223372036854775808 might be in a block of unevaluated code */
	  b = b + 1;
	}
      else
	{
	  a /= 2;
	  b /= 2;
	}
    }
#endif

  if (b < 0)
    {
      a = -a;
      b = -b;
    }
  divisor = c_gcd(a, b);
  if (divisor != 1)
    {
      a /= divisor;
      b /= divisor;
    }
  if (b == 1)
    return(make_integer(sc, a));
  
  NEW_CELL(sc, x);
  set_type(x, T_RATIO);
  numerator(x) = a;
  denominator(x) = b;

  return(x);
}



static s7_pointer exact_to_inexact(s7_scheme *sc, s7_pointer x)
{
  /* this is tricky because a big int can mess up when turned into a double:
   *   (truncate (exact->inexact most-positive-fixnum)) -> -9223372036854775808
   */
  switch (type(x))
    {
    case T_INTEGER:
      return(s7_make_real(sc, (s7_Double)(integer(x))));

    case T_RATIO:
      return(s7_make_real(sc, (s7_Double)(fraction(x))));

    case T_REAL:
    case T_COMPLEX:              /* apparently (exact->inexact 1+i) is not an error */
      return(x);

    default:
      return(s7_wrong_type_arg_error(sc, "exact->inexact", 0, x, "a number"));
    }
}



static double default_rationalize_error = 1.0e-12;

static s7_pointer inexact_to_exact(s7_scheme *sc, s7_pointer x)
{
  switch (type(x))
    {
    case T_INTEGER:
    case T_RATIO:
      return(x);

    case T_REAL:
      {
	s7_Int numer = 0, denom = 1;
	s7_Double val;

	val = s7_real(x);
	if ((isinf(val)) || (is_NaN(val)))
	  return(s7_wrong_type_arg_error(sc, "inexact->exact", 0, x, "a normal real"));

	if ((val > S7_LLONG_MAX) ||
	    (val < S7_LLONG_MIN))
	  return(s7_out_of_range_error(sc, "inexact->exact", 0, x, "too large to express as an integer"));

	if (c_rationalize(val, default_rationalize_error, &numer, &denom))
	  return(s7_make_ratio(sc, numer, denom));
      }

    default:
      return(s7_wrong_type_arg_error(sc, "inexact->exact", 0, x, "a real"));
    }

  return(x);
}


#if (!WITH_GMP)
s7_Double s7_number_to_real(s7_pointer x)
{
  switch (type(x))
    {
    case T_INTEGER: return((s7_Double)integer(x));
    case T_RATIO:   return((s7_Double)numerator(x) / (s7_Double)denominator(x));
    case T_REAL:    return(real(x));
    case T_COMPLEX: return(real_part(x));
    default:        return(0.0);
      /* what to do with a non-number? -- to return #f or throw an error, we need the s7_scheme pointer
       *   some sort of check is needed for FFI calls -- not a number -> segfault
       */
    }
}


s7_Int s7_number_to_integer(s7_pointer x)
{
  switch (type(x))
    {
    case T_INTEGER: return(integer(x));
    case T_RATIO:   return((s7_Int)((s7_Double)numerator(x) / (s7_Double)denominator(x)));
    case T_REAL:    return((s7_Int)real(x));
    case T_COMPLEX: return((s7_Int)real_part(x));
    default:        return(0);
    }
}


s7_Int s7_numerator(s7_pointer x)
{
  if (type(x) == T_RATIO)
    return(numerator(x));
  return(integer(x));
}


s7_Int s7_denominator(s7_pointer x)
{
  if (type(x) == T_RATIO)
    return(denominator(x));
  return(1);
}


s7_Double s7_real_part(s7_pointer x)
{
  return(number_to_real_part(x));
}


s7_Double s7_imag_part(s7_pointer x)
{
  return(number_to_imag_part(x));
}


s7_Int s7_integer(s7_pointer p)
{
  return(integer(p));
}


s7_Double s7_real(s7_pointer p)
{
  return(real(p));
}


static s7_Complex s7_complex(s7_pointer p)
{
  return(number_to_real_part(p) + number_to_imag_part(p) * _Complex_I);
}


static s7_pointer s7_from_c_complex(s7_scheme *sc, s7_Complex z)
{
  return(s7_make_complex(sc, creal(z), cimag(z)));
}
#endif


static int integer_length(s7_Int a)
{
  static int bits[256] =
    {0, 1, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 
     6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 
     7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8};

  #define I_8 256LL
  #define I_16 65536LL
  #define I_24 16777216LL
  #define I_32 4294967296LL
  #define I_40 1099511627776LL
  #define I_48 281474976710656LL
  #define I_56 72057594037927936LL

  if (a < 0) a = -a;
  if (a < I_8) return(bits[a]);
  if (a < I_16) return(8 + bits[a >> 8]);
  if (a < I_24) return(16 + bits[a >> 16]);
  if (a < I_32) return(24 + bits[a >> 24]);
  if (a < I_40) return(32 + bits[a >> 32]);
  if (a < I_48) return(40 + bits[a >> 40]);
  if (a < I_56) return(48 + bits[a >> 48]);
  return(56 + bits[a >> 56]);
}


static int s7_int_max = 0, s7_int_min = 0, s7_int_bits = 0, s7_int_digits = 0; /* initialized later */
static int s7_int_digits_by_radix[17];


#if (!WITH_GMP)
static s7_pointer s7_negate(s7_scheme *sc, s7_pointer p)     /* can't use "negate" because it confuses C++! */
{
  switch (type(p))
    {
    case T_INTEGER: 
      return(make_integer(sc, -integer(p)));
      
    case T_RATIO:
      return(s7_make_ratio(sc, -numerator(p), denominator(p)));
      
    case T_REAL:
      return(s7_make_real(sc, -real(p)));
      
    default:
      return(s7_make_complex(sc, -real_part(p), -imag_part(p)));
    }
}
#endif


static s7_pointer s7_invert(s7_scheme *sc, s7_pointer p)      /* s7_ to be consistent... */
{
  switch (type(p))
    {
    case T_INTEGER:
      return(s7_make_ratio(sc, 1, integer(p)));      /* a already checked, not 0 */
      
    case T_RATIO:
      return(s7_make_ratio(sc, denominator(p), numerator(p)));

    case T_REAL:
      return(s7_make_real(sc, 1.0 / real(p)));

    default:
      {
	s7_Double r2, i2, den;
	r2 = real_part(p);
	i2 = imag_part(p);
	den = (r2 * r2 + i2 * i2);
	return(s7_make_complex(sc, r2 / den, -i2 / den));
      }
    }
}


#if WITH_OPTIMIZATION
static s7_pointer subtract_ratios(s7_scheme *sc, s7_pointer x, s7_pointer y)
{
  s7_Int d1, d2, n1, n2;
  d1 = number_to_denominator(x);
  n1 = number_to_numerator(x);
  d2 = number_to_denominator(y);
  n2 = number_to_numerator(y);
  
  if (d1 == d2)                                     /* the easy case -- if overflow here, it matches the int case */
    return(s7_make_ratio(sc, n1 - n2, d1));

#if (!WITH_GMP)
  if ((d1 > s7_int_max) || (d2 > s7_int_max) ||     /* before counting bits, check that overflow is possible */
      (n1 > s7_int_max) || (n2 > s7_int_max) ||
      (n1 < s7_int_min) || (n2 < s7_int_min))
    {
      int d1bits, d2bits;
      d1bits = integer_length(d1);
      d2bits = integer_length(d2);
      if (((d1bits + d2bits) > s7_int_bits) ||
	  ((d1bits + integer_length(n2)) > (s7_int_bits - 1)) ||
	  ((d2bits + integer_length(n1)) > (s7_int_bits - 1)))
	return(s7_make_real(sc, ((long double)n1 / (long double)d1) - ((long double)n2 / (long double)d2)));
      return(s7_make_ratio(sc, n1 * d2 - n2 * d1, d1 * d2));
    }
#endif
  return(s7_make_ratio(sc, n1 * d2 - n2 * d1, d1 * d2));
}
#endif


#if (!WITH_GMP)
static bool s7_is_negative(s7_pointer obj)
{
  switch (type(obj))
    {
    case T_INTEGER: return(integer(obj) < 0);
    case T_RATIO:   return(numerator(obj) < 0);
    default:        return(real(obj) < 0);
    }
}


static bool s7_is_positive(s7_pointer x)
{
  switch (type(x))
    {
    case T_INTEGER: return(integer(x) > 0);
    case T_RATIO:   return(numerator(x) > 0);
    default:        return(real(x) > 0.0);
    }
}
#endif


static bool s7_is_zero(s7_pointer x)
{
  switch (type(x))
    {
    case T_INTEGER: return(integer(x) == 0);
    case T_REAL:    return(real(x) == 0.0);
    default:        return(false); /* ratios and complex numbers here are already collapsed into integers and reals */
    }
}


static bool s7_is_one(s7_pointer x)
  {
  switch (type(x))
    {
    case T_INTEGER: return(integer(x) == 1);
    case T_REAL:    return(real(x) == 1.0);
    default:        return(false);
    }
}


/* optimize exponents */
#define MAX_POW 32
static double pepow[17][MAX_POW], mepow[17][MAX_POW];

static void init_pows(void)
{
  int i, j;
  for (i = 2; i < 17; i++)        /* radix between 2 and 16 */
    for (j = 0; j < MAX_POW; j++) /* saved exponent between 0 and +/- MAX_POW */
      {
	pepow[i][j] = pow((double)i, (double)j);
	mepow[i][j] = pow((double)i, (double)(-j));
      }
}

static double ipow(int x, int y)
{
  if ((y < MAX_POW) && (y > (-MAX_POW)))
    {
      if (y >= 0)
	return(pepow[x][y]);
      return(mepow[x][-y]);
    }
  return(pow((double)x, (double)y));
}


static void s7_Int_to_string(char *p, s7_Int n, int radix, int width)
{
  static char dignum[] = "0123456789abcdef";
  int i = 2, len, start = 0, end = 0;
  s7_Int pown = (s7_Int)1;
  bool sign;

  if ((radix < 2) || (radix > 16))
    return;
  if (n == 0)
    {
      if (width <= 1)
	{
	  p[0] = '0';
	  p[1] = '\0';
	}
      else
	{
	  for (i = 0; i < width - 1; i++) 
	    p[i] = ' ';
	  p[width - 1] = '0';
	  p[width] = '\0';
	}
      return;
    }

  if (n == S7_LLONG_MIN)
    {
      /* a special case -- we can't use abs on this because it goes to 0, we won't get here if gmp.
       * (number->string most-negative-fixnum 2) -> "-0" unless we do something special 
       */
      int j;
      p[0] = '-';
      /* build it backwards (will reverse digits below) */
      p[1] = dignum[-(n % (s7_Int)radix)];
      n /= (s7_Int)radix;
      n = -n;
      for (i = 2; n >= (s7_Int)radix; i++)
	{
	  p[i] = dignum[n % (s7_Int)radix];
	  n /= (s7_Int)radix;
	}
      p[i] = dignum[n];
      len = i;
      /* reverse digits (leave sign alone) */
      for (i = 1, j = len; i < j; i++, j--)
	{
	  char tmp;
	  tmp = p[i];
	  p[i] = p[j];
	  p[j] = tmp;
	}
      p[len + 1] = 0;
      return;
      /* there has to be a better way... */
    }
      
  sign = (n < 0);
  n = s7_Int_abs(n); 

  /* the previous version that counted up to n, rather than dividing down below n, as here,
   *   could be confused by large ints on 64 bit machines
   */
  pown = n;
  for (i = 1; i < 100; i++)
    {
      if (pown < radix)
	break;
      pown /= (s7_Int)radix;
    }
  len = i - 1;

  if (sign) len++;

  if (width > len)                  /* (format #f "~10B" 123) */
    {
      start = width - len - 1;
      end += start;
      memset((void *)p, (int)' ', start);
      /*
      for (i = 0; i < start; i++) 
	p[i] = ' ';
      */
    }

  if (sign)
    {
      p[start] = '-';
      end++;
    }

  for (i = start + len; i >= end; i--)
    {
      p[i] = dignum[n % (s7_Int)radix];
      n /= (s7_Int)radix;
    }
  p[len + start + 1] = '\0';
}


static char *pad_number(const char *p, int len, int width)
{
  char *p1;
  int spaces;
  spaces = width - len;
  p1 = (char *)malloc((width + 1) * sizeof(char));
  p1[width] = '\0';
  memset((void *)p1, (int)' ', spaces);
  memcpy((void *)(p1 + spaces), (void *)p, len);
  return(p1);
}


#define BASE_10 10

static char *number_to_string_base_10(s7_pointer obj, int width, int precision, char float_choice)
{
  char *p;
  int len;

#if WITH_GMP
  if (s7_is_bignum(obj))
    return(big_number_to_string_with_radix(obj, BASE_10, width));
  /* this ignores precision because it's way too hard to get the mpfr string to look like
   *   C's output -- we either have to call mpfr_get_str twice (the first time just to 
   *   find out what the exponent is and how long the string actually is), or we have
   *   to do messy string manipulations.  So (format #f "",3F" pi) ignores the "3" and
   *   prints the full string.
   */
#endif

  switch (type(obj))
    {
    case T_INTEGER:
      len = 64 + width;
      p = (char *)malloc(len * sizeof(char));
      snprintf(p, len, 
	       (sizeof(int) >= sizeof(s7_Int)) ? "%*d" : "%*lld",
	       width, integer(obj));
      break;
      
    case T_RATIO:
      p = (char *)malloc(128 * sizeof(char));
      len = snprintf(p, 128,
		     (sizeof(int) >= sizeof(s7_Int)) ? "%d/%d" : "%lld/%lld", 
		     numerator(obj), denominator(obj));
      if (width > len)
	{
	  char *p1;
	  p1 = pad_number(p, len, width);
	  free(p);
	  return(p1);
	}
      break;
      
    case T_REAL:
      {
	int i, loc = -1;
	const char *frmt;
	p = (char *)malloc((256 + width) * sizeof(char));

	if (sizeof(double) >= sizeof(s7_Double))
	  frmt = (float_choice == 'g') ? "%*.*g" : ((float_choice == 'f') ? "%*.*f" : "%*.*e");
	else frmt = (float_choice == 'g') ? "%*.*Lg" : ((float_choice == 'f') ? "%*.*Lf" : "%*.*Le");

	len = snprintf(p, 256 + width, frmt, width, precision, s7_real(obj));
	for (i = 0; i < len; i++) /* does it have an exponent (if so, it's already a float) */
	  if (p[i] == 'e')
	    {
	      loc = i;
	      break;
	    }
	if (loc == -1)            /* no, so make it explicitly a float! */
	  {
	    for (i = 0; i < len; i++)  
	      if (p[i] == '.') break;
	    if (i == len)
	      {
		p[i]='.';
		p[i+1]='0';
		p[i+2]='\0';
	      }
	  }
      }
      break;
      
    default:
      {
	const char *frmt;
	p = (char *)malloc(256 * sizeof(char));

	if (sizeof(double) >= sizeof(s7_Double))
	  {
	    if (imag_part(obj) >= 0.0)
	      frmt = (float_choice == 'g') ? "%.*g+%.*gi" : ((float_choice == 'f') ? "%.*f+%.*fi" : "%.*e+%.*ei"); 
	    else frmt = (float_choice == 'g') ? "%.*g%.*gi" : ((float_choice == 'f') ? "%.*f-%.*fi" :"%.*e-%.*ei");
	  }
	else 
	  {
	    if (imag_part(obj) >= 0.0)
	      frmt = (float_choice == 'g') ? "%.*Lg+%.*Lgi" : ((float_choice == 'f') ? "%.*Lf+%.*Lfi" : "%.*Le+%.*Lei");
	    else frmt = (float_choice == 'g') ? "%.*Lg%.*Lgi" : ((float_choice == 'f') ? "%.*Lf-%.*Lfi" : "%.*Le-%.*Lei");
	  }

	len = snprintf(p, 256, frmt, precision, real_part(obj), precision, imag_part(obj));
	if (width > len)
	  {                             /* (format #f "~20g" 1+i) */
	    char *p1;
	    p1 = pad_number(p, len, width);
	    free(p);
	    return(p1);
	  }
      }
      break;
    }
  return(p);
}


static char *number_to_string_with_radix(s7_scheme *sc, s7_pointer obj, int radix, int width, int precision, char float_choice)
{
  char *p, *n, *d;
  int len;

#if WITH_GMP
  if (s7_is_bignum(obj))
    return(big_number_to_string_with_radix(obj, radix, width));
#endif

  if (radix == 10)
    return(number_to_string_base_10(obj, width, precision, float_choice));

  switch (type(obj))
    {
    case T_INTEGER:
      p = (char *)malloc((128 + width) * sizeof(char));
      s7_Int_to_string(p, s7_integer(obj), radix, width);
      return(p);
      break;
      
    case T_RATIO:
      {
	char n[128], d[128];
	s7_Int_to_string(n, numerator(obj), radix, 0);
	s7_Int_to_string(d, denominator(obj), radix, 0);
	p = (char *)malloc(256 * sizeof(char));
	len = snprintf(p, 256, "%s/%s", n, d);
      }
      break;
      
    case T_REAL:
      {
	int i;
	s7_Int int_part;
	s7_Double x, frac_part, min_frac, base;
	bool sign = false;
	char n[128], d[256];

	x = s7_real(obj);

	if (is_NaN(x))
	  return(copy_string("nan.0"));
	if (isinf(x))
	  {
	    if (x < 0.0)
	      return(copy_string("-inf.0"));    
	    return(copy_string("inf.0"));    
	  }

	if (x < 0.0)
	  {
	    sign = true;
	    x = -x;
	  }

	int_part = (s7_Int)floor(x);
	frac_part = x - int_part;
	s7_Int_to_string(n, int_part, radix, 0);
	min_frac = (s7_Double)ipow(radix, -precision);

	for (i = 0, base = radix; (i < precision) && (frac_part > min_frac); i++, base *= radix)
	  {
	    s7_Int ipart;
	    ipart = (int)(frac_part * base);
	    if (ipart >= radix)         /* rounding confusion */
	      ipart = radix - 1;
	    frac_part -= (ipart / base);
	    if (ipart < 10)
	      d[i] = (char)('0' + ipart);
	    else d[i] = (char)('a' + ipart -  10);
	  }
	if (i == 0)
	  d[i++] = '0';
	d[i] = '\0';
	p = (char *)malloc(256 * sizeof(char));
	len = snprintf(p, 256, "%s%s.%s", (sign) ? "-" : "", n, d);
      }
      break;

    default:
      p = (char *)malloc(512 * sizeof(char));
      n = number_to_string_with_radix(sc, s7_make_real(sc, real_part(obj)), radix, 0, precision, float_choice);
      d = number_to_string_with_radix(sc, s7_make_real(sc, imag_part(obj)), radix, 0, precision, float_choice);
      len = snprintf(p, 512, "%s%s%si", n, (imag_part(obj) < 0.0) ? "" : "+", d);
      free(n);
      free(d);
      break;
    }

  if (width > len)
    {
      char *p1;
      p1 = pad_number(p, len, width);
      free(p);
      return(p1);
    }
  return(p);
}


char *s7_number_to_string(s7_scheme *sc, s7_pointer obj, int radix)
{
  return(number_to_string_with_radix(sc, obj, radix, 0, 20, 'g')); 
  /* (log top 10) so we get all the digits in base 10 (??) */
}


static s7_pointer g_number_to_string(s7_scheme *sc, s7_pointer args)
{
  #define H_number_to_string "(number->string num (radix 10)) converts the number num into a string."
  s7_Int radix = 10;
  int size = 20;
  char *res;
  s7_pointer x;

  x = car(args);
  if (!s7_is_number(x))
    return(s7_wrong_type_arg_error(sc, "number->string", (is_null(cdr(args))) ? 0 : 1, x, "a number"));

  if (is_pair(cdr(args)))
    {
      s7_pointer y;
      y = cadr(args);
      if (s7_is_integer(y))
	radix = s7_integer(y);
      else return(s7_wrong_type_arg_error(sc, "number->string radix,", 2, y, "an integer"));
      if ((radix < 2) || (radix > 16))
	return(s7_out_of_range_error(sc, "number->string radix,", 2, y, "should be between 2 and 16"));
    }

#if WITH_GMP
  if (s7_is_bignum(x))
    return(make_string_uncopied(sc, big_number_to_string_with_radix(x, radix, 0)));
#endif

  if (!is_rational(x))
    {
      /* if size = 20, (number->string .1) gives "0.10000000000000000555", but if it's less than 20,
       *    large numbers (or very small numbers) mess up the less significant digits.
       */
      if (is_real(x))
	{
	  s7_Double val;
	  val = s7_Double_abs(s7_real(x));
	  if ((val < (S7_LONG_MAX / 4)) && (val > 1.0e-6))
	    size = 14;
	}
      else
	{
	  s7_Double rl, im;
	  rl = s7_Double_abs(s7_real_part(x));
	  if ((rl < (S7_LONG_MAX / 4)) && (rl > 1.0e-6))
	    {
	      im = s7_Double_abs(s7_imag_part(x));
	      if ((im < (S7_LONG_MAX / 4)) && (im > 1.0e-6))
		size = 14;
	    }
	}
    }

  if (radix != 10)
    res = number_to_string_with_radix(sc, x, radix, 0, (radix == 10) ? size : 20, 'g');
  else res = number_to_string_base_10(x, 0, size, 'g');
  
  return(make_string_uncopied(sc, res));
}


#define CTABLE_SIZE 256
static bool *exponent_table, *slashify_table, *char_ok_in_a_name, *white_space, *not_eol;
static int *digits;

static void init_ctables(void)
{
  int i;

  not_eol = (bool *)permanent_calloc(CTABLE_SIZE * sizeof(bool));
  exponent_table = (bool *)permanent_calloc(CTABLE_SIZE * sizeof(bool));
  slashify_table = (bool *)permanent_calloc(CTABLE_SIZE * sizeof(bool));
  char_ok_in_a_name = (bool *)permanent_calloc(CTABLE_SIZE * sizeof(bool));
  white_space = (bool *)permanent_calloc((CTABLE_SIZE + 1) * sizeof(bool));
  white_space++;    /* leave white_space[-1] false for white_space[EOF] */
  
  for (i = 1; i < CTABLE_SIZE; i++)
    not_eol[i] = true;
  not_eol[0] = false;
  not_eol['\n'] = false;

  for (i = 1; i < CTABLE_SIZE; i++)
    char_ok_in_a_name[i] = true;
  char_ok_in_a_name[0] = false;
  char_ok_in_a_name['('] = false;
  char_ok_in_a_name[')'] = false;
  char_ok_in_a_name[';'] = false;
  char_ok_in_a_name['\t'] = false;
  char_ok_in_a_name['\n'] = false;
  char_ok_in_a_name['\r'] = false;
  char_ok_in_a_name[' '] = false;
  char_ok_in_a_name['"'] = false;
  /* double-quote is recent, but I want '(1 ."hi") to be parsed as '(1 . "hi") 
   * what about stuff like vertical tab?  or comma?
   */

  for (i = 0; i < CTABLE_SIZE; i++)
    white_space[i] = false;
  white_space['\t'] = true;
  white_space['\n'] = true;
  white_space['\r'] = true;
  white_space['\f'] = true;
  white_space['\v'] = true;
  white_space[' '] = true;

  /* surely only 'e' is needed... */
  exponent_table['e'] = true; exponent_table['E'] = true;
#if WITH_EXTRA_EXPONENT_MARKERS
  exponent_table['s'] = true; exponent_table['S'] = true; 
  exponent_table['f'] = true; exponent_table['F'] = true;
  exponent_table['d'] = true; exponent_table['D'] = true;
  exponent_table['l'] = true; exponent_table['L'] = true;
#endif

  for (i = 0; i < 32; i++)
    slashify_table[i] = true;
  for (i = 127; i < 160; i++)
    slashify_table[i] = true;
  slashify_table['\\'] = true;
  slashify_table['"'] = true;
  slashify_table['\n'] = false;

  digits = (int *)permanent_calloc(CTABLE_SIZE * sizeof(int));
  for (i = 0; i < CTABLE_SIZE; i++)
    digits[i] = 256;

  digits['0'] = 0; digits['1'] = 1; digits['2'] = 2; digits['3'] = 3; digits['4'] = 4;
  digits['5'] = 5; digits['6'] = 6; digits['7'] = 7; digits['8'] = 8; digits['9'] = 9;
  digits['a'] = 10; digits['A'] = 10;
  digits['b'] = 11; digits['B'] = 11;
  digits['c'] = 12; digits['C'] = 12;
  digits['d'] = 13; digits['D'] = 13;
  digits['e'] = 14; digits['E'] = 14;
  digits['f'] = 15; digits['F'] = 15;
}


#define is_white_space(C) white_space[C]
  /* this is much faster than C's isspace, and does not depend on the current locale.
   * if c == EOF (-1), it indexes into the empty (0) slot we preallocated below white_space 
   */


static s7_pointer check_sharp_readers(s7_scheme *sc, const char *name)
{
  s7_pointer reader, value, args;
  int args_loc = -1;
  value = sc->F;
  args = sc->F; /* make g++ happy */

  /* *#reader* is assumed to be an alist of (char . proc)
   *    where each proc takes one argument, the string from just beyond the "#" to the next delimiter.
   *    The procedure can call read-char to read ahead in the current-input-port.
   *    If it returns anything other than #f, that is the value of the sharp expression.
   * This search happens after #|, #t, and #f.
   */

  for (reader = slot_value(sc->sharp_readers); is_not_null(reader); reader = cdr(reader))
    {
      if (name[0] == s7_character(caar(reader)))
	{
	  if (args_loc == -1)
	    {
	      args = list_1(sc, s7_make_string(sc, name));
	      args_loc = s7_gc_protect(sc, args);
	    }
	  value = s7_call(sc, cdar(reader), args);
	  if (value != sc->F)
	    break;
	}
    }
  if (args_loc != -1)
    s7_gc_unprotect_at(sc, args_loc);

  return(value);
}


static s7_pointer g_sharp_readers_set(s7_scheme *sc, s7_pointer args)
{
  /* new value must be either '() or a proper list of conses (char . func) */
  if (is_null(cadr(args))) return(cadr(args));
  if (is_pair(cadr(args)))
    {
      s7_pointer x;
      for (x = cadr(args); is_pair(x); x = cdr(x))
	{
	  if ((!is_pair(car(x))) ||
	      (!s7_is_character(caar(x))) ||
	      (!s7_is_procedure(cdar(x))))
	    return(sc->ERROR);
	}
      if (is_null(x))
	return(cadr(args));
    }
  return(sc->ERROR);
}


static bool is_abnormal(s7_pointer x)
{
  switch (type(x))
    {
    case T_INTEGER:
    case T_RATIO:
      return(false);

    case T_REAL:
      return(isinf(real(x)) || 
	     is_NaN(real(x)));

    case T_COMPLEX:
      return(((isinf(s7_real_part(x))) || 
	      (isinf(s7_imag_part(x))) ||
	      (is_NaN(s7_real_part(x))) || 
	      (is_NaN(s7_imag_part(x)))));

#if WITH_GMP
    case T_BIG_INTEGER:
    case T_BIG_RATIO:
      return(false);
      
    case T_BIG_REAL:
      return((isinf(s7_real_part(x))) || 
	     (is_NaN(s7_real_part(x))));

    case T_BIG_COMPLEX:
      return((isinf(s7_real_part(x))) || 
	     (isinf(s7_imag_part(x))) ||
	     (isnan(s7_real_part(x))) || 
	     (isnan(s7_imag_part(x))));
#endif

    default:
      return(true);
    }
}


#define NESTED_SHARP false
#define UNNESTED_SHARP true

#define SYMBOL_OK true
#define NO_SYMBOLS false

static s7_pointer make_sharp_constant(s7_scheme *sc, char *name, bool at_top, int radix) 
{
  /* name is the stuff after the '#', return sc->NIL if not a recognized #... entity */
  int len;
  s7_pointer x;

  if ((name[0] == 't') && (name[1] == '\0'))
    return(sc->T);
  
  if ((name[0] == 'f') && (name[1] == '\0'))
    return(sc->F);

  if (is_not_null(slot_value(sc->sharp_readers)))
    {
      x = check_sharp_readers(sc, name);
      if (x != sc->F)
	return(x);
    }
  
  len = safe_strlen(name);
  if (len < 2)
    return(sc->NIL);
      
  switch (name[0])
    {
      /* -------- #< ... > -------- */
    case '<':
      if (strings_are_equal(name, "<unspecified>"))
	return(sc->UNSPECIFIED);

      if (strings_are_equal(name, "<undefined>"))
	return(sc->UNDEFINED);

      if (strings_are_equal(name, "<eof>"))
	return(sc->EOF_OBJECT);

      return(sc->NIL);
      break;
      

      /* -------- #o #d #x #b -------- */
    case 'o':   /* #o (octal) */
    case 'd':   /* #d (decimal) */
    case 'x':   /* #x (hex) */
    case 'b':   /* #b (binary) */
      {
	bool to_inexact = false, to_exact = false;
	int num_at = 1;
  
	if (name[1] == '#')
	  {
	    if (!at_top)
	      return(sc->NIL);
	    if ((len > 2) && ((name[2] == 'e') || (name[2] == 'i'))) /* r6rs includes caps here */
	      {
		if ((len > 3) && (name[3] == '#'))
		  return(sc->NIL);
		to_inexact = (name[2] == 'i');
		to_exact = (name[2] == 'e');
		num_at = 3;
	      }
	    else return(sc->NIL);
	  }
	/* the #b or whatever overrides any radix passed in earlier */
	x = make_atom(sc, (char *)(name + num_at), (name[0] == 'o') ? 8 : ((name[0] == 'x') ? 16 : ((name[0] == 'b') ? 2 : 10)), NO_SYMBOLS);

	/* #x#i1 apparently makes sense, so #x1.0 should also be accepted.
	 * here we can get #b#e0/0 or #b#e+1/0 etc.
	 * surely if #e1+i is an error (or #f), and #e#x1+i is an error,
	 *   #x#e1+i should also be an error, but #e1+0i is not an error I guess since there actually isn't any imaginary part
	 */
	if (is_abnormal(x))
	  return(sc->NIL);

	if ((!to_exact) && (!to_inexact))
	  return(x);

	if ((s7_imag_part(x) != 0.0) && (to_exact))  /* #x#e1+i */
	  return(sc->NIL);

#if WITH_GMP
	if (s7_is_bignum(x))
	  {
	    if (to_exact)
	      return(big_inexact_to_exact(sc, list_1(sc, x)));
	    return(big_exact_to_inexact(sc, list_1(sc, x)));
	  }
#endif
	if (to_exact)
	  return(inexact_to_exact(sc, x));
	return(exact_to_inexact(sc, x));
      }
      break;


      /* -------- #i -------- */
    case 'i':   /* #i<num> = ->inexact (see token for table of choices here) */
      if (name[1] == '#')
	{
	  /* there are special cases here: "#e0/0" or "#e#b0/0" -- all infs are complex: 
	   *    #i1/0=nan.0 but #i1/0+i=inf+1i so e->i is a no-op but i->e is not
	   *
	   * even trickier: a *#reader* like #t<num> could be used as #e#t13.25 so make_sharp_constant
	   *   needs to be willing to call the readers even when not at_top (i.e. when NESTED_SHARP).
	   */

	  if ((name[2] == 'e') ||                        /* #i#e1 -- assume these aren't redefinable? */
	      (name[2] == 'i'))
	    return(sc->NIL);

	  x = make_sharp_constant(sc, (char *)(name + 2), NESTED_SHARP, radix);
	  if (s7_is_number(x))
	    {
	      if (is_abnormal(x))
		return(sc->NIL);
#if WITH_GMP
	      if (s7_is_bignum(x))                        /* (string->number "#b#e-11e+111") */
		return(big_exact_to_inexact(sc, list_1(sc, x)));
#endif
	      return(exact_to_inexact(sc, x));
	    }
	  return(sc->NIL);
	}
      x = make_atom(sc, (char *)(name + 1), radix, NO_SYMBOLS);
      if (!s7_is_number(x))  /* not is_abnormal(x) -- #i0/0 -> nan etc */
	return(sc->NIL);
#if WITH_GMP
      if (s7_is_bignum(x))
	return(big_exact_to_inexact(sc, list_1(sc, x)));
#endif
      return(exact_to_inexact(sc, x));
      break;
  

      /* -------- #e -------- */
    case 'e':   /* #e<num> = ->exact */
      if (name[1] == '#')
	{
	  if ((name[2] == 'e') ||                        /* #e#e1 */
	      (name[2] == 'i'))
	    return(sc->NIL);

	  x = make_sharp_constant(sc, (char *)(name + 2), NESTED_SHARP, radix);
	  if (s7_is_number(x))
	    {
	      if (is_abnormal(x))                        /* (string->number "#e#b0/0") */
		return(sc->NIL);
	      if (!s7_is_real(x))                        /* (string->number "#e#b1+i") */
		return(sc->NIL);
#if WITH_GMP
	      return(big_inexact_to_exact(sc, list_1(sc, x)));
#endif
	      return(inexact_to_exact(sc, x));
	    }
	  return(sc->NIL);
	}

      x = make_atom(sc, (char *)(name + 1), radix, NO_SYMBOLS);
#if WITH_GMP
      /* #e1e310 is a simple case */
      if (s7_is_bignum(x))
	return(big_inexact_to_exact(sc, list_1(sc, x)));
#endif	
      if (is_abnormal(x))                                /* (string->number "#e0/0") */
	return(sc->NIL);
      if (!s7_is_real(x))                                /* (string->number "#e1+i") */
	return(sc->NIL);
      
#if WITH_GMP
      /* there are non-big floats that are greater than most-positive-fixnum:
       *    :(> .1e20 most-positive-fixnum) -> #t
       *    :(bignum? .1e20) -> #f
       * so we have to check that, not just is it a bignum.
       */
      return(big_inexact_to_exact(sc, list_1(sc, x)));
#endif
      return(inexact_to_exact(sc, x));
      break;


      /* -------- #\... -------- */
    case '\\':
      if (name[2] == 0)                             /* the most common case: #\a */
	return(chars[(unsigned char)(name[1])]);
      /* not unsigned int here!  (unsigned int)255 (as a char) returns -1!! */

      if (strings_are_equal(name + 1, "space")) 
	return(chars[' ']);

      if ((strings_are_equal(name + 1, "newline")) || 
	  (strings_are_equal(name + 1, "linefeed")))
	return(chars['\n']);

      if (strings_are_equal(name + 1, "return")) 
	return(chars['\r']);

      if (strings_are_equal(name + 1, "tab")) 
	return(chars['\t']);

      if ((strings_are_equal(name + 1, "null")) || 
	  (strings_are_equal(name + 1, "nul")))
	return(chars[0]);

      /* the next 4 are for r7rs */
      if (strings_are_equal(name + 1, "alarm")) 
	return(chars[7]);

      if (strings_are_equal(name + 1, "backspace")) 
	return(chars[8]);

      if (strings_are_equal(name + 1, "escape")) 
	return(chars[0x1b]);

      if (strings_are_equal(name + 1, "delete")) 
	return(chars[0x7f]);

      if (name[1] == 'x')     /* #\x is just x, but apparently #\x<num> is int->char? #\x65 -> #\e -- Guile doesn't have this
			       *    (it is from r6rs -- perhaps it is a bad idea...)
			       */
	{
	  /* sscanf here misses errors like #\x1.4, but make_atom misses #\x6/3,
	   *   #\x#b0, #\x#e0.0, #\x-0, #\x#e0e100 etc, so we have to do it at
	   *   an even lower level.
	   */
	  bool happy = true;
	  char *tmp;
	  int lval = 0;

	  tmp = (char *)(name + 2);
	  while ((*tmp) && (happy))
	    {
	      int dig;
	      dig = digits[(int)(*tmp++)];
	      if (dig < 16)
		lval = dig + (lval * 16);
	      else happy = false;
	    }
	  if ((happy) &&
	      (lval < 256))
	    return(chars[lval]);
	}
    }
  return(sc->NIL);
}


static s7_Int string_to_integer(const char *str, int radix, bool *overflow)
{
  bool negative = false;
  s7_Int lval = 0;
  int dig;
  char *tmp = (char *)str;
  char *tmp1;

  if (str[0] == '+')
    tmp++;
  else 
    {
      if (str[0] == '-')
	{
	  negative = true;
	  tmp++;
	}
    }
  while (*tmp == '0') {tmp++;};
  tmp1 = tmp;

 if (radix == 10)
    {
      while (true)
	{
	  dig = digits[(unsigned char)(*tmp++)];
	  if (dig < 10)
	    lval = dig + (lval * 10);
	  else break;
	  dig = digits[(unsigned char)(*tmp++)];
	  if (dig < 10)
	    lval = dig + (lval * 10);
	  else break;
	}
    }
  else
    {
      while (true)
	{
	  dig = digits[(unsigned char)(*tmp++)];
	  if (dig < radix)
	    lval = dig + (lval * radix);
	  else break;
	  dig = digits[(unsigned char)(*tmp++)];
	  if (dig < radix)
	    lval = dig + (lval * radix);
	  else break;
	}
    }

#if WITH_GMP
  (*overflow) = ((lval > S7_LONG_MAX) ||
		 ((tmp - tmp1) > s7_int_digits_by_radix[radix]));
  /* this tells the string->number readers to create a bignum.  We need to be very
   *    conservative here to catch contexts such as (/ 1/524288 19073486328125)
   */
#else
  if ((tmp - tmp1 - 2) > s7_int_digits_by_radix[radix])
    {
      /* I can't decide what to do with these non-gmp overflows.  Perhaps NAN in all cases? 
       *     overflow: 9223372036854775810 -> -9223372036854775806 -- this is not caught currently
       */
      (*overflow) = true;
      if (negative)
	return(s7_Int_min);       /* or INFINITY? */
      return(s7_Int_max);         /* 0/100000000000000000000000000000000000000000000000000000000000000000000 */
    }
#endif

  if (negative)
    return(-lval);
  return(lval);
}


/*  9223372036854775807                9223372036854775807
 * -9223372036854775808               -9223372036854775808
 * 0000000000000000000000000001.0     1.0
 * 1.0000000000000000000000000000     1.0
 * 1000000000000000000000000000.0e-40 1.0e-12
 * 0.0000000000000000000000000001e40  1.0e12
 * 1.0e00000000000000000001           10.0
 */

static s7_Double string_to_double_with_radix(const char *ur_str, int radix, bool *overflow)
{
  /* strtod follows LANG which is not what we want (only "." is decimal point in Scheme).
   *   To overcome LANG in strtod would require screwing around with setlocale which never works.
   *   So we use our own code -- according to valgrind, this function is much faster than strtod.
   *
   * comma as decimal point causes ambiguities: `(+ ,1 2) etc
   */

  int i, sign = 1, frac_len, int_len, dig, max_len, exponent = 0;
  long long int int_part = 0, frac_part = 0;
  char *str;
  char *ipart, *fpart;
  s7_Double dval = 0.0;

  /* there's an ambiguity in number notation here if we allow "1e1" or "1.e1" in base 16 (or 15) -- is e a digit or an exponent marker?
   *   but 1e+1, for example disambiguates it -- kind of messy! -- the scheme spec says "e" can only occur in base 10. 
   *   mpfr says "e" as exponent only in bases <= 10 -- else use '@' which works in any base.  This can only cause confusion
   *   in scheme, unfortunately, due to the idiotic scheme polar notation.  But we accept "s" and "l" as exponent markers
   *   so, perhaps for radix > 10, the exponent, if any, has to use one of S s L l?  Not "l"!  And "s" originally meant "short".
   *
   * Another slight ambiguity: 1+1/2i is parsed as 1 + 0.5i, not 1+1/(2i), or (1+1)/(2i) or (1+1/2)i etc
   */

  max_len = s7_int_digits_by_radix[radix];
  str = (char *)ur_str;

  if (*str == '+')
    str++;
  else
    {
      if (*str == '-')
	{
	  str++;
	  sign = -1;
	}
    }
  while (*str == '0') {str++;};

  ipart = str;
  while (digits[(int)(*str)] < radix) str++;
  int_len = str - ipart;

  if (*str == '.') str++;
  fpart = str;
  while (digits[(int)(*str)] < radix) str++;
  frac_len = str - fpart;

  if ((*str) && (exponent_table[(unsigned char)(*str)]))
    {
      int exp_negative = false;
      str++;
      if (*str == '+') 
	str++;
      else
	{
	  if (*str == '-')
	    {
	      str++;
	      exp_negative = true;
	    }
	}
      while ((dig = digits[(int)(*str++)]) < 10) /* exponent is always base 10 */
	exponent = dig + (exponent * 10);
      if (exp_negative) exponent = -exponent;
    }

#if WITH_GMP
  /* 9007199254740995.0 */

  if (int_len + frac_len >= max_len)
    {
      (*overflow) = true;
      return(0.0);
    }
#endif
      
  str = ipart;
  if ((int_len + exponent) > max_len)
    {
      /*  12341234.56789e12                   12341234567889999872.0              1.234123456789e+19
       * -1234567890123456789.0              -1234567890123456768.0              -1.2345678901235e+18
       *  12345678901234567890.0              12345678901234567168.0              1.2345678901235e+19
       *  123.456e30                          123456000000000012741097792995328.0 1.23456e+32
       *  12345678901234567890.0e12           12345678901234569054409354903552.0  1.2345678901235e+31
       *  1.234567890123456789012e30          1234567890123456849145940148224.0   1.2345678901235e+30
       *  1e20                                100000000000000000000.0             1e+20
       *  1234567890123456789.0               1234567890123456768.0               1.2345678901235e+18
       *  123.456e16                          1234560000000000000.0               1.23456e+18
       *  98765432101234567890987654321.0e-5  987654321012345728401408.0          9.8765432101235e+23
       *  98765432101234567890987654321.0e-10 9876543210123456512.0               9.8765432101235e+18
       *  0.00000000000000001234e20           1234.0
       *  0.000000000000000000000000001234e30 1234.0
       *  0.0000000000000000000000000000000000001234e40 1234.0
       *  0.000000000012345678909876543210e15 12345.678909877
       *  0e1000                              0.0
       */

      for (i = 0; i < max_len; i++)
	{
	  dig = digits[(int)(*str++)];
	  if (dig < radix)
	    int_part = dig + (int_part * radix);
	  else break;
	}

      /* if the exponent is huge, check for 0 int_part and frac_part before complaining (0e1000 or 0.0e1000) */
      if ((int_part == 0) &&
	  (exponent > max_len))
	{
	  /* if frac_part is also 0, return 0.0 */
	  if (frac_len == 0)
	    return(0.0);

	  str = fpart;
	  while ((dig = digits[(int)(*str++)]) < radix)
	    frac_part = dig + (frac_part * radix);
	  if (frac_part == 0)
	    return(0.0);

#if WITH_GMP
	  (*overflow) = true;
#endif
	}
      
#if WITH_GMP
      (*overflow) = ((int_part > 0) || (exponent > 20));    /* .1e310 is a tricky case */
#endif
      
      if (int_part != 0) /* 0.<310 zeros here>1e310 for example --
			  *   pow (via ipow) thinks it has to be too big, returns Nan,
			  *   then Nan * 0 -> Nan and the NaN progogates
			  */
	{
	  if (int_len <= max_len)
	    dval = int_part * ipow(radix, exponent);
	  else dval = int_part * ipow(radix, exponent + int_len - max_len);
	}
      else dval = 0.0;

      /* shift by exponent, but if int_len > max_len then we assumed (see below) int_len - max_len 0's on the left */
      /*   using int_to_int or table lookups here instead of pow did not make any difference in speed */

      if (int_len < max_len)
	{
	  int k, flen;
	  str = fpart;
	  
	  for (k = 0; (frac_len > 0) && (k < exponent); k += max_len)
	    {
	      if (frac_len > max_len) flen = max_len; else flen = frac_len;
	      frac_len -= max_len;

	      frac_part = 0;
	      for (i = 0; i < flen; i++)
		frac_part = digits[(int)(*str++)] + (frac_part * radix);

	      if (frac_part != 0)                                /* same pow->NaN problem as above can occur here */
		dval += frac_part * ipow(radix, exponent - flen - k);
	    }
	}
      else
	{
	  /* some of the fraction is in the integer part before the negative exponent shifts it over */
	  if (int_len > max_len)
	    {
	      int ilen;
	      /* str should be at the last digit we read */
	      ilen = int_len - max_len;                          /* we read these above */
	      if (ilen > max_len)
		ilen = max_len;

	      for (i = 0; i < ilen; i++)
		frac_part = digits[(int)(*str++)] + (frac_part * radix);

	      dval += frac_part * ipow(radix, exponent - ilen);
	    }
	}

      return(sign * dval);
    }

  /* int_len + exponent <= max_len */

  if (int_len <= max_len)
    {
      char *iend;
      int int_exponent;

      /* a better algorithm (since the inaccuracies are in the radix^exponent portion):
       *   strip off leading zeros and possible sign,
       *   strip off digits beyond max_len, then remove any trailing zeros.
       *     (maybe fiddle with the lowest order digit here for rounding, but I doubt it matters)
       *   read digits until end of number or max_len reached, ignoring the decimal point
       *   get exponent and use it and decimal point location to position the current result integer
       * this always combines the same integer and the same exponent no matter how the number is expressed.
       */

      int_exponent = exponent;
      if (int_len > 0)
	{
	  iend = (char *)(str + int_len - 1);
	  while ((*iend == '0') && (iend != str)) {iend--; int_exponent++;}

	  while (str <= iend)
	    int_part = digits[(int)(*str++)] + (int_part * radix);
	}
      if (int_exponent != 0)
	dval = int_part * ipow(radix, int_exponent);
      else dval = (s7_Double)int_part;
    }
  else
    {
      int len, flen;
      long long int fpart = 0;

      /* 98765432101234567890987654321.0e-20    987654321.012346  
       * 98765432101234567890987654321.0e-29    0.98765432101235
       * 98765432101234567890987654321.0e-30    0.098765432101235
       * 98765432101234567890987654321.0e-28    9.8765432101235
       */

      len = int_len + exponent;
      for (i = 0; i < len; i++)
	int_part = digits[(int)(*str++)] + (int_part * radix);
      
      flen = -exponent;
      if (flen > max_len)
	flen = max_len;

      for (i = 0; i < flen; i++)
	fpart = digits[(int)(*str++)] + (fpart * radix);

      if (len <= 0)
	dval = int_part + fpart * ipow(radix, len - flen);
      else dval = int_part + fpart * ipow(radix, -flen);
    }

  if (frac_len > 0)
    {
      str = fpart;
      if (frac_len <= max_len)
	{
	  /* splitting out base 10 case saves very little here */
	  /* this ignores trailing zeros, so that 0.3 equals 0.300 */
	  char *fend;
	  
	  fend = (char *)(str + frac_len - 1);
	  while ((*fend == '0') && (fend != str)) {fend--; frac_len--;} /* (= .6 0.6000) */
	  
	  while (str <= fend)
	    frac_part = digits[(int)(*str++)] + (frac_part * radix);
	  dval += frac_part * ipow(radix, exponent - frac_len);
	  
	  /* fprintf(stderr, "frac: %lld, exp: (%d %d) %.20f, val: %.20f\n", frac_part, exponent, frac_len, ipow(radix, exponent - frac_len), dval); 
	   * 0.6:    frac:    6, exp: 0.10000000000000000555, val: 0.60000000000000008882
	   * 0.60:   frac:   60, exp: 0.01000000000000000021, val: 0.59999999999999997780
	   * 0.6000: frac: 6000, exp: 0.00010000000000000000, val: 0.59999999999999997780
	   * :(= 0.6 0.60)
	   * #f
	   * :(= #i3/5 0.6)
	   * #f
	   * so (string->number (number->string num)) == num only if both num's are the same text (or you get lucky)
	   * :(= 0.6 6e-1) ; but not 60e-2
	   * #t
	   *
	   * to fix the 0.60 case, we need to ignore trailing post-dot zeros.
	   */
	}
      else
	{
	  if (exponent <= 0)
	    {
	      for (i = 0; i < max_len; i++)
		frac_part = digits[(int)(*str++)] + (frac_part * radix);
	      
	      dval += frac_part * ipow(radix, exponent - max_len);
	    }
	  else
	    {
	      /* 1.0123456789876543210e1         10.12345678987654373771  
	       * 1.0123456789876543210e10        10123456789.87654304504394531250
	       * 0.000000010000000000000000e10   100.0
	       * 0.000000010000000000000000000000000000000000000e10 100.0
	       * 0.000000012222222222222222222222222222222222222e10 122.22222222222222
	       * 0.000000012222222222222222222222222222222222222e17 1222222222.222222
	       */
	      
	      int_part = 0;
	      for (i = 0; i < exponent; i++)
		int_part = digits[(int)(*str++)] + (int_part * radix);
	      
	      frac_len -= exponent;
	      if (frac_len > max_len)
		frac_len = max_len;
	      
	      for (i = 0; i < frac_len; i++)
		frac_part = digits[(int)(*str++)] + (frac_part * radix);
	      
	      dval += int_part + frac_part * ipow(radix, -frac_len);
	    }
	}
    }

#if WITH_GMP
  if ((int_part == 0) &&
      (frac_part == 0))
    return(0.0);
  (*overflow) = ((frac_len - exponent) > max_len);
#endif
  
  return(sign * dval);
}


/* make symbol or number atom from string */

static s7_pointer make_atom(s7_scheme *sc, char *q, int radix, bool want_symbol) 
{
  #define ISDIGIT(Chr, Rad) (digits[(unsigned char)Chr] < Rad)

  char c, *p;
  bool has_dec_point1 = false;

  p = q;
  c = *p++; 
  
  /* a number starts with + - . or digit, but so does 1+ for example */
  
  switch (c)
    {
    case '#':
      return(make_sharp_constant(sc, p, UNNESTED_SHARP, radix)); /* make_sharp_constant expects the '#' to be removed */

    case '+':
    case '-':
      c = *p++; 
      if (c == '.') 
	{ 
	  has_dec_point1 = true; 
	  c = *p++; 
	} 
      if ((!c) || (!ISDIGIT(c, radix)))
	return((want_symbol) ? make_symbol(sc, q) : sc->F);
      break;

    case '.':
      has_dec_point1 = true; 
      c = *p++; 

      if ((!c) || (!ISDIGIT(c, radix)))
	return((want_symbol) ? make_symbol(sc, q) : sc->F); 
      break;

    case '0':        /* these two are always digits */
    case '1':
      break;

    default:
      if (!ISDIGIT(c, radix))
	return((want_symbol) ? make_symbol(sc, q) : sc->F); 
      break;
    }

  /* now it's possibly a number -- the 1st character(s) could be part of a number in the current radix */

  {
    char *slash1 = NULL, *slash2 = NULL, *plus = NULL, *ex1 = NULL, *ex2 = NULL;
    bool has_i = false, has_dec_point2 = false;
    int has_plus_or_minus = 0, current_radix;

#if (!WITH_GMP)
    bool overflow = false;
#endif
    current_radix = radix;  /* current_radix is 10 for the exponent portions, but radix for all the rest */
  
    for ( ; (c = *p) != 0; ++p)
      {
	/* what about embedded null? (string->number (string #\1 (integer->char 0) #\0)) 
	 *   currently we stop and return 1, but Guile returns #f
	 */
	if (!ISDIGIT(c, current_radix))         /* moving this inside the switch statement was much slower */
	  {
	    current_radix = radix;
	    
	    switch (c)
	      {
		/* -------- decimal point -------- */
	      case '.':
		if (((has_dec_point1) ||
		     (slash1)) &&
		    (has_plus_or_minus == 0)) /* 1.. or 1/2. */
		  return((want_symbol) ? make_symbol(sc, q) : sc->F); 
		
		if (((has_dec_point2) ||
		     (slash2)) &&
		    (has_plus_or_minus != 0)) /* 1+1.. or 1+1/2. */
		  return((want_symbol) ? make_symbol(sc, q) : sc->F); 
		
		if ((!ISDIGIT(p[1], current_radix)) &&
		    (!ISDIGIT(p[-1], current_radix))) 
		  return((want_symbol) ? make_symbol(sc, q) : sc->F); 
		
		if (has_plus_or_minus == 0)
		  has_dec_point1 = true;
		else has_dec_point2 = true;
		continue;

		
		/* -------- exponent marker -------- */
	      case 'e': case 'E':
#if WITH_EXTRA_EXPONENT_MARKERS
	      case 's': case 'S':
	      case 'd': case 'D':
	      case 'f': case 'F':
	      case 'l': case 'L':
#endif
		if (current_radix > 10)
		  return((want_symbol) ? make_symbol(sc, q) : sc->F); 
		/* see note above */
		
		current_radix = 10;
		
		if (((ex1) ||
		     (slash1)) &&
		    (has_plus_or_minus == 0)) /* ee */
		  return((want_symbol) ? make_symbol(sc, q) : sc->F); 
		
		if (((ex2) ||
		     (slash2)) &&
		    (has_plus_or_minus != 0)) /* 1+1.0ee */
		  return((want_symbol) ? make_symbol(sc, q) : sc->F); 
		
		if ((!ISDIGIT(p[-1], current_radix)) &&
		    (p[-1] != '.'))
		  return((want_symbol) ? make_symbol(sc, q) : sc->F); 
		
		if (has_plus_or_minus == 0)
		  {
		    ex1 = p;
		    has_dec_point1 = true; /* decimal point illegal from now on */
		  }
		else 
		  {
		    ex2 = p;
		    has_dec_point2 = true;
		  }
		p++;
		if ((*p == '-') || (*p == '+')) p++;
		if (ISDIGIT(*p, current_radix))
		  continue;
		break;


		/* -------- internal + or - -------- */
	      case '+':
	      case '-':
		if (has_plus_or_minus != 0) /* already have the separator */
		  return((want_symbol) ? make_symbol(sc, q) : sc->F);
		
		if (c == '+') has_plus_or_minus = 1; else has_plus_or_minus = -1;
		plus = (char *)(p + 1);
		continue;
		
		/* ratio marker */
	      case '/':
		if ((has_plus_or_minus == 0) &&
		    ((ex1) ||
		     (slash1) ||
		     (has_dec_point1)))
		  return((want_symbol) ? make_symbol(sc, q) : sc->F);
		
		if ((has_plus_or_minus != 0) &&
		    ((ex2) ||
		     (slash2) ||
		     (has_dec_point2)))
		  return((want_symbol) ? make_symbol(sc, q) : sc->F);
		
		if (has_plus_or_minus == 0)
		  slash1 = (char *)(p + 1);
		else slash2 = (char *)(p + 1);
		
		if ((!ISDIGIT(p[1], current_radix)) ||
		    (!ISDIGIT(p[-1], current_radix)))
		  return((want_symbol) ? make_symbol(sc, q) : sc->F);
		
		continue;


		/* -------- i for the imaginary part -------- */
	      case 'i':
		if ((has_plus_or_minus != 0) && 
		    (!has_i))
		  {
		    has_i = true;
		    continue;
		  }
		break;

	      default:
		break;
	      }

	    return((want_symbol) ? make_symbol(sc, q) : sc->F);
	  }
      }
    
    if ((has_plus_or_minus != 0) &&        /* that is, we have an internal + or - */
	(!has_i))                          /*   but no i for the imaginary part */
      return((want_symbol) ? make_symbol(sc, q) : sc->F);

    if (has_i)
      {
#if (!WITH_GMP)
	s7_Double rl = 0.0, im = 0.0;
#else
	char e1 = 0, e2 = 0;
#endif
	s7_pointer result;
	int len;
	char ql1, pl1;
	
	len = safe_strlen(q);
	
	if (q[len - 1] != 'i')
	  return((want_symbol) ? make_symbol(sc, q) : sc->F);
	
	/* save original string */
	ql1 = q[len - 1];
	pl1 = (*(plus - 1));
#if WITH_GMP
	if (ex1) {e1 = *ex1; (*ex1) = 'e';} /* for mpfr */
	if (ex2) {e2 = *ex2; (*ex2) = 'e';}
#endif
	
	/* look for cases like 1+i */
	if ((q[len - 2] == '+') || (q[len - 2] == '-'))
	  q[len - 1] = '1';
	else q[len - 1] = '\0'; /* remove 'i' */
	
	(*((char *)(plus - 1))) = '\0';
	
	/* there is a slight inconsistency here:
	   1/0      -> nan.0
           1/0+0i   -> inf.0 (0/1+0i is 0.0)
	   #i1/0+0i -> inf.0
	   0/0      -> nan.0
	   0/0+0i   -> -nan.0
	*/

#if (!WITH_GMP)
	if ((has_dec_point1) ||
	    (ex1))
	  {
	    /* (string->number "1100.1+0.11i" 2) -- need to split into 2 honest reals before passing to non-base-10 str->dbl */
	    rl = string_to_double_with_radix(q, radix, &overflow);
	  }
	else
	  {
	    if (slash1)
	      {
		/* here the overflow could be innocuous if it's in the denominator and the numerator is 0
		 *    0/100000000000000000000000000000000000000-0i
		 */
		s7_Int num, den;
		num = string_to_integer(q, radix, &overflow);
		den = string_to_integer(slash1, radix, &overflow);
		
		if ((num == 0) && (den != 0))
		  {
		    rl = 0.0;
		    overflow = false;
		  }
		else rl = (s7_Double)num / (s7_Double)den;
	      }
	    else rl = (s7_Double)string_to_integer(q, radix, &overflow);
	    if (overflow) return(real_NaN);
	  }
	if (rl == -0.0) rl = 0.0;
	
	if ((has_dec_point2) ||
	    (ex2))
	  im = string_to_double_with_radix(plus, radix, &overflow);
	else
	  {
	    if (slash2)
	      {
		/* same as above: 0-0/100000000000000000000000000000000000000i
		 */
		s7_Int num, den;
		num = string_to_integer(plus, radix, &overflow);
		den = string_to_integer(slash2, radix, &overflow);
		if ((num == 0) && (den != 0))
		  {
		    im = 0.0;
		    overflow = false;
		  }
		else im = (s7_Double)num / (s7_Double)den;
	      }
	    else im = (s7_Double)string_to_integer(plus, radix, &overflow);
	    if (overflow) return(real_NaN);
	  }
	if ((has_plus_or_minus == -1) && 
	    (im != 0.0))
	  im = -im;
	result = s7_make_complex(sc, rl, im);
#else
	result = string_to_either_complex(sc, q, slash1, ex1, has_dec_point1, plus, slash2, ex2, has_dec_point2, radix, has_plus_or_minus);
#endif
	
	/* restore original string */
	q[len - 1] = ql1;
	(*((char *)(plus - 1))) = pl1;
#if WITH_GMP
	if (ex1) (*ex1) = e1;
	if (ex2) (*ex2) = e2;
#endif
	
	return(result);
      }
    
    /* not complex */
    if ((has_dec_point1) ||
	(ex1))
      {
	s7_pointer result;
	
	if (slash1)  /* not complex, so slash and "." is not a number */
	  return((want_symbol) ? make_symbol(sc, q) : sc->F);
	
#if (!WITH_GMP)
	result = s7_make_real(sc, string_to_double_with_radix(q, radix, &overflow));
#else
	{
	  char old_e = 0;
	  if (ex1)
	    {
	      old_e = (*ex1);
	      (*ex1) = 'e';
	    }
	  result = string_to_either_real(sc, q, radix);
	  if (ex1)
	    (*ex1) = old_e;
	}
#endif
	return(result);
      }
    
    /* not real */
    if (slash1)
#if (!WITH_GMP)
      {
	s7_Int n, d;

	n = string_to_integer(q, radix, &overflow);
	d = string_to_integer(slash1, radix, &overflow);

	if ((n == 0) && (d != 0))                        /* 0/100000000000000000000000000000000000000 */
	  return(small_int(0));
	if ((d == 0) || (overflow))
	  return(real_NaN);
	/* it would be neat to return 1 from 10000000000000000000000000000/10000000000000000000000000000 
	 *   but q is the entire number ('/' included) and slash1 is the stuff after the '/', and every 
	 *   big number comes through here, so there's no clean and safe way to check that q == slash1.
	 */
	return(s7_make_ratio(sc, n, d));
      }
#else
    return(string_to_either_ratio(sc, q, slash1, radix));
#endif
    
    /* integer */
#if (!WITH_GMP)
    {
      s7_Int x;
      x = string_to_integer(q, radix, &overflow);
      if (overflow) 
	return(s7_make_real(sc, (q[0] == '-') ? -INFINITY : INFINITY)); /* was NaN */
      return(make_integer(sc, x));
    }
#else
    return(string_to_either_integer(sc, q, radix));
#endif
  }
}


static s7_pointer s7_string_to_number(s7_scheme *sc, char *str, int radix)
{
  s7_pointer x;
  x = make_atom(sc, str, radix, NO_SYMBOLS);
  if (s7_is_number(x))  /* only needed because str might start with '#' and not be a number (#t for example) */
    return(x);
  return(sc->F);
}


static s7_pointer g_string_to_number_1(s7_scheme *sc, s7_pointer args, const char *caller)
{
  #define H_string_to_number "(string->number str (radix 10)) converts str into a number. \
If str does not represent a number, string->number returns #f.  If 'str' has an embedded radix, \
the 'radix' argument is ignored: (string->number \"#x11\" 2) -> 17 not 3."

  s7_Int radix = 0;
  char *str;

  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, caller, (is_null(cdr(args))) ? 0 : 1, car(args), "a string"));

  str = (char *)string_value(car(args));
  if ((!str) || (!(*str)))
    return(sc->F);

  if (is_pair(cdr(args)))
    {
      if (!s7_is_integer(cadr(args)))
	return(s7_wrong_type_arg_error(sc, caller, 2, cadr(args), "an integer"));

      if (s7_is_integer(cadr(args)))
	radix = s7_integer(cadr(args));
      if ((radix < 2) ||              /* what about negative int as base (Knuth), reals such as phi, and some complex like -1+i */
	  (radix > 16))               /* the only problem here is printing the number; perhaps put each digit in "()" in base 10: (123)(0)(34) */
	return(s7_out_of_range_error(sc, caller, 2, cadr(args), "should be between 2 and 16"));
    }
  else radix = 10;

  switch (str[0])
    {
    case 'n':
      if (safe_strcmp(str, "nan.0") == 0)
	return(real_NaN);
      break;

    case 'i':
      if (safe_strcmp(str, "inf.0") == 0)
	return(s7_make_real(sc, INFINITY));
      break;

    case '-':
      if ((str[1] == 'i') && (safe_strcmp((const char *)(str + 1), "inf.0") == 0))
	 return(s7_make_real(sc, -INFINITY));
       break;

    case '+':
      if ((str[1] == 'i') && (safe_strcmp((const char *)(str + 1), "inf.0") == 0))
	 return(s7_make_real(sc, INFINITY));
       break;
    }

  return(s7_string_to_number(sc, str, radix));
}


static s7_pointer g_string_to_number(s7_scheme *sc, s7_pointer args)
{
  return(g_string_to_number_1(sc, args, "string->number"));
}


static bool numbers_are_eqv(s7_pointer a, s7_pointer b)
{
  if (type(a) != type(b)) /* (eqv? 1 1.0) -> #f! */
    return(false);

  switch (type(a))
    {
    case T_INTEGER:
      return((integer(a) == integer(b)));
  
    case T_RATIO:
      return((numerator(a) == numerator(b)) &&
	     (denominator(a) == denominator(b)));

    case T_REAL:
      return(real(a) == real(b));
  
    default:
      return((real_part(a) == real_part(b)) &&
	     (imag_part(a) == imag_part(b)));
    }
  
  return(false);
}


#if (!WITH_GMP)
static s7_pointer g_abs(s7_scheme *sc, s7_pointer args)
{
  #define H_abs "(abs x) returns the absolute value of the real number x"

  s7_pointer x;
  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
      if (integer(x) < 0)
	return(make_integer(sc, -integer(x)));
      return(x);

    case T_RATIO:
      if (numerator(x) < 0)
	return(s7_make_ratio(sc, -numerator(x), denominator(x)));
      return(x);

    case T_REAL:
      if (real(x) < 0.0)
	return(s7_make_real(sc, -real(x)));
      return(x);

    default:
      return(s7_wrong_type_arg_error(sc, "abs", 0, x, "a real"));
    }
}


#if WITH_OPTIMIZATION
static s7_pointer abs_sub_ss;
static s7_pointer g_abs_sub_ss(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, y;
  int ret_type;

  x = finder(sc, cadr(car(args)));
  if (!s7_is_real(x))
    return(s7_wrong_type_arg_error(sc, "-", 1, x, "a real"));

  y = finder(sc, caddr(car(args)));
  if (!s7_is_real(y))
    return(s7_wrong_type_arg_error(sc, "-", 2, y, "a real"));

  if (type(x) > type(y))
    ret_type = type(x);
  else ret_type = type(y);

  switch (ret_type)
    {
    case T_INTEGER: 
      {
	s7_Int diff;
	diff = integer(x) - integer(y);
	return(make_integer(sc, s7_Int_abs(diff)));
      }
      break;
      
    case T_RATIO:
      {
	s7_pointer diff;
	diff = subtract_ratios(sc, x, y);
	if (numerator(diff) < 0)
	  return(s7_make_ratio(sc, -numerator(diff), denominator(diff)));
	return(diff);
      }
      break;
      
    case T_REAL:
      {
	s7_Double diff;
	diff = number_to_real(x) - number_to_real(y);
	return(s7_make_real(sc, s7_Double_abs(diff)));
      }
      break;
    }
  return(x);
}
#endif

static s7_pointer g_magnitude(s7_scheme *sc, s7_pointer args)
{
  #define H_magnitude "(magnitude z) returns the magnitude of z"
  s7_pointer x;
  x = car(args);

  switch (type(x))
    {
    case T_INTEGER:
      if (integer(x) < 0)
	return(make_integer(sc, -integer(x)));
      return(x);

    case T_RATIO:
      if (numerator(x) < 0)
	return(s7_make_ratio(sc, -numerator(x), denominator(x)));
      return(x);

    case T_REAL:
      if (real(x) < 0.0)
	return(s7_make_real(sc, -real(x)));
      return(x);

    case T_COMPLEX:
      return(s7_make_real(sc, hypot(imag_part(x), real_part(x))));

    default:
      return(s7_wrong_type_arg_error(sc, "magnitude", 0, x, "a number"));
    }
}


static s7_pointer g_angle(s7_scheme *sc, s7_pointer args)
{
  #define H_angle "(angle z) returns the angle of z"
  s7_pointer x;
  /* (angle inf+infi) -> 0.78539816339745 ? */

  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
      if (integer(x) < 0)
	return(real_pi);
      return(small_int(0));

    case T_RATIO:
      if (numerator(x) < 0)
	return(real_pi);
      return(small_int(0));

    case T_REAL:
      if (is_NaN(real(x))) return(x);
      if (real(x) < 0.0)
	return(real_pi);
      return(real_zero);

    case T_COMPLEX:
      return(s7_make_real(sc, atan2(imag_part(x), real_part(x))));

    default:
      return(s7_wrong_type_arg_error(sc, "angle", 0, x, "a number"));
    }
}


static s7_pointer g_rationalize(s7_scheme *sc, s7_pointer args)
{
  #define H_rationalize "(rationalize x err) returns the ratio with lowest denominator within err of x"
  s7_Double err;
  s7_pointer x;

  if (is_not_null(cdr(args)))
    {
      if (!s7_is_real(cadr(args)))
	return(s7_wrong_type_arg_error(sc, "rationalize error limit,", 2, cadr(args), "a real"));
      err = s7_number_to_real(cadr(args));
      if (is_NaN(err))
	return(s7_out_of_range_error(sc, "rationalize", 2, cadr(args), "error term is NaN"));
      if (err < 0.0) err = -err;
    }
  else err = default_rationalize_error;

  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
      {
	s7_Int a, b, pa;
	if (err < 1.0) return(x);
	a = s7_integer(x);
	if (a < 0) pa = -a; else pa = a;
	if (err >= pa) return(small_int(0));
	b = (s7_Int)err;
	pa -= b;
	if (a < 0)
	  return(make_integer(sc, -pa));
	return(make_integer(sc, pa));
      }

    case T_RATIO:
      if (err == 0.0)
	return(x);

    case T_REAL:
      {
	s7_Double rat;
	s7_Int numer = 0, denom = 1;

	rat = s7_number_to_real(x);

	if ((is_NaN(rat)) || (isinf(rat)))
	  return(s7_wrong_type_arg_error(sc, "rationalize", (is_null(cdr(args))) ? 0 : 1, x, "a normal real"));

	if (err >= s7_Double_abs(rat)) 
	  return(small_int(0));

	if ((rat > 9.2233720368548e+18) || (rat < -9.2233720368548e+18))
	  return(s7_out_of_range_error(sc, "rationalize", 1, x, "a real between most-negative-fixnum and most-positive-fixnum"));

	if ((s7_Double_abs(rat) + s7_Double_abs(err)) < 1.0e-18)
	  err = 1.0e-18;
	/* (/ 1.0 most-positive-fixnum) is 1.0842021e-19, so if we let err be less than that,
	 * (rationalize 1e-19 1e-20) hangs, but this only affects the initial ceiling, I believe.
	 */

	if (s7_Double_abs(rat) < s7_Double_abs(err))
	  return(small_int(0));

	if (c_rationalize(rat, err, &numer, &denom))
	  return(s7_make_ratio(sc, numer, denom));

	return(sc->F);
      }

    default:
      return(s7_wrong_type_arg_error(sc, "rationalize", (is_null(cdr(args))) ? 0 : 1, x, "a real"));
    }
}


static s7_pointer g_make_polar(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, y;
  s7_Double ang, mag;
  #define H_make_polar "(make-polar mag ang) returns a complex number with magnitude mag and angle ang"
  
  x = car(args);
  y = cadr(args);

  switch (type(x))
    {
    case T_INTEGER:
      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(x) == 0) return(x);            /* (make-polar 0 1) -> 0 */
	  if (integer(y) == 0) return(x);            /* (make-polar 1 0) -> 1 */
	  mag = (s7_Double)integer(x);
	  ang = (s7_Double)integer(y);
	  break;

	case T_RATIO:
	  if (integer(x) == 0) return(x);
	  mag = (s7_Double)integer(x);
	  ang = (s7_Double)fraction(y);
	  break;

	case T_REAL:
	  ang = real(y);
	  if (ang == 0.0) return(x);
	  if (is_NaN(ang)) return(y);
	  if (isinf(ang)) return(real_NaN);
	  if ((ang == M_PI) || (ang == -M_PI)) return(make_integer(sc, -integer(x)));
	  mag = (s7_Double)integer(x);
	  break;

	default:
	  return(s7_wrong_type_arg_error(sc, "make-polar angle,", 2, y, "a real"));  
	}
      break;

    case T_RATIO:
      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(y) == 0) return(x);
	  mag = (s7_Double)fraction(x);
	  ang = (s7_Double)integer(y);
	  break;

	case T_RATIO:
	  mag = (s7_Double)fraction(x);
	  ang = (s7_Double)fraction(y);
	  break;

	case T_REAL:
	  ang = real(y);
	  if (ang == 0.0) return(x);
	  if (is_NaN(ang)) return(y);
	  if (isinf(ang)) return(real_NaN);
	  if ((ang == M_PI) || (ang == -M_PI)) return(s7_make_ratio(sc, -numerator(x), denominator(x)));
	  mag = fraction(x);
	  break;

	default:
	  return(s7_wrong_type_arg_error(sc, "make-polar angle,", 2, y, "a real"));  
	}
      break;

    case T_REAL:
      mag = real(x);
      if (is_NaN(mag)) return(x);
      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(y) == 0) return(x);
	  ang = (s7_Double)integer(y);
	  break;

	case T_RATIO:
	  ang = (s7_Double)fraction(y);
	  break;

	case T_REAL:
	  ang = real(y);
	  if (ang == 0.0) return(x);
	  if (is_NaN(ang)) return(y);
	  if (isinf(ang)) return(real_NaN);
	  break;

	default:
	  return(s7_wrong_type_arg_error(sc, "make-polar angle,", 2, y, "a real"));  
	}
      break;

    default:
      return(s7_wrong_type_arg_error(sc, "make-polar magnitude,", 1, x, "a real"));
    }

  return(s7_make_complex(sc, mag * cos(ang), mag * sin(ang)));

  /* since sin is inaccurate for large arguments, so is make-polar:
   *    (make-polar 1.0 1e40) -> -0.76267273202438+0.64678458842683i, not 8.218988919070239214448025364432557517335E-1-5.696334009536363273080341815735687231337E-1i
   */
}


static s7_pointer g_make_rectangular(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, y;
  #define H_make_rectangular "(make-rectangular x1 x2) returns a complex number with real-part x1 and imaginary-part x2"
  
  x = car(args);
  y = cadr(args);

  switch (type(y))
    {
    case T_INTEGER:
      switch (type(x))
	{
	case T_INTEGER: if (integer(y) == 0) return(x); return(s7_make_complex(sc, (s7_Double)integer(x), (s7_Double)integer(y)));
	case T_RATIO:   if (integer(y) == 0) return(x); return(s7_make_complex(sc, (s7_Double)fraction(x), (s7_Double)integer(y)));
	case T_REAL:    if (integer(y) == 0) return(x); return(s7_make_complex(sc, real(x), (s7_Double)integer(y)));
	default:        return(s7_wrong_type_arg_error(sc, "make-rectangular real part,", 1, x, "a real"));
	}
      
    case T_RATIO:
      switch (type(x))
	{
	case T_INTEGER: return(s7_make_complex(sc, (s7_Double)integer(x), (s7_Double)fraction(y)));
	case T_RATIO:   return(s7_make_complex(sc, (s7_Double)fraction(x), (s7_Double)fraction(y)));
	case T_REAL:    return(s7_make_complex(sc, real(x), (s7_Double)fraction(y)));
	default:        return(s7_wrong_type_arg_error(sc, "make-rectangular real part,", 1, x, "a real"));
	}

    case T_REAL:
      switch (type(x))
	{
	case T_INTEGER: if (real(y) == 0.0) return(x); return(s7_make_complex(sc, (s7_Double)integer(x), real(y)));
	case T_RATIO:   if (real(y) == 0.0) return(x); return(s7_make_complex(sc, (s7_Double)fraction(x), real(y)));
	case T_REAL:    if (real(y) == 0.0) return(x); return(s7_make_complex(sc, real(x), real(y)));
	default:        return(s7_wrong_type_arg_error(sc, "make-rectangular real part,", 1, x, "a real"));
	}

    default:
      return(s7_wrong_type_arg_error(sc, "make-rectangular imaginary part,", 2, y, "a real"));  
    }
}


static s7_pointer g_exp(s7_scheme *sc, s7_pointer args)
{
  #define H_exp "(exp z) returns e^z, (exp 1) is 2.718281828459"
  s7_pointer x;

  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
      if (integer(x) == 0) return(small_int(1));                       /* (exp 0) -> 1 */
      return(s7_make_real(sc, exp((s7_Double)(integer(x)))));

    case T_RATIO:
      return(s7_make_real(sc, exp((s7_Double)fraction(x))));

    case T_REAL:
      return(s7_make_real(sc, exp(real(x))));

    case T_COMPLEX:
      return(s7_from_c_complex(sc, cexp(as_c_complex(x))));
      /* this is inaccurate for large arguments:
       *   (exp 0+1e20i) -> -0.66491178990701-0.74692189125949i, not 7.639704044417283004001468027378811228331E-1-6.45251285265780844205811711312523007406E-1i
       */

    default:
      return(s7_wrong_type_arg_error(sc, "exp", 0, x, "a number"));
    }
}


static s7_pointer g_log(s7_scheme *sc, s7_pointer args)
{
  #define H_log "(log z1 (z2 e)) returns log(z1) / log(z2) where z2 (the base) defaults to e: (log 8 2) = 3"
  s7_pointer x;
  
  x = car(args);

  if (!s7_is_number(x))
    return(s7_wrong_type_arg_error(sc, "log", (is_null(cdr(args))) ? 0 : 1, x, "a number"));

  if ((is_pair(cdr(args))) &&
      (!(s7_is_number(cadr(args)))))
    return(s7_wrong_type_arg_error(sc, "log base,", 2, cadr(args), "a number"));

  if (is_pair(cdr(args)))
    {
      s7_pointer y;

      y = cadr(args);
      if ((x == small_int(1)) && (y == small_int(1)))  /* (log 1 1) -> 0 (this is NaN in the bignum case) */
	return(small_int(0));

      /* (log 1 0) must be 0 since everyone says (expt 0 0) is 1 */
      if (s7_is_zero(y))
	{
	  if ((y == small_int(0)) &&
	      (x == small_int(1)))
	    return(y);
	  return(s7_out_of_range_error(sc, "log base,", 2, y, "can't be 0"));
	}

      if (s7_is_one(y))      /* this used to raise an error, but the bignum case is simpler if we return inf */
	{
	  if (s7_is_one(x))  /* but (log 1.0 1.0) -> 0.0 */
	    return(real_zero);
	  return(s7_make_real(sc, INFINITY));
	}

      if ((s7_is_real(x)) &&
	  (s7_is_real(y)) &&
	  (s7_is_positive(x)) &&
	  (s7_is_positive(y)))
	{
	  if ((s7_is_rational(x)) &&
	      (s7_is_rational(y)))
	    {
	      s7_Double res;
	      s7_Int ires;
	      res = log(number_to_real(x)) / log(number_to_real(y));
	      ires = (s7_Int)res;
	      if (res - ires == 0.0)
		return(make_integer(sc, ires));   /* (log 8 2) -> 3 or (log 1/8 2) -> -3 */
	      return(s7_make_real(sc, res));         /* perhaps use rationalize here? (log 2 8) -> 1/3 */
	    }
	  return(s7_make_real(sc, log(number_to_real(x)) / log(number_to_real(y))));
	}
      return(s7_from_c_complex(sc, clog(s7_complex(x)) / clog(s7_complex(y))));
    }
  
  if (s7_is_real(x))
    {
      if (s7_is_positive(x))
	return(s7_make_real(sc, log(number_to_real(x))));
      return(s7_make_complex(sc, log(-number_to_real(x)), M_PI));
    }

  return(s7_from_c_complex(sc, clog(s7_complex(x))));
}


static s7_pointer g_sin(s7_scheme *sc, s7_pointer args)
{
  #define H_sin "(sin z) returns sin(z)"
  s7_pointer x;

  x = car(args);
  switch (type(x))
    {
    case T_REAL:
      return(s7_make_real(sc, sin(real(x))));

    case T_INTEGER:
      if (integer(x) == 0) return(small_int(0));                      /* (sin 0) -> 0 */
      return(s7_make_real(sc, sin((s7_Double)integer(x))));
      
    case T_RATIO:
      return(s7_make_real(sc, sin((s7_Double)(fraction(x)))));

    case T_COMPLEX:
#if WITH_COMPLEX
      return(s7_from_c_complex(sc, csin(as_c_complex(x))));
#else
      return(real_zero);
#endif

    default:
      return(s7_wrong_type_arg_error(sc, "sin", 0, x, "a number"));
    }

  /* sin is totally inaccurate over about 1e18.  There's a way to get true results,
   *   but it involves fancy "range reduction" techniques. 
   *   This mean lots of things are inaccurate:
   * (sin (remainder 1e22 (* 2 pi)))
   * -0.57876806033477
   * but it should be -8.522008497671888065747423101326159661908E-1
   * ---
   * (remainder 1e22 (* 2 pi)) -> 1.0057952155665e+22 !!
   *   it should be 5.263007914620499494429139986095833592117E0
   */
}


static s7_pointer g_cos(s7_scheme *sc, s7_pointer args)
{
  #define H_cos "(cos z) returns cos(z)"
  s7_pointer x;

  x = car(args);
  switch (type(x))
    {
    case T_REAL:
      return(s7_make_real(sc, cos(real(x))));

    case T_INTEGER:
      if (integer(x) == 0) return(small_int(1));                     /* (cos 0) -> 1 */
      return(s7_make_real(sc, cos((s7_Double)integer(x))));
      
    case T_RATIO:
      return(s7_make_real(sc, cos((s7_Double)(fraction(x)))));

    case T_COMPLEX:
#if WITH_COMPLEX
      return(s7_from_c_complex(sc, ccos(as_c_complex(x))));
#else
      return(real_zero);
#endif

    default:
      return(s7_wrong_type_arg_error(sc, "cos", 0, x, "a number"));
    }
}


static s7_pointer g_tan(s7_scheme *sc, s7_pointer args)
{
  #define H_tan "(tan z) returns tan(z)"
  s7_pointer x;

  x = car(args);
  switch (type(x))
    {
    case T_REAL:
      return(s7_make_real(sc, tan(real(x))));
      
    case T_INTEGER:
      if (integer(x) == 0) return(small_int(0));                      /* (tan 0) -> 0 */
      return(s7_make_real(sc, tan((s7_Double)(integer(x)))));

    case T_RATIO:
      return(s7_make_real(sc, tan((s7_Double)(fraction(x)))));

    case T_COMPLEX:
      if (imag_part(x) > 350.0)
	return(s7_make_complex(sc, 0.0, 1.0));
      if (imag_part(x) < -350.0)
	return(s7_make_complex(sc, 0.0, -1.0));
      return(s7_from_c_complex(sc, ctan(as_c_complex(x))));

    default:
      return(s7_wrong_type_arg_error(sc, "tan", 0, x, "a number"));
    }
}


static s7_pointer g_asin(s7_scheme *sc, s7_pointer args)
{
  #define H_asin "(asin z) returns asin(z); (sin (asin 1)) = 1"
  s7_pointer n;

  n = car(args);
  switch (type(n))
    {
    case T_INTEGER:
      if (integer(n) == 0) return(small_int(0));                    /* (asin 0) -> 0 */
      
    case T_REAL:
    case T_RATIO:
      {
	s7_Double x, absx, recip;
	s7_Complex result;
	x = number_to_real(n);
	absx = s7_Double_abs(x);
	if (absx <= 1.0)
	  return(s7_make_real(sc, asin(x)));
      
	/* otherwise use maxima code: */
	recip = 1.0 / absx;
	result = (M_PI / 2.0) - (_Complex_I * clog(absx * (1.0 + (sqrt(1.0 + recip) * csqrt(1.0 - recip)))));
	if (x < 0.0)
	  return(s7_from_c_complex(sc, -result));
	return(s7_from_c_complex(sc, result));
      }

    case T_COMPLEX:
#if WITH_COMPLEX
      /* if either real or imag part is very large, use explicit formula, not casin */
      /*   this code taken from sbcl's src/code/irrat.lisp */
      /* break is around x+70000000i */

      if ((s7_Double_abs(real_part(n)) > 1.0e7) ||
	  (s7_Double_abs(imag_part(n)) > 1.0e7))
	{
	  s7_Complex sq1mz, sq1pz, z;
	  z = as_c_complex(n);
	  sq1mz = csqrt(1.0 - z);
	  sq1pz = csqrt(1.0 + z);
	  return(s7_make_complex(sc, 
				 atan(real_part(n) / creal(sq1mz * sq1pz)),
				 asinh(cimag(sq1pz * conj(sq1mz)))));
	}
#endif
      return(s7_from_c_complex(sc, casin(as_c_complex(n))));

    default:
      return(s7_wrong_type_arg_error(sc, "asin", 0, n, "a number"));
    }
}


static s7_pointer g_acos(s7_scheme *sc, s7_pointer args)
{
  #define H_acos "(acos z) returns acos(z); (cos (acos 1)) = 1"
  s7_pointer n;

  n = car(args);
  switch (type(n))
    {
    case T_INTEGER:
      if (integer(n) == 1) return(small_int(0));

    case T_REAL:
    case T_RATIO:
      {
	s7_Double x, absx, recip;
	s7_Complex result;
	x = number_to_real(n);
	absx = s7_Double_abs(x);
	if (absx <= 1.0)
	  return(s7_make_real(sc, acos(x)));
      
	/* else follow maxima again: */
	recip = 1.0 / absx;
	if (x > 0.0)
	  result = _Complex_I * clog(absx * (1.0 + (sqrt(1.0 + recip) * csqrt(1.0 - recip))));
	else result = M_PI - _Complex_I * clog(absx * (1.0 + (sqrt(1.0 + recip) * csqrt(1.0 - recip))));
	return(s7_from_c_complex(sc, result));
      }

    case T_COMPLEX:
#if WITH_COMPLEX
      /* if either real or imag part is very large, use explicit formula, not cacos */
      /*   this code taken from sbcl's src/code/irrat.lisp */

      if ((s7_Double_abs(real_part(n)) > 1.0e7) ||
	  (s7_Double_abs(imag_part(n)) > 1.0e7))
	{
	  s7_Complex sq1mz, sq1pz, z;
	  z = as_c_complex(n);
	  sq1mz = csqrt(1.0 - z);
	  sq1pz = csqrt(1.0 + z);
	  return(s7_make_complex(sc, 
				 2.0 * atan(creal(sq1mz) / creal(sq1pz)),
				 asinh(cimag(sq1mz * conj(sq1pz)))));
	}
#endif
      return(s7_from_c_complex(sc, cacos(s7_complex(n))));

    default:
      return(s7_wrong_type_arg_error(sc, "acos", 0, n, "a number"));
    }
}


static s7_pointer g_atan(s7_scheme *sc, s7_pointer args)
{
  #define H_atan "(atan z) returns atan(z), (atan y x) returns atan(y/x)"
  s7_pointer x, y;

  /* currently (atan inf.0 inf.0) -> 0.78539816339745, and (atan inf.0 -inf.0) -> 2.3561944901923 (etc) */

  x = car(args);
  if (!is_pair(cdr(args)))
    {
      switch (type(x))
	{
	case T_INTEGER:
	  if (integer(x) == 0) return(small_int(0));                /* (atan 0) -> 0 */
	  
	case T_REAL:
	case T_RATIO:
	  return(s7_make_real(sc, atan(number_to_real(x))));

	case T_COMPLEX:
	  return(s7_from_c_complex(sc, catan(as_c_complex(x))));

	default:
	  return(s7_wrong_type_arg_error(sc, "atan", 0, x, "a number"));
	}
    }

  if (!s7_is_real(x))
    return(s7_wrong_type_arg_error(sc, "atan", 1, x, "a real"));

  y = cadr(args);
  if (!s7_is_real(y))
    return(s7_wrong_type_arg_error(sc, "atan", 2, y, "a real"));

  return(s7_make_real(sc, atan2(number_to_real(x), number_to_real(y))));
}  


static s7_pointer g_sinh(s7_scheme *sc, s7_pointer args)
{
  #define H_sinh "(sinh z) returns sinh(z)"
  s7_pointer x;

  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
      if (integer(x) == 0) return(small_int(0));                    /* (sinh 0) -> 0 */

    case T_REAL:
    case T_RATIO:
      return(s7_make_real(sc, sinh(number_to_real(x))));

    case T_COMPLEX:
      return(s7_from_c_complex(sc, csinh(as_c_complex(x))));

    default:
      return(s7_wrong_type_arg_error(sc, "sinh", 0, x, "a number"));
    }
}


static s7_pointer g_cosh(s7_scheme *sc, s7_pointer args)
{
  #define H_cosh "(cosh z) returns cosh(z)"
  s7_pointer x;

  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
      if (integer(x) == 0) return(small_int(1));                   /* (cosh 0) -> 1 */

    case T_REAL:
    case T_RATIO:
      /* this is not completely correct when optimization kicks in.
       * :(define (hi) (do ((i 0 (+ i 1))) ((= i 1)) (display (cosh i))))
       * hi
       * :(hi)
       * 1.0()
       * :(cosh 0)
       * 1
       */
      return(s7_make_real(sc, cosh(number_to_real(x))));

    case T_COMPLEX:
      return(s7_from_c_complex(sc, ccosh(as_c_complex(x))));

    default:
      return(s7_wrong_type_arg_error(sc, "cosh", 0, x, "a number"));
    }
}


static s7_pointer g_tanh(s7_scheme *sc, s7_pointer args)
{
  #define H_tanh "(tanh z) returns tanh(z)"
  s7_pointer x;

  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
      if (integer(x) == 0) return(small_int(0));                   /* (tanh 0) -> 0 */

    case T_REAL:
    case T_RATIO:
      return(s7_make_real(sc, tanh(number_to_real(x))));

    case T_COMPLEX:
      if (real_part(x) > 350.0)
	return(real_one);               /* closer than 0.0 which is what ctanh is about to return! */
      if (real_part(x) < -350.0)
	return(s7_make_real(sc, -1.0));              /* closer than -0.0 which is what ctanh is about to return! */
      return(s7_from_c_complex(sc, ctanh(as_c_complex(x))));

    default:
      return(s7_wrong_type_arg_error(sc, "tanh", 0, x, "a number"));
    }
}


static s7_pointer g_asinh(s7_scheme *sc, s7_pointer args)
{
  #define H_asinh "(asinh z) returns asinh(z)"
  s7_pointer x;

  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
      if (integer(x) == 0) return(small_int(0));

    case T_REAL:
    case T_RATIO:
      return(s7_make_real(sc, asinh(number_to_real(x))));

    case T_COMPLEX:
      return(s7_from_c_complex(sc, casinh(as_c_complex(x))));

    default:
      return(s7_wrong_type_arg_error(sc, "asinh", 0, x, "a number"));
    }
}


static s7_pointer g_acosh(s7_scheme *sc, s7_pointer args)
{
  #define H_acosh "(acosh z) returns acosh(z)"
  s7_pointer x;

  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
      if (integer(x) == 1) return(small_int(0));

    case T_REAL:
    case T_RATIO:
      if (number_to_real(x) >= 1.0)
	return(s7_make_real(sc, acosh(number_to_real(x))));

    case T_COMPLEX:
      return(s7_from_c_complex(sc, cacosh(s7_complex(x)))); /* not as_c_complex because x might not be complex */

    default:
      return(s7_wrong_type_arg_error(sc, "acosh", 0, x, "a number"));
    }
}


static s7_pointer g_atanh(s7_scheme *sc, s7_pointer args)
{
  #define H_atanh "(atanh z) returns atanh(z)"
  s7_pointer x;

  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
      if (integer(x) == 0) return(small_int(0));                    /* (atanh 0) -> 0 */

    case T_REAL:
    case T_RATIO:
      if (s7_Double_abs(number_to_real(x)) < 1.0)
	return(s7_make_real(sc, atanh(number_to_real(x))));

      /* if we can't distinguish x from 1.0 even with long doubles, we'll get inf.0:
       *    (atanh 9223372036854775/9223372036854776) -> 18.714973875119
       *    (atanh 92233720368547758/92233720368547757) -> inf.0
       */

    case T_COMPLEX:
      return(s7_from_c_complex(sc, catanh(s7_complex(x))));

    default:
      return(s7_wrong_type_arg_error(sc, "atanh", 0, x, "a number"));
    }
}


static s7_pointer g_sqrt(s7_scheme *sc, s7_pointer args)
{
  #define H_sqrt "(sqrt z) returns the square root of z"
  s7_pointer n;
  s7_Double sqx;

  n = car(args);
  switch (type(n))
    {
    case T_INTEGER:
      if (integer(n) >= 0)
	{
	  s7_Int ix;
	  sqx = sqrt((s7_Double)integer(n));
	  ix = (s7_Int)sqx;
	  if ((ix * ix) == integer(n))
	    return(make_integer(sc, ix));
	  return(s7_make_real(sc, sqx));
	}
      sqx = (s7_Double)integer(n); /* we're trying to protect against (sqrt -9223372036854775808) where we can't negate the integer argument */
      return(s7_make_complex(sc, 0.0, sqrt((s7_Double)(-sqx))));
      break;

    case T_RATIO:
      sqx = fraction(n);
      if (sqx > 0.0) /* else it's complex, so it can't be a ratio */
	{
	  s7_Int nm = 0, dn = 1;
	  if (c_rationalize(sqx, 1.0e-16, &nm, &dn)) /* 1e-16 so that (sqrt 1/1099511627776) returns 1/1048576 */
	    {
	      if ((nm * nm == numerator(n)) &&
		  (dn * dn == denominator(n)))
		return(s7_make_ratio(sc, nm, dn));
	    }
	  return(s7_make_real(sc, sqrt(sqx)));
	}
      return(s7_make_complex(sc, 0.0, sqrt(-sqx)));
      break;

    case T_REAL:
      if (is_NaN(real(n)))
	return(real_NaN);
      if (real(n) >= 0.0)
	return(s7_make_real(sc, sqrt(real(n))));
      return(s7_make_complex(sc, 0.0, sqrt(-real(n))));
      break;

    case T_COMPLEX:
      return(s7_from_c_complex(sc, csqrt(as_c_complex(n))));
      break;

    default:
      return(s7_wrong_type_arg_error(sc, "sqrt", 0, n, "a number"));
    }
}


static s7_Int int_to_int(s7_Int x, s7_Int n)
{
  /* from GSL */
  s7_Int value = 1;
  do {
    if (n & 1) value *= x;
    n >>= 1;
    x *= x;
  } while (n);
  return(value);
}


static long long int nth_roots[63] = {
  S7_LLONG_MAX, S7_LLONG_MAX, 3037000499LL, 2097151, 55108, 6208, 1448, 511, 234, 127, 78, 52, 38, 28, 22, 
  18, 15, 13, 11, 9, 8, 7, 7, 6, 6, 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 
  2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2};

static long int_nth_roots[31] = {
  S7_LONG_MAX, S7_LONG_MAX, 46340, 1290, 215, 73, 35, 21, 14, 10, 8, 7, 5, 5, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2};

static bool int_pow_ok(s7_Int x, s7_Int y)
{
  if (s7_int_bits > 31)
    return((y < 63) &&
	   (nth_roots[y] >= s7_Int_abs(x)));
  return((y < 31) &&
	 (int_nth_roots[y] >= s7_Int_abs(x)));
}


static s7_pointer g_expt(s7_scheme *sc, s7_pointer args)
{
  #define H_expt "(expt z1 z2) returns z1^z2"
  s7_pointer n, pw;
  
  n = car(args);
  if (!s7_is_number(n))
    return(s7_wrong_type_arg_error(sc, "expt", 1, n, "a number"));

  pw = cadr(args);
  if (!s7_is_number(pw))
    return(s7_wrong_type_arg_error(sc, "expt power,", 2, pw, "a number"));

  /* this provides more than 2 args to expt:
   *  if (is_not_null(cddr(args)))
   *    return(g_expt(sc, list_2(sc, car(args), g_expt(sc, cdr(args)))));
   *
   * but it's unusual in scheme to process args in reverse order, and the
   * syntax by itself is ambiguous (does (expt 2 2 3) = 256 or 64?)
   */

  if (s7_is_zero(n))
    {
      if (s7_is_zero(pw))
	{
	  if ((s7_is_integer(n)) && (s7_is_integer(pw)))       /* (expt 0 0) -> 1 */
	    return(small_int(1));
	  return(real_zero);                                   /* (expt 0.0 0) -> 0.0 */
	}

      if (s7_is_real(pw))
	{
	  if (s7_is_negative(pw))                              /* (expt 0 -1) */
	    return(division_by_zero_error(sc, "expt", args));  
	  /* (Clisp gives divide-by-zero error here, Guile returns inf.0) */

	  if ((!s7_is_rational(pw)) &&                         /* (expt 0 most-positive-fixnum) */
	      (is_NaN(s7_real(pw))))                           /* (expt 0 +nan.0) */
	    return(pw);
	}
      else
	{                                                      /* (expt 0 a+bi) */
	  if (real_part(pw) < 0.0)                             /* (expt 0 -1+i) */
	    return(division_by_zero_error(sc, "expt", args));  
	}

      if ((s7_is_integer(n)) && (s7_is_integer(pw)))           /* pw != 0, (expt 0 2312) */
	return(small_int(0));
      return(real_zero);                                       /* (expt 0.0 123123) */
    }

  if (s7_is_one(pw))
    {
      if (s7_is_integer(pw))
	return(n);
      if (is_rational(n))
	return(s7_make_real(sc, number_to_real(n)));
      return(n);
    }
  
  if (type(pw) == T_INTEGER)
    {
      s7_Int y;
      y = s7_integer(pw);
      if (y == 0)
	{
	  /* (expt +nan.0 0) ?? */
	  if (is_rational(n))
	    return(small_int(1));
	  return(real_one);
	}

      switch (type(n))
	{
	case T_INTEGER:
	  {
	    s7_Int x;
	    x = s7_integer(n);
	    if (x == 1)
	      return(n);
	    
	    if (x == -1)
	      {
		if (y == S7_LLONG_MIN)
		  return(small_int(1));
		
		if (s7_Int_abs(y) & 1)
		  return(n);
		return(small_int(1));
	      }
	    
	    if (y == S7_LLONG_MIN)
	      return(small_int(0));                      /* (expt x most-negative-fixnum) !! */
	    
	    if (int_pow_ok(x, s7_Int_abs(y)))
	      {
		if (y > 0)
		  return(make_integer(sc, int_to_int(x, y)));
		return(s7_make_ratio(sc, 1, int_to_int(x, -y)));
	      }
	  }
	  break;

	case T_RATIO:
	  {
	    s7_Int nm, dn;
	    
	    nm = numerator(n);
	    dn = denominator(n);
	    
	    if (y == S7_LLONG_MIN)
	      {
		if (s7_Int_abs(nm) > dn)
		  return(small_int(0));              /* (expt 4/3 most-negative-fixnum) -> 0? */
		return(s7_make_real(sc, INFINITY));  /* (expt 3/4 most-negative-fixnum) -> inf? */
	      }
	    
	    if ((int_pow_ok(nm, s7_Int_abs(y))) &&
		(int_pow_ok(dn, s7_Int_abs(y))))
	      {
		if (y > 0)
		  return(s7_make_ratio(sc, int_to_int(nm, y), int_to_int(dn, y)));
		return(s7_make_ratio(sc, int_to_int(dn, -y), int_to_int(nm, -y)));
	      }
	  }
	  break;
	  /* occasionally int^rat can be int but it happens so infrequently it's probably not worth checking
	   *  one possibly easy case: (expt 1 1/2) -> 1 (-1?) etc
	   */

	case T_REAL:
	  /* (expt -1.0 most-positive-fixnum) should be -1.0
	   * (expt -1.0 (+ (expt 2 53) 1)) -> -1.0
	   * (expt -1.0 (- 1 (expt 2 54))) -> -1.0
	   */
	  if (s7_real(n) == -1.0)
	    {
	      if (y == S7_LLONG_MIN)
		return(real_one);
		
	      if (s7_Int_abs(y) & 1)
		return(n);
	      return(real_one);
	    }
	  break;

	case T_COMPLEX:
	  if ((s7_real_part(n) == 0.0) &&
	      ((s7_imag_part(n) == 1.0) ||
	       (s7_imag_part(n) == -1.0)))
	    {
	      bool yp, np;
	      yp = (y > 0);
	      np = (s7_imag_part(n) > 0.0);
	      switch (s7_Int_abs(y) % 4)
		{
		case 0: return(real_one);
		case 1: return(s7_make_complex(sc, 0.0, (yp == np) ? 1.0 : -1.0));
		case 2: return(s7_make_real(sc, -1.0));
		case 3: return(s7_make_complex(sc, 0.0, (yp == np) ? -1.0 : 1.0));
		}
	    }
	  break;
	}
    }

  if ((s7_is_real(n)) &&
      (s7_is_real(pw)))
    {
      s7_Double x, y;

      if ((type(pw) == T_RATIO) &&
	  (numerator(pw) == 1))
	{
	  if (denominator(pw) == 2)
	    return(g_sqrt(sc, args));
	  if (denominator(pw) == 3)
	    return(s7_make_real(sc, cbrt(number_to_real(n)))); /* (expt 27 1/3) should be 3, not 3.0... */
 
	  /* but: (expt 512/729 1/3) -> 0.88888888888889
	   */
	  /* and 4 -> sqrt(sqrt...) etc? */
	}

      x = number_to_real(n);
      y = number_to_real(pw);

      if (is_NaN(x)) return(n);
      if (is_NaN(y)) return(pw);
      if (y == 0.0) return(real_one);

      if ((x > 0.0) ||
	  ((y - floor(y)) < 1.0e-16))
	return(s7_make_real(sc, pow(x, y)));
    }
  
  /* (expt 0+i 1e+16) = 0.98156860153485-0.19111012657867i ? 
   * (expt 0+i 1+1/0i) = 0.0 ??
   */
  return(s7_from_c_complex(sc, cpow(s7_complex(n), s7_complex(pw))));
}


static s7_Int c_lcm(s7_Int a, s7_Int b)
{
  if ((a == 0) || (b == 0)) return(0);
  if (a < 0) a = -a;
  if (b < 0) b = -b;
  return((a / c_gcd(a, b)) * b);
}


static s7_pointer g_lcm(s7_scheme *sc, s7_pointer args)
{
  #define H_lcm "(lcm ...) returns the least common multiple of its rational arguments"
  s7_Int n = 1, d = 0;
  bool rats = false;
  s7_pointer x;

  if (!is_pair(args))
    return(small_int(1));

  if (!is_pair(cdr(args)))
    {
      if (!is_rational(car(args)))
	return(s7_wrong_type_arg_error(sc, "lcm", 1, car(args), "an integer or ratio"));
      return(g_abs(sc, args));
    }

  for (x = args; is_not_null(x); x = cdr(x)) 
    if (!is_rational(car(x)))
      return(s7_wrong_type_arg_error(sc, "lcm", position_of(x, args), car(x), "an integer or ratio"));
    else rats = ((rats) || (type(car(x)) == T_RATIO));

  if (!rats)
    {
      for (x = args; is_not_null(x); x = cdr(x)) 
	{
	  n = c_lcm(n, integer(car(x)));
	  if (n < 0) return(s7_out_of_range_error(sc, "lcm from", 0, args, "result is too large"));
	  if (n == 0)
	    return(small_int(0));
	}
      return(make_integer(sc, n));
    }

  /* from A Jaffer */
  for (x = args; is_not_null(x); x = cdr(x)) 
    {
      n = c_lcm(n, s7_numerator(car(x)));
      if (n < 0) return(s7_out_of_range_error(sc, "lcm from", 0, args, "result is too large"));
      if (n == 0)
	return(small_int(0));
      d = c_gcd(d, s7_denominator(car(x)));
    }
  return(s7_make_ratio(sc, n, d));
}


static s7_pointer g_gcd(s7_scheme *sc, s7_pointer args)
{
  #define H_gcd "(gcd ...) returns the greatest common divisor of its rational arguments"
  bool rats = false;
  s7_Int n = 0, d = 1;
  s7_pointer x;

  if (!is_pair(args))
    return(small_int(0));

  if (!is_pair(cdr(args)))
    {
      if (!is_rational(car(args)))
	return(s7_wrong_type_arg_error(sc, "gcd", 1, car(args), "an integer or ratio"));
      return(g_abs(sc, args));
    }

  for (x = args; is_not_null(x); x = cdr(x)) 
    if (!is_rational(car(x)))
      return(s7_wrong_type_arg_error(sc, "gcd", position_of(x, args), car(x), "an integer"));
    else rats = ((rats) || (type(car(x)) == T_RATIO));
  
  if (!rats)
    {
      for (x = args; is_not_null(x); x = cdr(x)) 
	{
	  n = c_gcd(n, integer(car(x)));
	  if (n < 0) return(s7_out_of_range_error(sc, "gcd from", 0, args, "intermediate result is too large"));
	  if (n == 1) return(small_int(1));
	}
      return(make_integer(sc, n));
    }

  /* from A Jaffer */
  for (x = args; is_not_null(x); x = cdr(x)) 
    {
      n = c_gcd(n, s7_numerator(car(x)));
      if (n < 0) return(s7_out_of_range_error(sc, "gcd from", 0, args, "intermediate result is too large"));
      d = c_lcm(d, s7_denominator(car(x)));
      if (d < 0) return(s7_out_of_range_error(sc, "gcd from", 0, args, "intermediate result is too large"));
    }
  return(s7_make_ratio(sc, n, d));
}


static s7_pointer s7_truncate(s7_scheme *sc, const char *caller, s7_Double xf)   /* can't use "truncate" -- it's in unistd.h */
{
  if ((xf > S7_LLONG_MAX) ||
      (xf < S7_LLONG_MIN))
    return(s7_out_of_range_error(sc, caller, 0, s7_make_real(sc, xf), "too large to express as an integer"));

  if (xf > 0.0)
    return(make_integer(sc, (s7_Int)floor(xf)));
  return(make_integer(sc, (s7_Int)ceil(xf)));
}



static s7_pointer g_quotient(s7_scheme *sc, s7_pointer args)
{
  #define H_quotient "(quotient x1 x2) returns the integer quotient of x1 and x2; (quotient 4 3) = 1"
  /* (define (quo x1 x2) (truncate (/ x1 x2))) ; slib 
   */  
  s7_pointer x, y;
  s7_Int d1, d2, n1, n2;

  x = car(args);
  y = cadr(args);

  switch (type(x))
    {
    case T_INTEGER:
      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(y) == 0) 
	    return(division_by_zero_error(sc, "quotient", args));
	  return(make_integer(sc, integer(x) / integer(y)));

	case T_RATIO:
	  n1 = integer(x);
	  d1 = 1;
	  n2 = numerator(y);
	  d2 = denominator(y);
	  goto RATIO_QUO_RATIO;

	case T_REAL:
	  if (real(y) == 0.0) 
	    return(division_by_zero_error(sc, "quotient", args));
	  if ((isinf(real(y))) || (is_NaN(real(y))))
	    return(s7_wrong_type_arg_error(sc, "quotient", 2, y, "a normal real"));
	  return(s7_truncate(sc, "quotient", (s7_Double)integer(x) / real(y)));

	default:        
	  return(s7_wrong_type_arg_error(sc, "quotient", 2, y, "a real"));
	}

    case T_RATIO:
      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(y) == 0) 
	    return(division_by_zero_error(sc, "quotient", args));
	  n1 = numerator(x);
	  d1 = denominator(x);
	  n2 = integer(y);
	  d2 = 1;
	  goto RATIO_QUO_RATIO;

	case T_RATIO:
	  n1 = numerator(x);
	  d1 = denominator(x);
	  n2 = numerator(y);
	  d2 = denominator(y);
	RATIO_QUO_RATIO:
	  if (d1 == d2)
	    return(make_integer(sc, n1 / n2));              /* (quotient 3/9223372036854775807 1/9223372036854775807) */
	  if (n1 == n2)
	    return(make_integer(sc, d2 / d1));              /* (quotient 9223372036854775807/2 9223372036854775807/8) */

	  if ((integer_length(n1) + integer_length(d2) >= s7_int_bits) ||
	      (integer_length(n2) + integer_length(d1) >= s7_int_bits))
	    return(s7_truncate(sc, "quotient", ((long double)n1 / (long double)n2) * ((long double)d2 / (long double)d1)));
	  /* this can lose:
	   *   (quotient 1 2305843009213693952/4611686018427387903) -> 2, not 1
	   *   (quotient 21053343141/6701487259 3587785776203/1142027682075) -> 1, not 0
	   */
	  return(make_integer(sc, (n1 * d2) / (n2 * d1)));  /* (quotient 922337203685477580 1/3) */

	case T_REAL:
	  if (real(y) == 0.0) 
	    return(division_by_zero_error(sc, "quotient", args));
	  if ((isinf(real(y))) || (is_NaN(real(y))))
	    return(s7_wrong_type_arg_error(sc, "quotient", 2, y, "a normal real"));
	  return(s7_truncate(sc, "quotient", (s7_Double)fraction(x) / real(y)));

	default:        
	  return(s7_wrong_type_arg_error(sc, "quotient", 2, y, "a real"));
	}

    case T_REAL:
      if ((isinf(real(x))) || (is_NaN(real(x))))
	return(s7_wrong_type_arg_error(sc, "quotient", 1, x, "a normal real"));

      /* if infs allowed we need to return infs/nans, else:
       *    (quotient inf.0 1e-309) -> -9223372036854775808
       *    (quotient inf.0 inf.0) -> -9223372036854775808
       */

      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(y) == 0) 
	    return(division_by_zero_error(sc, "quotient", args));
	  return(s7_truncate(sc, "quotient", real(x) / (s7_Double)integer(y)));

	case T_RATIO:
	  return(s7_truncate(sc, "quotient", real(x) / (s7_Double)fraction(y)));

	case T_REAL:
	  if (real(y) == 0.0) 
	    return(division_by_zero_error(sc, "quotient", args));
	  if ((isinf(real(y))) || (is_NaN(real(y))))
	    return(s7_wrong_type_arg_error(sc, "quotient", 2, y, "a normal real"));
	  return(s7_truncate(sc, "quotient", real(x) / real(y)));

	default:        
	  return(s7_wrong_type_arg_error(sc, "quotient", 2, y, "a real"));
	}

    default:
      return(s7_wrong_type_arg_error(sc, "quotient", 1, x, "a real"));
    }
}


static s7_pointer g_remainder(s7_scheme *sc, s7_pointer args)
{
  #define H_remainder "(remainder x1 x2) returns the integer remainder of x1 and x2; (remainder 10 3) = 1"
  /* (define (rem x1 x2) (- x1 (* x2 (quo x1 x2)))) ; slib 
   */
  s7_pointer x, y;
  s7_Int quo, d1, d2, n1, n2;
  s7_Double pre_quo;

  x = car(args);
  y = cadr(args);

  switch (type(x))
    {
    case T_INTEGER:
      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(y) == 0) 
	    return(division_by_zero_error(sc, "remainder", args));
	  return(make_integer(sc, integer(x) % integer(y)));

	case T_RATIO:
	  n1 = integer(x);
	  d1 = 1;
	  n2 = numerator(y);
	  d2 = denominator(y);
	  goto RATIO_REM_RATIO;

	case T_REAL:
	  if (real(y) == 0.0) 
	    return(division_by_zero_error(sc, "remainder", args));
	  if ((isinf(real(y))) || (is_NaN(real(y))))
	    return(s7_wrong_type_arg_error(sc, "remainder", 2, y, "a normal real"));

	  pre_quo = (s7_Double)integer(x) / real(y);
	  if ((pre_quo > S7_LLONG_MAX) || (pre_quo < S7_LLONG_MIN))
	    return(s7_out_of_range_error(sc, "remainder", 0, args, "too large to express as an integer"));
	  if (pre_quo > 0.0) quo = (s7_Int)floor(pre_quo); else quo = (s7_Int)ceil(pre_quo);
	  return(s7_make_real(sc, integer(x) - real(y) * quo));

	default:        
	  return(s7_wrong_type_arg_error(sc, "remainder", 2, y, "a real"));
	}

    case T_RATIO:
      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(y) == 0) 
	    return(division_by_zero_error(sc, "remainder", args));
	  n1 = numerator(x);
	  d1 = denominator(x);
	  n2 = integer(y);
	  d2 = 1;
	  goto RATIO_REM_RATIO;

	case T_RATIO:
	  n1 = numerator(x);
	  d1 = denominator(x);
	  n2 = numerator(y);
	  d2 = denominator(y);
	RATIO_REM_RATIO:
	  if (d1 == d2)
	    quo = (s7_Int)(n1 / n2);
	  else
	    {
	      if (n1 == n2)
		quo = (s7_Int)(d2 / d1);
	      else
		{
		  if ((integer_length(n1) + integer_length(d2) >= s7_int_bits) ||
		      (integer_length(n2) + integer_length(d1) >= s7_int_bits))
		    {
		      pre_quo = ((long double)n1 / (long double)n2) * ((long double)d2 / (long double)d1);
		      if ((pre_quo > S7_LLONG_MAX) || (pre_quo < S7_LLONG_MIN))
			return(s7_out_of_range_error(sc, "remainder", 0, args, "too large to express as an integer"));
		      if (pre_quo > 0.0) quo = (s7_Int)floor(pre_quo); else quo = (s7_Int)ceil(pre_quo);
		    }
		  else quo = (n1 * d2) / (n2 * d1);
		}
	    }
	  if (quo == 0)
	    return(x);

	  if ((d1 == d2) &&
	      ((integer_length(n2) + integer_length(quo)) < s7_int_bits))
	    return(s7_make_ratio(sc, n1 - n2 * quo, d1));
      
	  if ((integer_length(n1) + integer_length(d2) < s7_int_bits) &&
	      (integer_length(d1) + integer_length(d2) < s7_int_bits) &&
	      (integer_length(n2) + integer_length(d1) + integer_length(quo) < s7_int_bits))
	    return(s7_make_ratio(sc, n1 * d2 - n2 * d1 * quo, d1 * d2));
	  return(s7_out_of_range_error(sc, "remainder", 0, args, "intermediate (a/b) is too large"));

	case T_REAL:
	  {
	    s7_Double frac;
	    if (real(y) == 0.0) 
	      return(division_by_zero_error(sc, "remainder", args));
	    if ((isinf(real(y))) || (is_NaN(real(y))))
	      return(s7_wrong_type_arg_error(sc, "remainder", 2, y, "a normal real"));
	    frac = fraction(x);
	    pre_quo = frac / real(y);
	    if ((pre_quo > S7_LLONG_MAX) || (pre_quo < S7_LLONG_MIN))
	      return(s7_out_of_range_error(sc, "remainder", 0, args, "too large to express as an integer"));
	    if (pre_quo > 0.0) quo = (s7_Int)floor(pre_quo); else quo = (s7_Int)ceil(pre_quo);
	    return(s7_make_real(sc, frac - real(y) * quo));
	  }

	default:        
	  return(s7_wrong_type_arg_error(sc, "remainder", 2, y, "a real"));
	}

    case T_REAL:
      if ((isinf(real(x))) || (is_NaN(real(x))))
	return(s7_wrong_type_arg_error(sc, "remainder", 1, x, "a normal real"));

      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(y) == 0) 
	    return(division_by_zero_error(sc, "remainder", args));
	  pre_quo = real(x) / (s7_Double)integer(y);
	  if ((pre_quo > S7_LLONG_MAX) || (pre_quo < S7_LLONG_MIN))
	    return(s7_out_of_range_error(sc, "remainder", 0, args, "too large to express as an integer"));
	  if (pre_quo > 0.0) quo = (s7_Int)floor(pre_quo); else quo = (s7_Int)ceil(pre_quo);
	  return(s7_make_real(sc, real(x) - integer(y) * quo));

	case T_RATIO:
	  {
	    s7_Double frac;
	    frac = fraction(y);
	    pre_quo = real(x) / frac;
	    if ((pre_quo > S7_LLONG_MAX) || (pre_quo < S7_LLONG_MIN))
	      return(s7_out_of_range_error(sc, "remainder", 0, args, "too large to express as an integer"));
	    if (pre_quo > 0.0) quo = (s7_Int)floor(pre_quo); else quo = (s7_Int)ceil(pre_quo);
	    return(s7_make_real(sc, real(x) - frac * quo));
	  }

	case T_REAL:
	  if (real(y) == 0.0) 
	    return(division_by_zero_error(sc, "remainder", args));
	  if ((isinf(real(y))) || (is_NaN(real(y))))
	    return(s7_wrong_type_arg_error(sc, "remainder", 2, y, "a normal real"));

	  pre_quo = real(x) / real(y);
	  if ((pre_quo > S7_LLONG_MAX) || (pre_quo < S7_LLONG_MIN))
	    return(s7_out_of_range_error(sc, "remainder", 0, args, "too large to express as an integer"));
	  if (pre_quo > 0.0) quo = (s7_Int)floor(pre_quo); else quo = (s7_Int)ceil(pre_quo);

	  return(s7_make_real(sc, real(x) - real(y) * quo));

	  /* see under sin -- this calculation is completely bogus if "a" is large
	   * (quotient 1e22 (* 2 pi)) -> -9223372036854775808 -- should this return arithmetic-overflow?
	   *          but it should be 1591549430918953357688, 
	   * (remainder 1e22 (* 2 pi)) -> 1.0057952155665e+22
	   * -- the "remainder" is greater than the original argument!
	   * Clisp gives 0.0 here, as does sbcl
	   * currently s7 throws an error (out-of-range).
	   */

	default:        
	  return(s7_wrong_type_arg_error(sc, "remainder", 2, y, "a real"));
	}

    default:
      return(s7_wrong_type_arg_error(sc, "remainder", 1, x, "a real"));
    }
}



static s7_pointer g_floor(s7_scheme *sc, s7_pointer args)
{
  #define H_floor "(floor x) returns the integer closest to x toward -inf"
  s7_pointer x;

  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:   
      return(x);

    case T_RATIO: 
      {
	s7_Int val;
	val = numerator(x) / denominator(x); 
	/* C "/" truncates? -- C spec says "truncation toward 0" */
	/* we're avoiding "floor" here because the int->double conversion introduces inaccuracies for big numbers */
	if (numerator(x) < 0) /* not "val" because it might be truncated to 0 */
	  return(make_integer(sc, val - 1));
	return(make_integer(sc, val));
      }

    case T_REAL:
      {
	s7_Double z;
	z = real(x);
	if (is_NaN(z))
	  return(s7_out_of_range_error(sc, "floor", 0, x, "argument is NaN"));
	if (isinf(z))
	  return(s7_out_of_range_error(sc, "floor", 0, x, "argument is infinite"));

	/* I used to check for a big real arg here and throw an error, but that
	 *   can't work in general (see s7test).
	 */
	return(make_integer(sc, (s7_Int)floor(real(x)))); 
      }

    case T_COMPLEX:
      return(s7_wrong_type_arg_error(sc, "floor", 0, x, "a real"));

    default:
      return(s7_wrong_type_arg_error(sc, "floor", 0, x, "a number"));
    }
}


static s7_pointer g_ceiling(s7_scheme *sc, s7_pointer args)
{
  #define H_ceiling "(ceiling x) returns the integer closest to x toward inf"
  s7_pointer x;

  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:   
      return(x);

    case T_RATIO:
      {
	s7_Int val;
	val = numerator(x) / denominator(x);
	if (numerator(x) < 0)
	  return(make_integer(sc, val));
	return(make_integer(sc, val + 1));
      }

    case T_REAL:
      {
	s7_Double z;
	z = real(x);
	if (is_NaN(z))
	  return(s7_out_of_range_error(sc, "ceiling", 0, x, "argument is NaN"));
	if (isinf(z))
	  return(s7_out_of_range_error(sc, "ceiling", 0, x, "argument is infinite"));
	return(make_integer(sc, (s7_Int)ceil(real(x)))); 
      }

    case T_COMPLEX:
      return(s7_wrong_type_arg_error(sc, "ceiling", 0, x, "a real"));

    default:
      return(s7_wrong_type_arg_error(sc, "ceiling", 0, x, "a number"));
    }
}


static s7_pointer g_truncate(s7_scheme *sc, s7_pointer args)
{
  #define H_truncate "(truncate x) returns the integer closest to x toward 0"
  s7_pointer x;

  x = car(args);
  switch (type(x))
    {
    case T_INTEGER: 
      return(x);

    case T_RATIO: 
      return(make_integer(sc, (s7_Int)(numerator(x) / denominator(x)))); /* C "/" already truncates */

    case T_REAL:
      {
	s7_Double z;
	z = real(x);
	if (is_NaN(z))
	  return(s7_out_of_range_error(sc, "truncate", 0, x, "argument is NaN"));
	if (isinf(z))
	  return(s7_out_of_range_error(sc, "truncate", 0, x, "argument is infinite"));
	return(s7_truncate(sc, "truncate", real(x))); 
      }

    case T_COMPLEX:
      return(s7_wrong_type_arg_error(sc, "truncate", 0, x, "a real"));

    default:
      return(s7_wrong_type_arg_error(sc, "truncate", 0, x, "a number"));
    }
}


static s7_Double round_per_R5RS(s7_Double x) 
{
  s7_Double fl = floor(x);
  s7_Double ce = ceil(x);
  s7_Double dfl = x - fl;
  s7_Double dce = ce - x;
  
  if (dfl > dce) return(ce);
  if (dfl < dce) return(fl);
  if (fmod(fl, 2.0) == 0.0) return(fl);
  return(ce);
}


static s7_pointer g_round(s7_scheme *sc, s7_pointer args)
{
  #define H_round "(round x) returns the integer closest to x"
  s7_pointer x;

  x = car(args);
  switch (type(x))
    {
    case T_INTEGER: 
      return(x);

    case T_RATIO: 
      {
	s7_Int truncated, remains;
	long double frac;

	truncated = numerator(x) / denominator(x);
	remains = numerator(x) % denominator(x);
	frac = s7_fabsl((long double)remains / (long double)denominator(x));

	if ((frac > 0.5) ||
	    ((frac == 0.5) &&
	     (truncated % 2 != 0)))
	  {
	    if (numerator(x) < 0)
	      return(make_integer(sc, truncated - 1));
	    return(make_integer(sc, truncated + 1));
	  }
	return(make_integer(sc, truncated));
      }

    case T_REAL:
      {
	s7_Double z;
	z = real(x);
	if (is_NaN(z))
	  return(s7_out_of_range_error(sc, "round", 0, x, "argument is NaN"));
	if (isinf(z))
	  return(s7_out_of_range_error(sc, "round", 0, x, "argument is infinite"));
	return(make_integer(sc, (s7_Int)round_per_R5RS(z)));
      }


    case T_COMPLEX:
      return(s7_wrong_type_arg_error(sc, "round", 0, x, "a real"));

    default:
      return(s7_wrong_type_arg_error(sc, "round", 0, x, "a number"));
    }
}


static s7_Int c_mod(s7_Int x, s7_Int y)
{
  s7_Int z;
  /* if (y == 0) return(x); */ /* else arithmetic exception, but we're checking for this elsewhere */
  z = x % y;
  if (((y < 0) && (z > 0)) ||
      ((y > 0) && (z < 0)))
    return(z + y);
  return(z);
}


static s7_pointer g_modulo(s7_scheme *sc, s7_pointer args)
{
  #define H_modulo "(modulo x1 x2) returns x1 mod x2; (modulo 4 3) = 1.  The arguments can be real numbers."
  /* (define (mod x1 x2) (- x1 (* x2 (floor (/ x1 x2))))) from slib 
   * (mod x 0) = x according to "Concrete Mathematics"
   */
  s7_pointer x, y;
  s7_Double a, b;
  s7_Int n1, n2, d1, d2;

  x = car(args);
  y = cadr(args);

  switch (type(x))
    {
    case T_INTEGER:
      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(y) == 0) return(x);
	  return(make_integer(sc, c_mod(integer(x), integer(y))));

	case T_RATIO:
	  n1 = integer(x);
	  d1 = 1;
	  n2 = numerator(y);
	  d2 = denominator(y);
	  goto RATIO_MOD_RATIO;

	case T_REAL:
	  b = real(y);
	  if (b == 0.0) return(x);
	  if (is_NaN(b)) return(y);
	  if (isinf(b)) return(real_NaN);
	  a = (s7_Double)integer(x);
	  return(s7_make_real(sc, a - b * (s7_Int)floor(a / b)));

	default:
	  return(s7_wrong_type_arg_error(sc, "modulo", 2, y, "a real"));
	}

    case T_RATIO:
      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(y) == 0) return(x);
	  n1 = numerator(x);
	  d1 = denominator(x);
	  n2 = integer(y);
	  d2 = 1;
	  goto RATIO_MOD_RATIO;

	case T_RATIO:
	  n1 = numerator(x);
	  d1 = denominator(x);
	  n2 = numerator(y); /* can't be 0 */
	  d2 = denominator(y);
	  if (d1 == d2)
	    return(s7_make_ratio(sc, c_mod(n1, n2), d1));

	RATIO_MOD_RATIO:

	  if ((n1 == n2) &&
	      (d1 > d2))
	    return(x);                 /* signs match so this should be ok */

	  if ((integer_length(n1) + integer_length(d2) < s7_int_bits) &&
	      (integer_length(n2) + integer_length(d1) < s7_int_bits) &&
	      (integer_length(d1) + integer_length(d2) < s7_int_bits))
	    {
	      s7_Int n1d2, n2d1, fl;
	      n1d2 = n1 * d2;
	      n2d1 = n2 * d1;
	      
	      if (n2d1 == 1)
		return(small_int(0));
	      
	      /* can't use "floor" here (int->float ruins everything) */
	      fl = (s7_Int)(n1d2 / n2d1);
	      if (((n1 < 0) && (n2 > 0)) ||
		  ((n1 > 0) && (n2 < 0)))
		fl -= 1;
	      
	      if (fl == 0)
		return(x);
	      
	      if (integer_length(n2d1) + integer_length(fl) < s7_int_bits)
		return(s7_make_ratio(sc, n1d2 - (n2d1 * fl), d1 * d2));
	    }

	  /* there are cases here we might want to catch:
	   *    (modulo 1/9223372036 9223372036) -> error, not 1/9223372036?
	   *    (modulo 9223372036 1/9223372036) -> error, not 0?
	   *    (modulo 1 1/9223372036854775807) -> error, not 0?
	   *    (modulo 1/9223372036854775807 9223372036854775807) -> error, not 1/9223372036854775807?
	   */
	  return(s7_out_of_range_error(sc, "modulo", 0, x, "intermediate (a/b) is too large"));	
      
	case T_REAL:
	  b = real(y);
	  if (b == 0.0) return(x);
	  if (is_NaN(b)) return(y);
	  if (isinf(b)) return(real_NaN);
	  a = fraction(x);
	  return(s7_make_real(sc, a - b * (s7_Int)floor(a / b)));

	default:
	  return(s7_wrong_type_arg_error(sc, "modulo", 2, y, "a real"));
	}

    case T_REAL:
      a = real(x);
      if (is_NaN(a)) return(x);
      if (isinf(a)) return(real_NaN);

      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(y) == 0) return(x);
	  b = (s7_Double)integer(y);
	  return(s7_make_real(sc, a - b * (s7_Int)floor(a / b)));	  

	case T_RATIO:
	  b = fraction(y);
	  return(s7_make_real(sc, a - b * (s7_Int)floor(a / b)));	  

	case T_REAL:
	  b = real(y);
	  if (b == 0.0) return(x);
	  if (is_NaN(b)) return(y);
	  if (isinf(b)) return(real_NaN);
	  return(s7_make_real(sc, a - b * (s7_Int)floor(a / b)));

	default:
	  return(s7_wrong_type_arg_error(sc, "modulo", 2, y, "a real"));
	}

    default:
      return(s7_wrong_type_arg_error(sc, "modulo", 1, x, "a real"));
    }
}
#endif
/* !WITH_GMP */


static int reduce_fraction(s7_Int *numer, s7_Int *denom)
{
  s7_Int divisor;

  if (*numer == 0)
    {
      *denom = 1;
      return(T_INTEGER);
    }
  divisor = c_gcd(*numer, *denom);
  if (divisor != 1)
    {
      *numer /= divisor;
      *denom /= divisor;
    }
  if (*denom == 1)
    return(T_INTEGER);
  return(T_RATIO);
}



/* ---------------------------------------- add ---------------------------------------- */

static s7_pointer g_add(s7_scheme *sc, s7_pointer args)
{
  #define H_add "(+ ...) adds its arguments"
  s7_pointer x, p;
  s7_Int num_a, den_a;
  s7_Double rl_a, im_a;

#if (!WITH_GMP)
  if (is_null(args))
    return(small_int(0));
#endif
    
  x = car(args);
  p = cdr(args);
  if (is_null(p))
    {
      if (!is_number(x))
	return(s7_wrong_type_arg_error(sc, "+", 0, x, "a number"));
      return(x);
    }

  switch (type(x))
    {
    case T_INTEGER:
      num_a = integer(x);

    ADD_INTEGERS:
#if WITH_GMP
      if ((num_a > S7_LONG_MAX) ||
	  (num_a < S7_LONG_MIN))
	return(big_add(sc, cons(sc, s7_Int_to_big_integer(sc, num_a), p)));
#endif
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  if (is_null(p)) return(make_integer(sc, num_a + integer(x)));
	  num_a += integer(x);
	  goto ADD_INTEGERS;
	  
	case T_RATIO:
	  if (is_null(p)) return(s7_make_ratio(sc, numerator(x) + (num_a * denominator(x)), denominator(x)));
	  den_a = denominator(x);
	  num_a = numerator(x) + (num_a * denominator(x));
	  if (reduce_fraction(&num_a, &den_a) == T_INTEGER)
	    goto ADD_INTEGERS;
	  goto ADD_RATIOS;
	  
	case T_REAL:
	  if (is_null(p)) return(s7_make_real(sc, num_a + real(x)));
	  rl_a = (s7_Double)num_a + real(x);
	  goto ADD_REALS;
	  
	case T_COMPLEX:
	  if (is_null(p)) return(s7_make_complex(sc, num_a + real_part(x), imag_part(x)));
	  rl_a = (s7_Double)num_a + real_part(x);
	  im_a = imag_part(x);
	  goto ADD_COMPLEX;
	  
	default:
	  return(s7_wrong_type_arg_error(sc, "+", position_of(p, args) - 1, x, "a number")); /* p is 1 past the current arg */
	}
      break;

    case T_RATIO:
      num_a = numerator(x);
      den_a = denominator(x);
    ADD_RATIOS:
#if WITH_GMP
      if ((num_a > S7_LONG_MAX) ||
	  (den_a > S7_LONG_MAX) ||
	  (num_a < S7_LONG_MIN))
	return(big_add(sc, cons(sc, s7_ratio_to_big_ratio(sc, num_a, den_a), p)));
#endif
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  if (is_null(p)) return(s7_make_ratio(sc, num_a + (den_a * integer(x)), den_a));
	  num_a += (den_a * integer(x));
	  if (reduce_fraction(&num_a, &den_a) == T_INTEGER)
	    goto ADD_INTEGERS;
	  goto ADD_RATIOS;

	case T_RATIO:
	  {
	    s7_Int d1, d2, n1, n2;
	    d1 = den_a;
	    n1 = num_a;
	    d2 = denominator(x);
	    n2 = numerator(x);
	    if (d1 == d2)                                     /* the easy case -- if overflow here, it matches the int case */
	      {
		if (is_null(p))
		  return(s7_make_ratio(sc, n1 + n2, d1));
		num_a += n2;                  /* d1 can't be zero */
	      }
	    else
	      {
#if (!WITH_GMP)
		if ((d1 > s7_int_max) || (d2 > s7_int_max) ||     /* before counting bits, check that overflow is possible */
		    (n1 > s7_int_max) || (n2 > s7_int_max) ||
		    (n1 < s7_int_min) || (n2 < s7_int_min))
		  {
		    int d1bits, d2bits;
		    d1bits = integer_length(d1);
		    d2bits = integer_length(d2);
		    if (((d1bits + d2bits) > s7_int_bits) ||
			((d1bits + integer_length(n2)) > (s7_int_bits - 1)) ||
			((d2bits + integer_length(n1)) > (s7_int_bits - 1)))
		      {
			if (is_null(p))
			  return(s7_make_real(sc, ((long double)n1 / (long double)d1) + ((long double)n2 / (long double)d2)));
			rl_a = ((long double)n1 / (long double)d1) + ((long double)n2 / (long double)d2);
			/* this can lose:
			 *   (+ 1 1/9223372036854775807 -1) -> 0.0 not 1/9223372036854775807
			 */
			goto ADD_REALS;
		      }
		  }
#endif
		if (is_null(p))
		  return(s7_make_ratio(sc, n1 * d2 + n2 * d1, d1 * d2));
		num_a = n1 * d2 + n2 * d1;
		den_a = d1 * d2;
	      }
	    /* (+ 1/100 99/100 (- most-positive-fixnum 2)) should not be converted to real
	     */
	  if (reduce_fraction(&num_a, &den_a) == T_INTEGER)
	    goto ADD_INTEGERS;
	  goto ADD_RATIOS;
	  }

	case T_REAL:
	  if (is_null(p)) return(s7_make_real(sc, ((long double)num_a / (long double)den_a) + real(x)));
	  rl_a = ((long double)num_a / (long double)den_a) + real(x);
	  goto ADD_REALS;
	  
	case T_COMPLEX:
	  if (is_null(p)) return(s7_make_complex(sc, ((long double)num_a / (long double)den_a) + real_part(x), imag_part(x)));
	  rl_a = ((long double)num_a / (long double)den_a) + real_part(x);
	  im_a = imag_part(x);
	  goto ADD_COMPLEX;
	  
	default:
	  return(s7_wrong_type_arg_error(sc, "+", position_of(p, args) - 1, x, "a number"));
	}
      break;

    case T_REAL:
      rl_a = real(x);

    ADD_REALS:
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  if (is_null(p)) return(s7_make_real(sc, rl_a + integer(x)));
	  rl_a += (s7_Double)integer(x);
	  goto ADD_REALS;

	case T_RATIO:
	  if (is_null(p)) return(s7_make_real(sc, rl_a + fraction(x)));
	  rl_a += (s7_Double)fraction(x);
	  goto ADD_REALS;

	case T_REAL:
	  if (is_null(p)) return(s7_make_real(sc, rl_a + real(x)));
	  rl_a += real(x);
	  goto ADD_REALS;
	  
	case T_COMPLEX:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a + real_part(x), imag_part(x)));
	  rl_a += real_part(x);
	  im_a = imag_part(x);
	  goto ADD_COMPLEX;
	  
	default:
	  return(s7_wrong_type_arg_error(sc, "+", position_of(p, args) - 1, x, "a number"));
	}
      break;

    case T_COMPLEX:
      rl_a = real_part(x);
      im_a = imag_part(x);

    ADD_COMPLEX:
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a + integer(x), im_a));
	  rl_a += (s7_Double)integer(x);
	  goto ADD_COMPLEX;

	case T_RATIO:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a + fraction(x), im_a));
	  rl_a += (s7_Double)fraction(x);
	  goto ADD_COMPLEX;

	case T_REAL:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a + real(x), im_a));
	  rl_a += real(x);
	  goto ADD_COMPLEX;
	  
	case T_COMPLEX:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a + real_part(x), im_a + imag_part(x)));
	  rl_a += real_part(x);
	  im_a += imag_part(x);
	  if (im_a == 0.0)
	    goto ADD_REALS;
	  goto ADD_COMPLEX;
	  
	default:
	  return(s7_wrong_type_arg_error(sc, "+", position_of(p, args) - 1, x, "a number"));
	}
      break;

    default:
      return(s7_wrong_type_arg_error(sc, "+", 1, x, "a number"));
    }
}



#if WITH_OPTIMIZATION
static s7_pointer add_1, add_2, add_1s, add_s1, add_cs1, add_si, add_is, add_sf, add_fs;

static s7_pointer add_ratios(s7_scheme *sc, s7_pointer x, s7_pointer y)
{
  s7_Int d1, d2, n1, n2;
  d1 = number_to_denominator(x);
  n1 = number_to_numerator(x);
  d2 = number_to_denominator(y);
  n2 = number_to_numerator(y);

  if (d1 == d2)                                     /* the easy case -- if overflow here, it matches the int case */
    return(s7_make_ratio(sc, n1 + n2, d1));

  if ((d1 > s7_int_max) || (d2 > s7_int_max) ||     /* before counting bits, check that overflow is possible */
      (n1 > s7_int_max) || (n2 > s7_int_max) ||
      (n1 < s7_int_min) || (n2 < s7_int_min))
    {
      int d1bits, d2bits;
      d1bits = integer_length(d1);
      d2bits = integer_length(d2);
      if (((d1bits + d2bits) > s7_int_bits) ||
	  ((d1bits + integer_length(n2)) > (s7_int_bits - 1)) ||
	  ((d2bits + integer_length(n1)) > (s7_int_bits - 1)))
	return(s7_make_real(sc, ((long double)n1 / (long double)d1) + ((long double)n2 / (long double)d2)));
    }
  return(s7_make_ratio(sc, n1 * d2 + n2 * d1, d1 * d2));
}


static s7_pointer g_add_1(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_number(car(args)))
    return(s7_wrong_type_arg_error(sc, "+", 1, car(args), "a number"));
  return(car(args));
}

static s7_pointer g_add_2(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, y;

  x = car(args);
  y = cadr(args);

  /* this costs more than it saves, and we need to clear typeflag to be safe
   *     if ((x == (*(sc->free_heap_top))) || (y == (*(sc->free_heap_top)))) {sc->free_heap_top++;}
   */

  switch (type(x))
    {
    case T_INTEGER: 
      switch (type(y))
	{
	case T_INTEGER: return(make_integer(sc, integer(x) + integer(y)));
	case T_RATIO:   return(add_ratios(sc, x, y));
	case T_REAL:    return(s7_make_real(sc, integer(x) + real(y)));
	case T_COMPLEX: return(s7_make_complex(sc, integer(x) + real_part(y), imag_part(y)));
	default:        return(s7_wrong_type_arg_error(sc, "+", 2, y, "a number"));
	}
      
    case T_RATIO:
      switch (type(y))
	{
	case T_INTEGER:
	case T_RATIO:   return(add_ratios(sc, x, y));
	case T_REAL:    return(s7_make_real(sc, fraction(x) + real(y)));
	case T_COMPLEX: return(s7_make_complex(sc, fraction(x) + real_part(y), imag_part(y)));
	default:        return(s7_wrong_type_arg_error(sc, "+", 2, y, "a number"));
	}
      
    case T_REAL:
      switch (type(y))
	{
	case T_INTEGER: return(s7_make_real(sc, real(x) + integer(y)));
	case T_RATIO:   return(s7_make_real(sc, real(x) + fraction(y)));
	case T_REAL:    return(s7_make_real(sc, real(x) + real(y)));
	case T_COMPLEX: return(s7_make_complex(sc, real(x) + real_part(y), imag_part(y)));
	default:        return(s7_wrong_type_arg_error(sc, "+", 2, y, "a number"));
	}

    case T_COMPLEX:
      switch (type(y))
	{
	case T_INTEGER: return(s7_make_complex(sc, real_part(x) + integer(y), imag_part(x)));
	case T_RATIO:   return(s7_make_complex(sc, real_part(x) + fraction(y), imag_part(x)));
	case T_REAL:    return(s7_make_complex(sc, real_part(x) + real(y), imag_part(x)));
	case T_COMPLEX: return(s7_make_complex(sc, real_part(x) + real_part(y), imag_part(x) + imag_part(y)));
	default:        return(s7_wrong_type_arg_error(sc, "+", 2, y, "a number"));
	}

    default:
      return(s7_wrong_type_arg_error(sc, "+", 1, x, "a number"));
    }
  return(x);
}


static s7_pointer add1(s7_scheme *sc, s7_pointer x, int arg_num)
{
  switch (type(x))
    {
    case T_INTEGER: return(make_integer(sc, integer(x) + 1));
    case T_RATIO:   return(add_ratios(sc, x, small_int(1)));
    case T_REAL:    return(s7_make_real(sc, real(x) + 1.0));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) + 1.0, imag_part(x)));
    default:        return(s7_wrong_type_arg_error(sc, "+", arg_num, x, "a number"));
    }
  return(x);
}

static s7_pointer g_add_s1(s7_scheme *sc, s7_pointer args)
{
  return(add1(sc, car(args), 1));
}

static s7_pointer g_add_cs1(s7_scheme *sc, s7_pointer args)
{
  return(add1(sc, finder(sc, car(args)), 1));
}

static s7_pointer g_add_1s(s7_scheme *sc, s7_pointer args)
{
  return(add1(sc, cadr(args), 2));
}

static s7_pointer g_add_si(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  s7_Int n;

  x = finder(sc, car(args));
  n = integer(cadr(args));
  switch (type(x))
    {
    case T_INTEGER: return(make_integer(sc, integer(x) + n));
    case T_RATIO:   return(add_ratios(sc, x, cadr(args)));
    case T_REAL:    return(s7_make_real(sc, real(x) + n));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) + n, imag_part(x)));
    default:        return(s7_wrong_type_arg_error(sc, "+", 1, x, "a number"));
    }
  return(x);
}

static s7_pointer g_add_is(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  s7_Int n;

  x = finder(sc, cadr(args));
  n = integer(car(args));
  switch (type(x))
    {
    case T_INTEGER: return(make_integer(sc, integer(x) + n));
    case T_RATIO:   return(add_ratios(sc, x, cadr(args)));
    case T_REAL:    return(s7_make_real(sc, real(x) + n));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) + n, imag_part(x)));
    default:        return(s7_wrong_type_arg_error(sc, "+", 2, x, "a number"));
    }
  return(x);
}

static s7_pointer g_add_sf(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  s7_Double n;

  x = finder(sc, car(args));
  n = real(cadr(args));
  switch (type(x))
    {
    case T_INTEGER: return(s7_make_real(sc, integer(x) + n));
    case T_RATIO:   return(s7_make_real(sc, fraction(x) + n));
    case T_REAL:    return(s7_make_real(sc, real(x) + n));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) + n, imag_part(x)));
    default:        return(s7_wrong_type_arg_error(sc, "+", 1, x, "a number"));
    }
  return(x);
}

static s7_pointer g_add_fs(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  s7_Double n;

  x = finder(sc, cadr(args));
  n = real(car(args));
  switch (type(x))
    {
    case T_INTEGER: return(s7_make_real(sc, integer(x) + n));
    case T_RATIO:   return(s7_make_real(sc, fraction(x) + n));
    case T_REAL:    return(s7_make_real(sc, real(x) + n));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) + n, imag_part(x)));
    default:        return(s7_wrong_type_arg_error(sc, "+", 2, x, "a number"));
    }
  return(x);
}

/* (let () (define (hi a) (+ a 1)) ((apply let '((x 32)) (list (procedure-source hi))) 12))
 */
/* other standard cases (fft-related): (+|- (vector-ref ...) s) -- actually (vector-set! v i (+|- (vector-ref v i) s))
 *                                     (+|- (* s (vector-ref ...)) (* s (vector-ref ...)))
 */
#endif



/* ---------------------------------------- subtract ---------------------------------------- */

static s7_pointer g_subtract(s7_scheme *sc, s7_pointer args)
{
  #define H_subtract "(- x1 ...) subtracts its trailing arguments from the first, or negates the first if only one argument is given"
  s7_pointer x, p;
  s7_Int num_a, den_a;
  s7_Double rl_a, im_a;

  x = car(args);
  p = cdr(args);

#if (!WITH_GMP)
  if (is_null(p))
    {
      if (!is_number(x))
	return(s7_wrong_type_arg_error(sc, "-", 0, x, "a number"));
      return(s7_negate(sc, x));
    }
#endif

  switch (type(x))
    {
    case T_INTEGER:
      num_a = integer(x);

    SUBTRACT_INTEGERS:
#if WITH_GMP
      if ((num_a > S7_LONG_MAX) ||
	  (num_a < S7_LONG_MIN))
	return(big_subtract(sc, cons(sc, s7_Int_to_big_integer(sc, num_a), p)));
#endif
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  if (is_null(p)) return(make_integer(sc, num_a - integer(x)));
	  num_a -= integer(x);
	  goto SUBTRACT_INTEGERS;
	  
	case T_RATIO:
	  if (is_null(p)) return(s7_make_ratio(sc, (num_a * denominator(x)) - numerator(x), denominator(x)));
	  den_a = denominator(x);
	  num_a = (num_a * denominator(x)) - numerator(x);
	  if (reduce_fraction(&num_a, &den_a) == T_INTEGER)
	    goto SUBTRACT_INTEGERS;
	  goto SUBTRACT_RATIOS;
	  
	case T_REAL:
	  if (is_null(p)) return(s7_make_real(sc, num_a - real(x)));
	  rl_a = (s7_Double)num_a - real(x);
	  goto SUBTRACT_REALS;
	  
	case T_COMPLEX:
	  if (is_null(p)) return(s7_make_complex(sc, num_a - real_part(x), -imag_part(x)));
	  rl_a = (s7_Double)num_a - real_part(x);
	  im_a = -imag_part(x);
	  goto SUBTRACT_COMPLEX;
	  
	default:
	  return(s7_wrong_type_arg_error(sc, "-", position_of(p, args) - 1, x, "a number")); /* p is 1 past the current arg */
	}
      break;

    case T_RATIO:
      num_a = numerator(x);
      den_a = denominator(x);
    SUBTRACT_RATIOS:
#if WITH_GMP
      if ((num_a > S7_LONG_MAX) ||
	  (den_a > S7_LONG_MAX) ||
	  (num_a < S7_LONG_MIN))
	return(big_subtract(sc, cons(sc, s7_ratio_to_big_ratio(sc, num_a, den_a), p)));
#endif
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  if (is_null(p)) return(s7_make_ratio(sc, num_a - (den_a * integer(x)), den_a));
	  num_a -= (den_a * integer(x));
	  if (reduce_fraction(&num_a, &den_a) == T_INTEGER)
	    goto SUBTRACT_INTEGERS;
	  goto SUBTRACT_RATIOS;

	case T_RATIO:
	  {
	    s7_Int d1, d2, n1, n2;
	    d1 = den_a;
	    n1 = num_a;
	    d2 = denominator(x);
	    n2 = numerator(x);
	    if (d1 == d2)                                     /* the easy case -- if overflow here, it matches the int case */
	      {
		if (is_null(p))
		  return(s7_make_ratio(sc, n1 - n2, d1));
		num_a -= n2;                  /* d1 can't be zero */
	      }
	    else
	      {
#if (!WITH_GMP)
		if ((d1 > s7_int_max) || (d2 > s7_int_max) ||     /* before counting bits, check that overflow is possible */
		    (n1 > s7_int_max) || (n2 > s7_int_max) ||
		    (n1 < s7_int_min) || (n2 < s7_int_min))
		  {
		    int d1bits, d2bits;
		    d1bits = integer_length(d1);
		    d2bits = integer_length(d2);
		    if (((d1bits + d2bits) > s7_int_bits) ||
			((d1bits + integer_length(n2)) > (s7_int_bits - 1)) ||
			((d2bits + integer_length(n1)) > (s7_int_bits - 1)))
		      {
			if (is_null(p))
			  return(s7_make_real(sc, ((long double)n1 / (long double)d1) - ((long double)n2 / (long double)d2)));
			rl_a = ((long double)n1 / (long double)d1) - ((long double)n2 / (long double)d2);
			goto SUBTRACT_REALS;
		      }
		  }
#endif
		if (is_null(p))
		  return(s7_make_ratio(sc, n1 * d2 - n2 * d1, d1 * d2));
		num_a = n1 * d2 - n2 * d1;
		den_a = d1 * d2;
	      }
	  if (reduce_fraction(&num_a, &den_a) == T_INTEGER)
	    goto SUBTRACT_INTEGERS;
	  goto SUBTRACT_RATIOS;
	  }

	case T_REAL:
	  if (is_null(p)) return(s7_make_real(sc, ((long double)num_a / (long double)den_a) - real(x)));
	  rl_a = ((long double)num_a / (long double)den_a) - real(x);
	  goto SUBTRACT_REALS;
	  
	case T_COMPLEX:
	  if (is_null(p)) return(s7_make_complex(sc, ((long double)num_a / (long double)den_a) - real_part(x), -imag_part(x)));
	  rl_a = ((long double)num_a / (long double)den_a) - real_part(x);
	  im_a = -imag_part(x);
	  goto SUBTRACT_COMPLEX;
	  
	default:
	  return(s7_wrong_type_arg_error(sc, "-", position_of(p, args) - 1, x, "a number"));
	}
      break;

    case T_REAL:
      rl_a = real(x);

    SUBTRACT_REALS:
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  if (is_null(p)) return(s7_make_real(sc, rl_a - integer(x)));
	  rl_a -= (s7_Double)integer(x);
	  goto SUBTRACT_REALS;

	case T_RATIO:
	  if (is_null(p)) return(s7_make_real(sc, rl_a - fraction(x)));
	  rl_a -= (s7_Double)fraction(x);
	  goto SUBTRACT_REALS;

	case T_REAL:
	  if (is_null(p)) return(s7_make_real(sc, rl_a - real(x)));
	  rl_a -= real(x);
	  goto SUBTRACT_REALS;
	  
	case T_COMPLEX:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a - real_part(x), -imag_part(x)));
	  rl_a -= real_part(x);
	  im_a = -imag_part(x);
	  goto SUBTRACT_COMPLEX;
	  
	default:
	  return(s7_wrong_type_arg_error(sc, "-", position_of(p, args) - 1, x, "a number"));
	}
      break;

    case T_COMPLEX:
      rl_a = real_part(x);
      im_a = imag_part(x);

    SUBTRACT_COMPLEX:
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a - integer(x), im_a));
	  rl_a -= (s7_Double)integer(x);
	  goto SUBTRACT_COMPLEX;

	case T_RATIO:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a - fraction(x), im_a));
	  rl_a -= (s7_Double)fraction(x);
	  goto SUBTRACT_COMPLEX;

	case T_REAL:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a - real(x), im_a));
	  rl_a -= real(x);
	  goto SUBTRACT_COMPLEX;
	  
	case T_COMPLEX:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a - real_part(x), im_a - imag_part(x)));
	  rl_a -= real_part(x);
	  im_a -= imag_part(x);
	  if (im_a == 0.0)
	    goto SUBTRACT_REALS;
	  goto SUBTRACT_COMPLEX;
	  
	default:
	  return(s7_wrong_type_arg_error(sc, "-", position_of(p, args) - 1, x, "a number"));
	}
      break;

    default:
      return(s7_wrong_type_arg_error(sc, "-", 1, x, "a number"));
    }
}


#if WITH_OPTIMIZATION
static s7_pointer subtract_1, subtract_s1, subtract_cs1, subtract_2, subtract_csn;

static s7_pointer g_subtract_1(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;

  p = car(args);
  switch (type(p))
    {
    case T_INTEGER: 
#if WITH_GMP
      if (integer(p) == S7_LLONG_MIN)
	return(big_negate(sc, list_1(sc, promote_number(sc, T_BIG_INTEGER, p))));
#endif	
      return(make_integer(sc, -integer(p)));
      
    case T_RATIO:
      return(s7_make_ratio(sc, -numerator(p), denominator(p)));
      
    case T_REAL:
      return(s7_make_real(sc, -real(p)));
      
    case T_COMPLEX:
      return(s7_make_complex(sc, -real_part(p), -imag_part(p)));

    default:
      return(s7_wrong_type_arg_error(sc, "-", 1, car(args), "a number"));
    }
}

static s7_pointer g_subtract_2(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, y;

  x = car(args);
  y = cadr(args);
  switch (type(x))
    {
    case T_INTEGER: 
      switch (type(y))
	{
	case T_INTEGER: return(make_integer(sc, integer(x) - integer(y)));
	case T_RATIO:   return(g_subtract(sc, args));
	case T_REAL:    return(s7_make_real(sc, integer(x) - real(y)));
	case T_COMPLEX: return(s7_make_complex(sc, integer(x) - real_part(y), -imag_part(y)));
	default:        return(s7_wrong_type_arg_error(sc, "-", 2, y, "a number"));
	}
      
    case T_RATIO:
      switch (type(y))
	{
	case T_INTEGER:
	case T_RATIO:   return(g_subtract(sc, args));
	case T_REAL:    return(s7_make_real(sc, fraction(x) - real(y)));
	case T_COMPLEX: return(s7_make_complex(sc, fraction(x) - real_part(y), -imag_part(y)));
	default:        return(s7_wrong_type_arg_error(sc, "-", 2, y, "a number"));
	}
      
    case T_REAL:
      switch (type(y))
	{
	case T_INTEGER: return(s7_make_real(sc, real(x) - integer(y)));
	case T_RATIO:   return(s7_make_real(sc, real(x) - fraction(y)));
	case T_REAL:    return(s7_make_real(sc, real(x) - real(y)));
	case T_COMPLEX: return(s7_make_complex(sc, real(x) - real_part(y), -imag_part(y)));
	default:        return(s7_wrong_type_arg_error(sc, "-", 2, y, "a number"));
	}

    case T_COMPLEX:
      switch (type(y))
	{
	case T_INTEGER: return(s7_make_complex(sc, real_part(x) - integer(y), imag_part(x)));
	case T_RATIO:   return(s7_make_complex(sc, real_part(x) - fraction(y), imag_part(x)));
	case T_REAL:    return(s7_make_complex(sc, real_part(x) - real(y), imag_part(x)));
	case T_COMPLEX: return(s7_make_complex(sc, real_part(x) - real_part(y), imag_part(x) - imag_part(y)));
	default:        return(s7_wrong_type_arg_error(sc, "-", 2, y, "a number"));
	}

    default:
      return(s7_wrong_type_arg_error(sc, "-", 1, x, "a number"));
    }
  return(x);
}


static s7_pointer sub1(s7_scheme *sc, s7_pointer x, int arg_num)
{
  switch (type(x))
    {
    case T_INTEGER: return(make_integer(sc, integer(x) - 1));
    case T_RATIO:   return(subtract_ratios(sc, x, small_int(1)));
    case T_REAL:    return(s7_make_real(sc, real(x) - 1.0));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) - 1.0, imag_part(x)));
    default:        return(s7_wrong_type_arg_error(sc, "-", arg_num, x, "a number"));
    }
  return(x);
}

static s7_pointer g_subtract_cs1(s7_scheme *sc, s7_pointer args)
{
  return(sub1(sc, finder(sc, car(args)), 1));
}

static s7_pointer g_subtract_s1(s7_scheme *sc, s7_pointer args)
{
  return(sub1(sc, car(args), 1));
}

static s7_pointer g_subtract_csn(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  s7_Int n;

  x = finder(sc, car(args));
  n = s7_integer(cadr(args));
  switch (type(x))
    {
    case T_INTEGER: return(make_integer(sc, integer(x) - n));
    case T_RATIO:   return(subtract_ratios(sc, x, cadr(args)));
    case T_REAL:    return(s7_make_real(sc, real(x) - n));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) - n, imag_part(x)));
    default:        return(s7_wrong_type_arg_error(sc, "-", 1, x, "a number"));
    }
  return(x);
}
#endif



/* ---------------------------------------- multiply ---------------------------------------- */

static s7_pointer g_multiply(s7_scheme *sc, s7_pointer args)
{
  #define H_multiply "(* ...) multiplies its arguments"
  s7_pointer x, p;
  s7_Int num_a, den_a;
  s7_Double rl_a, im_a;

#if (!WITH_GMP)
  if (is_null(args))
    return(small_int(1));
#endif
    
  x = car(args);
  p = cdr(args);
  if (is_null(p))
    {
      if (!is_number(x))
	return(s7_wrong_type_arg_error(sc, "*", 0, x, "a number"));
      return(x);
    }

  switch (type(x))
    {
    case T_INTEGER:
      num_a = integer(x);

    MULTIPLY_INTEGERS:
#if WITH_GMP
      if ((num_a > S7_LONG_MAX) ||
	  (num_a < S7_LONG_MIN))
	return(big_multiply(sc, cons(sc, s7_Int_to_big_integer(sc, num_a), p)));
#endif
      x = car(p);
      p = cdr(p);
      switch (type(x))
	{
	case T_INTEGER:
#if WITH_GMP
	  if ((integer(x) > S7_LONG_MAX) ||
	      (integer(x) < S7_LONG_MIN))
	    return(big_multiply(sc, cons(sc, s7_Int_to_big_integer(sc, num_a), cons(sc, x, p))));
#endif
#if 0
	  if (sc->safety != 0)
	    {
	      bool a_signed;                                 /* (* 524288 19073486328125) -> -8446744073709551616 */
	      a_signed = (((num_a < 0) && (integer(x) > 0)) ||
			  ((num_a > 0) && (integer(x) < 0)));
	      num_a *= integer(x);
	      if (a_signed != (num_a < 0))
		return(s7_out_of_range_error(sc, "* with ", 0, args, "result is too large"));
	      if (is_null(p))
		return(make_integer(sc, num_a));
	    }
#endif
	  if (is_null(p)) return(make_integer(sc, num_a * integer(x)));
	  num_a *= integer(x);
	  goto MULTIPLY_INTEGERS;

	case T_RATIO:
	  if (is_null(p)) return(s7_make_ratio(sc, numerator(x) * num_a, denominator(x)));
	  den_a = denominator(x);
	  num_a *= numerator(x);
	  if (reduce_fraction(&num_a, &den_a) == T_INTEGER)
	    goto MULTIPLY_INTEGERS;
	  goto MULTIPLY_RATIOS;
	  
	case T_REAL:
	  if (is_null(p)) return(s7_make_real(sc, num_a * real(x)));
	  rl_a = num_a * real(x);
	  goto MULTIPLY_REALS;
	  
	case T_COMPLEX:
	  if (is_null(p)) return(s7_make_complex(sc, num_a * real_part(x), num_a * imag_part(x)));
	  rl_a = num_a * real_part(x);
	  im_a = num_a * imag_part(x);
	  goto MULTIPLY_COMPLEX;
	  
	default:
	  return(s7_wrong_type_arg_error(sc, "*", position_of(p, args) - 1, x, "a number")); /* p is 1 past the current arg */
	}
      break;

    case T_RATIO:
      num_a = numerator(x);
      den_a = denominator(x);
    MULTIPLY_RATIOS:
#if WITH_GMP
      if ((num_a > S7_LONG_MAX) ||
	  (den_a > S7_LONG_MAX) ||
	  (num_a < S7_LONG_MIN))
	return(big_multiply(sc, cons(sc, s7_ratio_to_big_ratio(sc, num_a, den_a), p)));
#endif
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  if (is_null(p)) return(s7_make_ratio(sc, num_a * integer(x), den_a));
	  num_a *= integer(x);
	  if (reduce_fraction(&num_a, &den_a) == T_INTEGER)
	    goto MULTIPLY_INTEGERS;
	  goto MULTIPLY_RATIOS;

	case T_RATIO:
	  {
	    s7_Int d1, d2, n1, n2;
	    d1 = den_a;
	    n1 = num_a;
	    d2 = denominator(x);
	    n2 = numerator(x);
#if (!WITH_GMP)
	    if ((d1 > s7_int_max) || (d2 > s7_int_max) ||     /* before counting bits, check that overflow is possible */
		(n1 > s7_int_max) || (n2 > s7_int_max) ||     /*    (* 1/524288 1/19073486328125) for example */
		(n1 < s7_int_min) || (n2 < s7_int_min))
	      {
		if ((integer_length(d1) + integer_length(d2) > s7_int_bits) ||
		    (integer_length(n1) + integer_length(n2) > s7_int_bits))
		  {
		    if (is_null(p))
		      return(s7_make_real(sc, ((long double)n1 / (long double)d1) * ((long double)n2 / (long double)d2)));
		    rl_a = ((long double)n1 / (long double)d1) * ((long double)n2 / (long double)d2);
		    goto MULTIPLY_REALS;
		  }
	      }
#endif
	    if (is_null(p)) return(s7_make_ratio(sc, n1 * n2, d1 * d2));
	    num_a *= n2;
	    den_a *= d2;
	    if (reduce_fraction(&num_a, &den_a) == T_INTEGER)
	      goto MULTIPLY_INTEGERS;
	    goto MULTIPLY_RATIOS;
	  }

	case T_REAL:
	  if (is_null(p)) return(s7_make_real(sc, ((long double)num_a / (long double)den_a) * real(x)));
	  rl_a = ((long double)num_a / (long double)den_a) * real(x);
	  goto MULTIPLY_REALS;
	  
	case T_COMPLEX:
	  {
	    s7_Double frac;
	    frac = ((long double)num_a / (long double)den_a);
	    if (is_null(p)) return(s7_make_complex(sc, frac * real_part(x), frac * imag_part(x)));
	    rl_a = frac * real_part(x);
	    im_a = frac * imag_part(x);
	    goto MULTIPLY_COMPLEX;
	  }
	  
	default:
	  return(s7_wrong_type_arg_error(sc, "*", position_of(p, args) - 1, x, "a number"));
	}
      break;

    case T_REAL:
      rl_a = real(x);

    MULTIPLY_REALS:
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  if (is_null(p)) return(s7_make_real(sc, rl_a * integer(x)));
	  rl_a *= integer(x);
	  goto MULTIPLY_REALS;

	case T_RATIO:
	  if (is_null(p)) return(s7_make_real(sc, rl_a * fraction(x)));
	  rl_a *= (s7_Double)fraction(x);
	  goto MULTIPLY_REALS;

	case T_REAL:
	  if (is_null(p)) return(s7_make_real(sc, rl_a * real(x)));
	  rl_a *= real(x);
	  goto MULTIPLY_REALS;
	  
	case T_COMPLEX:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a * real_part(x), rl_a * imag_part(x)));
	  im_a = rl_a * imag_part(x);
	  rl_a *= real_part(x);
	  goto MULTIPLY_COMPLEX;
	  
	default:
	  return(s7_wrong_type_arg_error(sc, "*", position_of(p, args) - 1, x, "a number"));
	}
      break;

    case T_COMPLEX:
      rl_a = real_part(x);
      im_a = imag_part(x);

    MULTIPLY_COMPLEX:
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a * integer(x), im_a * integer(x)));
	  rl_a *= integer(x);
	  im_a *= integer(x);
	  goto MULTIPLY_COMPLEX;

	case T_RATIO:
	  {
	    s7_Double frac;
	    frac = fraction(x);
	    if (is_null(p)) return(s7_make_complex(sc, rl_a * frac, im_a * frac));
	    rl_a *= frac;
	    im_a *= frac;
	    goto MULTIPLY_COMPLEX;
	  }

	case T_REAL:
	  if (is_null(p)) return(s7_make_complex(sc, rl_a * real(x), im_a * real(x)));
	  rl_a *= real(x);
	  im_a *= real(x);
	  goto MULTIPLY_COMPLEX;
	  
	case T_COMPLEX:
	  {
	    s7_Double r1, r2, i1, i2;
	    r1 = rl_a;
	    i1 = im_a;
	    r2 = real_part(x);
	    i2 = imag_part(x);
	    if (is_null(p))
	      return(s7_make_complex(sc, r1 * r2 - i1 * i2, r1 * i2 + r2 * i1));
	    rl_a = r1 * r2 - i1 * i2;
	    im_a = r1 * i2 + r2 * i1;
	    if (im_a == 0.0)
	      goto MULTIPLY_REALS;
	    goto MULTIPLY_COMPLEX;
	  }

	default:
	  return(s7_wrong_type_arg_error(sc, "*", position_of(p, args) - 1, x, "a number"));
	}
      break;

    default:
      return(s7_wrong_type_arg_error(sc, "*", 1, x, "a number"));
    }
}


#if WITH_OPTIMIZATION
static s7_pointer multiply_2, multiply_fs, multiply_sf, multiply_is, multiply_si;
#if (!WITH_GMP)
  static s7_pointer sqr_ss;
#endif
static s7_pointer g_multiply_2(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, y;

  x = car(args);
  y = cadr(args);
  switch (type(x))
    {
    case T_INTEGER: 
      switch (type(y))
	{
	case T_INTEGER: return(make_integer(sc, integer(x) * integer(y)));
	case T_RATIO:   return(g_multiply(sc, args));
	case T_REAL:    return(s7_make_real(sc, integer(x) * real(y)));
	case T_COMPLEX: return(s7_make_complex(sc, integer(x) * real_part(y), integer(x) * imag_part(y)));
	default:        return(s7_wrong_type_arg_error(sc, "*", 2, y, "a number"));
	}
      
    case T_RATIO:
      switch (type(y))
	{
	case T_INTEGER:
	case T_RATIO:    return(g_multiply(sc, args));
	case T_REAL:     return(s7_make_real(sc, fraction(x) * real(y)));
	case T_COMPLEX:  
	  {
	    s7_Double frac;
	    frac = fraction(x);
	    return(s7_make_complex(sc, frac * real_part(y), frac * imag_part(y)));
	  }
	default:        return(s7_wrong_type_arg_error(sc, "*", 2, y, "a number"));
	}
      
    case T_REAL:
      switch (type(y))
	{
	case T_INTEGER: return(s7_make_real(sc, real(x) * integer(y)));
	case T_RATIO:   return(s7_make_real(sc, real(x) * fraction(y)));
	case T_REAL:    return(s7_make_real(sc, real(x) * real(y)));
	case T_COMPLEX: return(s7_make_complex(sc, real(x) * real_part(y), real(x) * imag_part(y)));
	default:        return(s7_wrong_type_arg_error(sc, "*", 2, y, "a number"));
	}

    case T_COMPLEX:
      switch (type(y))
	{
	case T_INTEGER: return(s7_make_complex(sc, real_part(x) * integer(y), imag_part(x) * integer(y)));
	case T_RATIO:   
	  {
	    s7_Double frac;
	    frac = fraction(y);
	    return(s7_make_complex(sc, real_part(x) * frac, imag_part(x) * frac));
	  }
	case T_REAL:    return(s7_make_complex(sc, real_part(x) * real(y), imag_part(x) * real(y)));
	case T_COMPLEX:
	  {
	    s7_Double r1, r2, i1, i2;
	    r1 = real_part(x);
	    r2 = real_part(y);
	    i1 = imag_part(x);
	    i2 = imag_part(y);
	    return(s7_make_complex(sc, r1 * r2 - i1 * i2, r1 * i2 + r2 * i1));
	  }
	default:        return(s7_wrong_type_arg_error(sc, "*", 2, y, "a number"));
	}

    default:
      return(s7_wrong_type_arg_error(sc, "*", 1, x, "a number"));      
    }
  return(x);
}


static s7_pointer g_multiply_si(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  s7_Int n;

  x = finder(sc, car(args));
  n = integer(cadr(args));

  switch (type(x))
    {
    case T_INTEGER: return(make_integer(sc, integer(x) * n));
    case T_RATIO:   return(s7_make_ratio(sc, numerator(x) * n, denominator(x)));
    case T_REAL:    return(s7_make_real(sc, real(x) * n));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) * n, imag_part(x) * n));
    default:        return(s7_wrong_type_arg_error(sc, "*", 1, x, "a number"));
    }
  return(x);
}

static s7_pointer g_multiply_is(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  s7_Int n;

  x = finder(sc, cadr(args));
  n = integer(car(args));

  switch (type(x))
    {
    case T_INTEGER: return(make_integer(sc, integer(x) * n));
    case T_RATIO:   return(s7_make_ratio(sc, numerator(x) * n, denominator(x)));
    case T_REAL:    return(s7_make_real(sc, real(x) * n));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) * n, imag_part(x) * n));
    default:        return(s7_wrong_type_arg_error(sc, "*", 2, x, "a number"));
    }
  return(x);
}

static s7_pointer g_multiply_fs(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  s7_Double scl;

  scl = real(car(args));
  x = finder(sc, cadr(args));

  switch (type(x))
    {
    case T_INTEGER: return(s7_make_real(sc, integer(x) * scl));
    case T_RATIO:   return(s7_make_real(sc, numerator(x) * scl / denominator(x)));
    case T_REAL:    return(s7_make_real(sc, real(x) * scl));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) * scl, imag_part(x) * scl));
    default:        return(s7_wrong_type_arg_error(sc, "*", 1, x, "a number"));
    }
  return(x);
}

static s7_pointer g_multiply_sf(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  s7_Double scl;

  scl = real(cadr(args));
  x = finder(sc, car(args));

  switch (type(x))
    {
    case T_INTEGER: return(s7_make_real(sc, integer(x) * scl));
    case T_RATIO:   return(s7_make_real(sc, numerator(x) * scl / denominator(x)));
    case T_REAL:    return(s7_make_real(sc, real(x) * scl));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) * scl, imag_part(x) * scl));
    default:        return(s7_wrong_type_arg_error(sc, "*", 2, x, "a number"));
    }
  return(x);
}


#if (!WITH_GMP)
static s7_pointer g_sqr_ss(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  x = finder(sc, car(args));

  switch (type(x))
    {
    case T_INTEGER: return(s7_make_integer(sc, integer(x) * integer(x)));
    case T_RATIO:   return(s7_make_ratio(sc, numerator(x) * numerator(x), denominator(x) * denominator(x)));
    case T_REAL:    return(s7_make_real(sc, real(x) * real(x)));
    case T_COMPLEX: return(s7_make_complex(sc, real_part(x) * real_part(x) - imag_part(x) * imag_part(x), 2.0 * real_part(x) * imag_part(x)));
    default:        return(s7_wrong_type_arg_error(sc, "*", 1, x, "a number"));
    }
  return(x);
}
#endif

#endif




/* ---------------------------------------- divide ---------------------------------------- */

static s7_pointer g_divide(s7_scheme *sc, s7_pointer args)
{
  #define H_divide "(/ x1 ...) divides its first argument by the rest, or inverts the first if there is only one argument"
  s7_pointer x, p;
  s7_Int num_a, den_a;
  s7_Double rl_a, im_a;

  x = car(args);
  p = cdr(args);
  if (is_null(p))
    {
      if (!is_number(x))
	return(s7_wrong_type_arg_error(sc, "/", 0, x, "a number"));
      if (s7_is_zero(x))
	return(division_by_zero_error(sc, "/", args));
      return(s7_invert(sc, x));
    }

  switch (type(x))
    {
    case T_INTEGER:
      num_a = integer(x);
      if (num_a == 0)
	{
	  bool return_nan = false, return_real_zero = false;
	  for (; is_pair(p); p = cdr(p))
	    {
	      if (!is_number(car(p)))
		return(s7_wrong_type_arg_error(sc, "/", position_of(p, args), car(p), "a number"));
	      if (type(car(p)) > T_RATIO)
		{
		  return_real_zero = true;		  
		  if (is_NaN(real(car(p))))
		    return_nan = true;
		}
	    }
	  if (return_nan)
	    return(real_NaN);
	  if (return_real_zero)
	    return(real_zero);
	  return(small_int(0));
	}

    DIVIDE_INTEGERS:
#if WITH_GMP
      if ((num_a > S7_LONG_MAX) ||
	  (num_a < S7_LONG_MIN))
	return(big_divide(sc, cons(sc, s7_Int_to_big_integer(sc, num_a), p)));
#endif
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  if (integer(x) == 0)
	    return(division_by_zero_error(sc, "/", args));

	  /* to be consistent, I suppose we should search first for NaNs in the divisor list.
	   *   (* 0 0/0) is NaN, so (/ 1 0 0/0) should equal (/ 1 0/0) = NaN.  But the whole
	   *   thing is ridiculous.
	   */
	  
	  if (is_null(p)) return(s7_make_ratio(sc, num_a, integer(x)));
	  den_a = integer(x);
	  if (reduce_fraction(&num_a, &den_a) == T_INTEGER)
	    goto DIVIDE_INTEGERS;
	  goto DIVIDE_RATIOS;
	  
	case T_RATIO:
	  if ((num_a > S7_LONG_MAX) ||
	      (denominator(x) > S7_LONG_MAX) ||
	      (num_a < S7_LONG_MIN))
	    {
	      den_a = 1;
	      goto RATIO_OVER_RATIO;
	    }

	  if (is_null(p)) return(s7_make_ratio(sc, num_a * denominator(x), numerator(x)));
	  den_a = numerator(x);
	  num_a *= denominator(x);
	  if (reduce_fraction(&num_a, &den_a) == T_INTEGER)
	    goto DIVIDE_INTEGERS;
	  goto DIVIDE_RATIOS;
	  
	case T_REAL:
	  rl_a = (s7_Double)num_a;
	  if (real(x) == 0.0)
	    return(division_by_zero_error(sc, "/", args));
	  if (is_null(p)) return(s7_make_real(sc, rl_a / real(x)));
	  rl_a /= real(x);
	  goto DIVIDE_REALS;
	  
	case T_COMPLEX:
	  {
	    s7_Double i2, r2, den;
	    rl_a = (s7_Double)num_a;
	    r2 = real_part(x);
	    i2 = imag_part(x);
	    den = 1.0 / (r2 * r2 + i2 * i2);
	    /* we could avoid the squaring (see Knuth II p613 16)
	     *    not a big deal: (/ 1.0e308+1.0e308i 2.0e308+2.0e308i) => nan
	     *    (gmp case is ok here) 
	     */
	    if (is_null(p))
	      return(s7_make_complex(sc, rl_a * r2 * den, -(rl_a * i2 * den)));
	    im_a = -rl_a * i2 * den;
	    rl_a *= r2 * den;
	    goto DIVIDE_COMPLEX;
	  }

	default:
	  return(s7_wrong_type_arg_error(sc, "/", position_of(p, args) - 1, x, "a number")); /* p is 1 past the current arg */
	}
      break;

    case T_RATIO:
      num_a = numerator(x);
      den_a = denominator(x);
    DIVIDE_RATIOS:
#if WITH_GMP
      if ((num_a > S7_LONG_MAX) ||
	  (den_a > S7_LONG_MAX) ||
	  (num_a < S7_LONG_MIN))
	return(big_divide(sc, cons(sc, s7_ratio_to_big_ratio(sc, num_a, den_a), p)));
#endif
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  if (integer(x) == 0)
	    return(division_by_zero_error(sc, "/", args));
	  if (is_null(p)) return(s7_make_ratio(sc, num_a, den_a * integer(x)));
	  den_a *= integer(x);
	  if (reduce_fraction(&num_a, &den_a) == T_INTEGER)
	    goto DIVIDE_INTEGERS;
	  goto DIVIDE_RATIOS;

	RATIO_OVER_RATIO:
	case T_RATIO:
	  {
	    s7_Int d1, d2, n1, n2;
	    d1 = den_a;
	    n1 = num_a;
	    d2 = denominator(x);
	    n2 = numerator(x);
	    if (d1 == d2)  
	      {
		if (is_null(p))
		  return(s7_make_ratio(sc, n1, n2));
		den_a = n2;
	      }
	    else
	      {
#if (!WITH_GMP)
		if ((d1 > s7_int_max) || (d2 > s7_int_max) ||     /* before counting bits, check that overflow is possible */
		    (n1 > s7_int_max) || (n2 > s7_int_max) ||
		    (n1 < s7_int_min) || (n2 < s7_int_min))
		  {
		    if ((integer_length(d1) + integer_length(n2) > s7_int_bits) ||
			(integer_length(d2) + integer_length(n1) > s7_int_bits))
		      {
			s7_Double r1, r2;
			r1 = ((long double)num_a / (long double)den_a);
			r2 = inverted_fraction(x);
			if (is_null(p))
			  return(s7_make_real(sc, r1 * r2));
			rl_a = r1 * r2;
			goto DIVIDE_REALS;
		      }
		  }
#endif
		if (is_null(p))
		  return(s7_make_ratio(sc, n1 * d2, d1 * n2));
		num_a *= d2;
		den_a *= n2;
	      }
	    if (reduce_fraction(&num_a, &den_a) == T_INTEGER)
	      goto DIVIDE_INTEGERS;
	    goto DIVIDE_RATIOS;
	  }

	case T_REAL:
	  {
	    s7_Double r1;
	    if (real(x) == 0.0)
	      return(division_by_zero_error(sc, "/", args));
	    r1 = ((long double)num_a / (long double)den_a);
	    if (is_null(p)) return(s7_make_real(sc, r1 / real(x)));
	    rl_a = r1 / real(x);
	    goto DIVIDE_REALS;
	  }
	  
	case T_COMPLEX:
	  {
	    s7_Double den, i2, r2;
	    rl_a = ((long double)num_a / (long double)den_a);
	    r2 = real_part(x);
	    i2 = imag_part(x);
	    den = 1.0 / (r2 * r2 + i2 * i2);
	    if (is_null(p))
	      return(s7_make_complex(sc, rl_a * r2 * den, -rl_a * i2 * den));
	    im_a = -rl_a * i2 * den;
	    rl_a *= r2 * den;
	    goto DIVIDE_COMPLEX;
	  }
	  
	default:
	  return(s7_wrong_type_arg_error(sc, "/", position_of(p, args) - 1, x, "a number"));
	}
      break;

    case T_REAL:
      rl_a = real(x);
      if (rl_a == 0)
	{
	  bool return_nan = false;
	  for (; is_pair(p); p = cdr(p))
	    {
	      if (!is_number(car(p)))
		return(s7_wrong_type_arg_error(sc, "/", position_of(p, args), car(p), "a number"));
	      if ((type(car(p)) == T_REAL) && 
		  (is_NaN(real(car(p)))))
		return_nan = true;
	    }
	  if (return_nan)
	    return(real_NaN);
	  return(real_zero);
	}

    DIVIDE_REALS:
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  if (integer(x) == 0)
	    return(division_by_zero_error(sc, "/", args));
	  if (is_null(p)) return(s7_make_real(sc, rl_a / integer(x)));
	  rl_a /= (s7_Double)integer(x);
	  goto DIVIDE_REALS;

	case T_RATIO:
	  if (is_null(p)) return(s7_make_real(sc, rl_a * inverted_fraction(x)));
	  rl_a *= (s7_Double)inverted_fraction(x);
	  goto DIVIDE_REALS;

	case T_REAL:
	  if (real(x) == 0.0)
	    return(division_by_zero_error(sc, "/", args));
	  if (is_null(p)) return(s7_make_real(sc, rl_a / real(x)));
	  rl_a /= real(x);
	  goto DIVIDE_REALS;
	  
	case T_COMPLEX:
	  {
	    s7_Double den, r2, i2;
	    r2 = real_part(x);
	    i2 = imag_part(x);
	    den = 1.0 / (r2 * r2 + i2 * i2);
	    if (is_null(p))
	      return(s7_make_complex(sc, rl_a * r2 * den, -rl_a * i2 * den));
	    im_a = -rl_a * i2 * den;
	    rl_a *= r2 * den;
	    goto DIVIDE_COMPLEX;
	  }
	  
	default:
	  return(s7_wrong_type_arg_error(sc, "/", position_of(p, args) - 1, x, "a number"));
	}
      break;

    case T_COMPLEX:
      rl_a = real_part(x);
      im_a = imag_part(x);

    DIVIDE_COMPLEX:
      x = car(p);
      p = cdr(p);

      switch (type(x))
	{
	case T_INTEGER:
	  {
	    s7_Double r1;
	    if (integer(x) == 0)
	      return(division_by_zero_error(sc, "/", args));
	    r1 = 1.0 / (s7_Double)integer(x);
	    if (is_null(p)) return(s7_make_complex(sc, rl_a * r1, im_a * r1));
	    rl_a *= r1;
	    im_a *= r1;
	    goto DIVIDE_COMPLEX;
	  }

	case T_RATIO:
	  {
	    s7_Double frac;
	    frac = inverted_fraction(x);
	    if (is_null(p)) return(s7_make_complex(sc, rl_a * frac, im_a * frac));
	    rl_a *= frac;
	    im_a *= frac;
	    goto DIVIDE_COMPLEX;
	  }

	case T_REAL:
	  {
	    s7_Double r1;
	    if (real(x) == 0.0)
	      return(division_by_zero_error(sc, "/", args));
	    r1 = 1.0 / real(x);
	    if (is_null(p)) return(s7_make_complex(sc, rl_a * r1, im_a * r1));
	    rl_a *= r1;
	    im_a *= r1;
	    goto DIVIDE_COMPLEX;
	  }

	case T_COMPLEX:
	  {
	    s7_Double r1, r2, i1, i2, den;
	    r1 = rl_a;
	    i1 = im_a;
	    r2 = real_part(x);
	    i2 = imag_part(x);
	    den = 1.0 / (r2 * r2 + i2 * i2);
	    if (is_null(p))
	      return(s7_make_complex(sc, (r1 * r2 + i1 * i2) * den, (r2 * i1 - r1 * i2) * den));
	    rl_a = (r1 * r2 + i1 * i2) * den;
	    im_a = (r2 * i1 - r1 * i2) * den;
	    goto DIVIDE_COMPLEX;
	  }
	  
	default:
	  return(s7_wrong_type_arg_error(sc, "/", position_of(p, args) - 1, x, "a number"));
	}
      break;

    default:
      return(s7_wrong_type_arg_error(sc, "/", 1, x, "a number"));
    }
}



/* ---------------------------------------- max/min ---------------------------------------- */

static s7_pointer g_max(s7_scheme *sc, s7_pointer args)
{
  #define H_max "(max ...) returns the maximum of its arguments"

  s7_pointer x, y, p;
  s7_Int num_a, num_b, den_a, den_b;

  x = car(args);
  p = cdr(args);
  
  switch (type(x))
    {
    case T_INTEGER:
    MAX_INTEGERS:
      if (is_null(p)) return(x);
      y = car(p);
      p = cdr(p);
      
      switch (type(y))
	{
	case T_INTEGER: 
	  if (integer(x) < integer(y)) x = y;
	  goto MAX_INTEGERS;

	case T_RATIO:
	  num_a = integer(x);
	  den_a = 1;
	  num_b = numerator(y);
	  den_b = denominator(y);
	  goto RATIO_MAX_RATIO;
	  
	case T_REAL:    
	  if (is_NaN(real(y)))
	    {
	      for (; is_not_null(p); p = cdr(p))
		if (!is_real(car(p)))
		  return(s7_wrong_type_arg_error(sc, "max", position_of(p, args), car(p), "a real"));
	      return(y);
	    }
	  if (integer(x) < real(y)) 
	    {
	      x = y; 
	      goto MAX_REALS;
	    }
	  goto MAX_INTEGERS;

	default:        
	  return(s7_wrong_type_arg_error(sc, "max", position_of(p, args) - 1, y, "a real"));
	}


    case T_RATIO:
    MAX_RATIOS:
      if (is_null(p)) return(x);
      y = car(p);
      p = cdr(p);
      
      switch (type(y))
	{
	case T_INTEGER:
	  num_a = numerator(x);
	  den_a = denominator(x);
	  num_b = integer(y);
	  den_b = 1;
	  goto RATIO_MAX_RATIO;
	  
	case T_RATIO:
	  num_a = numerator(x);
	  den_a = denominator(x);
	  num_b = numerator(y);
	  den_b = denominator(y);
	  
	RATIO_MAX_RATIO:
	  /* there are tricky cases here where long ints outrun doubles:
	   *   (max 92233720368547758/9223372036854775807 92233720368547757/9223372036854775807)
	   * which should be 92233720368547758/9223372036854775807) but 1st the fraction gets reduced
	   * to 13176245766935394/1317624576693539401, so we fall into the double comparison, and
	   * there we should be comparing 
	   *    9.999999999999999992410584792601468961145E-3 and
	   *    9.999999999999999883990367544051025548645E-3
	   * but if using doubles we get 
	   *    0.010000000000000000208166817117 and
	   *    0.010000000000000000208166817117
	   * that is, we can't distinguish these two fractions once they're coerced to doubles.
	   *
	   * Even long doubles fail in innocuous-looking cases:
	   *     (min 21053343141/6701487259 3587785776203/1142027682075) -> 3587785776203/1142027682075
	   *     (max 21053343141/6701487259 3587785776203/1142027682075) -> 3587785776203/1142027682075
	   *
	   * Another consequence: outside gmp, we can't handle cases like
	   *    (max 9223372036854776/9223372036854775807 #i9223372036854775/9223372036854775000)
	   *    (max #i9223372036854776/9223372036854775807 9223372036854775/9223372036854775000)
	   * I guess if the user is using "inexact" numbers (#i...), he accepts their inexactness.
	   */
	  
	  if (den_a == den_b)
	    {
	      if (num_a < num_b)
		x = y;
	    }
	  else
	    {
	      if (num_a == num_b)
		{
		  if (((num_a >= 0) &&
		       (den_a > den_b)) ||
		      ((num_a < 0) &&
		       (den_a < den_b)))
		    x = y;
		}
	      else
		{
		  s7_Int vala, valb;
		  vala = num_a / den_a;
		  valb = num_b / den_b;
		  
		  if (!((vala > valb) ||
			((vala == valb) && (type(y) == T_INTEGER))))
		    {
		      if ((valb > vala) ||
			  ((vala == valb) && (type(x) == T_INTEGER)) ||
			  /* sigh -- both are ratios and the int parts are equal */
			  (((long double)(num_a % den_a) / (long double)den_a) <= ((long double)(num_b % den_b) / (long double)den_b)))
			x = y;
		    }
		}
	    }
	  if (type(x) == T_RATIO)
	    goto MAX_RATIOS;
	  goto MAX_INTEGERS;

	case T_REAL:
	  if (fraction(x) < real(y))
	    {
	      x = y;
	      goto MAX_REALS;
	    }
	  goto MAX_RATIOS;

	default:        
	  return(s7_wrong_type_arg_error(sc, "max", position_of(p, args) - 1, y, "a real"));
	}


    case T_REAL:
      if (is_NaN(real(x)))
	{
	  for (; is_not_null(p); p = cdr(p))
	    if (!is_real(car(p)))
	      return(s7_wrong_type_arg_error(sc, "max", position_of(p, args), car(p), "a real"));
	  return(x);
	}

    MAX_REALS:
      if (is_null(p)) return(x);
      y = car(p);
      p = cdr(p);
      
      switch (type(y))
	{
	case T_INTEGER: 
	  if (real(x) < integer(y))
	    {
	      x = y;
	      goto MAX_INTEGERS;
	    }
	  goto MAX_REALS;

	case T_RATIO:
	  if (real(x) < fraction(y))
	    {
	      x = y;
	      goto MAX_RATIOS;
	    }
	  goto MAX_REALS;

	case T_REAL:
	  if (is_NaN(real(y)))
	    {
	      for (; is_not_null(p); p = cdr(p))
		if (!is_real(car(p)))
		  return(s7_wrong_type_arg_error(sc, "max", position_of(p, args), car(p), "a real"));
	      return(y);
	    }
	  if (real(x) < real(y)) x = y;
	  goto MAX_REALS;

	default:
	  return(s7_wrong_type_arg_error(sc, "max", position_of(p, args) - 1, y, "a real"));
	}

    default:
      return(s7_wrong_type_arg_error(sc, "max", 1, x, "a real"));
    }
}


static s7_pointer g_min(s7_scheme *sc, s7_pointer args)
{
  #define H_min "(min ...) returns the minimum of its arguments"

  s7_pointer x, y, p;
  s7_Int num_a, num_b, den_a, den_b;

  x = car(args);
  p = cdr(args);
  
  switch (type(x))
    {
    case T_INTEGER:
    MIN_INTEGERS:
      if (is_null(p)) return(x);
      y = car(p);
      p = cdr(p);
      
      switch (type(y))
	{
	case T_INTEGER: 
	  if (integer(x) > integer(y)) x = y;
	  goto MIN_INTEGERS;

	case T_RATIO:
	  num_a = integer(x);
	  den_a = 1;
	  num_b = numerator(y);
	  den_b = denominator(y);
	  goto RATIO_MIN_RATIO;
	  
	case T_REAL:    
	  if (is_NaN(real(y)))
	    {
	      for (; is_not_null(p); p = cdr(p))
		if (!is_real(car(p)))
		  return(s7_wrong_type_arg_error(sc, "min", position_of(p, args), car(p), "a real"));
	      return(y);
	    }
	  if (integer(x) > real(y)) 
	    {
	      x = y; 
	      goto MIN_REALS;
	    }
	  goto MIN_INTEGERS;

	default:        
	  return(s7_wrong_type_arg_error(sc, "min", position_of(p, args) - 1, y, "a real"));
	}


    case T_RATIO:
    MIN_RATIOS:
      if (is_null(p)) return(x);
      y = car(p);
      p = cdr(p);
      
      switch (type(y))
	{
	case T_INTEGER:
	  num_a = numerator(x);
	  den_a = denominator(x);
	  num_b = integer(y);
	  den_b = 1;
	  goto RATIO_MIN_RATIO;
	  
	case T_RATIO:
	  num_a = numerator(x);
	  den_a = denominator(x);
	  num_b = numerator(y);
	  den_b = denominator(y);
	  
	RATIO_MIN_RATIO:
	    if (den_a == den_b)
	      {
		if (num_a > num_b)
		  x = y;
	      }
	    else
	      {
		if (num_a == num_b)
		  {
		    if (((num_a >= 0) &&
			 (den_a < den_b)) ||
			((num_a < 0) &&
			 (den_a > den_b)))
		      x = y;
		  }
		else
		  {
		    s7_Int vala, valb;
		    vala = num_a / den_a;
		    valb = num_b / den_b;

		    if (!((vala < valb) ||
			  ((vala == valb) && (type(x) == T_INTEGER))))
		      {
			if ((valb < vala) ||
			    ((vala == valb) && (type(y) == T_INTEGER)) ||
			    (((long double)(num_a % den_a) / (long double)den_a) >= ((long double)(num_b % den_b) / (long double)den_b)))
			  x = y;
		      }
		  }
	      }
	  if (type(x) == T_RATIO)
	    goto MIN_RATIOS;
	  goto MIN_INTEGERS;

	case T_REAL:
	  if (fraction(x) > real(y))
	    {
	      x = y;
	      goto MIN_REALS;
	    }
	  goto MIN_RATIOS;

	default:        
	  return(s7_wrong_type_arg_error(sc, "min", position_of(p, args) - 1, y, "a real"));
	}


    case T_REAL:
      if (is_NaN(real(x)))
	{
	  for (; is_not_null(p); p = cdr(p))
	    if (!is_real(car(p)))
	      return(s7_wrong_type_arg_error(sc, "min", position_of(p, args), car(p), "a real"));
	  return(x);
	}

    MIN_REALS:
      if (is_null(p)) return(x);
      y = car(p);
      p = cdr(p);
      
      switch (type(y))
	{
	case T_INTEGER: 
	  if (real(x) > integer(y))
	    {
	      x = y;
	      goto MIN_INTEGERS;
	    }
	  goto MIN_REALS;

	case T_RATIO:
	  if (real(x) > fraction(y))
	    {
	      x = y;
	      goto MIN_RATIOS;
	    }
	  goto MIN_REALS;

	case T_REAL:
	  if (is_NaN(real(y)))
	    {
	      for (; is_not_null(p); p = cdr(p))
		if (!is_real(car(p)))
		  return(s7_wrong_type_arg_error(sc, "min", position_of(p, args), car(p), "a real"));
	      return(y);
	    }
	  if (real(x) > real(y)) x = y;
	  goto MIN_REALS;

	default:
	  return(s7_wrong_type_arg_error(sc, "min", position_of(p, args) - 1, y, "a real"));
	}

    default:
      return(s7_wrong_type_arg_error(sc, "min", 1, x, "a real"));
    }
}



/* ---------------------------------------- = > < >= <= ---------------------------------------- */

static s7_pointer g_equal(s7_scheme *sc, s7_pointer args)
{
  #define H_equal "(= z1 ...) returns #t if all its arguments are equal"
  s7_pointer x, p;
  s7_Int num_a, den_a;
  s7_Double rl_a, im_a;

  x = car(args);
  p = cdr(args);

  switch (type(x))
    {
    case T_INTEGER:
      num_a = integer(x);
      while (true)
	{
	  x = car(p);
	  p = cdr(p);
	  switch (type(x))
	    {
	    case T_INTEGER: 
	      if (num_a != integer(x)) goto NOT_EQUAL; 
	      break;

	    case T_RATIO:   
	    case T_COMPLEX: 
	      goto NOT_EQUAL;                          
	      break;
	      
	    case T_REAL:    
	      if (num_a != real(x)) goto NOT_EQUAL;    
	      break;

	    default:
	      return(s7_wrong_type_arg_error(sc, "=", position_of(p, args) - 1, x, "a number"));	      
	    }
	  if (is_null(p))
	    return(sc->T);
	}

    case T_RATIO:
      num_a = numerator(x);
      den_a = denominator(x);
      rl_a = 0.0;
      while (true)
	{
	  x = car(p);
	  p = cdr(p);
	  switch (type(x))
	    {
	    case T_INTEGER: 
	    case T_COMPLEX: 
	      goto NOT_EQUAL;
	      break;

	    case T_RATIO:   
	      if ((num_a != numerator(x)) || (den_a != denominator(x)))	goto NOT_EQUAL;
	      break;

	    case T_REAL:
	      if (rl_a == 0.0)
		rl_a = ((long double)num_a) / ((long double)den_a);
	      if (rl_a != real(x)) goto NOT_EQUAL;
	      break;

	    default:        
	      return(s7_wrong_type_arg_error(sc, "=", position_of(p, args) - 1, x, "a number"));	      
	    }
	  if (is_null(p))
	    return(sc->T);
	}

    case T_REAL:
      rl_a = real(x);
      while (true)
	{
	  x = car(p);
	  p = cdr(p);
	  switch (type(x))
	    {
	    case T_INTEGER: 
	      if (rl_a != integer(x)) goto NOT_EQUAL; 
	      break;

	    case T_RATIO:
	      if (rl_a != fraction(x)) goto NOT_EQUAL;
	      break;
	      
	    case T_REAL:    
	      if (rl_a != real(x)) goto NOT_EQUAL;    
	      break;

	    case T_COMPLEX: 
	      goto NOT_EQUAL;                          
	      break;

	    default:
	      return(s7_wrong_type_arg_error(sc, "=", position_of(p, args) - 1, x, "a number"));	      
	    }
	  if (is_null(p))
	    return(sc->T);
	}
      
    case T_COMPLEX:
      rl_a = real_part(x);
      im_a = imag_part(x);
      while (true)
	{
	  x = car(p);
	  p = cdr(p);
	  switch (type(x))
	    {
	    case T_INTEGER: 
	    case T_RATIO:
	    case T_REAL:
	      goto NOT_EQUAL;
	      break;

	    case T_COMPLEX: 
	      if ((rl_a != real_part(x)) || (im_a != imag_part(x)))
		goto NOT_EQUAL;
	      break;

	    default:
	      return(s7_wrong_type_arg_error(sc, "=", position_of(p, args) - 1, x, "a number"));	      
	    }
	  if (is_null(p))
	    return(sc->T);
	}
      
    default:
      return(s7_wrong_type_arg_error(sc, "=", 1, x, "a number"));
    }

 NOT_EQUAL:
  for (; is_pair(p); p = cdr(p))
    if (!s7_is_number(car(p)))
      return(s7_wrong_type_arg_error(sc, "=", position_of(p, args), car(p), "a number"));
  return(sc->F);
}


#if WITH_OPTIMIZATION
static s7_pointer equal_s_ic, equal_length_ic, equal_2;
static s7_pointer g_equal_s_ic(s7_scheme *sc, s7_pointer args)
{
  s7_Int y;
  s7_pointer val;

  val = finder(sc, car(args));
  y = s7_integer(cadr(args));
  switch (type(val))
    {
    case T_INTEGER:
      return(make_boolean(sc, integer(val) == y));
      break;
      
    case T_RATIO:  
      return(sc->F);
      
    case T_REAL:    
      return(make_boolean(sc, real(val) == y));
      break;

    case T_COMPLEX:
      return(sc->F);

    default:
      return(s7_wrong_type_arg_error(sc, "=", 1, car(args), "a number"));
    }
  return(sc->T);
}

static s7_pointer object_length(s7_scheme *sc, s7_pointer obj);

static s7_pointer g_equal_length_ic(s7_scheme *sc, s7_pointer args)
{
  /* an experiment to avoid make_integer (and telescope opts)
   *   we get here with car=length expr, cadr=int
   */
  s7_Int ilen;
  s7_pointer val;
  
  val = finder(sc, cadar(args));
  ilen = s7_integer(cadr(args));

  switch (type(val))
    {
    case T_PAIR:
      return(make_boolean(sc, s7_list_length(sc, val) == ilen));

    case T_NIL:
      return(make_boolean(sc, ilen == 0));

    case T_VECTOR:
      return(make_boolean(sc, vector_length(val) == ilen));

    case T_STRING:
      return(make_boolean(sc, string_length(val) == ilen));

    case T_HASH_TABLE:
      return(make_boolean(sc, hash_table_length(val) == ilen));

    case T_C_OBJECT:
      return(make_boolean(sc, s7_integer(object_length(sc, val)) == ilen));

    case T_ENVIRONMENT:
      return(make_boolean(sc, environment_length(sc, val) == ilen));

    default:
      return(s7_wrong_type_arg_error(sc, "length", 0, val, "a list, vector, string, or hash-table"));
    }
  return(sc->F);
}


static s7_pointer g_equal_2(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, y;

  x = car(args);
  y = cadr(args);

  switch (type(x))
    {
    case T_INTEGER:
      switch (type(y))
	{
	case T_INTEGER: return(make_boolean(sc, integer(x) == integer(y)));
	case T_RATIO:   return(sc->F);
	case T_REAL:    return(make_boolean(sc, integer(x) == real(y)));
	case T_COMPLEX: return(sc->F);
	default:        return(s7_wrong_type_arg_error(sc, "=", 2, y, "a number"));	  
	}
      break;
      
    case T_RATIO:  
      switch (type(y))
	{
	case T_INTEGER: return(sc->F);
	case T_RATIO:   return(make_boolean(sc, (numerator(x) == numerator(y)) && (denominator(x) == denominator(y))));
	case T_REAL:    return(make_boolean(sc, fraction(x) == real(y)));            /* this could avoid the divide via numerator == denominator * x */
	case T_COMPLEX: return(sc->F);
	default:        return(s7_wrong_type_arg_error(sc, "=", 2, y, "a number"));	  
	}
      break;
      
    case T_REAL:    
      switch (type(y))
	{
	case T_INTEGER: return(make_boolean(sc, real(x) == integer(y)));
	case T_RATIO:   return(make_boolean(sc, real(x) == fraction(y)));
	case T_REAL:    return(make_boolean(sc, real(x) == real(y)));
	case T_COMPLEX: return(sc->F);
	default:        return(s7_wrong_type_arg_error(sc, "=", 2, y, "a number"));	  
	}
      break;
      
    case T_COMPLEX:
      switch (type(y))
	{
	case T_INTEGER:
	case T_RATIO:
	case T_REAL:
	  return(sc->F);

#ifndef _MSC_VER
	case T_COMPLEX: 
	  return(make_boolean(sc, (real_part(x) == real_part(y)) && (imag_part(x) == imag_part(y))));
#else
	case T_COMPLEX: 
	  if ((real_part(x) == real_part(y)) && (imag_part(x) == imag_part(y))) return(sc->T); else return(sc->F);
#endif
	default:        
	  return(s7_wrong_type_arg_error(sc, "=", 2, y, "a number"));	  
	}
      break;

    default:
      return(s7_wrong_type_arg_error(sc, "=", 1, x, "a number"));
    }
  return(sc->F);
}

#endif



#if (!WITH_GMP)

static s7_pointer g_less(s7_scheme *sc, s7_pointer args)
{
  #define H_less "(< x1 ...) returns #t if its arguments are in increasing order"
  s7_pointer x, y, p;

  x = car(args);
  p = cdr(args);

  switch (type(x))
    {
    case T_INTEGER:
    INTEGER_LESS:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(x) >= integer(y)) goto NOT_LESS;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_LESS;

	case T_RATIO:
	  /* no gmp here, but this can overflow: (< 9223372036 1/9223372036), but conversion to real is also problematic
	   */
	  if ((integer(x) >= 0) && (numerator(y) < 0)) goto NOT_LESS;  /* (< 1 -1/2), ratio numerator can't be 0 */
	  if ((integer(x) <= 0) && (numerator(y) > 0))                 /* (< 0 1/2) */
	    {
	      if (is_null(p)) return(sc->T);
	      x = y;
	      goto RATIO_LESS;
	    }
	  if ((integer(x) < S7_LONG_MAX) && 
	      (integer(x) > S7_LONG_MIN) && 
	      (denominator(y) < S7_LONG_MAX) && 
	      (denominator(y) > S7_LONG_MIN))
	    {
	      if ((integer(x) * denominator(y)) >= numerator(y)) goto NOT_LESS;
	    }
	  else
	    {
	      if (integer(x) >= fraction(y)) goto NOT_LESS;
	    }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_LESS;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_LESS;
	  if (integer(x) >= real(y)) goto NOT_LESS;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_LESS;

	case T_COMPLEX:
	default:
	  return(s7_wrong_type_arg_error(sc, "<", position_of(p, args) - 1, y, "a real"));
	}


    case T_RATIO:
    RATIO_LESS:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if ((numerator(x) > 0) && (integer(y) <= 0)) goto NOT_LESS;
	  if ((numerator(x) < 0) && (integer(y) >= 0)) 
	    {
	      if (is_null(p)) return(sc->T);
	      x = y;
	      goto INTEGER_LESS;
	    }
	  if ((integer(y) < S7_LONG_MAX) && 
	      (integer(y) > S7_LONG_MIN) && 
	      (denominator(x) < S7_LONG_MAX) && 
	      (denominator(x) > S7_LONG_MIN))
	    {
	      if (numerator(x) >= (integer(y) * denominator(x))) goto NOT_LESS;
	    }
	  else
	    {
	      if (fraction(x) >= integer(y)) goto NOT_LESS;
	    }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_LESS;

	case T_RATIO:
	  /* conversion to real and >= is not safe here (see comment under g_greater) */
	  {
	    s7_Int d1, d2, n1, n2;
	    d1 = denominator(x);
	    n1 = numerator(x);
	    d2 = denominator(y);
	    n2 = numerator(y);
	    if (d1 == d2)       
	      {
		if (n1 >= n2) goto NOT_LESS;
	      }
	    else
	      {
		if ((d1 > s7_int_max) || (d2 > s7_int_max) ||     /* before counting bits, check that overflow is possible */
		    (n1 > s7_int_max) || (n2 > s7_int_max) ||
		    (n1 < s7_int_min) || (n2 < s7_int_min))
		  {
		    int d1bits, d2bits;
		    d1bits = integer_length(d1);
		    d2bits = integer_length(d2);
		    if (((d1bits + d2bits) > s7_int_bits) ||
			((d1bits + integer_length(n2)) > (s7_int_bits - 1)) ||
			((d2bits + integer_length(n1)) > (s7_int_bits - 1)))
		      {
			if (fraction(x) >= fraction(y)) goto NOT_LESS;
      
			/* (< 21053343141/6701487259 3587785776203/1142027682075) -> #f because even long doubles aren't enough here 
			 * (= 21053343141/6701487259 3587785776203/1142027682075) is #f because it checks the actual ints and
			 * (> 21053343141/6701487259 3587785776203/1142027682075) is #f just like the < case.
			 * similarly
			 * (min 21053343141/6701487259 3587785776203/1142027682075) -> 3587785776203/1142027682075
			 * (max 21053343141/6701487259 3587785776203/1142027682075) -> 3587785776203/1142027682075
			 *
			 * if we print the long double results as integers, both are -3958705157555305931
			 *    so there's not a lot I can do in the non-gmp case.
			 */
		      }
		    else 
		      {
			if ((n1 * d2) >= (n2 * d1)) goto NOT_LESS;
		      }
		  }
		else
		  {
		    if ((n1 * d2) >=  (n2 * d1)) goto NOT_LESS;
		  }
	      }
	  }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_LESS;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_LESS;
	  if (fraction(x) >= real(y)) goto NOT_LESS;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_LESS;
	  
	case T_COMPLEX:
	default:
	  return(s7_wrong_type_arg_error(sc, "<", position_of(p, args) - 1, y, "a real"));
	}


    case T_REAL:
      if (is_NaN(real(x))) goto NOT_LESS;

    REAL_LESS:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if (real(x) >= integer(y)) goto NOT_LESS;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_LESS;

	case T_RATIO:
	  if (real(x) >= fraction(y)) goto NOT_LESS;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_LESS;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_LESS;
	  if (real(x) >= real(y)) goto NOT_LESS;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_LESS;

	case T_COMPLEX:
	default:
	  return(s7_wrong_type_arg_error(sc, "<", position_of(p, args) - 1, y, "a real"));
	}


    case T_COMPLEX:
    default:
      return(s7_wrong_type_arg_error(sc, "<", 1, x, "a real"));
    }

 NOT_LESS:
  for (; is_pair(p); p = cdr(p))
    if (!is_real(car(p)))
      return(s7_wrong_type_arg_error(sc, "<", position_of(p, args), car(p), "a real"));
  return(sc->F);
}



static s7_pointer g_less_or_equal(s7_scheme *sc, s7_pointer args)
{
  #define H_less_or_equal "(<= x1 ...) returns #t if its arguments are in increasing order"
  s7_pointer x, y, p;

  x = car(args);
  p = cdr(args);

  switch (type(x))
    {
    case T_INTEGER:
    INTEGER_LEQ:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(x) > integer(y)) goto NOT_LEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_LEQ;

	case T_RATIO:
	  /* no gmp here, but this can overflow: (< 9223372036 1/9223372036), but conversion to real is also problematic
	   */
	  if ((integer(x) >= 0) && (numerator(y) < 0)) goto NOT_LEQ;  /* (< 1 -1/2), ratio numerator can't be 0 */
	  if ((integer(x) <= 0) && (numerator(y) > 0))                 /* (< 0 1/2) */
	    {
	      if (is_null(p)) return(sc->T);
	      x = y;
	      goto RATIO_LEQ;
	    }
	  if ((integer(x) < S7_LONG_MAX) && 
	      (integer(x) > S7_LONG_MIN) && 
	      (denominator(y) < S7_LONG_MAX) && 
	      (denominator(y) > S7_LONG_MIN))
	    {
	      if ((integer(x) * denominator(y)) > numerator(y)) goto NOT_LEQ;
	    }
	  else
	    {
	      if (integer(x) > fraction(y)) goto NOT_LEQ;
	    }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_LEQ;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_LEQ;
	  if (integer(x) > real(y)) goto NOT_LEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_LEQ;

	case T_COMPLEX:
	default:
	  return(s7_wrong_type_arg_error(sc, "<=", position_of(p, args) - 1, y, "a real"));
	}


    case T_RATIO:
    RATIO_LEQ:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if ((numerator(x) > 0) && (integer(y) <= 0)) goto NOT_LEQ;
	  if ((numerator(x) < 0) && (integer(y) >= 0)) 
	    {
	      if (is_null(p)) return(sc->T);
	      x = y;
	      goto INTEGER_LEQ;
	    }
	  if ((integer(y) < S7_LONG_MAX) && 
	      (integer(y) > S7_LONG_MIN) && 
	      (denominator(x) < S7_LONG_MAX) && 
	      (denominator(x) > S7_LONG_MIN))
	    {
	      if (numerator(x) > (integer(y) * denominator(x))) goto NOT_LEQ;
	    }
	  else
	    {
	      if (fraction(x) > integer(y)) goto NOT_LEQ;
	    }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_LEQ;

	case T_RATIO:
	  {
	    s7_Int d1, d2, n1, n2;
	    d1 = denominator(x);
	    n1 = numerator(x);
	    d2 = denominator(y);
	    n2 = numerator(y);
	    if (d1 == d2)       
	      {
		if (n1 > n2) goto NOT_LEQ;
	      }
	    else
	      {
		if ((d1 > s7_int_max) || (d2 > s7_int_max) ||     /* before counting bits, check that overflow is possible */
		    (n1 > s7_int_max) || (n2 > s7_int_max) ||
		    (n1 < s7_int_min) || (n2 < s7_int_min))
		  {
		    int d1bits, d2bits;
		    d1bits = integer_length(d1);
		    d2bits = integer_length(d2);
		    if (((d1bits + d2bits) > s7_int_bits) ||
			((d1bits + integer_length(n2)) > (s7_int_bits - 1)) ||
			((d2bits + integer_length(n1)) > (s7_int_bits - 1)))
		      {
			if (fraction(x) > fraction(y)) goto NOT_LEQ;
		      }
		    else 
		      {
			if ((n1 * d2) > (n2 * d1)) goto NOT_LEQ;
		      }
		  }
		else
		  {
		    if ((n1 * d2) >  (n2 * d1)) goto NOT_LEQ;
		  }
	      }
	  }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_LEQ;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_LEQ;
	  if (fraction(x) > real(y)) goto NOT_LEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_LEQ;
	  
	case T_COMPLEX:
	default:
	  return(s7_wrong_type_arg_error(sc, "<=", position_of(p, args) - 1, y, "a real"));
	}


    case T_REAL:
      if (is_NaN(real(x))) goto NOT_LEQ;

    REAL_LEQ:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if (real(x) > integer(y)) goto NOT_LEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_LEQ;

	case T_RATIO:
	  if (real(x) > fraction(y)) goto NOT_LEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_LEQ;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_LEQ;
	  if (real(x) > real(y)) goto NOT_LEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_LEQ;

	case T_COMPLEX:
	default:
	  return(s7_wrong_type_arg_error(sc, "<=", position_of(p, args) - 1, y, "a real"));
	}


    case T_COMPLEX:
    default:
      return(s7_wrong_type_arg_error(sc, "<=", 1, x, "a real"));
    }

 NOT_LEQ:
  for (; is_pair(p); p = cdr(p))
    if (!is_real(car(p)))
      return(s7_wrong_type_arg_error(sc, "<=", position_of(p, args), car(p), "a real"));
  return(sc->F);
}


static s7_pointer g_greater(s7_scheme *sc, s7_pointer args)
{
  #define H_greater "(> x1 ...) returns #t if its arguments are in decreasing order"
  s7_pointer x, y, p;

  x = car(args);
  p = cdr(args);

  switch (type(x))
    {
    case T_INTEGER:
    INTEGER_GREATER:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(x) <= integer(y)) goto NOT_GREATER;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_GREATER;

	case T_RATIO:
	  /* no gmp here, but this can overflow: (< 9223372036 1/9223372036), but conversion to real is also problematic
	   */
	  if ((integer(x) <= 0) && (numerator(y) > 0)) goto NOT_GREATER; 
	  if ((integer(x) >= 0) && (numerator(y) < 0)) 
	    {
	      if (is_null(p)) return(sc->T);
	      x = y;
	      goto RATIO_GREATER;
	    }
	  if ((integer(x) < S7_LONG_MAX) && 
	      (integer(x) > S7_LONG_MIN) && 
	      (denominator(y) < S7_LONG_MAX) && 
	      (denominator(y) > S7_LONG_MIN))
	    {
	      if ((integer(x) * denominator(y)) <= numerator(y)) goto NOT_GREATER;
	    }
	  else
	    {
	      if (integer(x) <= fraction(y)) goto NOT_GREATER;
	    }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_GREATER;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_GREATER;
	  if (integer(x) <= real(y)) goto NOT_GREATER;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_GREATER;

	case T_COMPLEX:
	default:
	  return(s7_wrong_type_arg_error(sc, ">", position_of(p, args) - 1, y, "a real"));
	}


    case T_RATIO:
    RATIO_GREATER:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if ((numerator(x) < 0) && (integer(y) >= 0)) goto NOT_GREATER;
	  if ((numerator(x) > 0) && (integer(y) <= 0))
	    {
	      if (is_null(p)) return(sc->T);
	      x = y;
	      goto INTEGER_GREATER;
	    }
	  if ((integer(y) < S7_LONG_MAX) && 
	      (integer(y) > S7_LONG_MIN) && 
	      (denominator(x) < S7_LONG_MAX) && 
	      (denominator(x) > S7_LONG_MIN))
	    {
	      if (numerator(x) <= (integer(y) * denominator(x))) goto NOT_GREATER;
	    }
	  else
	    {
	      if (fraction(x) <= integer(y)) goto NOT_GREATER;
	    }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_GREATER;

	case T_RATIO:
	  {
	    s7_Int d1, d2, n1, n2;
	    d1 = denominator(x);
	    n1 = numerator(x);
	    d2 = denominator(y);
	    n2 = numerator(y);
	    if (d1 == d2)       
	      {
		if (n1 <= n2) goto NOT_GREATER;
	      }
	    else
	      {
		if ((d1 > s7_int_max) || (d2 > s7_int_max) ||     /* before counting bits, check that overflow is possible */
		    (n1 > s7_int_max) || (n2 > s7_int_max) ||
		    (n1 < s7_int_min) || (n2 < s7_int_min))
		  {
		    int d1bits, d2bits;
		    d1bits = integer_length(d1);
		    d2bits = integer_length(d2);
		    if (((d1bits + d2bits) > s7_int_bits) ||
			((d1bits + integer_length(n2)) > (s7_int_bits - 1)) ||
			((d2bits + integer_length(n1)) > (s7_int_bits - 1)))
		      {
			if (fraction(x) <= fraction(y)) goto NOT_GREATER;
      
			/* (< 21053343141/6701487259 3587785776203/1142027682075) -> #f because even long doubles aren't enough here 
			 * (= 21053343141/6701487259 3587785776203/1142027682075) is #f because it checks the actual ints and
			 * (> 21053343141/6701487259 3587785776203/1142027682075) is #f just like the < case.
			 * similarly
			 * (min 21053343141/6701487259 3587785776203/1142027682075) -> 3587785776203/1142027682075
			 * (max 21053343141/6701487259 3587785776203/1142027682075) -> 3587785776203/1142027682075
			 *
			 * if we print the long double results as integers, both are -3958705157555305931
			 *    so there's not a lot I can do in the non-gmp case.
			 */
		      }
		    else 
		      {
			if ((n1 * d2) <= (n2 * d1)) goto NOT_GREATER;
		      }
		  }
		else
		  {
		    if ((n1 * d2) <=  (n2 * d1)) goto NOT_GREATER;
		  }
	      }
	  }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_GREATER;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_GREATER;
	  if (fraction(x) <= real(y)) goto NOT_GREATER;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_GREATER;
	  
	case T_COMPLEX:
	default:
	  return(s7_wrong_type_arg_error(sc, ">", position_of(p, args) - 1, y, "a real"));
	}


    case T_REAL:
      if (is_NaN(real(x))) goto NOT_GREATER;

    REAL_GREATER:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if (real(x) <= integer(y)) goto NOT_GREATER;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_GREATER;

	case T_RATIO:
	  if (real(x) <= fraction(y)) goto NOT_GREATER;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_GREATER;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_GREATER;
	  if (real(x) <= real(y)) goto NOT_GREATER;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_GREATER;

	case T_COMPLEX:
	default:
	  return(s7_wrong_type_arg_error(sc, ">", position_of(p, args) - 1, y, "a real"));
	}


    case T_COMPLEX:
    default:
      return(s7_wrong_type_arg_error(sc, ">", 1, x, "a real"));
    }

 NOT_GREATER:
  for (; is_pair(p); p = cdr(p))
    if (!is_real(car(p)))
      return(s7_wrong_type_arg_error(sc, ">", position_of(p, args), car(p), "a real"));
  return(sc->F);
}


static s7_pointer g_greater_or_equal(s7_scheme *sc, s7_pointer args)
{
  #define H_greater_or_equal "(>= x1 ...) returns #t if its arguments are in decreasing order"
  /* (>= 1+i 1+i) is an error which seems unfortunate
   */
  s7_pointer x, y, p;

  x = car(args);
  p = cdr(args);

  switch (type(x))
    {
    case T_INTEGER:
    INTEGER_GEQ:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if (integer(x) < integer(y)) goto NOT_GEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_GEQ;

	case T_RATIO:
	  /* no gmp here, but this can overflow: (< 9223372036 1/9223372036), but conversion to real is also problematic
	   */
	  if ((integer(x) <= 0) && (numerator(y) > 0)) goto NOT_GEQ; 
	  if ((integer(x) >= 0) && (numerator(y) < 0)) 
	    {
	      if (is_null(p)) return(sc->T);
	      x = y;
	      goto RATIO_GEQ;
	    }
	  if ((integer(x) < S7_LONG_MAX) && 
	      (integer(x) > S7_LONG_MIN) && 
	      (denominator(y) < S7_LONG_MAX) && 
	      (denominator(y) > S7_LONG_MIN))
	    {
	      if ((integer(x) * denominator(y)) < numerator(y)) goto NOT_GEQ;
	    }
	  else
	    {
	      if (integer(x) < fraction(y)) goto NOT_GEQ;
	    }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_GEQ;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_GEQ;
	  if (integer(x) < real(y)) goto NOT_GEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_GEQ;

	case T_COMPLEX:
	default:
	  return(s7_wrong_type_arg_error(sc, ">=", position_of(p, args) - 1, y, "a real"));
	}


    case T_RATIO:
    RATIO_GEQ:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if ((numerator(x) < 0) && (integer(y) >= 0)) goto NOT_GEQ;
	  if ((numerator(x) > 0) && (integer(y) <= 0))
	    {
	      if (is_null(p)) return(sc->T);
	      x = y;
	      goto INTEGER_GEQ;
	    }
	  if ((integer(y) < S7_LONG_MAX) && 
	      (integer(y) > S7_LONG_MIN) && 
	      (denominator(x) < S7_LONG_MAX) && 
	      (denominator(x) > S7_LONG_MIN))
	    {
	      if (numerator(x) < (integer(y) * denominator(x))) goto NOT_GEQ;
	    }
	  else
	    {
	      if (fraction(x) < integer(y)) goto NOT_GEQ;
	    }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_GEQ;

	case T_RATIO:
	  {
	    s7_Int d1, d2, n1, n2;
	    d1 = denominator(x);
	    n1 = numerator(x);
	    d2 = denominator(y);
	    n2 = numerator(y);
	    if (d1 == d2)       
	      {
		if (n1 < n2) goto NOT_GEQ;
	      }
	    else
	      {
		if ((d1 > s7_int_max) || (d2 > s7_int_max) ||     /* before counting bits, check that overflow is possible */
		    (n1 > s7_int_max) || (n2 > s7_int_max) ||
		    (n1 < s7_int_min) || (n2 < s7_int_min))
		  {
		    int d1bits, d2bits;
		    d1bits = integer_length(d1);
		    d2bits = integer_length(d2);
		    if (((d1bits + d2bits) > s7_int_bits) ||
			((d1bits + integer_length(n2)) > (s7_int_bits - 1)) ||
			((d2bits + integer_length(n1)) > (s7_int_bits - 1)))
		      {
			if (fraction(x) < fraction(y)) goto NOT_GEQ;
		      }
		    else 
		      {
			if ((n1 * d2) < (n2 * d1)) goto NOT_GEQ;
		      }
		  }
		else
		  {
		    if ((n1 * d2) <  (n2 * d1)) goto NOT_GEQ;
		  }
	      }
	  }
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_GEQ;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_GEQ;
	  if (fraction(x) < real(y)) goto NOT_GEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_GEQ;
	  
	case T_COMPLEX:
	default:
	  return(s7_wrong_type_arg_error(sc, ">=", position_of(p, args) - 1, y, "a real"));
	}


    case T_REAL:
      if (is_NaN(real(x))) goto NOT_GEQ;

    REAL_GEQ:
      y = car(p);
      p = cdr(p);
      switch (type(y))
	{
	case T_INTEGER:
	  if (real(x) < integer(y)) goto NOT_GEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto INTEGER_GEQ;

	case T_RATIO:
	  if (real(x) < fraction(y)) goto NOT_GEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto RATIO_GEQ;

	case T_REAL:
	  if (is_NaN(real(y))) goto NOT_GEQ;
	  if (real(x) < real(y)) goto NOT_GEQ;
	  if (is_null(p)) return(sc->T);
	  x = y;
	  goto REAL_GEQ;

	case T_COMPLEX:
	default:
	  return(s7_wrong_type_arg_error(sc, ">=", position_of(p, args) - 1, y, "a real"));
	}


    case T_COMPLEX:
    default:
      return(s7_wrong_type_arg_error(sc, ">=", 1, x, "a real"));
    }

 NOT_GEQ:
  for (; is_pair(p); p = cdr(p))
    if (!is_real(car(p)))
      return(s7_wrong_type_arg_error(sc, ">=", position_of(p, args), car(p), "a real"));
  return(sc->F);

}



#if WITH_OPTIMIZATION
static s7_pointer less_s_ic;
static s7_pointer g_less_s_ic(s7_scheme *sc, s7_pointer args)
{
  s7_Int y;
  s7_pointer x;
  
  x = car(args);
  y = s7_integer(cadr(args));
  switch (type(x))
    {
    case T_INTEGER:
      return(make_boolean(sc, integer(x) < y));
      break;
      
    case T_RATIO:  
      if ((y >= 0) && (numerator(x) < 0))
	return(sc->T);
      if ((y <= 0) && (numerator(x) > 0))
	return(sc->F);
      if ((y < S7_LONG_MAX) && 
	  (y > S7_LONG_MIN) && 
	  (denominator(x) < S7_LONG_MAX) && 
	  (denominator(x) > S7_LONG_MIN))
	return(make_boolean(sc, (numerator(x) < (y * denominator(x)))));
      return(make_boolean(sc, fraction(x) < y));
      break;
      
    case T_REAL:    
      return(make_boolean(sc, real(x) < y));
      break;

    case T_COMPLEX:
    default:
      return(s7_wrong_type_arg_error(sc, "<", 1, x, "a real number"));  
    }
  return(sc->T);
}

static s7_pointer less_length_ic;
static s7_pointer g_less_length_ic(s7_scheme *sc, s7_pointer args)
{
  s7_Int ilen;
  s7_pointer val;
  
  val = finder(sc, cadar(args));
  ilen = s7_integer(cadr(args));

  switch (type(val))
    {
    case T_PAIR:
      return(make_boolean(sc, s7_list_length(sc, val) < ilen));

    case T_NIL:
      return(make_boolean(sc, ilen > 0));

    case T_VECTOR:
      return(make_boolean(sc, vector_length(val) < ilen));

    case T_STRING:
      return(make_boolean(sc, string_length(val) < ilen));

    case T_HASH_TABLE:
      return(make_boolean(sc, hash_table_length(val) < ilen));

    case T_C_OBJECT:
      return(make_boolean(sc, s7_integer(object_length(sc, val)) < ilen));

    case T_ENVIRONMENT:
      return(make_boolean(sc, environment_length(sc, val) < ilen));

    default:
      return(s7_wrong_type_arg_error(sc, "length", 0, val, "a list, vector, string, or hash-table"));
    }
  return(sc->F);
}

static s7_pointer less_2;
static s7_pointer g_less_2(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, y;

  x = car(args);
  y = cadr(args);
  switch (type(x))
    {
    case T_INTEGER:
      switch (type(y))
	{
	case T_INTEGER:
	  return(make_boolean(sc, integer(x) < integer(y)));
	  break;
	  
	case T_RATIO: 
	  return(g_less(sc, args));
	  break;
	  
	case T_REAL:
	  if (is_NaN(real(y))) return(sc->F);
	  return(make_boolean(sc, integer(x) < real(y)));
	  break;

	default:
	  return(s7_wrong_type_arg_error(sc, "<", 2, y, "a real"));	  
	}
      break;
      
    case T_RATIO:
      return(g_less(sc, args));
      break;
      
    case T_REAL:
      switch (type(y))
	{
	case T_INTEGER: 
	  if (is_NaN(real(x))) return(sc->F);
	  return(make_boolean(sc, real(x) < integer(y)));
	  break;
	  
	case T_RATIO:
	  if (is_NaN(real(x))) return(sc->F);
	  return(make_boolean(sc, real(x) < fraction(y)));
	  break;
	  
	case T_REAL:
	  if (is_NaN(real(x))) return(sc->F);
	  if (is_NaN(real(y))) return(sc->F);
	  return(make_boolean(sc, real(x) < real(y)));
	  break;

	default:
	  return(s7_wrong_type_arg_error(sc, "<", 2, y, "a real"));	  
	}
      break;

    default:
      return(s7_wrong_type_arg_error(sc, "<", 1, x, "a real"));      
    }

  return(sc->T);
}

static s7_pointer leq_s_ic;
static s7_pointer g_leq_s_ic(s7_scheme *sc, s7_pointer args)
{
  s7_Int y;
  s7_pointer x;
  
  x = car(args);
  y = s7_integer(cadr(args));

  switch (type(x))
    {
    case T_INTEGER:
      return(make_boolean(sc, integer(x) <= y));
      break;
      
    case T_RATIO:  
      if ((y >= 0) && (numerator(x) <= 0))
	return(sc->T);
      if ((y <= 0) && (numerator(x) > 0))
	return(sc->F);
      if ((y < S7_LONG_MAX) && 
	  (y > S7_LONG_MIN) && 
	  (denominator(x) < S7_LONG_MAX) && 
	  (denominator(x) > S7_LONG_MIN))
	return(make_boolean(sc, (numerator(x) <= (y * denominator(x)))));
      return(make_boolean(sc, fraction(x) <= y));
      break;
      
    case T_REAL:    
      return(make_boolean(sc, real(x) <= y));
      break;

    default:
      return(s7_wrong_type_arg_error(sc, "<=", 1, x, "a real"));  
    }
  return(sc->T);
}

static s7_pointer leq_2;
static s7_pointer g_leq_2(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, y;

  x = car(args);
  y = cadr(args);

  switch (type(x))
    {
    case T_INTEGER:
      switch (type(y))
	{
	case T_INTEGER:
	  return(make_boolean(sc, integer(x) <= integer(y)));
	  break;
	  
	case T_RATIO: 
	  return(g_less_or_equal(sc, args));
	  break;
	  
	case T_REAL:
	  if (is_NaN(real(y))) return(sc->F);
	  return(make_boolean(sc, integer(x) <= real(y)));
	  break;

	default:
	  return(s7_wrong_type_arg_error(sc, "<=", 2, y, "a real"));	  
	}
      break;
      
    case T_RATIO:
      return(g_less_or_equal(sc, args));
      break;
      
    case T_REAL:
      switch (type(y))
	{
	case T_INTEGER: 
	  if (is_NaN(real(x))) return(sc->F);
	  return(make_boolean(sc, real(x) <= integer(y)));
	  break;
	  
	case T_RATIO:
	  if (is_NaN(real(x))) return(sc->F);
	  return(make_boolean(sc, real(x) <= fraction(y)));
	  break;
	  
	case T_REAL:
	  if (is_NaN(real(x))) return(sc->F);
	  if (is_NaN(real(y))) return(sc->F);
	  return(make_boolean(sc, real(x) <= real(y)));
	  break;

	default:
	  return(s7_wrong_type_arg_error(sc, "<=", 2, y, "a real"));	  
	}
      break;

    default:
      return(s7_wrong_type_arg_error(sc, "<=", 1, x, "a real"));
    }

  return(sc->T);
}

static s7_pointer greater_s_ic;
static s7_pointer g_greater_s_ic(s7_scheme *sc, s7_pointer args)
{
  s7_Int y;
  s7_pointer x;
  
  x = car(args);
  y = s7_integer(cadr(args));
  switch (type(x))
    {
    case T_INTEGER:
      return(make_boolean(sc, integer(x) > y));
      break;
      
    case T_RATIO:  
      if ((y >= 0) && (numerator(x) < 0))
	return(sc->F);
      if ((y <= 0) && (numerator(x) > 0))
	return(sc->T);
      if ((y < S7_LONG_MAX) && 
	  (y > S7_LONG_MIN) && 
	  (denominator(x) < S7_LONG_MAX) && 
	  (denominator(x) > S7_LONG_MIN))
	return(make_boolean(sc, (numerator(x) > (y * denominator(x)))));
      return(make_boolean(sc, fraction(x) > y));
      break;
      
    case T_REAL:    
      return(make_boolean(sc, real(x) > y));
      break;

    default:
      return(s7_wrong_type_arg_error(sc, ">", 1, car(args), "a number"));  
    }
  return(sc->T);
}


static s7_pointer greater_2;
static s7_pointer g_greater_2(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, y;

  x = car(args);
  y = cadr(args);

  switch (type(x))
    {
    case T_INTEGER:
      switch (type(y))
	{
	case T_INTEGER: 
	  return(make_boolean(sc, integer(x) > integer(y)));

	case T_RATIO:   
	  return(g_greater(sc, args));

	case T_REAL:
	  if (is_NaN(real(y))) return(sc->F);
	  return(make_boolean(sc, integer(x) > real(y)));
	  break;

	default:       
	  return(s7_wrong_type_arg_error(sc, ">", 2, y, "a real"));
	}
      break;
      
    case T_RATIO:
      return(g_greater(sc, args));
      break;
      
    case T_REAL:
      switch (type(y))
	{
	case T_INTEGER: 
	  if (is_NaN(real(x))) return(sc->F);
	  return(make_boolean(sc, real(x) > integer(y)));
	  break;
	  
	case T_RATIO:
	  if (is_NaN(real(x))) return(sc->F);
	  return(make_boolean(sc, real(x) > fraction(y)));
	  break;
	  
	case T_REAL:
	  if (is_NaN(real(x))) return(sc->F);
	  if (is_NaN(real(y))) return(sc->F);
	  return(make_boolean(sc, real(x) > real(y)));
	  break;

	default:
	  return(s7_wrong_type_arg_error(sc, ">", 2, y, "a real"));
	}
      break;

    default:
      return(s7_wrong_type_arg_error(sc, ">", 1, x, "a real"));
    }

  return(sc->T);
}


/* (define (hi a b) (> (abs (- a b)) .1)) */
static s7_pointer greater_abs;
static s7_pointer g_greater_abs(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, y;
  s7_Double a;

  x = finder(sc, cadr(cadr(car(args))));
  if (!s7_is_real(x))
    return(s7_wrong_type_arg_error(sc, "-", 1, x, "a real"));

  y = finder(sc, caddr(cadr(car(args))));
  if (!s7_is_real(y))
    return(s7_wrong_type_arg_error(sc, "-", 2, y, "a real"));

  a = s7_number_to_real(x) - s7_number_to_real(y);

  return(make_boolean(sc, s7_Double_abs(a) > s7_real(cadr(args))));
}


static s7_pointer geq_2;
static s7_pointer g_geq_2(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, y;

  x = car(args);
  y = cadr(args);

  switch (type(x))
    {
    case T_INTEGER:
      switch (type(y))
	{
	case T_INTEGER:
	  return(make_boolean(sc, integer(x) >= integer(y)));
	  break;
	  
	case T_RATIO: 
	  return(g_greater_or_equal(sc, args));
	  break;
	  
	case T_REAL:
	  if (is_NaN(real(y))) return(sc->F);
	  return(make_boolean(sc, integer(x) >= real(y)));
	  break;

	default:
	  return(s7_wrong_type_arg_error(sc, ">=", 2, y, "a real"));	  
	}
      break;
      
    case T_RATIO:
      return(g_greater_or_equal(sc, args));
      break;
      
    case T_REAL:
      switch (type(y))
	{
	case T_INTEGER: 
	  if (is_NaN(real(x))) return(sc->F);
	  return(make_boolean(sc, real(x) >= integer(y)));
	  break;
	  
	case T_RATIO:
	  if (is_NaN(real(x))) return(sc->F);
	  return(make_boolean(sc, real(x) >= fraction(y)));
	  break;
	  
	case T_REAL:
	  if (is_NaN(real(x))) return(sc->F);
	  if (is_NaN(real(y))) return(sc->F);
	  return(make_boolean(sc, real(x) >= real(y)));
	  break;

	default:
	  return(s7_wrong_type_arg_error(sc, ">=", 2, y, "a real"));	  
	}
      break;

    default:
      return(s7_wrong_type_arg_error(sc, ">=", 1, x, "a real"));
    }

  return(sc->T);
}

static s7_pointer geq_length_ic;
static s7_pointer g_geq_length_ic(s7_scheme *sc, s7_pointer args)
{
  return(make_boolean(sc, is_false(sc, g_less_length_ic(sc, args))));
}


static s7_pointer geq_s_ic;
static s7_pointer g_geq_s_ic(s7_scheme *sc, s7_pointer args)
{
  s7_Int y;
  s7_pointer x;
  
  x = car(args);
  y = s7_integer(cadr(args));

  switch (type(x))
    {
    case T_INTEGER:
      return(make_boolean(sc, integer(x) >= y));
      break;
      
    case T_RATIO:  
      if ((y >= 0) && (numerator(x) < 0))
	return(sc->F);
      if ((y <= 0) && (numerator(x) >= 0))
	return(sc->T);
      if ((y < S7_LONG_MAX) && 
	  (y > S7_LONG_MIN) && 
	  (denominator(x) < S7_LONG_MAX) && 
	  (denominator(x) > S7_LONG_MIN))
	return(make_boolean(sc, (numerator(x) >= (y * denominator(x)))));
      return(make_boolean(sc, fraction(x) >= y));
      break;
      
    case T_REAL:    
      return(make_boolean(sc, real(x) >= y));
      break;

    default:
      return(s7_wrong_type_arg_error(sc, ">=", 1, x, "a real"));  
    }
  return(sc->T);
}
#endif
#endif



/* ---------------------------------------- real-part imag-part ---------------------------------------- */

#if (!WITH_GMP)
static s7_pointer g_real_part(s7_scheme *sc, s7_pointer args)
{
  #define H_real_part "(real-part num) returns the real part of num"
  s7_pointer p;

  p = car(args);
  switch (type(p))
    {
    case T_INTEGER:
    case T_RATIO:
    case T_REAL:
      return(p);

    case T_COMPLEX:
      return(s7_make_real(sc, real_part(p)));

    default:
      return(s7_wrong_type_arg_error(sc, "real-part", 0, p, "a number"));
    }
}


static s7_pointer g_imag_part(s7_scheme *sc, s7_pointer args)
{
  #define H_imag_part "(imag-part num) returns the imaginary part of num"
  s7_pointer p;
  /* currently (imag-part nan.0) -> 0.0 ? it's true but maybe confusing */

  p = car(args);
  switch (type(p))
    {
    case T_INTEGER:   
    case T_RATIO:   return(small_int(0));
    case T_REAL:    return(real_zero);
    case T_COMPLEX: return(s7_make_real(sc, imag_part(p)));
    default:        return(s7_wrong_type_arg_error(sc, "imag-part", 0, p, "a number"));
    }
}
#endif



/* ---------------------------------------- numerator denominator ---------------------------------------- */

static s7_pointer g_numerator(s7_scheme *sc, s7_pointer args)
{
  #define H_numerator "(numerator rat) returns the numerator of the rational number rat"
  s7_pointer x;

  x = car(args);
  switch (type(x))
    {
    case T_RATIO:
      return(make_integer(sc, numerator(x)));
      
    case T_INTEGER:
      return(x);

    case T_REAL:
    case T_COMPLEX:
    default:
      return(s7_wrong_type_arg_error(sc, "numerator", 0, x, "a rational number"));
    }
}


static s7_pointer g_denominator(s7_scheme *sc, s7_pointer args)
{
  #define H_denominator "(denominator rat) returns the denominator of the rational number rat"
  s7_pointer x;

  x = car(args);
  switch (type(x))
    {
    case T_RATIO:
      return(make_integer(sc, denominator(x)));
      
    case T_INTEGER:
      return(small_int(1));

    case T_REAL:
    case T_COMPLEX:
    default:
      return(s7_wrong_type_arg_error(sc, "denominator", 0, x, "a rational number"));
    }
}


/* ---------------------------------------- nan? infinite? ---------------------------------------- */

#if (!WITH_GMP)
static s7_pointer g_is_nan(s7_scheme *sc, s7_pointer args) 
{
  #define H_is_nan "(nan? obj) returns #t if obj is a NaN"
  /* s7 thinks a NaN is a real. */
  s7_pointer x;

  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
    case T_RATIO:
      return(sc->F);
	  
    case T_REAL:
      return(make_boolean(sc, is_NaN(real(x))));
	  
    case T_COMPLEX:
      return(make_boolean(sc, (is_NaN(real_part(x))) || (is_NaN(imag_part(x)))));

    default:
      return(sc->F);
    }
}


static s7_pointer g_is_infinite(s7_scheme *sc, s7_pointer args) 
{
  #define H_is_infinite "(infinite? obj) returns #t if obj is an infinite real"
  s7_pointer x;
  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
    case T_RATIO:
      return(sc->F);

    case T_REAL:
      return(make_boolean(sc, isinf(real(x))));

    case T_COMPLEX:
      return(make_boolean(sc, (isinf(real_part(x))) || (isinf(imag_part(x)))));

    default:
      return(sc->F);
    }
}
#endif



/* ---------------------------------------- number? complex? integer? rational? real?  ---------------------------------------- */

static s7_pointer g_is_number(s7_scheme *sc, s7_pointer args) 
{
  #define H_is_number "(number? obj) returns #t if obj is a number"
  return(make_boolean(sc, s7_is_number(car(args))));    /* we need the s7_* versions here for the GMP case */
}


static s7_pointer g_is_integer(s7_scheme *sc, s7_pointer args) 
{
  #define H_is_integer "(integer? obj) returns #t if obj is an integer"
  return(make_boolean(sc, s7_is_integer(car(args))));
}


static s7_pointer g_is_real(s7_scheme *sc, s7_pointer args) 
{
  #define H_is_real "(real? obj) returns #t if obj is a real number"
  return(make_boolean(sc, s7_is_real(car(args))));
}


static s7_pointer g_is_complex(s7_scheme *sc, s7_pointer args) 
{
  #define H_is_complex "(complex? obj) returns #t if obj is a number"
  return(make_boolean(sc, s7_is_number(car(args))));
}


static s7_pointer g_is_rational(s7_scheme *sc, s7_pointer args) 
{
  #define H_is_rational "(rational? obj) returns #t if obj is a rational number (either an integer or a ratio)"
  return(make_boolean(sc, s7_is_rational(car(args))));

  /* in the non-gmp case, (rational? 455702434782048082459/86885567283849955830) -> #f, not #t
   *  and similarly for exact? etc.
   */
}


/* ---------------------------------------- even? odd?---------------------------------------- */

static s7_pointer g_is_even(s7_scheme *sc, s7_pointer args)
{
  #define H_is_even "(even? int) returns #t if the integer int is even"
  if (!s7_is_integer(car(args)))
    return(s7_wrong_type_arg_error(sc, "even?", 0, car(args), "an integer"));
  return(make_boolean(sc, (integer(car(args)) & 1) == 0));

  /* extension to gaussian integers: odd if a+b is odd? */
}


static s7_pointer g_is_odd(s7_scheme *sc, s7_pointer args)
{
  #define H_is_odd "(odd? int) returns #t if the integer int is odd"
  if (!s7_is_integer(car(args)))
    return(s7_wrong_type_arg_error(sc, "odd?", 0, car(args), "an integer"));
  return(make_boolean(sc, (integer(car(args)) & 1) == 1));
}


/* ---------------------------------------- zero? positive? negative? ---------------------------------------- */

#if (!WITH_GMP)
static s7_pointer g_is_zero(s7_scheme *sc, s7_pointer args)
{
  #define H_is_zero "(zero? num) returns #t if the number num is zero"
  s7_pointer x;

  x = car(args);
  switch (type(x))
    {
    case T_INTEGER: 
      return(make_boolean(sc, integer(x) == 0));

    case T_REAL:    
      return(make_boolean(sc, real(x) == 0.0));

    case T_RATIO:
    case T_COMPLEX:
      /* ratios and complex numbers are already collapsed into integers and reals */
      return(sc->F);

    default:
      return(s7_wrong_type_arg_error(sc, "zero?", 0, x, "a number"));
    }
}


static s7_pointer g_is_positive(s7_scheme *sc, s7_pointer args)
{
  #define H_is_positive "(positive? num) returns #t if the real number num is positive (greater than 0)"
  s7_pointer x;
  
  x = car(args);
  switch (type(x))
    {
    case T_INTEGER: return(make_boolean(sc, integer(x) > 0));
    case T_RATIO:   return(make_boolean(sc, numerator(x) > 0));
    case T_REAL:    return(make_boolean(sc, real(x) > 0.0));
    default:        return(s7_wrong_type_arg_error(sc, "positive?", 0, x, "a real"));
    }
}


#if WITH_OPTIMIZATION
static s7_pointer is_negative_length;
static s7_pointer g_is_negative_length(s7_scheme *sc, s7_pointer args)
{
  s7_pointer val;
  val = finder(sc, cadar(args));

  switch (type(val))
    {
    case T_PAIR:
      return(make_boolean(sc, s7_list_length(sc, val) < 0));

    case T_VECTOR:
    case T_NIL:
    case T_STRING:
    case T_HASH_TABLE:
    case T_ENVIRONMENT:
      return(sc->F);

    case T_C_OBJECT:
      return(make_boolean(sc, s7_integer(object_length(sc, val)) < 0));

    default:
      return(s7_wrong_type_arg_error(sc, "length", 0, val, "a list, vector, string, or hash-table"));
    }
  return(sc->F);
}

static s7_pointer is_zero_logand_ss;
static s7_pointer g_is_zero_logand_ss(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, y;

  x = finder(sc, cadr(car(args)));
  if (!s7_is_integer(x))
    return(s7_wrong_type_arg_error(sc, "logand", 1, cadr(car(args)), "an integer"));

  y = finder(sc, caddr(car(args)));
  if (!s7_is_integer(y))
    return(s7_wrong_type_arg_error(sc, "logand", 2, caddr(car(args)), "an integer"));

  return(make_boolean(sc, (s7_integer(x) & s7_integer(y)) == 0));
}
#endif


static s7_pointer g_is_negative(s7_scheme *sc, s7_pointer args)
{
  #define H_is_negative "(negative? num) returns #t if the real number num is negative (less than 0)"
  s7_pointer x;
  
  x = car(args);
  switch (type(x))
    {
    case T_INTEGER: return(make_boolean(sc, integer(x) < 0));
    case T_RATIO:   return(make_boolean(sc, numerator(x) < 0));
    case T_REAL:    return(make_boolean(sc, real(x) < 0.0));
    default:        return(s7_wrong_type_arg_error(sc, "negative?", 0, x, "a real"));
    }
}


/* ---------------------------------------- exact<->inexact exact? inexact? ---------------------------------------- */

static s7_pointer g_exact_to_inexact(s7_scheme *sc, s7_pointer args)
{
  #define H_exact_to_inexact "(exact->inexact num) converts num to an inexact number; (exact->inexact 3/2) = 1.5"
  return(exact_to_inexact(sc, car(args)));
}


static s7_pointer g_inexact_to_exact(s7_scheme *sc, s7_pointer args)
{
  #define H_inexact_to_exact "(inexact->exact num) converts num to an exact number; (inexact->exact 1.5) = 3/2"
  return(inexact_to_exact(sc, car(args)));
}

#endif
/* (!WITH_GMP) */


static s7_pointer g_is_exact(s7_scheme *sc, s7_pointer args)
{
  #define H_is_exact "(exact? num) returns #t if num is exact (an integer or a ratio)"
  s7_pointer x;
  
  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
    case T_RATIO:
      return(sc->T);

    case T_REAL:
    case T_COMPLEX:
      return(sc->F);

#if WITH_GMP
    case T_BIG_INTEGER:
    case T_BIG_RATIO:
      return(sc->T);

    case T_BIG_REAL:
    case T_BIG_COMPLEX:
      return(sc->F);
#endif

    default:
      return(s7_wrong_type_arg_error(sc, "exact?", 0, x, "a number"));
    }
}


static s7_pointer g_is_inexact(s7_scheme *sc, s7_pointer args)
{
  #define H_is_inexact "(inexact? num) returns #t if num is inexact (neither an integer nor a ratio)"
  s7_pointer x;
  
  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
    case T_RATIO:
      return(sc->F);

    case T_REAL:
    case T_COMPLEX:
      return(sc->T);

#if WITH_GMP
    case T_BIG_INTEGER:
    case T_BIG_RATIO:
      return(sc->F);

    case T_BIG_REAL:
    case T_BIG_COMPLEX:
      return(sc->T);
#endif

    default:
      return(s7_wrong_type_arg_error(sc, "inexact?", 0, x, "a number"));
    }
}


bool s7_is_ulong(s7_pointer arg) 
{
  return(is_integer(arg));
}


unsigned long s7_ulong(s7_pointer p) 
{
  return(p->object.number.ul_value);
}


s7_pointer s7_make_ulong(s7_scheme *sc, unsigned long n) 
{
  s7_pointer x;
  NEW_CELL(sc, x);
  set_type(x, T_INTEGER);
  x->object.number.ul_value = n;
  return(x);
}


bool s7_is_ulong_long(s7_pointer arg) 
{
  return(is_integer(arg));
}


unsigned long long s7_ulong_long(s7_pointer p) 
{
  return(p->object.number.ull_value);
}


s7_pointer s7_make_ulong_long(s7_scheme *sc, unsigned long long n) 
{
  s7_pointer x;
  NEW_CELL(sc, x);
  set_type(x, T_INTEGER);
  x->object.number.ull_value = n;
  return(x);
}


/* ---------------------------------------- integer-length, integer-decode-float ---------------------------------------- */

static s7_pointer g_integer_length(s7_scheme *sc, s7_pointer args)
{
  #define H_integer_length "(integer-length arg) returns the number of bits required to represent the integer 'arg': (ceiling (log (abs arg) 2))"
  s7_Int x;
  if (!s7_is_integer(car(args)))
      return(s7_wrong_type_arg_error(sc, "integer-length", 0, car(args), "an integer"));
    
  x = s7_integer(car(args));
  if (x < 0)
    return(make_integer(sc, integer_length(-(x + 1))));
  return(make_integer(sc, integer_length(x)));
}


#if WITH_GMP
typedef struct decode_float_t {
  union {
    s7_Int integer_value;
    s7_Double real_value;
  } value;
} decode_float_t;
#endif

static s7_pointer g_integer_decode_float(s7_scheme *sc, s7_pointer args)
{
  #define H_integer_decode_float "(integer-decode-float x) returns a list containing the significand, exponent, and \
sign of 'x' (1 = positive, -1 = negative).  (integer-decode-float 0.0): (0 0 1)"

  s7_Int ix;
  s7_pointer arg;
  arg = car(args);

  /* frexp doesn't work in edge cases.  Since the double and long long int fields are equivalenced
   *   in the s7_num struct, we can get the actual bits of the double from the int.  The problem with doing this
   *   is that bignums don't use that struct.  Assume IEEE 754 and double = s7_Double.
   */

  if ((!s7_is_real(arg)) ||
      (s7_is_rational(arg)))
    return(s7_wrong_type_arg_error(sc, "integer-decode-float", 0, arg, "a non-rational real"));

  if (s7_real(arg) == 0.0)
    return(list_3(sc, small_int(0), small_int(0), small_int(1)));

#if WITH_GMP
  if (s7_is_bignum(arg)) 
    {
      decode_float_t num;
      num.value.real_value = s7_number_to_real(arg);
      if ((is_NaN(num.value.real_value)) || (isinf(num.value.real_value)))   /* (integer-decode-float (bignum "1e310")) */
	return(s7_out_of_range_error(sc, "integer-decode-float", 0, arg, "a real that s7_Double can handle"));
      ix = num.value.integer_value;
    }
  else
#endif

  ix = integer(arg);
  return(list_3(sc,
		make_integer(sc, (s7_Int)((ix & 0xfffffffffffffLL) | 0x10000000000000LL)),
		make_integer(sc, (s7_Int)(((ix & 0x7fffffffffffffffLL) >> 52) - 1023 - 52)),
		make_integer(sc, ((ix & 0x8000000000000000LL) != 0) ? -1 : 1)));
}


/* ---------------------------------------- logand, logior, logxor, lognot, logbit?, ash ---------------------------------------- */

static s7_pointer g_logior(s7_scheme *sc, s7_pointer args)
{
  #define H_logior "(logior int ...) returns the bitwise OR of its integer arguments (the bits that are on in any of the arguments)"
  s7_Int result = 0;
  s7_pointer x;

  for (x = args; is_not_null(x); x = cdr(x))
    if (!s7_is_integer(car(x)))
      return(s7_wrong_type_arg_error(sc, "logior", position_of(x, args), car(x), "an integer"));
    else result |= s7_integer(car(x));

  return(make_integer(sc, result));
}


static s7_pointer g_logxor(s7_scheme *sc, s7_pointer args)
{
  #define H_logxor "(logxor int ...) returns the bitwise XOR of its integer arguments (the bits that are on in an odd number of the arguments)"
  s7_Int result = 0;
  s7_pointer x;

  for (x = args; is_not_null(x); x = cdr(x))
    if (!s7_is_integer(car(x)))
      return(s7_wrong_type_arg_error(sc, "logxor", position_of(x, args), car(x), "an integer"));
    else result ^= s7_integer(car(x));

  return(make_integer(sc, result));
}


static s7_pointer g_logand(s7_scheme *sc, s7_pointer args)
{
  #define H_logand "(logand int ...) returns the bitwise AND of its integer arguments (the bits that are on in every argument)"
  s7_Int result = -1;
  s7_pointer x;

  for (x = args; is_not_null(x); x = cdr(x))
    if (!s7_is_integer(car(x)))
      return(s7_wrong_type_arg_error(sc, "logand", position_of(x, args), car(x), "an integer"));
    else result &= s7_integer(car(x));

  return(make_integer(sc, result));
}


static s7_pointer g_lognot(s7_scheme *sc, s7_pointer args)
{
  #define H_lognot "(lognot num) returns the bitwise negation (the complement, the bits that are not on) in num: (lognot 0) -> -1"
  if (!s7_is_integer(car(args)))
    return(s7_wrong_type_arg_error(sc, "lognot", 0, car(args), "an integer"));
  return(make_integer(sc, ~s7_integer(car(args))));
}


/* logbit?  CL is (logbitp index int) using 2^index, but that order strikes me as backwards
 *   at least gmp got the arg order right!
 */

static s7_pointer g_logbit(s7_scheme *sc, s7_pointer args)
{
  #define H_logbit "(logbit? int index) returns #t if the index-th bit is on in int, otherwise #f. The argument \
order here follows gmp, and is the opposite of the CL convention.  (logbit? int bit) is the same as (not (zero? (logand int (ash 1 bit))))."

  s7_pointer x, y;
  s7_Int index;      /* index in gmp is mp_bitcnt which is an unsigned long int */

  x = car(args);
  y = cadr(args);

  if (!s7_is_integer(y))
    return(s7_wrong_type_arg_error(sc, "logbit?", 2, y, "an integer"));
  index = s7_integer(y);
  if (index < 0)
    return(s7_out_of_range_error(sc, "logbit?", 2, y, "index must be non-negative"));

#if WITH_GMP
  if (type(x) == T_BIG_INTEGER)
    return(make_boolean(sc, (mpz_tstbit(big_integer(x), index) != 0)));
#endif
  
  if (index >= s7_int_bits)           /* not sure about the >: (logbit? -1 64) ?? */
    return(make_boolean(sc, integer(x) < 0));

  /* :(zero? (logand most-positive-fixnum (ash 1 63)))
   *   -> ash argument 2, 63, is out of range (shift is too large) 
   *   so logbit? has a wider range than the logand/ash shuffle above.
   */

  if (!s7_is_integer(x))
    return(s7_wrong_type_arg_error(sc, "logbit?", 1, x, "an integer"));

  /* all these long long ints are necessary, else C turns it into an int, gets confused about signs etc */
  return(make_boolean(sc, ((((long long int)(1LL << (long long int)index)) & (long long int)integer(x)) != 0)));
}


static s7_pointer g_ash(s7_scheme *sc, s7_pointer args)
{
  #define H_ash "(ash i1 i2) returns i1 shifted right or left i2 times, i1 << i2, (ash 1 3) -> 8, (ash 8 -3) -> 1"
  s7_Int arg1, arg2;
  s7_pointer x, y;

  x = car(args);
  if (!s7_is_integer(x))
    return(s7_wrong_type_arg_error(sc, "ash", 1, x, "an integer"));

  y = cadr(args);
  if (!s7_is_integer(y))
    return(s7_wrong_type_arg_error(sc, "ash", 2, y, "an integer"));
  
  arg1 = s7_integer(x);
  if (arg1 == 0) return(small_int(0));

  arg2 = s7_integer(y);
  if (arg2 >= s7_int_bits)
    return(s7_out_of_range_error(sc, "ash", 2, cadr(args), "shift is too large"));

  if (arg2 < -s7_int_bits)
    {
      if (arg1 < 0)                      /* (ash -31 -100) */
	return(make_integer(sc, -1));
      return(small_int(0));
    }

  /* I can't see any point in protecting this:
   *    (ash 9223372036854775807 1) -> -2
   * but anyone using ash must know something about bits...
   */
  if (arg2 >= 0)
    return(make_integer(sc, arg1 << arg2));
  return(make_integer(sc, arg1 >> -arg2));
}


/* ---------------------------------------- random ---------------------------------------- */

/* random numbers.  The simple version used in clm.c is probably adequate,
 *   but here I'll use Marsaglia's MWC algorithm as an experiment.
 *     (random num) -> a number (0..num), if num == 0 return 0, use global default state
 *     (random num state) -> same but use this state
 *     (make-random-state seed) -> make a new state
 *     (make-random-state seed type) ??
 *   to save the current seed, use copy
 *   to save it across load, random-state->list and list->random-state.
 */

typedef struct {
  unsigned long long int ran_seed, ran_carry;
} s7_rng_t;

static int rng_tag = 0;

#if WITH_GMP
static int big_rng_tag = 0;
#endif

static char *print_rng(s7_scheme *sc, void *val)
{
  char *buf;
  s7_rng_t *r = (s7_rng_t *)val;
  buf = (char *)malloc(64 * sizeof(char));
  snprintf(buf, 64, "#<rng %d %d>", (unsigned int)(r->ran_seed), (unsigned int)(r->ran_carry));
  return(buf);
}


static void free_rng(void *val)
{
  free(val);
}


static bool equal_rng(void *val1, void *val2)
{
  return(val1 == val2);
}


s7_pointer s7_make_random_state(s7_scheme *sc, s7_pointer args)
{
  #define H_make_random_state "(make-random-state seed) returns a new random number state initialized with 'seed'. \
Pass this as the second argument to 'random' to get a repeatable random number sequence:\n\
    (let ((seed (make-random-state 1234))) (random 1.0 seed))"

  s7_rng_t *r;

  if (!(s7_is_integer(car(args))))
    return(s7_wrong_type_arg_error(sc, "make-random-state,", 1, car(args), "an integer"));

  if (is_null(cdr(args)))
    {
      r = (s7_rng_t *)calloc(1, sizeof(s7_rng_t));
      r->ran_seed = s7_integer(car(args));
      r->ran_carry = 1675393560;  /* should this be dependent on the seed? */
      return(s7_make_object(sc, rng_tag, (void *)r));
    }

  if (!(s7_is_integer(cadr(args))))
    return(s7_wrong_type_arg_error(sc, "make-random-state,", 2, cadr(args), "an integer"));
  
  r = (s7_rng_t *)calloc(1, sizeof(s7_rng_t));
  r->ran_seed = s7_integer(car(args));
  r->ran_carry = s7_integer(cadr(args));
  return(s7_make_object(sc, rng_tag, (void *)r));
}


static s7_pointer copy_random_state(s7_scheme *sc, s7_pointer obj)
{
  if (object_type(obj) == rng_tag)
    {
      s7_rng_t *r, *new_r;
      r = (s7_rng_t *)s7_object_value(obj);
      new_r = (s7_rng_t *)calloc(1, sizeof(s7_rng_t));
      new_r->ran_seed = r->ran_seed;
      new_r->ran_carry = r->ran_carry;
      return(s7_make_object(sc, rng_tag, (void *)new_r));
    }
  /* I can't find a way to copy a gmp random generator */
  return(sc->F);
}


static double next_random(s7_rng_t *r)
{
  /* The multiply-with-carry generator for 32-bit integers: 
   *        x(n)=a*x(n-1) + carry mod 2^32 
   * Choose multiplier a from this list: 
   *   1791398085 1929682203 1683268614 1965537969 1675393560 
   *   1967773755 1517746329 1447497129 1655692410 1606218150 
   *   2051013963 1075433238 1557985959 1781943330 1893513180 
   *   1631296680 2131995753 2083801278 1873196400 1554115554 
   * ( or any 'a' for which both a*2^32-1 and a*2^31-1 are prime) 
   */
  double result;
  unsigned long long int temp;
  #define RAN_MULT 2131995753UL

  temp = r->ran_seed * RAN_MULT + r->ran_carry;
  r->ran_seed = (temp & 0xffffffffUL);
  r->ran_carry = (temp >> 32);
  result = (double)((unsigned int)(r->ran_seed)) / 4294967295.5;
  /* divisor was 2^32-1 = 4294967295.0, but somehow this can round up once in a billion tries? 
   *   do we want the double just less than 2^32?
   */

  /* (let ((mx 0) (mn 1000)) (do ((i 0 (+ i 1))) ((= i 10000)) (let ((val (random 123))) (set! mx (max mx val)) (set! mn (min mn val)))) (list mn mx)) */
  return(result);
}


static s7_rng_t *s7_default_rng(s7_scheme *sc)
{
  if (!sc->default_rng)
    {
      sc->default_rng = (s7_rng_t *)calloc(1, sizeof(s7_rng_t));
      ((s7_rng_t *)(sc->default_rng))->ran_seed = (unsigned int)time(NULL);
      ((s7_rng_t *)(sc->default_rng))->ran_carry = 1675393560;
    }
  return((s7_rng_t *)(sc->default_rng));
}


s7_pointer s7_random_state_to_list(s7_scheme *sc, s7_pointer args)
{
  #define H_random_state_to_list "(random-state->list r) returns the random state object as a list.\
You can later apply make-random-state to this list to continue a random number sequence from any point."

  s7_rng_t *r = NULL;
  s7_pointer obj;
  if (is_null(args))
    r = s7_default_rng(sc);
  else
    {
      obj = car(args);
      if ((!is_c_object(obj)) ||
	  (object_type(obj) != rng_tag))
	return(s7_wrong_type_arg_error(sc, "random-state->list,", 1, obj, "a random state as returned by make-random-state"));

	r = (s7_rng_t *)s7_object_value(obj);
    }
  
  if (r)
    return(list_2(sc, 
		  make_integer(sc, r->ran_seed), 
		  make_integer(sc, r->ran_carry)));
  return(sc->F);
}


void s7_set_default_random_state(s7_scheme *sc, s7_Int seed, s7_Int carry)
{
  sc->default_rng = (s7_rng_t *)calloc(1, sizeof(s7_rng_t));
  ((s7_rng_t *)(sc->default_rng))->ran_seed = (unsigned long long)seed;
  ((s7_rng_t *)(sc->default_rng))->ran_carry = (unsigned long long)carry;
}


s7_Double s7_random(s7_scheme *sc, s7_pointer state)
{
  if (!state)
    return(next_random(s7_default_rng(sc)));
  return(next_random((s7_rng_t *)s7_object_value(state)));
}


static s7_pointer g_random(s7_scheme *sc, s7_pointer args)
{
  #define H_random "(random num (state #f)) returns a random number between 0 and num (0 if num=0)."
  s7_pointer num, state;
  s7_rng_t *r;

  if (is_not_null(cdr(args)))
    {
      state = cadr(args);
      if (!is_c_object(state))
	return(s7_wrong_type_arg_error(sc, "random state,", 2, state, "a random state as returned by make-random-state"));

      if (object_type(state) == rng_tag)
	r = (s7_rng_t *)s7_object_value(state);
      else
	{
#if WITH_GMP
	  if (object_type(state) == big_rng_tag)
	    return(big_random(sc, args));
#endif
	  return(s7_wrong_type_arg_error(sc, "random state,", 2, state, "a random state as returned by make-random-state"));
	}
    }
  else r = s7_default_rng(sc);

  num = car(args);
  switch (type(num))
    {
    case T_INTEGER:
      return(make_integer(sc, (s7_Int)(integer(num) * next_random(r))));

    case T_RATIO:
      {
	s7_Double x, error;
	s7_Int numer = 0, denom = 1;

	/* the error here needs to take the size of the fraction into account.  Otherwise, if
	 *    error is (say) 1e-6 and the fraction is (say) 9000000/9223372036854775807,
	 *    c_rationalize will always return 0.
	 */
	x = (s7_Double)numerator(num) / (s7_Double)denominator(num);

	error = 1e-6 * s7_Double_abs(x);
	if (error > 1e-6)
	  error = 1e-6;
	else
	  {
	    if (error < 1e-18)
	      error = 1e-18;
	  }
	c_rationalize(x * next_random(r), error, &numer, &denom);
	return(s7_make_ratio(sc, numer, denom));
      }

    case T_REAL:
      return(s7_make_real(sc, real(num) * next_random(r)));

    case T_COMPLEX:
      return(s7_make_complex(sc, real_part(num) * next_random(r), imag_part(num) * next_random(r)));

    default:
      return(s7_wrong_type_arg_error(sc, "random bounds,", (is_null(cdr(args))) ? 0 : 1, num, "a number"));
    }
  return(sc->F);
}




/* -------------------------------- characters -------------------------------- */

static s7_pointer g_char_to_integer(s7_scheme *sc, s7_pointer args)
{
  #define H_char_to_integer "(char->integer c) converts the character c to an integer"
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char->integer", 0, car(args), "a character"));
  return(small_int(character(car(args))));
}


static s7_pointer g_integer_to_char(s7_scheme *sc, s7_pointer args)
{
  #define H_integer_to_char "(integer->char i) converts the non-negative integer i to a character"
  s7_pointer x;
  s7_Int ind;                   /* not int here!  (integer->char (expt 2 32)) -> #\null */
  x = car(args);

  if (!s7_is_integer(x))
    return(s7_wrong_type_arg_error(sc, "integer->char", 0, x, "an integer"));
  ind = s7_integer(x);
  if ((ind < 0) || (ind >= NUM_CHARS))
    return(s7_wrong_type_arg_error(sc, "integer->char", 0, x, "an integer that can represent a character"));

  return(s7_make_character(sc, (unsigned char)ind));
}


static unsigned char uppers[256], lowers[256];
static void init_uppers(void)
{
  int i;
  for (i = 0; i < 256; i++)
    {
      uppers[i] = (unsigned char)toupper(i);
      lowers[i] = (unsigned char)tolower(i);
    }
  /* setlocale can change these choices, but I think if we avoid both toupper and tolower after
   *   this initialization, we'll at least be internally consistent (what about isupper and islower? see s7test.scm)
   */
}

static s7_pointer g_char_upcase(s7_scheme *sc, s7_pointer args)
{
  #define H_char_upcase "(char-upcase c) converts the character c to upper case"
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char-upcase", 0, car(args), "a character"));

  return(s7_make_character(sc, uppers[character(car(args))]));
}


static s7_pointer g_char_downcase(s7_scheme *sc, s7_pointer args)
{
  #define H_char_downcase "(char-downcase c) converts the character c to lower case"
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char-downcase", 0, car(args), "a character"));
  return(s7_make_character(sc, lowers[character(car(args))]));
}


static s7_pointer g_is_char_alphabetic(s7_scheme *sc, s7_pointer args)
{
  #define H_is_char_alphabetic "(char-alphabetic? c) returns #t if the character c is alphabetic"
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char-alphabetic?", 0, car(args), "a character"));
  return(make_boolean(sc, isalpha(character(car(args)))));

  /* isalpha returns #t for (integer->char 226) and others in that range */
}


static s7_pointer g_is_char_numeric(s7_scheme *sc, s7_pointer args)
{
  #define H_is_char_numeric "(char-numeric? c) returns #t if the character c is a digit"
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char-numeric?", 0, car(args), "a character"));
  return(make_boolean(sc, isdigit(character(car(args)))));
}


static s7_pointer g_is_char_whitespace(s7_scheme *sc, s7_pointer args)
{
  #define H_is_char_whitespace "(char-whitespace? c) returns #t if the character c is non-printing character"
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char-whitespace?", 0, car(args), "a character"));
  return(make_boolean(sc, white_space[character(car(args))]));
}


static s7_pointer g_is_char_upper_case(s7_scheme *sc, s7_pointer args)
{
  #define H_is_char_upper_case "(char-upper-case? c) returns #t if the character c is in upper case"
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char-upper-case?", 0, car(args), "a character"));
  return(make_boolean(sc, isupper(character(car(args)))));
}


static s7_pointer g_is_char_lower_case(s7_scheme *sc, s7_pointer args)
{
  #define H_is_char_lower_case "(char-lower-case? c) returns #t if the character c is in lower case"
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char-lower-case?", 0, car(args), "a character"));
  return(make_boolean(sc, islower(character(car(args)))));
}


static s7_pointer g_is_char(s7_scheme *sc, s7_pointer args)
{
  #define H_is_char "(char? obj) returns #t if obj is a character"
  return(make_boolean(sc, s7_is_character(car(args))));
}


s7_pointer s7_make_character(s7_scheme *sc, unsigned int c) 
{
  return(chars[c]);
}


bool s7_is_character(s7_pointer p) 
{ 
  return(type(p) == T_CHARACTER);
}


char s7_character(s7_pointer p)  
{ 
  return(character(p));
}


static int charcmp(unsigned char c1, unsigned char c2, bool ci)
{
  if (ci)
    return(charcmp(uppers[c1], uppers[c2], false)); 
  /* not tolower here -- the single case is apparently supposed to be upper case
   *   this matters in a case like (char-ci<? #\_ #\e) which Guile and Gauche say is #f
   *   although (char<? #\_ #\e) is #t -- the spec does not say how to interpret this!
   */
  if (c1 == c2)
    return(0);
  if (c1 < c2)
    return(-1);
  return(1);
}


static s7_pointer g_char_cmp(s7_scheme *sc, s7_pointer args, int val, const char *name, bool ci)
{
  s7_pointer x;
  unsigned char last_chr;
  
  for (x = args; is_not_null(x); x = cdr(x))  
    if (!s7_is_character(car(x)))
      return(s7_wrong_type_arg_error(sc, name, position_of(x, args), car(x), "a character"));
  
  last_chr = character(car(args));
  for (x = cdr(args); is_not_null(x); x = cdr(x))
    {
      if (charcmp(last_chr, character(car(x)), ci) != val)
	return(sc->F);
      last_chr = character(car(x));
    }
  return(sc->T);
}


static s7_pointer g_char_cmp_not(s7_scheme *sc, s7_pointer args, int val, const char *name, bool ci)
{
  s7_pointer x;
  unsigned char last_chr;
  
  for (x = args; is_not_null(x); x = cdr(x))  
    if (!s7_is_character(car(x)))
      return(s7_wrong_type_arg_error(sc, name, position_of(x, args), car(x), "a character"));
  
  last_chr = character(car(args));
  for (x = cdr(args); is_not_null(x); x = cdr(x))
    {
      if (charcmp(last_chr, character(car(x)), ci) == val)
	return(sc->F);
      last_chr = character(car(x));
    }
  return(sc->T);
}


static s7_pointer g_chars_are_equal(s7_scheme *sc, s7_pointer args)
{
  #define H_chars_are_equal "(char=? char ...) returns #t if all the character arguments are equal"

  s7_pointer x, y;
  
  for (x = args; is_not_null(x); x = cdr(x))  
    if (!s7_is_character(car(x)))
      return(s7_wrong_type_arg_error(sc, "char=?", position_of(x, args), car(x), "a character"));
  
  y = car(args);
  for (x = cdr(args); is_not_null(x); x = cdr(x))
    if (car(x) != y)
      return(sc->F);
  return(sc->T);

}	


static s7_pointer g_chars_are_less(s7_scheme *sc, s7_pointer args)
{
  #define H_chars_are_less "(char<? char ...) returns #t if all the character arguments are increasing"
  return(g_char_cmp(sc, args, -1, "char<?", false));
}	


static s7_pointer g_chars_are_greater(s7_scheme *sc, s7_pointer args)
{
  #define H_chars_are_greater "(char>? char ...) returns #t if all the character arguments are decreasing"
  return(g_char_cmp(sc, args, 1, "char>?", false));
}


static s7_pointer g_chars_are_geq(s7_scheme *sc, s7_pointer args)
{
  #define H_chars_are_geq "(char>=? char ...) returns #t if all the character arguments are equal or decreasing"
  return(g_char_cmp_not(sc, args, -1, "char>=?", false));
}	


static s7_pointer g_chars_are_leq(s7_scheme *sc, s7_pointer args)
{
  #define H_chars_are_leq "(char<=? char ...) returns #t if all the character arguments are equal or increasing"
  return(g_char_cmp_not(sc, args, 1, "char<=?", false));
}


#if WITH_OPTIMIZATION

static s7_pointer char_equal_s_ic, char_equal_2;
static s7_pointer g_char_equal_s_ic(s7_scheme *sc, s7_pointer args)
{
  s7_pointer c;
  c = finder(sc, car(args));
  if (c == cadr(args))
    return(sc->T);
  if (!s7_is_character(c))
    return(s7_wrong_type_arg_error(sc, "char=?", 1, car(args), "a character"));
  return(sc->F);
}

static s7_pointer g_char_equal_2(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char=?", 1, car(args), "a character"));
  if (!s7_is_character(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "char=?", 2, cadr(args), "a character"));
  return(make_boolean(sc, car(args) == cadr(args)));
}


static s7_pointer char_less_s_ic, char_less_2;
static s7_pointer g_char_less_s_ic(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char<?", 1, car(args), "a character"));
  return(make_boolean(sc, character(car(args)) < character(cadr(args))));
}

static s7_pointer g_char_less_2(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char<?", 1, car(args), "a character"));
  if (!s7_is_character(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "char<?", 2, cadr(args), "a character"));
  return(make_boolean(sc, character(car(args)) < character(cadr(args))));
}


static s7_pointer char_greater_s_ic, char_greater_2;
static s7_pointer g_char_greater_s_ic(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char>?", 1, car(args), "a character"));
  return(make_boolean(sc, character(car(args)) > character(cadr(args))));
}

static s7_pointer g_char_greater_2(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char>?", 1, car(args), "a character"));
  if (!s7_is_character(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "char>?", 2, cadr(args), "a character"));
  return(make_boolean(sc, character(car(args)) > character(cadr(args))));
}


static s7_pointer char_geq_s_ic, char_geq_2;
static s7_pointer g_char_geq_s_ic(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char>=?", 1, car(args), "a character"));
  return(make_boolean(sc, character(car(args)) >= character(cadr(args))));
}

static s7_pointer g_char_geq_2(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char>=?", 1, car(args), "a character"));
  if (!s7_is_character(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "char>=?", 2, cadr(args), "a character"));
  return(make_boolean(sc, character(car(args)) >= character(cadr(args))));
}


static s7_pointer char_leq_s_ic, char_leq_2;
static s7_pointer g_char_leq_s_ic(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char<=?", 1, car(args), "a character"));
  return(make_boolean(sc, character(car(args)) <= character(cadr(args))));
}

static s7_pointer g_char_leq_2(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char<=?", 1, car(args), "a character"));
  if (!s7_is_character(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "char<=?", 2, cadr(args), "a character"));
  return(make_boolean(sc, character(car(args)) <= character(cadr(args))));
}
#endif



static s7_pointer g_chars_are_ci_equal(s7_scheme *sc, s7_pointer args)
{
  #define H_chars_are_ci_equal "(char-ci=? char ...) returns #t if all the character arguments are equal, ignoring case"
  return(g_char_cmp(sc, args, 0, "char-ci=?", true));
}


static s7_pointer g_chars_are_ci_less(s7_scheme *sc, s7_pointer args)
{
  #define H_chars_are_ci_less "(char-ci<? char ...) returns #t if all the character arguments are increasing, ignoring case"
  return(g_char_cmp(sc, args, -1, "char-ci<?", true));
}	


static s7_pointer g_chars_are_ci_greater(s7_scheme *sc, s7_pointer args)
{
  #define H_chars_are_ci_greater "(char-ci>? char ...) returns #t if all the character arguments are decreasing, ignoring case"
  return(g_char_cmp(sc, args, 1, "char-ci>?", true));
}	


static s7_pointer g_chars_are_ci_geq(s7_scheme *sc, s7_pointer args)
{
  #define H_chars_are_ci_geq "(char-ci>=? char ...) returns #t if all the character arguments are equal or decreasing, ignoring case"
  return(g_char_cmp_not(sc, args, -1, "char-ci>=?", true));
}


static s7_pointer g_chars_are_ci_leq(s7_scheme *sc, s7_pointer args)
{
  #define H_chars_are_ci_leq "(char-ci<=? char ...) returns #t if all the character arguments are equal or increasing, ignoring case"
  return(g_char_cmp_not(sc, args, 1, "char-ci<=?", true));
}


#if WITH_OPTIMIZATION

static s7_pointer char_ci_equal_s_ic, char_ci_equal_2;
static s7_pointer g_char_ci_equal_s_ic(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char-ci=?", 1, car(args), "a character"));
  return(make_boolean(sc, uppers[character(car(args))] == uppers[character(cadr(args))]));
}

static s7_pointer g_char_ci_equal_2(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char-ci=?", 1, car(args), "a character"));
  if (!s7_is_character(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "char-ci=?", 2, cadr(args), "a character"));
  return(make_boolean(sc, uppers[character(car(args))] == uppers[character(cadr(args))]));
}


static s7_pointer char_ci_less_s_ic, char_ci_less_2;
static s7_pointer g_char_ci_less_s_ic(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char-ci<?", 1, car(args), "a character"));
  return(make_boolean(sc, uppers[character(car(args))] < uppers[character(cadr(args))]));
}

static s7_pointer g_char_ci_less_2(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char-ci<?", 1, car(args), "a character"));
  if (!s7_is_character(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "char-ci<?", 2, cadr(args), "a character"));
  return(make_boolean(sc, uppers[character(car(args))] < uppers[character(cadr(args))]));
}


static s7_pointer char_ci_greater_s_ic, char_ci_greater_2;
static s7_pointer g_char_ci_greater_s_ic(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char-ci>?", 1, car(args), "a character"));
  return(make_boolean(sc, uppers[character(car(args))] > uppers[character(cadr(args))]));
}

static s7_pointer g_char_ci_greater_2(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char-ci>?", 1, car(args), "a character"));
  if (!s7_is_character(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "char-ci>?", 2, cadr(args), "a character"));
  return(make_boolean(sc, uppers[character(car(args))] > uppers[character(cadr(args))]));
}


static s7_pointer char_ci_geq_s_ic, char_ci_geq_2;
static s7_pointer g_char_ci_geq_s_ic(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char-ci>=?", 1, car(args), "a character"));
  return(make_boolean(sc, uppers[character(car(args))] >= uppers[character(cadr(args))]));
}

static s7_pointer g_char_ci_geq_2(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char-ci>=?", 1, car(args), "a character"));
  if (!s7_is_character(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "char-ci>=?", 2, cadr(args), "a character"));
  return(make_boolean(sc, uppers[character(car(args))] >= uppers[character(cadr(args))]));
}


static s7_pointer char_ci_leq_s_ic, char_ci_leq_2;
static s7_pointer g_char_ci_leq_s_ic(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char-ci<=?", 1, car(args), "a character"));
  return(make_boolean(sc, uppers[character(car(args))] <= uppers[character(cadr(args))]));
}

static s7_pointer g_char_ci_leq_2(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_character(car(args)))
    return(s7_wrong_type_arg_error(sc, "char-ci<=?", 1, car(args), "a character"));
  if (!s7_is_character(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "char-ci<=?", 2, cadr(args), "a character"));
  return(make_boolean(sc, uppers[character(car(args))] <= uppers[character(cadr(args))]));
}
#endif



/* -------------------------------- strings -------------------------------- */


s7_pointer s7_make_string_with_length(s7_scheme *sc, const char *str, int len) 
{
  s7_pointer x;
  NEW_CELL(sc, x);
  set_type(x, T_STRING | T_SAFE_PROCEDURE); /* should this follow the malloc? */
  string_value(x) = (char *)malloc((len + 1) * sizeof(char)); 
  if (len != 0)                                             /* memcpy can segfault if string_value(x) is NULL */
    memcpy((void *)string_value(x), (void *)str, len + 1);
  else string_value(x)[0] = 0;
  string_length(x) = len;
  string_hash(x) = 0;
  string_needs_free(x) = true;
  add_string(sc, x);
  return(x);
}


static s7_pointer s7_make_terminated_string_with_length(s7_scheme *sc, const char *str, int len) 
{
  s7_pointer x;
  NEW_CELL(sc, x);
  set_type(x, T_STRING | T_SAFE_PROCEDURE); /* should this follow the malloc? */
  string_value(x) = (char *)malloc((len + 1) * sizeof(char)); 
  if (len != 0)                                             /* memcpy can segfault if string_value(x) is NULL */
    memcpy((void *)string_value(x), (void *)str, len);
  string_value(x)[len] = 0;
  string_length(x) = len;
  string_hash(x) = 0;
  string_needs_free(x) = true;
  add_string(sc, x);
  return(x);
}


static s7_pointer make_string_uncopied_with_length(s7_scheme *sc, char *str, int len) 
{
  s7_pointer x;
  NEW_CELL(sc, x);
  set_type(x, T_STRING | T_SAFE_PROCEDURE);
  string_value(x) = str;
  string_length(x) = len;
  string_hash(x) = 0;
  string_needs_free(x) = true; /*assume caller has passed us a copy */
  add_string(sc, x);
  return(x);
}


static s7_pointer make_protected_string(s7_scheme *sc, const char *str)
{
  s7_pointer x;
  NEW_CELL(sc, x);
  set_type(x, T_STRING | T_IMMUTABLE | T_SAFE_PROCEDURE);
  string_value(x) = (char *)str;
  string_length(x) = safe_strlen(str);
  string_hash(x) = 0;
  string_needs_free(x) = false;
  return(x);
}


static s7_pointer make_empty_string(s7_scheme *sc, int len, char fill) 
{
  /* a permanent nil_string object saved almost nothing */
  s7_pointer x;
  NEW_CELL(sc, x);
  set_type(x, T_STRING);
  
  if (fill == 0)
    string_value(x) = (char *)calloc((len + 1), sizeof(char));
  else
    {
      string_value(x) = (char *)malloc((len + 1) * sizeof(char));
      memset((void *)(string_value(x)), fill, len);
    }
  string_value(x)[len] = 0;
  string_hash(x) = 0;
  string_length(x) = len;
  string_needs_free(x) = true;
  add_string(sc, x);
  return(x);
}


s7_pointer s7_make_string(s7_scheme *sc, const char *str) 
{
  return(s7_make_string_with_length(sc, str, safe_strlen(str)));
}


static s7_pointer make_string_uncopied(s7_scheme *sc, char *str) 
{
  return(make_string_uncopied_with_length(sc, str, safe_strlen(str)));
}


static char *make_permanent_string(const char *str)
{
  char *x;
  int len;
  len = safe_strlen(str);
  x = (char *)permanent_calloc((len + 1) * sizeof(char)); 
  memcpy((void *)x, (void *)str, len);
  return(x);
}


s7_pointer s7_make_permanent_string(const char *str) 
{
  /* for the symbol table which is never GC'd */
  s7_pointer x;
  x = (s7_cell *)permanent_calloc(sizeof(s7_cell));
  x->hloc = NOT_IN_HEAP;
  set_type(x, T_STRING | T_IMMUTABLE);
  if (str)
    {
      string_length(x) = safe_strlen(str);
      string_value(x) = (char *)permanent_calloc((string_length(x) + 1) * sizeof(char)); 
      memcpy((void *)string_value(x), (void *)str, string_length(x)); 
    }
  else 
    {
      string_value(x) = NULL;
      string_length(x) = 0;
    }
  string_hash(x) = 0;
  string_needs_free(x) = false;
  return(x);
}


bool s7_is_string(s7_pointer p)
{
  return((type(p) == T_STRING)); 
}


const char *s7_string(s7_pointer p) 
{ 
  return(string_value(p));
}


static s7_pointer g_is_string(s7_scheme *sc, s7_pointer args)
{
  #define H_is_string "(string? obj) returns #t if obj is a string"
  return(make_boolean(sc, s7_is_string(car(args))));
}


#define MAX_STRING_LENGTH 1073741824

static s7_pointer g_make_string(s7_scheme *sc, s7_pointer args)
{
  #define H_make_string "(make-string len (val #\\space)) makes a string of length len filled with the character val (default: space)"
  s7_Int len;
  char fill = ' ';
  
  if (!s7_is_integer(car(args)))
    return(s7_wrong_type_arg_error(sc, "make-string length,", (is_null(cdr(args))) ? 0 : 1, car(args), "an integer"));
  
  len = s7_integer(car(args));
  if (len < 0)
    return(s7_out_of_range_error(sc, "make-string length,", (is_null(cdr(args))) ? 0 : 1, car(args), "a non-negative integer"));
  if (len > MAX_STRING_LENGTH)
    return(s7_out_of_range_error(sc, "make-string length,", (is_null(cdr(args))) ? 0 : 1, car(args), "a reasonable integer!"));

  if (is_not_null(cdr(args))) 
    {
      if (!s7_is_character(cadr(args)))
	return(s7_wrong_type_arg_error(sc, "make-string filler,", 2, cadr(args), "a character"));
      fill = s7_character(cadr(args));
    }
  return(make_empty_string(sc, (int)len, fill));
}


static s7_pointer g_string_length(s7_scheme *sc, s7_pointer args)
{
  #define H_string_length "(string-length str) returns the length of the string str"
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "string-length", 0, car(args), "string"));
  return(make_integer(sc, string_length(car(args))));
}


unsigned int s7_string_length(s7_pointer str)
{
  return(string_length(str));
}


static s7_pointer string_ref_1(s7_scheme *sc, s7_pointer strng, s7_pointer index)
{
  char *str;
  s7_Int ind;

  if (!s7_is_integer(index))
    return(s7_wrong_type_arg_error(sc, "string-ref", 2, index, "an integer"));

  ind = s7_integer(index);

  if (ind < 0)
    return(s7_wrong_type_arg_error(sc, "string-ref index,", 2, index, "a non-negative integer"));
  if (ind >= string_length(strng))
    return(s7_out_of_range_error(sc, "string-ref index,", 2, index, "should be less than string length"));

  str = string_value(strng);
  return(s7_make_character(sc, ((unsigned char *)str)[ind]));
}


static s7_pointer g_string_ref(s7_scheme *sc, s7_pointer args)
{
  #define H_string_ref "(string-ref str index) returns the character at the index-th element of the string str"
  
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "string-ref", 1, car(args), "a string"));

  return(string_ref_1(sc, car(args), cadr(args)));
}


static s7_pointer g_string_set(s7_scheme *sc, s7_pointer args)
{
  #define H_string_set "(string-set! str index chr) sets the index-th element of the string str to the character chr"
  
  s7_pointer x, index;
  char *str;
  s7_Int ind;

  x = car(args);
  index = cadr(args);
  
  if (!s7_is_string(x))
    return(s7_wrong_type_arg_error(sc, "string-set!", 1, x, "a string"));
  if (!s7_is_character(caddr(args)))
    return(s7_wrong_type_arg_error(sc, "string-set!", 3, caddr(args), "a character"));
  if (!s7_is_integer(index))
    return(s7_wrong_type_arg_error(sc, "string-set! index,", 2, index, "an integer"));
  
  ind = s7_integer(index);

  if (ind < 0)
    return(s7_wrong_type_arg_error(sc, "string-set! index,", 2, index, "a non-negative integer"));
  if (ind >= string_length(x))
    return(s7_out_of_range_error(sc, "string-set! index,", 2, index, "should be less than string length"));

  str = string_value(x);
  str[ind] = (char)s7_character(caddr(args));
  return(caddr(args));
}


static s7_pointer g_string_append_1(s7_scheme *sc, s7_pointer args, const char *name)
{
  int len = 0;
  s7_pointer x, newstr;
  char *pos;
  
  if (is_null(args))
    return(s7_make_string_with_length(sc, "", 0));
  
  /* get length for new string */
  for (x = args; is_not_null(x); x = cdr(x)) 
    {
      if (!s7_is_string(car(x)))
	return(s7_wrong_type_arg_error(sc, name, position_of(x, args), car(x), "a string"));
      len += string_length(car(x));
    }
  
  /* store the contents of the argument strings into the new string */
  newstr = make_empty_string(sc, len + 1, 0); 
  /* +1 here because valgrind (but only in FC15) thinks we occasionally go one past the end,
   *   but I think it'c confused by low level copy operations that do 8 bytes at a time.
   */
  string_length(newstr) = len;
  for (pos = string_value(newstr), x = args; is_not_null(x); pos += string_length(car(x)), x = cdr(x)) 
    memcpy(pos, string_value(car(x)), string_length(car(x)));
  
  return(newstr);
}


static s7_pointer g_string_append(s7_scheme *sc, s7_pointer args)
{
  #define H_string_append "(string-append str1 ...) appends all its string arguments into one string"
  return(g_string_append_1(sc, args, "string-append"));
}


static s7_pointer g_string_copy(s7_scheme *sc, s7_pointer args)
{
  #define H_string_copy "(string-copy str) returns a copy of its string argument"
  if (is_null(args))
    return(s7_wrong_type_arg_error(sc, "string-copy", 0, car(args), "a string"));
  
  return(g_string_append_1(sc, args, "string-copy"));
}


static s7_pointer g_substring(s7_scheme *sc, s7_pointer args)
{
  #define H_substring "(substring str start (end (length str))) returns the portion of the string str between start and \
end: (substring \"01234\" 1 2) -> \"1\""
  
  s7_pointer x, start, end, str;
  s7_Int i0, i1;
  int len;
  char *s;
  
  str = car(args);
  start = cadr(args);
  
  if (!s7_is_string(str))
    return(s7_wrong_type_arg_error(sc, "substring", 1, str, "a string"));
  
  if (!s7_is_integer(start))
    return(s7_wrong_type_arg_error(sc, "substring start point,", 2, start, "an integer"));
  i0 = s7_integer(start);
  if (i0 < 0)
    return(s7_wrong_type_arg_error(sc, "substring start point,", 2, start, "a non-negative integer"));
  if (i0 > string_length(str))            /* (substring "012" 10) */
    return(s7_out_of_range_error(sc, "substring start point,", 2, start, "start <= string length"));
  /* this is how guile handles it: (substring "012" 3) -> "" */

  if (is_not_null(cddr(args)))
    {
      end = caddr(args);
      if (!s7_is_integer(end))
	return(s7_wrong_type_arg_error(sc, "substring end point,", 3, end, "an integer"));
      i1 = s7_integer(end);
      if (i1 < i0)
	return(s7_wrong_type_arg_error(sc, "substring end point,", 3, end, "an integer >= start"));
      if (i1 > string_length(str))
	return(s7_out_of_range_error(sc, "substring end point,", 3, end, "end <= string length"));
    }
  else i1 = string_length(str);
  
  s = string_value(str);
  len = i1 - i0;
  x = make_empty_string(sc, len, 0);
  memcpy(string_value(x), s + i0, len);
  string_value(x)[len] = 0;
  return(x);
}


/* (set! (substring...) ...)? -- might require allocation
 *   to implement substring_uncopied we'd need to make sure the lack of a trailing null can't confuse anyone
 *   this is also the case for read-line.
 */


#define USE_WRITE true
#define USE_DISPLAY false

static s7_pointer g_object_to_string(s7_scheme *sc, s7_pointer args)
{
  #define H_object_to_string "(object->string obj (write true)) returns a string representation of obj."
  
  if (is_not_null(cdr(args)))
    {
      if (s7_is_boolean(cadr(args)))
	return(s7_object_to_string(sc, car(args), s7_boolean(sc, cadr(args))));
      return(s7_wrong_type_arg_error(sc, "object->string", 2, cadr(args), "a boolean"));
    }
  return(s7_object_to_string(sc, car(args), USE_WRITE));
}


static int scheme_strcmp(s7_pointer s1, s7_pointer s2)
{
  /* tricky here because str[i] must be treated as unsigned
   *   (string<? (string (integer->char #xf0)) (string (integer->char #x70)))
   * also null or lack thereof does not say anything about the string end
   *   so we have to go by its length.
   */
  int i, len, len1, len2;
  char *str1, *str2;

  len1 = string_length(s1);
  len2 = string_length(s2);
  if (len1 > len2)
    len = len2;
  else len = len1;

  str1 = string_value(s1);
  str2 = string_value(s2);

  for (i = 0; i < len; i++)
    if ((unsigned char)(str1[i]) < (unsigned char )(str2[i]))
      return(-1);
    else
      {
	if ((unsigned char)(str1[i]) > (unsigned char)(str2[i]))
	  return(1);
      }

  if (len1 < len2) 
    return(-1);
  if (len1 > len2)
    return(1);
  return(0);
}


static s7_pointer g_string_cmp(s7_scheme *sc, s7_pointer args, int val, const char *name)
{
  s7_pointer x, y;
  
  for (x = args; is_not_null(x); x = cdr(x))  
    if (!s7_is_string(car(x)))
      return(s7_wrong_type_arg_error(sc, name, position_of(x, args), car(x), "a string"));
  
  y = car(args);
  for (x = cdr(args); is_not_null(x); x = cdr(x))
    {
      if (scheme_strcmp(y, car(x)) != val)
	return(sc->F);
      y = car(x);
    }
  return(sc->T);
}


static s7_pointer g_string_cmp_not(s7_scheme *sc, s7_pointer args, int val, const char *name)
{
  s7_pointer x, y;
  
  for (x = args; is_not_null(x); x = cdr(x))  
    if (!s7_is_string(car(x)))
      return(s7_wrong_type_arg_error(sc, name, position_of(x, args), car(x), "a string"));
  
  y = car(args);
  for (x = cdr(args); is_not_null(x); x = cdr(x))
    {
      if (scheme_strcmp(y, car(x)) == val)
	return(sc->F);
      y = car(x);
    }
  return(sc->T);
}


static bool scheme_strings_are_equal(s7_pointer x, s7_pointer y)
{
  char *xstr, *ystr;
  unsigned int i, len;

  len = string_length(x);
  if (string_length(y) != len) 
    return(false);

  xstr = string_value(x);
  ystr = string_value(y);

  for (i = 0; i < len; i++)
    if (xstr[i] != ystr[i])
      return(false);
  return(true);
}


static s7_pointer g_strings_are_equal(s7_scheme *sc, s7_pointer args)
{
  #define H_strings_are_equal "(string=? str ...) returns #t if all the string arguments are equal"

  /* C-based check stops at null, but we can have embedded nulls.
   *   (let ((s1 "1234") (s2 "1245")) (string-set! s1 1 #\null) (string-set! s2 1 #\null) (string=? s1 s2))
   */
  
  s7_pointer x, y;
  unsigned int len;
  bool happy = true;
  char *ystr;
  
  y = car(args);
  if (!s7_is_string(y))
    return(s7_wrong_type_arg_error(sc, "string=?", 1, y, "a string"));

  len = string_length(y);
  ystr = string_value(y);

  for (x = cdr(args); is_not_null(x); x = cdr(x)) 
    {
      if (y != car(x))
	{
	  if (!s7_is_string(car(x)))
	    return(s7_wrong_type_arg_error(sc, "string=?", position_of(x, args), car(x), "a string"));
	  if (happy)
	    {
	      if (string_length(car(x)) != len)
		happy = false;
	      else
		{
		  unsigned int k;
		  char *xstr;
		  xstr = string_value(car(x));
		  for (k = 0; k < len; k++)
		    if (ystr[k] != xstr[k])
		      {
			happy = false;
			break;
		      }
		}
	    }
	}
    }

  if (!happy)
    return(sc->F);
  return(sc->T);
}	


static s7_pointer g_strings_are_less(s7_scheme *sc, s7_pointer args)
{
  #define H_strings_are_less "(string<? str ...) returns #t if all the string arguments are increasing"
  return(g_string_cmp(sc, args, -1, "string<?"));
}	


static s7_pointer g_strings_are_greater(s7_scheme *sc, s7_pointer args)
{
  #define H_strings_are_greater "(string>? str ...) returns #t if all the string arguments are decreasing"
  return(g_string_cmp(sc, args, 1, "string>?"));
}	


static s7_pointer g_strings_are_geq(s7_scheme *sc, s7_pointer args)
{
  #define H_strings_are_geq "(string>=? str ...) returns #t if all the string arguments are equal or decreasing"
  return(g_string_cmp_not(sc, args, -1, "string>=?"));
}	


static s7_pointer g_strings_are_leq(s7_scheme *sc, s7_pointer args)
{
  #define H_strings_are_leq "(string<=? str ...) returns #t if all the string arguments are equal or increasing"
  return(g_string_cmp_not(sc, args, 1, "string<=?"));
}	

#if WITH_OPTIMIZATION

static s7_pointer string_equal_s_ic, string_equal_2, string_equal_vref_s;
static s7_pointer g_string_equal_s_ic(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "string=?", 1, car(args), "a string"));
  return(make_boolean(sc, scheme_strings_are_equal(car(args), cadr(args))));
}

static s7_pointer g_string_equal_vref_s(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p, vect, ind, vstr, target;
  s7_Int index;

  p = cdar(args);
  vect = finder(sc, car(p));
  if (!s7_is_vector(vect))
    return(s7_wrong_type_arg_error(sc, "vector-ref", 1, vect, "a vector"));

  ind = finder(sc, cadr(p));
  if (!s7_is_integer(ind))
    return(s7_wrong_type_arg_error(sc, "vector-ref", 2, ind, "an integer"));
  index = s7_integer(ind);
  if ((index < 0) ||
      (index >= vector_length(vect)))
    return(s7_out_of_range_error(sc, "vector-ref index,", 2, ind, "should non-negative and less than the vector length"));
  
  vstr = vector_element(vect, index);
  if (!s7_is_string(vstr))
    return(s7_wrong_type_arg_error(sc, "string=?", 1, vstr, "a string"));

  target = finder(sc, cadr(args));
  if (!s7_is_string(target))
    return(s7_wrong_type_arg_error(sc, "string=?", 2, target, "a string"));
  return(make_boolean(sc, scheme_strings_are_equal(vstr, target)));
}

static s7_pointer g_string_equal_2(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "string=?", 1, car(args), "a string"));
  if (!s7_is_string(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "string=?", 2, cadr(args), "a string"));
  return(make_boolean(sc, scheme_strings_are_equal(car(args), cadr(args))));
}


static s7_pointer string_less_s_ic, string_less_2;
static s7_pointer g_string_less_s_ic(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "string<?", 1, car(args), "a string"));
  return(make_boolean(sc, scheme_strcmp(car(args), cadr(args)) == -1));
}

static s7_pointer g_string_less_2(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "string<?", 1, car(args), "a string"));
  if (!s7_is_string(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "string<?", 2, cadr(args), "a string"));
  return(make_boolean(sc, scheme_strcmp(car(args), cadr(args)) == -1));
}


static s7_pointer string_greater_s_ic, string_greater_2;
static s7_pointer g_string_greater_s_ic(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "string>?", 1, car(args), "a string"));
  return(make_boolean(sc, scheme_strcmp(car(args), cadr(args)) == 1));
}

static s7_pointer g_string_greater_2(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "string>?", 1, car(args), "a string"));
  if (!s7_is_string(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "string>?", 2, cadr(args), "a string"));
  return(make_boolean(sc, scheme_strcmp(car(args), cadr(args)) == 1));
}


static s7_pointer string_geq_s_ic, string_geq_2;
static s7_pointer g_string_geq_s_ic(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "string>=?", 1, car(args), "a string"));
  return(make_boolean(sc, scheme_strcmp(car(args), cadr(args)) != -1));
}

static s7_pointer g_string_geq_2(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "string>=?", 1, car(args), "a string"));
  if (!s7_is_string(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "string>=?", 2, cadr(args), "a string"));
  return(make_boolean(sc, scheme_strcmp(car(args), cadr(args)) != -1));
}


static s7_pointer string_leq_s_ic, string_leq_2;
static s7_pointer g_string_leq_s_ic(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "string<=?", 1, car(args), "a string"));
  return(make_boolean(sc, scheme_strcmp(car(args), cadr(args)) != 1));
}

static s7_pointer g_string_leq_2(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "string<=?", 1, car(args), "a string"));
  if (!s7_is_string(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "string<=?", 2, cadr(args), "a string"));
  return(make_boolean(sc, scheme_strcmp(car(args), cadr(args)) != 1));
}
#endif


static int scheme_strcasecmp(s7_pointer s1, s7_pointer s2)
{
  /* same as scheme_strcmp -- watch out for unwanted sign! and lack of trailing null (length sets string end).
   */
  int i, len, len1, len2;
  char *str1, *str2;

  len1 = string_length(s1);
  len2 = string_length(s2);
  if (len1 > len2)
    len = len2;
  else len = len1;

  str1 = string_value(s1);
  str2 = string_value(s2);

  for (i = 0; i < len; i++)
    if (uppers[(int)str1[i]] < uppers[(int)str2[i]])
      return(-1);
    else
      {
	if (uppers[(int)str1[i]] > uppers[(int)str2[i]])
	  return(1);
      }

  if (len1 < len2) 
    return(-1);
  if (len1 > len2)
    return(1);
  return(0);
}


static bool scheme_strequal_ci(s7_pointer s1, s7_pointer s2)
{
  /* same as scheme_strcmp -- watch out for unwanted sign! */
  int i, len, len2;
  char *str1, *str2;

  len = string_length(s1);
  len2 = string_length(s2);
  if (len != len2)
    return(false);

  str1 = string_value(s1);
  str2 = string_value(s2);

  for (i = 0; i < len; i++)
    if (uppers[(int)str1[i]] != uppers[(int)str2[i]])
      return(false);
  return(true);
}


static s7_pointer g_string_ci_cmp(s7_scheme *sc, s7_pointer args, int val, const char *name)
{
  s7_pointer x, y;
  
  for (x = args; is_not_null(x); x = cdr(x))  
    if (!s7_is_string(car(x)))
      return(s7_wrong_type_arg_error(sc, name, position_of(x, args), car(x), "a string"));
  
  y = car(args);
  if (val == 0)
    {
      for (x = cdr(args); is_not_null(x); x = cdr(x))
	if (!scheme_strequal_ci(y, car(x)))
	  return(sc->F);
    }
  else
    {
      for (x = cdr(args); is_not_null(x); x = cdr(x))
	{
	  if (scheme_strcasecmp(y, car(x)) != val)
	    return(sc->F);
	  y = car(x);
	}
    }
  return(sc->T);
}


static s7_pointer g_string_ci_cmp_not(s7_scheme *sc, s7_pointer args, int val, const char *name)
{
  s7_pointer x, y;
  
  for (x = args; is_not_null(x); x = cdr(x))  
    if (!s7_is_string(car(x)))
      return(s7_wrong_type_arg_error(sc, name, position_of(x, args), car(x), "a string"));
  
  y = car(args);
  for (x = cdr(args); is_not_null(x); x = cdr(x))
    {
      if (scheme_strcasecmp(y, car(x)) == val)
	return(sc->F);
      y = car(x);
    }
  return(sc->T);
}


static s7_pointer g_strings_are_ci_equal(s7_scheme *sc, s7_pointer args)
{
  #define H_strings_are_ci_equal "(string-ci=? str ...) returns #t if all the string arguments are equal, ignoring case"
  return(g_string_ci_cmp(sc, args, 0, "string-ci=?"));
}	


static s7_pointer g_strings_are_ci_less(s7_scheme *sc, s7_pointer args)
{
  #define H_strings_are_ci_less "(string-ci<? str ...) returns #t if all the string arguments are increasing, ignoring case"
  return(g_string_ci_cmp(sc, args, -1, "string-ci<?"));
}	


static s7_pointer g_strings_are_ci_greater(s7_scheme *sc, s7_pointer args)
{
  #define H_strings_are_ci_greater "(string-ci>? str ...) returns #t if all the string arguments are decreasing, ignoring case"
  return(g_string_ci_cmp(sc, args, 1, "string-ci>?"));
}	


static s7_pointer g_strings_are_ci_geq(s7_scheme *sc, s7_pointer args)
{
  #define H_strings_are_ci_geq "(string-ci>=? str ...) returns #t if all the string arguments are equal or decreasing, ignoring case"
  return(g_string_ci_cmp_not(sc, args, -1, "string-ci>=?"));
}	


static s7_pointer g_strings_are_ci_leq(s7_scheme *sc, s7_pointer args)
{
  #define H_strings_are_ci_leq "(string-ci<=? str ...) returns #t if all the string arguments are equal or increasing, ignoring case"
  return(g_string_ci_cmp_not(sc, args, 1, "string-ci<=?"));
}	


#if WITH_OPTIMIZATION

static s7_pointer string_ci_equal_s_ic, string_ci_equal_2;
static s7_pointer g_string_ci_equal_s_ic(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "string-ci=?", 1, car(args), "a string"));
  return(make_boolean(sc, scheme_strequal_ci(car(args), cadr(args))));
}

static s7_pointer g_string_ci_equal_2(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "string-ci=?", 1, car(args), "a string"));
  if (!s7_is_string(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "string-ci=?", 2, cadr(args), "a string"));
  return(make_boolean(sc, scheme_strequal_ci(car(args), cadr(args))));
}


static s7_pointer string_ci_less_s_ic, string_ci_less_2;
static s7_pointer g_string_ci_less_s_ic(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "string-ci<?", 1, car(args), "a string"));
  return(make_boolean(sc, scheme_strcasecmp(car(args), cadr(args)) == -1));
}

static s7_pointer g_string_ci_less_2(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "string-ci<?", 1, car(args), "a string"));
  if (!s7_is_string(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "string-ci<?", 2, cadr(args), "a string"));
  return(make_boolean(sc, scheme_strcasecmp(car(args), cadr(args)) == -1));
}


static s7_pointer string_ci_greater_s_ic, string_ci_greater_2;
static s7_pointer g_string_ci_greater_s_ic(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "string-ci>?", 1, car(args), "a string"));
  return(make_boolean(sc, scheme_strcasecmp(car(args), cadr(args)) == 1));
}

static s7_pointer g_string_ci_greater_2(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "string-ci>?", 1, car(args), "a string"));
  if (!s7_is_string(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "string-ci>?", 2, cadr(args), "a string"));
  return(make_boolean(sc, scheme_strcasecmp(car(args), cadr(args)) == 1));
}


static s7_pointer string_ci_geq_s_ic, string_ci_geq_2;
static s7_pointer g_string_ci_geq_s_ic(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "string-ci>=?", 1, car(args), "a string"));
  return(make_boolean(sc, scheme_strcasecmp(car(args), cadr(args)) != -1));
}

static s7_pointer g_string_ci_geq_2(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "string-ci>=?", 1, car(args), "a string"));
  if (!s7_is_string(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "string-ci>=?", 2, cadr(args), "a string"));
  return(make_boolean(sc, scheme_strcasecmp(car(args), cadr(args)) != -1));
}


static s7_pointer string_ci_leq_s_ic, string_ci_leq_2;
static s7_pointer g_string_ci_leq_s_ic(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "string-ci<=?", 1, car(args), "a string"));
  return(make_boolean(sc, scheme_strcasecmp(car(args), cadr(args)) != 1));
}

static s7_pointer g_string_ci_leq_2(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "string-ci<=?", 1, car(args), "a string"));
  if (!s7_is_string(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "string-ci<=?", 2, cadr(args), "a string"));
  return(make_boolean(sc, scheme_strcasecmp(car(args), cadr(args)) != 1));
}
#endif



static s7_pointer g_string_fill(s7_scheme *sc, s7_pointer args)
{
  #define H_string_fill "(string-fill! str chr) fills the string str with the character chr"
  s7_pointer x;
  x = car(args);

  if (!s7_is_string(x))
    return(s7_wrong_type_arg_error(sc, "string-fill!", 1, x, "a string"));
  if (!s7_is_character(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "string-fill! filler,", 2, cadr(args), "a character"));

  /* strlen and so on here is probably not right -- a scheme string has a length
   *   set when it is created, and (apparently) can contain an embedded 0, so its
   *   print length is not its length.
   *         char *str; char c; str = string_value(car(args)); c = character(cadr(args));
   *         int i, len = 0; if (str) len = safe_strlen(str); if (len > 0) for (i = 0; i < len; i++) str[i] = c; 
   */
  if (string_length(x) > 0)
    memset((void *)(string_value(x)), (int)character(cadr(args)), string_length(x));
  return(cadr(args)); 
}


static s7_pointer g_string_1(s7_scheme *sc, s7_pointer args, const char *name)
{
  int i, len;
  s7_pointer x, newstr;
  
  /* get length for new string and check arg types */
  for (len = 0, x = args; is_not_null(x); len++, x = cdr(x)) 
    if (!s7_is_character(car(x)))
      return(s7_wrong_type_arg_error(sc, name, len + 1, car(x), "a character"));
  
  newstr = make_empty_string(sc, len, 0);
  for (i = 0, x = args; is_not_null(x); i++, x = cdr(x)) 
    string_value(newstr)[i] = character(car(x));
  
  return(newstr);
}


static s7_pointer g_string(s7_scheme *sc, s7_pointer args)
{
  #define H_string "(string chr...) appends all its character arguments into one string"
  if (is_null(args))                                /* (string) but not (string '()) */
    return(s7_make_string_with_length(sc, "", 0));
  return(g_string_1(sc, args, "string"));
}


static s7_pointer g_list_to_string(s7_scheme *sc, s7_pointer args)
{
  #define H_list_to_string "(list->string lst) appends all the list's characters into one string; (apply string lst)"
  if (is_null(car(args)))
    return(s7_make_string_with_length(sc, "", 0));
  
  if (!is_proper_list(sc, car(args)))
    return(s7_wrong_type_arg_error(sc, "list->string", 0, car(args), "a (proper, non-circular) list of characters"));
  
  return(g_string_1(sc, car(args), "list->string"));
}


static s7_pointer s7_string_to_list(s7_scheme *sc, const char *str, int len)
{
  int i;
  s7_pointer p;

  if (len == 0)
    return(sc->NIL);

  if (len < (sc->free_heap_top - sc->free_heap))
    {
      sc->w = s7_cons (sc, s7_make_character(sc, ((unsigned char)str[0])), sc->NIL);
      for (i = 1; i < len; i++)
	sc->w = cons_unchecked(sc, s7_make_character(sc, ((unsigned char)str[i])), sc->w);
    }
  else
    {
      sc->w = sc->NIL;
      for (i = 0; i < len; i++)
	sc->w = cons(sc, s7_make_character(sc, ((unsigned char)str[i])), sc->w);
    }
  p = sc->w;
  sc->w = sc->NIL;

  return(safe_reverse_in_place(sc, p));
}


static s7_pointer g_string_to_list(s7_scheme *sc, s7_pointer args)
{
  #define H_string_to_list "(string->list str) returns the elements of the string str in a list; (map values str)"
  s7_pointer str;

  str = car(args);
  if (!s7_is_string(str))
    return(s7_wrong_type_arg_error(sc, "string->list", 0, str, "a string"));

  return(s7_string_to_list(sc, string_value(str), string_length(str)));
}




/* -------------------------------- ports -------------------------------- 
 *
 * originally nil served as stdin and friends, but that made it impossible to catch an error
 *   like (read-line (current-output-port)) when the latter was stdout.  So we now have
 *   the built-in constant ports *stdin*, *stdout*, and *stderr*.  Some way is needed to
 *   refer to these directly so that (read-line *stdin*) for example can insist on reading
 *   from the terminal, or whatever stdin is.
 */

static char *describe_port(s7_scheme *sc, s7_pointer p)
{
  char *desc;
  if ((p == sc->standard_input) || (p == sc->standard_output) || (p == sc->standard_error))
    return(copy_string(port_filename(p)));

  desc = (char *)malloc(64 * sizeof(char));
  snprintf(desc, 64, "<port%s%s%s>",
  	   (is_file_port(p)) ? " file" : ((is_string_port(p)) ? " string" : " function"),
	   (is_input_port(p)) ? " input" : " output",
	   (port_is_closed(p)) ? " (closed)" : "");
  return(desc);
}


static s7_pointer g_is_port_closed(s7_scheme *sc, s7_pointer args)
{
  #define H_is_port_closed "(port-closed? p) returns #t if the port p is closed."
  s7_pointer x;

  x = car(args);
  if ((is_input_port(x)) || (is_output_port(x)))
    return(make_boolean(sc, port_is_closed(x)));

  return(s7_wrong_type_arg_error(sc, "port-closed?", 0, x, "a port"));      
}


static s7_pointer g_port_line_number(s7_scheme *sc, s7_pointer args)
{
  #define H_port_line_number "(port-line-number input-file-port) returns the current read line number of port"
  s7_pointer x;

  if (is_null(args))
    x = sc->input_port;
  else x = car(args);

  if ((!(is_input_port(x))) ||
      (port_is_closed(x)))
    return(s7_wrong_type_arg_error(sc, "port-line-number", 0, x, "an open input port"));

  return(make_integer(sc, port_line_number(x)));
}


const char *s7_port_filename(s7_pointer x)
{
  if (((is_input_port(x)) || 
       (is_output_port(x))) &&
      (!port_is_closed(x)))
    return(port_filename(x));
  return(NULL);
}


static s7_pointer g_port_filename(s7_scheme *sc, s7_pointer args)
{
  #define H_port_filename "(port-filename file-port) returns the filename associated with port"
  s7_pointer x;

  if (is_null(args))
    x = sc->input_port;
  else x = car(args);

  if (((is_input_port(x)) ||
       (is_output_port(x))) &&
      (!port_is_closed(x)))
    {
      if (port_filename(x))
	return(make_protected_string(sc, port_filename(x)));
      return(s7_make_string_with_length(sc, "", 0));   
      /* otherwise (eval-string (port-filename)) and (string->symbol (port-filename)) segfault */
    }

  return(s7_wrong_type_arg_error(sc, "port-filename", 0, x, "an open port"));
}


bool s7_is_input_port(s7_scheme *sc, s7_pointer p)   
{ 
  return(is_input_port(p));
}


static s7_pointer g_is_input_port(s7_scheme *sc, s7_pointer args)
{
  #define H_is_input_port "(input-port? p) returns #t if p is an input port"
  return(make_boolean(sc, is_input_port(car(args))));
}


bool s7_is_output_port(s7_scheme *sc, s7_pointer p)     
{ 
  return(is_output_port(p));
}


static s7_pointer g_is_output_port(s7_scheme *sc, s7_pointer args)
{
  #define H_is_output_port "(output-port? p) returns #t if p is an output port"
  return(make_boolean(sc, is_output_port(car(args))));
}


s7_pointer s7_current_input_port(s7_scheme *sc)
{
  return(sc->input_port);
}


static s7_pointer g_current_input_port(s7_scheme *sc, s7_pointer args)
{
  #define H_current_input_port "(current-input-port) returns the current input port"
  return(sc->input_port);
}


static s7_pointer g_set_current_input_port(s7_scheme *sc, s7_pointer args)
{
  #define H_set_current_input_port "(set-current-input-port port) sets the current-input port to port and returns the previous value of the input port"
  s7_pointer old_port, port;

  old_port = sc->input_port;
  port = car(args);
  if ((is_input_port(port)) &&
      (!port_is_closed(port)))
    sc->input_port = port;
  else return(s7_wrong_type_arg_error(sc, "set-current-input-port", 0, port, "an open input port"));

  return(old_port);
}


s7_pointer s7_set_current_input_port(s7_scheme *sc, s7_pointer port)
{
  s7_pointer old_port;
  old_port = sc->input_port;
  sc->input_port = port;
  return(old_port);
}


s7_pointer s7_current_output_port(s7_scheme *sc)
{
  return(sc->output_port);
}


s7_pointer s7_set_current_output_port(s7_scheme *sc, s7_pointer port)
{
  s7_pointer old_port;
  old_port = sc->output_port;
  sc->output_port = port;
  return(old_port);
}


static s7_pointer g_current_output_port(s7_scheme *sc, s7_pointer args)
{
  #define H_current_output_port "(current-output-port) returns the current output port"
  return(sc->output_port);
}


static s7_pointer g_set_current_output_port(s7_scheme *sc, s7_pointer args)
{
  #define H_set_current_output_port "(set-current-output-port port) sets the current-output port to port and returns the previous value of the output port"
  s7_pointer old_port, port;

  old_port = sc->output_port;
  port = car(args);
  if ((is_output_port(port)) &&
      (!port_is_closed(port)))
    sc->output_port = port;
  else return(s7_wrong_type_arg_error(sc, "set-current-output-port", 0, port, "an open output port"));

  return(old_port);
}


s7_pointer s7_current_error_port(s7_scheme *sc)
{
  return(sc->error_port);
}


s7_pointer s7_set_current_error_port(s7_scheme *sc, s7_pointer port)
{
  s7_pointer old_port;
  old_port = sc->error_port;
  sc->error_port = port;
  return(old_port);
}


static s7_pointer g_current_error_port(s7_scheme *sc, s7_pointer args)
{
  #define H_current_error_port "(current-error-port) returns the current error port"
  return(sc->error_port);
}


static s7_pointer g_set_current_error_port(s7_scheme *sc, s7_pointer args)
{
  #define H_set_current_error_port "(set-current-error-port port) sets the current-error port to port and returns the previous value of the error port"
  s7_pointer old_port, port;

  old_port = sc->error_port;
  port = car(args);
  if ((is_output_port(port)) &&
      (!port_is_closed(port)))
    sc->error_port = port;
  else return(s7_wrong_type_arg_error(sc, "set-current-error-port", 0, port, "an open output port"));

  return(old_port);
}


static s7_pointer g_is_char_ready(s7_scheme *sc, s7_pointer args)
{
  #define H_is_char_ready "(char-ready? (port (current-input-port))) returns #t if a character is ready for input on the given port"
  if (is_not_null(args))
    {
      s7_pointer pt = car(args);
      if (!is_input_port(pt))
	return(s7_wrong_type_arg_error(sc, "char-ready?", 0, pt, "an input port"));
      if (port_is_closed(pt))
	return(s7_wrong_type_arg_error(sc, "char-ready?", 0, pt, "an open input port"));

      if (is_function_port(pt))
	return((*(port_input_function(pt)))(sc, S7_IS_CHAR_READY, pt));
      return(make_boolean(sc, is_string_port(pt)));
    }
  return(make_boolean(sc, (is_input_port(sc->input_port)) && (is_string_port(sc->input_port))));
}      


static s7_pointer g_is_eof_object(s7_scheme *sc, s7_pointer args)
{
  #define H_is_eof_object "(eof-object? val) returns #t if val is the end-of-file object"
  return(make_boolean(sc, car(args) == sc->EOF_OBJECT));
}


static int closed_port_read_char(s7_scheme *sc, s7_pointer port);
static s7_pointer closed_port_read_line(s7_scheme *sc, s7_pointer port, bool with_eol);
static void closed_port_write_char(s7_scheme *sc, int c, s7_pointer port);
static void closed_port_display(s7_scheme *sc, const char *s, s7_pointer port);

void s7_close_input_port(s7_scheme *sc, s7_pointer p)
{
  if ((is_immutable(p)) ||
      ((is_input_port(p)) && (port_is_closed(p))))
    return;

  if (port_filename(p))
    {
      free(port_filename(p));
      port_filename(p) = NULL;
    }
  
  if ((is_file_port(p)) &&
      (port_file(p)))
    {
      fclose(port_file(p));
      port_file(p) = NULL;
    }

  if (port_needs_free(p))
    {
      if (port_string(p))
	{
	  free(port_string(p));
	  port_string(p) = NULL;
	}
      port_needs_free(p) = false;
    }

  /* if input string, someone else is dealing with GC */
  port_read_character(p) = closed_port_read_char;
  port_read_line(p) = closed_port_read_line;
  port_write_character(p) = closed_port_write_char;
  port_display(p) = closed_port_display;
  port_is_closed(p) = true;
}


static s7_pointer g_close_input_port(s7_scheme *sc, s7_pointer args)
{
  #define H_close_input_port "(close-input-port port) closes the port"
  s7_pointer pt;

  pt = car(args);
  if (!is_input_port(pt))
    return(s7_wrong_type_arg_error(sc, "close-input-port", 0, pt, "an input port"));

  if (!is_immutable(pt))
    s7_close_input_port(sc, pt);

  return(sc->UNSPECIFIED);
}


void s7_close_output_port(s7_scheme *sc, s7_pointer p)
{
  if ((is_immutable(p)) ||
      ((is_output_port(p)) && (port_is_closed(p))))
    return;
  
  if (port_filename(p))
    {
      free(port_filename(p));
      port_filename(p) = NULL;
    }
  
  if (is_file_port(p))
    {
      if (port_file(p))
	{
	  if (port_position(p) > 0)
	    fwrite((void *)(port_data(p)), 1, port_position(p), port_file(p));
	  port_position(p) = 0;
	  free(port_data(p));
	  fflush(port_file(p));
	  fclose(port_file(p));
	  port_file(p) = NULL;
	}
    }
  else
    {
      if ((is_string_port(p)) && (port_string(p)))
	{
	  free(port_string(p));
	  port_string(p) = NULL;
	  port_needs_free(p) = false;
	}
    }
  port_read_character(p) = closed_port_read_char;
  port_read_line(p) = closed_port_read_line;
  port_write_character(p) = closed_port_write_char;
  port_display(p) = closed_port_display;
  port_is_closed(p) = true;
}


static s7_pointer g_close_output_port(s7_scheme *sc, s7_pointer args)
{
  #define H_close_output_port "(close-output-port port) closes the port"
  s7_pointer pt;
  pt = car(args);
  if (!is_output_port(pt))
    return(s7_wrong_type_arg_error(sc, "close-output-port", 0, pt, "an output port"));
  if (!(is_immutable(pt)))
    s7_close_output_port(sc, pt);
  return(sc->UNSPECIFIED);
}


/* -------- read character functions -------- */

static int file_read_char(s7_scheme *sc, s7_pointer port)
{
  return(fgetc(port_file(port)));
}


static int function_read_char(s7_scheme *sc, s7_pointer port)
{
  return(character((*(port_input_function(port)))(sc, S7_READ_CHAR, port)));
}


static int string_read_char(s7_scheme *sc, s7_pointer port)
{
  if ((!(port_string(port))) ||
      (port_string_length(port) <= port_string_point(port)))
    return(EOF);
  return((unsigned char)port_string(port)[port_string_point(port)++]);
}


static int output_read_char(s7_scheme *sc, s7_pointer port)
{
  s7_wrong_type_arg_error(sc, "read-char", 0, port, "an input port");
  return(0);
}


static int closed_port_read_char(s7_scheme *sc, s7_pointer port)
{
  s7_wrong_type_arg_error(sc, "read-char", 0, port, "an open port");
  return(0);
}



/* -------- read line functions -------- */

static s7_pointer output_read_line(s7_scheme *sc, s7_pointer port, bool with_eol)
{
  return(s7_wrong_type_arg_error(sc, "read-line", 0, port, "an input port"));
}


static s7_pointer closed_port_read_line(s7_scheme *sc, s7_pointer port, bool with_eol)
{
  return(s7_wrong_type_arg_error(sc, "read-line", 0, port, "an open port"));
}


static s7_pointer function_read_line(s7_scheme *sc, s7_pointer port, bool with_eol)
{
  return((*(port_input_function(port)))(sc, S7_READ_LINE, port));
}


static s7_pointer stdin_read_line(s7_scheme *sc, s7_pointer port, bool with_eol)
{
  if (sc->read_line_buf == NULL)
    {
      sc->read_line_buf_size = 256;
      sc->read_line_buf = (char *)malloc(sc->read_line_buf_size * sizeof(char));
    }

  if (fgets(sc->read_line_buf, sc->read_line_buf_size, stdin) != NULL)
    return(s7_make_string(sc, sc->read_line_buf)); /* fgets adds the trailing '\0' */
  return(s7_make_string_with_length(sc, NULL, 0));
}


static s7_pointer file_read_line(s7_scheme *sc, s7_pointer port, bool with_eol)
{
  unsigned int i;
  if (sc->read_line_buf == NULL)
    {
      sc->read_line_buf_size = 256;
      sc->read_line_buf = (char *)malloc(sc->read_line_buf_size * sizeof(char));
    }

  for (i = 0; ; i++)
    {
      int c;
      if (i >= sc->read_line_buf_size)
	{
	  sc->read_line_buf_size *= 2;
	  sc->read_line_buf = (char *)realloc(sc->read_line_buf, sc->read_line_buf_size * sizeof(char));
	}
      c = fgetc(port_file(port)); /* not unsigned char! -- could be EOF */
      if (c == EOF)
	{
	  if (i == 0)
	    return(sc->EOF_OBJECT);
	  return(s7_make_terminated_string_with_length(sc, sc->read_line_buf, i));
	}
      sc->read_line_buf[i] = (char)c;
      if (c == '\n')
	{
	  port_line_number(port)++;
	  if (!with_eol) i--;
	  return(s7_make_terminated_string_with_length(sc, sc->read_line_buf, i + 1));
	}
    }
  return(sc->EOF_OBJECT);
}


static s7_pointer string_read_line(s7_scheme *sc, s7_pointer port, bool with_eol)
{
  unsigned int i, port_len, port_start;
  char *port_str, *cur, *end;

  /* this string can be left uncopied if port_needs_free(port), then the sc->strings
   *   array checked when the port is closed, and any held strings copied.  But 
   *   even in a best case (make-index.scm), the savings is small.
   *
   * perhaps dont free port string until GC sweep, sweep strings ahead of ports, do the holder check just before free
   *   but this fails because FFI code (snd-xen.c) assumes strings end in null.  This will mess up uncopied substring too...
   */

  port_len = port_string_length(port);
  port_start = port_string_point(port);
  port_str = port_string(port);
  end = (char *)(port_str + port_len);

  for (cur = (char *)(port_str + port_start); cur < end; cur++)
    if (*cur == '\n')
      {
	port_line_number(port)++;
	i = cur - port_str;
	port_string_point(port) = i + 1;
	return(s7_make_terminated_string_with_length(sc, (char *)(port_str + port_start), ((with_eol) ? i + 1 : i) - port_start));
      }
  i = cur - port_str;
  port_string_point(port) = port_len;
  if (i == port_start)
    return(sc->EOF_OBJECT);

#if 0
  if (port_needs_free(port))
    {
      /* here and above like this, and don't free port string until GC */
      s7_pointer result;
      result = make_string_uncopied_with_length(sc, (char *)(port_str + port_start), i - port_start);
      string_needs_free(result) = false;
      return(result);
    }
#endif
  
  return(s7_make_terminated_string_with_length(sc, (char *)(port_str + port_start), i - port_start));
}


/* -------- write character functions -------- */

static void string_write_char(s7_scheme *sc, int c, s7_pointer pt)
{
  if (port_string_point(pt) >= port_string_length(pt))
    {
      int loc;
      loc = port_string_length(pt);
      port_string_length(pt) *= 2;
      port_string(pt) = (char *)realloc(port_string(pt), port_string_length(pt) * sizeof(char));
      memset((void *)(port_string(pt) + loc), 0, loc);
    }
  port_string(pt)[port_string_point(pt)++] = c;
}

static void stdout_write_char(s7_scheme *sc, int c, s7_pointer port)
{
  fputc(c, stdout);
}

static void stderr_write_char(s7_scheme *sc, int c, s7_pointer port)
{
  fputc(c, stderr);
}

static void function_write_char(s7_scheme *sc, int c, s7_pointer port)
{
  (*(port_output_function(port)))(sc, c, port);
}


#define PORT_DATA_SIZE 256
static void file_write_char(s7_scheme *sc, int c, s7_pointer port)
{
  if (port_position(port) == PORT_DATA_SIZE)
    {
      fwrite((void *)(port_data(port)), 1, PORT_DATA_SIZE, port_file(port));
      port_position(port) = 0;
    }
  port_data(port)[port_position(port)++] = (unsigned char)c;
}


static void input_write_char(s7_scheme *sc, int c, s7_pointer port)
{
  s7_wrong_type_arg_error(sc, "write-char", 0, port, "an output port");
}


static void closed_port_write_char(s7_scheme *sc, int c, s7_pointer port)
{
  s7_wrong_type_arg_error(sc, "write-char", 0, port, "an open port");
}



/* -------- write string functions -------- */

static void input_display(s7_scheme *sc, const char *s, s7_pointer port)
{
  s7_wrong_type_arg_error(sc, "write", 0, port, "an output port");
}

static void closed_port_display(s7_scheme *sc, const char *s, s7_pointer port)
{
  s7_wrong_type_arg_error(sc, "write", 0, port, "an open port");
}


static void file_display(s7_scheme *sc, const char *s, s7_pointer port)
{
  if (s)
    {
      if (port_position(port) > 0)
	{
	  fwrite((void *)(port_data(port)), 1, port_position(port), port_file(port));
	  port_position(port) = 0;
	}
      if (fputs(s, port_file(port)) == EOF)
	fprintf(stderr, "write to %s: %s\n", port_filename(port), strerror(errno));
    }
}


static void string_display(s7_scheme *sc, const char *s, s7_pointer port)
{
  if (s)
    {
      for (; *s; s++)
	string_write_char(sc, *s, port);
    }
}


static void function_display(s7_scheme *sc, const char *s, s7_pointer port)
{
  if (s)
    {
      for (; *s; s++)
	(*(port_output_function(port)))(sc, *s, port);
    }
}


static void stdout_display(s7_scheme *sc, const char *s, s7_pointer port)
{
  if (s) fputs(s, stdout);
}


static void stderr_display(s7_scheme *sc, const char *s, s7_pointer port)
{
  if (s) fputs(s, stderr);
}



/* -------- skip to semicolon readers -------- */

static token_t file_read_semicolon(s7_scheme *sc, s7_pointer pt)
{
  int c;
  do (c = fgetc(port_file(pt))); while ((c != '\n') && (c != EOF));
  port_line_number(pt)++;
  if (c == EOF)
    return(TOKEN_EOF);
  return(token(sc));
}


static token_t string_read_semicolon(s7_scheme *sc, s7_pointer pt)
{
  unsigned char c;
  char *orig_str, *str;      
  str = (char *)(port_string(pt) + port_string_point(pt));
  orig_str = str;
  do {c = (unsigned char)(*str++);} while (not_eol[c]);
  /* the type casts are not otiose -- funny chars > 127 can become negative integers 
   *   use of an array lookup here is 25% faster than the equivalent 'or'
   *   earlier version: do {c = *str++;} while ((c != '\n') && (c != 0));
   */
  port_string_point(pt) += (str - orig_str);
  port_line_number(pt)++;
  if (c == 0)
    return(TOKEN_EOF);
  return(token(sc));
}


/* -------- white space readers -------- */

static int file_read_white_space(s7_scheme *sc, s7_pointer port)
{
  int c;
  while (is_white_space(c = fgetc(port_file(port))))
    if (c == '\n')
      port_line_number(port)++;
  return(c);
}

static int string_read_white_space(s7_scheme *sc, s7_pointer pt)
{
  char *orig_str, *str;
  unsigned char c1;

  str = (char *)(port_string(pt) + port_string_point(pt));
  if (!(*str)) return(EOF);

  /* we can't depend on the extra 0 of padding at the end of an input string port --
   *   eval_string and others take the given string without copying or padding.
   */
  orig_str = str;
  while (white_space[c1 = (unsigned char)(*str++)]) /* (let ((a 1)) a) -- 255 is not -1 = EOF */
    if (c1 == '\n')
      port_line_number(pt)++;
  
  if (c1 == 0)
    {
      port_string_point(pt) += (str - orig_str - 1);
      return(EOF);
    }
  port_string_point(pt) += (str - orig_str);
  return((int)c1);
}


/* name (alphanumeric token) readers */

static void resize_strbuf(s7_scheme *sc)
{
  unsigned int i, old_size;
  old_size = sc->strbuf_size;
  sc->strbuf_size *= 2;
  sc->strbuf = (char *)realloc(sc->strbuf, sc->strbuf_size * sizeof(char));
  for (i = old_size; i < sc->strbuf_size; i++) sc->strbuf[i] = '\0';
}


static s7_pointer file_read_name(s7_scheme *sc, s7_pointer pt, bool atom_case)
{
  int c;
  unsigned i = 1;
  /* sc->strbuf[0] has the 1st char of the string we're reading */

  do {
    c = fgetc(port_file(pt)); /* might return EOF */
    if (c == '\n')
      port_line_number(pt)++;
    
    sc->strbuf[i++] = c;
    if (i >= sc->strbuf_size)
      resize_strbuf(sc);
  } while ((c != EOF) && (char_ok_in_a_name[c]));
  
  if ((i == 2) && 
      (sc->strbuf[0] == '\\'))
    sc->strbuf[2] = '\0';
  else 
    {
      if (c != EOF)
	{
	  if (c == '\n')
	    port_line_number(pt)--;
	  ungetc(c, port_file(pt));
	}
      sc->strbuf[i - 1] = '\0';
    }

  if (atom_case)
    return(make_atom(sc, sc->strbuf, BASE_10, SYMBOL_OK));

  return(make_sharp_constant(sc, sc->strbuf, UNNESTED_SHARP, BASE_10));
}


#define WITH_SHARP false
#define NO_SHARP true

static s7_pointer string_read_name_no_free(s7_scheme *sc, s7_pointer pt, bool atom_case)
{
  /* sc->strbuf[0] has the 1st char of the string we're reading */

  unsigned int k = 0;
  char *orig_str, *str;
  
  orig_str = (char *)(port_string(pt) + port_string_point(pt) - 1);
  str = (char *)(orig_str + 1);
  
  while (char_ok_in_a_name[(unsigned char)(*str)]) {str++;}
  k = str - orig_str;
  port_string_point(pt) += k;
  
  if ((!atom_case) &&             /* there's a bizarre special case here \ with the next char #\null: (eval-string "(list \\\x00 1)") */
      (k == 1) && 
      (*orig_str == '\\'))
    {
      /* must be from #\( and friends -- a character that happens to be not ok-in-a-name */
      sc->strbuf[1] = orig_str[1];
      sc->strbuf[2] = '\0';
      return(make_sharp_constant(sc, sc->strbuf, UNNESTED_SHARP, BASE_10));
    }

  /* eval_c_string string is a constant so we can't set and unset the token's end char */
  if ((k + 1) >= sc->strbuf_size)
    resize_strbuf(sc);
  
  memcpy((void *)(sc->strbuf), (void *)orig_str, k);
  if (*str != 0) port_string_point(pt)--;
  sc->strbuf[k] = '\0';

  if (atom_case)
    return(make_atom(sc, sc->strbuf, BASE_10, SYMBOL_OK));
  return(make_sharp_constant(sc, sc->strbuf, UNNESTED_SHARP, BASE_10));
}


static s7_pointer string_read_name(s7_scheme *sc, s7_pointer pt, bool atom_case)
{
  /* port_string was allocated (and read from a file) so we can mess with it directly */
  unsigned int k = 0;
  char *orig_str, *str;
  s7_pointer result;
  char endc;
  
  orig_str = (char *)(port_string(pt) + port_string_point(pt) - 1);
  str = (char *)(orig_str + 1);
  
  while (char_ok_in_a_name[(unsigned char)(*str)]) {str++;}
  k = str - orig_str;
  port_string_point(pt) += k;
  
  if (!atom_case)             /* there's a bizarre special case here \ with the next char #\null: (eval-string "(list \\\x00 1)") */
    {
      s7_pointer result;
      char endc;

      if ((k == 1) && 
	  (*orig_str == '\\'))
	{
	  /* must be from #\( and friends -- a character that happens to be not ok-in-a-name */
	  sc->strbuf[1] = orig_str[1];
	  sc->strbuf[2] = '\0';
	  return(make_sharp_constant(sc, sc->strbuf, UNNESTED_SHARP, BASE_10));
	}

      /* port_string was allocated (and read from a file) so we can mess with it directly */
      
      endc = orig_str[k];
      orig_str[k] = '\0';
      result = make_sharp_constant(sc, orig_str, UNNESTED_SHARP, BASE_10);
      
      orig_str[k] = endc;
      if (*str != 0) port_string_point(pt)--;
      return(result);
    }

#if 0
  endc = orig_str[k];
  orig_str[k] = '\0';
#else
  endc = (*str);
  (*str) = '\0';
#endif
  
  switch (*orig_str)
    {
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
    case '.': case '+': case '-':
      result = make_atom(sc, orig_str, BASE_10, SYMBOL_OK);
      break;
      
    case '#':
      result = make_sharp_constant(sc, (char *)(orig_str + 1), UNNESTED_SHARP, BASE_10);
      break;
      
    default:
      /* result = make_symbol(sc, orig_str); 
       *    expanded for speed
       */
      {
	int location;
	s7_pointer x; 
	unsigned int loc = 0;
	const char *c; 
	
	/* expanding these two calls saves a lot of time */
	/* location = symbol_table_hash(orig_str, &loc); */
	/* result = symbol_table_find_by_name(sc, orig_str, location); */
	
	for (c = orig_str; *c; c++) 
	  loc = *c + loc * HASH_MULT;
	location = loc % SYMBOL_TABLE_SIZE; 
	
	result = sc->NIL;
	for (x = vector_element(sc->symbol_table, location); is_not_null(x); x = cdr(x)) 
	  { 
	    const char *s; 
	    s = symbol_name(car(x)); 
	    if ((s[0] == orig_str[0]) &&
		(strings_are_equal(orig_str, s)))
	      /* the string lengths here are k (for orig_str) and symbol_name_length(car(x)),
	       *   but checking that these are equal before calling strcmp is slower?!?
	       */
	      {
		result = car(x);
		break;
	      }
	  }
	if (is_null(result))
	  {
	    if (sc->symbol_table_is_locked)
	      result = sc->F;
	    else 
	      {
		result = new_symbol(sc, orig_str, location); 
		symbol_hash(result) = loc;
	      }
	  }
      }
      break;
    }
  
  (*str) = endc;
  if (endc != 0) port_string_point(pt)--;

  /* skipping the null has one minor consequence:
   *    (let ((str "(+ 1 2 3)")) (set! (str 2) #\null) (eval-string str)) ; "(+\x001 2 3)" -> 6
   *    (let ((str "(+ 1 2 3)")) (set! (str 3) #\null) (eval-string str)) ; "(+ \x00 2 3)" -> missing paren error
   */
  return(result);
}



static s7_pointer read_file(s7_scheme *sc, FILE *fp, const char *name, long max_size, const char *caller)
{
  s7_pointer port;
  long size;
  int port_loc;
  char *content = NULL;

  NEW_CELL(sc, port);
  port_loc = s7_gc_protect(sc, port);
  set_type(port, T_INPUT_PORT);
  port->object.port = (s7_port_t *)calloc(1, sizeof(s7_port_t));
  port_is_closed(port) = false;
  port_original_input_string(port) = sc->NIL;
  port_write_character(port) = input_write_char;

  /* if we're constantly opening files, and each open saves the file name in permanent
   *   memory, we gradually core-up.  
   */
  port_filename(port) = copy_string(name);
  port_line_number(port) = 1;  /* 1st line is numbered 1 */
  add_input_port(sc, port);

  fseek(fp, 0, SEEK_END);
  size = ftell(fp);
  rewind(fp);

  /* pseudo files (under /proc for example) have size=0, but we can read them, so don't assume a 0 length file is empty 
   */

  if ((size > 0) &&                /* if (size != 0) we get (open-input-file "/dev/tty") -> (open "/dev/tty") read 0 bytes of an expected -1? */
      ((max_size < 0) || (size < max_size)))
    {
      size_t bytes;
      content = (char *)malloc((size + 2) * sizeof(char));
      bytes = fread(content, sizeof(char), size, fp);
      if (bytes != (size_t)size)
	{
	  char tmp[256];
	  snprintf(tmp, 256, "(%s \"%s\") read %ld bytes of an expected %ld?", caller, name, (long)bytes, size);
	  port_display(sc->output_port)(sc, tmp, sc->output_port);
	}
      content[size] = '\0';
      content[size + 1] = '\0';
      fclose(fp);

      port_type(port) = STRING_PORT;
      port_string(port) = content;
      port_string_length(port) = size;
      port_string_point(port) = 0;
      port_needs_free(port) = true;
      port_read_character(port) = string_read_char;
      port_read_line(port) = string_read_line;
      port_display(port) = input_display;
      port_read_semicolon(port) = string_read_semicolon;
      port_read_white_space(port) = string_read_white_space;
      port_read_name(port) = string_read_name;
    }
  else
    {
      port_file(port) = fp;
      port_type(port) = FILE_PORT;
      port_needs_free(port) = false;
      port_read_character(port) = file_read_char;
      port_read_line(port) = file_read_line;
      port_display(port) = input_display;
      port_read_semicolon(port) = file_read_semicolon;
      port_read_white_space(port) = file_read_white_space;
      port_read_name(port) = file_read_name;
    }

  s7_gc_unprotect_at(sc, port_loc);
  return(port);
}


static s7_pointer make_input_file(s7_scheme *sc, const char *name, FILE *fp)
{
  #define MAX_SIZE_FOR_STRING_PORT 5000000
  return(read_file(sc, fp, name, MAX_SIZE_FOR_STRING_PORT, "open"));
}


static s7_pointer open_input_file_1(s7_scheme *sc, const char *name, const char *mode, const char *caller)
{
  FILE *fp;
  /* see if we can open this file before allocating a port */

  errno = 0;
  fp = fopen(name, mode);
  if (!fp)
    {
#ifndef _MSC_VER
      if (errno == EINVAL)
	return(file_error(sc, caller, "invalid mode", mode));
#endif
      return(file_error(sc, caller, strerror(errno), name));
    }

  return(make_input_file(sc, name, fp));
}


s7_pointer s7_open_input_file(s7_scheme *sc, const char *name, const char *mode)
{
  return(open_input_file_1(sc, name, mode, "open-input-file"));
}


static s7_pointer g_open_input_file(s7_scheme *sc, s7_pointer args)
{
  #define H_open_input_file "(open-input-file filename (mode \"r\")) opens filename for reading"
  s7_pointer name = car(args);

  if (!s7_is_string(name))
    return(s7_wrong_type_arg_error(sc, "open-input-file filename,", (is_null(cdr(args))) ? 0 : 1, name, "a string"));
  
  if (is_pair(cdr(args)))
    {
      if (!s7_is_string(cadr(args)))
	return(s7_wrong_type_arg_error(sc, "open-input-file mode,", 2, cadr(args), "a string (a mode such as \"r\")"));
      return(open_input_file_1(sc, s7_string(name), s7_string(cadr(args)), "open-input-file"));
    }
  return(open_input_file_1(sc, s7_string(name), "r", "open-input-file"));
}


static void make_standard_ports(s7_scheme *sc)
{
  s7_pointer x;

  /* standard output */
  x = (s7_cell *)permanent_calloc(sizeof(s7_cell));
  x->hloc = NOT_IN_HEAP;
  set_type(x, T_OUTPUT_PORT | T_IMMUTABLE);
  x->object.port = (s7_port_t *)permanent_calloc(sizeof(s7_port_t));
  port_type(x) = FILE_PORT;
  port_is_closed(x) = false;
  port_filename(x) = copy_string("*stdout*");
  port_file_number(x) = remember_file_name(sc, port_filename(x)); /* these numbers need to be correct for the evaluator (__FUNC__ data) */
  port_line_number(x) = 0;
  port_file(x) = stdout;
  port_needs_free(x) = false;
  port_read_character(x) = output_read_char;
  port_read_line(x) = output_read_line;
  port_display(x) = stdout_display;
  port_write_character(x) = stdout_write_char;
  sc->standard_output = x;

  /* standard error */
  x = (s7_cell *)permanent_calloc(sizeof(s7_cell));
  x->hloc = NOT_IN_HEAP;
  set_type(x, T_OUTPUT_PORT | T_IMMUTABLE);
  x->object.port = (s7_port_t *)permanent_calloc(sizeof(s7_port_t));
  port_type(x) = FILE_PORT;
  port_is_closed(x) = false;
  port_filename(x) = copy_string("*stderr*");
  port_file_number(x) = remember_file_name(sc, port_filename(x));
  port_line_number(x) = 0;
  port_file(x) = stderr;
  port_needs_free(x) = false;
  port_read_character(x) = output_read_char;
  port_read_line(x) = output_read_line;
  port_display(x) = stderr_display;
  port_write_character(x) = stderr_write_char;
  sc->standard_error = x;

  /* standard input */
  x = (s7_cell *)permanent_calloc(sizeof(s7_cell));
  x->hloc = NOT_IN_HEAP;
  set_type(x, T_INPUT_PORT | T_IMMUTABLE);
  x->object.port = (s7_port_t *)permanent_calloc(sizeof(s7_port_t));
  port_type(x) = FILE_PORT;
  port_is_closed(x) = false;
  port_original_input_string(x) = sc->NIL;
  port_filename(x) = copy_string("*stdin*");
  port_file_number(x) = remember_file_name(sc, port_filename(x));
  port_line_number(x) = 0;
  port_file(x) = stdin;
  port_needs_free(x) = false;
  port_read_character(x) = file_read_char;
  port_read_line(x) = stdin_read_line;
  port_display(x) = input_display;
  port_read_semicolon(x) = file_read_semicolon;
  port_read_white_space(x) = file_read_white_space;
  port_read_name(x) = file_read_name;
  port_write_character(x) = input_write_char;
  sc->standard_input = x;

  s7_define_constant(sc, "*stdin*", sc->standard_input);
  s7_define_constant(sc, "*stdout*", sc->standard_output);
  s7_define_constant(sc, "*stderr*", sc->standard_error);

  sc->input_port = sc->standard_input;
  sc->output_port = sc->standard_output;
  sc->error_port = sc->standard_error;
  sc->current_file = NULL;
  sc->current_line = -1;
}


s7_pointer s7_open_output_file(s7_scheme *sc, const char *name, const char *mode)
{
  FILE *fp;
  s7_pointer x;
  /* see if we can open this file before allocating a port */
  
  errno = 0;
  fp = fopen(name, mode);
  if (!fp)
    {
#ifndef _MSC_VER
      if (errno == EINVAL)
	return(file_error(sc, "open-output-file", "invalid mode", mode));
#endif
      return(file_error(sc, "open-output-file", strerror(errno), name));
    }

  NEW_CELL(sc, x);
  set_type(x, T_OUTPUT_PORT);
  
  x->object.port = (s7_port_t *)calloc(1, sizeof(s7_port_t));
  port_type(x) = FILE_PORT;
  port_is_closed(x) = false;
  port_filename(x) = copy_string(name);
  port_line_number(x) = 1;
  port_file(x) = fp;
  port_needs_free(x) = false;
  port_read_character(x) = output_read_char;
  port_read_line(x) = output_read_line;
  port_display(x) = file_display;
  port_write_character(x) = file_write_char;
  port_position(x) = 0;
  port_data(x) = (unsigned char *)malloc(PORT_DATA_SIZE);
  add_output_port(sc, x);
  return(x);
}


static s7_pointer g_open_output_file(s7_scheme *sc, s7_pointer args)
{
  #define H_open_output_file "(open-output-file filename (mode \"w\")) opens filename for writing"
  s7_pointer name = car(args);

  if (!s7_is_string(name))
    return(s7_wrong_type_arg_error(sc, "open-output-file filename,", (is_null(cdr(args))) ? 0 : 1, name, "a string"));
  
  if (is_pair(cdr(args)))
    {
      if (!s7_is_string(cadr(args)))
	return(s7_wrong_type_arg_error(sc, "open-output-file mode,", 2, cadr(args), "a string (a mode such as \"w\")"));
      return(s7_open_output_file(sc, s7_string(name), s7_string(cadr(args))));
    }
  return(s7_open_output_file(sc, s7_string(name), "w"));
}


s7_pointer s7_open_input_string(s7_scheme *sc, const char *input_string)
{
  s7_pointer x;
  NEW_CELL(sc, x);
  set_type(x, T_INPUT_PORT);
  
  x->object.port = (s7_port_t *)calloc(1, sizeof(s7_port_t));
  port_type(x) = STRING_PORT;
  port_is_closed(x) = false;
  port_original_input_string(x) = sc->NIL;
  port_string(x) = (char *)input_string;
  port_string_length(x) = safe_strlen(input_string);
  port_string_point(x) = 0;
  port_filename(x) = NULL;
  port_file_number(x) = -1;
  port_needs_free(x) = false;
  port_read_character(x) = string_read_char;
  port_read_line(x) = string_read_line;
  port_display(x) = input_display;
  port_read_semicolon(x) = string_read_semicolon;
  port_read_white_space(x) = string_read_white_space;
  port_read_name(x) = string_read_name_no_free;
  port_write_character(x) = input_write_char;
  add_input_port(sc, x);
  return(x);
}


static s7_pointer g_open_input_string(s7_scheme *sc, s7_pointer args)
{
  #define H_open_input_string "(open-input-string str) opens an input port reading str"
  s7_pointer input_string = car(args);

  if (!s7_is_string(input_string))
    return(s7_wrong_type_arg_error(sc, "open-input-string", 0, input_string, "a string"));
  
  return(s7_open_input_string(sc, s7_string(input_string))); /* presumably the caller is protecting the input string?? */
}


#define STRING_PORT_INITIAL_LENGTH 128

s7_pointer s7_open_output_string(s7_scheme *sc)
{
  s7_pointer x;
  NEW_CELL(sc, x);
  set_type(x, T_OUTPUT_PORT);
  
  x->object.port = (s7_port_t *)calloc(1, sizeof(s7_port_t));
  port_type(x) = STRING_PORT;
  port_is_closed(x) = false;
  port_string_length(x) = STRING_PORT_INITIAL_LENGTH;
  port_string(x) = (char *)calloc(STRING_PORT_INITIAL_LENGTH, sizeof(char));
  port_string_point(x) = 0;
  port_needs_free(x) = true;
  port_read_character(x) = output_read_char;
  port_read_line(x) = output_read_line;
  port_display(x) = string_display;
  port_write_character(x) = string_write_char;
  add_output_port(sc, x);
  return(x);
}


static s7_pointer g_open_output_string(s7_scheme *sc, s7_pointer args)
{
  #define H_open_output_string "(open-output-string) opens an output string port"
  return(s7_open_output_string(sc));
}


const char *s7_get_output_string(s7_scheme *sc, s7_pointer p)
{
  return(port_string(p));
}


static s7_pointer g_get_output_string(s7_scheme *sc, s7_pointer args)
{
  #define H_get_output_string "(get-output-string port) returns the output accumulated in port"
  s7_pointer p = car(args);

  if ((!is_output_port(p)) ||
      (!is_string_port(p)))
    return(s7_wrong_type_arg_error(sc, "get-output-string", 0, p, "an output string port"));
  if (port_is_closed(p))
    return(s7_wrong_type_arg_error(sc, "get-output-string", 0, p, "an active (open) string port"));

  return(s7_make_string(sc, s7_get_output_string(sc, p)));
}


s7_pointer s7_open_input_function(s7_scheme *sc, s7_pointer (*function)(s7_scheme *sc, s7_read_t read_choice, s7_pointer port))
{
  s7_pointer x;
  NEW_CELL(sc, x);
  set_type(x, T_INPUT_PORT);
  
  x->object.port = (s7_port_t *)calloc(1, sizeof(s7_port_t));
  port_type(x) = FUNCTION_PORT;
  port_is_closed(x) = false;
  port_original_input_string(x) = sc->NIL;
  port_needs_free(x) = false;
  port_input_function(x) = function;
  port_read_character(x) = function_read_char;
  port_read_line(x) = function_read_line;
  port_display(x) = input_display;
  port_write_character(x) = input_write_char;
  add_input_port(sc, x);
  return(x);
}


s7_pointer s7_open_output_function(s7_scheme *sc, void (*function)(s7_scheme *sc, unsigned char c, s7_pointer port))
{
  s7_pointer x;
  NEW_CELL(sc, x);
  set_type(x, T_OUTPUT_PORT);
  
  x->object.port = (s7_port_t *)calloc(1, sizeof(s7_port_t));
  port_type(x) = FUNCTION_PORT;
  port_is_closed(x) = false;
  port_needs_free(x) = false;
  port_output_function(x) = function;
  port_read_character(x) = output_read_char;
  port_read_line(x) = output_read_line;
  port_display(x) = function_display;
  port_write_character(x) = function_write_char;
  add_output_port(sc, x);
  return(x);
}


static void push_input_port(s7_scheme *sc, s7_pointer new_port)
{
  sc->input_port_stack = cons(sc, sc->input_port, sc->input_port_stack);
  sc->input_port = new_port;
}


static void pop_input_port(s7_scheme *sc)
{
  if (is_pair(sc->input_port_stack))
    {
      sc->input_port = car(sc->input_port_stack);

      /* is this safe? */
      typeflag(sc->input_port_stack) = 0;
      (*(sc->free_heap_top++)) = sc->input_port_stack;

      sc->input_port_stack = cdr(sc->input_port_stack);
    }
  else sc->input_port = sc->standard_input;
}


static int inchar(s7_pointer pt)
{
  int c;
  if (is_file_port(pt))
    c = fgetc(port_file(pt)); /* not unsigned char! -- could be EOF */
  else 
    {
      if ((!(port_string(pt))) ||
	  (port_string_length(pt) <= port_string_point(pt)))
	return(EOF);
      c = (unsigned char)port_string(pt)[port_string_point(pt)++];
    }

  if (c == '\n')
    port_line_number(pt)++;

  return(c);
}


static void backchar(char c, s7_pointer pt) 
{
  if (c == '\n')
    port_line_number(pt)--;

  if (is_file_port(pt))
    ungetc(c, port_file(pt));
  else 
    {
      if (port_string_point(pt) > 0)
	port_string_point(pt)--;
    }
}


int s7_read_char(s7_scheme *sc, s7_pointer port)
{
  /* needs to be int return value so EOF=-1, but not 255 */
  return(port_read_character(port)(sc, port));
}


int s7_peek_char(s7_scheme *sc, s7_pointer port)
{
  int c;              /* needs to be an int so EOF=-1, but not 255 */
  c = port_read_character(port)(sc, port);
  if (c != EOF)
    backchar(c, port);
  return(c);
}


static s7_pointer g_read_char(s7_scheme *sc, s7_pointer args)
{
  #define H_read_char "(read-char (port (current-input-port))) returns the next character in the input port"

  s7_pointer port;

  if (is_not_null(args))
    port = car(args);
  else port = sc->input_port;
  if (!is_input_port(port))
    return(s7_wrong_type_arg_error(sc, "read-char", 0, port, "an input port"));

  return(chars[port_read_character(port)(sc, port)]);
}


static s7_pointer g_peek_char(s7_scheme *sc, s7_pointer args)
{
  #define H_peek_char "(peek-char (port (current-input-port))) returns the next character in the input port, but does not remove it from the input stream"
  int c;
  s7_pointer port;

  if (is_not_null(args))
    port = car(args);
  else port = sc->input_port;

  if (!is_input_port(port))
    return(s7_wrong_type_arg_error(sc, "peek-char", 0, port, "an input port"));
  if (port_is_closed(port))
    return(s7_wrong_type_arg_error(sc, "peek-char", 0, port, "an open input port"));
      
  c = s7_peek_char(sc, port);
  return(chars[c]);
}


static s7_pointer g_read_line(s7_scheme *sc, s7_pointer args)
{
  #define H_read_line "(read-line port (with-eol #f)) returns the next line from port, or #<eof> (use the function eof-object?).\
If 'with-eol' is not #f, read-line includes the trailing end-of-line character."

  s7_pointer port;
  bool with_eol = false;

  if (is_not_null(args))
    {
      port = car(args);
      if (!is_input_port(port))
	return(s7_wrong_type_arg_error(sc, "read-line", (is_null(cdr(args))) ? 0 : 1, port, "an input port"));

      if (is_not_null(cdr(args)))
	{
	  /* support (read-line fp 'concat) for compatibility with guile */
	  if ((!s7_is_boolean(cadr(args))) &&
	      (!is_symbol(cadr(args))))
	    return(s7_wrong_type_arg_error(sc, "read-line", 2, cadr(args), "#f or #t"));
	  if (cadr(args) != sc->F)
	    with_eol = true;
	}
    }
  else port = sc->input_port;
  return(port_read_line(port)(sc, port, with_eol));
}



s7_pointer s7_read(s7_scheme *sc, s7_pointer port)
{
  if (is_input_port(port))
    {
      bool old_longjmp;
      old_longjmp = sc->longjmp_ok;
      if (!sc->longjmp_ok)
	{
	  sc->longjmp_ok = true;
	  if (setjmp(sc->goto_start) != 0)
	    return(sc->value);
	}
      push_input_port(sc, port);
      push_stack(sc, OP_EVAL_DONE, port, sc->NIL);
      eval(sc, OP_READ_INTERNAL);
      sc->longjmp_ok = old_longjmp;
      pop_input_port(sc);
      return(sc->value);
    }
  return(s7_wrong_type_arg_error(sc, "read", 0, port, "an input port"));
}


static s7_pointer g_read(s7_scheme *sc, s7_pointer args)
{
  #define H_read "(read (port (current-input-port))) returns the next object in the input port, or #<eof> at the end"
  s7_pointer port;
  
  if (is_not_null(args))
    port = car(args);
  else port = sc->input_port;
  
  if (!is_input_port(port)) /* was also not stdin */
    return(s7_wrong_type_arg_error(sc, "read", 0, port, "an input port"));
  if (port_is_closed(port))
    return(s7_wrong_type_arg_error(sc, "read", 0, port, "an open input port"));

  if (is_function_port(port))
    return((*(port_input_function(port)))(sc, S7_READ, port));
  
  if ((is_string_port(port)) &&
      (port_string_length(port) <= port_string_point(port)))
    return(sc->EOF_OBJECT);

  push_input_port(sc, port);

  push_stack(sc, OP_READ_DONE, sc->NIL, sc->NIL); /* this stops the internal read process so we only get one form */
  push_stack(sc, OP_READ_INTERNAL, sc->NIL, sc->NIL);
  return(port);
}


static FILE *search_load_path(s7_scheme *sc, const char *name)
{
  int i, len, name_len;
  s7_pointer lst;

  lst = s7_load_path(sc);
  len = s7_list_length(sc, lst);
  name_len = safe_strlen(name);

  for (i = 0; i < len; i++)
    {
      const char *new_dir;
      int size;
      new_dir = s7_string(s7_list_ref(sc, lst, i));
      if (new_dir)
	{
	  char *new_name;
	  FILE *fp;
	  size = name_len + safe_strlen(new_dir) + 2;
	  new_name = (char *)malloc(size * sizeof(char));
	  snprintf(new_name, size, "%s/%s", new_dir, name);
	  fp = fopen(new_name, "r");
	  free(new_name);
	  if (fp) return(fp);
	}
    }

  return(NULL);
}


static s7_pointer load_file(s7_scheme *sc, FILE *fp, const char *name)
{
  return(read_file(sc, fp, name, -1, "load"));  /* -1 means always read its contents into a local string */
}


s7_pointer s7_load(s7_scheme *sc, const char *filename)
{
  bool old_longjmp;
  s7_pointer port;
  FILE *fp;
  
  fp = fopen(filename, "r");
  if (!fp)
    fp = search_load_path(sc, filename);
  if (!fp)
    return(file_error(sc, "load", "can't open", filename));

  if (is_pair(hook_functions(sc->load_hook)))
    s7_hook_apply(sc, sc->load_hook, list_1(sc, s7_make_string(sc, filename)));

  port = load_file(sc, fp, filename);
  port_file_number(port) = remember_file_name(sc, filename);
  push_input_port(sc, port);
  
  /* it's possible to call this recursively (s7_load is XEN_LOAD_FILE which can be invoked via s7_call)
   *   but in that case, we actually want it to behave like g_load and continue the evaluation upon completion
   */
  sc->envir = sc->NIL;
  
  if (!sc->longjmp_ok)
    {
      push_stack(sc, OP_LOAD_RETURN_IF_EOF, port, sc->code);
      
      old_longjmp = sc->longjmp_ok;
      if (!sc->longjmp_ok)
	{
	  sc->longjmp_ok = true;
	  if (setjmp(sc->goto_start) != 0)
	    eval(sc, sc->op);
	  else eval(sc, OP_READ_INTERNAL);
	}
      sc->longjmp_ok = old_longjmp;  
      pop_input_port(sc);
      s7_close_input_port(sc, port);
    }
  else
    {
      /* caller here is assuming the load will be complete before this function returns */
      push_stack(sc, OP_LOAD_RETURN_IF_EOF, sc->args, sc->code);

      eval(sc, OP_READ_INTERNAL);
      pop_input_port(sc);
      s7_close_input_port(sc, port);
    }

  return(sc->UNSPECIFIED);
}


#include <sys/stat.h>

static bool is_directory(const char *filename)
{
#if HAVE_WINDOZE
  return(false);

#else
  /* from snd-file.c */
#ifdef S_ISDIR
  struct stat statbuf;
#if HAVE_LSTAT
  return((lstat(filename, &statbuf) >= 0) &&
	 (S_ISDIR(statbuf.st_mode)));
  return(false);
#else
  return((stat(filename, &statbuf) == 0) && 
	 (S_ISDIR(statbuf.st_mode)));
#endif
#endif
#endif
}


static s7_pointer g_load(s7_scheme *sc, s7_pointer args)
{
  #define H_load "(load file (env (global-environment))) loads the scheme file 'file'. The 'env' argument \
defaults to the global environment.  To load into the current environment instead, pass (current-environment)."

  FILE *fp = NULL;
  s7_pointer name, port;
  const char *fname;
  
  name = car(args);
  if (!s7_is_string(name))
    return(s7_wrong_type_arg_error(sc, "load filename,", (is_null(cdr(args))) ? 0 : 1, name, "a string"));

  if (is_not_null(cdr(args))) 
    {
      s7_pointer e;
      e = cadr(args);
      if (!is_environment(e))
	return(s7_wrong_type_arg_error(sc, "load", 2, e, "an environment"));
      if (e == sc->global_env)
	sc->envir = sc->NIL;
      else sc->envir = e;
    }
  else sc->envir = sc->NIL;
  
  fname = s7_string(name);
  if ((!fname) || (!(*fname)))                 /* fopen("", "r") returns a file pointer?? */
    return(s7_error(sc, sc->OUT_OF_RANGE, 
		    list_2(sc, 
			   make_protected_string(sc, "load's first argument, ~S, should be a filename"),
			   name)));

  if (is_directory(fname))
    return(s7_error(sc, sc->WRONG_TYPE_ARG, 
		    list_2(sc, make_protected_string(sc, "load argument, ~S, is a directory"), name)));

  fp = fopen(fname, "r");

#if WITH_GCC
  if (!fp)
    {
      /* catch one special case, "~/..." since it causes 99.9% of the "can't load ..." errors
       */
      if ((fname[0] == '~') &&
	  (fname[1] == '/'))
	{
	  char *home;
	  home = getenv("HOME");
	  if (home)
	    {
	      char *filename;
	      filename = (char *)calloc(safe_strlen(fname) + safe_strlen(home) + 1, sizeof(char));
	      strcpy(filename, home);
	      strcat(filename, (char *)(fname + 1));
	      fp = fopen(filename, "r");
	      free(filename);
	    }
	}
    }
#endif

  if (!fp)
    {
      fp = search_load_path(sc, fname);
      if (!fp)
	return(file_error(sc, "load", "can't open", fname));
    }

  port = load_file(sc, fp, fname);
  port_file_number(port) = remember_file_name(sc, fname);
  push_input_port(sc, port);

  push_stack(sc, OP_LOAD_CLOSE_AND_POP_IF_EOF, sc->NIL, sc->NIL);  /* was pushing args and code, but I don't think they're used later */
  push_stack(sc, OP_READ_INTERNAL, sc->NIL, sc->NIL);
  
  /* now we've opened and moved to the file to be loaded, and set up the stack to return
   *   to where we were when it is read.  Call *load-hook* if it is a procedure.
   */
  
  if (is_not_null(hook_functions(sc->load_hook)))
    push_stack(sc, OP_HOOK_APPLY, list_1(sc, s7_make_string(sc, fname)), hook_functions(sc->load_hook));

  return(sc->UNSPECIFIED);
}


s7_pointer s7_load_path(s7_scheme *sc)
{
  return(s7_symbol_value(sc, make_symbol(sc, "*load-path*")));
}


s7_pointer s7_add_to_load_path(s7_scheme *sc, const char *dir)
{
  s7_pointer load_path;
  load_path = make_symbol(sc, "*load-path*");

  s7_symbol_set_value(sc, 
		      load_path,
		      cons(sc, 
			      s7_make_string(sc, dir), 
			      s7_symbol_value(sc, load_path)));
  return(s7_load_path(sc));
}


static s7_pointer g_load_path_set(s7_scheme *sc, s7_pointer args)
{
  /* new value must be either '() or a proper list of strings */
  if (is_null(cadr(args))) return(cadr(args));
  if (is_pair(cadr(args)))
    {
      s7_pointer x;
      for (x = cadr(args); is_pair(x); x = cdr(x))
	if (!s7_is_string(car(x)))
	  return(sc->ERROR);
      if (is_null(x))
	return(cadr(args));
    }
  return(sc->ERROR);
}


static s7_pointer eval_string_1(s7_scheme *sc, const char *str)
{
  s7_pointer port;

  port = s7_open_input_string(sc, str);
  push_input_port(sc, port);

  push_stack(sc, OP_BARRIER, port, sc->NIL);
  /* we're being called directly from C here, not as part of a scheme program.
   *    Use this op to protect the port, I guess.
   */
  push_stack(sc, OP_EVAL_STRING, sc->args, sc->code);  
  /* eval-string is not tail-recursive because it pushes markers in eval to catch
   *    multiple statements in one eval-string call.
   */
  eval(sc, OP_READ_INTERNAL);

  pop_input_port(sc);
  s7_close_input_port(sc, port);
  if (is_multiple_value(sc->value))                    /* (+ 1 (eval-string "(values 2 3)")) */
    sc->value = splice_in_values(sc, multiple_value(sc->value));

  return(sc->value);
}


s7_pointer s7_eval_c_string(s7_scheme *sc, const char *str)
{
  bool old_longjmp;
  s7_pointer port, old_envir;
  /* this can be called recursively via s7_call */

  old_envir = sc->envir;
  sc->envir = sc->NIL; /* C call assumes top level, I think.  This is needed in any case
			       *   by dlinit -- the init function will be called in some local environment,
			       *   but the library entities it defines should obviously be top level,
			       *   as if via load.
			       */
  if (sc->longjmp_ok)
    {
      s7_pointer result;
      result = eval_string_1(sc, str);
      sc->envir = old_envir;
      return(result);
    }
  
  stack_reset(sc); 
  push_stack(sc, OP_EVAL_STRING, old_envir, sc->NIL); /* GC protect envir */

  port = s7_open_input_string(sc, str);
  push_input_port(sc, port);
  
  old_longjmp = sc->longjmp_ok;
  if (!sc->longjmp_ok)
    {
      sc->longjmp_ok = true;
      if (setjmp(sc->goto_start) != 0)
	eval(sc, sc->op);
      else eval(sc, OP_READ_INTERNAL);
    }
  
  sc->longjmp_ok = old_longjmp;
  pop_input_port(sc);
  s7_close_input_port(sc, port);
  sc->envir = old_envir;  

  return(sc->value);
}


static s7_pointer g_eval_string(s7_scheme *sc, s7_pointer args)
{
  #define H_eval_string "(eval-string str (env (current-environment))) returns the result of evaluating the string str as Scheme code"
  s7_pointer port;
  
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "eval-string", (is_null(cdr(args))) ? 0 : 1, car(args), "a string"));
  
  if (is_not_null(cdr(args)))
    {
      s7_pointer e;
      e = cadr(args);
      if (!is_environment(e))
 	return(s7_wrong_type_arg_error(sc, "eval-string", 2, e, "an environment"));
      if (e == sc->global_env)
	sc->envir = sc->NIL;
      else sc->envir = e;
    }

  port = s7_open_input_string(sc, s7_string(car(args)));
  push_input_port(sc, port);
  
  push_stack(sc, OP_EVAL_STRING_1, args, sc->code); /* was sc->args */
  push_stack(sc, OP_READ_INTERNAL, sc->NIL, sc->NIL);
  
  return(sc->F);
}



static s7_pointer call_with_input(s7_scheme *sc, s7_pointer port, s7_pointer args)
{
  port_original_input_string(port) = car(args);
  push_stack(sc, OP_UNWIND_INPUT, sc->input_port, port);
  push_stack(sc, OP_APPLY, list_1(sc, port), cadr(args));
  return(sc->F);
}


static s7_pointer g_call_with_input_string(s7_scheme *sc, s7_pointer args)
{
  #define H_call_with_input_string "(call-with-input-string str proc) opens a string port for str and applies proc to it"
  
  /* (call-with-input-string "44" (lambda (p) (+ 1 (read p)))) -> 45
   */
  
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "call-with-input-string", 1, car(args), "a string"));
  if (!is_procedure(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "call-with-input-string", 2, cadr(args), "a procedure"));
  if ((is_continuation(cadr(args))) || is_goto(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "call-with-input-string", 2, cadr(args), "a normal procedure (not a continuation)"));
  
  return(call_with_input(sc, s7_open_input_string(sc, s7_string(car(args))), args));
}


static s7_pointer g_call_with_input_file(s7_scheme *sc, s7_pointer args)
{
  #define H_call_with_input_file "(call-with-input-file filename proc) opens filename and calls proc with the input port as its argument"
  
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "call-with-input-file", 1, car(args), "a string (a filename)"));
  if (!is_procedure(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "call-with-input-file", 2, cadr(args), "a procedure"));
  if ((is_continuation(cadr(args))) || is_goto(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "call-with-input-file", 2, cadr(args), "a normal procedure (not a continuation)"));
  
  return(call_with_input(sc, open_input_file_1(sc, s7_string(car(args)), "r", "call-with-input-file"), args));
}


static s7_pointer with_input(s7_scheme *sc, s7_pointer port, s7_pointer args)
{
  s7_pointer old_input_port;
  old_input_port = sc->input_port;
  sc->input_port = port;
  port_original_input_string(port) = car(args);
  
  push_stack(sc, OP_UNWIND_INPUT, old_input_port, port);
  push_stack(sc, OP_APPLY, sc->NIL, cadr(args));
  return(sc->F);
}


static s7_pointer g_with_input_from_string(s7_scheme *sc, s7_pointer args)
{
  #define H_with_input_from_string "(with-input-from-string str thunk) opens str as the temporary current-input-port and calls thunk"
  
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "with-input-from-string", 1, car(args), "a string"));
  if (!is_thunk(sc, cadr(args)))
    return(s7_wrong_type_arg_error(sc, "with-input-from-string", 2, cadr(args), "a thunk"));
  
  /* since the arguments are evaluated before we get here, we can get some confusing situations:
   *   (with-input-from-string "#x2.1" (read))
   *   (read) -> whatever it can get from the current input port!
   *   ";with-input-from-string argument 2, #<eof>, is untyped but should be a thunk"
   */
  
  return(with_input(sc, s7_open_input_string(sc, s7_string(car(args))), args));
}


static s7_pointer g_with_input_from_file(s7_scheme *sc, s7_pointer args)
{
  #define H_with_input_from_file "(with-input-from-file filename thunk) opens filename as the temporary current-input-port and calls thunk"
  
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "with-input-from-file", 1, car(args), "a string (a filename)"));
  if (!is_thunk(sc, cadr(args)))
    return(s7_wrong_type_arg_error(sc, "with-input-from-file", 2, cadr(args), "a thunk"));
  
  return(with_input(sc, open_input_file_1(sc, s7_string(car(args)), "r", "with-input-from-file"), args));
}


void s7_write_char(s7_scheme *sc, int c, s7_pointer pt) 
{
  port_write_character(pt)(sc, c, pt);
}



#define IN_QUOTES true
#define NOT_IN_QUOTES false

static char *slashify_string(const char *p, int len, bool quoted, bool *slashified)
{
  int i, j = 0, cur_size;
  char *s;

  cur_size = len + 256;
  s = (char *)calloc(cur_size + 2, sizeof(char));
  /* this can be non-null even if there's not enough memory, but I think I'll check in the caller */
  if (quoted) s[j++] = '"';

  for (i = 0; i < len; i++) 
    {
      if (slashify_table[((unsigned char)(p[i]))])
	{
	  s[j++] = '\\';
	  (*slashified) = true;
	  switch (p[i]) 
	    {
	    case '"':
	      s[j++] = '"';
	      break;
	      
	    case '\\':
	      s[j++] = '\\';
	      break;
	      
	    default:               /* this is the "\x01" stuff */
	      { 
		unsigned int n;
		static char dignum[] = "0123456789abcdef";
		s[j++] = 'x';
		n = (unsigned int)(p[i]);
		if (n < 16)
		  s[j++] = '0';
		else s[j++] = dignum[(n / 16) % 16];
		s[j++] = dignum[n % 16];
	      }
	      break;
	    }
	}
      else s[j++] = p[i];
      if (j >= cur_size) /* even with 256 extra, we can overflow (for example, inordinately many tabs in ALSA output) */
	{
	  int k;
	  cur_size *= 2;
	  s = (char *)realloc(s, (cur_size + 2) * sizeof(char));
	  for (k = j; k < cur_size + 2; k++) s[k] = 0;
	}
    }
  if (quoted) s[j++] = '"';
  return(s);
}


/* (let () (define (hi a) (+ a 1)) (object->string hi))
 * (let () (define (hi a) (asd a 1)) (environment->list (procedure-environment hi)))
 */

static const char *c_closure_name(s7_scheme *sc, s7_pointer closure)
{
  s7_pointer x;

  if (is_safe_closure(closure_body(closure)))
    x = find_local_symbol(sc, next_environment(closure_environment(closure)), sc->__FUNC__);  /* skip the built-in arg list */
  else x = find_local_symbol(sc, closure_environment(closure), sc->__FUNC__);  /* returns nil if no __func__ */

  if (is_slot(x))
    {
      x = slot_value(x);
      if (is_symbol(x))
	return(symbol_name(x));
      if ((is_pair(x)) &&
	  (is_symbol(car(x))))
	return(symbol_name(car(x)));
    }
  return("#<closure>");
}


static char *describe_type_bits(s7_pointer obj)
{
  /* extracted from atom_to_c_string for debugging (gdb printout)
   */
  char *buf;
  buf = (char *)calloc(512, sizeof(char));
  snprintf(buf, 512, "type: %d (%s), flags: #x%x%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s", 
	   type(obj), 
	   type_name(obj),
	   typeflag(obj),
	   is_procedure(obj) ?          " procedure" : "",
	   is_marked(obj) ?             " gc-marked" : "",
	   is_immutable(obj) ?          " immutable" : "",
	   is_global(obj) ?             " global" : "",
	   is_any_macro(obj) ?          " anymac" : "",
	   is_expansion(obj) ?          " expansion" : "",
	   is_multiple_value(obj) ?     " values" : "",
	   is_keyword(obj) ?            " keyword" : "",
	   dont_eval_args(obj) ?        " dont-eval-args" : "",
	   is_syntactic(obj) ?          " syntactic" : "",
	   is_overlaid(obj) ?           " overlay" : "",
	   is_checked(obj) ?            " checked" : "",
	   is_unsafe(obj) ?             " unsafe" : "",
	   is_optimized(obj) ?          " optimized" : "",
	   is_safe_closure(obj) ?       " safe-closure" : "",
	   is_safe_procedure(obj) ?     " safe-procedure" : "",
	   is_setter(obj) ?             " setter" : "",
	   has_table(obj) ?             " function-table" : "",
	   is_one_liner(obj) ?          " one-liner" : "",
	   needs_copied_args(obj) ?     " copy-args" : "",
	   is_tail_call(obj) ?          " tail-call" : "",
	   ((typeflag(obj) & UNUSED_BITS) != 0) ? " bad bits!" : "");
  return(buf);
}

#define WITH_ELLIPSES false

static char *atom_to_c_string(s7_scheme *sc, s7_pointer obj, bool use_write)
{
  switch (type(obj))
    {
    case T_BOOLEAN:
      if (obj == sc->T)
	return(copy_string("#t"));
      return(copy_string("#f"));

    case T_NIL:
      return(copy_string("()"));
  
    case T_UNTYPED:
      if (obj == sc->EOF_OBJECT)
	return(copy_string("#<eof>"));
  
      if (obj == sc->UNDEFINED) 
	return(copy_string("#<undefined>"));
  
      if ((is_unspecified(obj)) || (obj == sc->NO_VALUE))
	return(copy_string("#<unspecified>"));

      if (obj == sc->ELSE)
	return(copy_string("else"));
      break;

    case T_INPUT_PORT:
    case T_OUTPUT_PORT:
      return(describe_port(sc, obj));

    case T_HOOK:
      return(copy_string("#<hook>"));

    case T_COUNTER:
      {
	char *str;
	str = (char *)malloc(64 * sizeof(char));
	snprintf(str, 64, "#<counter: %lld>", counter_count(obj));
	return(str);
      }

    case T_SLOT:
      {
	int len;
	char *str, *sym, *val;
	sym = symbol_name(slot_symbol(obj));
	val = atom_to_c_string(sc, slot_value(obj), use_write);
	len = symbol_name_length(slot_symbol(obj)) + safe_strlen(val) + 16;
	str = (char *)malloc(len * sizeof(char));
	snprintf(str, len, "#<slot: %s %s>", sym, val);
	free(val);
	return(str);
      }

    case T_INTEGER:
    case T_RATIO:
    case T_REAL:
    case T_COMPLEX:
      return(number_to_string_base_10(obj, 0, 14, 'g')); /* 20 digits is excessive in this context */

#if WITH_GMP
    case T_BIG_INTEGER:
    case T_BIG_RATIO:
    case T_BIG_REAL:
    case T_BIG_COMPLEX:
      return(big_number_to_string_with_radix(obj, BASE_10, 0));
#endif

    case T_SYMBOL:
      {
	bool slashified = false;
	char *str;
	/* I think this is the only place we print a symbol's name */
	/* return(copy_string_with_len(symbol_name(obj), symbol_name_length(obj))); */

	str = slashify_string(symbol_name(obj), symbol_name_length(obj), NOT_IN_QUOTES, &slashified);
	if (slashified)
	  {
	    char *symstr;
	    int len;
	    len = safe_strlen(str) + 16;
	    symstr = (char *)calloc(len, sizeof(char));
	    snprintf(symstr, len, "(symbol \"%s\")", str);
	    free(str);
	    return(symstr);
	  }
	return(str);
      }

    case T_SYNTAX:
      return(copy_string(op_names[(int)syntax_opcode(obj)]));

    case T_STRING:
      if (string_length(obj) > 0)
	{
	  /* if string_length is enormous, this can cause an eventual segfault.
	   * for now, print enough chars to make anyone happy
	   */
	  int len;
	  bool slashified = false;
	  len = string_length(obj);
	  if (len > (1 << 24))
	    len = (1 << 24);
	  if (!use_write) 
	    return(copy_string_with_len(string_value(obj), len));
	  return(slashify_string(string_value(obj), len, IN_QUOTES, &slashified));
	}
      if (!use_write)
	return(NULL);
      return(copy_string("\"\""));

    case T_CHARACTER:
      {
	#define P_SIZE 16
	char *p;
	unsigned char c;
	p = (char *)malloc(P_SIZE * sizeof(char));
	c = (unsigned char)s7_character(obj);             /* if not unsigned, (write (integer->char 212) -> #\xffffffd4! */
	if (!use_write) 
	  {
	    p[0]= c;
	    p[1]= 0;
	  } 
	else 
	  {
	    switch (c) 
	      {
	      case ' ':
		snprintf(p, P_SIZE, "#\\space"); 
		break;

	      case '\n':
		snprintf(p, P_SIZE, "#\\newline"); 
		break;

	      case '\r':
		snprintf(p, P_SIZE, "#\\return"); 
		break;

	      case '\t':
		snprintf(p, P_SIZE, "#\\tab"); 
		break;

	      case '\0':
		snprintf(p, P_SIZE, "#\\null");
		break;

	      default:
		if ((c < 32) || (c >= 127))
		  snprintf(p, P_SIZE, "#\\x%x", c);
		else snprintf(p, P_SIZE, "#\\%c", c); 
		break;
	      }
	  }
	return(p);
      }

    case T_MACRO:
      return(copy_string("#<macro>"));
    case T_BACRO:
      return(copy_string("#<bacro>"));
  
    case T_CLOSURE:
    case T_CLOSURE_STAR:
      return(copy_string(c_closure_name(sc, obj))); /* this looks for __func__ */
  
    case T_C_OPT_ARGS_FUNCTION:
    case T_C_RST_ARGS_FUNCTION:
    case T_C_LST_ARGS_FUNCTION:
    case T_C_ANY_ARGS_FUNCTION:
    case T_C_FUNCTION:
      return(copy_string(c_function_name(obj)));

    case T_C_MACRO:
      return(copy_string(c_macro_name(obj)));
  
    case T_C_POINTER:
      {
	char *str;
	str = (char *)calloc(32, sizeof(char));
	snprintf(str, 32, "#<c_pointer %p>", raw_pointer(obj));
	return(str);
      }
  
    case T_CONTINUATION:
      return(copy_string("#<continuation>"));
  
    case T_GOTO:
      return(copy_string("#<goto>"));
  
    case T_CATCH:                        /* this can't happen */
      return(copy_string("#<catch>"));
  
    case T_DYNAMIC_WIND:                 /* this can't happen */
      return(copy_string("#<dynamic-wind>"));
  
    case T_C_OBJECT:
      return(object_print(sc, obj));     /* this allocates already */

    case T_VECTOR: 
      return(copy_string("#<vector>"));

    case T_ENVIRONMENT:
      return(copy_string("#<environment>"));

    case T_PAIR: 
      return(copy_string("#<pair>"));

    default:
      break;
    }
  {
    char *str, *tmp;
    int len;
    tmp = describe_type_bits(obj);
    len = 32 + safe_strlen(tmp);
    str = (char *)malloc(len * sizeof(char));
    snprintf(str, len, "<unknown object! %s>", tmp);
    free(tmp);
    return(str);
  }
}


bool s7_is_valid_pointer(s7_pointer arg)
{
  return((arg) &&
	 (type(arg) > T_UNTYPED) && (type(arg) < BUILT_IN_TYPES));
}


static int display_multivector(s7_scheme *sc, s7_pointer vec, int out_len, int flat_ref, int dimension, int dimensions, char *out_str, char **elements, char *last)
{
  int i;

  if (*last == ')')
    strcat(out_str, " ");

  strcat(out_str, "(");
  (*last) = '(';

  for (i = 0; i < vector_dimension(vec, dimension); i++)
    {
      if (dimension == (dimensions - 1))
	{
	  if (flat_ref < out_len)
	    strcat(out_str, elements[flat_ref++]);
	  else
	    {
	      strcat(out_str, "...)");
	      return(flat_ref);
	    }
	  if (i < (vector_dimension(vec, dimension) - 1))
	    strcat(out_str, " ");
	}
      else 
	{
	  if (flat_ref < out_len)
	    flat_ref = display_multivector(sc, vec, out_len, flat_ref, dimension + 1, dimensions, out_str, elements, last);
	  else 
	    {
	      strcat(out_str, "...)");
	      return(flat_ref);
	    }
	}
    }
  strcat(out_str, ")");
  (*last) = ')';
  return(flat_ref);
}


static bool has_structure(s7_pointer p) 
{
  return((type(p) == T_PAIR) ||
	 (type(p) == T_VECTOR) ||
	 (type(p) == T_HASH_TABLE));
}

#define INITIAL_SHARED_INFO_SIZE 8

static int shared_ref(shared_info *ci, s7_pointer p)
{
  int i;
  s7_pointer *objs;
  objs = ci->objs;
  for (i = 0; i < ci->top; i++)
    if (objs[i] == p)
      {
	int val;
	val = ci->refs[i];
	if (val > 0)
	  ci->refs[i] = -ci->refs[i];
	return(val);
      }
  return(0);
}


static int peek_shared_ref(shared_info *ci, s7_pointer p)
{
  /* returns 0 if not found, otherwise the ref value for p */
  int i;
  s7_pointer *objs;
  objs = ci->objs;
  for (i = 0; i < ci->top; i++)
    if (objs[i] == p)
      return(ci->refs[i]);
  return(0);
}


static void enlarge_shared_info(shared_info *ci)
{
  int i;
  ci->size *= 2;
  ci->objs = (s7_pointer *)realloc(ci->objs, ci->size * sizeof(s7_pointer));
  ci->refs = (int *)realloc(ci->refs, ci->size * sizeof(int));
  for (i = ci->top; i < ci->size; i++)
    {
      ci->refs[i] = 0;
      ci->objs[i] = NULL;
    }
}


static void add_equal_ref(shared_info *ci, s7_pointer x, s7_pointer y)
{
  /* assume neither x nor y is in the table, and that they should share a ref value */

  if ((ci->top + 2) >= ci->size)
    enlarge_shared_info(ci);

  ci->ref++;
  ci->objs[ci->top] = x;
  ci->refs[ci->top++] = ci->ref;
  ci->objs[ci->top] = y;
  ci->refs[ci->top++] = ci->ref;
}


static void add_shared_ref(shared_info *ci, s7_pointer x, int ref_x)
{
  if (ci->top == ci->size)
    enlarge_shared_info(ci);

  ci->objs[ci->top] = x;
  ci->refs[ci->top++] = ref_x;
}


static shared_info *collect_shared_info(s7_scheme *sc, shared_info *ci, s7_pointer top)
{
  int i, ref = -1;
  s7_pointer *p, *objs_end;
  
  objs_end = (s7_pointer *)(ci->objs + ci->top);

  /* look for top in current list */
  for (p = ci->objs; p < objs_end; p++)
    if ((*p) == top)
      {
	i = (int)(p - ci->objs);
	if (ci->refs[i] == 0)
	  ci->refs[i] = ++ci->ref;  /* if found, set the ref number */
	ref = ci->refs[i];
	break;
      }

  if (ref == -1)
    {
      /* top not found -- add it to the list */
      if (ci->top == ci->size)
	enlarge_shared_info(ci);
      ci->objs[ci->top++] = top;

      /* now search the rest of this structure */
      if (is_pair(top))
	{
	  if (has_structure(car(top)))
	    collect_shared_info(sc, ci, car(top));
	  if (has_structure(cdr(top)))
	    collect_shared_info(sc, ci, cdr(top));
	}
      else
	{
	  int i, plen;
	  plen = s7_vector_print_length(sc);
	  if (plen > vector_length(top))
	    plen = vector_length(top);
	  for (i = 0; i < plen; i++)
	    if (has_structure(vector_element(top, i)))
	      collect_shared_info(sc, ci, vector_element(top, i));
	}
    }
  return(ci);
}


static shared_info *new_shared_info(s7_scheme *sc)
{
  shared_info *ci;
  if (sc->circle_info == NULL)
    {
      ci = (shared_info *)calloc(1, sizeof(shared_info));
      ci->size = INITIAL_SHARED_INFO_SIZE;
      ci->objs = (s7_pointer *)malloc(ci->size * sizeof(s7_pointer));
      ci->refs = (int *)calloc(ci->size, sizeof(int));   /* finder expects 0 = unseen previously */
      sc->circle_info = ci;
    }
  else 
    {
      ci = sc->circle_info;
      memset((void *)(ci->refs), 0, ci->top * sizeof(int));
    }
  ci->top = 0;
  ci->ref = 0;
  return(ci);
}


static shared_info *make_shared_info(s7_scheme *sc, s7_pointer top)
{
  shared_info *ci;
  int i, refs;
  s7_pointer *ci_objs;
  int *ci_refs;

  ci = new_shared_info(sc);

  /* collect all pointers associated with top */
  collect_shared_info(sc, ci, top);
  ci_objs = ci->objs;
  ci_refs = ci->refs;

  /* find if any were referenced twice */
  for (i = 0, refs = 0; i < ci->top; i++)
    if (ci_refs[i] > 0)
      {
	if (i == refs)
	  refs++;
	else
	  {
	    ci_objs[refs] = ci_objs[i];
	    ci_refs[refs++] = ci_refs[i];

	    ci_refs[i] = 0;
	    ci_objs[i] = NULL;
	  }
      }
  ci->top = refs;

  if (refs == 0)
    return(NULL);
  return(ci);
}



static char *vector_to_c_string(s7_scheme *sc, s7_pointer vect, bool to_file, shared_info *ci);
static char *hash_table_to_c_string(s7_scheme *sc, s7_pointer hash, bool to_file, shared_info *ci);
static char *list_to_c_string(s7_scheme *sc, s7_pointer lst, shared_info *ci);

static char *s7_object_to_c_string_1(s7_scheme *sc, s7_pointer obj, bool use_write, bool to_file, shared_info *ci)
{
  if (s7_is_vector(obj))
    return(vector_to_c_string(sc, obj, to_file, ci));

  if (is_pair(obj))
    return(list_to_c_string(sc, obj, ci));

  if (s7_is_hash_table(obj))
    return(hash_table_to_c_string(sc, obj, to_file, ci));

  /* it might be nice to print environments here as slots, but circle checks become a nightmare:
   *   (let ((lst (list 1 2 3))) (list-set! lst 1 (current-environment)) lst) 
   *   if we add T_ENVIRONMENT to has_structure, we need to extend all the circle check code.
   */

  return(atom_to_c_string(sc, obj, use_write));
}


static char *object_to_c_string_with_circle_check(s7_scheme *sc, s7_pointer vr, bool use_write, bool to_file, shared_info *ci)
{
  if (ci)
    {
      int ref;
      ref = shared_ref(ci, vr);
      if (ref != 0)
	{
	  char *name;
	  if (ref > 0)
	    {
	      char *element;
	      element = s7_object_to_c_string_1(sc, vr, USE_WRITE, WITH_ELLIPSES, ci);
	      name = (char *)calloc(strlen(element) + 32, sizeof(char));
	      sprintf(name, "#%d=%s", ref, element);
	      free(element);
	      return(name);
	    }
	  else
	    {
	      name = (char *)calloc(32, sizeof(char));
	      snprintf(name, 32, "#%d#", -ref);
	      return(name);
	    }
	}
    }
  return(s7_object_to_c_string_1(sc, vr, use_write, to_file, ci));
}


static char *vector_to_c_string(s7_scheme *sc, s7_pointer vect, bool to_file, shared_info *ci)
{
  s7_Int i, len, bufsize = 0;
  bool too_long = false;
  char **elements = NULL;
  char *buf;
  
  len = vector_length(vect);
  if (len == 0)
    {
      if (vector_is_multidimensional(vect))
	{
	  buf = (char *)calloc(16, sizeof(char));
	  snprintf(buf, 16, "#%dD()", vector_ndims(vect));
	  return(buf);
	}
      else return(copy_string("#()"));
    }
  
  if (!to_file)
    {
      int plen;
      /* if to_file we ignore *vector-print-length* so a subsequent read will be ok
       *
       * (with-output-to-file "test.test" 
       *   (lambda () 
       *     (let ((vect (make-vector (+ *vector-print-length* 2) 1.0))) 
       *       (write vect))))
       */

      plen = s7_vector_print_length(sc);
      if (plen <= 0)
	return(copy_string("#(...)"));

      if (len > plen)
	{
	  too_long = true;
	  len = plen;
	}
    }

  elements = (char **)calloc(len, sizeof(char *));
  for (i = 0; i < len; i++)
    {
      elements[i] = object_to_c_string_with_circle_check(sc, vector_element(vect, i), USE_WRITE, WITH_ELLIPSES, ci);
      bufsize += safe_strlen(elements[i]);
      
      if (bufsize > sc->print_width) 
	{
	  if ((i + 1) < len) too_long = true;
	  len = i + 1;
	  break;
	}
    }

  if (vector_is_multidimensional(vect))
    {
      char c;

      bufsize += (len * 4 * vector_ndims(vect) + 16);
      buf = (char *)malloc(bufsize * sizeof(char));

      c = '#';
      if (vector_ndims(vect) > 1)
	snprintf(buf, bufsize, "#%dD", vector_ndims(vect));
      else snprintf(buf, bufsize, "#");

      display_multivector(sc, vect, len, 0, 0, vector_ndims(vect), buf, elements, &c);

      for (i = 0; i < len; i++)
	if (elements[i]) free(elements[i]);
      free(elements);
      return(buf);
    }

  bufsize += (len * 4 + 16);                   /* might be 2 parens per element + space, so at least len*4 here */
  buf = (char *)malloc(bufsize * sizeof(char));

  sprintf(buf, "#(");
  for (i = 0; i < len - 1; i++)
    {
      if (elements[i])
	{
	  strcat(buf, elements[i]);
	  free(elements[i]);
	  strcat(buf, " ");
	}
    }

  if (elements[len - 1])
    {
      strcat(buf, elements[len - 1]);
      free(elements[len - 1]);
    }

  free(elements);
  if (too_long)
    strcat(buf, " ...");
  strcat(buf, ")");

  return(buf);
}


static s7_pointer structure_to_string(s7_scheme *sc, s7_pointer vect)
{
  return(make_string_uncopied(sc, object_to_c_string_with_circle_check(sc, vect, USE_WRITE, WITH_ELLIPSES, make_shared_info(sc, vect))));
}


static int circular_list_entries(s7_pointer lst)
{
  int i;
  s7_pointer x;
  for (i = 1, x = cdr(lst); ; i++, x = cdr(x))
    {
      int j;
      s7_pointer y;
      for (y = lst, j = 0; j < i; y = cdr(y), j++)
	if (x == y)
	  return(i + 1);
    }
}


static char *list_to_c_string(s7_scheme *sc, s7_pointer lst, shared_info *ci)
{
  s7_pointer x;
  int i, len, true_len, bufsize = 0, start = 0;
  char **elements = NULL;
  char *buf;

  true_len = s7_list_length(sc, lst);
  if (true_len < 0)                    /* a dotted list -- handle cars, then final cdr */
    len = (-true_len + 1);
  else
    {
      if (true_len == 0)               /* either '() or a circular list */
	{
	  if (is_not_null(lst))
	    len = circular_list_entries(lst);
	  else return(copy_string("()"));
	}
      else len = true_len;
    }

  elements = (char **)calloc(len, sizeof(char *));

  for (x = lst, i = 0; (is_not_null(x)) && (i < len); i++, x = cdr(x))
    {
      if (is_pair(x))
	{
	  if ((ci) && (i != 0) && (peek_shared_ref(ci, x) != 0))
	    {
	      elements[i] = object_to_c_string_with_circle_check(sc, x, USE_WRITE, WITH_ELLIPSES, ci);
	      len = i + 1;
	      bufsize += safe_strlen(elements[i]);
	      break;
	    }
	  else elements[i] = object_to_c_string_with_circle_check(sc, car(x), USE_WRITE, WITH_ELLIPSES, ci);
	}
      else 
	{
	  elements[i] = object_to_c_string_with_circle_check(sc, x, USE_WRITE, WITH_ELLIPSES, ci);
	  len = i + 1;
	  bufsize += safe_strlen(elements[i]);
	  break;
	}
      bufsize += safe_strlen(elements[i]);

      if (bufsize > sc->print_width) 
	{
	  len = i + 1;
	  x = sc->NIL;
	  break;
	}
    }
  
  bufsize += (16 + len * 2); /* len spaces; was 256 rather than 16, but the extra space is not needed */
  if (ci) bufsize += (ci->top * 16);
  buf = (char *)malloc(bufsize * sizeof(char));

  if (((car(lst) == sc->QUOTE) || (car(lst) == sc->QUOTE_UNCHECKED)) &&
      (true_len == 2))                    
    {
      /* len == 1 is important, otherwise (list 'quote 1 2) -> '1 2 which looks weird 
       *   or (object->string (apply . `''1)) -> "'quote 1"
       * so (quote x) = 'x but (quote x y z) should be left alone (if evaluated, it's an error)
       *
       * in CL:
       *    [2]> (list 'quote 1 2)
       *    (QUOTE 1 2)
       *    [3]> (list 'quote 1)
       *    '1
       *    [4]> (cons 'quote 1)
       *    (QUOTE . 1)
       *
       * in s7:
       *    :(list 'quote 1 2)
       *    (quote 1 2)
       *    :(list 'quote 1)
       *    '1
       *    :(cons 'quote 1)
       *    (quote . 1)
       */
      sprintf(buf, "'");
      start = 1;
    }
  else sprintf(buf, "(");
  if (is_multiple_value(lst))
    strcat(buf, "values ");

  for (i = start; i < len - 1; i++)
    {
      if (elements[i])
	{
	  strcat(buf, elements[i]);
	  strcat(buf, " ");
	}
    }

  if (is_not_null(x))
    strcat(buf, ". ");

  if (elements[len - 1])
    {
      strcat(buf, elements[len - 1]);
      if (((car(lst) != sc->QUOTE) && (car(lst) != sc->QUOTE_UNCHECKED)) ||
	  (true_len != 2))
	strcat(buf, ")");
    }

  for (i = 0; i < len; i++)
    if (elements[i])
      free(elements[i]);
  free(elements);

  return(buf);
}


char *s7_object_to_c_string(s7_scheme *sc, s7_pointer obj)
{
  shared_info *ci = NULL;
  if (has_structure(obj))
    ci = make_shared_info(sc, obj);
  return(object_to_c_string_with_circle_check(sc, obj, USE_WRITE, WITH_ELLIPSES, ci));
}


s7_pointer s7_object_to_string(s7_scheme *sc, s7_pointer obj, bool use_write)
{
  char *str;
  if ((s7_is_vector(obj)) ||
      (is_pair(obj)) ||
      (s7_is_hash_table(obj)))
    return(structure_to_string(sc, obj));

  str = atom_to_c_string(sc, obj, use_write);
  if (str)
    return(make_string_uncopied(sc, str));

  return(s7_make_string_with_length(sc, "", 0)); 
  /* else segfault in (string->symbol (object->string "" #f))
   *   this can't be optimized to make_string_uncopied -- gc trouble (attempt to free unallocated pointer)
   */
}


void s7_newline(s7_scheme *sc, s7_pointer port)
{
  s7_write_char(sc, '\n', port);
}


static s7_pointer g_newline(s7_scheme *sc, s7_pointer args)
{
  #define H_newline "(newline (port (current-output-port))) writes a carriage return to the port"
  s7_pointer port;
  
  if (is_not_null(args))
    {
      port = car(args);
      if (!is_output_port(port))
	return(s7_wrong_type_arg_error(sc, "newline", 0, port, "an output port"));
    }
  else port = sc->output_port;
  
  s7_newline(sc, port);
  return(sc->UNSPECIFIED);
}


static s7_pointer g_write_char(s7_scheme *sc, s7_pointer args)
{
  #define H_write_char "(write-char char (port (current-output-port))) writes char to the output port"
  s7_pointer port, chr;
  
  chr = car(args);
  if (!s7_is_character(chr))
    return(s7_wrong_type_arg_error(sc, "write-char", (is_null(cdr(args))) ? 0 : 1, chr, "a character"));
  
  if (is_pair(cdr(args)))
    {
      port = cadr(args);
      if (!is_output_port(port))
	return(s7_wrong_type_arg_error(sc, "write-char port", 2, port, "an output port"));
    }
  else port = sc->output_port;

  s7_write_char(sc, s7_character(chr), port);
  return(sc->UNSPECIFIED);
}


static void write_or_display(s7_scheme *sc, s7_pointer obj, s7_pointer port, bool use_write)
{
  char *val;
  shared_info *ci = NULL;
  if (has_structure(obj))
    ci = make_shared_info(sc, obj);
  val = object_to_c_string_with_circle_check(sc, obj, use_write, is_file_port(port), ci);
  port_display(port)(sc, val, port);
  if (val) free(val);
}


void s7_write(s7_scheme *sc, s7_pointer obj, s7_pointer port)
{
  write_or_display(sc, obj, port, USE_WRITE);
}


static s7_pointer g_write(s7_scheme *sc, s7_pointer args)
{
  #define H_write "(write str (port (current-output-port))) writes str (a string) to the output port"
  s7_pointer port;
  
  if (is_pair(cdr(args)))
    {
      port = cadr(args);
      if (!is_output_port(port))
	return(s7_wrong_type_arg_error(sc, "write port", 2, port, "an output port"));
    }
  else port = sc->output_port;
  write_or_display(sc, car(args), port, USE_WRITE);
  return(sc->UNSPECIFIED);
}


void s7_display(s7_scheme *sc, s7_pointer obj, s7_pointer port)
{
  write_or_display(sc, obj, port, USE_DISPLAY);
}


static s7_pointer g_display(s7_scheme *sc, s7_pointer args)
{
  #define H_display "(display str (port (current-output-port))) writes str (a string) to the output port"
  s7_pointer port;
  
  if (is_pair(cdr(args)))
    {
      port = cadr(args);
      if (!is_output_port(port))
	return(s7_wrong_type_arg_error(sc, "display port", 2, port, "an output port"));
    }
  else port = sc->output_port;
  write_or_display(sc, car(args), port, USE_DISPLAY);
  return(sc->UNSPECIFIED); /* wouldn't it be more useful to return car(args) here and in write? */
}


static s7_pointer g_read_byte(s7_scheme *sc, s7_pointer args)
{
  #define H_read_byte "(read-byte (port (current-input-port))): reads a byte from the input port"
  s7_pointer port;
  int c;
  
  if (is_not_null(args))
    port = car(args);
  else port = sc->input_port;

  if (!is_input_port(port))
    return(s7_wrong_type_arg_error(sc, "read-byte", 0, port, "an input port"));

  c = port_read_character(port)(sc, port);
  if (c == EOF)
    return(sc->EOF_OBJECT);
  return(small_int(c));
}


static s7_pointer g_write_byte(s7_scheme *sc, s7_pointer args)
{
  #define H_write_byte "(write-byte byte (port (current-output-port))): writes byte to the output port"
  s7_pointer port;
  
  if (!s7_is_integer(car(args)))
    return(s7_wrong_type_arg_error(sc, "write-byte", (is_null(cdr(args))) ? 0 : 1, car(args), "an integer"));
  
  if (is_pair(cdr(args)))
    port = cadr(args);
  else port = sc->output_port;

  if ((!is_output_port(port)) ||
      (is_string_port(port)))
    return(s7_wrong_type_arg_error(sc, "write-byte port", 2, port, "an output file or function port"));
  if (port_is_closed(port))
    return(s7_wrong_type_arg_error(sc, "write-byte port", 2, port, "an open output port"));

  s7_write_char(sc, (int)(s7_integer(car(args))), port);
#if 0
  /* this segfaults if port is *stdout* */
  if (is_file_port(port))
    file_write_char(sc, (unsigned char)s7_integer(car(args)), port);
  else (*(port_output_function(port)))(sc, (char)s7_integer(car(args)), port);
#endif
  return(car(args));
}


static s7_pointer g_call_with_output_string(s7_scheme *sc, s7_pointer args)
{
  #define H_call_with_output_string "(call-with-output-string proc) opens a string port applies proc to it, then returns the collected output"
  s7_pointer port;

  if (!is_procedure(car(args)))
    return(s7_wrong_type_arg_error(sc, "call-with-output-string", 1, car(args), "a procedure"));
  if ((is_continuation(car(args))) || is_goto(car(args)))
    return(s7_wrong_type_arg_error(sc, "call-with-output-string", 2, car(args), "a normal procedure (not a continuation)"));
  if (is_thunk(sc, car(args)))
    return(s7_wrong_type_arg_error(sc, "call-with-output-string", 2, car(args), "a procedure of one argument (the port)"));
  
  port = s7_open_output_string(sc);
  push_stack(sc, OP_UNWIND_OUTPUT, sc->F, port);
  push_stack(sc, OP_GET_OUTPUT_STRING, sc->F, port);
  push_stack(sc, OP_APPLY, list_1(sc, port), car(args));
  return(sc->F);
}


static s7_pointer g_call_with_output_file(s7_scheme *sc, s7_pointer args)
{
  #define H_call_with_output_file "(call-with-output-file filename proc) opens filename and calls proc with the output port as its argument"
  s7_pointer port;
  
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "call-with-output-file filename,", 1, car(args), "a string"));
  if (!is_procedure(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "call-with-output-file", 2, cadr(args), "a procedure"));
  if ((is_continuation(cadr(args))) || is_goto(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "call-with-output-file", 2, cadr(args), "a normal procedure (not a continuation)"));
  if (is_thunk(sc, car(args)))
    return(s7_wrong_type_arg_error(sc, "call-with-output-file", 2, car(args), "a procedure of one argument (the port)"));
  
  port = s7_open_output_file(sc, s7_string(car(args)), "w");
  push_stack(sc, OP_UNWIND_OUTPUT, sc->F, port);
  push_stack(sc, OP_APPLY, list_1(sc, port), cadr(args));
  return(sc->F);
}


static s7_pointer g_with_output_to_string(s7_scheme *sc, s7_pointer args)
{
  #define H_with_output_to_string "(with-output-to-string thunk) opens a string as a temporary current-output-port, calls thunk, then returns the collected output"
  s7_pointer old_output_port;

  if (!is_thunk(sc, car(args)))
    return(s7_wrong_type_arg_error(sc, "with-output-to-string", 1, car(args), "a thunk"));
  
  old_output_port = sc->output_port;
  sc->output_port = s7_open_output_string(sc);
  push_stack(sc, OP_UNWIND_OUTPUT, old_output_port, sc->output_port);
  push_stack(sc, OP_GET_OUTPUT_STRING, sc->F, sc->output_port);
  push_stack(sc, OP_APPLY, sc->NIL, car(args));
  return(sc->F);
}

/* (string-ref (with-output-to-string (lambda () (write "1234") (values (get-output-string) 1)))) */


static s7_pointer g_with_output_to_file(s7_scheme *sc, s7_pointer args)
{
  #define H_with_output_to_file "(with-output-to-file filename thunk) opens filename as the temporary current-output-port and calls thunk"
  s7_pointer old_output_port;
  
  if (!s7_is_string(car(args)))
    return(s7_wrong_type_arg_error(sc, "with-output-to-file filename,", 1, car(args), "a string"));
  if (!is_thunk(sc, cadr(args)))
    return(s7_wrong_type_arg_error(sc, "with-output-to-file", 2, cadr(args), "a thunk"));
  
  old_output_port = sc->output_port;
  sc->output_port = s7_open_output_file(sc, s7_string(car(args)), "w");
  push_stack(sc, OP_UNWIND_OUTPUT, old_output_port, sc->output_port);
  push_stack(sc, OP_APPLY, sc->NIL, cadr(args));
  return(sc->F);
}



/* -------------------------------- lists -------------------------------- */

s7_pointer s7_cons(s7_scheme *sc, s7_pointer a, s7_pointer b) 
{
  s7_pointer x;
  NEW_CELL(sc, x);
  car(x) = a;
  cdr(x) = b;
  set_type(x, T_PAIR | T_SAFE_PROCEDURE);
  return(x);
}


static s7_pointer cons_unchecked(s7_scheme *sc, s7_pointer a, s7_pointer b) 
{
  /* apparently slightly faster as a function? */
  s7_pointer x;
  NEW_CELL_NO_CHECK(sc, x);
  car(x) = a;
  cdr(x) = b;
  set_type(x, T_PAIR | T_SAFE_PROCEDURE);
  return(x);
}


static s7_pointer permanent_cons(s7_pointer a, s7_pointer b, int type)
{
  /* for the symbol table which is never GC'd (and its contents aren't marked) */
  s7_pointer x;
  x = (s7_cell *)permanent_calloc(sizeof(s7_cell));
  x->hloc = NOT_IN_HEAP;
  car(x) = a;
  cdr(x) = b;
  set_type(x, type);
  return(x);
}


bool s7_is_pair(s7_pointer p) 
{
  return(is_pair(p));
}


s7_pointer s7_car(s7_pointer p) {return(car(p));}
s7_pointer s7_cdr(s7_pointer p) {return(cdr(p));}

s7_pointer s7_cadr(s7_pointer p) {return(cadr(p));}
s7_pointer s7_cddr(s7_pointer p) {return(cddr(p));}
s7_pointer s7_cdar(s7_pointer p) {return(cdar(p));}
s7_pointer s7_caar(s7_pointer p) {return(caar(p));}

s7_pointer s7_caadr(s7_pointer p) {return(caadr(p));}
s7_pointer s7_caddr(s7_pointer p) {return(caddr(p));}
s7_pointer s7_cadar(s7_pointer p) {return(cadar(p));}
s7_pointer s7_caaar(s7_pointer p) {return(caaar(p));}
s7_pointer s7_cdadr(s7_pointer p) {return(cdadr(p));}
s7_pointer s7_cdddr(s7_pointer p) {return(cdddr(p));}
s7_pointer s7_cddar(s7_pointer p) {return(cddar(p));}
s7_pointer s7_cdaar(s7_pointer p) {return(cdaar(p));}

s7_pointer s7_caaadr(s7_pointer p) {return(caaadr(p));}
s7_pointer s7_caaddr(s7_pointer p) {return(caaddr(p));}
s7_pointer s7_caadar(s7_pointer p) {return(caadar(p));}
s7_pointer s7_caaaar(s7_pointer p) {return(caaaar(p));}
s7_pointer s7_cadadr(s7_pointer p) {return(cadadr(p));}
s7_pointer s7_cadddr(s7_pointer p) {return(cadddr(p));}
s7_pointer s7_caddar(s7_pointer p) {return(caddar(p));}
s7_pointer s7_cadaar(s7_pointer p) {return(cadaar(p));}

s7_pointer s7_cdaadr(s7_pointer p) {return(cdaadr(p));}
s7_pointer s7_cdaddr(s7_pointer p) {return(cdaddr(p));}
s7_pointer s7_cdadar(s7_pointer p) {return(cdadar(p));}
s7_pointer s7_cdaaar(s7_pointer p) {return(cdaaar(p));}
s7_pointer s7_cddadr(s7_pointer p) {return(cddadr(p));}
s7_pointer s7_cddddr(s7_pointer p) {return(cddddr(p));}
s7_pointer s7_cdddar(s7_pointer p) {return(cdddar(p));}
s7_pointer s7_cddaar(s7_pointer p) {return(cddaar(p));}


s7_pointer s7_set_car(s7_pointer p, s7_pointer q) 
{ 
  car(p) = q;
  return(p);
}


s7_pointer s7_set_cdr(s7_pointer p, s7_pointer q) 
{ 
  cdr(p) = q;
  return(p);
}


s7_pointer s7_list_ref(s7_scheme *sc, s7_pointer lst, int num)
{
  int i;
  s7_pointer x;

  for (x = lst, i = 0; (i < num) && (is_pair(x)); i++, x = cdr(x)) {}
  if ((i == num) && (is_pair(x)))
    return(car(x));

  return(sc->NIL);
}


s7_pointer s7_list_set(s7_scheme *sc, s7_pointer lst, int num, s7_pointer val)
{
  int i;
  s7_pointer x;

  for (x = lst, i = 0; (i < num) && (is_pair(x)); i++, x = cdr(x)) {}
  if ((i == num) &&
      (is_pair(x)))
    car(x) = val;

  return(val);
}


static void list_set(s7_scheme *sc, s7_pointer lst, int num, s7_pointer val)
{
  /* internal form where we know the list is ok */
  int i;
  s7_pointer x;
  for (x = lst, i = 0; i < num; i++, x = cdr(x)) {}
  car(x) = val;
}


s7_pointer s7_member(s7_scheme *sc, s7_pointer sym, s7_pointer lst)
{
  s7_pointer x;

  for (x = lst; is_pair(x); x = cdr(x))
    if (s7_is_equal(sc, sym, car(x)))
      return(x);

  return(sc->F);
}


static bool symbol_is_in_arg_list(s7_pointer sym, s7_pointer lst)
{
  s7_pointer x;
  for (x = lst; is_pair(x); x = cdr(x))
    if ((sym == car(x)) ||
	((is_pair(car(x))) &&
	 (sym == caar(x))))
	return(true);
  return(sym == x);
}


s7_pointer s7_assoc(s7_scheme *sc, s7_pointer sym, s7_pointer lst)
{
  s7_pointer x, y;

  if (!is_pair(lst))
    return(sc->F);

  x = lst;
  y = lst;
  while (true)
    {
      if ((is_pair(car(x))) && (s7_is_equal(sc, sym, caar(x)))) return(car(x));
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
      
      if ((is_pair(car(x))) && (s7_is_equal(sc, sym, caar(x)))) return(car(x));
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      y = cdr(y);
      if (x == y) return(sc->F);
    }
  return(sc->F);
}


s7_pointer s7_reverse(s7_scheme *sc, s7_pointer a) 
{
  /* reverse list -- produce new list (other code assumes this function does not return the original!) */
  s7_pointer x, p;

  if (is_null(a)) return(a);

  if (!is_pair(cdr(a)))
    {
      if (is_not_null(cdr(a)))
	return(cons(sc, cdr(a), car(a)));
      return(cons(sc, car(a), sc->NIL)); /* don't return 'a' itself */
    }

  sc->w = list_1(sc, car(a));

  for (x = cdr(a), p = a; is_pair(x); x = cdr(x), p = cdr(p))
    {
      sc->w = cons(sc, car(x), sc->w);
      if (is_pair(cdr(x)))
	{
	  x = cdr(x);
	  sc->w = cons(sc, car(x), sc->w);
	}
      if (x == p) /* this can take awhile to notice there's a cycle, but what does the caller expect? */
	break;
    }

  if (is_not_null(x))
    p = cons(sc, x, sc->w);    /* ?? this means that (reverse '(1 2 . 3)) returns '(3 2 1) -- we used to return '() here */
  else p = sc->w;

  sc->w = sc->NIL;
  return(p);
}

/* s7_reverse sometimes tacks extra nodes on the end of a reversed circular list (it detects the cycle too late) 
 *  (let ((lst (list 0))) (set! (cdr lst) lst) (reverse lst)) -> (#1=(0 . #1#) 0 0 0)
 */


static s7_pointer reverse_in_place(s7_scheme *sc, s7_pointer term, s7_pointer list) 
{
  s7_pointer p = list, result = term, q;

  while (is_not_null(p))
    {
      q = cdr(p);
      if ((!is_pair(q)) &&
	  (is_not_null(q)))
	return(sc->NIL); /* improper list? */
      cdr(p) = result;
      result = p;
      p = q;
    }

  return(result);
}


static s7_pointer safe_reverse_in_place(s7_scheme *sc, s7_pointer list) /* "safe" here means we guarantee this list is unproblematic */
{
  s7_pointer p = list, result, q;
  result = sc->NIL;

  while (is_not_null(p))
    {
      q = cdr(p);
      /*   also if (is_null(list)) || (is_null(cdr(list))) return(list) */
      cdr(p) = result;
      result = p;
      p = q;

      /* unroll the loop for speed */
      if (is_null(p)) break;
      q = cdr(p);
      cdr(p) = result;
      result = p;
      p = q;

      if (is_null(p)) break;
      q = cdr(p);
      cdr(p) = result;
      result = p;
      p = q;

      if (is_null(p)) break;
      q = cdr(p);
      cdr(p) = result;
      result = p;
      p = q;
    }

  return(result);
}


/* is this correct? (let ((x (list 1 2))) (eq? x (append '() x))) -> #t
 */

s7_pointer s7_append(s7_scheme *sc, s7_pointer a, s7_pointer b) 
{
  s7_pointer p = b, q;
  
  if (is_not_null(a)) 
    {
      a = s7_reverse(sc, a);
      while (is_not_null(a)) 
	{
	  q = cdr(a);
	  cdr(a) = p;
	  p = a;
	  a = q;
	}
    }
  return(p);
}


static s7_pointer copy_list(s7_scheme *sc, s7_pointer lst)
{
  /* the non-recursive version of this was only slightly faster */
  if (is_null(lst))
    return(sc->NIL);
  return(cons(sc, car(lst), copy_list(sc, cdr(lst))));
}


static s7_pointer revappend(s7_scheme *sc, s7_pointer a, s7_pointer b) 
{
  /* (map (lambda (x) (if (odd? x) (apply values '(1 2 3)) (values))) (list 1 2 3 4)) 
   *   is a bad case -- we have to copy the incoming list.
   */
   
  s7_pointer p = b, q;
  
  if (is_not_null(a)) 
    {
      a = copy_list(sc, a); 
      while (is_not_null(a)) 
	{
	  q = cdr(a);
	  cdr(a) = p;
	  p = a;
	  a = q;
	}
    }
  return(p);
}

static int safe_list_length(s7_scheme *sc, s7_pointer a)
{
  /* assume that "a" is a proper list */
  int i = 0;
  s7_pointer b;
  for (b = a; is_pair(b); i++, b = cdr(b)) {};
  return(i);
}


int s7_list_length(s7_scheme *sc, s7_pointer a) 
{
  /* returns -len if list is dotted, 0 if it's (directly) circular */
  int i;
  s7_pointer slow, fast;

  slow = fast = a;
  for (i = 0; ; i += 2)
    {
      if (!is_pair(fast))
	{
	  if (is_null(fast))
	    return(i);
	  return(-i);
	}
      
      fast = cdr(fast);
      if (!is_pair(fast)) 
	{
	  if (is_null(fast))
	    return(i + 1);
	  return(-i - 1);
	}
      /* if unrolled further, it's a lot slower? */
      
      fast = cdr(fast);
      slow = cdr(slow);
      if (fast == slow) 
	return(0);
    }
  return(0);
}


static s7_pointer g_is_null(s7_scheme *sc, s7_pointer args)
{
  #define H_is_null "(null? obj) returns #t if obj is the empty list"
  return(make_boolean(sc, is_null(car(args))));

  /* as a generic this could be: has_structure and length == 0 */
}


static s7_pointer g_is_pair(s7_scheme *sc, s7_pointer args)
{
  #define H_is_pair "(pair? obj) returns #t if obj is a pair (a non-empty list)"
  return(make_boolean(sc, is_pair(car(args))));
}


bool s7_is_list(s7_scheme *sc, s7_pointer p)
{
  return((is_pair(p)) ||
	 (is_null(p)));
}


static bool is_proper_list(s7_scheme *sc, s7_pointer lst)
{
  s7_pointer slow, fast;

  fast = lst;
  slow = lst;
  while (true)
    {
      if (!is_pair(fast)) 
	return(is_null(fast)); /* else it's an improper list */

      fast = cdr(fast);
      if (!is_pair(fast)) return(is_null(fast));

      fast = cdr(fast);
      if (!is_pair(fast)) return(is_null(fast));

      fast = cdr(fast);
      slow = cdr(slow);
      if (fast == slow) return(false);
    }
  return(true);
}


static s7_pointer g_is_list(s7_scheme *sc, s7_pointer args)
{
  #define H_is_list "(list? obj) returns #t if obj is a proper list"
  return(make_boolean(sc, is_proper_list(sc, car(args))));
}


#define MAX_LIST_LENGTH 1073741824

static s7_pointer g_make_list(s7_scheme *sc, s7_pointer args)
{
  #define H_make_list "(make-list length (initial-element #f)) returns a list of 'length' elements whose value is 'initial-element'."

  s7_pointer init, p;
  int i, ilen;
  s7_Int len;

  if (!s7_is_integer(car(args)))
    return(s7_wrong_type_arg_error(sc, "make-list", (is_null(cdr(args))) ? 0 : 1, car(args), "an integer"));

  len = s7_integer(car(args));            /* needs to be s7_Int here so that (make-list most-negative-fixnum) is handled correctly */
  if (len < 0)
    return(s7_out_of_range_error(sc, "make-list length,", (is_null(cdr(args))) ? 0 : 1, car(args), "should be non-negative"));
  if (len == 0) return(sc->NIL);          /* what about (make-list 0 123)? */
  if (len > MAX_LIST_LENGTH)
    return(s7_out_of_range_error(sc, "make-list length,", (is_null(cdr(args))) ? 0 : 1, car(args), "should be a reasonable integer"));

  if (is_pair(cdr(args)))
    init = cadr(args);
  else init = sc->F;

  sc->w = sc->NIL;
  ilen = (int)len;

  if (len < (sc->free_heap_top - sc->free_heap))
    {
      sc->w = cons(sc, init, sc->NIL);
      for (i = 1; i < ilen; i++)
	sc->w = cons_unchecked(sc, init, sc->w);
    }
  else
    {
      for (i = 0; i < ilen; i++)
	sc->w = cons(sc, init, sc->w);
    }
  p = sc->w;
  sc->w = sc->NIL;

  return(p);
}


static s7_pointer list_ref_1(s7_scheme *sc, s7_pointer lst, s7_pointer ind)
{
  s7_Int i, index;
  s7_pointer p;

  if (!s7_is_integer(ind))
    return(s7_wrong_type_arg_error(sc, "list-ref index,", 2, ind, "an integer"));
  
  index = s7_integer(ind);
  if (index < 0)
    return(s7_out_of_range_error(sc, "list-ref index,", 2, ind, "should be non-negative"));

  if (index > MAX_LIST_LENGTH)
    return(s7_out_of_range_error(sc, "list-ref index,", 2, ind, "should be a reasonable integer"));
  
  for (i = 0, p = lst; (i < index) && is_pair(p); i++, p = cdr(p)) {}
  
  if (!is_pair(p))
    {
      if (is_null(p))
	return(s7_out_of_range_error(sc, "list-ref index,", 2, ind, "should be less than list length"));
      return(s7_wrong_type_arg_error(sc, "list-ref", 1, lst, "a proper list"));
    }
  return(car(p));
}


static s7_pointer g_list_ref(s7_scheme *sc, s7_pointer args)
{
  #define H_list_ref "(list-ref lst i ...) returns the i-th element (0-based) of the list"
  
  /* (let ((L '((1 2 3) (4 5 6)))) (list-ref L 1 2)) 

    (define (lref L . args) 
      (if (null? (cdr args))
          (list-ref L (car args))
          (apply lref (list-ref L (car args)) (cdr args))))
  */

  s7_pointer lst, inds;
  lst = car(args);

  inds = cdr(args);
  while (true)
    {
      if (!is_pair(lst))
	return(s7_wrong_type_arg_error(sc, "list-ref", 1, lst, "a pair"));

      if (is_null(cdr(inds)))
	return(list_ref_1(sc, lst, car(inds)));
      lst = list_ref_1(sc, lst, car(inds));
      inds = cdr(inds);
    }
}


static s7_pointer g_list_set_1(s7_scheme *sc, s7_pointer lst, s7_pointer args, int arg_num)
{
  #define H_list_set "(list-set! lst i ... val) sets the i-th element (0-based) of the list to val"
  
  int i;
  s7_Int index;
  s7_pointer p, ind;

  /* (let ((L '((1 2 3) (4 5 6)))) (list-set! L 1 2 32) L) */

  if (!is_pair(lst))
    return(s7_wrong_type_arg_error(sc, "list-set!", 1, lst, "a pair"));

  ind = car(args);
  if (!s7_is_integer(ind))
    return(s7_wrong_type_arg_error(sc, "list-set! index,", arg_num, ind, "an integer"));
  
  index = s7_integer(ind);
  if (index < 0)
    return(s7_out_of_range_error(sc, "list-set!", arg_num, ind, "index should be non-negative"));

  if (index > MAX_LIST_LENGTH)
    return(s7_out_of_range_error(sc, "list-set! index,", arg_num, ind, "should be a reasonable integer"));
  
  for (i = 0, p = lst; (i < index) && is_pair(p); i++, p = cdr(p)) {}
  
  if (!is_pair(p))
    {
      if (is_null(p))
	return(s7_out_of_range_error(sc, "list-set! index,", arg_num, ind, "should be less than list length"));
      return(s7_wrong_type_arg_error(sc, "list-set!", 1, lst, "a proper list"));
    }
  if (is_null(cddr(args)))
    car(p) = cadr(args);
  else return(g_list_set_1(sc, car(p), cdr(args), arg_num + 1));
  return(cadr(args));
}


static s7_pointer g_list_set(s7_scheme *sc, s7_pointer args)
{
  return(g_list_set_1(sc, car(args), cdr(args), 2));
}


#if WITH_OPTIMIZATION
static s7_pointer list_set_ic;
static s7_pointer g_list_set_ic(s7_scheme *sc, s7_pointer args)
{
  int i;
  s7_Int index;
  s7_pointer p, lst;

  lst = car(args);
  if (!is_pair(lst))
    return(s7_wrong_type_arg_error(sc, "list-set!", 1, lst, "a pair"));

  index = s7_integer(cadr(args));
  for (i = 0, p = lst; (i < index) && is_pair(p); i++, p = cdr(p)) {}
  
  if (!is_pair(p))
    {
      if (is_null(p))
	return(s7_out_of_range_error(sc, "list-set! index,", 2, cadr(args), "should be less than list length"));
      return(s7_wrong_type_arg_error(sc, "list-set!", 1, lst, "a proper list"));
    }

  car(p) = caddr(args);
  return(car(p));
}
#endif


static s7_pointer g_list_tail(s7_scheme *sc, s7_pointer args)
{
  #define H_list_tail "(list-tail lst i) returns the list from the i-th element on"
  
  int i;
  s7_Int index;
  s7_pointer p;

  if (!s7_is_list(sc, car(args)))
    return(s7_wrong_type_arg_error(sc, "list-tail", 1, car(args), "a list"));
  if (!s7_is_integer(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "list-tail", 2, cadr(args), "an integer"));
  
  index = s7_integer(cadr(args));
  if (index < 0)
    return(s7_out_of_range_error(sc, "list-tail index,", 2, cadr(args), "should be non-negative"));
  if (index > MAX_LIST_LENGTH)
    return(s7_out_of_range_error(sc, "list-tail index,", 2, cadr(args), "should be a reasonable integer"));
  
  for (i = 0, p = car(args); (i < index) && (is_pair(p)); i++, p = cdr(p)) {}
  
  if (i < index)
    return(s7_out_of_range_error(sc, "list-tail", 2, cadr(args), "index should be less than list length"));

  /* I guess this would make sense with more than one index, but I'm not sure it's very important */
  
  return(p);
}


static s7_pointer g_car(s7_scheme *sc, s7_pointer args)
{
  #define H_car "(car pair) returns the first element of the pair"
  if (!is_pair(car(args))) return(s7_wrong_type_arg_error(sc, "car", 0, car(args), "a pair"));
  
  return(caar(args));
}


static s7_pointer g_cdr(s7_scheme *sc, s7_pointer args)
{
  #define H_cdr "(cdr pair) returns the second element of the pair"
  if (!is_pair(car(args))) return(s7_wrong_type_arg_error(sc, "cdr", 0, car(args), "a pair"));
  
  return(cdar(args));
}


static s7_pointer g_cons(s7_scheme *sc, s7_pointer args)
{
  /* n-ary cons could be the equivalent of CL's list*? */
  /*   it would be neater to have a single cons cell able to contain (directly) any number of elements */
  /*   (set! (cadr (cons 1 2 3)) 4) -> (1 4 . 3) */

  #define H_cons "(cons a b) returns a pair containing a and b"
  
  /* cdr(args) = cadr(args);
   * this is not safe -- it changes a variable's value directly:
   *   (let ((lst (list 1 2))) (list (apply cons lst) lst)) -> '((1 . 2) (1 . 2))
   */
  s7_pointer x;

  NEW_CELL(sc, x);
  car(x) = car(args);
  cdr(x) = cadr(args);
  set_type(x, T_PAIR | T_SAFE_PROCEDURE);
  return(x);
}


static s7_pointer g_set_car(s7_scheme *sc, s7_pointer args)
{
  #define H_set_car "(set-car! pair val) sets the pair's first element to val"
  s7_pointer p;

  p = car(args);
  if (!is_pair(p))
    return(s7_wrong_type_arg_error(sc, "set-car!", 1, p, "a pair"));
  
  car(p) = cadr(args);
  return(cadr(args));
}


static s7_pointer g_set_cdr(s7_scheme *sc, s7_pointer args)
{
  #define H_set_cdr "(set-cdr! pair val) sets the pair's second element to val"
  s7_pointer p;

  p = car(args);
  if (!is_pair(p))
    return(s7_wrong_type_arg_error(sc, "set-cdr!", 1, p, "a pair"));
  
  cdr(p) = cadr(args);
  return(cadr(args));
}


static s7_pointer g_caar(s7_scheme *sc, s7_pointer args)
{
  #define H_caar "(caar lst) returns (car (car lst)): (caar '((1 2))) -> 1"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "caar", 0, lst, "a pair"));
  if (!is_pair(car(lst))) return(s7_wrong_type_arg_error(sc, "caar", 0, lst, "a list whose car is also a list"));

 return(car(car(lst)));
}

static s7_pointer g_cadr(s7_scheme *sc, s7_pointer args)
{
  #define H_cadr "(cadr lst) returns (car (cdr lst)): (cadr '(1 2 3)) -> 2"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "cadr", 0, lst, "a pair"));
  if (!is_pair(cdr(lst))) return(s7_wrong_type_arg_error(sc, "cadr", 0, lst, "a list whose cdr is also a list"));

  return(car(cdr(lst)));
}


static s7_pointer g_cdar(s7_scheme *sc, s7_pointer args)
{
  #define H_cdar "(cdar lst) returns (cdr (car lst)): (cdar '((1 2 3))) -> '(2 3)"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "cdar", 0, lst, "a pair"));
  if (!is_pair(car(lst))) return(s7_wrong_type_arg_error(sc, "cdar", 0, lst, "a list whose car is also a list"));

  return(cdr(car(lst)));
}


static s7_pointer g_cddr(s7_scheme *sc, s7_pointer args)
{
  #define H_cddr "(cddr lst) returns (cdr (cdr lst)): (cddr '(1 2 3 4)) -> '(3 4)"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "cddr", 0, lst, "a pair"));
  if (!is_pair(cdr(lst))) return(s7_wrong_type_arg_error(sc, "cddr", 0, lst, "a list whose cdr is also a list"));

  return(cdr(cdr(lst)));
}


static s7_pointer g_caaar(s7_scheme *sc, s7_pointer args)
{
  #define H_caaar "(caaar lst) returns (car (car (car lst))): (caaar '(((1 2)))) -> 1"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "caaar", 0, lst, "a pair"));
  if (!is_pair(car(lst))) return(s7_wrong_type_arg_error(sc, "caaar", 0, lst, "a list whose car is also a list"));
  if (!is_pair(car(car(lst)))) return(s7_wrong_type_arg_error(sc, "caaar", 0, lst, "a list whose caar is also a list"));

  return(car(car(car(lst))));
}


static s7_pointer g_caadr(s7_scheme *sc, s7_pointer args)
{
  #define H_caadr "(caadr lst) returns (car (car (cdr lst))): (caadr '(1 (2 3))) -> 2"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "caadr", 0, lst, "a pair"));
  if (!is_pair(cdr(lst))) return(s7_wrong_type_arg_error(sc, "caadr", 0, lst, "a list whose cdr is also a list"));
  if (!is_pair(car(cdr(lst)))) return(s7_wrong_type_arg_error(sc, "caadr", 0, lst, "a list whose cadr is also a list"));

  return(car(car(cdr(lst))));
}


static s7_pointer g_cadar(s7_scheme *sc, s7_pointer args)
{
  #define H_cadar "(cadar lst) returns (car (cdr (car lst))): (cadar '((1 2 3))) -> 2"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "cadar", 0, lst, "a pair"));
  if (!is_pair(car(lst))) return(s7_wrong_type_arg_error(sc, "cadar", 0, lst, "a list whose car is also a list"));
  if (!is_pair(cdr(car(lst)))) return(s7_wrong_type_arg_error(sc, "cadar", 0, lst, "a list whose cdar is also a list"));

  return(car(cdr(car(lst))));
}


static s7_pointer g_cdaar(s7_scheme *sc, s7_pointer args)
{
  #define H_cdaar "(cdaar lst) returns (cdr (car (car lst))): (cdaar '(((1 2 3)))) -> '(2 3)"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "cdaar", 0, lst, "a pair"));
  if (!is_pair(car(lst))) return(s7_wrong_type_arg_error(sc, "cdaar", 0, lst, "a list whose car is also a list"));
  if (!is_pair(car(car(lst)))) return(s7_wrong_type_arg_error(sc, "cdaar", 0, lst, "a list whose caar is also a list"));

  return(cdr(car(car(lst))));
}


static s7_pointer g_caddr(s7_scheme *sc, s7_pointer args)
{
  #define H_caddr "(caddr lst) returns (car (cdr (cdr lst))): (caddr '(1 2 3 4)) -> 3"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "caddr", 0, lst, "a pair"));
  if (!is_pair(cdr(lst))) return(s7_wrong_type_arg_error(sc, "caddr", 0, lst, "a list whose cdr is also a list"));
  if (!is_pair(cdr(cdr(lst)))) return(s7_wrong_type_arg_error(sc, "caddr", 0, lst, "a list whose cddr is also a list"));

  return(car(cdr(cdr(lst))));
}


static s7_pointer g_cdddr(s7_scheme *sc, s7_pointer args)
{
  #define H_cdddr "(cdddr lst) returns (cdr (cdr (cdr lst))): (cdddr '(1 2 3 4)) -> '(4)"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "cdddr", 0, lst, "a pair"));
  if (!is_pair(cdr(lst))) return(s7_wrong_type_arg_error(sc, "cdddr", 0, lst, "a list whose cdr is also a list"));
  if (!is_pair(cdr(cdr(lst)))) return(s7_wrong_type_arg_error(sc, "cdddr", 0, lst, "a list whose cddr is also a list"));

  return(cdr(cdr(cdr(lst))));
}


static s7_pointer g_cdadr(s7_scheme *sc, s7_pointer args)
{
  #define H_cdadr "(cdadr lst) returns (cdr (car (cdr lst))): (cdadr '(1 (2 3 4))) -> '(3 4)"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "cdadr", 0, lst, "a pair"));
  if (!is_pair(cdr(lst))) return(s7_wrong_type_arg_error(sc, "cdadr", 0, lst, "a list whose cdr is also a list"));
  if (!is_pair(car(cdr(lst)))) return(s7_wrong_type_arg_error(sc, "cdadr", 0, lst, "a list whose cadr is also a list"));

  return(cdr(car(cdr(lst))));
}


static s7_pointer g_cddar(s7_scheme *sc, s7_pointer args)
{
  #define H_cddar "(cddar lst) returns (cdr (cdr (car lst))): (cddar '((1 2 3 4))) -> '(3 4)"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "cddar", 0, lst, "a pair"));
  if (!is_pair(car(lst))) return(s7_wrong_type_arg_error(sc, "cddar", 0, lst, "a list whose car is also a list"));
  if (!is_pair(cdr(car(lst)))) return(s7_wrong_type_arg_error(sc, "cddar", 0, lst, "a list whose cdar is also a list"));

  return(cdr(cdr(car(lst))));
}


static s7_pointer g_caaaar(s7_scheme *sc, s7_pointer args)
{
  #define H_caaaar "(caaaar lst) returns (car (car (car (car lst)))): (caaaar '((((1 2))))) -> 1"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "caaaar", 0, lst, "a pair"));
  if (!is_pair(car(lst))) return(s7_wrong_type_arg_error(sc, "caaaar", 0, lst, "a list whose car is also a list"));
  if (!is_pair(caar(lst))) return(s7_wrong_type_arg_error(sc, "caaaar", 0, lst, "a list whose caar is also a list"));
  if (!is_pair(caaar(lst))) return(s7_wrong_type_arg_error(sc, "caaaar", 0, lst, "a list whose caaar is also a list"));

  return(car(car(car(car(lst)))));
}


static s7_pointer g_caaadr(s7_scheme *sc, s7_pointer args)
{
  #define H_caaadr "(caaadr lst) returns (car (car (car (cdr lst)))): (caaadr '(1 ((2 3)))) -> 2"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "caaadr", 0, lst, "a pair"));
  if (!is_pair(cdr(lst))) return(s7_wrong_type_arg_error(sc, "caaadr", 0, lst, "a list whose cdr is also a list"));
  if (!is_pair(cadr(lst))) return(s7_wrong_type_arg_error(sc, "caaadr", 0, lst, "a list whose cadr is also a list"));
  if (!is_pair(caadr(lst))) return(s7_wrong_type_arg_error(sc, "caaadr", 0, lst, "a list whose caadr is also a list"));

  return(car(car(car(cdr(lst)))));
}


static s7_pointer g_caadar(s7_scheme *sc, s7_pointer args)
{
  #define H_caadar "(caadar lst) returns (car (car (cdr (car lst)))): (caadar '((1 (2 3)))) -> 2"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "caadar", 0, lst, "a pair"));
  if (!is_pair(car(lst))) return(s7_wrong_type_arg_error(sc, "caadar", 0, lst, "a list whose car is also a list"));
  if (!is_pair(cdar(lst))) return(s7_wrong_type_arg_error(sc, "caadar", 0, lst, "a list whose cdar is also a list"));
  if (!is_pair(cadar(lst))) return(s7_wrong_type_arg_error(sc, "caadar", 0, lst, "a list whose cadar is also a list"));

  return(car(car(cdr(car(lst)))));
}


static s7_pointer g_cadaar(s7_scheme *sc, s7_pointer args)
{
  #define H_cadaar "(cadaar lst) returns (car (cdr (car (car lst)))): (cadaar '(((1 2 3)))) -> 2"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "cadaar", 0, lst, "a pair"));
  if (!is_pair(car(lst))) return(s7_wrong_type_arg_error(sc, "cadaar", 0, lst, "a list whose car is also a list"));
  if (!is_pair(caar(lst))) return(s7_wrong_type_arg_error(sc, "cadaar", 0, lst, "a list whose caar is also a list"));
  if (!is_pair(cdaar(lst))) return(s7_wrong_type_arg_error(sc, "cadaar", 0, lst, "a list whose cdaar is also a list"));

  return(car(cdr(car(car(lst)))));
}


static s7_pointer g_caaddr(s7_scheme *sc, s7_pointer args)
{
  #define H_caaddr "(caaddr lst) returns (car (car (cdr (cdr lst)))): (caaddr '(1 2 (3 4))) -> 3"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "caaddr", 0, lst, "a pair"));
  if (!is_pair(cdr(lst))) return(s7_wrong_type_arg_error(sc, "caaddr", 0, lst, "a list whose cdr is also a list"));
  if (!is_pair(cddr(lst))) return(s7_wrong_type_arg_error(sc, "caaddr", 0, lst, "a list whose cddr is also a list"));
  if (!is_pair(caddr(lst))) return(s7_wrong_type_arg_error(sc, "caaddr", 0, lst, "a list whose caddr is also a list"));

  return(car(car(cdr(cdr(lst)))));
}


static s7_pointer g_cadddr(s7_scheme *sc, s7_pointer args)
{
  #define H_cadddr "(cadddr lst) returns (car (cdr (cdr (cdr lst)))): (cadddr '(1 2 3 4 5)) -> 4"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "cadddr", 0, lst, "a pair"));
  if (!is_pair(cdr(lst))) return(s7_wrong_type_arg_error(sc, "cadddr", 0, lst, "a list whose cdr is also a list"));
  if (!is_pair(cddr(lst))) return(s7_wrong_type_arg_error(sc, "cadddr", 0, lst, "a list whose cddr is also a list"));
  if (!is_pair(cdddr(lst))) return(s7_wrong_type_arg_error(sc, "cadddr", 0, lst, "a list whose cdddr is also a list"));

  return(car(cdr(cdr(cdr(lst)))));
}


static s7_pointer g_cadadr(s7_scheme *sc, s7_pointer args)
{
  #define H_cadadr "(cadadr lst) returns (car (cdr (car (cdr lst)))): (cadadr '(1 (2 3 4))) -> 3"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "cadadr", 0, lst, "a pair"));
  if (!is_pair(cdr(lst))) return(s7_wrong_type_arg_error(sc, "cadadr", 0, lst, "a list whose cdr is also a list"));
  if (!is_pair(cadr(lst))) return(s7_wrong_type_arg_error(sc, "cadadr", 0, lst, "a list whose cadr is also a list"));
  if (!is_pair(cdadr(lst))) return(s7_wrong_type_arg_error(sc, "cadadr", 0, lst, "a list whose cdadr is also a list"));

  return(car(cdr(car(cdr(lst)))));
}


static s7_pointer g_caddar(s7_scheme *sc, s7_pointer args)
{
  #define H_caddar "(caddar lst) returns (car (cdr (cdr (car lst)))): (caddar '((1 2 3 4))) -> 3"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "caddar", 0, lst, "a pair"));
  if (!is_pair(car(lst))) return(s7_wrong_type_arg_error(sc, "caddar", 0, lst, "a list whose car is also a list"));
  if (!is_pair(cdar(lst))) return(s7_wrong_type_arg_error(sc, "caddar", 0, lst, "a list whose cdar is also a list"));
  if (!is_pair(cddar(lst))) return(s7_wrong_type_arg_error(sc, "caddar", 0, lst, "a list whose cddar is also a list"));

  return(car(cdr(cdr(car(lst)))));
}


static s7_pointer g_cdaaar(s7_scheme *sc, s7_pointer args)
{
  #define H_cdaaar "(cdaaar lst) returns (cdr (car (car (car lst)))): (cdaaar '((((1 2 3))))) -> '(2 3)"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "cdaaar", 0, lst, "a pair"));
  if (!is_pair(car(lst))) return(s7_wrong_type_arg_error(sc, "cdaaar", 0, lst, "a list whose car is also a list"));
  if (!is_pair(caar(lst))) return(s7_wrong_type_arg_error(sc, "cdaaar", 0, lst, "a list whose caar is also a list"));
  if (!is_pair(caaar(lst))) return(s7_wrong_type_arg_error(sc, "cdaaar", 0, lst, "a list whose caaar is also a list"));

  return(cdr(car(car(car(lst)))));
}


static s7_pointer g_cdaadr(s7_scheme *sc, s7_pointer args)
{
  #define H_cdaadr "(cdaadr lst) returns (cdr (car (car (cdr lst)))): (cdaadr '(1 ((2 3 4)))) -> '(3 4)"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "cdaadr", 0, lst, "a pair"));
  if (!is_pair(cdr(lst))) return(s7_wrong_type_arg_error(sc, "cdaadr", 0, lst, "a list whose cdr is also a list"));
  if (!is_pair(cadr(lst))) return(s7_wrong_type_arg_error(sc, "cdaadr", 0, lst, "a list whose cadr is also a list"));
  if (!is_pair(caadr(lst))) return(s7_wrong_type_arg_error(sc, "cdaadr", 0, lst, "a list whose caadr is also a list"));

  return(cdr(car(car(cdr(lst)))));
}


static s7_pointer g_cdadar(s7_scheme *sc, s7_pointer args)
{
  #define H_cdadar "(cdadar lst) returns (cdr (car (cdr (car lst)))): (cdadar '((1 (2 3 4)))) -> '(3 4)"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "cdadar", 0, lst, "a pair"));
  if (!is_pair(car(lst))) return(s7_wrong_type_arg_error(sc, "cdadar", 0, lst, "a list whose car is also a list"));
  if (!is_pair(cdar(lst))) return(s7_wrong_type_arg_error(sc, "cdadar", 0, lst, "a list whose cdar is also a list"));
  if (!is_pair(cadar(lst))) return(s7_wrong_type_arg_error(sc, "cdadar", 0, lst, "a list whose cadar is also a list"));

  return(cdr(car(cdr(car(lst)))));
}


static s7_pointer g_cddaar(s7_scheme *sc, s7_pointer args)
{
  #define H_cddaar "(cddaar lst) returns (cdr (cdr (car (car lst)))): (cddaar '(((1 2 3 4)))) -> '(3 4)"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "cddaar", 0, lst, "a pair"));
  if (!is_pair(car(lst))) return(s7_wrong_type_arg_error(sc, "cddaar", 0, lst, "a list whose car is also a list"));
  if (!is_pair(caar(lst))) return(s7_wrong_type_arg_error(sc, "cddaar", 0, lst, "a list whose caar is also a list"));
  if (!is_pair(cdaar(lst))) return(s7_wrong_type_arg_error(sc, "cddaar", 0, lst, "a list whose cdaar is also a list"));

  return(cdr(cdr(car(car(lst)))));
}


static s7_pointer g_cdaddr(s7_scheme *sc, s7_pointer args)
{
  #define H_cdaddr "(cdaddr lst) returns (cdr (car (cdr (cdr lst)))): (cdaddr '(1 2 (3 4 5))) -> '(4 5)"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "cdaddr", 0, lst, "a pair"));
  if (!is_pair(cdr(lst))) return(s7_wrong_type_arg_error(sc, "cdaddr", 0, lst, "a list whose cdr is also a list"));
  if (!is_pair(cddr(lst))) return(s7_wrong_type_arg_error(sc, "cdaddr", 0, lst, "a list whose cddr is also a list"));
  if (!is_pair(caddr(lst))) return(s7_wrong_type_arg_error(sc, "cdaddr", 0, lst, "a list whose caddr is also a list"));

  return(cdr(car(cdr(cdr(lst)))));
}


static s7_pointer g_cddddr(s7_scheme *sc, s7_pointer args)
{
  #define H_cddddr "(cddddr lst) returns (cdr (cdr (cdr (cdr lst)))): (cddddr '(1 2 3 4 5)) -> '(5)"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "cddddr", 0, lst, "a pair"));
  if (!is_pair(cdr(lst))) return(s7_wrong_type_arg_error(sc, "cddddr", 0, lst, "a list whose cdr is also a list"));
  if (!is_pair(cddr(lst))) return(s7_wrong_type_arg_error(sc, "cddddr", 0, lst, "a list whose cddr is also a list"));
  if (!is_pair(cdddr(lst))) return(s7_wrong_type_arg_error(sc, "cddddr", 0, lst, "a list whose cdddr is also a list"));

  return(cdr(cdr(cdr(cdr(lst)))));
}


static s7_pointer g_cddadr(s7_scheme *sc, s7_pointer args)
{
  #define H_cddadr "(cddadr lst) returns (cdr (cdr (car (cdr lst)))): (cddadr '(1 (2 3 4 5))) -> '(4 5)"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "cddadr", 0, lst, "a pair"));
  if (!is_pair(cdr(lst))) return(s7_wrong_type_arg_error(sc, "cddadr", 0, lst, "a list whose cdr is also a list"));
  if (!is_pair(cadr(lst))) return(s7_wrong_type_arg_error(sc, "cddadr", 0, lst, "a list whose cadr is also a list"));
  if (!is_pair(cdadr(lst))) return(s7_wrong_type_arg_error(sc, "cddadr", 0, lst, "a list whose cdadr is also a list"));

  return(cdr(cdr(car(cdr(lst)))));
}


static s7_pointer g_cdddar(s7_scheme *sc, s7_pointer args)
{
  #define H_cdddar "(cdddar lst) returns (cdr (cdr (cdr (car lst)))): (cdddar '((1 2 3 4 5))) -> '(4 5)"
  s7_pointer lst = car(args);

  if (!is_pair(lst)) return(s7_wrong_type_arg_error(sc, "cdddar", 0, lst, "a pair"));
  if (!is_pair(car(lst))) return(s7_wrong_type_arg_error(sc, "cdddar", 0, lst, "a list whose car is also a list"));
  if (!is_pair(cdar(lst))) return(s7_wrong_type_arg_error(sc, "cdddar", 0, lst, "a list whose cdar is also a list"));
  if (!is_pair(cddar(lst))) return(s7_wrong_type_arg_error(sc, "cdddar", 0, lst, "a list whose cddar is also a list"));

  return(cdr(cdr(cdr(car(lst)))));
}


/* reverse is in the generic function section */
static s7_pointer g_reverse_in_place(s7_scheme *sc, s7_pointer args)
{
  #define H_reverse_in_place "(reverse! lst) reverses lst in place"
  s7_pointer p, np;
  
  p = car(args);
  if (is_null(p))
    return(sc->NIL);
  
  if (!is_pair(p))
    return(s7_wrong_type_arg_error(sc, "reverse!", 0, p, "a list"));
  
  np = reverse_in_place(sc, sc->NIL, p);
  if (is_null(np))
    return(s7_wrong_type_arg_error(sc, "reverse!", 0, p, "a proper list"));
  
  return(np);
}


s7_pointer s7_assq(s7_scheme *sc, s7_pointer obj, s7_pointer x)
{
  s7_pointer y;
  y = x;
  while (true)
    {
      /* we can blithely take the car of anything, since we're not treating it as an object,
       *   then if we get a bogus match, the following check that caar made sense ought to catch it.
       */
      if ((obj == caar(x)) && (is_pair(car(x)))) return(car(x));
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if ((obj == caar(x)) && (is_pair(car(x)))) return(car(x));
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if ((obj == caar(x)) && (is_pair(car(x)))) return(car(x));
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if ((obj == caar(x)) && (is_pair(car(x)))) return(car(x));
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if ((obj == caar(x)) && (is_pair(car(x)))) return(car(x));
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if ((obj == caar(x)) && (is_pair(car(x)))) return(car(x));
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      y = cdr(y);
      if (x == y) return(sc->F);
    }
  return(sc->F); /* not reached */
}


static s7_pointer g_assq(s7_scheme *sc, s7_pointer args)
{
  #define H_assq "(assq obj alist) returns the key-value pair associated (via eq?) with the key obj in the association list alist"
  /* this version accepts any kind of list 
   *   my little essay: the scheme standard should not unnecessarily restrict the kinds of arguments
   *                    a function can take (such as saying memq only accepts proper lists).  It is
   *                    trivial for the programmer to add such a check to a built-in function, but
   *                    not trivial to re-invent the built-in function with that restriction removed.
   *                    If some structure exists as a normal scheme object (a dotted or circular list),
   *                    every built-in function should be able to deal with it, if it makes sense at all.
   */

  s7_pointer x;

  x = cadr(args);
  if (!is_pair(x))
    {
      if (is_null(x)) return(sc->F);
      return(s7_wrong_type_arg_error(sc, "assq", 2, x, "a list"));
    }
  return(s7_assq(sc, car(args), x));
}


static s7_pointer g_assv(s7_scheme *sc, s7_pointer args)
{
  #define H_assv "(assv obj alist) returns the key-value pair associated (via eqv?) with the key obj in the association list alist"
  s7_pointer x, y, obj;

  x = cadr(args);
  if (!is_pair(x))
    {
      if (is_null(x)) return(sc->F);
      return(s7_wrong_type_arg_error(sc, "assv", 2, x, "a list"));
    }
  obj = car(args);
  if (is_simple(obj))
    return(s7_assq(sc, obj, x));

  y = x;
  while (true)
    {
      /* here we can't play the assq == game because s7_is_eqv thinks it's getting a legit s7 object */
      if ((is_pair(car(x))) && (s7_is_eqv(obj, caar(x)))) return(car(x));
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if ((is_pair(car(x))) && (s7_is_eqv(obj, caar(x)))) return(car(x));
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      y = cdr(y);
      if (x == y) return(sc->F);
    }
  return(sc->F); /* not reached */
}


static s7_pointer g_assoc(s7_scheme *sc, s7_pointer args)
{
  #define H_assoc "(assoc obj alist (func #f)) returns the key-value pair associated (via equal?) with the key obj in the association list alist.\
If 'func' is a function of 2 arguments, it is used for the comparison instead of 'equal?"

  s7_pointer x, y, obj;

  x = cadr(args);      

  if (!is_pair(x))
    {
      if (is_null(x)) 
	return(sc->F);
      return(s7_wrong_type_arg_error(sc, "assoc", 2, x, "a list"));
    }
  if (!is_pair(car(x)))
    return(s7_wrong_type_arg_error(sc, "assoc", 2, x, "an a-list")); /* we're assuming caar below so it better exist */
      
  if (is_not_null(cddr(args))) 
    {
      s7_pointer eq_func;
      /* check 3rd arg before 2nd (trailing arg error check) */

      eq_func = caddr(args);
      if (!is_procedure(eq_func))
	return(s7_wrong_type_arg_error(sc, "assoc function,", 3, eq_func, "a function"));

      if (!args_match(sc, eq_func, 2))
	return(s7_wrong_type_arg_error(sc, "assoc", 3, eq_func, "a procedure that can take 2 arguments"));

      /* now maybe there's a simple case */
      if (s7_list_length(sc, x) > 0)
	{
	  if ((is_safe_procedure(eq_func)) &&
	      (is_c_function(eq_func)))
	    {
	      s7_function func;
	      
	      func = c_function_call(eq_func);
	      car(sc->T2_1) = car(args);
	      for (; is_pair(x); x = cdr(x))
		{
		  if (is_pair(car(x)))
		    {
		      car(sc->T2_2) = caar(x);
		      if (is_true(sc, (*func)(sc, sc->T2_1)))
			return(car(x));
		    }
		  else return(s7_wrong_type_arg_error(sc, "assoc", 2, cadr(args), "an a-list"));
		}
	      return(sc->F);
	    }
	}

      /* sc->value = sc->F; */
      y = cons(sc, args, sc->NIL);
      ecdr(y) = x;
      fcdr(y) = x;
      push_stack(sc, OP_ASSOC_IF, y, eq_func);
      push_stack(sc, OP_APPLY, list_2(sc, car(args), caar(x)), eq_func);
      return(sc->UNSPECIFIED);
    }

  x = cadr(args);
  if (!is_pair(x))
    {
      if (is_null(x)) return(sc->F);
      return(s7_wrong_type_arg_error(sc, "assoc", 2, x, "a list"));
    }
  obj = car(args);

  if (is_simple(obj))
    return(s7_assq(sc, obj, x));

  y = x;
  if (s7_is_string(obj))
    {
      s7_pointer val;
      while (true)
	{
	  if (is_pair(car(x)))
	    {
	      val = caar(x);
	      if ((val == obj) ||
		  ((s7_is_string(val)) &&
		   (scheme_strings_are_equal(obj, val))))
		return(car(x));
	    }
	  x = cdr(x);
	  if (!is_pair(x)) return(sc->F);

	  if (is_pair(car(x)))
	    {
	      val = caar(x);
	      if ((val == obj) ||
		  ((s7_is_string(val)) &&
		   (scheme_strings_are_equal(obj, val))))
		return(car(x));
	    }
	  x = cdr(x);
	  if (!is_pair(x)) return(sc->F);
      
	  y = cdr(y);
	  if (x == y) return(sc->F);
	}
      return(sc->F);
    }

  while (true)
    {
      if ((is_pair(car(x))) && (s7_is_equal(sc, obj, caar(x)))) return(car(x));
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
      
      if ((is_pair(car(x))) && (s7_is_equal(sc, obj, caar(x)))) return(car(x));
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      y = cdr(y);
      if (x == y) return(sc->F);
    }
  return(sc->F); /* not reached */
}


/* ---------------- member, memv, memq ---------------- */

s7_pointer s7_memq(s7_scheme *sc, s7_pointer obj, s7_pointer x)
{
  s7_pointer y;
  y = x;
  while (true)
    {
      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      y = cdr(y);
      if (x == y) return(sc->F);
    }
  return(sc->F);
}


static s7_pointer g_memq(s7_scheme *sc, s7_pointer args)
{
  #define H_memq "(memq obj list) looks for obj in list and returns the list from that point if it is found, otherwise #f. memq uses eq?"
  s7_pointer x;

  x = cadr(args);
  if (!is_pair(x))
    {
      if (is_null(x)) return(sc->F);
      return(s7_wrong_type_arg_error(sc, "memq", 2, x, "a list"));
    }
  return(s7_memq(sc, car(args), x));
}

/* I think (memq 'c '(a b . c)) should return #f because otherwise
 *   (memq '() ...) would return the '() at the end.
 */


#if WITH_OPTIMIZATION
/* if memq's list is a quoted list, it won't be changing, so we can tell ahead of time that it is
 *   a proper list, and what its length is.
 */
static s7_pointer memq_3, memq_4, memq_any;

static s7_pointer g_memq_3(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, obj;
  x = cadr(args);
  obj = car(args);
  while (true)
    {
      if (obj == car(x)) return(x);
      x = cdr(x);
      if (obj == car(x)) return(x);
      x = cdr(x);
      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
    }
  return(sc->F);
}

static s7_pointer g_memq_4(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, obj;
  x = cadr(args);
  obj = car(args);
  /* using expression_data here for cadr(args) and doing a linear search is slower!
   */
  while (true)
    {
      if (obj == car(x)) return(x);
      x = cdr(x);
      if (obj == car(x)) return(x);
      x = cdr(x);
      if (obj == car(x)) return(x);
      x = cdr(x);
      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
    }
  return(sc->F);
}

static s7_pointer g_memq_any(s7_scheme *sc, s7_pointer args)
{
  /* no circular list check needed in this case */
  s7_pointer x, obj;
  x = cadr(args);
  obj = car(args);
  while (true)
    {
      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
    }
  return(sc->F);
}


static s7_pointer memq_car;
static s7_pointer g_memq_car(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, obj;

  obj = finder(sc, cadar(args));
  if (!is_pair(obj))
    return(s7_wrong_type_arg_error(sc, "car", 0, cadar(args), "a pair"));
  obj = car(obj);
  x = cadr(cadr(args));

  while (true)
    {
      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);

      if (obj == car(x)) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
    }
  return(sc->F);
}

static s7_pointer memq_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if ((is_pair(caddr(expr))) &&
      (car(caddr(expr)) == sc->QUOTE) &&
      (is_pair(cadr(caddr(expr)))))
    {
      int len;

      if ((is_pair(cadr(expr))) &&
	  (optimize_data(cadr(expr)) == HOP_SAFE_C_S) &&
	  (c_call(cadr(expr)) == g_car))
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(memq_car);
	}

      len = s7_list_length(sc, cadr(caddr(expr)));
      if (len > 0)
	{
	  if ((len % 4) == 0)
	    return(memq_4);
	  if ((len % 3) == 0)
	    return(memq_3);
	  return(memq_any);
	}
    }
  return(f);
}

#endif

static s7_pointer memv_number(s7_scheme *sc, s7_pointer obj, s7_pointer x)
{
  s7_pointer y;
  y = x;
  while (true)
    {
      if ((s7_is_number(car(x))) && (numbers_are_eqv(obj, car(x)))) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
      if ((s7_is_number(car(x))) && (numbers_are_eqv(obj, car(x)))) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
      if ((s7_is_number(car(x))) && (numbers_are_eqv(obj, car(x)))) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
      if ((s7_is_number(car(x))) && (numbers_are_eqv(obj, car(x)))) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
      y = cdr(y);
      if (x == y) return(sc->F);
    }
  return(sc->F);
}


static s7_pointer g_memv(s7_scheme *sc, s7_pointer args)
{
  #define H_memv "(memv obj list) looks for obj in list and returns the list from that point if it is found, otherwise #f. memv uses eqv?"
  s7_pointer x, y, obj;

  x = cadr(args);
  if (!is_pair(x))
    {
      if (is_null(x)) return(sc->F);
      return(s7_wrong_type_arg_error(sc, "memv", 2, x, "a list"));
    }
  obj = car(args);

  if (is_simple(obj))
    return(s7_memq(sc, obj, x));

  if (s7_is_number(obj))
    return(memv_number(sc, obj, x));

  y = x;
  while (true)
    {
      if (s7_is_eqv(obj, car(x))) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
      
      if (s7_is_eqv(obj, car(x))) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
      
      y = cdr(y);
      if (x == y) return(sc->F);
    }
  return(sc->F); /* not reached */
}


static s7_pointer member(s7_scheme *sc, s7_pointer obj, s7_pointer x)
{
  s7_pointer y;

  y = x;
  if (s7_is_string(obj))
    {
      while (true)
	{
	  if ((obj == car(x)) ||
	      ((s7_is_string(car(x))) &&
	       (scheme_strings_are_equal(obj, car(x)))))
	    return(x);
	  x = cdr(x);
	  if (!is_pair(x)) return(sc->F);

	  if ((obj == car(x)) ||
	      ((s7_is_string(car(x))) &&
	       (scheme_strings_are_equal(obj, car(x)))))
	    return(x);
	  x = cdr(x);
	  if (!is_pair(x)) return(sc->F);
      
	  y = cdr(y);
	  if (x == y) return(sc->F);
	}
      return(sc->F);
    }

  while (true)
    {
      if (s7_is_equal(sc, obj, car(x))) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
      
      if (s7_is_equal(sc, obj, car(x))) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
      
      if (s7_is_equal(sc, obj, car(x))) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
      
      if (s7_is_equal(sc, obj, car(x))) return(x);
      x = cdr(x);
      if (!is_pair(x)) return(sc->F);
      
      y = cdr(y);
      if (x == y) return(sc->F);
    }
  return(sc->F); /* not reached */
}


static s7_pointer g_member(s7_scheme *sc, s7_pointer args)
{
  #define H_member "(member obj list (func #f)) looks for obj in list and returns the list from that point if it is found, otherwise #f. \
member uses equal?  If 'func' is a function of 2 arguments, it is used for the comparison instead of 'equal?"

  /* this could be extended to accept sequences:
   *    (member #\a "123123abnfc" char=?) -> "abnfc"
   *    (member "abc" "123abc321" string=?) -> "abc321" but there's the string length complication
   *    (member 1 #(0 1 2) =) ? -- for c_objects (ffi) we would need a substring equivalent
   *      and what would it do for a hash-table? or an environment?
   * also the 3 arg version can implement position, find, etc
   */

  s7_pointer x, y, obj;

  x = cadr(args);

  if (!is_pair(x))
    {
      if (is_null(x)) return(sc->F);
      return(s7_wrong_type_arg_error(sc, "member", 2, x, "a list"));
    }

  if (is_not_null(cddr(args))) 
    {
      s7_pointer eq_func;
      /* check 3rd arg before 2nd (trailing arg error check) */

      eq_func = caddr(args);
      if (!is_procedure(eq_func))
	return(s7_wrong_type_arg_error(sc, "member function,", 3, eq_func, "a function"));

      if (!args_match(sc, eq_func, 2))
	return(s7_wrong_type_arg_error(sc, "member", 3, eq_func, "a procedure that can take 2 arguments"));

      /* now maybe there's a simple case */
      if (s7_list_length(sc, x) > 0)
	{
	  if ((is_safe_procedure(eq_func)) &&
	      (is_c_function(eq_func)))
	    {
	      s7_function func;

	      func = c_function_call(eq_func);
	      car(sc->T2_1) = car(args);

	      for (; is_pair(x); x = cdr(x))
		{
		  car(sc->T2_2) = car(x);
		  if (is_true(sc, (*func)(sc, sc->T2_1)))
		    return(x);
		}
	      return(sc->F);
	    }
	}

      /* sc->value = sc->F; */
      y = cons(sc, args, sc->NIL); /* this could probably be handled with a counter cell (cdr here is unused) */
      ecdr(y) = x;
      fcdr(y) = x;
      push_stack(sc, OP_MEMBER_IF, y, eq_func);
      push_stack(sc, OP_APPLY, list_2(sc, car(args), car(x)), eq_func);
      return(sc->UNSPECIFIED);
    }

  obj = car(args);

  if (is_simple(obj))
    return(s7_memq(sc, obj, x));

  /* the only things that aren't simply == here are c_object, string, number, vector, hash-table, pair, and c_pointer
   *   but all the other cases are unlikely.
   */

  if (s7_is_number(obj))
    return(memv_number(sc, obj, x));

  return(member(sc, obj, x));
}



#if WITH_OPTIMIZATION
static s7_pointer member_sq;
static s7_pointer g_member_sq(s7_scheme *sc, s7_pointer args)
{
  s7_pointer obj, lst;
  lst = cadr(cadr(args));
  obj = finder(sc, car(args));

  if (is_simple(obj))
    return(s7_memq(sc, obj, lst));

  if (s7_is_number(obj))
    return(memv_number(sc, obj, lst));

  return(member(sc, obj, lst));
}

static s7_pointer member_sym_s;
static s7_pointer g_member_sym_s(s7_scheme *sc, s7_pointer args)
{
  /* 1st arg is quoted symbol */
  s7_pointer lst;

  lst = finder(sc, cadr(args));
  if (!is_pair(lst))
    {
      if (is_null(lst)) return(sc->F);
      return(s7_wrong_type_arg_error(sc, "member", 2, lst, "a list"));
    }
  return(s7_memq(sc, cadar(args), lst));
}

static s7_pointer member_num_s;
static s7_pointer g_member_num_s(s7_scheme *sc, s7_pointer args)
{
  s7_pointer lst;

  lst = finder(sc, cadr(args));
  if (!is_pair(lst))
    {
      if (is_null(lst)) return(sc->F);
      return(s7_wrong_type_arg_error(sc, "member", 2, lst, "a list"));
    }
  return(memv_number(sc, car(args), lst));
}

static s7_pointer member_ss;
static s7_pointer g_member_ss(s7_scheme *sc, s7_pointer args)
{
  s7_pointer obj, x;

  obj = finder(sc, car(args));
  x = finder(sc, cadr(args));
  if (!is_pair(x))
    {
      if (is_null(x)) return(sc->F);
      return(s7_wrong_type_arg_error(sc, "member", 2, x, "a list"));
    }

  if (is_simple(obj))
    return(s7_memq(sc, obj, x));

  if (s7_is_number(obj))
    return(memv_number(sc, obj, x));

  return(member(sc, obj, x));
}

static s7_pointer member_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (is_symbol(caddr(expr)))
	{
	  if ((is_pair(cadr(expr))) &&
	      (car(cadr(expr)) == sc->QUOTE) &&
	      (is_symbol(cadr(cadr(expr)))))
	    {
	      /* (member 'a odd) */
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(member_sym_s);
	    }

	  if (s7_is_number(cadr(expr)))
	    {
	      /* (member 4 lst) */
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(member_num_s);
	    }

	  if (is_symbol(cadr(expr)))
	    {
	      /* (member obj lst) */
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(member_ss);
	    }
	}
      else
	{
	  if ((is_symbol(cadr(expr))) &&
	      (is_pair(caddr(expr))) &&
	      (car(caddr(expr)) == sc->QUOTE) &&
	      (is_pair(cadr(caddr(expr)))))
	    {
	      /* (member q '(quote lambda case)) */
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(member_sq);
	    }
	}
    }
  return(f);
}
#endif


static bool is_member(s7_pointer sym, s7_pointer lst)
{
  s7_pointer x;
  for (x = lst; is_pair(x); x = cdr(x))
    if (sym == car(x))
      return(true);
  return(false);
}


static s7_pointer g_is_provided(s7_scheme *sc, s7_pointer args)
{
  #define H_is_provided "(provided? symbol) returns #t if symbol is a member of the *features* list"
  if (!is_symbol(car(args)))
    return(s7_wrong_type_arg_error(sc, "provided?", 0, car(args), "a symbol"));

  return(make_boolean(sc, is_member(car(args), s7_name_to_value(sc, "*features*"))));
}


static s7_pointer g_provide(s7_scheme *sc, s7_pointer args)
{
  #define H_provide "(provide symbol) adds symbol to the *features* list"
  s7_pointer features;

  if (!is_symbol(car(args)))
    return(s7_wrong_type_arg_error(sc, "provide", 0, car(args), "a symbol"));

  features = make_symbol(sc, "*features*");
  if (!is_member(car(args), s7_symbol_value(sc, features)))
    s7_symbol_set_value(sc, 
			features,
			cons(sc, 
				car(args), 
				s7_symbol_value(sc, features)));
  return(car(args));
}


void s7_provide(s7_scheme *sc, const char *feature)
{
  g_provide(sc, cons(sc, s7_make_symbol(sc, feature), sc->NIL));
}


static s7_pointer g_features_set(s7_scheme *sc, s7_pointer args)
{
  if (s7_is_list(sc, cadr(args)))
    return(cadr(args));
  return(sc->ERROR);
}



static s7_pointer g_list(s7_scheme *sc, s7_pointer args)
{
  #define H_list "(list ...) returns its arguments in a list"

  return(args);
}


static s7_pointer g_append(s7_scheme *sc, s7_pointer args)
{
  #define H_append "(append ...) returns its argument lists appended into one list"
  /* but weirdly (append '() 1) returns 1 
   */
  s7_pointer x, y;

  if (is_null(args)) 
    return(sc->NIL);

  if (is_null(cdr(args)))
    return(car(args)); 
  
  x = sc->NIL;
  for (y = args; is_not_null(y); y = cdr(y)) 
    {
      /* the original version used s7_append but that copies the arguments too many times if there are 3 or more lists */
      s7_pointer p;
      if (is_null(cdr(y)))
	return(reverse_in_place(sc, car(y), x)); /* i.e. tack car(y) onto end of x copied and reversed */

      p = car(y);
      if (!is_proper_list(sc, p))
	return(s7_wrong_type_arg_error(sc, "append", position_of(y, args), p, "a proper list"));

      while (is_not_null(p))
	{
	  x = cons(sc, car(p), x);
	  p = cdr(p);
	}
    }
  return(x);
}


static s7_pointer append_in_place(s7_scheme *sc, s7_pointer a, s7_pointer b)
{
  /* tack b onto the end of a without copying either -- 'a' is changed! */
  s7_pointer p;
  if (is_null(a))
    return(b);
  p = a;
  while (is_not_null(cdr(p))) p = cdr(p);
  cdr(p) = b;
  return(a);
}



/* -------------------------------- vectors -------------------------------- */

bool s7_is_vector(s7_pointer p)    
{ 
  return(type(p) == T_VECTOR);
}

#define FILLED true
#define NOT_FILLED false

static s7_pointer make_vector_1(s7_scheme *sc, s7_Int len, bool filled, bool add_to_cache) 
{
  s7_pointer x;
  if (len > 134217728)
    {
      /* len is an "int" currently */
      float ilog2;

      ilog2 = log((double)len) / log(2.0);
      if (sizeof(size_t) > 4)
	{
	  if (ilog2 > 56.0)
	    return(s7_out_of_range_error(sc, "make-vector length,", 1, make_integer(sc, len), "should be less than about 2^56 probably"));
	}
      else
	{
	  if (ilog2 > 28.0)
	    return(s7_out_of_range_error(sc, "make-vector length,", 1, make_integer(sc, len), "should be less than about 2^28 probably"));
	}
    }

  /* this has to follow the error checks!  (else garbage in free_heap temps portion confuses GC when "vector" is finalized) */

  NEW_CELL(sc, x);
  vector_length(x) = 0;
  vector_elements(x) = NULL;
  set_type(x, T_VECTOR | T_SAFE_PROCEDURE); /* (v 0) as vector-ref is safe */

  if (len > 0)
    {
      vector_elements(x) = (s7_pointer *)malloc(len * sizeof(s7_pointer));
      if (!(vector_elements(x)))
	return(s7_error(sc, make_symbol(sc, "out-of-memory"), 
			list_1(sc, make_protected_string(sc, "make-vector allocation failed!"))));

      vector_length(x) = len;
      if (filled) s7_vector_fill(sc, x, sc->NIL); /* make_hash_table assumes nil as the default value */
    }

  x->object.vector.vextra.dim_info = NULL;
  if (add_to_cache) add_vector(sc, x);
  return(x);
}


s7_pointer s7_make_vector(s7_scheme *sc, s7_Int len)
{
  return(make_vector_1(sc, len, FILLED, true));
}


s7_Int s7_vector_length(s7_pointer vec)
{
  return(vector_length(vec));
}


s7_Int s7_vector_print_length(s7_scheme *sc)
{
  return(s7_integer(slot_value(sc->vector_print_length)));
}


s7_Int s7_set_vector_print_length(s7_scheme *sc, s7_Int new_len)
{
  s7_Int old_len;
  old_len = s7_integer(slot_value(sc->vector_print_length));
  slot_set_value(sc->vector_print_length, make_integer(sc, new_len));
  return(old_len);
}


static s7_pointer g_vector_print_length_set(s7_scheme *sc, s7_pointer args)
{
  if (s7_is_integer(cadr(args)))
    {
      s7_Int len;
      len = s7_integer(cadr(args));
      if (len >= 0)
	return(cadr(args));
    }
  return(sc->ERROR);
}


#if (!WITH_GMP)
void s7_vector_fill(s7_scheme *sc, s7_pointer vec, s7_pointer obj)
#else
static void vector_fill(s7_scheme *sc, s7_pointer vec, s7_pointer obj)
#endif
{
  s7_Int len, i = 1, left;
  s7_pointer *orig, *cur;

  len = vector_length(vec);
  if (len == 0) return;

  orig = vector_elements(vec);
  orig[0] = obj;

  while (i < len)
    {
      cur = (s7_pointer *)(orig + i);
      left = len - i;
      if (left < i)
	memcpy((void *)cur, (void *)orig, sizeof(s7_pointer) * left);
      else memcpy((void *)cur, (void *)orig, sizeof(s7_pointer) * i);
      /* is there a faster way to do this? */
      i *= 2;
    }
}


static s7_pointer g_vector_fill(s7_scheme *sc, s7_pointer args)
{
  #define H_vector_fill "(vector-fill! v val) sets all elements of the vector v to val"
  s7_pointer x;
  x = car(args);

  if (!s7_is_vector(x))
    return(s7_wrong_type_arg_error(sc, "vector-fill!", 1, x, "a vector"));

  s7_vector_fill(sc, x, cadr(args));
  return(cadr(args));
}


s7_pointer s7_vector_ref(s7_scheme *sc, s7_pointer vec, s7_Int index) 
{
  if (index >= vector_length(vec))
    return(s7_out_of_range_error(sc, "vector-ref index,", 2, make_integer(sc, index), "should be less than vector length"));

  return(vector_element(vec, index));
}


s7_pointer s7_vector_set(s7_scheme *sc, s7_pointer vec, s7_Int index, s7_pointer a) 
{
  if (index >= vector_length(vec))
    return(s7_out_of_range_error(sc, "vector-set! index,", 2, make_integer(sc, index), "should be less than vector length"));

  vector_element(vec, index) = a;
  return(a);
}


s7_pointer *s7_vector_elements(s7_pointer vec)
{
  return(vector_elements(vec));
}


s7_Int *s7_vector_dimensions(s7_pointer vec)
{
  s7_Int *dims;
  if (vector_is_multidimensional(vec))
    return(vec->object.vector.vextra.dim_info->dims);
  dims = (s7_Int *)malloc(sizeof(s7_Int));
  dims[0] = vector_length(vec);
  return(dims);
}


s7_Int *s7_vector_offsets(s7_pointer vec)
{
  s7_Int *offs;
  if (vector_is_multidimensional(vec))
    return(vec->object.vector.vextra.dim_info->offsets);
  offs = (s7_Int *)malloc(sizeof(s7_Int));
  offs[0] = 1;
  return(offs);
}


s7_pointer s7_vector_to_list(s7_scheme *sc, s7_pointer vect)
{
  s7_Int i, len;
  s7_pointer p;

  len = vector_length(vect);
  if (len == 0) return(sc->NIL);

  if (len < (sc->free_heap_top - sc->free_heap))
    {
      sc->w = cons(sc, vector_element(vect, len - 1), sc->NIL);
      for (i = len - 2; i >= 0; i--)
	sc->w = cons_unchecked(sc, vector_element(vect, i), sc->w);
    }
  else
    {
      sc->w = sc->NIL;
      for (i = len - 1; i >= 0; i--)
	sc->w = cons(sc, vector_element(vect, i), sc->w);
    }
  p = sc->w;
  sc->w = sc->NIL;
  return(p);
}


static s7_pointer g_vector_to_list(s7_scheme *sc, s7_pointer args)
{
  #define H_vector_to_list "(vector->list v) returns the elements of the vector v as a list; (map values v)"
  if (!s7_is_vector(car(args)))
    return(s7_wrong_type_arg_error(sc, "vector->list", 0, car(args), "a vector"));
  return(s7_vector_to_list(sc, car(args)));
}


s7_pointer s7_make_and_fill_vector(s7_scheme *sc, s7_Int len, s7_pointer fill)
{
  s7_pointer vect;
  vect = make_vector_1(sc, len, NOT_FILLED, true);
  s7_vector_fill(sc, vect, fill);
  return(vect);
}


static s7_pointer g_vector(s7_scheme *sc, s7_pointer args)
{
  #define H_vector "(vector ...) returns a vector whose elements are the arguments"
  s7_Int i, len;
  s7_pointer vec;
  
  len = s7_list_length(sc, args);
  vec = make_vector_1(sc, len, NOT_FILLED, true);
  if (len > 0)
    {
      s7_pointer x;
      for (x = args, i = 0; is_pair(x); x = cdr(x), i++) 
	vector_element(vec, i) =  car(x);
    }
  return(vec);
}


static s7_pointer g_list_to_vector(s7_scheme *sc, s7_pointer args)
{
  #define H_list_to_vector "(list->vector lst) returns a vector containing the elements of lst; (apply vector lst)"
  
  if (is_null(car(args)))
    return(s7_make_vector(sc, 0));
  if (!is_proper_list(sc, car(args)))
    return(s7_wrong_type_arg_error(sc, "list->vector", 0, car(args), "a proper list"));
  return(g_vector(sc, car(args)));
}


static s7_pointer g_vector_length(s7_scheme *sc, s7_pointer args)
{
  #define H_vector_length "(vector-length v) returns the length of vector v"
  if (!s7_is_vector(car(args)))
    return(s7_wrong_type_arg_error(sc, "vector-length", 0, car(args), "a vector"));
  return(make_integer(sc, vector_length(car(args))));
}


static s7_pointer make_shared_vector(s7_scheme *sc, s7_pointer vect, int skip_dims, s7_Int index)
{
  s7_pointer x;
  s7_vdims_t *v;

  /* (let ((v #2d((1 2) (3 4)))) (v 1)) 
   * (let ((v (make-vector '(2 3 4) 0))) (v 1 2))
   * (let ((v #3d(((0 1 2 3) (4 5 6 7) (8 9 10 11)) ((12 13 14 15) (16 17 18 19) (20 21 22 23))))) (v 0 1))
   */

  NEW_CELL(sc, x);
  vector_length(x) = 0;
  vector_elements(x) = NULL;
  set_type(x, T_VECTOR);

  v = (s7_vdims_t *)malloc(sizeof(s7_vdims_t));
 
  v->ndims = vector_ndims(vect) - skip_dims;
  v->dims = (s7_Int *)((vect)->object.vector.vextra.dim_info->dims + skip_dims);
  v->offsets = (s7_Int *)((vect)->object.vector.vextra.dim_info->offsets + skip_dims);
  v->original = vect;
  x->object.vector.vextra.dim_info = v;  

  vector_length(x) = vector_offset(vect, skip_dims - 1);
  vector_elements(x) = (s7_pointer *)(vector_elements(vect) + index);
  add_vector(sc, x);
  return(x);
}


static s7_pointer vector_ref_1(s7_scheme *sc, s7_pointer vect, s7_pointer indices)
{
  s7_Int index = 0;
  if (vector_length(vect) == 0)
    return(s7_out_of_range_error(sc, "vector-ref", 1, vect, "this vector has no elements, so vector-ref is hopeless"));

  if (vector_is_multidimensional(vect))
    {
      unsigned int i;
      s7_pointer x;
      for (x = indices, i = 0; (is_not_null(x)) && (i < vector_ndims(vect)); x = cdr(x), i++)
	{
	  s7_Int n;
	  if (!s7_is_integer(car(x)))
	    return(s7_wrong_type_arg_error(sc, "vector-ref index,", i + 2, car(x), "an integer"));

	  n = s7_integer(car(x));
	  if ((n < 0) || 
	      (n >= vector_dimension(vect, i)))
	    return(s7_out_of_range_error(sc, "vector-ref", i + 2, car(x), "index should be between 0 and the dimension size"));

	  index += n * vector_offset(vect, i);
	}
      if (is_not_null(x))
	return(s7_wrong_number_of_args_error(sc, "too many indices for vector-ref: ~A", indices));

      /* if not enough indices, return a shared vector covering whatever is left */
      if (i < vector_ndims(vect))
	return(make_shared_vector(sc, vect, i, index));
    }
  else
    {
      /* (let ((hi (make-vector 3 0.0)) (sum 0.0)) (do ((i 0 (+ i 1))) ((= i 3)) (set! sum (+ sum (hi i)))) sum) */

      if (!s7_is_integer(car(indices)))
	return(s7_wrong_type_arg_error(sc, "vector-ref index,", 2, car(indices), "an integer"));

      index = s7_integer(car(indices));
      if ((index < 0) ||
	  (index >= vector_length(vect)))
	return(s7_out_of_range_error(sc, "vector-ref index,", 2, car(indices), "should be between 0 and the vector length"));
      
      if (is_not_null(cdr(indices)))                /* (let ((L '#(#(1 2 3) #(4 5 6)))) (vector-ref L 1 2)) */
	{
	  s7_pointer new_vect;
	  new_vect = vector_element(vect, index);
	  if (!s7_is_vector(new_vect))             /* (vector-ref #(1) 0 0) */
	    return(s7_wrong_type_arg_error(sc, "vector-ref", 1, new_vect, "a vector"));

	  return(vector_ref_1(sc, new_vect, cdr(indices))); 
	}
    }

  return(vector_element(vect, index));
}


static s7_pointer g_vector_ref(s7_scheme *sc, s7_pointer args)
{
  #define H_vector_ref "(vector-ref v ... i) returns the i-th element of vector v.  If v \
is a multidimensional vector, you can also use (vector-ref v ...) where the trailing args \
are the indices, or omit 'vector-ref': (v ...)."

  s7_pointer vec;

  vec = car(args);
  if (!s7_is_vector(vec))
    return(s7_wrong_type_arg_error(sc, "vector-ref", 1, vec, "a vector"));

  return(vector_ref_1(sc, vec, cdr(args)));
}

#if WITH_OPTIMIZATION
static s7_pointer vector_ref_ic, vector_ref_add1;
static s7_pointer g_vector_ref_ic(s7_scheme *sc, s7_pointer args)
{
  s7_pointer vec;
  s7_Int index;

  vec = finder(sc, car(args));
  if (!s7_is_vector(vec))
    return(s7_wrong_type_arg_error(sc, "vector-ref", 1, car(args), "a vector"));

  index = s7_integer(cadr(args));
  if (index >= vector_length(vec))
    return(s7_out_of_range_error(sc, "vector-ref index,", 2, cadr(args), "should be less than vector length"));

  if (vector_is_multidimensional(vec))
    return(make_shared_vector(sc, vec, 1, index));

  return(vector_element(vec, index));
}

static s7_pointer g_vector_ref_add1(s7_scheme *sc, s7_pointer args)
{
  s7_pointer vec, x;
  s7_Int index;

  vec = finder(sc, car(args));
  if (!s7_is_vector(vec))
    return(s7_wrong_type_arg_error(sc, "vector-ref", 1, car(args), "a vector"));

  x = finder(sc, cadr(cadr(args)));
  if (!s7_is_integer(x))
    return(s7_wrong_type_arg_error(sc, "vector-ref", 2, cadr(args), "an integer"));

  index = s7_integer(x) + 1;
  if ((index < 0) ||
      (index >= vector_length(vec)))
    return(s7_out_of_range_error(sc, "vector-ref index,", 2, cadr(args), "should be between 0 and the vector length"));

  if (vector_is_multidimensional(vec))
    return(make_shared_vector(sc, vec, 1, index));

  return(vector_element(vec, index));
}

static s7_pointer vector_ref_2;
static s7_pointer g_vector_ref_2(s7_scheme *sc, s7_pointer args)
{
  s7_pointer vec, ind;
  s7_Int index;

  vec = car(args);
  if (!s7_is_vector(vec))
    return(s7_wrong_type_arg_error(sc, "vector-ref", 1, vec, "a vector"));
  if (vector_is_multidimensional(vec))
    return(g_vector_ref(sc, args));
  
  ind = cadr(args);
  if (!s7_is_integer(ind))
    return(s7_wrong_type_arg_error(sc, "vector-ref index,", 2, ind, "an integer"));
  index = s7_integer(ind);
  if ((index < 0) ||
      (index >= vector_length(vec)))
    return(s7_out_of_range_error(sc, "vector-ref index,", 2, ind, "should be between 0 and the vector length"));

  return(vector_element(vec, index));
}
#endif


static s7_pointer g_vector_set(s7_scheme *sc, s7_pointer args)
{
  #define H_vector_set "(vector-set! v i ... value) sets the i-th element of vector v to value.  If 'v' is \
multidimensional you can also use (vector-set! v ... val) where the ellipsis refers to the indices.  You \
can also use 'set!' instead of 'vector-set!': (set! (v ...) val) -- I find this form much easier to read."

  s7_pointer vec, val;
  s7_Int index;

  vec = car(args);
  if (!s7_is_vector(vec))
    return(s7_wrong_type_arg_error(sc, "vector-set!", 1, vec, "a vector"));
  if (vector_length(vec) == 0)
    return(s7_out_of_range_error(sc, "vector-set!", 1, vec, "this vector has no elements, so vector-set! is hopeless"));
  
  if (vector_is_multidimensional(vec))
    {
      unsigned int i;
      s7_pointer x;
      index = 0;
      for (x = cdr(args), i = 0; (is_not_null(cdr(x))) && (i < vector_ndims(vec)); x = cdr(x), i++)
	{
	  s7_Int n;
	  if (!s7_is_integer(car(x)))
	    return(s7_wrong_type_arg_error(sc, "vector-set! index,", i + 2, car(x), "an integer"));

	  n = s7_integer(car(x));
	  if ((n < 0) || 
	      (n >= vector_dimension(vec, i)))
	    return(s7_out_of_range_error(sc, "vector-set!", i, car(x), "index should be between 0 and the dimension size"));

	  index += n * vector_offset(vec, i);
	}

      if (is_not_null(cdr(x)))
	return(s7_wrong_number_of_args_error(sc, "too many args for vector-set!: ~A", args));
      if (i != vector_ndims(vec))
	return(s7_wrong_number_of_args_error(sc, "not enough args for vector-set!: ~A", args));

      val = car(x);
    }
  else
    {
      if (!s7_is_integer(cadr(args)))
	return(s7_wrong_type_arg_error(sc, "vector-set! index,", 2, cadr(args), "an integer"));

      index = s7_integer(cadr(args));
      if ((index < 0) ||
	  (index >= vector_length(vec)))
	return(s7_out_of_range_error(sc, "vector-set! index,", 2, cadr(args), "should be between 0 and the vector length"));

      if (is_not_null(cdddr(args)))
	return(g_vector_set(sc, cons(sc, vector_element(vec, index), cddr(args))));

      val = caddr(args);
    }
  
  vector_element(vec, index) = val;
  return(val);
}


#if WITH_OPTIMIZATION
static s7_pointer vector_set_ic;
static s7_pointer g_vector_set_ic(s7_scheme *sc, s7_pointer args)
{
  s7_pointer vec, val;
  s7_Int index;

  vec = finder(sc, car(args));
  if (!s7_is_vector(vec))
    return(s7_wrong_type_arg_error(sc, "vector-set!", 1, car(args), "a vector"));
  if (vector_is_multidimensional(vec))
    return(g_vector_set(sc, list_3(sc, vec, cadr(args), finder(sc, caddr(args)))));

  index = s7_integer(cadr(args));
  if (index >= vector_length(vec))
    return(s7_out_of_range_error(sc, "vector-set! index,", 2, cadr(args), "should be less than vector length"));

  val = finder(sc, caddr(args));
  vector_element(vec, index) = val;
  return(val);
}

static s7_pointer vector_set_ref;
static s7_pointer g_vector_set_ref(s7_scheme *sc, s7_pointer args)
{
  s7_pointer vec1, vec2, ind, val;
  s7_Int index;

  vec1 = finder(sc, car(args));
  if (!s7_is_vector(vec1))
    return(s7_wrong_type_arg_error(sc, "vector-set!", 1, car(args), "a vector"));

  vec2 = finder(sc, cadr(caddr(args)));
  if (!s7_is_vector(vec2))
    return(s7_wrong_type_arg_error(sc, "vector-ref", 1, cadr(caddr(args)), "a vector"));
  
  ind = finder(sc, cadr(args));
  if (!s7_is_integer(ind))
    return(s7_wrong_type_arg_error(sc, "vector-set! index,", 2, cadr(args), "an integer"));

  index = s7_integer(ind);
  if ((index < 0) ||
      (index >= vector_length(vec1)) ||
      (index >= vector_length(vec2)))
    return(s7_out_of_range_error(sc, "vector-set! index,", 2, ind, "should be between 0 and the vector length"));

  val = vector_element(vec2, index);

  if (vector_is_multidimensional(vec1))
    return(g_vector_set(sc, list_3(sc, vec1, ind, val)));

  vector_element(vec1, index) = val;
  return(val);
}

static s7_pointer vector_set_3;
static s7_pointer g_vector_set_3(s7_scheme *sc, s7_pointer args)
{
  s7_pointer vec, ind, val;
  s7_Int index;

  vec = car(args);
  if (!s7_is_vector(vec))
    return(s7_wrong_type_arg_error(sc, "vector-set!", 1, vec, "a vector"));
  if (vector_is_multidimensional(vec))
    return(g_vector_set(sc, args));
  
  ind = cadr(args);
  if (!s7_is_integer(ind))
    return(s7_wrong_type_arg_error(sc, "vector-set! index,", 2, ind, "an integer"));
  index = s7_integer(ind);
  if ((index < 0) ||
      (index >= vector_length(vec)))
    return(s7_out_of_range_error(sc, "vector-set! index,", 2, ind, "should be between 0 and the vector length"));

  val = caddr(args);
  vector_element(vec, index) = val;
  return(val);
}

static s7_pointer vector_set_ssc;
static s7_pointer g_vector_set_ssc(s7_scheme *sc, s7_pointer args)
{
  s7_pointer vec, ind, val;
  s7_Int index;

  vec = finder(sc, car(args));
  if (!s7_is_vector(vec))
    return(s7_wrong_type_arg_error(sc, "vector-set!", 1, car(args), "a vector"));
  if (vector_is_multidimensional(vec))
    return(g_vector_set(sc, list_3(sc, vec, finder(sc, cadr(args)), caddr(args))));

  ind = finder(sc, cadr(args));
  if (!s7_is_integer(ind))
    return(s7_wrong_type_arg_error(sc, "vector-set! index,", 2, ind, "an integer"));
  index = s7_integer(ind);
  if (index < 0)
    return(s7_out_of_range_error(sc, "vector-set! index,", 2, ind, "should not be negative"));
  if (index >= vector_length(vec))
    return(s7_out_of_range_error(sc, "vector-set! index,", 2, ind, "should be less than vector length"));

  val = caddr(args);
  vector_element(vec, index) = val;
  return(val);
}

/* other standard cases (fft-related): (+|- (vector-ref ...) s) -- actually (vector-set! v i (+|- (vector-ref v i) s))
 *                                     (+|- (* s (vector-ref ...)) (* s (vector-ref ...)))
 * (vset v i s)
 * (vset v i (vref v j)) (vset v (+ i 1) (vref v (+ j 1)))
 * (vset v (+ i 1) s)
 */
#endif



#define MAX_VECTOR_DIMENSIONS 512

static s7_pointer g_make_vector(s7_scheme *sc, s7_pointer args)
{
  #define H_make_vector "(make-vector len (value #f)) returns a vector of len elements initialized to value. \
To create a multidimensional vector, put the dimension bounds in a list (this is to avoid ambiguities such as \
(make-vector 1 2) where it's not clear whether the '2' is an initial value or a dimension size).  (make-vector '(2 3) 1.0) \
returns a 2 dimensional vector of 6 total elements, all initialized to 1.0."

  s7_Int len;
  s7_pointer x, fill, vec;
  fill = sc->UNSPECIFIED;

  x = car(args);
  if (s7_is_integer(x))
    {
      len = s7_integer(x);
      if (len < 0)
	return(s7_wrong_type_arg_error(sc, "make-vector length,", (is_null(cdr(args))) ? 0 : 1, x, "a non-negative integer"));
    }
  else
    {
      s7_pointer y;
      if (!(is_pair(x)))
	return(s7_wrong_type_arg_error(sc, "make-vector", (is_null(cdr(args))) ? 0 : 1, x, "an integer or a list of integers"));

      if (!s7_is_integer(car(x)))
	return(s7_wrong_type_arg_error(sc, "make-vector", (is_null(cdr(args))) ? 0 : 1, car(x), "each dimension should be an integer"));

      if (is_null(cdr(x)))
	len = s7_integer(car(x));
      else
	{
	  int dims;

	  dims = s7_list_length(sc, x);
	  if (dims <= 0)                /* 0 if circular, negative if dotted */
	    return(s7_wrong_type_arg_error(sc, "make-vector", (is_null(cdr(args))) ? 0 : 1, x, "a proper list of dimensions"));
	  if (dims > MAX_VECTOR_DIMENSIONS)
	    return(s7_out_of_range_error(sc, "make-vector dimension list,", (is_null(cdr(args))) ? 0 : 1, x, "less than 512 dimensions"));

	  for (len = 1, y = x; is_not_null(y); y = cdr(y))
	    {
	      if (!s7_is_integer(car(y)))
		return(s7_wrong_type_arg_error(sc, "make-vector", position_of(y, x), car(y), "an integer"));
	      len *= s7_integer(car(y));
	      if (len < 0)
		return(s7_wrong_type_arg_error(sc, "make-vector", position_of(y, x), car(y), "a non-negative integer"));
	    }
	}
    }
  
  if (is_not_null(cdr(args))) 
    fill = cadr(args);

  vec = make_vector_1(sc, len, NOT_FILLED, true);
  if (len > 0) s7_vector_fill(sc, vec, fill);

  if ((is_pair(x)) &&
      (is_pair(cdr(x))))
    {
      int i;
      s7_Int offset = 1;
      s7_pointer y;
      s7_vdims_t *v;

      v = (s7_vdims_t *)malloc(sizeof(s7_vdims_t));
      v->ndims = safe_list_length(sc, x);
      v->dims = (s7_Int *)malloc(v->ndims * sizeof(s7_Int));
      v->offsets = (s7_Int *)malloc(v->ndims * sizeof(s7_Int));
      v->original = sc->F;

      for (i = 0, y = x; is_not_null(y); i++, y = cdr(y))
	v->dims[i] = s7_integer(car(y));

      for (i = v->ndims - 1; i >= 0; i--)
	{
	  v->offsets[i] = offset;
	  offset *= v->dims[i];
	}

      vec->object.vector.vextra.dim_info = v;
    }

  return(vec);
}


static s7_pointer g_is_vector(s7_scheme *sc, s7_pointer args)
{
  #define H_is_vector "(vector? obj) returns #t if obj is a vector"
  return(make_boolean(sc, s7_is_vector(car(args))));
}


int s7_vector_rank(s7_pointer vect)
{
  if (vector_is_multidimensional(vect))
    return(vector_ndims(vect));
  return(1);
}


static s7_pointer g_vector_dimensions(s7_scheme *sc, s7_pointer args)
{
  #define H_vector_dimensions "(vector-dimensions vect) returns a list of vect's dimensions.  In srfi-63 terms:\n\
    (define array-dimensions vector-dimensions)\n\
    (define (array-rank v) (length (vector-dimensions v)))"

  s7_pointer x;

  x = car(args);
  if (!s7_is_vector(x))
    return(s7_wrong_type_arg_error(sc, "vector-dimensions", 0, x, "a vector"));

  if (vector_is_multidimensional(x))
    {
      int i;
      sc->w = sc->NIL;
      for (i = vector_ndims(x) - 1; i >= 0; i--)
	sc->w = cons(sc, make_integer(sc, vector_dimension(x, i)), sc->w);
      x = sc->w;
      sc->w = sc->NIL;
      return(x);
    }
  
  return(list_1(sc, make_integer(sc, vector_length(x))));
}


#define MV_TOO_MANY_ELEMENTS -1
#define MV_NOT_ENOUGH_ELEMENTS -2

static int traverse_vector_data(s7_scheme *sc, s7_pointer vec, int flat_ref, int dimension, int dimensions, int *sizes, s7_pointer lst)
{
  /* we're filling vec, we're currently looking for element (flat-wise) flat_ref,
   *   we're at ref in dimension of dimensions, where sizes gives the bounds, and lst is our data
   *   #3D(((1 2 3) (4 5 6)) ((7 8 9) (10 11 12)))
   */
  int i;
  s7_pointer x;

  for (i = 0, x = lst; i < sizes[dimension]; i++, x = cdr(x))
    {
      if (!is_pair(x))
	return(MV_NOT_ENOUGH_ELEMENTS);

      if (dimension == (dimensions - 1))
	vector_element(vec, flat_ref++) = car(x);
      else 
	{
	  flat_ref = traverse_vector_data(sc, vec, flat_ref, dimension + 1, dimensions, sizes, car(x));
	  if (flat_ref < 0) return(flat_ref);
	}
    }

  if (is_not_null(x))
    return(MV_TOO_MANY_ELEMENTS);
  return(flat_ref);
}


static s7_pointer s7_multivector_error(s7_scheme *sc, const char *message, s7_pointer data)
{
  return(s7_error(sc, sc->READ_ERROR,
		  list_3(sc, 
			 make_protected_string(sc, "reading constant vector, ~A: ~A"),
			 make_protected_string(sc, message),
			 data)));
}


static s7_pointer g_multivector(s7_scheme *sc, int dims, s7_pointer data)
{
  /* get the dimension bounds from data, make the new vector, fill it from data */
  s7_pointer vec, x;
  int i, total_size = 1, vec_loc, err;
  int *sizes;
  
  /* (#2d((1 2 3) (4 5 6)) 0 0) -> 1
   * (#2d((1 2 3) (4 5 6)) 0 1) -> 2
   * (#2d((1 2 3) (4 5 6)) 1 1) -> 5
   * (#3D(((1 2) (3 4)) ((5 6) (7 8))) 0 0 0) -> 1
   * (#3D(((1 2) (3 4)) ((5 6) (7 8))) 1 1 0) -> 7
   * #3D(((1 2) (3 4)) ((5 6) (7))) -> error, #3D(((1 2) (3 4)) ((5 6) (7 8 9))), #3D(((1 2) (3 4)) (5 (7 8 9))) etc
   *
   * but a special case: #nD() is an n-dimensional empty vector
   */

  if (dims <= 0)      /* #0d(...) */
    return(s7_out_of_range_error(sc, "#nD(...) dimensions,", 1, make_integer(sc, dims), "must be 1 or more"));

  sc->w = sc->NIL;
  if (is_null(data))  /* dims are already 0 (calloc above) */
    return(g_make_vector(sc, list_1(sc, g_make_list(sc, list_2(sc, make_integer(sc, dims), small_int(0))))));

  sizes = (int *)calloc(dims, sizeof(int));
  for (x = data, i = 0; i < dims; i++)
    {
      sizes[i] = safe_list_length(sc, x);
      total_size *= sizes[i];
      sc->w = cons(sc, make_integer(sc, sizes[i]), sc->w);
      x = car(x);
      if ((i < (dims - 1)) && 
	  (!is_pair(x)))
	{
	  free(sizes);
	  return(s7_multivector_error(sc, "we need a list that fully specifies the vector's elements", data));
	}
    }

  vec = g_make_vector(sc, list_1(sc, safe_reverse_in_place(sc, sc->w)));
  vec_loc = s7_gc_protect(sc, vec);
  sc->w = sc->NIL;

  /* now fill the vector checking that all the lists match */
  err = traverse_vector_data(sc, vec, 0, 0, dims, sizes, data);

  free(sizes);
  s7_gc_unprotect_at(sc, vec_loc);
  if (err < 0) 
    return(s7_multivector_error(sc, (err == MV_TOO_MANY_ELEMENTS) ? "found too many elements" : "not enough elements found", data));

  return(vec);
}


static s7_pointer g_qq_multivector(s7_scheme *sc, s7_pointer args)
{
  /* `#2d((1 2) ,(list 3 4)) */
  #define H_qq_multivector "quasiquote internal support for multidimensional vector constants"
  return(g_multivector(sc, s7_integer(car(args)), cdr(args)));
}


static s7_pointer vector_copy(s7_scheme *sc, s7_pointer old_vect)
{
  s7_Int len;
  s7_pointer new_vect;
  len = vector_length(old_vect);

  if (vector_is_multidimensional(old_vect))
    new_vect = g_make_vector(sc, list_1(sc, g_vector_dimensions(sc, list_1(sc, old_vect))));
  else new_vect = make_vector_1(sc, len, NOT_FILLED, true);

  /* here and in vector-fill! we have a problem with bignums -- should new bignums be allocated? (copy_list also) */

  memcpy((void *)(vector_elements(new_vect)), (void *)(vector_elements(old_vect)), len * sizeof(s7_pointer));
  return(new_vect);
}





/* -------- sort! -------- */

static s7_scheme *compare_sc;
static s7_function compare_func;
static s7_pointer compare_args;

static int vector_compare(const void *v1, const void *v2)
{
  car(compare_args) = (*(s7_pointer *)v1);
  cadr(compare_args) = (*(s7_pointer *)v2);
  if (is_true(compare_sc, (*(compare_func))(compare_sc, compare_args)))
    return(-1);
  return(1);
}


static s7_pointer g_sort(s7_scheme *sc, s7_pointer args)
{
  #define H_sort "(sort! list-or-vector less?) sorts a list or vector using the function 'less?' to compare elements.\
If its first argument is a list, the list is copied (despite the '!')."

  s7_pointer data, lessp;
  s7_Int len = 0, n, k;

  data = car(args);
  if (is_null(data)) return(sc->NIL);
  if ((!is_pair(data)) && 
      (!s7_is_vector(data)) &&
      (!is_c_object(data)))
    return(s7_wrong_type_arg_error(sc, "sort! data,", 1, data, "a vector, list, or an object from make-type and friends"));

  lessp = cadr(args);
  if (!is_procedure(lessp))
    return(s7_wrong_type_arg_error(sc, "sort! function,", 2, lessp, "a function"));
  if ((is_continuation(lessp)) || is_goto(lessp))
    return(s7_wrong_type_arg_error(sc, "sort!", 2, lessp, "a normal procedure (not a continuation)"));
  if (!args_match(sc, lessp, 2))
    return(s7_wrong_type_arg_error(sc, "sort!", 2, lessp, "a procedure that can take 2 arguments"));

  switch (type(data))
    {
    case T_PAIR:
      len = s7_list_length(sc, data);            /* nil disposed of above, so 0 here == infinite */
      if (len <= 0)
	return(s7_error(sc, sc->WRONG_TYPE_ARG, 
			list_2(sc, make_protected_string(sc, "sort! argument 1 should be a proper list: ~S"), data)));
      if (len < 2) return(data);
      if (len == 2)
	{
	  push_stack(sc, OP_SORT_TWO, data, args);  /* this will return a list -- was sc->args  */
	  push_stack(sc, OP_APPLY, data, lessp);
	  return(sc->F);
	}
      push_stack(sc, OP_SORT4, args, sc->code);    /* gc protect the original list, OP_SORT4 calls s7_vector_to_list, was sc->args */
      car(args) = g_list_to_vector(sc, sc->x = list_1(sc, data));
      break;

    case T_VECTOR:
      /* (sort! (symbol-table) (lambda (a b) (< (length a) (length b)))) 
       * but this is harmless because (symbol-table) copies the vector -- are there any other cases?
       *
       * if (is_immutable(data)) 
       *   return(s7_wrong_type_arg_error(sc, "sort!", 1, data, "a mutable vector"));
       */
      len = vector_length(data);

      if ((is_safe_procedure(lessp)) &&
	  (is_c_function(lessp)))
	{
	  compare_sc = sc;
	  compare_func = c_function_call(lessp);
	  compare_args = sc->T2_1;
	  qsort((void *)s7_vector_elements(data), len, sizeof(s7_pointer), vector_compare);
	  return(data);
	}
      break;

      /* SOMEDAY: to sort c/s objects requires another arg to s7_new_type_x and make-type
       */
    }

  if (len < 2) return(data);
  n = len - 1;
  k = ((int)(n / 2)) + 1;

  sc->x = s7_make_vector(sc, (sc->safety == 0) ? 4 : 6); 
  vector_element(sc->x, 0) = make_mutable_integer(sc, n);
  vector_element(sc->x, 1) = make_mutable_integer(sc, k);
  vector_element(sc->x, 2) = make_mutable_integer(sc, 0);
  vector_element(sc->x, 3) = make_mutable_integer(sc, 0);
  if (sc->safety != 0)
    {
      vector_element(sc->x, 4) = make_mutable_integer(sc, 0);
      vector_element(sc->x, 5) = make_integer(sc, n * n);
    }

  push_stack(sc, OP_SORT, args, sc->x);
  sc->x = sc->NIL;
  return(sc->F);
  
  /* if the comparison function waffles, sort! can hang: (sort! '(1 2 3) (lambda (a b) (= a b)))
   * set *safety* to 1 to add a check for this loop, but the "safe" procedures are direct, so unchecked.
   */
}




/* -------- hash tables -------- */

bool s7_is_hash_table(s7_pointer p)
{
  return(type(p) == T_HASH_TABLE);
}


static s7_pointer g_is_hash_table(s7_scheme *sc, s7_pointer args)
{
  #define H_is_hash_table "(hash-table? obj) returns #t if obj is a hash-table"
  return(make_boolean(sc, s7_is_hash_table(car(args))));
}


static s7_pointer g_hash_table_size(s7_scheme *sc, s7_pointer args)
{
  #define H_hash_table_size "(hash-table-size obj) returns the size of the hash-table obj"
  if (!s7_is_hash_table(car(args)))
    return(s7_wrong_type_arg_error(sc, "hash-table-size", 0, car(args), "a hash-table"));
  return(make_integer(sc, hash_table_length(car(args))));
}


static s7_Int hash_loc(s7_scheme *sc, s7_pointer key)
{
  s7_Int loc = 0;
  const char *c; 

  switch (type(key))
    {
    case T_STRING:
      if (string_hash(key) != 0)
	return(string_hash(key));
      for (c = string_value(key); *c; c++) 
	loc = *c + loc * HASH_MULT;
      string_hash(key) = loc;
      return(loc);

    case T_INTEGER:
      loc = s7_integer(key);
      if (loc < 0) return(-loc);
      return(loc);

    case T_REAL:
      loc = (s7_Int)floor(s7_real(key));
      if (loc < 0) loc = -loc;
      return(loc);
      
      /* ratio or complex -- use type */

    case T_SYMBOL:
      return(symbol_hash(key));

    case T_SYNTAX:
      return((s7_Int)syntax_opcode(key));

    case T_CHARACTER:
      return((s7_Int)character(key));

    case T_VECTOR:
      return(vector_length(key));

    case T_PAIR:
      return(s7_list_length(sc, key));

    default:
      break;
    }

  return(type(key));
}


static s7_pointer hash_empty(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  return(sc->NIL);
}


static s7_pointer hash_equal(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  s7_pointer x;
  unsigned int hash_len, loc;

  hash_len = (int)hash_table_length(table) - 1;
  loc = hash_loc(sc, key) & hash_len;

  for (x = hash_table_elements(table)[loc]; is_pair(x); x = cdr(x))
    if (s7_is_equal(sc, fcdr(x), key))
      return(car(x));
  return(sc->NIL);
}


static s7_pointer hash_int(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  if (s7_is_integer(key))
    {
      s7_Int keyval;
      s7_pointer x;
      unsigned int hash_len, loc;
      hash_len = (int)hash_table_length(table) - 1;
      keyval = s7_integer(key);
      if (keyval < 0)
	loc = (-keyval) & hash_len;
      else loc = keyval & hash_len;
      for (x = hash_table_elements(table)[loc]; is_pair(x); x = cdr(x))
	if (s7_integer(fcdr(x)) == keyval)
	  return(car(x));
    }
  return(sc->NIL);
}


static s7_pointer hash_string(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  if (s7_is_string(key))
    {
      s7_pointer x;
      int hash_len, loc;
      hash_len = (int)hash_table_length(table) - 1;
      loc = string_hash(key);
      if (loc == 0)
	{
	  const char *c; 
	  for (c = string_value(key); *c; c++) 
	    loc = *c + loc * HASH_MULT;
	  string_hash(key) = loc;
	}
      loc &= hash_len;
      for (x = hash_table_elements(table)[loc]; is_pair(x); x = cdr(x))
	if (strings_are_equal(string_value(fcdr(x)), string_value(key)))
	  return(car(x));
    }
  return(sc->NIL);
}


static s7_pointer hash_char(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  if (s7_is_character(key))
    {
      s7_pointer x;
      int hash_len, loc;
      hash_len = (int)hash_table_length(table) - 1;
      loc = character(key) & hash_len;
      for (x = hash_table_elements(table)[loc]; is_pair(x); x = cdr(x))
	if (fcdr(x) == key)
	  return(car(x));
    }
  return(sc->NIL);
}


static s7_pointer hash_float(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  #define HASH_FLOAT_EPSILON 1.0e-12
  /* give the equality check some room.  We only get here if key is T_REAL.
   */

  if (type(key) == T_REAL)
    {
      s7_Double keyval;
      s7_pointer x;
      int hash_len, loc;

      hash_len = (int)hash_table_length(table) - 1;
      loc = (int)floor(real(key));
      if (loc < 0) loc = -loc;
      loc &= hash_len;
      keyval = real(key);

      for (x = hash_table_elements(table)[loc]; is_pair(x); x = cdr(x))
	if (fabs(s7_real(fcdr(x)) - keyval) < HASH_FLOAT_EPSILON)
	  return(car(x));
    }
  return(sc->NIL);
}


static s7_pointer hash_symbol(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  if (is_symbol(key))
    {
      s7_pointer x;
      int hash_len, loc;
      hash_len = (int)hash_table_length(table) - 1;
      loc = symbol_hash(key) & hash_len;
      for (x = hash_table_elements(table)[loc]; is_pair(x); x = cdr(x))
	if (fcdr(x) == key)
	  return(car(x));
    }
  return(sc->NIL);
}


#define DEFAULT_HASH_TABLE_SIZE 511
/* this may be too big -- smaller is faster in lint (even 31).  Perhaps we need *hash-table-default-size*?
 */

s7_pointer s7_make_hash_table(s7_scheme *sc, s7_Int size)
{
  s7_pointer table;
  /* size is rounded up to the next power of 2 */

  if ((size & (size + 1)) != 0)      /* already 2^n - 1 ? */
    {
      size--;
      size |= (size >> 1);
      size |= (size >> 2);
      size |= (size >> 4);
      size |= (size >> 8);
      size |= (size >> 16);
      if (s7_int_bits > 31) /* this is either 31 or 63 */
	size |= (size >> 32);
    }

  table = make_vector_1(sc, size + 1, FILLED, false); /* nil is the default value, don't add to vector cache! */
  /* size + 1 can be fooled if we don't catch most-positive-fixnum */

  set_type(table, T_HASH_TABLE | T_SAFE_PROCEDURE);
  hash_table_function(table) = hash_empty;
  hash_table_entries(table) = 0;
  add_hash_table(sc, table);
  return(table);
}


static s7_pointer g_make_hash_table(s7_scheme *sc, s7_pointer args)
{
  #define H_make_hash_table "(make-hash-table (size 511)) returns a new hash table"
  s7_Int size = DEFAULT_HASH_TABLE_SIZE;

  if (is_not_null(args))
    {
      if (s7_is_integer(car(args)))
	{
	  size = s7_integer(car(args));
	  if (size <= 0)
	    return(s7_out_of_range_error(sc, "make-hash-table size,", 0, car(args), "should be a positive integer"));
	  if (size > MAX_LIST_LENGTH)
	    return(s7_out_of_range_error(sc, "make-hash-table size,", 0, car(args), "should be a reasonable integer"));
	}
      else return(s7_wrong_type_arg_error(sc, "make-hash-table size,", 0, car(args), "an integer"));
    }
  
  return(s7_make_hash_table(sc, size));
}


s7_pointer s7_hash_table_ref(s7_scheme *sc, s7_pointer table, s7_pointer key)
{
  s7_pointer x;
  x = (*hash_table_function(table))(sc, table, key);

  if (is_not_null(x))
    return(cdr(x));
  return(sc->F);
}


s7_pointer s7_hash_table_set(s7_scheme *sc, s7_pointer table, s7_pointer key, s7_pointer value)
{
  s7_pointer x;
  x = (*hash_table_function(table))(sc, table, key);

  if (is_not_null(x))
    cdr(x) = value;
  else
    {
      s7_Int hash_len, loc;
      int typ;

      hash_len = hash_table_length(table) - 1;
      loc = hash_loc(sc, key) & hash_len;
      hash_table_entries(table)++;
      typ = type(key);

      if (hash_table_function(table) == hash_empty)
	{
	  switch (typ)
	    {
	    case T_STRING:
	      hash_table_function(table) = hash_string;
	      break;

	    case T_INTEGER:
	      hash_table_function(table) = hash_int;
	      break;

	    case T_REAL:
	      hash_table_function(table) = hash_float;
	      break;

	    case T_RATIO:
	    case T_COMPLEX:
	      hash_table_function(table) = hash_equal;
	      break;

	    case T_SYMBOL:
	      hash_table_function(table) = hash_symbol;
	      break;

	    case T_CHARACTER:
	      hash_table_function(table) = hash_char;
	      break;
	      
	    default:
	      hash_table_function(table) = hash_equal;
	      break;
	    }
	}
      else
	{
	  switch (typ)
	    {
	    case T_STRING:
	      if (hash_table_function(table) != hash_string) 
		hash_table_function(table) = hash_equal;
	      break;
	      
	    case T_INTEGER:
	      if (hash_table_function(table) != hash_int)
		hash_table_function(table) = hash_equal;
	      break;

	    case T_REAL:
	      if (hash_table_function(table) != hash_float)
		hash_table_function(table) = hash_equal;
	      break;

	    case T_RATIO:
	    case T_COMPLEX:
	      hash_table_function(table) = hash_equal;
	      break;

	    case T_CHARACTER:
	      if (hash_table_function(table) != hash_char) 
		hash_table_function(table) = hash_equal;
	      break;

	    case T_SYMBOL:
	      if (hash_table_function(table) != hash_symbol) 
		hash_table_function(table) = hash_equal;
	      break;

	    default:
	      hash_table_function(table) = hash_equal;
	      break;
	    }
	}
      x = cons_unchecked(sc, cons(sc, key, value), hash_table_elements(table)[loc]);
      hash_table_elements(table)[loc] = x;
      fcdr(x) = key;
    }
  return(value);
}


static s7_pointer g_hash_table_ref(s7_scheme *sc, s7_pointer args)
{
  /* basically the same layout as the symbol table */
  #define H_hash_table_ref "(hash-table-ref table key) returns the value associated with key in the hash table"
  s7_pointer table;

  table = car(args);
  
  if (!s7_is_hash_table(table))
    return(s7_wrong_type_arg_error(sc, "hash-table-ref", 1, table, "a hash-table"));

  /*
    (define (href H . args) 
      (if (null? (cdr args))
          (hash-table-ref H (car args))
          (apply href (hash-table-ref H (car args)) (cdr args))))
  */

  if (is_null(cddr(args)))
    return(s7_hash_table_ref(sc, table, cadr(args)));
  return(g_hash_table_ref(sc, cons(sc, s7_hash_table_ref(sc, table, cadr(args)), cddr(args))));
}

#if WITH_OPTIMIZATION
static s7_pointer hash_table_ref_2;
static s7_pointer g_hash_table_ref_2(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, table;
  table = car(args);
  
  if (!s7_is_hash_table(table))
    return(s7_wrong_type_arg_error(sc, "hash-table-ref", 1, table, "a hash-table"));
  
  if (hash_table_entries(table) == 0)
    return(sc->F);

  x = (*hash_table_function(table))(sc, table, cadr(args));
  if (is_not_null(x))
    return(cdr(x));
  return(sc->F);
}

static s7_pointer hash_table_ref_ss;
static s7_pointer g_hash_table_ref_ss(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, table;
  table = finder(sc, car(args));
  
  if (!s7_is_hash_table(table))
    return(s7_wrong_type_arg_error(sc, "hash-table-ref", 1, car(args), "a hash-table"));
  
  if (hash_table_entries(table) == 0)
    return(sc->F);

  x = (*hash_table_function(table))(sc, table, finder(sc, cadr(args)));
  if (is_not_null(x))
    return(cdr(x));
  return(sc->F);
}

static s7_pointer hash_table_ref_car;
static s7_pointer g_hash_table_ref_car(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x, y, table;
  table = finder(sc, car(args));
  
  if (!s7_is_hash_table(table))
    return(s7_wrong_type_arg_error(sc, "hash-table-ref", 1, car(args), "a hash-table"));
  
  if (hash_table_entries(table) == 0)
    return(sc->F);

  y = finder(sc, cadadr(args));
  if (!is_pair(y))
    return(s7_wrong_type_arg_error(sc, "car", 1, y, "a pair"));

  x = (*hash_table_function(table))(sc, table, car(y));
  if (is_not_null(x))
    return(cdr(x));
  return(sc->F);
}
#endif


static s7_pointer g_hash_table_set(s7_scheme *sc, s7_pointer args)
{
  #define H_hash_table_set "(hash-table-set! table key value) sets the value associated with key in the hash table to value"
  s7_pointer table;

  table = car(args);
  
  if (!s7_is_hash_table(table))
    return(s7_wrong_type_arg_error(sc, "hash-table-set!", 1, table, "a hash-table"));

  /* how would (set! (ht a b) c) choose the inner table if (ht a b) is not found?
   *   I'm not sure the multi-index case makes sense here
   */

  return(s7_hash_table_set(sc, table, cadr(args), caddr(args)));
}


static s7_pointer g_hash_table(s7_scheme *sc, s7_pointer args)
{
  #define H_hash_table "(hash-table ...) returns a hash-table containing the cons's passed as its arguments. \
That is, (hash-table '(\"hi\" . 3) (\"ho\" . 32)) returns a new hash-table with the two key/value pairs preinstalled."

  int len, ht_loc;
  s7_pointer x, ht;
  
  len = s7_list_length(sc, args);
  ht = s7_make_hash_table(sc, (len > 512) ? 4095 : 511);
  if (is_not_null(args))
    {
      ht_loc = s7_gc_protect(sc, ht); /* hash_table_set can cons, so we need to protect this */
      for (x = args; is_pair(x); x = cdr(x)) 
	{
	  if (is_pair(car(x)))
	    s7_hash_table_set(sc, ht, caar(x), cdar(x));
	  else
	    {
	      if (is_not_null(car(x)))
		{
		  s7_gc_unprotect_at(sc, ht_loc);
		  return(s7_wrong_type_arg_error(sc, "hash-table", position_of(x, args), car(x), "a pair: (key value)"));
		}
	    }
	}
      s7_gc_unprotect_at(sc, ht_loc);
    }
  return(ht);
}


static s7_pointer hash_list_copy(s7_scheme *sc, s7_pointer obj)
{
  if (is_pair(obj))
    {
      s7_pointer p;
      p = cons(sc, s7_copy(sc, car(obj)), hash_list_copy(sc, cdr(obj)));
      fcdr(p) = caar(p);
      return(p);
    }
  return(obj);
}


static s7_pointer hash_table_copy(s7_scheme *sc, s7_pointer old_hash)
{
  /* this has to copy not only the lists but the cons's in the lists! */
  s7_Int i, len;
  s7_pointer new_hash;
  s7_pointer *old_lists, *new_lists;
  int gc_loc;

  len = vector_length(old_hash);
  new_hash = s7_make_hash_table(sc, len);
  gc_loc = s7_gc_protect(sc, new_hash);

  old_lists = vector_elements(old_hash);
  new_lists = vector_elements(new_hash);

  for (i = 0; i < len; i++)
    if (is_not_null(old_lists[i]))
      new_lists[i] = hash_list_copy(sc, old_lists[i]);

  hash_table_entries(new_hash) = hash_table_entries(old_hash);
  hash_table_function(new_hash) = hash_table_function(old_hash);

  s7_gc_unprotect_at(sc, gc_loc);
  return(new_hash);
}


static s7_pointer hash_table_reverse(s7_scheme *sc, s7_pointer old_hash)
{
  s7_Int i, len;
  s7_pointer new_hash;
  s7_pointer *old_lists;
  int gc_loc;

  len = vector_length(old_hash);
  new_hash = s7_make_hash_table(sc, len);
  gc_loc = s7_gc_protect(sc, new_hash);

  old_lists = vector_elements(old_hash);
  /* don't set entries or function -- s7_hash_table_set below will handle those */

  for (i = 0; i < len; i++)
    {
      s7_pointer x;
      for (x = old_lists[i]; is_not_null(x); x = cdr(x))
	s7_hash_table_set(sc, new_hash, cdar(x), caar(x));
    }

  s7_gc_unprotect_at(sc, gc_loc);
  return(new_hash);
}


static s7_pointer hash_table_clear(s7_scheme *sc, s7_pointer table)
{
  int i, len;

  len = vector_length(table);
  for (i = 0; i < len; i++)
    vector_element(table, i) = sc->NIL;
  hash_table_entries(table) = 0;
  hash_table_function(table) = hash_empty;
  return(table);
}


static bool hash_tables_are_equal(s7_scheme *sc, s7_pointer x, s7_pointer y)
{
  s7_pointer *lists;
  int i, len;

  if (hash_table_entries(x) != hash_table_entries(y))
    return(false);
  if (hash_table_entries(x) == 0)
    return(true);
  if (hash_table_function(x) != hash_table_function(y))
    return(false);

  len = vector_length(x);
  lists = vector_elements(x);

  for (i = 0; i < len; i++)
    {
      s7_pointer p;
      for (p = lists[i]; is_not_null(p); p = cdr(p))
	{
	  s7_pointer key, x_val, y_val;

	  key = caar(p);
	  x_val = cdar(p);
	  y_val = (*hash_table_function(y))(sc, y, key);

	  if (is_null(y_val))
	    return(false);
	  if (!s7_is_equal(sc, x_val, cdr(y_val)))
	    return(false);
	}
    }

  /* if we get here, every key/value in x has a corresponding key/value in y, and the number of entries match,
   *   so surely the tables are equal??
   */

  return(true);
}


static s7_pointer g_hash_table_iterate(s7_scheme *sc, s7_pointer args)
{
  /* internal func pointed to by sc->HASH_TABLE_ITERATE */
  s7_pointer lst, loc, table;
  s7_Int vloc, len;
  s7_pointer *elements;

  lst = caar(args);
  if (is_pair(lst))
    {
      caar(args) = cdr(lst);
      return(car(lst));
    }

  table = cadar(args);
  len = hash_table_length(table);
  elements = hash_table_elements(table);

  loc = caddar(args);
  for (vloc = integer(loc) + 1; vloc < len;  vloc++)
    {
      s7_pointer x;
      x = elements[vloc];
      if (is_not_null(x))
	{
	  integer(loc) = vloc;
	  caar(args) = cdr(x);
	  return(car(x));
	}
    }

  integer(loc) = len;
  return(sc->NIL);
}


static s7_pointer g_make_hash_table_iterator(s7_scheme *sc, s7_pointer args)
{
  #define H_make_hash_table_iterator "(make-hash-table-iterator table) returns a function of no arguments that \
returns the next (key . value) pair in the hash-table each time it is called.  When there are no more pairs, it returns nil."

  if (!s7_is_hash_table(car(args)))
    return(s7_wrong_type_arg_error(sc, "make-hash-table-iterator", 0, car(args), "a hash-table"));

  return(make_closure(sc, 
		      sc->NIL,                             /* no args to the new function */
		      list_1(sc, list_2(sc, sc->HASH_TABLE_ITERATE,
					list_2(sc, sc->QUOTE, 
					       list_3(sc, sc->NIL, car(args), make_mutable_integer(sc, -1))))),
		      T_CLOSURE));
}


#define hash_table_iterate_args(H) cdadar(closure_body(H))

static char *hash_table_to_c_string(s7_scheme *sc, s7_pointer hash, bool to_file, shared_info *ci)
{
  s7_Int i, len, bufsize = 0, gc_iter;
  bool too_long = false;
  char **elements = NULL;
  char *buf;
  s7_pointer iterator, iter_loc;
  
  len = hash_table_entries(hash);
  if (len == 0)
    return(copy_string("#<hash-table>"));

  if (!to_file)
    {
      int plen;
      plen = s7_vector_print_length(sc);
      if (plen <= 0)
	return(copy_string("#<hash-table ...>"));

      if (len > plen)
	{
	  too_long = true;
	  len = plen;
	}
    }

  iterator = g_make_hash_table_iterator(sc, list_1(sc, hash));

  gc_iter = s7_gc_protect(sc, iterator);
  iter_loc = hash_table_iterate_args(iterator);

  elements = (char **)calloc(len, sizeof(char *));

  for (i = 0; i < len; i++)
    {
      elements[i] = object_to_c_string_with_circle_check(sc, g_hash_table_iterate(sc, iter_loc), USE_WRITE, WITH_ELLIPSES, ci);
      bufsize += safe_strlen(elements[i]);

      if (bufsize > sc->print_width) 
	{
	  if ((i + 1) < len) too_long = true;
	  len = i + 1;
	  break;
	}
    }
  s7_gc_unprotect_at(sc, gc_iter);

  bufsize += (len * 4 + 32);                   /* might be 2 parens per element + space + quote, so at least len*4 here */
  buf = (char *)malloc(bufsize * sizeof(char));

  sprintf(buf, "#<hash-table ");
  for (i = 0; i < len - 1; i++)
    {
      if (elements[i])
	{
	  /* strcat(buf, "'"); -- it's a constant so do we need a quote? #(0 (1 2)) for example */
	  strcat(buf, elements[i]);
	  free(elements[i]);
	  strcat(buf, " ");
	}
    }

  if (elements[len - 1])
    {
      /* strcat(buf, "'"); */
      strcat(buf, elements[len - 1]);
      free(elements[len - 1]);
    }

  free(elements);
  if (too_long)
    strcat(buf, " ...");
  strcat(buf, ">");

  return(buf);
}




/* -------------------------------- objects and functions -------------------------------- */

bool s7_is_function(s7_pointer p)  
{ 
  return(is_c_function(p));
}


bool s7_is_object(s7_pointer p) 
{ 
  return(is_c_object(p));
}



#if WITH_OPTIMIZATION
static unsigned int f_class = 0;

static s7_pointer fallback_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  return(f);
}


unsigned int s7_function_class(s7_pointer f)
{
  return(c_function_class(f));
}

void s7_function_set_class(s7_pointer f, unsigned int c)
{
  c_function_class(f) = c;
}


s7_pointer (*s7_function_chooser(s7_pointer fnc))(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  return(c_function_chooser(fnc));
}

void s7_function_set_chooser(s7_pointer fnc,  s7_pointer (*chooser)(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr))
{
  c_function_chooser(fnc) = chooser;
}


void *s7_function_chooser_data(s7_pointer expr)
{
  if ((ecdr(expr)) &&
      (is_c_function(ecdr(expr))))
    return(c_function_chooser_data(ecdr(expr)));
  return(NULL);
}

void s7_function_chooser_set_data(s7_pointer f, void *data)
{
  c_function_chooser_data(f) = data;
}


s7_function s7_function_choice(s7_pointer expr)
{
  if ((ecdr(expr)) &&
      (is_c_function(ecdr(expr))))
    return(c_function_call(ecdr(expr)));
  return(NULL);
}

void s7_function_choice_set_direct(s7_pointer expr)
{
  optimize_data(expr) = HOP_SAFE_C_C;
  clear_unsafe(expr);
}

bool s7_function_choice_is_direct(s7_pointer expr)
{
  return(optimize_data(expr) == HOP_SAFE_C_C);
}

s7_pointer s7_call_direct(s7_scheme *sc, s7_pointer expr)
{
  return(c_function_call(ecdr(expr))(sc, cdr(expr)));
}


static unsigned short table_top = 1, tables_size = 0;
static void ***tables = NULL;
static int *table_sizes = NULL;

static void clear_table(s7_pointer expr)
{
  int loc;
  loc = optimize_data_index(expr);
  if (tables[loc])
    memset((void *)(tables[loc]), 0, table_sizes[loc] * sizeof(void **));
}


void **s7_expression_make_data(s7_scheme *sc, s7_pointer expr, int size)
{
  if (tables == NULL)
    {
      tables_size = 8;
      tables = (void ***)calloc(tables_size, sizeof(void **));
      table_sizes = (int *)calloc(tables_size, sizeof(int));
    }
  else
    {
      if (table_top == tables_size)
	{
	  int i;
	  tables_size += 8;
	  tables = (void ***)realloc(tables, tables_size * sizeof(void **));
	  table_sizes = (int *)realloc(table_sizes, tables_size * sizeof(int));
	  for (i = table_top; i < tables_size; i++) {tables[i] = NULL; table_sizes[i] = 0;}
	}
    }
  tables[table_top] = (void **)calloc(size, sizeof(void *));
  table_sizes[table_top] = size;

  optimize_data_index(expr) = table_top++;
  set_has_table(expr);
  return(tables[optimize_data_index(expr)]);
}


void **s7_expression_data(s7_pointer expr)
{
  /* the has_table check should protect against leftover garbage in the optimize_data_index field */
  if (has_table(expr))
    return(tables[optimize_data_index(expr)]);
  return(NULL);
}

#else

unsigned int s7_function_class(s7_pointer f)
{
  return(0);
}

void s7_function_set_class(s7_pointer f, unsigned int c)
{
}

s7_pointer (*s7_function_chooser(s7_pointer fnc))(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  return(NULL); /* not called, I hope */
}

void s7_function_set_chooser(s7_pointer f,  s7_pointer (*chooser)(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr))
{
}

void *s7_function_chooser_data(s7_pointer f)
{
  return(NULL);
}

void s7_function_chooser_set_data(s7_pointer f, void *data)
{
}

s7_function s7_function_choice(s7_pointer expr)
{
  return(NULL);
}

bool s7_function_choice_is_direct(s7_pointer expr)
{
  return(false);
}

void s7_function_choice_set_direct(s7_pointer expr)
{
}

s7_pointer s7_call_direct(s7_scheme *sc, s7_pointer expr)
{
  return(sc->UNSPECIFIED);
}

void **s7_expression_make_data(s7_scheme *sc, s7_pointer expr, int size)
{
  return(NULL);
}

void **s7_expression_data(s7_pointer expr)
{
  return(NULL);
}
#endif


s7_pointer s7_make_function(s7_scheme *sc, const char *name, s7_function f, int required_args, int optional_args, bool rest_arg, const char *doc)
{
  s7_func_t *ptr;
  int ftype = T_C_FUNCTION;
  s7_pointer x;

  x = (s7_cell *)permanent_calloc(sizeof(s7_cell));
  x->hloc = NOT_IN_HEAP;

  ptr = (s7_func_t *)permanent_calloc(sizeof(s7_func_t));
  if (required_args == 0)
    {
      if (rest_arg)
	ftype = T_C_ANY_ARGS_FUNCTION;
      else 
	{
	  if (optional_args != 0)
	    ftype = T_C_OPT_ARGS_FUNCTION;
	  /* a thunk needs to check for no args passed */
	}
    }
  else
    {
      if (rest_arg)
	ftype = T_C_RST_ARGS_FUNCTION;
    }
  
  set_type(x, ftype | T_PROCEDURE);

  c_function(x) = ptr;
  c_function_call(x) = f;
  c_function_setter(x) = sc->F;
  c_function_name(x) = name;   /* (procedure-name proc) => (format #f "~A" proc) */
  if (doc)
    c_function_documentation(x) = make_permanent_string(doc);

  c_function_required_args(x) = required_args;
  c_function_optional_args(x) = optional_args;
  c_function_has_rest_arg(x) = rest_arg;
  if (rest_arg)
    c_function_all_args(x) = 10000000;
  else c_function_all_args(x) = required_args + optional_args;

#if WITH_OPTIMIZATION
  c_function_class(x) = ++f_class;
  c_function_chooser(x) = fallback_chooser;
  c_function_chooser_data(x) = NULL;
#endif

  return(x);
}


s7_pointer s7_apply_function(s7_scheme *sc, s7_pointer fnc, s7_pointer args)
{
  if (is_symbol(fnc))
    fnc = s7_symbol_value(sc, fnc);

  if (is_c_function(fnc))
    return(c_function_call(fnc)(sc, args));

  /* fprintf(stderr, "%s %s\n", DISPLAY(closure_body(fnc)), DISPLAY(args)); */

  push_stack(sc, OP_EVAL_DONE, sc->args, sc->code); 
  sc->args = args;
  sc->code = fnc;
  eval(sc, OP_APPLY);
  return(sc->value);
}


bool s7_is_procedure(s7_pointer x)
{
  return(is_procedure(x)); /* this returns "is applicable" so it is true for applicable c_objects, macros, etc */
}


static s7_pointer g_is_procedure(s7_scheme *sc, s7_pointer args)
{
  #define H_is_procedure "(procedure? obj) returns #t if obj is a procedure"
  s7_pointer x;
  int typ;

  x = car(args);
  if (!is_procedure(x)) return(sc->F);
  typ = type(x);

  /* make_object sets the T_PROCEDURE bit if the object has an apply function,
   *   but we currently return (procedure? "hi") -> #f, so we can't simply use
   *   is_procedure. 
   * 
   * Unfortunately much C code depends on s7_is_procedure treating applicable
   *  objects and macros as procedures.  Ideally we'd have s7_is_applicable.
   */
  return(make_boolean(sc,
		      (typ == T_CLOSURE) || 
		      (typ == T_CLOSURE_STAR) ||
		      (typ >= T_C_FUNCTION) ||
		      (typ == T_GOTO) ||
		      (typ == T_CONTINUATION) ||
		      (is_pws(x))));
}


static void s7_function_set_setter(s7_scheme *sc, const char *getter, const char *setter)
{
  c_function_setter(s7_name_to_value(sc, getter)) = s7_name_to_value(sc, setter);
}


static char *pws_documentation(s7_pointer x);
static s7_pointer pws_source(s7_scheme *sc, s7_pointer x);
static s7_pointer pws_arity(s7_scheme *sc, s7_pointer obj);


s7_pointer s7_procedure_source(s7_scheme *sc, s7_pointer p)
{
  /* in this context, there's no way to distinguish between:
   *    (procedure-source (let ((b 1)) (lambda (a) (+ a b))))
   * and
   *    (let ((b 1)) (procedure-source (lambda (a) (+ a b))))
   * both become:
   * ((a) (+ a b)) (((b . 1)) #(() () () () () ((make-filtered-comb . make-filtered-comb)) () () ...))
   */
  
  if (is_closure(p) || is_closure_star(p) || is_macro(p) || is_bacro(p))
    {
      return(cons(sc, 
		     append_in_place(sc, 
				     list_2(sc, 
					    (is_closure_star(p)) ? sc->LAMBDA_STAR : sc->LAMBDA, 
					    closure_args(p)),
				     closure_body(p)),
		     closure_environment(p)));
    }
  
  if (is_pws(p))
    return(pws_source(sc, p));
  return(sc->NIL);
}


static s7_pointer g_procedure_source(s7_scheme *sc, s7_pointer args)
{
  /* make it look like a scheme-level lambda */
  s7_pointer p;
  
  #define H_procedure_source "(procedure-source func) tries to return the definition of func"
  
  p = car(args);

  if (is_symbol(p))
    {
      p = s7_symbol_value(sc, p);
      if (p == sc->UNDEFINED)
	return(s7_error(sc, sc->WRONG_TYPE_ARG, 
			list_2(sc, make_protected_string(sc, "procedure-source arg, '~S, is unbound"), car(args))));
    }
  if (is_c_function(p))
    return(sc->NIL);

  if (is_c_macro(p))
    return(s7_wrong_type_arg_error(sc, "procedure-source", 0, p, "a scheme macro"));
  if ((!is_procedure(p)) &&
      (!is_macro(p)) &&
      (!is_bacro(p)))
    return(s7_wrong_type_arg_error(sc, "procedure-source", 0, car(args), "a procedure or a macro"));

  if (is_closure(p) || is_closure_star(p) || is_macro(p) || is_bacro(p))
    {
      s7_pointer body;
      body = closure_body(p);
      if (is_safe_closure(body))
	clear_safe_closure(body);
      return(append_in_place(sc, 
			     list_2(sc, 
				    (is_closure_star(p)) ? sc->LAMBDA_STAR : sc->LAMBDA, 
				    closure_args(p)),
			     body));
    }

  if (is_pws(p))
    return(pws_source(sc, p));
  return(sc->NIL);
}


s7_pointer s7_procedure_environment(s7_pointer p)    
{ 
  if (is_closure(p) || is_closure_star(p) || is_macro(p) || is_bacro(p))
    return(closure_environment(p));
  return(p); /* we need an s7_scheme arg! */
}


static s7_pointer g_procedure_environment(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;
  #define H_procedure_environment "(procedure-environment func) tries to return func's environment"

  /* this procedure gives direct access to a function's closure -- see s7test.scm 
   *   for some wild examples.  At least it provides a not-too-kludgey way for several functions
   *   to share a closure.
   */ 
  
  p = car(args);
  if (is_symbol(p))
    {
      p = s7_symbol_value(sc, p);
      if (p == sc->UNDEFINED)
	return(s7_error(sc, sc->WRONG_TYPE_ARG, 
			list_2(sc, make_protected_string(sc, "procedure-environment arg, '~S, is unbound"), car(args))));
    }

  if ((typeflag(p) & (T_PROCEDURE | T_ANY_MACRO)) == 0)
    return(s7_wrong_type_arg_error(sc, "procedure-environment", 0, car(args), "a procedure or a macro"));

  if ((is_closure(p) || is_closure_star(p) || is_macro(p) || is_bacro(p)) &&
      (is_not_null(closure_environment(p))))
    return(closure_environment(p));
  return(sc->global_env);
}


void s7_define_function(s7_scheme *sc, const char *name, s7_function fnc, int required_args, int optional_args, bool rest_arg, const char *doc)
{
  s7_pointer func;
  func = s7_make_function(sc, name, fnc, required_args, optional_args, rest_arg, doc);
  s7_define(sc, sc->NIL, make_symbol(sc, name), func);
}


void s7_define_safe_function(s7_scheme *sc, const char *name, s7_function fnc, int required_args, int optional_args, bool rest_arg, const char *doc)
{
  s7_pointer func;
  func = s7_make_function(sc, name, fnc, required_args, optional_args, rest_arg, doc);
  typeflag(func) |= T_SAFE_PROCEDURE;
  s7_define(sc, sc->NIL, make_symbol(sc, name), func);
}


static void s7_define_constant_function(s7_scheme *sc, const char *name, s7_function fnc, int required_args, int optional_args, bool rest_arg, const char *doc)
{
  s7_pointer func, sym;
  sym = make_symbol(sc, name);
  func = s7_make_function(sc, name, fnc, required_args, optional_args, rest_arg, doc);
  s7_define(sc, sc->NIL, sym, func);
  set_immutable(slot_symbol(global_slot(sym)));
  {
    s7_pointer p;
    p = s7_symbol_value(sc, make_symbol(sc, name));
    typeflag(p) &= ~(T_SAFE_PROCEDURE);
  }
}


void s7_define_macro(s7_scheme *sc, const char *name, s7_function fnc, int required_args, int optional_args, bool rest_arg, const char *doc)
{
  s7_pointer func;
  func = s7_make_function(sc, name, fnc, required_args, optional_args, rest_arg, doc);
  set_type(func, T_C_MACRO | T_ANY_MACRO | T_DONT_EVAL_ARGS); /* this used to include T_PROCEDURE */
  s7_define(sc, sc->NIL, make_symbol(sc, name), func);
}


bool s7_is_macro(s7_scheme *sc, s7_pointer x)
{
  if ((is_macro(x)) || (is_bacro(x)) || (is_c_macro(x)))
    return(true);

  if (is_symbol(x))
    {
      x = s7_symbol_value(sc, x);
      return(is_macro(x) || is_bacro(x) || is_c_macro(x));
    }
  return(false);
}


static s7_pointer g_is_macro(s7_scheme *sc, s7_pointer args)
{
  #define H_is_macro "(macro? arg) returns #t if 'arg' is a macro"
  return(make_boolean(sc, s7_is_macro(sc, car(args))));
  /* it would be more consistent (with procedure? for example) if this returned #f for a symbol,
   *   but fully-expand expects this version.
   */
}


void s7_define_function_star(s7_scheme *sc, const char *name, s7_function fnc, const char *arglist, const char *doc)
{
  /* make an internal function of any args that calls fnc, then wrap it in define* and use eval_c_string */
  /* should (does) this ignore :key and other such noise? */

  char *internal_function, *internal_arglist;
  int arglist_len, len, args;
  const char *local_sym;
  s7_pointer local_args;
  
  arglist_len = safe_strlen(arglist);
  internal_arglist = (char *)calloc(arglist_len + 64, sizeof(char));
  snprintf(internal_arglist, arglist_len + 64, "(map (lambda (arg) (if (symbol? arg) arg (car arg))) '(%s))", arglist);
  local_args = s7_eval_c_string(sc, internal_arglist);
  free(internal_arglist);
  
  args = safe_list_length(sc, local_args); /* since local_args is from map (above), I think it can't be improper */
  internal_arglist = s7_object_to_c_string(sc, local_args);
  /* this has an opening paren which we don't want */
  internal_arglist[0] = ' ';

  local_sym = symbol_name(s7_gensym(sc, "define*"));
  s7_define_function(sc, local_sym, fnc, args, 0, 0, NULL);

  len = 32 + 2 * arglist_len + safe_strlen(doc) + safe_strlen(name) + safe_strlen(local_sym);
  internal_function = (char *)calloc(len, sizeof(char));
  snprintf(internal_function, len, "(define* (%s %s) \"%s\" (%s %s)", name, arglist, doc, local_sym, internal_arglist);
  s7_eval_c_string(sc, internal_function);

  free(internal_function);
  free(internal_arglist);
}


const char *s7_procedure_documentation(s7_scheme *sc, s7_pointer x)
{
  if (is_symbol(x))
    x = s7_symbol_value(sc, x); /* this is needed by Snd */

  if ((s7_is_function(x)) ||
      (is_c_macro(x)))
    return((char *)c_function_documentation(x));
  
  if (((is_closure(x)) || 
       (is_closure_star(x))) &&
      (s7_is_string(car(closure_body(x)))))
    return(s7_string(car(closure_body(x))));
  
  if (is_pws(x))
    return(pws_documentation(x));
  
  if ((s7_is_macro(sc, x)) &&
      (s7_is_string(caddr(cadr(car(closure_body(x)))))))
    return(s7_string(caddr(cadr(car(closure_body(x))))));

  return(""); /* not NULL here so that (string=? "" (procedure-documentation no-doc-func)) -> #t */
}


static s7_pointer g_procedure_documentation(s7_scheme *sc, s7_pointer args)
{
  #define H_procedure_documentation "(procedure-documentation func) returns func's documentation string"
  s7_pointer p;

  p = car(args);
  if (is_symbol(p))
    p = s7_symbol_value(sc, p);

  if ((!is_procedure(p)) &&
      (!s7_is_macro(sc, p)))
    return(s7_wrong_type_arg_error(sc, "procedure-documentation", 0, car(args), "a procedure"));
  return(s7_make_string(sc, s7_procedure_documentation(sc, p)));
}


const char *s7_help(s7_scheme *sc, s7_pointer obj)
{
  if (is_syntax(obj))
    {
      switch (syntax_opcode(obj))
	{
	case OP_QUOTE:
	  return("(quote obj) returns obj unevaluated.  'obj is an abbreviation for (quote obj).");
	  
	case OP_IF:
	  return("(if expr true-stuff optional-false-stuff) evaluates expr, then \
if it is true, evaluates true-stuff; otherwise, if optional-false-stuff exists, it is evaluated.");
	  
	case OP_BEGIN:
	  return("(begin ...) evaluates each form in its body, returning the value of the last one");

	case OP_SET:
	  return("(set! variable value) sets the value of variable to value.");

	case OP_LET:
	  return("(let ((var val)...) ...) binds each variable to its initial value, then evaluates its body,\
returning the value of the last form.  The let variables are local to it, and in the case of let (as opposed to let* for example), \
the variables are not available for use until all have been initialized.");

	case OP_LET_STAR:
	  return("(let* ((var val)...) ...) binds each variable to its initial value, then evaluates its body,\
returning the value of the last form.  The let* variables are local to it, and are available immediately (unlike let, for example).");

	case OP_LETREC:
	  return("(letrec ((var (lambda ...)))...) is like let, but var can refer to itself in \
its value (i.e. you can define local recursive functions)");

	case OP_COND:
	  return("(cond (expr clause...)...) is like if..then.  Each expr is evaluated in \
order, and if one is not #f, the associated clauses are evaluated, whereupon cond returns.");

	case OP_AND:
	  return("(and expr expr ...) evaluates each of its arguments in order, quitting (and returning #f) as soon \
as one of them returns #f.  If all are non-#f, it returns the last value.");

	case OP_OR:
	  return("(or expr expr ...) evaluates each of its argments in order, quitting as soon as \
one of them is not #f.  If all are #f, or returns #f.");

	case OP_CASE:
	  return("(case val ((key...) clause...)...) looks for val in the various lists of keys, \
and if a match is found (via eqv?), the associated clauses are evaluated, and case returns.");

	case OP_DO:
	  return("(do (vars...) (loop control and return value) ...) is a do-loop.");

	case OP_WITH_ENV:
	  return("(with-environment env ...) evaluates its body in the environment env.");

	case OP_LAMBDA:
	  return("(lambda args ...) returns a function.");

	case OP_LAMBDA_STAR:
	  return("(lambda* args ...) returns a function; the args list can have default values, \
the parameters themselves can be accessed via keywords.");

	case OP_DEFINE:
	  return("(define var val) assigns val to the variable (symbol) var.  (define (func args) ...) is \
shorthand for (define func (lambda args ...))");

	case OP_DEFINE_STAR:
	  return("(define* (func args) ...) defines a function with optional/keyword arguments.");

	case OP_DEFINE_CONSTANT:
	  return("(define-constant var val) defines var to be a constant (it can't be set or bound), with the value val.");

	case OP_DEFMACRO:
	  return("(defmacro mac (args) ...) defines mac to be a macro.");

	case OP_DEFMACRO_STAR:
	  return("(defmacro* mac (args) ...) defines mac to be a macro with optional/keyword arguments.");

	case OP_DEFINE_MACRO:
	  return("(define-macro (mac args) ...) defines mac to be a macro.");

	case OP_DEFINE_MACRO_STAR:
	  return("(define-macro* (mac args) ...) defines mac to be a macro with optional/keyword arguments.");

	case OP_DEFINE_EXPANSION:
	  return("(define-expansion (mac args) ...) defines mac to be a read-time macro.");

	case OP_DEFINE_BACRO:
	  return("(define-bacro (mac args) ...) defines mac to be a bacro.");

	case OP_DEFINE_BACRO_STAR:
	  return("(define-bacro* (mac args) ...) defines mac to be a bacro with optional/keyword arguments.");
	}
      
      return(NULL);

      /* others: macroexpand
            :(define-macro (hiho a) "a test" `(+ 1 ,a))
            hiho
            :(hiho 2)
            3
            :(s7-help hiho)
            "a test"
            with-sound #f etc -- all macros at least
       */
    }

  if (is_symbol(obj))
    {
      /* here look for name */
    obj = s7_symbol_value(sc, obj);
    }

  if ((typeflag(obj) & (T_ANY_MACRO | T_PROCEDURE)) != 0)
    return(s7_procedure_documentation(sc, obj));

  if ((is_hook(obj)) &&
      (s7_is_string(hook_documentation(obj))))
    return(s7_string(hook_documentation(obj)));

  /* if is string, apropos? (can scan symbol table)
   */
  /* here keep a table as in xen.c? need s7_help and s7_set_help in C + maybe s7_define_constant_with_documentation
   */

  return(NULL);
}


static s7_pointer g_help(s7_scheme *sc, s7_pointer args)
{
  #define H_help "(help obj) returns obj's documentation"
  const char *doc;
  doc = s7_help(sc, car(args));
  if (!doc)
    return(sc->F);
  return(s7_make_string(sc, doc));
}


s7_pointer s7_procedure_arity(s7_scheme *sc, s7_pointer x)
{
  if (is_c_function(x))
    return(list_3(sc, 
		  make_integer(sc, c_function_required_args(x)),
		  make_integer(sc, c_function_optional_args(x)),
		  make_boolean(sc, c_function_has_rest_arg(x))));
  /* this was optimized to be a permanent immutable list, but makes a mess of all the setters,
   *   and then there are cases like: (fill! (append (list 1) (procedure-arity abs)) 0)
   */

  if ((is_closure(x)) ||
      (is_closure_star(x)) ||
      (is_pair(x)))
    {
      int len;
      
      if (is_pair(x))
	len = s7_list_length(sc, car(x));
      else 
	{
	  if (is_symbol(closure_args(x)))
	    return(list_3(sc, small_int(0), small_int(0), sc->T));
	  len = s7_list_length(sc, closure_args(x));
	}

      if (is_closure_star(x))
	{
	  s7_pointer tmp;        /* make sure we aren't counting :optional and friends as arguments */
	  int opts = 0;

	  if (is_pair(x))
	    tmp = car(x);
	  else tmp = closure_args(x);

	  for (; is_pair(tmp); tmp = cdr(tmp))
	    {
	      if ((car(tmp) == sc->KEY_KEY) ||
		  (car(tmp) == sc->KEY_OPTIONAL) ||
		  (car(tmp) == sc->KEY_ALLOW_OTHER_KEYS))
		opts++;
	      if (car(tmp) == sc->KEY_REST)
		{
		  opts += 2;     /* both :rest and the arg name are not counted as optional args */
		  if (len > 0) len = -len;
		}
	    }
	  return(list_3(sc, small_int(0), make_integer(sc, abs(len) - opts), make_boolean(sc, len < 0)));
	}

      return(list_3(sc, make_integer(sc, abs(len)), small_int(0), make_boolean(sc, len < 0)));
    }
  
  if (is_pws(x))
    {
      if (is_not_null(s7_procedure_getter(sc, x)))
	return(s7_procedure_arity(sc, s7_procedure_getter(sc, x)));
      return(pws_arity(sc, x));
    }

  if ((object_is_applicable(x)) ||
      (s7_is_continuation(x)) ||
      (is_goto(x)))
    return(list_3(sc, small_int(0), small_int(0), sc->T));

  /* it's not straightforward to add support for macros here -- the arity from the
   *   user's point of view refers to the embedded lambda, not the outer lambda.
   */
  return(sc->NIL);
}


static s7_pointer g_procedure_arity(s7_scheme *sc, s7_pointer args)
{
  #define H_procedure_arity "(procedure-arity func) returns a list describing func's arguments: '(required optional rest)"
  s7_pointer p;

  p = car(args);
  if (is_symbol(p))
    {
      p = s7_symbol_value(sc, p);
      if (p == sc->UNDEFINED)
	return(s7_error(sc, sc->WRONG_TYPE_ARG, 
			list_2(sc, make_protected_string(sc, "procedure-arity arg, '~S, is unbound"), car(args))));
    }

  if (!is_procedure(p))
    return(s7_wrong_type_arg_error(sc, "procedure-arity", 0, car(args), "a procedure"));
  return(s7_procedure_arity(sc, p));
}


static s7_pointer closure_name(s7_scheme *sc, s7_pointer closure)
{
  s7_pointer x;

  x = find_local_symbol(sc, closure_environment(closure), sc->__FUNC__);  /* returns nil if no __func__ */
  if (is_slot(x))
    {
      x = slot_value(x);
      if (is_symbol(x))
	return(x);
      if ((is_pair(x)) &&
	  (is_symbol(car(x))))
	return(car(x));
    }

  if (is_pair(sc->cur_code))
    return(sc->cur_code);

  return(caar(closure)); /* desperation -- this is the parameter list */
}

/* (define* (hi (a 1) (b 2)) a) (hi :b 1 2) */
/* (let ((x (lambda* ((a 1) (b 2)) a))) (x :b 1 2)) */




/* -------------------------------- new types -------------------------------- */

typedef struct {
  int type;
  const char *name;
  char *(*print)(s7_scheme *sc, void *value);
  void (*free)(void *value);
  bool (*equal)(void *val1, void *val2);
  bool (*scheme_equal)(s7_scheme *sc, void *val1, void *val2);
  void (*gc_mark)(void *val);
  s7_pointer (*apply)(s7_scheme *sc, s7_pointer obj, s7_pointer args);
  s7_pointer (*set)(s7_scheme *sc, s7_pointer obj, s7_pointer args);
  s7_pointer (*length)(s7_scheme *sc, s7_pointer obj);
  s7_pointer (*copy)(s7_scheme *sc, s7_pointer obj);
  s7_pointer (*reverse)(s7_scheme *sc, s7_pointer obj);
  s7_pointer (*fill)(s7_scheme *sc, s7_pointer obj, s7_pointer args);
  s7_pointer print_func, equal_func, getter_func, setter_func, length_func, copy_func, reverse_func, fill_func;
} s7_object_t;


static s7_object_t *object_types = NULL;
static int object_types_size = 0;
static int num_types = 0;

static char *fallback_print(s7_scheme *sc, void *val) /* obj is object_value(s7_pointer_obj) */
{
  return(copy_string("#<unprintable object>"));
}

static void fallback_free(void *value)
{
}

static bool fallback_equal(void *val1, void *val2)
{
  return(val1 == val2);
}

static s7_pointer fallback_ref(s7_scheme *sc, s7_pointer obj, s7_pointer args)
{
  return(apply_error(sc, obj, args));
}

static s7_pointer fallback_set(s7_scheme *sc, s7_pointer obj, s7_pointer args)
{
  return(eval_error(sc, "attempt to set ~A?", obj));
}

static s7_pointer fallback_length(s7_scheme *sc, s7_pointer obj)
{
  return(eval_error(sc, "attempt to get length of ~A?", obj));
}


int s7_new_type(const char *name, 
		char *(*print)(s7_scheme *sc, void *value), 
		void (*free)(void *value), 
		bool (*equal)(void *val1, void *val2),
		void (*gc_mark)(void *val),
                s7_pointer (*apply)(s7_scheme *sc, s7_pointer obj, s7_pointer args),
                s7_pointer (*set)(s7_scheme *sc, s7_pointer obj, s7_pointer args))
{
  int tag;
  tag = num_types++;
  if (tag >= object_types_size)
    {
      if (object_types_size == 0)
	{
	  object_types_size = 8;
	  object_types = (s7_object_t *)calloc(object_types_size, sizeof(s7_object_t));
	}
      else
	{
	  object_types_size = tag + 8;
	  object_types = (s7_object_t *)realloc((void *)object_types, object_types_size * sizeof(s7_object_t));
	}
    }
  object_types[tag].type = tag;
  object_types[tag].name = copy_string(name);

  if (free)
    object_types[tag].free = free;
  else object_types[tag].free = fallback_free;

  if (print)
    object_types[tag].print = print;
  else object_types[tag].print = fallback_print;

  if (equal)
    object_types[tag].equal = equal;
  else object_types[tag].equal = fallback_equal;

  object_types[tag].scheme_equal = NULL;
  object_types[tag].gc_mark = gc_mark;

  if (apply)
    object_types[tag].apply = apply;
  else object_types[tag].apply = fallback_ref;

  if (set)
    object_types[tag].set = set;
  else object_types[tag].set = fallback_set;

  object_types[tag].length = fallback_length;
  object_types[tag].copy = NULL;
  object_types[tag].reverse = NULL;
  object_types[tag].fill = NULL;
  object_types[tag].print_func = NULL;
  object_types[tag].length_func = NULL;
  object_types[tag].equal_func = NULL;
  object_types[tag].getter_func = NULL;
  object_types[tag].setter_func = NULL;
  object_types[tag].copy_func = NULL;
  object_types[tag].reverse_func = NULL;
  object_types[tag].fill_func = NULL;
  return(tag);
}


int s7_new_type_x(const char *name, 
		  char *(*print)(s7_scheme *sc, void *value), 
		  void (*free)(void *value), 
		  bool (*equal)(void *val1, void *val2),
		  void (*gc_mark)(void *val),
		  s7_pointer (*apply)(s7_scheme *sc, s7_pointer obj, s7_pointer args),
		  s7_pointer (*set)(s7_scheme *sc, s7_pointer obj, s7_pointer args),
		  s7_pointer (*length)(s7_scheme *sc, s7_pointer obj),
		  s7_pointer (*copy)(s7_scheme *sc, s7_pointer obj),
		  s7_pointer (*reverse)(s7_scheme *sc, s7_pointer obj),
		  s7_pointer (*fill)(s7_scheme *sc, s7_pointer obj, s7_pointer val))
{
  int tag;
  tag = s7_new_type(name, print, free, equal, gc_mark, apply, set);

  if (length)
    object_types[tag].length = length;
  else object_types[tag].length = fallback_length;

  object_types[tag].copy = copy;
  object_types[tag].reverse = reverse;
  object_types[tag].fill = fill;
  return(tag);
}


static char *object_print(s7_scheme *sc, s7_pointer a)
{
  return((*(object_types[object_type(a)].print))(sc, object_value(a))); /* assume allocation here (so we'll free the string later) */
}


static void free_object(s7_pointer a)
{
  (*(object_types[object_type(a)].free))(object_value(a));
}


static bool objects_are_equal(s7_scheme *sc, s7_pointer a, s7_pointer b)
{
  if (object_type(a) == object_type(b))
    {
      int tag;
      tag = object_type(a);
      if (object_types[tag].equal)
	return((*(object_types[tag].equal))(object_value(a), object_value(b)));
      if (object_types[tag].scheme_equal)
	return((*(object_types[tag].scheme_equal))(sc, object_value(a), object_value(b)));
      return(a == b);
    }
  return(false);
}


static void mark_embedded_objects(s7_pointer a) /* called by gc, calls fobj's mark func */
{
  int tag;
  tag = object_type(a);
  if (tag < num_types)
    {
      if (object_types[tag].gc_mark)
	(*(object_types[tag].gc_mark))(object_value(a));
    }
}


static bool object_is_applicable(s7_pointer x)
{
  return((is_c_object(x)) &&
	 (object_types[object_type(x)].apply));
}


void *s7_object_value(s7_pointer obj)
{
  return(object_value(obj));
}


int s7_object_type(s7_pointer obj)
{
  if (is_c_object(obj))
    return(object_type(obj));
  return(-1);
}


static int pws_tag;

s7_pointer s7_make_object(s7_scheme *sc, int type, void *value)
{
  s7_pointer x;

  NEW_CELL(sc, x);
  object_type(x) = type;
  object_value(x) = value;
  set_type(x, T_C_OBJECT); /* free_object checks that the free function exists */

  if (object_types[type].apply)
    {
      object_ref(x) = object_types[type].apply;
      if (object_ref(x) != fallback_ref)
	{
	  if (type != pws_tag)
	    typeflag(x) |= (T_PROCEDURE | T_SAFE_PROCEDURE);
	  else typeflag(x) |= T_PROCEDURE;
	}
    }
  else object_ref(x) = fallback_ref;

  if (object_types[type].set)
    object_set(x) = object_types[type].set;
  else object_set(x) = fallback_set;

  add_c_object(sc, x);
  return(x);
}


static s7_pointer object_length(s7_scheme *sc, s7_pointer obj)
{
  return((*(object_types[object_type(obj)].length))(sc, obj));
}


static s7_pointer object_copy(s7_scheme *sc, s7_pointer obj)
{
  int tag;
  tag = object_type(obj);
  if (object_types[tag].copy)
    return((*(object_types[tag].copy))(sc, obj));

  return(eval_error(sc, "attempt to copy ~A?", obj));
}


static s7_pointer object_reverse(s7_scheme *sc, s7_pointer obj)
{
  int tag;
  tag = object_type(obj);
  if (object_types[tag].reverse)
    return((*(object_types[tag].reverse))(sc, obj));

  return(eval_error(sc, "attempt to reverse ~A?", obj));
}




/* ---------------- scheme-level new types ---------------- */

typedef struct {
  int type;
  s7_pointer value;
} s_type_t;

/* this just holds the type tag locally!  Very bad planning!
 */

static char *call_s_object_print(s7_scheme *sc, void *value)
{
  /* value here is the s_type_t object, the (scheme) function to call is object_types[tag].print_func */
  /*   it will be passed the value, not the original object */

  s_type_t *obj = (s_type_t *)value;
  car(sc->s_function_args) = obj->value;
  return(copy_string((char *)s7_string(s7_call(sc, object_types[obj->type].print_func, sc->s_function_args))));
  /* object_print assumes the value returned here can be freed */
}


static  bool s_type_equal(s7_scheme *sc, void *a, void *b)
{
  /* this is the fallback if no equal func is specified */
  s_type_t *obj1 = (s_type_t *)a;
  s_type_t *obj2 = (s_type_t *)b;
  return(s7_is_equal(sc, obj1->value, obj2->value));
}


static bool call_s_object_equal(s7_scheme *sc, void *a, void *b)
{
  /* we get here if an equal func was specified in make-type */
  s_type_t *obj1 = (s_type_t *)a;
  s_type_t *obj2 = (s_type_t *)b;
  return(s7_boolean(sc, s7_call(sc, object_types[obj1->type].equal_func, list_2(sc, obj1->value, obj2->value))));
}


static s7_pointer call_s_object_getter(s7_scheme *sc, s7_pointer a, s7_pointer args)
{
  /* still accessible via for-each */
  s_type_t *obj;
  obj = (s_type_t *)s7_object_value(a);
  return(s7_call(sc, object_types[obj->type].getter_func, cons(sc, obj->value, args))); /* ?? */
}


static s7_pointer call_s_object_setter(s7_scheme *sc, s7_pointer a, s7_pointer args)
{
  /* still accessible via reverse, for-each */
  s_type_t *obj;
  obj = (s_type_t *)s7_object_value(a);
  return(s7_call(sc, object_types[obj->type].setter_func, cons(sc, obj->value, args))); /* ?? */
}


/* generalized set! calls g_internal_object_set which then calls the object's set function 
 */

static s7_pointer g_internal_object_set(s7_scheme *sc, s7_pointer args)
{
  return((*(object_set(car(args))))(sc, car(args), cdr(args)));
}


#define SAVE_X_Y_Z(X, Y, Z)	     \
  do {                               \
    X = ((is_null(sc->x)) ? -1 : s7_gc_protect(sc, sc->x));	\
    Y = ((is_null(sc->y)) ? -1 : s7_gc_protect(sc, sc->y));	\
    Z = ((is_null(sc->z)) ? -1 : s7_gc_protect(sc, sc->z));	\
  } while (0)

#define RESTORE_X_Y_Z(X, Y, Z)                \
  do {                                        \
    if (X == -1) sc->x = sc->NIL; else {sc->x = s7_gc_protected_at(sc, X); s7_gc_unprotect_at(sc, X);} \
    if (Y == -1) sc->y = sc->NIL; else {sc->y = s7_gc_protected_at(sc, Y); s7_gc_unprotect_at(sc, Y);} \
    if (Z == -1) sc->z = sc->NIL; else {sc->z = s7_gc_protected_at(sc, Z); s7_gc_unprotect_at(sc, Z);} \
    } while (0)

static s7_pointer call_s_object_length(s7_scheme *sc, s7_pointer a)
{
  s_type_t *obj;
  s7_pointer result;
  int save_x = -1, save_y = -1, save_z = -1;

  obj = (s_type_t *)s7_object_value(a);
  car(sc->s_function_args) = obj->value;
  SAVE_X_Y_Z(save_x, save_y, save_z);
  result = s7_call(sc, object_types[obj->type].length_func, sc->s_function_args);
  RESTORE_X_Y_Z(save_x, save_y, save_z);

  return(result);
}


/* s7_call in this context can lead to segfaults:
 *
 *    (call-with-exit (lambda (exit) (length ((cadr (make-type :length (lambda (a) (exit 32)))) 1))))
 *      [partly fixed; still callable via object_reverse and applicable_length]
 * 
 *    (call-with-exit (lambda (exit) (object->string ((cadr (make-type :print (lambda (a) (exit 32)))) 1))))
 *      [hard to fix -- very low level access to the method (atom_to_c_string)]
 *
 *    (call-with-exit (lambda (exit) (copy ((cadr (make-type :copy (lambda (a) (exit 32)))) 1))))
 *      [called in s7_copy, g_copy calls s7_copy]
 *      [hard to fix because hash-tables use s7_copy -- needs at least expansion of g_copy]
 *      [  and in g_copy we'd need another operator OP_S7_MAKE_OBJECT maybe, to handle the ]
 *      [  result = s7_make_object(sc, new_obj->type, (void *)new_obj) business after the  ]
 *      [  value has been copied.]
 *
 *    (call-with-exit (lambda (exit) (fill! ((cadr (make-type :fill (lambda (a n) (exit 32)))) 1) 0)))
 *      [fixed]
 *
 *    (call-with-exit (lambda (exit) (let ((typ (make-type :equal (lambda (a n) (exit 32))))) (equal? ((cadr typ) 1) ((cadr typ) 1)))))
 *      [callable via s7_is_equal and s7_is_equal_ci]
 *      [hard to fix: g_is_equal calls s7_is_equal, but here I think we could split out s_object_equal if equal_func exists]
 *      [  but that requires a separate copy of s7_is_equal etc]
 *
 * the *#readers*, *unbound-variable-hook* funcs have the same problem [symbol-bind]
 *   [reader funcs are s7_call'ed in check_sharp_readers called from make_sharp_constant]
 */ 


static s7_pointer call_s_object_copy(s7_scheme *sc, s7_pointer a)
{
  s_type_t *obj, *new_obj;

  obj = (s_type_t *)s7_object_value(a);
  car(sc->s_function_args) = obj->value;

  new_obj = (s_type_t *)calloc(1, sizeof(s_type_t));
  new_obj->type = obj->type;

  new_obj->value = s7_call(sc, object_types[new_obj->type].copy_func, sc->s_function_args);
  return(s7_make_object(sc, new_obj->type, (void *)new_obj));
}


static s7_pointer call_s_object_reverse(s7_scheme *sc, s7_pointer a)
{
  s_type_t *obj, *new_obj;

  obj = (s_type_t *)s7_object_value(a);
  car(sc->s_function_args) = obj->value;

  new_obj = (s_type_t *)calloc(1, sizeof(s_type_t));
  new_obj->type = obj->type;

  new_obj->value = s7_call(sc, object_types[new_obj->type].reverse_func, sc->s_function_args);
  return(s7_make_object(sc, new_obj->type, (void *)new_obj));
}


static s7_pointer call_s_object_fill(s7_scheme *sc, s7_pointer a, s7_pointer val)
{
  /* I think this is no longer accessible */
  s_type_t *obj;
  obj = (s_type_t *)s7_object_value(a);
  return(s7_call(sc, object_types[obj->type].fill_func, list_2(sc, obj->value, val)));
}


static char *s_type_print(s7_scheme *sc, void *val)
{
  /* object_print assumes the string is allocated here */
  s_type_t *obj = (s_type_t *)val;
  char *str, *full_str;
  int len, tag;
  
  tag = obj->type;
  str = s7_object_to_c_string(sc, obj->value);
  len = safe_strlen(str) + safe_strlen(object_types[tag].name) + 16;
  full_str = (char *)calloc(len, sizeof(char));
  snprintf(full_str, len, "#<%s %s>", object_types[tag].name, str);
  free(str);
  return(full_str);
}


static void s_type_free(void *val)
{
  free(val);
}


static void s_type_gc_mark(void *val)
{
  s_type_t *obj = (s_type_t *)val;
  s7_mark_object(obj->value);
}


static s7_pointer s_is_type(s7_scheme *sc, s7_pointer args)
{
  if (is_c_object(cadr(args)))
    {
      s_type_t *obj;
      obj = (s_type_t *)s7_object_value(cadr(args));
      return(s7_make_boolean(sc, obj->type == s7_integer(car(args))));
    }
  return(sc->F);
}


static s7_pointer s_type_make(s7_scheme *sc, s7_pointer args)
{
  s_type_t *obj;
  obj = (s_type_t *)calloc(1, sizeof(s_type_t));
  obj->type = s7_integer(car(args));
  obj->value = cadr(args);
  return(s7_make_object(sc, obj->type, (void *)obj));
}


static s7_pointer s_type_ref(s7_scheme *sc, s7_pointer args)
{
  s7_pointer x;
  int tag;

  if (s7_is_integer(car(args)))
    {
      tag = s7_integer(car(args));
      x = cadr(args);
      if (is_c_object(x))
	{
	  s_type_t *obj;
	  obj = (s_type_t *)s7_object_value(x);
	  if (obj->type == tag)
	    return(obj->value);
	}

      return(s7_error(sc, sc->WRONG_TYPE_ARG, 
		      list_4(sc, 
			     make_protected_string(sc, "~A type's 'ref' function argument, ~S, is ~A?"),
			     make_protected_string(sc, object_types[tag].name),
			     x,
			     make_protected_string(sc, type_name(x)))));
    }
  return(sc->F); /* someone has completely messed up */
}


static s7_pointer g_make_type(s7_scheme *sc, s7_pointer args)
{
  #define H_make_type "(make-type print equal getter setter length name copy reverse fill) returns a new type object, \
a list of three functions: ?, make, and ref.  The '?' function returns #t if passed an argument of the new type, \
the 'make' function creates a new object of that type, and the 'ref' function returns the value of that object.\
The optional arguments to make-type are functions that specify how objects of the new type display themselves (print, 1 argument), \
check for equality (equal, 2 args, both will be of the new type), apply themselves to arguments, (getter, any number \
of args, see vector for an example), respond to the generalized set! and length generic functions, and finally, \
one special case: name sets the type name (a string), which only matters if you're not specifying the print function. \
In each case, the argument is the value of the object, not the object itself."

  int tag;

  tag = s7_new_type("anonymous-type", s_type_print, s_type_free, NULL, s_type_gc_mark, NULL, NULL);
  object_types[tag].equal = NULL;
  object_types[tag].scheme_equal = s_type_equal;

  if (is_not_null(args))
    {
      int i, args_loc = -1, func_loc = -1;
      s7_pointer x;

      args_loc = s7_gc_protect(sc, args);

      /* if any of the special functions are specified, store them in the type object so we can find them later.
       *    they also need to be GC-protected:
       *    (let ((ctr ((cadr (make-type :getter (lambda (a b) b))))))
       *      (gc)
       *      ;; any reference here to the getter is likely to fail if it hasn't been protected
       */
      for (i = 0, x = args; is_not_null(x); i++, x = cdr(x))
	{
	  s7_pointer func, proc_args;
	  int nargs = 0;
	  bool rest_arg = false;

	  /* the closure_star mechanism passes the args in declaration order */
	  func = car(x);
	  if (func != sc->F)            /* #f means arg was not set */
	    {
	      if (i != 5)
		{
		  if (!is_procedure(func))
		    {
		      s7_gc_unprotect_at(sc, args_loc);
		      return(s7_error(sc, sc->WRONG_TYPE_ARG, 
				      list_2(sc, 
					     make_protected_string(sc, "make-type arg, ~A, should be a function"),
					     func)));
		    }
		  func_loc = s7_gc_protect(sc, func); /* this ought to be faster in the mark phase than checking every function field of every scheme type(?) */
		  proc_args = s7_procedure_arity(sc, func);
		  nargs = s7_integer(car(proc_args)) + s7_integer(cadr(proc_args));
		  rest_arg = (caddr(proc_args) != sc->F);
		}

	      switch (i)
		{
		case 0:                 /* print, ((cadr (make-type :print (lambda (a) (format #f "#<typo: ~S>" a)))) "gypo") -> #<typo: "gypo"> */
		  if ((s7_integer(car(proc_args)) > 1) || 
		      ((nargs == 0) && (!rest_arg)))
		    {
		      s7_gc_unprotect_at(sc, args_loc);
		      if (func_loc != -1) s7_gc_unprotect_at(sc, func_loc);
		      return(s7_error(sc, sc->WRONG_TYPE_ARG, 
				      list_2(sc, make_protected_string(sc, "make-type :print procedure, ~A, should take one argument"), func)));
		    }

		  object_types[tag].print_func = func;
		  object_types[tag].print = call_s_object_print;
		  break;

		case 1:                 /* equal */
		  /* (let ((typo (make-type :equal (lambda (a b) (equal? a b))))) (let ((a ((cadr typo) 123)) (b ((cadr typo) 321))) (equal? a b))) */
		  if ((s7_integer(car(proc_args)) > 2) || 
		      ((nargs < 2) && (!rest_arg)))
		    {
		      s7_gc_unprotect_at(sc, args_loc);
		      if (func_loc != -1) s7_gc_unprotect_at(sc, func_loc);
		      return(s7_error(sc, sc->WRONG_TYPE_ARG, 
				      list_2(sc, make_protected_string(sc, "make-type :equal procedure, ~A, should take two arguments"), func)));
		    }
		  object_types[tag].equal = NULL;
		  object_types[tag].equal_func = func;
		  object_types[tag].scheme_equal = call_s_object_equal;
		  break;

		case 2:                 /* getter: (((cadr (make-type :getter (lambda (a b) (vector-ref a b)))) (vector 1 2 3)) 1) -> 2 */
		  if ((nargs == 0) && (!rest_arg))
		    {
		      s7_gc_unprotect_at(sc, args_loc);
		      if (func_loc != -1) s7_gc_unprotect_at(sc, func_loc);
		      return(s7_error(sc, sc->WRONG_TYPE_ARG, 
				      list_2(sc, make_protected_string(sc, "make-type :getter procedure, ~A, should take at least one argument"), func)));
		    }
		  object_types[tag].getter_func = func;
		  object_types[tag].apply = call_s_object_getter;
		  break;

		case 3:                 /* setter: (set! (((cadr (make-type :setter (lambda (a b c) (vector-set! a b c)))) (vector 1 2 3)) 1) 23) */
		  if ((nargs < 2) && (!rest_arg))
		    {
		      s7_gc_unprotect_at(sc, args_loc);
		      if (func_loc != -1) s7_gc_unprotect_at(sc, func_loc);
		      return(s7_error(sc, sc->WRONG_TYPE_ARG, 
				      list_2(sc, make_protected_string(sc, "make-type :setter procedure, ~A, should take at least two arguments"), func)));
		    }
		  object_types[tag].setter_func = func;
		  object_types[tag].set = call_s_object_setter;
		  break;

		case 4:                 /* length: (length ((cadr (make-type :length (lambda (a) (vector-length a)))) (vector 1 2 3))) -> 3 */
		  if ((s7_integer(car(proc_args)) > 1) || 
		      ((nargs == 0) && (!rest_arg)))
		    {
		      s7_gc_unprotect_at(sc, args_loc);
		      if (func_loc != -1) s7_gc_unprotect_at(sc, func_loc);
		      return(s7_error(sc, sc->WRONG_TYPE_ARG, 
				      list_2(sc, make_protected_string(sc, "make-type :length procedure, ~A, should take at one argument"), func)));
		    }

		  object_types[tag].length_func = func;
		  object_types[tag].length = call_s_object_length;
		  break;

		case 5:                 /* name, ((cadr (make-type :name "hiho")) 123) -> #<hiho 123> */
		  if (!s7_is_string(func))
		    {
		      s7_gc_unprotect_at(sc, args_loc);
		      if (func_loc != -1) s7_gc_unprotect_at(sc, func_loc);
		      return(s7_error(sc, sc->WRONG_TYPE_ARG, 
				      list_2(sc, make_protected_string(sc, "make-type :name arg, ~S, should be a string"), func)));
		    }
		  object_types[tag].name = copy_string(s7_string(func));
		  break;

		case 6:                 /* copy */
		  if ((s7_integer(car(proc_args)) > 1) || 
		      ((nargs == 0) && (!rest_arg)))
		    {
		      s7_gc_unprotect_at(sc, args_loc);
		      if (func_loc != -1) s7_gc_unprotect_at(sc, func_loc);
		      return(s7_error(sc, sc->WRONG_TYPE_ARG, 
				      list_2(sc, make_protected_string(sc, "make-type :copy procedure, ~A, should take at one argument"), func)));
		    }
		  object_types[tag].copy_func = func;
		  object_types[tag].copy = call_s_object_copy;
		  break;

		case 7:                 /* reverse */
		  if ((s7_integer(car(proc_args)) > 1) || 
		      ((nargs == 0) && (!rest_arg)))
		    {
		      s7_gc_unprotect_at(sc, args_loc);
		      if (func_loc != -1) s7_gc_unprotect_at(sc, func_loc);
		      return(s7_error(sc, sc->WRONG_TYPE_ARG, 
				      list_2(sc, make_protected_string(sc, "make-type :reverse procedure, ~A, should take at one argument"), func)));
		    }
		  object_types[tag].reverse_func = func;
		  object_types[tag].reverse = call_s_object_reverse;
		  break;

		case 8:                 /* fill */
		  if ((s7_integer(car(proc_args)) > 2) || 
		      ((nargs < 2) && (!rest_arg)))
		    {
		      s7_gc_unprotect_at(sc, args_loc);
		      if (func_loc != -1) s7_gc_unprotect_at(sc, func_loc);
		      return(s7_error(sc, sc->WRONG_TYPE_ARG, 
				      list_2(sc, make_protected_string(sc, "make-type :fill procedure, ~A, should take at two arguments"), func)));
		    }
		  object_types[tag].fill_func = func;
		  object_types[tag].fill = call_s_object_fill;
		  break;
		}
	    }
	}
      s7_gc_unprotect_at(sc, args_loc);
    }

  {
    s7_pointer result;
    int result_loc;
    result = list_3(sc, sc->NIL, sc->NIL, sc->NIL);
    result_loc = s7_gc_protect(sc, result);

    /* ? method: (lambda (arg) (s_is_type tag arg)) 
     *     returns #t if arg is of the new type
     */
    car(result) = make_closure(sc, list_1(sc, sc->S_TYPE_ARG),
			       list_1(sc, list_3(sc, sc->S_IS_TYPE, make_integer(sc, tag), sc->S_TYPE_ARG)),
			       T_CLOSURE);

    /* make method: (lambda* (arg) (s_type_make tag arg))
     *   returns an object of the new type with its value specified by arg (defaults to #f)
     */
    cadr(result) = make_closure(sc, list_1(sc, sc->S_TYPE_ARG),
				list_1(sc, list_3(sc, sc->S_TYPE_MAKE, make_integer(sc, tag), sc->S_TYPE_ARG)),
				T_CLOSURE_STAR);

    /* ref method: (lambda (arg) (s_type_ref arg))
     *   returns the value passed to make above 
     */
    caddr(result) = make_closure(sc, list_1(sc, sc->S_TYPE_ARG),
				 list_1(sc, list_3(sc, sc->S_TYPE_REF, make_integer(sc, tag), sc->S_TYPE_ARG)),
				 T_CLOSURE);
    s7_gc_unprotect_at(sc, result_loc);
    return(result);
  }
}

/* here it would be neat if we allowed any keywords, and those not handled explicitly could
 *    be added to the methods list under the key-word->symbol name.  define* without the need
 *    to state in advance what keys -- undeclared key would be bound in the func env under its
 *    name and value -- define+?  -- the extra args would be in an alist accessible under
 *    the rest arg name?  
 */




/* -------- procedure-with-setter -------- */

s7_pointer s7_make_procedure_with_setter(s7_scheme *sc, 
					 const char *name,
					 s7_pointer (*getter)(s7_scheme *sc, s7_pointer args), 
					 int get_req_args, int get_opt_args,
					 s7_pointer (*setter)(s7_scheme *sc, s7_pointer args),
					 int set_req_args, int set_opt_args,
					 const char *documentation)
{
  /* on the C side, we merely define the two functions, and set the getter's setter, so there
   *   is no separate pws object.  
   */
  s7_pointer get_func, set_func;
  char *internal_set_name;
  int len;

  len = 16 + safe_strlen(name);
  internal_set_name = (char *)calloc(len, sizeof(char));
  snprintf(internal_set_name, len, "[set-%s]", name);

  get_func = s7_make_function(sc, name, getter, get_req_args, get_opt_args, false, documentation); 
  typeflag(get_func) |= T_SAFE_PROCEDURE;
  s7_define(sc, sc->NIL, make_symbol(sc, name), get_func);
  set_func = s7_make_function(sc, internal_set_name, setter, set_req_args, set_opt_args, false, documentation); 
  typeflag(set_func) |= T_SAFE_PROCEDURE;
  s7_define(sc, sc->NIL, make_symbol(sc, internal_set_name), set_func);
  c_function_setter(get_func) = set_func;

  return(get_func);
}
  

typedef struct {
  s7_pointer (*getter)(s7_scheme *sc, s7_pointer args);
  int get_req_args, get_opt_args;
  s7_pointer (*setter)(s7_scheme *sc, s7_pointer args);
  int set_req_args, set_opt_args;
  s7_pointer scheme_getter;
  s7_pointer scheme_setter;
  char *documentation;
  char *name;
} s7_pws_t;


static bool is_pws(s7_pointer obj)
{
  return((is_c_object(obj)) &&
	 (object_type(obj) == pws_tag));
}


static char *pws_print(s7_scheme *sc, void *obj)
{
  s7_pws_t *f = (s7_pws_t *)obj;
  if (f->name)
    return(copy_string(f->name));
  return(copy_string((char *)"#<procedure-with-setter>"));
}


static void pws_free(void *obj)
{
  s7_pws_t *f = (s7_pws_t *)obj;
  if (f)
    {
      if (f->documentation)
	free(f->documentation);
      if (f->name)
	free(f->name);
      free(f);
    }
}


static void pws_mark(void *val)
{
  s7_pws_t *f = (s7_pws_t *)val;
  S7_MARK(f->scheme_getter);
  S7_MARK(f->scheme_setter);
}


static bool pws_equal(void *obj1, void *obj2)
{
  return(obj1 == obj2);
}


static s7_pointer pws_apply(s7_scheme *sc, s7_pointer obj, s7_pointer args)
{
  /* this is called as the pws object apply method, not as the actual getter */
  s7_pws_t *f;

  f = (s7_pws_t *)s7_object_value(obj);
  if (f->getter != NULL)
    {
      /* this is the c_function case */
      int len;

      len = safe_list_length(sc, args);
      if (len < f->get_req_args)
	return(s7_error(sc, sc->WRONG_NUMBER_OF_ARGS, 
			list_3(sc, sc->NOT_ENOUGH_ARGUMENTS, obj, args)));

      if (len > (f->get_req_args + f->get_opt_args))
	return(s7_error(sc, sc->WRONG_NUMBER_OF_ARGS, 
			list_3(sc, sc->TOO_MANY_ARGUMENTS, obj, args)));

      return((*(f->getter))(sc, args));
    }

  push_stack(sc, OP_APPLY, args, f->scheme_getter);
  return(sc->F);
}


static s7_pointer pws_set(s7_scheme *sc, s7_pointer obj, s7_pointer args)
{
  /* this is the pws set method, not the actual setter */
  s7_pws_t *f;

  f = (s7_pws_t *)s7_object_value(obj);
  if (f->setter != NULL)
    {
      /* this is the c_function case */
      int len;

      len = safe_list_length(sc, args);
      if (len < f->set_req_args)
	return(s7_error(sc, sc->WRONG_NUMBER_OF_ARGS, 
			list_3(sc, sc->NOT_ENOUGH_ARGUMENTS, obj, args)));

      if (len > (f->set_req_args + f->set_opt_args))
	return(s7_error(sc, sc->WRONG_NUMBER_OF_ARGS, 
			list_3(sc, sc->TOO_MANY_ARGUMENTS, obj, args)));

      return((*(f->setter))(sc, args));
    }

  push_stack(sc, OP_APPLY, args, f->scheme_setter);
  return(sc->F);
}


static s7_pointer pws_arity(s7_scheme *sc, s7_pointer obj)
{
  s7_pws_t *f;
  f = (s7_pws_t *)s7_object_value(obj);
  return(list_3(sc, make_integer(sc, f->get_req_args), make_integer(sc, f->get_opt_args), sc->F));
}


static s7_pointer g_make_procedure_with_setter(s7_scheme *sc, s7_pointer args)
{
  #define H_make_procedure_with_setter "(make-procedure-with-setter getter setter) combines its \
two function arguments as a procedure-with-setter.  The 'getter' is called unless the procedure \
occurs as the object of set!."

  s7_pointer p, getter, setter, arity;
  s7_pws_t *f;
  int gc_loc;
  /* the two args should be functions, the setter taking one more arg than the getter */

  getter = car(args);
  if (!is_procedure(getter))
    return(s7_wrong_type_arg_error(sc, "make-procedure-with-setter getter,", 1, getter, "a procedure"));
  setter = cadr(args);
  if (!is_procedure(setter))
    return(s7_wrong_type_arg_error(sc, "make-procedure-with-setter setter,", 2, setter, "a procedure"));

  f = (s7_pws_t *)calloc(1, sizeof(s7_pws_t));
  p = s7_make_object(sc, pws_tag, (void *)f);
  typeflag(p) |= T_PROCEDURE;
  gc_loc = s7_gc_protect(sc, p);

  f->scheme_getter = getter;
  f->scheme_setter = setter;

  arity = s7_procedure_arity(sc, getter); /* calls cons -- might trigger GC, so setter/getter fields need to be set for s7_mark */
  if (is_pair(arity))
    f->get_req_args = s7_integer(car(arity));
  f->documentation = copy_string(s7_procedure_documentation(sc, getter)); /* pws might be GC'd whereupon the doc string is freed */
  
  arity = s7_procedure_arity(sc, setter);
  if (is_pair(arity))
    f->set_req_args = s7_integer(car(arity));
  
  s7_gc_unprotect_at(sc, gc_loc);
  return(p);
}


bool s7_is_procedure_with_setter(s7_pointer obj)
{
  /* C-side make-pws does not create an object -- it is a function with a setter, but
   *   that include car -- should procedure-with-setter be #t of it?
   * but here in s7 we're using this procedure with the assumption that it is a pws-object 
   */
  return((is_pws(obj)) ||
	 ((is_c_function(obj)) &&
	  (is_c_function(c_function_setter(obj)))));
}


s7_pointer s7_procedure_getter(s7_scheme *sc, s7_pointer obj)
{
  if (is_c_function(obj))
    return(obj);

  if (is_pws(obj))
    {
      s7_pws_t *f;
      f = (s7_pws_t *)s7_object_value(obj);
      return(f->scheme_getter);
    }
  
  return(sc->NIL);
}


s7_pointer s7_procedure_setter(s7_scheme *sc, s7_pointer obj)
{
  if (is_c_function(obj))     /* this includes pws via ffi */
    return(c_function_setter(obj));

  if (is_pws(obj))
    {
      s7_pws_t *f;
      f = (s7_pws_t *)s7_object_value(obj);
      return(f->scheme_setter);
    }

  return(sc->NIL);
}


static s7_pointer g_procedure_setter(s7_scheme *sc, s7_pointer args)
{
  #define H_procedure_setter "(procedure-setter obj) returns the setter associated with obj, or #f"
  return(s7_procedure_setter(sc, car(args)));
}

/* if this were settable, 
 *    (define (make-procedure-with-setter g s) (set! (procedure-setter g) s) g)
 *    (define (procedure-with-setter? g) (procedure? (procedure-setter g)))
 *
 * for c-function both sides? there's 
 *   c_function_setter(getter) = setter;
 * for scheme both sides
 *   getter->scheme_setter = setter;
 * but this assumes getter has a pws struct already -- perhaps append after env in closure list?
 *
 * pws is really aimed at C FFI but there we have C both sides!
 *   in scheme, it's currently used in with-sound for wsdat refs and in generators for field access
 *              dlocsig for fields, enved for channel-envelope, examp for cursor stuff,
 *              extensions for channel-sync, and similarly elsewhere.
 *
 * hooks are also clumsy and not the right thing, but aimed at C
 *   could hooks be environments?  watchers, or-hooks, error (cerror) handlers
 */


static s7_pointer g_is_procedure_with_setter(s7_scheme *sc, s7_pointer args)
{
  #define H_is_procedure_with_setter "(procedure-with-setter? obj) returns #t if obj is a procedure-with-setter"
  return(make_boolean(sc, s7_is_procedure_with_setter(car(args))));
}


static char *pws_documentation(s7_pointer x)
{
  s7_pws_t *f = (s7_pws_t *)s7_object_value(x);
  return(f->documentation);
}


static s7_pointer pws_source(s7_scheme *sc, s7_pointer x)
{
  s7_pws_t *f;
  f = (s7_pws_t *)s7_object_value(x);
  if ((is_closure(f->scheme_getter)) ||
      (is_closure_star(f->scheme_getter)))
    return(append_in_place(sc, 
			   list_2(sc,
				  (is_closure(f->scheme_getter)) ? sc->LAMBDA : sc->LAMBDA_STAR,
				  closure_args(f->scheme_getter)),
			   closure_body(f->scheme_getter)));
  return(sc->NIL);
}


void s7_define_function_with_setter(s7_scheme *sc, const char *name, s7_function get_fnc, s7_function set_fnc, int req_args, int opt_args, const char *doc)
{
  s7_make_procedure_with_setter(sc, name, get_fnc, req_args, opt_args, set_fnc, req_args + 1, opt_args, doc);
}


const char *s7_procedure_name(s7_scheme *sc, s7_pointer proc)
{
  switch (type(proc))
    {
    case T_CLOSURE:
    case T_CLOSURE_STAR:
      return(c_closure_name(sc, proc));
  
    case T_C_OPT_ARGS_FUNCTION:
    case T_C_RST_ARGS_FUNCTION:
    case T_C_LST_ARGS_FUNCTION:
    case T_C_ANY_ARGS_FUNCTION:
    case T_C_FUNCTION:
      return(c_function_name(proc));

    case T_C_MACRO:
      return(c_macro_name(proc));

    case T_SYMBOL:
      {
	s7_pointer slot; 
	slot = find_symbol(sc, proc); /* we don't want unbound-variable errors here */
	if ((!is_null(slot)) &&
	    (proc != slot_value(slot)) &&       /* (procedure-name :hi) -> infinite loop */
	    (is_procedure(slot_value(slot))))   /* a symbol -> symbol -> back to 1st?? */
	  return(s7_procedure_name(sc, slot_value(slot)));
	return(NULL);
      }

    case T_C_OBJECT:
      if (object_type(proc) == pws_tag)
	{
	  s7_pws_t *f;
	  f = (s7_pws_t *)object_value(proc);
	  return(f->name);
	}
      return(NULL);

      /* T_MACRO et all don't have an easily accessible name -- I suppose we could
       *   search the current environment?
       */
    }
  return(NULL);
}


static s7_pointer g_procedure_name(s7_scheme *sc, s7_pointer args)
{
  #define H_procedure_name "(procedure-name proc) returns the name of the procedure or closure proc, if it can be found."
  return(s7_make_string(sc, s7_procedure_name(sc, car(args))));
}


static bool args_match(s7_scheme *sc, s7_pointer x, int args)
{
  switch (type(x))
    {
    case T_C_ANY_ARGS_FUNCTION:
    case T_C_OPT_ARGS_FUNCTION:
    case T_C_RST_ARGS_FUNCTION:
    case T_C_LST_ARGS_FUNCTION:
    case T_C_FUNCTION:
      return(((int)c_function_required_args(x) <= args) &&
	     ((int)c_function_all_args(x) >= args));

    case T_CLOSURE:
      return((is_symbol(closure_args(x))) ||
	     (safe_list_length(sc, closure_args(x)) == args));

    case T_CLOSURE_STAR:
    return((is_symbol(closure_args(x))) ||
	   (safe_list_length(sc, closure_args(x)) >= args));

    case T_C_OBJECT:
      if (object_type(x) == pws_tag)
	{
	  s7_pws_t *f;
	  if (is_not_null(s7_procedure_getter(sc, x))) /* a scheme function in this case */
	    return(args_match(sc, s7_procedure_getter(sc, x), 2));

	  f = (s7_pws_t *)s7_object_value(x);	  
	  return((f->get_req_args <= args) &&
		 ((f->get_req_args + f->get_opt_args) >= args));
	}
      break;

    case T_STRING:
    case T_HASH_TABLE:
    case T_ENVIRONMENT:
      return(args == 1);
      
    case T_PAIR:
      return(true);

    case T_VECTOR:
      if (vector_is_multidimensional(x))
	return(args <= (int)vector_ndims(x));
      return(args == 1);
    }
  return(false);
}


static bool is_thunk(s7_scheme *sc, s7_pointer x)
{
  switch (type(x))
    {
    case T_C_FUNCTION:
      return(c_function_all_args(x) == 0);

    case T_CLOSURE:
    case T_CLOSURE_STAR:
      return(is_null(closure_args(x)));

    case T_C_OBJECT:
      if (object_type(x) == pws_tag)
	{
	  s7_pws_t *f;
	  if (is_not_null(s7_procedure_getter(sc, x))) /* a scheme function in this case */
	    return(is_thunk(sc, s7_procedure_getter(sc, x)));

	  f = (s7_pws_t *)s7_object_value(x);	  
	  return((f->get_req_args == 0) &&
		 (f->get_opt_args == 0));
	}
    }
  return(false);
}




/* -------------------------------- symbol-access ------------------------------------------------ */
/*
 * originally in Snd I wanted notification when a variable was set, and it's not very pretty
 *      to have to use pws's everywhere.  Here (s7) we have constants and tracing.
 *
 * these are in the same realm:
 *   typed var: restrict set! to the desired type or do auto-conversions
 *   constant: disallow set!
 *   traced var: report either read/write
 *   keywords: can't set or bind, always return self as value
 *
 * and related (but much messier to implement):
 *   fluid-let: dynamic until exit (call/cc error normal)
 *   dynamic variables: insist any ref is to the current binding [dynamic-let]
 *
 * a value wrapper or transparent object won't work:
 * (define-macro (trace sym)
 *   `(set! ,sym (wrap ,sym :setter (lambda (binding a) (format #t "~A set to ~A~%" (car binding) a)))))
 * 
 * (define-macro (untrace sym)
 *   `(set! ,sym ,sym) ??? -- oops...
 *
 * (symbol-access sym) -- a pws, if set! it affects the current binding actions.
 *   the actions are local to the current environment, so
 *   we automatically undo the local accessors when leaving the current scope.
 *   a list of 3 funcs: getter setter binder -- rest is ignored so that
 *   trace can save the current accessors in cadddr of the symbol-access list, 
 *   untrace uses that to restore the old form, etc
 *
 * (define (notify-if-set var notifier) ; returns #t if it's ok to set
 *   (set! (symbol-access) 
 *         (list #f (lambda (symbol new-value) (or (notifier symbol new-value) new-value)) #f)))
 */


s7_pointer s7_symbol_access(s7_scheme *sc, s7_pointer sym)
{
  if (symbol_has_accessor(sym))
    return(s7_gc_protected_at(sc, symbol_accessor(sym)));
  return(sc->F);
}


s7_pointer s7_symbol_set_access(s7_scheme *sc, s7_pointer symbol, s7_pointer funcs)
{
  if (symbol_has_accessor(symbol))
    s7_gc_unprotect_at(sc, symbol_accessor(symbol));

  if ((is_pair(funcs)) &&
      (s7_list_length(sc, funcs) >= 3) &&
      ((is_procedure(car(funcs))) ||
       (is_procedure(cadr(funcs))) ||
       (is_procedure(caddr(funcs)))))
    symbol_accessor(symbol) = s7_gc_protect(sc, funcs);
  else symbol_accessor(symbol) = -1;
  return(funcs);
}


static s7_pointer g_symbol_get_access(s7_scheme *sc, s7_pointer args)
{
  #define H_symbol_access "(symbol-access sym) is a procedure-with-setter that adds or removes controls on how a \
symbol accesses its current binding."

  if (!is_symbol(car(args)))
    return(s7_wrong_type_arg_error(sc, "symbol-access,", 0, car(args), "a symbol"));
  return(s7_symbol_access(sc, car(args)));
}


static s7_pointer g_symbol_set_access(s7_scheme *sc, s7_pointer args)
{
  s7_pointer sym, funcs;
  sym = car(args);
  if (!is_symbol(sym))
    return(s7_wrong_type_arg_error(sc, "set! symbol-access,", 1, sym, "a symbol"));

  funcs = cadr(args);
  if (funcs != sc->F)
    {
      if ((!is_pair(funcs)) ||
	  (s7_list_length(sc, funcs) != 3))
	return(s7_wrong_type_arg_error(sc, "set! symbol-access,", 2, funcs, "a list of 3 settings"));	
      if ((is_procedure(car(funcs))) && (!args_match(sc, car(funcs), 2)))
	return(s7_wrong_type_arg_error(sc, "set! symbol-access get function,", 2, car(funcs), "a procedure of 2 arguments"));	
      if ((is_procedure(cadr(funcs))) && (!args_match(sc, cadr(funcs), 2)))
	return(s7_wrong_type_arg_error(sc, "set! symbol-access set function,", 2, cadr(funcs), "a procedure of 2 arguments"));	
      if ((is_procedure(caddr(funcs))) && (!args_match(sc, caddr(funcs), 2)))
	return(s7_wrong_type_arg_error(sc, "set! symbol-access bind function,", 2, caddr(funcs), "a procedure of 2 arguments"));	
    }
  return(s7_symbol_set_access(sc, sym, funcs));
}


static s7_pointer call_symbol_bind(s7_scheme *sc, s7_pointer symbol, s7_pointer new_value)
{
  s7_pointer func;
  func = caddr(s7_gc_protected_at(sc, symbol_accessor(symbol)));
  if (is_procedure(func))
    {
      s7_pointer old_value;
      old_value = new_value;

      /* this happens in contexts that are tricky to implement with a clean use of the evaluator stack (as in
       *   the parallel symbol set case), so we need to use s7_call.  But if an uncaught error
       *   occurs in s7_call, the error handler marches up the stack looking for a catch, unwinding the stack
       *   past the point of the call.  In the worst case, we can segfault because any subsequent pop_stack
       *   (i.e. an unchecked goto START), walks off the start of the stack. 
       */
      
      if (is_c_function(func))
	{
	  car(sc->T2_1) = symbol;
	  car(sc->T2_2) = new_value;
      	  new_value = c_function_call(func)(sc, sc->T2_1);
	}
      else
	{
	  bool old_off;
	  old_off = sc->gc_off;
	  sc->gc_off = true;
	  new_value = s7_call(sc, func, list_2(sc, symbol, new_value));
	  sc->gc_off = old_off;
	}
      if (new_value == sc->ERROR)
	return(s7_error(sc, sc->ERROR,
			list_3(sc, make_protected_string(sc, "can't bind ~S to ~S"), symbol, old_value)));
    }
  return(new_value);
}


static s7_pointer bind_accessed_symbol(s7_scheme *sc, opcode_t op, s7_pointer symbol, s7_pointer new_value)
{
  s7_pointer func;
  func = caddr(s7_gc_protected_at(sc, symbol_accessor(symbol)));
  if (is_procedure(func))
    {
      if (is_c_function(func))
	{
	  s7_pointer old_value;
	  old_value = new_value;
	  car(sc->T2_1) = symbol;
	  car(sc->T2_2) = new_value;
	  new_value = c_function_call(func)(sc, sc->T2_1);
	  if (new_value == sc->ERROR)
	    return(s7_error(sc, sc->ERROR,
			    list_3(sc, make_protected_string(sc, "can't bind ~S to ~S"), symbol, old_value)));
	}
      else
	{
	  sc->args = list_2(sc, symbol, new_value);
	  push_stack(sc, op, sc->args, sc->code);
	  sc->code = func;
	  return(sc->NO_VALUE);
	}
    }
  return(new_value);
}


void *s7_symbol_accessor_data(s7_pointer sym)
{
  return(symbol_accessor_data(sym));
}

void s7_symbol_set_accessor_data(s7_pointer sym, void *val)
{
  symbol_accessor_data(sym) = val;
}


#if WITH_OPTIMIZATION

bool s7_in_safe_do(s7_scheme *sc)
{
  return(sc->safe_do_level > 0);
}

static void safe_do_set_id(s7_scheme *sc, long long int id)
{
  if (sc->safe_do_level >= sc->safe_do_ids_size)
    {
      sc->safe_do_ids_size = sc->safe_do_level * 2;
      sc->safe_do_ids = (long long int *)realloc(sc->safe_do_ids, sc->safe_do_ids_size * sizeof(long long int));
    }
  sc->safe_do_ids[sc->safe_do_level] = id;
}

bool s7_is_do_local(s7_scheme *sc, s7_pointer symbol)
{
  return(sc->safe_do_ids[sc->safe_do_level] == symbol_id(symbol));
}

bool s7_is_do_local_or_global(s7_scheme *sc, s7_pointer symbol)
{
  return(sc->safe_do_ids[sc->safe_do_level] >= symbol_id(symbol));
}

bool s7_is_do_global(s7_scheme *sc, s7_pointer symbol)
{
  return(sc->safe_do_ids[sc->safe_do_level] > symbol_id(symbol));
}

#else

bool s7_in_safe_do(s7_scheme *sc)
{
  return(false);
}

bool s7_is_do_local(s7_scheme *sc, s7_pointer symbol)
{
  return(false);
}

bool s7_is_do_local_or_global(s7_scheme *sc, s7_pointer symbol)
{
  return(false);
}

bool s7_is_do_global(s7_scheme *sc, s7_pointer symbol)
{
  return(false);
}

#endif




/* -------------------------------- hooks -------------------------------- */


static bool is_function_with_arity(s7_pointer x)
{
  /* hook function lists are more restrictive than s7_is_procedure which accepts things like continuations */
  int typ;
  typ = type(x);
  return((typ == T_CLOSURE) || 
	 (typ == T_CLOSURE_STAR) ||
	 (typ >= T_C_FUNCTION) ||
	 (is_pws(x)));
}


static bool function_arity_ok(s7_scheme *sc, s7_pointer hook, s7_pointer func)
{
  /* when a function is added to a hook, we need to check that its arity is compatible
   *   with the hook arity.  The function must accept the hook's required number of
   *   arguments, and optionally accept any optional hook arguments.  If the hook
   *   has a rest argument, the function must have one too.
   */
  s7_pointer func_args, hook_args;
  int hook_req = 0, hook_opt = 0, func_req = 0, func_opt = 0;
  bool hook_rst = false, func_rst = false;

  func_args = s7_procedure_arity(sc, func);
  func_req = s7_integer(car(func_args));

  hook_args = hook_arity(hook);
  hook_req = s7_integer(car(hook_args));

  if (hook_req < func_req) return(false); /* func requires too many args */

  func_rst = is_true(sc, caddr(func_args));
  hook_rst = is_true(sc, caddr(hook_args));

  if (func_rst) return(true);             /* func required args are ok, and it has a rest arg, so it matches */
  if (hook_rst) return(false);            /* func has no rest, hook has rest -- can't be safe */

  /* both rest args are false, hook-req >= func-req */
  func_opt = s7_integer(cadr(func_args));
  hook_opt = s7_integer(cadr(hook_args));

  /* most args hook handles must be <= most func handles */
  if ((hook_req + hook_opt) <= (func_req + func_opt)) return(true);
  return(false);
}


bool s7_is_hook(s7_pointer p)
{
  return(is_hook(p));
}


s7_pointer s7_hook_functions(s7_pointer hook)
{
  return(hook_functions(hook));
}


s7_pointer s7_hook_set_functions(s7_pointer hook, s7_pointer functions)
{
  if (is_pair(functions))
    hook_functions(hook) = functions;
  return(hook_functions(hook));
}


s7_pointer s7_hook_arity(s7_pointer hook)
{
  return(hook_arity(hook));
}


const char *s7_hook_documentation(s7_pointer hook)
{
  return(string_value(hook_documentation(hook)));
}


s7_pointer s7_make_hook(s7_scheme *sc, int required_args, int optional_args, bool rest_arg, const char *documentation) 
{
  /* arg order follows s7_make_function */
  s7_pointer x;
  NEW_CELL(sc, x);
  hook_arity(x) = list_3(sc, 
			 make_integer(sc, required_args), 
			 make_integer(sc, optional_args), 
			 make_boolean(sc, rest_arg));
  hook_functions(x) = sc->NIL;
  hook_documentation(x) = s7_make_string(sc, documentation);
  set_type(x, T_HOOK); /* not sure about this */
  return(x);
}


static s7_pointer hook_copy(s7_scheme *sc, s7_pointer hook)
{
  s7_pointer new_hook, arity;
  int gc_loc;

  arity = hook_arity(hook);
  new_hook = s7_make_hook(sc, s7_integer(car(arity)), s7_integer(cadr(arity)), s7_boolean(sc, caddr(arity)), s7_string(hook_documentation(hook)));
  if (is_null(hook_functions(hook)))
    return(new_hook);

  gc_loc = s7_gc_protect(sc, new_hook);
  hook_functions(new_hook) = copy_list(sc, hook_functions(hook));
  s7_gc_unprotect_at(sc, gc_loc);
  return(new_hook);
}


s7_pointer s7_hook_apply(s7_scheme *sc, s7_pointer hook, s7_pointer args)
{
  if (is_pair(hook_functions(hook)))
    {
      int gc_loc;
      s7_pointer x;
      gc_loc = s7_gc_protect(sc, args);
      for (x = hook_functions(hook); is_not_null(x); x = cdr(x))
	s7_call(sc, car(x), args);
      s7_gc_unprotect_at(sc, gc_loc);
    }
  return(sc->UNSPECIFIED);
}


static s7_pointer g_is_hook(s7_scheme *sc, s7_pointer args)
{
  #define H_is_hook "(hook? obj) returns #t if obj is a hook"
  return(make_boolean(sc, is_hook(car(args))));
}


static s7_pointer g_make_hook(s7_scheme *sc, s7_pointer args)
{
  #define H_make_hook "(make-hook (arity (1 0 #f)) (doc \"\")) returns a new hook.  'arity' is a list \
describing the argument list that the hook-functions will see: (list required optional rest). \
It defaults to no arguments: '(0 0 #f).  Any function added to the hook's list has to be compatible \
with the hook arity.  'doc' is a documentation string."

  s7_pointer x;
  NEW_CELL(sc, x);

  if (is_not_null(args))
    {
      s7_pointer arity;
      arity = car(args);
      if (is_pair(arity))
	{
	  s7_Int req, opt;
	  if ((s7_list_length(sc, arity) != 3) ||
	      (!s7_is_integer(car(arity))) ||
	      (!s7_is_integer(cadr(arity))) ||
	      (!s7_is_boolean(caddr(arity))))
	    return(s7_wrong_type_arg_error(sc, "make-hook", (is_null(cdr(args))) ? 0 : 1, arity, "an arity list: (required optional rest)"));
	  req = s7_integer(car(arity));
	  opt = s7_integer(cadr(arity));
	  if ((req < 0) ||
	      (opt < 0))
	    return(s7_wrong_type_arg_error(sc, "make-hook", (is_null(cdr(args))) ? 0 : 1, arity, "number of args can't be negative"));
	  hook_arity(x) = arity; 
	}
      else 
	{
	  /* backwards compatibility -- this used to be just an integer => required args */
	  if (s7_is_integer(arity))
	    {
	      if (s7_integer(arity) < 0)
		return(s7_wrong_type_arg_error(sc, "make-hook", (is_null(cdr(args))) ? 0 : 1, arity, "a non-negative integer, or an arity list: (required optional rest)"));
	      hook_arity(x) = list_3(sc, arity, small_int(0), sc->F);
	    }
	  else return(s7_wrong_type_arg_error(sc, "make-hook", (is_null(cdr(args))) ? 0 : 1, arity, "an arity list: (required optional rest)"));
	}

      if (is_not_null(cdr(args)))
	{
	  s7_pointer doc;
	  doc = cadr(args);
	  if (s7_is_string(doc))
	    hook_documentation(x) = doc;
	  else return(s7_wrong_type_arg_error(sc, "make-hook", 2, doc, "a string"));
	}
      else hook_documentation(x) = s7_make_string(sc, "");
    }
  else 
    {
      hook_arity(x) = list_3(sc, small_int(0), small_int(0), sc->F);
      hook_documentation(x) = s7_make_string(sc, "");
    }

  hook_functions(x) = sc->NIL;

  set_type(x, T_HOOK);
  return(x);
}


static s7_pointer g_hook(s7_scheme *sc, s7_pointer args)
{
  #define H_hook "(hook ...) returns a new hook object with its arguments (all functions) \
as the initial hook-functions list, and taking its arity from those functions.  This is a \
convenient short-hand similar to (vector ...) or (list ...).  The hook arity is that of the \
first function in the list, or '(0 0 #f) if there are no functions.  All the other functions \
must be compatible with the arity of the first."

  s7_pointer x, hook;
  int gc_loc;
  
  if (is_null(args))
    return(s7_make_hook(sc, 0, 0, false, NULL));

  for (x = args; is_pair(x); x = cdr(x))
    if (!is_function_with_arity(car(x)))
      return(s7_wrong_type_arg_error(sc, "hook", position_of(x, args), car(x), "a function"));

  hook = g_make_hook(sc, cons(sc, s7_procedure_arity(sc, car(args)), sc->NIL));
  hook_functions(hook) = args;
  gc_loc = s7_gc_protect(sc, hook);
  
  for (x = cdr(args); is_pair(x); x = cdr(x))
    if (!function_arity_ok(sc, hook, car(x)))
      {
	s7_gc_unprotect_at(sc, gc_loc);
	return(s7_wrong_type_arg_error(sc, "hook", position_of(x, args), car(x), "compatible function"));
      }
  s7_gc_unprotect_at(sc, gc_loc);
  return(hook);
}


static s7_pointer g_hook_functions(s7_scheme *sc, s7_pointer args)
{
  #define H_hook_functions "(hook-functions hook) returns the list of functions on the hook. \
It is settable;  (set! (hook-functions hook) (cons func (hook-functions hook))) adds func \
to the current list."

  if (!is_hook(car(args)))
    return(s7_wrong_type_arg_error(sc, "hook-functions", 0, car(args), "a hook"));

  return(hook_functions(car(args)));
}


static s7_pointer g_hook_set_functions(s7_scheme *sc, s7_pointer args)
{
  s7_pointer hook, funcs;
  hook = car(args);

  if (!is_hook(hook))
    return(s7_wrong_type_arg_error(sc, "hook-functions", 1, hook, "a hook"));

  funcs = cadr(args);
  if (!s7_is_list(sc, funcs))
    return(s7_wrong_type_arg_error(sc, "hook-functions", 2, funcs, "a list of functions or '()"));

  if (is_pair(funcs))
    {
      s7_pointer x, y;
      bool cdr_y = false;
      for (x = funcs, y = funcs; is_pair(x); x = cdr(x))
	{
	  if (!is_function_with_arity(car(x)))
	    return(s7_wrong_type_arg_error(sc, "hook-functions", 2, funcs, "a list of functions"));
	  if (!function_arity_ok(sc, hook, car(x)))
	    return(s7_wrong_type_arg_error(sc, "hook-functions", 2, funcs, "a list of functions of the correct arity"));

	  if (cdr_y)
	    {
	      if (x == y)
		return(s7_wrong_type_arg_error(sc, "hook-functions", 2, funcs, "a proper (non-circular) list of functions"));
	      y = cdr(y);
	    }
	  cdr_y = !cdr_y;
	}
      if (is_not_null(x))
	return(s7_wrong_type_arg_error(sc, "hook-functions", 2, funcs, "a proper list of functions"));
    }

  hook_functions(hook) = funcs;
  return(funcs);
}


static s7_pointer g_hook_arity(s7_scheme *sc, s7_pointer args)
{
  #define H_hook_arity "(hook-arity hook) returns the hook's arity, a list giving the number \
of required arguments, optional arguments, and whether there is a rest argument.  Each function \
on the hook's function list has to be applicable to a list of arguments compatible with this description."

  if (!is_hook(car(args)))
    return(s7_wrong_type_arg_error(sc, "hook-arity", 0, car(args), "a hook"));

  return(hook_arity(car(args)));
}


static s7_pointer g_hook_documentation(s7_scheme *sc, s7_pointer args)
{
  #define H_hook_documentation "(hook-documentation hook) returns the documentation associated \
with the hook."

  if (!is_hook(car(args)))
    return(s7_wrong_type_arg_error(sc, "hook-documentation", 0, car(args), "a hook"));

  return(hook_documentation(car(args)));
}


static s7_pointer g_hook_apply(s7_scheme *sc, s7_pointer args)
{
  #define H_hook_apply "(hook-apply hook ...) applies each function in the hook's function \
list to the trailing arguments of hook-apply."

  s7_pointer hook, hook_args;

  hook = car(args);
  if (!is_hook(hook))
    return(s7_wrong_type_arg_error(sc, "hook-apply", 1, hook, "a hook"));

  if (is_null(cdr(args)))
    hook_args = sc->NIL;
  else 
    {
      if (is_null(cddr(args)))
	hook_args = cadr(args);
      else hook_args = apply_list_star(sc, cdr(args));

      if (!is_proper_list(sc, hook_args))        /* (hook-apply + #f) etc */
	return(s7_error(sc, sc->WRONG_TYPE_ARG, 
			list_2(sc, 
			       make_protected_string(sc, "hook-apply's last argument should be a proper list: ~A"),
			       hook_args)));
    }

  if (caddr(hook_arity(hook)) == sc->F)
    {
      int arg_num;
      arg_num = safe_list_length(sc, hook_args);
      if ((arg_num < s7_integer(car(hook_arity(hook)))) ||
	  (arg_num > (s7_integer(car(hook_arity(hook))) + s7_integer(cadr(hook_arity(hook))))))
	return(s7_error(sc, sc->WRONG_NUMBER_OF_ARGS, 
			list_3(sc, 
			       make_protected_string(sc, "hook passed wrong number of args: ~A (arity: ~A)"),
			       hook_args,
			       hook_arity(hook))));
    }

  if (is_pair(hook_functions(hook)))
    push_stack(sc, OP_HOOK_APPLY, hook_args, hook_functions(hook));

  return(sc->UNSPECIFIED);
}


static bool hooks_are_equal(s7_scheme *sc, s7_pointer x, s7_pointer y)
{
  return(s7_is_equal(sc, hook_arity(x), hook_arity(y)) &&
	 s7_is_equal(sc, hook_functions(x), hook_functions(y)));
}


static bool internal_hook_arity_ok(s7_scheme *sc, s7_pointer hook, s7_pointer funcs)
{
  s7_pointer x;
  for (x = funcs; is_pair(x); x = cdr(x))
    if ((!is_function_with_arity(car(x))) ||
	(!function_arity_ok(sc, hook, car(x))))
      return(false);
  return(is_null(x));
}


static s7_pointer g_trace_hook_set(s7_scheme *sc, s7_pointer args)
{
  /* in normal use, we'd (set! (hook-functions *trace-hook*) ...), but for backwards compatibility,
   *   we also need to support (set! *trace-hook* func).
   */
  s7_pointer funcs;
  funcs = cadr(args);
  if (s7_is_list(sc, funcs)) /* pair or null */
    {
      if (internal_hook_arity_ok(sc, sc->trace_hook, funcs))
	hook_functions(sc->trace_hook) = funcs;
      else return(sc->ERROR);
    }
  else
    {
      if (s7_is_procedure(funcs))
	hook_functions(sc->trace_hook) = cons(sc, funcs, sc->NIL);
      else return(sc->ERROR);
    }
  return(sc->trace_hook); /* kinda pointless... */
}


static s7_pointer g_load_hook_set(s7_scheme *sc, s7_pointer args)
{
  /* in normal use, we'd (set! (hook-functions *load-hook*) ...), but for backwards compatibility,
   *   we also need to support (set! *load-hook* func).
   */
  s7_pointer funcs;
  funcs = cadr(args);
  if (s7_is_list(sc, funcs))
    {
      if (internal_hook_arity_ok(sc, sc->load_hook, funcs))
	hook_functions(sc->load_hook) = funcs; 
      else return(sc->ERROR);
    }
  else
    {
      if (s7_is_procedure(funcs))
	hook_functions(sc->load_hook) = cons(sc, funcs, sc->NIL);
      else return(sc->ERROR);
    }
  return(sc->load_hook);
}


static s7_pointer g_unbound_variable_hook_set(s7_scheme *sc, s7_pointer args)
{
  /* in normal use, we'd (set! (hook-functions *unbound-variable-hook*) ...), but for backwards compatibility,
   *   we also need to support (set! *unbound-variable-hook* func).
   */
  s7_pointer funcs;
  funcs = cadr(args);
  if (s7_is_list(sc, funcs))
    {
      if (internal_hook_arity_ok(sc, sc->unbound_variable_hook, funcs))
	hook_functions(sc->unbound_variable_hook) = funcs;
      else return(sc->ERROR);
    }
  else
    {
      if (s7_is_procedure(funcs))
	hook_functions(sc->unbound_variable_hook) = cons(sc, funcs, sc->NIL);
      else return(sc->ERROR);
    }
  return(sc->unbound_variable_hook);
}


static s7_pointer g_error_hook_set(s7_scheme *sc, s7_pointer args)
{
  /* in normal use, we'd (set! (hook-functions *error-hook*) ...), but for backwards compatibility,
   *   we also need to support (set! *error-hook* func).
   */
  s7_pointer funcs;

  funcs = cadr(args);
  if (s7_is_list(sc, funcs))
    {
      if (internal_hook_arity_ok(sc, sc->error_hook, funcs))
	hook_functions(sc->error_hook) = funcs;
      else return(sc->ERROR);
    }
  else
    {
      if (s7_is_procedure(funcs))
	hook_functions(sc->error_hook) = cons(sc, funcs, sc->NIL);
      else return(sc->ERROR);
    }
  return(sc->error_hook);
}






/* -------------------------------- eq etc -------------------------------- */

bool s7_is_eq(s7_pointer obj1, s7_pointer obj2)
{
  return(obj1 == obj2);
}


bool s7_is_eqv(s7_pointer a, s7_pointer b) 
{
  if (a == b) 
    return(true);
  
#if WITH_GMP
  if (big_numbers_are_eqv(a, b))
    return(true);                                     /* T_INTEGER != T_BIG_INTEGER etc */
#endif

  if (type(a) != type(b)) 
    return(false);
  
  if (s7_is_string(a)) 
    return(string_value(a) == string_value(b));
  
  if (s7_is_number(a))
    return(numbers_are_eqv(a, b));
  
  return(false);
}


/* -------- structure equality -------- 
 *
 * equal? examines the entire structure (possibly a tree etc), which might contain
 *   cycles (vector element is the vector etc), so list/vector/hash-table equality
 *   needs to carry along a list of pointers seen so far.
 */

static bool structures_are_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci);

static bool s7_is_equal_tracking_circles(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  if (x == y) 
    return(true);
  
#if WITH_GMP
  if (big_numbers_are_eqv(x, y))
    return(true);
#endif

  if (type(x) != type(y)) 
    return(false);
  
  switch (type(x))
    {
    case T_STRING:
      return(scheme_strings_are_equal(x, y));

    case T_C_OBJECT:
      return(objects_are_equal(sc, x, y));

    case T_INTEGER:
    case T_RATIO:
    case T_REAL:
    case T_COMPLEX:
      return(numbers_are_eqv(x, y));

    case T_COUNTER:
      return(counter_count(x) == counter_count(y)); /* ?? */

    case T_ENVIRONMENT:
    case T_VECTOR:
    case T_HASH_TABLE:
    case T_PAIR:
      return(structures_are_equal(sc, x, y, ci));

    case T_HOOK:
      return(hooks_are_equal(sc, x, y));

    case T_C_POINTER:       /* might have a list of these for example */
      return(raw_pointer(x) == raw_pointer(y));
    }

  return(false); /* we already checked that x != y (port etc) */
}


static bool structures_are_equal(s7_scheme *sc, s7_pointer x, s7_pointer y, shared_info *ci)
{
  /* here we know x and y are pointers to the same type of structure */
  int ref_x, ref_y;

  ref_x = peek_shared_ref(ci, x);
  ref_y = peek_shared_ref(ci, y);

  if ((ref_x != 0) && (ref_y != 0))
    return(ref_x == ref_y);
  
  if ((ref_x != 0) || (ref_y != 0))
    {
      /* try to harmonize the new guy -- there can be more than one structure equal to the current one */
      if (ref_x != 0)
	add_shared_ref(ci, y, ref_x);
      else add_shared_ref(ci, x, ref_y);
    }
  else add_equal_ref(ci, x, y);
  
  /* now compare the elements of the structures. */
  switch (type(x))
    {
    case T_PAIR:
      return((s7_is_equal_tracking_circles(sc, car(x), car(y), ci)) &&
	     (s7_is_equal_tracking_circles(sc, cdr(x), cdr(y), ci)));

    case T_VECTOR:
    case T_HASH_TABLE:
      {
	s7_Int i, len;
	len = vector_length(x);
	if (len != vector_length(y)) return(false);
	
	if (s7_is_vector(x))
	  {
	    /* there's one special case: shared vectors can have 1 dimension but include the dimension info */
	    int x_dims = 1, y_dims = 1, j;
	    
	    if (vector_is_multidimensional(x))
	      x_dims = vector_ndims(x);
	    if (vector_is_multidimensional(y))
	      y_dims = vector_ndims(y);
	    
	    if (x_dims != y_dims)
	      return(false);
	    
	    if (x_dims > 1)
	      for (j = 0; j < x_dims; j++)
		if (vector_dimension(x, j) != vector_dimension(y, j))
		  return(false);
	  }
	
	for (i = 0; i < len; i++)
	  if (!(s7_is_equal_tracking_circles(sc, vector_element(x, i), vector_element(y, i), ci)))
	    return(false);
      }

    case T_ENVIRONMENT:
      {
	/*
	  (let ((e1 #f) (e2 #f))
	    (let ((a 1)) (set! e1 (current-environment)))
            (let ((a 1)) (set! e2 (current-environment))) 
            (equal? e1 e2))
	*/
	s7_pointer ex, ey;
	for (ex = x, ey = y; is_environment(ex) && is_environment(ey); ex = next_environment(ex), ey = next_environment(ey))
	  {
	    s7_pointer px, py;
	    if (ex == ey)
	      return(true);
	    if ((ex == sc->global_env) ||
		(ey == sc->global_env))
	      return(false);
	    /* currently order matters, unlike hash-tables */
	    for (px = environment_slots(ex), py = environment_slots(ey); is_slot(px) && is_slot(py); px = next_slot(px), py = next_slot(py))
	      if ((slot_symbol(px) != slot_symbol(py)) ||
		  (!(s7_is_equal_tracking_circles(sc, slot_value(px), slot_value(py), ci))))
		return(false);
	    if ((is_slot(px)) ||
		(is_slot(py)))
	      return(false);
	  }
	if ((is_environment(ex)) ||
	    (is_environment(ey)))
	  return(false);
	return(true);
      }
    }
  return(true);
}


bool s7_is_equal(s7_scheme *sc, s7_pointer x, s7_pointer y)
{
  if (x == y) 
    return(true);
  
#if WITH_GMP
  if (big_numbers_are_eqv(x, y))
    return(true);
#endif

  if (type(x) != type(y)) 
    return(false);
  
  switch (type(x))
    {
      /* one problematic case: #<unspecified> is currently not equal to (values) but they print the same.
       *   case T_UNTYPED: return((s7_is_unspecified(sc, x)) && (s7_is_unspecified(sc, y)))
       */
    case T_STRING:
      return(scheme_strings_are_equal(x, y));

    case T_C_OBJECT:
      return(objects_are_equal(sc, x, y));

    case T_INTEGER:
    case T_RATIO:
    case T_REAL:
    case T_COMPLEX:
      return(numbers_are_eqv(x, y));

    case T_ENVIRONMENT:
      return(structures_are_equal(sc, x, y, new_shared_info(sc)));

    case T_VECTOR:
      return((vector_length(x) == vector_length(y)) &&
	     (structures_are_equal(sc, x, y, new_shared_info(sc))));

    case T_HASH_TABLE:
      return(hash_tables_are_equal(sc, x, y));

    case T_PAIR:
#if (!WITH_GMP)
      return((type(car(x)) == type(car(y))) &&
	     (structures_are_equal(sc, x, y, new_shared_info(sc))));
#else
      return(structures_are_equal(sc, x, y, new_shared_info(sc)));
#endif

    case T_HOOK:
      return(hooks_are_equal(sc, x, y));

    case T_C_POINTER:
      return(raw_pointer(x) == raw_pointer(y));
    }
  return(false); /* we already checked that x != y (port etc) */
}


static s7_pointer g_is_eq(s7_scheme *sc, s7_pointer args)
{
  #define H_is_eq "(eq? obj1 obj2) returns #t if obj1 is eq to (the same object as) obj2"
  return(make_boolean(sc, car(args) == cadr(args)));
}


static s7_pointer g_is_eqv(s7_scheme *sc, s7_pointer args)
{
  #define H_is_eqv "(eqv? obj1 obj2) returns #t if obj1 is equivalent to obj2"
  return(make_boolean(sc, s7_is_eqv(car(args), cadr(args))));
}


static s7_pointer g_is_equal(s7_scheme *sc, s7_pointer args)
{
  #define H_is_equal "(equal? obj1 obj2) returns #t if obj1 is equal to obj2"
  return(make_boolean(sc, s7_is_equal(sc, car(args), cadr(args))));
}



/* ---------------------------------------- length, copy, fill ---------------------------------------- */

static s7_pointer g_length(s7_scheme *sc, s7_pointer args)
{
  #define H_length "(length obj) returns the length of obj, which can be a list, vector, string, or hash-table. \
The length of a dotted list does not include the final cdr, and is returned as a negative number.  A circular \
list has infinite length."

  s7_pointer lst;
  lst = car(args);

  switch (type(lst))
    {
    case T_PAIR:
      {
	int len;
	len = s7_list_length(sc, lst);
	/* len < 0 -> dotted and (abs len) is length not counting the final cdr
	 * len == 0, circular so length is infinite
	 */
	if (len == 0)
	  return(s7_make_real(sc, INFINITY));
	return(make_integer(sc, len));
      }

    case T_NIL:
      return(small_int(0));

    case T_VECTOR:
      return(g_vector_length(sc, args));

    case T_STRING:
      return(g_string_length(sc, args));

    case T_HASH_TABLE:
      return(g_hash_table_size(sc, args));

    case T_C_OBJECT:
      return(object_length(sc, car(args)));

    case T_ENVIRONMENT:
      return(make_integer(sc, environment_length(sc, car(args))));

    default:
      return(s7_wrong_type_arg_error(sc, "length", 0, lst, "a list, vector, string, or hash-table"));
    }
  
  return(small_int(0));
}

/* what about (length file)?  input port, read_file gets the file length, so perhaps save it
 */


static s7_pointer list_copy(s7_scheme *sc, s7_pointer x, s7_pointer y, bool step)
{
  if ((!is_pair(x)) ||
       (x == y))
    return(x);
  return(cons(sc, car(x), list_copy(sc, cdr(x), (step) ? cdr(y) : y, !step)));
}


static s7_pointer s7_copy(s7_scheme *sc, s7_pointer obj)
{
  switch (type(obj))
    {
    case T_STRING:
      return(s7_make_string_with_length(sc, string_value(obj), string_length(obj)));

    case T_C_OBJECT:
      return(object_copy(sc, obj));

    case T_HASH_TABLE:              /* this has to copy nearly everything */
      return(hash_table_copy(sc, obj));
      
    case T_ENVIRONMENT:             /* this copies only the local env and points to outer envs */
      return(environment_copy(sc, obj));

    case T_VECTOR:
      return(vector_copy(sc, obj)); /* "shallow" copy */

    case T_PAIR:
      return(cons(sc, car(obj), list_copy(sc, cdr(obj), obj, true)));  /* this is the only use of list_copy */

    case T_HOOK:
      return(hook_copy(sc, obj));

#if WITH_GMP
    case T_BIG_INTEGER:
      return(mpz_to_big_integer(sc, big_integer(obj)));

    case T_BIG_RATIO:
      return(mpq_to_big_ratio(sc, big_ratio(obj)));

    case T_BIG_REAL:
      return(mpfr_to_big_real(sc, big_real(obj)));

    case T_BIG_COMPLEX:
      return(mpc_to_big_complex(sc, big_complex(obj)));
#endif
    }
  return(obj);
}


static s7_pointer g_copy(s7_scheme *sc, s7_pointer args)
{
  #define H_copy "(copy obj) returns a copy of obj"
  return(s7_copy(sc, car(args)));
}


static s7_pointer g_reverse(s7_scheme *sc, s7_pointer args)
{
  #define H_reverse "(reverse lst) returns a list with the elements of lst in reverse order.  reverse \
also accepts a string or vector argument."
  s7_pointer p, np;
  
  p = car(args);
  np = sc->NIL;

  switch (type(p))
    {
    case T_NIL:
      return(sc->NIL);

    case T_PAIR:
      return(s7_reverse(sc, p));
      /* if (is_null(np)) return(s7_wrong_type_arg_error(sc, "reverse", 0, p, "a proper list")); */
      break;

    case T_STRING:
      {
	int i, j, len;
	len = string_length(p);
	np = make_empty_string(sc, len, 0);
	if (len > 0)
	  for (i = 0, j = len - 1; i < len; i++, j--)
	    string_value(np)[i] = string_value(p)[j];
      }
      break;

    case T_VECTOR:
      {
	s7_Int i, j, len;
	len = vector_length(p);
	if (vector_is_multidimensional(p))
	  np = g_make_vector(sc, list_1(sc, g_vector_dimensions(sc, list_1(sc, p))));
	else np = make_vector_1(sc, len, NOT_FILLED, true);
	if (len > 0)
	  for (i = 0, j = len - 1; i < len; i++, j--)
	    vector_element(np, i) = vector_element(p, j);
      }
      break;

    case T_HASH_TABLE:
      return(hash_table_reverse(sc, p));

    case T_C_OBJECT:
      return(object_reverse(sc, p));

    default:
      return(s7_wrong_type_arg_error(sc, "reverse", 0, p, "a list, string, vector, or hash-table"));
    }
  
  return(np);
}


static s7_pointer list_fill(s7_scheme *sc, s7_pointer obj, s7_pointer val)
{
  /* ambiguous ("tree-fill"?) but if it's like vector-fill, we just stomp on the top level */
  s7_pointer x, y;

  x = obj;
  y = obj;

  while (true)
    {
      if (!is_pair(x)) return(val);
      car(x) = val;
      if (is_pair(cdr(x)))
	{
	  x = cdr(x);
	  car(x) = val;
	  if (is_pair(cdr(x)))
	    {
	      x = cdr(x);
	      y = cdr(y);
	      if (x == y) return(val);
	    }
	  else
	    {
	      if (is_not_null(cdr(x)))
		cdr(x) = val;
	      return(val);
	    }
	}
      else
	{
	  if (is_not_null(cdr(x)))
	    cdr(x) = val;
	  return(val);
	}
    }
  return(val);
}


static s7_pointer g_fill(s7_scheme *sc, s7_pointer args)
{
  #define H_fill "(fill! obj val) fills obj with the value val"

  switch (type(car(args)))
    {
    case T_STRING:
      return(g_string_fill(sc, args));

    case T_HASH_TABLE:
      if (is_not_null(cadr(args)))
	return(s7_wrong_type_arg_error(sc, "fill! hash-table value,", 2, cadr(args), "nil"));
      return(hash_table_clear(sc, car(args)));

    case T_VECTOR:
      return(g_vector_fill(sc, args));

    case T_C_OBJECT:
      {
	int tag;
	tag = object_type(car(args));
	if (object_types[tag].fill_func)
	  {
	    s_type_t *obj;
	    obj = (s_type_t *)s7_object_value(car(args));
	    car(args) = obj->value;
	    push_stack(sc, OP_APPLY, args, object_types[obj->type].fill_func);
	    return(sc->UNSPECIFIED);
	  }

	if (object_types[tag].fill)
	  return((*(object_types[tag].fill))(sc, car(args), cadr(args)));
	return(eval_error(sc, "attempt to fill ~A?", car(args)));
      }

    case T_PAIR:
      return(list_fill(sc, car(args), cadr(args)));

    case T_NIL:
      return(cadr(args));        /* this parallels the empty vector case */
    }

  return(s7_wrong_type_arg_error(sc, "fill!", 1, car(args), "a fillable object")); /* (fill! 1 0) */
}


static s7_pointer object_to_list(s7_scheme *sc, s7_pointer obj)
{
  switch (type(obj))
    {
    case T_VECTOR:
      return(s7_vector_to_list(sc, obj));

    case T_STRING:
      return(s7_string_to_list(sc, string_value(obj), string_length(obj)));

    case T_HASH_TABLE:
      {
	s7_pointer x, iterator, iter_loc;
	int gc_iter;
	/* (format #f "~{~A ~}" (hash-table '(a . 1) '(b . 2))) */

	iterator = g_make_hash_table_iterator(sc, list_1(sc, obj));
	gc_iter = s7_gc_protect(sc, iterator);
	iter_loc = hash_table_iterate_args(iterator);

	sc->w = sc->NIL;
	while (true)
	  {
	    x = g_hash_table_iterate(sc, iter_loc);
	    if (is_null(x)) break;
	    sc->w = cons(sc, x, sc->w);
	  }

	x = sc->w;
	sc->w = sc->NIL;
	s7_gc_unprotect_at(sc, gc_iter);
	return(x);
      }
      
    case T_HOOK:
      return(hook_functions(obj));

    case T_ENVIRONMENT:
      return(s7_environment_to_list(sc, obj));

    case T_C_OBJECT:
      {
	long int i, len; /* the "long" matters on 64-bit machines */
	s7_pointer x, z, result;
	int save_x = -1, save_y = -1, save_z = -1, gc_res = -1, gc_z = -1;
	/* (format #f "~{~A ~}" (vct 1 2 3)) */

	x = object_length(sc, obj);
	if (s7_is_integer(x))
	  len = s7_integer(x);
	else return(sc->F);

	if (len < 0)
	  return(sc->F);
	if (len == 0)
	  return(sc->NIL);

	result = g_make_list(sc, list_1(sc, make_integer(sc, len)));
	gc_res = s7_gc_protect(sc, result);
	z = list_1(sc, sc->F);
	gc_z = s7_gc_protect(sc, z);

	for (i = 0, x = result; i < len; i++, x = cdr(x))
	  {
	    car(z) = make_integer(sc, i);
	    SAVE_X_Y_Z(save_x, save_y, save_z);
	    car(x) = (*(object_ref(obj)))(sc, obj, z);
	    RESTORE_X_Y_Z(save_x, save_y, save_z);
	  }
	
	s7_gc_unprotect_at(sc, gc_z);
	s7_gc_unprotect_at(sc, gc_res);
	return(result);
      }
    }
  return(obj);
}




/* -------------------------------- format -------------------------------- */

static char *format_error(s7_scheme *sc, const char *msg, const char *str, s7_pointer args, format_data *dat, bool in_error_handler)
{
  int len;
  char *errmsg;
  s7_pointer x;

  if (in_error_handler)       /* actually, this must be an internal bug, but don't call s7_error recursively! */
    return(copy_string(msg)); /* this will be freed in by the string sweep in the GC */

  if (dat->loc == 0)
    {
      len = safe_strlen(msg) + 32;
      errmsg = (char *)malloc(len * sizeof(char));
      if (is_pair(args))
	snprintf(errmsg, len, "format ~S ~{~S~^ ~}: %s", msg);
      else snprintf(errmsg, len, "format ~S: %s", msg);
    }
  else 
    {
      char *filler;
      int i;
      filler = (char *)calloc(dat->loc + 12, sizeof(char));
      for (i = 0; i < dat->loc + 11; i++)
	filler[i] = ' ';
      len = safe_strlen(msg) + 32 + dat->loc + 12;
      errmsg = (char *)malloc(len * sizeof(char));
      if (is_pair(args))
	snprintf(errmsg, len, "format: ~S ~{~S~^ ~}\n%s^: %s", filler, msg);
      else snprintf(errmsg, len, "format: ~S\n%s^: %s", filler, msg);
      free(filler);
    }

  if (is_pair(args))
    x = list_3(sc, make_string_uncopied(sc, errmsg), make_protected_string(sc, str), args);
  else x = list_2(sc, make_string_uncopied(sc, errmsg), make_protected_string(sc, str));

  if (dat->str) {free(dat->str); dat->str = NULL;}

  s7_error(sc, sc->FORMAT_ERROR, x);
  return(NULL);
}


static void format_append_char(format_data *dat, char c)
{
  if (dat->len <= dat->loc + 2)
    {
      dat->len *= 2;
      dat->str = (char *)realloc(dat->str, dat->len * sizeof(char));
    }
  dat->str[dat->loc++] = c;
}


static void format_append_string(format_data *dat, const char *str)
{
  const char *s;
  int loc, len;
  if (!str) return;
  len = dat->len - 3;
  for (loc = dat->loc, s = str; (*s) != 0; loc++, s++)
    {
      if (len < loc)
	{
	  dat->len *= 2;
	  dat->str = (char *)realloc(dat->str, dat->len * sizeof(char));
	  len = dat->len - 3;
	}
      dat->str[loc] = (*s);
    }
  dat->loc = loc;
}


static int format_read_integer(s7_scheme *sc, int *cur_i, int str_len, const char *str, s7_pointer args, format_data *fdat)
{
  int i, arg1 = -1;
  char *tmp;
  i = *cur_i;
  if (isdigit(str[i]))
    {
      tmp = (char *)(str + i);
      if (sscanf(tmp, "%d", &arg1) < 1)
	format_error(sc, "bad number?", str, args, fdat, false);

      for (i = i + 1; i < str_len - 1; i++)
	if (!isdigit(str[i]))
	  break;
      if (i >= str_len)
	format_error(sc, "numeric argument, but no directive!", str, args, fdat, false);
    }
  *cur_i = i;
  return(arg1);
}


static void format_number(s7_scheme *sc, format_data *fdat, int radix, int width, int precision, char float_choice, char pad)
{
  char *tmp;
  if (width < 0) width = 0;

  /* precision choice depends on float_choice if it's -1 */
  if (precision < 0)
    {
      if ((float_choice == 'e') ||
	  (float_choice == 'f') ||
	  (float_choice == 'g'))
	precision = 6;
      else
	{
	  /* in the "int" cases, precision depends on the arg type */
	  switch (type(car(fdat->args)))
	    {
	    case T_INTEGER: 
	    case T_RATIO:
	      precision = 0; 
	      break;

	    default:
	      precision = 6;
	      break;
	    }
	}
    }

  /* should (format #f "~F" 1/3) return "1/3"?? in CL it's "0.33333334" */

  tmp = number_to_string_with_radix(sc, car(fdat->args), radix, width, precision, float_choice);

  if (pad != ' ')
    {
      char *padtmp;
      padtmp = tmp;
      while (*padtmp == ' ') (*(padtmp++)) = pad;
    }
  format_append_string(fdat, tmp);
  free(tmp);
  fdat->args = cdr(fdat->args);
}


#if WITH_GMP
static bool s7_is_one_or_big_one(s7_pointer p);
#else
#define s7_is_one_or_big_one(Num) s7_is_one(Num)
#endif
static char *truncate_string(char *form, int len, bool use_write);


static char *format_to_c_string(s7_scheme *sc, const char *str, s7_pointer args, s7_pointer *next_arg, int fdepth, bool in_error_handler)
{
  #define INITIAL_FORMAT_LENGTH 128
  int i = 0, str_len = 0;
  format_data *fdat = NULL;
  char *result, *tmp = NULL;

  str_len = safe_strlen(str);

  if (fdepth >= sc->num_fdats)
    {
      int i, new_num_fdats;
      new_num_fdats = fdepth * 2;
      sc->fdats = (format_data **)realloc(sc->fdats, sizeof(format_data *) * new_num_fdats);
      for (i = sc->num_fdats; i < new_num_fdats; i++) sc->fdats[i] = NULL;
      sc->num_fdats = new_num_fdats;
    }

  fdat = sc->fdats[fdepth];

  if (!fdat)
    {
      fdat = (format_data *)malloc(sizeof(format_data));
      fdat->loc = 0;
      fdat->args = args;
      sc->fdats[fdepth] = fdat;
    }
  else
    {
      fdat->loc = 0;
      fdat->args = args;
    }

  fdat->len = INITIAL_FORMAT_LENGTH;
  fdat->str = (char *)calloc(fdat->len, sizeof(char)); /* ~nT col checks need true current string length, so malloc here is messy */
  /* fdat->str[0] = '\0'; */

  if (str_len == 0)
    {
      if ((is_not_null(args)) &&
	  (next_arg == NULL))
	return(format_error(sc, "too many arguments", str, args, fdat, in_error_handler));
    }
  else
    {
      for (i = 0; i < str_len; i++)
	{
	  if (str[i] == '~')
	    {
	      if (i == str_len - 1)
		return(format_error(sc, "control string ends in tilde", str, args, fdat, in_error_handler));

	      switch (str[i + 1])
		{
		case '%':                           /* -------- newline -------- */
		  /* sbcl apparently accepts numeric args here (including 0) */
		  format_append_char(fdat, '\n');
		  i++;
		  break;

		case '&':                           /* -------- conditional newline -------- */
		  if ((fdat->loc > 0) &&
		      (fdat->str[fdat->loc - 1] != '\n'))
		    format_append_char(fdat, '\n');
		  i++;
		  break;
		  
		case '~':                           /* -------- tilde -------- */
		  format_append_char(fdat, '~');
		  i++;
		  break;

		case '\n':                          /* -------- trim white-space -------- */
		  for (i = i + 2; i <str_len - 1; i++)
		    if (!(white_space[(unsigned char)(str[i])]))
		      {
			i--;
			break;
		      }
		  break;
		  
		case '*':                           /* -------- ignore arg -------- */
		  i++;
		  fdat->args = cdr(fdat->args);
		  break;

		case '^':                           /* -------- exit -------- */
		  if (is_null(fdat->args))
		    {
		      i = str_len;
		      goto ALL_DONE;
		    }
		  i++;
		  break;

		case '@':                           /* -------- plural, 'y' or 'ies' -------- */
		  i += 2;
		  if ((str[i] != 'P') && (str[i] != 'p'))
		    return(format_error(sc, "unknown '@' directive", str, args, fdat, in_error_handler));
		  if (!s7_is_real(car(fdat->args)))        /* CL accepts non numbers here */
		    return(format_error(sc, "'@P' directive argument is not a real number", str, args, fdat, in_error_handler));

		  if (!s7_is_one_or_big_one(car(fdat->args)))
		    format_append_string(fdat, "ies");
		  else format_append_char(fdat, 'y');

		  fdat->args = cdr(fdat->args);
		  break;

		case 'P': case 'p':                 /* -------- plural in 's' -------- */
		  if (!s7_is_real(car(fdat->args)))
		    return(format_error(sc, "'P' directive argument is not a real number", str, args, fdat, in_error_handler));
		  if (!s7_is_one_or_big_one(car(fdat->args)))
		    format_append_char(fdat, 's');
		  i++;
		  fdat->args = cdr(fdat->args);
		  break;
		  
		case '{':                           /* -------- iteration -------- */
		  {
		    int k, curly_len = -1, curly_nesting = 1;

		    if (is_null(fdat->args))
		      return(format_error(sc, "missing argument", str, args, fdat, in_error_handler));

		    for (k = i + 2; k < str_len - 1; k++)
		      if (str[k] == '~')
			{
			  if (str[k + 1] == '}')
			    {
			      curly_nesting--;
			      if (curly_nesting == 0)
				{
				  curly_len = k - i - 1;
				  break;
				}
			    }
			  else
			    {
			      if (str[k + 1] == '{')
				curly_nesting++;
			    }
			}

		    if (curly_len == -1)
		      return(format_error(sc, "'{' directive, but no matching '}'", str, args, fdat, in_error_handler));

		    /* what about cons's here?  I can't see any way in CL either to specify the car or cdr of a cons within the format string 
		     *   (cons 1 2) is applicable: ((cons 1 2) 0) -> 1
		     *   also there can be applicable objects that won't work in the map context (arg not integer etc)
		     */
 		    if (is_not_null(car(fdat->args)))               /* (format #f "~{~A ~}" '()) -> "" */
		      {
			s7_pointer curly_arg;

			curly_arg = object_to_list(sc, car(fdat->args)); 
			if (is_not_null(curly_arg))                 /* (format #f "~{~A ~}" #()) -> "" */
			  {
			    char *curly_str = NULL;               /* this is the local (nested) format control string */
			    int curly_gc;

			    if (!is_proper_list(sc, curly_arg))
			      return(format_error(sc, "'{' directive argument should be a proper list or something we can turn into a list", 
						  str, args, fdat, in_error_handler));
			    curly_gc = s7_gc_protect(sc, curly_arg);

			    curly_str = (char *)malloc(curly_len * sizeof(char));
			    for (k = 0; k < curly_len - 1; k++)
			      curly_str[k] = str[i + 2 + k];
			    curly_str[curly_len - 1] = '\0';

			    while (is_not_null(curly_arg))
			      {
				s7_pointer new_arg = sc->NIL;
				tmp = format_to_c_string(sc, curly_str, curly_arg, &new_arg, fdepth + 1, in_error_handler);
				format_append_string(fdat, tmp);
				if (tmp) free(tmp);
				if (curly_arg == new_arg)
				  {
				    if (curly_str) free(curly_str);
				    s7_gc_unprotect_at(sc, curly_gc);
				    return(format_error(sc, "'{...}' doesn't consume any arguments!", str, args, fdat, in_error_handler));
				  }
				curly_arg = new_arg;
			      }

			    free(curly_str);
			    s7_gc_unprotect_at(sc, curly_gc);
			  }
		      }

		    i += (curly_len + 2); /* jump past the ending '}' too */
		    fdat->args = cdr(fdat->args);
		  }
		  break;
		  
		case '}':
		  return(format_error(sc, "unmatched '}'", str, args, fdat, in_error_handler));

		case 'C': case 'c':
		  {
		    s7_pointer obj;
		    
		    if (is_null(fdat->args))
		      return(format_error(sc, "~~C: missing argument", str, args, fdat, in_error_handler));
		    /* the "~~" here and below protects against "~C" being treated as a directive */
		    i++;
		    obj = car(fdat->args);
		    
		    if (!s7_is_character(obj))
		      return(format_error(sc, "'C' directive requires a character argument", str, args, fdat, in_error_handler));
		    
		    tmp = atom_to_c_string(sc, obj, false);
		    format_append_string(fdat, tmp);
		    if (tmp) free(tmp);
		    fdat->args = cdr(fdat->args);
		  }
		  break;
		  
		  /* -------- numeric args -------- */
		case '0': case '1': case '2': case '3': case '4': case '5':
		case '6': case '7': case '8': case '9': case ',':

		case 'B': case 'b':
		case 'D': case 'd':
		case 'E': case 'e':
		case 'F': case 'f':
		case 'G': case 'g':
		case 'O': case 'o':
		case 'X': case 'x':

		case 'A': case 'a': 
		case 'S': case 's':
		  
		case 'T': case 't':
		  {
		    int width = -1, precision = -1;
		    char pad = ' ';
		    i++;

		    if (isdigit(str[i]))
		      {
			width = format_read_integer(sc, &i, str_len, str, args, fdat);
			if ((width < 0) ||  /* sccanf is 32 bit so (format #f "~922337203685477580F" pi) gets a width of -1 or some such number */
			    (width > MAX_STRING_LENGTH))
			  return(format_error(sc, "numeric argument too large", str, args, fdat, in_error_handler));
		      }

		    if (str[i] == ',')
		      {
			i++;
			if (isdigit(str[i]))
			  {
			    precision = format_read_integer(sc, &i, str_len, str, args, fdat);
			    if ((precision < 0) ||
				(precision > MAX_STRING_LENGTH)) /* needed in 32-bit case */
			      return(format_error(sc, "numeric argument too large", str, args, fdat, in_error_handler));
			  }
			/* is (format #f "~12,12D" 1) an error?  The precision has no use here. */
			else
			  {
			    if (str[i] == '\'')       /* (format #f "~12,'xD" 1) -> "xxxxxxxxxxx1" */
			      {
				pad = str[i + 1];
				i += 2;
			      }
			    /* is (let ((str "~12,'xD")) (set! (str 5) #\null) (format #f str 1)) an error? */
			  }
		      }

		    switch (str[i])
		      {
			/* object->string */
		      case 'A': case 'a':
		      case 'S': case 's':
			{
			  shared_info *ci = NULL;
			  s7_pointer obj;
			  
			  if (is_null(fdat->args))
			    return(format_error(sc, "missing argument", str, args, fdat, in_error_handler));
			  if (precision != -1)
			    return(format_error(sc, "extra numeric argument", str, args, fdat, in_error_handler));

			  obj = car(fdat->args);
			  
			  if (has_structure(obj))
			    ci = make_shared_info(sc, obj);

			  if (width > 0)
			    sc->print_width = width;
			  else sc->print_width = MAX_STRING_LENGTH;

			  /* what should width=0 do here?
			   *   (format #f "~0,0F" pi) -> "3.0"
			   *   (format #f "~0,0D" 123) -> "123"
			   */

			  tmp = object_to_c_string_with_circle_check(sc, obj, (str[i] == 'S') || (str[i] == 's'), WITH_ELLIPSES, ci);

			  sc->print_width = MAX_STRING_LENGTH;
			  if (width > 0)
			    truncate_string(tmp, width, ((str[i] == 'S') || (str[i] == 's')));

			  format_append_string(fdat, tmp);
			  if (tmp) free(tmp);
			  fdat->args = cdr(fdat->args);
			}
			break;


			/* -------- pad to column -------- */
			/*   are columns numbered from 1 or 0?  there seems to be disagreement about this directive */
			/*   does "space over to" mean including? */
		      case 'T': case 't':
			if (width == -1) width = 0;
			if (precision == -1) precision = 0;
			if ((width > 0) || (precision > 0))         /* (format #f "a~8Tb") */
			  {
			    int j, k, outstr_len;
			    outstr_len = safe_strlen(fdat->str);
			    for (k = outstr_len - 1; k >= 0; k--)
			      if (fdat->str[k] == '\n')
				break;

			    /* (length (substring (format #f "~%~10T.") 1)) == (length (format #f "~10T."))
			     * (length (substring (format #f "~%-~10T.~%") 1)) == (length (format #f "-~10T.~%"))
			     */

			    if (precision > 0)
			      {
				int mult;
				mult = (int)(ceil((s7_Double)(outstr_len - k - width) / (s7_Double)precision)); /* CLtL2 ("least positive int") */
				if (mult < 1) mult = 1;
				width += (precision * mult);
			      }
			    for (j = outstr_len - k; j < width; j++)
			      format_append_char(fdat, pad);
			  }
			break;


			/* -------- numbers -------- */
		      case 'F': case 'f':
			if (is_null(fdat->args))
			  return(format_error(sc, "~~F: missing argument", str, args, fdat, in_error_handler));
			if (!(s7_is_number(car(fdat->args))))
			  return(format_error(sc, "~~F: numeric argument required", str, args, fdat, in_error_handler));

			format_number(sc, fdat, 10, width, precision, 'f', pad);
			break;

		      case 'G': case 'g':
			if (is_null(fdat->args))
			  return(format_error(sc, "~~G: missing argument", str, args, fdat, in_error_handler));
			if (!(s7_is_number(car(fdat->args))))
			  return(format_error(sc, "~~G: numeric argument required", str, args, fdat, in_error_handler));

			format_number(sc, fdat, 10, width, precision, 'g', pad);
			break;

		      case 'E': case 'e':
			if (is_null(fdat->args))
			  return(format_error(sc, "~~E: missing argument", str, args, fdat, in_error_handler));
			if (!(s7_is_number(car(fdat->args))))
			  return(format_error(sc, "~~E: numeric argument required", str, args, fdat, in_error_handler));

			format_number(sc, fdat, 10, width, precision, 'e', pad);
			break;

			/* how to handle non-integer arguments in the next 4 cases?  clisp just returns
			 *   the argument: (format nil "~X" 1.25) -> "1.25" which is perverse (ClTl2 p 581:
			 *   "if arg is not an integer, it is printed in ~A format and decimal base")!!
			 *   Guile raises an error ("argument is not an integer").  slib also raise an error.
			 *   I think I'll use the type of the number to choose the output format.
			 */
		      case 'D': case 'd':
			if (is_null(fdat->args))
			  return(format_error(sc, "~~D: missing argument", str, args, fdat, in_error_handler));
			if (!(s7_is_number(car(fdat->args))))
			  return(format_error(sc, "~~D: numeric argument required", str, args, fdat, in_error_handler));

			format_number(sc, fdat, 10, width, precision, 'd', pad);
			break;

		      case 'O': case 'o':
			if (is_null(fdat->args))
			  return(format_error(sc, "~~O: missing argument", str, args, fdat, in_error_handler));
			if (!(s7_is_number(car(fdat->args))))
			  return(format_error(sc, "~~O: numeric argument required", str, args, fdat, in_error_handler));

			format_number(sc, fdat, 8, width, precision, 'o', pad);
			break;

		      case 'X': case 'x':
			if (is_null(fdat->args))
			  return(format_error(sc, "~~X: missing argument", str, args, fdat, in_error_handler));
			if (!(s7_is_number(car(fdat->args))))
			  return(format_error(sc, "~~X: numeric argument required", str, args, fdat, in_error_handler));

			format_number(sc, fdat, 16, width, precision, 'x', pad);
			break;

		      case 'B': case 'b':
			if (is_null(fdat->args))
			  return(format_error(sc, "~~B: missing argument", str, args, fdat, in_error_handler));
			if (!(s7_is_number(car(fdat->args))))
			  return(format_error(sc, "~~B: numeric argument required", str, args, fdat, in_error_handler));

			format_number(sc, fdat, 2, width, precision, 'b', pad);
			break;
		      
		      default:
			if (width > 0)
			  return(format_error(sc, "format directive does not take a numeric argument", str, args, fdat, in_error_handler));
			return(format_error(sc, "unimplemented format directive", str, args, fdat, in_error_handler));
		      }
		  }
		  break;

		default:
		  return(format_error(sc, "unimplemented format directive", str, args, fdat, in_error_handler));
		}
	    }
	  else 
	    {
	      /* format_append_char(fdat, str[i]); */
	      if (fdat->len <= fdat->loc + 2)
		{
		  fdat->len *= 2;
		  fdat->str = (char *)realloc(fdat->str, fdat->len * sizeof(char));
		}
	      fdat->str[fdat->loc++] = str[i];
	    }
	}
    }

 ALL_DONE:
  if (next_arg)
    (*next_arg) = fdat->args;
  else
    {
      if (is_not_null(fdat->args))
	return(format_error(sc, "too many arguments", str, args, fdat, in_error_handler));
    }
  if (i < str_len)
    format_append_char(fdat, str[i]);    /* possible trailing ~ is sent out */
  format_append_char(fdat, '\0');

  result = fdat->str;
  fdat->str = NULL;

  return(result);
}


static s7_pointer format_to_output(s7_scheme *sc, s7_pointer out_loc, const char *in_str, s7_pointer args, bool in_error_handler)
{
  s7_pointer result;

  if ((!in_str) || (!(*in_str)))
    {
      if (is_not_null(args))
	return(s7_error(sc, 
			sc->FORMAT_ERROR, 
			list_2(sc, make_protected_string(sc, "format control string is null, but there are other arguments: ~A"), args)));
      return(make_protected_string(sc, ""));
    }

  result = make_string_uncopied(sc, format_to_c_string(sc, in_str, args, NULL, 0, in_error_handler));

  if (out_loc != sc->F)
    s7_display(sc, result, out_loc);

  return(result);
}


static s7_pointer g_format(s7_scheme *sc, s7_pointer args)
{
  #define H_format "(format out str . args) substitutes args into str sending the result to out. Most of \
s7's format directives are taken from CL: ~% = newline, ~& = newline if the preceding output character was \
no a newline, ~~ = ~, ~<newline> trims white space, ~* skips an argument, ~^ exits {} iteration if the arg list is exhausted, \
~nT spaces over to column n, ~A prints a representation of any object, ~S is the same, but puts strings in double quotes, \
~C prints a character, numbers are handled by ~F, ~E, ~G, ~B, ~O, ~D, and ~X with preceding numbers giving \
spacing (and spacing character) and precision.  ~{ starts an embedded format directive which is ended by ~}: \n\
\n\
  >(format #f \"dashed: ~{~A~^-~}\" '(1 2 3))\n\
  \"dashed: 1-2-3\"\n\
\n\
~P inserts \"s\" if the current argument is not 1 or 1.0 (use ~@P for \"ies\" or \"y\").\n\
~B is number->string in base 2, ~O in base 8, ~D base 10, ~X base 16,\n\
~E: (format #f \"~E\" 100.1) -&gt; \"1.001000e+02\" (%e in C)\n\
~F: (format #f \"~F\" 100.1) -&gt; \"100.100000\"   (%f in C)\n\
~G: (format #f \"~G\" 100.1) -&gt; \"100.1\"        (%g in C)"

  s7_pointer pt;
  pt = car(args);

  if (s7_is_string(pt))
    return(format_to_output(sc, sc->F, s7_string(pt), cdr(args), false));

  if (!s7_is_string(cadr(args)))
    return(s7_wrong_type_arg_error(sc, "format control string,", 2, cadr(args), "a string"));
    
  if (!((s7_is_boolean(pt)) ||               /* #f or #t */
	((is_output_port(pt)) &&             /* (current-output-port) or call-with-open-file arg, etc */
	 (!port_is_closed(pt)))))
    return(s7_wrong_type_arg_error(sc, "format", 1, pt, "#f, #t, or an open output port"));

  return(format_to_output(sc, (pt == sc->T) ? sc->output_port : pt, s7_string(cadr(args)), cddr(args), false));
}


const char *s7_format(s7_scheme *sc, s7_pointer args)
{
  return(s7_string(g_format(sc, args))); /* for the run macro in run.c */
}



/* -------- trace -------- */

/* 
    (define (hiho arg) (if (> arg 0) (+ 1 (hiho (- arg 1))) 0))
    (trace hiho)
    (hiho 3)

    [hiho 3]
     [hiho 2]
      [hiho 1]
       [hiho 0]
        0
       1
      2
     3
*/

/* to trace an internal function, put the trace/untrace pair in the enclosing function
 */

static s7_pointer g_trace(s7_scheme *sc, s7_pointer args)
{
  #define H_trace "(trace . args) adds each function in its argument list to the trace list.\
Each argument can be a function, symbol, macro, or any applicable object: (trace abs '+ v) where v is a vct \
prints out data about any call on abs or +, and any reference to the vct v. Trace output is sent \
to the current-output-port.  If trace is called without any arguments, everything is traced -- use \
untrace without arguments to turn this off."

  int i;
  s7_pointer x;

  if (is_null(args))
    {
      trace_all = true;
      tracing = true;
      return(sc->F);
    }
  
  for (i = 1, x = args; is_not_null(x); i++, x = cdr(x)) 
    if ((!is_symbol(car(x))) &&
	(!is_procedure(car(x))) &&
	(!is_any_macro(car(x))))
      return(s7_wrong_type_arg_error(sc, "trace", i, car(x), "a symbol, a function, or some other applicable object"));

  for (x = args; is_not_null(x); x = cdr(x)) 
    {
      if (is_symbol(car(x)))
	sc->trace_list[sc->trace_top++] = eval_symbol(sc, car(x));
      else sc->trace_list[sc->trace_top++] = car(x);
      if (sc->trace_top >= sc->trace_list_size)
	{
	  sc->trace_list_size *= 2;
	  sc->trace_list = (s7_pointer *)realloc(sc->trace_list, sc->trace_list_size * sizeof(s7_pointer));
	}
    }

  tracing = (sc->trace_top > 0);
  return(sc->T);
}


static s7_pointer g_untrace(s7_scheme *sc, s7_pointer args)
{
  #define H_untrace "(untrace . args) removes each function in its arg list from the trace list. \
If untrace is called with no arguments, all functions are removed, turning off all tracing."
  int i, j, ctr;
  s7_pointer x;

  if (is_null(args))
    {
      trace_all = false;
      for (i = 0; i < sc->trace_top; i++)
	sc->trace_list[i] = sc->NIL;
      sc->trace_top = 0;
      tracing = false;
      return(sc->F);
    }

  for (ctr = 1, x = args; is_not_null(x); ctr++, x = cdr(x)) 
    {
      s7_pointer value;
      if (is_symbol(car(x)))
	value = eval_symbol(sc, car(x));
      else 
	{
	  if ((is_procedure(car(x))) ||
	      (is_any_macro(car(x))))
	    value = car(x);
	  else return(s7_wrong_type_arg_error(sc, "untrace", ctr, car(x), "a symbol or procedure")); /* (untrace "hi") */
	}
      for (i = 0; i < sc->trace_top; i++)
	if (value == sc->trace_list[i])
	  sc->trace_list[i] = sc->NIL;
    }
  
  /* now collapse list and reset trace_top (and possibly tracing) */
  for (i = 0, j = 0; i < sc->trace_top; i++)
    if (is_not_null(sc->trace_list[i]))
      sc->trace_list[j++] = sc->trace_list[i];
  
  sc->trace_top = j;
  tracing = (sc->trace_top > 0);
  return(sc->T);
}


static void trace_apply(s7_scheme *sc)
{
  int i;
  bool trace_it = false;

  if (trace_all)
    trace_it = true;
  else
    {
      for (i = 0; i < sc->trace_top; i++)
	if (sc->code == sc->trace_list[i])
	  {
	    trace_it = true;
	    break;
	  }
    }

  if (trace_it)
    {
      int k, len;
      char *tmp1, *tmp2, *str;
      push_stack(sc, OP_TRACE_RETURN, sc->code, sc->NIL);
      tmp1 = s7_object_to_c_string(sc, sc->code);
      tmp2 = s7_object_to_c_string(sc, sc->args);

      len = safe_strlen(tmp2);
      tmp2[0] = ' ';
      tmp2[len - 1] = ']';
      
      len += (safe_strlen(tmp1) + sc->trace_depth + 64);
      str = (char *)calloc(len, sizeof(char));
      
      for (k = 0; k < sc->trace_depth; k++) str[k] = ' ';
      str[k] = '[';
      strcat(str, tmp1);
      strcat(str, tmp2);
      free(tmp1);
      free(tmp2);
      
      strcat(str, "\n");
      port_display(sc->output_port)(sc, str, sc->output_port);
      free(str);
      
      sc->trace_depth++;
 
      if (is_not_null(hook_functions(sc->trace_hook)))
	{
	  push_stack(sc, OP_TRACE_HOOK_QUIT, copy_list(sc, sc->args), sc->code); 
	  /* this restores the current state after dealing with the trace hook func.
	   *   we have to copy the arglist because the evaluator now uses the arg list
	   *   as the local environment -- the assumption when we get back to TRACE_HOOK_QUIT
	   *   is that eval's sc->args is no longer of interest.
	   *
	   * we have to turn off tracing while evaluating the trace hook functions
	   */
	  tracing = false;
	  s7_hook_apply(sc, sc->trace_hook, list_2(sc, sc->code, sc->args));

	  /* it would be nice if *trace-hook* could return #f to turn off trace printout.
	   *   then it could be used (more cleanly) for a call-history list (a circular list)
	   */
	}
    }
}


static void trace_return(s7_scheme *sc)
{
  int k, len;
  char *str, *tmp;

  tmp = s7_object_to_c_string(sc, sc->value);  

  len = sc->trace_depth + safe_strlen(tmp) + 3;
  str = (char *)calloc(len, sizeof(char));

  for (k = 0; k < sc->trace_depth; k++) str[k] = ' ';
  strcat(str, tmp);
  strcat(str, "\n");
  free(tmp);

  port_display(sc->output_port)(sc, str, sc->output_port);
  free(str);

  sc->trace_depth--;
  if (sc->trace_depth < 0) sc->trace_depth = 0;
}





/* -------- error handlers -------- */

static const char *type_name(s7_pointer arg)
{
  switch (type(arg))
    {
    case T_NIL:          return("nil");
    case T_UNTYPED:      return("untyped");
    case T_BOOLEAN:      return("boolean");
    case T_STRING:       return("string");
    case T_SYMBOL:       return("symbol");
    case T_SYNTAX:       return("syntax");
    case T_PAIR:         return("pair");
    case T_CLOSURE:      return("closure");
    case T_CLOSURE_STAR: return("closure*");
    case T_GOTO:         return("goto");
    case T_CONTINUATION: return("continuation");
    case T_C_OPT_ARGS_FUNCTION:
    case T_C_RST_ARGS_FUNCTION:
    case T_C_LST_ARGS_FUNCTION:
    case T_C_ANY_ARGS_FUNCTION:
    case T_C_FUNCTION:   return("function");
    case T_C_MACRO:      return("macro");
    case T_C_POINTER:    return("c-pointer");
    case T_CHARACTER:    return("character");
    case T_VECTOR:       return("vector");
    case T_BACRO:        return("bacro");
    case T_MACRO:        return("macro");
    case T_CATCH:        return("catch");
    case T_DYNAMIC_WIND: return("dynamic-wind");
    case T_HASH_TABLE:   return("hash-table");
    case T_C_OBJECT:     return(object_types[object_type(arg)].name);
    case T_HOOK:         return("hook");
    case T_COUNTER:      return("counter");
    case T_SLOT:         return("slot");
    case T_ENVIRONMENT:  return("environment");

    case T_INTEGER:      return("integer");
    case T_RATIO:        return("ratio");
    case T_REAL:         return("real");
    case T_COMPLEX:      return("complex number"); /* "a complex" doesn't sound right */

    case T_BIG_INTEGER:  return("big integer");
    case T_BIG_RATIO:    return("big ratio");
    case T_BIG_REAL:     return("big real");
    case T_BIG_COMPLEX:  return("big complex number");

    case T_INPUT_PORT:
      {
	if (is_file_port(arg))
	  return("input file port");
	if (is_string_port(arg))
	  return("input string port");
	return("input port");
      }

    case T_OUTPUT_PORT:
      {
	if (is_file_port(arg))
	  return("output file port");
	if (is_string_port(arg))
	  return("output string port");
	return("output port");
      }
    }
  return("messed up object");
}


static bool is_vowel(const char *name)
{
  return((name[0] == 'a') || 
	 (name[0] == 'e') || 
	 (name[0] == 'i') || 
	 (name[0] == 'o') || 
	 (name[0] == 'u'));
}


s7_pointer s7_wrong_type_arg_error(s7_scheme *sc, const char *caller, int arg_n, s7_pointer arg, const char *descr)
{
  /* info list is '(format_string caller arg_n arg type_name descr) */
  if (arg_n < 0) arg_n = 0;

  if (arg_n > 0)
    {
      list_set(sc, sc->WRONG_TYPE_ARG_INFO, 1, make_protected_string(sc, caller));
      list_set(sc, sc->WRONG_TYPE_ARG_INFO, 2, make_integer(sc, arg_n));
      list_set(sc, sc->WRONG_TYPE_ARG_INFO, 3, arg);
      list_set(sc, sc->WRONG_TYPE_ARG_INFO, 4, make_protected_string(sc, (is_vowel(type_name(arg))) ? "n" : ""));
      list_set(sc, sc->WRONG_TYPE_ARG_INFO, 5, make_protected_string(sc, type_name(arg)));
      list_set(sc, sc->WRONG_TYPE_ARG_INFO, 6, make_protected_string(sc, descr));
      return(s7_error(sc, sc->WRONG_TYPE_ARG, sc->WRONG_TYPE_ARG_INFO));

      /* it's possible to avoid the cell allocations here, but at the cost of memcpy instead --
       *   almost a toss-up.
       */
    }
  list_set(sc, sc->SIMPLE_WRONG_TYPE_ARG_INFO, 1, make_protected_string(sc, caller));
  list_set(sc, sc->SIMPLE_WRONG_TYPE_ARG_INFO, 2, arg);
  list_set(sc, sc->SIMPLE_WRONG_TYPE_ARG_INFO, 3, make_protected_string(sc, (is_vowel(type_name(arg))) ? "n" : ""));
  list_set(sc, sc->SIMPLE_WRONG_TYPE_ARG_INFO, 4, make_protected_string(sc, type_name(arg)));
  list_set(sc, sc->SIMPLE_WRONG_TYPE_ARG_INFO, 5, make_protected_string(sc, descr));
  return(s7_error(sc, sc->WRONG_TYPE_ARG, sc->SIMPLE_WRONG_TYPE_ARG_INFO));
}


s7_pointer s7_out_of_range_error(s7_scheme *sc, const char *caller, int arg_n, s7_pointer arg, const char *descr)
{
  /* info list is '(format_string caller arg_n arg descr) */
  if (arg_n < 0) arg_n = 0;

  if (arg_n > 0)
    {
      list_set(sc, sc->OUT_OF_RANGE_INFO, 1, make_protected_string(sc, caller));
      list_set(sc, sc->OUT_OF_RANGE_INFO, 2, make_integer(sc, arg_n));
      list_set(sc, sc->OUT_OF_RANGE_INFO, 3, arg);
      list_set(sc, sc->OUT_OF_RANGE_INFO, 4, make_protected_string(sc, descr));
      return(s7_error(sc, sc->OUT_OF_RANGE, sc->OUT_OF_RANGE_INFO));
    }
  list_set(sc, sc->SIMPLE_OUT_OF_RANGE_INFO, 1, make_protected_string(sc, caller));
  list_set(sc, sc->SIMPLE_OUT_OF_RANGE_INFO, 2, arg);
  list_set(sc, sc->SIMPLE_OUT_OF_RANGE_INFO, 3, make_protected_string(sc, descr));
  return(s7_error(sc, sc->OUT_OF_RANGE, sc->SIMPLE_OUT_OF_RANGE_INFO));
}


s7_pointer s7_wrong_number_of_args_error(s7_scheme *sc, const char *caller, s7_pointer args)
{
  return(s7_error(sc, sc->WRONG_NUMBER_OF_ARGS, 
		  list_2(sc, 
			 make_protected_string(sc, caller), /* "caller" includes the format directives */
			 args)));
}


static s7_pointer division_by_zero_error(s7_scheme *sc, const char *caller, s7_pointer arg)
{
  return(s7_error(sc, make_symbol(sc, "division-by-zero"), 
		  list_3(sc, 
			 make_protected_string(sc, "~A: division by zero, ~A"), 
			 make_protected_string(sc, caller),
			 arg)));
}


static s7_pointer file_error(s7_scheme *sc, const char *caller, const char *descr, const char *name)
{
  return(s7_error(sc, make_symbol(sc, "io-error"), 
		  list_4(sc, 
			 make_protected_string(sc, "~A: ~A ~S"),
			 make_protected_string(sc, caller),
			 make_protected_string(sc, descr),
			 make_protected_string(sc, name))));
}


void s7_set_error_exiter(s7_scheme *sc, void (*error_exiter)(void))
{
  sc->error_exiter = error_exiter;
}


static s7_pointer g_dynamic_wind(s7_scheme *sc, s7_pointer args)
{
  #define H_dynamic_wind "(dynamic-wind init body finish) calls init, then body, then finish, \
each a function of no arguments, guaranteeing that finish is called even if body is exited"
  s7_pointer p;

  if (!is_thunk(sc, car(args)))
    return(s7_wrong_type_arg_error(sc, "dynamic-wind", 1, car(args), "a thunk"));
  if (!is_thunk(sc, cadr(args)))
    return(s7_wrong_type_arg_error(sc, "dynamic-wind", 2, cadr(args), "a thunk"));
  if (!is_thunk(sc, caddr(args)))
    return(s7_wrong_type_arg_error(sc, "dynamic-wind", 3, caddr(args), "a thunk"));

  /* this won't work:

       (let ((final (lambda (a b c) (list a b c))))
         (dynamic-wind
           (lambda () #f)
           (lambda () (set! final (lambda () (display "in final"))))
           final))

   * but why not?  'final' is a thunk by the time it is evaluated.
   *   catch (the error handler) is similar.
   *
   * It can't work here because we set up the dynamic_wind_out slot below and
   *   even if the thunk check was removed, we'd still be trying to apply the original function.
   */
  
  NEW_CELL(sc, p);
  dynamic_wind_in(p) = car(args);
  dynamic_wind_body(p) = cadr(args);
  dynamic_wind_out(p) = caddr(args);
  dynamic_wind_state(p) = DWIND_INIT;
  set_type(p, T_DYNAMIC_WIND);            /* atom -> don't mark car/cdr, don't copy */

  push_stack(sc, OP_DYNAMIC_WIND, sc->NIL, p);          /* args will be the saved result, code = s7_dynwind_t obj */
  push_stack(sc, OP_APPLY, sc->NIL, car(args));
  return(sc->F);
}

/* C-side dynamic-wind would need at least void* context pointer passed to each function,
 *   and to fit with the scheme-side stuff above, the functions need to be s7 functions,
 *   so I wonder if it could be s7_dynamic_wind(s7_scheme *sc, s7_pointer init, s7_pointer body, s7_pointer finish)
 *   and the caller would use the C-closure idea (s7.html) to package up C-side data.
 *   Then, the caller would probably assume a return value, requiring s7_call?
 *   -> g_dynamic_wind(sc, list_3(sc, init, body, finish)) but with eval(sc, OP_APPLY) at end?
 */


static s7_pointer g_catch(s7_scheme *sc, s7_pointer args)
{
  #define H_catch "(catch tag thunk handler) evaluates thunk; if an error occurs that matches the tag (#t matches all), the handler is called"
  s7_pointer p;

  /* should this check for a tag that can't possibly be eq? to anything that error might throw? (a string for example)
   */

  if (!is_thunk(sc, cadr(args)))
    return(s7_wrong_type_arg_error(sc, "catch", 2, cadr(args), "a thunk"));
  if (!is_procedure(caddr(args)))
    return(s7_wrong_type_arg_error(sc, "catch", 3, caddr(args), "a procedure"));
  
  NEW_CELL(sc, p);
  catch_tag(p) = car(args);
  catch_goto_loc(p) = s7_stack_top(sc);
  catch_op_loc(p) = (int)(sc->op_stack_now - sc->op_stack);
  catch_handler(p) = caddr(args);
  set_type(p, T_CATCH);

  push_stack(sc, OP_CATCH, args, p);          /* args ignored but maybe safer for GC? */

  if (is_closure(cadr(args)))
    {
      sc->code = closure_body(cadr(args));
      NEW_FRAME(sc, closure_environment(cadr(args)), sc->envir);
      push_stack(sc, OP_BEGIN, sc->args, sc->code);
    }
  else push_stack(sc, OP_APPLY, sc->NIL, cadr(args));
  return(sc->F);
}



s7_pointer s7_catch_all(s7_scheme *sc, s7_pointer thunk, s7_pointer error_handler)
{
  s7_pointer p;

  NEW_CELL(sc, p);
  catch_tag(p) = sc->T;                   /* if we catch everything, the error handling stuff in s7_call is not needed */
  catch_goto_loc(p) = s7_stack_top(sc);
  catch_op_loc(p) = (int)(sc->op_stack_now - sc->op_stack);
  catch_handler(p) = error_handler;
  set_type(p, T_CATCH);

  push_stack(sc, OP_EVAL_DONE, sc->args, sc->code); 
  push_stack(sc, OP_CATCH, sc->NIL, p);
  sc->args = sc->NIL;
  sc->code = thunk;
  eval(sc, OP_APPLY);

  return(sc->value);
}


#if 0
static s7_pointer s7_apply_function_with_catch(s7_scheme *sc, s7_pointer func, s7_pointer args, s7_pointer error_handler)
{
  s7_pointer p;
  NEW_CELL(sc, p);
  catch_tag(p) = sc->T; 
  catch_goto_loc(p) = s7_stack_top(sc);
  catch_op_loc(p) = (int)(sc->op_stack_now - sc->op_stack);
  catch_handler(p) = error_handler;
  set_type(p, T_CATCH);

  push_stack(sc, OP_EVAL_DONE, sc->args, sc->code); 
  push_stack(sc, OP_CATCH, sc->NIL, p);
  sc->args = args;
  sc->code = func;
  eval(sc, OP_APPLY);

  return(sc->value);
}
#endif


#if 0
/* here's an example:
 *
    static s7_pointer error_handler(s7_scheme *sc, s7_pointer args)
    {
      fprintf(stderr, "got an error! %s\n", s7_object_to_c_string(sc, args));
      return(s7_car(args));
    }
    
    static s7_pointer adder(s7_scheme *sc, s7_pointer args)
    {
      return(s7_f(sc));
    }
    
    int main(int argc, char **argv)
    {
      s7_scheme *s7;
      s7 = s7_init();  
      fprintf(stderr, "catch: %s\n",
        s7_object_to_c_string(s7, 
    	  s7_catch_all(s7, 
    	    s7_make_function(s7, "adder", adder, 0, 0, false, "+"),
    	    s7_make_function(s7, "error-handler", error_handler, 0, 0, true, "error handler"))));
    }
 *
 * the function names are local (for error handling etc).  In general, it would be safer to
 *    call s7_gc_protect on the functions (s7_make_function does not protect them).  The
 *    body function ("adder") is a thunk -- args will always be nil.
 */
#endif


/* error reporting info -- save filename and line number */

#define INITIAL_FILE_NAMES_SIZE 8
static s7_pointer *file_names = NULL;
static int file_names_size = 0;
static int file_names_top = -1;

#define remembered_line_number(Line) (Line & 0xfffff)
#define remembered_file_name(Line)   (((Line >> 20) <= file_names_top) ? file_names[Line >> 20] : sc->F)
/* this gives room for 4000 files each of 1000000 lines */


static int remember_file_name(s7_scheme *sc, const char *file)
{
  int i, old_size = 0;

  file_names_top++;
  if (file_names_top >= file_names_size)
    {
      if (file_names_size == 0)
	{
	  file_names_size = INITIAL_FILE_NAMES_SIZE;
	  file_names = (s7_pointer *)calloc(file_names_size, sizeof(s7_pointer));
	}
      else
	{
	  old_size = file_names_size;
	  file_names_size *= 2;
	  file_names = (s7_pointer *)realloc(file_names, file_names_size * sizeof(s7_pointer));
	}
      for (i = old_size; i < file_names_size; i++)
	file_names[i] = sc->F;
    }
  file_names[file_names_top] = s7_make_permanent_string(file);

  return(file_names_top);
}


#define ERROR_INFO_DEFAULT sc->F
#define ERROR_TYPE 0
#define ERROR_DATA 1
#define ERROR_CODE 2
#define ERROR_CODE_LINE 3
#define ERROR_CODE_FILE 4
#define ERROR_ENVIRONMENT 5
#define ERROR_STACK_SIZE 8
#define ERROR_INFO_SIZE (6 + ERROR_STACK_SIZE)

/* *error-info* is a vector of 6 or more elements:
 *    0: the error type or tag ('division-by-zero)
 *    1: the message or information passed by the error function
 *    2: if not #f, the code that s7 thinks triggered the error
 *    3: if not #f, the line number of that code
 *    4: if not #f, the file name of that code
 *    5: the environment at the point of the error
 *    6..top: stack enviroment pointers (giving enough info to reconstruct the current call stack), ending in #f
 */

/* slightly ugly:

(define-macro (cerror . args)
  `(call/cc
    (lambda (continue)
      (apply error continue ',args))))

;;; now ((vector-ref *error-info* 0)) will continue from the error

(define (cerror . args)
  (format #t "error: ~A" (car args))
  (if (not (null? (cdr args)))
      (if (and (string? (cadr args))
	       (not (null? (cddr args))))
	  (let ((str (apply format (cdr args))))
	    (format #t "~S~%" str))
	  (format #t "~S~%" (cadr args))))
  (format #t "continue? (<cr>=yes) ")
  (let ((val (read-line ())))
    (if (not (char=? (val 0) #\newline))
	(error (car args)))))

;;; so perhaps wrap the caller-passed stuff in "continue?" etc?
*/


static s7_pointer s7_error_1(s7_scheme *sc, s7_pointer type, s7_pointer info, bool exit_eval)
{
  int i;
  bool reset_error_hook = false;
  s7_pointer catcher;
  const char *call_name = NULL, *call_file = NULL;
  int call_line = 0;

  /* fprintf(stderr, "error: %s %s\n", DISPLAY(type), DISPLAY(info)); */

  /* null info can mean symbol table is locked so make-symbol uses s7_error to get out */

  /* set up *error-info*, look for a catch that matches 'type', if found
   *   call its error-handler, else if *error-hook* is bound, call it,
   *   else send out the error info ourselves.
   */
  sc->no_values = 0; 
#if WITH_OPTIMIZATION
  sc->safe_do_level = 0; 
  finder = find_symbol_or_bust;
#endif
  catcher = sc->F;

  if (sc->s7_call_name)
    {
      call_name = sc->s7_call_name;
      call_file = sc->s7_call_file;
      call_line = sc->s7_call_line;
      sc->s7_call_name = NULL;
      sc->s7_call_file = NULL;
      sc->s7_call_line = -1;
    }

  vector_element(sc->error_info, ERROR_TYPE) = type;
  vector_element(sc->error_info, ERROR_DATA) = info;
  vector_element(sc->error_info, ERROR_CODE) = sc->cur_code;
  vector_element(sc->error_info, ERROR_CODE_LINE) = ERROR_INFO_DEFAULT;
  vector_element(sc->error_info, ERROR_CODE_FILE) = ERROR_INFO_DEFAULT;
  vector_element(sc->error_info, ERROR_ENVIRONMENT) = sc->envir;
  s7_gc_on(sc, true);  /* this is in case we were triggered from the sort function -- clumsy! */

  /* (let ((x 32)) (define (h1 a) (* a "hi")) (define (h2 b) (+ b (h1 b))) (h2 1)) */
  if (is_pair(sc->cur_code))
    {
      int line, j, top;
      line = pair_line_number(sc->cur_code);

      if ((line > 0) &&
	  (remembered_line_number(line) != 0) &&
	  (remembered_file_name(line) != sc->F))
	{
	  vector_element(sc->error_info, ERROR_CODE_LINE) = make_integer(sc, remembered_line_number(line)); 
	  vector_element(sc->error_info, ERROR_CODE_FILE) = remembered_file_name(line);	  
	}

      for (top = s7_stack_top(sc) - 1, j = ERROR_ENVIRONMENT + 1; (top > 0) && (j < ERROR_INFO_SIZE); top -= 4, j++)
	vector_element(sc->error_info, j) = stack_environment(sc->stack, top);
      if (j < ERROR_INFO_SIZE)
	vector_element(sc->error_info, j) = ERROR_INFO_DEFAULT;
    }

  sc->cur_code = ERROR_INFO_DEFAULT;

  /* if (!s7_is_continuation(type))... */

  /* top is 1 past actual top, top - 1 is op, if op = OP_CATCH, top - 4 is the cell containing the catch struct */
  for (i = s7_stack_top(sc) - 1; i >= 3; i -= 4)
    {
      opcode_t op;
      s7_pointer x;

      op = stack_op(sc->stack, i);
      switch (op)
	{
	case OP_DYNAMIC_WIND:
	  x = stack_code(sc->stack, i);
	  if (dynamic_wind_state(x) == DWIND_BODY)
	    {
	      dynamic_wind_state(x) = DWIND_FINISH;   /* make sure an uncaught error in the exit thunk doesn't cause us to loop */
	      push_stack(sc, OP_EVAL_DONE, sc->args, sc->code); 
	      sc->args = sc->NIL;
	      sc->code = dynamic_wind_out(x);
	      eval(sc, OP_APPLY);                     /* I guess this means no call/cc out of the exit thunk in an error-catching context */
	    }
	  break;

	case OP_CATCH:
	  x = stack_code(sc->stack, i);
	  if ((type == sc->T) ||
	      (catch_tag(x) == sc->T) ||
	      (catch_tag(x) == type))
	    {
	      catcher = x;
	      goto GOT_CATCH;
	    }
	  break;

	case OP_UNWIND_OUTPUT:
	  x = stack_code(sc->stack, i);                /* "code" = port that we opened */
	  s7_close_output_port(sc, x);
	  x = stack_args(sc->stack, i);                /* "args" = port that we shadowed, if not #f */
	  if (x != sc->F)
	    sc->output_port = x;
	  break;

	case OP_UNWIND_INPUT:
	  s7_close_input_port(sc, stack_code(sc->stack, i)); /* "code" = port that we opened */
	  sc->input_port = stack_args(sc->stack, i);         /* "args" = port that we shadowed */
	  break;

	case OP_READ_DONE:        /* perhaps an error during (read) */
	  pop_input_port(sc);
	  break;

	case OP_EVAL_STRING_1:    /* perhaps an error happened before we could push the OP_EVAL_STRING_2 */
	case OP_EVAL_STRING_2:
	  s7_close_input_port(sc, sc->input_port);
	  pop_input_port(sc);
	  break;

	case OP_BARRIER:
	  if (is_input_port(stack_args(sc->stack, i)))      /* (eval-string "'(1 .)") */
	    {
	      if (sc->input_port == stack_args(sc->stack, i))
		pop_input_port(sc);
	      s7_close_input_port(sc, stack_args(sc->stack, i));
	    }
	  break;

	case OP_DEACTIVATE_GOTO:
	  call_exit_active(stack_args(sc->stack, i)) = false;
	  break;

	case OP_TRACE_RETURN:
	  sc->trace_depth--;
	  if (sc->trace_depth < 0) sc->trace_depth = 0;
	  break;

	  /* perhaps also OP_LOAD_CLOSE_AND_POP_IF_EOF 
	   *  currently an error during a nested load stops all loads
	   */

	case OP_ERROR_HOOK_QUIT:
	  hook_functions(sc->error_hook) = stack_code(sc->stack, i);

	  /* apparently there was an error during *error-hook* evaluation, but Rick wants the hook re-established anyway */
	  reset_error_hook = true;
	  /* avoid infinite loop -- don't try to (re-)evaluate (buggy) *error-hook*! */
	  break;

	default:
	  break;
	}
    }
  
GOT_CATCH:
  if (catcher != sc->F)
    {
      int loc;

      sc->args = list_2(sc, type, info);
      sc->code = catch_handler(catcher);
      loc = catch_goto_loc(catcher);
      sc->op_stack_now = (s7_pointer *)(sc->op_stack + catch_op_loc(catcher));
      sc->stack_end = (s7_pointer *)(sc->stack_start + loc);

      /* if user (i.e. yers truly!) copies/pastes the preceding lambda () into the
       *   error handler portion of the catch, he gets the inexplicable message:
       *       ;(): too many arguments: (a1 ())
       *   when this apply tries to call the handler.  So, we need a special case
       *   error check here!
       */
      if (!args_match(sc, sc->code, 2))
	return(s7_wrong_number_of_args_error(sc, "catch error handler has wrong number of args: ~A", sc->args));

      sc->op = OP_APPLY;

      /* explicit eval needed if s7_call called into scheme where a caught error occurred (ex6 in exs7.c)
       *  but putting it here (via eval(sc, OP_APPLY)) means the C stack is not cleared correctly in non-s7-call cases, 
       *  so defer it until s7_call 
       */
    }
  else
    {
      /* (set! *error-hook* (list (lambda (tag args) (apply format (cons #t args))))) */

      if ((!reset_error_hook) && 
	  (is_pair(hook_functions(sc->error_hook))))
	{
	  s7_pointer error_list;
	  /* (set! (hook-functions *error-hook*) (list (lambda args (format *stderr* "got error ~A~%" args)))) */

	  error_list = hook_functions(sc->error_hook);
	  hook_functions(sc->error_hook) = sc->NIL;
	  /* if the *error-hook* functions trigger an error, we had better not have *error-hook* still set! */

	  push_stack(sc, OP_ERROR_HOOK_QUIT, sc->NIL, error_list); /* restore *error-hook* upon successful (or any!) evaluation */
	  sc->args = list_2(sc, type, info);
	  sc->code = error_list;
	  /* push_stack(sc, OP_HOOK_APPLY, sc->args, sc->code); */

	  /* if we drop into the longjmp below, the hook functions are not called!
	   *   OP_ERROR_HOOK_QUIT performs the longjmp, so it should be safe to go to eval.
	   */
	  eval(sc, OP_HOOK_APPLY);
	}
      else
	{
	  /* if info is not a list, send object->string to current error port,
	   *   else assume car(info) is a format control string, and cdr(info) are its args
	   *
	   * if at all possible, get some indication of where we are!
	   */
	  s7_pointer x, error_port;
	  error_port = s7_current_error_port(sc);

	  if ((!s7_is_list(sc, info)) ||
	      (!s7_is_string(car(info))))
	    format_to_output(sc, error_port, "\n;~A ~A", list_2(sc, type, info), type == sc->FORMAT_ERROR);
	  else
	    {
	      const char *carstr;
	      int i, len;
	      bool use_format = false;
	      
	      /* it's possible that the error string is just a string -- not intended for format */
	      carstr = s7_string(car(info));
	      len = string_length(car(info));
	      for (i = 0; i < len; i++)
		if (carstr[i] == '~')
		  {
		    use_format = true;
		    break;
		  }
	      
	      if (use_format)
		{
		  char *errstr;
		  len += 8;
		  errstr = (char *)malloc(len * sizeof(char));
		  snprintf(errstr, len, "\n;%s", s7_string(car(info)));
		  format_to_output(sc, error_port, errstr, cdr(info), type == sc->FORMAT_ERROR);
		  free(errstr);
		}
	      else format_to_output(sc, error_port, "\n;~A ~A", list_2(sc, type, info), type == sc->FORMAT_ERROR);
	    }
	  
	  /* now display location at end */
	  
	  if ((is_input_port(sc->input_port)) &&
	      (port_file(sc->input_port) != stdin))
	    {
	      const char *filename = NULL;
	      int line;
	      
	      filename = port_filename(sc->input_port);
	      line = port_line_number(sc->input_port);
	      
	      if (filename)
		format_to_output(sc, error_port, "\n;  ~A[~D]",
				 list_2(sc, make_protected_string(sc, filename), make_integer(sc, line)), 
				 type == sc->FORMAT_ERROR);
	      else 
		{
		  if (line > 0)
		    format_to_output(sc, error_port, "\n;  line ~D", 
				     list_1(sc, make_integer(sc, line)), 
				     type == sc->FORMAT_ERROR);
		}
	    }
	  else
	    {
	      if ((call_file != NULL) &&
		  (call_name != NULL) &&
		  (call_line >= 0))
		{
		  format_to_output(sc, error_port, "\n;  ~A ~A[~D]",
				   list_3(sc, 
					  make_protected_string(sc, call_name), 
					  make_protected_string(sc, call_file), 
					  make_integer(sc, call_line)), 
				   type == sc->FORMAT_ERROR);
		}
	    }
	  s7_newline(sc, error_port);

	  if (is_pair(vector_element(sc->error_info, ERROR_CODE)))
	    {
	      format_to_output(sc, error_port, ";    ~A", 
			       list_1(sc, vector_element(sc->error_info, ERROR_CODE)),
			       type == sc->FORMAT_ERROR);
	      s7_newline(sc, error_port);

	      if (s7_is_string(vector_element(sc->error_info, ERROR_CODE_FILE)))
		{
		  format_to_output(sc, error_port, ";    [~S, line ~D]",
				   list_2(sc, 
					  vector_element(sc->error_info, ERROR_CODE_FILE), 
					  vector_element(sc->error_info, ERROR_CODE_LINE)), 
				   type == sc->FORMAT_ERROR);
		  s7_newline(sc, error_port);
		}
	    }

	  /* look for __func__ in the error environment etc
	   */
	  x = find_local_symbol(sc, vector_element(sc->error_info, ERROR_ENVIRONMENT), sc->__FUNC__);  /* returns nil if no __func__ */

	  if ((is_slot(x)) &&
	      (error_port != sc->F))
	    {
	      s7_display(sc, make_protected_string(sc, ";    "), error_port);
	      s7_display(sc, slot_value(x), error_port);
	      s7_newline(sc, error_port);
	    }
	  
	  if ((exit_eval) &&
	      (sc->error_exiter))
	    (*(sc->error_exiter))();

	  /* if (s7_is_continuation(type))
	   *   go into repl here with access to continuation?  Or expect *error-handler* to deal with it?
	   */
	  sc->value = type;
	  stack_reset(sc);
	  sc->op = OP_ERROR_QUIT;
	}
    }

  if (sc->longjmp_ok)
    {
      longjmp(sc->goto_start, 1); /* this is trying to clear the C stack back to some clean state */
    }

  return(type);
}


s7_pointer s7_error(s7_scheme *sc, s7_pointer type, s7_pointer info)
{
  return(s7_error_1(sc, type, info, false));
}


s7_pointer s7_error_and_exit(s7_scheme *sc, s7_pointer type, s7_pointer info)
{
  return(s7_error_1(sc, type, info, true));
}


static s7_pointer apply_error(s7_scheme *sc, s7_pointer obj, s7_pointer args)
{
  /* the operator type is needed here else the error message is confusing:
   *    (apply '+ (list 1 2))) -> ;attempt to apply + to (1 2)?
   */
  if (is_null(obj))
    return(s7_error(sc, sc->SYNTAX_ERROR, list_2(sc, make_protected_string(sc, "attempt to apply nil to ~S?"), args)));
  return(s7_error(sc, sc->SYNTAX_ERROR, 
		  list_4(sc, 
			 make_protected_string(sc, "attempt to apply the ~A ~S to ~S?"), 
			 make_protected_string(sc, type_name(obj)), obj, args)));
}


static s7_pointer eval_error(s7_scheme *sc, const char *errmsg, s7_pointer obj)
{
  return(s7_error(sc, sc->SYNTAX_ERROR, list_2(sc, make_protected_string(sc, errmsg), obj)));
}


static s7_pointer eval_error_with_name(s7_scheme *sc, const char *errmsg, s7_pointer obj)
{
  return(s7_error(sc, sc->SYNTAX_ERROR, 
		  list_3(sc, 
			 make_protected_string(sc, errmsg),
			 make_protected_string(sc, op_names[(int)(sc->op)]),
			 obj)));
}


static s7_pointer eval_error_no_arg(s7_scheme *sc, const char *errmsg)
{
  return(s7_error(sc, sc->SYNTAX_ERROR, list_1(sc, make_protected_string(sc, errmsg))));
}


static s7_pointer read_error(s7_scheme *sc, const char *errmsg)
{
  /* reader errors happen before the evaluator gets involved, so forms such as:
   *   (catch #t (lambda () (car '( . ))) (lambda arg 'error))
   * do not catch the error if we simply signal an error when we encounter it.
   */
  char *msg;
  int len;
  s7_pointer pt;
  pt = sc->input_port;

  /* make an heroic effort to find where we slid off the tracks */

  if (is_string_port(sc->input_port))
    {
      #define QUOTE_SIZE 40
      unsigned int i, j, start = 0, end, slen;
      char *recent_input = NULL;

      /* we can run off the end in cases like (eval-string "(. . ,.)") or (eval-string " (@ . ,.)") */
      if (port_string_point(pt) >= port_string_length(pt))        
	port_string_point(pt) = port_string_length(pt) - 1;

      /* start at current position and look back a few chars */
      for (i = port_string_point(pt), j = 0; (i > 0) && (j < QUOTE_SIZE); i--, j++)
	if ((port_string(pt)[i] == '\0') ||
	    (port_string(pt)[i] == '\n') ||
	    (port_string(pt)[i] == '\r'))
	  break;
      start = i;

      /* start at current position and look ahead a few chars */
      for (i = port_string_point(pt), j = 0; (i < port_string_length(pt)) && (j < QUOTE_SIZE); i++, j++)
	if ((port_string(pt)[i] == '\0') ||
	    (port_string(pt)[i] == '\n') ||
	    (port_string(pt)[i] == '\r'))
	  break;

      end = i;
      slen = end - start;
      /* hopefully this is more or less the current line where the read error happened */

      if (slen > 0)
	{
	  recent_input = (char *)calloc((slen + 9), sizeof(char));
	  for (i = 0; i < (slen + 8); i++) recent_input[i] = '.';
	  recent_input[3] = ' ';
	  recent_input[slen + 4] = ' ';
	  for (i = 0; i < slen; i++) recent_input[i + 4] = port_string(pt)[start + i];
	}

      if ((port_line_number(pt) > 0) &&
	  (port_filename(pt)))
	{
	  len = safe_strlen(recent_input) + safe_strlen(errmsg) + safe_strlen(port_filename(pt)) + safe_strlen(sc->current_file) + 64;
	  msg = (char *)malloc(len * sizeof(char));
	  len = snprintf(msg, len, "%s: %s %s[%d], last top-level form at: %s[%d]", 
			 errmsg, (recent_input) ? recent_input : "", port_filename(pt), port_line_number(pt),
			 sc->current_file, sc->current_line);
	}
      else
	{
	  len = safe_strlen(recent_input) + safe_strlen(errmsg) + safe_strlen(sc->current_file) + 64;
	  msg = (char *)malloc(len * sizeof(char));

	  if ((sc->current_file) &&
	      (sc->current_line >= 0))
	    len = snprintf(msg, len, "%s: %s, last top-level form at %s[%d]", 
			   errmsg, (recent_input) ? recent_input : "",
			   sc->current_file, sc->current_line);
	  else len = snprintf(msg, len, "%s: %s", errmsg, (recent_input) ? recent_input : "");
	}
      
      if (recent_input) free(recent_input);
      return(s7_error(sc, sc->READ_ERROR, 
		      list_1(sc, make_string_uncopied_with_length(sc, msg, len))));
    }

  if ((port_line_number(pt) > 0) &&
      (port_filename(pt)))
    {
      len = safe_strlen(errmsg) + safe_strlen(port_filename(pt)) + safe_strlen(sc->current_file) + 64;
      msg = (char *)malloc(len * sizeof(char));
      len = snprintf(msg, len, "%s %s[%d], last top-level form at %s[%d]", 
		     errmsg, port_filename(pt), port_line_number(pt), 
		     sc->current_file, sc->current_line);
      return(s7_error(sc, sc->READ_ERROR, 
		      list_1(sc, make_string_uncopied_with_length(sc, msg, len))));
    }

  return(s7_error(sc, sc->READ_ERROR, 
		  list_1(sc, make_protected_string(sc, (char *)errmsg))));
}


static s7_pointer g_error(s7_scheme *sc, s7_pointer args)
{
  #define H_error "(error type ...) signals an error.  The 'type' can be used with catch to trap \
particular errors.  If the error is not caught, s7 treats the 2nd argument as a format control string, \
and applies it to the rest of the arguments."

  if (is_not_null(args))
    {
      if (s7_is_string(car(args)))                    /* CL-style error? -- use tag = 'no-catch */
	{
	  s7_error(sc, make_symbol(sc, "no-catch"), args);
	  return(sc->UNSPECIFIED);
	}
      return(s7_error(sc, car(args), cdr(args)));
    }
  return(s7_error(sc, sc->NIL, sc->NIL));
}


static char *truncate_string(char *form, int len, bool use_write)
{
  int form_len;

  form_len = safe_strlen(form);
  if (form_len <= len)
    return(form);

  if (use_write)
    {
      /* I guess we need to protect the outer double quotes in this case */
      int i;
      for (i = len - 4; i >= (len / 2); i--)
	if (is_white_space((int)form[i]))
	  {
	    form[i] = '.';
	    form[i + 1] = '.';
	    form[i + 2] = '.';
	    form[i + 3] = '"';
	    form[i + 4] = '\0';
	    return(form);
	  }
      i = len - 4;
      if (i > 0)
	{
	  form[i] = '.';
	  form[i + 1] = '.';
	  form[i + 2] = '.';
	  form[i + 3] = '"';
	  form[i + 4] = '\0';
	}
      else 
	{
	  if (len >= 2)
	    {
	      form[len - 1] = '"';
	      form[len] = '\0';
	    }
	}
    }
  else
    {
      int i;
      for (i = len - 3; i >= (len / 2); i--)
	if (is_white_space((int)form[i]))
	  {
	    form[i] = '.';
	    form[i + 1] = '.';
	    form[i + 2] = '.';
	    form[i + 3] = '\0';
	    return(form);
	  }
      i = len - 3;
      if (i >= 0)
	{
	  form[i] = '.';
	  form[i + 1] = '.';
	  form[i + 2] = '.';
	  form[i + 3] = '\0';
	}
      else form[len] = '\0';
    }
  return(form);
}


static char *object_to_truncated_string(s7_scheme *sc, s7_pointer p, int len)
{
  return(truncate_string(s7_object_to_c_string(sc, p), len, false));
}


static bool too_many_arguments(s7_scheme *sc, s7_pointer proc, int len)
{
  s7_pointer proc_args;
  proc_args = s7_procedure_arity(sc, proc);
  return((is_pair(proc_args)) &&
	 (caddr(proc_args) == sc->F) &&
	 ((s7_integer(car(proc_args)) + s7_integer(cadr(proc_args))) < len));
}


#define SYMBOL_TO_VALUE(Sc, Sym) ((is_global(Sym)) ? slot_value(global_slot(Sym)) : finder(Sc, Sym))

static char *missing_close_paren_syntax_check(s7_scheme *sc, s7_pointer lst)
{
  s7_pointer p;
  char *msg = NULL;
  for (p = lst; is_pair(p); p = cdr(p))
    {
      if (is_pair(car(p)))
	{
	  if (is_symbol(caar(p)))
	    {
	      int len;
	      
	      len = s7_list_length(sc, car(p));
	      if (((s7_is_eq(caar(p), s7_make_symbol(sc, "if"))) &&
		   (len > 4)) ||
		  ((s7_is_procedure(SYMBOL_TO_VALUE(sc, caar(p)))) &&
		   (too_many_arguments(sc, SYMBOL_TO_VALUE(sc, caar(p)), len))) ||
		  ((s7_is_eq(caar(p), s7_make_symbol(sc, "define"))) &&
		   (is_pair(cdr(p))) &&
		   (is_symbol(cadr(p))) &&
		   (len > 3)))
		{
		  int msg_len, form_len;
		  char *form;
		  /* it's very tricky to try to see other errors here, especially because 'case'
		   *   can have syntax names in its key lists.  Even this may get fooled, but
		   *   I'm hoping that more often than not, it will help track down the missing
		   *   close paren.
		   */

		  form = object_to_truncated_string(sc, car(p), 80);
		  form_len = safe_strlen(form);
		  msg_len = form_len + 128;
		  msg = (char *)calloc(msg_len, sizeof(char));
		  snprintf(msg, msg_len, ";  this looks bogus: %s", form);
		  free(form);

		  return(msg);
		}
	    }
	  msg = missing_close_paren_syntax_check(sc, car(p));
	  if (msg) 
	    return(msg);
	}
    }
  return(NULL);
}


static s7_pointer missing_close_paren_error(s7_scheme *sc)
{
  int len;
  char *msg, *syntax_msg = NULL;
  s7_pointer pt;
  pt = sc->input_port;

  /* it's hard to give a good idea here of where the missing paren is because we've already
   *   popped off all the stacked info, following ')' until eof.
   * but the current incoming program code is in sc->args, reversed at its top level,
   *   so it's worth looking for some problem involving too many clauses (if) or arguments, etc.
   * this can be a hard bug to track down in a large program, so s7 really has to make an effort to help.
   */

  if (is_pair(sc->args))
    {
      sc->x = s7_reverse(sc, sc->args);
      syntax_msg = missing_close_paren_syntax_check(sc, sc->x);
      /* if syntax_msg is null, we didn't find the problem, so perhaps show it indented?
       */
    }

  if ((port_line_number(pt) > 0) &&
      (port_filename(pt)))
    {
      len = safe_strlen(port_filename(pt)) + safe_strlen(sc->current_file) + safe_strlen(syntax_msg) + 128;
      msg = (char *)malloc(len * sizeof(char));
      if (syntax_msg)
	{
	  len = snprintf(msg, len, "missing close paren, %s[%d], last top-level form at %s[%d]\n%s", 
			 port_filename(pt), port_line_number(pt), 
			 sc->current_file, sc->current_line, syntax_msg);
	  free(syntax_msg);
	}
      else len = snprintf(msg, len, "missing close paren, %s[%d], last top-level form at %s[%d]", 
			  port_filename(pt), port_line_number(pt), 
			  sc->current_file, sc->current_line);
      return(s7_error(sc, sc->READ_ERROR, make_string_uncopied_with_length(sc, msg, len)));
    }

  if (syntax_msg)
    {
      len = safe_strlen(syntax_msg) + 128;
      msg = (char *)malloc(len * sizeof(char));
      len = snprintf(msg, len, "missing close paren\n%s\n", syntax_msg);
      free(syntax_msg);
      return(s7_error(sc, sc->READ_ERROR, make_string_uncopied_with_length(sc, msg, len)));
    }
  return(s7_error(sc, sc->READ_ERROR, 
		  list_1(sc, make_protected_string(sc, "missing close paren"))));
}


static void improper_arglist_error(s7_scheme *sc)
{
  /* sc->code is the last (dotted) arg, sc->args is the arglist reversed not including sc->code
   *   the original was `(,@(reverse args) . ,code) essentially
   */
  s7_error(sc, sc->SYNTAX_ERROR, 
	   list_2(sc,
		  make_protected_string(sc, "improper list of arguments: ~A"),
		  append_in_place(sc, safe_reverse_in_place(sc, sc->args), sc->code)));
}


/* (let ((a 43) (c 123)) (define (hi b) (stacktrace) b) (hi a)) */

static void display_frame(s7_scheme *sc, s7_pointer envir, s7_pointer port)
{
  s7_pointer frame;
  int i;
  for (i = 0, frame = envir; is_slot(environment_slots(frame)) && (i < 4); i++, frame = next_environment(frame))
    {
      s7_pointer arg;
      arg = environment_slots(frame);
      while (is_slot(arg))
	{
	  if ((slot_symbol(arg) == sc->__FUNC__) ||
	      (!s7_is_procedure(slot_value(arg))))
	    {
	      s7_write(sc, arg, port);
	      s7_write_char(sc, ' ', port);
	    }
	  arg = next_slot(arg);
	}
      s7_newline(sc, port);    
    }
}


static s7_pointer g_stacktrace(s7_scheme *sc, s7_pointer args)
{
  /* 3 cases currently: 
   *    if args=nil, show current stack
   *           =vector, assume it is a vector of envs from *error-info*
   *           =continuation, show its stack
   * if trailing arg is a port, it sets where the output goes
   */
  #define H_stacktrace "(stacktrace (obj #f) (port (current-output-port))) displays a stacktrace.  If obj is not \
given, the current stack is displayed, if obj is *error-info*, the stack at the point of the error is displayed, and if obj \
is a continuation, its stack is displayed.  If the trailing port argument is not given, \
output is sent to the current-output-port."

  int i, top = 0;
  s7_pointer stk = sc->F, port, obj;
  port = s7_current_output_port(sc);

  if (is_not_null(args))
    {
      if (is_output_port(car(args)))
	{
	  port = car(args);
	  args = cdr(args);
	}
      else
	{
	  if (is_not_null(cdr(args)))
	    {
	      if ((is_output_port(cadr(args))) &&
		  (!port_is_closed(cadr(args))))
		port = cadr(args);
	      else return(s7_wrong_type_arg_error(sc, "stacktrace", 2, cadr(args), "an open output port"));
	    }
	}
      obj = car(args);
    }
  else obj = sc->F;

  /* *error-info* is the special case here
   *   TODO: stacktrace + *error-info* is too hard to use, and appears to repeat the local vars
   */
  if (s7_is_vector(obj))
    {
      if (vector_length(obj) < ERROR_INFO_SIZE)
	return(s7_wrong_type_arg_error(sc, "stacktrace", 1, obj, "*error-info*"));

      for (i = ERROR_ENVIRONMENT; i < ERROR_INFO_SIZE; i++)
	{
	  if (vector_element(obj, i) == ERROR_INFO_DEFAULT)
	    break;
	  display_frame(sc, vector_element(obj, i), port);
	}
      return(sc->UNSPECIFIED);
    }

  if (is_null(args))
    {
      top = s7_stack_top(sc);
      stk = sc->stack;
    }
  else
    {
      if (s7_is_continuation(obj))
	{
	  top = continuation_stack_top(obj);
	  stk = continuation_stack(obj);
	}
    }
  if (stk == sc->F)
    return(s7_wrong_type_arg_error(sc, "stacktrace", 0, args, "a vector, or continuation"));
  
  if (sc->envir != stack_environment(stk, top - 1))
    display_frame(sc, sc->envir, port);
  for (i = top - 1; i > 0; i -= 4)
    display_frame(sc, stack_environment(stk, i), port);

  return(sc->UNSPECIFIED);
}


s7_pointer s7_stacktrace(s7_scheme *sc, s7_pointer arg)
{
  if (s7_is_list(sc, arg))
    return(g_stacktrace(sc, arg));
  return(g_stacktrace(sc, list_1(sc, arg)));
}




/* -------------------------------- leftovers -------------------------------- */


bool (*s7_begin_hook(s7_scheme *sc))(s7_scheme *sc)
{
  return(sc->begin_hook);
}


void s7_set_begin_hook(s7_scheme *sc, bool (*hook)(s7_scheme *sc))
{
  sc->begin_hook = hook;
}


void s7_quit(s7_scheme *sc)
{
  sc->longjmp_ok = false;
  pop_input_port(sc);

  stack_reset(sc);
  push_stack(sc, OP_EVAL_DONE, sc->NIL, sc->NIL);
}


static s7_pointer apply_list_star(s7_scheme *sc, s7_pointer d) 
{
  s7_pointer p, q;
  /* we check this ahead of time: if (is_null(cdr(d))) return(car(d)); */
  p = cons(sc, car(d), cdr(d));
  q = p;
  while (is_not_null(cdr(cdr(p))))
    {
      d = cons(sc, car(p), cdr(p));
      if (is_not_null(cdr(cdr(p))))
	p = cdr(d);
    }
  cdr(p) = car(cdr(p));
  return(q);
}


static s7_pointer g_apply(s7_scheme *sc, s7_pointer args)
{
  #define H_apply "(apply func ...) applies func to the rest of the arguments"

  sc->code = car(args);
  if (is_null(cdr(args)))
    sc->args = sc->NIL;
  else 
    {
      if (is_safe_procedure(sc->code))
	{
	  s7_pointer p, q;

	  for (q = args, p = cdr(args); is_not_null(cdr(p)); q = p, p = cdr(p));
	  /* the last arg is supposed to be a list, it will be spliced onto the end of the previous arg list (if any) below */

	  if (!is_proper_list(sc, car(p)))        /* (apply + #f) etc */
	    return(s7_error(sc, sc->WRONG_TYPE_ARG, 
			    list_2(sc, 
				   make_protected_string(sc, "apply's last argument should be a proper list: ~A"),
				   args)));
	  cdr(q) = car(p);                       
	  push_stack(sc, OP_APPLY, cdr(args), sc->code);
	  return(sc->NIL);
	}
      else
	{
	  /* here we have to copy the arg list */
	  if (is_null(cddr(args)))
	    sc->args = cadr(args);
	  else sc->args = apply_list_star(sc, cdr(args));

	  if (!is_proper_list(sc, sc->args))        /* (apply + #f) etc */
	    return(s7_error(sc, sc->WRONG_TYPE_ARG, 
			    list_2(sc, 
				   make_protected_string(sc, "apply's last argument should be a proper list: ~A"),
				   args)));
	}
    }

  if (is_any_macro(sc->code))                   /* (apply mac '(3)) -> (apply mac '((mac 3))) */
    {
      push_stack(sc, OP_EVAL_MACRO, sc->NIL, sc->args);
      sc->args = list_1(sc, cons(sc, sc->code, sc->args));
    }
  push_stack(sc, OP_APPLY, sc->args, sc->code);
  return(sc->NIL);
}


s7_pointer s7_eval(s7_scheme *sc, s7_pointer code, s7_pointer e)
{
  /* TODO: error handling */
  push_stack(sc, OP_EVAL_DONE, sc->args, sc->code);
  sc->code = code;
  sc->envir = e;
  eval(sc, OP_BEGIN);
  return(sc->value);
}


s7_pointer s7_eval_form(s7_scheme *sc, s7_pointer form, s7_pointer e)
{
  /* TODO: error handling */
  push_stack(sc, OP_EVAL_DONE, sc->args, sc->code);
  sc->code = form;
  sc->envir = e;
  eval(sc, OP_EVAL);
  return(sc->value);
}


static s7_pointer g_eval(s7_scheme *sc, s7_pointer args)
{
  #define H_eval "(eval code (env (current-environment))) evaluates code in the environment env. 'env' \
defaults to the current environment; to evaluate something in the top-level environment instead, \
pass (global-environment):\n\
\n\
  (define x 32) \n\
  (let ((x 3))\n\
    (eval 'x (global-environment)))\n\
\n\
  returns 32"
  
  if (is_not_null(cdr(args)))
    {
      s7_pointer e;
      e = cadr(args);
      if (!is_environment(e))
	return(s7_wrong_type_arg_error(sc, "eval", 2, e, "an environment"));
      if (e == sc->global_env)
	sc->envir = sc->NIL;
      else sc->envir = e;
    }
  sc->code = car(args);

  if (s7_stack_top(sc) < 12)
    push_stack(sc, OP_BARRIER, sc->NIL, sc->NIL);
  push_stack(sc, OP_EVAL, sc->args, sc->code);
  return(sc->NIL);
}


s7_pointer s7_call(s7_scheme *sc, s7_pointer func, s7_pointer args)
{
  bool old_longjmp;
  jmp_buf old_goto_start;

  /* this can be called while we are in the eval loop (within eval_c_string for instance),
   *   and if we reset the stack, the previously running evaluation steps off the end
   *   of the stack == segfault. 
   */

  if (is_c_function(func))
    return(c_function_call(func)(sc, args));  /* no check for wrong-number-of-args -- is that reasonable? */

  sc->temp1 = func;
  sc->temp2 = args;

  old_longjmp = sc->longjmp_ok;
  memcpy((void *)old_goto_start, (void *)(sc->goto_start), sizeof(jmp_buf));

  /* if an error occurs during s7_call, and it is caught, catch (above) wants to longjmp
   *   to its caller to complete the error handler evaluation so that the C stack is
   *   cleaned up -- this means we have to have the longjmp location set here, but
   *   we could get here from anywhere, so we need to save and restore the incoming
   *   longjmp location.
   */

  sc->longjmp_ok = true;
  if (setjmp(sc->goto_start) != 0) /* returning from s7_error catch handler */
    {
      sc->longjmp_ok = old_longjmp;
      memcpy((void *)(sc->goto_start), (void *)old_goto_start, sizeof(jmp_buf));
      
      if ((sc->op == OP_ERROR_QUIT) &&
	  (sc->longjmp_ok))
	{
	  longjmp(sc->goto_start, 1); /* this is trying to clear the C stack back to some clean state */
	}

      eval(sc, sc->op); 
      /* sc->op can be OP_APPLY if s7_call raised an error that was caught (via catch) -- we're about to go to the error handler */
      return(sc->value);
    }

  push_stack(sc, OP_EVAL_DONE, sc->args, sc->code); /* this saves the current evaluation and will eventually finish this (possibly) nested call */
  sc->args = args; 
  sc->code = func; 

  /* besides a closure, "func" can also be an object (T_C_OBJECT) -- in Snd, a generator for example
   */

  eval(sc, OP_APPLY);

  sc->longjmp_ok = old_longjmp;
  memcpy((void *)(sc->goto_start), (void *)old_goto_start, sizeof(jmp_buf));

  return(sc->value);
} 


s7_pointer s7_call_with_location(s7_scheme *sc, s7_pointer func, s7_pointer args, const char *caller, const char *file, int line)
{ 
  s7_pointer result;

  if (caller)
    {
      sc->s7_call_name = caller;
      sc->s7_call_file = file;
      sc->s7_call_line = line;
    }

  result = s7_call(sc, func, args);
  
  if (caller)
    {
      sc->s7_call_name = NULL;
      sc->s7_call_file = NULL;
      sc->s7_call_line = -1;
    }

  return(result);
}


static s7_pointer g_s7_version(s7_scheme *sc, s7_pointer args)
{
  #define H_s7_version "(s7-version) returns some string describing the current s7"
#if WITH_COUNTS
  report_counts();
#endif
  return(s7_make_string(sc, "s7 " S7_VERSION ", " S7_DATE));
}



/* ---------------------------------------- map and for-each ---------------------------------------- */

static s7_Int applicable_length(s7_scheme *sc, s7_pointer obj)
{
  switch (type(obj))
    {
    case T_PAIR:
      {
	s7_Int len;
	len = s7_list_length(sc, obj);
	if (len < 0)             /* dotted (does not include the final cdr -- perhaps this is a bug) */
	  return(-len);          /*         it means that (map abs '(1 . "hi")) returns '(1) */
	if (len == 0)            /* circular */
	  return(S7_LONG_MAX);
	return(len);
      }

    case T_C_OBJECT:
      {
	/* both map and for-each assume sc->x|y|z are unchanged across this call */
	s7_pointer result;

	result = object_length(sc, obj);
	if (s7_is_integer(result))   /* we need to check, else misinterpreting it as an integer can lead to a infinite loop */
	  return(s7_integer(result));
	return(-1);
      }

    case T_STRING:
      return(string_length(obj));

    case T_VECTOR:
      return(vector_length(obj));

    case T_HASH_TABLE:
      return(hash_table_entries(obj));

    case T_ENVIRONMENT:
      return(environment_length(sc, obj));

    case T_NIL:
      return(0);
    }
  
  return(-1);
}


static bool next_for_each(s7_scheme *sc)
{
  s7_pointer x, y, z, vargs, fargs;
  int loc, zloc = -1;

  z = sc->NIL;

  vargs = counter_list(sc->args);
  fargs = counter_result(sc->args);
  loc = counter_count(sc->args);

  /* for-each func ... -- each trailing sequence arg contributes one arg to the current call on func, 
   *   so in the next loop, gather one arg from each sequence.
   */

  for (x = fargs, y = vargs; is_not_null(x); x = cdr(x), y = cdr(y))
    switch (type(car(y)))
      {
      case T_PAIR:
	car(x) = caar(y);
	car(y) = cdar(y);
	break;

      case T_C_OBJECT: 
	{
	  int save_x = -1, save_y = -1, save_z = -1;
	  if (is_null(z))
	    {
	      z = list_1(sc, make_integer(sc, loc));

	      /* we can't use car(sc->args) directly here -- it is a counter.
	       *   the object application (the getter function) might return the index.
	       *   Then, if we pre-increment, and the for-each application sees the incremented value.
	       *
               *    (let ((ctr ((cadr (make-type :getter (lambda (a b) b) :length (lambda (a) 4))))) (sum 0))
               *      (for-each (lambda (a b) (set! sum (+ sum a b))) ctr ctr) sum)
	       */
	      zloc = s7_gc_protect(sc, z);
	    }

	  SAVE_X_Y_Z(save_x, save_y, save_z);
	  car(x) = (*(object_ref(car(y))))(sc, car(y), z);
	  RESTORE_X_Y_Z(save_x, save_y, save_z);
	}
	break;

      case T_VECTOR:
	car(x) = vector_element(car(y), loc); 
	break;

      case T_CLOSURE: /* hash-table via an iterator */
	car(x) = g_hash_table_iterate(sc, hash_table_iterate_args(car(y)));  /* cdadadar? I suppose this accessor could be optimized */
	break;

      case T_STRING:
	car(x) = s7_make_character(sc, ((unsigned char)(string_value(car(y))[loc])));
	break;

      default:           /* see comment in next_map: (let ((L (list 1 2 3 4 5))) (for-each (lambda (x) (set-cdr! (cddr L) 5) (display x)) L)) */
	if (is_not_null(z))
	  s7_gc_unprotect_at(sc, zloc);
	return(false);
	break;
      }

  if (zloc != -1)
    s7_gc_unprotect_at(sc, zloc);

  counter_count(sc->args) = loc + 1;
  push_stack(sc, OP_FOR_EACH, sc->args, sc->code);
  sc->args = fargs;

  if (is_macro(sc->code))
    {
      push_stack(sc, OP_EVAL_MACRO, sc->NIL, sc->args);
      /* we can use a temp here because the arg list will be copied in OP_FOR_EACH */
      car(sc->TEMP_CELL_1) = sc->code;
      cdr(sc->TEMP_CELL_1) = sc->args;
      sc->args = sc->TEMP_CELL;
    }
  return(true);
}


static bool is_sequence(s7_scheme *sc, s7_pointer p)
{
  return((s7_is_list(sc, p)) ||
	 (s7_is_vector(p)) ||
	 (s7_is_string(p)) ||
	 (s7_is_hash_table(p)) ||
	 (is_environment(p)) ||
	 (is_c_object(p)));
}


static s7_pointer make_counter(s7_scheme *sc, s7_pointer result, s7_pointer lst, long long int count, long long int length)
{
  s7_pointer x;
  NEW_CELL(sc, x);
  counter_result(x) = result;
  counter_list(x) = lst;
  counter_count(x) = count;
  counter_length(x) = length;
  set_type(x, T_COUNTER);
  return(x);
}


#if WITH_OPTIMIZATION
static void initialize_safe_do(s7_scheme *sc, s7_pointer tree);
#endif


static s7_pointer g_for_each(s7_scheme *sc, s7_pointer args)
{
  #define H_for_each "(for-each proc object . objects) applies proc to each element of the objects traversed in parallel. \
Each object can be a list (the normal case), string, vector, hash-table, or any applicable object."

  /* (for-each (lambda (n) (format #t "~A " n)) (vct 1.0 2.0 3.0)) */
  s7_Int len = 0; 
  /* "int" here is unhappy on 64-bit machines, and "long int" is unhappy on 32-bit machines, so try long long int.
   *    string_length is an int.
   */
  s7_pointer obj, x;
  sc->code = car(args); /* the function */

  /* macro application requires the entire call as the argument to apply, but apply itself fixes this up.
   *  that is, g_apply checks, then goes to OP_EVAL_MACRO after OP_APPLY with the fixed up list,
   *  but here we are simply sending it to OP_APPLY, so
   *
   *    (define-macro (hi a) `(+ ,a 1))
   *    (apply hi '(1))                 ; internally rewritten as (apply hi '((hi 1)))
   *    2                               ; apply adds the evaluation if its 1st arg is a macro
   *    (map hi '((hi 1) (hi 2)))       ; here we've rewritten the arg lists by hand
   *    ((+ 1 1) (+ 2 1))               ; but no evaluation
   *
   * ideally I think it should be
   *
   *    (map hi '(1 2))
   *    (2 3)
   *
   * OP_APPLY knows this is happening (in the 2nd case) and raises an error -- what would break
   *   if we handle it locally instead?  This actually affects only T_C_MACRO (quasiquote) --
   *   normal macros/bacros would still be broken.
   * or could we rewrite the args to fit just as in apply? (also need the evaluation)
   */

  /* before checking len=0, we need to check that the arguments are all sequences (this is like our handling of args to + for example)
   *   otherwise (for-each = "" 123) -> #<unspecified> 
   * the function may not actually be applicable to its sequence elements, but that isn't an error:
   *   (map abs "") -> '()
   */
  obj = cadr(args); 

  if ((is_pair(obj)) &&                                               /* arg is a list */
      (is_null(cddr(args))))                                          /* only one list arg */
    {
      len = s7_list_length(sc, obj);
      if (len > 0)                                                    /* a proper list arg */
	{
	  if ((type(sc->code) == T_CLOSURE) &&                        /* not lambda* that might get confused about arg names */
	      (is_proper_list(sc, closure_args(sc->code))) &&         /* not a rest arg: not is_pair: (lambda (x . args) arg) */
	      (!is_immutable(car(closure_args(sc->code)))) &&         /* not a bad arg name! */
	      (!symbol_has_accessor(car(closure_args(sc->code)))) &&  /* not wrapped in an accessor */
	      (safe_list_length(sc, closure_args(sc->code)) == 1))    /* closure takes just one arg */
	    {
	      /* one list arg -- special, but very common case
	       *
	       * checking for and (only rarely) using the safe_do machinery here was much slower.
	       * one-liner here plus separate loop in eval was only slightly faster 
	       */
	      push_stack(sc, OP_FOR_EACH_SIMPLE, obj, sc->code);
	      return(sc->UNSPECIFIED);

	      /* this, and map, across a string (say) or other sequence involving
	       *   only 1 arg could be optimized in the same way. OP_FOR_EACH|MAP_STRING|VECTOR|C_OBJECT
	       */

	      /* if len is 1 can't we just call apply?
	       */
	    }

	  if ((is_safe_procedure(sc->code)) &&
	      (is_c_function(sc->code)) &&
	      (args_match(sc, sc->code, 1)))
	    {
	      s7_function func;
	      s7_pointer p;
	      /* sc->z = cons(sc, sc->F, sc->NIL); */ /* what is this about? */
	      func = c_function_call(sc->code);
	      for (p = obj; is_pair(p); p = cdr(p))
		{
		  car(sc->T1_1) = car(p);
		  (*func)(sc, sc->T1_1);
		}
	      return(sc->UNSPECIFIED);
	    }
	}

      if (len < 0)
	len = -len; 
      else
	if (len == 0) 
	  len = S7_LONG_MAX;
    }
  else 
    {
      for (x = cdr(args); is_not_null(x); x = cdr(x))
	{
	  s7_Int nlen;
	  if (!is_sequence(sc, car(x)))
	    return(s7_wrong_type_arg_error(sc, "for-each", position_of(x, args), car(x), "a sequence"));

	  nlen = applicable_length(sc, car(x));
	  if (nlen < 0)
	    return(s7_wrong_type_arg_error(sc, "for-each", position_of(x, args), car(x), "a vector, list, string, hash-table, or applicable object"));
	  if ((x == cdr(args)) || (nlen < len))
	    len = nlen;
	  if (len == 0) 
	    {
	      for (x = cdr(x); is_not_null(x); x = cdr(x))
		if (!is_sequence(sc, car(x)))
		  return(s7_wrong_type_arg_error(sc, "for-each", position_of(x, args), car(x), "a sequence"));
	      break;   /* need error check below */
	    }
	}
    }

  if (len != 0)
    {
      sc->x = list_1(sc, sc->NIL);
      if (s7_is_hash_table(obj))
	sc->z = list_1(sc, g_make_hash_table_iterator(sc, cdr(args)));
      else 
	{ 
	  /* (let ((a 1) (b 2)) (for-each (lambda (slot) (format #t "~A~%" slot)) (current-environment))) */
	  if (is_environment(obj))
	    sc->z = list_1(sc, s7_environment_to_list(sc, obj));
	  else sc->z = list_1(sc, obj);
	}

      /* we have to copy the args if any of them is a list:
       *     (let* ((x (list (list 1 2 3))) (y (apply for-each abs x))) (list x y))
       *  (is this trying to say that the for-each loop might otherwise change the original list as it cdrs down it?)
       */

      if (is_not_null(cddr(args)))
	{
	  for (x = cddr(args); is_not_null(x); x = cdr(x))
	    {
	      sc->x = cons(sc, sc->NIL, sc->x);          /* we're making a list to be filled in later with the individual args */

	      if (s7_is_hash_table(car(x)))
		sc->z = cons_unchecked(sc, g_make_hash_table_iterator(sc, x), sc->z);
	      else
		{
		  if (is_environment(car(x)))
		    sc->z = cons(sc, s7_environment_to_list(sc, car(x)), sc->z);
		  else sc->z = cons_unchecked(sc, car(x), sc->z);
		}
	    }
	}
    }
  else /* len == 0 */
    {
      /* here we can't depend on OP_APPLY to do the error check on the 1st arg:
       *   (map 0 '()) -> '()
       * so we check by hand before returning #<unspecified>
       */
      if (((typeflag(sc->code) & (T_ANY_MACRO | T_PROCEDURE)) != 0) ||
	  (is_pair(sc->code)) ||                   /* if this used is_sequence (above), (map '()...) would not be an error */
	  (s7_is_string(sc->code)) ||
	  (s7_is_vector(sc->code)) ||
	  (s7_is_hash_table(sc->code)) ||
	  (is_environment(sc->code)) ||
	  (is_hook(sc->code)) ||
	  (is_syntax(sc->code)))
	return(sc->UNSPECIFIED);    /* circular -> S7_LONG_MAX in this case, so 0 -> nil */
      return(s7_wrong_type_arg_error(sc, "for-each", 1, sc->code, "a procedure or something applicable"));
    }

  if (len == S7_LONG_MAX)
    {
      /* if at this point len == S7_LONG_MAX, then all args are circular lists, assuming that
       *    we're not looking at some enormous vector or string -- perhaps -1 would be a
       *    better marker.  This actually might not be an error (the for-each function could
       *    have call-with-exit), but it seems better to complain about it.
       * 
       * this means that a make-type generator's length is tricky:
       *    (let ((ctr ((cadr (make-type :getter (lambda (a b) b) :length (lambda (a) (- (expt 2 31) 1)))))) (sum 0))
       *      (call-with-exit (lambda (go) (for-each (lambda (a) (set! sum (+ sum a)) (if (> sum 100) (go sum))) ctr))))
       * returns an error about circular lists, but should return 105.
       *
       * I think I'll at least check that the args were in fact lists.
       */
      for (x = cdr(args); (is_pair(x)) && (is_pair(car(x))); x = cdr(x)) {}
      if (!is_pair(x))
	return(s7_error(sc, sc->WRONG_TYPE_ARG, 
			list_2(sc, make_protected_string(sc, "for-each's arguments are circular lists! ~A"), cdr(args))));
    }

  sc->y = args;
  sc->args = make_counter(sc, sc->x, safe_reverse_in_place(sc, sc->z), 0, len);
  /* counter_result: func-arglist, counter_list: objects, counter_count: current count, counter_length: applicable length */

  sc->x = sc->NIL;
  sc->y = sc->NIL;
  sc->z = sc->NIL;

  push_stack(sc, OP_FOR_EACH, sc->args, sc->code);
  return(sc->UNSPECIFIED);
}


#if WITH_OPTIMIZATION

static bool is_lambda(s7_scheme *sc, s7_pointer sym)
{
  return((sym == sc->LAMBDA) && (symbol_id(sym) == 0));
  /* symbol_id==0 means it has never been rebound (T_GLOBAL might not be set for initial stuff) */
}


/* also call-with-exit/catch/map etc */

static s7_pointer for_each_1;
static s7_pointer g_for_each_1(s7_scheme *sc, s7_pointer args)
{
  /* 1st arg is (lambda (sym) ...) when sym has no complications, only 2 args passed
   */
  s7_pointer obj;
  s7_Int len;

  obj = cadr(args);
  if (!is_pair(obj))
    return(g_for_each(sc, args));
  
  len = s7_list_length(sc, obj);
  if (len == 0)                   /* circular list */
    return(g_for_each(sc, args)); /* this is faster than calling s7_error here -- weird! */
  
  push_stack(sc, OP_FOR_EACH_SIMPLE, obj, car(args));
  return(sc->UNSPECIFIED);
}

static s7_pointer for_each_2;
static s7_pointer g_for_each_2(s7_scheme *sc, s7_pointer args)
{
  /* 1st arg is (lambda (sym) (...)) -- one pair in lambda, when sym has no complications, only 2 args passed
   */
  s7_pointer obj;
  s7_Int len;

  obj = cadr(args);
  if (!is_pair(obj))
    return(g_for_each(sc, args));
  
  len = s7_list_length(sc, obj);
  if (len == 0)                   /* circular list */
    return(g_for_each(sc, args)); /* this is faster than calling s7_error here -- weird! */
  
  push_stack(sc, OP_FOR_EACH_SIMPLER, obj, car(args));
  return(sc->UNSPECIFIED);
}

static s7_pointer for_each_3;
static s7_pointer g_for_each_3(s7_scheme *sc, s7_pointer args)
{
  /* 1st arg is (lambda (sym) (...)) when sym has no complications, lambda body is safe, only 2 args passed
   */
  s7_pointer arg, obj, func;
  s7_Int len;

  func = car(args);
  obj = cadr(args);
  if (!is_pair(obj))
    return(g_for_each(sc, args));
  
  len = s7_list_length(sc, obj);
  if (len == 0)                   /* circular list */
    return(g_for_each(sc, args)); /* this is faster than calling s7_error here -- weird! */

  closure_environment(func) = new_frame_in_env(sc, closure_environment(func));
  for (arg = closure_args(func); is_pair(arg); arg = cdr(arg))
    s7_make_slot(sc, closure_environment(func), car(arg), sc->NIL);
  environment_slots(closure_environment(func)) = reverse_slots(sc, environment_slots(closure_environment(func)));
  
  push_stack(sc, OP_SAFE_FOR_EACH, obj, car(args));
  return(sc->UNSPECIFIED);
}

/* (let () (define (hi) (let ((lst '(1 2 3))) (for-each (lambda (a) a) lst))) (hi)) */

static s7_pointer for_each_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      s7_pointer lexpr;
      lexpr = cadr(expr);
      if (is_lambda(sc, car(lexpr)))
	{
	  s7_pointer largs, lbody;
	  largs = cadr(lexpr);
	  lbody = cddr(lexpr);
      	  if ((is_pair(largs)) &&
	      (is_null(cdr(largs))) &&
	      (!is_immutable(car(largs))) &&
	      (!symbol_has_accessor(car(largs))))
	    {
	      bool bad_set = false;
	      if (body_is_safe(sc, sc->NIL, lbody, is_null(cdr(lbody)), &bad_set))
		return(for_each_3);

	      if ((is_pair(car(lbody))) &&
		  (is_null(cdr(lbody))))
		return(for_each_2);

	      return(for_each_1); 
	    }
	}
      /* (define (hi) (for-each (lambda (a) a) (values '(1 2 3) '(3 2 1))))
       *   currently won't happen because the OP_C_L* cases are restricted here to S and opSq
       */
    }
  return(f);
}

/* extending this to map made little difference */
#endif


static bool next_map(s7_scheme *sc)
{
  s7_pointer y, z, vargs;
  int loc, zloc = -1;

  z = sc->NIL;
  vargs = counter_list(sc->args);
  loc = counter_count(sc->args);
  sc->x = sc->NIL;                     /* can't use preset args list here (as in for-each): (map list '(a b c)) */

  for (y = vargs; is_not_null(y); y = cdr(y))
    {
      s7_pointer x;
      switch (type(car(y)))
	{
	case T_PAIR:
	  x = caar(y);
	  car(y) = cdar(y);
	  break;
	  
	case T_C_OBJECT: 
	  {
	    int save_x = -1, save_y = -1, save_z = -1;
	    if (is_null(z))
	      {
		z = list_1(sc, make_integer(sc, loc));
		zloc = s7_gc_protect(sc, z);
	      }
	    SAVE_X_Y_Z(save_x, save_y, save_z);
	    x = (*(object_ref(car(y))))(sc, car(y), z);
	    RESTORE_X_Y_Z(save_x, save_y, save_z);
	  }
	  break;
	
	case T_VECTOR:
	  x = vector_element(car(y), loc); 
	  break;

	case T_STRING:
	  x = s7_make_character(sc, ((unsigned char)(string_value(car(y))[loc])));
	  break;

	case T_CLOSURE:   /* hash-table via an iterator */
	  x = g_hash_table_iterate(sc, hash_table_iterate_args(car(y)));
	  break;

	default: 
	  /* this can happen if one of the args is clobbered by the map function, so our initial
	   *   length is messed up:
	   *   (let ((L (list 1 2 3 4 5))) (map (lambda (x) (set-cdr! (cddr L) 5) x) L))
	   */

	  if (is_not_null(z))
	    s7_gc_unprotect_at(sc, zloc);
	  return(false);                  /* this stops the map process, so the code mentioned above returns '(1 2 3) */
	  break;
	}
      
      sc->x = cons(sc, x, sc->x);
    }

  if (is_not_null(z))
    s7_gc_unprotect_at(sc, zloc);
  sc->x = safe_reverse_in_place(sc, sc->x);

  counter_count(sc->args) = loc + 1;
  push_stack(sc, OP_MAP, sc->args, sc->code);
  sc->args = sc->x;
  sc->x = sc->NIL;

  if (is_macro(sc->code))
    {
      /* (let ((lst '(1 2 3))) 
       *   (define-macro (hiho a) `(+ 1 ,a)) 
       *   (map hiho lst)) 
       * -> '(2 3 4)
       */
      push_stack(sc, OP_EVAL_MACRO, sc->NIL, sc->args);
      car(sc->TEMP_CELL_1) = sc->code;
      cdr(sc->TEMP_CELL_1) = sc->args;
      sc->args = sc->TEMP_CELL;
    }

  return(true);
}


static s7_pointer g_map(s7_scheme *sc, s7_pointer args)
{
  #define H_map "(map proc object . objects) applies proc to a list made up of the next element of each of its arguments, returning \
a list of the results.  Its arguments can be lists, vectors, strings, hash-tables, or any applicable objects."

  s7_Int len;
  s7_pointer obj, x;
  sc->code = car(args);

  sc->y = args;                     /* gc protect */
  obj = cadr(args); 

  /* there are two simple cases here: a safe function can be handled immediately, and a closure
   *   with one arg called with one list can be handled without all the elaborate run-time checks.
   */

  if ((is_pair(obj)) &&                                               /* arg is a list */
      (is_null(cddr(args))))                                          /* only one list arg */
    {
      len = s7_list_length(sc, obj);
      if (len > 0)                                                    /* a proper list arg */
	{
	  if ((type(sc->code) == T_CLOSURE) &&                        /* not lambda* that might get confused about arg names */
	      (is_proper_list(sc, closure_args(sc->code))) &&         /* not a rest arg, not is_pair here: (lambda (x . args) args) */
	      (!is_immutable(car(closure_args(sc->code)))) &&         /* not a bad arg name! */
	      (!symbol_has_accessor(car(closure_args(sc->code)))) &&  /* not wrapped in an accessor */
	      (safe_list_length(sc, closure_args(sc->code)) == 1))    /* closure takes just one arg */
	    {
	      push_stack(sc, OP_MAP_SIMPLE, make_counter(sc, sc->NIL, obj, len, 0), sc->code);
	      return(sc->NO_VALUE);
	    }

	  if ((is_safe_procedure(sc->code)) &&
	      (is_c_function(sc->code)) &&
	      (args_match(sc, sc->code, 1)))
	    {
	      s7_pointer p;
	      s7_function func;

	      func = c_function_call(sc->code);
	      sc->x = sc->NIL;
	      for (p = obj; is_pair(p); p = cdr(p))
		{
		  car(sc->T1_1) = car(p);
		  sc->x = cons(sc, (*func)(sc, sc->T1_1), sc->x); /* can we assume a safe function won't return multiple values? */
		}
	      p = safe_reverse_in_place(sc, sc->x);
	      sc->x = sc->NIL;

	      return(p);
	    }
	}

      if (len < 0)
	len = -len; 
      else
	if (len == 0) 
	  len = S7_LONG_MAX;
    }
  else len = applicable_length(sc, obj);

  if (len < 0)
    return(s7_wrong_type_arg_error(sc, "map", 2, obj, "a vector, list, string, hash-table, or applicable object"));

  for (x = cddr(args); is_not_null(x); x = cdr(x))
    if (!is_sequence(sc, car(x)))
      return(s7_wrong_type_arg_error(sc, "map", position_of(x, args), car(x), "a sequence"));

  if (len != 0)
    {
      if (s7_is_hash_table(obj))
	sc->z = list_1(sc, g_make_hash_table_iterator(sc, cdr(args)));
      else 
	{
	  if (is_environment(obj))
	    sc->z = list_1(sc, s7_environment_to_list(sc, obj));
	  else sc->z = list_1(sc, obj);
	}

      /* we have to copy the args if any of them is a list:
       * (let* ((x (list (list 1 2 3))) (y (apply map abs x))) (list x y))
       */
      
      if (is_not_null(cddr(args)))
	{
	  for (x = cddr(args); is_not_null(x); x = cdr(x))
	    {
	      s7_Int nlen;
	      
	      nlen = applicable_length(sc, car(x));
	      if (nlen < 0)
		return(s7_wrong_type_arg_error(sc, "map", position_of(x, args), car(x), "a vector, list, string, hash-table, or applicable object"));
	      if (nlen < len) len = nlen;
	      if (len == 0) break; /* need error check below */
	      
	      if (s7_is_hash_table(car(x)))
		sc->z = cons(sc, g_make_hash_table_iterator(sc, x), sc->z);
	      else
		{
		  if (is_environment(car(x)))
		    sc->z = cons_unchecked(sc, s7_environment_to_list(sc, car(x)), sc->z);
		  else sc->z = cons_unchecked(sc, car(x), sc->z);
		}
	    }
	}
    }

  if (len == 0)   /* (map 1 "hi" '()) */
    {
      if (((typeflag(sc->code) & (T_ANY_MACRO | T_PROCEDURE)) != 0) ||
	  (is_pair(sc->code)) ||
	  (s7_is_string(sc->code)) ||
	  (s7_is_vector(sc->code)) ||
	  (s7_is_hash_table(sc->code)) ||
	  (is_environment(sc->code)) ||
	  (is_hook(sc->code)) ||
	  (is_syntax(sc->code)))
	return(sc->NIL);    /* obj has no elements (the circular list case will return S7_LONG_MAX here) */
      return(s7_wrong_type_arg_error(sc, "map", 1, sc->code, "a procedure or something applicable"));
    }

  if (len == S7_LONG_MAX)
    {
      /* all args are circular lists, or perhaps an odd scheme type (see comment under for-each) */
      for (x = cdr(args); (is_pair(x)) && (is_pair(car(x))); x = cdr(x)) {}
      if (!is_pair(x))
	return(s7_error(sc, sc->WRONG_TYPE_ARG, 
			list_2(sc, make_protected_string(sc, "map's arguments are circular lists! ~A"), cdr(args))));
    }

  sc->args = make_counter(sc, sc->NIL, safe_reverse_in_place(sc, sc->z), 0, len);
  sc->y = sc->NIL;
  sc->z = sc->NIL;
  if (next_map(sc))
    push_stack(sc, OP_APPLY, sc->args, sc->code);

  return(sc->NIL);
}



/* -------------------------------- multiple-values -------------------------------- */

static s7_pointer splice_in_values(s7_scheme *sc, s7_pointer args)
{
  if (sc->stack_end > sc->stack_start)
    {
      /* code = args yet to eval in order, args = evalled args reversed */
      int top;
      s7_pointer x;
      top = s7_stack_top(sc) - 1;

      switch (stack_op(sc->stack, top))
	{
	  /* the normal case -- splice values into caller's args */
	case OP_EVAL_ARGS1:
	case OP_EVAL_ARGS2:
	case OP_EVAL_ARGS3:
	case OP_EVAL_ARGS4:
#if WITH_OPTIMIZATION
	case OP_EVAL_ARGS_P_1:
	case OP_EVAL_ARGS_P_2:
	case OP_EVAL_ARGS_P_3:
	case OP_EVAL_ARGS_P_4:
	case OP_SAFE_C_opSq_P_1:
	case OP_SAFE_C_PP_1:
	case OP_SAFE_C_PP_2:
	case OP_C_P_1:
#endif
	  /* it's not safe to simply reverse args and tack the current stacked args onto its (new) end,
	   *   setting stacked args to cdr of reversed-args and returning car because the list (args)
	   *   can be some variable's value in a macro expansion via ,@ and reversing it in place
	   *   (all this to avoid consing), clobbers the variable's value.
	   */
	  for (x = args; is_not_null(cdr(x)); x = cdr(x))
	    stack_args(sc->stack, top) = cons(sc, car(x), stack_args(sc->stack, top));
	  return(car(x));

	case OP_EVAL_ARGS5:
	  /* code = previous arg saved, args = ante-previous args reversed
	   *   we'll take value->code->args and reverse in args5
	   *   if one value, return it, else
	   *      put code onto args, splice as above until there are 2 left
	   *      set code to 1st and value to last
	   */
	  if (is_null(args))
	    return(sc->UNSPECIFIED);

	  if (is_null(cdr(args)))
	    return(car(args));

	  stack_args(sc->stack, top) = cons(sc, stack_code(sc->stack, top), stack_args(sc->stack, top));
	  for (x = args; is_not_null(cddr(x)); x = cdr(x))
	    stack_args(sc->stack, top) = cons(sc, car(x), stack_args(sc->stack, top));
	  stack_code(sc->stack, top) = car(x);
	  return(cadr(x));

	  /* look for errors here rather than glomming up the set! and let code */
	case OP_SET1:                                             /* (set! var (values 1 2 3)) */
	  set_multiple_value(args);
	  return(eval_error(sc, "can't set! some variable to ~A", args));

	case OP_LET1:                                             /* (let ((var (values 1 2 3))) ...) */
	case OP_LET_STAR1:
	case OP_LETREC1:
	  set_multiple_value(args);
	  return(eval_error_with_name(sc, "~A: can't bind some variable to ~A", args));

	  /* handle 'and' and 'or' specially */
	case OP_AND1:
	  for (x = args; is_not_null(cdr(x)); x = cdr(x))
	    if (car(x) == sc->F)
	      return(sc->F);
	  return(car(x));

	case OP_OR1:
	  for (x = args; is_not_null(cdr(x)); x = cdr(x))
	    if (car(x) != sc->F)
	      return(car(x));
	  return(car(x));

	case OP_BARRIER: 
	  pop_stack(sc);
	  return(splice_in_values(sc, args));

	default:
	  break;
	}
    }

  /* let it meander back up the call chain until someone knows where to splice it */
  set_multiple_value(args);
  return(args);
}


static s7_pointer g_values(s7_scheme *sc, s7_pointer args)
{
  #define H_values "(values obj ...) splices its arguments into whatever list holds it (its 'continuation')"

  if (is_null(args))
    {
      if (stack_op(sc->stack, s7_stack_top(sc) - 1) == OP_SET1)  /* (set! var (values)) */
	return(eval_error(sc, "set!: can't assign (values) to something", args));

      /* fprintf(stderr, "values nil: %s %s %s\n", real_op_names[stack_op(sc->stack, s7_stack_top(sc) - 1)], DISPLAY_80(sc->code), DISPLAY_80(sc->args)); */
      return(sc->NO_VALUE); 
    }

  /* this was sc->NIL until 16-Jun-10, 
   *   nil is consistent with the implied values call in call/cc (if no args, the continuation function returns '())
   *   hmmm... 
   *   Guile complains ("too few values returned to continuation") in the call/cc case, and
   *   (equal? (if #f #f) (* (values))) complains "Zero values returned to single-valued continuation"
   *   so perhaps call/cc should also return #<unspecified> -- I don't know what is best.
   *
   * a note in the scheme bboard:
   *
   *  This would work in s7:
   *
   *  (define (print-concat . args) 
   *    (if (or (null? args)               ; (print-concat)
   *            (eq? (car args) (values))) ; (print-concat arg1 ...)
   *        (newline) 
   *      (begin 
   *        (display (car args)) 
   *        (print-concat (apply values (cdr args))))))
   *  
   *  but it's a bit ugly.  I think (values) should be the same as
   *  (apply values '()). It's currently #<unspecified>, mainly for
   *  historical reasons (a lot of the code s7 is used with
   *  assumes that behavior).  If (values) simply vanished,
   *  then code like (abs -1 (values)) is not an error.
   *  If it returned nil, you get inconsistencies elsewhere:
   *  (values) now is the same as (values '()) which is
   *  (apply values '(())).  This also affects quasiquote 
   *  (s7 uses (apply values ...) for unquote-splicing).
   *
   * But that makes me think that (eval) should return #<unspecified>. What about (eval-string)?
   */
  
  if (is_null(cdr(args)))
    return(car(args));

  return(splice_in_values(sc, args));
}


s7_pointer s7_values(s7_scheme *sc, int num_values, ...)
{
  int i;
  va_list ap;
  s7_pointer p;
  
  if (num_values == 0)
    return(sc->NIL);

  sc->w = sc->NIL;
  va_start(ap, num_values);
  for (i = 0; i < num_values; i++)
    sc->w = cons(sc, va_arg(ap, s7_pointer), sc->w);
  va_end(ap);

  p = sc->w;
  sc->w = sc->NIL;

  return(g_values(sc, safe_reverse_in_place(sc, p)));
}



/* -------------------------------- quasiquote -------------------------------- */

static s7_pointer g_qq_list(s7_scheme *sc, s7_pointer args)
{
  #define H_qq_list "({list} ...) returns its arguments in a list (internal to quasiquote)"

  s7_pointer x, y, px;

  for (x = args; is_pair(x); x = cdr(x))
    if (car(x) == sc->NO_VALUE) 
      break;
  
  if (is_null(x))
    return(args);

  /* this is not maximally efficient, but it's not important:
   *   we've hit the rare special case where ({apply} {values} '())) needs to be ignored
   *   in the splicing process (i.e. the arglist acts as if the thing never happened)
   */
  px = sc->NIL;
  for (x = args, y = args; is_pair(y); y = cdr(y))
    if (car(y) != sc->NO_VALUE)
      {
	car(x) = car(y);
	px = x;
	x = cdr(x);
      }

  if ((is_not_null(y)) &&
      (y != sc->NO_VALUE))
    cdr(x) = cdr(y);
  else 
    {
      sc->no_values--;
      if (is_null(px))
	return(sc->NIL);
      cdr(px) = sc->NIL;
    }

  return(args);
}


static s7_pointer g_qq_values(s7_scheme *sc, s7_pointer args)
{
  #define H_qq_values "(apply {values} arg) is the quasiquote internal form for \",@arg\""

  /* for quasiquote handling: (apply values car(args)) if args not nil, else nil
   *    (values) -> #<unspecified> which is not wanted in this context.
   */
  if (is_null(args))
    {
      sc->no_values++;
      return(sc->NO_VALUE);
    }
  return(g_values(sc, args));
}


static s7_pointer g_apply_values(s7_scheme *sc, s7_pointer args)
{
  #define H_apply_values "({apply-values} var) applies values to var.  This is an internal function."
  s7_pointer x;

  if (is_null(args))
    {
      sc->no_values++;
      return(sc->NO_VALUE);
    }
  if (is_null(cdr(args)))
    x = car(args);
  else x = apply_list_star(sc, args);

  if (!is_proper_list(sc, x))
    return(s7_error(sc, sc->WRONG_TYPE_ARG, 
		    list_2(sc, 
			   make_protected_string(sc, "apply's last argument should be a proper list: ~A"),
			   args)));

  return(g_qq_values(sc, x));
}

/* (apply values ...) replaces (unquote_splicing ...)
 *
 * (define-macro (hi a) `(+ 1 ,a) == (list '+ 1 a)
 * (define-macro (hi a) ``(+ 1 ,,a) == (list list '+ 1 (list quote a)))
 *
 * (define-macro (hi a) `(+ 1 ,@a) == (list '+ 1 (apply values a))
 * (define-macro (hi a) ``(+ 1 ,,@a) == (list list '+ 1 (apply values a))
 *
 * this is not the same as CL's quasiquote; for example:
 *   [1]> (let ((a 1) (b 2)) `(,a ,@b))
 *   (1 . 2)
 *   in s7 this is an error.  
 *
 * also in CL the target of ,@ can apparently be a circular list
 */

static bool is_simple_code(s7_scheme *sc, s7_pointer form)
{
  s7_pointer tmp;
  for (tmp = form; is_pair(tmp); tmp = cdr(tmp))
    if (is_pair(car(tmp)))
      {
	if (!is_simple_code(sc, car(tmp)))
	  return(false);
      }
    else
      {
	if ((car(tmp) == sc->UNQUOTE) ||
	    ((is_null(car(tmp))) && (is_null(cdr(tmp)))))
	  return(false);
      }
  return(is_null(tmp));
}


static s7_pointer g_quasiquote_1(s7_scheme *sc, s7_pointer form)
{
  #define H_quasiquote "(quasiquote arg) is the same as `arg.  If arg is a list, it can contain \
comma (\"unquote\") and comma-atsign (\"unquote-splicing\") to pre-evaluate portions of the list. \
unquoted expressions are evaluated and plugged into the list, unquote-splicing evaluates the expression \
and splices the resultant list into the outer list. `(1 ,(+ 1 1) ,@(list 3 4)) -> (1 2 3 4)."

  if (!is_pair(form))
    {
      if (!is_symbol(form))
	{
	  /* things that evaluate to themselves don't need to be quoted. 
	   *    but this means `() -> () whereas below `(1) -> '(1) -- should nil here return '()?
	   *    (this also affects vector constants since they call g_quasiquote at run time in OP_READ_QUASIQUOTE_VECTOR)
	   */
	  return(form);
	}
      return(list_2(sc, sc->QUOTE, form));
    }

  if (car(form) == sc->UNQUOTE)
    {
      if (is_not_null(cddr(form)))
	return(eval_error(sc, "unquote: too many arguments, ~S", form));
      return(cadr(form));
    }

  /* it's a list, so return the list with each element handled as above.
   *    we try to support dotted lists which makes the code much messier.
   */
  
  /* if no element of the list is a list or unquote, just return the original quoted */
  if (is_simple_code(sc, form))
    return(list_2(sc, sc->QUOTE, form));

  {
    int len, i, loc;
    s7_pointer orig, bq, old_scw;
    bool dotted = false;

    len = s7_list_length(sc, form);
    if (len < 0)
      {
	len = -len;
	dotted = true;
      }

    old_scw = sc->w;
    loc = s7_gc_protect(sc, old_scw);

    sc->w = sc->NIL;
    for (i = 0; i <= len; i++)
      sc->w = cons(sc, sc->NIL, sc->w);

    car(sc->w) = sc->QQ_LIST;
    
    if (!dotted)
      {
	for (orig = form, bq = cdr(sc->w), i = 0; i < len; i++, orig = cdr(orig), bq = cdr(bq))
	  {
	    if ((is_pair(cdr(orig))) &&   /* this was is_pair(orig) which seems to be always the case */
		(cadr(orig) == sc->UNQUOTE))
	      {
		/* `(1 . ,(+ 1 1)) -> '(1 unquote (+ 1 1)) -> '(1 . 2) 
		 * `(1 . ,@'((2 3))) -> (1 unquote ({apply} {values} '((2 3)))) -> ({append} ({list} 1) ({apply} {values} '((2 3)))) -> '(1 2 3)
		 * this used to be `(1 . ,@('(2 3))).  
		 *     This now becomes (1 unquote ({apply} {values} ('(2 3)))) -> ({append} ({list} 1) ({apply} {values} ('(2 3)))) -> error
		 * `(1 . (,@'(2 3))) works in both cases, and `(1 . (,(+ 1 1)))
		 * so do we actually need this block? `(1 ,@'(2 3)) if undotted
		 */
		car(bq) = g_quasiquote_1(sc, car(orig));
		cdr(bq) = sc->NIL;
		sc->w = list_3(sc, sc->QQ_APPEND, sc->w, caddr(orig));
		break;
	      }
	    else car(bq) = g_quasiquote_1(sc, car(orig));
	  }
      }
    else
      {
	/* `(1 2 . 3) */
	len--;
	for (orig = form, bq = cdr(sc->w), i = 0; i < len; i++, orig = cdr(orig), bq = cdr(bq))
	  car(bq) = g_quasiquote_1(sc, car(orig));
	car(bq) = g_quasiquote_1(sc, car(orig));

	sc->w = list_3(sc, sc->QQ_APPEND, sc->w, g_quasiquote_1(sc, cdr(orig)));
	/* quasiquote might quote a symbol in cdr(orig), so it's not completely pointless */
      }

    bq = sc->w;
    sc->w = old_scw;
    s7_gc_unprotect_at(sc, loc);
    return(bq);
  }
}


static s7_pointer g_quasiquote(s7_scheme *sc, s7_pointer args)
{
  /* this is for explicit quasiquote support, not the backquote stuff in macros */
  return(g_quasiquote_1(sc, car(args)));
}



/* ---------------- reader funcs for eval ---------------- */

static void back_up_stack(s7_scheme *sc)
{
  opcode_t top_op;
  top_op = stack_op(sc->stack, s7_stack_top(sc) - 1);
  if (top_op == OP_READ_DOT)
    {
      pop_stack(sc);
      top_op = stack_op(sc->stack, s7_stack_top(sc) - 1);
    }
  if (top_op == OP_READ_VECTOR)
    {
      pop_stack(sc);
      top_op = stack_op(sc->stack, s7_stack_top(sc) - 1);
    }
  if (top_op == OP_READ_QUOTE)
    pop_stack(sc);
}


static token_t read_sharp(s7_scheme *sc, s7_pointer pt)
{
  int c;
  /* inchar can return EOF, so it can't be used directly as an index into the digits array */
  c = inchar(pt);
  if (c == EOF)
    s7_error(sc, sc->READ_ERROR,
	     list_1(sc, make_protected_string(sc, "unexpected '#' at end of input")));

  sc->w = small_int(1);
  if (c == '(') 
    return(TOKEN_VECTOR);

  if (isdigit(c)) /* #2D(...) */
    {
      int dims, dig, d, loc = 0;
      sc->strbuf[loc++] = c;
      dims = digits[c];

      while (true)
	{
	  d = inchar(pt);
	  if (d == EOF)
	    s7_error(sc, sc->READ_ERROR,
		     list_1(sc, make_protected_string(sc, "unexpected end of input while reading #n...")));

	  dig = digits[d];
	  if (dig >= 10) break;
	  dims = dig + (dims * 10);
	  sc->strbuf[loc++] = d;
	}
      sc->strbuf[loc++] = d;
      if ((d == 'D') || (d == 'd'))
	{
	  d = inchar(pt);
	  if (d == EOF)
	    s7_error(sc, sc->READ_ERROR,
		     list_1(sc, make_protected_string(sc, "unexpected end of input while reading #nD...")));
	  sc->strbuf[loc++] = d;
	  if (d == '(')
	    {
	      sc->w = make_integer(sc, dims);
	      return(TOKEN_VECTOR);
	    }
	}

      /* try to back out */
      for (d = loc - 1; d > 0; d--)
	backchar(sc->strbuf[d], pt);
    }

#if (!S7_DISABLE_DEPRECATED)
  if (c == ':')  /* turn #: into : -- this is for compatiblity with Guile, #:optional in particular */
    {
      sc->strbuf[0] = ':';
      return(TOKEN_ATOM);
    }

  /* block comments in either #! ... !# */
  if (c == '!') 
    {
      char last_char;
      last_char = ' ';
      while ((c = inchar(pt)) != EOF)
	{
	  if ((c == '#') &&
	      (last_char == '!'))
	    break;
	  last_char = c;
	}
      if (c == EOF)
	s7_error(sc, sc->READ_ERROR,
		 list_1(sc, make_protected_string(sc, "unexpected end of input while reading #!")));
      return(token(sc));
    }
#endif
      
  /* block comments in #| ... |# 
   *   since we ignore everything until the |#, internal semicolon comments are ignored,
   *   meaning that ;|# is as effective as |#
   */
  if (c == '|') 
    {
      char last_char;
      last_char = ' ';
      while ((c = inchar(pt)) != EOF)
	{
	  if ((c == '#') &&
	      (last_char == '|'))
	    break;
#if 0
	  /* can't decide about this... */
	  if ((c == '|') &&
	      (last_char == '#'))
	    fprintf(stderr, "#| encountered inside #| block\n");
#endif
	  last_char = c;
	}
      if (c == EOF)
	s7_error(sc, sc->READ_ERROR,
		 list_1(sc, make_protected_string(sc, "unexpected end of input while reading #|")));
      return(token(sc));
    }
      
  sc->strbuf[0] = c; 
  return(TOKEN_SHARP_CONST); /* next stage notices any errors */
}    



static token_t read_comma(s7_scheme *sc, s7_pointer pt)
{
  int c;
  /* here we probably should check for symbol names that start with "@":
     
     :(defmacro hi (@foo) `(+ ,@foo 1))
     hi
     :(hi 2)
     ;foo: unbound variable
	 
     but

     :(defmacro hi (.foo) `(+ ,.foo 1))
     hi
     :(hi 2)
     3

     and ambiguous:
     :(define-macro (hi @foo . foo) `(list ,@foo))

     what about , @foo -- is the space significant?  We accept ,@ foo.
  */

  if ((c = inchar(pt)) == '@') 
    return(TOKEN_AT_MARK);

  if (c == EOF)
    {
      sc->strbuf[0] = ',';  /* was '@' which doesn't make any sense */
      return(TOKEN_COMMA);  /* was TOKEN_ATOM, which also doesn't seem sensible */
    }
  backchar(c, pt);
  return(TOKEN_COMMA);
}


static token_t read_dot(s7_scheme *sc, s7_pointer pt)
{
  int c;
  c = inchar(pt);
  if (c != EOF)
    {
      backchar(c, pt);

      if ((!char_ok_in_a_name[c]) && (c != 0))
	return(TOKEN_DOT);
    }
  else
    {
      sc->strbuf[0] = '.'; 
      return(TOKEN_DOT);
    }
  sc->strbuf[0] = '.'; 
  return(TOKEN_ATOM);  /* i.e. something that can start with a dot like a number */
}


static token_t token(s7_scheme *sc)
{
  int c;
  s7_pointer pt;

  pt = sc->input_port;
  c = port_read_white_space(pt)(sc, pt);

  switch (c) 
    {
    case '(':
      return(TOKEN_LEFT_PAREN);
      
    case ')':
      return(TOKEN_RIGHT_PAREN);
      
    case '.':
      return(read_dot(sc, pt));

    case '\'':
      return(TOKEN_QUOTE);
      
    case ';':
      return(port_read_semicolon(pt)(sc, pt));

    case '"':
      return(TOKEN_DOUBLE_QUOTE);
      
    case '`':
      return(TOKEN_BACK_QUOTE);
      
    case ',':
      return(read_comma(sc, pt));
      
    case '#':
      return(read_sharp(sc, pt));

    case EOF:
      return(TOKEN_EOF);

    default: 
      sc->strbuf[0] = c; /* every TOKEN_ATOM return goes to port_read_name, so we save a backchar/inchar shuffle by starting the read here */
      return(TOKEN_ATOM);
    }
}



#define NOT_AN_X_CHAR -1

static int read_x_char(s7_pointer pt)
{
  /* possible "\xnn" char (write creates these things, so we have to read them) 
   *   but we could have crazy input like "\x -- with no trailing double quote
   */
  int d1, d2, c;

  c = inchar(pt);
  if (c == EOF)
    return(NOT_AN_X_CHAR);

  d1 = digits[c];
  if (d1 < 16)
    {
      c = inchar(pt);
      if (c == EOF)
	return(NOT_AN_X_CHAR);
      d2 = digits[c];
      if (d2 < 16)
	return(16 * d1 + d2);           /* following char can be anything, including a number -- we ignore it */
      /* apparently one digit is also ok */
      backchar(c, pt);
      return(d1);
    }
  return(NOT_AN_X_CHAR);
}


static s7_pointer read_string_constant(s7_scheme *sc, s7_pointer pt)
{
  /* sc->F => error 
   *   no check needed here for bad input port and so on
   */
  unsigned int i = 0;
  int c;

  if (is_string_port(pt))
    {
      /* try the most common case first */
      char *s, *start, *end;
      start = (char *)(port_string(pt) + port_string_point(pt));
      end = (char *)(port_string(pt) + port_string_length(pt));
      for (s = start; s < end; s++)
	{
	  if (*s == '"')                         /* switch here no faster */
	    {
	      s7_pointer result;
	      int len;
	      len = s - start;
	      result = s7_make_terminated_string_with_length(sc, start, len);
	      port_string_point(pt) += (len + 1);
	      return(result);
	    }
	  else
	    {
	      if (*s == '\\')
		{
		  if ((unsigned int)(s - start) >= sc->strbuf_size)
		    resize_strbuf(sc);
		  for (i = 0; i < (unsigned int)(s - start); i++)
		    sc->strbuf[i] = port_string(pt)[port_string_point(pt)++];
		  break;
		}
	      else
		{
		  if (*s == '\n')
		    port_line_number(pt)++; 
		}
	    }
	}
    }

  while (true)
    {
      /* splitting this check out and duplicating the loop was slower?!? */
      c = port_read_character(pt)(sc, pt);

      switch (c)
	{
	case '\n': 
	  port_line_number(pt)++; 
	  sc->strbuf[i++] = c;
	  break;

	case EOF:
	  return(sc->F);

	case '"':
	  return(s7_make_terminated_string_with_length(sc, sc->strbuf, i));

	case '\\':
	  c = inchar(pt);

	  if (c == EOF) 
	    return(sc->F);

	  if (c == '\\')
	    sc->strbuf[i++] = '\\';
	  else
	    {
	      if (c == '"')
		sc->strbuf[i++] = '"';
	      else
		{
		  if (c == 'n')
		    sc->strbuf[i++] = '\n';
		  else 
		    {
		      if (c == 't') /* this is for compatibility with other Schemes */
			sc->strbuf[i++] = '\t';
		      else 
			{
			  if (c == 'x')
			    {
			      c = read_x_char(pt);
			      if (c == NOT_AN_X_CHAR)
				return(sc->T);
			      sc->strbuf[i++] = (unsigned char)c;
			    }
			  else
			    {
			      if (!is_white_space(c))
				return(sc->T); 

			      /* #f here would give confusing error message "end of input", so return #t=bad backslash.
			       *     this is not optimal. It's easy to forget that backslash needs to be backslashed. 
			       *
			       * the white_space business implements Scheme's \<newline> or \<space> feature -- the character after \ is flushed.
			       *   It may be that r6rs expects all white space after \ to be flushed, but in that case
			       *   (string->number "1\   2") is 12??  Too bizarre.
			       */
			    }
			}
		    }
		}
	    }
	  break;

	default:
	  sc->strbuf[i++] = c;
	  break;
	}

      if (i >= sc->strbuf_size)
	resize_strbuf(sc);
    }
}


/* static const char *tokens[12] = {"eof", "left_paren", "right_paren", "dot", "atom", "quote", 
                                    "double_quote", "back_quote", "comma", "at_mark", "sharp_const", "vector"}; 
*/

static s7_pointer read_expression(s7_scheme *sc)
{ 
  while (true) 
    {
      int c;
      switch (sc->tok) 
	{
	case TOKEN_EOF:
	  return(sc->EOF_OBJECT);
	  
	case TOKEN_VECTOR:  /* already read #( -- TOKEN_VECTOR is triggered by #( */
	    push_stack(sc, OP_READ_VECTOR, sc->w, sc->NIL);   /* sc->w is the dimensions */
	    /* fall through */
	  
	case TOKEN_LEFT_PAREN:
	  sc->tok = token(sc);

	  if (sc->tok == TOKEN_RIGHT_PAREN)
	    return(sc->NIL);

	  if (sc->tok == TOKEN_DOT) 
	    {
	      back_up_stack(sc);
	      do {c = inchar(sc->input_port);} while ((c != ')') && (c != EOF));
	      return(read_error(sc, "stray dot after '('?"));         /* (car '( . )) */
	    }

	  if (sc->tok == TOKEN_EOF)
	    return(missing_close_paren_error(sc));

	  push_stack(sc, OP_READ_LIST, sc->NIL, sc->NIL);
	  /* here we need to clear args, but code is ignored */

	  if (sc->stack_end >= sc->stack_resize_trigger)
	    increase_stack_size(sc);
	  break;
	  
	case TOKEN_QUOTE:
	  push_stack(sc, OP_READ_QUOTE, sc->NIL, sc->NIL);
	  sc->tok = token(sc);
	  break;
	  
	case TOKEN_BACK_QUOTE:
	  sc->tok = token(sc);
	  if (sc->tok == TOKEN_VECTOR) 
	    {
	      push_stack(sc, OP_READ_QUASIQUOTE_VECTOR, sc->w, sc->NIL);
	      sc->tok = TOKEN_LEFT_PAREN; 
	    } 
	  else push_stack(sc, OP_READ_QUASIQUOTE, sc->NIL, sc->NIL);
	  break;
	  
	case TOKEN_COMMA:
	  push_stack(sc, OP_READ_UNQUOTE, sc->NIL, sc->NIL);
	  sc->tok = token(sc);
	  break;
	  
	case TOKEN_AT_MARK:
	  push_stack(sc, OP_READ_APPLY_VALUES, sc->NIL, sc->NIL);
	  sc->tok = token(sc);
	  break;
	  
	case TOKEN_ATOM:
	  return(port_read_name(sc->input_port)(sc, sc->input_port, NO_SHARP));
	  /* If reading list (from lparen), this will finally get us to op_read_list */
	  
	case TOKEN_DOUBLE_QUOTE:
	  sc->value = read_string_constant(sc, sc->input_port);

	  if (sc->value == sc->F)                                   /* can happen if input code ends in the middle of a string */
	    return(read_error(sc, "end of input encountered while in a string"));
	  if (sc->value == sc->T)
	    return(read_error(sc, "unknown backslash usage -- perhaps you meant two backslashes?"));

	  return(sc->value);
	  
	case TOKEN_SHARP_CONST:
	  sc->value = port_read_name(sc->input_port)(sc, sc->input_port, WITH_SHARP);

	  /* here we need the following character and form 
	   *   strbuf[0] == '#', false above = # case, not an atom
	   */
	  if (is_null(sc->value))
	    {
	      return(read_error(sc, "undefined # expression"));
	      /* a read error here seems draconian -- this unknown constant doesn't otherwise get in our way
	       *   but how to alert the caller to the problem without stopping the read?
	       */
	    }
	  return(sc->value);

	case TOKEN_DOT:                                             /* (catch #t (lambda () (+ 1 . . )) (lambda args 'hiho)) */
	  back_up_stack(sc);
	  do {c = inchar(sc->input_port);} while ((c != ')') && (c != EOF));
	  return(read_error(sc, "stray dot in list?"));             /* (+ 1 . . ) */

	case TOKEN_RIGHT_PAREN:                                     /* (catch #t (lambda () '(1 2 . )) (lambda args 'hiho)) */
	  back_up_stack(sc);
	  return(read_error(sc, "unexpected close paren"));         /* (+ 1 2)) or (+ 1 . ) */
	}
    }

  /* we never get here */
  return(sc->NIL);
}



/* ---------------- */

static s7_pointer unbound_variable(s7_scheme *sc, s7_pointer sym)
{
  /* handle *unbound-variable-hook* */

  /* this always occurs in a context where we're trying anything, so I'll move a couple of those tests here */
  if (is_keyword(sym))
    return(sym);

  if (sym == sc->UNQUOTE)
    return(eval_error_no_arg(sc, "unquote (',') occurred outside quasiquote"));
  
  if (safe_strcmp(symbol_name(sym), "|#") == 0)
    return(read_error(sc, "unmatched |#"));

  if (is_not_null(hook_functions(sc->unbound_variable_hook)))
    {
      int save_x = -1, save_y = -1, save_z = -1, cur_code_loc = -1;
      s7_pointer x, cur_code;

      cur_code = sc->cur_code;
      if (!is_pair(cur_code))
	{
	  /* isolated typo perhaps -- no pair to hold the position info, so make one.
	   *   sc->cur_code is GC-protected, so this should be safe.
	   */
	  cur_code = cons(sc, sym, sc->NIL);     /* the error will say "(sym)" which is not too misleading */
	  pair_line_number(cur_code) = port_line_number(sc->input_port) | (port_file_number(sc->input_port) << 20);
	}
      cur_code_loc = s7_gc_protect(sc, cur_code);   /* we need to save this because it has the file/line number of the unbound symbol */

      SAVE_X_Y_Z(save_x, save_y, save_z);

      /* not s7_hook_apply here because we need the value that the hook function returns
       *   should we call the entire list?  or just call trailing funcs if x is #<unspecified>?
       */

      x = s7_call(sc, 
		  car(hook_functions(sc->unbound_variable_hook)),
		  list_1(sc, sym));

      RESTORE_X_Y_Z(save_x, save_y, save_z);

      sc->cur_code = cur_code;
      s7_gc_unprotect_at(sc, cur_code_loc);

      if (x != sc->UNDEFINED)
	return(x);
    }
  
  return(eval_error(sc, "~A: unbound variable", sym));
}


static s7_pointer eval_symbol(s7_scheme *sc, s7_pointer sym)
{
  s7_pointer x;

  x = find_symbol(sc, sym);
  if (is_slot(x)) 
    return(slot_value(x));

  return(unbound_variable(sc, sym));
}


static s7_pointer assign_syntax(s7_scheme *sc, const char *name, opcode_t op) 
{
  s7_pointer x, syn;
  unsigned int loc = 0;

  x = new_symbol(sc, name, symbol_table_hash(name, &loc)); 
  symbol_hash(x) = loc;

  syn = (s7_cell *)permanent_calloc(sizeof(s7_cell));
  set_type(syn, T_SYNTAX | T_SYNTACTIC | T_DONT_EVAL_ARGS); 
  syntax_opcode(syn) = op;
  cdr(syn) = syn; /* this saves us an error check in the main eval section */
  car(syn) = x;
  global_slot(x) = permanent_slot(x, syn);

  typeflag(x) = SYNTACTIC_TYPE;
  symbol_id(x) = 0;
  syntax_opcode(x) = op;
  return(x);
}


static s7_pointer assign_internal_syntax(s7_scheme *sc, const char *name, opcode_t op) 
{
  s7_pointer x, str, syn; 
  unsigned int loc = 0;

  str = s7_make_permanent_string(name);

  x = (s7_cell *)permanent_calloc(sizeof(s7_extended_cell));
  x->hloc = NOT_IN_HEAP;
  symbol_name_cell(x) = str;
  set_type(x, T_SYMBOL);

  global_slot(x) = sc->NIL;
  local_slot(x) = sc->NIL;
  symbol_table_hash(name, &loc); 
  symbol_hash(x) = loc;
  symbol_id(x) = 0;
  symbol_accessor(x) = -1;

  syn = (s7_cell *)permanent_calloc(sizeof(s7_cell));
  set_type(syn, T_SYNTAX | T_SYNTACTIC | T_DONT_EVAL_ARGS); 
  syntax_opcode(syn) = op;
  cdr(syn) = syn; 
  car(syn) = s7_make_symbol(sc, name);
  global_slot(x) = permanent_slot(x, syn);

  typeflag(x) = SYNTACTIC_TYPE;
  symbol_id(x) = 0;
  syntax_opcode(x) = op;
  return(x);
}


static bool direct_memq(s7_pointer symbol, s7_pointer list)
{
  s7_pointer x;
  for (x = list; is_pair(x); x = cdr(x))
    if (car(x) == symbol)
      return(true);
  return(false);
}


#if 0
static s7_pointer remq(s7_scheme *sc, s7_pointer a, s7_pointer obj) 
{
  s7_pointer p;

  sc->w = sc->NIL;
  for ( ; is_pair(a); a = cdr(a))
    if (car(a) != obj)
      sc->w = cons(sc, car(a), sc->w);
  p = sc->w;
  sc->w = sc->NIL;

  return(s7_reverse(sc, p));
}
#endif



static s7_pointer g_pair_line_number(s7_scheme *sc, s7_pointer args)
{
  #define H_pair_line_number "(pair-line-number pair) returns the line number at which it read 'pair'"

  if (!is_pair(car(args)))
    return(s7_wrong_type_arg_error(sc, "pair-line-number", 0, car(args), "a pair"));	
  return(make_integer(sc, (s7_Int)(remembered_line_number(pair_line_number(car(args))))));
}


static s7_pointer quotify(s7_scheme *sc, s7_pointer pars)
{
  /* the default parameter values of define-macro* and define-bacro* should not be evaluated until
   * the expansion is evaluated.  That is, 
   *
   *   (let ((x 0)) (define-macro* (hi (a (let () (set! x (+ x 1)) x))) `(let ((x -1)) (+ x ,a))) (list (hi) x))
   *
   * should return the same value as the equivalent explicit form:
   *
   *   (let ((x 0)) (define-macro (hi a) `(let ((x -1)) (+ x ,a))) (list (hi (let () (set! x (+ x 1)) x)) x))
   *
   * '(-1 0) in both cases.
   *
   * But at the point in eval where we handle lambda* arguments, we can't easily tell whether we're part of
   * a function or a macro, so at definition time of a macro* we scan the parameter list for an expression
   * as a default value, and replace it with (quote expr).
   *
   * and... (define-macro* ((a x)) ...) should behave the same as (define-macro* ((a (+ x 0))) ...)
   */
  s7_pointer tmp;
  for (tmp = pars; is_pair(tmp); tmp = cdr(tmp))
    if ((is_pair(car(tmp))) &&
	(is_pair(cdar(tmp))) &&
	((is_pair(cadar(tmp))) ||
	 (is_symbol(cadar(tmp))) ||
	 (is_syntax(cadar(tmp)))))
      cadar(tmp) = list_2(sc, sc->QUOTE, cadar(tmp));
  return(pars);
}


static s7_pointer lambda_star_argument_set_value(s7_scheme *sc, s7_pointer sym, s7_pointer val)
{
  s7_pointer x;

  for (x = environment_slots(sc->envir) /* presumably the arglist */; is_slot(x); x = next_slot(x))
    if (slot_symbol(x) == sym)   /* slot_symbol(x) won't be sc->NIL here even if no args and no locals because we at least have __func__ */
      {
	/* x is our binding (symbol . value) */
	if (is_not_checked(x))
	  set_checked(x); /* this is a special use of this bit, I think */
	else 
	  {
	    return(s7_error(sc, sc->WRONG_TYPE_ARG,
			    list_4(sc,
				   make_protected_string(sc, "~A: parameter set twice, ~A in ~A"),
				   closure_name(sc, sc->code), sc->y, sc->args)));
	  }
	slot_set_value(x, val);
	return(val);
      }
  return(sc->NO_VALUE);
}


static s7_pointer lambda_star_set_args(s7_scheme *sc)
{
  /* sc->code is a closure: ((args body) envir)
   * (define* (hi a (b 1)) (+ a b))
   * (procedure-source hi) -> (lambda* (a (b 1)) (+ a b))
   *
   * so rather than spinning through the args binding names to values in the
   *   procedure's new environment (as in the usual closure case above),
   *   we scan the current args, and match against the
   *   template in the car of the closure, binding as we go.
   *
   * for each actual arg, if it's not a keyword that matches a member of the 
   *   template, bind it to its current (place-wise) arg, else bind it to
   *   that arg.  If it's the symbol :key or :optional, just go on.
   *   If it's :rest bind the next arg to the trailing args at this point.
   *   All args can be accessed by their name as a keyword.
   *   In other words (define* (hi (a 1)) ...) is the same as (define* (hi :key (a 1)) ...) etc.
   *
   * all args are optional, any arg with no default value defaults to #f.
   *   but the rest arg should default to '().
   *
   * I later decided to add two warnings: if a parameter is set twice and if
   *   an unknown keyword is seen in a keyword position and there is no rest arg.
   */

  bool allow_other_keys = false;
  
  /* get the current args, re-setting args that have explicit values */
  sc->x = closure_args(sc->code);
  sc->y = sc->args; 
  sc->z = sc->NIL;
  while ((is_pair(sc->x)) &&
	 (is_pair(sc->y)))
    {
      if ((car(sc->x) == sc->KEY_KEY) ||
	  (car(sc->x) == sc->KEY_OPTIONAL))
	sc->x = cdr(sc->x);                         /* everything is :key and :optional, so these are ignored */
      else
	{
	  if (car(sc->x) == sc->KEY_ALLOW_OTHER_KEYS)
	    {
	      allow_other_keys = true;
	      sc->x = cdr(sc->x);
	    }
	  else
	    {
	      if (car(sc->x) == sc->KEY_REST)           /* the rest arg */
		{
		  /* next arg is bound to trailing args from this point as a list */
		  sc->z = sc->KEY_REST;
		  sc->x = cdr(sc->x);
		  
		  if (is_pair(car(sc->x)))
		    lambda_star_argument_set_value(sc, caar(sc->x), sc->y);
		  else lambda_star_argument_set_value(sc, car(sc->x), sc->y);
		  
		  sc->y = cdr(sc->y);
		  sc->x = cdr(sc->x);
		  if (car(sc->x) == sc->KEY_ALLOW_OTHER_KEYS)
		    break;
		}
	      else
		{
		  if (is_keyword(car(sc->y)))
		    {
		      char *name;                       /* found a keyword, need to remove the ':' before checking the lambda args */
		      s7_pointer sym;
		      
		      name = symbol_name(car(sc->y));
		      if (name[0] == ':')
			sym = make_symbol(sc, (const char *)(name + 1));
		      else
			{
			  /* must be a trailing ':' here, else not is_keyword */
			  name[symbol_name_length(car(sc->y)) - 1] = '\0';
			  sym = make_symbol(sc, name);
			  name[symbol_name_length(car(sc->y)) - 1] = ':';
			}
		      
		      if ((is_null(cdr(sc->y))) ||
			  (lambda_star_argument_set_value(sc, sym, car(cdr(sc->y))) == sc->NO_VALUE))
			{
			  /* if default value is a key, go ahead and use this value.
			   *    (define* (f (a :b)) a) (f :c) 
			   * this has become much trickier than I anticipated...
			   */
			  
			  if ((allow_other_keys) ||
			      (direct_memq(sc->KEY_ALLOW_OTHER_KEYS, sc->x)))
			    {
			      allow_other_keys = true;
			      /* in CL: (defun hi (&key (a 1) &allow-other-keys) a) (hi :b :a :a 3) -> 3 
			       * in s7: (define* (hi (a 1) :allow-other-keys) a)    (hi :b :a :a 3) -> 3
			       */
			      sc->y = cddr(sc->y);
			      continue;
			    }
			  else
			    {
			      if ((is_pair(car(sc->x))) &&
				  (is_keyword(cadar(sc->x))))
				{
				  /* sc->x is the closure args list, not the copy of it in the current environment */
				  s7_pointer x;
				  
				  x = find_symbol(sc, caar(sc->x));
				  if (is_slot(x))
				    {
				      if (is_not_checked(x))
					{
					  set_checked(x);
					  slot_set_value(x, car(sc->y));
					}
				      else 
					{
					  return(s7_error(sc, sc->WRONG_TYPE_ARG,
							  list_4(sc,
								 make_protected_string(sc, "~A: parameter set twice, ~A in ~A"),
								 closure_name(sc, sc->code), sc->y, sc->args)));
					}
				    }
				  else
				    {
				      return(s7_error(sc, sc->WRONG_TYPE_ARG,
						      list_4(sc,
							     make_protected_string(sc, "~A: unknown key: ~A in ~A"),
							     closure_name(sc, sc->code), sc->y, sc->args)));
				    }
				  /* (define* (f a (b :c)) b) (f :b 1 :d) */
				}
			      else
				{
				  return(s7_error(sc, sc->WRONG_TYPE_ARG,
						  list_4(sc,
							 make_protected_string(sc, "~A: unknown key: ~A in ~A"),
							 closure_name(sc, sc->code), sc->y, sc->args)));
				}
			    }
			}
		      sc->y = cdr(sc->y);
		      if (is_pair(sc->y)) sc->y = cdr(sc->y);
		    }
		  else                                  /* not a key/value pair */
		    {
		      if (is_pair(car(sc->x)))
			lambda_star_argument_set_value(sc, caar(sc->x), car(sc->y));
		      else lambda_star_argument_set_value(sc, car(sc->x), car(sc->y));
		      sc->y = cdr(sc->y);
		    }
		  sc->x = cdr(sc->x);
		}
	    }
	}
    }

  /* (let () (define* (hi (a 1) :allow-other-keys) a) (hi :a 2 32)) */
  /* (let () (define* (f (a :b)) a) (list (f) (f 1) (f :c) (f :a :c) (f :a 1) (f))) */

  /* check for trailing args with no :rest arg */
  if (is_not_null(sc->y))
    {
      if ((is_not_null(sc->x)) ||
	  (sc->z == sc->KEY_REST))
	{
	  if (is_symbol(sc->x))
	    add_slot(sc, sc->x, sc->y); 
	}
      else
	{
	  if ((!allow_other_keys) ||
	      (!is_keyword(car(sc->y))))
	    {
	      return(s7_error(sc, sc->WRONG_NUMBER_OF_ARGS, 
			      list_3(sc, sc->TOO_MANY_ARGUMENTS, closure_name(sc, sc->code), sc->args)));
	    }
	} 
    }
  return(sc->NIL);
}



/* even with the frame_id optimization, this is still about 10% of our total computing: 146/1594 in lg
 *   but the separate finders no longer buy us much: 1608 vs 1583.
 * by using 1 function here, I can specialize safe do lookups without any overhead --
 *   so it's worth the 1.5% slow down elsewhere.  Or so it seemed... I am not hitting
 *   safe do much outside of CLM, and when I do, no big win from the specialized lookup.
 */


static s7_pointer find_symbol_or_bust(s7_scheme *sc, s7_pointer hdl) 
{
  s7_pointer x;

  if (frame_id(sc->envir) == symbol_id(hdl))
    return(slot_value(local_slot(hdl)));

  for (x = sc->envir; symbol_id(hdl) < frame_id(x); x = next_environment(x));

  if (frame_id(x) == symbol_id(hdl))
    return(slot_value(local_slot(hdl)));

  for (; is_environment(x); x = next_environment(x))
    {
      s7_pointer y;
      for (y = environment_slots(x); is_slot(y); y = next_slot(y))
	if (slot_symbol(y) == hdl)
	  return(slot_value(y));
    }

  x = global_slot(hdl);
  if (is_not_null(x)) return(slot_value(x));

  return(unbound_variable(sc, hdl));
}


#if WITH_OPTIMIZATION

/* work in progress, to say the least... */

static bool is_h_safe_c_s(s7_pointer test)
{
  int op;
  op = optimize_data(test);
  return((op == HOP_SAFE_C_S) ||
	 (op == HOP_SAFE_IS_PAIR_S) ||
	 (op == HOP_SAFE_IS_SYMBOL_S) ||
	 (op == HOP_SAFE_NOT_S));
}

static bool is_safe_c_s(s7_pointer test)
{
  int op;
  op = optimize_data(test) & 0xfffe;
  return((op == OP_SAFE_C_S) ||
	 (op == OP_SAFE_IS_PAIR_S) ||
	 (op == OP_SAFE_IS_SYMBOL_S) ||
	 (op == OP_SAFE_NOT_S));
}

static bool is_h_optimized(s7_pointer p)
{
  return((is_optimized(p)) &&
	 ((optimize_data(p) & 1) != 0) &&
	 ((optimize_data(p) < OP_UNKNOWN_C) ||
	  (optimize_data(p) >= OP_SAFE_C_P)));
}



static s7_pointer find_safe_do_symbol_or_bust(s7_scheme *sc, s7_pointer sym) 
{
  s7_pointer val;
  val = (s7_pointer)symbol_op_data(sym);
  if (!val)
    {
      val = find_symbol(sc, sym);
      if (!is_slot(val)) 
	return(unbound_variable(sc, sym));
      if (s7_is_do_local_or_global(sc, sym))
	symbol_op_data(sym) = (void *)val;
    }
  return(slot_value(val));
}


static s7_pointer abs_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
#if (!WITH_GMP)
  if ((is_optimized(cadr(expr))) &&
      (optimize_data(cadr(expr)) == HOP_SAFE_C_SS) &&
      (ecdr(cadr(expr)) == subtract_2))
    {
      optimize_data(expr) = HOP_SAFE_C_C;
      return(abs_sub_ss);
    }
#endif
  return(f);
}

static s7_pointer is_pair_car, is_pair_cdr, is_pair_cadr;
static s7_pointer g_is_pair_car(s7_scheme *sc, s7_pointer args) 
{
  s7_pointer val;
  val = finder(sc, cadar(args));
  if (!is_pair(val))
    return(s7_wrong_type_arg_error(sc, "car", 0, val, "a pair"));
  return(make_boolean(sc, is_pair(car(val))));
}

static s7_pointer g_is_pair_cdr(s7_scheme *sc, s7_pointer args) 
{
  s7_pointer val;
  val = finder(sc, cadar(args));
  if (!is_pair(val))
    return(s7_wrong_type_arg_error(sc, "cdr", 0, val, "a pair"));
  return(make_boolean(sc, is_pair(cdr(val))));
}

static s7_pointer g_is_pair_cadr(s7_scheme *sc, s7_pointer args) 
{
  s7_pointer val;
  val = finder(sc, cadar(args));
  if (!is_pair(val))
    return(s7_wrong_type_arg_error(sc, "cadr", 0, val, "a pair"));
  return(make_boolean(sc, is_pair(cadr(val))));
}

static s7_pointer is_pair_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (is_optimized(cadr(expr)))
    {
      if (optimize_data(cadr(expr)) == HOP_SAFE_C_S)
	{
	  s7_function g;
	  g = c_call(cadr(expr));
	  if (g == g_car)
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(is_pair_car);
	    }
	  if (g == g_cdr)
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(is_pair_cdr);
	    }
	  if (g == g_cadr)
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(is_pair_cadr);
	    }
	}
    }
  if ((is_symbol(cadr(expr))) &&
      (optimize_data(expr) == HOP_SAFE_C_S))
    {
      optimize_data(expr) = HOP_SAFE_IS_PAIR_S;
    }
  return(f);
}

static s7_pointer is_null_cdr;
static s7_pointer g_is_null_cdr(s7_scheme *sc, s7_pointer args) 
{
  s7_pointer val;
  val = finder(sc, cadar(args));
  if (!is_pair(val))
    return(s7_wrong_type_arg_error(sc, "cdr", 0, val, "a pair"));
  return(make_boolean(sc, is_null(cdr(val))));
}

static s7_pointer is_null_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if ((is_optimized(cadr(expr))) &&
      (optimize_data(cadr(expr)) == HOP_SAFE_C_S))
    {
      s7_function g;
      g = c_call(cadr(expr));
      if (g == g_cdr)
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(is_null_cdr);
	}
    }
  return(f);
}

static s7_pointer is_symbol_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if ((is_symbol(cadr(expr))) &&
      (optimize_data(expr) == HOP_SAFE_C_S))
    {
      optimize_data(expr) = HOP_SAFE_IS_SYMBOL_S;
    }
  return(f);
}

static s7_pointer is_eq_car, is_eq_car_q, is_eq_caar_q;
static s7_pointer g_is_eq_car(s7_scheme *sc, s7_pointer args) 
{
  s7_pointer lst, val;
  lst = finder(sc, cadar(args));
  if (!is_pair(lst))
    return(s7_wrong_type_arg_error(sc, "car", 0, lst, "a pair"));
  val = finder(sc, cadr(args));
  return(make_boolean(sc, car(lst) == val));
}

static s7_pointer g_is_eq_car_q(s7_scheme *sc, s7_pointer args) 
{
  s7_pointer lst;
  lst = finder(sc, cadar(args));
  if (!is_pair(lst))
    return(s7_wrong_type_arg_error(sc, "car", 0, lst, "a pair"));
  return(make_boolean(sc, car(lst) == cadr(cadr(args))));
}

static s7_pointer g_is_eq_caar_q(s7_scheme *sc, s7_pointer args) 
{
  s7_pointer lst;
  lst = finder(sc, cadar(args));
  if (!is_pair(lst))
    return(s7_wrong_type_arg_error(sc, "caar", 0, lst, "a pair"));
  return(make_boolean(sc, caar(lst) == cadr(cadr(args))));
}

static s7_pointer is_eq_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if ((is_pair(cadr(expr))) &&
      (is_optimized(cadr(expr))) &&
      (optimize_data(cadr(expr)) == HOP_SAFE_C_S))
    {
      if ((is_symbol(caddr(expr))) &&
	  (c_call(cadr(expr)) == g_car))
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(is_eq_car);
	}
      if ((is_pair(caddr(expr))) &&
	  (caaddr(expr) == sc->QUOTE))
	{
	  if (c_call(cadr(expr)) == g_car)
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(is_eq_car_q);
	    }
	  if (c_call(cadr(expr)) == g_caar)
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(is_eq_caar_q);
	    }
	}
    }
  return(f);
}


static s7_pointer is_zero_logand_s_ash_cs; /* yow */
static s7_pointer g_is_zero_logand_s_ash_cs(s7_scheme *sc, s7_pointer args) 
{
  /* not logbitp */
  s7_pointer s1, s2, x;
  s7_Int i1, i2, i3;
  x = cdar(args);

  s1 = finder(sc, car(x)); /* car(args) = (logand sym (ash 1 pos)), cadar = sym */
  if (!s7_is_integer(s1))
    return(s7_wrong_type_arg_error(sc, "logand", 1, s1, "an integer"));

  s2 = finder(sc, caddr(cadr(x)));
  if (!s7_is_integer(s2))
    return(s7_wrong_type_arg_error(sc, "ash", 2, s2, "an integer"));

  i1 = s7_integer(cadr(cadr(x)));
  i2 = s7_integer(s2);
  i3 = s7_integer(s1);

  return(make_boolean(sc, ((i1 << i2) & i3) == 0));
}

static s7_pointer is_zero_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  /* (define (tst a b) (zero? (logand a (ash 1 b))))
   *     cadr(expr) = (logand a (ash 1 b))
   *     caddr(cadr(expr)) = (ash 1 b)
   */
  s7_pointer x;
  x = cadr(expr);

  if ((is_optimized(x)) &&
      (c_call(x) == g_logand))
    {
#if (!WITH_GMP)
      if (optimize_data(cadr(expr)) == HOP_SAFE_C_SS)
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(is_zero_logand_ss);
	}
#endif
      if ((optimize_data(x) == HOP_SAFE_C_S_opCSq) &&
	  (c_call(caddr(x)) == g_ash) &&
	  (s7_is_integer(cadr(caddr(x)))))
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(is_zero_logand_s_ash_cs);
	}
    }
  return(f);
}


/* also not-chooser for all the ? procs, ss case for not equal? etc
 */
static s7_pointer not_is_pair, not_is_symbol, not_is_null, not_is_list, not_is_number, not_is_real, not_is_rational, not_is_integer;
static s7_pointer not_is_boolean, not_is_char, not_is_string, not_is_eof;
static s7_pointer not_is_eq_sq, not_is_eq_ss;

static s7_pointer g_not_is_pair(s7_scheme *sc, s7_pointer args) {return(make_boolean(sc, !is_pair(finder(sc, cadar(args)))));}
static s7_pointer g_not_is_null(s7_scheme *sc, s7_pointer args) {return(make_boolean(sc, !is_null(finder(sc, cadar(args)))));}
static s7_pointer g_not_is_symbol(s7_scheme *sc, s7_pointer args) {return(make_boolean(sc, !is_symbol(finder(sc, cadar(args)))));}
static s7_pointer g_not_is_number(s7_scheme *sc, s7_pointer args) {return(make_boolean(sc, !s7_is_number(finder(sc, cadar(args)))));}
static s7_pointer g_not_is_real(s7_scheme *sc, s7_pointer args) {return(make_boolean(sc, !s7_is_real(finder(sc, cadar(args)))));}
static s7_pointer g_not_is_integer(s7_scheme *sc, s7_pointer args) {return(make_boolean(sc, !s7_is_integer(finder(sc, cadar(args)))));}
static s7_pointer g_not_is_rational(s7_scheme *sc, s7_pointer args) {return(make_boolean(sc, !s7_is_rational(finder(sc, cadar(args)))));}
static s7_pointer g_not_is_list(s7_scheme *sc, s7_pointer args) {return(make_boolean(sc, !is_proper_list(sc, finder(sc, cadar(args)))));}
static s7_pointer g_not_is_boolean(s7_scheme *sc, s7_pointer args) {return(make_boolean(sc, !s7_is_boolean(finder(sc, cadar(args)))));}
static s7_pointer g_not_is_char(s7_scheme *sc, s7_pointer args) {return(make_boolean(sc, !s7_is_character(finder(sc, cadar(args)))));}
static s7_pointer g_not_is_string(s7_scheme *sc, s7_pointer args) {return(make_boolean(sc, !s7_is_string(finder(sc, cadar(args)))));}
static s7_pointer g_not_is_eof(s7_scheme *sc, s7_pointer args) {return(make_boolean(sc, finder(sc, cadar(args)) != sc->EOF_OBJECT));}

static s7_pointer g_not_is_eq_sq(s7_scheme *sc, s7_pointer args) 
{
  return(make_boolean(sc, finder(sc, cadr(car(args))) != cadr(caddr(car(args)))));
}

static s7_pointer g_not_is_eq_ss(s7_scheme *sc, s7_pointer args) 
{
  return(make_boolean(sc, finder(sc, cadr(car(args))) != finder(sc, caddr(car(args)))));
}

static s7_pointer not_is_pair_car;
static s7_pointer g_not_is_pair_car(s7_scheme *sc, s7_pointer args) 
{
  s7_pointer val;
  val = finder(sc, cadr(cadar(args)));
  if (!is_pair(val))
    return(s7_wrong_type_arg_error(sc, "car", 0, val, "a pair"));
  return(make_boolean(sc, !is_pair(car(val))));
}

static s7_pointer not_is_null_cdr;
static s7_pointer g_not_is_null_cdr(s7_scheme *sc, s7_pointer args) 
{
  s7_pointer val;
  val = finder(sc, cadr(cadar(args)));
  if (!is_pair(val))
    return(s7_wrong_type_arg_error(sc, "cdr", 0, val, "a pair"));
  return(make_boolean(sc, !is_null(cdr(val))));
}

static s7_pointer not_c_c;
static s7_pointer g_not_c_c(s7_scheme *sc, s7_pointer args) 
{
  /* args: ( (null? l) ) */
  return(make_boolean(sc, is_false(sc, c_call(car(args))(sc, cdar(args)))));
}

static s7_pointer not_chooser(s7_scheme *sc, s7_pointer g, int args, s7_pointer expr)
{
  if (is_optimized(cadr(expr))) /* cadr(expr) might be a symbol, for example; is_optimized includes is_pair */
    {
      if (optimize_data(cadr(expr)) == HOP_SAFE_IS_PAIR_S)
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(not_is_pair);
	}
      if (optimize_data(cadr(expr)) == HOP_SAFE_IS_SYMBOL_S)
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(not_is_symbol);
	}

      if (optimize_data(cadr(expr)) == HOP_SAFE_C_S)
	{
	  s7_function f;
	  f = c_call(cadr(expr));
	  
	  if (f == g_is_pair)
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(not_is_pair);
	    }
	  if (f == g_is_null)
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(not_is_null);
	    }
	  if (f == g_is_symbol)
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(not_is_symbol);
	    }
	  if (f == g_is_list)
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(not_is_list);
	    }
	  if ((f == g_is_number) || (f == g_is_complex))
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(not_is_number);
	    }
	  if (f == g_is_rational)
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(not_is_rational);
	    }
	  if (f == g_is_real)
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(not_is_real);
	    }
	  if (f == g_is_integer)
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(not_is_integer);
	    }
	  if (f == g_is_char)
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(not_is_char);
	    }
	  if (f == g_is_boolean)
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(not_is_boolean);
	    }
	  if (f == g_is_string)
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(not_is_string);
	    }
	  if (f == g_is_eof_object)
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(not_is_eof);
	    }
	}

      if ((optimize_data(cadr(expr)) == HOP_SAFE_C_SQ) &&
	  (c_call(cadr(expr)) == g_is_eq))
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(not_is_eq_sq);
	}

      if (optimize_data(cadr(expr)) == HOP_SAFE_C_SS)
	{
	  if (c_call(cadr(expr)) == g_is_eq)
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(not_is_eq_ss);
	    }
	}

      if (optimize_data(cadr(expr)) == HOP_SAFE_C_C)
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  if (c_call(cadr(expr)) == g_is_pair_car)
	    return(not_is_pair_car);
	  if (c_call(cadr(expr)) == g_is_null_cdr)
	    return(not_is_null_cdr);
	  return(not_c_c);
	}
    }

  if ((is_symbol(cadr(expr))) &&
      (optimize_data(expr) == HOP_SAFE_C_S))
    {
      optimize_data(expr) = HOP_SAFE_NOT_S;
    }
  return(g);
}


static s7_pointer vector_ref_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (is_symbol(cadr(expr)))
	{
	  if ((s7_is_integer(caddr(expr))) &&
	      (s7_integer(caddr(expr)) >= 0))
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(vector_ref_ic);
	    }

	  if ((is_pair(caddr(expr))) &&
	      (is_optimized(caddr(expr))) &&
	      (c_call(caddr(expr)) == g_add_cs1))
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(vector_ref_add1);
	      /* vector_ref_sub1 was not worth the code, and few other easily optimized expressions happen here */
	    }
	}
      return(vector_ref_2);
    }
  return(f);
}


static s7_pointer vector_set_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 3)
    {
      if ((s7_is_integer(caddr(expr))) &&
	  (s7_integer(caddr(expr)) >= 0) &&
	  (is_symbol(cadr(expr))) &&
	  (is_symbol(cadddr(expr))))
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(vector_set_ic);
	}
      if ((is_symbol(cadr(expr))) &&
	  (is_symbol(caddr(expr))) &&
	  (!is_pair(cadddr(expr))) &&
	  (!is_symbol(cadddr(expr))))
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(vector_set_ssc);
	}
      if ((is_symbol(cadr(expr))) &&
	  (is_symbol(caddr(expr))) &&
	  (is_pair(cadddr(expr))) &&
	  (is_optimized(cadddr(expr))))
	{
	  s7_pointer vref;
	  vref = cadddr(expr);
	  if ((is_symbol(cadr(vref))) &&
	      (caddr(vref) == caddr(expr)) &&
	      (c_call(vref) == g_vector_ref_2))
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(vector_set_ref);
	    }
	}
      return(vector_set_3);
    }
  return(f);
}


static s7_pointer list_set_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if ((args == 3) &&
      (s7_is_integer(caddr(expr))) &&
      (s7_integer(caddr(expr)) >= 0) &&
      (s7_integer(caddr(expr)) < MAX_LIST_LENGTH))
    return(list_set_ic);
  return(f);
}


static s7_pointer hash_table_ref_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if ((is_symbol(cadr(expr))) &&
	  (is_symbol(caddr(expr))))
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(hash_table_ref_ss);
	}
      if ((is_symbol(cadr(expr))) &&
	  (is_pair(caddr(expr))) &&
	  (is_optimized(caddr(expr))) &&
	  (optimize_data(caddr(expr)) == HOP_SAFE_C_S) &&
	  (c_call(caddr(expr)) == g_car))
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(hash_table_ref_car);
	}
      return(hash_table_ref_2);
    }
  return(f);
}


static s7_pointer add_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  /* fprintf(stderr, "add %d %s\n", args, DISPLAY(expr)); */
  /* (+ s f) (+ (* s s) s) (+ s s) (+ s (* s s))
   */
  if (args == 1)
    return(add_1);

  if (args == 2)
    {
      if (cadr(expr) == small_int(1))
	return(add_1s);
      if (caddr(expr) == small_int(1))
	{
	  if (is_symbol(cadr(expr)))
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(add_cs1);
	    }
	  return(add_s1);
	}

      if ((s7_is_integer(caddr(expr))) &&
	  (is_symbol(cadr(expr))))
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(add_si);
	}

      if ((s7_is_integer(cadr(expr))) &&
	  (is_symbol(caddr(expr))))
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(add_is);
	}

      if ((type(caddr(expr)) == T_REAL) &&
	  (is_symbol(cadr(expr))))
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(add_sf);
	}

      if ((type(cadr(expr)) == T_REAL) &&
	  (is_symbol(caddr(expr))))
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(add_fs);
	}

#if 0
      /* in snd-test, apparently (+ int (* int int)) is very common, but I think it is fake: (sound-data-set! sd chn i (+ i (* chn 10))) and so on
       */
      if ((!is_pair(cadr(expr))) &&
	  (is_pair(caddr(expr))) &&
	  (is_optimized(caddr(expr))) &&
	  (optimize_data(caddr(expr)) == HOP_SAFE_C_C) &&
	  ((c_call(caddr(expr)) == g_multiply_is) || (c_call(caddr(expr)) == g_multiply_si)))
	{
	  fprintf(stderr, "add2 si + p %s: %s\n", (is_optimized(caddr(expr))) ? opt_names[optimize_data(caddr(expr))] : "unopt", DISPLAY_80(expr)); 
	}
#endif

      return(add_2);
    }

  return(f);
}


static s7_pointer multiply_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  /* it's possible to precompute constant multiplies, use OP_CONSTANT as the optimize_data, 
   *    put the value in fcdr, remove it from the heap, return NULL (checking for ecdr null in 
   *    set_c_function and c_function_is_ok), then simply return fcdr in OP_CONSTANT, but
   *    that sort of thing almost never happens in real code.
   */
  if (args == 2)
    {
      if ((is_symbol(cadr(expr))) &&
	  (s7_is_integer(caddr(expr))))
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(multiply_si);
	}

      if ((is_symbol(caddr(expr))) &&
	  (s7_is_integer(cadr(expr))))
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(multiply_is);
	}

      if ((type(cadr(expr)) == T_REAL) &&
	  (is_symbol(caddr(expr))))
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(multiply_fs);
	}
      if ((type(caddr(expr)) == T_REAL) &&
	  (is_symbol(cadr(expr))))
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(multiply_sf);
	}
#if (!WITH_GMP)
      if ((is_symbol(cadr(expr))) &&
	  (cadr(expr) == caddr(expr)))
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(sqr_ss);
	}
#endif
      /* fprintf(stderr, "mul2: %s\n", DISPLAY_80(expr)); */

      /* in clm, gen * gen of any sort is real * real: (* (env ampf) (polywave gen1 (+ (env frqf) ...)))
       *       or even (* mx (polynomial pcoeffs (* amp sig)))
       *       how to handle constants? (* 0.25 3030)
       *       (* r (sin mx)) (* 0.5 (+ x0 x1)) (* x2 (+ 38.027264 x2))
       *     !!  (* (env indf1) (oscil fmosc1 (+ (* fm1-rat vib) fuzz)))
       *       and here one is known to be int: (* ak (floor (/ r ak))) (* 1.25 (frames))
       */
      return(multiply_2);
    }
  return(f);
}

/* (define (hi a) (- a 1.0)) (hi (string->number "#e1."))
 */

static s7_pointer subtract_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 1)
    return(subtract_1);

  if (args == 2)
    {
      if (caddr(expr) == small_int(1))
	{
	  if (is_symbol(cadr(expr)))
	    {
	      /* fprintf(stderr, "subtract_cs1: %s\n", DISPLAY(expr)); */
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(subtract_cs1);
	    }
	  /* fprintf(stderr, "subtract_s1: %s\n", DISPLAY(expr));  */
	  return(subtract_s1);
	}
      if ((s7_is_integer(caddr(expr))) &&
	  (is_symbol(cadr(expr))))
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(subtract_csn);
	}
      return(subtract_2);
    }
  return(f);
}


static s7_pointer equal_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (s7_is_integer(caddr(expr)))
	{
	  if (is_optimized(cadr(expr)))
	    {
	      s7_function f;
	      f = c_call(cadr(expr));
	      if (f == g_length)
		{
		  if (optimize_data(cadr(expr)) == HOP_SAFE_C_S)
		    {
		      optimize_data(expr) = HOP_SAFE_C_C;
		      return(equal_length_ic);
		    }
		}
	    }
	  if (is_symbol(cadr(expr)))
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(equal_s_ic);
	    }
	}
      /* fprintf(stderr, "=2: %s\n", DISPLAY_80(expr)); */
      return(equal_2);
    }
  return(f);
}


#if (!WITH_GMP)
static s7_pointer less_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (s7_is_integer(caddr(expr)))
	{
	  if ((is_optimized(cadr(expr))) &&
	      (optimize_data(cadr(expr)) == HOP_SAFE_C_S))
	    {
	      s7_function f;
	      f = c_call(cadr(expr));
	      if (f == g_length)
		{
		  optimize_data(expr) = HOP_SAFE_C_C;
		  return(less_length_ic);
		}
	    }
	  return(less_s_ic);
	}
      return(less_2);
    }
  return(f);
}


static s7_pointer leq_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (s7_is_integer(caddr(expr)))
	return(leq_s_ic);
      return(leq_2);
    }
  return(f);
}


static s7_pointer greater_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if ((s7_is_real(caddr(expr))) &&
	  (!s7_is_rational(caddr(expr))) &&
	  (is_optimized(cadr(expr))) &&
	  (optimize_data(cadr(expr)) == HOP_SAFE_C_C) &&
	  (ecdr(cadr(expr)) == abs_sub_ss))
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(greater_abs);
	}

      if (s7_is_integer(caddr(expr)))
	return(greater_s_ic);
      return(greater_2);
    }
  return(f);
}


static s7_pointer geq_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (s7_is_integer(caddr(expr)))
	{
	  if ((is_optimized(cadr(expr))) &&
	      (optimize_data(cadr(expr)) == HOP_SAFE_C_S))
	    {
	      s7_function f;
	      f = c_call(cadr(expr));
	      if (f == g_length)
		{
		  optimize_data(expr) = HOP_SAFE_C_C;
		  return(geq_length_ic);
		}
	    }
	  return(geq_s_ic);
	}
      /* fprintf(stderr, ">=2: %s\n", DISPLAY_80(expr)); */
      return(geq_2);
    }
  return(f);
}


static s7_pointer is_negative_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 1)
    {
      if ((is_optimized(cadr(expr))) &&
	  (optimize_data(cadr(expr)) == HOP_SAFE_C_S))
	{
	  s7_function f;
	  f = c_call(cadr(expr));
	  if (f == g_length)
	    {
	      optimize_data(expr) = HOP_SAFE_C_C;
	      return(is_negative_length);
	    }
	}
    }
  return(f);
}
#endif


static s7_pointer char_equal_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if ((is_symbol(cadr(expr))) &&
	  (s7_is_character(caddr(expr))))
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(char_equal_s_ic);
	}
      return(char_equal_2);
    }
  return(f);
}

static s7_pointer char_less_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (s7_is_character(caddr(expr)))
	return(char_less_s_ic);
      return(char_less_2);
    }
  return(f);
}

static s7_pointer char_greater_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (s7_is_character(caddr(expr)))
	return(char_greater_s_ic);
      return(char_greater_2);
    }
  return(f);
}

static s7_pointer char_geq_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (s7_is_character(caddr(expr)))
	return(char_geq_s_ic);
      return(char_geq_2);
    }
  return(f);
}

static s7_pointer char_leq_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (s7_is_character(caddr(expr)))
	return(char_leq_s_ic);
      return(char_leq_2);
    }
  return(f);
}


static s7_pointer char_ci_equal_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (s7_is_character(caddr(expr)))
	return(char_ci_equal_s_ic);
      return(char_ci_equal_2);
    }
  return(f);
}

static s7_pointer char_ci_less_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (s7_is_character(caddr(expr)))
	return(char_ci_less_s_ic);
      return(char_ci_less_2);
    }
  return(f);
}

static s7_pointer char_ci_greater_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (s7_is_character(caddr(expr)))
	return(char_ci_greater_s_ic);
      return(char_ci_greater_2);
    }
  return(f);
}

static s7_pointer char_ci_geq_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (s7_is_character(caddr(expr)))
	return(char_ci_geq_s_ic);
      return(char_ci_geq_2);
    }
  return(f);
}

static s7_pointer char_ci_leq_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (s7_is_character(caddr(expr)))
	return(char_ci_leq_s_ic);
      return(char_ci_leq_2);
    }
  return(f);
}


static s7_pointer string_equal_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (s7_is_string(caddr(expr)))
	return(string_equal_s_ic); 

      if ((is_symbol(caddr(expr))) &&
	  (is_pair(cadr(expr))) &&
	  (is_optimized(cadr(expr))) &&
	  (optimize_data(cadr(expr)) == HOP_SAFE_C_SS) &&
	  (c_call(cadr(expr)) == g_vector_ref_2))
	{
	  optimize_data(expr) = HOP_SAFE_C_C;
	  return(string_equal_vref_s);
	}
      return(string_equal_2);
    }
  return(f);
}

static s7_pointer string_less_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (s7_is_string(caddr(expr)))
	return(string_less_s_ic);
      
      if ((is_pair(cadr(expr))) &&
	  (is_optimized(cadr(expr))) &&
	  (c_call(cadr(expr)) == g_symbol_to_string))
	set_c_function(cadr(expr), symbol_to_string_uncopied);

      if ((is_pair(caddr(expr))) &&
	  (is_optimized(caddr(expr))) &&
	  (c_call(caddr(expr)) == g_symbol_to_string))
	set_c_function(caddr(expr), symbol_to_string_uncopied);

      return(string_less_2);
    }
  return(f);
}

static s7_pointer string_greater_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (s7_is_string(caddr(expr)))
	return(string_greater_s_ic);
      return(string_greater_2);
    }
  return(f);
}

static s7_pointer string_geq_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (s7_is_string(caddr(expr)))
	return(string_geq_s_ic);
      return(string_geq_2);
    }
  return(f);
}

static s7_pointer string_leq_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (s7_is_string(caddr(expr)))
	return(string_leq_s_ic);
      return(string_leq_2);
    }
  return(f);
}


static s7_pointer string_ci_equal_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (s7_is_string(caddr(expr)))
	return(string_ci_equal_s_ic);
      return(string_ci_equal_2);
    }
  return(f);
}

static s7_pointer string_ci_less_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (s7_is_string(caddr(expr)))
	return(string_ci_less_s_ic);
      return(string_ci_less_2);
    }
  return(f);
}

static s7_pointer string_ci_greater_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (s7_is_string(caddr(expr)))
	return(string_ci_greater_s_ic);
      return(string_ci_greater_2);
    }
  return(f);
}

static s7_pointer string_ci_geq_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (s7_is_string(caddr(expr)))
	return(string_ci_geq_s_ic);
      return(string_ci_geq_2);
    }
  return(f);
}

static s7_pointer string_ci_leq_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if (args == 2)
    {
      if (s7_is_string(caddr(expr)))
	return(string_ci_leq_s_ic);
      return(string_ci_leq_2);
    }
  return(f);
}


static s7_pointer string_ref_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if ((is_pair(cadr(expr))) &&
      (is_optimized(cadr(expr))) &&
      (c_call(cadr(expr)) == g_symbol_to_string))
    set_c_function(cadr(expr), symbol_to_string_uncopied);
  return(f);
}

static s7_pointer string_append_chooser(s7_scheme *sc, s7_pointer f, int args, s7_pointer expr)
{
  if ((is_pair(cadr(expr))) &&
      (is_optimized(cadr(expr))) &&
      (c_call(cadr(expr)) == g_symbol_to_string))
    set_c_function(cadr(expr), symbol_to_string_uncopied);
  return(f);
}



static unsigned int add_class = 0, equal_class = 0;

static void init_choosers(s7_scheme *sc)
{
  s7_pointer f;

  /* + */
  f = slot_value(global_slot(make_symbol(sc, "+")));
  add_class = c_function_class(f);
  c_function_chooser(f) = add_chooser;

  add_1 = s7_make_function(sc, "+", g_add_1, 1, 0, false, "+ optimization");
  c_function_class(add_1) = c_function_class(f);
  add_2 = s7_make_function(sc, "+", g_add_2, 2, 0, false, "+ optimization");
  c_function_class(add_2) = c_function_class(f);
  add_1s = s7_make_function(sc, "+", g_add_1s, 2, 0, false, "+ optimization");
  c_function_class(add_1s) = c_function_class(f);
  add_s1 = s7_make_function(sc, "+", g_add_s1, 2, 0, false, "+ optimization");
  c_function_class(add_s1) = c_function_class(f);
  add_cs1 = s7_make_function(sc, "+", g_add_cs1, 2, 0, false, "+ optimization");
  c_function_class(add_cs1) = c_function_class(f);
  add_si = s7_make_function(sc, "+", g_add_si, 2, 0, false, "+ optimization");
  c_function_class(add_si) = c_function_class(f);
  add_is = s7_make_function(sc, "+", g_add_is, 2, 0, false, "+ optimization");
  c_function_class(add_is) = c_function_class(f);
  add_sf = s7_make_function(sc, "+", g_add_sf, 2, 0, false, "+ optimization");
  c_function_class(add_sf) = c_function_class(f);
  add_fs = s7_make_function(sc, "+", g_add_fs, 2, 0, false, "+ optimization");
  c_function_class(add_fs) = c_function_class(f);
  

  /* - */
  f = slot_value(global_slot(make_symbol(sc, "-")));
  c_function_chooser(f) = subtract_chooser;

  subtract_1 = s7_make_function(sc, "-", g_subtract_1, 1, 0, false, "- optimization");
  c_function_class(subtract_1) = c_function_class(f);
  subtract_2 = s7_make_function(sc, "-", g_subtract_2, 2, 0, false, "- optimization");
  c_function_class(add_2) = c_function_class(f);
  subtract_s1 = s7_make_function(sc, "-", g_subtract_s1, 2, 0, false, "- optimization");
  c_function_class(subtract_s1) = c_function_class(f);
  subtract_cs1 = s7_make_function(sc, "-", g_subtract_cs1, 2, 0, false, "- optimization");
  c_function_class(subtract_cs1) = c_function_class(f);
  subtract_csn = s7_make_function(sc, "-", g_subtract_csn, 2, 0, false, "- optimization");
  c_function_class(subtract_csn) = c_function_class(f);
  

  /* * */
  f = slot_value(global_slot(make_symbol(sc, "*")));
  c_function_chooser(f) = multiply_chooser;

  multiply_2 = s7_make_function(sc, "*", g_multiply_2, 2, 0, false, "* optimization");
  c_function_class(multiply_2) = c_function_class(f);
  multiply_is = s7_make_function(sc, "*", g_multiply_is, 2, 0, false, "* optimization");
  c_function_class(multiply_is) = c_function_class(f);
  multiply_si = s7_make_function(sc, "*", g_multiply_si, 2, 0, false, "* optimization");
  c_function_class(multiply_si) = c_function_class(f);
  multiply_fs = s7_make_function(sc, "*", g_multiply_fs, 2, 0, false, "* optimization");
  c_function_class(multiply_fs) = c_function_class(f);
  multiply_sf = s7_make_function(sc, "*", g_multiply_sf, 2, 0, false, "* optimization");
  c_function_class(multiply_sf) = c_function_class(f);
#if (!WITH_GMP)
  sqr_ss = s7_make_function(sc, "*", g_sqr_ss, 2, 0, false, "* optimization");
  c_function_class(sqr_ss) = c_function_class(f);
#endif

  /* = */
  f = slot_value(global_slot(make_symbol(sc, "=")));
  equal_class = c_function_class(f);
  c_function_chooser(f) = equal_chooser;

  equal_s_ic = s7_make_function(sc, "=", g_equal_s_ic, 2, 0, false, "= optimization");
  c_function_class(equal_s_ic) = c_function_class(f);
  equal_length_ic = s7_make_function(sc, "=", g_equal_length_ic, 2, 0, false, "= optimization");
  c_function_class(equal_length_ic) = c_function_class(f);
  equal_2 = s7_make_function(sc, "=", g_equal_2, 2, 0, false, "= optimization");
  c_function_class(equal_2) = c_function_class(f);


#if (!WITH_GMP)
  /* negative? */
  f = slot_value(global_slot(make_symbol(sc, "negative?")));
  c_function_chooser(f) = is_negative_chooser;

  is_negative_length = s7_make_function(sc, "negative?", g_is_negative_length, 1, 0, false, "negative? optimization");
  c_function_class(is_negative_length) = c_function_class(f);


  /* < */
  f = slot_value(global_slot(make_symbol(sc, "<")));
  c_function_chooser(f) = less_chooser;

  less_s_ic = s7_make_function(sc, "<", g_less_s_ic, 2, 0, false, "< optimization");
  c_function_class(less_s_ic) = c_function_class(f);
  less_2 = s7_make_function(sc, "<", g_less_2, 2, 0, false, "< optimization");
  c_function_class(less_2) = c_function_class(f);
  less_length_ic = s7_make_function(sc, "<", g_less_length_ic, 2, 0, false, "< optimization");
  c_function_class(less_length_ic) = c_function_class(f);


  /* > */
  f = slot_value(global_slot(make_symbol(sc, ">")));
  c_function_chooser(f) = greater_chooser;

  greater_s_ic = s7_make_function(sc, ">", g_greater_s_ic, 2, 0, false, "> optimization");
  c_function_class(greater_s_ic) = c_function_class(f);
  greater_abs = s7_make_function(sc, ">", g_greater_abs, 2, 0, false, "> optimization");
  c_function_class(greater_abs) = c_function_class(f);
  greater_2 = s7_make_function(sc, ">", g_greater_2, 2, 0, false, "> optimization");
  c_function_class(greater_2) = c_function_class(f);


  /* <= */
  f = slot_value(global_slot(make_symbol(sc, "<=")));
  c_function_chooser(f) = leq_chooser;

  leq_s_ic = s7_make_function(sc, "<=", g_leq_s_ic, 2, 0, false, "<= optimization");
  c_function_class(leq_s_ic) = c_function_class(f);
  leq_2 = s7_make_function(sc, "<=", g_leq_2, 2, 0, false, "<= optimization");
  c_function_class(leq_2) = c_function_class(f);


  /* >= */
  f = slot_value(global_slot(make_symbol(sc, ">=")));
  c_function_chooser(f) = geq_chooser;

  geq_s_ic = s7_make_function(sc, ">=", g_geq_s_ic, 2, 0, false, ">= optimization");
  c_function_class(geq_s_ic) = c_function_class(f);
  geq_2 = s7_make_function(sc, ">=", g_geq_2, 2, 0, false, ">= optimization");
  c_function_class(geq_2) = c_function_class(f);
  geq_length_ic = s7_make_function(sc, ">=", g_geq_length_ic, 2, 0, false, ">= optimization");
  c_function_class(geq_length_ic) = c_function_class(f);
#endif

  /* char=? */
  f = slot_value(global_slot(make_symbol(sc, "char=?")));
  c_function_chooser(f) = char_equal_chooser;

  char_equal_s_ic = s7_make_function(sc, "char=?", g_char_equal_s_ic, 2, 0, false, "char=? optimization");
  c_function_class(char_equal_s_ic) = c_function_class(f);
  char_equal_2 = s7_make_function(sc, "char=?", g_char_equal_2, 2, 0, false, "char=? optimization");
  c_function_class(char_equal_2) = c_function_class(f);


  /* char>? */
  f = slot_value(global_slot(make_symbol(sc, "char>?")));
  c_function_chooser(f) = char_greater_chooser;

  char_greater_s_ic = s7_make_function(sc, "char>?", g_char_greater_s_ic, 2, 0, false, "char>? optimization");
  c_function_class(char_greater_s_ic) = c_function_class(f);
  char_greater_2 = s7_make_function(sc, "char>?", g_char_greater_2, 2, 0, false, "char>? optimization");
  c_function_class(char_greater_2) = c_function_class(f);


  /* char<? */
  f = slot_value(global_slot(make_symbol(sc, "char<?")));
  c_function_chooser(f) = char_less_chooser;

  char_less_s_ic = s7_make_function(sc, "char<?", g_char_less_s_ic, 2, 0, false, "char<? optimization");
  c_function_class(char_less_s_ic) = c_function_class(f);
  char_less_2 = s7_make_function(sc, "char<?", g_char_less_2, 2, 0, false, "char<? optimization");
  c_function_class(char_less_2) = c_function_class(f);


  /* char<=? */
  f = slot_value(global_slot(make_symbol(sc, "char<=?")));
  c_function_chooser(f) = char_leq_chooser;

  char_leq_s_ic = s7_make_function(sc, "char<=?", g_char_leq_s_ic, 2, 0, false, "char<=? optimization");
  c_function_class(char_leq_s_ic) = c_function_class(f);
  char_leq_2 = s7_make_function(sc, "char<=?", g_char_leq_2, 2, 0, false, "char<=? optimization");
  c_function_class(char_leq_2) = c_function_class(f);


  /* char>=? */
  f = slot_value(global_slot(make_symbol(sc, "char>=?")));
  c_function_chooser(f) = char_geq_chooser;

  char_geq_s_ic = s7_make_function(sc, "char>=?", g_char_geq_s_ic, 2, 0, false, "char>=? optimization");
  c_function_class(char_geq_s_ic) = c_function_class(f);
  char_geq_2 = s7_make_function(sc, "char>=?", g_char_geq_2, 2, 0, false, "char>=? optimization");
  c_function_class(char_geq_2) = c_function_class(f);


  /* char-ci=? */
  f = slot_value(global_slot(make_symbol(sc, "char-ci=?")));
  c_function_chooser(f) = char_ci_equal_chooser;

  char_ci_equal_s_ic = s7_make_function(sc, "char-ci=?", g_char_ci_equal_s_ic, 2, 0, false, "char-ci=? optimization");
  c_function_class(char_ci_equal_s_ic) = c_function_class(f);
  char_ci_equal_2 = s7_make_function(sc, "char-ci=?", g_char_ci_equal_2, 2, 0, false, "char-ci=? optimization");
  c_function_class(char_ci_equal_2) = c_function_class(f);


  /* char-ci>? */
  f = slot_value(global_slot(make_symbol(sc, "char-ci>?")));
  c_function_chooser(f) = char_ci_greater_chooser;

  char_ci_greater_s_ic = s7_make_function(sc, "char-ci>?", g_char_ci_greater_s_ic, 2, 0, false, "char-ci>? optimization");
  c_function_class(char_ci_greater_s_ic) = c_function_class(f);
  char_ci_greater_2 = s7_make_function(sc, "char-ci>?", g_char_ci_greater_2, 2, 0, false, "char-ci>? optimization");
  c_function_class(char_ci_greater_2) = c_function_class(f);


  /* char-ci<? */
  f = slot_value(global_slot(make_symbol(sc, "char-ci<?")));
  c_function_chooser(f) = char_ci_less_chooser;

  char_ci_less_s_ic = s7_make_function(sc, "char-ci<?", g_char_ci_less_s_ic, 2, 0, false, "char-ci<? optimization");
  c_function_class(char_ci_less_s_ic) = c_function_class(f);
  char_ci_less_2 = s7_make_function(sc, "char-ci<?", g_char_ci_less_2, 2, 0, false, "char-ci<? optimization");
  c_function_class(char_ci_less_2) = c_function_class(f);


  /* char-ci<=? */
  f = slot_value(global_slot(make_symbol(sc, "char-ci<=?")));
  c_function_chooser(f) = char_ci_leq_chooser;

  char_ci_leq_s_ic = s7_make_function(sc, "char-ci<=?", g_char_ci_leq_s_ic, 2, 0, false, "char-ci<=? optimization");
  c_function_class(char_ci_leq_s_ic) = c_function_class(f);
  char_ci_leq_2 = s7_make_function(sc, "char-ci<=?", g_char_ci_leq_2, 2, 0, false, "char-ci<=? optimization");
  c_function_class(char_ci_leq_2) = c_function_class(f);


  /* char-ci>=? */
  f = slot_value(global_slot(make_symbol(sc, "char-ci>=?")));
  c_function_chooser(f) = char_ci_geq_chooser;

  char_ci_geq_s_ic = s7_make_function(sc, "char-ci>=?", g_char_ci_geq_s_ic, 2, 0, false, "char-ci>=? optimization");
  c_function_class(char_ci_geq_s_ic) = c_function_class(f);
  char_ci_geq_2 = s7_make_function(sc, "char-ci>=?", g_char_ci_geq_2, 2, 0, false, "char-ci>=? optimization");
  c_function_class(char_ci_geq_2) = c_function_class(f);


  /* string=? */
  f = slot_value(global_slot(make_symbol(sc, "string=?")));
  c_function_chooser(f) = string_equal_chooser;

  string_equal_s_ic = s7_make_function(sc, "string=?", g_string_equal_s_ic, 2, 0, false, "string=? optimization");
  c_function_class(string_equal_s_ic) = c_function_class(f);
  string_equal_vref_s = s7_make_function(sc, "string=?", g_string_equal_vref_s, 2, 0, false, "string=? optimization");
  c_function_class(string_equal_vref_s) = c_function_class(f);
  string_equal_2 = s7_make_function(sc, "string=?", g_string_equal_2, 2, 0, false, "string=? optimization");
  c_function_class(string_equal_2) = c_function_class(f);


  /* string>? */
  f = slot_value(global_slot(make_symbol(sc, "string>?")));
  c_function_chooser(f) = string_greater_chooser;

  string_greater_s_ic = s7_make_function(sc, "string>?", g_string_greater_s_ic, 2, 0, false, "string>? optimization");
  c_function_class(string_greater_s_ic) = c_function_class(f);
  string_greater_2 = s7_make_function(sc, "string>?", g_string_greater_2, 2, 0, false, "string>? optimization");
  c_function_class(string_greater_2) = c_function_class(f);


  /* string<? */
  f = slot_value(global_slot(make_symbol(sc, "string<?")));
  c_function_chooser(f) = string_less_chooser;

  string_less_s_ic = s7_make_function(sc, "string<?", g_string_less_s_ic, 2, 0, false, "string<? optimization");
  c_function_class(string_less_s_ic) = c_function_class(f);
  string_less_2 = s7_make_function(sc, "string<?", g_string_less_2, 2, 0, false, "string<? optimization");
  c_function_class(string_less_2) = c_function_class(f);


  /* string<=? */
  f = slot_value(global_slot(make_symbol(sc, "string<=?")));
  c_function_chooser(f) = string_leq_chooser;

  string_leq_s_ic = s7_make_function(sc, "string<=?", g_string_leq_s_ic, 2, 0, false, "string<=? optimization");
  c_function_class(string_leq_s_ic) = c_function_class(f);
  string_leq_2 = s7_make_function(sc, "string<=?", g_string_leq_2, 2, 0, false, "string<=? optimization");
  c_function_class(string_leq_2) = c_function_class(f);


  /* string>=? */
  f = slot_value(global_slot(make_symbol(sc, "string>=?")));
  c_function_chooser(f) = string_geq_chooser;

  string_geq_s_ic = s7_make_function(sc, "string>=?", g_string_geq_s_ic, 2, 0, false, "string>=? optimization");
  c_function_class(string_geq_s_ic) = c_function_class(f);
  string_geq_2 = s7_make_function(sc, "string>=?", g_string_geq_2, 2, 0, false, "string>=? optimization");
  c_function_class(string_geq_2) = c_function_class(f);


  /* string-ci=? */
  f = slot_value(global_slot(make_symbol(sc, "string-ci=?")));
  c_function_chooser(f) = string_ci_equal_chooser;

  string_ci_equal_s_ic = s7_make_function(sc, "string-ci=?", g_string_ci_equal_s_ic, 2, 0, false, "string-ci=? optimization");
  c_function_class(string_ci_equal_s_ic) = c_function_class(f);
  string_ci_equal_2 = s7_make_function(sc, "string-ci=?", g_string_ci_equal_2, 2, 0, false, "string-ci=? optimization");
  c_function_class(string_ci_equal_2) = c_function_class(f);


  /* string-ci>? */
  f = slot_value(global_slot(make_symbol(sc, "string-ci>?")));
  c_function_chooser(f) = string_ci_greater_chooser;

  string_ci_greater_s_ic = s7_make_function(sc, "string-ci>?", g_string_ci_greater_s_ic, 2, 0, false, "string-ci>? optimization");
  c_function_class(string_ci_greater_s_ic) = c_function_class(f);
  string_ci_greater_2 = s7_make_function(sc, "string-ci>?", g_string_ci_greater_2, 2, 0, false, "string-ci>? optimization");
  c_function_class(string_ci_greater_2) = c_function_class(f);


  /* string-ci<? */
  f = slot_value(global_slot(make_symbol(sc, "string-ci<?")));
  c_function_chooser(f) = string_ci_less_chooser;

  string_ci_less_s_ic = s7_make_function(sc, "string-ci<?", g_string_ci_less_s_ic, 2, 0, false, "string-ci<? optimization");
  c_function_class(string_ci_less_s_ic) = c_function_class(f);
  string_ci_less_2 = s7_make_function(sc, "string-ci<?", g_string_ci_less_2, 2, 0, false, "string-ci<? optimization");
  c_function_class(string_ci_less_2) = c_function_class(f);


  /* string-ci<=? */
  f = slot_value(global_slot(make_symbol(sc, "string-ci<=?")));
  c_function_chooser(f) = string_ci_leq_chooser;

  string_ci_leq_s_ic = s7_make_function(sc, "string-ci<=?", g_string_ci_leq_s_ic, 2, 0, false, "string-ci<=? optimization");
  c_function_class(string_ci_leq_s_ic) = c_function_class(f);
  string_ci_leq_2 = s7_make_function(sc, "string-ci<=?", g_string_ci_leq_2, 2, 0, false, "string-ci<=? optimization");
  c_function_class(string_ci_leq_2) = c_function_class(f);


  /* string-ci>=? */
  f = slot_value(global_slot(make_symbol(sc, "string-ci>=?")));
  c_function_chooser(f) = string_ci_geq_chooser;

  string_ci_geq_s_ic = s7_make_function(sc, "string-ci>=?", g_string_ci_geq_s_ic, 2, 0, false, "string-ci>=? optimization");
  c_function_class(string_ci_geq_s_ic) = c_function_class(f);
  string_ci_geq_2 = s7_make_function(sc, "string-ci>=?", g_string_ci_geq_2, 2, 0, false, "string-ci>=? optimization");
  c_function_class(string_ci_geq_2) = c_function_class(f);


  /* string-ref */
  f = slot_value(global_slot(make_symbol(sc, "string-ref")));
  c_function_chooser(f) = string_ref_chooser;


  /* string-append */
  f = slot_value(global_slot(make_symbol(sc, "string-append")));
  c_function_chooser(f) = string_append_chooser;


  /* symbol->string experiment */
  f = slot_value(global_slot(make_symbol(sc, "symbol->string")));
  symbol_to_string_uncopied = s7_make_function(sc, "symbol->string", g_symbol_to_string_uncopied, 1, 0, false, "symbol->string optimization");
  c_function_class(symbol_to_string_uncopied) = c_function_class(f);


  /* abs */
  f = slot_value(global_slot(make_symbol(sc, "abs")));
  c_function_chooser(f) = abs_chooser;
#if (!WITH_GMP)
  abs_sub_ss = s7_make_function(sc, "abs", g_abs_sub_ss, 2, 0, false, "abs(- a b) optimization");
  c_function_class(abs_sub_ss) = c_function_class(f);
#endif


  /* for-each */
  f = slot_value(global_slot(make_symbol(sc, "for-each")));
  c_function_chooser(f) = for_each_chooser;
  for_each_1 = s7_make_function(sc, "for-each", g_for_each_1, 2, 0, false, "for-each optimization");
  c_function_class(for_each_1) = c_function_class(f);
  for_each_2 = s7_make_function(sc, "for-each", g_for_each_2, 2, 0, false, "for-each optimization");
  c_function_class(for_each_2) = c_function_class(f);
  for_each_3 = s7_make_function(sc, "for-each", g_for_each_3, 2, 0, false, "for-each optimization");
  c_function_class(for_each_3) = c_function_class(f);


  /* vector-ref */
  f = slot_value(global_slot(make_symbol(sc, "vector-ref")));
  c_function_chooser(f) = vector_ref_chooser;

  vector_ref_ic = s7_make_function(sc, "vector-ref", g_vector_ref_ic, 2, 0, false, "vector-ref optimization");
  c_function_class(vector_ref_ic) = c_function_class(f);
  vector_ref_add1 = s7_make_function(sc, "vector-ref", g_vector_ref_add1, 2, 0, false, "vector-ref optimization");
  c_function_class(vector_ref_add1) = c_function_class(f);
  vector_ref_2 = s7_make_function(sc, "vector-ref", g_vector_ref_2, 2, 0, false, "vector-ref optimization");
  c_function_class(vector_ref_2) = c_function_class(f);


  /* vector-set! */
  f = slot_value(global_slot(make_symbol(sc, "vector-set!")));
  c_function_chooser(f) = vector_set_chooser;

  vector_set_ic = s7_make_function(sc, "vector-set!", g_vector_set_ic, 3, 0, false, "vector-set! optimization");
  c_function_class(vector_set_ic) = c_function_class(f);
  vector_set_ssc = s7_make_function(sc, "vector-set!", g_vector_set_ssc, 3, 0, false, "vector-set! optimization");
  c_function_class(vector_set_ssc) = c_function_class(f);
  vector_set_3 = s7_make_function(sc, "vector-set!", g_vector_set_3, 3, 0, false, "vector-set! optimization");
  c_function_class(vector_set_3) = c_function_class(f);
  vector_set_ref = s7_make_function(sc, "vector-set!", g_vector_set_ref, 3, 0, false, "vector-set! optimization");
  c_function_class(vector_set_ref) = c_function_class(f);


  /* list-set! */
  f = slot_value(global_slot(make_symbol(sc, "list-set!")));
  c_function_chooser(f) = list_set_chooser;

  list_set_ic = s7_make_function(sc, "list-set!", g_list_set_ic, 3, 0, false, "list-set! optimization");
  c_function_class(list_set_ic) = c_function_class(f);


  /* hash-table-ref */
  f = slot_value(global_slot(make_symbol(sc, "hash-table-ref")));
  c_function_chooser(f) = hash_table_ref_chooser;

  hash_table_ref_2 = s7_make_function(sc, "hash-table-ref", g_hash_table_ref_2, 2, 0, false, "hash-table-ref optimization");
  c_function_class(hash_table_ref_2) = c_function_class(f);
  hash_table_ref_ss = s7_make_function(sc, "hash-table-ref", g_hash_table_ref_ss, 2, 0, false, "hash-table-ref optimization");
  c_function_class(hash_table_ref_ss) = c_function_class(f);
  hash_table_ref_car = s7_make_function(sc, "hash-table-ref", g_hash_table_ref_car, 2, 0, false, "hash-table-ref optimization");
  c_function_class(hash_table_ref_car) = c_function_class(f);


  /* not */
  f = slot_value(global_slot(make_symbol(sc, "not")));
  c_function_chooser(f) = not_chooser;

  not_is_pair = s7_make_function(sc, "not", g_not_is_pair, 1, 0, false, "not optimization");
  c_function_class(not_is_pair) = c_function_class(f);
  not_is_null = s7_make_function(sc, "not", g_not_is_null, 1, 0, false, "not optimization");
  c_function_class(not_is_null) = c_function_class(f);
  not_is_list = s7_make_function(sc, "not", g_not_is_list, 1, 0, false, "not optimization");
  c_function_class(not_is_list) = c_function_class(f);
  not_is_symbol = s7_make_function(sc, "not", g_not_is_symbol, 1, 0, false, "not optimization");
  c_function_class(not_is_symbol) = c_function_class(f);
  not_is_number = s7_make_function(sc, "not", g_not_is_number, 1, 0, false, "not optimization");
  c_function_class(not_is_number) = c_function_class(f);
  not_is_real = s7_make_function(sc, "not", g_not_is_real, 1, 0, false, "not optimization");
  c_function_class(not_is_real) = c_function_class(f);
  not_is_rational = s7_make_function(sc, "not", g_not_is_rational, 1, 0, false, "not optimization");
  c_function_class(not_is_rational) = c_function_class(f);
  not_is_integer = s7_make_function(sc, "not", g_not_is_integer, 1, 0, false, "not optimization");
  c_function_class(not_is_integer) = c_function_class(f);
  not_is_boolean = s7_make_function(sc, "not", g_not_is_boolean, 1, 0, false, "not optimization");
  c_function_class(not_is_boolean) = c_function_class(f);
  not_is_string = s7_make_function(sc, "not", g_not_is_string, 1, 0, false, "not optimization");
  c_function_class(not_is_string) = c_function_class(f);
  not_is_char = s7_make_function(sc, "not", g_not_is_char, 1, 0, false, "not optimization");
  c_function_class(not_is_char) = c_function_class(f);
  not_is_eof = s7_make_function(sc, "not", g_not_is_eof, 1, 0, false, "not optimization");
  c_function_class(not_is_eof) = c_function_class(f);
  not_is_eq_ss = s7_make_function(sc, "not", g_not_is_eq_ss, 1, 0, false, "not optimization");
  c_function_class(not_is_eq_ss) = c_function_class(f);
  not_is_eq_sq = s7_make_function(sc, "not", g_not_is_eq_sq, 1, 0, false, "not optimization");
  c_function_class(not_is_eq_sq) = c_function_class(f);
  not_is_pair_car = s7_make_function(sc, "not", g_not_is_pair_car, 1, 0, false, "not optimization");
  c_function_class(not_is_pair_car) = c_function_class(f);
  not_is_null_cdr = s7_make_function(sc, "not", g_not_is_null_cdr, 1, 0, false, "not optimization");
  c_function_class(not_is_null_cdr) = c_function_class(f);
  not_c_c = s7_make_function(sc, "not", g_not_c_c, 1, 0, false, "not optimization");
  c_function_class(not_c_c) = c_function_class(f);


  /* pair? */
  f = slot_value(global_slot(make_symbol(sc, "pair?")));
  c_function_chooser(f) = is_pair_chooser;

  is_pair_car = s7_make_function(sc, "pair?", g_is_pair_car, 1, 0, false, "pair? optimization");
  c_function_class(is_pair_car) = c_function_class(f);
  is_pair_cdr = s7_make_function(sc, "pair?", g_is_pair_cdr, 1, 0, false, "pair? optimization");
  c_function_class(is_pair_cdr) = c_function_class(f);
  is_pair_cadr = s7_make_function(sc, "pair?", g_is_pair_cadr, 1, 0, false, "pair? optimization");
  c_function_class(is_pair_cadr) = c_function_class(f);


  /* null? */
  f = slot_value(global_slot(make_symbol(sc, "null?")));
  c_function_chooser(f) = is_null_chooser;

  is_null_cdr = s7_make_function(sc, "null?", g_is_null_cdr, 1, 0, false, "null? optimization");
  c_function_class(is_null_cdr) = c_function_class(f);


  /* symbol? */
  f = slot_value(global_slot(make_symbol(sc, "symbol?")));
  c_function_chooser(f) = is_symbol_chooser;


  /* eq? */
  f = slot_value(global_slot(make_symbol(sc, "eq?")));
  c_function_chooser(f) = is_eq_chooser;

  is_eq_car = s7_make_function(sc, "eq?", g_is_eq_car, 2, 0, false, "eq? optimization");
  c_function_class(is_eq_car) = c_function_class(f);
  is_eq_car_q = s7_make_function(sc, "eq?", g_is_eq_car_q, 2, 0, false, "eq? optimization");
  c_function_class(is_eq_car_q) = c_function_class(f);
  is_eq_caar_q = s7_make_function(sc, "eq?", g_is_eq_caar_q, 2, 0, false, "eq? optimization");
  c_function_class(is_eq_caar_q) = c_function_class(f);


  /* zero? */
  f = slot_value(global_slot(make_symbol(sc, "zero?")));
  c_function_chooser(f) = is_zero_chooser;

  is_zero_logand_s_ash_cs = s7_make_function(sc, "zero?", g_is_zero_logand_s_ash_cs, 1, 0, false, "zero? optimization");
  c_function_class(is_zero_logand_s_ash_cs) = c_function_class(f);
#if (!WITH_GMP)
  is_zero_logand_ss = s7_make_function(sc, "zero?", g_is_zero_logand_ss, 1, 0, false, "zero? optimization");
  c_function_class(is_zero_logand_ss) = c_function_class(f);
#endif


  /* memq */
  f = slot_value(global_slot(make_symbol(sc, "memq")));
  c_function_chooser(f) = memq_chooser;
  
  memq_3 = s7_make_function(sc, "memq", g_memq_3, 2, 0, false, "memq optimization");
  c_function_class(memq_3) = c_function_class(f);
  memq_4 = s7_make_function(sc, "memq", g_memq_4, 2, 0, false, "memq optimization");
  c_function_class(memq_4) = c_function_class(f);
  memq_any = s7_make_function(sc, "memq", g_memq_any, 2, 0, false, "memq optimization");
  c_function_class(memq_any) = c_function_class(f);
  memq_car = s7_make_function(sc, "memq", g_memq_car, 2, 0, false, "memq optimization");
  c_function_class(memq_car) = c_function_class(f);


  /* member */
  f = slot_value(global_slot(make_symbol(sc, "member")));
  c_function_chooser(f) = member_chooser;
  
  member_ss = s7_make_function(sc, "member", g_member_ss, 2, 0, false, "member optimization");
  c_function_class(member_ss) = c_function_class(f);
  member_sq = s7_make_function(sc, "member", g_member_sq, 2, 0, false, "member optimization");
  c_function_class(member_sq) = c_function_class(f);
  member_sym_s = s7_make_function(sc, "member", g_member_sym_s, 2, 0, false, "member optimization");
  c_function_class(member_sym_s) = c_function_class(f);
  member_num_s = s7_make_function(sc, "member", g_member_num_s, 2, 0, false, "member optimization");
  c_function_class(member_num_s) = c_function_class(f);
}


static bool optimize_expression(s7_scheme *sc, s7_pointer x, int hop, s7_pointer e);
static int combine_ops(s7_scheme *sc, int op1, s7_pointer e1, s7_pointer e2);

static s7_pointer collect_collisions(s7_scheme *sc, s7_pointer lst, s7_pointer e)
{
  s7_pointer p;
  sc->w = e;
  for (p = lst; is_pair(p); p = cdr(p))
    {
      if ((is_symbol(car(p))) &&
	  (global_slot(car(p)) != sc->NIL))
	sc->w = cons(sc, car(p), sc->w);
      else
	{
	  if ((is_pair(car(p))) &&
	      (is_symbol(caar(p))) &&
	      (global_slot(caar(p)) != sc->NIL))
	    sc->w = cons(sc, caar(p), sc->w);
	}
    }
  return(sc->w);
}


#if WITH_OPTIMIZATION
static bool is_orx_safe(s7_pointer p)
{
  int op;
  op = optimize_data(p);
  return((op == HOP_SAFE_C_S) ||
	 (op == HOP_SAFE_C_C) ||
	 (op == HOP_SAFE_IS_PAIR_S) ||
	 (op == HOP_SAFE_IS_SYMBOL_S) ||
	 (op == HOP_SAFE_NOT_S) ||
	 (op == HOP_SAFE_C_SS) ||
	 (op == HOP_SAFE_C_SC) ||
	 (op == HOP_SAFE_C_CS) ||
	 (op == HOP_SAFE_C_SQ) ||
	 (op == HOP_SAFE_C_opSq) ||
	 (op == HOP_SAFE_C_opSq_opSq));
}
#endif


static int c_s_count(s7_pointer x)
{
  int count = 0;
  s7_pointer p;
  for (p = cdr(x); is_pair(p); p = cdr(p))
    if ((is_pair(car(p))) &&
	(is_optimized(car(p))) &&
	(is_safe_c_s(car(p))))
      count++;
  return(count);
}

static int orx_count(s7_pointer x)
{
  int count = 0;
  s7_pointer p;
  for (p = cdr(x); is_pair(p); p = cdr(p))
    if ((is_pair(car(p))) &&
	(is_optimized(car(p))) &&
	(is_orx_safe(car(p))))
      count++;
  return(count);
}


static bool optimize_function(s7_scheme *sc, s7_pointer x, s7_pointer func, int hop, s7_pointer e)
{
  int pairs = 0, symbols = 0, args = 0, bad_pairs = 0, quotes = 0;
  s7_pointer p;
  bool func_is_safe, func_is_c_function, func_is_closure;

  func_is_safe = is_safe_procedure(func);
  func_is_c_function = is_c_function(func);
  func_is_closure = is_closure(func);

  if (func_is_closure) /* can't depend on ecdr here because it might not be global, or might be redefined locally */
    {
      if ((!is_global(caar(x))) ||
	  (direct_memq(caar(x), e)))
	hop = 0;
      /* this is very tricky!  See s7test for some cases.  Basically, we need to protect a recursive call
       *   of the current function being optimized from being confused with some previous definition
       *   of the same name.
       */
    }

  for (p = cdar(x); is_pair(p); p = cdr(p), args++)
    {
      if (is_pair(car(p)))
	{
	  pairs++;
	  if (!is_checked(car(p)))
	    {
	      bool opt;
	      opt = optimize_expression(sc, p, hop, e);
 	      if (!opt) 
		{
		  bad_pairs++;
		  /* fprintf(stderr, "    optexp: %s: unsafe: %d, opt: %d\n", DISPLAY_80(car(p)), is_unsafe(car(p)), is_optimized(car(p)));  */
		  if ((caar(p) == sc->QUOTE) &&
		      (is_pair(cdar(p))) &&
		      (is_null(cddar(p))))
		    quotes++;
		}
	    }
	  else 
	    {
	      if ((!is_optimized(car(p))) ||
		  (is_unsafe(car(p))))
		{
		  bad_pairs++;
		  /* fprintf(stderr, "    checked before %s: unsafe: %d, opt: %d\n", DISPLAY_80(car(p)), is_unsafe(car(p)), is_optimized(car(p))); */
		  if ((caar(p) == sc->QUOTE) &&
		      (is_pair(cdar(p))) &&
		      (is_null(cddar(p))))
		    quotes++;
		}
	    }
	}
      else
	{
	  if (is_symbol(car(p)))
	    symbols++;
	}
    }
  
  /*
  fprintf(stderr, "%d %d %s %s\n", pairs, bad_pairs, DISPLAY(car(x)), (is_optimized(cadar(x))) ? opt_names[optimize_data(cadar(x))] : "");

  if ((is_not_null(p)) ||
      (!args_match(sc, func, args)))
    fprintf(stderr, "%s null: %d, match: %d\n", DISPLAY_80(car(x)), is_null(p), args_match(sc, func, args));
  */
  /* TODO: args_match is confused by scheme-side pws -- we need is_pws (for scheme-side pws) -- c-side is safe c function already,
   *          then check the getter or setter depending on the context, then branches in the optimizer.
   *          Most of these are just list-ref or vector-ref.
   */

  if ((is_null(p)) &&                /* if not null, dotted list of args? */
      (args_match(sc, func, args)))  /* we have a legit call, at least syntactically */
    {
      /*
      if (is_closure_star(func))
	fprintf(stderr, "closure*: %s\n", DISPLAY_80(car(x)));
      */
      switch (args)
	{
	  /* -------------------------------------------------------------------------------- */
	case 0: 
	  if (func_is_c_function)
	    {
	      set_optimized(car(x));
	      set_optimize_data(car(x), OP_SAFE_C_C);
	      set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
	      if (func_is_safe)
		return(true);
	      set_unsafe(car(x));
	      return(false);
	    }
	  if ((func_is_closure) &&
	      (is_null(closure_args(func)))) /* no rest arg funny business */
	    {
	      set_optimized(car(x));
	      set_unsafe(car(x));
	      if (is_safe_closure(closure_body(func)))
		set_optimize_data(car(x), OP_SAFE_THUNK);
	      else set_optimize_data(car(x), OP_THUNK);
	      ecdr(car(x)) = func;
	      return(false); /* false because currently the C_PP stuff assumes safe procedure calls */
	    }
	  /* call-with-exit and call/cc require an argument */
	  
	  break;
	  
	  /* for a defined function, if the body is completely optimizable, and
	   *   there is no values object or anything that goofs with the stack,
	   *   can it be declared safe? no -- s7_call is a killer.
	   *
	   * if all constant args, and func is 1->1, save actual result?
	   */

	  /* -------------------------------------------------------------------------------- */
	case 1:
	  if (pairs == 0)
	    {
	      if (func_is_c_function)
		{
		  if (func_is_safe)
		    {     
		      set_optimized(car(x));
		      if (symbols == 0)
			set_optimize_data(car(x), OP_SAFE_C_C);
		      else set_optimize_data(car(x), OP_SAFE_C_S);

		      /* we can't simply check is_global here to forego symbol value lookup later because we aren't
		       *    tracking local vars, so the global bit may be on right now, but won't be when
		       *    this code is evaluated.  But memq(sym, e) would catch such cases.
		       *    I think it has already been checked for func, so we only need to look for cadar(x).
		       *    But global symbols are rare, and I don't see a huge savings in the lookup time --
		       *    in callgrind it's about 7/lookup in both cases.
		       */
		      set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
		      return(true);
		    }
		  else
		    {
		      set_optimized(car(x));
		      set_unsafe(car(x));
		      if (symbols == 0)
			set_optimize_data(car(x), OP_C_ALL_G);
		      else 
			{
			  if (c_function_call(func) == g_read)
			    set_optimize_data(car(x), OP_READ_S);
			  else set_optimize_data(car(x), OP_C_S);
			}
		      set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
		      return(false); 
		    }
		}
	      else
		{
		  if (func_is_closure)
		    {
		      set_optimized(car(x));
		      set_unsafe(car(x));
		      if (is_symbol(cadar(x)))
			set_optimize_data(car(x), ((is_safe_closure(closure_body(func))) ? OP_SAFE_CLOSURE_S : OP_CLOSURE_S));
		      else set_optimize_data(car(x), ((is_safe_closure(closure_body(func))) ? OP_SAFE_CLOSURE_C : OP_CLOSURE_C));
		      ecdr(car(x)) = func;
		      fcdr(car(x)) = cadar(x);

		      if ((optimize_data_match(car(x), OP_SAFE_CLOSURE_S)) &&
			  (is_one_liner(closure_body(func))))
			{
			  s7_pointer body;
			  body = car(closure_body(func));
			  if ((is_optimized(body)) &&
			      (optimize_data(body) == HOP_SAFE_C_C) &&
			      (ecdr(body) == vector_ref_ic) &&
			      (is_proper_list(sc, closure_args(func))) &&
			      (cadr(body) == car(closure_args(func))))
			    set_optimize_data(car(x), OP_SAFE_CLOSURE_S_vref); /* structs often use vectors + constant offsets */
			  else set_optimize_data(car(x), OP_SAFE_CLOSURE_S_one);
			}

		      return(false); 
		    }
		  else
		    {
		      if (s7_is_vector(func))
			{
			  set_optimized(car(x));
			  if (is_symbol(cadar(x)))
			    set_optimize_data(car(x), OP_VECTOR_S);
			  else set_optimize_data(car(x), OP_VECTOR_C);
			  ecdr(car(x)) = func;
			  return(true);
			}

		      /* perhaps also closure*?
		       */

		    }
		}
	    }
	  
	  if (pairs == 1)
	    {
	      if (bad_pairs == 0)
		{
		  /* we can optimize whatever the arg involves */
		  if (func_is_c_function)
		    {
		      if (func_is_safe)
			{
			  int op;
			  op = combine_ops(sc, SAFE_C_P, car(x), cadar(x));
			  if (op != SAFE_C_P)
			    {
			      set_optimized(car(x));
			      set_optimize_data(car(x), op);
			      set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
			    }
			}
		    }
		  else
		    {
		      if ((func_is_closure) &&
			  ((is_safe_c_s(cadar(x))) ||
			   (optimize_data_match(cadar(x), OP_SAFE_C_C))))
			{
			  set_optimized(car(x));
			  set_unsafe(car(x));
			  if (is_safe_c_s(cadar(x)))
			    {
			      if (is_safe_closure(closure_body(func)))
				set_optimize_data(car(x), OP_SAFE_CLOSURE_opSq);
			      else
				{
				  if (c_call(cadar(x)) == g_cdr)
				    set_optimize_data(car(x), OP_CLOSURE_CDR_S);
				  else set_optimize_data(car(x), OP_CLOSURE_opSq);
				}
			      fcdr(car(x)) = cadr(cadar(x));
			    }
			  else 
			    {
			      set_optimize_data(car(x), ((is_safe_closure(closure_body(func)) ? OP_SAFE_CLOSURE_opCq : OP_CLOSURE_opCq)));
			      fcdr(car(x)) = cadar(x);
			    }
			  ecdr(car(x)) = func;
			  return(false); 
			}
		      /* fprintf(stderr, "%s %s\n", DISPLAY(car(x)), (is_optimized(cadar(x))) ? opt_names[optimize_data(cadar(x))] : ""); */
		    }
		}
	      else
		{
		  if (quotes == 1)
		    {
		      if (func_is_c_function)
			{
			  if (func_is_safe)
			    {
			      set_optimized(car(x));
			      set_optimize_data(car(x), OP_SAFE_C_Q);
			      set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
			    }
			  else
			    {
			      set_optimized(car(x));
			      set_unsafe(car(x));
			      set_optimize_data(car(x), OP_C_ALL_G);
			      set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
			      return(false); 
			    }
			}
		      else
			{
			  if (func_is_closure)
			    {
			      set_optimized(car(x));
			      set_unsafe(car(x));
			      set_optimize_data(car(x), ((is_safe_closure(closure_body(func))) ? OP_SAFE_CLOSURE_Q : OP_CLOSURE_Q));
			      ecdr(car(x)) = func;
			      return(false); 
			    }
			}
		    }
		  else
		    {
		      if (func_is_c_function)
			{
			  if (func_is_safe)
			    {
			      /* fprintf(stderr, "%s: safe of bad 1: %s\n", opt_names[optimize_data(cadar(x))], DISPLAY_80(car(x))); */
			      
			      if (is_optimized(cadar(x))) 
				{
				  if (optimize_data_match(cadar(x), OP_SAFE_CLOSURE_SS))
				    {
				      set_optimized(car(x));
				      set_unsafe(car(x));
				      set_optimize_data(car(x), OP_SAFE_C_opSAFE_CLOSURE_SSq);
				      set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
				      return(false); 
				    }
				  
				  if (optimize_data_match(cadar(x), OP_SAFE_CLOSURE_opSq_S))
				    {
				      set_optimized(car(x));
				      set_unsafe(car(x));
				      set_optimize_data(car(x), OP_SAFE_C_opSAFE_CLOSURE_opSq_Sq);
				      set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
				      return(false); 
				    }
#if 0
				  if (optimize_data_match(cadar(x), OP_UNKNOWN_S))
				    {
				      fprintf(stderr, "1 arg, it's unknown\n");
				    }
				  /* clm2xen wants to try to handle (oscil (oscs i)) and similar cases,
				   *   even though they might be problematic in general.  If an outside
				   *   chooser returns anything other than func here, we'll assume
				   *   it is OP_SAFE_C_C and safe.
				   */
				  set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
				  if (ecdr(car(x)) != func)
				    {
				      set_optimized(car(x));
				      set_optimize_data(car(x), OP_SAFE_C_C);
				      return(true);
				    }
				  /* but... this requires that all the outside choosers know how to handle
				   *   unsafe args like (values ...)!  
				   */
#endif
				}
			    }
			  else
			    {
			      if ((is_pair(cadar(x))) &&
				  (is_lambda(sc, caadar(x))))
				{
				  set_optimized(car(x));
				  set_unsafe(car(x));
				  if (c_function_call(func) == g_call_with_exit)
				    set_optimize_data(car(x), OP_CALL_WITH_EXIT);
				  else set_optimize_data(car(x), OP_C_L);
				  set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
				  return(false);
				}
			    }
			}
		    }
		}
	    }

	  if ((!is_optimized(car(x))) &&
	      (pairs == 1) &&
	      /* (bad_pairs == 1) && */
	      (quotes == 0) &&
	      (func_is_c_function))
	    {
	      if (func_is_safe)
		{
		  set_optimized(car(x));
		  if (bad_pairs == 0)
		    {
		      if (is_orx_safe(cadar(x)))
			set_optimize_data(car(x), OP_SAFE_C_ALL_X);
		      else set_optimize_data(car(x), OP_SAFE_C_Z);
		      set_c_function(car(x), c_function_chooser(func)(sc, func, 1, car(x))); /* was func */
		      return(true);
		    }
		  else
		    {
		      set_unsafe(car(x));
		      set_optimize_data(car(x), OP_SAFE_C_P);
		      set_c_function(car(x), func);
		      return(false);
		    }
		}
	      else
		{
		  if ((is_optimized(cadar(x))) &&
		      (c_function_has_rest_arg(func)))
		    {
		      set_optimized(car(x));
		      set_unsafe(car(x));
		      set_optimize_data(car(x), OP_C_P);
		      set_c_function(car(x), func);
		      return(false);
		    }
		}
	    }
	  /* -- left now are unsafe callers -- arg might be a symbol or constant or quoted thing -- should these be optimized?
	   * closure:  (quot (car (car args))) (1- (1- x)) 
	   * closure*: (fill-in score)
	   * unknown arg: (alphanumeric? (line (+ i 1)))
	   */

	  /*
	  if (!is_optimized(car(x)))
	    fprintf(stderr, "1 case: %s\n", DISPLAY_80(car(x)));
	  */
	  break; 


	  /* -------------------------------------------------------------------------------- */
	case 2:

	  /* fprintf(stderr, "quote: %d, pair: %d (bad: %d), sym: %d in %s\n", quotes, pairs, bad_pairs, symbols, s7_object_to_c_string(sc, car(x)));  */
	  if (pairs == 0)
	    {
	      if (func_is_c_function)
		{
		  if ((func_is_safe) ||
		      (c_function_call(func) == g_member) || /* the unsafe case has a 3rd arg */
		      (c_function_call(func) == g_assoc))
		    {
		      if (symbols == 0)
			{
			  set_optimize_data(car(x), OP_SAFE_C_C);
			}
		      else
			{
			  if (symbols == 2)
			    {
			      set_optimize_data(car(x), OP_SAFE_C_SS); /* these two symbols are almost never the same, (sqrt (+ (* x x) (* y y))) */
			    }
			  else
			    {
			      if (is_symbol(cadar(x)))
				{
				  set_optimize_data(car(x), OP_SAFE_C_SC);
				}
			      else 
				{
				  set_optimize_data(car(x), OP_SAFE_C_CS);
				}
			    }
			}
		      set_optimized(car(x));
		      set_c_function(car(x), c_function_chooser(func)(sc, func, 2, car(x)));
		      return(true);
		    }
		  else
		    {
		      set_optimized(car(x));
		      set_unsafe(car(x));
		      set_optimize_data(car(x), OP_C_ALL_G);
		      set_c_function(car(x), c_function_chooser(func)(sc, func, 2, car(x)));
		      return(false); 
		    }
		}
	      else
		{
		  if ((func_is_closure) &&
		      (s7_list_length(sc, closure_args(func)) == 2))
		    {
		      set_optimized(car(x));
		      set_unsafe(car(x));
		      if (symbols == 2)
			set_optimize_data(car(x), ((is_safe_closure(closure_body(func)) ? OP_SAFE_CLOSURE_SS : OP_CLOSURE_SS)));
		      else
			{
			  if (symbols == 0)
			    set_optimize_data(car(x), ((is_safe_closure(closure_body(func)) ? OP_SAFE_CLOSURE_CC : OP_CLOSURE_CC)));
			  else
			    {
			      if (is_symbol(cadar(x)))
				set_optimize_data(car(x), ((is_safe_closure(closure_body(func)) ? OP_SAFE_CLOSURE_SC : OP_CLOSURE_SC)));
			      else set_optimize_data(car(x), ((is_safe_closure(closure_body(func)) ? OP_SAFE_CLOSURE_CS : OP_CLOSURE_CS)));
			    }
			}
		      ecdr(car(x)) = func;
		      fcdr(car(x)) = caddar(x);
		      return(false);
		    }
		}
	    }
	  else /* pairs != 0 */
	    {
	      if (bad_pairs == 0)
		{
		  if (pairs == 2)
		    {
		      if (func_is_c_function)
			{
			  if ((func_is_safe) ||
			      (c_function_call(func) == g_member) ||
			      (c_function_call(func) == g_assoc))
			    {
			      int op;
			      op = combine_ops(sc, SAFE_C_PP, cadar(x), caddar(x));
			      if (op != SAFE_C_PP)
				{
				  set_optimized(car(x));
				  set_optimize_data(car(x), op);
				  set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));

				  if (op == OP_SAFE_C_ZZ)
				    {
#if 0
				      fprintf(stderr, "pp->zz: %s %s %s\n", opt_names[optimize_data(cadar(x))], opt_names[optimize_data(caddar(x))], DISPLAY(car(x)));
				      fprintf(stderr, " ecdrs: %p %p\n", ecdr(cadar(x)), ecdr(caddar(x)));
#endif
				      if ((is_orx_safe(cadar(x))) &&
					  (is_orx_safe(caddar(x))))
					set_optimize_data(car(x), OP_SAFE_C_ALL_X);
				    }

				}
			      else
				{
				  /* fprintf(stderr, "safe c_pp?? %s\n", DISPLAY_80(car(x))); */

				  set_optimized(car(x));
				  if ((optimize_data_match(cadar(x), OP_SAFE_C_C)) &&
				      (optimize_data_match(caddar(x), OP_SAFE_C_C)))
				    set_optimize_data(car(x), OP_SAFE_C_DD);
				  else 
				    {
				      if ((is_orx_safe(cadar(x))) &&
					  (is_orx_safe(caddar(x))))
					set_optimize_data(car(x), OP_SAFE_C_ALL_X);
				      else
					{
					  /*
					    fprintf(stderr, "zz: %s %s %s\n", opt_names[optimize_data(cadar(x))], opt_names[optimize_data(caddar(x))], DISPLAY(car(x)));
					    fprintf(stderr, " ecdrs: %p %p\n", ecdr(cadar(x)), ecdr(caddar(x)));
					  */
					  set_optimize_data(car(x), OP_SAFE_C_ZZ);
					}
				    }
				  set_c_function(car(x), c_function_chooser(func)(sc, func, 2, car(x))); /* was func */
				}
			    }
			}
		      else
			{
			  if ((func_is_closure) &&
			      (is_safe_c_s(cadar(x))) &&
			      (is_safe_c_s(caddar(x))))
			    {
			      set_optimized(car(x));
			      set_unsafe(car(x));
			      set_optimize_data(car(x), ((is_safe_closure(closure_body(func)) ? OP_SAFE_CLOSURE_opSq_opSq : OP_CLOSURE_opSq_opSq)));
			      ecdr(car(x)) = func;
			      fcdr(car(x)) = cadr(caddar(x));
			      return(false);
			    }
			  else
			    {
			      /* fprintf(stderr, "2 case: %s\n", DISPLAY_80(car(x))); */
			      /* opSq_opSSq? (fneq (car b) (vct-ref a i)
			       */
			    }
			}
		    }
		  if (pairs == 1)
		    {
		      if (func_is_c_function)
			{
			  if ((func_is_safe) ||
			      (c_function_call(func) == g_member) ||
			      (c_function_call(func) == g_assoc))
			    {
			      int orig_op, op;

			      if (is_pair(cadar(x)))
				{
				  if (is_symbol(caddar(x)))
				    orig_op = SAFE_C_PS;
				  else orig_op = SAFE_C_PC;
				  op = combine_ops(sc, orig_op, car(x), cadar(x));
				}
			      else 
				{
				  if (is_symbol(cadar(x)))
				    orig_op = SAFE_C_SP;
				  else orig_op = SAFE_C_CP;
				  op = combine_ops(sc, orig_op, car(x), caddar(x));
				}
				  
			      if (op != orig_op)
				{
				  set_optimized(car(x));
				  set_optimize_data(car(x), op);
				  set_c_function(car(x), c_function_chooser(func)(sc, func, 2, car(x)));
				}
			    }
			  else
			    {
			      if ((symbols == 1) &&
				  (is_symbol(cadar(x))) &&
				  (is_safe_c_s(caddar(x))))
				{
				  set_optimized(car(x));
				  set_unsafe(car(x));
				  set_optimize_data(car(x), OP_C_S_opSq);
				  set_c_function(car(x), c_function_chooser(func)(sc, func, 2, car(x))); /* added */
				  return(false);
				}
			    }
			}
		      else
			{
			  if ((func_is_closure) &&
			      (s7_list_length(sc, closure_args(func)) == 2))
			    {
			      if ((is_symbol(caddar(x))) &&
				  (is_safe_c_s(cadar(x))))
				{
				  set_optimized(car(x));
				  set_unsafe(car(x));
				  set_optimize_data(car(x), ((is_safe_closure(closure_body(func)) ? OP_SAFE_CLOSURE_opSq_S : OP_CLOSURE_opSq_S)));
				  ecdr(car(x)) = func;
				  fcdr(car(x)) = caddar(x);
				  return(false); 
				}
			      if ((is_symbol(cadar(x))) &&
				  (is_safe_c_s(caddar(x))))
				{
				  set_optimized(car(x));
				  set_unsafe(car(x));
				  set_optimize_data(car(x), ((is_safe_closure(closure_body(func)) ? OP_SAFE_CLOSURE_S_opSq : OP_CLOSURE_S_opSq)));
				  ecdr(car(x)) = func;
				  return(false); 
				}
			    }
			}
		    }
		}
	      else /* pairs != 0 and bad_pairs != 0 */
		{
		  if ((bad_pairs == 1) && (quotes == 1))
		    {
		      if (func_is_c_function)
			{
			  if ((func_is_safe) ||
			      (c_function_call(func) == g_member) ||
			      (c_function_call(func) == g_assoc))
			    {
			      if  (symbols == 1)
				{
				  set_optimized(car(x));
				  if (is_symbol(cadar(x)))
				    set_optimize_data(car(x), OP_SAFE_C_SQ);
				  else set_optimize_data(car(x), OP_SAFE_C_QS);
				  set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
				}
			      else
				{
				  if (pairs == 2)
				    {
				      int orig_op, op;
				      if (car(cadar(x)) == sc->QUOTE)
					orig_op = SAFE_C_QP;
				      else orig_op = SAFE_C_PQ;
				      op = combine_ops(sc, orig_op, car(x), cadar(x));
				      
				      if (op != orig_op)
					{
					  set_optimized(car(x));
					  set_optimize_data(car(x), op);
					  set_c_function(car(x), c_function_chooser(func)(sc, func, 2, car(x)));
					}
				    }
				}
			    }
			  else
			    {
			      if (pairs == 1)
				{
				  set_optimized(car(x));
				  set_unsafe(car(x));
				  set_optimize_data(car(x), OP_C_ALL_G);
				  set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
				  return(false); 
				}
			    }
			}
		    }
		  else
		    {
		      if (quotes == 2)
			{
			  if (func_is_c_function)
			    {
			      if ((func_is_safe) ||
				  (c_function_call(func) == g_member) ||
				  (c_function_call(func) == g_assoc))
				{
				  set_optimized(car(x));
				  set_optimize_data(car(x), OP_SAFE_C_QQ);
				  set_c_function(car(x), c_function_chooser(func)(sc, func, 2, car(x)));
				}
			      else
				{
				  set_optimized(car(x));
				  set_unsafe(car(x));
				  set_optimize_data(car(x), OP_C_ALL_G);
				  set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
				  return(false); 
				}
			    }
			}
		      else
			{
			  if ((bad_pairs == 1) &&
			      (is_pair(cadar(x))) &&
			      (is_lambda(sc, caadar(x))))
			    {
			      if (func_is_c_function)
				{
				  if (is_symbol(caddar(x)))
				    {
				      set_optimized(car(x));
				      set_unsafe(car(x));
				      set_optimize_data(car(x), OP_C_LS);
				      set_c_function(car(x), c_function_chooser(func)(sc, func, 2, car(x)));
				      if (c_call(car(x)) == g_for_each_1)
					set_optimize_data(car(x), OP_C_FOR_EACH_LS);
				      else
					{
					  if (c_call(car(x)) == g_for_each_2)
					    set_optimize_data(car(x), OP_C_FOR_EACH_LS_2);
					}
				      return(false);
				    }
				  if ((pairs == 2) &&
				      (is_optimized(caddar(x))) &&
				      (is_safe_c_s(caddar(x))))
				    {
				      set_optimized(car(x));
				      set_unsafe(car(x));
				      set_optimize_data(car(x), OP_C_L_opSq);
				      set_c_function(car(x), c_function_chooser(func)(sc, func, 2, car(x)));
				      if (c_call(car(x)) == g_for_each_2)
					set_optimize_data(car(x), OP_C_FOR_EACH_L_opSq);
				      return(false);
				    }
				}
			    }
			  else
			    {
			      if ((bad_pairs == 2) &&
				  (quotes == 0) &&
				  (is_pair(cadar(x))) &&
				  (is_lambda(sc, caadar(x))) &&
				  (is_pair(caddar(x))) &&
				  (is_lambda(sc, car(caddar(x)))))
				{
				  if (func_is_c_function)
				    {
				      set_optimized(car(x));
				      set_unsafe(car(x));
				      set_optimize_data(car(x), OP_C_LL);
				      set_c_function(car(x), c_function_chooser(func)(sc, func, 2, car(x)));
				      return(false);
				    }
				}
			    }
			}
		    }
		}
	    }
	  /*
	  if (!is_optimized(car(x)))
	    fprintf(stderr, "%s not optimized (%d %d)\n", DISPLAY_80(car(x)), pairs, bad_pairs);
	  */
	  if ((!is_optimized(car(x))) &&
	      (pairs == 1) &&
	      /* (bad_pairs == 1) && */
	      (quotes == 0) &&
	      (symbols == 1) &&
	      (func_is_c_function) &&
	      ((func_is_safe) ||
	       (c_function_call(func) == g_member) ||
	       (c_function_call(func) == g_assoc)))
	    {
	      set_optimized(car(x));
	      if (bad_pairs == 0)
		{
		  if (is_symbol(cadar(x)))
		    set_optimize_data(car(x), OP_SAFE_C_SZ);
		  else set_optimize_data(car(x), OP_SAFE_C_ZS);
		  set_c_function(car(x), c_function_chooser(func)(sc, func, 2, car(x)));
		  return(true);
		}
	      else
		{
		  set_unsafe(car(x));
		  if (is_symbol(cadar(x)))
		    set_optimize_data(car(x), OP_SAFE_C_SP);
		  else set_optimize_data(car(x), OP_SAFE_C_PS);
		  set_c_function(car(x), func);
		  return(false);
		}
	    }

	  if ((!is_optimized(car(x))) &&
	      (pairs == 1) &&
	      /* (bad_pairs == 1) && */
	      (quotes == 0) &&
	      (symbols == 0) &&
	      (func_is_c_function) &&
	      ((func_is_safe) ||
	       (c_function_call(func) == g_member) ||
	       (c_function_call(func) == g_assoc)))
	    {
	      set_optimized(car(x));
	      if (bad_pairs == 0)
		{
		  if (is_symbol(cadar(x)))
		    set_optimize_data(car(x), OP_SAFE_C_CZ);
		  else set_optimize_data(car(x), OP_SAFE_C_ZC);
		  set_c_function(car(x), c_function_chooser(func)(sc, func, 2, car(x)));
		  return(true);
		}
	      else
		{
		  set_unsafe(car(x));
		  if (!is_pair(cadar(x)))
		    set_optimize_data(car(x), OP_SAFE_C_CP);
		  else set_optimize_data(car(x), OP_SAFE_C_PC);
		  set_c_function(car(x), func);
		  return(false);
		}
	    }
	  /*
	  fprintf(stderr, "%d %d %d %d %d\n", 
		  !is_optimized(car(x)),
		  (pairs == 2),
		  (bad_pairs == 0),
		  (func_is_c_function),
		  (func_is_safe));
	  */
	  
	  if ((!is_optimized(car(x))) &&
	      (pairs == 2) &&
	      (func_is_c_function) &&
	      ((func_is_safe) ||
	       (c_function_call(func) == g_member) ||
	       (c_function_call(func) == g_assoc)))
	    {
	      if ((bad_pairs == 1) &&
		  (is_optimized(cadar(x))) &&
		  (is_safe_c_s(cadar(x))))
		{
		  /* unsafe func here won't work unless we check that later and make the new arg list (for {list} etc)
		   *   (and it has to be the last pair else the unknown_s stuff can mess up)
		   */
		  set_unsafe(car(x));
		  set_optimized(car(x));
		  set_optimize_data(car(x), OP_SAFE_C_opSq_P);
		  set_c_function(car(x), func);
		  /* not chooser here and below because the choosers don't know about values */
		  return(false);
		}
	      else
		{
		  if ((is_optimized(cadar(x))) &&
		      (is_optimized(caddar(x))) &&
		      (c_function_has_rest_arg(func))) /* so I don't have to worry about values here */
		    {
		      set_optimized(car(x));
		      set_unsafe(car(x));
		      set_optimize_data(car(x), OP_SAFE_C_PP);
		      /* set_c_function(car(x), func); */
		      set_c_function(car(x), c_function_chooser(func)(sc, func, 2, car(x))); 
		      return(false);
		    }
		}
	    }
	  
	  /*
	    also g_member/assoc with safe closure 3rd arg is safe
	    closure* all arg choices (CS SS SOX in index)
	    (fneq (car b) (vct-ref a i))
	    SP case in all: (try-peval example1 (list ...))
	    QS case: (reduce-global '+ const), SQ case: (gen-type boolean-con '())
	    unopt arg cases: (memq (const-value (cadr exp)) '(#f ()))
	    set-car! and set-cdr!?  (set-cdr! (cdr elem) info)
	  */
	  /*
	  if (!is_optimized(car(x)))
	    fprintf(stderr, "2 case: %s %d %d\n", DISPLAY_80(car(x)), is_optimized(cadar(x)), is_optimized(caddar(x)));
	  */
	  break;
	  
	  /* -------------------------------------------------------------------------------- */
	case 3:
	  /* here quotes in safe functions are somewhat rare: list-ref, format, etc */
	  if (pairs == 0)
	    {
	      if (func_is_c_function)
		{
		  if (func_is_safe)
		    {
		      set_optimized(car(x));
		      if (symbols == 0)
			set_optimize_data(car(x), OP_SAFE_C_C);
		      else 
			{
			  if (symbols == 3)
			    set_optimize_data(car(x), OP_SAFE_C_SSS);
			  else set_optimize_data(car(x), OP_SAFE_C_XXX);
			}
		      set_c_function(car(x), c_function_chooser(func)(sc, func, 3, car(x)));
		    }
		}
	      else
		{
		  if ((func_is_closure) &&
		      (s7_list_length(sc, closure_args(func)) == 3) &&
		      (symbols == 3))
		    {
		      set_optimized(car(x));
		      set_unsafe(car(x));
		      set_optimize_data(car(x), ((is_safe_closure(closure_body(func)) ? OP_SAFE_CLOSURE_SSS : OP_CLOSURE_SSS)));
		      ecdr(car(x)) = func;
		      fcdr(car(x)) = cadddr(car(x));
		      return(false);
		    }
		}
	    }

	  if ((pairs == 1) &&
	      (bad_pairs == 0) &&
	      (func_is_c_function))
	    {
	      if (func_is_safe)
		{
		  set_optimized(car(x));		  
		  if (is_pair(cadar(x)))
		    {
		      if (optimize_data_match(cadar(x), OP_SAFE_C_C))
			set_optimize_data(car(x), OP_SAFE_C_DXX);
		      else
			{/* this is slower
			  if (is_orx_safe(cadar(x)))
			    set_optimize_data(car(x), OP_SAFE_C_ALL_X);
			  else 
			 */
			    set_optimize_data(car(x), OP_SAFE_C_ZXX);
			}
		    }
		  else
		    {
		      if (is_pair(caddar(x)))
			{
			  if (optimize_data_match(caddar(x), OP_SAFE_C_C))
			    set_optimize_data(car(x), OP_SAFE_C_XDX);
			  else
			    {
			      /*
			      if (is_orx_safe(caddar(x)))
				set_optimize_data(car(x), OP_SAFE_C_ALL_X);
			      else 
			      */
			      set_optimize_data(car(x), OP_SAFE_C_XZX);
			    }
			}
		      else 
			{
			  if (optimize_data_match(cadddar(x), OP_SAFE_C_C))
			    set_optimize_data(car(x), OP_SAFE_C_XXD);
			  else
			    {
			      /*
			      if (is_orx_safe(cadddar(x)))
				set_optimize_data(car(x), OP_SAFE_C_ALL_X);
			      else 
			      */
			      set_optimize_data(car(x), OP_SAFE_C_XXZ);
			    }
			}
		    }
		  set_c_function(car(x), c_function_chooser(func)(sc, func, 3, car(x)));
		}
	      else
		{
		  if ((symbols == 0) &&
		      (is_pair(cadar(x))) &&
		      (is_optimized(cadar(x))) &&
		      (is_safe_c_s(cadar(x))))
		    {
		      set_optimized(car(x));
		      set_unsafe(car(x));
		      set_optimize_data(car(x), OP_C_opSq_CC);
		      set_c_function(car(x), c_function_chooser(func)(sc, func, 3, car(x)));
		      return(false);
		    }
		}
	    }

	  if ((pairs == 2) &&
	      (bad_pairs == 0) &&
	      (func_is_c_function) &&
	      (func_is_safe))
	    {
	      set_optimized(car(x));		  
	      if (!is_pair(cadddr(car(x))))
		{
		  if ((optimize_data_match(cadar(x), OP_SAFE_C_C)) &&
		      (optimize_data_match(caddar(x), OP_SAFE_C_C)))
		    set_optimize_data(car(x), OP_SAFE_C_DDX);
		  else
		    {
		      /*
		      if ((is_orx_safe(cadar(x))) &&
			  (is_orx_safe(caddar(x))))
			set_optimize_data(car(x), OP_SAFE_C_ALL_X);
		      else 
		      */
		      set_optimize_data(car(x), OP_SAFE_C_ZZX);
		    }
		}
	      else
		{
		  if (!is_pair(caddar(x)))
		    {
		      if ((optimize_data_match(cadar(x), OP_SAFE_C_C)) &&
			  (optimize_data_match(cadddar(x), OP_SAFE_C_C)))
			set_optimize_data(car(x), OP_SAFE_C_DXD);
		      else
			{
			  /*
			  if ((is_orx_safe(cadar(x))) &&
			      (is_orx_safe(cadddar(x))))
			    set_optimize_data(car(x), OP_SAFE_C_ALL_X);
			  else 
			  */
			  set_optimize_data(car(x), OP_SAFE_C_ZXZ);
			}
		    }
		  else 
		    {
		      if ((optimize_data_match(caddar(x), OP_SAFE_C_C)) &&
			  (optimize_data_match(cadddar(x), OP_SAFE_C_C)))
			set_optimize_data(car(x), OP_SAFE_C_XDD);
		      else
			{
			  /*
			  if ((is_orx_safe(caddar(x))) &&
			      (is_orx_safe(cadddar(x))))
			    set_optimize_data(car(x), OP_SAFE_C_ALL_X);
			  else 
			  */
			  set_optimize_data(car(x), OP_SAFE_C_XZZ);
			}
		    }
		}
	      set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
	    }

	  if ((pairs == 3) &&
	      (bad_pairs == 0) &&
	      (func_is_c_function) &&
	      (func_is_safe))
	    {
	      set_optimized(car(x));		  
	      if ((optimize_data_match(cadar(x), OP_SAFE_C_C)) &&
		  (optimize_data_match(caddar(x), OP_SAFE_C_C)) &&
		  (optimize_data_match(cadddar(x), OP_SAFE_C_C)))
		set_optimize_data(car(x), OP_SAFE_C_DDD);
	      else
		{
		  /*
		  if ((is_orx_safe(cadar(x))) && 
		      (is_orx_safe(caddar(x))) &&
		      (is_orx_safe(cadddar(x))))
		    set_optimize_data(car(x), OP_SAFE_C_ALL_X);
		  else 
		  */
		  set_optimize_data(car(x), OP_SAFE_C_ZZZ);
		}
	      set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
	      /*
	      if (optimize_data_match(car(x), OP_SAFE_C_ZZZ))
		fprintf(stderr, "use zzz: %s\n", DISPLAY_80(car(x)));
	      */
	    }


	  /* (define (hi) (catch #t (lambda () 1) (lambda args 2))) */

	  if ((bad_pairs == 2) &&
	      (symbols == 0) &&
	      (quotes == 0) &&
	      (is_pair(caddar(x))) &&
	      (is_lambda(sc, car(caddar(x)))) &&
	      (is_pair(car(cdddr(car(x))))) &&
	      (is_lambda(sc, car(car(cdddr(car(x)))))))
	    {
	      if (func_is_c_function)
		{
		  set_optimized(car(x));
		  set_unsafe(car(x));
		  if (c_function_call(func) == g_catch)
		    set_optimize_data(car(x), OP_C_CATCH);
		  else set_optimize_data(car(x), OP_C_CLL);
		  set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
		  return(false);
		}
	    }

	  if (bad_pairs > quotes) return(false);

	  if ((!is_optimized(car(x))) &&
	      (pairs == (quotes + c_s_count(car(x)))))
	    {
	      if (func_is_c_function)
		{
		  if (func_is_safe)
		    {
		      set_optimized(car(x));
		      set_optimize_data(car(x), OP_SAFE_C_ALL_G);
		      set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
		    }
		  else
		    {
		      set_optimized(car(x));
		      set_optimize_data(car(x), OP_C_ALL_G);
		      set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
		      if (optimize_data(car(x)) != HOP_SAFE_C_C) /* did chooser fix it up? */
			{
			  set_unsafe(car(x));
			  return(false); 
			}
		      else return(true);
		    }
		}
	      else
		{
		  if ((func_is_closure) &&
		      (s7_list_length(sc, closure_args(func)) == 3))
		    {
		      set_optimized(car(x));
		      set_unsafe(car(x));
		      set_optimize_data(car(x), ((is_safe_closure(closure_body(func)) ? OP_SAFE_CLOSURE_ALL_G : OP_CLOSURE_ALL_G)));
		      ecdr(car(x)) = func;
		      return(false); 
		    }
		}
	    }

	  if ((!is_optimized(car(x))) &&
	      (func_is_c_function) &&
	      (func_is_safe) &&
	      (pairs == (quotes + orx_count(car(x)))))
	    {
	      set_optimized(car(x));
	      set_optimize_data(car(x), OP_SAFE_C_ALL_X);
	      set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
	    }

#if 0
	  if ((!is_optimized(car(x))) &&
	      (pairs == 3) &&
	      (func_is_c_function) &&
	      (func_is_safe) &&
	      (is_optimized(cadar(x))) &&
	      (is_optimized(caddar(x))) &&
	      (is_optimized(cadddar(x))) &&
	      (c_function_has_rest_arg(func))) /* so I don't have to worry about values here */
	    {
	      fprintf(stderr, "3 case: %s\n", DISPLAY_80(car(x)));
#if 0
		      set_optimized(car(x));
		      set_unsafe(car(x));
		      set_optimize_data(car(x), OP_SAFE_C_PP);
		      /* set_c_function(car(x), func); */
		      set_c_function(car(x), c_function_chooser(func)(sc, func, 2, car(x))); 
		      return(false);
#endif
	    }
#endif
	  
	  /* closure*: (checked-substring dline 0 epos)
	   * (safe-)closure(?): (matrix-ref a mm (vector-ref l1 k))  (matrix-ref a (+ i 1) 0) (href harr (+ x 3) (+ y 1))
	   * unsafe: (list 'implies (- n 1) n)
	   */
	  /*
	  if (!is_optimized(car(x)))
	    fprintf(stderr, "3 case: %s\n", DISPLAY_80(car(x)));
	  */

	  break;
	  
	  /* -------------------------------------------------------------------------------- */
	default:
	  if (bad_pairs > quotes) return(false);
	  if (func_is_c_function)
	    {
	      if (func_is_safe)
		{
		  if (pairs == 0)
		    {
		      if (symbols == 0)
			{
			  set_optimized(car(x));
			  set_optimize_data(car(x), OP_SAFE_C_C);
			  set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
			}
		      else
			{
			  if ((symbols == args) &&
			      (args < GC_TRIGGER_SIZE))
			    {
			      set_optimized(car(x));
			      set_optimize_data(car(x), OP_SAFE_C_ALL_S);
			      set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
			    }
			  /* all_g case handled below */
			}
		    }
		  /* all_g case handled below */
		}
	      /* all_g case handled below */
	    }
	  else
	    {
	      if ((func_is_closure) &&
		  (pairs == 0) &&
		  ((symbols == args) || (symbols == 0)) &&
		  (args < GC_TRIGGER_SIZE))
		{
		  set_optimized(car(x));
		  set_unsafe(car(x));
		  if (symbols == 0)
		    set_optimize_data(car(x), ((is_safe_closure(closure_body(func)) ? OP_SAFE_CLOSURE_ALL_C : OP_CLOSURE_ALL_C)));
		  else set_optimize_data(car(x), ((is_safe_closure(closure_body(func)) ? OP_SAFE_CLOSURE_ALL_S : OP_CLOSURE_ALL_S)));
		  ecdr(car(x)) = func;
		  return(false);
		}
	    }

	  if ((!is_optimized(car(x))) &&
	      (args < GC_TRIGGER_SIZE) &&
	      (pairs == (quotes + c_s_count(car(x)))))
	    {
	      if (func_is_c_function)
		{
		  if (func_is_safe)
		    {
		      set_optimized(car(x));
		      set_optimize_data(car(x), OP_SAFE_C_ALL_G);
		      set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
		    }
		  else
		    {
		      set_optimized(car(x));
		      set_unsafe(car(x));
		      set_optimize_data(car(x), OP_C_ALL_G);
		      set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
		      return(false); 
		    }
		}
	      else
		{
		  /* none of the closures can skip the ecdr=func check via hop
		   *    unless they are globally defined.
		   */
		  if ((func_is_closure) &&
		      (s7_list_length(sc, closure_args(func)) == args))
		    {
		      set_optimized(car(x));
		      set_unsafe(car(x));
		      set_optimize_data(car(x), ((is_safe_closure(closure_body(func)) ? OP_SAFE_CLOSURE_ALL_G : OP_CLOSURE_ALL_G)));
		      ecdr(car(x)) = func;
		      return(false); 
		    }
		}
	    }

	  if ((!is_optimized(car(x))) &&
	      (func_is_c_function) &&
	      (func_is_safe) &&
	      (args < GC_TRIGGER_SIZE) &&
	      (pairs == (quotes + orx_count(car(x)))))
	    {
	      set_optimized(car(x));
	      set_optimize_data(car(x), OP_SAFE_C_ALL_X);
	      set_c_function(car(x), c_function_chooser(func)(sc, func, args, car(x)));
	    }

	  if ((!is_optimized(car(x))) &&
	      (bad_pairs == 0) &&
	      (func_is_c_function) &&
	      (func_is_safe))
	    {
	      /* give caller's optimizer a chance at it */
	      s7_pointer new_func;
	      set_optimize_data(car(x), 0);
	      new_func = c_function_chooser(func)(sc, func, args, car(x));
	      if (new_func != func)
		{
		  set_optimized(car(x));
		  if (optimize_data(car(x)) == 0)
		    set_optimize_data(car(x), OP_SAFE_C_C);
		  set_c_function(car(x), new_func);
		}
	    }

	  /* closure*: (offset-channel off beg dur index chn)
	   * closure with opt args: (lint-walk-body name line-number head (cddr form) (append vars env))
	   * why this: (apply format #t (string-append "  ~A (line ~D): " str "~%") name line args)
	   * why this: (+ -1 b g (- (* a g))) (* 0.5 index (- r r1) (+ one (cos modphase)))
	   */
	  /*
	  if (!is_optimized(car(x)))
	    fprintf(stderr, "%d case: %s\n", s7_list_length(sc, cdar(x)), DISPLAY_80(car(x)));
	  */
	  break;
	}
      return(is_optimized(car(x)));
    }
  return(false);
}


static s7_pointer check_for_global_set(s7_scheme *sc, s7_pointer tree, s7_pointer e)
{
  if (is_checked_1(tree))
    return(e);
  set_checked_1(tree);
  if ((is_pair(tree)) &&
      (car(tree) != sc->QUOTE))
    {
      if (car(tree) == sc->SET)
	{
	  if ((is_pair(cdr(tree))) &&
	      (is_symbol(cadr(tree))) &&
	      (is_global(cadr(tree))))
	    return(cons(sc, cadr(tree), e));
	}
      if (is_pair(car(tree)))
	e = check_for_global_set(sc, car(tree), e);
      return(check_for_global_set(sc, cdr(tree), e));
    }
  return(e);
}


static bool optimize_syntax(s7_scheme *sc, s7_pointer x, s7_pointer func, int hop, s7_pointer e)
{
  s7_pointer p;
  opcode_t op;

  p = car(x);

  /* fprintf(stderr, "optimize syntax %s\n     e: %s\n", DISPLAY(p), DISPLAY(e)); */

  if (!is_pair(cdr(p))) /* cddr(p) might be null if, for example, (begin (let ...)) */
    return(false);

  op = (opcode_t)syntax_opcode(func);

  if ((op == OP_QUOTE) || 
      (op == OP_WITH_ENV))
    return(false);

  sc->w = e;
  if (op == OP_LET)
    {
      /* fprintf(stderr, "let\n"); */
      if (is_symbol(cadr(p)))
	{
	  if (global_slot(cadr(p)) != sc->NIL)
	    sc->w = cons(sc, cadr(p), sc->w);
	  sc->w = collect_collisions(sc, caddr(p), sc->w);
	}
      else 
	{
	  sc->w = collect_collisions(sc, cadr(p), sc->w);
	}
    }
  else
    {
      if ((op == OP_LET_STAR) ||
	  (op == OP_LETREC))
	{
	  sc->w = collect_collisions(sc, cadr(p), sc->w);
	}
      else
	{
	  if ((op == OP_DEFINE) || 
	      (op == OP_DEFINE_STAR) ||
	      (op == OP_LAMBDA) ||
	      (op == OP_LAMBDA_STAR))
	    {
	      /* fprintf(stderr, "define: %s\n", DISPLAY_80(p)); */
	      if (is_symbol(cadr(p)))
		{
		  if (global_slot(cadr(p)) != sc->NIL)
		    sc->w = cons(sc, cadr(p), sc->w);
		}
	      else sc->w = collect_collisions(sc, cadr(p), sc->w);

	      if ((op == OP_DEFINE) &&
		  (is_pair(cadr(p))))
		sc->w = cons(sc, p, sc->w);
	    }
	  else
	    {
	      if (op == OP_SET)
		{
		  if ((is_symbol(cadr(p))) &&
		      (global_slot(cadr(p)) != sc->NIL))
		    sc->w = cons(sc, cadr(p), sc->w);
		}
	      else
		{
		  if (op == OP_DO)
		    {
		      sc->w = collect_collisions(sc, cadr(p), sc->w);
		    }
		}
	    }
	}
    }
  
  sc->w = check_for_global_set(sc, cdar(x), sc->w);
  /* (define (hi a) (if a (begin (+ 1 a) (display a) (+ a a))))
   */

  for (p = cdar(x); is_pair(p); p = cdr(p))
    {
      if (is_pair(car(p)))
	{
	  if (!is_checked(car(p)))
	    optimize_expression(sc, p, hop, sc->w);

	  if ((is_optimized(car(p))) &&
	      (op == OP_BEGIN) &&
	      (!is_null(cdr(p))))
	    {
	      /* fprintf(stderr, "%s %s\n", opt_names[optimize_data(car(p))], DISPLAY_80(car(p))); */
	      if (optimize_data(car(p)) == HOP_SAFE_C_SS)
		optimize_data(car(p)) = HOP_SAFE_C_SS_2;
	    }
	}
    }

  return(false);
}


static s7_pointer find_uncomplicated_symbol(s7_scheme *sc, s7_pointer hdl, s7_pointer e)
{ 
  s7_pointer x;
  long long int id;

  if (direct_memq(hdl, e))
    return(sc->NIL);

  if (is_global(hdl))
    return(global_slot(hdl));

  id = symbol_id(hdl);
  for (x = sc->envir; id < frame_id(x); x = next_environment(x));
  for (; is_environment(x); x = next_environment(x))
    {
      s7_pointer y;
      if (frame_id(x) == id)
	return(local_slot(hdl));

      for (y = environment_slots(x); is_slot(y); y = next_slot(y))
	if (slot_symbol(y) == hdl)
	  return(y);
    }
  return(sc->NIL);
} 

/* (define (hi) (let ((old+ +) (j 0)) (do ((i 0 (+ i 1))) ((or (< i -3) (> i 3))) (set! + -) (set! j (old+ j i))) (set! + old+) j)) */

static bool optimize_expression(s7_scheme *sc, s7_pointer x, int hop, s7_pointer e)
{
  s7_pointer y;

  set_checked(car(x));
  y = caar(x);

  /* fprintf(stderr, "[%p]\noptimize_expression %s\n     e: %s, y: %s %s\n", car(x), DISPLAY(car(x)), DISPLAY(e), DISPLAY(y), describe_type_bits(y)); */

  if (is_symbol(y))
    {
      s7_pointer func;
      
      if (is_syntactic(y))
	return(optimize_syntax(sc, x, slot_value(global_slot(y)), hop, e));

      if (y == sc->QUOTE)
	return(false);

      func = find_uncomplicated_symbol(sc, y, e);
      /* fprintf(stderr, "find: %s -> %s in %s\n", DISPLAY(y), DISPLAY(func), DISPLAY(e)); */

      if (is_not_null(func))
	func = slot_value(func);
      else func = sc->F;

      if (is_syntactic(func))
	return(optimize_syntax(sc, x, func, hop, e));

      /* we miss implicit indexing here because at this time, the data are not set */

      if ((is_procedure(func)) ||
	  (is_c_function(func)) ||
	  (is_safe_procedure(func))) /* built-in applicable objects like vectors */
	return(optimize_function(sc, x, func, hop, e));
      else                           /* caar(x) is a symbol but it's not a procedure or a "safe" case = vector etc */
	{
	  /* else it's something like a let variable binding: (sqrtfreq (sqrt frequency)) */
	  s7_pointer p;
	  int len = 0, pairs = 0, symbols = 0;
	  for (p = cdar(x); is_pair(p); p = cdr(p), len++)
	    {
	      if (is_pair(car(p)))
		{
		  pairs++;
		  if (!is_checked(car(p)))
		    {
		      optimize_expression(sc, p, hop, e);
		    }
		}
	      else
		{
		  if (is_symbol(car(p)))
		    symbols++;
		}
	    }

	  /* fprintf(stderr, "%s: pairs: %d, symbols: %d, len: %d, is_optimized: %d\n", DISPLAY_80(car(x)), pairs, symbols, len, is_optimized(car(x)));  */
	  /* (define (hi) (let ((v (vector 1 2 3)) (i 1)) (v i)))
	   * (define (ho v) (let ((i 1)) (v i))) (ho (vector 1 2 3)) (ho (list 1 22 3))
	   */
	  if ((is_null(p)) &&              /* (+ 1 . 2) */
	      (!is_optimized(car(x))))
	    {
	      /* fprintf(stderr, "opt expr unknown %s\n", DISPLAY_80(car(x))); */

	      if (pairs == 0)
		{
		  if (len == 1)
		    {
		      if (caar(x) != sc->QUOTE) /* !! quote can be redefined locally, unsetting the T_SYNTACTIC flag -- can this happen elsewhere? */
			{
			  set_optimized(car(x));
			  set_unsafe(car(x));
			  if (symbols == 1)
			    set_optimize_data(car(x), OP_UNKNOWN_S);
			  else set_optimize_data(car(x), OP_UNKNOWN_C);
			  ecdr(car(x)) = NULL;

			  /* hooboy -- we get here in let bindings...
			   *
			   * to save access to the caller, we'd need to pass it as an arg to optimize_expression
			   */
			}
		      return(false); 
		    }
		  
		  if (len == 2)
		    {
		      set_optimized(car(x));
		      set_unsafe(car(x));
		      if (symbols == 2)
			set_optimize_data(car(x), OP_UNKNOWN_SS);
		      else
			{
			  if (symbols == 0)
			    set_optimize_data(car(x), OP_UNKNOWN_CC);
			  else
			    {
			      if (is_symbol(cadar(x)))
				set_optimize_data(car(x), OP_UNKNOWN_SC);
			      else set_optimize_data(car(x), OP_UNKNOWN_CS);
			    }
			}
		      ecdr(car(x)) = NULL;
		      return(false); 
		    }
		  
		  if ((len == 3) &&
		      (symbols == 3))
		    {
		      set_optimized(car(x));
		      set_unsafe(car(x));
		      set_optimize_data(car(x), OP_UNKNOWN_SSS);
		      ecdr(car(x)) = NULL;
		      return(false); 
		    }
		}
	      else
		{
		  /*
		  fprintf(stderr, "%s is unknown (%d %d %d %s)\n", DISPLAY_80(car(x)), pairs, len, 
			  is_optimized(cadar(x)), (is_optimized(cadar(x))) ? opt_names[optimize_data(cadar(x))] : "");
		  */
		  if ((pairs == 1) &&
		      (len == 1) &&
		      (is_optimized(cadar(x))) &&
		      (is_safe_c_s(cadar(x))))
		    {
		      set_optimized(car(x));
		      set_unsafe(car(x));
		      set_optimize_data(car(x), OP_UNKNOWN_opSq);
		      ecdr(car(x)) = NULL;
		      return(false); 
		    }
		  
		  if ((pairs == 1) &&
		      (len == 1) &&
		      (is_optimized(cadar(x))) &&
		      (optimize_data_match(cadar(x), OP_SAFE_C_C)))
		    {
		      set_optimized(car(x));
		      set_unsafe(car(x));
		      /* fprintf(stderr, "opCq: %s\n", DISPLAY(car(x)));  */
		      set_optimize_data(car(x), OP_UNKNOWN_opCq);
		      ecdr(car(x)) = NULL;
		      return(false); 
		    }
		  
		  if ((pairs == 1) &&
		      (len == 2) &&
		      (symbols == 1))
		    {
		      if ((is_pair(cadar(x))) &&
			  (is_optimized(cadar(x))) &&
			  (is_safe_c_s(cadar(x))))
			{
			  set_optimized(car(x));
			  set_unsafe(car(x));
			  set_optimize_data(car(x), OP_UNKNOWN_opSq_S);
			  ecdr(car(x)) = NULL;
			  return(false); 
			}
		      if ((is_pair(caddar(x))) &&
			  (is_optimized(caddar(x))) &&
			  (is_safe_c_s(caddar(x))))
			{
			  set_optimized(car(x));
			  set_unsafe(car(x));
			  set_optimize_data(car(x), OP_UNKNOWN_S_opSq);
			  ecdr(car(x)) = NULL;
			  return(false); 
			}
		    }
		  if ((pairs == 2) &&
		      (len == 2) &&
		      (is_optimized(cadar(x))) &&
		      (is_optimized(caddar(x))) &&
		      (is_safe_c_s(cadar(x))) &&
		      (is_safe_c_s(caddar(x))))
		    {
		      set_optimized(car(x));
		      set_unsafe(car(x));
		      set_optimize_data(car(x), OP_UNKNOWN_opSq_opSq);
		      ecdr(car(x)) = NULL;
		      return(false); 
		    }
		}
	    }
	}
    }
  else
    {
      /* caar(x) is not a symbol, but there might be interesting stuff here */
      /* (define (hi a) (case 1 ((1) (if (> a 2) a 2)))) */

      s7_pointer p;
      for (p = car(x); is_pair(p); p = cdr(p))
	{
	  if (is_pair(car(p)))
	    {
	      if (!is_checked(car(p)))
		{
		  optimize_expression(sc, p, hop, e);
		}
	    }
	}
    }
  return(false);
}


static bool optimize(s7_scheme *sc, s7_pointer code, int hop, s7_pointer e)
{
  s7_pointer x;
  bool happy = true;

  /* fprintf(stderr, "optimize: %s\n", DISPLAY_80(code)); */

  for (x = code; is_pair(x) && (!is_checked(x)); x = cdr(x))
    {
      set_checked(x);
      if (is_pair(car(x)))
	{
	  bool opt;
	  opt = optimize_expression(sc, x, hop, e);
	  if (!opt) happy = false;
	}
      else
	{
	  if (!is_optimized(x))
	    happy = false;
	}
    }
#if 0
  if (happy) 
    {
      if (!is_optimized(code))
	fprintf(stderr, "opt: %s %d\n", DISPLAY_80(code), optimize_data(code));
      set_optimized(code);
    }
#endif
  return(happy);
}


static int combine_ops(s7_scheme *sc, int op1, s7_pointer e1, s7_pointer e2)
{
  int op2;
  op2 = optimize_data(e2) & 0xfffe;

  switch (op1)
    {
    case SAFE_C_P:
      switch (op2)
	{
	case OP_SAFE_C_C: 
	  return(OP_SAFE_C_opCq); /* this includes the multi-arg C_C cases */

	case OP_SAFE_IS_PAIR_S:
	case OP_SAFE_IS_SYMBOL_S:
	case OP_SAFE_NOT_S:

	case OP_SAFE_C_S: 
	  return(OP_SAFE_C_opSq);

	case OP_SAFE_C_Q: 
	  return(OP_SAFE_C_opQq);

	case OP_SAFE_C_SS: 
	  return(OP_SAFE_C_opSSq);

	case OP_SAFE_C_SQ: 
	  return(OP_SAFE_C_opSQq);

	case OP_SAFE_C_QS: 
	  return(OP_SAFE_C_opQSq);

	case OP_SAFE_C_SC: 
	  return(OP_SAFE_C_opSCq);

	case OP_SAFE_C_CS: 
	  return(OP_SAFE_C_opCSq);

	  /* CQ QQ QC */

	case OP_SAFE_C_XXX:
	  return(OP_SAFE_C_opXXXq);

	  /* safe_c_op_opCq_q
	   */

	default:
	  return(OP_SAFE_C_Z);
	  break;
	}
      break;
      

    case SAFE_C_SP:
      switch (op2)
	{
	case OP_SAFE_IS_PAIR_S:
	case OP_SAFE_IS_SYMBOL_S:
	case OP_SAFE_NOT_S:

	case OP_SAFE_C_S:
	  return(OP_SAFE_C_S_opSq);

	case OP_SAFE_C_C:
	  return(OP_SAFE_C_S_opCq);

	case OP_SAFE_C_SC:
	  return(OP_SAFE_C_S_opSCq);
	  
	case OP_SAFE_C_CS:
	  return(OP_SAFE_C_S_opCSq);

	case OP_SAFE_C_SS:
	  return(OP_SAFE_C_S_opSSq);

	default:
	  return(OP_SAFE_C_SZ);
	  break;
	}
      break;


    case SAFE_C_QP:
      return(OP_SAFE_C_QZ);

    case SAFE_C_PQ:
      return(OP_SAFE_C_ZQ);


    case SAFE_C_PS:
      switch (op2)
	{
	case OP_SAFE_C_C:
	  return(OP_SAFE_C_opCq_S);

	case OP_SAFE_IS_PAIR_S:
	case OP_SAFE_IS_SYMBOL_S:
	case OP_SAFE_NOT_S:

	case OP_SAFE_C_S:
	  return(OP_SAFE_C_opSq_S);
	  
	case OP_SAFE_C_CS:
	  return(OP_SAFE_C_opCSq_S);

	case OP_SAFE_C_SC:
	  return(OP_SAFE_C_opSCq_S);

	case OP_SAFE_C_SS:
	  return(OP_SAFE_C_opSSq_S);

	default:
	  return(OP_SAFE_C_ZS);
	  break;
	}
      break;


    case SAFE_C_PC:
      switch (op2)
	{
	case OP_SAFE_C_C:
	  return(OP_SAFE_C_opCq_C);

	case OP_SAFE_IS_PAIR_S:
	case OP_SAFE_IS_SYMBOL_S:
	case OP_SAFE_NOT_S:

	case OP_SAFE_C_S:
	  return(OP_SAFE_C_opSq_C);

	case OP_SAFE_C_CS:
	  return(OP_SAFE_C_opCSq_C);

	case OP_SAFE_C_SS:
	  return(OP_SAFE_C_opSSq_C);

	case OP_SAFE_C_SC:
	  return(OP_SAFE_C_opSCq_C);

	default:
	  return(OP_SAFE_C_ZC);
	  break;
	}
      break;


    case SAFE_C_CP:
      switch (op2)
	{
	case OP_SAFE_C_C:
	  return(OP_SAFE_C_C_opCq);

	case OP_SAFE_IS_PAIR_S:
	case OP_SAFE_IS_SYMBOL_S:
	case OP_SAFE_NOT_S:

	case OP_SAFE_C_S:
	  return(OP_SAFE_C_C_opSq);

	case OP_SAFE_C_CS:
	  return(OP_SAFE_C_C_opCSq);

	case OP_SAFE_C_SC:
	  return(OP_SAFE_C_C_opSCq);
	  
	case OP_SAFE_C_SS:
	  return(OP_SAFE_C_C_opSSq);

	default:
	  return(OP_SAFE_C_CZ);
	  break;
	}
      break;


    case SAFE_C_PP:
      switch (op2)
	{
	case OP_SAFE_IS_PAIR_S:
	case OP_SAFE_IS_SYMBOL_S:
	case OP_SAFE_NOT_S:

	case OP_SAFE_C_S:
	  if (optimize_data_match(e1, OP_SAFE_C_S))
	    return(OP_SAFE_C_opSq_opSq);
	  break;

	case OP_SAFE_C_C:
	  if (optimize_data_match(e1, OP_SAFE_C_C))
	    return(OP_SAFE_C_opCq_opCq);
	  break;

	case OP_SAFE_C_SC:
	  if (optimize_data_match(e1, OP_SAFE_C_SC))
	    return(OP_SAFE_C_opSCq_opSCq);
	  break;

	case OP_SAFE_C_SS:
	  if (optimize_data_match(e1, OP_SAFE_C_SS))
	    return(OP_SAFE_C_opSSq_opSSq);
	  break;

	default:
	  return(OP_SAFE_C_ZZ);
	  break;
	}
      break;


    default:
      break;
    }
  return(op1);
}


#if WITH_GCC

#define c_function_is_ok(Sc, X) ({ s7_pointer _p_, _X_; \
                                   _X_ = X;  \
                                    _p_ = SYMBOL_TO_VALUE(Sc, car(_X_));			 \
				    ((_p_ == ecdr(_X_)) || \
                                    ((is_c_function(_p_)) && (is_c_function(ecdr(_X_))) && \
                                     (c_function_class(_p_) == c_function_class(ecdr(_X_))))); })

#else

static bool c_function_is_ok(s7_scheme *sc, s7_pointer x) 
{
  s7_pointer p;
  p = SYMBOL_TO_VALUE(sc, car(x));
  return((p == ecdr(x)) ||
	 ((is_c_function(p)) &&
	  (c_function_class(p) == c_function_class(ecdr(x)))));
}

#endif

#define indirect_c_function_is_ok(Sc, X) (((optimize_data(X) & 0x1) != 0) || (c_function_is_ok(Sc, X)))


static bool sequence_is_safe_for_opteval(s7_scheme *sc, s7_pointer body)
{
  s7_pointer p;
  for (p = body; is_pair(p); p = cdr(p))
    {
      s7_pointer expr;
      expr = car(p);
      
      /* only pairs have the optimized bit set
	if ((!is_pair(expr)) || 
	  (!is_optimized(expr)))
	return(false);
      */
      if (!is_optimized(expr))
	return(false);

      /* we're checking twice in some cases, but the 2nd (later) check can't be
       *   omitted, (by using is_checked), because the function might 
       *   have been redefined since the initial check.  
       */
      if ((optimize_data(expr) & 1) != 1)
	{
	  switch (optimize_data(expr))
	    {
	    case OP_VECTOR_C:
	    case OP_VECTOR_S:
	    case OP_STRING_C:
	    case OP_STRING_S:
	    case OP_PAIR_C:
	    case OP_PAIR_S:
	    case OP_HASH_TABLE_C:
	    case OP_HASH_TABLE_S:
	    case OP_C_OBJECT_C:
	    case OP_C_OBJECT_S:
	    case OP_ENVIRONMENT_S:
	      break;
	      
	    case OP_SAFE_IS_PAIR_S:
	    case OP_SAFE_IS_SYMBOL_S:
	    case OP_SAFE_NOT_S:
	    case OP_SAFE_C_C: 
	    case OP_SAFE_C_S: 
	    case OP_SAFE_C_Q:
	    case OP_SAFE_C_SS:
	    case OP_SAFE_C_SS_2:
	    case OP_SAFE_C_SC:
	    case OP_SAFE_C_CS:
	    case OP_SAFE_C_SQ:
	    case OP_SAFE_C_QS:
	    case OP_SAFE_C_QQ:
	    case OP_SAFE_C_XXX:
	    case OP_SAFE_C_SSS:
	    case OP_SAFE_C_ALL_S:
	    case OP_SAFE_C_ALL_G:
	    case OP_SAFE_C_ALL_X:
	      if (!c_function_is_ok(sc, expr)) return(false);
	      break;
	      
	      
	    case OP_SAFE_C_opCq:
	    case OP_SAFE_C_opQq: 
	    case OP_SAFE_C_opSq: 
	    case OP_SAFE_C_opSSq:
	    case OP_SAFE_C_opSCq:
	    case OP_SAFE_C_opSQq:
	    case OP_SAFE_C_opQSq:
	    case OP_SAFE_C_opCSq:
	    case OP_SAFE_C_opCq_S:
	    case OP_SAFE_C_opSq_S:
	    case OP_SAFE_C_opCq_C:
	    case OP_SAFE_C_opSq_C:
	    case OP_SAFE_C_opCSq_C:
	    case OP_SAFE_C_opSSq_C:
	    case OP_SAFE_C_opSCq_S:
	    case OP_SAFE_C_opSCq_C:
	    case OP_SAFE_C_opCSq_S:
	    case OP_SAFE_C_opSSq_S:
	    case OP_SAFE_C_opXXXq:
	    case OP_SAFE_C_Z:
	    case OP_SAFE_C_ZS:
	    case OP_SAFE_C_ZC:
	    case OP_SAFE_C_ZQ:
	    case OP_SAFE_C_ZXX:
	    case OP_SAFE_C_DXX:
	      if (!c_function_is_ok(sc, expr)) return(false);
	      if (!c_function_is_ok(sc, cadr(expr))) return(false);
	      break;
	      
	      
	    case OP_SAFE_C_C_opSq:
	    case OP_SAFE_C_C_opCq:
	    case OP_SAFE_C_C_opCSq:
	    case OP_SAFE_C_C_opSSq:
	    case OP_SAFE_C_S_opSq:
	    case OP_SAFE_C_S_opCq:
	    case OP_SAFE_C_S_opSCq:
	    case OP_SAFE_C_C_opSCq:
	    case OP_SAFE_C_S_opSSq:
	    case OP_SAFE_C_S_opCSq:
	    case OP_SAFE_C_SZ:
	    case OP_SAFE_C_CZ:
	    case OP_SAFE_C_QZ:
	    case OP_SAFE_C_XZX:
	    case OP_SAFE_C_XDX:
	      if (!c_function_is_ok(sc, expr)) return(false);
	      if (!c_function_is_ok(sc, caddr(expr))) return(false);
	      break;
	      
	      
	    case OP_SAFE_C_opCq_opCq:
	    case OP_SAFE_C_opSq_opSq:
	    case OP_SAFE_C_opSCq_opSCq:
	    case OP_SAFE_C_opSSq_opSSq:
	    case OP_SAFE_C_ZZ:
	    case OP_SAFE_C_ZZX:
	    case OP_SAFE_C_DD:
	    case OP_SAFE_C_DDX:
	      /* (op (op S) (op S)) */
	      if (!c_function_is_ok(sc, expr)) return(false);
	      if (!c_function_is_ok(sc, cadr(expr))) return(false);
	      if (!c_function_is_ok(sc, caddr(expr))) return(false);
	      break;

	    case OP_SAFE_C_XXZ:
	    case OP_SAFE_C_XXD:
	      if (!c_function_is_ok(sc, expr)) return(false);
	      if (!c_function_is_ok(sc, cadddr(expr))) return(false);
	      break;

	    case OP_SAFE_C_XZZ:
	    case OP_SAFE_C_XDD:
	      if (!c_function_is_ok(sc, expr)) return(false);
	      if (!c_function_is_ok(sc, caddr(expr))) return(false);
	      if (!c_function_is_ok(sc, cadddr(expr))) return(false);
	      break;

	    case OP_SAFE_C_ZXZ:
	    case OP_SAFE_C_DXD:
	      if (!c_function_is_ok(sc, expr)) return(false);
	      if (!c_function_is_ok(sc, cadr(expr))) return(false);
	      if (!c_function_is_ok(sc, cadddr(expr))) return(false);
	      break;

	    case OP_SAFE_C_ZZZ:
	    case OP_SAFE_C_DDD:
	      if (!c_function_is_ok(sc, expr)) return(false);
	      if (!c_function_is_ok(sc, cadr(expr))) return(false);
	      if (!c_function_is_ok(sc, caddr(expr))) return(false);
	      if (!c_function_is_ok(sc, cadddr(expr))) return(false);
	      break;
	      
	    default:
	      return(false);
	    }
	}
    }
  return(true);
}


static bool is_hop_safe_closure(s7_pointer p)
{
  int op;
  op = optimize_data(p);
  return(((op & 1) == 1) &&
	 (op >= HOP_SAFE_THUNK) &&
	 (op <= HOP_SAFE_CLOSURE_opCq));
}


static bool form_is_safe(s7_scheme *sc, s7_pointer func, s7_pointer x, bool at_end, bool *bad_set)
{
  /* fprintf(stderr, "    f: %s %s %d %d %d\n", DISPLAY(func), DISPLAY_80(x), at_end, is_syntactic(car(x)), is_optimized(car(x))); */
  if (!is_proper_list(sc, x))
    return(false);
  if (is_syntactic(car(x)))
    {
      switch (syntax_opcode(car(x)))
	{
	case OP_OR:
	case OP_AND:
	case OP_BEGIN:
	  if (!body_is_safe(sc, func, cdr(x), at_end, bad_set))
	    return(false);
	  break;
	  
	case OP_QUOTE:
	  break;
	  
	  /* in the name binders, we first have to check that "func" actually is the same thing as the caller's func */
	case OP_LET:
	  if (is_symbol(cadr(x)))
	    {
	      s7_pointer vars, p;
	      for (vars = caddr(x); is_pair(vars); vars = cdr(vars))
		{
		  if ((caar(vars) == func) ||
		      (caar(vars) == cadr(x)))
		    return(false); /* it's shadowed -- all bets are off! */
		}
	      
	      for (p = cdddr(x); is_not_null(cdr(p)); p = cdr(p));
	      if (is_pair(car(p)))
		{
		  form_is_safe(sc, cadr(x), car(p), true, bad_set);
		  if ((at_end) && (!*bad_set))
		    form_is_safe(sc, func, car(p), true, bad_set);
		}
	      return(false);
	    }
	  /* fall through */
	  
	case OP_LET_STAR:
	case OP_LETREC:
	  {
	    bool happy = true;
	    if (is_pair(cadr(x)))
	      {
		s7_pointer vars;
		for (vars = cadr(x); is_pair(vars); vars = cdr(vars))
		  {
		    if ((!is_pair(car(vars))) ||
			(!is_pair(cdar(vars))))
		      return(false);
		    
		    if ((is_pair(cadar(vars))) &&
			(!form_is_safe(sc, func, cadar(vars), false, bad_set)))
		      happy = false;
		    
		    if (caar(vars) == func)
		      return(false); /* it's shadowed */
		  }
	      }
	    if ((!body_is_safe(sc, func, cddr(x), at_end, bad_set)) ||
		(!happy))
	      return(false);
	  }
	  break;
	  
	case OP_IF:
	  {
	    bool t, e1, e2;
	    t = ((!is_pair(cadr(x))) || (form_is_safe(sc, func, cadr(x), false, bad_set)));
	    e1 = ((!is_pair(caddr(x))) || (form_is_safe(sc, func, caddr(x), at_end, bad_set)));
	    e2 = ((!is_pair(cdddr(x))) || (!is_pair(cadddr(x))) || (form_is_safe(sc, func, cadddr(x), at_end, bad_set)));
	    return((t) && (e1) && (e2));
	  }
	  break;
	  
	case OP_COND:
	  {
	    s7_pointer p;
	    bool happy = true;
	    for (p = cdr(x); is_pair(p); p = cdr(p))
	      {
		s7_pointer expr;
		expr = car(p);
		if (is_pair(expr)) /* ?? */
		  {
		    if ((is_pair(car(expr))) && (!form_is_safe(sc, func, car(expr), false, bad_set)))
		      happy = false;
		    if ((is_pair(cdr(expr))) && (!body_is_safe(sc, func, cdr(expr), at_end, bad_set)))
		      happy = false;
		  }
	      }
	    if ((!happy) || (is_not_null(p)))
	      return(false);
	  }
	  break;
	  
	case OP_CASE:
	  {
	    s7_pointer p;
	    bool happy = true;
	    if ((is_pair(cadr(x))) && (!form_is_safe(sc, func, cadr(x), false, bad_set))) return(false);
	    for (p = cddr(x); is_pair(p); p = cdr(p))
	      {
		if ((is_pair(car(p))) && (!body_is_safe(sc, func, cdar(p), at_end, bad_set)))
		  happy = false;
	      }
	    return(happy);
	  }
	  break;
	  
	case OP_DO:
	  /* (do (...) (...) ...) */
	  if (!is_pair(cddr(x)))
	    return(false);
	  if (!body_is_safe(sc, func, cdddr(x), false, bad_set))
	    return(false);
	  if (is_pair(cadr(x)))
	    {
	      s7_pointer vars;
	      for (vars = cadr(x); is_pair(vars); vars = cdr(vars))
		{
		  if (!is_pair(car(vars)))
		    return(false);
		  
		  if (caar(vars) == func)
		    return(false);
		  /* fprintf(stderr, "cadar: %s, cadr: %s\n", DISPLAY_80(cadar(vars)), DISPLAY_80(cadr(vars))); */
		  
		  if ((is_pair(cadar(vars))) &&
		      (!form_is_safe(sc, func, cadar(vars), false, bad_set)))
		    return(false);
		  
		  if ((is_pair(cddar(vars))) &&
		      (is_pair(caddar(vars))) &&
		      (!form_is_safe(sc, func, caddar(vars), false, bad_set)))
		    return(false);
		}
	    }
	  if ((is_pair(caddr(x))) &&
	      (!body_is_safe(sc, func, caddr(x), at_end, bad_set)))
	    return(false);
	  break;
	  
	case OP_SET:
	  /* if we set func, we have to make sure we abandon the tail call scan:
	   * (let () (define (hi a) (let ((v (vector 1 2 3))) (set! hi v) (hi a))) (hi 1))
	   */
	  /* fprintf(stderr, "%s: %s %d\n", DISPLAY(x), DISPLAY(func), cadr(x) == func); */
	  if (cadr(x) == func)
	    (*bad_set) = true;
	  return(false);
	  
	default:
	  /* try to catch weird cases like: 
	   * (let () (define (hi1 a) (define (hi1 b) (+ b 1)) (hi1 a)) (hi1 1))
	   * TODO: what about trailers?
	   * (let () (define (hi1 a) (define (ho1 b) b) (define (hi1 b) (+ b 1)) (hi1 a)) (hi1 1))
	   */
	  /*
	    if (syntax_opcode(car(x)) >= OP_SAFE_IF1)
	    fprintf(stderr, "%s %s no good: %s, %s %s, %p %p%s\n", 
	    BOLD_TEXT, real_op_names[syntax_opcode(car(x))], DISPLAY(x), DISPLAY(func), DISPLAY(caadr(x)), caadr(x), func, UNBOLD_TEXT);
	  */
	  
	  if ((is_pair(cdr(x))) &&
	      ((cadr(x) == func) ||
	       ((is_pair(cadr(x))) && (caadr(x) == func))))
	    (*bad_set) = true;
	  return(false);
	}
    }
  else
    {
      if ((!is_optimized(x)) || 
	  ((is_unsafe(x)) &&
	   (!is_hop_safe_closure(x))))
	{
	  if (car(x) == func) /* try to catch tail call */
	    {
	      s7_pointer p;
	      for (p = cdr(x); is_pair(p); p = cdr(p))
		if ((is_pair(car(p))) &&
		    (!is_optimized(car(p))))
		  return(false);

	      if ((at_end) && (is_null(p) && (!*bad_set)))
		{
		  /* we still miss a few cases, and need more unknown optimizations, or set them here
		   *   (let () (define (hi1 a) (define (hi1 b) (+ b 1)) (hi1 a)) (hi1 1)) 2)
		   */
		  /*
		  fprintf(stderr, "%s%s tail-call: %s x: %s, arg1: %s%s\n", 
			  BOLD_TEXT, 
			  DISPLAY(func), 
			  DISPLAY(x), 
			  (is_optimized(x)) ? opt_names[optimize_data(x)] : "none",
			  ((is_not_null(cdr(x))) && (is_pair(cadr(x))) && (is_optimized(cadr(x)))) ? opt_names[optimize_data(cadr(x))] : "none",
			  UNBOLD_TEXT);
		  */
		  set_tail_call(x);
		  /* fcdr(x) = NULL; */
		  return(true);
		}
	      return(false);
	    }
	  
	  if (is_symbol(car(x)))
	    {
	      if (is_global(car(x)))
		{
		  s7_pointer f;
		  f = finder(sc, car(x));
		  if (((is_c_function(f)) &&
		       ((is_safe_procedure(f)) ||
			((is_null(cdddr(x))) &&
			 ((c_function_call(f) == g_member) || (c_function_call(f) == g_assoc))))) ||
		      ((is_closure(f)) &&
		       (is_safe_closure(closure_body(f)))))
		    {
		      s7_pointer p;
		      for (p = cdr(x); is_pair(p); p = cdr(p))
			if ((is_pair(car(p))) &&
			    ((!is_optimized(car(p))) ||
			     ((is_unsafe(car(p))) &&
			      (!is_hop_safe_closure(car(p))))))
			  {
			    /* fprintf(stderr, "unopt trailing arg: %s\n", DISPLAY(car(p))); */
			    if ((caar(p) != func) ||
				(!is_null(cdr(p))))
			      return(false);
			  }
		      return(is_null(p));
		    }
		}
	      else
		{
		  s7_pointer f;
		  f = find_symbol(sc, car(x));
		  if ((is_slot(f)) &&
		      ((is_syntax(slot_value(f))) || (is_any_macro(slot_value(f)))))
		    (*bad_set) = true;
		}
	    }
	  return(false);
	}
    }
  return(true);
}


static bool body_is_safe(s7_scheme *sc, s7_pointer func, s7_pointer body, bool at_end, bool *bad_set)
{
  s7_pointer p;
  /* fprintf(stderr, "b: %s %s %d\n", DISPLAY(func), DISPLAY_80(body), at_end); */
  for (p = body; is_pair(p); p = cdr(p))
    {
      if ((is_pair(car(p))) &&
	  (!form_is_safe(sc, func, car(p), (at_end) && (is_null(cdr(p))), bad_set)))
	{
	  if ((at_end) && (!is_null(cdr(p))) && (!*bad_set))
	    {
	      /* we checked in check_define that this is not a circular list */
	      for (p = cdr(p); is_not_null(cdr(p)); p = cdr(p));
	      if (is_pair(car(p)))
		form_is_safe(sc, func, car(p), at_end, bad_set);
	    }
	  return(false);
	}
    }
  return(true);
}
#endif


void s7_unoptimize(s7_scheme *sc, s7_pointer code)
{
  /* needed by the run macro 
   */
  if (is_pair(code))
    {
      if ((typeflag(car(code)) == SYNTACTIC_TYPE) &&
	  (syntax_opcode(car(code)) >= OP_QUOTE_UNCHECKED))
	car(code) = car(slot_value(global_slot(car(code))));
      else s7_unoptimize(sc, car(code));
      s7_unoptimize(sc, cdr(code));
    }
}


static s7_pointer g_unoptimize(s7_scheme *sc, s7_pointer args)
{
  #define H_unoptimize "(unoptimize code) erases all the optimizer info in code"

  if (typeflag(car(args)) == SYNTACTIC_TYPE)
    return(car(slot_value(global_slot(car(args)))));

  s7_unoptimize(sc, car(args));
  return(car(args));
}




/* ---------------------------------------- error checks ---------------------------------------- */

static s7_pointer check_lambda_args(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_list(sc, args))
    {
      if (s7_is_constant(args))                       /* (lambda :a ...) */
	return(eval_error(sc, "lambda parameter '~S is a constant", args)); /* not ~A here, (lambda #\null do) for example */
      
      /* we currently accept (lambda i i . i) (lambda quote i)  (lambda : : . #()) (lambda : 1 . "")
       *   at this level, but when the lambda form is evaluated, it will trigger an error.
       */
      if (is_symbol(args))
	set_local(args);
    }
  else
    {
      s7_pointer x;
      for (x = args; is_pair(x); x = cdr(x))
	{
	  if (s7_is_constant(car(x)))                      /* (lambda (pi) pi) */
	    return(eval_error(sc, "lambda parameter '~S is a constant", car(x)));

	  if (symbol_is_in_arg_list(car(x), cdr(x)))       /* (lambda (a a) ...) or (lambda (a . a) ...) */
	    return(eval_error(sc, "lambda parameter '~S is used twice in the parameter list", car(x)));
	  set_local(car(x));
	}
      if ((is_not_null(x)) &&
	  (s7_is_constant(x)))                             /* (lambda (a . 0.0) a) or (lambda (a . :b) a) */
	return(eval_error(sc, "lambda :rest parameter '~A is a constant", x));
    }
  return(sc->F);
}


static s7_pointer check_lambda_star_args(s7_scheme *sc, s7_pointer args)
{
  if (!s7_is_list(sc, args))
    {
      if (s7_is_constant(args))                                  /* (lambda* :a ...) */
	return(eval_error(sc, "lambda* parameter '~A is a constant", args));
      if (is_symbol(args))
	set_local(args);
    }
  else
    { 
      s7_pointer w;
      for (w = args; is_pair(w); w = cdr(w))
	{
	  if (is_pair(car(w)))
	    {
	      if (s7_is_constant(caar(w)))                            /* (lambda* ((:a 1)) ...) */
		return(eval_error(sc, "lambda* parameter '~A is a constant", caar(w)));
	      if (symbol_is_in_arg_list(caar(w), cdr(w)))             /* (lambda* ((a 1) a) ...) */
		return(eval_error(sc, "lambda* parameter '~A is used twice in the argument list", caar(w)));

	      if (!is_pair(cdar(w)))                                  /* (lambda* ((a . 0.0)) a) */
		{
		  if (is_null(cdar(w)))                             /* (lambda* ((a)) ...) */
		    return(eval_error(sc, "lambda* parameter default value missing? '~A", car(w)));
		  return(eval_error(sc, "lambda* parameter is a dotted pair? '~A", car(w)));
		}
	      if (is_not_null(cddar(w)))                               /* (lambda* ((a 0.0 "hi")) a) */
		return(eval_error(sc, "lambda* parameter has multiple default values? '~A", car(w)));

	      set_local(caar(w));
	    }
	  else 
	    {
	      if (car(w) != sc->KEY_REST)
		{
		  if ((s7_is_constant(car(w))) &&
		      (car(w) != sc->KEY_KEY) &&
		      (car(w) != sc->KEY_OPTIONAL) &&
		      (car(w) != sc->KEY_ALLOW_OTHER_KEYS))           /* (lambda* (pi) ...) */
		    return(eval_error(sc, "lambda* parameter '~A is a constant", car(w)));

		  if (symbol_is_in_arg_list(car(w), cdr(w)))          /* (lambda* (a a) ...) or (lambda* (a . a) ...) */
		    return(eval_error(sc, "lambda* parameter '~A is used twice in the argument list", car(w)));

		  if ((car(w) == sc->KEY_ALLOW_OTHER_KEYS) &&         /* (lambda* (:allow-other-keys x) x) */
		      (is_not_null(cdr(w))))
		    eval_error(sc, ":allow-other-keys should be the last parameter: ~A", args);

		  if (!is_keyword(car(w))) set_local(car(w));
		}
	      else
		{
		  if (!is_pair(cdr(w)))                               /* (lambda* (:rest) ...) */
		    return(eval_error(sc, "lambda* :rest parameter missing? ~A", w));
		  if (!is_symbol(cadr(w)))                         /* (lambda* (:rest (a 1)) ...) */
		    {
		      if (!is_pair(cadr(w)))                          /* (lambda* (:rest 1) ...) */
			return(eval_error(sc, "lambda* :rest parameter is not a symbol? ~A", w));
		      return(eval_error(sc, "lambda* :rest parameter can't have a default value. ~A", w));
		    }
		  else
		    {
		      if (is_immutable(cadr(w)))
			return(s7_error(sc, sc->WRONG_TYPE_ARG,
					list_2(sc, make_protected_string(sc, "can't bind an immutable object: ~S"), w)));
		    }
		  set_local(cadr(w));
		}
	    }
	}
      if (is_not_null(w))
	{
	  if (s7_is_constant(w))                             /* (lambda* (a . 0.0) a) or (lambda* (a . :b) a) */
	    return(eval_error(sc, "lambda* :rest parameter '~A is a constant", w));
	  if (is_symbol(w))
	    set_local(w);
	}
    }
  return(sc->F);
}


static s7_pointer check_case(s7_scheme *sc)
{
  bool keys_simple = true, has_else = false, has_feed_to = false, keys_single = true, keys_int = true;
  s7_pointer x, y;

  if (!is_pair(sc->code))                                            /* (case) or (case . 1) */
    return(eval_error(sc, "case has no selector:  ~A", sc->code));
  if (!is_pair(cdr(sc->code)))                                       /* (case 1) or (case 1 . 1) */
    return(eval_error(sc, "case has no clauses?:  ~A", sc->code));
  if (!is_pair(cadr(sc->code)))                                      /* (case 1 1) */
    return(eval_error(sc, "case clause is not a list? ~A", sc->code));
  
  for (x = cdr(sc->code); is_not_null(x); x = cdr(x))
    {
      if ((!is_pair(x)) ||                                        /* (case 1 ((2) 1) . 1) */
	  (!is_pair(car(x))))
	return(eval_error(sc, "case clause ~A messed up", x));	 
      if (!is_pair(cdar(x)))                                      /* (case 1 ((1))) */
	return(eval_error(sc, "case clause result missing: ~A", car(x)));
      
      y = caar(x);
      if (!is_pair(y))
	{
	  if ((y != sc->ELSE) &&                                  /* (case 1 (2 1)) */
	      ((!is_symbol(y)) ||
	       (s7_symbol_value(sc, y) != sc->ELSE)))             /* "proper list" below because: (case 1 (() 2) ... */
	    return(eval_error(sc, "case clause key list ~A is not a proper list or 'else'", y));
	  if (is_not_null(cdr(x)))                                  /* (case 1 (else 1) ((2) 1)) */
	    return(eval_error(sc, "case 'else' clause, ~A, is not the last clause", x));
	  has_else = true;
	}
      else
	{
	  /* what about (case 1 ((1) #t) ((1) #f)) [this is ok by guile]
	   *            (case 1 ((1) #t) ())
	   *            (case 1 ((2 2 2) 1)): guile says #<unspecified>
	   */
	  
	  /* the selector (sc->value) is evaluated, but the search key is not
	   *    (case '2 ((2) 3) (else 1)) -> 3
	   *    (case '2 (('2) 3) (else 1)) -> 1
	   */
	  if (!is_simple(car(y)))
	    keys_simple = false;
	  if (!is_integer(car(y)))
	    keys_int = false;
	  if (!is_null(cdr(y)))
	    keys_single = false;

	  for (y = cdr(y); is_not_null(y); y = cdr(y)) 
	    {
	      if (!is_pair(y))                                        /* (case () ((1 . 2) . hi) . hi) */
		return(eval_error(sc, "case key list is improper? ~A", x));
	      if (!is_simple(car(y)))
		keys_simple = false;
	      if (!is_integer(car(y)))
		keys_int = false;
	    }
	}

      y = car(x);
      if ((cadr(y) == sc->FEED_TO) &&
	  (s7_symbol_value(sc, sc->FEED_TO) == sc->UNDEFINED))
	{
	  has_feed_to = true;
	  if (!is_pair(cddr(y)))                                  /* (case 1 (else =>)) */
	    return(eval_error(sc, "case: '=>' target missing?  ~A", y));
	  if (is_pair(cdddr(y)))                                  /* (case 1 (else => + - *)) */
	    return(eval_error(sc, "case: '=>' has too many targets: ~A", y));
	}
    }

  if ((is_overlaid(sc->code)) &&
      (cdr(ecdr(sc->code)) == sc->code))
    {
      if (is_pair(car(sc->code)))
	{
#if WITH_OPTIMIZATION
	  if ((is_optimized(car(sc->code))) &&
	      (is_h_safe_c_s(car(sc->code))))
	    car(ecdr(sc->code)) = sc->SAFE_CASE_S;
	  else car(ecdr(sc->code)) = sc->CASE_UNCHECKED;
#endif
	}
      else 
	{
	  if ((is_symbol(car(sc->code))) &&
	      (!has_feed_to) &&
	      (keys_simple))
	    {
	      if (has_else)
		{
		  /* if (keys_single) fprintf(stderr, "simple else: %s\n", DISPLAY_80(sc->code)); */
		  /* this case does happen -- perhaps CASE_SIMPLEST_ELSE */
		  car(ecdr(sc->code)) = sc->CASE_SIMPLE;
		}
	      else 
		{
		  if (keys_single)
		    {
		      car(ecdr(sc->code)) = sc->CASE_SIMPLEST;
		      for (x = cdr(sc->code); is_not_null(x); x = cdr(x))
			fcdr(x) = caaar(x);
		    }
		  else car(ecdr(sc->code)) = sc->CASE_SIMPLER;
		}
	    }
	  else 
	    {
#if (!WITH_GMP)
	      /* OP_CASE_INT assumes non-gmp ints for the selector */
	      if ((is_symbol(car(sc->code))) &&
		  (!has_feed_to) &&
		  (keys_int) &&
		  (keys_single))
		car(ecdr(sc->code)) = sc->CASE_INT;
	      else 
#endif
		car(ecdr(sc->code)) = sc->CASE_UNCHECKED;
	    }
	}
    }

  return(sc->code);
}


#if WITH_OPTIMIZATION
static bool is_letx_safe(s7_pointer p)
{
  int op;
  if (!is_optimized(p))
    return(false);
  op = optimize_data(p);
  return((op == HOP_SAFE_C_C) ||
	 (op == HOP_SAFE_C_S) ||
	 (op == HOP_SAFE_C_SS) ||
	 (op == HOP_SAFE_C_SC) ||
	 (op == HOP_SAFE_C_CS) ||
	 (op == HOP_SAFE_C_opSq));
}
#endif


static s7_pointer check_let(s7_scheme *sc)
{
  s7_pointer x;
  bool named_let;
  int vars;
  /* fprintf(stderr, "let: %s\n", DISPLAY_80(sc->code)); */

  if (!is_pair(sc->code))               /* (let . 1) */
    {
      if (is_null(sc->code))            /* (let) */
	return(eval_error(sc, "let has no variables or body: ~A", sc->code));
      return(eval_error(sc, "let form is an improper list? ~A", sc->code));
    }
  
  if (!is_pair(cdr(sc->code)))          /* (let () ) */
    return(eval_error(sc, "let has no body: ~A", sc->code));

  if ((!s7_is_list(sc, car(sc->code))) && /* (let 1 ...) */
      (!is_symbol(car(sc->code))))
    return(eval_error(sc, "let variable list is messed up or missing: ~A", sc->code));
  
  /* we accept these (other schemes complain, but I can't see why -- a no-op is the user's business!):
   *   (let () (define (hi) (+ 1 2)))
   *   (let () (begin (define x 3)))
   *   (let () 3 (begin (define x 3)))
   *   (let () (define x 3))
   *   (let () (if #t (define (x) 3)))
   *
   * similar cases:
   *   (case 0 ((0) (define (x) 3) (x)))
   *   (cond (0 (define (x) 3) (x)))
   *   (and (define (x) x) 1)
   *   (begin (define (x y) y) (x (define (x y) y)))
   *   (if (define (x) 1) 2 3)
   *   (do () ((define (x) 1) (define (y) 2)))
   *
   * but we can get some humorous results: 
   *   (let ((x (lambda () 3))) (if (define (x) 4) (x) 0)) -> 4
   */
  
  named_let = (is_symbol(car(sc->code)));

  if (named_let)
    {
      if ((!s7_is_list(sc, cadr(sc->code))) ||  /* (let hi #t) */
	  (is_null(cddr(sc->code))))           /* (let hi ()) */
	return(eval_error(sc, "named let variable list is messed up or missing: ~A", sc->code));

      if (is_immutable(car(sc->code)))
	return(s7_error(sc, sc->WRONG_TYPE_ARG,
			list_2(sc, make_protected_string(sc, "can't bind an immutable object: ~S"), sc->code)));
      set_local(car(sc->code));
    }
  
  for (vars = 0, x = ((named_let) ? cadr(sc->code) : car(sc->code)); is_pair(x); vars++, x = cdr(x))
    {
      s7_pointer y, carx;
      
      carx = car(x);
      
      if (!is_pair(carx))                /* (let ((x)) ...) or (let ((x 1) . (y 2)) ...) */
	return(eval_error(sc, "let variable declaration, but no value?: ~A", x));
      
      if (!(is_pair(cdr(carx))))         /* (let ((x . 1))...) */
	return(eval_error(sc, "let variable declaration is not a proper list?: ~A", x));
      
      if (is_not_null(cddr(carx)))       /* (let ((x 1 2 3)) ...) */
	return(eval_error(sc, "let variable declaration has more than one value?: ~A", x));
      
      /* currently if the extra value involves a read error, we get a kind of panicky-looking message:
       *   (let ((x . 2 . 3)) x)
       *   ;let variable declaration has more than one value?: (x error error "stray dot?: ...  ((x . 2 . 3)) x) ..")
       */
      
      if (!(is_symbol(car(carx))))
	return(eval_error(sc, "bad variable ~S in let bindings", carx));
      
      if (is_immutable(car(carx)))
	return(s7_error(sc, sc->WRONG_TYPE_ARG,
			list_2(sc, make_protected_string(sc, "can't bind an immutable object: ~S"), x)));

      /* check for name collisions -- not sure this is required by Scheme */
      for (y = is_symbol(car(sc->code)) ? cadr(sc->code) : car(sc->code); y != x; y = cdr(y))
	if (car(carx) == caar(y))
	  return(eval_error(sc, "duplicate identifier in let: ~A", carx));

      set_local(car(carx));
    }
  
  /* we accept (let ((:hi 1)) :hi)
   *           (let ('1) quote) [guile accepts this]
   */
  
  if (is_not_null(x))                  /* (let* ((a 1) . b) a) */
    return(eval_error(sc, "let var list improper?: ~A", sc->code));

  if ((is_overlaid(sc->code)) &&
      (cdr(ecdr(sc->code)) == sc->code))
    {
      if (named_let)
	{
	  if (is_null(cadr(sc->code)))
	    car(ecdr(sc->code)) = sc->NAMED_LET_NO_VARS;
	  else car(ecdr(sc->code)) = sc->NAMED_LET;
	}
      else
	{
	  if ((is_pair(cdr(sc->code))) &&
	      (is_pair(cadr(sc->code))) &&
	      (is_null(cddr(sc->code))))
	    set_one_liner(cdr(sc->code));

	  if (is_null(car(sc->code)))
	    car(ecdr(sc->code)) = sc->LET_NO_VARS;
	  else 
	    {
	      if (is_null(cdar(sc->code)))
		{
		  s7_pointer binding;
		  binding = caar(sc->code);
		  if (symbol_has_accessor(car(binding)))
		    car(ecdr(sc->code)) = sc->LET_UNCHECKED;
		  else
		    {
		      if (is_pair(cadr(binding)))
			{
			  if ((car(cadr(binding)) == sc->QUOTE) || 
			      (car(cadr(binding)) == sc->QUOTE_UNCHECKED))
			    car(ecdr(sc->code)) = sc->LET_Q;
			  else
			    {
#if WITH_OPTIMIZATION
			      if ((is_optimized(cadr(binding))) &&
				  (is_h_safe_c_s(cadr(binding))))
				{
				  if (is_one_liner(cdr(sc->code)))
				    {
				      car(ecdr(sc->code)) = sc->LET_R_P;
				      if (is_syntactic(caadr(sc->code)))
					{
					  if (c_call(cadr(binding)) == g_read_char)
					    car(ecdr(sc->code)) = sc->LET_READ_CHAR_P;
					  else 
					    {
					      if (c_call(cadr(binding)) == g_car)
						car(ecdr(sc->code)) = sc->LET_CAR_P;
					    }
					}
				    }
				  else car(ecdr(sc->code)) = sc->LET_R;
				  fcdr(sc->code) = cadr(cadr(binding));
				}
			      else 
				{
				  fcdr(sc->code) = cadr(binding);
				  if (is_one_liner(cdr(sc->code)))
				    {
				      /* (define (hi a b) (let ((x (a b))) (char=? x #\a))) */
				      if (is_letx_safe(cadr(binding)))
					car(ecdr(sc->code)) = sc->LET_OX_P;
				      else 
					{
					  if ((is_optimized(cadr(binding))) &&
					      (optimize_data_match(cadr(binding), OP_UNKNOWN_S)))
					    {
					      car(ecdr(sc->code)) = sc->LET_UNKNOWN_S_P;
					      fcdr(sc->code) = caar(sc->code);
					    }
					  else car(ecdr(sc->code)) = sc->LET_O_P;
					}
				    }
				  else car(ecdr(sc->code)) = sc->LET_O;
				}
#else
			      car(ecdr(sc->code)) = sc->LET_UNCHECKED;
#endif
			    }
			}
		      else
			{
			  if (is_symbol(cadaar(sc->code)))
			    {
			      if (is_one_liner(cdr(sc->code)))
				car(ecdr(sc->code)) = sc->LET_S_P;
			      else car(ecdr(sc->code)) = sc->LET_S;
			    }
			  else 
			    {
			      if (is_one_liner(cdr(sc->code)))
				{
#if WITH_OPTIMIZATION
				  if ((!is_syntactic(car(cadr(sc->code)))) &&
				      (is_optimized(cadr(sc->code))))
				    car(ecdr(sc->code)) = sc->LET_C_D;
				  else
#endif
				  car(ecdr(sc->code)) = sc->LET_C_P;
				}
			      else car(ecdr(sc->code)) = sc->LET_C;
			    }
			}
		    }
		}
	      else 
		{
		  if (vars < GC_TRIGGER_SIZE)
		    {
		      s7_pointer p, x, op;
		      /* car(ecdr(sc->code)) = sc->LET_UNCHECKED; */
		      op = sc->NIL;
		      for (p = car(sc->code); is_pair(p); p = cdr(p))
			{
			  x = car(p);
			  if (symbol_has_accessor(car(x)))
			    {
			      op = sc->LET_UNCHECKED;
			      break;
			    }
			  if (is_pair(cadr(x)))
			    {
			      if ((car(cadr(x)) == sc->QUOTE) || 
				  (car(cadr(x)) == sc->QUOTE_UNCHECKED))
				{
				  if ((op == sc->NIL) || (op == sc->LET_ALL_Q))
				    op = sc->LET_ALL_Q;
				  else op = sc->LET_ALL_G;
				}
			      else
				{
#if WITH_OPTIMIZATION
				  if ((is_optimized(cadr(x))) &&
				      (is_h_safe_c_s(cadr(x))))
				    {
				      if ((op == sc->NIL) || (op == sc->LET_ALL_R))
					op = sc->LET_ALL_R;
				      else op = sc->LET_ALL_G;
				    }
				  else 
#endif
				    {
				      op = sc->LET_UNCHECKED;
				      break;
				    }
				}
			    }
			  else
			    {
			      if (is_symbol(cadr(x)))
				{
				  if ((op == sc->NIL) || (op == sc->LET_ALL_S))
				    op = sc->LET_ALL_S;
				  else op = sc->LET_ALL_G;
				}
			      else
				{
				  if ((op == sc->NIL) || (op == sc->LET_ALL_C))
				    op = sc->LET_ALL_C;
				  else op = sc->LET_ALL_G;
				}
			    }
			}
		      car(ecdr(sc->code)) = op;
		    }
		  else car(ecdr(sc->code)) = sc->LET_UNCHECKED;
		}
	    }
	}
    }
  /*
  if (car(ecdr(sc->code)) == sc->LET_UNCHECKED)
    fprintf(stderr, "%s\n", DISPLAY_80(sc->code));
  */
  return(sc->code);
}


static s7_pointer check_let_star(s7_scheme *sc)
{
  s7_pointer y;
  if (!is_pair(sc->code))                    /* (let* . 1) */
    return(eval_error(sc, "let* variable list is messed up: ~A", sc->code));
  
  if ((!is_pair(cdr(sc->code))) ||           /* (let*) */
      (!s7_is_list(sc, car(sc->code))))      /* (let* 1 ...), also there's no named let* */
    {
      if (is_symbol(car(sc->code)))
	return(eval_error(sc, "there is no named let*: ~A", sc->code));
      return(eval_error(sc, "let* variable list is messed up: ~A", sc->code));
    }
  
  if ((!is_null(car(sc->code))) &&
      ((!is_pair(car(sc->code))) ||            /* (let* x ... ) */
       (!is_pair(caar(sc->code))) ||           /* (let* (x) ...) */
       (!is_pair(cdaar(sc->code)))))            /* (let* ((x . 1)) ...) */
    return(eval_error(sc, "let* variable declaration value is missing: ~A", sc->code));
  
  for (y = car(sc->code); is_pair(y); y = cdr(y))
    {
      s7_pointer x, z;
      x = car(y);
      if (!(is_symbol(car(x))))     /* (let* ((3 1)) 1) */
	return(eval_error(sc, "bad variable ~S in let* bindings", x));

      z = car(x);
      if (is_immutable(z))
	return(s7_error(sc, sc->WRONG_TYPE_ARG,
			list_2(sc, make_protected_string(sc, "can't bind an immutable object: ~S"), x)));

      if (!is_pair(x))                 /* (let* ((x)) ...) */
	return(eval_error(sc, "let* variable declaration, but no value?: ~A", x));
      
      if (!(is_pair(cdr(x))))          /* (let* ((x . 1))...) */
	return(eval_error(sc, "let* variable declaration is not a proper list?: ~A", x));
      
      if (is_not_null(cddr(x)))        /* (let* ((x 1 2 3)) ...) */
	return(eval_error(sc, "let* variable declaration has more than one value?: ~A", x));
      
      x = cdr(y);
      if (is_pair(x)) 
	{ 
	  if (!is_pair(car(x)))             /* (let* ((x -1) 2) 3) */
	    return(eval_error(sc, "let* variable/binding is ~S?", car(x)));
	  
	  if (!is_pair(cdar(x)))            /* (let* ((a 1) (b . 2)) ...) */
	    return(eval_error(sc, "let* variable list is messed up? ~A", x));
	}
      else
	{
	  if (is_not_null(x))               /* (let* ((a 1) . b) a) */
	    return(eval_error(sc, "let* var list improper?: ~A", x));
	}
      set_local(z);
    }

  if ((is_overlaid(sc->code)) &&
      (cdr(ecdr(sc->code)) == sc->code))
    {
      car(ecdr(sc->code)) = sc->LET_STAR_UNCHECKED;
      if (is_null(car(sc->code)))
	car(ecdr(sc->code)) = sc->LET_NO_VARS; /* (let* () ...) */
      else 
	{
	  if (is_null(cdar(sc->code)))
	    {
	      s7_pointer binding;
	      binding = caar(sc->code);
	      if (symbol_has_accessor(car(binding)))
		car(ecdr(sc->code)) = sc->LET_STAR_UNCHECKED;
	      else
		{
		  if (is_pair(cadr(binding)))
		    {
		      if ((car(cadr(binding)) == sc->QUOTE) || 
			  (car(cadr(binding)) == sc->QUOTE_UNCHECKED))
			car(ecdr(sc->code)) = sc->LET_Q;
		      else
			{
#if WITH_OPTIMIZATION
			  if ((is_optimized(cadr(binding))) &&
			      (is_h_safe_c_s(cadr(binding))))
			    {
			      if (is_one_liner(cdr(sc->code)))
				{
				  car(ecdr(sc->code)) = sc->LET_R_P;
				  if (is_syntactic(caadr(sc->code)))
				    {
				      if (c_call(cadr(binding)) == g_read_char)
					car(ecdr(sc->code)) = sc->LET_READ_CHAR_P;
				      else 
					{
					  if (c_call(cadr(binding)) == g_car)
					    car(ecdr(sc->code)) = sc->LET_CAR_P;
					}
				    }
				}
			      else car(ecdr(sc->code)) = sc->LET_R;
			      fcdr(sc->code) = cadr(cadr(binding));
			    }
			  else 
			    {
			      fcdr(sc->code) = cadr(binding);
			      if (is_one_liner(cdr(sc->code)))
				{
				  if (is_letx_safe(cadr(binding)))
				    car(ecdr(sc->code)) = sc->LET_OX_P;
				  else
				    {
				      if ((is_optimized(cadr(binding))) &&
					  (optimize_data_match(cadr(binding), OP_UNKNOWN_S)))
					{
					  car(ecdr(sc->code)) = sc->LET_UNKNOWN_S_P;
					  fcdr(sc->code) = caar(sc->code);
					}
				      else car(ecdr(sc->code)) = sc->LET_O_P;
				    }
				}
			      else car(ecdr(sc->code)) = sc->LET_O;
			    }
#else
			  car(ecdr(sc->code)) = sc->LET_STAR_UNCHECKED;
#endif
			}
		    }
		  else
		    {
		      if (is_symbol(cadaar(sc->code)))
			{
			  if (is_one_liner(cdr(sc->code)))
			    car(ecdr(sc->code)) = sc->LET_S_P;
			  else car(ecdr(sc->code)) = sc->LET_S;
			}
		      else 
			{
			  if (is_one_liner(cdr(sc->code)))
			    {
#if WITH_OPTIMIZATION
			      if ((!is_syntactic(car(cadr(sc->code)))) &&
				  (is_optimized(cadr(sc->code))))
				car(ecdr(sc->code)) = sc->LET_C_D;
			      else
#endif
				car(ecdr(sc->code)) = sc->LET_C_P;
			    }
			  else car(ecdr(sc->code)) = sc->LET_C;
			}
		    }
		}
	    }
	  else 
	    {
	      car(ecdr(sc->code)) = sc->LET_STAR_UNCHECKED;
	    }
	}
    }

  return(sc->code);
}


static s7_pointer check_letrec(s7_scheme *sc)
{
  s7_pointer x;
  if ((!is_pair(sc->code)) ||                 /* (letrec . 1) */
      (!is_pair(cdr(sc->code))) ||            /* (letrec) */
      (!s7_is_list(sc, car(sc->code))))       /* (letrec 1 ...) */
    return(eval_error(sc, "letrec variable list is messed up: ~A", sc->code));
  
  for (x = car(sc->code); is_not_null(x); x = cdr(x))
    {
      if (!is_pair(x))                    /* (letrec ((a 1) . 2) ...) */
	return(eval_error(sc, "improper list of letrec variables? ~A", sc->code));
      
      if ((!is_pair(car(x))) ||           /* (letrec (1 2) #t) */
	  (!(is_symbol(caar(x)))))
	return(eval_error(sc, "bad variable ~S in letrec bindings", car(x)));
      
      if (is_immutable(caar(x)))
	return(s7_error(sc, sc->WRONG_TYPE_ARG,
			list_2(sc, make_protected_string(sc, "can't bind an immutable object: ~S"), x)));

      if (!is_pair(cdar(x)))              /* (letrec ((x . 1))...) */
	{
	  if (is_null(cdar(x)))               /* (letrec ((x)) x) -- perhaps this is legal? */
	    return(eval_error(sc, "letrec variable declaration has no value?: ~A", car(x)));
	  return(eval_error(sc, "letrec variable declaration is not a proper list?: ~A", car(x)));
	}
      if (is_not_null(cddar(x)))          /* (letrec ((x 1 2 3)) ...) */
	return(eval_error(sc, "letrec variable declaration has more than one value?: ~A", car(x)));

      set_local(caar(x));
    }

  if ((is_overlaid(sc->code)) &&
      (cdr(ecdr(sc->code)) == sc->code))
    {
      car(ecdr(sc->code)) = sc->LETREC_UNCHECKED;
    }

  return(sc->code);
}


static s7_pointer check_quote(s7_scheme *sc)
{
  if (!is_pair(sc->code))                    /* (quote . -1) */
    {
      if (is_null(sc->code))
	return(eval_error(sc, "quote: not enough arguments: ~A", sc->code));
      return(eval_error(sc, "quote: stray dot?: ~A", sc->code));
    }
  if (is_not_null(cdr(sc->code)))             /* (quote . (1 2)) or (quote 1 1) */
    return(eval_error(sc, "quote: too many arguments ~A", sc->code));

  if ((is_overlaid(sc->code)) &&
      (cdr(ecdr(sc->code)) == sc->code))
    {
      car(ecdr(sc->code)) = sc->QUOTE_UNCHECKED;
    }

  return(sc->code);
}


static s7_pointer check_and(s7_scheme *sc)
{
  s7_pointer p;
  bool all_pairs;

  if (is_null(sc->code))
    return(sc->code);

  all_pairs = is_pair(sc->code);
  for (p = sc->code; is_pair(p); p = cdr(p))
    {
      if (!is_pair(car(p)))
	all_pairs = false;
    }

  if (is_not_null(p))                                    /* (and . 1) (and #t . 1) */
    return(eval_error(sc, "and: stray dot?: ~A", sc->code));

  if ((is_overlaid(sc->code)) &&
      (cdr(ecdr(sc->code)) == sc->code))
    {
      if (all_pairs)
	{
#if WITH_OPTIMIZATION
  	  if (sequence_is_safe_for_opteval(sc, sc->code))
 	    {
	      bool opt = true, xopt = true;
	      s7_pointer sym;

 	      p = sc->code;
	      sym = cadar(p);

	      for (; is_pair(p) && ((opt) || (xopt)); p = cdr(p))
		{
		  if ((opt) &&
		      ((!is_safe_c_s(car(p))) ||
		       (sym != cadar(p))))
		    opt = false;
		  if ((xopt) &&
		      (!is_orx_safe(car(p))))
		    xopt = false;
		}

	      if (opt)
		car(ecdr(sc->code)) = sc->SAFE_AND_S;
	      else 
		{
		  if (xopt)
		    car(ecdr(sc->code)) = sc->SAFER_AND_X;
		  else car(ecdr(sc->code)) = sc->AND_P;
		}
 	    }
  	  else
#endif
	    car(ecdr(sc->code)) = sc->AND_P;
	}
      else car(ecdr(sc->code)) = sc->AND_UNCHECKED;
    }
  return(sc->code);
}


/* (define (hi a b) (if (or (pair? a) (= a b)) a))
 */

static s7_pointer check_or(s7_scheme *sc)
{
  s7_pointer p;
  bool all_pairs;

  if (is_null(sc->code))
    return(sc->code);

  all_pairs = is_pair(sc->code);
  for (p = sc->code; is_pair(p); p = cdr(p))
    {
      if (!is_pair(car(p)))
	all_pairs = false;
    }

  if (is_not_null(p))
    return(eval_error(sc, "or: stray dot?: ~A", sc->code));
  
  if ((is_overlaid(sc->code)) &&
      (cdr(ecdr(sc->code)) == sc->code))
    {
      if (all_pairs)
	{
#if WITH_OPTIMIZATION
  	  if (sequence_is_safe_for_opteval(sc, sc->code))
 	    {
	      bool opt = true, hopt = true, xopt = true, h_any = true, copt = true, csopt = true;
	      s7_pointer sym;

 	      p = sc->code;
	      sym = cadar(p);

	      for ( ; is_pair(p) && ((opt) || (hopt) || (xopt) || (h_any)); p = cdr(p))
		{
		  if ((opt) &&
		      ((!is_safe_c_s(car(p))) ||
		       (sym != cadar(p))))
		    opt = false;
		  if ((hopt) &&
		      ((!is_h_safe_c_s(car(p))) ||
		       (sym != cadar(p))))
		    hopt = false;
		  if ((xopt) &&
		      (!is_orx_safe(car(p))))
		    xopt = false;
		  if ((h_any) && 
		      (!is_h_optimized(car(p))))
		    h_any = false;
		  if ((copt) &&
		      (optimize_data(car(p)) != HOP_SAFE_C_C))
		    copt = false;
		  if ((csopt) &&
		      ((sym != cadar(p)) ||
		       (c_call(car(p)) != g_char_equal_s_ic)))
		    csopt = false;
		}

	      if (opt)
		{
		  if (hopt)
		    car(ecdr(sc->code)) = sc->SAFER_OR_S;
		  else car(ecdr(sc->code)) = sc->SAFE_OR_S;
		}
	      else 
		{
		  if (csopt)
		    car(ecdr(sc->code)) = sc->SAFER_OR_CEQ;
		  else
		    {
		      if (copt)
			car(ecdr(sc->code)) = sc->SAFER_OR_C;
		      else
			{
			  if (xopt)
			    car(ecdr(sc->code)) = sc->SAFER_OR_X;
			  else car(ecdr(sc->code)) = sc->OR_P;
			}
		    }
		}
 	    }
  	  else
#endif
	    car(ecdr(sc->code)) = sc->OR_P;
	}
      else car(ecdr(sc->code)) = sc->OR_UNCHECKED;
    }
  return(sc->code);
}


static s7_pointer check_if(s7_scheme *sc)
{
  s7_pointer cdr_code;

  if (!is_pair(sc->code))                               /* (if) or (if . 1) */
    return(eval_error(sc, "(if): if needs at least 2 expressions: ~A", sc->code));
  
  cdr_code = cdr(sc->code);
  if (!is_pair(cdr_code))                          /* (if 1) */
    return(eval_error(sc, "(if ~A): if needs another clause", car(sc->code)));
  
  if (is_pair(cdr(cdr_code)))
    {
      if (is_not_null(cddr(cdr_code)))                   /* (if 1 2 3 4) */
	return(eval_error(sc, "too many clauses for if: ~A", sc->code));
    }
  else
    {
      if (is_not_null(cdr(cdr_code)))                    /* (if 1 2 . 3) */
	return(eval_error(sc, "if: ~A has improper list?", sc->code));
    }

  if ((is_overlaid(sc->code)) &&
      (cdr(ecdr(sc->code)) == sc->code))
    {
      s7_pointer test, t, f;

      test = car(sc->code);
      t = car(cdr_code);
      if (is_pair(cdr(cdr_code)))
	f = cadr(cdr_code);
      else f = sc->UNSPECIFIED;

      if (is_pair(test))
	{
	  if (is_pair(t))
	    {
	      if (is_pair(f))
		{
#if WITH_OPTIMIZATION
 		  if (is_optimized(test))
		    {
		      if (optimize_data(test) == HOP_SAFE_C_C)
			{
			  if (c_call(test) == g_char_equal_s_ic)
			    {
			      car(ecdr(sc->code)) = sc->SAFE_IF_CEQ_P_P;
			      fcdr(sc->code) = f;
			    }
			  else car(ecdr(sc->code)) = sc->SAFE_IF_CC_P_P;
			}
		      else 
			{
			  if (is_h_safe_c_s(test))
			    {
			      if ((car(t) == sc->QUOTE) || 
				  (car(t) == sc->QUOTE_UNCHECKED))
				{
				  car(ecdr(sc->code)) = sc->SAFE_IF_CS_Q_P;
				  fcdr(sc->code) = cadr(test);
				}
			      else 
				{
				  if (optimize_data(test) == HOP_SAFE_IS_PAIR_S)
				    {
				      car(ecdr(sc->code)) = sc->SAFE_IF_IS_PAIR_P_P;
				      fcdr(sc->code) = cadr(test);
				    }
				  else
				    {
				      if (optimize_data(test) == HOP_SAFE_IS_SYMBOL_S)
					car(ecdr(sc->code)) = sc->SAFE_IF_IS_SYMBOL_P_P;
				      else
					{
					  if ((c_call(test) == g_is_eof_object) &&
					      (is_syntactic(car(f))))
					    car(ecdr(sc->code)) = sc->SAFE_IF_IS_EOF_P_P;
					  else car(ecdr(sc->code)) = sc->SAFE_IF_CS_P_P;
					}
				      fcdr(sc->code) = cadr(test);
				    }
				}
			    }
			  else
			    {
			      if (optimize_data(test) == HOP_SAFE_C_SQ)
				{
				  car(ecdr(sc->code)) = sc->SAFE_IF_CSQ_P_P;
				  fcdr(sc->code) = cadr(caddr(test));
				}
			      else
				{
				  if (optimize_data(test) == HOP_SAFE_C_SS)
				    {
				      car(ecdr(sc->code)) = sc->SAFE_IF_CSS_P_P;
				      fcdr(sc->code) = caddr(test);
				    }
				  else
				    {
				      if (optimize_data(test) == HOP_SAFE_C_SC)
					{
					  car(ecdr(sc->code)) = sc->SAFE_IF_CSC_P_P;
					  fcdr(sc->code) = caddr(test);
					}
				      else
					{
					  if ((is_optimized(t)) &&
					      (is_optimized(f)))
					    car(ecdr(sc->code)) = sc->SAFE_IF2;
					  else car(ecdr(sc->code)) = sc->IF_O_P_P;
					}
				    }
				}
			    }
			}
		    }
		  else
#endif
		    {
		      if (is_syntactic(car(test)))
			{
			  opcode_t new_op;
			  car(ecdr(sc->code)) = sc->IF_B_P_P;

			  if (syntax_opcode(car(test)) == OP_AND)
			    {
			      s7_pointer old_code;
			      old_code = sc->code;
			      sc->code = cdr(test);
			      check_and(sc);
			      new_op = syntax_opcode(car(test));
			      sc->code = old_code;
			      if (new_op == OP_AND_P)
				car(ecdr(sc->code)) = sc->IF_ANDP_P_P;
			    }
			  else
			    {
			      if (syntax_opcode(car(test)) == OP_OR)
				{
				  s7_pointer old_code;
				  old_code = sc->code;
				  sc->code = cdr(test);
				  check_or(sc);
				  new_op = syntax_opcode(car(test));
				  sc->code = old_code;
				      
				  if (new_op == OP_OR_P)
				    car(ecdr(sc->code)) = sc->IF_ORP_P_P;
#if WITH_OPTIMIZATION
				  else
				    {
				      if (new_op == OP_SAFER_OR_CEQ)
					{
					  s7_pointer p;
					  car(ecdr(sc->code)) = sc->IF_ORCEQ_P_P;
					  fcdr(sc->code) = caddr(sc->code);
					  for (p = cdar(sc->code); is_pair(p); p = cdr(p))
					    fcdr(p) = caddar(p);
					}
				    }
#endif
				}
			    }
			}
		      else car(ecdr(sc->code)) = sc->IF_P_P_P;
		    }
		}
	      else 
		{
		  if (is_null(cdr(cdr_code)))
		    {
#if WITH_OPTIMIZATION
		      if (is_optimized(test))
			{
			  if (optimize_data(test) == HOP_SAFE_C_C)
			    car(ecdr(sc->code)) = sc->SAFE_IF_CC_P;
			  else 
			    {
			      if (is_h_safe_c_s(test))
				{
				  if (optimize_data(test) == HOP_SAFE_IS_PAIR_S)
				    {
				      car(ecdr(sc->code)) = sc->SAFE_IF_IS_PAIR_P;
				      fcdr(sc->code) = cadr(test);
				    }
				  else
				    {
				      if (optimize_data(test) == HOP_SAFE_IS_SYMBOL_S)
					car(ecdr(sc->code)) = sc->SAFE_IF_IS_SYMBOL_P;
				      else car(ecdr(sc->code)) = sc->SAFE_IF_CS_P; 
				      fcdr(sc->code) = cadr(test);
				    }
				}
			      else
				{
				  if (optimize_data(test) == HOP_SAFE_C_SQ)
				    {
				      car(ecdr(sc->code)) = sc->SAFE_IF_CSQ_P;
				      fcdr(sc->code) = cadr(caddr(test));
				    }
				  else
				    {
				      if (optimize_data(test) == HOP_SAFE_C_SS)
					{
					  car(ecdr(sc->code)) = sc->SAFE_IF_CSS_P;
					  fcdr(sc->code) = caddr(test);
					}
				      else
					{
					  if (optimize_data(test) == HOP_SAFE_C_SC)
					    {
					      car(ecdr(sc->code)) = sc->SAFE_IF_CSC_P;
					      fcdr(sc->code) = caddr(test);
					    }
					  else				
					    {
					      if (optimize_data(test) == HOP_SAFE_C_opSSq)
						{
						  car(ecdr(sc->code)) = sc->SAFE_IF_C_SS_P;
						}
					      else
						{
						  if (is_optimized(t))
						    car(ecdr(sc->code)) = sc->SAFE_IF1;
						  else car(ecdr(sc->code)) = sc->IF_O_P;
						}
					    }
					}
				    }
				}
			    }
			}
		      else 
			{
			  if (is_syntactic(car(test)))
			    {
			      opcode_t new_op;
			      car(ecdr(sc->code)) = sc->IF_B_P;

			      if (syntax_opcode(car(test)) == OP_AND)
				{
				  s7_pointer old_code;
				  old_code = sc->code;
				  sc->code = cdr(test);
				  check_and(sc);
				  new_op = syntax_opcode(car(test));
				  sc->code = old_code;

				  if (new_op == OP_SAFER_AND_X)
				    car(ecdr(sc->code)) = sc->SAFE_IF_ANDX_P;
				  else
				    {
				      if (new_op == OP_AND_P)
					car(ecdr(sc->code)) = sc->IF_ANDP_P;
				    }
				}
			      else
				{
				  if (syntax_opcode(car(test)) == OP_OR)
				    {
				      s7_pointer old_code;
				      old_code = sc->code;
				      sc->code = cdr(test);
				      check_or(sc);
				      new_op = syntax_opcode(car(test));
				      sc->code = old_code;
				      
				      if (new_op == OP_SAFER_OR_X)
					car(ecdr(sc->code)) = sc->SAFE_IF_ORX_P;
				      else
					{
					  if (new_op == OP_OR_P)
					    car(ecdr(sc->code)) = sc->IF_ORP_P;
					  else
					    {
					      if (new_op == OP_SAFER_OR_CEQ)
						{
						  s7_pointer p;
						  car(ecdr(sc->code)) = sc->IF_ORCEQ_P;
						  fcdr(sc->code) = cdar(sc->code);
						  for (p = cdar(sc->code); is_pair(p); p = cdr(p))
						    fcdr(p) = caddar(p);
						}
					    }
					}
				    }
				}
			    }
			  else car(ecdr(sc->code)) = sc->IF_P_P;
			}
		    }
#else
		      {
			if (is_syntactic(car(test)))
			  car(ecdr(sc->code)) = sc->IF_B_P;
			else car(ecdr(sc->code)) = sc->IF_P_P;
		      }
		    }
#endif
		  else
		    {
#if WITH_OPTIMIZATION
		      if ((is_optimized(test)) &&
			  (is_h_safe_c_s(test)))
			{
			  if (optimize_data(test) == HOP_SAFE_IS_PAIR_S)
			    {
			      car(ecdr(sc->code)) = sc->SAFE_IF_IS_PAIR_P_X;
			      fcdr(sc->code) = cadr(test);
			    }
			  else
			    {
			      if (optimize_data(test) == HOP_SAFE_IS_SYMBOL_S)
				car(ecdr(sc->code)) = sc->SAFE_IF_IS_SYMBOL_P_X;
			      else car(ecdr(sc->code)) = sc->SAFE_IF_CS_P_X;
			    }
			}
		      else
#endif
		      car(ecdr(sc->code)) = sc->IF_P_P_X;
		    }
		}
	    }
	  else 
	    {

	      if (is_pair(f))
		{
#if WITH_OPTIMIZATION
		  if (is_optimized(test))
		    {
		      if (optimize_data(test) == HOP_SAFE_C_C)
			car(ecdr(sc->code)) = sc->SAFE_IF_CC_X_P;
		      else
			{
			  if (is_h_safe_c_s(test))
			    {
			      car(ecdr(sc->code)) = sc->SAFE_IF_CS_X_P;
			      fcdr(sc->code) = cadr(test);
			    }
			  else 
			    {
			      if (optimize_data(test) == HOP_SAFE_C_SS)
				{
				  car(ecdr(sc->code)) = sc->SAFE_IF_CSS_X_P;
				  fcdr(sc->code) = caddr(test);
				}
			      else
				{
				  if (optimize_data(test) == HOP_SAFE_C_SC)
				    {
				      if (t == cadr(test))
					car(ecdr(sc->code)) = sc->SAFE_IF_CSC_X_P_A;
				      else car(ecdr(sc->code)) = sc->SAFE_IF_CSC_X_P;
				      fcdr(sc->code) = caddr(test);
				    }
				  else car(ecdr(sc->code)) = sc->IF_P_X_P;
				}
			    }
			}
		    }
		  else
#endif
		    car(ecdr(sc->code)) = sc->IF_P_X_P;
		}
	      else 
		{

		  if (is_null(cdr(cdr_code)))
		    {
#if WITH_OPTIMIZATION
		      if (is_optimized(test))
			{
			  /* fprintf(stderr, "px: %s %s\n", (is_optimized(test)) ? opt_names[optimize_data(test)] : "unopt", DISPLAY_80(test)); */
			  car(ecdr(sc->code)) = sc->IF_P_X;
			}
		      else
#endif
			car(ecdr(sc->code)) = sc->IF_P_X;
		    }
		  else car(ecdr(sc->code)) = sc->IF_P_X_X;
		}
	    }
	}
      else /* test is symbol or constant */
	{
	  if (is_pair(t))
	    {
	      if (is_pair(f))
		car(ecdr(sc->code)) = sc->IF_X_P_P;
	      else 
		{
		  if (is_null(cdr(cdr_code)))
		    car(ecdr(sc->code)) = sc->IF_X_P;
		  else car(ecdr(sc->code)) = sc->IF_X_P_X;
		}
	    }
	  else 
	    {
	      if (is_pair(f))
		car(ecdr(sc->code)) = sc->IF_X_X_P;
	      else 
		{
		  if (is_null(cdr(cdr_code)))
		    car(ecdr(sc->code)) = sc->IF_X_X;
		  else car(ecdr(sc->code)) = sc->IF_X_X_X;
		}
	    }
	}
    }

  return(sc->code);
}


#if WITH_OPTIMIZATION
static bool is_safe_arg_list(s7_scheme *sc, s7_pointer lst)
{
  s7_pointer p;
  for (p = lst; is_pair(p); p = cdr(p))
    if ((is_symbol(car(p))) &&
	(symbol_has_accessor(car(p)))) /* this checks the extended portion, so it better be a symbol */
      return(false);
  return(true);
}
#endif


static s7_pointer check_define(s7_scheme *sc)
{
  s7_pointer x;
  if (!is_pair(sc->code))
    return(eval_error_with_name(sc, "~A: nothing to define? ~A", sc->code));   /* (define) */
  
  if (!is_pair(cdr(sc->code)))
    {
      if (is_null(cdr(sc->code)))
	return(eval_error_with_name(sc, "~A: no value? ~A", sc->code));        /* (define var) */
      return(eval_error_with_name(sc, "~A: bad form? ~A", sc->code));          /* (define var . 1) */
    }
  if (!is_pair(car(sc->code)))
    {
      if (is_not_null(cddr(sc->code)))                                           /* (define var 1 . 2) */
	return(eval_error_with_name(sc, "~A: more than 1 value? ~A", sc->code)); /* (define var 1 2) */
      if (sc->op == OP_DEFINE_STAR)
	return(eval_error(sc, "define* is restricted to functions: (define* ~{~S~^ ~})", sc->code));
      
      x = car(sc->code);
      if (!is_symbol(x))                                                        /* (define 3 a) */
	return(eval_error_with_name(sc, "~A: define a non-symbol? ~S", x));
      if (is_keyword(x))                                                        /* (define :hi 1) */
	return(eval_error_with_name(sc, "~A ~A: keywords are constants", x));
      if (is_syntactic(x))                                                      /* (define and a) */
	return(eval_error_with_name(sc, "~A ~A: syntactic keywords tend to behave badly if redefined", x));
    }
  else
    {
      int len;
      x = caar(sc->code);
      if (!is_symbol(x))                                                        /* (define (3 a) a) */
	return(eval_error_with_name(sc, "~A: define a non-symbol? ~S", x));
      if (is_syntactic(x))                                                      /* (define (and a) a) */
	return(eval_error_with_name(sc, "~A ~A: syntactic keywords tend to behave badly if redefined", x));
      
      if (sc->op == OP_DEFINE_STAR)
	check_lambda_star_args(sc, cdar(sc->code));
      else check_lambda_args(sc, cdar(sc->code));

      len = s7_list_length(sc, cdr(sc->code));
      if (len < 0)                                                               /* (define (hi) 1 . 2) */
	return(eval_error_with_name(sc, "~A: function body messed up, ~A", sc->code));

      if ((len == 1) &&
	  (is_pair(cadr(sc->code))))
	set_one_liner(cdr(sc->code));

#if WITH_OPTIMIZATION
      if (len > 0)  /* i.e. not circular */
	{
	  bool bad_set = false;
	  /* fprintf(stderr, "optimize: %s %s\n", DISPLAY(x), DISPLAY(sc->code)); */
	  optimize(sc, cdr(sc->code), 1, collect_collisions(sc, cdar(sc->code), list_1(sc, x)));
	  
	  /* if the body is safe, we can optimize the calling sequence */
	  if ((sc->op == OP_DEFINE) &&
	      (is_proper_list(sc, cdar(sc->code))) &&
	      (is_safe_arg_list(sc, cdar(sc->code))) &&
	      (body_is_safe(sc, caar(sc->code), cdr(sc->code), true, &bad_set)))
	    {
	      /* (define (hi a) (+ a 1) (hi (- a 1))) */
	      /* fprintf(stderr, "%s is safe\n", DISPLAY_80(sc->code)); */
	      set_safe_closure(cdr(sc->code));
	    }
	}
#endif
    }

  if ((is_overlaid(sc->code)) &&
      (cdr(ecdr(sc->code)) == sc->code))
    {
      if (sc->op == OP_DEFINE)
	car(ecdr(sc->code)) = sc->DEFINE_UNCHECKED;
      else car(ecdr(sc->code)) = sc->DEFINE_STAR_UNCHECKED;
    }

  return(sc->code);
}


static s7_pointer check_set(s7_scheme *sc)
{
  if (!is_pair(sc->code))
    {
      if (is_null(sc->code))                                             /* (set!) */
	return(eval_error(sc, "set!: not enough arguments: ~A", sc->code));
      return(eval_error(sc, "set!: stray dot? ~A", sc->code));           /* (set! . 1) */
    }
  if (!is_pair(cdr(sc->code)))                                
    {
      if (is_null(cdr(sc->code)))                                         /* (set! var) */
	return(eval_error(sc, "set!: not enough arguments: ~A", sc->code));
      return(eval_error(sc, "set!: stray dot? ~A", sc->code));           /* (set! var . 1) */
    }
  if (is_not_null(cddr(sc->code)))                                       /* (set! var 1 2) */
    return(eval_error(sc, "~A: too many arguments to set!", sc->code));
  
  /* cadr (the value) has not yet been evaluated */
  
  if (is_immutable(car(sc->code)))                                       /* (set! pi 3) */
    return(eval_error(sc, "set!: can't alter immutable object: ~S", car(sc->code)));
  
  if (is_pair(car(sc->code)))
    {
      if (is_pair(caar(sc->code)))
	{
	  if (!s7_is_list(sc, cdar(sc->code)))                          /* (set! ('(1 2) . 0) 1) */
	    eval_error(sc, "improper list of args to set!: ~A", sc->code);
	}
      if (!is_proper_list(sc, car(sc->code)))                           /* (set! ("hi" . 1) #\a) or (set! (#(1 2) . 1) 0) */
	eval_error(sc, "set! target is an improper list: (set! ~A ...)", car(sc->code));
    }
  else
    {
      if (!is_symbol(car(sc->code)))                                 /* (set! 12345 1) */
	return(eval_error(sc, "set! can't change ~S", car(sc->code)));
    }

  if ((is_overlaid(sc->code)) &&
      (cdr(ecdr(sc->code)) == sc->code))
    {
      if (is_pair(car(sc->code)))
	{
	  s7_pointer inner, value;
	  inner = car(sc->code);
	  value = cadr(sc->code);
	  if ((is_symbol(car(inner))) &&
	      (is_pair(cdr(inner))) &&
	      (!is_pair(cddr(inner))) &&
	      (!is_pair(cadr(inner))) &&
	      (!is_pair(cddr(sc->code))) &&
	      (!is_pair(value)))
	    car(ecdr(sc->code)) = sc->SET_PAIR;
	  else car(ecdr(sc->code)) = sc->SET_UNCHECKED;
	}
      else car(ecdr(sc->code)) = sc->SET_NORMAL;

      if (is_symbol(car(sc->code)))
	{
	  if (is_symbol(cadr(sc->code)))
	    car(ecdr(sc->code)) = sc->SET_SYMBOL_S;
	  else
	    {
	      if (!is_pair(cadr(sc->code)))
		car(ecdr(sc->code)) = sc->SET_SYMBOL_C;
	      else
		{
		  if ((car(cadr(sc->code)) == sc->QUOTE) || 
		      (car(cadr(sc->code)) == sc->QUOTE_UNCHECKED))
		    car(ecdr(sc->code)) = sc->SET_SYMBOL_Q;
		  else
		    {
		      /* if cadr(cadr) == car, or cdr(cadr) not null and cadr(cadr) == car, and cddr(cadr) == null,
		       *   it's (set! <var> (<op> <var> val)) or (<op> val <var>) or (<op> <var>)
		       *   in the set code, we get the slot as usual, then in case 1 above,
		       *   car(sc->T2_1) = slot_value(slot), car(sc->T2_2) = increment, call <op>, set slot_value(slot)
		       *
		       * this can be done in all combined cases where a symbol is repeated (do in particular)
		       */
		      
		      /* (define (hi) (let ((x 1)) (set! x (+ x 1)))) */
		      car(ecdr(sc->code)) = sc->SET_SYMBOL_P;
#if WITH_OPTIMIZATION
		      if ((!symbol_has_accessor(car(sc->code))) &&
			  /* (!is_global(car(sc->code))) && */ /* we use find_symbol to get the slot, so this should be ok */
			  (is_optimized(cadr(sc->code))) &&
			  (!is_unsafe(cadr(sc->code))) &&
			  (is_not_null(cdr(cadr(sc->code)))))               /* (set! x (y)) */
			{
			  /* the "!is_unsafe" check is needed in case OP_UNKNOWN_SC gets here --
			   *   it (formerly) did not clear or set the e|fcdr fields, so they may contain
			   *   garbage which can confuse the check below.  I now 
			   *   clear ecdr.  (We can get here with an unknown op if previously
			   *   + was redefined locally, for example, so the optimizer wants to
			   *   delay the decision until runtime).
			   */
			  if (is_not_null(cddr(cadr(sc->code))))
			    {
			      if ((caddr(cadr(sc->code)) == small_int(1)) &&
				  (cadr(cadr(sc->code)) == car(sc->code)))
				{
				  if ((ecdr(cadr(sc->code)) == add_s1) ||
				      (ecdr(cadr(sc->code)) == add_cs1))
				    car(ecdr(sc->code)) = sc->INCREMENT_1;
				  else 
				    {
				      if ((ecdr(cadr(sc->code)) == subtract_s1) ||
					  (ecdr(cadr(sc->code)) == subtract_cs1))
					{
					  /* fprintf(stderr, "%p use decrement for %s %p %p %p %p\n", cadr(sc->code), DISPLAY(sc->code), add_s1, add_cs1, subtract_s1, subtract_cs1); */
					  car(ecdr(sc->code)) = sc->DECREMENT_1;
					}
				    }
				}
			      else
				{
				  if ((cadr(cadr(sc->code)) == small_int(1)) &&
				      (caddr(cadr(sc->code)) == car(sc->code)) &&
				      (ecdr(cadr(sc->code)) == add_1s))
				    car(ecdr(sc->code)) = sc->INCREMENT_1;
				  else
				    {
				      if ((car(sc->code) == caddr(cadr(sc->code))) &&
					  (is_symbol(cadr(cadr(sc->code)))) &&
					  (caadr(sc->code) == s7_make_symbol(sc, "cons")))
					{
					  car(ecdr(sc->code)) = sc->SET_CONS;
					  fcdr(sc->code) = cadr(cadr(sc->code));
					}
				    }
				}
			    }
			  else
			    {
			      if ((car(sc->code) == cadr(cadr(sc->code))) &&
				  (caadr(sc->code) == sc->CDR))
				car(ecdr(sc->code)) = sc->SET_CDR;
			      else
				{
				  if (is_h_safe_c_s(cadr(sc->code)))
				    {
				      car(ecdr(sc->code)) = sc->SET_SYMBOL_R;
				      fcdr(sc->code) = cadr(cadr(sc->code));
				    }
				}
			    }
			}
#endif
		    }
		}
	    }
	}
    }

  return(sc->code);
}


#if WITH_OPTIMIZATION
static void initialize_safe_do_1(s7_scheme *sc, s7_pointer tree)
{
  if (is_pair(tree))
    {
      if (has_table(tree))
	clear_table(tree);

      initialize_safe_do_1(sc, car(tree));
      initialize_safe_do_1(sc, cdr(tree));
    }
  else
    {
      if (is_symbol(tree))
	{
	  symbol_accessor_data(tree) = NULL;
	  symbol_op_data(tree) = NULL;
	}
    }
}


static void initialize_safe_do(s7_scheme *sc, s7_pointer tree)
{
  initialize_safe_do_1(sc, tree);
  sc->safe_do_level++;
  finder = find_safe_do_symbol_or_bust;
  safe_do_set_id(sc, frame_id(sc->envir));
}


static void finalize_safe_do(s7_scheme *sc)
{
  sc->safe_do_level--;
  if (sc->safe_do_level <= 0) finder = find_symbol_or_bust;
}


/* to back out of a safe do, clear all tables, set safe do level to 0
 *   add a <0 check wherever we currently safe_do_level--;
 *   finish current non-safe call somehow -- clearing tables will mess this up.
 */


static bool safe_stepper(s7_scheme *sc, s7_pointer expr, s7_pointer vars)
{
  /* for now, just look for stepper as last element of any list
   *    any embedded set is handled by do_is_safe, so we don't need to descend into the depths
   */
  s7_pointer p;
  if (direct_memq(cadr(expr), vars))
    {
#if PRINTING
      fprintf(stderr, "      ;  target memq %s %s\n", DISPLAY(vars), DISPLAY_80(expr));
#endif
      return(false);
    }

  for (p = cdr(expr); is_pair(cdr(p)); p = cdr(p));

  if (is_pair(p))
    {
      if ((is_optimized(p)) &&
	  ((optimize_data(p) & 1) != 0) &&
	  (optimize_data(p) > OP_SAFE_C_C) &&
	  (optimize_data(p) < OP_C_LS))
	return(true);

      if (direct_memq(car(p), vars))
	{
#if PRINTING
	  fprintf(stderr, "      ;  result memq %s %s\n", DISPLAY(vars), DISPLAY_80(expr));
#endif
	  return(false);
	}
    }
  else 
    {
      if (direct_memq(p, vars))
	{
#if PRINTING
	  fprintf(stderr, "      ;  result memq %s %s\n", DISPLAY(vars), DISPLAY_80(expr));
#endif
	  return(false);
	}
    }
  return(true);
}

/* (define (hi) (do ((i 0 (+ i 1))) ((= i 3)) (let ((val 1)) (set! val 2) )))
 * (define (hi) (do ((i 0 (+ i 1))) ((= i 3)) (let ((val 3)) (set! val i) )))
 * (define (hi) (do ((i 0 (+ i 1))) ((= i 3)) (display i)) (newline))
 * (define (hi) (do ((i 0 (+ i 1))) ((= i 10)) (set! i (+ i 1)) (display i)) (newline))
 *
 * (define (hi a) (do ((i 0 (+ i 1))) ((= i a)) (display i)))
 */

static bool do_is_safe(s7_scheme *sc, s7_pointer body, s7_pointer steppers, s7_pointer var_list, bool *has_set)
{
  /* here any (unsafe?) closure or jumping-op (call/cc) or shadowed variable is trouble
   */
  s7_pointer p;
  /* if (s7_list_length(sc, body) <= 0) return(false); */

#if PRINTING
  fprintf(stderr, "      ;do check: %s\n", DISPLAY_80(body));
#endif

  for (p = body; is_pair(p); p = cdr(p))
    {
      s7_pointer expr;
      expr = car(p);
      if (is_pair(expr))
	{
	  s7_pointer x;
	  x = car(expr);
	  if (is_symbol(x))
	    {
	      if (is_syntactic(x))
		{
		  opcode_t op;
		  s7_pointer func, vars;
		  func = slot_value(global_slot(x));
		  op = (opcode_t)syntax_opcode(func);
		  switch (op)
		    {
		    case OP_QUOTE:
		      break;

		    case OP_LET:
		      if (is_symbol(cadr(expr)))
			{
#if PRINTING
			  fprintf(stderr, "    ;unsafe named_let %s\n", DISPLAY_80(cadr(expr)));
#endif
			  return(false);
			}

		    case OP_LET_STAR:
		    case OP_LETREC:
		    case OP_DO:
		      for (vars = cadr(expr); is_pair(vars); vars = cdr(vars))
			{
			  s7_pointer var;
			  var = caar(vars);
			  if ((direct_memq(var, var_list)) ||
			      (direct_memq(var, steppers)))
			    {
#if PRINTING
			      fprintf(stderr, "      ;stepper renamed %s\n", DISPLAY_80(cadr(expr)));
#endif
			      return(false);
			    }
			  var_list = cons(sc, var, var_list);
			  sc->x = var_list;
			  if ((is_pair(cdar(vars))) &&
			      (!do_is_safe(sc, cdar(vars), steppers, var_list, has_set)))
			    {
#if PRINTING
			      fprintf(stderr, "    ;%s unsafe\n", DISPLAY_80(cdar(vars)));
#endif
			      return(false);
			    }
			}

		      if (!do_is_safe(sc, (op == OP_DO) ? cdddr(expr) : cddr(expr), steppers, var_list, has_set))
			{
#if PRINTING
			  fprintf(stderr, "    ;bad %s\n", (op == OP_DO) ? DISPLAY_80(cdddr(expr)) : DISPLAY_80(cddr(expr)));
#endif
			  return(false);
			}
		      break;

		    case OP_SET:
		      if (!direct_memq(cadr(expr), var_list))
			(*has_set) = true;
		      if (!do_is_safe(sc, cddr(expr), steppers, var_list, has_set))
			{
#if PRINTING
			  fprintf(stderr, "    ;unsafe set: %s\n", DISPLAY_80(expr));
#endif
			  return(false);
			}
		      if (!safe_stepper(sc, expr, steppers))
			{
#if PRINTING
			  fprintf(stderr, "    ;unsafe stepper %s\n", DISPLAY_80(expr));
#endif
			  return(false);
			}
		      break;
		      
		    case OP_IF:
		    case OP_COND:
		    case OP_CASE:
		    case OP_AND:
		    case OP_OR:
		    case OP_BEGIN:
		      if (!do_is_safe(sc, cdr(expr), steppers, var_list, has_set))
			{
#if PRINTING
			  fprintf(stderr, "    ;unsafe %s %s\n", real_op_names[op], DISPLAY_80(cdr(expr)));
#endif
			  return(false);
			}
		      break;

		    default:
#if PRINTING
		      fprintf(stderr, "    ;%s unsafe\n", real_op_names[op]);
#endif
		      return(false);
		      break;
		    }
		}
	      else
		{
		  if ((!is_optimized(expr)) ||
		      (is_unsafe(expr)) ||
		      (!do_is_safe(sc, cdr(expr), steppers, var_list, has_set)))
		    {
#if PRINTING
		      fprintf(stderr, "      ;unsafe (%s): %s\n", 
			      (is_optimized(expr)) ? ((is_unsafe(expr)) ? "not safe" : "do is not safe") : "unopt",
			      DISPLAY_80(expr));
#endif
		      return(false);
		    }
		  else
		    {
		      if (is_setter(car(expr)))
			{
			  if (!direct_memq(cadr(expr), var_list))
			    {
#if PRINTING
			      fprintf(stderr, "      ;set: %s is global\n", DISPLAY(cadr(expr)));
#endif
			      (*has_set) = true;
			    }
			  if (!do_is_safe(sc, cddr(expr), steppers, var_list, has_set))
			    {
#if PRINTING
			      fprintf(stderr, "    ; set stepper unsafe: %s\n", DISPLAY_80(cadr(expr)));
#endif
			      return(false);		      
			    }
			  if (!safe_stepper(sc, expr, steppers))
			    {
#if PRINTING
			      fprintf(stderr, "      ;unsafe (stepper expr): %s\n", DISPLAY_80(expr)); 
#endif
			      return(false);
			    }
			}
		    }
		}
	    }
	  else 
	    {
#if PRINTING
	      fprintf(stderr, "      ;unsafe (?): %s\n", DISPLAY_80(expr));
#endif
	      return(false); 
	      /* car(expr) ("x") is not a symbol: ((mus-data loc) chan) for example
	       *   but that's actually safe since it's just in effect vector-ref
	       *   there are several examples in dlocsig: ((group-speakers group) i) etc
	       */
	    }
	}
    }
#if PRINTING
  fprintf(stderr, "        ;%s is ok\n", DISPLAY_80(body));
#endif
  return(true);
}
#endif


#if WITH_OPTIMIZATION
static bool is_init_dox_safe(s7_scheme *sc, s7_pointer p)
{
  int op;
  
  if (!is_pair(p)) return(true);
  if (car(p) == sc->QUOTE) return(true);
  if (!is_optimized(p)) return(false);

  op = optimize_data(p);
  return((op == HOP_SAFE_C_S) ||
	 (op == HOP_SAFE_C_C) ||
	 (op == HOP_SAFE_C_SS) ||
	 (op == HOP_SAFE_C_QS) ||
	 (op == HOP_SAFE_C_S_opCq));
}

static s7_pointer init_dox_eval(s7_scheme *sc, s7_pointer code)
{
  s7_pointer args, val;
  if (is_pair(code))
    {
      if (car(code) == sc->QUOTE)
	return(cadr(code));
      switch (optimize_data(code))
	{
	case HOP_SAFE_C_C:
	  return(c_call(code)(sc, cdr(code)));
	  
	case HOP_SAFE_C_S:
	  car(sc->T1_1) = finder(sc, cadr(code));
	  return(c_call(code)(sc, sc->T1_1));
	  
	case HOP_SAFE_C_SS:
	  args = cdr(code);
	  val = finder(sc, car(args));
	  car(sc->T2_2) = finder(sc, cadr(args));
	  car(sc->T2_1) = val;
	  return(c_call(code)(sc, sc->T2_1));

	case HOP_SAFE_C_QS:
	  args = cdr(code);
	  car(sc->T2_2) = finder(sc, cadr(args));
	  car(sc->T2_1) = cadr(car(args));
	  return(c_call(code)(sc, sc->T2_1));

	case HOP_SAFE_C_S_opCq:
	  args = cdr(code);
	  val = finder(sc, car(args));
	  car(sc->T2_2) = c_call(cadr(args))(sc, cdr(cadr(args))); /* any number of constants here */
	  car(sc->T2_1) = val;
	  return(c_call(code)(sc, sc->T2_1));
      
	default:
	  fprintf(stderr, "oops init dox %s\n", DISPLAY_80(code));
	  return(sc->UNSPECIFIED);
	}
    }
  if (is_symbol(code))
    return(finder(sc, code));
  return(code);
}

static bool is_step_dox_safe(s7_scheme *sc, s7_pointer p)
{
  int op;
  
  if (!is_pair(p)) return(false);
  if (car(p) == sc->QUOTE) return(false);
  if (!is_optimized(p)) return(false);

  op = optimize_data(p);
  return((op == HOP_SAFE_C_S) ||
	 (op == HOP_SAFE_C_C) ||
	 (op == HOP_SAFE_C_SS) ||
	 (op == HOP_SAFE_C_SC) ||
	 (op == HOP_SAFE_C_CS) ||
	 (op == HOP_SAFE_C_QS) ||
	 (op == HOP_SAFE_C_S_opCq));
}

static s7_pointer step_dox_eval(s7_scheme *sc, s7_pointer code)
{
  s7_pointer args, val;
  switch (optimize_data(code))
    {
    case HOP_SAFE_C_C:
      return(c_call(code)(sc, cdr(code)));
      
    case HOP_SAFE_C_S:
      car(sc->T1_1) = finder(sc, cadr(code));
      return(c_call(code)(sc, sc->T1_1));
      
    case HOP_SAFE_C_SC:
      args = cdr(code);
      car(sc->T2_1) = finder(sc, car(args));
      car(sc->T2_2) = cadr(args);
      return(c_call(code)(sc, sc->T2_1));
      
    case HOP_SAFE_C_CS:
      args = cdr(code);
      car(sc->T2_2) = finder(sc, cadr(args));
      car(sc->T2_1) = car(args);
      return(c_call(code)(sc, sc->T2_1));
      
    case HOP_SAFE_C_SS:
      args = cdr(code);
      val = finder(sc, car(args));
      car(sc->T2_2) = finder(sc, cadr(args));
      car(sc->T2_1) = val;
      return(c_call(code)(sc, sc->T2_1));
      
    case HOP_SAFE_C_QS:
      args = cdr(code);
      car(sc->T2_2) = finder(sc, cadr(args));
      car(sc->T2_1) = cadr(car(args));
      return(c_call(code)(sc, sc->T2_1));
      
    case HOP_SAFE_C_S_opCq:
      args = cdr(code);
      val = finder(sc, car(args));
      car(sc->T2_2) = c_call(cadr(args))(sc, cdr(cadr(args))); /* any number of constants here */
      car(sc->T2_1) = val;
      return(c_call(code)(sc, sc->T2_1));
      
    default:
      fprintf(stderr, "oops step dox %s\n", DISPLAY_80(code));
      return(sc->UNSPECIFIED);
    }
}

static bool is_end_dox_safe(s7_scheme *sc, s7_pointer p)
{
  int op;
  
  if (!is_pair(p)) return(false);
  if (car(p) == sc->QUOTE) return(false);
  if (!is_optimized(p)) return(false);

  op = optimize_data(p);
  return((op == HOP_SAFE_C_S) ||
	 (op == HOP_SAFE_C_C) ||
	 (op == HOP_SAFE_C_SS) ||
	 (op == HOP_SAFE_C_SC) ||
	 (op == HOP_SAFE_C_CS) ||
	 (op == HOP_SAFE_NOT_S) ||
	 (op == HOP_SAFE_C_QS) ||
	 (op == HOP_SAFE_C_S_opCq) ||
	 (op == HOP_SAFE_C_opSq_S) ||
	 (op == HOP_SAFE_C_S_opSq));
}

static s7_pointer end_dox_eval(s7_scheme *sc, s7_pointer code)
{
  s7_pointer args, val;
  switch (optimize_data(code))
    {
    case HOP_SAFE_C_C:
      return(c_call(code)(sc, cdr(code)));
      
    case HOP_SAFE_C_S:
      car(sc->T1_1) = finder(sc, cadr(code));
      return(c_call(code)(sc, sc->T1_1));
      
    case HOP_SAFE_C_SC:
      args = cdr(code);
      car(sc->T2_1) = finder(sc, car(args));
      car(sc->T2_2) = cadr(args);
      return(c_call(code)(sc, sc->T2_1));
      
    case HOP_SAFE_C_CS:
      args = cdr(code);
      car(sc->T2_2) = finder(sc, cadr(args));
      car(sc->T2_1) = car(args);
      return(c_call(code)(sc, sc->T2_1));
      
    case HOP_SAFE_C_SS:
      args = cdr(code);
      val = finder(sc, car(args));
      car(sc->T2_2) = finder(sc, cadr(args));
      car(sc->T2_1) = val;
      return(c_call(code)(sc, sc->T2_1));
      
    case HOP_SAFE_C_QS:
      args = cdr(code);
      car(sc->T2_2) = finder(sc, cadr(args));
      car(sc->T2_1) = cadr(car(args));
      return(c_call(code)(sc, sc->T2_1));
      
    case HOP_SAFE_C_S_opCq:
      args = cdr(code);
      val = finder(sc, car(args));
      car(sc->T2_2) = c_call(cadr(args))(sc, cdr(cadr(args))); /* any number of constants here */
      car(sc->T2_1) = val;
      return(c_call(code)(sc, sc->T2_1));
      
    case HOP_SAFE_C_opSq_S:
      args = cdr(code);
      car(sc->T1_1) = finder(sc, cadr(car(args)));
      sc->temp4 = c_call(car(args))(sc, sc->T1_1);
      car(sc->T2_2) = finder(sc, cadr(args));
      car(sc->T2_1) = sc->temp4;
      return(c_call(code)(sc, sc->T2_1));

    case HOP_SAFE_C_S_opSq:
      args = caddr(code);
      car(sc->T1_1) = finder(sc, cadr(args));
      sc->temp4 = c_call(args)(sc, sc->T1_1);
      car(sc->T2_1) = finder(sc, cadr(code));
      car(sc->T2_2) = sc->temp4;
      return(c_call(code)(sc, sc->T2_1));

    case HOP_SAFE_NOT_S:
      return(make_boolean(sc, is_false(sc, finder(sc, cadr(code)))));
      
    default:
      fprintf(stderr, "oops end dox %s\n", DISPLAY_80(code));
      return(sc->UNSPECIFIED);
    }
}


static bool body_is_direct(int op)
{
  return((op == HOP_SAFE_C_C) ||
	 (op == HOP_SAFE_C_S) ||
	 (op == HOP_SAFE_C_SS) ||
	 (op == HOP_SAFE_C_SSS));
}
#endif


static s7_pointer check_do(s7_scheme *sc)
{
  s7_pointer x;
  
  if ((!is_pair(sc->code)) ||                             /* (do . 1) */
      ((!is_pair(car(sc->code))) &&                       /* (do 123) */
       (is_not_null(car(sc->code)))))                     /* (do () ...) is ok */
    return(eval_error(sc, "do: var list is not a list: ~S", sc->code));
  
  if (!is_pair(cdr(sc->code)))                            /* (do () . 1) */
    return(eval_error(sc, "do body is messed up: ~A", sc->code));
  
  if ((!is_pair(cadr(sc->code))) &&                       /* (do ((i 0)) 123) */
      (is_not_null(cadr(sc->code))))                      /* no end-test? */
    return(eval_error(sc, "do: end-test and end-value list is not a list: ~A", sc->code));
  
  if (is_pair(car(sc->code)))
    {
      for (x = car(sc->code); is_pair(x); x = cdr(x))
	{
	  if (!(is_pair(car(x))))                         /* (do (4) (= 3)) */
	    return(eval_error(sc, "do: variable name missing? ~A", sc->code));

	  if (!is_symbol(caar(x)))                     /* (do ((3 2)) ()) */
	    return(eval_error(sc, "do step variable: ~S is not a symbol?", x));

	  if (is_immutable(caar(x)))                     /* (do ((pi 3 (+ pi 1))) ((= pi 4)) pi) */
	    return(eval_error(sc, "do step variable: ~S is immutable", x));
	  
	  if (is_pair(cdar(x)))
	    {
	      if ((!is_pair(cddar(x))) &&
		  (is_not_null(cddar(x))))               /* (do ((i 0 . 1)) ...) */
		return(eval_error(sc, "do: step variable info is an improper list?: ~A", sc->code));
	      
	      if ((is_pair(cddar(x))) && 
		  (is_not_null(cdr(cddar(x)))))          /* (do ((i 0 1 (+ i 1))) ...) */
		return(eval_error(sc, "do: step variable info has extra stuff after the increment: ~A", sc->code));
	    }
	  else return(eval_error(sc, "do: step variable has no initial value: ~A", x));

	  set_local(caar(x));

	  /* (do ((i)) ...) */
	}

      if (is_not_null(x))                                /* (do ((i 0 i) . 1) ((= i 1))) */
	return(eval_error(sc, "do: list of variables is improper: ~A", sc->code));
    }

  if (is_pair(cadr(sc->code)))
    {
      for (x = cadr(sc->code); is_pair(x); x = cdr(x));
      if (is_not_null(x))
	return(eval_error(sc, "stray dot in do end section?", sc->code));
    }

  for (x = cddr(sc->code); is_pair(x); x = cdr(x));
  if (is_not_null(x))
    return(eval_error(sc, "stray dot in do body?", sc->code));

  if ((is_overlaid(sc->code)) &&
      (cdr(ecdr(sc->code)) == sc->code))
    {
      car(ecdr(sc->code)) = sc->DO_UNCHECKED;


#if WITH_OPTIMIZATION
      {
	/* (define (hi) (do ((i 0 (+ i 1))) ((= i 3)) (display i)) (newline)) */
	s7_pointer vars, end, body;

#if PRINTING
	  fprintf(stderr, "\n----------------------------------------\ncheck %s\n", DISPLAY_80(sc->code));
#endif	  

	vars = car(sc->code);
	end = cadr(sc->code);
	body = cddr(sc->code);

#if PRINTING
	/* (define (hi) (do ((i 1.5 (+ i 1))) ((= i 2.5)) (display i) (newline)))
	 *   in OP_DOTIMES, for example, if init value is not an integer, it goes to OP_SIMPLE_DO
 	 * remaining optimizable cases: we can step by 1 and use = for end, and yet simple_do(_p) calls the functions
 	 * geq happens as often as =, and -1 as step
 	 * also cdr as step to is_null as end
	 * also what about no do-var cases? (do () ...)
	 *
	 * also do body is optimized expr: vector_set_3 via hop_safe_c_sss for example or (vset v i (vref w i))
	 */
	if ((!is_pair(end)) || (!is_pair(car(end))) ||
	    (!is_pair(vars)) ||
	    (!is_null(cdr(vars))))
	  fprintf(stderr, "    ;end or vars not simple: %s\n", DISPLAY_80(sc->code));
#endif

	if ((is_pair(end)) && (is_pair(car(end))) &&
	    (is_pair(vars)) &&
	    (is_null(cdr(vars))))
	  {
	    /* loop has one step variable, and normal-looking end test
	     */
	    vars = car(vars);

#if PRINTING
	    /* c_ss c_s [end/step also: (= end limit)
	     */
	    if ((safe_list_length(sc, vars) != 3) ||
		((is_pair(cadr(vars))) &&
		 ((!is_optimized(cadr(vars))) ||
		  (optimize_data(cadr(vars)) != HOP_SAFE_C_C))))
	      fprintf(stderr, "    ;var init unsafe: %s %s\n", DISPLAY_80(vars), (is_optimized(cadr(vars))) ? opt_names[optimize_data(cadr(vars))] : "");
#endif

	    if ((safe_list_length(sc, vars) == 3) &&
		((!is_pair(cadr(vars))) ||
		 ((is_optimized(cadr(vars))) &&
		  (optimize_data(cadr(vars)) == HOP_SAFE_C_C))))
	      {
		s7_pointer step_expr;
		step_expr = caddr(vars);
#if PRINTING
		if (!is_optimized(step_expr))
		  fprintf(stderr, "    ;step expr unsafe: %s\n", DISPLAY_80(step_expr));
		else
		  {
		    if (!((((optimize_data(step_expr) == HOP_SAFE_C_SC) && (car(vars) == cadr(step_expr))) ||
			   ((optimize_data(step_expr) == HOP_SAFE_C_C) && (car(vars) == cadr(step_expr)) && 
			    ((ecdr(step_expr) == add_cs1) || (ecdr(step_expr) == subtract_cs1))) ||
			   ((optimize_data(step_expr) == HOP_SAFE_C_CS) && (car(vars) == caddr(step_expr))))))
		      fprintf(stderr, "    ;step_expr %s %s\n", opt_names[optimize_data(step_expr)], DISPLAY_80(step_expr));
		  }
#endif

		if ((is_optimized(step_expr)) &&
		    (((optimize_data(step_expr) == HOP_SAFE_C_SC) && (car(vars) == cadr(step_expr))) ||
		     ((optimize_data(step_expr) == HOP_SAFE_C_C) && (car(vars) == cadr(step_expr)) && 
		      ((ecdr(step_expr) == add_cs1) || (ecdr(step_expr) == subtract_cs1))) ||
		     ((optimize_data(step_expr) == HOP_SAFE_C_CS) && (car(vars) == caddr(step_expr)))))
		  {
		    /* step var is (var const|symbol (op var const)|(op const var)) 
		     */
		    end = car(end);

#if PRINTING
		    if (!((is_optimized(end)) &&
			  (car(vars) == cadr(end)) &&
			  ((ecdr(end) == equal_s_ic) ||
			   (optimize_data(end) == HOP_SAFE_C_SS) || 
			   (optimize_data(end) == HOP_SAFE_C_SC))))
		      fprintf(stderr, "    ;end unsafe: %s (var: %d) %s\n", 
			      (is_optimized(end)) ? opt_names[optimize_data(end)] : "unopt", car(vars) == cadr(end),
			      DISPLAY_80(end));
#endif

		    if ((is_optimized(end)) &&
			(car(vars) == cadr(end)) &&
			((ecdr(end) == equal_s_ic) ||
			 (optimize_data(end) == HOP_SAFE_C_SS) || 
			 (optimize_data(end) == HOP_SAFE_C_SC)))
		      {
			/* end var is (op var const|symbol) using same var as step 
			 *   so at least we can use SIMPLE_DO
			 */
			bool has_set = false;

#if PRINTING
			fprintf(stderr, "      ;step: %s %s, end: %s %s\n",
				opt_names[optimize_data(step_expr)], c_function_name(ecdr(step_expr)),
				opt_names[optimize_data(end)], c_function_name(ecdr(end)));
#endif

			if (ecdr(step_expr) == add_cs1)
			  {
			    set_c_function(step_expr, add_s1);
			    optimize_data(step_expr) = HOP_SAFE_C_SC;
			  }
			if (ecdr(step_expr) == subtract_cs1)
			  {
			    set_c_function(step_expr, subtract_s1);
			    optimize_data(step_expr) = HOP_SAFE_C_SC;
			  }
			if (ecdr(end) == equal_s_ic)
			  {
			    set_c_function(end, equal_2);
			    optimize_data(end) = HOP_SAFE_C_SC;
			  }
			car(ecdr(sc->code)) = sc->SIMPLE_DO;
			if ((safe_list_length(sc, body) == 1) &&
			    (is_pair(car(body))) &&
			    (is_syntactic(caar(body))))
			  {
			    car(ecdr(sc->code)) = sc->SIMPLE_DO_P;
			    fcdr(sc->code) = caddr(caar(sc->code));
			  }

			if ((safe_list_length(sc, body) == 1) &&
			    (is_pair(car(body))) &&
			    (is_optimized(car(body))))
			  {
			    if (body_is_direct(optimize_data(car(body))))
			      car(ecdr(sc->code)) = sc->SIMPLE_DO_OPT;
#if PRINTING
			    /* c_s c_c */
			    else fprintf(stderr, "      ;optimizable as simple: %s %s\n", opt_names[optimize_data(car(body))], DISPLAY_80(car(body)));
#endif
			  }

			if (do_is_safe(sc, body, sc->w = list_1(sc, car(vars)), sc->NIL, &has_set))
			  {
			    if ((safe_list_length(sc, body) == 1) &&
				(is_pair(car(body))) &&
				(is_optimized(car(body))))
			      {
				if (body_is_direct(optimize_data(car(body))))
				  car(ecdr(sc->code)) = sc->SAFE_SIMPLE_DO;
#if PRINTING
				/* c_s c_ss */
				else fprintf(stderr, "      ;optimizable as safe_simple: %s %s\n", opt_names[optimize_data(car(body))], DISPLAY_80(car(body)));
#endif
			      }

			    /* now look for the very common dotimes case
			     */
			    if ((((s7_is_integer(caddr(step_expr))) &&
				  (s7_integer(caddr(step_expr)) == 1)) ||
				 ((s7_is_integer(cadr(step_expr))) &&
				  (s7_integer(cadr(step_expr)) == 1))) &&
				(c_function_class(ecdr(step_expr)) == add_class) &&
#if WITH_GMP
				(c_function_class(ecdr(end)) == equal_class)
#else
				((c_function_class(ecdr(end)) == equal_class) ||
				 (ecdr(end) == geq_2))
#endif
				)
			      {
				/* we're stepping by +1 and going to =
				 *   the final integer check has to wait until run time (symbol value dependent)
				 */

				car(ecdr(sc->code)) = sc->SAFE_DO;
				if ((!has_set) &&
				    (c_function_class(ecdr(end)) == equal_class))
				{
				  car(ecdr(sc->code)) = sc->DOTIMES;

				  /* what are the most common cases here?
				   */
				  
				  if ((safe_list_length(sc, body) == 1) &&
				      (is_pair(car(body))))
				    {
#if PRINTING
				      /* (define (hi a b) (do ((i 0 (+ i 1))) ((= i a)) (vector-set! b i a))) */
				      fprintf(stderr, "    dotimes: %s\n", opt_names[optimize_data(car(body))]);
#endif				      

				      if ((is_optimized(car(body))) &&
					  (optimize_data(car(body)) == HOP_SAFE_C_C))
					{
					  car(ecdr(sc->code)) = sc->DOTIMES_C_C;
					}
				      else
					{
					  if ((is_syntactic(caar(body))) &&
					      (syntax_opcode(caar(body)) == OP_LET))
					    {
					      s7_pointer x;
					      x = cddar(body);
					      if ((is_pair(car(x))) &&
						  (is_optimized(car(x))) &&
						  (is_null(cdr(x))) &&
						  (optimize_data(car(x)) == HOP_SAFE_C_C))
						{
						  bool happy = true;

						  /* fprintf(stderr, "safe let?: %s\n", DISPLAY_80(sc->code));  */
						  
						  for (x = cadar(body); is_pair(x); x = cdr(x))
						    {
						      /* fprintf(stderr, "let var: %s\n", DISPLAY(car(x)));  */
						      if ((!is_pair(cadar(x))) ||
							  (!is_optimized(cadar(x))) ||
							  (optimize_data(cadar(x)) != HOP_SAFE_C_C))
							{
							  /*
							  fprintf(stderr, "%d %d %s\n", 
								  is_pair(cadar(x)),
								  is_optimized(cadar(x)),
								  opt_names[optimize_data(cadar(x))]);
							  */
							  happy = false;
							  break;
							}
						    }
						  if (happy)
						    {
						      /* fprintf(stderr, "safe: %s\n", DISPLAY_80(sc->code));  */
						      car(ecdr(sc->code)) = sc->SIMPLE_DOTIMES;
						    }
						}
					    }
					}
				    }
				}
      
			      }
			  }
#if PRINTING
			fprintf(stderr, "%s: %s\n\n", real_op_names[syntax_opcode(car(ecdr(sc->code)))], DISPLAY_80(sc->code));
#endif
			return(sc->NIL); /* tell OP_DO that this is a special case */
		      }
		  }
	      }
	  }
	/* we get here if there is more than one local var or anything "non-simple" about the rest 
	 */
	/* (define (hi) (do ((i 0 (+ i 1))) ((= i 3)) (display i)) (newline)) 
	 * (define (hi) (do ((i 0 (+ i 1)) (j 1 (+ j 1))) ((= i 3)) (display j))(newline))
	 */
	{
	  vars = car(sc->code);
	  end = cadr(sc->code);
	  body = cddr(sc->code);

	  /* check end expression 1st */
	  if (!is_end_dox_safe(sc, car(end)))
	    {
#if PRINTING
	      fprintf(stderr, "%send expr %s is trouble (%s)%s\n", 
		      BOLD_TEXT,
		      DISPLAY_80(car(end)), 
		      ((is_pair(car(end))) && (is_optimized(car(end)))) ? opt_names[optimize_data(car(end))] : "unopt",
		      UNBOLD_TEXT);
#endif
	      return(sc->code);
	    }
	  /* vars can be nil (no steppers) */
	  if (is_pair(vars))
	    {
	      s7_pointer p;
	      for (p = vars; is_pair(p); p = cdr(p))
		{
		  s7_pointer var;
		  var = car(p);
		  if (!is_init_dox_safe(sc, cadr(var)))
		    {
#if PRINTING
		      fprintf(stderr, "%s%s init expr %s is trouble (%s)%s\n", 
			      BOLD_TEXT,
			      DISPLAY(car(var)), 
			      DISPLAY_80(cadr(var)),
			      ((is_pair(cadr(var))) && (is_optimized(cadr(var)))) ? opt_names[optimize_data(cadr(var))] : "unopt",
			      UNBOLD_TEXT);
			      
#endif
		      return(sc->code);
		    }
		  if ((is_pair(cddr(var))) &&
		      (!is_step_dox_safe(sc, caddr(var))))
		    {
#if PRINTING
		      fprintf(stderr, "%s%s step expr %s is trouble (%s)%s\n", 
			      BOLD_TEXT,
			      DISPLAY(car(var)), 
			      DISPLAY_80(caddr(var)),
			      ((is_pair(caddr(var))) && (is_optimized(caddr(var)))) ? opt_names[optimize_data(caddr(var))] : "unopt",
			      UNBOLD_TEXT);
#endif
		      return(sc->code);
		    }
		}
	    }
	  
	  /* end and steps look ok! */
	  car(ecdr(sc->code)) = sc->DOX;

#if PRINTING
	  fprintf(stderr, "dox optimizable: %s\n", DISPLAY_80(sc->code));
	  if ((safe_list_length(sc, body) == 1) &&
	      (is_pair(car(body))) &&
	      (is_optimized(car(body))))
	    {
	      /* this is very rare and we need to check do is safe etc */
	      fprintf(stderr, "      ;dox perhaps optimizable as simple: %s %s\n", opt_names[optimize_data(car(body))], DISPLAY_80(car(body)));
	    }
#endif
#if 0
	  /* as far as I can tell, this happens very rarely, and p_case below is about 1/2 the cases
	   */
	  if ((is_null(cdr(end))) &&
	      (safe_list_length(sc, body) == 1) &&
	      (is_pair(car(body))) &&
	      (is_syntactic(caar(body))))
	    {
	      bool has_set = false;
	      s7_pointer p, var_list;
	      
	      for (p = vars, var_list = sc->NIL; is_pair(p); p = cdr(p))
		var_list = cons(sc, caar(p), var_list);
	      sc->w = var_list;
	      
	      if (do_is_safe(sc, body, var_list, sc->NIL, &has_set))
		{
#if PRINTING
		  fprintf(stderr, "    safe: %s\n", DISPLAY_80(sc->code));
#endif
		  car(ecdr(sc->code)) = sc->SAFE_DOX_P;
		}
	      
	      sc->w = sc->NIL;
	    }
#endif

	  return(sc->NIL);
	}
      }
#endif
    }

#if PRINTING
  fprintf(stderr, "not simple: %d %s\n\n", is_null(cdr(cddr(sc->code))), DISPLAY_80(sc->code));
#endif
  return(sc->code);
}


static s7_pointer check_defmacro(s7_scheme *sc)
{
  s7_pointer x, y, z;
  if (!is_pair(sc->code))                                               /* (defmacro . 1) */
    return(eval_error_with_name(sc, "~A name missing (stray dot?): ~A", sc->code));
  
  x = car(sc->code);
  if (!is_symbol(x))                                             /* (defmacro) or (defmacro 1 ...) */
    return(eval_error_with_name(sc, "~A name: ~S is not a symbol?", x));     
  
  if (is_immutable(x))
    return(eval_error_with_name(sc, "~A: ~S is immutable", x)); /* (defmacro pi (a) `(+ ,a 1)) */
  
  z = cdr(sc->code);
  if (!is_pair(z))                                                  /* (defmacro a) */
    return(eval_error_with_name(sc, "~A ~A, but no args or body?", sc->code));
  
  y = car(z);            /* the arglist */
  if ((!s7_is_list(sc, y)) &&
      (!is_symbol(y)))
    return(s7_error(sc, sc->SYNTAX_ERROR,                               /* (defmacro mac "hi" ...) */
		    list_3(sc, make_protected_string(sc, "defmacro ~A argument list is ~S?"), x, y)));
  
  for ( ; is_pair(y); y = cdr(y))
    if ((!is_symbol(car(y))) &&
	(sc->op == OP_DEFMACRO))
      return(s7_error(sc, sc->SYNTAX_ERROR,                             /* (defmacro mac (1) ...) */
		      list_3(sc, make_protected_string(sc, "defmacro ~A argument name is not a symbol: ~S"), x, y)));
  
  /* other parameter error checks are handled by lambda/lambda* (see OP_LAMBDA above) at macro expansion time */
  
  if (!is_pair(cdr(z)))
    {
      if (is_null(cdr(z)))                                             /* (defmacro hi ()) */
	return(eval_error_with_name(sc, "~A ~A has no body?", x));
      return(eval_error_with_name(sc, "~A ~A has stray dot?", x));
    }
  return(sc->code);
}


static s7_pointer check_define_macro(s7_scheme *sc)
{
  s7_pointer x;
  if (!is_pair(sc->code))                                               /* (define-macro . 1) */
    return(eval_error_with_name(sc, "~A name missing (stray dot?): ~A", sc->code));
  if (!is_pair(car(sc->code)))                                          /* (define-macro a ...) */
    return(s7_wrong_type_arg_error(sc, op_names[(int)(sc->op)], 1, car(sc->code), "a list (name ...)"));
  
  x = caar(sc->code);
  if (!is_symbol(x))
    return(eval_error_with_name(sc, "~A: ~S is not a symbol?", x));
  if (dont_eval_args(x))                                            /* (define-macro (quote a) quote) */
    return(eval_error_with_name(sc, "~A: syntactic keywords (such as ~S) tend to behave badly if redefined", x));
  
  if (is_immutable(x))
    return(eval_error_with_name(sc, "~A: ~S is immutable", x));
  
  if (!is_pair(cdr(sc->code)))                /* (define-macro (...)) */
    return(eval_error_with_name(sc, "~A ~A, but no body?", x));
  
  sc->y = cdar(sc->code);            /* the arglist */
  if ((!s7_is_list(sc, sc->y)) &&
      (!is_symbol(sc->y)))
    return(s7_error(sc, sc->SYNTAX_ERROR,                                      /* (define-macro (mac . 1) ...) */
		    list_3(sc, make_protected_string(sc, "define-macro ~A argument list is ~S?"), x, sc->y)));
  
  for ( ; is_pair(sc->y); sc->y = cdr(sc->y))
    if ((!is_symbol(car(sc->y))) &&
	((sc->op == OP_DEFINE_MACRO) || (sc->op == OP_DEFINE_BACRO)))
      return(s7_error(sc, sc->SYNTAX_ERROR,                                    /* (define-macro (mac 1) ...) */
		      list_3(sc, make_protected_string(sc, "define-macro ~A argument name is not a symbol: ~S"), x, sc->y)));
  
  return(sc->code);
}


static s7_pointer check_cond(s7_scheme *sc)
{
  bool has_feed_to = false;
  s7_pointer x;
  if (!is_pair(sc->code))                                             /* (cond) or (cond . 1) */
    return(eval_error(sc, "cond, but no body: ~A", sc->code));

  for (x = sc->code; is_pair(x); x = cdr(x))
    {
      if (!is_pair(car(x)))                                         /* (cond 1) or (cond (#t 1) 3) */
	return(eval_error(sc, "every clause in cond must be a list: ~A", car(x)));
      else
	{
	  s7_pointer y;
	  y = car(x);
	  if ((cadr(y) == sc->FEED_TO) &&
	      (s7_symbol_value(sc, sc->FEED_TO) == sc->UNDEFINED))
	    {
	      has_feed_to = true;
	      if (!is_pair(cddr(y)))                                  /* (cond (#t =>)) or (cond (#t => . 1)) */
		return(eval_error(sc, "cond: '=>' target missing?  ~A", x));
	      if (is_pair(cdddr(y)))                                  /* (cond (1 => + abs)) */
		return(eval_error(sc, "cond: '=>' has too many targets: ~A", x));
	    }
	  /* currently we accept:
	   *     (cond (1 2) (=> . =>)) and all variants thereof, e.g. (cond (1 2) (=> 1 . 2) (1 2)) or 
	   *     (cond (1) (=>)) but Guile accepts this?
	   *     (cond (1) (1 =>))
	   * amusing (correct) case: (cond (1 => "hi")) -> #\i
	   */
	}
    }
  if (is_not_null(x))                                             /* (cond ((1 2)) . 1) */
    return(eval_error(sc, "cond: stray dot? ~A", sc->code));

  if ((is_overlaid(sc->code)) &&
      (cdr(ecdr(sc->code)) == sc->code))
    {
      if (has_feed_to)
	car(ecdr(sc->code)) = sc->COND_UNCHECKED;
      else 
	{
	  car(ecdr(sc->code)) = sc->COND_SIMPLE;
#if WITH_OPTIMIZATION
	  {
	    s7_pointer p, clause;
	    bool xopt = true;
	    for (p = sc->code; is_pair(p) && (xopt); p = cdr(p))
	      {
		clause = caar(p);
		if (is_pair(clause))
		  {
		    if (!is_orx_safe(clause))
		      xopt = false;
		  }
		else
		  {
		    if ((clause != sc->T) &&
			((!is_symbol(clause)) ||
			 ((clause != sc->ELSE) &&
			  (s7_symbol_value(sc, clause) != sc->ELSE))))
		      xopt = false;
		  }
	      }
	    if (xopt)
	      car(ecdr(sc->code)) = sc->COND_SIMPLER;
	  }
	  /* perhaps cond as if then else or 2 choices (ie no loop in cond) COND|CASE_2 and COND|CASE_3
	   *   similarly in case if (say) 2 or 3 choices unroll the loop
	   *   but these are rare and the current overhead is not huge.
	   */
#endif
	}
    }
  return(sc->code);
}


#if WITH_OPTIMIZATION

#if (!WITH_GCC)

static s7_pointer fs(s7_scheme *sc, s7_pointer hdl)
{
  /* like finder, but no unbound-variable error */
  s7_pointer x;	

  if (frame_id(sc->envir) == symbol_id(hdl))
    return(slot_value(local_slot(hdl)));

  for (x = sc->envir; symbol_id(hdl) < frame_id(x); x = next_environment(x));

  if (frame_id(x) == symbol_id(hdl))
    return(slot_value(local_slot(hdl)));	

  for (; is_environment(x); x = next_environment(x))
    {
      s7_pointer y; 
      for (y = environment_slots(x); is_slot(y); y = next_slot(y))	
	if (slot_symbol(y) == hdl)
	  return(slot_value(y)); 
    }
  x = global_slot(hdl);	
  if (is_not_null(x)) 
    return(slot_value(x)); 
  return(sc->UNDEFINED);
}

#define function_is_ok(Code) (fs(sc, car(Code)) == ecdr(Code))

#define check_tfunction(Code) \
  if ((frame_id(sc->envir) != symbol_id(car(Code))) || \
      (slot_value(local_slot(car(Code))) != ecdr(Code))) \
    ecdr(Code) = fs(sc, car(Code))

#else

static s7_pointer fs(s7_scheme *sc, s7_pointer hdl)
{
  /* like finder, but no unbound-variable error */
  s7_pointer x;	

  for (x = sc->envir; symbol_id(hdl) < frame_id(x); x = next_environment(x));

  if (frame_id(x) == symbol_id(hdl))
    return(slot_value(local_slot(hdl)));	

  for (; is_environment(x); x = next_environment(x))
    {
      s7_pointer y; 
      for (y = environment_slots(x); is_slot(y); y = next_slot(y))	
	if (slot_symbol(y) == hdl)
	  return(slot_value(y)); 
    }
  x = global_slot(hdl);	
  if (is_not_null(x)) 
    return(slot_value(x)); 
  return(sc->UNDEFINED);
}

static bool function_is_ok_1(s7_scheme *sc, s7_pointer p, s7_pointer val)
{
  if ((is_closure(val)) && (is_closure(ecdr(p))))
    {
      if (closure_body(val) == closure_body(ecdr(p)))
	{
	  ecdr(p) = val;
	  return(true);
	}
      if ((is_pair(closure_args(val))) && 
	  (is_pair(closure_args(ecdr(p)))) &&
	  (safe_list_length(sc, closure_args(val)) == safe_list_length(sc, closure_args(ecdr(p)))))
	{
	  if ((optimize_data(p) < OP_SAFE_THUNK) ||
	      (!is_safe_closure(closure_body(ecdr(p)))))
	    {
	      ecdr(p) = val;
	      return(true);
	    }
	}
    }
  return(false);
}

#define function_is_ok(Code) \
  ({ s7_pointer _code_; \
     _code_ = Code; \
     (((frame_id(sc->envir) == symbol_id(car(_code_))) && \
       (slot_value(local_slot(car(_code_))) == ecdr(_code_))) || \
      ({ s7_pointer _val_; \
         _val_ = fs(sc, car(_code_)); \
         ((_val_ == ecdr(_code_)) || \
          (function_is_ok_1(sc, _code_, _val_))); })); })

#define check_tfunction(Code) \
  if ((frame_id(sc->envir) != symbol_id(car(Code))) || \
      (slot_value(local_slot(car(Code))) != ecdr(Code))) \
    ecdr(Code) = fs(sc, car(Code))

#endif
#endif



static void annotate_expansion(s7_scheme *sc, s7_pointer p)
{
  s7_pointer x;
  if ((is_symbol(car(p))) &&
      (is_pair(cdr(p))))
    {
      ecdr(cdr(p)) = p;
      set_overlay(cdr(p));
    }
  else
    {
      if (is_pair(car(p)))
	annotate_expansion(sc, car(p));
    }
  for (x = cdr(p); is_pair(x); x = cdr(x))
    if (is_pair(car(x)))
      annotate_expansion(sc, car(x));
}




/* -------------------------------- eval -------------------------------- */

static s7_pointer eval(s7_scheme *sc, opcode_t first_op) 
{
  sc->cur_code = ERROR_INFO_DEFAULT;
  sc->op = first_op;
  
  /* this procedure can be entered recursively (via s7_call for example), so it's no place for a setjmp
   *   I don't think the recursion can hurt our continuations because s7_call is coming from hooks and
   *   callbacks that are implicit in our stack.
   */
  
  goto START_WITHOUT_POP_STACK;
  /* this ugly two-step is actually noticeably faster than other ways of writing this code
   */

  /* syntax_opcode can be optimize_data, the field can be set at read time, we could
   *   probably combine the optimized and normal case statements, jump here if eval (eval_pair, opt_eval),
   *   and thereby save the is_syntactic and is_pair check in op_eval (still need is_optimized).
   *   To get there, the cdr(code) step in is_syntactic needs to be spread out to all the branches but
   *   all the current pops and jumps assume the cdr, op_begin would explicitly jump back here, no op_eval, 
   *   current trailers would be outside? and where would eval args go?  Huge change, might save 1% if lucky.
   *
   * a better savings: about half the cases don't care about args or op, so perhaps pop_stack could be
   *   spread over the branches, but then how to handle the no-pop-stack case?  pop_stack_no_args here
   *   might work -- the direct cases are all syntactic, so args won't matter.
   */

 START:
  pop_stack(sc);

 START_WITHOUT_POP_STACK:
  switch (sc->op) 
    {
    case OP_READ_INTERNAL:
      /* if we're loading a file, and in the file we evaluate something like:
       *
       *    (let ()
       *      (set-current-input-port (open-input-file "tmp2.r5rs"))
       *      (close-input-port (current-input-port)))
       *    ... (with no reset of input port to its original value)
       *
       * the load process tries to read the loaded string, but the sc->input_port is now closed,
       * and the original is inaccessible!  So we get a segfault in token.  We don't want to put
       * a port_is_closed check there because token only rarely is in this danger.  I think this
       * is the only place where we can be about to call token, and someone has screwed up our port.
       *
       * We can't call read_error here because it assumes the input string is ok!
       */

      if (port_is_closed(sc->input_port))
	return(s7_error(sc, sc->READ_ERROR, 
			list_1(sc, make_protected_string(sc, "our input port got clobbered!"))));

      sc->tok = token(sc);

      switch (sc->tok)
	{
	case TOKEN_EOF:
	  goto START;

	case TOKEN_RIGHT_PAREN:
	  read_error(sc, "unexpected close paren");

	case TOKEN_COMMA:
	  read_error(sc, "unexpected comma");

	default:
	  sc->value = read_expression(sc);
	  sc->current_line = port_line_number(sc->input_port);  /* this info is used to track down missing close parens */
	  sc->current_file = port_filename(sc->input_port);
	  goto START;
	}

      
      /* (read p) from scheme
       *    "p" becomes current input port for eval's duration, then pops back before returning value into calling expr
       */
    case OP_READ_DONE:
      pop_input_port(sc);

      if (sc->tok == TOKEN_EOF)
	sc->value = sc->EOF_OBJECT;
      sc->current_file = NULL;
      goto START;
      
      
      /* load("file"); from C (g_load) -- assume caller will clean up
       *   read and evaluate exprs until EOF that matches (stack reflects nesting)
       */
    case OP_LOAD_RETURN_IF_EOF:  /* loop here until eof (via push stack below) */
      if (sc->tok != TOKEN_EOF)
	{
	  push_stack(sc, OP_LOAD_RETURN_IF_EOF, sc->NIL, sc->NIL);
	  push_stack(sc, OP_READ_INTERNAL, sc->NIL, sc->NIL);
	  sc->code = sc->value;
	  goto EVAL;             /* we read an expression, now evaluate it, and return to read the next */
	}
      sc->current_file = NULL;
      return(sc->F);
      
      
      /* (load "file") in scheme 
       *    read and evaluate all exprs, then upon EOF, close current and pop input port stack
       */
    case OP_LOAD_CLOSE_AND_POP_IF_EOF:
      if (sc->tok != TOKEN_EOF)
	{
	  push_stack(sc, OP_LOAD_CLOSE_AND_POP_IF_EOF, sc->NIL, sc->NIL); /* was push args, code */
	  push_stack(sc, OP_READ_INTERNAL, sc->NIL, sc->NIL);
	  sc->code = sc->value;
	  goto EVAL;             /* we read an expression, now evaluate it, and return to read the next */
	}
      s7_close_input_port(sc, sc->input_port);
      pop_input_port(sc);
      sc->current_file = NULL;

      if (is_multiple_value(sc->value))                    /* (load "file") where "file" is (values 1 2 3) */
	sc->value = splice_in_values(sc, multiple_value(sc->value));
      goto START;
      

      /* read and evaluate string expression(s?)
       *    assume caller (C via g_eval_c_string) is dealing with the string port
       */
    case OP_EVAL_STRING:
      /* this is the C side s7_eval_c_string. 
       */

      /* (eval-string (string-append "(list 1 2 3)" (string #\newline) (string #\newline))) 
       *    needs to be sure to get rid of the trailing white space before checking for EOF
       *    else it tries to eval twice and gets "attempt to apply 1?, line 2"
       */
      if ((sc->tok != TOKEN_EOF) && 
	  (port_string_point(sc->input_port) < port_string_length(sc->input_port))) /* ran past end somehow? */
	{
	  unsigned char c;
	  while (white_space[c = port_string(sc->input_port)[port_string_point(sc->input_port)++]])
	    if (c == '\n')
	      port_line_number(sc->input_port)++;

	  if (c != 0)
	    {
	      backchar(c, sc->input_port);
	      push_stack(sc, OP_EVAL_STRING, sc->NIL, sc->value);
	      push_stack(sc, OP_READ_INTERNAL, sc->NIL, sc->NIL);
	    }
	  else push_stack(sc, OP_EVAL_DONE, sc->NIL, sc->value);
	}
      else push_stack(sc, OP_EVAL_DONE, sc->NIL, sc->value);
      sc->code = sc->value;
      goto EVAL;

      
    case OP_EVAL_STRING_2:
      s7_close_input_port(sc, sc->input_port);
      pop_input_port(sc);

      if (is_multiple_value(sc->value))
	sc->value = splice_in_values(sc, multiple_value(sc->value));

      goto START;

      
    case OP_EVAL_STRING_1:
      if ((sc->tok != TOKEN_EOF) && 
	  (port_string_point(sc->input_port) < port_string_length(sc->input_port))) /* ran past end somehow? */
	{
	  unsigned char c;
	  while (white_space[c = port_string(sc->input_port)[port_string_point(sc->input_port)++]])
	    if (c == '\n')
	      port_line_number(sc->input_port)++;

	  if (c != 0)
	    {
	      backchar(c, sc->input_port);
	      push_stack(sc, OP_EVAL_STRING_1, sc->NIL, sc->value);
	      push_stack(sc, OP_READ_INTERNAL, sc->NIL, sc->NIL);
	    }
	  else push_stack(sc, OP_EVAL_STRING_2, sc->NIL, sc->NIL);
	}
      else push_stack(sc, OP_EVAL_STRING_2, sc->NIL, sc->NIL);
      sc->code = sc->value;
      goto EVAL;


      /* -------------------- sort! (heapsort, done directly so that call/cc in the sort function will work correctly) -------------------- */

    #define SORT_N integer(vector_element(sc->code, 0))
    #define SORT_K integer(vector_element(sc->code, 1))
    #define SORT_J integer(vector_element(sc->code, 2))
    #define SORT_K1 integer(vector_element(sc->code, 3))
    #define SORT_CALLS integer(vector_element(sc->code, 4))
    #define SORT_STOP integer(vector_element(sc->code, 5))
    #define SORT_DATA(K) vector_element(car(sc->args), K)
    #define SORT_LESSP cadr(sc->args)

    HEAPSORT:
      {
	s7_Int n, j, k;
	n = SORT_N;
	k = SORT_K1;

	if ((n == k) || (k > ((s7_Int)(n / 2)))) /* k == n == 0 is the first case */
	  goto START;

	if (sc->safety != 0)
	  {
	    SORT_CALLS++;
	    if (SORT_CALLS > SORT_STOP)
	      return(eval_error(sc, "sort! is caught in an infinite loop, comparison: ~S", SORT_LESSP));
	  }
	j = 2 * k;
	SORT_J = j;
	if (j < n)
	  {
	    push_stack(sc, OP_SORT1, sc->args, sc->code);
	    sc->x = SORT_LESSP; /* cadr of sc->args */
	    sc->args = list_2(sc, SORT_DATA(j), SORT_DATA(j + 1));
	    sc->code = sc->x;
	    goto APPLY;
	  }
	else sc->value = sc->F;
      }

    case OP_SORT1:
      {
	s7_Int j, k;
	k = SORT_K1;
	j = SORT_J;
	if (is_true(sc, sc->value))
	  {
	    j = j + 1;
	    SORT_J = j;
	  }
	push_stack(sc, OP_SORT2, sc->args, sc->code);
	sc->x = SORT_LESSP;
	sc->args = list_2(sc, SORT_DATA(k), SORT_DATA(j));
	sc->code = sc->x;
	goto APPLY;
      }

    case OP_SORT2:
      {
	s7_Int j, k;
	k = SORT_K1;
	j = SORT_J;
	if (is_true(sc, sc->value))
	  {
	    sc->x = SORT_DATA(j);
	    SORT_DATA(j) = SORT_DATA(k);
	    SORT_DATA(k) = sc->x;
	  }
	else goto START;
	SORT_K1 = SORT_J;
	goto HEAPSORT;
      }

    case OP_SORT:
      /* coming in sc->args is sort args (data less?), sc->code = '(n k 0)
       *
       * here we call the inner loop until k <= 0 [the local k! -- this is tricky because scheme passes args by value]
       */
      {
	s7_Int k;
	k = SORT_K;
	if (k <= 0)
	  goto SORT3;

	SORT_K = k - 1;
	SORT_K1 = k - 1;

	push_stack(sc, OP_SORT, sc->args, sc->code);
	goto HEAPSORT;
      }

      SORT3:
      case OP_SORT3:
	{
	  s7_Int n;
	  n = SORT_N;
	  if (n <= 0)
	    {
	      sc->value = car(sc->args);
	      goto START;
	    }
	  sc->x = SORT_DATA(0);
	  SORT_DATA(0) = SORT_DATA(n);
	  SORT_DATA(n) = sc->x;
	  SORT_N = n - 1;
	  SORT_K1 = 0;
	  push_stack(sc, OP_SORT3, sc->args, sc->code);
	  goto HEAPSORT;
	}

    case OP_SORT4:
      /* sc->value is the sort vector which needs to be turned into a list */
      sc->value = s7_vector_to_list(sc, sc->value);
      goto START;

    case OP_SORT_TWO:
      /* here we're sorting a list of 2 items */
      if (is_true(sc, sc->value))
	sc->value = sc->args;
      else sc->value = list_2(sc, cadr(sc->args), car(sc->args));
      goto START;

      /* batcher networks:
       *    ((0 2) (0 1) (1 2))
       *    ((0 2) (1 3) (0 1) (2 3) (1 2))
       *    ((0 4) (0 2) (1 3) (2 4) (0 1) (2 3) (1 4) (1 2) (3 4))
       *    ((0 4) (1 5) (0 2) (1 3) (2 4) (3 5) (0 1) (2 3) (4 5) (1 4) (1 2) (3 4))
       *    ((0 4) (1 5) (2 6) (0 2) (1 3) (4 6) (2 4) (3 5) (0 1) (2 3) (4 5) (1 4) (3 6) (1 2) (3 4) (5 6))
       *    ((0 4) (1 5) (2 6) (3 7) (0 2) (1 3) (4 6) (5 7) (2 4) (3 5) (0 1) (2 3) (4 5) (6 7) (1 4) (3 6) (1 2) (3 4) (5 6))
       *
       * but since it has to be done here by hand, it turns into too much code, 3 is:
       *    < l0 l2 ?
       *    no goto L1
       *    < l0 l1 ?
       *    no  return 1 0 2
       *    < l1 l2?
       *    yes return 0 1 2 (direct)
       *    no  return 0 2 1
       *  L1:
       *    < l0 l1 ?
       *    yes return 2 0 1
       *    < l1 l2 ?
       *    yes return 1 2 0
       *    no  return 2 1 0
       * since each "<" op above goes to OP_APPLY, we have ca 5 labels, and ca 25-50 lines
       */


    case OP_MAP_SIMPLE:
      /* func = sc->code, func takes one arg, args = '(nil arglist) at the start with symbol_id = len
       *   the func is prechecked, as is the list.
       */
      {
	s7_pointer args;
	args = sc->args;
	if (sc->value != sc->NO_VALUE)
	  {
  	    if (is_multiple_value(sc->value))
 	      counter_result(args) = revappend(sc, multiple_value(sc->value), counter_result(args));
 	    else counter_result(args) = cons(sc, sc->value, counter_result(args));
  	  }
 	if (is_pair(counter_list(args)))
  	  {
 	    if ((--counter_count(args)) >= 0) /* protect against circular arg lists created by the map function! */
  	      {
  		s7_pointer code;
  		code = sc->code;
 		sc->x = car(counter_list(args));
 		counter_list(args) = cdr(counter_list(args));  /* move down list of args */
  		push_stack(sc, OP_MAP_SIMPLE, args, code);
		
		/* now call the function directly -- we know it is a T_CLOSURE
		 *   and that it can be called with one argument, so there's no need to go to
		 *   apply.
		 */
		NEW_FRAME_WITH_SLOT(sc, closure_environment(code), sc->envir, car(closure_args(code)), sc->x);
		sc->code = closure_body(code);
		goto BEGIN;
	      }
	  }
	sc->value = safe_reverse_in_place(sc, counter_result(args));
	typeflag(args) = 0;
	(*(sc->free_heap_top++)) = args;
	sc->args = sc->NIL;
	goto START;
      }


    case OP_MAP:
      if (sc->value != sc->NO_VALUE)                   /* (map (lambda (x) (values)) (list 1)) */
	{
	  if (is_multiple_value(sc->value))            /* (map (lambda (x) (if (odd? x) (values x (* x 20)) (values))) (list 1 2 3 4)) */
	    counter_result(sc->args) = revappend(sc, multiple_value(sc->value), counter_result(sc->args));
	  /* not append_in_place here because sc->value has the multiple-values bit set */
	  else counter_result(sc->args) = cons(sc, sc->value, counter_result(sc->args));
	}

      if (counter_count(sc->args) < counter_length(sc->args))
	{
	  if (next_map(sc))
	    {
	      if (needs_copied_args(sc->code))
		sc->args = copy_list(sc, sc->args);
	      goto APPLY;
	    }
	}
      sc->value = safe_reverse_in_place(sc, counter_result(sc->args));
      typeflag(sc->args) = 0;
      (*(sc->free_heap_top++)) = sc->args;
      sc->args = sc->NIL;
      goto START;

      
      /* -------------------------------- FOR-EACH -------------------------------- */

#if WITH_OPTIMIZATION
    FOR_EACH_SIMPLE:
#endif
    case OP_FOR_EACH_SIMPLE:
      /* func = sc->code, func takes one arg, args = arglist 
       * (for-each (lambda (x) (display x)) (list 1 2 3)) 
       */
      {
	s7_pointer args;
	args = sc->args;
	if (is_pair(args))
	  {
	    /* now call the function directly as in map above. 
	     *
	     * it's possible to preset the frame and its entry so that this allocation happens only
	     *   once per for-each call, but then for-each is like do in that exported args are equivalenced.
	     *   Since the savings isn't huge, I guess I'll go ahead and allocate every time.
	     */
	    s7_pointer code;
	    code = sc->code;
	    NEW_FRAME_WITH_SLOT(sc, closure_environment(code), sc->envir, car(closure_args(code)), car(args));  /* set function arg value */
	    push_stack(sc, OP_FOR_EACH_SIMPLE, cdr(args), code);
	    sc->code = closure_body(code);
	    goto BEGIN;
	  }
	sc->value = sc->UNSPECIFIED;
	goto START;
      }

#if WITH_OPTIMIZATION
    FOR_EACH_SIMPLER:
#endif
    case OP_FOR_EACH_SIMPLER:
      {
	s7_pointer args;
	args = sc->args;
	if (is_pair(args))
	  {
	    s7_pointer code;
	    code = sc->code;
	    NEW_FRAME_WITH_SLOT(sc, closure_environment(code), sc->envir, car(closure_args(code)), car(args));
	    push_stack(sc, OP_FOR_EACH_SIMPLER, cdr(args), code);
	    sc->code = car(closure_body(code));
	    goto EVAL_PAIR; /* lg: mostly syn */
	  }
	sc->value = sc->UNSPECIFIED;
	goto START;
      }


    case OP_FOR_EACH:
      if (counter_count(sc->args) < counter_length(sc->args))
	{
	  if (next_for_each(sc)) 
	    {
	      if (needs_copied_args(sc->code))
		sc->args = copy_list(sc, sc->args);
	      goto APPLY;
	    }
	}
      sc->value = sc->UNSPECIFIED;
      typeflag(sc->args) = 0;
      (*(sc->free_heap_top++)) = sc->args;
      sc->args = sc->NIL;
      goto START;


#if WITH_OPTIMIZATION
    case OP_SAFE_FOR_EACH:
      if (is_pair(sc->args))
	{
	  /* car(sc->args) is next closure arg
	   * closure is sc->code, known safe, its env has been set up (in g_for_each_3) for use here
	   */
	  s7_pointer x, arg, env;
	  unsigned long long int id;

	  id = ++frame_number;
	  env = closure_environment(sc->code);
	  frame_id(env) = id;

	  arg = car(sc->args);
	  push_stack(sc, OP_SAFE_FOR_EACH, cdr(sc->args), sc->code);

	  x = environment_slots(env);
	  slot_set_value(x, arg);
	  symbol_id(slot_symbol(x)) = id;
	  local_slot(slot_symbol(x)) = x;
	  sc->envir = env;
	  sc->code = closure_body(sc->code);
	  goto BEGIN;
	}
      sc->value = sc->UNSPECIFIED;
      goto START;
#endif


      /* -------------------------------- MEMBER -------------------------------- */
    case OP_MEMBER_IF:
      /* code=func, args = (list original args) with ecdr->position in cadr (the list), value = result of comparison
       */
      if (sc->value != sc->F)            /* previous comparison was not #f -- return list */
	{
	  sc->value = ecdr(sc->args);
	  goto START;
	}

      if (!is_pair(cdr(ecdr(sc->args))))     /* no more args -- return #f */
	{
	  sc->value = sc->F;
	  goto START;
	}
      ecdr(sc->args) = cdr(ecdr(sc->args));  /* cdr down arg list */
      push_stack(sc, OP_MEMBER_IF1, sc->args, sc->code);
      sc->args = list_2(sc, caar(sc->args), car(ecdr(sc->args)));
      goto APPLY;


    case OP_MEMBER_IF1:
      if (sc->value != sc->F)            /* previous comparison was not #f -- return list */
	{
	  sc->value = ecdr(sc->args);
	  goto START;
	}

      if (!is_pair(cdr(ecdr(sc->args))))     /* no more args -- return #f */
	{
	  sc->value = sc->F;
	  goto START;
	}
      ecdr(sc->args) = cdr(ecdr(sc->args));  /* cdr down arg list */

      /* circular list check */
      if (ecdr(sc->args) == fcdr(sc->args))
	{
	  sc->value = sc->F;
	  goto START;
	}
      fcdr(sc->args) = cdr(fcdr(sc->args));  /* cdr down the slow list (check for circular list) */

      push_stack(sc, OP_MEMBER_IF, sc->args, sc->code);
      sc->args = list_2(sc, caar(sc->args), car(ecdr(sc->args)));
      goto APPLY;



      /* -------------------------------- ASSOC -------------------------------- */
    case OP_ASSOC_IF1:
    case OP_ASSOC_IF:
      /* code=func, args=(list args) with f/ecdr=list, value=result of comparison
       *   (assoc 3 '((1 . a) (2 . b) (3 . c) (4 . d)) =)
       */
      if (sc->value != sc->F)            /* previous comparison was not #f -- return (car list) */
	{
	  sc->value = car(ecdr(sc->args));
	  goto START;
	}

      if (!is_pair(cdr(ecdr(sc->args))))     /* (assoc 3 '((1 . 2) . 3) =) or nil */
	{
	  sc->value = sc->F;
	  goto START;
	}
      ecdr(sc->args) = cdr(ecdr(sc->args));  /* cdr down arg list */

      if (sc->op == OP_ASSOC_IF1)
	{
	  /* circular list check */
	  if (ecdr(sc->args) == fcdr(sc->args))
	    {
	      sc->value = sc->F;
	      goto START;
	    }
	  fcdr(sc->args) = cdr(fcdr(sc->args));  /* cdr down the slow list */
	  push_stack(sc, OP_ASSOC_IF, sc->args, sc->code);
	}
      else push_stack(sc, OP_ASSOC_IF1, sc->args, sc->code);

      if (!is_pair(car(ecdr(sc->args))))     /* (assoc 1 '((2 . 2) 3) =) -- we access caaadr below */
	return(eval_error(sc, "assoc: 2nd arg is not an alist: ~S", sc->args));
      /* not sure about this -- we could simply skip the entry both here and in g_assoc
       *   (assoc 1 '((2 . 2) 3)) -> #f
       *   (assoc 1 '((2 . 2) 3) =) -> error currently
       */

      sc->args = list_2(sc, caar(sc->args), caar(ecdr(sc->args)));
      goto APPLY;



      /* -------------------------------- HOOK-APPLY -------------------------------- */
    case OP_HOOK_APPLY:
      /* args = function args, code = function list */
      if (is_null(sc->code))
	goto START;

      push_stack(sc, OP_HOOK_APPLY, sc->args, cdr(sc->code));
      sc->code = car(sc->code);
      /* fall through */



      /* ---------------- OP_APPLY ---------------- */
    case OP_APPLY:      /* apply 'code' to 'args' */
      if (tracing) 
	trace_apply(sc);
      if (needs_copied_args(sc->code))
	sc->args = copy_list(sc, sc->args);
      goto APPLY;




      /* -------------------------------- DO -------------------------------- */

#if WITH_OPTIMIZATION
      /* (define (hi) (do ((i 0 (+ i 1))) ((= i 3)) (display i)) (newline)) 
       * (with-sound () (nrev 0 .1))
       */

    SIMPLE_DOTIMES:
    case OP_SIMPLE_DOTIMES:
      /* set up as in DOTIMES, pull in all the let vars
       *   run body without any jumps: update and check step,
       *   SAFE_C_C of each let var, SAFE_C_C of body
       *   goto START
       */
      {
	s7_pointer vars, init_val, func, body, stepper, lets;
	vars = car(sc->code);
	init_val = cadr(car(vars));
	if (is_symbol(init_val))
	  init_val = finder(sc, init_val);
	if (s7_is_integer(init_val))
	  {
	    s7_pointer end_expr, end_val, let_var, p;
	    end_expr = caadr(sc->code);
	    end_val = caddr(end_expr);
	    if (is_symbol(end_val))
	      end_val = finder(sc, end_val);
	    if (s7_is_integer(end_val))
	      {
		int num_vars = 0;
		if (s7_integer(init_val) == s7_integer(end_val))
		  {
		    sc->code = cdr(cadr(sc->code));
		    goto BEGIN;
		  }
		sc->envir = new_frame_in_env(sc, sc->envir); 
		sc->args = add_slot(sc, caar(vars), make_mutable_integer(sc, s7_integer(init_val)));
		denominator(slot_value(sc->args)) = s7_integer(end_val);
		initialize_safe_do(sc, sc->code);

		/* add the let vars but not initialized yet 
		 */
		for (let_var = car(cdaddr(sc->code)); is_pair(let_var); let_var = cdr(let_var), num_vars++)
		  add_slot(sc, caar(let_var), sc->UNDEFINED);
		
		stepper = slot_value(sc->args);
		/* func = ecdr(caddr(caddr(sc->code))); */
		func = caddr(caddr(sc->code));
		body = cdr(caddr(caddr(sc->code)));
		lets = car(cdaddr(sc->code));
		
		if (num_vars != 1)
		  {
		  SIMPLE_DOTIMES_LET_LOOP:
		    
		    /* eval let + body  (cddr(sc->code)) */
		    for (let_var = lets; is_pair(let_var); let_var = cdr(let_var))
		      {
			p = find_local_symbol(sc, sc->envir, caar(let_var));
			slot_set_value(p, c_call(cadar(let_var))(sc, cdadar(let_var)));
		      }
		    
		    /* (with-sound () (fm-violin 0 .0001 440 .1))
		     */
		    c_call(func)(sc, body);
		    numerator(stepper)++;
		    if (numerator(stepper) == denominator(stepper))
		      {
			sc->code = cdr(cadr(sc->code));
			finalize_safe_do(sc);
			goto BEGIN;
		      }
		    goto SIMPLE_DOTIMES_LET_LOOP;
		  }
		else
		  {
		    p = environment_slots(sc->envir);
		    let_var = cdadar(lets);
		    lets = cadar(lets);
		    
		  SIMPLE_DOTIMES_LOOP:
		    slot_set_value(p, c_call(lets)(sc, let_var));
		    c_call(func)(sc, body);
		    numerator(stepper)++;
		    if (numerator(stepper) == denominator(stepper))
		      {
			sc->code = cdr(cadr(sc->code));
			finalize_safe_do(sc);
			goto BEGIN;
		      }
		    goto SIMPLE_DOTIMES_LOOP;
		  }
	      }
	  }
	car(ecdr(sc->code)) = sc->SIMPLE_DO;
	goto SIMPLE_DO;
      }


      /* ---------------- */



    DOTIMES_C_C:
    case OP_DOTIMES_C_C:
      {
	s7_pointer vars, init_val, func, body, stepper;

	vars = car(sc->code);
	init_val = cadr(car(vars));
	if (is_symbol(init_val))
	  init_val = finder(sc, init_val);
	else
	  {
	    if (is_pair(init_val))
	      init_val = c_call(init_val)(sc, cdr(init_val));
	  }
	if (s7_is_integer(init_val))
	  {
	    s7_pointer end_expr, end_val;
	    end_expr = caadr(sc->code);
	    end_val = caddr(end_expr);
	    if (is_symbol(end_val))
	      end_val = finder(sc, end_val);
	    if (s7_is_integer(end_val))
	      {
		if (s7_integer(init_val) == s7_integer(end_val))
		  {
		    sc->code = cdr(cadr(sc->code));
		    goto BEGIN;
		  }

		sc->envir = new_frame_in_env(sc, sc->envir); 
		sc->args = add_slot(sc, caar(vars), make_mutable_integer(sc, s7_integer(init_val)));

		denominator(slot_value(sc->args)) = s7_integer(end_val);
		initialize_safe_do(sc, sc->code);

		/* func = ecdr(caddr(sc->code)); */
		func = caddr(sc->code);
		body = cdr(caddr(sc->code));
		stepper = slot_value(sc->args);
		
	      DOTIMES_C_C_LOOP:
		/* eval body  (cddr(sc->code)) */
		c_call(func)(sc, body);
		numerator(stepper)++;
		if (numerator(stepper) == denominator(stepper))
		  {
		    sc->code = cdr(cadr(sc->code));
		    finalize_safe_do(sc);
		    goto BEGIN;
		  }
		goto DOTIMES_C_C_LOOP;
	      }
	  }
	car(ecdr(sc->code)) = sc->SIMPLE_DO;
	goto SIMPLE_DO;
      }



    DOTIMES:
    case OP_DOTIMES:
      {
	s7_pointer vars, init_val;
	vars = car(sc->code);
	init_val = cadr(car(vars));
	if (is_symbol(init_val))
	  init_val = finder(sc, init_val);
	else
	  {
	    if (is_pair(init_val))
	      init_val = c_call(init_val)(sc, cdr(init_val));
	  }
	if (s7_is_integer(init_val))
	  {
	    s7_pointer end_expr, end_val;
	    end_expr = caadr(sc->code);
	    end_val = caddr(end_expr);
	    if (is_symbol(end_val))
	      end_val = finder(sc, end_val);
	    if (s7_is_integer(end_val))
	      {
		if (s7_integer(init_val) == s7_integer(end_val))
		  {
		    sc->code = cdr(cadr(sc->code));
		    goto BEGIN;
		  }
		sc->envir = new_frame_in_env(sc, sc->envir); 
		sc->args = add_slot(sc, caar(vars), make_mutable_integer(sc, s7_integer(init_val)));

		denominator(slot_value(sc->args)) = s7_integer(end_val);
		initialize_safe_do(sc, sc->code);

		push_stack(sc, OP_DOTIMES_STEP, sc->args, sc->code);
		sc->code = cddr(sc->code);

		goto BEGIN;
	      }
	  }

	car(ecdr(sc->code)) = sc->SIMPLE_DO;
	goto SIMPLE_DO;
      }

    case OP_DOTIMES_STEP:
      numerator(slot_value(sc->args))++;
      if (numerator(slot_value(sc->args)) == denominator(slot_value(sc->args)))
	{
	  sc->code = cdr(cadr(sc->code));
	  finalize_safe_do(sc);
	  goto BEGIN;
	}
      push_stack(sc, OP_DOTIMES_STEP, sc->args, sc->code);
      sc->code = cddr(sc->code);
      goto BEGIN;


    SAFE_DO:
    case OP_SAFE_DO:
      {
	/* body is safe, step = +1, end is =, but stepper and end might be set in the body
	 */
	s7_pointer end, init_val, end_val;

	init_val = cadaar(sc->code);
	if (is_symbol(init_val))
	  init_val = finder(sc, init_val);
	else
	  {
	    if (is_pair(init_val))
	      init_val = c_call(init_val)(sc, cdr(init_val));
	  }
	
	end = caddr(car(cadr(sc->code)));
	if (is_symbol(end))
	  end_val = finder(sc, end);
	else end_val = end;

#if WITH_GMP
	if (s7_integer(init_val) == s7_integer(end_val))
#else
	if ((s7_integer(init_val) == s7_integer(end_val)) ||
	    ((s7_integer(init_val) > s7_integer(end_val)) &&
	     (ecdr(car(cadr(sc->code))) == geq_2)))
#endif
	  {
	    sc->code = cdr(cadr(sc->code));
	    goto BEGIN;
	  }
	sc->envir = new_frame_in_env(sc, sc->envir);

	if (is_symbol(end))
	  sc->args = cons(sc, add_slot(sc, caaar(sc->code), init_val), find_symbol(sc, end));
	else sc->args = cons(sc, add_slot(sc, caaar(sc->code), init_val), cons_unchecked(sc, end, end));
	
	initialize_safe_do(sc, sc->code);

	push_stack(sc, OP_SAFE_DO_STEP, sc->args, sc->code);
	sc->code = cddr(sc->code);
	goto BEGIN;
      }

    case OP_SAFE_DO_STEP:
      {
	s7_Int step;
	s7_pointer args;
	args = sc->args;

	step = s7_integer(slot_value(car(args))) + 1;
	slot_set_value(car(args), make_integer(sc, step));
#if WITH_GMP
	if (step == s7_integer(slot_value(cdr(args))))
#else
	if ((step == s7_integer(slot_value(cdr(args)))) ||
	    ((step > s7_integer(slot_value(cdr(args)))) &&
	     (ecdr(car(cadr(sc->code))) == geq_2)))
#endif	      
	  {
	    finalize_safe_do(sc);
	    sc->code = cdr(cadr(sc->code));
	    goto BEGIN;
	  }
	push_stack(sc, OP_SAFE_DO_STEP, args, sc->code);
	sc->code = cddr(sc->code);
	goto BEGIN;
      }


    SIMPLE_DO:
    case OP_SIMPLE_DO:
      {
	/* body might not be safe in this case, but the step and end exprs are easy
	 */
	s7_pointer init, end;
	sc->envir = new_frame_in_env(sc, sc->envir);
	init = cadaar(sc->code);
	if (is_symbol(init))
	  sc->value = finder(sc, init);
	else 
	  {
	    if (is_pair(init))
	      sc->value = c_call(init)(sc, cdr(init));
	    else sc->value = init;
	  }
	end = caddr(car(cadr(sc->code)));
	if (is_symbol(end))
	  sc->args = cons_unchecked(sc, add_slot(sc, caaar(sc->code), sc->value), find_symbol(sc, end));
	else 
	  {
	    s7_pointer slot;
	    NEW_CELL_NO_CHECK(sc, slot);
	    slot_symbol(slot) = end;
	    slot_set_value(slot, end);
	    set_type(slot, T_SLOT | T_IMMUTABLE);
	    sc->args = cons_unchecked(sc, add_slot(sc, caaar(sc->code), sc->value), slot);
	  }
	goto SIMPLE_DO_END;
      }

    case OP_SIMPLE_DO_STEP:
      {
	s7_pointer step, car_args;
	car_args = car(sc->args);
	step = caddr(caar(sc->code));
	if (is_symbol(cadr(step)))
	  {
	    car(sc->T2_1) = slot_value(car_args);
	    car(sc->T2_2) = caddr(step);
	  }
	else
	  {
	    car(sc->T2_2) = slot_value(car_args);
	    car(sc->T2_1) = cadr(step);
	  }
	slot_set_value(car_args, c_call(step)(sc, sc->T2_1));
      }

    SIMPLE_DO_END:
      {
	s7_pointer end_test, args, code;
	args = sc->args;
	code = sc->code;
	end_test = car(cadr(code));
	car(sc->T2_1) = slot_value(car(args));
	car(sc->T2_2) = slot_value(cdr(args));
	if (is_true(sc, c_call(end_test)(sc, sc->T2_1)))
	  {
	    sc->code = cdr(cadr(code));
	    goto BEGIN;
	  }
	push_stack(sc, OP_SIMPLE_DO_STEP, args, code);
	sc->code = cddr(code);
	goto BEGIN;
      }



    SAFE_SIMPLE_DO:
    case OP_SAFE_SIMPLE_DO:
      {
	/* body is safe in this case, step and end exprs are easy
	 */

	s7_pointer init, end, end_test, body, step, args, val, val2, s1, s2, cadr_step, caddr_step;
	s7_pointer car_args = sc->NIL, cadr_args = sc->NIL, caddr_args = sc->NIL;
	bool cadr_step_is_symbol;

	sc->envir = new_frame_in_env(sc, sc->envir);
	
	init = cadaar(sc->code);
	step = caddr(caar(sc->code));
	end_test = car(cadr(sc->code));
	body = caddr(sc->code);

	args = cdr(body);
	if (!is_null(args))
	  {
	    car_args = car(args);
	    if (!is_null(cdr(args)))
	      {
		cadr_args = cadr(args);
		if (!is_null(cddr(args)))
		  caddr_args = caddr(args);
	      }
	  }

	cadr_step = cadr(step);
	cadr_step_is_symbol = (is_symbol(cadr_step));
	caddr_step = caddr(step);

	if (is_symbol(init))
	  sc->value = finder(sc, init);
	else 
	  {
	    if (is_pair(init))
	      sc->value = c_call(init)(sc, cdr(init));
	    else sc->value = init;
	  }
	
	end = caddr(end_test);
	if (is_symbol(end))
	  {
	    s1 = add_slot(sc, caaar(sc->code), sc->value);
	    s2 = find_symbol(sc, end);
	  }
	else 
	  {
	    NEW_CELL_NO_CHECK(sc, s2);
	    slot_symbol(s2) = end;
	    slot_set_value(s2, end);
	    set_type(s2, T_SLOT | T_IMMUTABLE);
	    s1 = add_slot(sc, caaar(sc->code), sc->value);
	  }
	initialize_safe_do(sc, sc->code);

      SAFE_SIMPLE_DO_END:
	car(sc->T2_1) = slot_value(s1);
	car(sc->T2_2) = slot_value(s2);
	if (is_true(sc, c_call(end_test)(sc, sc->T2_1)))
	  {
	    finalize_safe_do(sc);
	    sc->code = cdr(cadr(sc->code));
	    goto BEGIN;
	  }

	switch (optimize_data(body))
	  {
	  case HOP_SAFE_C_C:
	    c_call(body)(sc, args);
	    break;
	    
	  case HOP_SAFE_C_S:
	    car(sc->T1_1) = finder(sc, car_args);
	    c_call(body)(sc, sc->T1_1);
	    break;
	    
	  case HOP_SAFE_C_SS:
	    val = finder(sc, car_args);
	    car(sc->T2_2) = finder(sc, cadr_args);
	    car(sc->T2_1) = val;
	    c_call(body)(sc, sc->T2_1);
	    break;
	    
	  case HOP_SAFE_C_SSS:
	    val = find_symbol_or_bust(sc, car_args);
	    val2 = find_symbol_or_bust(sc, cadr_args);
	    car(sc->T3_3) = find_symbol_or_bust(sc, caddr_args);
	    car(sc->T3_1) = val;
	    car(sc->T3_2) = val2;
	    c_call(body)(sc, sc->T3_1);
	    break;
	  }
      
	if (cadr_step_is_symbol)
	  {
	    car(sc->T2_1) = slot_value(s1);
	    car(sc->T2_2) = caddr_step;
	  }
	else
	  {
	    car(sc->T2_2) = slot_value(s1);
	    car(sc->T2_1) = cadr_step;
	  }
	slot_set_value(s1, c_call(step)(sc, sc->T2_1));
	goto SAFE_SIMPLE_DO_END;
      }



    SIMPLE_DO_OPT:
    case OP_SIMPLE_DO_OPT:
      {
	s7_pointer init, end, end_test, body, step, args, val, val2, s1, s2, cadr_step, caddr_step;
	s7_pointer car_args = sc->NIL, cadr_args = sc->NIL, caddr_args = sc->NIL;
	bool cadr_step_is_symbol;
	sc->envir = new_frame_in_env(sc, sc->envir);
	
	init = cadaar(sc->code);
	step = caddr(caar(sc->code));
	end_test = car(cadr(sc->code));
	body = caddr(sc->code);

	args = cdr(body);
	if (!is_null(args))
	  {
	    car_args = car(args);
	    if (!is_null(cdr(args)))
	      {
		cadr_args = cadr(args);
		if (!is_null(cddr(args)))
		  caddr_args = caddr(args);
	      }
	  }

	cadr_step = cadr(step);
	cadr_step_is_symbol = (is_symbol(cadr_step));
	caddr_step = caddr(step);

	if (is_symbol(init))
	  sc->value = finder(sc, init);
	else 
	  {
	    if (is_pair(init))
	      sc->value = c_call(init)(sc, cdr(init));
	    else sc->value = init;
	  }
	
	end = caddr(end_test);
	if (is_symbol(end))
	  {
	    s1 = add_slot(sc, caaar(sc->code), sc->value);
	    s2 = find_symbol(sc, end);
	  }
	else 
	  {
	    NEW_CELL_NO_CHECK(sc, s2);
	    slot_symbol(s2) = end;
	    slot_set_value(s2, end);
	    set_type(s2, T_SLOT | T_IMMUTABLE);
	    s1 = add_slot(sc, caaar(sc->code), sc->value);
	  }

      SIMPLE_DO_END_OPT:
	car(sc->T2_1) = slot_value(s1);
	car(sc->T2_2) = slot_value(s2);
	if (is_true(sc, c_call(end_test)(sc, sc->T2_1)))
	  {
	    sc->code = cdr(cadr(sc->code));
	    goto BEGIN;
	  }

	switch (optimize_data(body))
	  {
	  case HOP_SAFE_C_C:
	    c_call(body)(sc, args);
	    break;
	    
	  case HOP_SAFE_C_S:
	    car(sc->T1_1) = finder(sc, car_args);
	    c_call(body)(sc, sc->T1_1);
	    break;
	    
	  case HOP_SAFE_C_SS:
	    val = finder(sc, car_args);
	    car(sc->T2_2) = finder(sc, cadr_args);
	    car(sc->T2_1) = val;
	    c_call(body)(sc, sc->T2_1);
	    break;
	    
	  case HOP_SAFE_C_SSS:
	    val = find_symbol_or_bust(sc, car_args);
	    val2 = find_symbol_or_bust(sc, cadr_args);
	    car(sc->T3_3) = find_symbol_or_bust(sc, caddr_args);
	    car(sc->T3_1) = val;
	    car(sc->T3_2) = val2;
	    c_call(body)(sc, sc->T3_1);
	    break;
	  }

	/* this is actually not safe.  Perhaps I should not use define_safe_procedure in Snd for
	 *    functions like g_open_sound.  It can call both a set of hooks, and s7_load, and
	 *    even sc->code is not guaranteed to survive it.  Since the calls would be recursive,
	 *    perhaps our locals will be ok.
	 */
      
	if (cadr_step_is_symbol)
	  {
	    car(sc->T2_1) = slot_value(s1);
	    car(sc->T2_2) = caddr_step;
	  }
	else
	  {
	    car(sc->T2_2) = slot_value(s1);
	    car(sc->T2_1) = cadr_step;
	  }
	slot_set_value(s1, c_call(step)(sc, sc->T2_1));
	goto SIMPLE_DO_END_OPT;
      }


    SIMPLE_DO_P:
    case OP_SIMPLE_DO_P:
      {
	s7_pointer init, end;
	sc->envir = new_frame_in_env(sc, sc->envir);

	init = cadaar(sc->code);
	if (is_symbol(init))
	  sc->value = finder(sc, init);
	else 
	  {
	    if (is_pair(init))
	      sc->value = c_call(init)(sc, cdr(init));
	    else sc->value = init;
	  }
	
	end = caddr(car(cadr(sc->code)));
	if (is_symbol(end))
	  sc->args = cons_unchecked(sc, add_slot(sc, caaar(sc->code), sc->value), find_symbol(sc, end));
	else 
	  {
	    s7_pointer slot;
	    NEW_CELL_NO_CHECK(sc, slot);
	    slot_symbol(slot) = end;
	    slot_set_value(slot, end);
	    set_type(slot, T_SLOT | T_IMMUTABLE);
	    sc->args = cons_unchecked(sc, add_slot(sc, caaar(sc->code), sc->value), slot);
	  }
	goto SIMPLE_DO_END_P;
      }

    case OP_SIMPLE_DO_STEP_P:
      {
	s7_pointer step, car_args;
	car_args = car(sc->args);
	step = fcdr(sc->code); /* caddr(caar(sc->code)) */
	if (is_symbol(cadr(step)))
	  {
	    car(sc->T2_1) = slot_value(car_args);
	    car(sc->T2_2) = caddr(step);
	  }
	else
	  {
	    car(sc->T2_2) = slot_value(car_args);
	    car(sc->T2_1) = cadr(step);
	  }
	slot_set_value(car_args, c_call(step)(sc, sc->T2_1));
      }

    SIMPLE_DO_END_P:
      {
	s7_pointer end_test, code, args;
	args = sc->args;
	code = sc->code;
	end_test = car(cadr(code));
	car(sc->T2_1) = slot_value(car(args));
	car(sc->T2_2) = slot_value(cdr(args));
	if (is_true(sc, c_call(end_test)(sc, sc->T2_1)))
	  {
	    sc->code = cdr(cadr(code));
	    goto BEGIN;
	  }
	push_stack(sc, OP_SIMPLE_DO_STEP_P, args, code);
	code = caddr(code);
	sc->cur_code = code;
	sc->op = (opcode_t)syntax_opcode(car(code));
	sc->code = cdr(code);
	goto START_WITHOUT_POP_STACK;
      }



      /* DOX cases */
    DOX:
    case OP_DOX:
      /* any number of steppers using dox exprs, end also dox, body and end result arbitrary.
       *    since all these exprs are local, we don't need to jump until the body
       */
      {
	s7_pointer frame, vars, slot;
	NEW_FRAME(sc, sc->envir, frame); /* new frame is not tied into the symbol lookup process yet */
	for (vars = car(sc->code); is_pair(vars); vars = cdr(vars))
	  {
	    NEW_CELL_NO_CHECK(sc, slot);	
	    slot_symbol(slot) = caar(vars);
	    slot_set_value(slot, init_dox_eval(sc, cadar(vars)));
	    slot_expression(slot) = cddar(vars);
	    set_type(slot, T_SLOT | T_IMMUTABLE);
	    next_slot(slot) = environment_slots(frame);
	    environment_slots(frame) = slot;
	  }
	sc->envir = frame;
	for (slot = environment_slots(frame); is_slot(slot); slot = next_slot(slot))
	  {
	    symbol_id(slot_symbol(slot)) = frame_id(frame);    
	    local_slot(slot_symbol(slot)) = slot;
	  }
      }
      goto DOX_END;
	

      case OP_DOX_STEP:
	{
	  s7_pointer slot;
	  for (slot = environment_slots(sc->envir); is_slot(slot); slot = next_slot(slot))
	    {
	      if (is_pair(slot_expression(slot)))
		slot_pending_value(slot) = step_dox_eval(sc, car(slot_expression(slot)));
	      else slot_pending_value(slot) = slot_value(slot);
	    }
	  for (slot = environment_slots(sc->envir); is_slot(slot); slot = next_slot(slot))
	    slot_set_value(slot, slot_pending_value(slot));
	}

      DOX_END:
	if (is_true(sc, end_dox_eval(sc, caadr(sc->code))))
	  {
	    /* if no end result exprs, we return nil, but others probably #<unspecified>
	     *    (let ((x (do ((i 0 (+ i 1))) (#t)))) x) -> '()
	     */
	    sc->code = cdadr(sc->code);
	    goto BEGIN;
	  }
	push_stack(sc, OP_DOX_STEP, sc->NIL, sc->code);
	sc->code = cddr(sc->code);
	goto BEGIN;
#endif

      
	/* we could use slot_pending_value, slot_expression, not this extra list, but the list seems simpler.
	 */
    #define DO_VAR_SLOT(P) ecdr(P)
    #define DO_VAR_NEW_VALUE(P) cdr(P)
    #define DO_VAR_STEP_EXPR(P) car(P)

    case OP_DO_STEP:
      /* increment all vars, return to endtest 
       *   these are also updated in parallel at the end, so we gather all the incremented values first
       */
      
      /* here we know car(sc->args) is not null */
      push_stack(sc, OP_DO_END, sc->args, sc->code);
      sc->args = car(sc->args);                /* the var data lists */
      sc->code = sc->args;                     /* save the top of the list */


    DO_STEP1:
      /* on each iteration, we first get here with args as the list of var bindings, exprs, and init vals
       *   e.g. (((i . 0) (+ i 1) 0))
       * each arg incr expr is evaluated and the value placed in caddr while we cdr down args
       * finally args is nil...
       */

      if (is_null(sc->args))
	{
	  s7_pointer x;
	  for (x = sc->code; is_not_null(x); x = cdr(x))
	    slot_set_value(DO_VAR_SLOT(car(x)), DO_VAR_NEW_VALUE(car(x)));

	  /* some schemes rebind here, rather than reset, but that is expensive,
	   *    and only matters once in a blue moon (closure over enclosed lambda referring to a do var)
	   *    and the caller can easily mimic the correct behavior in that case by adding a let or using a named let,
	   *    making the rebinding explicit.
	   *
	   * Hmmm... I'll leave this alone, but there are other less cut-and-dried cases:
	   *
	   *   (let ((j (lambda () 0))
	   *         (k 0))
	   *     (do ((i (j) (j))
	   *          (j (lambda () 1) (lambda () (+ i 1)))) ; bind here hits different "i" than set!
	   *         ((= i 3) k)
	   *       (set! k (+ k i))))
	   *
	   *   is it 6 or 3?
	   *
	   * if we had a way to tell that there were no lambdas in the do expression, would that
	   *   guarantee that set was ok?  Here's a bad case:
	   *
	   *   (let ((f #f))
	   *     (do ((i 0 (+ i 1)))
	   *         ((= i 3))
	   *       (let () ; so that the define is ok
	   *         (define (x) i)
	   *         (if (= i 1) (set! f x))))
	   *    (f))
	   *
	   * s7 says 3, guile says 1.
	   *
	   * I wonder if what they're actually talking about is a kind of shared value problem.  If we
	   *   set the value directly (not the cdr(binding) but, for example, integer(cdr(binding))), then
	   *   every previous reference gets changed as a side-effect.  In the current code, we're "binding" 
	   *   the value in the sense that on each step, a new value is assigned to the step variable.
	   *   In the "direct" case, (let ((v #(0 0 0))) (do ((i 0 (+ i 1))) ((= i 3) v) (set! (v i) i)) 
	   *   would return #(3 3 3).
	   */
	  
	  sc->value = sc->NIL;
	  pop_stack_no_op(sc); 
	  goto DO_END;
	}
      push_stack(sc, OP_DO_STEP2, sc->args, sc->code);
      
      /* here sc->args is a list like (((i . 0) (+ i 1) 0) ...)
       *   so sc->code becomes (+ i 1) in this case 
       */
      sc->code = DO_VAR_STEP_EXPR(car(sc->args));
      goto EVAL;
      

    case OP_DO_STEP2:
      DO_VAR_NEW_VALUE(car(sc->args)) = sc->value;            /* save current value */
      sc->args = cdr(sc->args);                               /* go to next step var */
      goto DO_STEP1;


    case OP_DO: 
      /* setup is very similar to let */
      /* sc->code is the stuff after "do" */
#if WITH_OPTIMIZATION
      if (is_null(check_do(sc)))
	{
	  if (car(ecdr(sc->code)) == sc->SIMPLE_DO)
	    goto SIMPLE_DO;
	  if (car(ecdr(sc->code)) == sc->SIMPLE_DO_P)
	    goto SIMPLE_DO_P;
	  if (car(ecdr(sc->code)) == sc->SIMPLE_DO_OPT)
	    goto SIMPLE_DO_OPT;
	  if (car(ecdr(sc->code)) == sc->SAFE_SIMPLE_DO)
	    goto SAFE_SIMPLE_DO;
	  if (car(ecdr(sc->code)) == sc->DOTIMES_C_C)
	    goto DOTIMES_C_C;
	  if (car(ecdr(sc->code)) == sc->SIMPLE_DOTIMES)
	    goto SIMPLE_DOTIMES;
	  if (car(ecdr(sc->code)) == sc->SAFE_DO)
	    goto SAFE_DO;
	  if (car(ecdr(sc->code)) == sc->DOX)
	    goto DOX;
	  goto DOTIMES;
	}
#else
      check_do(sc);
#endif


    case OP_DO_UNCHECKED:
      if (is_null(car(sc->code)))                           /* (do () ...) -- (let ((i 0)) (do () ((= i 1)) (set! i 1))) */
	{
	  sc->envir = new_frame_in_env(sc, sc->envir); 
	  sc->args = cons_unchecked(sc, sc->NIL, cadr(sc->code));
	  sc->code = cddr(sc->code);
	  goto DO_END;
	}
      
      /* eval each init value, then set up the new frame (like let, not let*) */
      sc->args = sc->NIL;                             /* the evaluated var-data */
      sc->value = sc->code;                           /* protect it */
      sc->code = car(sc->code);                       /* the vars */
      

    DO_INIT:
    case OP_DO_INIT:
      sc->args = cons(sc, sc->value, sc->args);    /* code will be last element (first after reverse) */
      if (is_pair(sc->code))
	{
	  /* here sc->code is a list like: ((i 0 (+ i 1)) ...)
	   *   so cadar gets the init value.
	   *
	   * we accept:
	   *       (do ((i 1) (i 2)) (#t i)) -> 2
	   *       (do () (1) . "hi") -- this is like (do () (#t #t) (asdf))
	   */
	  s7_pointer init;
	  init = cadar(sc->code);
	  if (is_pair(init))
	    {
	      push_stack(sc, OP_DO_INIT, sc->args, cdr(sc->code));
	      sc->code = init;
	      goto EVAL_PAIR;
	    }
	  if (is_symbol(init))
	    sc->value = finder(sc, init);
	  else sc->value = init;
	  sc->code = cdr(sc->code);
	  goto DO_INIT;
	}

      /* all the initial values are now in the args list */
      {
	s7_pointer x, y, z;
	sc->args = safe_reverse_in_place(sc, sc->args);
	sc->code = car(sc->args);                       /* saved at the start */
	z = sc->args;
	sc->args = cdr(sc->args);                       /* init values */

	/* sc->envir = new_frame_in_env(sc, sc->envir); */
	/* sc->args was cons'd above, so it should be safe to reuse it as the new frame */
	environment_slots(z) = sc->NIL;                  
	next_environment(z) = sc->envir;
	set_type(z, T_ENVIRONMENT | T_IMMUTABLE); 
	frame_id(z) = ++frame_number; 
	sc->envir = z;
	
	/* run through sc->code and sc->args adding '( caar(car(code)) . car(args) ) to sc->envir,
	 *    also reuse the value cells as the new frame slots.
	 */
	sc->value = sc->NIL;
	y = sc->args;
	for (x = car(sc->code); is_not_null(y); x = cdr(x)) 
	  {
	    s7_pointer sym, args, val;
	    sym = caar(x);
	    val = car(y);
	    args = cdr(y);
	    
	    slot_symbol(y) = sym;

	    if (symbol_has_accessor(sym))
	      {
		s7_pointer old_value;
		old_value = sc->value;
		slot_set_value(y, call_symbol_bind(sc, sym, val)); /* this might change sc->value */
		sc->value = old_value;
	      }
	    else slot_set_value(y, val);

	    /* var access checks during the step section? 
	     *
	     * to embed here, we need x, y, (sc->args? sc->code?), sc->value
	     *
	     * label at start of loop plus jump back at end
	     * if accessor and binder is scheme func
	     *   push_stack DO_INIT_WITH_ACCESSOR, args=x y value args, code, jumping back after apply
	     *   args='(sym val), code=scheme func, goto APPLY
	     *   DO_INIT_WITH_ACCESSOR checks for 'error and jumps to
	     * DO_INIT_ACCESS:
	     *   reset x y code args from saved stuff
	     *   set cdr(y) from new value
	     *   restore old sc->value
	     * [before saving y, set its type?][will a gc clobber anything here?]
	     */

	    set_type(y, T_SLOT | T_IMMUTABLE);
	    next_slot(y) = environment_slots(sc->envir);
	    environment_slots(sc->envir) = y;
	    symbol_id(sym) = frame_id(sc->envir);
	    local_slot(sym) = y;

	    if (is_not_null(cddar(x)))                /* else no incr expr, so ignore it henceforth */
	      {
		s7_pointer p;
		p = cons(sc, caddar(x), val);
		ecdr(p) = y;
					              /* val is just a place-holder -- this is where we store the new value */
		sc->value = cons_unchecked(sc, p, sc->value);
	      }
	    y = args;
	  }
	sc->args = cons(sc, safe_reverse_in_place(sc, sc->value), cadr(sc->code));
	sc->code = cddr(sc->code);
	
	/* here args is a list of 2 or 3 lists, 1st is (list (list (var . binding) incr-expr init-value) ...), 2nd is end-expr, 3rd can be result expr
	 *   so for (do ((i 0 (+ i 1))) ((= i 3) (+ i 1)) ...) args is ((((i . 0) (+ i 1) 0 #f)) (= i 3) (+ i 1))
	 */
      }


    DO_END:
    case OP_DO_END:
      /* here vars have been init'd or incr'd
       *    args = (list var-data end-expr return-expr-if-any)
       *      if (do ((i 0 (+ i 1))) ((= i 3) 10)),            args: (vars (= i 3) 10)
       *      if (do ((i 0 (+ i 1))) ((= i 3))),               args: (vars (= i 3)) and result expr is () == (begin)
       *      if (do ((i 0 (+ i 1))) (#t 10 12)),              args: (vars #t 10 12), result: ([begin] 10 12) -> 12 
       *      if (call-with-exit (lambda (r) (do () () (r)))), args: '(())
       *    code = body
       */

      if (is_not_null(cdr(sc->args)))
	{
	  push_stack(sc, OP_DO_END1, sc->args, sc->code);
	  sc->code = cadr(sc->args);               /* evaluate the end expr */
	  goto EVAL;
	}
      else 
	{
	  /* (do ((...)) () ...) -- no endtest */

	  if (is_null(car(sc->args)))
	    push_stack(sc, OP_DO_END, sc->args, sc->code);
	  else push_stack(sc, OP_DO_STEP, sc->args, sc->code);

	  goto BEGIN;
	}


    case OP_DO_END1:
      /* sc->value is the result of end-test evaluation */
      if (is_true(sc, sc->value))
	{
	  /* we're done -- deal with result exprs 
	   *   if there isn't an end test, there also isn't a result (they're in the same list)
	   */
	  sc->code = cddr(sc->args);                /* result expr (a list -- implicit begin) */
	  
	  typeflag(sc->args) = 0;
	  (*(sc->free_heap_top++)) = sc->args;
	  sc->args = sc->NIL;

	  if (is_null(sc->code))
	    {
	      sc->value = sc->NIL;
	      goto START;
	    }
	}
      else
	{
	  /* evaluate the body and step vars, etc */
	  if (is_null(car(sc->args)))
	    push_stack(sc, OP_DO_END, sc->args, sc->code);
	  else push_stack(sc, OP_DO_STEP, sc->args, sc->code);
	  /* sc->code is ready to go */
	}
      /* fall through
       *
       *    the other most common goto BEGIN's are OP_LET and the closure branch of OP_APPLY,
       *    but moving OP_LET here was slightly slower even in lg which has few do-loops.
       *    goto BEGIN2 in these cases (to avoid the error checks below) was much slower, apparently
       *    because there was one extra goto (to EVAL or EVAL_PAIR).
       */



      /* -------------------------------- BEGIN -------------------------------- */

    BEGIN:
    case OP_BEGIN:
      /* sc->args is not used here */
      if (sc->begin_hook)
	{
	  opcode_t op;
	  op = sc->op;
	  push_stack(sc, OP_BARRIER, sc->args, sc->code);
	  if ((*(sc->begin_hook))(sc))
	    {
	      s7_quit(sc);
	      /* don't call gc here -- perhaps at restart somehow? */
	      return(sc->F);
	    }
	  pop_stack_no_op(sc);
	  sc->op = op; /* for better error handling.  otherwise we get "barrier" as the offending function name in eval_error_with_name */
	}
      /* fall through */

      /* 1st statement of more than 1:
       * s7: [75467, 240203] OP_DEFINE_UNCHECKED: 18129 OP_SET_UNCHECKED: 6842 OP_SET_SYMBOL_P: 3125 OP_SAFE_IF_CC_P: 11265 OP_SAFE_IF1: 5325 h_safe_c_sss: 4400
       * lg: [828113, 286974] OP_IF_B_P: 173866 closure_ss: 105271 closure_all_s: 83812 OP_SAFE_IF_CC_P_P: 67182
       * in: [750321, 7873843] OP_IF_B_P: 207387 OP_SIMPLE_DO: 82858 OP_IF_X_P_P: 78918 OP_DECREMENT_1: 79191 OP_IF_X_P: 66396
       */
      /* in gcc we can use computed gotos, setting begin_label to this position if
       *   begin_hook is not in use, thereby avoiding that check above.  This saves
       *   about 4 in lg -- sort of a toss-up given the added complexity.
       */
    case OP_BEGIN1:
      /* use of local for sc->code here saves about .5
       */
      if (!is_pair(sc->code))                   /* (begin) -> () */
	{
	  if (is_not_null(sc->code))            /* (begin . 1), (cond (#t . 1)) */
	    return(eval_error_with_name(sc, "~A: unexpected dot or '() at end of body? ~A", sc->code));
	  sc->value = sc->NIL;
	  goto START;
	}

      if (type(cdr(sc->code)) != T_NIL)
	{
	  /* if there are more expressions following, and this is not a pair, it's a no-op!
	   *   but it also might be an error, which I suppose we should catch: (begin +> 1) 
	   *
	   * this double type check is hard to get around: everything would work if we pushed
	   *    nil and caught it on the rebound except multiple values and tail recursion.
	   *    (+ (begin 3 (values 1 2) 4)), (+ 1 (begin (values 5 6) (values 2 3))) etc
	   * moving the is_pair check above up to OP_BEGIN and so on is slower.
	   * also checking above for is_pair (to try to avoid extra push_stacks) is
	   *    slightly slower, and the jump to EVAL_PAIR messes up the cur_code error info
	   * and moving the is_pair(code) check out (via check_begin etc) is troublesome because
	   *    do loops (perhaps others) assume they can fall into begin without any null body check.
	   */
	  push_stack_no_args(sc, OP_BEGIN1, cdr(sc->code)); 
	  /* this is slightly less than optimal because a stacktrace may show bogus args, but
	   *    I think the speed differences outweighs that.  In most cases, this is by far the
	   *    main push_stack call.
	   */
	}
      sc->code = car(sc->code);

    EVAL:
    case OP_EVAL: 
      /* main part of evaluation 
       *   at this point, it's sc->code we care about; sc->args is not relevant.
       */
      /* fprintf(stderr, "    eval: %s\n", DISPLAY_80(sc->code)); */

      if (is_pair(sc->code))
	{
	  /* we jump here when we already know sc->code is a pair */

	EVAL_PAIR:
	  /* fprintf(stderr, "    eval_pair: %s\n", DISPLAY_80(sc->code)); */

	  if (typeflag(car(sc->code)) == SYNTACTIC_TYPE) /* this is much faster than checking is_synactic */
	    {
	      sc->cur_code = sc->code;               /* in case an error occurs, this helps tell us where we are */

	      sc->op = (opcode_t)syntax_opcode(car(sc->code));
	      sc->code = cdr(sc->code);
	      goto START_WITHOUT_POP_STACK;
	      /* it is only slightly faster to use labels as values (computed gotos) here
	       */
	    }

	  /* sc->code is a pair, car(sc->code) is not syntactic */
#if WITH_OPTIMIZATION
	  if (is_optimized(sc->code))
	    {
	      s7_pointer code;

	    OPT_EVAL:
	      code = sc->code;
	      sc->cur_code = code;

	      switch (optimize_data(code))
		{
		case OP_NOT_AN_OP:
		case HOP_NOT_AN_OP:
		  break;
		  
		case OP_THUNK:
		  if (!function_is_ok(code))
		    break;
		  
		case HOP_THUNK:
		  NEW_FRAME(sc, closure_environment(ecdr(code)), sc->envir);
		  sc->code = closure_body(ecdr(code));
		  goto BEGIN;


		case OP_SAFE_THUNK:
		  if (!function_is_ok(code))
		    break;
		  
		case HOP_SAFE_THUNK:
		  /* (let ((x 1)) (let () (define (f) x) (let ((x 0)) (define (g) (set! x 32) (f)) (g)))) */
		  NEW_FRAME(sc, closure_environment(ecdr(code)), sc->envir);
		  sc->code = closure_body(ecdr(code));
		  goto BEGIN;
		  
		  
		  /* if we actually use the HOP_* mechanism with closures (see optimize_function), and
		   *   we use ecdr below, the nested closure is not looked up again at
		   *   run-time in the current environment, so
		   *      (let () (define (f2 a) (+ a 1)) (define (f1 a) (f2 a)) (define (f2 a) (- a)) (f1 12))
		   *   returns 13, whereas in the unoptimized version of s7 it will return -12.
		   * is this a bug in either case?  what about:
		   *      (let () (define (f2 a) (+ a 1)) (define (f1 a) (f2 a)) (define (f2 a b) (- a b)) (f1 12))
		   * but... we can't use ecdr reliably -- the saved pointer might be to an old version not GC'd etc.
		   */
		  
		case OP_SAFE_CLOSURE_S:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_S;
		      goto OPT_EVAL;
		    }
		  
		case HOP_SAFE_CLOSURE_S:
		  /* since a tail call is safe, we can't change the current env's frame_id until
		   *   after we do the lookup -- it might be the current func's arg, and we're
		   *   about to call the same func.
		   */
		  {
		    s7_pointer env, x;

		    env = closure_environment(ecdr(code));
		    x = environment_slots(env);
		    slot_set_value(x, finder(sc, fcdr(code)));

		    frame_id(env) = ++frame_number;
		    symbol_id(slot_symbol(x)) = frame_number;
		    local_slot(slot_symbol(x)) = x;
		    sc->envir = env;
		    sc->code = closure_body(ecdr(code));
		    if (is_one_liner(sc->code))
		      {
			sc->code = car(sc->code);
			/* lg: none
			 */
			goto EVAL_PAIR;
		      }
		    goto BEGIN;
		  }
		  
		  
		case OP_SAFE_CLOSURE_S_one:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_S;
		      goto OPT_EVAL;
		    }
		  
		case HOP_SAFE_CLOSURE_S_one:
		  {
		    s7_pointer env, x;

		    env = closure_environment(ecdr(code));
		    x = environment_slots(env);
		    slot_set_value(x, finder(sc, fcdr(code)));

		    frame_id(env) = ++frame_number;
		    symbol_id(slot_symbol(x)) = frame_number;
		    local_slot(slot_symbol(x)) = x;
		    sc->envir = env;
		    sc->code = car(closure_body(ecdr(code)));
		    /* lg: syn: 39750 h_safe_c_sq: 45690
		     * in: syn: 23871
		     * al: syn: 4534616 h_safe_c_c: 456347 h_safe_c_ss: 863554 [h_safe_c_s: 4467]
		     * 
		     * if syn:
		     * lg  OP_SAFER_OR_S: 34582
		     * in: OP_SAFER_OR_S: 18391
		     * al: OP_LET_READ_CHAR_P: 4460056 OP_AND_P: 70741
		     */
		    goto EVAL_PAIR;
		  }


		case OP_SAFE_CLOSURE_S_vref:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_S;
		      goto OPT_EVAL;
		    }
		  
		case HOP_SAFE_CLOSURE_S_vref:
		  {
		    s7_pointer v;
		    v = finder(sc, fcdr(code));
		    if (!s7_is_vector(v))
		      return(s7_wrong_type_arg_error(sc, "vector-ref", 1, v, "a vector"));
		    if (!vector_is_multidimensional(v))
		      {
			s7_Int index;
			index = s7_integer(caddr(car(closure_body(ecdr(code)))));
			if (index < vector_length(v))
			  {
			    sc->value = vector_elements(v)[index];
			    goto START;
			  }
		      }
		    sc->value = vector_ref_1(sc, v, cdr(code)); /* ok because it's just an integer constant */
		    goto START;
		  }

		  
		case OP_SAFE_CLOSURE_C:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_C;
		      goto OPT_EVAL;
		    }
		  
		case HOP_SAFE_CLOSURE_C:
		  sc->value = cadr(code);
		  goto SAFE_CLOSURE_ONE;
		  
		  
		case OP_SAFE_CLOSURE_Q:
		  if (!function_is_ok(code))
		    break;
		  
		case HOP_SAFE_CLOSURE_Q:
		  sc->value = cadr(cadr(code));
		  goto SAFE_CLOSURE_ONE;
		  
		  
		case OP_SAFE_CLOSURE_opCq:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_opCq;
		      goto OPT_EVAL;
		    }
		  if (!indirect_c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_SAFE_CLOSURE_opCq:
		  sc->value = c_call(fcdr(code))(sc, cdr(fcdr(code)));
		  goto SAFE_CLOSURE_ONE;


		case OP_SAFE_CLOSURE_opSq:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_opSq;
		      goto OPT_EVAL;
		    }
		  if (!indirect_c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_SAFE_CLOSURE_opSq:
		  car(sc->T1_1) = finder(sc, fcdr(code));
		  sc->value = c_call(cadr(code))(sc, sc->T1_1);
		  /* goto SAFE_CLOSURE_ONE; */
		  
		  
		SAFE_CLOSURE_ONE:
		  {
		    s7_pointer env, x;
		    env = closure_environment(ecdr(code));
		    frame_id(env) = ++frame_number;
		    x = environment_slots(env);
		    slot_set_value(x, sc->value);
		    symbol_id(slot_symbol(x)) = frame_number;
		    local_slot(slot_symbol(x)) = x;
		    sc->envir = env;
		    sc->code = closure_body(ecdr(code));

		    /* if (is_one_liner(sc->code))
		     * lg: 20243 syn [OP_SAFER_OR_S: 17530]
		     * in: none
		     * al: unopt: 124269
		     */
		    goto BEGIN;
		  }


		case OP_SAFE_CLOSURE_SS:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_SS;
		      goto OPT_EVAL;
		    }
		  
		case HOP_SAFE_CLOSURE_SS:
		  {
		    s7_pointer val;
		    val = finder(sc, fcdr(code));
		    car(sc->T2_1) = finder(sc, cadr(code));
		    car(sc->T2_2) = val;
		    sc->code = ecdr(code);
		    /* goto SAFE_CLOSURE_TWO; */
		  }
		  
		SAFE_CLOSURE_TWO:
		  {
		    s7_pointer env, x;
		    unsigned long long int id;
		    
		    id = ++frame_number;
		    env = closure_environment(sc->code);
		    frame_id(env) = id;
		    
		    x = environment_slots(env);
		    slot_set_value(x, car(sc->T2_1));
		    symbol_id(slot_symbol(x)) = id;
		    local_slot(slot_symbol(x)) = x; /* this is not redundant */
		    
		    x = next_slot(x);
		    slot_set_value(x, car(sc->T2_2));
		    symbol_id(slot_symbol(x)) = id;
		    local_slot(slot_symbol(x)) = x; 
		    
		    sc->envir = env;
		    sc->code = closure_body(sc->code);
		    
		    if (is_one_liner(sc->code))
		      {
			sc->code = car(sc->code);
			goto EVAL_PAIR; 
			/* lg: syn: OP_SAFER_OR_X: 319892 
			 * al: OP_IF_P_X_P: 24986
			 */
		      }
		    goto BEGIN;
		  }
		  
		  
		case OP_SAFE_CLOSURE_SC:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_SC;
		      goto OPT_EVAL;
		    }
		  
		case HOP_SAFE_CLOSURE_SC:
		  car(sc->T2_1) = finder(sc, cadr(code));
		  car(sc->T2_2) = fcdr(code);
		  sc->code = ecdr(code);
		  goto SAFE_CLOSURE_TWO;
		  
		  
		case OP_SAFE_CLOSURE_CS:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_CS;
		      goto OPT_EVAL;
		    }
		  
		case HOP_SAFE_CLOSURE_CS:
		  car(sc->T2_2) = finder(sc, fcdr(code));
		  car(sc->T2_1) = cadr(code);
		  sc->code = ecdr(code);
		  goto SAFE_CLOSURE_TWO;
		  
		  
		case OP_SAFE_CLOSURE_CC:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_CC;
		      goto OPT_EVAL;
		    }
		  
		case HOP_SAFE_CLOSURE_CC:
		  car(sc->T2_1) = cadr(code);
		  car(sc->T2_2) = fcdr(code);
		  sc->code = ecdr(code);
		  goto SAFE_CLOSURE_TWO;
		  
		  
		case OP_SAFE_CLOSURE_opSq_opSq:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_opSq_opSq;
		      goto OPT_EVAL;
		    }
		  if (!indirect_c_function_is_ok(sc, cadr(code)))
		    break;
		  if (!indirect_c_function_is_ok(sc, caddr(code)))
		    break;
		  
		case HOP_SAFE_CLOSURE_opSq_opSq:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    
		    car(sc->T1_1) = finder(sc, cadr(car(args)));
		    sc->temp4 = c_call(car(args))(sc, sc->T1_1);
		    car(sc->T1_1) = finder(sc, fcdr(code));
		    
		    car(sc->T2_2) = c_call(cadr(args))(sc, sc->T1_1);
		    car(sc->T2_1) = sc->temp4;
		    sc->code = ecdr(code);
		    goto SAFE_CLOSURE_TWO;
		  }
		  
		  
		case OP_SAFE_CLOSURE_SSS:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_SSS;
		      goto OPT_EVAL;
		    }
		  
		case HOP_SAFE_CLOSURE_SSS:
		  {
		    s7_pointer x, y, z;
		    z = finder(sc, fcdr(code)); 
		    y = finder(sc, caddr(code));
		    x = finder(sc, cadr(code));
		    car(sc->T3_1) = x;
		    car(sc->T3_2) = y;
		    car(sc->T3_3) = z;
		    sc->args = sc->T3_1;
		    sc->code = ecdr(code);
		    /* goto SAFE_CLOSURE; */
		  }
		  
		  
		SAFE_CLOSURE:
		  {
		    s7_pointer env, x, z;
		    unsigned long long int id;
		    
		    id = ++frame_number;
		    env = closure_environment(sc->code);
		    frame_id(env) = id;
		    
		    for (x = environment_slots(env), z = sc->args; is_slot(x); x = next_slot(x), z = cdr(z))
		      {
			slot_set_value(x, car(z));
			symbol_id(slot_symbol(x)) = id;
			local_slot(slot_symbol(x)) = x;
		      }
		    sc->envir = env;
		    sc->code = closure_body(sc->code);
		    
		    if (is_one_liner(sc->code))
		      {
			sc->code = car(sc->code);
			goto EVAL_PAIR;
		      }
		    goto BEGIN;
		  }
		  
		  
		case OP_SAFE_CLOSURE_ALL_S:
		  if (!function_is_ok(code))
		    break;
		  
		case HOP_SAFE_CLOSURE_ALL_S:
		  {
		    s7_pointer args;
		    int i;
		    for (i = 0, args = cdr(code); is_pair(args); i++, args = cdr(args))
		      {
			car(sc->all_g_args[i]) = finder(sc, car(args));
			cdr(sc->all_g_args[i]) = sc->all_g_args[i + 1];
		      }
		    cdr(sc->all_g_args[i - 1]) = sc->NIL;
		    sc->args = sc->all_g_args[0];
		    sc->code = ecdr(code);
		    goto SAFE_CLOSURE;
		  }
		  
		  
		case OP_SAFE_CLOSURE_ALL_C:
		  if (!function_is_ok(code))
		    break;
		  
		case HOP_SAFE_CLOSURE_ALL_C:
		  sc->args = cdr(code);
		  sc->code = ecdr(code);
		  goto SAFE_CLOSURE;
		  

		case OP_SAFE_CLOSURE_ALL_G:
		  if (!function_is_ok(code))
		    break;
		  
		case HOP_SAFE_CLOSURE_ALL_G:
		  {
		    s7_pointer args;
		    int i;

		    for (i = 0, args = cdr(code); is_pair(args); i++, args = cdr(args))
		      {
			s7_pointer arg, val;
			arg = car(args);
			if (is_pair(arg))
			  {
			    if (is_optimized(arg))
			      {
				car(sc->T1_1) = finder(sc, cadr(arg));
				val = c_call(arg)(sc, sc->T1_1);
			      }
			    else val = cadr(arg);
			  }
			else
			  {
			    if (is_symbol(arg))
			      val = finder(sc, arg);
			    else val = arg;
			  }
			car(sc->all_g_args[i]) = val;
			cdr(sc->all_g_args[i]) = sc->all_g_args[i + 1];
		      }
		    cdr(sc->all_g_args[i - 1]) = sc->NIL;
		    sc->args = sc->all_g_args[0];
		    sc->code = ecdr(code);
		    goto SAFE_CLOSURE;
		  }
		  
		  
		case OP_SAFE_CLOSURE_opSq_S:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_opSq_S;
		      goto OPT_EVAL;
		    }
		  if (!indirect_c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_SAFE_CLOSURE_opSq_S:
		  {
		    s7_pointer val1;
		    val1 = finder(sc, fcdr(code));
		    car(sc->T1_1) = finder(sc, cadr(cadr(code)));
		    car(sc->T2_1) = c_call(cadr(code))(sc, sc->T1_1);
		    car(sc->T2_2) = val1;
		    sc->code = ecdr(code);
		    goto SAFE_CLOSURE_TWO;
		  }
		  break;
		  

		case OP_SAFE_CLOSURE_S_opSq:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_S_opSq;
		      goto OPT_EVAL;
		    }
		  if (!indirect_c_function_is_ok(sc, caddr(code)))
		    break;
		  
		case HOP_SAFE_CLOSURE_S_opSq:
		  {
		    s7_pointer val;
		    val = finder(sc, cadr(code)); /* the 1st S */

		    car(sc->T1_1) = finder(sc, cadr(caddr(code)));
		    car(sc->T2_2) = c_call(caddr(code))(sc, sc->T1_1);
		    car(sc->T2_1) = val;
		    sc->code = ecdr(code);
		    goto SAFE_CLOSURE_TWO;
		  }
		  
		  
		case OP_CLOSURE_C:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_C;
		      goto OPT_EVAL;
		    }
		  
		case HOP_CLOSURE_C:
		  sc->value = cadr(code);
		  goto UNSAFE_CLOSURE_ONE;
		  

		case OP_CLOSURE_Q:
		  if (!function_is_ok(code))
		    break;
		  
		case HOP_CLOSURE_Q:
		  sc->value = cadr(cadr(code));
		  goto UNSAFE_CLOSURE_ONE;
		  

		case OP_CLOSURE_opCq:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_opCq;
		      goto OPT_EVAL;
		    }
		  if (!indirect_c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_CLOSURE_opCq:
		  sc->value = c_call(fcdr(code))(sc, cdr(fcdr(code)));
		  goto UNSAFE_CLOSURE_ONE;
		  

		case OP_CLOSURE_opSq:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_opSq;
		      goto OPT_EVAL;
		    }
		  if (!indirect_c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_CLOSURE_opSq:
		  car(sc->T1_1) = find_symbol_or_bust(sc, fcdr(code));
		  sc->value = c_call(cadr(code))(sc, sc->T1_1);
		  goto UNSAFE_CLOSURE_ONE;
		  


		case OP_CLOSURE_CDR_S:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_opSq;
		      goto OPT_EVAL;
		    }
		  if (!indirect_c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_CLOSURE_CDR_S:
		  {
		    s7_pointer p;
		    p = find_symbol_or_bust(sc, cadr(cadr(code)));
		    if (!is_pair(p))
		      s7_wrong_type_arg_error(sc, "cdr", 0, p, "a pair");
		    sc->value = cdr(p);
		    goto UNSAFE_CLOSURE_ONE;
		  }


		case OP_CLOSURE_S:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_S;
		      goto OPT_EVAL;
		    }

		case HOP_CLOSURE_S:
		  sc->value = find_symbol_or_bust(sc, fcdr(code));
		  /* goto UNSAFE_CLOSURE_ONE; */
		  
		UNSAFE_CLOSURE_ONE:
		  {
		    s7_pointer frame, slot, args, sym, rest;
		    
		    if (sc->stack_end >= sc->stack_resize_trigger)
		      increase_stack_size(sc);
		    
		    sc->code = ecdr(sc->code);
		    args = closure_args(sc->code);
		    if (is_pair(args))
		      {
			sym = car(args);
			rest = cdr(args);
		      }
		    else
		      {
			sym = args;
			sc->value = list_1(sc, sc->value);
			rest = sc->NIL;
		      }
		    
		    NEW_CELL(sc, frame);                   
		    frame_id(frame) = ++frame_number; 
		    next_environment(frame) = closure_environment(sc->code);
		    set_type(frame, T_ENVIRONMENT | T_IMMUTABLE); 
		    sc->envir = frame;
		    
		    NEW_CELL_NO_CHECK(sc, slot);
		    slot_symbol(slot) = sym;
		    symbol_id(sym) = frame_number;	
		    local_slot(sym) = slot;
		    if (symbol_has_accessor(sym))
		      slot_set_value(slot, call_symbol_bind(sc, sym, sc->value));
		    else slot_set_value(slot, sc->value);
		    set_type(slot, T_SLOT | T_IMMUTABLE);
		    next_slot(slot) = sc->NIL;
		    environment_slots(frame) = slot;
		    
		    if (!is_null(rest))
		      add_slot(sc, rest, sc->NIL);
		    /* here we have a closure with a required arg and a rest arg, and we're called
		     *   with one arg, so the rest arg is nil.
		     */
		    
		    sc->code = closure_body(sc->code);
		    if (is_one_liner(sc->code))
		      {
			sc->code = car(sc->code);
			goto EVAL_PAIR; 
			/* lg: OP_LET_C: 27987 OP_IF_P_X_P: 27322 OP_LET_OX_P: 21230 OP_SAFE_IF_IS_PAIR_P_X: 27178
			 * al: OP_IF_P_X_P: 2692749 OP_COND_SIMPLER: 261496 OP_IF_P_X_P: 2692749 OP_IF_ORP_P_P: 219828
			 */
		      }
		    goto BEGIN;
		  }


		case OP_CLOSURE_SS:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_SS;
		      goto OPT_EVAL;
		    }
		  
		case HOP_CLOSURE_SS:
		  car(sc->T2_1) = find_symbol_or_bust(sc, cadr(code));
		  car(sc->T2_2) = find_symbol_or_bust(sc, fcdr(code));
		  /* goto UNSAFE_CLOSURE_TWO; */


		UNSAFE_CLOSURE_TWO:
		  {
		    s7_pointer frame, slot1, slot2, args, sym, val, rest;
		    /* the other common tclosure case that still comes here is opsq_opsq
		     */
		    if (sc->stack_end >= sc->stack_resize_trigger)
		      increase_stack_size(sc);
		    sc->code = ecdr(sc->code);
		    args = closure_args(sc->code);
		    if (is_pair(args))
		      {
			sym = car(args);
			val = car(sc->T2_1);
		      }
		    else
		      {
			sym = args;
			val = list_2(sc, car(sc->T2_1), car(sc->T2_2));
		      }
		    
		    NEW_CELL(sc, frame);                   
		    frame_id(frame) = ++frame_number; 
		    next_environment(frame) = closure_environment(sc->code);
		    set_type(frame, T_ENVIRONMENT | T_IMMUTABLE); 
		    sc->envir = frame;
		    
		    NEW_CELL_NO_CHECK(sc, slot1);
		    slot_symbol(slot1) = sym;
		    symbol_id(sym) = frame_number;	
		    local_slot(sym) = slot1;
		    if (symbol_has_accessor(sym))
		      slot_set_value(slot1, call_symbol_bind(sc, sym, val));
		    else slot_set_value(slot1, val);
		    set_type(slot1, T_SLOT | T_IMMUTABLE);
		    environment_slots(frame) = slot1;
		    
		    if (is_pair(args))
		      {
			args = cdr(args);
			if (is_pair(args))
			  {
			    sym = car(args);
			    val = car(sc->T2_2);
			    rest = cdr(args);
			  }
			else
			  {
			    sym = args;
			    val = list_1(sc, car(sc->T2_2));
			    rest = sc->NIL;
			  }
			
			NEW_CELL_NO_CHECK(sc, slot2);
			slot_symbol(slot2) = sym;
			symbol_id(sym) = frame_number;	
			local_slot(sym) = slot2;
			if (symbol_has_accessor(sym))
			  slot_set_value(slot2, call_symbol_bind(sc, sym, val));
			else slot_set_value(slot2, val);
			set_type(slot2, T_SLOT | T_IMMUTABLE);
			next_slot(slot2) = sc->NIL;
			next_slot(slot1) = slot2;
			
			if (!is_null(rest))
			  add_slot(sc, rest, sc->NIL);
			/* here we have a closure with two required args and a rest arg, and we're called
			 *   with two args, so the rest arg is nil.
			 */
		      }
		    else next_slot(slot1) = sc->NIL;
		    
		    sc->code = closure_body(sc->code);
		    if (is_one_liner(sc->code))
		      {
			sc->code = car(sc->code);
			goto EVAL_PAIR; 
			/* lg: OP_LET_O_P: 107534 OP_SAFE_IF_IS_PAIR_P_P: 87672
			 * al: OP_COND_SIMPLE: 219331 OP_NAMED_LET: 221049 OP_COND_SIMPLER: 201751
			 */
		      }
		    goto BEGIN;
		  }


		case OP_CLOSURE_SC:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_SC;
		      goto OPT_EVAL;
		    }
		  
		case HOP_CLOSURE_SC:
		  car(sc->T2_1) = find_symbol_or_bust(sc, cadr(code));
		  car(sc->T2_2) = fcdr(code);
		  goto UNSAFE_CLOSURE_TWO;
		  

		case OP_CLOSURE_CS:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_CS;
		      goto OPT_EVAL;
		    }
		  
		case HOP_CLOSURE_CS:
		  car(sc->T2_1) = cadr(code);
		  car(sc->T2_2) = find_symbol_or_bust(sc, fcdr(code));
		  goto UNSAFE_CLOSURE_TWO;
		  

		case OP_CLOSURE_CC:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_CC;
		      goto OPT_EVAL;
		    }
		  
		case HOP_CLOSURE_CC:
		  car(sc->T2_1) = cadr(code);
		  car(sc->T2_2) = fcdr(code);
		  goto UNSAFE_CLOSURE_TWO;
		  

		case OP_CLOSURE_opSq_opSq:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_opSq_opSq;
		      goto OPT_EVAL;
		    }
		  if (!indirect_c_function_is_ok(sc, cadr(code)))
		    break;
		  if (!indirect_c_function_is_ok(sc, caddr(code)))
		    break;
		  
		case HOP_CLOSURE_opSq_opSq:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    car(sc->T1_1) = find_symbol_or_bust(sc, cadr(car(args)));
		    car(sc->T2_1) = c_call(car(args))(sc, sc->T1_1);
		    car(sc->T1_1) = find_symbol_or_bust(sc, fcdr(code));
		    car(sc->T2_2) = c_call(cadr(args))(sc, sc->T1_1);
		    goto UNSAFE_CLOSURE_TWO;
		  }


		case OP_CLOSURE_SSS:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_SSS;
		      goto OPT_EVAL;
		    }
		  
		case HOP_CLOSURE_SSS:
		  {
		    s7_pointer x, y, z, v1, v2, v3, args;
		    args = cdr(code);

		    v1 = find_symbol_or_bust(sc, fcdr(code));
		    v2 = find_symbol_or_bust(sc, cadr(args));
		    v3 = find_symbol_or_bust(sc, car(args));
		    
		    NEW_CELL(sc, z);
		    set_type(z, T_PAIR);
		    car(z) = v1;
		    cdr(z) = sc->NIL;
		    
		    NEW_CELL_NO_CHECK(sc, y);
		    set_type(y, T_PAIR);
		    car(y) = v2;
		    cdr(y) = z;
		    
		    NEW_CELL_NO_CHECK(sc, x);
		    set_type(x, T_PAIR);
		    car(x) = v3;
		    cdr(x) = y;
		    sc->args = x;
		    sc->code = ecdr(code);
		    /* goto UNSAFE_CLOSURE; */
		  }


		UNSAFE_CLOSURE:
		  if (sc->stack_end >= sc->stack_resize_trigger)
		    increase_stack_size(sc);
		  
		  NEW_FRAME(sc, closure_environment(sc->code), sc->envir);
		  {
		    s7_pointer x, z, e;
		    unsigned long long int id;
		    
		    e = sc->envir;
		    id = frame_id(e);
		    
		    for (x = closure_args(sc->code), z = sc->args; is_pair(x); x = cdr(x))
		      {
			s7_pointer sym, args, val;
			
			sym = car(x);
			val = car(z);
			args = cdr(z);
			
			slot_symbol(z) = sym;
			symbol_id(sym) = id;
			local_slot(sym) = z;
			if (symbol_has_accessor(sym))
			  slot_set_value(z, call_symbol_bind(sc, sym, val));
			else slot_set_value(z, val);
			set_type(z, T_SLOT | T_IMMUTABLE);
			next_slot(z) = environment_slots(e);
			environment_slots(e) = z;
			z = args;
		      }
		    if (!is_null(x)) 
		      add_slot(sc, x, z); /* the rest arg */
		  }
		  sc->code = closure_body(sc->code);
		  if (is_one_liner(sc->code))
		    {
		      sc->code = car(sc->code);
		      goto EVAL_PAIR;
		    }
		  goto BEGIN;
		  
		  
		case OP_CLOSURE_ALL_S:
		  if (!function_is_ok(code))
		    break;
		  
		case HOP_CLOSURE_ALL_S:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    sc->w = cons(sc, find_symbol_or_bust(sc, car(args)), sc->NIL);
		    for (args = cdr(args); is_pair(args); args = cdr(args))
		      sc->w = cons_unchecked(sc, find_symbol_or_bust(sc, car(args)), sc->w);
		    sc->args = safe_reverse_in_place(sc, sc->w);
		    sc->code = ecdr(code);
		    sc->w = sc->NIL;
		    goto UNSAFE_CLOSURE;
		  }
		  
		  
		case OP_CLOSURE_ALL_C:
		  if (!function_is_ok(code))
		    break;
		  
		case HOP_CLOSURE_ALL_C:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    sc->w = cons(sc, car(args), sc->NIL);
		    for (args = cdr(args); is_pair(args); args = cdr(args))
		      sc->w = cons_unchecked(sc, car(args), sc->w);
		    sc->args = safe_reverse_in_place(sc, sc->w);
		    sc->code = ecdr(code);
		    sc->w = sc->NIL;
		    goto UNSAFE_CLOSURE;
		  }
		  
		  
		case OP_CLOSURE_ALL_G:
		  if (!function_is_ok(code))
		    break;
		  
		case HOP_CLOSURE_ALL_G:
		  {
		    s7_pointer args;

		    if (sc->free_heap_top <= sc->free_heap_trigger) try_to_call_gc(sc);
		    sc->w = sc->NIL;

		    for (args = cdr(code); is_pair(args); args = cdr(args))
		      {
			s7_pointer arg, val;
			arg = car(args);
			if (is_pair(arg))
			  {
			    if (is_optimized(arg))
			      {
				car(sc->T1_1) = find_symbol_or_bust(sc, cadr(arg));
				val = c_call(arg)(sc, sc->T1_1);
			      }
			    else val = cadr(arg);
			  }
			else
			  {
			    if (is_symbol(arg))
			      val = find_symbol_or_bust(sc, arg);
			    else val = arg;
			  }
			sc->w = cons_unchecked(sc, val, sc->w);
		      }

		    sc->args = safe_reverse_in_place(sc, sc->w);
		    sc->w = sc->NIL;
		    sc->code = ecdr(code);
		    goto UNSAFE_CLOSURE;
		  }


		case OP_CLOSURE_opSq_S:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_opSq_S;
		      goto OPT_EVAL;
		    }
		  if (!indirect_c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_CLOSURE_opSq_S:
		  {
		    s7_pointer args;
		    car(sc->T2_2) = find_symbol_or_bust(sc, fcdr(code));
		    args = cadr(code);
		    car(sc->T1_1) = find_symbol_or_bust(sc, cadr(args));
		    car(sc->T2_1) = c_call(args)(sc, sc->T1_1);
		    goto UNSAFE_CLOSURE_TWO;
		  }


		case OP_CLOSURE_S_opSq:
		  if (!function_is_ok(code))
		    {
		      optimize_data(code) = OP_UNKNOWN_S_opSq;
		      goto OPT_EVAL;
		    }
		  if (!indirect_c_function_is_ok(sc, caddr(code)))
		    break;
		  
		case HOP_CLOSURE_S_opSq:
		  {
		    s7_pointer args;
		    car(sc->T2_1) = find_symbol_or_bust(sc, cadr(code));
		    args = caddr(code);
		    car(sc->T1_1) = find_symbol_or_bust(sc, cadr(args));
		    car(sc->T2_2) = c_call(args)(sc, sc->T1_1);
		    goto UNSAFE_CLOSURE_TWO;
		  }


		  /* -------------------------------------------------------------------------------- */
		case OP_TCLOSURE_S:
		  check_tfunction(code);

		case HOP_TCLOSURE_S:
		  sc->value = find_symbol_or_bust(sc, fcdr(code)); 
		  goto TCLOSURE_ONE;


		case OP_TCLOSURE_opCq:
		  check_tfunction(code);
		  if (!indirect_c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_TCLOSURE_opCq:
		  sc->value = c_call(fcdr(code))(sc, cdr(fcdr(code)));
		  goto TCLOSURE_ONE;


		case OP_TCLOSURE_opSq:
		  check_tfunction(code);
		  if (!indirect_c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_TCLOSURE_opSq:
		  car(sc->T1_1) = find_symbol_or_bust(sc, fcdr(code));
		  sc->value = c_call(cadr(code))(sc, sc->T1_1);
		  goto TCLOSURE_ONE;


		case OP_TCLOSURE_CDR_S:
		  check_tfunction(code);
		  if (!indirect_c_function_is_ok(sc, cadr(code))) 
		    break;
		  
		case HOP_TCLOSURE_CDR_S:
		  {
		    s7_pointer p;
		    p = find_symbol_or_bust(sc, fcdr(code));
		    if (!is_pair(p))
		      s7_wrong_type_arg_error(sc, "cdr", 0, p, "a pair");
		    sc->value = cdr(p);
		    /* goto TCLOSURE_ONE; */
		  }


		TCLOSURE_ONE:
		  {
		    s7_pointer args, slot, e, target_env;

		    code = ecdr(code);
		    target_env = closure_environment(code);
		    e = sc->envir;
		    while (next_environment(e) != target_env)
		      {
			typeflag(e) = 0;
			(*(sc->free_heap_top++)) = e;
			e = next_environment(e);
		      }
		    slot = environment_slots(e);
		    sc->envir = e;
		    /* the extra slots can also be released, but in my tests that rarely happens, so checking for them slows us down */

		    args = closure_args(code);
		    if (is_pair(args))
		      {
			slot_set_value(slot, sc->value);
			args = cdr(args);
			if (!is_null(args))
			  {
			    slot = next_slot(slot);
			    slot_set_value(slot, sc->NIL);
			  }
		      }
		    else slot_set_value(slot, list_1(sc, sc->value));
		    next_slot(slot) = sc->NIL;
		    
		    code = closure_body(code);
		    if (is_one_liner(code))
		      {
			sc->code = car(code);
			goto EVAL_PAIR; 
			/* all: OP_LET_READ_CHAR_P: 4460056
			 */
		      }

		    /* here we can't possibly be a non-pair, I think (how else would there be a tail-call?)
		     *   but skipping the is_pair business gains very little, and I feel uneasy omitting
		     *   the sc->begin_hook check.
		     */
		    sc->code = code;
		    goto BEGIN;
		    
		    /* here is the alternate version:
		       
		    if (!is_null(cdr(code)))
		      push_stack(sc, OP_BEGIN1, sc->NIL, cdr(code)); 
		    sc->code = car(code);
		    goto EVAL;

		    */
		  }


		case OP_TCLOSURE_SS:
		  check_tfunction(code);
		  
		case HOP_TCLOSURE_SS:
		  car(sc->T2_1) = find_symbol_or_bust(sc, cadr(code));
		  car(sc->T2_2) = find_symbol_or_bust(sc, fcdr(code));
		  goto TCLOSURE_TWO;


		  /* (define (hi a b) (eval '(+ 1 2)) (if (null? a) b (hi (cdr a) b))) 
		   * (hi '(1 2) 321)
		   *
		   * (define (n1) (let hi ((a '(1 2)) (b 321)) (eval '(+ 1 2)) (if (null? a) b (hi (cdr a) b))))
		   * (n1)
		   */
		case OP_TCLOSURE_opSq_S:
		  check_tfunction(code);
		  if (!indirect_c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_TCLOSURE_opSq_S:
		  car(sc->T2_2) = find_symbol_or_bust(sc, fcdr(code));
		  car(sc->T1_1) = find_symbol_or_bust(sc, cadr(cadr(code)));
		  car(sc->T2_1) = c_call(cadr(code))(sc, sc->T1_1);
		  goto TCLOSURE_TWO;


		case OP_TCLOSURE_S_opSq:
		  check_tfunction(code);
		  if (!indirect_c_function_is_ok(sc, caddr(code)))
		    break;
		  
		case HOP_TCLOSURE_S_opSq:
		  {
		    s7_pointer args;
		    car(sc->T2_1) = find_symbol_or_bust(sc, cadr(code));
		    args = caddr(code);
		    car(sc->T1_1) = find_symbol_or_bust(sc, cadr(args));
		    car(sc->T2_2) = c_call(args)(sc, sc->T1_1);
		    /* goto TCLOSURE_TWO; */
		  }


		TCLOSURE_TWO:
		  {
		    s7_pointer args, slot, e, target_env;
		    /* call without new frame */

		    code = ecdr(code);
		    target_env = closure_environment(code);
		    e = sc->envir;
		    while (next_environment(e) != target_env)
		      {
			typeflag(e) = 0;
			(*(sc->free_heap_top++)) = e;
			e = next_environment(e);
		      }
		    slot = environment_slots(e);
		    sc->envir = e;

		    args = closure_args(code);
		    /* if func above, code: (hi (cdr a) b)
		     *                args: (a b)
		     *                sc->envir: #<slot: b 321> #<slot: a #<pair>> ()
		     *                next up:   #<slot: __func__ hi> ()
		     *
		     * if named let, code: #<closure>
		     *               args: (a b)
		     *               sc->envir: #<slot: b 321> #<slot: a #<pair>> ()
		     *               next up: #<slot: hi #<closure>>
		     */
		    if (is_pair(args))
		      {
			bool reversed;
			reversed = (slot_symbol(slot) != car(args));
			/*
			if (!reversed)
			  fprintf(stderr, "not reversed: %s\n", DISPLAY(code));
			*/

			slot_set_value(slot, car((reversed) ? sc->T2_2 : sc->T2_1));
			slot = next_slot(slot);
			args = cdr(args);
			if (is_pair(args))
			  slot_set_value(slot, car((reversed) ? sc->T2_1 : sc->T2_2));
			else slot_set_value(slot, list_1(sc, car((reversed) ? sc->T2_1: sc->T2_2)));

			args = cdr(args);
			if (!is_null(args))
			  {
			    slot = next_slot(slot);
			    slot_set_value(slot, sc->NIL);
			  }
		      }
		    else 
		      {
			slot_set_value(slot, list_2(sc, car(sc->T2_1), car(sc->T2_2)));
		      }
		    next_slot(slot) = sc->NIL;
		    
		    code = closure_body(code);
		    if (is_one_liner(code))
		      {
			sc->code = car(code);
			goto EVAL_PAIR; 
		      }

		    sc->code = code;
		    goto BEGIN;

		    /* see note above 
		    if (!is_null(cdr(code)))
		      push_stack(sc, OP_BEGIN1, sc->NIL, cdr(code)); 
		    sc->code = car(code);
		    goto EVAL;
		    */
		  }


		  /* -------------------------------------------------------------------------------- */

		case OP_UNKNOWN_S:
		case HOP_UNKNOWN_S:
		  {
		    s7_pointer f;
		    f = find_symbol_or_bust(sc, car(code));
		    /* fprintf(stderr, "%s, f: %s (%d)\n", DISPLAY(code), DISPLAY(f), s7_in_safe_do(sc)); */
		    
		    switch (type(f))
		      {
		      case T_VECTOR:
			optimize_data(code) = OP_VECTOR_S;
			/* to back out from here we need backpointers */
			ecdr(code) = f;
			goto OPT_EVAL;
			
		      case T_C_FUNCTION:
		      case T_C_ANY_ARGS_FUNCTION:
		      case T_C_OPT_ARGS_FUNCTION:
		      case T_C_RST_ARGS_FUNCTION:
		      case T_C_LST_ARGS_FUNCTION:
			if ((is_safe_procedure(f)) &&
			    (c_function_required_args(f) <= 1) &&
			    (c_function_all_args(f) >= 1))
			  {
			    optimize_data(code) = OP_SAFE_C_S;
			    set_c_function(code, f);
			    goto OPT_EVAL;
			  }
			break;
			
		      case T_STRING:
			optimize_data(code) = OP_STRING_S;
			ecdr(code) = f;
			goto OPT_EVAL;
			
		      case T_PAIR:
			optimize_data(code) = OP_PAIR_S;
			ecdr(code) = f;
			goto OPT_EVAL;
			
		      case T_CLOSURE:
			if ((is_symbol(closure_args(f))) ||
			    (safe_list_length(sc, closure_args(f)) == 1))
			  {
			    if (is_safe_closure(closure_body(f)))
			      {
				if (is_one_liner(closure_body(f)))
				  optimize_data(code) = OP_SAFE_CLOSURE_S_one;
				else optimize_data(code) = OP_SAFE_CLOSURE_S;
			      }
			    else 
			      {
				if (is_tail_call(code))
				  optimize_data(code) = OP_TCLOSURE_S;
				else optimize_data(code) = OP_CLOSURE_S;
			      }
			    if (is_global(car(code)))
			      optimize_data(code) |= 1;
			    ecdr(code) = f;
			    fcdr(code) = cadr(code);
			    goto OPT_EVAL;
			  }
			break;
			
		      case T_CLOSURE_STAR:
			break;
			
		      case T_C_OBJECT:
			if (args_match(sc, f, 1))
			  {
			    optimize_data(code) = OP_C_OBJECT_S;
			    ecdr(code) = f;
			    goto OPT_EVAL;
			  }
			break;

		      case T_ENVIRONMENT:
			optimize_data(code) = OP_ENVIRONMENT_S;
			ecdr(code) = f;
			goto OPT_EVAL;
			
		      case T_HASH_TABLE:
			optimize_data(code) = OP_HASH_TABLE_S;
			ecdr(code) = f;
			goto OPT_EVAL;
			

			/* another possibility: goto, also continuation, macro
			 * case T_GOTO:
			 *  sc->args = list_1(sc, finder(cadr(code))); ??
			 *  call_with_exit(sc)
			 *  goto START;
			 */

		      default:
			break;
		      }
		    break;
		  }
		  
		case OP_UNKNOWN_C:
		case HOP_UNKNOWN_C:
		  {
		    s7_pointer f;
		    f = find_symbol_or_bust(sc, car(code));
		    
		    if (args_match(sc, f, 1))
		      {
			switch (type(f))
			  {
			  case T_VECTOR:
			    optimize_data(code) = OP_VECTOR_C;
			    ecdr(code) = f;
			    goto OPT_EVAL;
			    
			  case T_C_FUNCTION:
			  case T_C_ANY_ARGS_FUNCTION:
			  case T_C_OPT_ARGS_FUNCTION:
			  case T_C_RST_ARGS_FUNCTION:
			  case T_C_LST_ARGS_FUNCTION:
			    if (is_safe_procedure(f))
			      {
				optimize_data(code) = OP_SAFE_C_C;
				set_c_function(code, f);
				goto OPT_EVAL;
			      }
			    break;
			    
			    /* here cadr can only be a non-negative integer, and in the pair case, car is optimizable
			     *   (so precheck could save check in OP_STRING|PAIR|VECTOR|etc|_C)
			     */
			  case T_STRING:
			    optimize_data(code) = OP_STRING_C;
			    ecdr(code) = f;
			    goto OPT_EVAL;
			    
			  case T_PAIR:
			    optimize_data(code) = OP_PAIR_C;
			    ecdr(code) = f;
			    goto OPT_EVAL;
			    break;
			    
			  case T_CLOSURE:
			    if ((is_symbol(closure_args(f))) ||
				(safe_list_length(sc, closure_args(f)) == 1))
			      {
				if (is_safe_closure(closure_body(f)))
				  optimize_data(code) = OP_SAFE_CLOSURE_C;
				else optimize_data(code) = OP_CLOSURE_C;

				if (is_global(car(code)))
				  optimize_data(code) |= 1;
				ecdr(code) = f;
				goto OPT_EVAL;
			      }
			    break;
			    
			  case T_CLOSURE_STAR:
			    break;
			    
			  case T_C_OBJECT:
			    optimize_data(code) = OP_C_OBJECT_C;
			    ecdr(code) = f;
			    goto OPT_EVAL;

			  case T_HASH_TABLE:
			    optimize_data(code) = OP_HASH_TABLE_C;
			    ecdr(code) = f;
			    goto OPT_EVAL;
			    break;

			    /* no need for t_environment here because its argument can't be a constant */
			    
			  default:
			    break;
			  }
		      }
		    break;
		  }

		  /* split current in 2: current is default, and runs until a problem is noticed, then throws us here
		   *   unknown gets the new value and goes to the old form, doing the lookup everytime. 
		   */
		  
		case OP_UNKNOWN_SS:
		case HOP_UNKNOWN_SS:
		  {
		    s7_pointer f;
		    f = find_symbol_or_bust(sc, car(code));
		    
		    if (args_match(sc, f, 2))
		      {
			switch (type(f))
			  {
			  case T_CLOSURE:
			    if (is_safe_closure(closure_body(f)))
			      optimize_data(code) = OP_SAFE_CLOSURE_SS;
			    else 
			      {
				if (is_tail_call(code))
				  optimize_data(code) = OP_TCLOSURE_SS;
				else optimize_data(code) = OP_CLOSURE_SS;
			      }
			    if (is_global(car(code)))
			      optimize_data(code) |= 1;
			    ecdr(code) = f;
			    fcdr(code) = caddr(code);
			    goto OPT_EVAL;

			  case T_C_FUNCTION:
			  case T_C_ANY_ARGS_FUNCTION:
			  case T_C_OPT_ARGS_FUNCTION:
			  case T_C_RST_ARGS_FUNCTION:
			  case T_C_LST_ARGS_FUNCTION:
			    if (is_safe_procedure(f))
			      {
				/* fprintf(stderr, "unknown ss: %s\n", DISPLAY(code)); */
				optimize_data(code) = OP_SAFE_C_SS;
				set_c_function(code, f);
				goto OPT_EVAL;
			      }
			    break;
			    
			  default:
			    break;
			  }
		      }
		  }
		  break;


		case OP_UNKNOWN_SC:
		case HOP_UNKNOWN_SC:
		  {
		    s7_pointer f;
		    f = find_symbol_or_bust(sc, car(code));
		    if (args_match(sc, f, 2))
		      {
			switch (type(f))
			  {
			  case T_CLOSURE:
			    if (is_safe_closure(closure_body(f)))
			      optimize_data(code) = OP_SAFE_CLOSURE_SC;
			    else optimize_data(code) = OP_CLOSURE_SC;

			    if (is_global(car(code)))
			      optimize_data(code) |= 1;
			    ecdr(code) = f;
			    fcdr(code) = caddr(code);
			    goto OPT_EVAL;
			    break;

			  case T_C_FUNCTION:
			  case T_C_ANY_ARGS_FUNCTION:
			  case T_C_OPT_ARGS_FUNCTION:
			  case T_C_RST_ARGS_FUNCTION:
			  case T_C_LST_ARGS_FUNCTION:
			    if (is_safe_procedure(f))
			      {
				optimize_data(code) = OP_SAFE_C_SC;
				set_c_function(code, f);
				goto OPT_EVAL;
			      }
			    break;
			    
			  default:
			    break;
			  }
		      }
		  }
		  break;


		case OP_UNKNOWN_CS:
		case HOP_UNKNOWN_CS:
		  {
		    s7_pointer f;
		    f = find_symbol_or_bust(sc, car(code));
		    if (args_match(sc, f, 2))
		      {
			switch (type(f))
			  {
			  case T_CLOSURE:
			    if (is_safe_closure(closure_body(f)))
			      optimize_data(code) = OP_SAFE_CLOSURE_CS;
			    else optimize_data(code) = OP_CLOSURE_CS;

			    if (is_global(car(code)))
			      optimize_data(code) |= 1;
			    ecdr(code) = f;
			    fcdr(code) = caddr(code);
			    goto OPT_EVAL;
			    break;

			  case T_C_FUNCTION:
			  case T_C_ANY_ARGS_FUNCTION:
			  case T_C_OPT_ARGS_FUNCTION:
			  case T_C_RST_ARGS_FUNCTION:
			  case T_C_LST_ARGS_FUNCTION:
			    if (is_safe_procedure(f))
			      {
				optimize_data(code) = OP_SAFE_C_CS;
				set_c_function(code, f);
				goto OPT_EVAL;
			      }
			    break;
			    
			  default:
			    break;
			  }
		      }
		  }
		  break;


		case OP_UNKNOWN_CC:
		case HOP_UNKNOWN_CC:
		  {
		    s7_pointer f;
		    f = find_symbol_or_bust(sc, car(code));
		    if (args_match(sc, f, 2))
		      {
			switch (type(f))
			  {
			  case T_CLOSURE:
			    if (is_safe_closure(closure_body(f)))
			      optimize_data(code) = OP_SAFE_CLOSURE_CC;
			    else optimize_data(code) = OP_CLOSURE_CC;

			    if (is_global(car(code)))
			      optimize_data(code) |= 1;
			    ecdr(code) = f;
			    fcdr(code) = caddr(code);
			    goto OPT_EVAL;
			    break;

			  case T_C_FUNCTION:
			  case T_C_ANY_ARGS_FUNCTION:
			  case T_C_OPT_ARGS_FUNCTION:
			  case T_C_RST_ARGS_FUNCTION:
			  case T_C_LST_ARGS_FUNCTION:
			    if (is_safe_procedure(f))
			      {
				optimize_data(code) = OP_SAFE_C_C;
				set_c_function(code, f);
				goto OPT_EVAL;
			      }
			    break;
			    
			  default:
			    break;
			  }
		      }
		  }
		  break;


		case OP_UNKNOWN_SSS:
		case HOP_UNKNOWN_SSS:
		  {
		    s7_pointer f;
		    f = find_symbol_or_bust(sc, car(code));
		    if (args_match(sc, f, 3))
		      {
			switch (type(f))
			  {
			  case T_CLOSURE:
			    if (is_safe_closure(closure_body(f)))
			      optimize_data(code) = OP_SAFE_CLOSURE_SSS;
			    else optimize_data(code) = OP_CLOSURE_SSS;

			    if (is_global(car(code)))
			      optimize_data(code) |= 1;
			    ecdr(code) = f;
			    fcdr(code) = cadddr(code);
			    goto OPT_EVAL;
			    break;

			  case T_C_FUNCTION:
			  case T_C_ANY_ARGS_FUNCTION:
			  case T_C_OPT_ARGS_FUNCTION:
			  case T_C_RST_ARGS_FUNCTION:
			  case T_C_LST_ARGS_FUNCTION:
			    if (is_safe_procedure(f))
			      {
				optimize_data(code) = OP_SAFE_C_SSS;
				set_c_function(code, f);
				goto OPT_EVAL;
			      }
			    break;
			    
			  default:
			    break;
			  }
		      }
		  }
		  break;


		case OP_UNKNOWN_opCq:
		case HOP_UNKNOWN_opCq:
		  {
		    s7_pointer f;
		    /* fprintf(stderr, "unknown: %s\n", DISPLAY(code)); */
		    f = find_symbol_or_bust(sc, car(code));
		    
		    if (args_match(sc, f, 1))
		      {
			switch (type(f))
			  {
			  case T_VECTOR:
			    optimize_data(code) = OP_VECTOR_opCq;
			    ecdr(code) = f;
			    goto OPT_EVAL;
			    
			  case T_C_FUNCTION:
			  case T_C_ANY_ARGS_FUNCTION:
			  case T_C_OPT_ARGS_FUNCTION:
			  case T_C_RST_ARGS_FUNCTION:
			  case T_C_LST_ARGS_FUNCTION:
			    if (is_safe_procedure(f))
			      {
				optimize_data(code) = OP_SAFE_C_opCq;
				set_c_function(code, f);
				goto OPT_EVAL;
			      }
			    break;
			    
			  case T_STRING:
			    optimize_data(code) = OP_STRING_opCq;
			    ecdr(code) = f;
			    goto OPT_EVAL;
			    
			  case T_PAIR:
			    optimize_data(code) = OP_PAIR_opCq;
			    ecdr(code) = f;
			    goto OPT_EVAL;
			    break;
			    
			  case T_CLOSURE:
			    if ((is_symbol(closure_args(f))) ||
				(safe_list_length(sc, closure_args(f)) == 1))
			      {
				if (is_safe_closure(closure_body(f)))
				  optimize_data(code) = OP_SAFE_CLOSURE_opCq;
				else 
				  {
				    if (is_tail_call(code))
				      optimize_data(code) = OP_TCLOSURE_opCq;
				    else optimize_data(code) = OP_CLOSURE_opCq;
				  }
				if ((is_global(car(code))) &&
				    (is_hopping(cadr(code))))
				  optimize_data(code) |= 1;
				ecdr(code) = f;
				fcdr(code) = cadr(code);
				goto OPT_EVAL;
			      }
			    break;
			    
			  case T_CLOSURE_STAR:
			    break;
			    
			  case T_C_OBJECT:
			    optimize_data(code) = OP_C_OBJECT_opCq;
			    ecdr(code) = f;
			    goto OPT_EVAL;

			  case T_HASH_TABLE:
			    optimize_data(code) = OP_HASH_TABLE_opCq;
			    ecdr(code) = f;
			    goto OPT_EVAL;
			    break;
			    
			  default:
			    break;
			  }
		      }
		    break;
		  }


		case OP_UNKNOWN_opSq:
		case HOP_UNKNOWN_opSq:
		  {
		    s7_pointer f;
		    f = find_symbol_or_bust(sc, car(code));
		    if (args_match(sc, f, 1))
		      {
			switch (type(f))
			  {
			  case T_C_FUNCTION:
			  case T_C_ANY_ARGS_FUNCTION:
			  case T_C_OPT_ARGS_FUNCTION:
			  case T_C_RST_ARGS_FUNCTION:
			  case T_C_LST_ARGS_FUNCTION:
			    if (is_safe_procedure(f))
			      {
				optimize_data(code) = OP_SAFE_C_opSq;
				set_c_function(code, f);
				goto OPT_EVAL;
			      }
			    break;

			  case T_CLOSURE:			    
			    if (is_safe_closure(closure_body(f)))
			      optimize_data(code) = OP_SAFE_CLOSURE_opSq;
			    else 
			      {
				if (is_tail_call(code))
				  {
				    if (c_call(cadr(code)) == g_cdr)
				      optimize_data(code) = OP_TCLOSURE_CDR_S;
				    else optimize_data(code) = OP_TCLOSURE_opSq;
				  }
				else
				  {
				    if (c_call(cadr(code)) == g_cdr)
				      optimize_data(code) = OP_CLOSURE_CDR_S;
				    else optimize_data(code) = OP_CLOSURE_opSq;
				  }
			      }
			    if ((is_global(car(code))) &&
				(is_hopping(cadr(code))))
			      optimize_data(code) |= 1;
			    ecdr(code) = f;
			    fcdr(code) = cadr(cadr(code));
			    goto OPT_EVAL;
			    break;

			  default:
			    break;
			  }
		      }
		  }
		  break;


		case OP_UNKNOWN_opSq_S:
		case HOP_UNKNOWN_opSq_S:
		  {
		    s7_pointer f;
		    f = find_symbol_or_bust(sc, car(code));
		    if (args_match(sc, f, 2))
		      {
			switch (type(f))
			  {
			  case T_C_FUNCTION:
			  case T_C_ANY_ARGS_FUNCTION:
			  case T_C_OPT_ARGS_FUNCTION:
			  case T_C_RST_ARGS_FUNCTION:
			  case T_C_LST_ARGS_FUNCTION:
			    if (is_safe_procedure(f))
			      {
				optimize_data(code) = OP_SAFE_C_opSq_S;
				set_c_function(code, f);
				goto OPT_EVAL;
			      }
			    break;

			  case T_CLOSURE:
			    if (is_safe_closure(closure_body(f)))
			      optimize_data(code) = OP_SAFE_CLOSURE_opSq_S;
			    else 
			      {
				if (is_tail_call(code))
				  optimize_data(code) = OP_TCLOSURE_opSq_S;
				else optimize_data(code) = OP_CLOSURE_opSq_S;
			      }
			    if ((is_global(car(code))) &&
				(is_hopping(cadr(code))))
			      optimize_data(code) |= 1;
			    ecdr(code) = f;
			    fcdr(code) = caddr(code);
			    goto OPT_EVAL;
			    break;

			  default:
			    break;
			  }
		      }
		  }
		  break;


		case OP_UNKNOWN_S_opSq:
		case HOP_UNKNOWN_S_opSq:
		  {
		    s7_pointer f;
		    f = find_symbol_or_bust(sc, car(code));

		    if (args_match(sc, f, 2))
		      {
			switch (type(f))
			  {
			  case T_CLOSURE:
			    if (is_safe_closure(closure_body(f)))
			      optimize_data(code) = OP_SAFE_CLOSURE_S_opSq;
			    else 
			      {
				if (is_tail_call(code))
				  optimize_data(code) = OP_TCLOSURE_S_opSq;
				else optimize_data(code) = OP_CLOSURE_S_opSq;
			      }
			    if ((is_global(car(code))) &&
				(is_hopping(caddr(code))))
			      optimize_data(code) |= 1;
			    ecdr(code) = f;
			    goto OPT_EVAL;
			    break;

			  case T_C_FUNCTION:
			  case T_C_ANY_ARGS_FUNCTION:
			  case T_C_OPT_ARGS_FUNCTION:
			  case T_C_RST_ARGS_FUNCTION:
			  case T_C_LST_ARGS_FUNCTION:
			    if (is_safe_procedure(f))
			      {
				optimize_data(code) = OP_SAFE_C_S_opSq;
				set_c_function(code, f);
				goto OPT_EVAL;
			      }
			    break;

			  default:
			    break;
			  }
		      }
		  }
		  break;


		case OP_UNKNOWN_opSq_opSq:
		case HOP_UNKNOWN_opSq_opSq:
		  {
		    s7_pointer f;
		    f = find_symbol_or_bust(sc, car(code));
		    if (args_match(sc, f, 2))
		      {
			switch (type(f))
			  {
			  case T_CLOSURE:
			    if (is_safe_closure(closure_body(f)))
			      optimize_data(code) = OP_SAFE_CLOSURE_opSq_opSq;
			    else optimize_data(code) = OP_CLOSURE_opSq_opSq;

			    if ((is_global(car(code))) &&
				(is_hopping(cadr(code))) &&
				(is_hopping(caddr(code))))
			      optimize_data(code) |= 1;
			    ecdr(code) = f;
			    fcdr(code) = cadr(caddr(code));
			    goto OPT_EVAL;
			    break;

			  case T_C_FUNCTION:
			  case T_C_ANY_ARGS_FUNCTION:
			  case T_C_OPT_ARGS_FUNCTION:
			  case T_C_RST_ARGS_FUNCTION:
			  case T_C_LST_ARGS_FUNCTION:
			    if (is_safe_procedure(f))
			      {
				optimize_data(code) = OP_SAFE_C_opSq_opSq;
				set_c_function(code, f);
				goto OPT_EVAL;
			      }
			    break;

			  default:
			    break;
			  }
		      }
		  }
		  break;


		case OP_VECTOR_C:
		case HOP_VECTOR_C:
		  {
		    /* this is the implicit indexing case (vector-ref is a normal safe op) 
		     *    (define (hi) (let ((v (vector 1 2 3))) (v 0)))
		     *    this starts as unknown_c [26608 in optimize_function] -> vector_c
		     *    but it still reports itself as unsafe, so there are higher levels possible
		     */
		    s7_pointer v; 
		    v = find_symbol_or_bust(sc, car(code));
		    if (!s7_is_vector(v))
		      break;

		    if (!vector_is_multidimensional(v))
		      {
			s7_Int index;
			index = s7_integer(cadr(code));
			if ((index < vector_length(v)) &&
			    (index >= 0))
			  {
			    sc->value = vector_elements(v)[index];
			    goto START;
			  }
		      }
		    sc->value = vector_ref_1(sc, v, cdr(code));
		    goto START;
		  }

		  
		case OP_VECTOR_opCq:
		  if (!indirect_c_function_is_ok(sc, cadr(code)))
		    break;
		case HOP_VECTOR_opCq:
		  {
		    s7_pointer v, x; 
		    v = find_symbol_or_bust(sc, car(code));
		    if (!s7_is_vector(v))
		      break;

		    x = c_call(cadr(code))(sc, cdadr(code));
		    if (s7_is_integer(x))
		      {
			if (!vector_is_multidimensional(v))
			  {
			    s7_Int index;
			    index = s7_integer(x);
			    if ((index < vector_length(v)) &&
				(index >= 0))
			      {
				sc->value = vector_elements(v)[index];
				goto START;
			      }
			    s7_out_of_range_error(sc, "vector-ref index,", 1, cadr(code), "between 0 and vector length");
			  }
			sc->value = vector_ref_1(sc, v, cons(sc, x, sc->NIL));
			goto START;
		      }
		    return(s7_wrong_type_arg_error(sc, "vector-ref index,", 1, cadr(code), "an integer"));
		  }
		  break;


		case OP_VECTOR_S:
		case HOP_VECTOR_S:
		  {
		    s7_pointer v, ind; 
		    v = find_symbol_or_bust(sc, car(code));
		    ind = find_symbol_or_bust(sc, cadr(code));
		    if ((!s7_is_vector(v)) ||
			(!s7_is_integer(ind)))
		      break;
		  
		    if (!vector_is_multidimensional(v))
		      {
			s7_Int index;
			index = s7_integer(ind);
			if ((index < vector_length(v)) &&
			    (index >= 0))
			  {
			    sc->value = vector_elements(v)[index];
			    goto START;
			  }
			s7_out_of_range_error(sc, "vector-ref index,", 1, cadr(code), "between 0 and vector length");
		      }
		    sc->value = vector_ref_1(sc, v, cons(sc, ind, sc->NIL));
		    goto START;
		  }

		case OP_STRING_C:
		case HOP_STRING_C:
		  {
		    s7_Int index;
		    s7_pointer s;
		    s = find_symbol_or_bust(sc, car(code));
		    if (!s7_is_string(s))
		      break;

		    index = s7_integer(cadr(code));
		    if ((index < string_length(s)) &&
			(index >= 0))
		      {
			sc->value = s7_make_character(sc, ((unsigned char *)string_value(s))[index]);
			goto START;
		      }
		    sc->value = string_ref_1(sc, s, cadr(code));
		    goto START;
		  }
		  

		case OP_STRING_opCq:
		  if (!indirect_c_function_is_ok(sc, cadr(code)))
		    break;
		case HOP_STRING_opCq:
		  {
		    s7_Int index;
		    s7_pointer s, x;
		    s = find_symbol_or_bust(sc, car(code));
		    if (!s7_is_string(s))
		      break;

		    x = c_call(cadr(code))(sc, cdadr(code));
		    if (s7_is_integer(x))
		      {
			index = s7_integer(x);
			if ((index < string_length(s)) &&
			    (index >= 0))
			  {
			    sc->value = s7_make_character(sc, ((unsigned char *)string_value(s))[index]);
			    goto START;
			  }
			s7_out_of_range_error(sc, "string-ref index,", 2, cadr(code), "between 0 and string length");
		      }
		    s7_wrong_type_arg_error(sc, "string-ref", 2, cadr(code), "an integer");
		  }
		  break;
		  

		case OP_STRING_S:
		case HOP_STRING_S:
		  {
		    s7_Int index;
		    s7_pointer s, ind;
		    s = find_symbol_or_bust(sc, car(code));
		    ind = find_symbol_or_bust(sc, cadr(code));
		    if ((!s7_is_string(s)) ||
			(!s7_is_integer(ind)))
		      break;

		    index = s7_integer(ind);
		    if ((index < string_length(s)) &&
			(index >= 0))
		      {
			sc->value = s7_make_character(sc, ((unsigned char *)string_value(s))[index]);
			goto START;
		      }
		    sc->value = string_ref_1(sc, s, cons(sc, ind, sc->NIL));
		    goto START;
		  }
		  

		case OP_HASH_TABLE_C:
		case HOP_HASH_TABLE_C:
		  {
		    s7_pointer s;
		    s = find_symbol_or_bust(sc, car(code));
		    if (!s7_is_hash_table(s))
		      break;

		    sc->value = s7_hash_table_ref(sc, s, cadr(code));
		    goto START;
		  }
		  

		case OP_HASH_TABLE_opCq:
		  if (!indirect_c_function_is_ok(sc, cadr(code)))
		    break;
		case HOP_HASH_TABLE_opCq:
		  {
		    s7_pointer s, x;
		    s = find_symbol_or_bust(sc, car(code));
		    if (!s7_is_hash_table(s))
		      break;

		    x = c_call(cadr(code))(sc, cdadr(code));
		    sc->value = s7_hash_table_ref(sc, s, x);
		    goto START;
		  }
		  

		case OP_HASH_TABLE_S:
		case HOP_HASH_TABLE_S:
		  {
		    s7_pointer s;
		    s = find_symbol_or_bust(sc, car(code));
		    if (!s7_is_hash_table(s))
		      break;

		    sc->value = s7_hash_table_ref(sc, s, find_symbol_or_bust(sc, cadr(code)));
		    goto START;
		  }
		  

		case OP_ENVIRONMENT_S:
		case HOP_ENVIRONMENT_S:
		  {
		    s7_pointer s;
		    s = find_symbol_or_bust(sc, car(code));
		    if (!is_environment(s))
		      break;

		    sc->value = environment_ref(sc, s, find_symbol_or_bust(sc, cadr(code)));
		    goto START;
		  }
		  

		case OP_PAIR_C:
		case HOP_PAIR_C:
		  {
		    s7_pointer s;
		    s = find_symbol_or_bust(sc, car(code));
		    if (!is_pair(s))
		      break;

		    sc->value = list_ref_1(sc, s, cadr(code));
		    goto START;
		  }
		  

		case OP_PAIR_opCq:
		  if (!indirect_c_function_is_ok(sc, cadr(code)))
		    break;
		case HOP_PAIR_opCq:
		  {
		    s7_pointer s, x;
		    s = find_symbol_or_bust(sc, car(code));
		    if (!is_pair(s))
		      break;
		    x = c_call(cadr(code))(sc, cdadr(code));
		    sc->value = list_ref_1(sc, s, x);
		    goto START;
		  }
		  

		case OP_PAIR_S:
		case HOP_PAIR_S:
		  {
		    s7_pointer s, ind;
		    s = find_symbol_or_bust(sc, car(code));
		    if (!is_pair(s))
		      break;
		    ind = find_symbol_or_bust(sc, cadr(code));
		    sc->value = list_ref_1(sc, s, ind);
		    goto START;
		  }
		  

		  /* are these actually safe? we can call either pws_apply or call_s_object_getter here,
		   *   which will copy args (via OP_APPLY I think) in the pws_apply case.
		   */
		case OP_C_OBJECT_C:
		case HOP_C_OBJECT_C:
		  {
		    s7_pointer c;
		    c = find_symbol_or_bust(sc, car(code));
		    if (!is_c_object(c))
		      break;
		    sc->value = (*(object_ref(c)))(sc, c, cdr(code));
		    goto START;
		  }

		case OP_C_OBJECT_opCq:
		  if (!indirect_c_function_is_ok(sc, cadr(code)))
		    break;
		case HOP_C_OBJECT_opCq:
		  {
		    s7_pointer c;
		    c = find_symbol_or_bust(sc, car(code));
		    if (!is_c_object(c))
		      break;
		    car(sc->T1_1) = c_call(cadr(code))(sc, cdadr(code));
		    sc->value = (*(object_ref(c)))(sc, c, sc->T1_1);
		    goto START;
		  }

		case OP_C_OBJECT_S:
		case HOP_C_OBJECT_S:
		  {
		    s7_pointer c;
		    c = find_symbol_or_bust(sc, car(code));
		    if (!is_c_object(c))
		      break;
		    car(sc->T1_1) = find_symbol_or_bust(sc, cadr(code));
		    sc->value = (*(object_ref(c)))(sc, c, sc->T1_1);
		    goto START;
		  }


		case OP_SAFE_C_C:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_SAFE_C_C:
		  sc->value = c_call(code)(sc, cdr(code)); /* this includes all safe calls where all args are constants */
		  goto START;
		  /* lg OP_AND1: 59608
		   * in OP_EVAL_ARGS3: 77304
		   * al OP_DO_STEP2: 624740
		   */

		case OP_SAFE_C_Q:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_SAFE_C_Q:
		  car(sc->T1_1) = cadar(cdr(code));
		  sc->value = c_call(code)(sc, sc->T1_1);
		  goto START;
		  

		case OP_SAFE_C_S:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_SAFE_C_S:
		  car(sc->T1_1) = finder(sc, cadr(code));
		  sc->value = c_call(code)(sc, sc->T1_1);
		  goto START;
		  /* al OP_COND1_SIMPLE: 378506
		   */

		  
		case OP_SAFE_IS_PAIR_S:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_SAFE_IS_PAIR_S:
		  sc->value = make_boolean(sc, is_pair(finder(sc, cadr(code))));
		  goto START;

		  
		case OP_SAFE_IS_SYMBOL_S:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_SAFE_IS_SYMBOL_S:
		  sc->value = make_boolean(sc, is_symbol(finder(sc, cadr(code))));
		  goto START;

		  
		case OP_SAFE_NOT_S:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_SAFE_NOT_S:
		  sc->value = make_boolean(sc, is_false(sc, finder(sc, cadr(code))));
		  goto START;

		  
		case OP_SAFE_C_SS:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_SAFE_C_SS:
		  {
		    s7_pointer val, args;
		    args = cdr(code);
		    val = finder(sc, car(args));
		    car(sc->T2_2) = finder(sc, cadr(args));
		    car(sc->T2_1) = val;
		    sc->value = c_call(code)(sc, sc->T2_1);
		    /* counts[stack_op(sc->stack, s7_stack_top(sc) - 1)]++;  */
		    goto START; 
		    /* al 6012910 OP_BEGIN1: 4460056
		     *     actually write-char
		     */

		  }
		  
		case OP_SAFE_C_SS_2:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_SAFE_C_SS_2:
		  {
		    s7_pointer val, args;
		    args = cdr(code);
		    val = finder(sc, car(args));
		    car(sc->T2_2) = finder(sc, cadr(args));
		    car(sc->T2_1) = val;
		    c_call(code)(sc, sc->T2_1);

		    sc->envir = sc->stack_end[-3];
		    code = sc->stack_end[-4];
		    sc->code =  car(code);

		    if (is_null(cdr(code)))
		      sc->stack_end -= 4;
		    else sc->stack_end[-4] = cdr(code);
		    
		    goto EVAL;
		  }
		  
		case OP_SAFE_C_ALL_S:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_SAFE_C_ALL_S:
		  {
		    s7_pointer args;
		    int i;

		    for (i = 0, args = cdr(code); is_pair(args); i++, args = cdr(args))
		      {
			car(sc->all_g_args[i]) = finder(sc, car(args));
			cdr(sc->all_g_args[i]) = sc->all_g_args[i + 1];
		      }
		    cdr(sc->all_g_args[i - 1]) = sc->NIL;
		    sc->value = c_call(code)(sc, sc->all_g_args[0]);
		    goto START;
		  }
		  
		case OP_SAFE_C_SC:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_SAFE_C_SC:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    car(sc->T2_1) = finder(sc, car(args));
		    car(sc->T2_2) = cadr(args);
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_CS:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_SAFE_C_CS:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    car(sc->T2_2) = finder(sc, cadr(args));
		    car(sc->T2_1) = car(args);
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }
		  

		case OP_SAFE_C_SQ:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_SAFE_C_SQ:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    car(sc->T2_1) = finder(sc, car(args));
		    car(sc->T2_2) = cadr(cadr(args));
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }
		  

		case OP_SAFE_C_QS:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_SAFE_C_QS:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    car(sc->T2_2) = finder(sc, cadr(args));
		    car(sc->T2_1) = cadr(car(args));
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }
		  

		case OP_SAFE_C_QQ:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_SAFE_C_QQ:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    car(sc->T2_1) = cadr(car(args));
		    car(sc->T2_2) = cadr(cadr(args));
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_XXX:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_SAFE_C_XXX:
		  {
		    s7_pointer arg, val1, val2, val3, args;
		    args = cdr(code);
		    
		    arg = car(args);
		    if (is_symbol(arg))
		      val1 = finder(sc, arg);
		    else val1 = arg;
		    
		    arg = cadr(args);
		    if (is_symbol(arg))
		      val2 = finder(sc, arg);
		    else val2 = arg;
		    
		    arg = caddr(args);
		    if (is_symbol(arg))
		      val3 = finder(sc, arg);
		    else val3 = arg;
		    
		    car(sc->T3_1) = val1;
		    car(sc->T3_2) = val2;
		    car(sc->T3_3) = val3;
		    sc->value = c_call(code)(sc, sc->T3_1);
		    goto START;
		  }
		  

		case OP_SAFE_C_opXXXq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_SAFE_C_opXXXq:
		  {
		    s7_pointer arg, val1, val2, val3, args;
		    args = cdr(cadr(code));
		    
		    arg = car(args);
		    if (is_symbol(arg))
		      val1 = finder(sc, arg);
		    else val1 = arg;
		    
		    arg = cadr(args);
		    if (is_symbol(arg))
		      val2 = finder(sc, arg);
		    else val2 = arg;
		    
		    arg = caddr(args);
		    if (is_symbol(arg))
		      val3 = finder(sc, arg);
		    else val3 = arg;
		    
		    car(sc->T3_1) = val1;
		    car(sc->T3_2) = val2;
		    car(sc->T3_3) = val3;
		    car(sc->T1_1) = c_call(cadr(code))(sc, sc->T3_1);

		    sc->value = c_call(code)(sc, sc->T1_1);
		    goto START;
		  }
		  
		case OP_SAFE_C_Z:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;

		case HOP_SAFE_C_Z:
		  push_stack(sc, OP_SAFE_C_P_1, sc->NIL, code);
		  sc->code = cadr(code);
		  goto OPT_EVAL;
		  

		case OP_SAFE_C_CZ:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;

		case HOP_SAFE_C_CZ:
		  push_stack(sc, OP_SAFE_C_SZ_1, cadr(code), code);
		  sc->code = caddr(code);
		  goto OPT_EVAL;
		  

		case OP_SAFE_C_ZC:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;

		case HOP_SAFE_C_ZC:
		  push_stack(sc, OP_SAFE_C_ZS_1, caddr(code), code);
		  sc->code = cadr(code);
		  goto OPT_EVAL;
		  

		case OP_SAFE_C_QZ:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;

		case HOP_SAFE_C_QZ:
		  push_stack(sc, OP_SAFE_C_SZ_1, cadr(cadr(code)), code);
		  sc->code = caddr(code);
		  goto OPT_EVAL;
		  

		case OP_SAFE_C_ZQ:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;

		case HOP_SAFE_C_ZQ:
		  push_stack(sc, OP_SAFE_C_ZS_1, cadr(caddr(code)), code);
		  sc->code = cadr(code);
		  goto OPT_EVAL;
		  

		case OP_SAFE_C_SZ:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;

		case HOP_SAFE_C_SZ:
		  push_stack(sc, OP_SAFE_C_SZ_1, finder(sc, cadr(code)), code);
		  sc->code = caddr(code);
		  goto OPT_EVAL;
		  

		case OP_SAFE_C_ZS:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;

		case HOP_SAFE_C_ZS:
		  push_stack(sc, OP_SAFE_C_ZS_1, finder(sc, caddr(code)), code);
		  sc->code = cadr(code);
		  goto OPT_EVAL;
		  

		case OP_SAFE_C_ZZ:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;

		case HOP_SAFE_C_ZZ:
		  push_stack(sc, OP_SAFE_C_ZZ_1, sc->NIL, code);
		  sc->code = cadr(code);
		  goto OPT_EVAL;
		  


		case OP_SAFE_C_ZXX:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;

		case HOP_SAFE_C_ZXX:
		  push_stack(sc, OP_SAFE_C_ZXX_1, (is_symbol(caddr(code))) ? finder(sc, caddr(code)) : caddr(code), code);
		  sc->code = cadr(code);
		  goto OPT_EVAL;
		  

		case OP_SAFE_C_XZX:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;

		case HOP_SAFE_C_XZX:
		  push_stack(sc, OP_SAFE_C_XZX_1, (is_symbol(cadr(code))) ? finder(sc, cadr(code)) : cadr(code), code);
		  sc->code = caddr(code);
		  goto OPT_EVAL;
		  

		case OP_SAFE_C_XXZ:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadddr(code)))
		    break;

		case HOP_SAFE_C_XXZ:
		  push_stack(sc, OP_SAFE_C_XXZ_1, (is_symbol(cadr(code))) ? finder(sc, cadr(code)) : cadr(code), code);
		  sc->code = cadddr(code);
		  goto OPT_EVAL;
		  

		case OP_SAFE_C_ZZX:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;

		case HOP_SAFE_C_ZZX:
		  push_stack(sc, OP_SAFE_C_ZZX_1, (is_symbol(cadddr(code))) ? finder(sc, cadddr(code)) : cadddr(code), code);
		  sc->code = cadr(code);
		  goto OPT_EVAL;
		  

		case OP_SAFE_C_ZXZ:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  if (!c_function_is_ok(sc, cadddr(code)))
		    break;

		case HOP_SAFE_C_ZXZ:
		  push_stack(sc, OP_SAFE_C_ZXZ_1, (is_symbol(caddr(code))) ? finder(sc, caddr(code)) : caddr(code), code);
		  sc->code = cadr(code);
		  goto OPT_EVAL;
		  

		case OP_SAFE_C_XZZ:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;
		  if (!c_function_is_ok(sc, cadddr(code)))
		    break;

		case HOP_SAFE_C_XZZ:
		  push_stack(sc, OP_SAFE_C_XZZ_1, (is_symbol(cadr(code))) ? finder(sc, cadr(code)) : cadr(code), code);
		  sc->code = caddr(code);
		  goto OPT_EVAL;
		  

		case OP_SAFE_C_ZZZ:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;
		  if (!c_function_is_ok(sc, cadddr(code)))
		    break;

		case HOP_SAFE_C_ZZZ:
		  push_stack(sc, OP_SAFE_C_ZZZ_1, sc->NIL, code);
		  sc->code = cadr(code);
		  goto OPT_EVAL;
		  


		  /* same as OP_SAFE_C_opCq_opCq */
		case OP_SAFE_C_DD:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;

		case HOP_SAFE_C_DD:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    car(sc->T2_1) = c_call(car(args))(sc, cdr(car(args)));
		    car(sc->T2_2) = c_call(cadr(args))(sc, cdr(cadr(args)));
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }


		case OP_SAFE_C_DXX:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;

		case HOP_SAFE_C_DXX:
		  {
		    s7_pointer args, val1, val2;
		    args = cdr(code);

		    if (is_symbol(cadr(args)))
		      val1 = finder(sc, cadr(args));
		    else val1 = cadr(args);
		    if (is_symbol(caddr(args)))
		      val2 = finder(sc, caddr(args));
		    else val2 = caddr(args);

		    car(sc->T3_1) = c_call(car(args))(sc, cdr(car(args)));
		    car(sc->T3_2) = val1;
		    car(sc->T3_3) = val2;
		    sc->value = c_call(code)(sc, sc->T3_1);
		    goto START;
		  }


		case OP_SAFE_C_XDX:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;

		case HOP_SAFE_C_XDX:
		  {
		    s7_pointer args, val1, val2;
		    args = cdr(code);

		    if (is_symbol(car(args)))
		      val1 = finder(sc, car(args));
		    else val1 = car(args);
		    if (is_symbol(caddr(args)))
		      val2 = finder(sc, caddr(args));
		    else val2 = caddr(args);

		    car(sc->T3_2) = c_call(cadr(args))(sc, cdr(cadr(args)));
		    car(sc->T3_1) = val1;
		    car(sc->T3_3) = val2;
		    sc->value = c_call(code)(sc, sc->T3_1);
		    goto START;
		  }
		  

		case OP_SAFE_C_XXD:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadddr(code)))
		    break;

		case HOP_SAFE_C_XXD:
		  {
		    s7_pointer args, val1, val2;
		    args = cdr(code);

		    if (is_symbol(car(args)))
		      val1 = finder(sc, car(args));
		    else val1 = car(args);
		    if (is_symbol(cadr(args)))
		      val2 = finder(sc, cadr(args));
		    else val2 = cadr(args);

		    car(sc->T3_3) = c_call(caddr(args))(sc, cdr(caddr(args)));
		    car(sc->T3_1) = val1;
		    car(sc->T3_2) = val2;
		    sc->value = c_call(code)(sc, sc->T3_1);
		    goto START;
		  }
		  

		case OP_SAFE_C_DDX:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;

		case HOP_SAFE_C_DDX:
		  {
		    s7_pointer args;
		    args = cdr(code);

		    sc->temp4 = c_call(car(args))(sc, cdr(car(args)));
		    if (is_symbol(caddr(args)))
		      car(sc->T3_3) = finder(sc, caddr(args));
		    else car(sc->T3_3) = caddr(args);

		    car(sc->T3_2) = c_call(cadr(args))(sc, cdr(cadr(args)));
		    car(sc->T3_1) = sc->temp4;
		    sc->value = c_call(code)(sc, sc->T3_1);
		    goto START;
		  }
		  

		case OP_SAFE_C_DXD:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  if (!c_function_is_ok(sc, cadddr(code)))
		    break;

		case HOP_SAFE_C_DXD:
		  {
		    s7_pointer args;
		    args = cdr(code);

		    sc->temp4 = c_call(car(args))(sc, cdr(car(args)));
		    if (is_symbol(cadr(args)))
		      car(sc->T3_2) = finder(sc, cadr(args));
		    else car(sc->T3_2) = cadr(args);

		    car(sc->T3_3) = c_call(caddr(args))(sc, cdr(caddr(args)));
		    car(sc->T3_1) = sc->temp4;
		    sc->value = c_call(code)(sc, sc->T3_1);
		    goto START;
		  }
		  

		case OP_SAFE_C_XDD:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;
		  if (!c_function_is_ok(sc, cadddr(code)))
		    break;

		case HOP_SAFE_C_XDD:
		  {
		    s7_pointer args;
		    args = cdr(code);

		    sc->temp4 = c_call(cadr(args))(sc, cdr(cadr(args)));
		    if (is_symbol(car(args)))
		      car(sc->T3_1) = finder(sc, car(args));
		    else car(sc->T3_1) = car(args);

		    car(sc->T3_3) = c_call(caddr(args))(sc, cdr(caddr(args)));
		    car(sc->T3_2) = sc->temp4;
		    sc->value = c_call(code)(sc, sc->T3_1);
		    goto START;
		  }
		  

		case OP_SAFE_C_DDD:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;
		  if (!c_function_is_ok(sc, cadddr(code)))
		    break;

		case HOP_SAFE_C_DDD:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    sc->temp4 = c_call(car(args))(sc, cdr(car(args)));
		    car(sc->T1_1) = c_call(cadr(args))(sc, cdr(cadr(args)));
		    car(sc->T3_3) = c_call(caddr(args))(sc, cdr(caddr(args)));
		    car(sc->T3_1) = sc->temp4;
		    car(sc->T3_2) = car(sc->T1_1);
		    sc->value = c_call(code)(sc, sc->T3_1);
		    goto START;
		  }


		case OP_SAFE_C_ALL_G:
		  if (!c_function_is_ok(sc, code))
		    break;

		case HOP_SAFE_C_ALL_G:
		  {
		    s7_pointer args;
		    int i;

		    for (i = 0, args = cdr(code); is_pair(args); i++, args = cdr(args))
		      {
			s7_pointer arg, val;
			arg = car(args);
			if (is_pair(arg))
			  {
			    if (is_optimized(arg))
			      {
				car(sc->T1_1) = finder(sc, cadr(arg));
				val = c_call(arg)(sc, sc->T1_1);
			      }
			    else val = cadr(arg);
			  }
			else
			  {
			    if (is_symbol(arg))
			      val = finder(sc, arg);
			    else val = arg;
			  }
			car(sc->all_g_args[i]) = val;
			cdr(sc->all_g_args[i]) = sc->all_g_args[i + 1];
		      }

		    cdr(sc->all_g_args[i - 1]) = sc->NIL;
		    sc->value = c_call(code)(sc, sc->all_g_args[0]);
		    goto START;
		  }


		case OP_SAFE_C_ALL_X:
		  if (!c_function_is_ok(sc, code))
		    break;

		case HOP_SAFE_C_ALL_X:
		  {
		    s7_pointer args, val;
		    int i;
		    val = sc->NIL; /* make gcc happy */

		    for (i = 0, args = cdr(code); is_pair(args); i++, args = cdr(args))
		      {
			s7_pointer arg;
			arg = car(args);
			if (is_pair(arg))
			  {
			    if (is_optimized(arg))
			      {
				s7_pointer largs;
				switch (optimize_data(arg))
				  {
				  case HOP_SAFE_C_C:
				    val = c_call(arg)(sc, cdr(arg));
				    break;
				    
				  case HOP_SAFE_C_S:
				    car(sc->T1_1) = finder(sc, cadr(arg));
				    val = c_call(arg)(sc, sc->T1_1);
				    break;
				    
				  case HOP_SAFE_IS_PAIR_S:
				    val = make_boolean(sc, is_pair(finder(sc, cadr(arg))));
				    break;
				    
				  case HOP_SAFE_IS_SYMBOL_S:
				    val = make_boolean(sc, is_symbol(finder(sc, cadr(arg))));
				    break;
				    
				  case HOP_SAFE_NOT_S:
				    val = make_boolean(sc, is_false(sc, finder(sc, cadr(arg))));
				    break;
				    
				  case HOP_SAFE_C_SC:
				    largs = cdr(arg);
				    car(sc->T2_1) = finder(sc, car(largs));
				    car(sc->T2_2) = cadr(largs);
				    val = c_call(arg)(sc, sc->T2_1);
				    break;
				    
				  case HOP_SAFE_C_CS:
				    largs = cdr(arg);
				    car(sc->T2_2) = finder(sc, cadr(largs));
				    car(sc->T2_1) = car(largs);
				    val = c_call(arg)(sc, sc->T2_1);
				    break;
				    
				  case HOP_SAFE_C_SS:
				    largs = cdr(arg);
				    val = finder(sc, car(largs));
				    car(sc->T2_2) = finder(sc, cadr(largs));
				    car(sc->T2_1) = val;
				    val = c_call(arg)(sc, sc->T2_1);
				    break;
				    
				  case HOP_SAFE_C_SQ:
				    largs = cdr(arg);
				    car(sc->T2_1) = finder(sc, car(largs));
				    car(sc->T2_2) = cadr(cadr(largs));
				    val = c_call(arg)(sc, sc->T2_1);
				    break;
				    
				  case HOP_SAFE_C_opSq:
				    largs = cadr(arg);
				    car(sc->T1_1) = finder(sc, cadr(largs));
				    car(sc->T1_1) = c_call(largs)(sc, sc->T1_1);
				    val = c_call(arg)(sc, sc->T1_1);
				    break;
				    
				  case HOP_SAFE_C_opSq_opSq:
				    largs = cdr(arg);
				    car(sc->T1_1) = finder(sc, cadr(car(largs)));
				    sc->temp4 = c_call(car(largs))(sc, sc->T1_1);
				    largs = cadr(largs);
				    car(sc->T1_1) = finder(sc, cadr(largs));
				    car(sc->T2_2) = c_call(largs)(sc, sc->T1_1);
				    car(sc->T2_1) = sc->temp4;
				    val = c_call(arg)(sc, sc->T2_1);
				    break;
				  }
			      }
			    else val = cadr(arg);
			  }
			else
			  {
			    if (is_symbol(arg))
			      val = finder(sc, arg);
			    else val = arg;
			  }
			car(sc->all_g_args[i]) = val;
			cdr(sc->all_g_args[i]) = sc->all_g_args[i + 1];
		      }

		    cdr(sc->all_g_args[i - 1]) = sc->NIL;
		    sc->value = c_call(code)(sc, sc->all_g_args[0]);
		    goto START;
		  }


		case OP_C_S_opSq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!indirect_c_function_is_ok(sc, caddr(code)))
		    break;

		case HOP_C_S_opSq:
		  {
		    s7_pointer args, val;
		    args = cdr(code);
		    val = find_symbol_or_bust(sc, car(args));
		    car(sc->T1_1) = find_symbol_or_bust(sc, cadr(cadr(args)));
		    sc->args = list_2(sc, val, c_call(cadr(args))(sc, sc->T1_1));
		    sc->value = c_call(code)(sc, sc->args);
		    goto START;
		  }


		case OP_C_opSq_CC:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!indirect_c_function_is_ok(sc, cadr(code)))
		    break;

		case HOP_C_opSq_CC:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    car(sc->T1_1) = find_symbol_or_bust(sc, cadr(car(args)));
		    sc->args = list_3(sc, c_call(car(args))(sc, sc->T1_1), cadr(args), caddr(args));
		    sc->value = c_call(code)(sc, sc->args);
		    goto START;
		  }


		case OP_C_S:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_C_S:
		  sc->args = list_1(sc, find_symbol_or_bust(sc, cadr(code)));
		  sc->value = c_call(code)(sc, sc->args);
		  goto START;


		case OP_READ_S:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_READ_S:
		  {
		    s7_pointer port;
		    port = find_symbol_or_bust(sc, cadr(code));
		    
		    if (!is_input_port(port)) /* was also not stdin */
		      s7_wrong_type_arg_error(sc, "read", 0, port, "an input port");
		    if (port_is_closed(port))
		      s7_wrong_type_arg_error(sc, "read", 0, port, "an open input port");
		    
		    if (is_function_port(port))
		      sc->value = (*(port_input_function(port)))(sc, S7_READ, port);
		    else
		      {
			if ((is_string_port(port)) &&
			    (port_string_length(port) <= port_string_point(port)))
			  sc->value = sc->EOF_OBJECT;
			else
			  {
			    push_input_port(sc, port);
			    push_stack(sc, OP_READ_DONE, sc->NIL, sc->NIL); /* this stops the internal read process so we only get one form */
			    sc->tok = token(sc);
			    switch (sc->tok)
			      {
			      case TOKEN_EOF:
				goto START;
				
			      case TOKEN_RIGHT_PAREN:
				read_error(sc, "unexpected close paren");
				
			      case TOKEN_COMMA:
				read_error(sc, "unexpected comma");
				
			      default:
				sc->value = read_expression(sc);
				sc->current_line = port_line_number(sc->input_port);  /* this info is used to track down missing close parens */
				sc->current_file = port_filename(sc->input_port);
			      }
			  }
		      }
		    goto START;
		  }


		case OP_C_P:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_C_P:
		  push_stack(sc, OP_C_P_1, sc->NIL, code);
		  sc->code = cadr(code);
		  if (is_optimized(sc->code))
		    goto OPT_EVAL;
		  goto EVAL_PAIR;


		case OP_C_ALL_G:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_C_ALL_G:
		  {
		    s7_pointer args;

		    if (sc->free_heap_top <= sc->free_heap_trigger) try_to_call_gc(sc);
		    sc->w = sc->NIL;

		    for (args = cdr(code); is_pair(args); args = cdr(args))
		      {
			s7_pointer arg, val;
			arg = car(args);
			if (is_pair(arg))
			  {
			    if (is_optimized(arg))
			      {
				car(sc->T1_1) = find_symbol_or_bust(sc, cadr(arg));
				val = c_call(arg)(sc, sc->T1_1);
			      }
			    else val = cadr(arg);
			  }
			else
			  {
			    if (is_symbol(arg))
			      val = find_symbol_or_bust(sc, arg);
			    else val = arg;
			  }
			sc->w = cons_unchecked(sc, val, sc->w);
		      }

		    sc->value = c_call(code)(sc, safe_reverse_in_place(sc, sc->w));
		    sc->w = sc->NIL;
		    goto START;
		  }


		case OP_SAFE_C_SSS:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_SAFE_C_SSS:
		  {
		    s7_pointer val1, val2, args;
		    args = cdr(code);

		    val1 = find_symbol_or_bust(sc, car(args));
		    val2 = find_symbol_or_bust(sc, cadr(args));
		    car(sc->T3_3) = find_symbol_or_bust(sc, caddr(args));
		    car(sc->T3_1) = val1;
		    car(sc->T3_2) = val2;
		    sc->value = c_call(code)(sc, sc->T3_1);
		    goto START;
		  }
		  
		  /* now ops that need extra op name checks */
		  
		case OP_SAFE_C_opCq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_SAFE_C_opCq:
		  car(sc->T1_1) = c_call(car(cdr(code)))(sc, cdar(cdr(code))); /* OP_SAFE_C_C can involve any number of ops */
		  sc->value = c_call(code)(sc, sc->T1_1);
		  goto START;

		  
		case OP_SAFE_C_opQq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_SAFE_C_opQq:
		  car(sc->T1_1) = cadr(cadr(car(cdr(code))));
		  car(sc->T1_1) = c_call(car(cdr(code)))(sc, sc->T1_1);
		  sc->value = c_call(code)(sc, sc->T1_1);
		  goto START;

		  
		case OP_SAFE_C_opSq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_SAFE_C_opSq:
		  {
		    s7_pointer args;
		    args = cadr(code);
		    car(sc->T1_1) = finder(sc, cadr(args));
		    car(sc->T1_1) = c_call(args)(sc, sc->T1_1);
		    sc->value = c_call(code)(sc, sc->T1_1);
		    goto START;
		  }
		  /* lg: (length (cdr)) here
		   */
		  
		  
		case OP_SAFE_C_P:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_SAFE_C_P:
		  push_stack(sc, OP_EVAL_ARGS_P_1, sc->NIL, code); /* catch values etc */
		  sc->code = cadr(code);
		  goto EVAL_PAIR; /* in: syn */
		  

		case OP_SAFE_C_PS:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_SAFE_C_PS:
		  push_stack(sc, OP_EVAL_ARGS_P_3, sc->NIL, code);
		  sc->code = cadr(code);
		  goto EVAL_PAIR; /* in: opt */
		  

		case OP_SAFE_C_PC:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_SAFE_C_PC:
		  push_stack(sc, OP_EVAL_ARGS_P_4, sc->NIL, code);
		  sc->code = cadr(code);
		  goto EVAL_PAIR; /* in: opt */
		  

		case OP_SAFE_C_SP:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_SAFE_C_SP:
		  push_stack(sc, OP_EVAL_ARGS_P_2, list_1(sc, finder(sc, cadr(code))), code);
		  sc->code = caddr(code);
		  goto EVAL_PAIR; /* lg/in: opt */
		  

		case OP_SAFE_C_CP:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_SAFE_C_CP:
		  push_stack(sc, OP_EVAL_ARGS_P_2, list_1(sc, cadr(code)), code);
		  sc->code = caddr(code);
		  goto EVAL_PAIR;


		case OP_SAFE_C_PP:
		  if (!c_function_is_ok(sc, code))
		    break;
		  
		case HOP_SAFE_C_PP:
		  push_stack(sc, OP_SAFE_C_PP_1, sc->NIL, code);
		  sc->code = cadr(code);
		  if (is_optimized(sc->code))
		    goto OPT_EVAL;
		  goto EVAL_PAIR;

		  
		case OP_SAFE_C_opSSq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_SAFE_C_opSSq:
		  {
		    s7_pointer args, val1;
		    args = cadr(code);
		    val1 = finder(sc, cadr(args));
		    car(sc->T2_2) = finder(sc, caddr(args));
		    car(sc->T2_1) = val1;
		    car(sc->T1_1) = c_call(args)(sc, sc->T2_1);
		    sc->value = c_call(code)(sc, sc->T1_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_opSCq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_SAFE_C_opSCq:
		  {
		    s7_pointer args;
		    args = cadr(code);
		    car(sc->T2_1) = finder(sc, cadr(args));
		    car(sc->T2_2) = caddr(args);
		    car(sc->T1_1) = c_call(args)(sc, sc->T2_1);
		    sc->value = c_call(code)(sc, sc->T1_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_opCSq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_SAFE_C_opCSq:
		  {
		    s7_pointer args;
		    args = cadr(code);
		    car(sc->T2_2) = finder(sc, caddr(args));
		    car(sc->T2_1) = cadr(args);
		    car(sc->T1_1) = c_call(args)(sc, sc->T2_1);
		    sc->value = c_call(code)(sc, sc->T1_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_opSQq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_SAFE_C_opSQq:
		  {
		    s7_pointer args;
		    args = cadr(code);
		    car(sc->T2_1) = finder(sc, cadr(args));
		    car(sc->T2_2) = cadr(caddr(args)); 
		    car(sc->T1_1) = c_call(args)(sc, sc->T2_1);
		    sc->value = c_call(code)(sc, sc->T1_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_opQSq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_SAFE_C_opQSq:
		  {
		    s7_pointer args;
		    args = cadr(code);
		    car(sc->T2_2) = finder(sc, caddr(args));
		    car(sc->T2_1) = cadr(cadr(args)); 
		    car(sc->T1_1) = c_call(args)(sc, sc->T2_1);
		    sc->value = c_call(code)(sc, sc->T1_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_S_opSq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;
		  
		case HOP_SAFE_C_S_opSq:
		  {
		    s7_pointer args;
		    args = caddr(code);
		    car(sc->T1_1) = finder(sc, cadr(args));
		    sc->temp4 = c_call(args)(sc, sc->T1_1);
		    car(sc->T2_1) = finder(sc, cadr(code));
		    car(sc->T2_2) = sc->temp4;
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }

		case OP_SAFE_C_S_opCq: 
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;
		  
		case HOP_SAFE_C_S_opCq: 
		  {
		    s7_pointer val1, args;
		    args = cdr(code);
		    val1 = finder(sc, car(args));
		    car(sc->T2_2) = c_call(cadr(args))(sc, cdr(cadr(args))); /* any number of constants here */
		    car(sc->T2_1) = val1;
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_C_opSq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;
		  
		case HOP_SAFE_C_C_opSq:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    car(sc->T1_1) = finder(sc, cadr(cadr(args)));
		    car(sc->T2_2) = c_call(cadr(args))(sc, sc->T1_1);
		    car(sc->T2_1) = car(args);
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_C_opCq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;
		  
		case HOP_SAFE_C_C_opCq:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    car(sc->T2_2) = c_call(cadr(args))(sc, cdr(cadr(args))); /* any # of args */
		    car(sc->T2_1) = car(args);
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_C_opCSq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;
		  
		case HOP_SAFE_C_C_opCSq:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    car(sc->T2_2) = finder(sc, caddr(cadr(args)));
		    car(sc->T2_1) = cadr(cadr(args));
		    car(sc->T2_2) = c_call(cadr(args))(sc, sc->T2_1);
		    car(sc->T2_1) = car(args);
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_opCSq_C:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_SAFE_C_opCSq_C:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    car(sc->T2_2) = finder(sc, caddr(car(args)));
		    car(sc->T2_1) = cadr(car(args));
		    car(sc->T2_1) = c_call(car(args))(sc, sc->T2_1);
		    car(sc->T2_2) = cadr(args);
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_C_opSSq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;
		  
		case HOP_SAFE_C_C_opSSq:
		  {
		    s7_pointer args, val;
		    args = cdr(code);
		    val = finder(sc, cadr(cadr(args)));
		    car(sc->T2_2) = finder(sc, caddr(cadr(args)));
		    car(sc->T2_1) = val;
		    car(sc->T2_2) = c_call(cadr(args))(sc, sc->T2_1);
		    car(sc->T2_1) = car(args);
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_opSSq_C:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_SAFE_C_opSSq_C:
		  {
		    s7_pointer args, val;
		    args = cdr(code);
		    val = finder(sc, cadr(car(args)));
		    car(sc->T2_2) = finder(sc, caddr(car(args)));
		    car(sc->T2_1) = val;
		    car(sc->T2_1) = c_call(car(args))(sc, sc->T2_1);
		    car(sc->T2_2) = cadr(args);
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_opSSq_S:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_SAFE_C_opSSq_S:
		  {
		    s7_pointer args, val, val1;
		    args = cdr(code);
		    val = finder(sc, cadr(car(args)));
		    val1 = finder(sc, cadr(args));
		    car(sc->T2_2) = finder(sc, caddr(car(args)));
		    car(sc->T2_1) = val;
		    car(sc->T2_1) = c_call(car(args))(sc, sc->T2_1);
		    car(sc->T2_2) = val1;
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_opSCq_S:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_SAFE_C_opSCq_S:
		  {
		    s7_pointer args, val1;
		    args = cdr(code);
		    val1 = finder(sc, cadr(args));
		    car(sc->T2_1) = finder(sc, cadr(car(args)));
		    car(sc->T2_2) = caddr(car(args));
		    car(sc->T2_1) = c_call(car(args))(sc, sc->T2_1);
		    car(sc->T2_2) = val1;
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_opSCq_C:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_SAFE_C_opSCq_C:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    car(sc->T2_1) = finder(sc, cadr(car(args)));
		    car(sc->T2_2) = caddr(car(args));
		    car(sc->T2_1) = c_call(car(args))(sc, sc->T2_1);
		    car(sc->T2_2) = cadr(args);
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_opCSq_S:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_SAFE_C_opCSq_S:
		  {
		    s7_pointer args, val1;
		    args = cdr(code);
		    val1 = finder(sc, cadr(args));
		    car(sc->T2_2) = finder(sc, caddr(car(args)));
		    car(sc->T2_1) = cadr(car(args));
		    car(sc->T2_1) = c_call(car(args))(sc, sc->T2_1);
		    car(sc->T2_2) = val1;
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_S_opSCq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;
		  
		case HOP_SAFE_C_S_opSCq:
		  {
		    s7_pointer val1, args;
		    args = cdr(code);
		    val1 = finder(sc, car(args));
		    car(sc->T2_1) = finder(sc, cadr(cadr(args)));
		    car(sc->T2_2) = caddr(cadr(args));
		    car(sc->T2_2) = c_call(cadr(args))(sc, sc->T2_1);
		    car(sc->T2_1) = val1;
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_C_opSCq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;
		  
		case HOP_SAFE_C_C_opSCq:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    car(sc->T2_1) = finder(sc, cadr(cadr(args)));
		    car(sc->T2_2) = caddr(cadr(args));
		    car(sc->T2_2) = c_call(cadr(args))(sc, sc->T2_1);
		    car(sc->T2_1) = car(args);
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_S_opSSq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;
		  
		case HOP_SAFE_C_S_opSSq:
		  {
		    s7_pointer val1, val2, args;
		    args = cdr(code);
		    val1 = finder(sc, car(args));
		    val2 = finder(sc, cadr(cadr(args)));
		    car(sc->T2_2) = finder(sc, caddr(cadr(args)));
		    car(sc->T2_1) = val2;
		    car(sc->T2_2) = c_call(cadr(args))(sc, sc->T2_1);
		    car(sc->T2_1) = val1;
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_S_opCSq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;
		  
		case HOP_SAFE_C_S_opCSq:
		  {
		    s7_pointer val1, args;
		    args = cdr(code);
		    val1 = finder(sc, car(args));
		    car(sc->T2_2) = finder(sc, caddr(cadr(args)));
		    car(sc->T2_1) = cadr(cadr(args));
		    car(sc->T2_2) = c_call(cadr(args))(sc, sc->T2_1);
		    car(sc->T2_1) = val1;
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_opSq_S:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_SAFE_C_opSq_S:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    car(sc->T1_1) = finder(sc, cadr(car(args)));
		    sc->temp4 = c_call(car(args))(sc, sc->T1_1);
		    car(sc->T2_2) = finder(sc, cadr(args));
		    car(sc->T2_1) = sc->temp4;
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_opSq_P:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_SAFE_C_opSq_P:
		  {
		    s7_pointer args;
		    args = cadr(code);
		    car(sc->T1_1) = finder(sc, cadr(args));
		    push_stack(sc, OP_SAFE_C_opSq_P_1, list_1(sc, c_call(args)(sc, sc->T1_1)), sc->code);
		    sc->code = caddr(code);
		    goto EVAL_PAIR;
		  }

		  
		case OP_SAFE_C_opCq_S:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_SAFE_C_opCq_S:
		  {
		    s7_pointer args, val;
		    args = cdr(code);
		    val = finder(sc, cadr(args));
		    car(sc->T2_1) = c_call(car(args))(sc, cdr(car(args)));
		    car(sc->T2_2) = val;
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_opCq_C:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_SAFE_C_opCq_C:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    car(sc->T2_1) = c_call(car(args))(sc, cdr(car(args)));
		    car(sc->T2_2) = cadr(args); /* the 2nd C stands for 1 arg? */
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_opSq_C:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  
		case HOP_SAFE_C_opSq_C:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    car(sc->T1_1) = finder(sc, cadr(car(args)));
		    car(sc->T2_1) = c_call(car(args))(sc, sc->T1_1);
		    car(sc->T2_2) = cadr(args);
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_opSq_opSq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;
		  
		case HOP_SAFE_C_opSq_opSq:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    car(sc->T1_1) = finder(sc, cadr(car(args)));
		    sc->temp4 = c_call(car(args))(sc, sc->T1_1);
		    args = cadr(args);
		    car(sc->T1_1) = finder(sc, cadr(args));
		    car(sc->T2_2) = c_call(args)(sc, sc->T1_1);
		    car(sc->T2_1) = sc->temp4;
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }


		case OP_SAFE_C_opCq_opCq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;
		  
		case HOP_SAFE_C_opCq_opCq:
		  {
		    s7_pointer args;
		    args = cdr(code);
		    car(sc->T2_1) = c_call(car(args))(sc, cdr(car(args)));
		    car(sc->T2_2) = c_call(cadr(args))(sc, cdr(cadr(args)));
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }


		case OP_SAFE_C_opSCq_opSCq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;
		  
		case HOP_SAFE_C_opSCq_opSCq:
		  {
		    s7_pointer args, val2;
		    args = cdr(code);
		    val2 = finder(sc, cadr(cadr(args)));
		    car(sc->T2_1) = finder(sc, cadr(car(args)));
		    car(sc->T2_2) = caddr(car(args));
		    sc->temp4 = c_call(car(args))(sc, sc->T2_1);
		    car(sc->T2_1) = val2;
		    car(sc->T2_2) = caddr(cadr(args));
		    car(sc->T2_2) = c_call(cadr(args))(sc, sc->T2_1);
		    car(sc->T2_1) = sc->temp4;
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }

		  
		case OP_SAFE_C_opSSq_opSSq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(code)))
		    break;
		  if (!c_function_is_ok(sc, caddr(code)))
		    break;
		  
		case HOP_SAFE_C_opSSq_opSSq:
		  {
		    s7_pointer args, val3, val4;
		    args = cdr(code);

		    val3 = finder(sc, caddr(car(args)));
		    val4 = finder(sc, caddr(cadr(args)));
		    
		    car(sc->T2_1) = finder(sc, cadr(car(args)));
		    car(sc->T2_2) = val3;
		    sc->temp4 = c_call(car(args))(sc, sc->T2_1);
		    car(sc->T2_1) = finder(sc, cadr(cadr(args)));
		    car(sc->T2_2) = val4;
		    car(sc->T2_2) = c_call(cadr(args))(sc, sc->T2_1);
		    car(sc->T2_1) = sc->temp4;
		    sc->value = c_call(code)(sc, sc->T2_1);
		    goto START;
		  }


		case OP_SAFE_C_opSAFE_CLOSURE_SSq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!function_is_ok(cadr(code)))
		    break;

		case HOP_SAFE_C_opSAFE_CLOSURE_SSq:
		  {
		    s7_pointer val, arg;
		    arg = cadr(code);
		    val = finder(sc, caddr(arg));
		    car(sc->T2_1) = finder(sc, cadr(arg));
		    car(sc->T2_2) = val;
		    push_stack(sc, OP_SAFE_C_P_1, sc->NIL, code);
		    sc->code = ecdr(arg);
		    goto SAFE_CLOSURE_TWO;
		  }

		case OP_SAFE_C_opSAFE_CLOSURE_opSq_Sq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!c_function_is_ok(sc, cadr(cadr(code))))
		    break;
		  if (!function_is_ok(cadr(code)))
		    break;

		case HOP_SAFE_C_opSAFE_CLOSURE_opSq_Sq:
		  {
		    s7_pointer val1, arg;
		    arg = cadr(code);
		    val1 = finder(sc, caddr(arg));
		    car(sc->T1_1) = finder(sc, cadr(cadr(arg)));
		    car(sc->T2_1) = c_call(cadr(arg))(sc, sc->T1_1);
		    car(sc->T2_2) = val1;
		    push_stack(sc, OP_SAFE_C_P_1, sc->NIL, code);
		    sc->code = ecdr(arg);
		    goto SAFE_CLOSURE_TWO;
		  }


		case OP_C_L:
		  if (!c_function_is_ok(sc, code))
		    break;
		  check_lambda_args(sc, cadr(cadr(code)));

		case HOP_C_L:
		  {
		    s7_pointer x, arg1;

		    MAKE_CLOSURE(sc, x, car(cdadr(code)), cdr(cdadr(code)), sc->envir);
		    NEW_CELL_NO_CHECK(sc, arg1);
		    set_type(arg1, T_PAIR);
		    car(arg1) = x;
		    cdr(arg1) = sc->NIL;
		    sc->args = arg1;

		    sc->value = c_call(code)(sc, sc->args);
		    goto START;
		  }


		case OP_CALL_WITH_EXIT:
		  if (!c_function_is_ok(sc, code))
		    break;
		  check_lambda_args(sc, cadr(cadr(code))); 
		  
		case HOP_CALL_WITH_EXIT:
		  {
		    s7_pointer go, args;
		    go = make_goto(sc);
		    args = car(cdadr(code));
		    push_stack(sc, OP_DEACTIVATE_GOTO, go, code); /* code arg is ignored, but perhaps this is safer in GC? */
		    if (is_symbol(args))
		      NEW_FRAME_WITH_SLOT(sc, sc->envir, sc->envir, args, list_1(sc, go));
		    else
		      {
			if (!is_pair(args))
			  return(s7_error(sc, 
					  sc->WRONG_NUMBER_OF_ARGS, 
					  list_3(sc, sc->NOT_ENOUGH_ARGUMENTS, args, sc->cur_code)));
			if (!is_null(cdr(args)))
			  return(s7_error(sc, 
					  sc->WRONG_NUMBER_OF_ARGS, 
					  list_3(sc, sc->TOO_MANY_ARGUMENTS, args, sc->cur_code)));
			NEW_FRAME_WITH_SLOT(sc, sc->envir, sc->envir, car(args), go);
		      }
		    sc->code = cdr(cdadr(code));
		    if (is_one_liner(sc->code))
		      {
			sc->code = car(sc->code);
			goto EVAL_PAIR;
		      }
		    goto BEGIN;
		  } 


		case OP_C_LL:
		  if (!c_function_is_ok(sc, code))
		    break;
		  check_lambda_args(sc, cadr(cadr(code)));
		  check_lambda_args(sc, cadr(caddr(code)));

		case HOP_C_LL:
		  {
		    s7_pointer x, y, z, arg1, arg2;

		    z = cdadr(code);
		    MAKE_CLOSURE(sc, x, car(z), cdr(z), sc->envir);
		    z = cdaddr(code);
		    MAKE_CLOSURE(sc, y, car(z), cdr(z), sc->envir);

		    NEW_CELL_NO_CHECK(sc, arg1);
		    set_type(arg1, T_PAIR);
		    car(arg1) = x;

		    NEW_CELL_NO_CHECK(sc, arg2);
		    set_type(arg2, T_PAIR);
		    car(arg2) = y;
		    cdr(arg1) = arg2;
		    cdr(arg2) = sc->NIL;
		    sc->args = arg1;

		    sc->value = c_call(code)(sc, sc->args);
		    goto START;
		  }


		case OP_C_FOR_EACH_LS:
		  if (!c_function_is_ok(sc, code))
		    break;
		  check_lambda_args(sc, cadr(cadr(code)));

		case HOP_C_FOR_EACH_LS:
		  {
		    s7_pointer x, y, z;

		    y = cdadr(code);
		    MAKE_CLOSURE(sc, x, car(y), cdr(y), sc->envir);

		    z = find_symbol_or_bust(sc, caddr(code));
		    if ((!is_pair(z)) ||
			(s7_list_length(sc, z) == 0))
		      {
			sc->value = g_for_each(sc, list_2(sc, x, z));
			goto START;
		      }
		    sc->code = x;
		    sc->args = z;
		    goto FOR_EACH_SIMPLE;
		  }


		case OP_C_FOR_EACH_LS_2:
		  if (!c_function_is_ok(sc, code))
		    break;
		  check_lambda_args(sc, cadr(cadr(code)));

		case HOP_C_FOR_EACH_LS_2:
		  {
		    s7_pointer x, y, z;

		    y = cdadr(code);
		    MAKE_CLOSURE(sc, x, car(y), cdr(y), sc->envir);

		    z = find_symbol_or_bust(sc, caddr(code));
		    if ((!is_pair(z)) ||
			(s7_list_length(sc, z) == 0))
		      {
			sc->value = g_for_each(sc, list_2(sc, x, z));
			goto START;
		      }
		    sc->code = x;
		    sc->args = z;
		    goto FOR_EACH_SIMPLER;
		  }


		case OP_C_FOR_EACH_L_opSq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!indirect_c_function_is_ok(sc, caddr(code)))
		    break;
		  check_lambda_args(sc, cadr(cadr(code)));

		case HOP_C_FOR_EACH_L_opSq:
		  {
		    s7_pointer x, y, z;

		    y = cdadr(code);
		    MAKE_CLOSURE(sc, x, car(y), cdr(y), sc->envir);

		    y = caddr(code);
		    car(sc->T1_1) = find_symbol_or_bust(sc, cadr(y));
		    z = c_call(y)(sc, sc->T1_1);

		    if ((!is_pair(z)) ||
			(s7_list_length(sc, z) == 0))
		      {
			sc->value = g_for_each(sc, list_2(sc, x, z));
			goto START;
		      }
		    sc->code = x;
		    sc->args = z;
		    goto FOR_EACH_SIMPLER;
		  }


		case OP_C_LS:
		  if (!c_function_is_ok(sc, code))
		    break;
		  check_lambda_args(sc, cadr(cadr(code)));

		case HOP_C_LS:
		  {
		    s7_pointer x, arg1, arg2, y;
		    y = cdadr(code);
		    MAKE_CLOSURE(sc, x, car(y), cdr(y), sc->envir);

		    NEW_CELL_NO_CHECK(sc, arg1);
		    set_type(arg1, T_PAIR);
		    car(arg1) = x;
		    
		    NEW_CELL_NO_CHECK(sc, arg2);
		    set_type(arg2, T_PAIR);
		    cdr(arg1) = arg2;
		    sc->args = arg1;
		    car(arg2) = find_symbol_or_bust(sc, caddr(code));
		    cdr(arg2) = sc->NIL;
		    /* (define (hi a) (for-each (lambda (n) (display n) (newline)) a)) (hi '(1 2 3)) */
		    
		    sc->value = c_call(code)(sc, sc->args);
		    goto START;
		  }


		case OP_C_L_opSq:
		  if (!c_function_is_ok(sc, code))
		    break;
		  if (!indirect_c_function_is_ok(sc, caddr(code)))
		    break;
		  check_lambda_args(sc, cadr(cadr(code)));

		case HOP_C_L_opSq:
		  {
		    s7_pointer x, y, arg1, arg2;
		    
		    y = cdadr(code);
		    MAKE_CLOSURE(sc, x, car(y), cdr(y), sc->envir);

		    NEW_CELL_NO_CHECK(sc, arg1);
		    set_type(arg1, T_PAIR);
		    car(arg1) = x;

		    NEW_CELL_NO_CHECK(sc, arg2);
		    set_type(arg2, T_PAIR);
		    cdr(arg1) = arg2;
		    sc->args = arg1;

		    car(sc->T1_1) = find_symbol_or_bust(sc, cadr(caddr(code)));
		    car(arg2) = c_call(caddr(code))(sc, sc->T1_1);
		    cdr(arg2) = sc->NIL;

		    /* (define (hi a) (for-each (lambda (n) (display n) (newline)) a)) (hi '(1 2 3)) */
		    
		    sc->value = c_call(code)(sc, sc->args);
		    goto START;
		  }


		case OP_C_CLL:
		  if (!c_function_is_ok(sc, code))
		    break;
		  check_lambda_args(sc, cadr(caddr(code)));
		  check_lambda_args(sc, cadr(cadddr(code)));

		case HOP_C_CLL:
		  {
		    s7_pointer x, y, z, arg1, arg2, arg3;

		    z = cdaddr(code);
		    MAKE_CLOSURE(sc, x, car(z), cdr(z), sc->envir);
		    z = cdr(cadddr(code));
		    MAKE_CLOSURE(sc, y, car(z), cdr(z), sc->envir);

		    NEW_CELL_NO_CHECK(sc, arg1);
		    set_type(arg1, T_PAIR);
		    car(arg1) = cadr(code);

		    NEW_CELL_NO_CHECK(sc, arg2);
		    set_type(arg2, T_PAIR);
		    car(arg2) = x;
		    cdr(arg1) = arg2;

		    NEW_CELL_NO_CHECK(sc, arg3);
		    set_type(arg3, T_PAIR);
		    car(arg3) = y;
		    cdr(arg2) = arg3;
		    cdr(arg3) = sc->NIL;
		    sc->args = arg1;

		    sc->value = c_call(code)(sc, sc->args);
		    goto START;
		  }


		case OP_C_CATCH:
		  if (!c_function_is_ok(sc, code))
		    break;
		  check_lambda_args(sc, cadr(caddr(code)));
		  check_lambda_args(sc, cadr(cadddr(code)));

		case HOP_C_CATCH:
		  {
		    s7_pointer y, p, z, args;
		    /* code is (catch #t (lambda () ....) (lambda args ....)) */

		    args = cddr(code);
		    if (!is_null(cadar(args)))
		      return(s7_wrong_type_arg_error(sc, "catch", 2, car(args), "a thunk"));
		    
		    z = cdadr(args);
		    MAKE_CLOSURE(sc, y, car(z), cdr(z), sc->envir);

		    NEW_CELL_NO_CHECK(sc, p);
		    catch_tag(p) = cadr(code);
		    catch_goto_loc(p) = s7_stack_top(sc);
		    catch_op_loc(p) = (int)(sc->op_stack_now - sc->op_stack);
		    catch_handler(p) = y;
		    set_type(p, T_CATCH);

		    push_stack(sc, OP_CATCH, code, p); /* code ignored here, except by GC */
		    sc->code = cddar(args);
		    NEW_FRAME(sc, sc->envir, sc->envir);
		    goto BEGIN;
		  }


 		default:
 		  fprintf(stderr, "bad op in opt_eval: op %d, is_opt: %d, %s\n",
 			  optimize_data(code), is_optimized(code), DISPLAY_80(code));
 		  break;
		}

	      /* else cancel all the optimization info -- someone stepped on our symbol */

	      /* fprintf(stderr, "clear %s %s\n", opt_names[optimize_data(code)], DISPLAY(code)); */
	      clear_optimized(code);
	      clear_optimize_data(code);
	      ecdr(code) = sc->NIL;
	      fcdr(code) = sc->NIL;
	      /* and fall into the normal evaluator */
	    }
#endif
	  
	  /* -------------------------------------------------------------------------------- */

	  /* fprintf(stderr, "unopt: %s\n", DISPLAY_80(sc->code));  */

	  /* trailers */
	  sc->cur_code = sc->code;

	  {
	    s7_pointer carc;
	    carc = car(sc->code);
	    if (is_symbol(carc))
	      {
		/* car is a symbol, sc->code a list */
		sc->value = SYMBOL_TO_VALUE(sc, carc); /* not finder here -- nearly 2/3's of these are globals */
		/*
		if ((is_c_function(sc->value)) &&
		    (is_safe_procedure(sc->value)))
		  fprintf(stderr, "unopt safe: %s\n", DISPLAY(sc->code));
		*/
		sc->code = cdr(sc->code);
		sc->args = sc->NIL;
		/* drop into eval args
		 */
	      }
	    else
	      {
		/* very uncommon case: car is either itself a pair or some non-symbol */
		if (is_pair(carc))
		  {
		    /* evaluate the inner list */
		    push_stack(sc, OP_EVAL_ARGS, sc->NIL, cdr(sc->code));
		    if (is_syntactic(car(carc)))
		      {
			if ((car(carc) == sc->QUOTE) &&        /* ('and #f) */
			    (is_syntactic(cadr(carc))))
			  return(apply_error(sc, cadr(carc), cdr(sc->code)));
			
			sc->op = (opcode_t)syntax_opcode(car(carc));
			sc->code = cdr(carc);
			goto START_WITHOUT_POP_STACK;
		      }
		    
		    push_stack(sc, OP_EVAL_ARGS, sc->NIL, cdr(carc));
		    sc->code = car(carc);
		    goto EVAL;
		  }
		else
		  {
		    /* car must be the function to be applied */
		    sc->value = carc;
		    sc->code = cdr(sc->code);
		    sc->args = sc->NIL;
		    /* drop into OP_EVAL_ARGS */
		  }
	      }
	  }
	}
      else
	{
	  /* here sc->code is not a pair */
	  if (is_symbol(sc->code))
	    {
	      sc->value = finder(sc, sc->code);
	      pop_stack(sc);
	      if (sc->op != OP_EVAL_ARGS)
		goto START_WITHOUT_POP_STACK;
	      /* drop into EVAL_ARGS */
	    }
	  else
	    {
	      /* sc->code is not a pair or a symbol */
	      sc->value = sc->code;
	      goto START;
	    }
	}
      /* sc->value is car=something applicable
       * sc->code = rest of expression
       * sc->args is nil (set by the drop-through cases above -- perhaps clearer to bring that down?)
       */
      
      
    case OP_EVAL_ARGS:
      if (dont_eval_args(sc->value))
	{
	  if (is_any_macro(sc->value))
	    {    
	      /* macro expansion */
	      push_stack(sc, OP_EVAL_MACRO, sc->NIL, sc->code); /* sc->code is here for (vital) GC protection */
	      /* 
	       * pass a list (mac . args) to the macro expander
	       */
#if 0
	      car(sc->TEMP_CELL_1) = sc->value; /* macro */
	      cdr(sc->TEMP_CELL_1) = sc->code;  /* args */
#endif
	      sc->args = list_1(sc, cons(sc, sc->value, sc->code));
	      sc->code = sc->value;
	      goto APPLY;                      /* not UNSAFE_CLOSURE because it might be a bacro */
	    }

	  /* (define progn begin)
	   * (progn (display "hi") (+ 1 23))
	   */
	  sc->op = (opcode_t)syntax_opcode(sc->value);
	  goto START_WITHOUT_POP_STACK;
	}

      /* sc->value is the func
       *   we don't have to delay lookup of the func because arg evaluation order is not specified, so
       *     (let ((func +)) (func (let () (set! func -) 3) 2))
       *   can return 5.
       */
      /* if (is_null(sc->code)) {sc->code = sc->value; goto APPLY;}
       *   this is hit very rarely so it costs more than it saves
       */

      push_op_stack(sc, sc->value);
      if (sc->op_stack_now >= sc->op_stack_end)
	resize_op_stack(sc);

      goto EVAL_ARGS;
      /* moving eval_args up here (to avoid this goto) was slightly slower, probably by chance.
       */
      /* this code can almost certainly be simplified -- "it just growed..." 
       */


    case OP_EVAL_ARGS5:
      /* sc->value is the last arg, sc->code is the previous
       */
      {
	s7_pointer x, y;

	sc->z = pop_op_stack(sc);
	if (is_safe_procedure(sc->z))
	  {
	    x = sc->TEMP_CELL_3;
	    y = sc->TEMP_CELL_2;
	  }
	else
	  {
	    NEW_CELL(sc, x); 
	    set_type(x, T_PAIR);
	    NEW_CELL_NO_CHECK(sc, y); 
	    set_type(y, T_PAIR);
	  }

	car(x) = sc->code;
	cdr(x) = sc->args;
	car(y) = sc->value;
	cdr(y) = x;
	sc->args = safe_reverse_in_place(sc, y); 
	sc->code = sc->z;
	goto APPLY;
      }

      
    case OP_EVAL_ARGS2:
      /* sc->value is the last arg, [so if is_null(cdr(sc->code) and current is pair, push args2]
       */
      {
	s7_pointer x;

	sc->code = pop_op_stack(sc);
	if (is_safe_procedure(sc->code))
	  x = sc->TEMP_CELL_2;
	else
	  {
	    NEW_CELL(sc, x);
	    set_type(x, T_PAIR);
	  }

	car(x) = sc->value;
	cdr(x) = sc->args;
	if (type(sc->args) != T_NIL)
	  sc->args = safe_reverse_in_place(sc, x);
	else sc->args = x;
	goto APPLY;
      }

#if WITH_OPTIMIZATION
      /* tricky cases here all involve values (i.e. multiple-values) */

    case OP_EVAL_ARGS_P_1:
      if (is_not_null(sc->args))
	{
	  car(sc->TEMP_CELL_2) = sc->value;
	  cdr(sc->TEMP_CELL_2) = sc->args;
	  /* we can't just call the c function directly here because the values call that
	   *   appended these new arguments might have added too many, so it's simpler
	   *   to let OP_APPLY sort that out.
	   */
	  sc->code = ecdr(sc->code);
 	  sc->args = safe_reverse_in_place(sc, sc->TEMP_CELL_2);
 	  goto APPLY;
	}
      car(sc->T1_1) = sc->value;
      sc->value = c_call(sc->code)(sc, sc->T1_1);
      goto START;


    case OP_EVAL_ARGS_P_2:
      if (is_not_null(cdr(sc->args)))
	{
	  /* P must have resulted in values being spliced into the arg list.
	   *   caller is being difficult, so we'll just mimic the code above.
	   */
	  car(sc->TEMP_CELL_2) = sc->value;
	  cdr(sc->TEMP_CELL_2) = sc->args;
	  sc->args = safe_reverse_in_place(sc, sc->TEMP_CELL_2);
	  /* see note above */
	  sc->code = ecdr(sc->code);
	  goto APPLY;
	}
      car(sc->T2_2) = sc->value;
      car(sc->T2_1) = car(sc->args);
      sc->value = c_call(sc->code)(sc, sc->T2_1);

      typeflag(sc->args) = 0;                          /* this was allocated above, so we can release it now */
      (*(sc->free_heap_top++)) = sc->args;
      sc->args = sc->NIL;

      goto START;


    case OP_EVAL_ARGS_P_3:
      sc->w = sc->value;
      if (is_not_null(sc->args))
	{
	  /* (define (hi a) (+ (values 1 2) a)) 
	   * (define (hi a) (log (values 1 2) a))
	   */
	  sc->z = sc->args;
	  car(sc->TEMP_CELL_2) = finder(sc, caddr(sc->code));
	  cdr(sc->TEMP_CELL_2) = sc->TEMP_CELL_3;
	  car(sc->TEMP_CELL_3) = sc->w;
	  cdr(sc->TEMP_CELL_3) = sc->z;
	  sc->args = safe_reverse_in_place(sc, sc->TEMP_CELL_2);
	  sc->code = ecdr(sc->code);
	  goto APPLY;
	}
      car(sc->T2_2) = finder(sc, caddr(sc->code));
      car(sc->T2_1) = sc->w;
      sc->value = c_call(sc->code)(sc, sc->T2_1);
      goto START;


    case OP_EVAL_ARGS_P_4:
	if (is_not_null(sc->args))
	  {
	    car(sc->TEMP_CELL_2) = caddr(sc->code);
	    cdr(sc->TEMP_CELL_2) = sc->TEMP_CELL_3;
	    car(sc->TEMP_CELL_3) = sc->value;
	    cdr(sc->TEMP_CELL_3) = sc->args;
	    sc->args = safe_reverse_in_place(sc, sc->TEMP_CELL_2);
	    sc->code = ecdr(sc->code);
	    goto APPLY;                    /* (define (hi) (log (values 1 2) 3)) ? */
	  }
      car(sc->T2_1) = sc->value;
      car(sc->T2_2) = caddr(sc->code);
      sc->value = c_call(sc->code)(sc, sc->T2_1);
      goto START;

    case OP_SAFE_C_ZS_1:
      car(sc->T2_1) = sc->value;
      car(sc->T2_2) = sc->args;
      sc->value = c_call(sc->code)(sc, sc->T2_1);
      goto START;

    case OP_SAFE_C_SZ_1:
      car(sc->T2_1) = sc->args;
      car(sc->T2_2) = sc->value;
      sc->value = c_call(sc->code)(sc, sc->T2_1);
      goto START;

    case OP_SAFE_C_ZZ_1:
      push_stack(sc, OP_SAFE_C_ZZ_2, sc->value, sc->code);
      sc->code = caddr(sc->code);
      goto OPT_EVAL;

    case OP_SAFE_C_ZZ_2:
      car(sc->T2_1) = sc->args;
      car(sc->T2_2) = sc->value;
      sc->value = c_call(sc->code)(sc, sc->T2_1);
      goto START;

    case OP_SAFE_C_ZXX_1:
      car(sc->T3_3) = (is_symbol(cadddr(sc->code))) ? finder(sc, cadddr(sc->code)) : cadddr(sc->code);
      car(sc->T3_1) = sc->value;
      car(sc->T3_2) = sc->args;
      sc->value = c_call(sc->code)(sc, sc->T3_1);
      goto START;

    case OP_SAFE_C_XZX_1:
      car(sc->T3_3) = (is_symbol(cadddr(sc->code))) ? finder(sc, cadddr(sc->code)) : cadddr(sc->code);
      car(sc->T3_2) = sc->value;
      car(sc->T3_1) = sc->args;
      sc->value = c_call(sc->code)(sc, sc->T3_1);
      goto START;

    case OP_SAFE_C_XXZ_1:
      car(sc->T3_2) = (is_symbol(caddr(sc->code))) ? finder(sc, caddr(sc->code)) : caddr(sc->code);
      car(sc->T3_3) = sc->value;
      car(sc->T3_1) = sc->args;
      sc->value = c_call(sc->code)(sc, sc->T3_1);
      goto START;

    case OP_SAFE_C_ZZX_1:
      push_op_stack(sc, sc->value);
      push_stack(sc, OP_SAFE_C_ZZX_2, sc->args, sc->code);
      sc->code = caddr(sc->code);
      goto OPT_EVAL;

    case OP_SAFE_C_ZZX_2:
      car(sc->T3_1) = pop_op_stack(sc);
      car(sc->T3_2) = sc->value;
      car(sc->T3_3) = sc->args;
      sc->value = c_call(sc->code)(sc, sc->T3_1);
      goto START;

    case OP_SAFE_C_ZXZ_1:
      push_op_stack(sc, sc->value);
      push_stack(sc, OP_SAFE_C_ZXZ_2, sc->args, sc->code);
      sc->code = cadddr(sc->code);
      goto OPT_EVAL;

    case OP_SAFE_C_ZXZ_2:
      car(sc->T3_1) = pop_op_stack(sc);
      car(sc->T3_2) = sc->args;
      car(sc->T3_3) = sc->value;
      sc->value = c_call(sc->code)(sc, sc->T3_1);
      goto START;

    case OP_SAFE_C_XZZ_1:
      push_op_stack(sc, sc->value);
      push_stack(sc, OP_SAFE_C_XZZ_2, sc->args, sc->code);
      sc->code = cadddr(sc->code);
      goto OPT_EVAL;

    case OP_SAFE_C_XZZ_2:
      car(sc->T3_1) = sc->args;
      car(sc->T3_2) = pop_op_stack(sc);
      car(sc->T3_3) = sc->value;
      sc->value = c_call(sc->code)(sc, sc->T3_1);
      goto START;

    case OP_SAFE_C_ZZZ_1:
      push_stack(sc, OP_SAFE_C_ZZZ_2, sc->value, sc->code);
      sc->code = caddr(sc->code);
      goto OPT_EVAL;

    case OP_SAFE_C_ZZZ_2:
      push_op_stack(sc, sc->value);
      push_stack(sc, OP_SAFE_C_ZZZ_3, sc->args, sc->code);
      sc->code = cadddr(sc->code);
      goto OPT_EVAL;

    case OP_SAFE_C_ZZZ_3:
      car(sc->T3_1) = sc->args;
      car(sc->T3_2) = pop_op_stack(sc);
      car(sc->T3_3) = sc->value;
      sc->value = c_call(sc->code)(sc, sc->T3_1);
      goto START;

    case OP_SAFE_C_opSq_P_1:
      if (is_null(cdr(sc->args))) /* values not called */
	{
	  car(sc->T2_2) = sc->value;
	  car(sc->T2_1) = car(sc->args);
	  sc->value = c_call(sc->code)(sc, sc->T2_1);
	}
      else
	{
	  /* here we need an argnum check since values could have appended any number of args
	   */
	  unsigned int len;
	  len = safe_list_length(sc, sc->args); 

	  /* can values return an improper or circular list?  I don't think so:
	   *   (values 1 . 2) -> improper arg list error (same with apply values)
	   *
	   * currently (values) does not simply erase itself:
	   *
	   *   :(let () (define (arg2 a) (let ((b 1)) (set! b (+ a b)) (values))) (define (hi c) (expt (abs c) (arg2 2))) (hi 2))
	   *   ;expt power, argument 2, #<unspecified>, is an untyped but should be a number
	   *  
	   *   :(s7-version (values))
	   *   ;s7-version: too many arguments: (#<unspecified>)
	   *
	   *   :(exp (values) 0.0)
	   *   ;exp: too many arguments: (#<unspecified> 0.0)
	   *
	   * map is explicitly a special case, and surely it is more confusing to have (values) scattered at random.
	   * also this is consistent with the unoptimized version
	   */

	  if (c_function_all_args(ecdr(sc->code)) < (len + 1)) /* + 1 because we have the last arg waiting */
	    return(s7_error(sc, 
			    sc->WRONG_NUMBER_OF_ARGS, 
			    list_3(sc, sc->TOO_MANY_ARGUMENTS, sc->code, sc->args)));

	  sc->args = safe_reverse_in_place(sc, cons(sc, sc->value, sc->args));
	  sc->value = c_call(sc->code)(sc, sc->args);
	}
      goto START;
#endif
      
    case OP_EVAL_ARGS3:
      /* sc->value is the next-to-last arg, and we know the last arg is not a list (so values can't mess us up!)
       */
      {
	s7_pointer x, y, val;

	if (type(sc->code) == T_SYMBOL)
	  val = finder(sc, sc->code);
	else val = sc->code;

	sc->x = pop_op_stack(sc);
	if (is_safe_procedure(sc->x))
	  {
	    x = sc->TEMP_CELL_3;
	    y = sc->TEMP_CELL_2;
	  }
	else
	  {
	    NEW_CELL(sc, x); 
	    set_type(x, T_PAIR);
	    NEW_CELL_NO_CHECK(sc, y);
	    set_type(y, T_PAIR);
	  }
	car(x) = sc->value;
	cdr(x) = sc->args;
	car(y) = val;
	cdr(y) = x;
	sc->args = safe_reverse_in_place(sc, y); 
	sc->code = sc->x;
	goto APPLY;
      }

      
    case OP_EVAL_ARGS4:
      /* sc->code is a pair, and either cdr(sc->code) is not null or car(sc->code) is a pair 
       */
      /* fprintf(stderr, "code: %s, value: %s, args: %s\n", DISPLAY_80(sc->code), DISPLAY_80(sc->value), DISPLAY_80(sc->args)); 
       *   (#f #f) (env #f) etc.  args is very often nil here, so we're looking at 3 simple args
       *   or even just 2 in some cases: (+ req opt) with value 2 and args ()
       */
      {
        s7_pointer x;
	NEW_CELL(sc, x);
	car(x) = sc->value;
	cdr(x) = sc->args;
	set_type(x, T_PAIR);
	sc->args = x;
	goto EVAL_ARGS_PAIR;
      }


    case OP_EVAL_ARGS1:
      {
        s7_pointer x;
	NEW_CELL(sc, x);
	car(x) = sc->value;
	cdr(x) = sc->args;
	set_type(x, T_PAIR);
	sc->args = x;
      }
      

    EVAL_ARGS:
      /* 1st time, value = op, args = nil, code is args */

      if (is_pair(sc->code))  /* evaluate current arg -- must check for pair here, not sc->NIL (improper list as args) */
	{ 
	  int typ;
	  s7_pointer car_code;

	EVAL_ARGS_PAIR:
	  car_code = car(sc->code);
	  typ = type(car_code);
	  
	  /* switch statement here is much slower for some reason */
	  if (typ == T_PAIR)
	    {
	      if (is_null(cdr(sc->code)))
		push_stack(sc, OP_EVAL_ARGS2, sc->args, sc->NIL);
	      else 
		{
		  if (!is_pair(cdr(sc->code)))            /* (= 0 '(1 . 2) . 3) */
		    improper_arglist_error(sc);

		  if ((is_null(cddr(sc->code))) &&
		      (!is_pair(cadr(sc->code))))
		    push_stack(sc, OP_EVAL_ARGS3, sc->args, cadr(sc->code)); 
		  else push_stack(sc, OP_EVAL_ARGS4, sc->args, cdr(sc->code));
		}
	      sc->code = car_code;
	      /* s7: unopt 455508, h_safe_c_c: 31971 c_all_g: 100309
	       * lg: unopt 80308, h_safe_c_s: 27119 closure_opsq: 24073
	       * in: unopt: 212977, h_safe_c_c: 77592
	       */
	      goto EVAL_PAIR;
	    }

	  /* car(sc->code) is not a pair */
	  if (is_pair(cdr(sc->code)))
	    {
	      sc->code = cdr(sc->code);
	      if (typ == T_SYMBOL)
		sc->value = finder(sc, car_code);
	      else sc->value = car_code;
	      /* sc->value is the current arg's value, sc->code is pointing to the next */

	      /* cdr(sc->code) may not be a pair or nil here! 
	       *   (eq? #f . 1) -> sc->code is 1
	       */
	      if (is_null(cdr(sc->code)))
		{
		  s7_pointer x, y, val;
		  /* we're at the last arg, sc->value is the previous one, not yet saved in the args list */
		  car_code = car(sc->code);
		  typ = type(car_code);

		  if (typ == T_PAIR)
		    {
		      push_stack(sc, OP_EVAL_ARGS5, sc->args, sc->value);
		      sc->code = car_code;
		      goto EVAL_PAIR;
		    }

		  /* get the last arg */
		  if (typ == T_SYMBOL)
		    val = finder(sc, car_code);
		  else val = car_code;
		  
		  /* get the current arg, which is not a list */
		  sc->code = pop_op_stack(sc);
		  if (is_safe_procedure(sc->code)) /* this check does not currently pay for itself, I think */
		    {
		      x = sc->TEMP_CELL_3; /* these are already pairs */
		      y = sc->TEMP_CELL_2;
		    }
		  else
		    {
		      NEW_CELL(sc, x); 
		      set_type(x, T_PAIR);
		      NEW_CELL_NO_CHECK(sc, y); 
		      set_type(y, T_PAIR);
		    }
		  car(x) = sc->value;
		  cdr(x) = sc->args;
		  car(y) = val;
		  cdr(y) = x;
		  sc->args = safe_reverse_in_place(sc, y); 
		  /* drop into APPLY */
		}
	      else 
		{
		  /* here we know sc->code is a pair, cdr(sc->code) is not null
		   *   sc->value is the previous arg's value
		   */
		  s7_pointer x;
		  NEW_CELL(sc, x); 
		  car(x) = sc->value;
		  cdr(x) = sc->args;
		  set_type(x, T_PAIR);
		  sc->args = x;
		  goto EVAL_ARGS_PAIR;
		}
	    }
	  else
	    {
	      /* here we've reached the last arg (sc->code == nil) 
	       *   it is not a pair (typ == T_PAIR was caught earlier)
	       *   if this is a safe function, we're going straight to
	       *   apply with no complications, so we can use TEMP_CELL_2.
	       */
	      s7_pointer x;

	      if (type(cdr(sc->code)) != T_NIL)
		improper_arglist_error(sc);

	      sc->code = pop_op_stack(sc);
	      if (is_safe_procedure(sc->code))
		{
		  x = sc->TEMP_CELL_2;
		  if (typ == T_SYMBOL)
		    car(x) = finder(sc, car_code);
		  else car(x) = car_code;
		  cdr(x) = sc->args;
		}
	      else
		{
		  s7_pointer val;
		  if (typ == T_SYMBOL)
		    val = finder(sc, car_code); /* this has to precede the set_type below */
		  else val = car_code;
		  NEW_CELL(sc, x); 
		  car(x) = val;
		  cdr(x) = sc->args;
		  set_type(x, T_PAIR);
		}

	      if (type(sc->args) != T_NIL)
		sc->args = safe_reverse_in_place(sc, x);
	      else sc->args = x;
	      /* drop into APPLY */
	    }
	}
      else                       /* got all args -- go to apply */
	{
	  if (is_not_null(sc->code))
	    improper_arglist_error(sc);
	  else
	    {
	      sc->code = pop_op_stack(sc);
	      sc->args = safe_reverse_in_place(sc, sc->args); 
	      /* fall through  */
	    }
	}

      /* we could omit the arg reversal in many cases, but lots of code assumes the args are in order;
       *   adding a bit for this in the type field saves some time in s7test (many + and * tests), but costs
       *   about the same time in other cases, so it's not a clear win.
       */

      /*
       * if (sc->stack_end >= sc->stack_resize_trigger)
       *   increase_stack_size(sc);
       *
       * the two places where the stack reaches it maximum size are in read_expression where TOKEN_LEFT_PAREN
       *   pushes OP_READ_LIST, and (the actual max) in OP_EVAL at the push of OP_EVAL_ARGS.  I've moved
       *   the stack size check from here (where it reflects the eval stack size) to read_expression (where
       *   it reflects nested list depth), and added it to the T_CLOSURE(*) parts of apply since (for example) 
       *   extremely deep recursion involving map or for-each can increase the stack size indefinitely:
       *
       * (define (tfe a b)
       *   (format #t "~A ~A -> ~A~%" a b (-s7-stack-size))
       *   (for-each
       *     (lambda (c)
       *       (if (< c b)
       *           (tfe (+ c 1) b)))
       *     (list a)))
       *
       * now (tfe 0 1000) triggers the stack increase.
       */

      if (tracing) 
	trace_apply(sc);

      /* SOMEDAY: trace needs to be either removed or re-invigorated (with optimization, it's not very useful)
       */

    APPLY:
      switch (type(sc->code))
	{
	case T_C_FUNCTION: 	                    /* -------- C-based function -------- */
	  {
	    unsigned int len;
	    len = safe_list_length(sc, sc->args);

	    if (len < c_function_required_args(sc->code))
	      return(s7_error(sc, 
			      sc->WRONG_NUMBER_OF_ARGS, 
			      list_3(sc, sc->NOT_ENOUGH_ARGUMENTS, sc->code, sc->args)));

	    if (c_function_all_args(sc->code) < len)
	      return(s7_error(sc, 
			      sc->WRONG_NUMBER_OF_ARGS, 
			      list_3(sc, sc->TOO_MANY_ARGUMENTS, sc->code, sc->args)));
	  }
	  /* drop into ... */

	case T_C_ANY_ARGS_FUNCTION:                 /* -------- C-based function that can take any number of arguments -------- */
	  sc->value = c_function_call(sc->code)(sc, sc->args);
	  goto START;


	case T_C_OPT_ARGS_FUNCTION:                 /* -------- C-based function that has n optional arguments -------- */
	  {
	    unsigned int len;
	    len = safe_list_length(sc, sc->args);
	    if (c_function_all_args(sc->code) < len)
	      return(s7_error(sc, 
			      sc->WRONG_NUMBER_OF_ARGS, 
			      list_3(sc, sc->TOO_MANY_ARGUMENTS, sc->code, sc->args)));
	    sc->value = c_function_call(sc->code)(sc, sc->args);
	    goto START;
	  }

	case T_C_RST_ARGS_FUNCTION:                 /* -------- C-based function that has n required args, then any others -------- */
	  {
	    unsigned int len;
	    len = safe_list_length(sc, sc->args);
	    if (len < c_function_required_args(sc->code))
	      return(s7_error(sc, 
			      sc->WRONG_NUMBER_OF_ARGS, 
			      list_3(sc, sc->NOT_ENOUGH_ARGUMENTS, sc->code, sc->args)));
	    sc->value = c_function_call(sc->code)(sc, sc->args);
	    /* sc->code here need not match sc->code before the function call (map for example) */
	    goto START;
	  }

	case T_C_LST_ARGS_FUNCTION:                 /* -------- {list} -------- */
	  if (sc->no_values == 0)
	    sc->value = sc->args;
	  else sc->value = g_qq_list(sc, sc->args);
	  goto START;

	case T_C_MACRO: 	                    /* -------- C-based macro -------- */
	  {
	    unsigned int len;

	    if (is_pair(car(sc->args)))            /* normally args is ((mac-name ...)) */
	      {                                    /*   but in a case like (call-with-exit quasiquote), args is (#<goto>) */
		if (!s7_is_list(sc, cdar(sc->args)))  /* (quasiquote . 1) */
		  eval_error(sc, "improper list of arguments: ~A", car(sc->args));
		sc->args = cdar(sc->args);         
	      }
	    else 
	      {
		/* I think this can only be triggered by map/for-each with a (c_)macro [in fact, quasiquote] argument,
		 *    so perhaps we can fix it here?
		 */
		eval_error(sc, "~A called as a function, but it's a macro!", sc->code);
	      }

	    len = s7_list_length(sc, sc->args);
	    if (len <= 0)                          /* (quasiquote 0 . 1) */
	      return(s7_error(sc, sc->SYNTAX_ERROR, 
			      list_3(sc, make_protected_string(sc, "~A: improper list of arguments: ~S"), sc->code, sc->args)));

	    if (len < c_macro_required_args(sc->code))
	      return(s7_error(sc, 
			      sc->WRONG_NUMBER_OF_ARGS, 
			      list_3(sc, sc->NOT_ENOUGH_ARGUMENTS, sc->code, sc->args)));
	    
	    if (c_macro_all_args(sc->code) < len)
	      return(s7_error(sc, 
			      sc->WRONG_NUMBER_OF_ARGS, 
			      list_3(sc, sc->TOO_MANY_ARGUMENTS, sc->code, sc->args)));

	    sc->value = c_macro_call(sc->code)(sc, sc->args);
	    goto START;
	  }
	  
	case T_BACRO:                                /* -------- bacro -------- */
	  NEW_FRAME(sc, sc->envir, sc->envir);       /* like let* -- we'll be adding macro args, so might as well sequester things here */
	  goto BACRO;
	  /* another choice is to do expansion and evaluation in the definition env,
	   *   so only the args come (unevaluated) from the current env.
	   *   I can't immediately see any need for this.
	   *
	   * but probably useful would be bacro without the following evaluation, sort of like a function with unevaluated args
	   *  (but expanded in the call-time env): "lacro"?
	   */

	case T_CLOSURE:                              /* -------- normal function (lambda), or macro -------- */
#if 0
	  /* this check takes longer than the code saves */

	  if (is_safe_closure(closure_body(sc->code)))
	    {
	      /* here we know nothing funny is going on in the closure body -- not even set!
	       *   so the new_frame with the args can be preallocated, and the arg values simply
	       *   stored in their slots.  We still need to update frame_id etc.  Also, we
	       *   checked earlier that none of the parameter names need access handling.
	       */
	      /* sc->args are the arg values, closure_environment is the preallocated env
	       */
	      s7_pointer x, z, env;
	      unsigned long long int id;

	      id = ++frame_number;
	      env = closure_environment(sc->code);
	      frame_id(env) = id;

	      for (x = environment_slots(env), z = sc->args; is_slot(x); x = next_slot(x), z = cdr(z))
		{
		  if (is_null(z))
		    return(s7_error(sc, 
				    sc->WRONG_NUMBER_OF_ARGS, 
				    list_3(sc, sc->NOT_ENOUGH_ARGUMENTS, closure_name(sc, sc->code), sc->args)));
		  slot_set_value(x, car(z));
		  symbol_id(slot_symbol(x)) = id;
		  local_slot(slot_symbol(x)) = x;
		}

	      if ((is_null(x)) &&
		  (is_not_null(z)))
		return(s7_error(sc, 
				sc->WRONG_NUMBER_OF_ARGS, 
				list_3(sc, sc->TOO_MANY_ARGUMENTS, closure_name(sc, sc->code), sc->args)));

	      sc->envir = env;
	      sc->code = closure_body(sc->code);

	      if (is_one_liner(sc->code))
		{
		  sc->code = car(sc->code);
		  goto EVAL_PAIR;
		}
	      goto BEGIN;
	    }
#endif
	  if (sc->stack_end >= sc->stack_resize_trigger)
	    increase_stack_size(sc);

	  /* fprintf(stderr, "unsafe: %s\n", DISPLAY_80(closure_body(sc->code)));  */

	case T_MACRO:
	  /* sc->envir = new_frame_in_env(sc, closure_environment(sc->code)); */
	  NEW_FRAME(sc, closure_environment(sc->code), sc->envir);

	BACRO:
	  /* load up the current args into the ((args) (lambda)) layout [via the current environment] */

	  /* (defmacro hi (a b) `(+ ,a ,b)) */
	  /*   -> code: #<macro>, args: ((hi 2 3)), closure args: (defmac-9) */
	  /*   then back again: code: #<closure>, args: (2 3), closure args: (a b) */

	  /* (define (hi a b) (+ a b))
	   * (hi 1 2)
	   * sc->args: (1 2)
	   */

	  /* it is possible to apply a macro:
	   *
	   * (define-macro (hi a) `(+ ,a 1))
	   * (apply hi '(4))
	   * 5
	   *
	   * (let ((x 32)) (define-macro* (mac (a (let () (display x) 3))) `(+ 1 ,a)) (let ((x 4)) (mac x)))
	   * displays 32, returns 5, so the evaluation of the default value happens in the definition environment
	   * (let ((x 32)) (define* (mac (a (let () (display x) 3))) (+ 1 a)) (let ((x 4)) (mac x)))
	   * is the same
	   * (let ((x 32)) (define-bacro* (mac (a (let () (display x) 3))) `(+ 1 ,a)) (let ((x 4)) (mac x)))
	   * displays 4 and returns 5
	   *
	   * so... we need a way to get the call environment from within either a macro or a function
	   */

	  /* (let* ((mac (let () (define-macro (mac1 a) `(+ ,a 1)) mac1)) (lst (list 1))) (set-cdr! lst lst) (apply mac lst ()))
	   * hangs -- this is equivalent to (eval <circular-list>) which also hangs.
	   */

	  {
	    s7_pointer x, z, e;
	    unsigned long long int id;
	    e = sc->envir;
	    id = frame_id(e);

	    for (x = closure_args(sc->code), z = sc->args; is_pair(x); x = cdr(x))
	      {
		s7_pointer sym, args, val;
		/* reuse the value cells as the new frame slots */
		
		if (is_null(z))
		  return(s7_error(sc, 
				  sc->WRONG_NUMBER_OF_ARGS, 
				  list_3(sc, sc->NOT_ENOUGH_ARGUMENTS, closure_name(sc, sc->code), sc->cur_code)));
		/* now that args are being reused as slots, the error message can't use sc->args, 
		 *  so fallback on sc->cur_code in this section.
		 */
		
		sym = car(x);
		val = car(z);
		args = cdr(z);
		
		slot_symbol(z) = sym;
		symbol_id(sym) = id;
		local_slot(sym) = z;
		if (symbol_has_accessor(sym))
		  slot_set_value(z, call_symbol_bind(sc, sym, val));
		else slot_set_value(z, val);
		set_type(z, T_SLOT | T_IMMUTABLE);
		next_slot(z) = environment_slots(e);
		environment_slots(e) = z;
		z = args;
	      }
	    if (is_null(x)) 
	      {
		if (is_not_null(z))
		  return(s7_error(sc, 
				  sc->WRONG_NUMBER_OF_ARGS, 
				  list_3(sc, sc->TOO_MANY_ARGUMENTS, closure_name(sc, sc->code), sc->cur_code)));
	      } 
	    else add_slot(sc, x, z); /* the rest arg */
	  }
	  sc->code = closure_body(sc->code);
	  if ((is_one_liner(sc->code)) &&
	      (!(sc->begin_hook))) /* perhaps this could be moved to the optimizer pass */
	    {
	      sc->code = car(sc->code);
	      goto EVAL_PAIR;
	    }
	  goto BEGIN;


	case T_CLOSURE_STAR:	                  /* -------- define* (lambda*) -------- */
	  { 
	    s7_pointer z;
	    if (sc->stack_end >= sc->stack_resize_trigger)
	      increase_stack_size(sc);

	    sc->envir = new_frame_in_env(sc, closure_environment(sc->code)); 

	    /* get default values, which may involve evaluation 
	     */
	    push_stack(sc, OP_LAMBDA_STAR_DEFAULT, sc->args, sc->code); /* op is just a placeholder (don't use OP_BARRIER here) */
	    sc->args = closure_args(sc->code);

	  LAMBDA_STAR_DEFAULT:
	    z = sc->args;
	    if (is_pair(z))
	      {
		/* bind all the args to something (default value or #f or maybe #undefined) */
		if (!((car(z) == sc->KEY_KEY) ||
		      (car(z) == sc->KEY_OPTIONAL) ||
		      (car(z) == sc->KEY_ALLOW_OTHER_KEYS)))
		  {
		    if (car(z) == sc->KEY_REST)
		      {
			add_slot(sc, cadr(z), sc->NIL);       /* set :rest arg to sc->NIL, not sc->F */
			sc->args = cdr(sc->args);
		      }
		    else
		      {
			if (is_pair(car(z)))                 /* (define* (hi (a mus-next)) a) */
			  {                                  /* or (define* (hi (a 'hi)) (list a (eq? a 'hi))) */
			    /* if default val is an expression, it needs to be evaluated in the definition environment
			     *   (let ((c 1)) (define* (a (b (+ c 1))) b) (set! c 2) (a)) -> 3
			     */
			    s7_pointer val;
			    val = cadar(z);
			    if (is_symbol(val))
			      add_slot(sc, caar(z), finder(sc, val));
			    else
			      {
				if (is_pair(val))
				  {
				    if ((car(val) == sc->QUOTE) ||
					(car(val) == sc->QUOTE_UNCHECKED))
				      add_slot(sc, caar(z), cadr(val));
				    else
				      {
					push_stack(sc, OP_LAMBDA_STAR_DEFAULT, sc->args, sc->code);
					sc->code = cadar(z);
					goto EVAL_PAIR;
				      }
				  }
				else add_slot(sc, caar(z), val);
			      }
			  }
			else add_slot(sc, car(z), sc->F);
		      }
		  }
		sc->args = cdr(sc->args);
		goto LAMBDA_STAR_DEFAULT;
	      }
	    else
	      {
		if (is_symbol(z))                      /* dotted (last) arg? -- make sure its name exists in the current environment */
		  add_slot(sc, z, sc->NIL); 
	      }
	    /* end of default arg evaluations */

	    pop_stack_no_op(sc);                          /* get original args and code back */
	    lambda_star_set_args(sc);                     /* load up current arg vals */
	    sc->code = closure_body(sc->code);            /* evaluate the function body */
	    if (is_one_liner(sc->code))
	      {
		sc->code = car(sc->code);
		goto EVAL_PAIR;
	      }
	    goto BEGIN;
	  }
		
	case T_CONTINUATION:	                  /* -------- continuation ("call-with-current-continuation") -------- */
	  call_with_current_continuation(sc);
	  goto START;

	case T_GOTO:	                          /* -------- goto ("call-with-exit") -------- */
	  call_with_exit(sc);
	  goto START;

	case T_HOOK:                              /* -------- hook -------- */
	  if (is_pair(hook_functions(sc->code)))
	    {
	      sc->code = hook_functions(sc->code);
	      push_stack(sc, OP_HOOK_APPLY, sc->args, cdr(sc->code));
	      sc->code = car(sc->code);
	      if (needs_copied_args(sc->code))
		sc->args = copy_list(sc, sc->args);
	      /* this copy is needed, unfortunately, but could it be done before the push, then skip the copy in OP_HOOK_APPLY? */
	      goto APPLY;
	    }
	  else sc->value = sc->UNSPECIFIED;
	  goto START;


	case T_C_OBJECT:                          /* -------- applicable (new-type) object -------- */
	  sc->value = (*(object_ref(sc->code)))(sc, sc->code, sc->args);
	  /* pws_apply, fallback_ref, or call_s_object_getter
	   *    pws_apply goes to op_apply, fallback_ref throws an error, call_s_object_getter calls s7_call
	   *    so it's not obvious how to avoid the jump.
	   */
	  goto START;


	case T_VECTOR:                            /* -------- vector as applicable object -------- */
	  /* sc->code is the vector, sc->args is the list of indices 
	   */
	  if ((is_null(cdr(sc->args))) &&       /* cdr(nil) -> #<unspecified> internally */
	      (s7_is_integer(car(sc->args))) &&
	      (!vector_is_multidimensional(sc->code)))
	    {
	      s7_Int index;
	      index = s7_integer(car(sc->args));
	      if ((index >= 0) &&
		  (index < vector_length(sc->code)))
		{
		  sc->value = vector_elements(sc->code)[index];
		  goto START;
		}
	    }

	  if (is_null(sc->args))                            /* (#2d((1 2) (3 4))) */
	    return(s7_wrong_number_of_args_error(sc, "not enough args for vector-ref: ~A", sc->args));

	  sc->value = vector_ref_1(sc, sc->code, sc->args);
	  goto START;


	case T_STRING:                            /* -------- string as applicable object -------- */
	  if (is_null(cdr(sc->args)))
	    {
	      if (s7_is_integer(car(sc->args)))
		{
		  int index;
		  index = (int)s7_integer(car(sc->args));
		  if ((index >= 0) &&
		      (index < (int)string_length(sc->code)))
		    {
		      sc->value = s7_make_character(sc, ((unsigned char *)string_value(sc->code))[index]);
		      goto START;
		    }
		}
	      sc->value = string_ref_1(sc, sc->code, car(sc->args));
	      goto START;
	    }
	  return(s7_error(sc, 
			  sc->WRONG_NUMBER_OF_ARGS, 
			  list_3(sc, (is_null(sc->args)) ? sc->NOT_ENOUGH_ARGUMENTS : sc->TOO_MANY_ARGUMENTS, sc->code, sc->args)));


	case T_PAIR:                              /* -------- list as applicable object -------- */
	  if (is_multiple_value(sc->code))                                  /* ((values 1 2 3) 0) */
	    {
	      /* car of values can be anything, so conjure up a new expression, and apply again */
	      sc->x = multiple_value(sc->code);                             /* ((values + 1 2) 3) */
	      sc->code = car(sc->x);
	      sc->args = s7_append(sc, cdr(sc->x), sc->args);
	      sc->x = sc->NIL;
	      goto APPLY;
	    }
	  if (is_null(cdr(sc->args)))
	    sc->value = list_ref_1(sc, sc->code, car(sc->args));            /* (L 1) */
	  else
	    {
	      if (is_null(sc->args))
		return(s7_wrong_number_of_args_error(sc, "not enough args for list-ref (via list as applicable object): ~A", sc->args));
	      sc->value = g_list_ref(sc, cons(sc, sc->code, sc->args)); /* (L 1 2) */
	    }
	  goto START;


	case T_HASH_TABLE:                        /* -------- hash-table as applicable object -------- */
	  if (is_null(cdr(sc->args)))
	    sc->value = s7_hash_table_ref(sc, sc->code, car(sc->args));
	  else
	    {
	      if (is_null(sc->args))
		return(s7_wrong_number_of_args_error(sc, "not enough args for hash-table-ref (via hash table as applicable object): ~A", sc->args));
	      sc->value = g_hash_table_ref(sc, cons(sc, sc->code, sc->args));
	    }
	  goto START;


	case T_ENVIRONMENT:
	  if (is_null(cdr(sc->args)))
	    {
	      if (is_symbol(car(sc->args)))
		sc->value = environment_ref(sc, sc->code, car(sc->args));
	      else return(s7_wrong_type_arg_error(sc, "environment application", 1, car(sc->args), "a symbol"));
	    }
	  else return(s7_wrong_number_of_args_error(sc, "environment as applicable object takes one argument: ~A", sc->args));
	  goto START;


	case T_SYNTAX:                            /* -------- syntactic keyword as applicable object -------- */
	  sc->op = (opcode_t)syntax_opcode(sc->code);                       /* (apply begin '((define x 3) (+ x 2))) */
	  sc->code = sc->args;
	  goto START_WITHOUT_POP_STACK;


	default:
	  return(apply_error(sc, sc->code, sc->args));
	}
      /* ---------------- end APPLY/EVAL ---------------- */


    case OP_LAMBDA_STAR_DEFAULT:
      /* sc->args is the current closure arg list position, sc->value is the default expression's value
       */
      add_slot(sc, caar(sc->args), sc->value);
      sc->args = cdr(sc->args);
      goto LAMBDA_STAR_DEFAULT;


    case OP_QUOTE:
      check_quote(sc);

    case OP_QUOTE_UNCHECKED:
      sc->value = car(sc->code);
      goto START;


    case OP_DEFINE_STAR:
      check_define(sc);

    case OP_DEFINE_STAR_UNCHECKED:
      {
	s7_pointer x;
	NEW_CELL(sc, x);
	closure_args(x) = cdar(sc->code);
	closure_body(x) = cdr(sc->code);
	closure_environment(x) = sc->envir;
	set_type(x, T_CLOSURE_STAR | T_PROCEDURE);
	sc->value = x;
	sc->code = caar(sc->code);
	goto DEFINE1;
      }

      
    case OP_DEFINE_CONSTANT1:
      /* define-constant -> OP_DEFINE_CONSTANT -> OP_DEFINE..1, then back to here */
      /*   at this point, sc->value is the symbol that we want to be immutable, sc->code is the original pair */

      sc->x = find_symbol(sc, sc->value);
      if (is_slot(sc->x))
	set_immutable(slot_symbol(sc->x));

      /* if we have define-constant of a symbol that is purely local, then reload the file containing that code,
       *   the symbol itself is immutable, but the original local slot has vanished, so we can't compare (below)
       *   against the old value.  Should define-constant always be global?  Otherwise we need to keep track of
       *   exactly where it was defined!
       *
       * but the following code is confusing because the definition leaks out -- not sure what to do here.
       *   perhaps a separate global table of these?  Also, redefinition of a constant local closure won't
       *   work unless we check equality of the functions!
       *
       * another question: (define-constant c (list 1 2 3)) but the list contents are settable
       */
      goto START;


    case OP_DEFINE_CONSTANT:
      push_stack(sc, OP_DEFINE_CONSTANT1, sc->NIL, sc->code);

      
    case OP_DEFINE:
      check_define(sc);

    case OP_DEFINE_UNCHECKED:
      if (!is_pair(car(sc->code)))
	{
	  s7_pointer x;
	  x = car(sc->code);
	  sc->code = cadr(sc->code);
	  if (is_pair(sc->code))
	    {
	      push_stack(sc, OP_DEFINE1, sc->NIL, x);
	      goto EVAL_PAIR;
	    }
	  
	  if (is_symbol(sc->code))
	    sc->value = SYMBOL_TO_VALUE(sc, sc->code);
	  else sc->value = sc->code;
	  sc->code = x;
	  /* fall through */
	}
      else
	{
	  s7_pointer x;
	  /* a closure.  If we called this same code earlier (a local define), the only thing
	   *   that is new here is the environment -- we can't blithely save the closure object
	   *   in fcdr somewhere, and pick it up the next time around (since call/cc might take
	   *   us back to the previous case).
	   */
	  MAKE_CLOSURE(sc, x, cdar(sc->code), cdr(sc->code), sc->envir);
	  sc->value = x;
	  sc->code = caar(sc->code);
	  /* fall through */
	}

    DEFINE1:
    case OP_DEFINE1:
      /* sc->code is the symbol being defined, sc->value is its value
       *   if sc->value is a closure, car is of the form ((args...) body...)
       *   so the doc string if any is (cadr (car value))
       *   and the arg list gives the number of optional args up to the dot
       */

      /* it's not possible to expand and replace macros at this point without evaluating
       *   the body.  Just as examples, say we have a macro "mac", 
       *   (define (hi) (call/cc (lambda (mac) (mac 1))))
       *   (define (hi) (quote (mac 1))) or macroexpand etc
       *   (define (hi mac) (mac 1)) assuming mac here is a function passed as an arg,
       * etc...  
       */

      /* if we're defining a function, add its symbol to the new function's environment under the name __func__ */

      /* the immutable constant check needs to wait until we have the actual new value because
       *   we want to ignore the rebinding (not raise an error) if it is the existing value.
       *   This happens when we reload a file that has a define-constant call.
       */
      if (is_immutable(sc->code))                                        /* (define pi 3) or (define (pi a) a) */
	{
	  s7_pointer x;
	  if (!is_symbol(sc->code))                                      /* (define "pi" 3) ? */
	    return(eval_error_with_name(sc, "~A: ~S is immutable", sc->code));

	  x = find_symbol(sc, sc->code);
	  if ((!is_slot(x)) ||
	      (!s7_is_equal(sc, sc->value, slot_value(x))))              /* if value is unchanged, just ignore this definition */
	    {
	      /* fprintf(stderr, "im: %s %s %s\n", DISPLAY(sc->code), DISPLAY(sc->value), DISPLAY(x)); */
	      return(eval_error_with_name(sc, "~A: ~S is immutable", sc->code));
	    }
	}
      if (symbol_has_accessor(sc->code))
	{
	  sc->value = bind_accessed_symbol(sc, OP_DEFINE_WITH_ACCESSOR, sc->code, sc->value);
	  if (sc->value == sc->NO_VALUE)
	    goto APPLY;
	  /* if all goes well, OP_DEFINE_WITH_ACCESSOR will jump to DEFINE2 */
	}

    DEFINE2:
      {
	s7_pointer val;
	val = sc->value;
	if ((is_closure(val)) || 
	    (is_closure_star(val)))
	  {
	    /* we can get here from let: (define (outer a) (let () (define (inner b) (+ a b)) (inner a)))
	     *   but the port info is not relevant here, so restrict the __func__ list making to top-level
	     *   cases (via sc->envir == sc->NIL).
	     */
	    if ((!is_environment(sc->envir)) &&
		(port_filename(sc->input_port)) &&                             /* add (__func__ (name file line)) to current env */
		(port_file(sc->input_port) != stdin))
	      sc->x = list_3(sc, sc->code,
			     file_names[port_file_number(sc->input_port)],
			     make_integer(sc, port_line_number(sc->input_port)));
	    else sc->x = sc->code;           /* fallback on (__func__ name) */
	    
	    {
	      s7_pointer x, y;
	      
	      NEW_CELL(sc, x);
	      frame_id(x) = ++frame_number;
	      next_environment(x) = closure_environment(val);
	      set_type(x, T_ENVIRONMENT | T_IMMUTABLE);
	      closure_environment(val) = x;
	      
	      NEW_CELL_NO_CHECK(sc, y);
	      slot_symbol(y) = sc->__FUNC__;
	      symbol_id(slot_symbol(y)) = frame_id(x);
	      local_slot(slot_symbol(y)) = y;
	      slot_set_value(y, sc->x);
	      set_type(y, T_SLOT | T_IMMUTABLE);
	      next_slot(y) = sc->NIL;
	      environment_slots(x) = y;
	      
	      if (is_safe_closure(closure_body(val)))
		{
		  s7_pointer arg;
		  closure_environment(val) = new_frame_in_env(sc, closure_environment(val));
		  for (arg = closure_args(val); is_pair(arg); arg = cdr(arg))
		    s7_make_slot(sc, closure_environment(val), car(arg), sc->NIL);
		  environment_slots(closure_environment(val)) = reverse_slots(sc, environment_slots(closure_environment(val)));
		}
	      
	      /* add the newly defined thing to the current environment */
	      if (is_environment(sc->envir))
		{
		  ADD_SLOT(sc->envir, sc->code, val);
		  set_local(sc->code);
		}
	      else s7_make_slot(sc, sc->envir, sc->code, val);
	    }
	  }
	else
	  {
	    if (is_pws(val))
	      {
		s7_pws_t *f = (s7_pws_t *)s7_object_value(val);
		if (!(f->name))
		  f->name = copy_string(symbol_name(sc->code));
	      }
	    
	    /* add the newly defined thing to the current environment */
	    sc->x = find_local_symbol(sc, sc->envir, sc->code);
	    if (is_slot(sc->x))
	      slot_set_value(sc->x, val); 
	    else s7_make_slot(sc, sc->envir, sc->code, val);
	  }
      }

      sc->value = sc->code;
      goto START;


    case OP_DEFINE_WITH_ACCESSOR:
      if (sc->value == sc->ERROR) /* backwards compatibility... */
	return(s7_error(sc, sc->ERROR,
			list_3(sc, make_protected_string(sc, "can't define ~S to ~S"), car(sc->args), cadr(sc->args))));
      goto DEFINE2;

      

      /* -------------------------------- SET! -------------------------------- */

    case OP_SET_PAIR:
      /* car(sc->code) is a pair, caar(code) is the object with a setter, it has one (safe) argument, and one safe value to set
       *   (set! (str i) #\a) in a function (both inner things need to be symbols to get here)
       *   the inner list is a proper list, with no embedded list at car.
       */
      {
	s7_pointer obj, arg, value, code;

	code = sc->code;
	if (is_symbol(caar(code)))
	  obj = s7_symbol_value(sc, caar(code));                     /* this can't be simply slot_value */
	else obj = caar(code);                                       /* might be the pws function itself */

	/* pairs won't happen here in either case -- see check_set */
	if (is_symbol(cadar(code)))
	  arg = finder(sc, cadar(code));
	else arg = cadar(code);

	if (is_symbol(cadr(code)))
	  value = finder(sc, cadr(code));
	else value = cadr(code);

	switch (type(obj))
	    {
	    case T_C_OBJECT:
	      car(sc->T2_1) = arg;
	      car(sc->T2_2) = value;
	      sc->value = (*(object_set(obj)))(sc, obj, sc->T2_1);
	      break;
   
	    case T_VECTOR:
	      car(sc->T3_1) = obj;
	      car(sc->T3_2) = arg;
	      car(sc->T3_3) = value;
	      sc->value = g_vector_set(sc, sc->T3_1);
	      break;
	      
	    case T_STRING:
	      car(sc->T3_1) = obj;
	      car(sc->T3_2) = arg;
	      car(sc->T3_3) = value;
	      sc->value = g_string_set(sc, sc->T3_1);
	      break;

	    case T_PAIR:
	      car(sc->T3_1) = obj;
	      car(sc->T3_2) = arg;
	      car(sc->T3_3) = value;
	      sc->value = g_list_set(sc, sc->T3_1);
	      break;

	    case T_HASH_TABLE:
	      car(sc->T3_1) = obj;
	      car(sc->T3_2) = arg;
	      car(sc->T3_3) = value;
	      sc->value = g_hash_table_set(sc, sc->T3_1);
	      break;

	    case T_ENVIRONMENT:
	      /* (set! (env sym) val) */
	      sc->value = environment_set(sc, obj, arg, value);
	      break;

	    case T_C_OPT_ARGS_FUNCTION:
	    case T_C_RST_ARGS_FUNCTION:
	    case T_C_ANY_ARGS_FUNCTION:                       /* (let ((lst (list 1 2))) (set! (list-ref lst 1) 2) lst) */
	    case T_C_FUNCTION:
	      if (is_procedure(c_function_setter(obj)))
		{
		  car(sc->T2_1) = arg;
		  car(sc->T2_2) = value;
		  sc->value = c_function_call(c_function_setter(obj))(sc, sc->T2_1);
		}
	      else return(eval_error(sc, "no generalized set for ~A", obj));
	      break;

	    default:                                         /* (set! (1 2) 3) */
	      return(eval_error(sc, "no generalized set for ~A", obj));
	    }
	goto START;
      }
      

#if WITH_OPTIMIZATION
    case OP_INCREMENT_1:
      if (indirect_c_function_is_ok(sc, cadr(sc->code)))
	{
	  s7_pointer y;
	  y = find_symbol(sc, car(sc->code));
	  if (is_slot(y))
	    {
	      s7_pointer val;	    
	      val = slot_value(y);
	      switch (type(val))
		{
		case T_INTEGER:
		  /* sc->value = make_integer(sc, integer(val) + 1); */
		  {
		    s7_Int n;
		    n = integer(val) + 1;
		    if ((n & ~(NUM_SMALL_INTS - 1)) == 0)
		      sc->value = small_int(n);
		    else
		      {
			s7_pointer x;
			NEW_CELL(sc, x);
			set_type(x, T_INTEGER);
			integer(x) = n;
			sc->value = x;
		      }
		    slot_set_value(y, sc->value);
		    /* this can't be optimized to treat y's value as a mutable integer 
		     * also, we have to set sc->value, since s7.html says set! returns the value.
		     */
		    goto START; 
		  }
		  
		case T_REAL:
		  sc->value = s7_make_real(sc, real(val) + 1.0);
		  slot_set_value(y, sc->value);
		  goto START;
		  
		default:
		  break;
		}
	    }
	}
      /* else use standard set */
      push_stack_no_args(sc, OP_SET1, car(sc->code));
      sc->code = cadr(sc->code);
      goto EVAL_PAIR;


    case OP_DECREMENT_1:
      if ((optimize_data(cadr(sc->code)) != OP_SAFE_C_SC) ||
	  (indirect_c_function_is_ok(sc, cadr(sc->code))))
	{
	  s7_pointer y;
	  y = find_symbol(sc, car(sc->code));
	  if (is_slot(y))
	    {
	      s7_pointer val;	    
	      val = slot_value(y);
	      switch (type(val))
		{
		case T_INTEGER:
		  sc->value = make_integer(sc, integer(val) - 1);
		  slot_set_value(y, sc->value);
		  goto START;
		  
		case T_REAL:
		  sc->value = s7_make_real(sc, real(val) - 1.0);
		  slot_set_value(y, sc->value);
		  goto START;
		  
		default:
		  break;
		}
	    }
	}
      /* else use standard set */
      push_stack_no_args(sc, OP_SET1, car(sc->code)); 
      sc->code = cadr(sc->code);
      goto EVAL_PAIR;


    case OP_SET_CDR:
      if ((optimize_data(cadr(sc->code)) != OP_SAFE_C_S) ||
	  (indirect_c_function_is_ok(sc, cadr(sc->code))))
	{
	  s7_pointer y;
	  y = find_symbol(sc, car(sc->code));
	  if (is_slot(y))
	    {
	      s7_pointer val;	    
	      val = slot_value(y);
	      if (!is_pair(val))
		return(s7_wrong_type_arg_error(sc, "cdr", 0, val, "a pair"));
	      sc->value = cdr(val);                     /* set! returns the value */
	      slot_set_value(y, sc->value);
	      goto START;
	    }
	}
      /* else use standard set */
      push_stack_no_args(sc, OP_SET1, car(sc->code)); 
      sc->code = cadr(sc->code);
      goto EVAL_PAIR;


    case OP_SET_CONS:
      if ((optimize_data(cadr(sc->code)) != OP_SAFE_C_SS) || /* it's safe if hop_safe_c_ss, so if op_safe_c_ss we check? */
	  (indirect_c_function_is_ok(sc, cadr(sc->code))))
	{
	  s7_pointer y;
	  y = find_symbol(sc, car(sc->code)); 
	  if (is_slot(y))
	    {
	      sc->value = cons(sc, finder(sc, fcdr(sc->code)), slot_value(y));
	      slot_set_value(y, sc->value);
	      goto START;
	    }
	}
      /* else use standard set */
      push_stack_no_args(sc, OP_SET1, car(sc->code)); 
      sc->code = cadr(sc->code);
      goto EVAL_PAIR;


    case OP_SET_SYMBOL_R:
      sc->temp4 = sc->code;
      car(sc->T1_1) = finder(sc, fcdr(sc->code));
      sc->value = c_call(cadr(sc->code))(sc, sc->T1_1);
      /*
       * I think this is supposed to be a safe c function (is_h_safe_c_s above), but
       *   it's possible for the call to change sc->code?  
       *   in snd-test: 
       *           (set! snd-output (open-sound output-1)) 
       *   becomes (set! (hook-functions *error-hook*) '())
       *   because open-sound is calling itself safe, but it can call all kinds of hooks and whatnot,
       *   and these can hit errors, setting *error-hook*, fallling into s7_call and so on.
       * (this is independent of c_function_call: it happens with optimization off, so set below
       * also needs to be protected, I guess).
       */
      if (sc->code != sc->temp4)
	sc->code = sc->temp4;

      sc->code = car(sc->code);
      goto SET1;
#endif      

    case OP_SET_SYMBOL_C:
      sc->value = cadr(sc->code);
      sc->code = car(sc->code);
      goto SET1;

    case OP_SET_SYMBOL_Q:
      sc->value = cadr(cadr(sc->code));
      sc->code = car(sc->code);
      goto SET1;

    case OP_SET_SYMBOL_S:
      {
	s7_pointer settee;
	settee = car(sc->code);
	sc->value = finder(sc, cadr(sc->code));
	sc->code = settee;
	goto SET1;
      }

    case OP_SET_SYMBOL_P:
      push_stack_no_args(sc, OP_SET1, car(sc->code)); 
      sc->code = cadr(sc->code);
      /* s7: mostly not opt
       * lg: closure_sss: 176901 h_safe_c_opssq_s: 12762 h_safe_c_c: 4882
       * in: h_safe_c_ps: 175863 h_safe_c_cs: 26220 but not opt: 116946
       * al: h_safe_closure_ss: 26487 h_safe_c_opssq_s: 22591
       */
      goto EVAL_PAIR; /* lg: opt */


    case OP_SET2:
      if (is_pair(sc->value))
	{
	  /* (let ((L '((1 2 3)))) (set! ((L 0) 1) 32) L)
	   * (let ((L '(((1 2 3))))) (set! ((L 0) 0 1) 32) L)
	   * any deeper nesting was handled already by the first eval
	   *   set! looks at its first argument, if it's a symbol, it sets the associated value,
	   *   if it's a list, it looks at the car of that list to decide which setter to call,
	   *   if it's a list of lists, it passes the embedded lists to eval, then looks at the
	   *   car of the result.  This means that we can do crazy things like:
	   *   (let ((x '(1)) (y '(2))) (set! ((if #t x y) 0) 32) x)
	   *
	   * the other args need to be evaluated (but not the list as if it were code):
	   *   (let ((L '((1 2 3))) (index 1)) (set! ((L 0) index) 32) L)
	   */

	  if (!is_proper_list(sc, sc->args))                                 /* (set! ('(1 2) 1 . 2) 1) */
	    eval_error(sc, "set! target arguments are an improper list: ~A", sc->args);

	  /* in all of these cases, we might need to GC protect the temporary lists */

	  if (is_multiple_value(sc->value))
	    sc->code = cons(sc, sc->SET, s7_append(sc, multiple_value(sc->value), s7_append(sc, sc->args, sc->code))); /* drop into OP_SET */
	  else
	    {
	      if (sc->args != sc->NIL)
		{
		  push_op_stack(sc, sc->LIST_SET);
		  push_stack(sc, OP_EVAL_ARGS1, list_1(sc, sc->value), s7_append(sc, cdr(sc->args), sc->code));
		  sc->code = car(sc->args);
		}
	      else return(eval_error(sc, "list set!: not enough arguments: ~S", sc->code));
	      goto EVAL;
	    }
	  /* timings: old 3.25, new 2.34 (1.93 direct), (12.2): 1.34
	   *    (time (let ((L '((1 2 3)))) (do ((i 0 (+ i 1))) ((= i 10000000)) (set! ((L 0) 1) i)) L))
	   */
	}
      else
	{
	  if (s7_is_vector(sc->value))
	    {
	      /* (let ((L '#(#(1 2 3) #(4 5 6)))) (set! ((L 1) 0) 32) L)
	       * bad case when args is nil: (let ((L '#(#(1 2 3) #(4 5 6)))) (set! ((L 1)) 32) L)
	       */
	      if (sc->args != sc->NIL)
		{
		  push_op_stack(sc, sc->VECTOR_SET);
		  push_stack(sc, OP_EVAL_ARGS1, list_1(sc, sc->value), s7_append(sc, cdr(sc->args), sc->code));
		  sc->code = car(sc->args);
		}
	      else return(eval_error(sc, "vector set!: not enough arguments: ~S", sc->code));
	      goto EVAL;
	    }
	  sc->code = cons_unchecked(sc, cons(sc, sc->value, sc->args), sc->code);
	}
      /* fall through */


    case OP_SET:                                                             /* entry for set! */
      check_set(sc);

    case OP_SET_UNCHECKED:
      if (is_pair(car(sc->code)))                                              /* has accessor */
	{
	  if (is_pair(caar(sc->code)))
	    {
	      push_stack(sc, OP_SET2, cdar(sc->code), cdr(sc->code));
	      sc->code = caar(sc->code);
	      goto EVAL_PAIR;
	    }
	  
	  if (is_symbol(caar(sc->code)))
	    sc->x = s7_symbol_value(sc, caar(sc->code));                     /* this can't be simply slot_value */
	  else sc->x = caar(sc->code);                                       /* might be the pws function itself */
	  
	  /* code here is the accessor and the value without the "set!": ((window-width) 800) */
	  /*    (set! (hi 0) (* 2 3)) -> ((hi 0) (* 2 3)) */

	  /* for these kinds of objects, some Schemes restrict set!
	   * (list-set! '(1 2 3) 1 32) is accepted but does it make sense?
	   * (set-car! '(1 . 2) 32)
	   * (string-set! "hiho" 1 #\z)
	   * (vector-set! #(1 2 3) 1 32)
	   * (let ((x (lambda () "hiho"))) (string-set! (x) 1 #\a))
	   * (let ((x (lambda () #(1 2 3)))) (vector-set! (x) 1 32))
	   * (let ((str "hiho")) (string-set! str 1 #\x) str)
	   * (let ((v #(1 2 3))) (vector-set! v 1 32) v)
	   * (let ((x (lambda () "hiho"))) (string-set! (x) 1 #\x) (x))
	   *
	   *   It seems weird that we can reach into both the function body, and its closure:
	   *   (let ((xx (let ((x '(1 2 3))) (lambda () x)))) (list-set! (xx) 1 32) (xx)) -> '(1 32 3)
	   *
	   * (let* ((x '(1 2)) (y (list x)) (z (car y))) (list-set! z 1 32) (list x y z))
	   * ((1 32) ((1 32)) (1 32))
	   *
	   * (string-set! (symbol->string 'symbol->string) 1 #\X) -> error currently also in Guile "string is read-only"
	   * (setf (elt (symbol-name 'xyz) 1) #\X) -> error in CL "read-only string"
	   */

	  switch (type(sc->x))
	    {
	    case T_C_OBJECT:
	      push_op_stack(sc, sc->OBJECT_SET);
	      if (is_pair(cdar(sc->code)))
		{
		  push_stack(sc, OP_EVAL_ARGS1, list_1(sc, sc->x), s7_append(sc, cddar(sc->code), cdr(sc->code)));
		  sc->code = cadar(sc->code);
		}
	      else /* (set! (window-width) 800) -> ((window-width) 800) so cdar(sc->code) is nil */
		{
		  push_stack(sc, OP_EVAL_ARGS1, list_1(sc, sc->x), cddr(sc->code));
		  sc->code = cadr(sc->code);
		}
	      break;

	    case T_VECTOR:
	      /* sc->x is the vector, sc->code is expr without the set! */
	      /*  args have not been evaluated! */
	      /* sc->code = cons(sc, sc->VECTOR_SET, s7_append(sc, car(sc->code), cdr(sc->code))); */

	      if (cdar(sc->code) != sc->NIL)
		{
		  push_op_stack(sc, sc->VECTOR_SET);
		  push_stack(sc, OP_EVAL_ARGS1, list_1(sc, sc->x), s7_append(sc, cddar(sc->code), cdr(sc->code)));
		  sc->code = cadar(sc->code);
		}
	      else sc->code = cons(sc, sc->VECTOR_SET, s7_append(sc, car(sc->code), cdr(sc->code)));
	      break;
	      
	    case T_STRING:
	      /* sc->code = cons(sc, sc->STRING_SET, s7_append(sc, car(sc->code), cdr(sc->code))); 
	       *  it is slower to break out the normal case and handle it here.
	       *
	       * here only one index makes sense, and it is required, so 
	       *   (set! ("str") #\a), (set! ("str" . 1) #\a) and (set! ("str" 1 2) #\a)
	       *   are all errors (but see below!).
	       *
	       * (time (let ((str "123")) (do ((i 0 (+ i 1))) ((= i 10000000)) (set! (str 1) #\1))))
	       *   (12.3): 0.618
	       *
	       * it costs a lot more to check here and avoid s7_append, but we can't avoid the check!
	       *   (set! ("hi") 1)
	       * code comes in as '(("hi") 1), so cdar(code) is nil, cddar is #<unspecified>, cadar is also #<unspecified>, cdr is '(1)
	       *   so eval gets: improper list of arguments: ("hi" #<unspecified> 1 #<unspecified> . #<unspecified>)
	       *   because (below) code is #<unspecified> and the append business returns '(1 #<unspecified> . #<unspecified>)
	       */

	      if (cdar(sc->code) != sc->NIL)
		{
		  push_stack(sc, OP_EVAL_ARGS1, list_1(sc, sc->x), sc->y = s7_append(sc, cddar(sc->code), cdr(sc->code)));
		  push_op_stack(sc, sc->STRING_SET);
		  sc->code = cadar(sc->code);
		}
	      else sc->code = cons(sc, sc->STRING_SET, s7_append(sc, car(sc->code), cdr(sc->code))); 
	      break;

	    case T_PAIR:
	      /* code: ((lst 1) 32) from (let ((lst '(1 2 3))) (set! (lst 1) 32))
	       * old form:
	       *    sc->code = cons(sc, sc->LIST_SET, s7_append(sc, car(sc->code), cdr(sc->code))); 
	       *    old: 2.32, new: 1.77 (1.50 direct), (12.2): 1.07
	       *    (time (let ((lst '(1 2 3))) (do ((i 0 (+ i 1))) ((= i 10000000)) (set! (lst 1) 32))))
	       */
	      if (cdar(sc->code) != sc->NIL)
		{
		  push_op_stack(sc, sc->LIST_SET);
		  push_stack(sc, OP_EVAL_ARGS1, list_1(sc, sc->x), s7_append(sc, cddar(sc->code), cdr(sc->code)));
		  sc->code = cadar(sc->code);
		}
	      else sc->code = cons(sc, sc->LIST_SET, s7_append(sc, car(sc->code), cdr(sc->code))); 
	      break;

	    case T_HASH_TABLE:
	      if (cdar(sc->code) != sc->NIL)
		{
		  push_op_stack(sc, sc->HASH_TABLE_SET);
		  push_stack(sc, OP_EVAL_ARGS1, list_1(sc, sc->x), s7_append(sc, cddar(sc->code), cdr(sc->code)));
		  sc->code = cadar(sc->code);
		}
	      else sc->code = cons(sc, sc->HASH_TABLE_SET, s7_append(sc, car(sc->code), cdr(sc->code)));
	      break;


	    case T_ENVIRONMENT:
	      sc->code = cons(sc, sc->ENVIRONMENT_SET, s7_append(sc, car(sc->code), cdr(sc->code)));
	      break;


	    case T_C_OPT_ARGS_FUNCTION:
	    case T_C_RST_ARGS_FUNCTION:
	    case T_C_ANY_ARGS_FUNCTION:                       /* (let ((lst (list 1 2))) (set! (list-ref lst 0) 2) lst) */
	    case T_C_FUNCTION:
	      /* perhaps it has a setter */
	      
	      if (is_procedure(c_function_setter(sc->x)))
		{
		  /* sc->code = cons(sc, c_function_setter(sc->x), s7_append(sc, cdar(sc->code), cdr(sc->code))); */
		  if (is_pair(cdar(sc->code)))
		    {
		      push_op_stack(sc, c_function_setter(sc->x));
		      push_stack(sc, OP_EVAL_ARGS1, sc->NIL, s7_append(sc, cddar(sc->code), cdr(sc->code)));
		      sc->code = cadar(sc->code);
		    }
		  else
		    {
		      if ((is_null(cddr(sc->code))) &&
			  (!is_pair(cadr(sc->code))))
			{
			  if (is_symbol(cadr(sc->code)))
			    car(sc->T1_1) = finder(sc, cadr(sc->code));
			  else car(sc->T1_1) = cadr(sc->code);
			  sc->args = sc->T1_1;
			  sc->code = c_function_setter(sc->x);
			  goto APPLY; /* check arg num etc */
			}
		      push_op_stack(sc, c_function_setter(sc->x));
		      push_stack(sc, OP_EVAL_ARGS1, sc->NIL, cddr(sc->code));
		      sc->code = cadr(sc->code);
		    }
		}
	      else return(eval_error(sc, "no generalized set for ~A", caar(sc->code)));
	      break;

	    default:                                         /* (set! (1 2) 3) */
	      return(eval_error(sc, "no generalized set for ~A", caar(sc->code)));
	    }
	  /* sc->x = sc->NIL; */
	  goto EVAL;
	}

    case OP_SET_NORMAL:
      {
	s7_pointer x;
	x = cadr(sc->code);
	if (is_pair(x))
	  {
	    push_stack_no_args(sc, OP_SET1, car(sc->code));
	    sc->code = x;
	    goto EVAL_PAIR;
	  }

	if (is_symbol(x))
	  sc->value = finder(sc, x);
	else sc->value = x;
	sc->code = car(sc->code);
      }

      
    SET1:
    case OP_SET1:  
      /* if unbound variable hook here, we need the binding, not the current value */
      if (is_global(sc->code))
	sc->y = global_slot(sc->code);
      else sc->y = find_symbol(sc, sc->code);
      if (is_slot(sc->y)) 
	{
	  if (symbol_has_accessor(sc->code))
	    {
	      s7_pointer func;
	      func = cadr(s7_gc_protected_at(sc, symbol_accessor(sc->code)));
	      if (is_procedure(func))
		{
		  if (is_c_function(func))
		    {
		      car(sc->T2_1) = sc->code;
		      car(sc->T2_2) = sc->value;
		      sc->value = c_function_call(func)(sc, sc->T2_1);
		      if (sc->value == sc->ERROR) /* backwards compatibility... */
			return(s7_error(sc, sc->ERROR,
					list_3(sc, make_protected_string(sc, "can't set ~S to ~S"), car(sc->T2_1), car(sc->T2_2))));
		    }
		  else
		    {
		      sc->args = list_2(sc, sc->code, sc->value);
		      push_stack(sc, OP_SET_WITH_ACCESSOR, sc->args, sc->y); /* op, args, code */
		      sc->code = func;
		      goto APPLY;
		    }
		}
	    }
	  else
	    {
	      if (is_syntax(slot_value(sc->y)))
		return(eval_error(sc, "can't set! ~A", sc->code));
	    }
	  slot_set_value(sc->y, sc->value); 
	  goto START;
	}
      return(eval_error(sc, "set! ~A: unbound variable", sc->code));


    case OP_SET_WITH_ACCESSOR:
      if (sc->value == sc->ERROR) /* backwards compatibility... */
	return(s7_error(sc, sc->ERROR,
			list_3(sc, make_protected_string(sc, "can't set ~S to ~S"), car(sc->args), cadr(sc->args))));
      slot_set_value(sc->code, sc->value); 
      goto START;
      


      /* -------------------------------- IF -------------------------------- */

    case OP_IF:
      check_if(sc);
      push_stack_no_args(sc, OP_IF1, cdr(sc->code));
      sc->code = car(sc->code);
      goto EVAL;

    case OP_IF1:
      if (is_true(sc, sc->value))
	sc->code = car(sc->code);
      else sc->code = cadr(sc->code);   /* even pre-optimization, (if #f #f) ==> #<unspecified> because car(sc->NIL) = sc->UNSPECIFIED */

      if (is_pair(sc->code))
	goto EVAL_PAIR;

      if (is_symbol(sc->code))
	sc->value = finder(sc, sc->code);
      else sc->value = sc->code;
      goto START;
      

    case OP_IF_P_P_P:
      push_stack_no_args(sc, OP_IF_PPP, cdr(sc->code));
      sc->code = car(sc->code);
      goto EVAL_PAIR;

#if WITH_OPTIMIZATION
    case OP_IF_O_P_P:
      push_stack_no_args(sc, OP_IF_PPP, cdr(sc->code));
      sc->code = car(sc->code);
      if (!is_optimized(sc->code))
	goto EVAL_PAIR;
      /* lg: h_safe_c_s_opsq: 30810 h_safe_c_opsq: 7814
       * s7: h_safe_c_cs: 6504
       * in: h_safe_c_sp: 37136  h_safe_c_opssq: 37136
       * al: h_safe_c_opssq: 63608 safe_c_s: 55105 safe_c_ss: 44838
       */
      goto OPT_EVAL;
#endif

    case OP_IF_B_P_P:
      {
	s7_pointer code;
	code = sc->code;
	push_stack_no_args(sc, OP_IF_PPP, cdr(code));
	code = car(code);
	sc->op = (opcode_t)syntax_opcode(car(code));
	/* lg: OP_SAFER_OR_C: 27195
	 * in: OP_OR_P: 231353, OP_AND_P: 123570
	 * al: OP_SAFER_OR_CEQ: 4460055, OP_OR_P: 220550
	 */
	sc->code = cdr(code);
	goto START_WITHOUT_POP_STACK;
      }


    case OP_IF_ANDP_P_P:
      push_stack_no_args(sc, OP_IF_PPP, cdr(sc->code));
      sc->code = cdar(sc->code);
      goto AND_P;


    case OP_IF_ORP_P_P:
      push_stack_no_args(sc, OP_IF_PPP, cdr(sc->code));
      sc->code = cdar(sc->code);
      goto OR_P;


#if WITH_OPTIMIZATION
    case OP_IF_ORCEQ_P_P:
      {
	s7_pointer sym, code;
	/* (define (hi c) (if (or (char=? c #\a) (char=? c #\b)) (string c) (string #\y)))
	 * sc->code: ((or (char=? c #\a) (char=? c #\b)) (string c) (string #\y))
	 */
	code = cdar(sc->code);           /* the sequence of char=? in the or */
	sym = finder(sc, cadar(code));   /* the symbol */
	do {
	  if (sym == fcdr(code))
	    {
	      sc->code = cadr(sc->code);
	      goto EVAL_PAIR;
	    }
	  code = cdr(code);
	} while (is_pair(code));
	  
	if (!s7_is_character(sym))
	  s7_wrong_type_arg_error(sc, "char=?", 1, cadar(code), "a character");
	sc->code = fcdr(sc->code);  /* caddr */
	goto EVAL_PAIR; /* all: syn */
      }
#endif


    case OP_IF_PPP:
      if (is_true(sc, sc->value))
	sc->code = car(sc->code);
      else sc->code = cadr(sc->code);
      goto EVAL_PAIR; /* all: syn */
	

    case OP_IF_P_P:
      push_stack_no_args(sc, OP_IF_PP, cadr(sc->code));
      sc->code = car(sc->code);
      goto EVAL_PAIR;

#if WITH_OPTIMIZATION
    case OP_IF_O_P:
      /* lg: [h_safe_c_opssq: 27603] h_safe_c_s_opcq: 22273
       * s7: [h_safe_c_opssq: 23662]
       * al: h_safe_c_opsq: 14578
       * sn: h_safe_c_s_opsq: 2752660
       * in: h_safe_c_opssq_s: 2153849 (string=? (vector-ref...) target)
       */
      push_stack_no_args(sc, OP_IF_PP, cadr(sc->code));
      sc->code = car(sc->code);
      if (!is_optimized(sc->code))
	goto EVAL_PAIR;
      goto OPT_EVAL;


    case OP_SAFE_IF_ANDX_P:
      {
	s7_pointer p;
	for (p = cdar(sc->code); is_pair(p); p = cdr(p))
	  {
	    s7_pointer val, code, args;
	    code = car(p);
	    switch (optimize_data(code))
	      {
	      case HOP_SAFE_C_C:
		sc->value = c_call(code)(sc, cdr(code));
		break;
		
	      case HOP_SAFE_C_S:
		car(sc->T1_1) = finder(sc, cadr(code));
		sc->value = c_call(code)(sc, sc->T1_1);
		break;
		
	      case HOP_SAFE_IS_PAIR_S:
		sc->value = make_boolean(sc, is_pair(finder(sc, cadr(code))));
		break;
		
	      case HOP_SAFE_IS_SYMBOL_S:
		sc->value = make_boolean(sc, is_symbol(finder(sc, cadr(code))));
		break;
		
	      case HOP_SAFE_NOT_S:
		sc->value = make_boolean(sc, is_false(sc, finder(sc, cadr(code))));
		break;
		
	      case HOP_SAFE_C_SC:
		args = cdr(code);
		car(sc->T2_1) = finder(sc, car(args));
		car(sc->T2_2) = cadr(args);
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_CS:
		args = cdr(code);
		car(sc->T2_2) = finder(sc, cadr(args));
		car(sc->T2_1) = car(args);
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_SS:
		args = cdr(code);
		val = finder(sc, car(args));
		car(sc->T2_2) = finder(sc, cadr(args));
		car(sc->T2_1) = val;
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_SQ:
		args = cdr(code);
		car(sc->T2_1) = finder(sc, car(args));
		car(sc->T2_2) = cadr(cadr(args));
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_opSq:
		args = cadr(code);
		car(sc->T1_1) = finder(sc, cadr(args));
		car(sc->T1_1) = c_call(args)(sc, sc->T1_1);
		sc->value = c_call(code)(sc, sc->T1_1);
		break;
		
	      case HOP_SAFE_C_opSq_opSq:
		args = cdr(code);
		car(sc->T1_1) = finder(sc, cadr(car(args)));
		sc->temp4 = c_call(car(args))(sc, sc->T1_1);
		args = cadr(args);
		car(sc->T1_1) = finder(sc, cadr(args));
		car(sc->T2_2) = c_call(args)(sc, sc->T1_1);
		car(sc->T2_1) = sc->temp4;
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      default:
		fprintf(stderr, "oops: if_and_x: %s\n", DISPLAY_80(code));
		break;
	      }
	    
	    if (is_false(sc, sc->value))
	      {
		sc->value = sc->UNSPECIFIED;
		goto START;
	      }
	  }
	if (is_true(sc, sc->value))
	  {
	    sc->code = cadr(sc->code);
	    goto EVAL_PAIR;
	  }
	sc->value = sc->UNSPECIFIED;
	goto START;
      }


    case OP_SAFE_IF_ORX_P:
      {
	s7_pointer p;
	for (p = cdar(sc->code); is_pair(p); p = cdr(p))
	  {
	    s7_pointer val, code, args;
	    code = car(p);
	    switch (optimize_data(code))
	      {
	      case HOP_SAFE_C_C:
		sc->value = c_call(code)(sc, cdr(code));
		break;
		
	      case HOP_SAFE_C_S:
		car(sc->T1_1) = finder(sc, cadr(code));
		sc->value = c_call(code)(sc, sc->T1_1);
		break;
		
	      case HOP_SAFE_IS_PAIR_S:
		sc->value = make_boolean(sc, is_pair(finder(sc, cadr(code))));
		break;
		
	      case HOP_SAFE_IS_SYMBOL_S:
		sc->value = make_boolean(sc, is_symbol(finder(sc, cadr(code))));
		break;
		
	      case HOP_SAFE_NOT_S:
		sc->value = make_boolean(sc, is_false(sc, finder(sc, cadr(code))));
		break;
		
	      case HOP_SAFE_C_SC:
		args = cdr(code);
		car(sc->T2_1) = finder(sc, car(args));
		car(sc->T2_2) = cadr(args);
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_CS:
		args = cdr(code);
		car(sc->T2_2) = finder(sc, cadr(args));
		car(sc->T2_1) = car(args);
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_SS:
		args = cdr(code);
		val = finder(sc, car(args));
		car(sc->T2_2) = finder(sc, cadr(args));
		car(sc->T2_1) = val;
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_SQ:
		args = cdr(code);
		car(sc->T2_1) = finder(sc, car(args));
		car(sc->T2_2) = cadr(cadr(args));
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_opSq:
		args = cadr(code);
		car(sc->T1_1) = finder(sc, cadr(args));
		car(sc->T1_1) = c_call(args)(sc, sc->T1_1);
		sc->value = c_call(code)(sc, sc->T1_1);
		break;
		
	      case HOP_SAFE_C_opSq_opSq:
		args = cdr(code);
		car(sc->T1_1) = finder(sc, cadr(car(args)));
		sc->temp4 = c_call(car(args))(sc, sc->T1_1);
		args = cadr(args);
		car(sc->T1_1) = finder(sc, cadr(args));
		car(sc->T2_2) = c_call(args)(sc, sc->T1_1);
		car(sc->T2_1) = sc->temp4;
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      default:
		fprintf(stderr, "oops: if_and_x: %s\n", DISPLAY_80(code));
		break;
	      }
	    
	    if (is_true(sc, sc->value))
	      {
		sc->code = cadr(sc->code);
		goto EVAL_PAIR;
	      }
	  }
	if (is_true(sc, sc->value))
	  {
	    sc->code = cadr(sc->code);
	    goto EVAL_PAIR;
	  }
	sc->value = sc->UNSPECIFIED;
	goto START;
      }
#endif


    case OP_IF_B_P:
      {
	s7_pointer code;
	code = sc->code;
	push_stack_no_args(sc, OP_IF_PP, cadr(code));
	code = car(code);
	sc->op = (opcode_t)syntax_opcode(car(code));

	/* s7: OP_OR_P: 20820 OP_AND_P: 3517
	 * lg: OP_AND_UNCHECKED: 290466 OP_AND_P: 112172
	 * in: OP_AND_P: 3099451, OP_AND_UNCHECKED: 204313, OP_SAFER_OR_CEQ: 2251149
	 * al: none to speak of
	 */
	sc->code = cdr(code);
	goto START_WITHOUT_POP_STACK;
      }


    case OP_IF_ANDP_P:
      push_stack_no_args(sc, OP_IF_PP, cadr(sc->code));
      sc->code = cdar(sc->code);
      goto AND_P;


    case OP_IF_ORP_P:
      push_stack_no_args(sc, OP_IF_PP, cadr(sc->code));
      sc->code = cdar(sc->code);
      goto OR_P;


#if WITH_OPTIMIZATION
    case OP_IF_ORCEQ_P:
      {
	s7_pointer code, sym;
	code = fcdr(sc->code);  /* cdar */
	sym = finder(sc, cadar(code));

	do {
	  if (sym == fcdr(code))
	    {
	      sc->code = cadr(sc->code);
	      goto EVAL_PAIR;
	    }
	  code = cdr(code);
	} while (is_pair(code));

	if (!s7_is_character(sym))
	  s7_wrong_type_arg_error(sc, "char=?", 1, cadar(code), "a character");
	sc->value = sc->UNSPECIFIED;
	goto START;
	/* in OP_SIMPLE_DO_STEP_P: 1884560 */
      }
#endif


    case OP_IF_PP:
      if (is_true(sc, sc->value))
	goto EVAL_PAIR; /* almost never happens */
      sc->value = sc->UNSPECIFIED;  
      goto START;  
      /* lg 469613  OP_BEGIN1: 446942
       * in 3128150 OP_BEGIN1: 717010 OP_SIMPLE_DO_STEP_P: 2400713
       */
      
    case OP_IF_P_P_X:
      push_stack_no_args(sc, OP_IF_PPX, cdr(sc->code));
      sc->code = car(sc->code);
      goto EVAL_PAIR;
      
    case OP_IF_PPX:
      if (is_true(sc, sc->value))
	{
	  sc->code = car(sc->code);
	  goto EVAL_PAIR;
	}
      if (is_symbol(cadr(sc->code)))
	sc->value = finder(sc, cadr(sc->code));
      else sc->value = cadr(sc->code);
      goto START;


    case OP_IF_P_X_P:
      push_stack_no_args(sc, OP_IF_PXP, cdr(sc->code));
      sc->code = car(sc->code);
      /* s7: h_safe_c_sc: 872
       * lg: closure_s: 32814
       * in: h_safe_c_ss: 124044 h_safe_c_opsq: 37136 [cs and ss are now separated]
       * al: h_safe_c_sc: 2955964 h_safe_c_ss: 222633
       */
      goto EVAL_PAIR; /* all: opt */

    case OP_IF_PXP:
      if (is_true(sc, sc->value))
	{
	  if (is_symbol(car(sc->code)))
	    sc->value = finder(sc, car(sc->code));
	  else sc->value = car(sc->code);
	  goto START;
	}
      sc->code = cadr(sc->code);
      /* al unopt: 500168 h_safe_c_pp: 1346267
       * lg and in: mostly unopt
       */
      goto EVAL_PAIR;


    case OP_IF_P_X:
      push_stack_no_args(sc, OP_IF_PX, cadr(sc->code));
      sc->code = car(sc->code);
      goto EVAL_PAIR;

    case OP_IF_PX:
      if (is_true(sc, sc->value))
	{
	  if (is_symbol(sc->code))
	    sc->value = finder(sc, sc->code);
	  else sc->value = sc->code;
	}
      else sc->value = sc->UNSPECIFIED;
      goto START;


    case OP_IF_P_X_X:
      push_stack_no_args(sc, OP_IF_PXX, cdr(sc->code));
      sc->code = car(sc->code);
      goto EVAL_PAIR;

    case OP_IF_PXX:
      if (is_true(sc, sc->value))
	{
	  if (is_symbol(car(sc->code)))
	    sc->value = finder(sc, car(sc->code));
	  else sc->value = car(sc->code);
	}
      else
	{
	  if (is_symbol(cadr(sc->code)))
	    sc->value = finder(sc, cadr(sc->code));
	  else sc->value = cadr(sc->code);
	}
      goto START;

      
    case OP_IF_X_P_P:
      if (is_symbol(car(sc->code)))
	{
	  if (is_true(sc, finder(sc, car(sc->code))))
	    sc->code = cadr(sc->code);
	  else sc->code = caddr(sc->code);
	}
      else
	{
	  if (is_true(sc, car(sc->code)))
	    sc->code = cadr(sc->code);
	  else sc->code = caddr(sc->code);
	}
      goto EVAL_PAIR;


    case OP_IF_X_P:
      if (is_symbol(car(sc->code)))
	{
	  if (is_true(sc, finder(sc, car(sc->code))))
	    {
	      sc->code = cadr(sc->code);
	      goto EVAL_PAIR;
	    }
	}
      else
	{
	  if (is_true(sc, car(sc->code)))
	    {
	      sc->code = cadr(sc->code);
	      goto EVAL_PAIR;
	    }
	}
      sc->value = sc->UNSPECIFIED;
      goto START;
      

    case OP_IF_X_P_X:
      if (is_symbol(car(sc->code)))
	{
	  if (is_true(sc, finder(sc, car(sc->code))))
	    {
	      sc->code = cadr(sc->code);
	      goto EVAL_PAIR;
	    }
	}
      else
	{
	  if (is_true(sc, car(sc->code)))
	    {
	      sc->code = cadr(sc->code);
	      goto EVAL_PAIR;
	    }
	}
      if (is_symbol(caddr(sc->code)))
	sc->value = finder(sc, caddr(sc->code));
      else sc->value = caddr(sc->code);
      goto START;


    case OP_IF_X_X_P:
      if (is_symbol(car(sc->code)))
	{
	  if (is_true(sc, finder(sc, car(sc->code))))
	    {
	      if (is_symbol(cadr(sc->code)))
		sc->value = finder(sc, cadr(sc->code));
	      else sc->value = cadr(sc->code);
	      goto START;
	    }
	}
      else
	{
	  if (is_true(sc, car(sc->code)))
	    {
	      if (is_symbol(cadr(sc->code)))
		sc->value = finder(sc, cadr(sc->code));
	      else sc->value = cadr(sc->code);
	      goto START;
	    }
	}
      sc->code = caddr(sc->code);
      goto EVAL_PAIR;


    case OP_IF_X_X:
      sc->value = sc->UNSPECIFIED;
      if (is_symbol(car(sc->code)))
	{
	  if (is_true(sc, finder(sc, car(sc->code))))
	    {
	      if (is_symbol(cadr(sc->code)))
		sc->value = finder(sc, cadr(sc->code));
	      else sc->value = cadr(sc->code);
	    }
	}
      else
	{
	  if (is_true(sc, car(sc->code)))
	    {
	      if (is_symbol(cadr(sc->code)))
		sc->value = finder(sc, cadr(sc->code));
	      else sc->value = cadr(sc->code);
	    }
	}
      /* this this amost always going to OP_EVAL_ARGS4 (once in a while it's OP_MAP_SIMPLE) */
      goto START;


    case OP_IF_X_X_X:
      sc->value = sc->UNSPECIFIED;
      if (is_symbol(car(sc->code)))
	{
	  if (is_true(sc, finder(sc, car(sc->code))))
	    {
	      if (is_symbol(cadr(sc->code)))
		sc->value = SYMBOL_TO_VALUE(sc, cadr(sc->code));
	      else sc->value = cadr(sc->code);
	    }
	  else
	    {
	      if (is_symbol(caddr(sc->code)))
		sc->value = SYMBOL_TO_VALUE(sc, caddr(sc->code));
	      else sc->value = caddr(sc->code);
	    }
	}
      else
	{
	  if (is_true(sc, car(sc->code)))
	    {
	      if (is_symbol(cadr(sc->code)))
		sc->value = SYMBOL_TO_VALUE(sc, cadr(sc->code));
	      else sc->value = cadr(sc->code);
	    }
	  else
	    {
	      if (is_symbol(caddr(sc->code)))
		sc->value = SYMBOL_TO_VALUE(sc, caddr(sc->code));
	      else sc->value = caddr(sc->code);
	    }
	}
      goto START;

#if WITH_OPTIMIZATION
    case OP_SAFE_IF2:
      push_stack_no_args(sc, OP_SAFE_IF2_1, cdr(sc->code));
      sc->code = car(sc->code);
      if (!is_optimized(sc->code))
	goto EVAL_PAIR;
      goto OPT_EVAL;

    case OP_SAFE_IF2_1:
      if (is_true(sc, sc->value))
	sc->code = car(sc->code);
      else sc->code = cadr(sc->code);
      if (!is_optimized(sc->code))
	goto EVAL_PAIR;
      goto OPT_EVAL;
	

    case OP_SAFE_IF1:
      push_stack_no_args(sc, OP_SAFE_IF1_1, cadr(sc->code));
      sc->code = car(sc->code); 
      if (!is_optimized(sc->code))
	goto EVAL_PAIR;
      /* lg: 1: h_safe_c_opsafe_closure_ssq 56202   1: closure_ss 50332
       */
      goto OPT_EVAL;

    case OP_SAFE_IF1_1:
      if (is_true(sc, sc->value))
	{
	  if (!is_optimized(sc->code))
	    goto EVAL_PAIR;
	  goto OPT_EVAL;          /* all c-xxx in lg */
	}
      sc->value = sc->UNSPECIFIED;
      goto START;


    case OP_SAFE_IF_CC_X_P:
      {
	s7_pointer code;
	code = sc->code;
	if (is_true(sc, c_call(car(code))(sc, cdar(code))))
	  {
	    if (is_symbol(cadr(code)))
	      sc->value = finder(sc, cadr(code));
	    else sc->value = cadr(code);
	    goto START;
	  }
	sc->code = caddr(code);
	goto EVAL_PAIR; /* all: opt */
      }


    case OP_SAFE_IF_CC_P:
      {
	s7_pointer code;
	code = sc->code;
	if (is_true(sc, c_call(car(code))(sc, cdar(code))))
	  {
	    sc->code = cadr(code);
	    goto EVAL_PAIR;
	  }
	else sc->value = sc->UNSPECIFIED;
	goto START;
	/* in OP_DOTIMES_STEP: 2153849
	 */ 
      }


    case OP_SAFE_IF_CC_P_P:
      {
	s7_pointer code;
	code = sc->code;
	if (is_true(sc, c_call(car(code))(sc, cdar(code))))
	  sc->code = cadr(code);
	else sc->code = caddr(code);
	goto EVAL_PAIR;
      }


    case OP_SAFE_IF_CEQ_P_P:
      {
	s7_pointer c, code;
	code = cdar(sc->code);
	c = finder(sc, car(code));
	if (c == cadr(code))           /* we know cadr(code) is a char, so if c equals it, c must be a char */
	  sc->code = cadr(sc->code);
	else 
	  {
	    if (!s7_is_character(c))
	      return(s7_wrong_type_arg_error(sc, "char=?", 1, c, "a character"));
	    sc->code = fcdr(sc->code); /* caddr */
	  }
	goto EVAL_PAIR;
      }


    case OP_SAFE_IF_CS_X_P:
      {
	s7_pointer code;
	code = sc->code;
	car(sc->T1_1) = finder(sc, fcdr(code));
	if (is_true(sc, c_call(car(code))(sc, sc->T1_1)))
	  {
	    if (is_symbol(cadr(code)))
	      sc->value = finder(sc, cadr(code));
	    else sc->value = cadr(code);
	    goto START;
	  }
	sc->code = caddr(code);
	goto EVAL_PAIR; /* all: syn */
      }

    case OP_SAFE_IF_CSS_X_P:
      {
	s7_pointer code;
	code = sc->code;
	car(sc->T2_1) = finder(sc, cadar(code));
	car(sc->T2_2) = finder(sc, fcdr(code));
	if (is_true(sc, c_call(car(code))(sc, sc->T2_1)))
	  {
	    if (is_symbol(cadr(code)))
	      sc->value = finder(sc, cadr(code));
	    else sc->value = cadr(code);
	    goto START;
	  }
	sc->code = caddr(code);
	goto EVAL_PAIR;
      }

    case OP_SAFE_IF_CSC_X_P:
      {
	s7_pointer code;
	code = sc->code;
	car(sc->T2_1) = finder(sc, cadar(code));
	car(sc->T2_2) = fcdr(code);
	if (is_true(sc, c_call(car(code))(sc, sc->T2_1)))
	  {
	    if (is_symbol(cadr(code)))
	      sc->value = finder(sc, cadr(code));
	    else sc->value = cadr(code);
	    goto START;
	  }
	sc->code = caddr(code);
	goto EVAL_PAIR;
      }

    case OP_SAFE_IF_CSC_X_P_A:
      {
	s7_pointer code;
	code = sc->code;
	sc->value = finder(sc, cadar(code));
	car(sc->T2_1) = sc->value;
	car(sc->T2_2) = fcdr(code);
	if (is_true(sc, c_call(car(code))(sc, sc->T2_1)))
	  goto START;
	sc->code = caddr(code);
	goto EVAL_PAIR;
      }


    case OP_SAFE_IF_CS_P_X:
      {
	s7_pointer code;
	code = sc->code;
	car(sc->T1_1) = finder(sc, cadar(code));
	if (is_true(sc, c_call(car(code))(sc, sc->T1_1)))
	  {
	    sc->code = cadr(code);
	    goto EVAL_PAIR;
	  }
	if (is_symbol(caddr(code)))
	  sc->value = finder(sc, caddr(code));
	else sc->value = caddr(code);
	goto START;
      }


    case OP_SAFE_IF_IS_PAIR_P_X:
      if (is_pair(finder(sc, fcdr(sc->code))))
	{
	  sc->code = cadr(sc->code);
	  goto EVAL_PAIR;
	}
      if (is_symbol(caddr(sc->code)))
	sc->value = finder(sc, caddr(sc->code));
      else sc->value = caddr(sc->code);
      goto START;


    case OP_SAFE_IF_IS_SYMBOL_P_X:
      if (is_symbol(finder(sc, cadar(sc->code))))
	{
	  sc->code = cadr(sc->code);
	  goto EVAL_PAIR;
	}
      if (is_symbol(caddr(sc->code)))
	sc->value = finder(sc, caddr(sc->code));
      else sc->value = caddr(sc->code);
      goto START;


    case OP_SAFE_IF_CS_P_P:
      car(sc->T1_1) = finder(sc, fcdr(sc->code));
      if (is_true(sc, c_call(car(sc->code))(sc, sc->T1_1)))
	sc->code = cadr(sc->code);
      else sc->code = caddr(sc->code);
      goto EVAL_PAIR;


    case OP_SAFE_IF_IS_PAIR_P_P:
      if (is_pair(finder(sc, fcdr(sc->code))))
	sc->code = cadr(sc->code);
      else sc->code = caddr(sc->code);
      goto EVAL_PAIR;

    case OP_SAFE_IF_IS_SYMBOL_P_P:
      if (is_symbol(finder(sc, fcdr(sc->code))))
	sc->code = cadr(sc->code);
      else sc->code = caddr(sc->code);
      goto EVAL_PAIR;

    case OP_SAFE_IF_IS_EOF_P_P:
      {
	s7_pointer code;
	code = sc->code;
	if (finder(sc, fcdr(code)) == sc->EOF_OBJECT)
	  {
	    sc->code = cadr(code);
	    goto EVAL_PAIR;
	  }
	code = caddr(code);
	sc->op = (opcode_t)syntax_opcode(car(code));
	sc->code = cdr(code);
	goto START_WITHOUT_POP_STACK;
      }

    case OP_SAFE_IF_CS_Q_P:
      car(sc->T1_1) = finder(sc, fcdr(sc->code));
      if (is_true(sc, c_call(car(sc->code))(sc, sc->T1_1)))
	{
	  sc->value = cadr(cadr(sc->code));
	  goto START;
	}
      sc->code = caddr(sc->code);
      goto EVAL_PAIR;


    case OP_SAFE_IF_CS_P:
      car(sc->T1_1) = finder(sc, fcdr(sc->code));
      if (is_true(sc, c_call(car(sc->code))(sc, sc->T1_1)))
	{
	  sc->code = cadr(sc->code);
	  goto EVAL_PAIR; /* in: syn */
	}
      sc->value = sc->UNSPECIFIED;
      goto START;


    case OP_SAFE_IF_CSQ_P:
      car(sc->T2_1) = finder(sc, cadr(car(sc->code)));
      car(sc->T2_2) = fcdr(sc->code);
      if (is_true(sc, c_call(car(sc->code))(sc, sc->T2_1)))
	{
	  sc->code = cadr(sc->code);
	  goto EVAL_PAIR;
	}
      sc->value = sc->UNSPECIFIED;
      goto START;
      /* lg OP_BEGIN1: 5153
       * al 
       */

    case OP_SAFE_IF_CSQ_P_P:
      car(sc->T2_1) = finder(sc, cadr(car(sc->code)));
      car(sc->T2_2) = fcdr(sc->code);
      if (is_true(sc, c_call(car(sc->code))(sc, sc->T2_1)))
	sc->code = cadr(sc->code);
      else sc->code = caddr(sc->code);
      goto EVAL_PAIR;


    case OP_SAFE_IF_CSS_P:
      car(sc->T2_1) = finder(sc, cadr(car(sc->code)));
      car(sc->T2_2) = finder(sc, fcdr(sc->code));
      if (is_true(sc, c_call(car(sc->code))(sc, sc->T2_1)))
	{
	  sc->code = cadr(sc->code);
	  goto EVAL_PAIR;
	}
      sc->value = sc->UNSPECIFIED;
      goto START;
      /* snd-test OP_SAFE_IF_CSS_P -> OP_BEGIN1: 4541525
       */


    case OP_SAFE_IF_CSS_P_P:
      car(sc->T2_1) = finder(sc, cadr(car(sc->code)));
      car(sc->T2_2) = finder(sc, fcdr(sc->code));
      if (is_true(sc, c_call(car(sc->code))(sc, sc->T2_1)))
	sc->code = cadr(sc->code);
      else sc->code = caddr(sc->code);
      goto EVAL_PAIR;


    case OP_SAFE_IF_CSC_P:
      car(sc->T2_1) = finder(sc, cadr(car(sc->code)));
      car(sc->T2_2) = fcdr(sc->code);
      if (is_true(sc, c_call(car(sc->code))(sc, sc->T2_1)))
	{
	  sc->code = cadr(sc->code);
	  goto EVAL_PAIR;
	}
      sc->value = sc->UNSPECIFIED;
      goto START;

    case OP_SAFE_IF_CSC_P_P:
      car(sc->T2_1) = finder(sc, cadr(car(sc->code)));
      car(sc->T2_2) = fcdr(sc->code);
      if (is_true(sc, c_call(car(sc->code))(sc, sc->T2_1)))
	sc->code = cadr(sc->code);
      else sc->code = caddr(sc->code);
      goto EVAL_PAIR;


    case OP_SAFE_IF_IS_PAIR_P:
      if (is_pair(finder(sc, fcdr(sc->code))))
	{
	  sc->code = cadr(sc->code);
	  goto EVAL_PAIR;
	}
      sc->value = sc->UNSPECIFIED;
      goto START;

    case OP_SAFE_IF_IS_SYMBOL_P:
      if (is_symbol(finder(sc, fcdr(sc->code))))
	{
	  sc->code = cadr(sc->code);
	  goto EVAL_PAIR;
	}
      sc->value = sc->UNSPECIFIED;
      goto START;


    case OP_SAFE_IF_C_SS_P:
      {
	s7_pointer args, val1;
	args = cadar(sc->code);
	val1 = finder(sc, cadr(args));
	car(sc->T2_2) = finder(sc, caddr(args));
	car(sc->T2_1) = val1;
	car(sc->T1_1) = c_call(args)(sc, sc->T2_1);
	if (is_true(sc, c_call(car(sc->code))(sc, sc->T1_1)))
	  {
	    sc->code = cadr(sc->code);
	    goto EVAL_PAIR;
	  }
	sc->value = sc->UNSPECIFIED;
	goto START;
      }


   case OP_SAFE_C_P_1:
      car(sc->T1_1) = sc->value;
      sc->value = c_call(sc->code)(sc, sc->T1_1);
      goto START;


    case OP_SAFE_C_PP_1:
      push_stack(sc, OP_SAFE_C_PP_2, cons(sc, sc->value, sc->args), sc->code);
      sc->code = caddr(sc->code);
      if (is_optimized(sc->code))
	{
	  /* lg: h_safe_c_opsq_opsq: 18797 h_safe_c_ss: 4197
	   * al: h_closure_opcq: 1346267
	   */
	  goto OPT_EVAL;
	}
      goto EVAL_PAIR;

    case OP_SAFE_C_PP_2:
      {
	s7_pointer p;
	p = sc->args;
	if (is_null(cdr(p)))
	  {
	    car(sc->T2_1) = car(p);
	    car(sc->T2_2) = sc->value;
	    sc->value = c_call(sc->code)(sc, sc->T2_1);

	    typeflag(p) = 0;                          /* this was allocated above in the cons, so we can release it now */
	    (*(sc->free_heap_top++)) = p;
	    sc->args = sc->NIL;

	    goto START;
	  }
	
	/* (define (hi) (+ (values 1 2) (values 3 4)))
	 *
	 * c_call(sc->code) here refers to the fcdr field which is g_add_2, but we have any number of args from a values call
	 *   the original (unoptimized) function is not directly accessible: ecdr is the holder of sc->code,
	 *   its type is T_C_FUNCTION, so ecdr(sc->code).object.ffptr points to the g_add_2 function data,
	 *   which is not helpful (the class doesn't give us a backpointer).  car(sc->code) is a symbol, its
	 *   local|global_slot points to a slot, slot_value(local_slot(car(sc-code))) is our original function?!?!
	 *   so we want c_function_call(slot_value(local_slot(car(sc->code)))), I guess.  Surely this could be simpler.
	 *   And safer -- what if someone defines + at top-level after using it in this case?  Perhaps add orig_func
	 *   field to ffptr info, then c_original_function_call(sc->code)(...).
	 */

	sc->value = c_function_call(slot_value(global_slot(car(sc->code))))(sc, safe_reverse_in_place(sc, cons(sc, sc->value, p)));
	goto START;
      }


    case OP_C_P_1:
      if (is_null(sc->args))
	sc->value = c_call(sc->code)(sc, list_1(sc, sc->value));
      else sc->value = c_call(sc->code)(sc, safe_reverse_in_place(sc, cons(sc, sc->value, sc->args)));
      goto START;

#endif

	     
      

      /* -------------------------------- LET -------------------------------- */

    case OP_LET_NO_VARS:
      NEW_FRAME(sc, sc->envir, sc->envir); 
      sc->code = cdr(sc->code);            /* ignore the () */
      goto BEGIN;


    case OP_NAMED_LET_NO_VARS:
      NEW_FRAME(sc, sc->envir, sc->envir); 
      sc->args = sc->NIL;                  /* the let args */
      sc->x = make_closure(sc, sc->NIL, cddr(sc->code), T_CLOSURE);
      add_slot(sc, car(sc->code), sc->x); 
      sc->code = cddr(sc->code);
      sc->x = sc->NIL;
      goto BEGIN;


    case OP_LET_C:
      /* one var, init is constant, incoming sc->code is '(((var val))...)!
       */
      NEW_FRAME_WITH_SLOT(sc, sc->envir, sc->envir, caaar(sc->code), cadaar(sc->code));
      sc->code = cdr(sc->code);
      goto BEGIN;


    case OP_LET_C_P:
      /* one var, init is constant, incoming sc->code is '(((var val))...)!
       *   body is one statement
       */
      NEW_FRAME_WITH_SLOT(sc, sc->envir, sc->envir, caaar(sc->code), cadaar(sc->code));
      sc->code = cadr(sc->code);
      goto EVAL_PAIR;


#if WITH_OPTIMIZATION
    case OP_LET_C_D:
      /* one var, init is constant, incoming sc->code is '(((var val))...)!
       *   body is one optimized statement
       *
       * if this were known to be safe, it could be preallocated, I think.
       */
      {
	s7_pointer binding;
	binding = caar(sc->code);
	NEW_FRAME_WITH_SLOT(sc, sc->envir, sc->envir, car(binding), cadr(binding));
	sc->code = cadr(sc->code);
	goto OPT_EVAL;
      }

      /* if the body here is hop_safe_c_s and the s matches the let var we could use:
       *
       *  car(sc->T1_1) = cadr(caar(sc->code));
       *  sc->value = c_call(cadr(sc->code))(sc, sc->T1_1);
       *  goto START;
       *
       * but this kind of code doesn't seem to happen very often
       */
#endif

 
    case OP_LET_S:
      /* one var, init is symbol, incoming sc->code is '(((var sym))...)
       */
      {
	s7_pointer val;
	val = finder(sc, cadaar(sc->code));
	NEW_FRAME_WITH_SLOT(sc, sc->envir, sc->envir, caaar(sc->code), val);
	sc->code = cdr(sc->code);
	goto BEGIN;
      }


    case OP_LET_S_P:
      /* one var, init is symbol, incoming sc->code is '(((var sym))...)
       */
      {
	s7_pointer binding, val;
	binding = caar(sc->code);
	val = finder(sc, cadr(binding));
	NEW_FRAME_WITH_SLOT(sc, sc->envir, sc->envir, car(binding), val);
	sc->code = cadr(sc->code);
	goto EVAL_PAIR; /* lg: opt -- entirely h_c_for_each_ls: 56202 */
      }


    case OP_LET_Q:
      /* one var, init is quoted, incoming sc->code is '(((var 'val))...)
       */
      {
	s7_pointer binding;
	binding = caar(sc->code);
	NEW_FRAME_WITH_SLOT(sc, sc->envir, sc->envir, car(binding), cadr(cadr(binding)));
	sc->code = cdr(sc->code);
	goto BEGIN;
      }

#if WITH_OPTIMIZATION
    case OP_LET_R:
      /* one var, init is safe (op sym), incoming sc->code is '(((var (op sym)))...)
       */
      {
	s7_pointer binding;
	binding = caar(sc->code);
	car(sc->T1_1) = finder(sc, fcdr(sc->code));
	sc->value = c_call(cadr(binding))(sc, sc->T1_1);
	NEW_FRAME_WITH_SLOT(sc, sc->envir, sc->envir, car(binding), sc->value);
	sc->code = cdr(sc->code);
	goto BEGIN;
      }

    case OP_LET_R_P:
      {
	s7_pointer binding;
	binding = caar(sc->code);
	car(sc->T1_1) = finder(sc, fcdr(sc->code));
	sc->value = c_call(cadr(binding))(sc, sc->T1_1);
	NEW_FRAME_WITH_SLOT(sc, sc->envir, sc->envir, car(binding), sc->value);
	sc->code = cadr(sc->code);
	goto EVAL_PAIR;
      }

    case OP_LET_CAR_P:
      {
	s7_pointer val, code;
	opcode_t op;

	code = sc->code;
	val = finder(sc, fcdr(code));
	if (!is_pair(val)) 
	  s7_wrong_type_arg_error(sc, "car", 0, val, "a pair");
	NEW_FRAME_WITH_SLOT(sc, sc->envir, sc->envir, caaar(code), car(val));

	code = cadr(code);
	op = (opcode_t)syntax_opcode(car(code));

	if (op == OP_SAFE_IF_CC_X_P)
	  {
	    code = cdr(code);
	    if (is_true(sc, c_call(car(code))(sc, cdar(code))))
	      {
		if (is_symbol(cadr(code)))
		  sc->value = finder(sc, cadr(code));
		else sc->value = cadr(code);
		goto START;
	      }
	    sc->code = caddr(code);
	    goto EVAL_PAIR; /* all: opt */
	  }
	sc->op = op;
	sc->code = cdr(code);
	goto START_WITHOUT_POP_STACK;
      }

    case OP_LET_READ_CHAR_P:
      {
	s7_pointer port, code, fc, c;
	opcode_t op;

	code = sc->code;
	port = finder(sc, fcdr(code));
	if (!is_input_port(port)) /* (read-char 123)? */
	  s7_wrong_type_arg_error(sc, "read-char", 0, port, "an input port");

	c = chars[port_read_character(port)(sc, port)];
	NEW_FRAME_WITH_SLOT(sc, sc->envir, sc->envir, fc = caaar(code), c);
	code = cadr(code);

	op = (opcode_t)syntax_opcode(car(code));
	if (op == OP_SAFE_IF_IS_EOF_P_P)
	  {
	    code = cdr(code);
	    if (fcdr(code) != fc)
	      c = finder(sc, fcdr(code));
	    if (c == sc->EOF_OBJECT)
	      {
		sc->code = cadr(code);
		goto EVAL_PAIR;
	      }
	    code = caddr(code);
	    sc->code = cdr(code);
	    op = (opcode_t)syntax_opcode(car(code));
	    if (op == OP_BEGIN)
	      goto BEGIN;
	    sc->op = op;
	    goto START_WITHOUT_POP_STACK;
	  }
	sc->op = op;
	sc->code = cdr(code);
	goto START_WITHOUT_POP_STACK;
      }


    case OP_LET_O:
      push_stack(sc, OP_LET_O1, caaar(sc->code), cdr(sc->code));
      sc->code = fcdr(sc->code);
      if (!is_optimized(sc->code))
	goto EVAL_PAIR;
      goto OPT_EVAL;

    case OP_LET_O1:
      NEW_FRAME_WITH_SLOT(sc, sc->envir, sc->envir, sc->args, sc->value);
      goto BEGIN;


    case OP_LET_O_P:
      push_stack(sc, OP_LET_O2, caaar(sc->code), cadr(sc->code));
      sc->code = fcdr(sc->code);
      if (!is_optimized(sc->code))
	goto EVAL_PAIR;
      /* this is primarily safe_closure_ss or other such stuff
       */
      goto OPT_EVAL;

    case OP_LET_O2:
      NEW_FRAME_WITH_SLOT(sc, sc->envir, sc->envir, sc->args, sc->value);
      goto EVAL_PAIR;
      /* handling syntax case explicitly here was a toss-up
       */


    case OP_LET_OX_P:
      {
	s7_pointer val, args, code, sc_code;
	sc_code = sc->code; /* save across possible embedded eval? */
	code = fcdr(sc_code);
	switch (optimize_data(code))
	  {
	  case HOP_SAFE_C_C:
	    sc->value = c_call(code)(sc, cdr(code));
	    break;
	    
	  case HOP_SAFE_C_SC:
	    args = cdr(code);
	    car(sc->T2_1) = finder(sc, car(args));
	    car(sc->T2_2) = cadr(args);
	    sc->value = c_call(code)(sc, sc->T2_1);
	    break;
	    
	  case HOP_SAFE_C_CS:
	    args = cdr(code);
	    car(sc->T2_2) = finder(sc, cadr(args));
	    car(sc->T2_1) = car(args);
	    sc->value = c_call(code)(sc, sc->T2_1);
	    break;
	    
	  case HOP_SAFE_C_SS:
	    args = cdr(code);
	    val = finder(sc, car(args));
	    car(sc->T2_2) = finder(sc, cadr(args));
	    car(sc->T2_1) = val;
	    sc->value = c_call(code)(sc, sc->T2_1);
	    break;
	    
	  case HOP_SAFE_C_opSq:
	    args = cadr(code);
	    car(sc->T1_1) = finder(sc, cadr(args));
	    car(sc->T1_1) = c_call(args)(sc, sc->T1_1);
	    sc->value = c_call(code)(sc, sc->T1_1);
	    break;
	    
	  default:
	    fprintf(stderr, "oops: letx_p: %s\n", DISPLAY_80(sc->code));
	    break;
	  }
	
	NEW_FRAME_WITH_SLOT(sc, sc->envir, sc->envir, caaar(sc_code), sc->value);
	sc->code = cadr(sc_code);
	/* lg: h_call_with_exit if opt
	 */
	goto EVAL_PAIR;
      }


    case OP_LET_UNKNOWN_S_P:
      {
	s7_pointer val, ind, binding, expr;
	s7_Int index;
	/* sc->code is the let statement without the let: (((char (string index))) body) where body is one expression
	 */
	binding = fcdr(sc->code);
	expr = cadr(binding);
	val = finder(sc, car(expr));
	switch (type(val))
	  {
	  case T_STRING:
	    ind = finder(sc, cadr(expr));
	    if (s7_is_integer(ind))
	      {
		index = s7_integer(ind);
		if ((index < string_length(val)) &&
		    (index >= 0))
		  {
		    NEW_FRAME_WITH_SLOT(sc, sc->envir, sc->envir, car(binding), chars[(unsigned char)(string_value(val)[index])]);
		    sc->code = cadr(sc->code);
		    goto EVAL_PAIR;
		  }
	      }
	    break;

	  case T_C_FUNCTION:
	  case T_C_ANY_ARGS_FUNCTION:
	  case T_C_OPT_ARGS_FUNCTION:
	  case T_C_RST_ARGS_FUNCTION:
	  case T_C_LST_ARGS_FUNCTION:
	    if ((is_safe_procedure(val)) &&
		(c_function_required_args(val) <= 1) &&
		(c_function_all_args(val) >= 1))
	      {
		car(sc->T1_1) = finder(sc, cadr(expr));
		sc->value = c_call(expr)(sc, sc->T1_1);
		NEW_FRAME_WITH_SLOT(sc, sc->envir, sc->envir, car(binding), sc->value);
		sc->code = cadr(sc->code);
		goto EVAL_PAIR; /* in: syn */
	      }
	    break;
	    
	  default:
	    break;
	  }

	car(ecdr(sc->code)) = sc->LET_UNCHECKED;
	goto LET_UNCHECKED;
      }
#endif

    case OP_LET_ALL_C:
      /* n vars, all inits are constants.  We need to GC-protect the new frame-list as it is being
       *    created without tying the new frame into sc->envir until the end.
       */
      {
	s7_pointer p, frame;

	NEW_CELL(sc, frame);
	frame_id(frame) = frame_number + 1;
	environment_slots(frame) = sc->NIL;
	next_environment(frame) = sc->envir;
	set_type(frame, T_ENVIRONMENT | T_IMMUTABLE);
	sc->args = frame;                    /* GC protect it */

	for (p = car(sc->code); is_pair(p); p = cdr(p))
	  ADD_SLOT(frame, caar(p), cadar(p)); 
	
	frame_number++;
	sc->envir = frame;
	sc->code = cdr(sc->code);
	goto BEGIN;
      }


    case OP_LET_ALL_Q:
      {
	s7_pointer p, frame;

	NEW_CELL(sc, frame);
	frame_id(frame) = frame_number + 1;
	environment_slots(frame) = sc->NIL;
	next_environment(frame) = sc->envir;
	set_type(frame, T_ENVIRONMENT | T_IMMUTABLE);
	sc->args = frame;

	for (p = car(sc->code); is_pair(p); p = cdr(p))
	  ADD_SLOT(frame, caar(p), cadr(cadar(p)));
	
	frame_number++;
	sc->envir = frame;
	sc->code = cdr(sc->code);
	goto BEGIN;
      }


    case OP_LET_ALL_S:
      {
	s7_pointer p, frame;

	NEW_CELL(sc, frame);
	frame_id(frame) = frame_number + 1;
	environment_slots(frame) = sc->NIL;
	next_environment(frame) = sc->envir;
	set_type(frame, T_ENVIRONMENT | T_IMMUTABLE);
	sc->args = frame;

	for (p = car(sc->code); is_pair(p); p = cdr(p))
	  ADD_SLOT(frame, caar(p), finder(sc, cadar(p)));
	
	frame_number++;
	sc->envir = frame;
	sc->code = cdr(sc->code);
	goto BEGIN;
      }

#if WITH_OPTIMIZATION
    case OP_LET_ALL_R:
      {
	s7_pointer p, frame;

	NEW_CELL(sc, frame);
	frame_id(frame) = frame_number + 1;
	environment_slots(frame) = sc->NIL;
	next_environment(frame) = sc->envir;
	set_type(frame, T_ENVIRONMENT | T_IMMUTABLE);
	sc->args = frame;

	for (p = car(sc->code); is_pair(p); p = cdr(p))
	  {
	    s7_pointer cp;
	    cp = cadar(p);
	    car(sc->T1_1) = finder(sc, cadr(cp));
	    ADD_SLOT(frame, caar(p), c_call(cp)(sc, sc->T1_1)); 
	  }
	
	frame_number++;
	sc->envir = frame;
	sc->code = cdr(sc->code);
	goto BEGIN;
      }
#endif


    case OP_LET_ALL_G:
      {
	s7_pointer p, frame;

	NEW_CELL(sc, frame);
	frame_id(frame) = frame_number + 1;
	environment_slots(frame) = sc->NIL;
	next_environment(frame) = sc->envir;
	set_type(frame, T_ENVIRONMENT | T_IMMUTABLE);
	sc->args = frame; 

	for (p = car(sc->code); is_pair(p); p = cdr(p))
	  {
	    s7_pointer x, val;
	    x = cadar(p);
	    if (is_pair(x))
	      {
		if ((car(x) == sc->QUOTE) || 
		    (car(x) == sc->QUOTE_UNCHECKED))
		  val = cadr(x);
		else
		  {
		    car(sc->T1_1) = finder(sc, cadr(x));
		    val = c_call(x)(sc, sc->T1_1); 
		  }
	      }
	    else
	      {
		if (is_symbol(x))
		  val = finder(sc, x);
		else val = x;
	      }
	    ADD_SLOT(frame, caar(p), val);
	  }
	
	frame_number++;
	sc->envir = frame;
	sc->code = cdr(sc->code);
	goto BEGIN;
      }


    case OP_NAMED_LET:
      sc->args = sc->NIL;
      sc->value = sc->code;
      sc->code = cadr(sc->code);
      goto LET1;

#if WITH_OPTIMIZATION
    LET_UNCHECKED:
#endif
    case OP_LET_UNCHECKED:
      /* not named, but has vars */
      {
	s7_pointer x;
	NEW_CELL(sc, x);
	car(x) = sc->code;
	cdr(x) = sc->NIL;
	set_type(x, T_PAIR);
	sc->args = x;
	sc->code = car(sc->code);
	goto LET1A;
      }

      
    case OP_LET:
      /* sc->code is everything after the let: (let ((a 1)) a) so sc->code is (((a 1)) a) */
      /*   car can be either a list or a symbol ("named let") */
      {
	bool named_let;
	check_let(sc);
	sc->args = sc->NIL;
	sc->value = sc->code;
	named_let = is_symbol(car(sc->code));

	sc->code = (named_let) ? cadr(sc->code) : car(sc->code);
	if (is_null(sc->code))                    /* (let () ...):  no bindings, so skip that step */
	  {
	    sc->code = sc->value;
	    NEW_FRAME(sc, sc->envir, sc->envir);
	    if (named_let)
	      {    
		s7_pointer x;
		for (x = cadr(sc->code), sc->args = sc->NIL; is_pair(x); x = cdr(x)) 
		  sc->args = cons(sc, caar(x), sc->args);
		sc->x = make_closure(sc, safe_reverse_in_place(sc, sc->args), cddr(sc->code), T_CLOSURE);
		add_slot(sc, car(sc->code), sc->x); 
		sc->code = cddr(sc->code);
		sc->x = sc->NIL;
	      }
	    else 
	      {
		sc->code = cdr(sc->code);
	      }
	    goto BEGIN;
	  }
      }


    LET1:
    case OP_LET1:       /* let -- calculate parameters */
      /* sc->args = cons(sc, sc->value, sc->args); */
      {
	s7_pointer x, y;

	NEW_CELL(sc, x);
	car(x) = sc->value; /* the 1st time (now handled above), this saves the entire let body across the evaluations -- we pick it up later */
	cdr(x) = sc->args;
	set_type(x, T_PAIR);
	sc->args = x;

	if (is_pair(sc->code)) 
	  { 
	  LET1A:
	    x = cadar(sc->code);

	    if (is_pair(x))
	      {
		push_stack(sc, OP_LET1, sc->args, cdr(sc->code));
		sc->code = x;
		goto EVAL_PAIR; 
		/* this push_stack/goto can't be optimized away via a local optimize_data case statement
		 *    because any c_call can trigger an embedded call on the evaluator (for example,
		 *    open-sound involves both hooks, and s7_load if the corresponding .scm code exists),
		 *    so we have to protect sc->code and sc->args via the stack.  (I subsequently added
		 *    some protection here, but debugging this is hard, and the gain is not huge).
		 */
	      }
	    if (is_symbol(x))
	      sc->value = finder(sc, x);
	    else sc->value = x;
	    sc->code = cdr(sc->code);
	    goto LET1;
	  }

	x = safe_reverse_in_place(sc, sc->args);
	sc->code = car(x); /* restore the original form */
	y = cdr(x);        /* use sc->args as the new frame */
	environment_slots(x) = sc->NIL;                  
	next_environment(x) = sc->envir;
	set_type(x, T_ENVIRONMENT | T_IMMUTABLE); 
	frame_id(x) = ++frame_number; 
	sc->envir = x;		   
	
	{
	  bool named_let;
	  named_let = is_symbol(car(sc->code));
	  if (named_let)
	    {
	      /* we need to check the current environment for ridiculous cases like
	       *    (let hiho ((hiho 4)) hiho) -- I guess hiho is 4
	       */
	      s7_pointer let_name;
	      let_name = car(sc->code);
	      sc->envir = new_frame_in_env(sc, sc->envir);

	      sc->w = sc->NIL;
	      for (x = cadr(sc->code); is_pair(x); x = cdr(x)) 
		sc->w = cons(sc, caar(x), sc->w);

	      sc->x = make_closure(sc, safe_reverse_in_place(sc, sc->w), cddr(sc->code), T_CLOSURE);
	      add_slot(sc, let_name, sc->x); 
	      sc->x = sc->NIL;

	      sc->envir = new_frame_in_env(sc, sc->envir);
	      for (x = cadr(sc->code); is_not_null(y); x = cdr(x)) 
		{
		  s7_pointer sym, args, val;
		  /* reuse the value cells as the new frame slots */
		  
		  sym = caar(x);
		  if (sym == let_name) let_name = sc->NIL;
		  val = car(y);
		  args = cdr(y);
		  
		  slot_symbol(y) = sym;
		  if (symbol_has_accessor(sym))
		    slot_set_value(y, call_symbol_bind(sc, sym, val));
		  else slot_set_value(y, val);
		  set_type(y, T_SLOT | T_IMMUTABLE);
		  next_slot(y) = environment_slots(sc->envir);
		  environment_slots(sc->envir) = y;
		  symbol_id(sym) = frame_id(sc->envir);
		  local_slot(sym) = y;
		  
		  y = args;
		}
	      sc->code = cddr(sc->code);
	    }
	  else
	    {
	      s7_pointer e;
	      unsigned long long int id;

	      e = sc->envir;
	      id = frame_id(e);

	      for (x = car(sc->code); is_not_null(y); x = cdr(x)) 
		{
		  s7_pointer sym, args, val;
		  /* reuse the value cells as the new frame slots */
		  
		  sym = caar(x);
		  val = car(y);
		  args = cdr(y);
		  
		  slot_symbol(y) = sym;
		  symbol_id(sym) = id;
		  local_slot(sym) = y;
		  if (symbol_has_accessor(sym))
		    slot_set_value(y, call_symbol_bind(sc, sym, val));
		  else slot_set_value(y, val);
		  set_type(y, T_SLOT | T_IMMUTABLE);
		  next_slot(y) = environment_slots(e);
		  environment_slots(e) = y;
		  
		  y = args;
		}
	      sc->code = cdr(sc->code);
	    }
	}
	if (is_one_liner(sc->code))
	  {
	    sc->code = car(sc->code);
	    goto EVAL_PAIR;
	  }
	goto BEGIN;
      }


      /* -------------------------------- LET* -------------------------------- */

    case OP_LET_STAR:
      check_let_star(sc);

    case OP_LET_STAR_UNCHECKED:
      if (is_null(car(sc->code)))
	{
	  sc->envir = new_frame_in_env(sc, sc->envir);
	  sc->code = cdr(sc->code);
	  goto BEGIN;
	}
      push_stack(sc, OP_LET_STAR1, cdr(sc->code), car(sc->code));
                                      /* args is the let body, saved for later, code is the list of vars+initial-values */
      sc->code = cadaar(sc->code);
                                      /* caar(code) = 1st var/val pair, we've checked that all these guys are legit, so cadr of that is the value */
      goto EVAL;
      
      
    LET_STAR1:
    case OP_LET_STAR1:    /* let* -- calculate parameters */
      /* we can't skip (or reuse) this new frame -- we have to imitate a nested let, otherwise
       *
       *   (let ((f1 (lambda (arg) (+ arg 1))))
       *     (let* ((x 32)
       *            (f1 (lambda (arg) (f1 (+ x arg)))))
       *       (f1 1)))
       *
       * will hang.
       */
      {
	s7_pointer x, y;
	NEW_CELL(sc, x); 
	frame_id(x) = ++frame_number;
	next_environment(x) = sc->envir;
	sc->envir = x;
	set_type(x, T_ENVIRONMENT | T_IMMUTABLE);

	NEW_CELL_NO_CHECK(sc, y); 
	slot_symbol(y) = caar(sc->code);
	symbol_id(slot_symbol(y)) = frame_id(x);
	local_slot(slot_symbol(y)) = y;
	if (symbol_has_accessor(slot_symbol(y)))
	  slot_set_value(y, call_symbol_bind(sc, slot_symbol(y), sc->value));
	else slot_set_value(y, sc->value);
	set_type(y, T_SLOT | T_IMMUTABLE);
	next_slot(y) = sc->NIL;
	environment_slots(x) = y;
      }
      /* NEW_FRAME_WITH_SLOT does not include the symbol_accessor check */


      sc->code = cdr(sc->code);
      if (is_pair(sc->code)) 
	{ 
	  s7_pointer x;
	  x = cadar(sc->code);

	  if (is_pair(x))
	    {
	      push_stack(sc, OP_LET_STAR1, sc->args, sc->code); 
	      sc->code = x;
	      goto EVAL_PAIR;
	    }

	  if (is_symbol(x))
	    sc->value = finder(sc, x);
	  else sc->value = x;
	  goto LET_STAR1;
	} 
      sc->code = sc->args; /* cdr of original sc->code = cddr(let), set in push_stack above */
      goto BEGIN;
      

      /* -------------------------------- LETREC -------------------------------- */
      
    case OP_LETREC:
      check_letrec(sc);

    case OP_LETREC_UNCHECKED:
      /* get all local vars and set to #<undefined>
       * get parallel list of values
       * eval each member of values list with env still full of #<undefined>'s
       * assign each value to its variable
       * eval body
       *
       * which means that (letrec ((x x)) x) is not an error!
       */
      sc->envir = new_frame_in_env(sc, sc->envir); 
      if (is_pair(car(sc->code)))
	{
	  s7_pointer x;
	  for (x = car(sc->code); is_not_null(x); x = cdr(x))
	    {
	      s7_pointer slot;
	      slot = add_slot(sc, caar(x), sc->UNDEFINED);
	      slot_pending_value(slot) = sc->UNDEFINED;
	      slot_expression(slot) = cadar(x);
	    }
	  sc->args = environment_slots(sc->envir);
	  push_stack(sc, OP_LETREC1, sc->args, sc->code);
	  sc->code = slot_expression(sc->args);
	  goto EVAL;
	}
      sc->code = cdr(sc->code);
      goto BEGIN;

    case OP_LETREC1:
      slot_pending_value(sc->args) = sc->value;
      sc->args = next_slot(sc->args);
      if (is_slot(sc->args))
	{
	  push_stack(sc, OP_LETREC1, sc->args, sc->code);
	  sc->code = slot_expression(sc->args);
	  goto EVAL;
	}
      else
	{
	  s7_pointer slot;
	  for (slot = environment_slots(sc->envir); is_slot(slot); slot = next_slot(slot))
	    {
	      if (symbol_has_accessor(slot_symbol(slot)))
		slot_set_value(slot, call_symbol_bind(sc, slot_symbol(slot), slot_pending_value(slot)));
	      else slot_set_value(slot, slot_pending_value(slot));
	    }
	  sc->code = cdr(sc->code);
	  goto BEGIN;
	}

      

      /* -------------------------------- COND -------------------------------- */      
    case OP_COND:
      check_cond(sc);

    case OP_COND_UNCHECKED:
      push_stack(sc, OP_COND1, sc->NIL, sc->code);
      sc->code = caar(sc->code);
      goto EVAL;
      
      
    case OP_COND1:
      if (is_true(sc, sc->value))     /* got a hit (is_true -> not false, so else is true even though it has no value) */
	{
	  sc->code = cdar(sc->code);
	  if (is_null(sc->code))
	    {
	      if (is_multiple_value(sc->value))                             /* (+ 1 (cond ((values 2 3)))) */
		sc->value = splice_in_values(sc, multiple_value(sc->value));
	      /* no result clause, so return test, (cond (#t)) -> #t, (cond ((+ 1 2))) -> 3 */
	      goto START;
	    }

	  if ((is_pair(sc->code)) &&
	      (car(sc->code) == sc->FEED_TO) &&
	      (s7_symbol_value(sc, sc->FEED_TO) == sc->UNDEFINED))
	    {
	      if (is_multiple_value(sc->value))                             /* (cond ((values 1 2) => +)) */
		sc->code = cons(sc, cadr(sc->code), multiple_value(sc->value));
	      else sc->code = list_2(sc, cadr(sc->code), list_2(sc, sc->QUOTE, sc->value)); 
	      goto EVAL_PAIR;
	    }

	  /* (cond (#t 1)) sc->code is (1), but we're leaving syntax checks to OP_BEGIN: (cond (#t . 1)) etc */
	  goto BEGIN;
	}

      sc->code = cdr(sc->code);
      if (is_null(sc->code))
	{
	  sc->value = sc->NIL;
	  goto START;
	} 
	  
      push_stack_no_args(sc, OP_COND1, sc->code);
      sc->code = caar(sc->code);
      goto EVAL;
      
      
    case OP_COND_SIMPLE:
      push_stack_no_args(sc, OP_COND1_SIMPLE, sc->code);
      sc->code = caar(sc->code);
      goto EVAL;
      
      
    COND1_SIMPLE:
    case OP_COND1_SIMPLE:
      if (is_true(sc, sc->value))
	{
	  sc->code = cdar(sc->code);
	  if (is_null(sc->code))
	    {
	      if (is_multiple_value(sc->value))
		sc->value = splice_in_values(sc, multiple_value(sc->value));
	      goto START;
	    }
	  goto BEGIN;
	}

      sc->code = cdr(sc->code);
      if (is_null(sc->code))
	{
	  sc->value = sc->NIL;
	  goto START;
	} 
      if (is_pair(caar(sc->code)))
	{
	  push_stack_no_args(sc, OP_COND1_SIMPLE, sc->code);
	  sc->code = caar(sc->code);
	  goto EVAL_PAIR;
	}
      sc->value = caar(sc->code);
      if (is_symbol(sc->value))
	sc->value = finder(sc, sc->value);
      goto COND1_SIMPLE;


#if WITH_OPTIMIZATION
    case OP_COND_SIMPLER:
      {
	s7_pointer p, code;
	for (p = sc->code; is_pair(p); p = cdr(p))
	  {
	    code = caar(p);
	    if (is_pair(code))
	      {
		s7_pointer args, val;
		switch (optimize_data(code))
		  {
		  case HOP_SAFE_C_C:
		    sc->value = c_call(code)(sc, cdr(code));
		    break;
		    
		  case HOP_SAFE_C_S:
		    car(sc->T1_1) = finder(sc, cadr(code));
		    sc->value = c_call(code)(sc, sc->T1_1);
		    break;
		    
		  case HOP_SAFE_IS_PAIR_S:
		    sc->value = make_boolean(sc, is_pair(finder(sc, cadr(code))));
		    break;
		    
		  case HOP_SAFE_IS_SYMBOL_S:
		    sc->value = make_boolean(sc, is_symbol(finder(sc, cadr(code))));
		    break;
		    
		  case HOP_SAFE_NOT_S:
		    sc->value = make_boolean(sc, is_false(sc, finder(sc, cadr(code))));
		    break;
		    
		  case HOP_SAFE_C_SC:
		    args = cdr(code);
		    car(sc->T2_1) = finder(sc, car(args));
		    car(sc->T2_2) = cadr(args);
		    sc->value = c_call(code)(sc, sc->T2_1);
		    break;
		    
		  case HOP_SAFE_C_CS:
		    args = cdr(code);
		    car(sc->T2_2) = finder(sc, cadr(args));
		    car(sc->T2_1) = car(args);
		    sc->value = c_call(code)(sc, sc->T2_1);
		    break;
		    
		  case HOP_SAFE_C_SS:
		    args = cdr(code);
		    val = finder(sc, car(args));
		    car(sc->T2_2) = finder(sc, cadr(args));
		    car(sc->T2_1) = val;
		    sc->value = c_call(code)(sc, sc->T2_1);
		    break;
		    
		  case HOP_SAFE_C_SQ:
		    args = cdr(code);
		    car(sc->T2_1) = finder(sc, car(args));
		    car(sc->T2_2) = cadr(cadr(args));
		    sc->value = c_call(code)(sc, sc->T2_1);
		    break;
		    
		  case HOP_SAFE_C_opSq:
		    args = cadr(code);
		    car(sc->T1_1) = finder(sc, cadr(args));
		    car(sc->T1_1) = c_call(args)(sc, sc->T1_1);
		    sc->value = c_call(code)(sc, sc->T1_1);
		    break;
		    
		  case HOP_SAFE_C_opSq_opSq:
		    args = cdr(code);
		    car(sc->T1_1) = finder(sc, cadr(car(args)));
		    sc->temp4 = c_call(car(args))(sc, sc->T1_1);
		    args = cadr(args);
		    car(sc->T1_1) = finder(sc, cadr(args));
		    car(sc->T2_2) = c_call(args)(sc, sc->T1_1);
		    car(sc->T2_1) = sc->temp4;
		    sc->value = c_call(code)(sc, sc->T2_1);
		    break;
		  }
	      }
	    else
	      {
		/* here we can assume either #t or else */
		sc->value = code;
		sc->code = cdar(p);
		goto BEGIN;
	      }
	    if (is_true(sc, sc->value))
	      {
		sc->code = cdar(p);
		goto BEGIN;
	      }
	  }
	sc->value = sc->NIL;
	goto START;
      }
#endif      
      


      /* -------------------------------- AND -------------------------------- */
    case OP_AND:
      check_and(sc);
      if (is_null(sc->code))
	{
	  sc->value = sc->T;
	  goto START;
	}
      goto AND1;

    case OP_AND1:
      if ((is_false(sc, sc->value)) ||
	  (is_null(sc->code)))
	goto START;

    AND1:
    case OP_AND_UNCHECKED:
      {
	s7_pointer p;
	p = car(sc->code);
	if (!is_pair(p))
	  {
	    if (is_symbol(p))
	      sc->value = SYMBOL_TO_VALUE(sc, p);
	    else sc->value = p;

	    if ((is_false(sc, sc->value)) ||
		(is_null(cdr(sc->code))))
	      goto START;

	    sc->code = cdr(sc->code);
	    goto AND1;
	  }

	if (is_not_null(cdr(sc->code)))
	  push_stack_no_args(sc, OP_AND1, cdr(sc->code));
	sc->code = p;
	goto EVAL_PAIR;
      }


#if WITH_OPTIMIZATION
    AND_P1:
#endif
    case OP_AND_P1:
      if ((is_false(sc, sc->value)) ||
	  (is_null(sc->code)))
	{
	  /* counts[stack_op(sc->stack, s7_stack_top(sc) - 1)]++; */
	  goto START; 
	  /* lg 329702, OP_IF_PP: 101657
	   * in 3044565 OP_IF_PP: 2928624
	   */
	}
      /* fall through */


    AND_P:
#if (!WITH_OPTIMIZATION)

    case OP_AND_P:
      if (is_not_null(cdr(sc->code)))
	push_stack_no_args(sc, OP_AND_P1, cdr(sc->code)); 
      sc->code = car(sc->code);
      goto EVAL_PAIR;

#else

    case OP_AND_P:
      {
	s7_pointer code;
	code = car(sc->code);
	if (is_optimized(code))
	  {
	    s7_pointer args, val;
	    switch (optimize_data(code))
	      {
	      case HOP_SAFE_C_C:
		sc->value = c_call(code)(sc, cdr(code));
		break;
		
	      case HOP_SAFE_C_S:
		car(sc->T1_1) = finder(sc, cadr(code));
		sc->value = c_call(code)(sc, sc->T1_1);
		break;
		
	      case HOP_SAFE_IS_PAIR_S:
		sc->value = make_boolean(sc, is_pair(finder(sc, cadr(code))));
		break;
		
	      case HOP_SAFE_IS_SYMBOL_S:
		sc->value = make_boolean(sc, is_symbol(finder(sc, cadr(code))));
		break;
		
	      case HOP_SAFE_NOT_S:
		sc->value = make_boolean(sc, is_false(sc, finder(sc, cadr(code))));
		break;
		
	      case HOP_SAFE_C_SC:
		args = cdr(code);
		car(sc->T2_1) = finder(sc, car(args));
		car(sc->T2_2) = cadr(args);
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_CS:
		args = cdr(code);
		car(sc->T2_2) = finder(sc, cadr(args));
		car(sc->T2_1) = car(args);
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_SS:
		args = cdr(code);
		val = finder(sc, car(args));
		car(sc->T2_2) = finder(sc, cadr(args));
		car(sc->T2_1) = val;
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_SQ:
		args = cdr(code);
		car(sc->T2_1) = finder(sc, car(args));
		car(sc->T2_2) = cadr(cadr(args));
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_opSq:
		args = cadr(code);
		car(sc->T1_1) = finder(sc, cadr(args));
		car(sc->T1_1) = c_call(args)(sc, sc->T1_1);
		sc->value = c_call(code)(sc, sc->T1_1);
		break;
		
	      case HOP_SAFE_C_opSq_opSq:
		args = cdr(code);
		car(sc->T1_1) = finder(sc, cadr(car(args)));
		sc->temp4 = c_call(car(args))(sc, sc->T1_1);
		args = cadr(args);
		car(sc->T1_1) = finder(sc, cadr(args));
		car(sc->T2_2) = c_call(args)(sc, sc->T1_1);
		car(sc->T2_1) = sc->temp4;
		sc->value = c_call(code)(sc, sc->T2_1);
		break;

	      default:
		if (is_not_null(cdr(sc->code)))
		  push_stack_no_args(sc, OP_AND_P1, cdr(sc->code)); 
		sc->code = code;
		goto OPT_EVAL;
		/* lg: h_safe_c_opsafe_closure_opsq_sq: 18989 h_safe_c_opsqq: 5475 h_safe_c_opssq: 5002
		 * in: h_safe_c_pc: 323136 h_safe_c_s_opcq: 156159 h_safe_c_p: 53026 h_safe_c_opcsq: 18164
		 * s7: h_safe_c_ps: 17573 safe_c_s: 6693
		 * al: h_safe_c_opssq_s: 538901 h_safe_c_s_opssq: 52230
		 */
		break;
	      }
	    sc->code = cdr(sc->code);
	    goto AND_P1;
	  }
	else
	  {
	    /* lg: OP_OR_P: 186930 OP_SAFER_OR_X: 21227 OP_LET_O_P: 24685
	     * in: OP_OR_P: 161987 
	     * s7: OP_SAFER_OR_X: 34484
	     * al: OP_SAFER_OR_X: 49599
	     */
	    if (is_not_null(cdr(sc->code)))
	      push_stack_no_args(sc, OP_AND_P1, cdr(sc->code)); 
	    sc->code = code;
	    goto EVAL_PAIR;
	  }
      }


    case OP_SAFE_AND_S:
      if (sequence_is_safe_for_opteval(sc, sc->code))
	{
	  s7_pointer code;
	  car(sc->T1_1) = finder(sc, cadar(sc->code));
	  for (code = sc->code; is_pair(code); code = cdr(code))
	    {
	      sc->value = c_call(car(code))(sc, sc->T1_1);
	      if (is_false(sc, sc->value))
		goto START;
	    }
	  goto START;
	}
      else
	{
	  /* optimizer has already checked the (and) case */
	  car(ecdr(sc->code)) = sc->AND_UNCHECKED;
	  goto AND1;
	}
      goto START;
      
      
    case OP_SAFER_AND_X:
      {
	s7_pointer p;
	for (p = sc->code; is_pair(p); p = cdr(p))
	  {
	    s7_pointer val, code, args;
	    code = car(p);
	    switch (optimize_data(code))
	      {
	      case HOP_SAFE_C_C:
		sc->value = c_call(code)(sc, cdr(code));
		break;
		
	      case HOP_SAFE_C_S:
		car(sc->T1_1) = finder(sc, cadr(code));
		sc->value = c_call(code)(sc, sc->T1_1);
		break;
		
	      case HOP_SAFE_IS_PAIR_S:
		sc->value = make_boolean(sc, is_pair(finder(sc, cadr(code))));
		break;
		
	      case HOP_SAFE_IS_SYMBOL_S:
		sc->value = make_boolean(sc, is_symbol(finder(sc, cadr(code))));
		break;
		
	      case HOP_SAFE_NOT_S:
		sc->value = make_boolean(sc, is_false(sc, finder(sc, cadr(code))));
		break;
		
	      case HOP_SAFE_C_SC:
		args = cdr(code);
		car(sc->T2_1) = finder(sc, car(args));
		car(sc->T2_2) = cadr(args);
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_CS:
		args = cdr(code);
		car(sc->T2_2) = finder(sc, cadr(args));
		car(sc->T2_1) = car(args);
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_SS:
		args = cdr(code);
		val = finder(sc, car(args));
		car(sc->T2_2) = finder(sc, cadr(args));
		car(sc->T2_1) = val;
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_SQ:
		args = cdr(code);
		car(sc->T2_1) = finder(sc, car(args));
		car(sc->T2_2) = cadr(cadr(args));
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_opSq:
		args = cadr(code);
		car(sc->T1_1) = finder(sc, cadr(args));
		car(sc->T1_1) = c_call(args)(sc, sc->T1_1);
		sc->value = c_call(code)(sc, sc->T1_1);
		break;
		
	      case HOP_SAFE_C_opSq_opSq:
		args = cdr(code);
		car(sc->T1_1) = finder(sc, cadr(car(args)));
		sc->temp4 = c_call(car(args))(sc, sc->T1_1);
		args = cadr(args);
		car(sc->T1_1) = finder(sc, cadr(args));
		car(sc->T2_2) = c_call(args)(sc, sc->T1_1);
		car(sc->T2_1) = sc->temp4;
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      default:
		fprintf(stderr, "oops: safer_and_x: %s\n", DISPLAY_80(code));
		break;
	      }
	    if (is_false(sc, sc->value))
	      goto START;
	  }
	goto START;
      }
#endif
      
      
      /* -------------------------------- OR -------------------------------- */      
    case OP_OR:
      check_or(sc);
      if (is_null(sc->code)) 
	{
	  sc->value = sc->F;
	  goto START;
	}
      goto OR1;
      
    case OP_OR1:
      if ((is_true(sc, sc->value)) ||
	  (is_null(sc->code)))
	goto START;

    OR1:
    case OP_OR_UNCHECKED:
      if (!is_pair(car(sc->code)))
	{
	  if (is_symbol(car(sc->code)))
	    sc->value = finder(sc, car(sc->code));
	  else sc->value = car(sc->code);

	  if ((is_true(sc, sc->value)) ||
	      (is_null(cdr(sc->code))))
	    goto START;

	  sc->code = cdr(sc->code);
	  goto OR1;
	}

      if (is_not_null(cdr(sc->code)))
	push_stack_no_args(sc, OP_OR1, cdr(sc->code)); 
      sc->code = car(sc->code);
      goto EVAL_PAIR;


#if WITH_OPTIMIZATION
    OR_P1:
#endif
    case OP_OR_P1:
      if ((is_true(sc, sc->value)) ||
	  (is_null(sc->code)))
	goto START;
      /* fall through */

    OR_P:
#if (!WITH_OPTIMIZATION)

    case OP_OR_P:
      if (is_not_null(cdr(sc->code)))
	push_stack_no_args(sc, OP_OR_P1, cdr(sc->code)); 
      sc->code = car(sc->code);
      goto EVAL_PAIR;

#else

    case OP_OR_P:
      {
	s7_pointer code;
	code = car(sc->code);
	if (is_optimized(code))
	  {
	    s7_pointer args, val;
	    switch (optimize_data(code))
	      {
	      case HOP_SAFE_C_C:
		sc->value = c_call(code)(sc, cdr(code));
		break;
		
	      case HOP_SAFE_C_S:
		car(sc->T1_1) = finder(sc, cadr(code));
		sc->value = c_call(code)(sc, sc->T1_1);
		break;
		
	      case HOP_SAFE_IS_PAIR_S:
		sc->value = make_boolean(sc, is_pair(finder(sc, cadr(code))));
		break;
		
	      case HOP_SAFE_IS_SYMBOL_S:
		sc->value = make_boolean(sc, is_symbol(finder(sc, cadr(code))));
		break;
		
	      case HOP_SAFE_NOT_S:
		sc->value = make_boolean(sc, is_false(sc, finder(sc, cadr(code))));
		break;
		
	      case HOP_SAFE_C_SC:
		args = cdr(code);
		car(sc->T2_1) = finder(sc, car(args));
		car(sc->T2_2) = cadr(args);
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_CS:
		args = cdr(code);
		car(sc->T2_2) = finder(sc, cadr(args));
		car(sc->T2_1) = car(args);
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_SS:
		args = cdr(code);
		val = finder(sc, car(args));
		car(sc->T2_2) = finder(sc, cadr(args));
		car(sc->T2_1) = val;
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_SQ:
		args = cdr(code);
		car(sc->T2_1) = finder(sc, car(args));
		car(sc->T2_2) = cadr(cadr(args));
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_opSq:
		args = cadr(code);
		car(sc->T1_1) = finder(sc, cadr(args));
		car(sc->T1_1) = c_call(args)(sc, sc->T1_1);
		sc->value = c_call(code)(sc, sc->T1_1);
		break;
		
	      case HOP_SAFE_C_opSq_opSq:
		args = cdr(code);
		car(sc->T1_1) = finder(sc, cadr(car(args)));
		sc->temp4 = c_call(car(args))(sc, sc->T1_1);
		args = cadr(args);
		car(sc->T1_1) = finder(sc, cadr(args));
		car(sc->T2_2) = c_call(args)(sc, sc->T1_1);
		car(sc->T2_1) = sc->temp4;
		sc->value = c_call(code)(sc, sc->T2_1);
		break;

	      default:
		if (is_not_null(cdr(sc->code)))
		  push_stack_no_args(sc, OP_OR_P1, cdr(sc->code)); 
		sc->code = code;
		goto OPT_EVAL;
		/* s7: h_safe_c_p: 4144 safe_c_opscq: 1632
		 * in: h_safe_c_p: 80268 
		 * al: safe_closure_s: 439875 h_safe_c_opssq: 23021
		 * lg: closure_opsq_s: 56130 h-call-with-exit: 35698
		 */
		break;
	      }
	    sc->code = cdr(sc->code);
	    goto OR_P1;
	  }
	else
	  {
	    /* s7: OP_AND_P: 69823 OP_SAFER_AND_X: 52659
	     * in: OP_AND_UNCHECKED: 91409
	     * al: OP_AND_P: 549112
	     * lg: OP_AND_P: 271373
	     */
	    if (is_not_null(cdr(sc->code)))
	      push_stack_no_args(sc, OP_OR_P1, cdr(sc->code)); 
	    sc->code = code;
	    goto EVAL_PAIR;
	  }
      }


    case OP_SAFE_OR_S:
      if (!sequence_is_safe_for_opteval(sc, sc->code))
	{
	  car(ecdr(sc->code)) = sc->OR_UNCHECKED;
	  goto OR1;
	}
      /* fall through */
      
    case OP_SAFER_OR_S:
      {
	s7_pointer code;
	car(sc->T1_1) = finder(sc, cadar(sc->code));
	for (code = sc->code; is_pair(code); code = cdr(code))
	  {
	    sc->value = c_call(car(code))(sc, sc->T1_1);
	    if (is_true(sc, sc->value))
	      goto START;
	  }
	goto START;
      }
      

    case OP_SAFER_OR_CEQ:
    case OP_SAFER_OR_C:
      {
	s7_pointer code;
	for (code = sc->code; is_pair(code); code = cdr(code))
	  {
	    sc->value = c_call(car(code))(sc, cdar(code));
	    if (is_true(sc, sc->value))
	      goto START;
	  }
	goto START;
      }
      
      
    case OP_SAFER_OR_X:
      {
	s7_pointer p;
	for (p = sc->code; is_pair(p); p = cdr(p))
	  {
	    s7_pointer val, code, args;
	    code = car(p);
	    switch (optimize_data(code))
	      {
	      case HOP_SAFE_C_C:
		sc->value = c_call(code)(sc, cdr(code));
		break;
		
	      case HOP_SAFE_C_S:
		car(sc->T1_1) = finder(sc, cadr(code));
		sc->value = c_call(code)(sc, sc->T1_1);
		break;
		
	      case HOP_SAFE_IS_PAIR_S:
		sc->value = make_boolean(sc, is_pair(finder(sc, cadr(code))));
		break;
		
	      case HOP_SAFE_IS_SYMBOL_S:
		sc->value = make_boolean(sc, is_symbol(finder(sc, cadr(code))));
		break;
		
	      case HOP_SAFE_NOT_S:
		sc->value = make_boolean(sc, is_false(sc, finder(sc, cadr(code))));
		break;
		
	      case HOP_SAFE_C_SC:
		args = cdr(code);
		car(sc->T2_1) = finder(sc, car(args));
		car(sc->T2_2) = cadr(args);
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_CS:
		args = cdr(code);
		car(sc->T2_2) = finder(sc, cadr(args));
		car(sc->T2_1) = car(args);
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_SS:
		args = cdr(code);
		val = finder(sc, car(args));
		car(sc->T2_2) = finder(sc, cadr(args));
		car(sc->T2_1) = val;
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_SQ:
		args = cdr(code);
		car(sc->T2_1) = finder(sc, car(args));
		car(sc->T2_2) = cadr(cadr(args));
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      case HOP_SAFE_C_opSq:
		args = cadr(code);
		car(sc->T1_1) = finder(sc, cadr(args));
		car(sc->T1_1) = c_call(args)(sc, sc->T1_1);
		sc->value = c_call(code)(sc, sc->T1_1);
		break;
		
	      case HOP_SAFE_C_opSq_opSq:
		args = cdr(code);
		car(sc->T1_1) = finder(sc, cadr(car(args)));
		sc->temp4 = c_call(car(args))(sc, sc->T1_1);
		args = cadr(args);
		car(sc->T1_1) = finder(sc, cadr(args));
		car(sc->T2_2) = c_call(args)(sc, sc->T1_1);
		car(sc->T2_1) = sc->temp4;
		sc->value = c_call(code)(sc, sc->T2_1);
		break;
		
	      default:
		fprintf(stderr, "oops: safer_or_x: %s\n", DISPLAY_80(code));
		break;
	      }
	    
	    if (is_true(sc, sc->value))
	      goto START;
	  }
	goto START;
      }
#endif
      
      
      /* (define* (char= . args) (or (< (length args) 2) (apply char=? args)))
       */
      
      /* by going direct without a push_stack on the last one we get "tail recursion",
       *   but if the last arg (also in "and" above) is "values", there is a slight
       *   inconsistency: the values are returned and spliced into the caller if trailing, but
       *   are spliced into the "or" if not trailing, so
       *
       *   (+ 10 (or (values 1 2) #f))
       *   11
       *   (+ 10 (or #f (values 1 2)))
       *   13
       *   (+ 10 (or (or #f (values 1 2)) #f))
       *   11
       *
       * The tail recursion is more important.  This behavior matches that of "begin" -- if the
       * values statement is last, it splices into the next outer arglist.
       *
       * I tried implementing dynamic-binding by walking up the stack looking at envs, but
       *   for that to work in some cases, every new frame needs to be on the stack, hence
       *   OP_LET_UNWIND.  But that means the stack grows on every let, so tail recursion is
       *   defeated.  I'd have to notice tail calls and pop the stack, but I decided that
       *   was too much work for a minor (unneeded) feature.  Although the stack does not
       *   grow, we're depending on the GC to take care of the old local envs after the
       *   tail call.  My test:
       *
       * (let ((max-stack 0))
       *   (define (tc-1 a c) 
       *     (let ((b (+ a 1)) 
       *           (d (make-vector 1000)))
       *       (if (> (s7-stack-size) max-stack)
       *           (set! max-stack (s7-stack-size)))
       *       (if (< b c) 
       *           (tc-1 b c))))
       * (tc-1 0 3200000) max-stack)
       * 4
       *
       * indicated (via "top") that the memory footprint was not growing, and the "4"
       * shows the stack was ok, so I currently believe that s7 is doing tail calls
       * correctly.
       */


      /* -------------------------------- macro evaluation -------------------------------- */

    case OP_EVAL_MACRO:    /* after (scheme-side) macroexpansion, evaluate the resulting expression */
      /* 
       * (define-macro (hi a) `(+ ,a 1))
       * (hi 2)
       * here with value: (+ 2 1)
       */
      sc->code = sc->value;
      goto EVAL;


    case OP_DEFMACRO:
    case OP_DEFMACRO_STAR:
      /* defmacro(*) could be defined in terms of define-macro(*), but I guess this gives us better error messages */
      check_defmacro(sc);

      if (symbol_has_accessor(car(sc->code)))
	{
	  sc->value = bind_accessed_symbol(sc, OP_DEFMACRO_WITH_ACCESSOR, car(sc->code), sc->code);
	  if (sc->value == sc->NO_VALUE)
	    goto APPLY;
	  sc->code = sc->value;
	}

    DEFMACRO2:
      sc->x = car(sc->code);
      sc->z = cdr(sc->code);
      sc->y = s7_gensym(sc, "defmac");
      /* (defmacro hi (a b) `(+ ,a ,b)) becomes:
       *   sc->x: hi
       *   code: (lambda (defmac-21) 
       *               (apply (lambda (a b) 
       *                        (cons (quote +) (cons a (cons b (quote ()))))) 
       *                      (cdr defmac-21)))
       */
      NEW_CELL_NO_CHECK(sc, sc->value);
      closure_args(sc->value) = list_1(sc, sc->y);
      closure_body(sc->value) = list_1(sc, 
				       cons(sc, 
					    sc->APPLY,
					    cons(sc, 
						 cons(sc, 
						      (sc->op == OP_DEFMACRO_STAR) ? sc->LAMBDA_STAR : sc->LAMBDA,
						      sc->z),
						 list_1(sc, list_2(sc, sc->CDR, sc->y)))));
      closure_environment(sc->value) = sc->envir;
      sc->code = sc->x;
      sc->y = sc->NIL;
      sc->z = sc->NIL;
      goto MACRO;


    case OP_DEFMACRO_WITH_ACCESSOR:
      if (sc->value == sc->ERROR) /* backwards compatibility... */
	return(s7_error(sc, sc->ERROR,
			list_3(sc, make_protected_string(sc, "can't defmacro ~S to ~S"), car(sc->args), cadr(sc->args))));
      sc->code = sc->value;
      goto DEFMACRO2;


    case OP_DEFINE_MACRO_WITH_ACCESSOR:
      if (sc->value == sc->ERROR) /* backwards compatibility... */
	return(s7_error(sc, sc->ERROR,
			list_3(sc, make_protected_string(sc, "can't define-macro ~S to ~S"), car(sc->args), cadr(sc->args))));
      sc->code = sc->value;
      goto DEFINE_MACRO2;


    case OP_DEFINE_BACRO:
    case OP_DEFINE_BACRO_STAR:
    case OP_DEFINE_EXPANSION:
    case OP_DEFINE_MACRO:
    case OP_DEFINE_MACRO_STAR:
      check_define_macro(sc);

      if (symbol_has_accessor(caar(sc->code)))
	{
	  sc->value = bind_accessed_symbol(sc, OP_DEFINE_MACRO_WITH_ACCESSOR, caar(sc->code), sc->code);
	  if (sc->value == sc->NO_VALUE)
	    goto APPLY;
	  sc->code = sc->value;
	}

      /* (define-macro (hi a) `(+ ,a 1))
       *   in this case we want cadr, not caddr of defmacro
       */

    DEFINE_MACRO2:
      sc->x = caar(sc->code);
      sc->z = cdr(sc->code);
      sc->y = s7_gensym(sc, "defmac");
      /* (define-macro (hi a b) `(+ ,a ,b)) becomes:
       *   sc->x: hi
       *   sc->code: (lambda ({defmac}-14) (apply (lambda (a b) ({list} '+ a b)) (cdr {defmac}-14)))
       */
      NEW_CELL_NO_CHECK(sc, sc->value);
      closure_args(sc->value) = list_1(sc, sc->y);
      closure_body(sc->value) = list_1(sc, 
					 cons(sc, 
					      sc->APPLY,
					      cons(sc, 
						   cons(sc, 
							((sc->op == OP_DEFINE_MACRO_STAR) || 
							 (sc->op == OP_DEFINE_BACRO_STAR)) ? sc->LAMBDA_STAR : sc->LAMBDA,
							cons(sc,                /* cdar(sc->code) is the parameter list */
							     ((sc->op == OP_DEFINE_MACRO_STAR) || 
							      (sc->op == OP_DEFINE_BACRO_STAR)) ? quotify(sc, cdar(sc->code)) : cdar(sc->code),
							     sc->z)),
						   list_1(sc, list_2(sc, sc->CDR, sc->y)))));
      closure_environment(sc->value) = sc->envir;
      sc->code = sc->x; 
      sc->y = sc->NIL;
      sc->z = sc->NIL;

    MACRO:
      /* symbol? macro name has already been checked */
      /* find name in environment, and define it */
      sc->x = find_local_symbol(sc, sc->envir, sc->code); 
      if (is_slot(sc->x))
	slot_set_value(sc->x, sc->value); 
      else s7_make_slot(sc, sc->envir, sc->code, sc->value); /* was current but we've checked immutable already */

      if ((sc->op == OP_DEFINE_BACRO) ||
	  (sc->op == OP_DEFINE_BACRO_STAR))
	set_type(sc->value, T_BACRO | T_ANY_MACRO | T_DONT_EVAL_ARGS | T_COPY_ARGS);
      else
	{
	  if (sc->op == OP_DEFINE_EXPANSION)
	    {
	      set_type(sc->value, T_MACRO | T_ANY_MACRO | T_EXPANSION | T_DONT_EVAL_ARGS | T_COPY_ARGS);
	      set_type(sc->code, EXPANSION_TYPE);
	    }
	  else set_type(sc->value, T_MACRO | T_ANY_MACRO | T_DONT_EVAL_ARGS | T_COPY_ARGS); 
	}

      sc->value = sc->code;
      goto START;
      
      
    case OP_LAMBDA: 
      /* code is a lambda form minus the "lambda": ((a b) (+ a b)) */
      /* this includes unevaluated symbols (direct symbol table refs) in macro arg list */
      {
	s7_pointer code;
	code = sc->code;

	if ((!is_pair(code)) ||
	    (!is_pair(cdr(code))))                               /* (lambda) or (lambda #f) or (lambda . 1) */
	  return(eval_error(sc, "lambda: no args or no body? ~A", code));
	check_lambda_args(sc, car(code));
	
	if ((is_pair(cdr(code))) &&
	    (is_pair(cadr(code))) &&
	    (is_null(cddr(code))))
	  set_one_liner(cdr(code));
	
#if WITH_OPTIMIZATION
	optimize(sc, cdr(code), 0, sc->NIL);
#endif
	if ((is_overlaid(code)) &&
	    (cdr(ecdr(code)) == code))
	  car(ecdr(code)) = sc->LAMBDA_UNCHECKED;
      }

    case OP_LAMBDA_UNCHECKED:
      MAKE_CLOSURE(sc, sc->value, car(sc->code), cdr(sc->code), sc->envir);
      goto START;


    case OP_LAMBDA_STAR:
      if ((!is_pair(sc->code)) ||
	  (!is_pair(cdr(sc->code))))                                          /* (lambda*) or (lambda* #f) */
	return(eval_error(sc, "lambda*: no args or no body? ~A", sc->code));
      check_lambda_star_args(sc, car(sc->code));

      if ((is_pair(cdr(sc->code))) &&
	  (is_pair(cadr(sc->code))) &&
	  (is_null(cddr(sc->code))))
	set_one_liner(cdr(sc->code));

#if WITH_OPTIMIZATION
      optimize(sc, cdr(sc->code), 0, sc->NIL);
#endif
      if ((is_overlaid(sc->code)) &&
	  (cdr(ecdr(sc->code)) == sc->code))
	{
	  car(ecdr(sc->code)) = sc->LAMBDA_STAR_UNCHECKED;
	}

    case OP_LAMBDA_STAR_UNCHECKED:
      sc->value = make_closure(sc, car(sc->code), cdr(sc->code), T_CLOSURE_STAR);
      goto START;
      


      /* -------------------------------- CASE -------------------------------- */

#if WITH_OPTIMIZATION
    case OP_SAFE_CASE_S:
      car(sc->T1_1) = finder(sc, cadr(car(sc->code)));
      sc->value = c_call(car(sc->code))(sc, sc->T1_1);
      sc->code = cdr(sc->code);
      goto CASE1;
#endif


    case OP_CASE:      /* case, car(sc->code) is the selector */
      check_case(sc);


    case OP_CASE_UNCHECKED:
      {
	s7_pointer carc;
	carc = car(sc->code);
	if (!is_pair(carc))
	  {
	    if (is_symbol(carc))
	      sc->value = finder(sc, carc);
	    else sc->value = carc;
	    sc->code = cdr(sc->code);
	    /* fall through */
	  }
	else
	  {
	    push_stack_no_args(sc, OP_CASE1, cdr(sc->code));
	    sc->code = carc;
	    goto EVAL_PAIR;
	  }
      }
      

#if WITH_OPTIMIZATION      
    CASE1:
#endif
    case OP_CASE1: 
      {
	s7_pointer x, y;
	if (is_simple(sc->value))
	  {
	    for (x = sc->code; is_pair(x); x = cdr(x))
	      {
		y = caar(x);
		if (!is_pair(y))
		  goto ELSE_CASE;
		do {
		  if (car(y) == sc->value)
		    goto ELSE_CASE;
		  y = cdr(y);
		} while (is_pair(y));
	      }
	  }
	else
	  {
	    for (x = sc->code; is_pair(x); x = cdr(x)) 
	      {
		y = caar(x);
		if (!is_pair(y))
		  goto ELSE_CASE;
		for (; is_pair(y); y = cdr(y)) 
		  if (s7_is_eqv(car(y), sc->value)) 
		    goto ELSE_CASE;
	      }
	  }
	/* x is the entire matching clause, (case 2 ((2) 3)), x: (((2) 3)) */
      ELSE_CASE:
	if (is_not_null(x))
	  {
	    sc->code = cdar(x);
	    
	    /* check for => */
	    if ((car(sc->code) == sc->FEED_TO) &&
		(s7_symbol_value(sc, sc->FEED_TO) == sc->UNDEFINED))
	      {
		sc->code = list_2(sc, cadr(sc->code), list_2(sc, sc->QUOTE, sc->value)); 
		goto EVAL_PAIR;
	      }
	    goto BEGIN;
	  }
	
	/* no match found */
	sc->value = sc->UNSPECIFIED; /* this was sc->NIL but the spec says case value is unspecified if no clauses match */
	goto START;
      }


    case OP_CASE_SIMPLE: 
      /* assume symbol as selector, all keys are simple, and no =>
       */
      {
	s7_pointer x, y, selector;
	selector = finder(sc, car(sc->code));
	for (x = cdr(sc->code); is_pair(x); x = cdr(x))
	  {
	    y = caar(x);
	    if (!is_pair(y)) /* else? */
	      {
		sc->code = cdar(x);
		goto BEGIN;
	      }
	    do {
	      if (car(y) == selector)
		{
		  sc->code = cdar(x);
		  goto BEGIN;
		}
	      y = cdr(y);
	    } while (is_pair(y));
	  }
	sc->value = sc->UNSPECIFIED;
	goto START;
      }


    case OP_CASE_INT: 
      /* assume symbol as selector, all keys are single and int, and no =>
       * ideally we'd just jump to the code we want indexed via the selector,
       * but I can't find any place to store the vector.
       */
      {
	s7_pointer x, y, selector;
	selector = finder(sc, car(sc->code));
	if (!is_integer(selector))
	  {
	    for (x = cdr(sc->code); is_pair(x); x = cdr(x))
	      {
		y = caar(x);
		if (!is_pair(y))
		  {
		    sc->code = cdar(x);
		    goto BEGIN;
		  }
	      }
	  }
	else
	  {
	    s7_Int index;
	    index = integer(selector);
	    for (x = cdr(sc->code); is_pair(x); x = cdr(x))
	      {
		y = caar(x);
		if (!is_pair(y))
		  {
		    sc->code = cdar(x);
		    goto BEGIN;
		  }
		if (integer(car(y)) == index)
		  {
		    sc->code = cdar(x);
		    goto BEGIN;
		  }
	      }
	  }
	sc->value = sc->UNSPECIFIED;
	goto START;
      }


    case OP_CASE_SIMPLER: 
      /* assume symbol as selector, all keys are simple, and no => and no else
       */
      {
	s7_pointer x, y, selector;
	selector = finder(sc, car(sc->code));
	for (x = cdr(sc->code); is_pair(x); x = cdr(x))
	  {
	    y = caar(x);
	    do {
	      if (car(y) == selector)
		{
		  sc->code = cdar(x);
		  goto BEGIN;
		}
	      y = cdr(y);
	    } while (is_pair(y));
	  }
	sc->value = sc->UNSPECIFIED;
	goto START;
      }


    case OP_CASE_SIMPLEST: 
      /* assume symbol as selector, all keys are simple and singletons, and no => and no else
       */
      {
	s7_pointer x, selector;
	selector = finder(sc, car(sc->code));
	for (x = cdr(sc->code); is_pair(x); x = cdr(x))
	  if (fcdr(x) == selector)
	    {
	      sc->code = cdar(x);
	      goto BEGIN;
	    }
	sc->value = sc->UNSPECIFIED;
	goto START;
      }



    case OP_ERROR_QUIT: 
    case OP_EVAL_DONE:
      /* this is the "time to quit" operator */
      return(sc->F);
      break;
      

    case OP_BARRIER:
    case OP_CATCH:
      goto START;


    case OP_DEACTIVATE_GOTO:
      call_exit_active(sc->args) = false;      /* as we leave the call-with-exit body, deactivate the exiter */
      goto START;


    case OP_TRACE_RETURN:
      trace_return(sc);
      goto START;
      
      
    case OP_TRACE_HOOK_QUIT:
      tracing = true;                 /* this was turned off before calling the *trace-hook* functions */
      /* sc->args was copied in trace_apply */
      goto APPLY; /* goto apply but not trace */

      
    case OP_ERROR_HOOK_QUIT:
      hook_functions(sc->error_hook) = sc->code;  /* restore old value */

      /* now mimic the end of the normal error handler.  Since this error hook evaluation can happen
       *   in an arbitrary s7_call nesting, we can't just return from the current evaluation --
       *   we have to jump to the original (top-level) call.  Otherwise '#<unspecified> or whatever
       *   is simply treated as the (non-error) return value, and the higher level evaluations
       *   get confused.
       */
      stack_reset(sc);
      sc->op = OP_ERROR_QUIT;
      if (sc->longjmp_ok)
	{
	  longjmp(sc->goto_start, 1);
	}
      return(sc->value); /* not executed I hope */

      
    case OP_GET_OUTPUT_STRING:          /* from call-with-output-string and with-output-to-string -- return the string */
      sc->value = s7_make_string(sc, s7_get_output_string(sc, sc->code));
      goto START;


    case OP_UNWIND_OUTPUT:
      {
	bool is_file;
	is_file = is_file_port(sc->code);

	if ((is_output_port(sc->code)) &&
	    (!port_is_closed(sc->code)))
	  s7_close_output_port(sc, sc->code); /* may call fflush */

	if ((is_output_port(sc->args)) &&
	    (!port_is_closed(sc->args)))
	  sc->output_port = sc->args;
       
       if (is_file)
	 {
	   if (is_multiple_value(sc->value)) 
	     sc->value = splice_in_values(sc, multiple_value(sc->value));
	 }
       goto START;
      }


    case OP_UNWIND_INPUT:
      if ((is_input_port(sc->code)) &&
	  (!port_is_closed(sc->code)))
	s7_close_input_port(sc, sc->code);

      if ((is_input_port(sc->args)) &&
	  (!port_is_closed(sc->args)))
	sc->input_port = sc->args;

      if (is_multiple_value(sc->value)) 
	sc->value = splice_in_values(sc, multiple_value(sc->value));
      goto START;


    case OP_DYNAMIC_WIND:
      if (dynamic_wind_state(sc->code) == DWIND_INIT)
	{
	  dynamic_wind_state(sc->code) = DWIND_BODY;
	  push_stack(sc, OP_DYNAMIC_WIND, sc->NIL, sc->code);
	  sc->args = sc->NIL;
	  sc->code = dynamic_wind_body(sc->code);
	  goto APPLY;
	}
      else
	{
	  if (dynamic_wind_state(sc->code) == DWIND_BODY)
	    {
	      dynamic_wind_state(sc->code) = DWIND_FINISH;
	      push_stack(sc, OP_DYNAMIC_WIND, sc->value, sc->code);
	      sc->args = sc->NIL;
	      sc->code = dynamic_wind_out(sc->code);
	      goto APPLY;
	    }
	  else
	    {
	      /* (+ 1 (dynamic-wind (lambda () #f) (lambda () (values 2 3 4)) (lambda () #f)) 5) */
	      if (is_multiple_value(sc->args))
		sc->value = splice_in_values(sc, multiple_value(sc->args));
	      else sc->value = sc->args;                         /* value saved above */ 
	      goto START;
	    }
	}
      break;
      

    case OP_WITH_ENV:
      /* (with-environment env . body) */
      if (!is_pair(sc->code))                            /* (with-environment . "hi") */
	return(eval_error(sc, "with-environment takes an environment argument: ~A", sc->code));
      if (!is_pair(cdr(sc->code)))
	return(eval_error(sc, "with-environment body is messed up: ~A", sc->code));

      push_stack(sc, OP_WITH_ENV1, sc->NIL, cdr(sc->code));
      sc->code = car(sc->code);                          /* eval env arg */
      goto EVAL;

      
    case OP_WITH_ENV1:
      if (!is_environment(sc->value))                    /* (with-environment . "hi") */
	return(eval_error(sc, "with-environment takes an environment argument: ~A", sc->value));

      if (sc->value == sc->global_env)
	{
	  NEW_FRAME(sc, sc->NIL, sc->envir);             /* otherwise, find_symbol_or_bust can die because it assumes sc->envir is ok */	
	}
      else sc->envir = sc->value;
      /* body is implicit in stack -- sc->code is ready to go */
      goto BEGIN;


    case OP_EXPANSION:
      /* after the expander has finished, we need to add some annotations
       */
      annotate_expansion(sc, sc->value);
      goto START;


    READ_LIST:
    case OP_READ_LIST: 
      /* sc->args is sc->NIL at first */
      /*    was: sc->args = cons(sc, sc->value, sc->args); */ 
      {
	s7_pointer x;
	NEW_CELL(sc, x);
	car(x) = sc->value;
	cdr(x) = sc->args;
	set_type(x, T_PAIR);
	sc->args = x;
      }
      
      /* sc->tok = token(sc); */
      /* this is 75% of the token calls, so expanding it saves lots of time */
      {
	int c;
	s7_pointer pt;

	pt = sc->input_port;
	c = port_read_white_space(pt)(sc, pt);

	switch (c) 
	  {
	  case '(':
	    sc->tok = TOKEN_LEFT_PAREN;
	    push_stack(sc, OP_READ_LIST, sc->args, sc->NIL);
	    sc->value = read_expression(sc);
	    goto START;
	    /* this is almost always OP_READ_LIST, but handling that explicitly is slower */
      
	  case ')':
	    sc->tok = TOKEN_RIGHT_PAREN;
	    break;
	    
	  case '.':
	    sc->tok = read_dot(sc, pt); /* dot or atom */
	    break;
	    
	  case '\'':
	    sc->tok = TOKEN_QUOTE;
	    push_stack(sc, OP_READ_LIST, sc->args, sc->NIL);
	    sc->value = read_expression(sc);
	    goto START;
	    
	  case ';':
	    sc->tok = port_read_semicolon(pt)(sc, pt);
	    break;
	    
	  case '"':
	    sc->tok = TOKEN_DOUBLE_QUOTE;
	    sc->value = read_string_constant(sc, sc->input_port);
	    if (sc->value == sc->F)                                /* can happen if input code ends in the middle of a string */
	      return(read_error(sc, "end of input encountered while in a string"));
	    if (sc->value == sc->T)
	      return(read_error(sc, "unknown backslash usage -- perhaps you meant two backslashes?"));
	    goto READ_LIST;
	    
	  case '`':
	    sc->tok = TOKEN_BACK_QUOTE;
	    push_stack(sc, OP_READ_LIST, sc->args, sc->NIL);
	    sc->value = read_expression(sc);
	    goto START;
	    
	  case ',':
	    sc->tok = read_comma(sc, pt); /* at_mark or comma */
	    push_stack(sc, OP_READ_LIST, sc->args, sc->NIL);
	    sc->value = read_expression(sc);
	    goto START;
	    
	  case '#':
	    sc->tok = read_sharp(sc, pt);
	    break;

	  case EOF:
	    return(missing_close_paren_error(sc));
	    
	  default: 
	    sc->strbuf[0] = c; 
	    /* sc->tok = TOKEN_ATOM; */
	    sc->value = port_read_name(pt)(sc, pt, NO_SHARP);
	    goto READ_LIST;
	  }
      }

    READ_TOK:
      switch (sc->tok)
	{
	case TOKEN_RIGHT_PAREN:
	  /* sc->args can't be null here */
	  
	  sc->value = safe_reverse_in_place(sc, sc->args);
	  pair_line_number(sc->value) = port_line_number(sc->input_port) | (port_file_number(sc->input_port) << 20);
	  /* I think sc->input_port can't be nil -- it falls back on stdin now */

	  /* read-time macro expansion
	   *
	   *   (defmacro hi (a) (format #t "hi...") `(+ ,a 1))
	   *   (define (ho b) (+ 1 (hi b)))
	   *
	   * here sc->value is: (ho b), (hi b), (+ 1 (hi b)), (define (ho b) (+ 1 (hi b)))
	   * 
	   * but... first we can't tell for sure at this point that "hi" really is a macro
	   *
	   *   (letrec ((hi ... (hi...))) will be confused about the 2nd hi,
	   *   or (call/cc (lambda (hi) (hi 1))) etc.
	   *
	   * second, figuring out that we're quoted is not easy -- we have to march all the
	   * way to the bottom of the stack looking for op_read_quote or op_read_vector
	   *
	   *    #(((hi)) 2) or '(((hi)))
	   *
	   * or op_read_list with args not equal (quote) or (macroexapand)
	   *
	   *    '(hi 3) or (macroexpand (hi 3) or (quote (hi 3))
	   *
	   * and those are only the problems I noticed!
	   *
	   * The hardest of these problems involve shadowing, so Rick asked for "define-expansion"
	   *   which is just like define-macro, but the programmer guarantees that the macro
	   *   name will not be shadowed.  So I'll also assume that the other funny cases are
	   *   being avoided -- see what happens!
	   *
	   *   (define-expansion (hi a) `(+ ,a 1))
	   *   (define (ho b) (+ 1 (hi b)))
	   *   (procedure-source ho) -> (lambda (b) (+ 1 (+ b 1)))
	   */
	  
	  if (is_expansion(car(sc->value)))                                  /* was checking the T_EXPANSION bit (car(sc->value) is a symbol) */
	    {
	      int loc;

	      loc = s7_stack_top(sc) - 1;
	      if ((loc >= 3) &&
		  (stack_op(sc->stack, loc) != OP_READ_QUOTE) &&             /* '(hi 1) for example */
		  (car(stack_args(sc->stack, loc)) != sc->QUOTE) &&          /* (quote (hi 1)) */
		  (car(stack_args(sc->stack, loc)) != sc->MACROEXPAND))      /* (macroexpand (hi 1)) */
		{
		  s7_pointer x;
		  x = slot_value(find_symbol(sc, car(sc->value)));
		  sc->args = list_1(sc, sc->value); 
		  sc->code = x;
		  push_stack(sc, OP_EXPANSION, sc->NIL, sc->NIL);
		  goto APPLY;
		}
	    }
	  
	  if ((is_symbol(car(sc->value))) &&
	      (is_pair(cdr(sc->value))))
	    {
	      ecdr(cdr(sc->value)) = sc->value;
	      set_overlay(cdr(sc->value));
	    }
	  /* I think it's too soon to scan every list at this point for symbols and pairs */
	  break;

	case TOKEN_EOF:
	  /* can't happen, I believe */
	  return(missing_close_paren_error(sc));

	case TOKEN_ATOM:
	  sc->value = port_read_name(sc->input_port)(sc, sc->input_port, NO_SHARP);
	  goto READ_LIST;
	  break;

	case TOKEN_SHARP_CONST:
	  sc->value = port_read_name(sc->input_port)(sc, sc->input_port, WITH_SHARP);
	  if (is_null(sc->value))
	    return(read_error(sc, "undefined # expression"));
	  if (sc->value == sc->NO_VALUE)
	    {
	      /* (set! *#readers* (cons (cons #\; (lambda (s) (read) (values))) *#readers*))
	       * (+ 1 #;(* 2 3) 4)
	       * so we need to get the next token, act on it without any assumptions about read list
	       */
	      sc->tok = token(sc);
	      goto READ_TOK;
	    }
	  goto READ_LIST;
	  break;

	case TOKEN_DOUBLE_QUOTE:
	  sc->value = read_string_constant(sc, sc->input_port);
	  if (sc->value == sc->F)                                /* can happen if input code ends in the middle of a string */
	    return(read_error(sc, "end of input encountered while in a string"));
	  if (sc->value == sc->T)
	    return(read_error(sc, "unknown backslash usage -- perhaps you meant two backslashes?"));
	  goto READ_LIST;
	  break;

	case TOKEN_DOT:
	  push_stack(sc, OP_READ_DOT, sc->args, sc->NIL);
	  sc->tok = token(sc);
	  sc->value = read_expression(sc);
	  break;

	default:
	  /* by far the main case here is TOKEN_LEFT_PAREN, but it doesn't save anything to move it to this level */
	  push_stack(sc, OP_READ_LIST, sc->args, sc->NIL);
	  sc->value = read_expression(sc);
	  /* check for op_read_list here and explicit pop_stack are slower */
	  break;
	}
      goto START;

      
    case OP_READ_DOT:
      if (token(sc) != TOKEN_RIGHT_PAREN)
	{
	  back_up_stack(sc);
	  read_error(sc, "stray dot?");            /* (+ 1 . 2 3) or (list . ) */
	}
      /* args = previously read stuff, value = thing just after the dot and before the ')':
       *   (list 1 2 . 3)
       *   value: 3, args: (2 1 list) 
       *   '(1 . 2)
       *   value: 2, args: (1)
       *
       * but we also get here in a lambda arg list:
       *   (lambda (a b . c) #f)
       *   value: c, args: (b a)
       *
       * so we have to leave any error checks until later, I guess
       *   -- in eval_args1, if we end with non-pair-not-nil then
       *      something is fishy
       */
      sc->value = reverse_in_place(sc, sc->value, sc->args);
      goto START;
      
      
    case OP_READ_QUOTE:
      /* can't check for sc->value = sc->NIL here because we want ''() to be different from '() */
      sc->value = list_2(sc, sc->QUOTE, sc->value);
      ecdr(cdr(sc->value)) = sc->value;
      set_overlay(cdr(sc->value));
      goto START;      
      
      
    case OP_READ_QUASIQUOTE:
      /* this was pushed when the backquote was seen, then eventually we popped back to it */
      sc->value = g_quasiquote_1(sc, sc->value);
      goto START;
      
      
    case OP_READ_VECTOR:
      if (!is_proper_list(sc, sc->value))       /* #(1 . 2) */
	return(read_error(sc, "vector constant data is not a proper list"));

      if (sc->args == small_int(1))
	sc->value = g_vector(sc, sc->value);
      else sc->value = g_multivector(sc, (int)s7_integer(sc->args), sc->value);
      goto START;

      
    case OP_READ_QUASIQUOTE_VECTOR:
      /* this works only if the quasiquoted list elements can be evaluated in the read-time environment.
       *
       *    `#(1 ,@(list 1 2) 4) -> (apply vector ({list} 1 ({apply} {values} (list 1 2)) 4)) -> #(1 1 2 4)
       *
       * Originally, I used:
       *
       *   sc->value = list_3(sc, sc->APPLY, sc->VECTOR, g_quasiquote_1(sc, sc->value));
       *   goto START;
       *
       * which means that #(...) makes a vector at read time, but `#(...) is just like (vector ...).
       *
       *   :(let ((f1 (lambda () (let ((x 32)) #(x 0))))
       *          (f2 (lambda () (let ((x 32)) `#(,x 0)))))
       *      (eq? (f1) (f1)))
       *   #t
       *   :(let ((f1 (lambda () (let ((x 32)) #(x 0))))
       *          (f2 (lambda () (let ((x 32)) `#(,x 0)))))
       *      (eq? (f2) (f2)))
       *   #f
       *
       * The tricky part in s7 is that we might have quasiquoted multidimensional vectors:
       */
      if (sc->args == small_int(1))
	sc->code = list_3(sc, sc->APPLY, sc->VECTOR, g_quasiquote_1(sc, sc->value)); /* qq result will be evaluated (might include {list} etc) */
      else sc->code = list_4(sc, sc->APPLY, sc->MULTIVECTOR, sc->args, g_quasiquote_1(sc, sc->value));
      goto EVAL_PAIR;

      
    case OP_READ_UNQUOTE:
      /* here if sc->value is a constant, the unquote is pointless */
      if ((is_pair(sc->value)) ||
	  (is_symbol(sc->value)))
	sc->value = list_2(sc, sc->UNQUOTE, sc->value);
      goto START;
      
      
    case OP_READ_APPLY_VALUES:
      sc->value = list_2(sc, sc->UNQUOTE, list_2(sc, sc->QQ_APPLY_VALUES, sc->value));
      goto START;
      
      
    default:
      fprintf(stderr, "unknown operator: %d, %s\n", 
	      (int)(sc->op), 
	      (((int)(sc->op) < OP_MAX_DEFINED) && ((int)(sc->op) >= 0)) ? op_names[(int)(sc->op)] : "?");
      return(eval_error(sc, "~A: unknown operator!", make_integer(sc, sc->op))); /* not small_int because it's bogus */
    }

  return(sc->F);
}




/* -------------------------------- multiprecision arithmetic -------------------------------- */

#if WITH_GMP

static mp_prec_t mpc_precision = 128;
static mp_prec_t mpc_set_default_precision(mp_prec_t prec) {mpc_precision = prec; return(prec);}

#define mpc_init(Z) mpc_init2(Z, mpc_precision)

static void mpc_init_set(mpc_ptr z, mpc_ptr y, mpc_rnd_t rnd)
{
  mpc_init(z);
  mpc_set(z, y, rnd);
}


mpfr_t *s7_big_real(s7_pointer x)    {return(&big_real(x));}
mpz_t  *s7_big_integer(s7_pointer x) {return(&big_integer(x));}
mpq_t  *s7_big_ratio(s7_pointer x)   {return(&big_ratio(x));}
mpc_t  *s7_big_complex(s7_pointer x) {return(&big_complex(x));}

#define is_big_number(p)  t_big_number_p[type(p)]
#define is_big_integer(p) (type(p) == T_BIG_INTEGER)
#define is_big_ratio(p)   (type(p) == T_BIG_RATIO)
#define is_big_real(p)    (type(p) == T_BIG_REAL)
#define is_big_complex(p) (type(p) == T_BIG_COMPLEX)


bool s7_is_number(s7_pointer p)
{
  return((is_number(p)) || (is_big_number(p)));
}


bool s7_is_integer(s7_pointer p) 
{ 
  return((type(p) == T_INTEGER) ||
	 (type(p) == T_BIG_INTEGER));
}


bool s7_is_real(s7_pointer p) 
{ 
  return((is_real(p)) ||
	 (type(p) == T_BIG_INTEGER) ||
	 (type(p) == T_BIG_RATIO) ||
	 (type(p) == T_BIG_REAL));
}


bool s7_is_rational(s7_pointer p)
{
  return((is_rational(p)) ||
	 (type(p) == T_BIG_INTEGER) ||
	 (type(p) == T_BIG_RATIO));
}


bool s7_is_ratio(s7_pointer p)
{
  return((type(p) == T_RATIO) ||
	 (type(p) == T_BIG_RATIO));
}


bool s7_is_complex(s7_pointer p)
{
  return(s7_is_number(p));
}


static char *mpfr_to_string(mpfr_t val, int radix)
{
  char *str, *tmp, *str1;
  mp_exp_t expptr;
  int i, len, ep;

  if (mpfr_zero_p(val))
    return(copy_string("0.0"));

  str1 = mpfr_get_str(NULL, &expptr, radix, 0, val, GMP_RNDN);
  /* 0 -> full precision, but it's too hard to make this look like C formatted output.
   *  
   *  :(format #f "~,3F" pi)
   *  "3.141592653589793238462643383279502884195E0" 
   *  :(format #f "~,3F" 1.1234567890123)  ; not a bignum
   *  "1.123"
   *  :(format #f "~,3F" 1.12345678901234) ; a bignum
   *  "1.123456789012339999999999999999999999999E0"
   * 
   * but we don't know the exponent or the string length until after we call mpfr_get_str.
   */

  str = str1;
  ep = (int)expptr;
  len = safe_strlen(str);

  if (radix <= 10)
    {
      /* remove trailing 0's */
      for (i = len - 1; i > 3; i--)
	if (str[i] != '0')
	  break;
      if (i < len - 1)
	str[i + 1] = '\0';

      len += 64;
      tmp = (char *)malloc(len * sizeof(char));
  
      if (str[0] == '-')
	snprintf(tmp, len, "-%c.%sE%d", str[1], (char *)(str + 2), ep - 1);
      else snprintf(tmp, len, "%c.%sE%d", str[0], (char *)(str + 1), ep - 1);
    }
  else
    {
      /* if radix > 10, we should not be using the 'E' business -- need to move the decimal point */
      /* (number->string 1234.5678909876543212345 16) "4d2.91614dc3ab1f80e55a563311b8f308"
       * (number->string -1234.5678909876543212345 16) "-4d2.91614dc3ab1f80e55a563311b8f308"
       * (number->string 1234.5678909876543212345e8 16) "1cbe991a6a.c3f35c11868cb7e3fb75536"
       * (number->string 1234.5678909876543212345e-8 16) "0.0000cf204983a27e1eff701c562a870641e50"
       * (number->string 123456789098765432.12345e-8 16) "499602d2.fcd6e9e1748ba5adccc12c5a8"
       * (number->string 123456789098765432.1e20 16) "949b0f70beeac8895e74b18b9680000.00"
       */
      int loc = 0;
      tmp = (char *)calloc(len + ep + 64, sizeof(char));
      if (str[0] == '-')
	tmp[loc++] = (*str++);
      if (ep < 0)
	{
	  ep = -ep;
	  tmp[loc++] = '0';
	  tmp[loc++] = '.';
	  for (i = 0; i < ep; i++)
	    tmp[loc++] = '0';
	}
      else
	{
	  for (i = 0; i < ep; i++)
	    tmp[loc++] = (*str++);
	  tmp[loc++] = '.';
	}
      while (*str) tmp[loc++] = (*str++);
    }
  mpfr_free_str(str1);
  return(tmp);
}


static char *mpc_to_string(mpc_t val, int radix)
{
  char *rl, *im, *tmp;
  int len;
  mpfr_t r;

  mpfr_init(r);
  mpc_real(r, val, GMP_RNDN);
  rl = mpfr_to_string(r, radix);
  mpc_imag(r, val, GMP_RNDN);
  im = mpfr_to_string(r, radix);

  len = safe_strlen(rl) + safe_strlen(im) + 128;
  tmp = (char *)malloc(len * sizeof(char));
  snprintf(tmp, len, "%s%s%si", rl, (im[0] == '-') ? "" : "+", im);

  free(rl);
  free(im);
  return(tmp);
}


static char *big_number_to_string_with_radix(s7_pointer p, int radix, int width)
{
  char *str = NULL;

  switch (type(p))
    {
    case T_BIG_INTEGER:
      str = mpz_get_str(NULL, radix, big_integer(p));
      break;

    case T_BIG_RATIO:
      str = mpq_get_str(NULL, radix, big_ratio(p));
      break;

    case T_BIG_REAL:
      str = mpfr_to_string(big_real(p), radix);
      break;

    default:
      str = mpc_to_string(big_complex(p), radix);
      break;
    }

  if (width > 0)
    {
      int len;
      len = safe_strlen(str);
      if (width > len)
	{
	  char *p1;
	  p1 = pad_number(str, len, width);
	  free(str);
	  return(p1);
	}
    }
  return(str);
}


static bool s7_is_one_or_big_one(s7_pointer p)
{
  bool result = false;

  if (!is_big_number(p))
    return(s7_is_one(p));

  if (type(p) == T_BIG_INTEGER)
    {
      mpz_t n;
      mpz_init_set_si(n, 1);
      result = (mpz_cmp(n, big_integer(p)) == 0);
      mpz_clear(n);
    }
  else
    {
      if (type(p) == T_BIG_REAL)
	{
	  mpfr_t n;
	  mpfr_init_set_d(n, 1.0, GMP_RNDN);
	  result = (mpfr_cmp(n, big_real(p)) == 0);
	  mpfr_clear(n);
	}
    }
  return(result);
}


static s7_pointer string_to_big_integer(s7_scheme *sc, const char *str, int radix)
{
  s7_pointer x;

  NEW_CELL(sc, x);
  set_type(x, T_BIG_INTEGER);
  mpz_init_set_str(big_integer(x), (str[0] == '+') ? (const char *)(str + 1) : str, radix);

  return(x);
}


static s7_pointer mpz_to_big_integer(s7_scheme *sc, mpz_t val)
{
  s7_pointer x;

  NEW_CELL(sc, x);
  set_type(x, T_BIG_INTEGER);
  mpz_init_set(big_integer(x), val);

  return(x);
}


s7_pointer s7_make_big_integer(s7_scheme *sc, mpz_t *val)
{
  return(mpz_to_big_integer(sc, *val));
}


static s7_pointer string_to_big_ratio(s7_scheme *sc, const char *str, int radix)
{
  s7_pointer x;
  mpq_t n;

  mpq_init(n);
  mpq_set_str(n, str, radix);
  mpq_canonicalize(n);

  if (mpz_cmp_ui(mpq_denref(n), 1) == 0)
    x = mpz_to_big_integer(sc, mpq_numref(n));
  else
    {
      NEW_CELL(sc, x);
      set_type(x, T_BIG_RATIO);
      mpq_init(big_ratio(x));
      mpq_set_num(big_ratio(x), mpq_numref(n));
      mpq_set_den(big_ratio(x), mpq_denref(n));
    }

  mpq_clear(n);
  return(x);
}


static s7_pointer mpq_to_big_ratio(s7_scheme *sc, mpq_t val)
{
  s7_pointer x;

  NEW_CELL(sc, x);
  set_type(x, T_BIG_RATIO);
  mpq_init(big_ratio(x));
  mpq_set_num(big_ratio(x), mpq_numref(val));
  mpq_set_den(big_ratio(x), mpq_denref(val));
  
  return(x);
}


s7_pointer s7_make_big_ratio(s7_scheme *sc, mpq_t *val)
{
  return(mpq_to_big_ratio(sc, *val));
}


static s7_pointer mpz_to_big_ratio(s7_scheme *sc, mpz_t val)
{
  s7_pointer x;

  NEW_CELL(sc, x);
  set_type(x, T_BIG_RATIO);
  mpq_init(big_ratio(x));
  mpq_set_num(big_ratio(x), val);
  
  return(x);
}


static s7_pointer make_big_integer_or_ratio(s7_scheme *sc, s7_pointer z)
{
  if (mpz_cmp_ui(mpq_denref(big_ratio(z)), 1) == 0)
    return(mpz_to_big_integer(sc, mpq_numref(big_ratio(z))));
  return(z);
}


static s7_pointer string_to_big_real(s7_scheme *sc, const char *str, int radix)
{
  s7_pointer x;

  NEW_CELL(sc, x);
  set_type(x, T_BIG_REAL);
  mpfr_init_set_str(big_real(x), str, radix, GMP_RNDN);

  return(x);
}

static void mpz_init_set_s7_Int(mpz_t n, s7_Int uval);

static s7_pointer s7_number_to_big_real(s7_scheme *sc, s7_pointer p)
{
  s7_pointer x;

  NEW_CELL(sc, x);
  set_type(x, T_BIG_REAL);

  switch (type(p))
    {
    case T_INTEGER: 
      {
	s7_Int val;
	val = s7_integer(p);
	if ((val <= S7_LONG_MAX) && (val >= S7_LONG_MIN))
	  mpfr_init_set_si(big_real(x), (int)val, GMP_RNDN);
	else mpfr_init_set_ld(big_real(x), (long double)val, GMP_RNDN);
      }
      break;

    case T_RATIO:
      /* here we can't use fraction(number(p)) even though that uses long double division because
       *   there are lots of long long int ratios that will still look the same.
       *   We have to do the actual bignum divide by hand.
       */
      {
	mpq_t rat;
	mpz_t n1, d1;

	mpz_init_set_s7_Int(n1, numerator(p));
	mpz_init_set_s7_Int(d1, denominator(p));
	mpq_init(rat);

	mpq_set_num(rat, n1);
	mpq_set_den(rat, d1);
	mpq_canonicalize(rat);
	mpfr_init_set_q(big_real(x), rat, GMP_RNDN);

	mpz_clear(n1);
	mpz_clear(d1);
	mpq_clear(rat);
      }
      break;

    default:
      mpfr_init_set_d(big_real(x), s7_real(p), GMP_RNDN);
      break;
    }

  return(x);
}


static s7_pointer mpz_to_big_real(s7_scheme *sc, mpz_t val)
{
  s7_pointer x;

  NEW_CELL(sc, x);
  set_type(x, T_BIG_REAL);
  mpfr_init_set_z(big_real(x), val, GMP_RNDN);

  return(x);
}


static s7_pointer mpq_to_big_real(s7_scheme *sc, mpq_t val)
{
  s7_pointer x;

  NEW_CELL(sc, x);
  set_type(x, T_BIG_REAL);
  mpfr_init_set_q(big_real(x), val, GMP_RNDN);

  return(x);
}


static s7_pointer mpfr_to_big_real(s7_scheme *sc, mpfr_t val)
{
  s7_pointer x;

  NEW_CELL(sc, x);
  set_type(x, T_BIG_REAL);
  mpfr_init_set(big_real(x), val, GMP_RNDN);
  
  /* fprintf(stderr, "%s: %s\n", __func__, DISPLAY(x)); */
  return(x);
}


s7_pointer s7_make_big_real(s7_scheme *sc, mpfr_t *val)
{
  return(mpfr_to_big_real(sc, *val));
}


static s7_pointer big_pi(s7_scheme *sc)
{
  s7_pointer x;

  NEW_CELL(sc, x);
  set_type(x, T_BIG_REAL);
  mpfr_init(big_real(x));
  mpfr_const_pi(big_real(x), GMP_RNDN);
  
  return(x);
}


static s7_pointer s7_number_to_big_complex(s7_scheme *sc, s7_pointer p)
{
  s7_pointer x;

  NEW_CELL(sc, x);
  set_type(x, T_BIG_COMPLEX);
  mpc_init(big_complex(x));

  switch (type(p))
    {
    case T_INTEGER:
      {
	s7_Int val;
	val = s7_integer(p);
	if ((val <= S7_LONG_MAX) && (val >= S7_LONG_MIN))
	  mpc_set_si(big_complex(x), (long int)val, MPC_RNDNN);
	else mpc_set_d(big_complex(x), (double)val, MPC_RNDNN);
      }
      break;

    case T_RATIO:
      /* can't use fraction here */
      {
	mpfr_t temp;
	mpq_t rat;
	mpz_t n1, d1;

	mpz_init_set_s7_Int(n1, numerator(p));
	mpz_init_set_s7_Int(d1, denominator(p));
	mpq_init(rat);

	mpq_set_num(rat, n1);
	mpq_set_den(rat, d1);
	mpq_canonicalize(rat);
	mpfr_init_set_q(temp, rat, GMP_RNDN);
	mpc_set_fr(big_complex(x), temp, MPC_RNDNN);

	mpz_clear(n1);
	mpz_clear(d1);
	mpq_clear(rat);
	mpfr_clear(temp);
      }
      break;

    case T_REAL:
      mpc_set_d(big_complex(x), s7_real(p), MPC_RNDNN);
      break;

    default:
      mpc_set_d_d(big_complex(x), real_part(p), imag_part(p), MPC_RNDNN);
      break;
    }
  return(x);
}


static s7_pointer make_big_real_or_complex(s7_scheme *sc, s7_pointer z)
{
  double ipart;

  ipart = mpfr_get_d(mpc_imagref(big_complex(z)), GMP_RNDN);
  /* not mpfr_cmp_ui to 0 here because that misleads us when imag_part is NaN or inf */
  if (ipart == 0.0)
    return(mpfr_to_big_real(sc, mpc_realref(big_complex(z))));
  return(z);
}


static s7_pointer mpz_to_big_complex(s7_scheme *sc, mpz_t val)
{
  mpfr_t temp;
  s7_pointer x;

  NEW_CELL(sc, x);
  set_type(x, T_BIG_COMPLEX);
  mpc_init(big_complex(x));
  mpfr_init_set_z(temp, val, GMP_RNDN);
  mpc_set_fr(big_complex(x), temp, MPC_RNDNN);

  mpfr_clear(temp);
  return(x);
}


static s7_pointer mpq_to_big_complex(s7_scheme *sc, mpq_t val)
{
  mpfr_t temp;
  s7_pointer x;

  NEW_CELL(sc, x);
  set_type(x, T_BIG_COMPLEX);
  mpc_init(big_complex(x));
  mpfr_init_set_q(temp, val, GMP_RNDN);
  mpc_set_fr(big_complex(x), temp, MPC_RNDNN);

  mpfr_clear(temp);
  return(x);
}


static s7_pointer mpfr_to_big_complex(s7_scheme *sc, mpfr_t val)
{
  s7_pointer x;

  NEW_CELL(sc, x);
  set_type(x, T_BIG_COMPLEX);
  mpc_init(big_complex(x));
  mpc_set_fr(big_complex(x), val, MPC_RNDNN);

  return(x);
}


static s7_pointer mpc_to_big_complex(s7_scheme *sc, mpc_t val)
{
  s7_pointer x;

  NEW_CELL(sc, x);
  set_type(x, T_BIG_COMPLEX);
  mpc_init(big_complex(x));
  mpc_set(big_complex(x), val, MPC_RNDNN);

  return(x);
}


s7_pointer s7_make_big_complex(s7_scheme *sc, mpc_t *val)
{
  return(mpc_to_big_complex(sc, *val));
}


static s7_pointer make_big_complex(s7_scheme *sc, mpfr_t rl, mpfr_t im)
{
  /* there is no mpc_get_str equivalent, so we need to split up str,
   *   use make_big_real to get the 2 halves, then mpc_init, then
   *   mpc_set_fr_fr.
   */
  s7_pointer x;

  NEW_CELL(sc, x);
  set_type(x, T_BIG_COMPLEX);
  mpc_init(big_complex(x));
  mpc_set_fr_fr(big_complex(x), rl ,im, MPC_RNDNN);

  return(x);
}


/* gmp.h mpz_init_set_si the "si" part is "signed long int", so in the 64-bit case,
 *   s7_Int already fits.  I guess we can catch this (since no configure script)
 *   by noticing that sizeof(s7_Int) == sizeof(long int).
 */

static void mpz_init_set_s7_Int(mpz_t n, s7_Int uval)
{
  if (sizeof(s7_Int) == sizeof(long int))
    mpz_init_set_si(n, uval);
  else
    {
      if ((uval <= S7_LONG_MAX) && (uval >= S7_LONG_MIN))
	mpz_init_set_si(n, (int)uval);
      else
	{ 
	  /* long long int to gmp mpz_t */
	  bool need_sign;
	  long long int val;
	  val = (long long int)uval;
	  /* handle one special case (sigh) */
	  if (val == S7_LLONG_MIN)
	    mpz_init_set_str(n, "-9223372036854775808", 10);
	  else
	    {
	      need_sign = (val < 0);
	      if (need_sign) val = -val;
	      mpz_init_set_si(n, val >> 32);
	      mpz_mul_2exp(n, n, 32);
	      mpz_add_ui(n, n, (unsigned int)(val & 0xffffffff));
	      if (need_sign) mpz_neg(n, n);
	    }
	}
    }
}


static s7_pointer s7_Int_to_big_integer(s7_scheme *sc, s7_Int val)
{
  s7_pointer x;

  NEW_CELL(sc, x);
  set_type(x, T_BIG_INTEGER);
  mpz_init_set_s7_Int(big_integer(x), val);

  return(x);
}


static s7_Int big_integer_to_s7_Int(mpz_t n)
{
  long long int high, low;
  mpz_t x;
  bool need_sign = false;

  if (sizeof(s7_Int) == sizeof(long int))
    return(mpz_get_si(n));

  if (mpz_fits_sint_p(n))
    return((s7_Int)mpz_get_si(n));
  /* special case as always is most-negative-fixnum */

  mpz_init_set(x, n);
  if (mpz_cmp_ui(x, 0) < 0)
    {
      need_sign = true;
      mpz_neg(x, x);
    }
  low = mpz_get_ui(x);
  if (low == S7_LLONG_MIN)
    return(S7_LLONG_MIN);

  mpz_fdiv_q_2exp(x, x, 32);
  high = mpz_get_ui(x);
  mpz_clear(x);
  if (need_sign)
    return(-(low + (high << 32)));
  return(low + (high << 32));
}

/* (expt 1/2 9223372036854775807) */


s7_Double s7_number_to_real(s7_pointer x)
{
  switch (type(x))
    {
    case T_INTEGER: 
      return((s7_Double)integer(x));

    case T_RATIO:   
      return((s7_Double)numerator(x) / (s7_Double)denominator(x));

    case T_REAL:    
      return(real(x));

    case T_COMPLEX:
      return(real_part(x));

    case T_BIG_INTEGER:
      return((s7_Double)big_integer_to_s7_Int(big_integer(x)));

    case T_BIG_RATIO:
      return((s7_Double)((long double)big_integer_to_s7_Int(mpq_numref(big_ratio(x))) / (long double)big_integer_to_s7_Int(mpq_denref(big_ratio(x)))));

    case T_BIG_REAL:
      return((s7_Double)mpfr_get_d(big_real(x), GMP_RNDN));

    case T_BIG_COMPLEX:
      return((s7_Double)mpfr_get_d(mpc_realref(big_complex(x)), GMP_RNDN));
    }
  return(0.0);  /* this is what the non-gmp case does */
}


s7_Int s7_number_to_integer(s7_pointer x)
{
  switch (type(x))
    {
    case T_INTEGER: 
      return(integer(x));

    case T_RATIO:   
      return((s7_Int)numerator(x) / (s7_Double)denominator(x));

    case T_REAL:    
      return((s7_Int)real(x));

    case T_COMPLEX: 
      return((s7_Int)real_part(x));

    case T_BIG_INTEGER:
      return(big_integer_to_s7_Int(big_integer(x)));

    case T_BIG_RATIO:
      return((s7_Int)((double)big_integer_to_s7_Int(mpq_numref(big_ratio(x))) / (double)big_integer_to_s7_Int(mpq_denref(big_ratio(x)))));

    case T_BIG_REAL:
      return((s7_Int)mpfr_get_d(big_real(x), GMP_RNDN));

    case T_BIG_COMPLEX:
      return((s7_Int)mpfr_get_d(mpc_realref(big_complex(x)), GMP_RNDN));
    }
  return(0);  /* this is what the non-gmp case does */
}


s7_Int s7_numerator(s7_pointer x)
{
  switch (type(x))
    {
    case T_INTEGER:
      return(integer(x));

    case T_RATIO:
      return(numerator(x));

    case T_BIG_INTEGER:
      return(big_integer_to_s7_Int(big_integer(x)));

    case T_BIG_RATIO:
      return(big_integer_to_s7_Int(mpq_numref(big_ratio(x))));
    }

  return(0);
}


s7_Int s7_denominator(s7_pointer x)
{
  switch (type(x))
    {
    case T_RATIO:
      return(denominator(x));

    case T_BIG_RATIO:
      return(big_integer_to_s7_Int(mpq_denref(big_ratio(x))));
    }

  return(1);
}


s7_Double s7_real_part(s7_pointer x)
{
  switch(type(x))
    {
    case T_INTEGER:     return((s7_Double)integer(x));
    case T_RATIO:       return(fraction(x));
    case T_REAL:        return(real(x));
    case T_COMPLEX:     return(real_part(x));
    case T_BIG_INTEGER: return((s7_Double)big_integer_to_s7_Int(big_integer(x)));
    case T_BIG_RATIO:   return((s7_Double)((long double)big_integer_to_s7_Int(mpq_numref(big_ratio(x))) / (long double)big_integer_to_s7_Int(mpq_denref(big_ratio(x)))));
    case T_BIG_REAL:    return((s7_Double)mpfr_get_d(big_real(x), GMP_RNDN));
    default:            return((s7_Double)mpfr_get_d(mpc_realref(big_complex(x)), GMP_RNDN));
    }
}


s7_Double s7_imag_part(s7_pointer x)
{
  switch (type(x))
    {
    case T_COMPLEX:     return(imag_part(x));
    case T_BIG_COMPLEX: return((s7_Double)mpfr_get_d(mpc_imagref(big_complex(x)), GMP_RNDN));
    default:            return(0.0);
    }
}


s7_Int s7_integer(s7_pointer p)
{
  if (type(p) == T_BIG_INTEGER)
    return(big_integer_to_s7_Int(big_integer(p)));
  return(integer(p));
}


s7_Double s7_real(s7_pointer p)
{
  if (type(p) == T_BIG_REAL)
    return((s7_Double)mpfr_get_d(big_real(p), GMP_RNDN));
  return(real(p));
}


static mpq_t *s7_Ints_to_mpq(s7_Int num, s7_Int den)
{
  mpq_t *n;
  n = (mpq_t *)malloc(sizeof(mpq_t));
  mpq_init(*n);
  if ((num <= S7_LONG_MAX) && (num >= S7_LONG_MIN) &&
      (den <= S7_LONG_MAX) && (den >= S7_LONG_MIN))
    mpq_set_si(*n, (long int)num, (long int)den);
  else
    {
      mpz_t n1, d1;
      mpz_init_set_s7_Int(n1, num);
      mpz_init_set_s7_Int(d1, den);
      mpq_set_num(*n, n1);
      mpq_set_den(*n, d1);
      mpq_canonicalize(*n);
      mpz_clear(n1);
      mpz_clear(d1);
    }
  return(n);
}


static mpfr_t *s7_Double_to_mpfr(s7_Double val)
{
  mpfr_t *n;
  n = (mpfr_t *)malloc(sizeof(mpfr_t));
  mpfr_init_set_d(*n, val, GMP_RNDN);
  return(n);
}


static mpc_t *s7_Doubles_to_mpc(s7_Double rl, s7_Double im)
{
  mpc_t *n;
  n = (mpc_t *)malloc(sizeof(mpc_t));
  mpc_init(*n);
  mpc_set_d_d(*n, rl, im, MPC_RNDNN);
  return(n);
}


static s7_pointer s7_ratio_to_big_ratio(s7_scheme *sc, s7_Int num, s7_Int den)
{
  s7_pointer x;

  NEW_CELL(sc, x);
  set_type(x, T_BIG_RATIO);
  mpq_init(big_ratio(x));

  if ((num <= S7_LONG_MAX) && (num >= S7_LONG_MIN) &&
      (den <= S7_LONG_MAX) && (den >= S7_LONG_MIN))
    mpq_set_si(big_ratio(x), (long int)num, (long int)den);
  else
    {
      mpz_t n1, d1;
      mpz_init_set_s7_Int(n1, num);
      mpz_init_set_s7_Int(d1, den);
      mpq_set_num(big_ratio(x), n1);
      mpq_set_den(big_ratio(x), d1);
      mpq_canonicalize(big_ratio(x));
      mpz_clear(n1);
      mpz_clear(d1);
    }

  return(x);
}


static bool big_numbers_are_eqv(s7_pointer a, s7_pointer b)
{
  bool result;
  /* either or both can be big here, but not neither */

  if ((!is_big_number(a)) && (!is_big_number(b)))
    return(false);
  
  if (s7_is_integer(a))
    {
      mpz_t a1, b1;
      if (!(s7_is_integer(b))) return(false);

      if ((is_big_number(a)) && (is_big_number(b)))
	return(mpz_cmp(big_integer(a), big_integer(b)) == 0);

      if (is_big_number(a))
	mpz_init_set(a1, big_integer(a));
      else mpz_init_set_s7_Int(a1, s7_integer(a));

      if (is_big_number(b))
	mpz_init_set(b1, big_integer(b));
      else mpz_init_set_s7_Int(b1, s7_integer(b));
      result = (mpz_cmp(a1, b1) == 0);

      mpz_clear(a1);
      mpz_clear(b1);
      return(result);
    }

  if (s7_is_ratio(a))
    {
      mpq_t *a1, *b1;
      if (!s7_is_ratio(b)) return(false);

      if ((is_big_number(a)) && (is_big_number(b)))
	return(mpq_cmp(big_ratio(a), big_ratio(b)) == 0);

      if (is_big_number(a))
	a1 = &big_ratio(a);
      else a1 = s7_Ints_to_mpq(numerator(a), denominator(a));
      if (is_big_number(b))
	b1 = &big_ratio(b);
      else b1 = s7_Ints_to_mpq(numerator(b), denominator(b));

      result = (mpq_cmp(*a1, *b1) == 0);

      if (!is_big_number(a))
	{
	  mpq_clear(*a1);
	  free(a1);
	}
      if (!is_big_number(b))
	{
	  mpq_clear(*b1);
	  free(b1);
	}
      return(result);
    }

  if (s7_is_real(a))
    {
      mpfr_t *a1, *b1;

      /* s7_is_real is not finicky enough here -- (eqv? 1.0 1) should return #f */
      if (is_big_number(b))
	{
	  if (type(b) != T_BIG_REAL)
	    return(false);
	}
      else
	{
	  if (type(b) != T_REAL)
	    return(false);
	}

      if ((is_big_number(a)) && (is_big_number(b)))
	return(mpfr_equal_p(big_real(a), big_real(b)));

      if (is_big_number(a))
	a1 = &big_real(a);
      else a1 = s7_Double_to_mpfr(s7_real(a));

      if (is_big_number(b))
	b1 = &big_real(b);
      else b1 = s7_Double_to_mpfr(s7_real(b));

      result = (mpfr_cmp(*a1, *b1) == 0);

      if (!is_big_number(a))
	{
	  mpfr_clear(*a1);
	  free(a1);
	}
      if (!is_big_number(b))
	{
	  mpfr_clear(*b1);
	  free(b1);
	}
      return(result);
    }

  if (s7_is_complex(a))
    {
      mpc_t *a1, *b1;
      /* s7_is_complex is not finicky enough here */
      if (is_big_number(b))
	{
	  if (type(b) != T_BIG_COMPLEX)
	    return(false);
	}
      else
	{
	  if (is_real(b))
	    return(false);
	}

      if ((is_big_number(a)) && (is_big_number(b)))
	return(mpc_cmp(big_complex(a), big_complex(b)) == 0);

      if (is_big_number(a))
	a1 = &big_complex(a);
      else a1 = s7_Doubles_to_mpc(real_part(a), imag_part(a));

      if (is_big_number(b))
	b1 = &big_complex(b);
      else b1 = s7_Doubles_to_mpc(real_part(b), imag_part(b));

      result = (mpc_cmp(*a1, *b1) == 0);

      if (!is_big_number(a))
	{
	  mpc_clear(*a1);
	  free(a1);
	}
      if (!is_big_number(b))
	{
	  mpc_clear(*b1);
	  free(b1);
	}
      return(result);
    }

  return(false);
}


static s7_pointer string_to_either_integer(s7_scheme *sc, const char *str, int radix)
{
  s7_Int val;
  bool overflow = false;

  val = string_to_integer(str, radix, &overflow);
  if (!overflow)
    return(make_integer(sc, val));
  
  return(string_to_big_integer(sc, str, radix));
}


static s7_pointer string_to_either_ratio(s7_scheme *sc, const char *nstr, const char *dstr, int radix)
{
  s7_Int n, d;
  bool overflow = false;

  /* gmp segfaults if passed a bignum/0 so this needs to check first that
   *   the denominator is not 0 before letting gmp screw up.  Also, if the
   *   first character is '+', gmp returns 0!
   */
  d = string_to_integer(dstr, radix, &overflow);
  if (!overflow)
    {
      if (d == 0)
	return(real_NaN);

      n = string_to_integer(nstr, radix, &overflow);
      if (!overflow)
	return(s7_make_ratio(sc, n, d));
    }
  if (nstr[0] == '+')
    return(string_to_big_ratio(sc, (const char *)(nstr + 1), radix));
  return(string_to_big_ratio(sc, nstr, radix));
}


static s7_pointer string_to_either_real(s7_scheme *sc, const char *str, int radix)
{
  bool overflow = false;
  s7_Double val;

  val = string_to_double_with_radix((char *)str, radix, &overflow);
  if (!overflow)
    return(s7_make_real(sc, val));

  return(string_to_big_real(sc, str, radix));
}


static s7_pointer string_to_either_complex_1(s7_scheme *sc, char *q, char *slash1, char *ex1, bool has_dec_point1, int radix, s7_Double *d_rl)
{
  bool overflow = false;
  /* there's a real problem here -- we don't want to promote s7_Double .1 to a bignum because
   *    its low order digits are garbage, causing (rationalize .1 0) to return 3602879701896397/36028797018963968
   *    no matter what the bignum-precision.  But we can't just fallback on gmp's reader because (for example)
   *    it reads 1/2+i or 1+0/0i as 1.0.  Also format gets screwed up.  And string->number signals an error
   *    where it should return #f.  I wonder what to do.
   */
  if ((has_dec_point1) ||
      (ex1))
    {
      (*d_rl) = string_to_double_with_radix(q, radix, &overflow);
      if (overflow)
	return(string_to_big_real(sc, q, radix));
    }
  else
    {
      if (slash1)
	{
	  s7_Int n, d;

	  /* q can include the slash and denominator */
	  n = string_to_integer(q, radix, &overflow);
	  if (overflow)
	    return(string_to_big_ratio(sc, q, radix));
	  else
	    {
	      d = string_to_integer(slash1, radix, &overflow);
	      if (!overflow)
		(*d_rl) = (s7_Double)n / (s7_Double)d;
	      else return(string_to_big_ratio(sc, q, radix));
	    }
	}
      else
	{
	  s7_Int val;

	  val = string_to_integer(q, radix, &overflow);
	  if (overflow)
	    return(string_to_big_integer(sc, q, radix));
	  (*d_rl) = (s7_Double)val;
	}
    }
  if ((*d_rl) == -0.0) (*d_rl) = 0.0;

  return(NULL);
}


static s7_pointer string_to_either_complex(s7_scheme *sc,
					   char *q, char *slash1, char *ex1, bool has_dec_point1, 
					   char *plus, char *slash2, char *ex2, bool has_dec_point2, 
					   int radix, int has_plus_or_minus)
{
  /* this can be just about anything involving 2 real/ratio/int portions, +/- in between and 'i' at the end */
  double d_rl = 0.0, d_im = 0.0;
  s7_pointer p_rl = NULL, p_im = NULL, result;
  mpfr_t m_rl, m_im;

  p_rl = string_to_either_complex_1(sc, q, slash1, ex1, has_dec_point1, radix, &d_rl);
  p_im = string_to_either_complex_1(sc, plus, slash2, ex2, has_dec_point2, radix, &d_im);
  
  if (d_im == 0.0)
    {
      /* 1.0+0.0000000000000000000000000000i */
      if ((!p_im) ||
	  (big_is_zero_1(sc, p_im) == sc->T))
	{
	  if (!p_rl)
	    return(s7_make_real(sc, d_rl));
	  return(p_rl);
	}
    }

  if ((!p_rl) && (!p_im))
    return(s7_make_complex(sc, d_rl, (has_plus_or_minus == -1) ? (-d_im) : d_im));

  if (p_rl)
    mpfr_init_set(m_rl, big_real(promote_number(sc, T_BIG_REAL, p_rl)), GMP_RNDN);
  else mpfr_init_set_d(m_rl, d_rl, GMP_RNDN);

  if (p_im)
    mpfr_init_set(m_im, big_real(promote_number(sc, T_BIG_REAL, p_im)), GMP_RNDN);
  else mpfr_init_set_d(m_im, d_im, GMP_RNDN);

  if (has_plus_or_minus == -1) 
    mpfr_neg(m_im, m_im, GMP_RNDN);

  result = make_big_complex(sc, m_rl, m_im);
  
  mpfr_clear(m_rl);
  mpfr_clear(m_im);
  return(result);
}


static int big_type_to_result_type(int cur_type, int next_type)
{
  if ((cur_type == T_BIG_COMPLEX) || 
      (cur_type == T_COMPLEX) ||
      (next_type == T_BIG_COMPLEX))
    return(T_BIG_COMPLEX);

  if ((cur_type == T_BIG_REAL) ||
      (cur_type == T_REAL) ||
      (next_type == T_BIG_REAL))
    return(T_BIG_REAL);

  if ((cur_type == T_BIG_RATIO) ||
      (cur_type == T_RATIO) ||
      (next_type == T_BIG_RATIO))
    return(T_BIG_RATIO);

  return(T_BIG_INTEGER);
}


static int normal_type_to_result_type(int cur_type, int next_type)
{
  if (cur_type > T_COMPLEX)
    next_type += 4;
  if (cur_type > next_type)
    return(cur_type);
  return(next_type);
}


static s7_pointer promote_number_1(s7_scheme *sc, int type, s7_pointer x, bool copy)
{
  /* x can be any number -- need to convert it to the current result type */

  switch (type)
    {
    case T_BIG_INTEGER:
      if (is_big_number(x)) 
	{
	  if (copy)
	    return(mpz_to_big_integer(sc, big_integer(x)));
	  return(x);                       /* can only be T_BIG_INTEGER here */
	}
      return(s7_Int_to_big_integer(sc, s7_integer(x))); /* can only be integer here */
      break;

    case T_BIG_RATIO:
      if (is_big_number(x))
	{
	  if (type(x) == T_BIG_RATIO) 
	    {
	      if (copy)
		return(mpq_to_big_ratio(sc, big_ratio(x)));
	      return(x);
	    }
	  return(mpz_to_big_ratio(sc, big_integer(x)));
	}
      if (type(x) == T_INTEGER)
	return(s7_ratio_to_big_ratio(sc, integer(x), 1));
      return(s7_ratio_to_big_ratio(sc, numerator(x), denominator(x)));
      break;

    case T_BIG_REAL:
      if (is_big_number(x))
	{
	  if (type(x) == T_BIG_REAL) 
	    {
	      if (copy)
		return(mpfr_to_big_real(sc, big_real(x)));
	      return(x);
	    }
	  if (type(x) == T_BIG_RATIO) 
	    return(mpq_to_big_real(sc, big_ratio(x)));
	  return(mpz_to_big_real(sc, big_integer(x)));
	}
      return(s7_number_to_big_real(sc, x));
      break;

    default:
      if (is_big_number(x))
	{
	  if (type(x) == T_BIG_COMPLEX) 
	    {
	      if (copy)
		return(mpc_to_big_complex(sc, big_complex(x)));
	      return(x);
	    }
	  if (type(x) == T_BIG_REAL) 
	    return(mpfr_to_big_complex(sc, big_real(x)));
	  if (type(x) == T_BIG_RATIO) 
	    return(mpq_to_big_complex(sc, big_ratio(x)));
	  return(mpz_to_big_complex(sc, big_integer(x)));
	}
      return(s7_number_to_big_complex(sc, x));
      break;
    }

  return(sc->NIL);
}


static s7_pointer promote_number(s7_scheme *sc, int type, s7_pointer x)
{
  return(promote_number_1(sc, type, x, false));
}


static s7_pointer to_big(s7_scheme *sc, s7_pointer x)
{
  if (is_big_number(x))
    return(x);
  switch (type(x))
    {
    case T_INTEGER:
      return(s7_Int_to_big_integer(sc, integer(x)));
      
    case T_RATIO:
      return(s7_ratio_to_big_ratio(sc, numerator(x), denominator(x)));

    case T_REAL:
      return(s7_number_to_big_real(sc, x));

    default:
      return(s7_number_to_big_complex(sc, x));
    }
}


static s7_pointer copy_and_promote_number(s7_scheme *sc, int type, s7_pointer x)
{
  return(promote_number_1(sc, type, x, true));
}


void s7_vector_fill(s7_scheme *sc, s7_pointer vec, s7_pointer obj) 
{
  s7_Int i, len;
  s7_pointer *tp;
  len = vector_length(vec);
  tp = (s7_pointer *)(vector_elements(vec));

  /* if the same bignum object is assigned to each element, different vector elements
   *    are actually the same -- we need to make a copy of obj for each one
   */

  if (is_big_number(obj))
    {
      int gc_loc;

      /* we'll be calling new_cell below, hence the GC, so make sure the elements are markable,
       *   and the vector itself is GC protected (we can be called within make-vector).
       */
      gc_loc = s7_gc_protect(sc, vec);
      vector_fill(sc, vec, sc->NIL); 

      switch (type(obj))
	{
	case T_BIG_INTEGER:
	  for (i = 0; i < len; i++) 
	    tp[i] = mpz_to_big_integer(sc, big_integer(obj));
	  break;

	case T_BIG_RATIO:
	  for (i = 0; i < len; i++) 
	    tp[i] = mpq_to_big_ratio(sc, big_ratio(obj));
	  break;

	case T_BIG_REAL:
	  for (i = 0; i < len; i++) 
	    tp[i] = mpfr_to_big_real(sc, big_real(obj));
	  break;

	default:
	  for (i = 0; i < len; i++) 
	    tp[i] = mpc_to_big_complex(sc, big_complex(obj));
	  break;
	}

      s7_gc_unprotect_at(sc, gc_loc);
    }
  else vector_fill(sc, vec, obj);
}


static s7_pointer g_bignum(s7_scheme *sc, s7_pointer args)
{
  #define H_bignum "(bignum val (radix 10)) returns a multiprecision version of the string 'val'"
  s7_pointer p;

  p = g_string_to_number_1(sc, args, "bignum"); 
  if (is_false(sc, p))                                       /* (bignum "1/3.0") */
    s7_error(sc, make_symbol(sc, "bignum-error"),
	     list_2(sc,
		    make_protected_string(sc, "bignum argument does not represent a number: ~S"),
		    car(args)));

  switch (type(p))
    {
    case T_INTEGER:  
      return(promote_number(sc, T_BIG_INTEGER, p));

    case T_RATIO:
      return(promote_number(sc, T_BIG_RATIO, p));

      /* we can't use promote_number here because it propogates C-double inaccuracies
       *    (rationalize (bignum "0.1") 0) should return 1/10 not 3602879701896397/36028797018963968
       */
    case T_REAL:
      if (is_NaN(real(p))) return(p);
      return(string_to_big_real(sc, s7_string(car(args)), (is_pair(cdr(args))) ? s7_integer(cadr(args)) : 10));

    case T_COMPLEX:
      /* PERHAPS: read complex via gmp, (rationalize (real-part (bignum "0.1+i")) 0) -> 3602879701896397/36028797018963968
       */
      return(promote_number(sc, T_BIG_COMPLEX, p));

    default:
      return(p);
    }
}


bool s7_is_bignum(s7_pointer obj)
{
  return(is_big_number(obj));
}


static s7_pointer g_is_bignum(s7_scheme *sc, s7_pointer args)
{
  #define H_is_bignum "(bignum? obj) returns #t if obj is a multiprecision number."
  return(s7_make_boolean(sc, is_big_number(car(args))));
}


static s7_Int add_max = 0;

static s7_pointer big_add(s7_scheme *sc, s7_pointer args)
{
  int result_type = T_INTEGER;
  s7_pointer x, result;

  if (is_null(args))
    return(small_int(0));

  if ((is_null(cdr(args))) && (s7_is_number(car(args))))
    return(car(args));

  for (x = args; is_not_null(x); x = cdr(x)) 
    {
      s7_pointer p;
      p = car(x);
      if (!is_number(p))
	{
	  if (!is_big_number(p))
	    return(s7_wrong_type_arg_error(sc, "+", position_of(x, args), p, "a number"));
	  else result_type = big_type_to_result_type(result_type, type(p));
	}
      else result_type = normal_type_to_result_type(result_type, type(p));
    }

  if (result_type < T_BIG_INTEGER)
    return(g_add(sc, args));

  result = copy_and_promote_number(sc, result_type, car(args));

  for (x = cdr(args); is_not_null(x); x = cdr(x)) 
    {
      s7_pointer arg;
      arg = promote_number(sc, result_type, car(x));

      switch (result_type)
	{
	case T_BIG_INTEGER:
	  mpz_add(big_integer(result), big_integer(result), big_integer(arg));
	  break;
	  
	case T_BIG_RATIO:
	  mpq_add(big_ratio(result), big_ratio(result), big_ratio(arg));
	  break;
	  
	case T_BIG_REAL:
	  mpfr_add(big_real(result), big_real(result), big_real(arg), GMP_RNDN);
	  break;
	  
	case T_BIG_COMPLEX:
	  mpc_add(big_complex(result), big_complex(result), big_complex(arg), MPC_RNDNN);
	  break;
	}
    }
  
  switch (result_type)
      {
      case T_BIG_RATIO:
	return(make_big_integer_or_ratio(sc, result));

      case T_BIG_COMPLEX:
	return(make_big_real_or_complex(sc, result));
      }
  return(result);
}


static s7_pointer big_negate(s7_scheme *sc, s7_pointer args)
{
  /* assume cdr(args) is nil and we're called from subtract, so check for big num else call g_subtract */
  s7_pointer p, x;

  p = car(args);
  switch (type(p))
    {
    case T_BIG_INTEGER:
      x = mpz_to_big_integer(sc, big_integer(p));
      mpz_neg(big_integer(x), big_integer(x));
      return(x);

    case T_BIG_RATIO:
      x = mpq_to_big_ratio(sc, big_ratio(p));
      mpq_neg(big_ratio(x), big_ratio(x));
      return(x);

    case T_BIG_REAL:
      x = mpfr_to_big_real(sc, big_real(p));
      mpfr_neg(big_real(x), big_real(x), GMP_RNDN);
      return(x);

    case T_BIG_COMPLEX:
      x = mpc_to_big_complex(sc, big_complex(p));
      mpc_neg(big_complex(x), big_complex(x), MPC_RNDNN);
      return(x);

    case T_INTEGER: 
      if (integer(p) == S7_LLONG_MIN)
	{
	  x = s7_Int_to_big_integer(sc, integer(p));
	  mpz_neg(big_integer(x), big_integer(x));
	  return(x);
	}
      return(make_integer(sc, -integer(p)));
      
    case T_RATIO:
      return(s7_make_ratio(sc, -numerator(p), denominator(p)));
      
    case T_REAL:
      return(s7_make_real(sc, -real(p)));
      
    default:
      return(s7_make_complex(sc, -real_part(p), -imag_part(p)));
    }
}


static s7_pointer big_subtract(s7_scheme *sc, s7_pointer args)
{
  int result_type = T_INTEGER;
  s7_pointer x, result;

  if (!s7_is_number(car(args)))
    return(s7_wrong_type_arg_error(sc, "-", 1, car(args), "a number"));
  
  if (is_null(cdr(args))) 
    return(big_negate(sc, args));

  for (x = args; is_not_null(x); x = cdr(x)) 
    {
      s7_pointer p;
      p = car(x);
      if (!is_number(p))
	{
	  if (!is_big_number(p))
	    return(s7_wrong_type_arg_error(sc, "-", position_of(x, args), p, "a number"));
	  else result_type = big_type_to_result_type(result_type, type(p));
	}
      else result_type = normal_type_to_result_type(result_type, type(p));
    }

  if (result_type < T_BIG_INTEGER)
    return(g_subtract(sc, args));

  result = copy_and_promote_number(sc, result_type, car(args));

  for (x = cdr(args); is_not_null(x); x = cdr(x)) 
    {
      s7_pointer arg;
      arg = promote_number(sc, result_type, car(x));

      switch (result_type)
	{
	case T_BIG_INTEGER:
	  mpz_sub(big_integer(result), big_integer(result), big_integer(arg));
	  break;
	  
	case T_BIG_RATIO:
	  mpq_sub(big_ratio(result), big_ratio(result), big_ratio(arg));
	  break;
	  
	case T_BIG_REAL:
	  mpfr_sub(big_real(result), big_real(result), big_real(arg), GMP_RNDN);
	  break;
	  
	case T_BIG_COMPLEX:
	  mpc_sub(big_complex(result), big_complex(result), big_complex(arg), MPC_RNDNN);
	  break;
	}
    }

  switch (result_type)
    {
      case T_BIG_RATIO:
	return(make_big_integer_or_ratio(sc, result));

      case T_BIG_COMPLEX:
	return(make_big_real_or_complex(sc, result));
      }
  return(result);
}


static s7_pointer big_multiply(s7_scheme *sc, s7_pointer args)
{
  int result_type = T_INTEGER;
  s7_pointer x, result;

  if (is_null(args))
    return(small_int(1));

  if ((is_null(cdr(args))) && (s7_is_number(car(args))))
    return(car(args));

  for (x = args; is_not_null(x); x = cdr(x)) 
    {
      s7_pointer p;
      p = car(x);
      if (!is_number(p))
	{
	  if (!is_big_number(p))
	    return(s7_wrong_type_arg_error(sc, "*", position_of(x, args), p, "a number"));
	  else result_type = big_type_to_result_type(result_type, type(p));
	}
      else result_type = normal_type_to_result_type(result_type, type(p));
    }

  if (result_type < T_BIG_INTEGER)
    return(g_multiply(sc, args));

  result = copy_and_promote_number(sc, result_type, car(args));

  for (x = cdr(args); is_not_null(x); x = cdr(x)) 
    {
      s7_pointer arg;
      arg = promote_number(sc, result_type, car(x));
      switch (result_type)
	{
	case T_BIG_INTEGER:
	  mpz_mul(big_integer(result), big_integer(result), big_integer(arg));
	  break;
	  
	case T_BIG_RATIO:
	  mpq_mul(big_ratio(result), big_ratio(result), big_ratio(arg));
	  break;

	case T_BIG_REAL:
	  mpfr_mul(big_real(result), big_real(result), big_real(arg), GMP_RNDN);
	  break;

	case T_BIG_COMPLEX:
	  mpc_mul(big_complex(result), big_complex(result), big_complex(arg), MPC_RNDNN);
	  break;
	}
    }

  switch (result_type)
    {
    case T_BIG_RATIO:
      return(make_big_integer_or_ratio(sc, result));

    case T_BIG_COMPLEX:
      return(make_big_real_or_complex(sc, result));
    }
  return(result);
}


static s7_pointer big_invert(s7_scheme *sc, s7_pointer args)
{
  /* assume cdr(args) is nil and we're called from divide, so check for big num else call g_divide */
  s7_pointer p, x;

  p = car(args);
  if (big_is_zero_1(sc, p) == sc->T)
    return(division_by_zero_error(sc, "/", p));
  
  switch (type(p))
    {
    case T_INTEGER:
      if (integer(p) == S7_LLONG_MIN)
	{
	  mpz_t n1, d1;

	  NEW_CELL(sc, x);
	  set_type(x, T_BIG_RATIO);

	  mpz_init_set_s7_Int(n1, 1);
	  mpz_init_set_s7_Int(d1, S7_LLONG_MIN);
	  mpq_set_num(big_ratio(x), n1);
	  mpq_set_den(big_ratio(x), d1);
	  mpq_canonicalize(big_ratio(x));
	  mpz_clear(n1);
	  mpz_clear(d1);

	  return(x);
	}
      return(s7_make_ratio(sc, 1, integer(p)));      /* a already checked, not 0 */
      
    case T_RATIO:
      return(s7_make_ratio(sc, denominator(p), numerator(p)));

    case T_REAL:
      return(s7_make_real(sc, 1.0 / real(p)));

    case T_COMPLEX:
      {
	s7_Double r2, i2, den;
	r2 = real_part(p);
	i2 = imag_part(p);
	den = (r2 * r2 + i2 * i2);
	return(s7_make_complex(sc, r2 / den, -i2 / den));
      }

    case T_BIG_INTEGER:
      /* p might be 1 or -1 */
      {
	mpz_t n;

	mpz_init_set_si(n, 1);
	if (mpz_cmp(n, big_integer(p)) == 0)
	  {
	    mpz_clear(n);
	    return(small_int(1));
	  }
	mpz_set_si(n, -1);
	if (mpz_cmp(n, big_integer(p)) == 0)
	  {
	    mpz_clear(n);
	    return(s7_make_integer(sc, -1));
	  }

	NEW_CELL(sc, x);
	set_type(x, T_BIG_RATIO);
	mpq_init(big_ratio(x));

	mpz_set_ui(n, 1);
	mpq_set_num(big_ratio(x), n);
	mpz_clear(n);

	mpq_set_den(big_ratio(x), big_integer(p));
	mpq_canonicalize(big_ratio(x));
	return(x);
      }

    case T_BIG_RATIO:
      {
	mpz_t n;

	mpz_init_set_si(n, 1);
	if (mpz_cmp(n, mpq_numref(big_ratio(p))) == 0)
	  {
	    mpz_clear(n);
	    return(mpz_to_big_integer(sc, mpq_denref(big_ratio(p))));
	  }
	mpz_set_si(n, -1);
	if (mpz_cmp(n, mpq_numref(big_ratio(p))) == 0)
	  {
	    mpz_clear(n);
	    x = mpz_to_big_integer(sc, mpq_denref(big_ratio(p)));
	    mpz_neg(big_integer(x), big_integer(x));
	    return(x);
	  }
	mpz_clear(n);

	NEW_CELL(sc, x);
	set_type(x, T_BIG_RATIO);

	mpq_init(big_ratio(x));
	mpq_set_num(big_ratio(x), mpq_denref(big_ratio(p)));
	mpq_set_den(big_ratio(x), mpq_numref(big_ratio(p)));
	mpq_canonicalize(big_ratio(x));
	return(x);
      }

    case T_BIG_REAL:
      x = mpfr_to_big_real(sc, big_real(p));
      mpfr_ui_div(big_real(x), 1, big_real(x), GMP_RNDN);
      return(x);

    default:
      x = mpc_to_big_complex(sc, big_complex(p));
      mpc_ui_div(big_complex(x), 1, big_complex(x), MPC_RNDNN);
      return(x);
    }
}


static s7_pointer big_divide(s7_scheme *sc, s7_pointer args)
{
  int result_type = T_INTEGER;
  s7_pointer x, divisor, result;

  if (!s7_is_number(car(args)))
    return(s7_wrong_type_arg_error(sc, "/", 1, car(args), "a number"));
  
  if (is_null(cdr(args))) 
    return(big_invert(sc, args));

  for (x = args; is_not_null(x); x = cdr(x)) 
    {
      s7_pointer p;
      p = car(x);
      if (!is_number(p))
	{
	  if (!is_big_number(p))
	    return(s7_wrong_type_arg_error(sc, "/", position_of(x, args), p, "a number"));
	  else result_type = big_type_to_result_type(result_type, type(p));
	}
      else result_type = normal_type_to_result_type(result_type, type(p));
    }

  if (result_type < T_BIG_INTEGER)
    return(g_divide(sc, args));

  divisor = copy_and_promote_number(sc, result_type, cadr(args));

  for (x = cddr(args); is_not_null(x); x = cdr(x)) 
    {
      s7_pointer arg;
      arg = promote_number(sc, result_type, car(x));
      switch (result_type)
	{
	case T_BIG_INTEGER:
	  mpz_mul(big_integer(divisor), big_integer(divisor), big_integer(arg));
	  break;
	  
	case T_BIG_RATIO:
	  mpq_mul(big_ratio(divisor), big_ratio(divisor), big_ratio(arg));
	  break;
	  
	case T_BIG_REAL:
	  mpfr_mul(big_real(divisor), big_real(divisor), big_real(arg), GMP_RNDN);
	  break;
	  
	case T_BIG_COMPLEX:
	  mpc_mul(big_complex(divisor), big_complex(divisor), big_complex(arg), MPC_RNDNN);
	  break;
	}
    }
    
  if (big_is_zero_1(sc, divisor) == sc->T)
    return(division_by_zero_error(sc, "/", args));

  result = copy_and_promote_number(sc, result_type, car(args));

  switch (result_type)
    {
    case T_BIG_INTEGER:
      {
	NEW_CELL(sc, x);
	set_type(x, T_BIG_RATIO);

	mpq_init(big_ratio(x));
	mpq_set_num(big_ratio(x), big_integer(result));
	mpq_set_den(big_ratio(x), big_integer(divisor));
	mpq_canonicalize(big_ratio(x));

	if (mpz_cmp_ui(mpq_denref(big_ratio(x)), 1) == 0)
	  return(mpz_to_big_integer(sc, mpq_numref(big_ratio(x))));
	return(x);
      }
      
    case T_BIG_RATIO:
      mpq_div(big_ratio(result), big_ratio(result), big_ratio(divisor));
      return(make_big_integer_or_ratio(sc, result));
      break;
      
    case T_BIG_REAL:
      mpfr_div(big_real(result), big_real(result), big_real(divisor), GMP_RNDN);
      break;
      
    case T_BIG_COMPLEX:
      mpc_div(big_complex(result), big_complex(result), big_complex(divisor), MPC_RNDNN);
      return(make_big_real_or_complex(sc, result));
      break;
    }

  return(result);
}


static s7_pointer big_numerator(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;

  p = car(args);
  switch (type(p))
    {
    case T_BIG_INTEGER:
      return(p);

    case T_BIG_RATIO:
      return(mpz_to_big_integer(sc, mpq_numref(big_ratio(p))));

    default:
      return(g_numerator(sc, args));
    }
}


static s7_pointer big_denominator(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;

  p = car(args);
  switch (type(p))
    {
    case T_BIG_INTEGER:
      return(small_int(1));

    case T_BIG_RATIO:
      return(mpz_to_big_integer(sc, mpq_denref(big_ratio(p))));

    default:
      return(g_denominator(sc, args));
    }
}


static s7_pointer big_is_even(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;
  p = car(args);
  if (type(p) == T_BIG_INTEGER)
    return(make_boolean(sc, mpz_even_p(big_integer(p))));
  return(g_is_even(sc, args));
}


static s7_pointer big_is_odd(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p;
  p = car(args);
  if (type(p) == T_BIG_INTEGER)
    return(make_boolean(sc, mpz_odd_p(big_integer(p))));
  return(g_is_odd(sc, args));
}


static s7_pointer big_real_part(s7_scheme *sc, s7_pointer args)
{
  #define H_real_part "(real-part num) returns the real part of num"
  s7_pointer p;

  p = car(args);
  switch (type(p))
    {
    case T_INTEGER:
    case T_RATIO:
    case T_REAL:
    case T_BIG_INTEGER:
    case T_BIG_RATIO:
    case T_BIG_REAL:
      return(p);

    case T_COMPLEX:
      return(s7_make_real(sc, real_part(p)));

    case T_BIG_COMPLEX:
      {
	s7_pointer x;
       
	NEW_CELL(sc, x);
	set_type(x, T_BIG_REAL);
	mpfr_init(big_real(x));
	mpc_real(big_real(x), big_complex(p), GMP_RNDN);
	
	return(x);
      }

    default:
      return(s7_wrong_type_arg_error(sc, "real-part", 0, p, "a number"));
    }
}


static s7_pointer big_imag_part(s7_scheme *sc, s7_pointer args)
{
  #define H_imag_part "(imag-part num) returns the imaginary part of num"
  s7_pointer p;
  p = car(args);

  switch (type(p))
    {
    case T_INTEGER:   
    case T_RATIO:   
    case T_BIG_INTEGER:   
    case T_BIG_RATIO:   
      return(small_int(0));

    case T_REAL:    
    case T_BIG_REAL:    
      return(real_zero);

    case T_COMPLEX: 
      return(s7_make_real(sc, imag_part(p)));

    case T_BIG_COMPLEX:
      {
	s7_pointer x;
       
	NEW_CELL(sc, x);
	set_type(x, T_BIG_REAL);
	mpfr_init(big_real(x));
	mpc_imag(big_real(x), big_complex(p), GMP_RNDN);
	
	return(x);
      }

    default:        
      return(s7_wrong_type_arg_error(sc, "imag-part", 0, p, "a number"));
    }
}


static s7_pointer big_abs(s7_scheme *sc, s7_pointer args)
{
  #define H_abs "(abs x) returns the absolute value of the real number x"
  s7_pointer p, x;

  p = car(args);
  switch (type(p))
    {
    case T_INTEGER:
      if (integer(p) < 0)
	{
	  if (integer(p) == S7_LLONG_MIN)
	    {
	      x = s7_Int_to_big_integer(sc, integer(p));
	      mpz_neg(big_integer(x), big_integer(x));
	      return(x);
	    }
	  return(make_integer(sc, -integer(p)));
	}
      return(p);

    case T_RATIO:
      if (numerator(p) < 0)
	return(s7_make_ratio(sc, -numerator(p), denominator(p)));
      return(p);

    case T_REAL:
      if (real(p) < 0.0)
	return(s7_make_real(sc, -real(p)));
      return(p);

    case T_BIG_INTEGER:
      x = mpz_to_big_integer(sc, big_integer(p));
      mpz_abs(big_integer(x), big_integer(x));
      return(x);

    case T_BIG_RATIO:
      x = mpq_to_big_ratio(sc, big_ratio(p));
      mpq_abs(big_ratio(x), big_ratio(x));
      return(x);

    case T_BIG_REAL:
      x = mpfr_to_big_real(sc, big_real(p));
      mpfr_abs(big_real(x), big_real(x), GMP_RNDN);
      return(x);

    default:
      return(s7_wrong_type_arg_error(sc, "abs", 0, p, "a real"));
    }
}


static s7_pointer big_magnitude(s7_scheme *sc, s7_pointer args)
{
  #define H_magnitude "(magnitude z) returns the magnitude of z"
  s7_pointer p;

  p = car(args);
  if (!s7_is_number(p))
    return(s7_wrong_type_arg_error(sc, "magnitude", 0, p, "a number"));

  if (type(p) == T_BIG_COMPLEX)
    {
      mpfr_t n;
      mpfr_init(n);
      mpc_abs(n, big_complex(p), GMP_RNDN);
      p = mpfr_to_big_real(sc, n);
      mpfr_clear(n);
      return(p);
    }

  if (type(p) == T_COMPLEX)
    return(s7_make_real(sc, hypot(imag_part(p), real_part(p))));

  return(big_abs(sc, args));
}


static s7_pointer big_angle(s7_scheme *sc, s7_pointer args)
{
  #define H_angle "(angle z) returns the angle of z"
  s7_pointer p;

  p = car(args);
  switch (type(p))
    {
    case T_INTEGER:
      if (integer(p) < 0)
	return(real_pi);
      return(small_int(0));

    case T_RATIO:
      if (numerator(p) < 0)
	return(real_pi);
      return(small_int(0));

    case T_REAL:
      if (is_NaN(real(p))) return(p);
      if (real(p) < 0.0)
	return(real_pi);
      return(real_zero);

    case T_COMPLEX:
      return(s7_make_real(sc, atan2(imag_part(p), real_part(p))));

    case T_BIG_INTEGER:
      if (mpz_cmp_ui(big_integer(p), 0) >= 0)
	return(small_int(0));
      return(big_pi(sc));

    case T_BIG_RATIO:
      if (mpq_cmp_ui(big_ratio(p), 0, 1) >= 0)
	return(small_int(0));
      return(big_pi(sc));

    case T_BIG_REAL:
      {
	double x;
	x = mpfr_get_d(big_real(p), GMP_RNDN);
	/* mpfr_get_d returns inf or -inf if the arg is too large for a double */
	if (is_NaN(x)) return(p);
	if (x >= 0.0)
	  return(real_zero);
	return(big_pi(sc));
      }

    case T_BIG_COMPLEX:
      {
	s7_pointer x;
	NEW_CELL(sc, x);
	set_type(x, T_BIG_REAL);
	mpfr_init(big_real(x));
	mpc_arg(big_real(x), big_complex(p), GMP_RNDN);
	return(x);
      }
      
    default:
      return(s7_wrong_type_arg_error(sc, "angle", 0, p, "a number"));
    }
}


static s7_pointer big_make_rectangular(s7_scheme *sc, s7_pointer args)
{
  #define H_make_rectangular "(make-rectangular x1 x2) returns a complex number with real-part x1 and imaginary-part x2"

  s7_pointer p0, p1, p;
  mpfr_t rl, im;
  double x;

  p0 = car(args);
  if (!s7_is_real(p0))
    return(s7_wrong_type_arg_error(sc, "make-rectangular real part,", 1, p0, "a real"));

  p1 = cadr(args);
  if (!s7_is_real(p1))
    return(s7_wrong_type_arg_error(sc, "make-rectangular imaginary part,", 2, p1, "a real"));
    
  if ((!is_big_number(p1)) && (s7_real(p1) == 0.0)) /* imag-part is not bignum and is 0.0 */
    return(p0);

  mpfr_init_set(im, big_real(promote_number(sc, T_BIG_REAL, p1)), GMP_RNDN);
  x = mpfr_get_d(im, GMP_RNDN);
  if (x == 0.0)                                   /* imag-part is bignum 0.0 */
    {
      mpfr_clear(im);
      return(p0);
    }

  mpfr_init_set(rl, big_real(promote_number(sc, T_BIG_REAL, p0)), GMP_RNDN);

  NEW_CELL(sc, p);
  set_type(p, T_BIG_COMPLEX);
  mpc_init(big_complex(p));
  mpc_set_fr_fr(big_complex(p), rl, im, MPC_RNDNN);

  mpfr_clear(rl);
  mpfr_clear(im);
  return(p);
}


/* (make-polar 0 (real-part (log 0))) = 0? or nan?
 */

static s7_pointer big_make_polar(s7_scheme *sc, s7_pointer args)
{
  #define H_make_polar "(make-polar mag ang) returns a complex number with magnitude mag and angle ang"

  s7_pointer p0, p1, p;
  mpfr_t ang, mag, rl, im;
  double x, y;

  p0 = car(args);
  if (!s7_is_real(p0))
    return(s7_wrong_type_arg_error(sc, "make-polar magnitude,", 1, p0, "a real"));

  p1 = cadr(args);
  if (!s7_is_real(p1))
    return(s7_wrong_type_arg_error(sc, "make-polar angle,", 2, p1, "a real"));
  
  mpfr_init_set(ang, big_real(promote_number(sc, T_BIG_REAL, p1)), GMP_RNDN);
  y = mpfr_get_d(ang, GMP_RNDN);

  if (is_NaN(y))
    {
      mpfr_clear(ang);
      return(real_NaN);
    }

  mpfr_init_set(mag, big_real(promote_number(sc, T_BIG_REAL, p0)), GMP_RNDN);
  x = mpfr_get_d(mag, GMP_RNDN);

  if (is_NaN(x))
    {
      mpfr_clear(ang);
      mpfr_clear(mag);
      return(real_NaN);
    }

  if ((x == 0.0) || (y == 0.0))
    {
      mpfr_clear(ang);
      mpfr_clear(mag);
      return(p0);
    }

  mpfr_init_set(im, ang, GMP_RNDN);
  mpfr_sin(im, im, GMP_RNDN);
  mpfr_mul(im, im, mag, GMP_RNDN);
      
  x = mpfr_get_d(im, GMP_RNDN);
  if (x == 0.0)
    {
      mpfr_clear(im);
      mpfr_clear(ang);
      mpfr_clear(mag);
      return(p0);
    }

  mpfr_init_set(rl, ang, GMP_RNDN);
  mpfr_cos(rl, rl, GMP_RNDN);
  mpfr_mul(rl, rl, mag, GMP_RNDN);

  NEW_CELL(sc, p);
  set_type(p, T_BIG_COMPLEX);
  mpc_init(big_complex(p));
  mpc_set_fr_fr(big_complex(p), rl, im, MPC_RNDNN);

  mpfr_clear(rl);
  mpfr_clear(im);
  mpfr_clear(ang);
  mpfr_clear(mag);
  return(p);
}


static s7_pointer big_log(s7_scheme *sc, s7_pointer args)
{
  #define H_log "(log z1 (z2 e)) returns log(z1) / log(z2) where z2 (the base) defaults to e: (log 8 2) = 3"
  /* either arg can be big, 2nd is optional */
  s7_pointer p0, p1 = NULL, p;

  p0 = car(args);
  if (!s7_is_number(p0))
    return(s7_wrong_type_arg_error(sc, "log", (is_null(cdr(args))) ? 0 : 1, p0, "a number"));

  if (is_not_null(cdr(args)))
    {
      p1 = cadr(args);
      if (!s7_is_number(p1))
	return(s7_wrong_type_arg_error(sc, "log base,", 2, p1, "a number"));
    }

  if ((s7_is_real(p0)) &&
      ((!p1) || (s7_is_real(p1))))
    {
      double x, y = 0.0;

      p0 = promote_number(sc, T_BIG_REAL, p0);
      x = mpfr_get_d(big_real(p0), GMP_RNDN);
      if (is_NaN(x))
	return(real_NaN);

      if (p1) 
	{
	  p1 = promote_number(sc, T_BIG_REAL, p1);
	  y = mpfr_get_d(big_real(p1), GMP_RNDN);

	  /* we can't check y here for 1.0 (check for 0.0 apparently is ok):
	   *  :(log 100.0 (+ 1.0 (bignum "1e-16")))
	   *  ;log base, argument 2, 1.000000000000000100000000000000000000002E0, is out of range (can't be 0.0 or 1.0)
	   *  :(= 1.0 (+ 1.0 (bignum "1e-16")))
	   *  #f
	   */
	  if (is_NaN(y))
	    return(real_NaN);
	  if (y == 0.0)
	    return(s7_out_of_range_error(sc, "log base,", 2, p1, "can't be 0.0"));
	}
      if (x == 0.0)
	return(s7_make_complex(sc, -INFINITY, M_PI));

      if ((x > 0.0) && (y >= 0.0))
	{
	  mpfr_t n, base;

	  mpfr_init_set(n, big_real(p0), GMP_RNDN);
	  mpfr_log(n, n, GMP_RNDN);

	  if (!p1)
	    {
	      /* presumably log is safe with regard to real-part overflow giving a bogus int? */
	      if ((s7_is_rational(car(args))) &&
		  (mpfr_integer_p(n) != 0))
		{
		  NEW_CELL(sc, p);
		  set_type(p, T_BIG_INTEGER);
		  mpz_init(big_integer(p));
		  mpfr_get_z(big_integer(p), n, GMP_RNDN);
		}
	      else p = mpfr_to_big_real(sc, n);
	      mpfr_clear(n);
	      return(p);
	    }

	  mpfr_init_set(base, big_real(p1), GMP_RNDN);
	  mpfr_log(base, base, GMP_RNDN);
	  mpfr_div(n, n, base, GMP_RNDN);
	  mpfr_clear(base);

	  if ((s7_is_rational(car(args))) &&
	      (s7_is_rational(cadr(args))) &&
	      (mpfr_integer_p(n) != 0))
	    {
	      NEW_CELL(sc, p);
	      set_type(p, T_BIG_INTEGER);
	      mpz_init(big_integer(p));
	      mpfr_get_z(big_integer(p), n, GMP_RNDN);
	    }
	  else p = mpfr_to_big_real(sc, n);
	  mpfr_clear(n);
	  return(p);
	}
    }

  p0 = promote_number(sc, T_BIG_COMPLEX, p0);
  if (p1) p1 = promote_number(sc, T_BIG_COMPLEX, p1);
  {
    mpc_t n, base;
    double x;

    mpc_init(n);
    mpc_set(n, big_complex(p0), MPC_RNDNN);
    mpc_log(n, n, MPC_RNDNN);
    if (!p1)
      {
	p = mpc_to_big_complex(sc, n);
	mpc_clear(n);
	return(p);
      }

    mpc_init(base);
    mpc_set(base, big_complex(p1), MPC_RNDNN);
    mpc_log(base, base, MPC_RNDNN);
    mpc_div(n, n, base, MPC_RNDNN);
    mpc_clear(base);

    x = mpfr_get_d(mpc_imagref(n), GMP_RNDN);
    if (x == 0.0)
      p = mpfr_to_big_real(sc, mpc_realref(n));
    else p = mpc_to_big_complex(sc, n);

    mpc_clear(n);
    return(p);
  }
}


static s7_pointer big_sqrt(s7_scheme *sc, s7_pointer args)
{
  /* real >= 0 -> real, else complex */
  #define H_sqrt "(sqrt z) returns the square root of z"

  s7_pointer p;

  p = car(args);
  if (!s7_is_number(p))
    return(s7_wrong_type_arg_error(sc, "sqrt", 0, p, "a number"));

  p = to_big(sc, p);
    
  /* if big integer, try to return int if perfect square */
  if (type(p) == T_BIG_INTEGER)
    {
      if (mpz_cmp_ui(big_integer(p), 0) < 0)
	p = promote_number(sc, T_BIG_COMPLEX, p);
      else
	{
	  mpz_t n, rem;

	  mpz_init(rem);
	  mpz_init_set(n, big_integer(p));
	  mpz_sqrtrem(n, rem, n);

	  if (mpz_cmp_ui(rem, 0) == 0)
	    {
	      p = mpz_to_big_integer(sc, n);
	      mpz_clear(n);
	      mpz_clear(rem);
	      return(p);
	    }
	  mpz_clear(n);
	  mpz_clear(rem);
	  p = promote_number(sc, T_BIG_REAL, p);
	}
    }
  
  /* if big ratio, check both num and den for squares */
  if (type(p) == T_BIG_RATIO)
    {
      if (mpq_cmp_ui(big_ratio(p), 0, 1) < 0)
	p = promote_number(sc, T_BIG_COMPLEX, p);
      else
	{
	  mpz_t n1, rem;
	  mpz_init(rem);
	  mpz_init_set(n1, mpq_numref(big_ratio(p)));
	  mpz_sqrtrem(n1, rem, n1);

	  if (mpz_cmp_ui(rem, 0) == 0)
	    {
	      mpz_t d1;
	      mpz_init_set(d1, mpq_denref(big_ratio(p)));
	      mpz_sqrtrem(d1, rem, d1);

	      if (mpz_cmp_ui(rem, 0) == 0)
		{
		  mpq_t n;
		  mpq_init(n);
		  mpq_set_num(n, n1);
		  mpq_set_den(n, d1);
		  mpq_canonicalize(n);
		  p = mpq_to_big_ratio(sc, n);
		  mpz_clear(n1);
		  mpz_clear(d1);
		  mpz_clear(rem);
		  mpq_clear(n);
		  return(p);
		}
	      mpz_clear(d1);
	    }

	  mpz_clear(n1);
	  mpz_clear(rem);
	  p = promote_number(sc, T_BIG_REAL, p);
	}
    }
  
  /* if real and not negative, use mpfr_sqrt */
  if (type(p) == T_BIG_REAL)
    {
      if (mpfr_cmp_ui(big_real(p), 0) < 0)
	p = promote_number(sc, T_BIG_COMPLEX, p);
      else
	{
	  mpfr_t n;
	  mpfr_init_set(n, big_real(p), GMP_RNDN);
	  mpfr_sqrt(n, n, GMP_RNDN);
	  p = mpfr_to_big_real(sc, n);
	  mpfr_clear(n);
	  return(p);
	}
    }
  
  /* p is a big number, so it must be complex at this point */
  {
    mpc_t n;
    mpc_init(n);
    mpc_set(n, big_complex(p), MPC_RNDNN);
    mpc_sqrt(n, n, MPC_RNDNN);
    p = mpc_to_big_complex(sc, n);
    mpc_clear(n);
    return(p);
  }
}


/* (define (diff f a) (magnitude (- (f a) (f (bignum (number->string a))))))
 * (sin 1e15+1e15i) hangs in mpc 0.8.2, but appears to be fixed in the current svn sources
 */

enum {TRIG_NO_CHECK, TRIG_TAN_CHECK, TRIG_TANH_CHECK};

static s7_pointer big_trig(s7_scheme *sc, s7_pointer args, 
			   int (*mpfr_trig)(mpfr_ptr, mpfr_srcptr, mpfr_rnd_t),
			   int (*mpc_trig)(mpc_ptr, mpc_srcptr, mpc_rnd_t),
			   int tan_case, const char *caller)
/* these declarations mimic the mpfr.h and mpc.h declarations.  It seems to me that
 *   they ought to be:
 *			   int (*mpfr_trig)(mpfr_t rop, mpfr_t op, mp_rnd_t rnd),
 *			   void (*mpc_trig)(mpc_t rop, mpc_t op, mpc_rnd_t rnd))
 */
{
  s7_pointer p;
  p = car(args);

  /* I think here we should always promote to bignum (otherwise, for example, (exp 800) -> inf)
   */
  if (!s7_is_number(p))
    return(s7_wrong_type_arg_error(sc, caller, 0, p, "a number"));
    
  if (s7_is_real(p))
    {
      mpfr_t n;
      mpfr_init_set(n, big_real(promote_number(sc, T_BIG_REAL, p)), GMP_RNDN);
      mpfr_trig(n, n, GMP_RNDN);
      /* it's confusing to check for ints here via mpfr_integer_p because it
       *   is dependent on the precision!  (exp 617/5) returns an integer if
       *   precision is 128, but a float if 512.
       */
      p = mpfr_to_big_real(sc, n);
      mpfr_clear(n);
      return(p);
    }

  if (!is_big_number(p))
    p = promote_number(sc, T_BIG_COMPLEX, p);

  if (tan_case == TRIG_TAN_CHECK)
    {
      if ((MPC_INEX_IM(mpc_cmp_si_si(big_complex(p), 1, 350))) > 0)
	return(s7_make_complex(sc, 0.0, 1.0));	
      if ((MPC_INEX_IM(mpc_cmp_si_si(big_complex(p), 1, -350))) < 0)
	return(s7_make_complex(sc, 0.0, -1.0));	
    }
  
  if (tan_case == TRIG_TANH_CHECK)
    {
      if ((MPC_INEX_RE(mpc_cmp_si_si(big_complex(p), 350, 1))) > 0)
	return(real_one);
      if ((MPC_INEX_RE(mpc_cmp_si_si(big_complex(p), -350, 1))) < 0)
	return(s7_make_real(sc, -1.0));
    }
  
  {
    mpc_t n;
    double ix;

    mpc_init(n);
    mpc_trig(n, big_complex(p), MPC_RNDNN);  
    /* (sin (bignum "1e15+1e15i")) causes mpc to hang (e9 is ok, but e10 hangs) 
     *   (sin (bignum "0+1e10i")) -> 0+inf (sin (bignum "1+1e10i")) hangs
     *
     * before comparing imag-part to 0, we need to look for NaN and inf, else:
     *    (sinh 0+0/0i) -> 0.0
     *    (sinh (log 0.0)) -> inf.0
     */

    ix = mpfr_get_d(mpc_imagref(n), GMP_RNDN);
    if (ix == 0.0)
      {
	mpfr_t z;

	mpfr_init_set(z, mpc_realref(n), GMP_RNDN);
	p = mpfr_to_big_real(sc, z);
	mpfr_clear(z);
      }
    else p = mpc_to_big_complex(sc, n);
    mpc_clear(n);
    return(p);
  }
}


static s7_pointer big_sin(s7_scheme *sc, s7_pointer args)
{
  #define H_sin "(sin z) returns sin(z)"
  return(big_trig(sc, args, mpfr_sin, mpc_sin, TRIG_NO_CHECK, "sin"));
}


static s7_pointer big_cos(s7_scheme *sc, s7_pointer args)
{
  #define H_cos "(cos z) returns cos(z)"
  return(big_trig(sc, args, mpfr_cos, mpc_cos, TRIG_NO_CHECK, "cos"));
}


static s7_pointer big_tan(s7_scheme *sc, s7_pointer args)
{
  #define H_tan "(tan z) returns tan(z)"
  return(big_trig(sc, args, mpfr_tan, mpc_tan, TRIG_TAN_CHECK, "tan"));
}


static s7_pointer big_sinh(s7_scheme *sc, s7_pointer args)
{
  #define H_sinh "(sinh z) returns sinh(z)"
  /* currently (sinh 0+0/0i) -> 0.0? */
  return(big_trig(sc, args, mpfr_sinh, mpc_sinh, TRIG_NO_CHECK, "sinh"));
}


static s7_pointer big_cosh(s7_scheme *sc, s7_pointer args)
{
  #define H_cosh "(cosh z) returns cosh(z)"
  return(big_trig(sc, args, mpfr_cosh, mpc_cosh, TRIG_NO_CHECK, "cosh"));
}


static s7_pointer big_tanh(s7_scheme *sc, s7_pointer args)
{
  #define H_tanh "(tanh z) returns tanh(z)"
  return(big_trig(sc, args, mpfr_tanh, mpc_tanh, TRIG_TANH_CHECK, "tanh"));
}


static s7_pointer big_exp(s7_scheme *sc, s7_pointer args)
{
  #define H_exp "(exp z) returns e^z, (exp 1) is 2.718281828459"
  return(big_trig(sc, args, mpfr_exp, mpc_exp, TRIG_NO_CHECK, "exp"));
}


static bool s7_is_negative(s7_pointer obj);
static bool s7_is_positive(s7_pointer obj);

static s7_pointer big_expt(s7_scheme *sc, s7_pointer args)
{
  #define H_expt "(expt z1 z2) returns z1^z2"
  s7_pointer x, y, p;

  /* see comment under g_expt
   *  if (is_not_null(cddr(args)))
   *    return(big_expt(sc, list_2(sc, car(args), big_expt(sc, cdr(args)))));
   */

  x = car(args);
  if (!s7_is_number(x))
    return(s7_wrong_type_arg_error(sc, "expt", 1, x, "a number"));

  y = cadr(args);
  if (!s7_is_number(y))
    return(s7_wrong_type_arg_error(sc, "expt power,", 2, y, "a number"));

  if (big_is_zero_1(sc, x) == sc->T)
    {
      if ((s7_is_integer(x)) && 
	  (s7_is_integer(y)) &&
	  (big_is_zero_1(sc, y) == sc->T))
	return(small_int(1));

      if (s7_is_real(y))
	{
	  if (s7_is_negative(y))
	    return(division_by_zero_error(sc, "expt", args));
	}
      else
	{
	  if (s7_is_negative(big_real_part(sc, cdr(args))))
	    return(division_by_zero_error(sc, "expt", args));
	}

      if ((s7_is_rational(x)) && 
	  (s7_is_rational(y)))
	return(small_int(0));

      return(real_zero);
    }

  if (s7_is_integer(y))
    {
      s7_Int yval;

      if (!is_big_number(y))
	yval = s7_integer(y);
      else yval = big_integer_to_s7_Int(big_integer(y));

      if (yval == 0)
	{
	  if (s7_is_rational(x))
	    return(small_int(1));
	  return(real_one);
	}

      if (yval == 1)
	return(x);

      if (!is_big_number(x))
	{
	  if ((s7_is_one(x)) || (s7_is_zero(x)))
	    return(x);
	}

      if ((yval < S7_LONG_MAX) &&
	  (yval > S7_LONG_MIN))
	{
	  if (s7_is_integer(x))
	    {
	      mpz_t n;
	      mpq_t r;

	      x = promote_number(sc, T_BIG_INTEGER, x);
	      mpz_init_set(n, big_integer(x));
	      if (yval >= 0)
		{
		  mpz_pow_ui(n, n, (unsigned int)yval);
		  p = mpz_to_big_integer(sc, n);
		  mpz_clear(n);
		  return(p);
		}

	      mpz_pow_ui(n, n, (unsigned int)(-yval));
	      mpq_init(r);
	      mpq_set_z(r, n);
	      mpq_inv(r, r);
	      if (mpz_cmp_ui(mpq_denref(r), 1) == 0)
		{
		  mpz_t z;
		  mpz_init_set(z, mpq_numref(r));
		  mpq_clear(r);
		  mpz_clear(n);
		  p = mpz_to_big_integer(sc, z);
		  mpz_clear(z);
		  return(p);
		}
	      mpz_clear(n);
	      p = mpq_to_big_ratio(sc, r);
	      mpq_clear(r);
	      return(p);
	    }
	}

      if (s7_is_ratio(x)) /* (here y is an integer) */
	{
	  mpz_t n, d;
	  mpq_t r;

	  if (yval > (S7_LLONG_MAX >> 4)) /* gmp aborts or segfaults if this is too large, but I don't know where the actual boundary is */
	    return(s7_out_of_range_error(sc, "integer expt exponent,", 1, y, "gmp will segfault"));

	  x = promote_number(sc, T_BIG_RATIO, x);
	  mpz_init_set(n, mpq_numref(big_ratio(x)));
	  mpz_init_set(d, mpq_denref(big_ratio(x)));
	  mpq_init(r);
	  if (yval >= 0)
	    {
	      mpz_pow_ui(n, n, (unsigned int)yval);
	      mpz_pow_ui(d, d, (unsigned int)yval);
	      mpq_set_num(r, n);
	      mpq_set_den(r, d);
	    }
	  else
	    {
	      yval = -yval;
	      mpz_pow_ui(n, n, (unsigned int)yval);
	      mpz_pow_ui(d, d, (unsigned int)yval);
	      mpq_set_num(r, d);
	      mpq_set_den(r, n);	      
	      mpq_canonicalize(r);
	    }
	  mpz_clear(n);
	  mpz_clear(d);
	  if (mpz_cmp_ui(mpq_denref(r), 1) == 0)
	    {
	      mpz_t z;
	      mpz_init_set(z, mpq_numref(r));
	      mpq_clear(r);
	      p = mpz_to_big_integer(sc, z);
	      mpz_clear(z);
	      return(p);
	    }
	  p = mpq_to_big_ratio(sc, r);
	  mpq_clear(r);
	  return(p);
	}

      if (s7_is_real(x))
	{
	  mpfr_t z;
	  x = promote_number(sc, T_BIG_REAL, x);
	  mpfr_init_set(z, big_real(x), GMP_RNDN);
	  mpfr_pow_si(z, z, yval, GMP_RNDN);
	  p = mpfr_to_big_real(sc, z);
	  mpfr_clear(z);
	  return(p);
	}
    }

  if ((s7_is_ratio(y)) &&
      (numerator(y) == 1))
    {
      if (denominator(y) == 2)
	return(big_sqrt(sc, args));

      if ((s7_is_real(x)) &&
	  (denominator(y) == 3))
	{
	  mpfr_t z;
	  mpfr_init_set(z, big_real(promote_number(sc, T_BIG_REAL, x)), GMP_RNDN);
	  mpfr_cbrt(z, z, GMP_RNDN);
	  p = mpfr_to_big_real(sc, z);
	  mpfr_clear(z);
	  return(p);
	}
    }

  if ((s7_is_real(x)) &&
      (s7_is_real(y)) &&
      (s7_is_positive(x)))
    {
      mpfr_t z;
      mpfr_init_set(z, big_real(promote_number(sc, T_BIG_REAL, x)), GMP_RNDN);
      mpfr_pow(z, z, big_real(promote_number(sc, T_BIG_REAL, y)), GMP_RNDN);
      p = mpfr_to_big_real(sc, z);
      mpfr_clear(z);
      return(p);
    }

  {
    mpc_t cy;
    mpc_t z;
    
    x = promote_number(sc, T_BIG_COMPLEX, x);
    y = promote_number(sc, T_BIG_COMPLEX, y);
    
    mpc_init(z);
    mpc_set(z, big_complex(x), MPC_RNDNN);
    
    if (mpc_cmp_si_si(z, 0, 0) == 0)
      {
	mpc_clear(z);
	return(small_int(0));
      }
    
    if (mpc_cmp_si_si(z, 1, 0) == 0)
      {
	mpc_clear(z);
	return(small_int(1));
      }
    
    mpc_init(cy);
    mpc_set(cy, big_complex(y), MPC_RNDNN);
    mpc_pow(z, z, cy, MPC_RNDNN);
    mpc_clear(cy);
    
    if (mpfr_cmp_ui(mpc_imagref(z), 0) == 0)
      {
	mpfr_t n;
	if ((s7_is_rational(car(args))) &&
	    (s7_is_rational(cadr(args))) &&
	    (mpfr_integer_p(mpc_realref(z)) != 0))
	  {
	    /* mpfr_integer_p can be confused: (expt 2718/1000 (bignum "617/5")) returns an int if precision=128, float if 512 */
	    /*   so first make sure we're within (say) 31 bits */
	    mpfr_t zi;
	    mpfr_init_set_ui(zi, S7_LONG_MAX, GMP_RNDN);
	    if (mpfr_cmpabs(mpc_realref(z), zi) < 0)
	      {
		mpz_t k;
		mpz_init(k);
		mpfr_get_z(k, mpc_realref(z), GMP_RNDN);
		mpc_clear(z);
		mpfr_clear(zi);
		p = mpz_to_big_integer(sc, k);
		mpz_clear(k);
		return(p);
	      }
	    mpfr_clear(zi);
	  }

	mpfr_init_set(n, mpc_realref(z), GMP_RNDN);
	mpc_clear(z);
	p = mpfr_to_big_real(sc, n);
	mpfr_clear(n);
	return(p);
      }
    
    p = mpc_to_big_complex(sc, z);
    mpc_clear(z);
    return(p);
  }
}


static s7_pointer big_asinh(s7_scheme *sc, s7_pointer args)
{
  #define H_asinh "(asinh z) returns asinh(z)"
  s7_pointer p;

  p = car(args);
  if (!s7_is_number(p))
    return(s7_wrong_type_arg_error(sc, "asinh", 0, p, "a number"));
  
  if (s7_is_real(p))
    {
      mpfr_t n;
      p = promote_number(sc, T_BIG_REAL, p);
      mpfr_init_set(n, big_real(p), GMP_RNDN);
      mpfr_asinh(n, n, GMP_RNDN);
      p = mpfr_to_big_real(sc, n);
      mpfr_clear(n);
      return(p);
    }

  {
    mpc_t n;
    p = promote_number(sc, T_BIG_COMPLEX, p);
    mpc_init(n);
    mpc_set(n, big_complex(p), MPC_RNDNN);
    mpc_asinh(n, n, MPC_RNDNN);
    p = mpc_to_big_complex(sc, n);
    mpc_clear(n);
    return(p);
  }
}


static s7_pointer big_acosh(s7_scheme *sc, s7_pointer args)
{
  #define H_acosh "(acosh z) returns acosh(z)"
  s7_pointer p;
  double x;
  mpc_t n;

  p = car(args);
  if (!s7_is_number(p))
    return(s7_wrong_type_arg_error(sc, "acosh", 0, p, "a number"));

  p = promote_number(sc, T_BIG_COMPLEX, p);
  
  mpc_init(n);
  mpc_set(n, big_complex(p), MPC_RNDNN);
  mpc_acosh(n, n, MPC_RNDNN);
  
  x = mpfr_get_d(mpc_imagref(n), GMP_RNDN);
  if (x == 0.0)
    p = mpfr_to_big_real(sc, mpc_realref(n));
  else p = mpc_to_big_complex(sc, n);
  mpc_clear(n);
  return(p);
}


static s7_pointer big_atanh(s7_scheme *sc, s7_pointer args)
{
  #define H_atanh "(atanh z) returns atanh(z)"
  s7_pointer p;

  p = car(args);
  if (!s7_is_number(p))
    return(s7_wrong_type_arg_error(sc, "atanh", 0, p, "a number"));

  if (s7_is_real(p))
    {
      bool ok;
      mpfr_t temp;
      p = promote_number(sc, T_BIG_REAL, p);
      mpfr_init_set_ui(temp, 1, GMP_RNDN);
      ok = (mpfr_cmpabs(big_real(p), temp) < 0);
      mpfr_clear(temp);
      if (ok)
	{
	  mpfr_t n;
	  mpfr_init_set(n, big_real(p), GMP_RNDN);
	  mpfr_atanh(n, n, GMP_RNDN);
	  p = mpfr_to_big_real(sc, n);
	  mpfr_clear(n);
	  return(p);
	}
    }

  {
    mpc_t n;
    p = promote_number(sc, T_BIG_COMPLEX, p);
    mpc_init(n);
    mpc_set(n, big_complex(p), MPC_RNDNN);
    mpc_atanh(n, n, MPC_RNDNN);
    p = mpc_to_big_complex(sc, n);
    mpc_clear(n);
    return(p);
  }
}


static s7_pointer big_atan(s7_scheme *sc, s7_pointer args)
{
  #define H_atan "(atan z) returns atan(z), (atan y x) returns atan(y/x)"
  s7_pointer p0, p1 = NULL, p;

  p0 = car(args);
  if (!s7_is_number(p0))
    return(s7_wrong_type_arg_error(sc, "atan", 0, p0, "a number"));

  if (is_not_null(cdr(args)))
    {
      p1 = cadr(args);
      if (!s7_is_real(p1))
	return(s7_wrong_type_arg_error(sc, "atan", 2, p1, "a real number"));
      if (!s7_is_real(p0))
	return(s7_wrong_type_arg_error(sc, "atan", 1, p0, "if 2 args, both should be real"));

      p1 = promote_number(sc, T_BIG_REAL, p1);
    }

  if (s7_is_real(p0))
    {
      mpfr_t n;
      p0 = promote_number(sc, T_BIG_REAL, p0);
      mpfr_init_set(n, big_real(p0), GMP_RNDN);
      if (!p1)
	mpfr_atan(n, n, GMP_RNDN);
      else mpfr_atan2(n, n, big_real(p1), GMP_RNDN);
      p = mpfr_to_big_real(sc, n);
      mpfr_clear(n);
      return(p);
    }

  {
    mpc_t n;
    p0 = promote_number(sc, T_BIG_COMPLEX, p0);
    mpc_init_set(n, big_complex(p0), MPC_RNDNN);
    mpc_atan(n, n, MPC_RNDNN);
    p = mpc_to_big_complex(sc, n);
    mpc_clear(n);
    return(p);
  }
}


static s7_pointer big_acos(s7_scheme *sc, s7_pointer args)
{
  #define H_acos "(acos z) returns acos(z); (cos (acos 1)) = 1"
  s7_pointer p;

  p = car(args);
  if (!s7_is_number(p))
    return(s7_wrong_type_arg_error(sc, "acos", 0, p, "a number"));

  if (s7_is_real(p))
    {
      bool ok;
      mpfr_t temp;
      mpfr_t n;
      p = promote_number(sc, T_BIG_REAL, p);
      mpfr_init_set(n, big_real(p), GMP_RNDN);
      mpfr_init_set_ui(temp, 1, GMP_RNDN);
      ok = (mpfr_cmpabs(n, temp) <= 0);
      mpfr_clear(temp);
      if (ok)
	{
	  mpfr_acos(n, n, GMP_RNDN);
	  p = mpfr_to_big_real(sc, n);
	  mpfr_clear(n);
	  return(p);
	}
      mpfr_clear(n);
    }

  {
    mpc_t n;
    p = promote_number(sc, T_BIG_COMPLEX, p);
    mpc_init_set(n, big_complex(p), MPC_RNDNN);
    mpc_acos(n, n, MPC_RNDNN);
    p = mpc_to_big_complex(sc, n);
    mpc_clear(n);
    return(p);
  }
}


static s7_pointer big_asin(s7_scheme *sc, s7_pointer args)
{
  #define H_asin "(asin z) returns asin(z); (sin (asin 1)) = 1"
  s7_pointer p;

  p = car(args);
  if (!s7_is_number(p))
    return(s7_wrong_type_arg_error(sc, "asin", 0, p, "a number"));

  if (s7_is_real(p))
    {
      bool ok;
      mpfr_t temp;
      mpfr_t n;
      p = promote_number(sc, T_BIG_REAL, p);
      mpfr_init_set(n, big_real(p), GMP_RNDN);
      mpfr_init_set_ui(temp, 1, GMP_RNDN);
      ok = (mpfr_cmpabs(n, temp) <= 0);
      mpfr_clear(temp);
      if (ok)
	{
	  mpfr_asin(n, n, GMP_RNDN);
	  p = mpfr_to_big_real(sc, n);
	  mpfr_clear(n);
	  return(p);
	}
      mpfr_clear(n);
    }

  {
    mpc_t n;
    p = promote_number(sc, T_BIG_COMPLEX, p);
    mpc_init_set(n, big_complex(p), MPC_RNDNN);
    mpc_asin(n, n, MPC_RNDNN);
    p = mpc_to_big_complex(sc, n);
    mpc_clear(n);
    return(p);
  }
}


static s7_pointer big_is_zero_1(s7_scheme *sc, s7_pointer p)
{
  switch (type(p))
    {
    case T_INTEGER: 
      return(make_boolean(sc, integer(p) == 0));

    case T_RATIO:
    case T_BIG_RATIO:
      return(sc->F);

    case T_REAL:    
      return(make_boolean(sc, real(p) == 0.0));

    case T_COMPLEX:
    case T_BIG_COMPLEX:
      return(sc->F);

    case T_BIG_INTEGER:
      return(make_boolean(sc, mpz_cmp_ui(big_integer(p), 0) == 0));

    case T_BIG_REAL:
      return(make_boolean(sc, mpfr_zero_p(big_real(p))));

    default:
      return(s7_wrong_type_arg_error(sc, "zero?", 0, p, "a number"));
    }
}


static s7_pointer big_is_zero(s7_scheme *sc, s7_pointer args)
{
  #define H_is_zero "(zero? num) returns #t if the number num is zero"
  return(big_is_zero_1(sc, car(args)));
}


static bool s7_is_positive(s7_pointer obj)
{
  switch (type(obj))
    {
    case T_INTEGER: 
      return(integer(obj) > 0);

    case T_RATIO: 
      return(numerator(obj) > 0);

    case T_REAL:
      return(real(obj) > 0);

    case T_BIG_INTEGER:
      return(mpz_cmp_ui(big_integer(obj), 0) > 0);

    case T_BIG_RATIO:
      return(mpq_cmp_ui(big_ratio(obj), 0, 1) > 0);

    case T_BIG_REAL:
      return(mpfr_cmp_ui(big_real(obj), 0) > 0);
    }

  return(false);
}


static s7_pointer big_is_positive(s7_scheme *sc, s7_pointer args)
{
  #define H_is_positive "(positive? num) returns #t if the real number num is positive (greater than 0)"
  if (!s7_is_real(car(args)))
    return(s7_wrong_type_arg_error(sc, "positive?", 0, car(args), "a real"));
  
  return(make_boolean(sc, s7_is_positive(car(args))));
}


static bool s7_is_negative(s7_pointer obj)
{
  switch (type(obj))
    {
    case T_INTEGER: 
      return(integer(obj) < 0);

    case T_RATIO:   
      return(numerator(obj) < 0);

    case T_REAL:
      return(real(obj) < 0);

    case T_BIG_INTEGER:
      return(mpz_cmp_ui(big_integer(obj), 0) < 0);

    case T_BIG_RATIO:
      return(mpq_cmp_ui(big_ratio(obj), 0, 1) < 0);

    case T_BIG_REAL:
      return(mpfr_cmp_ui(big_real(obj), 0) < 0);
    }

  return(false); /* need s7_scheme pointer to return an error */
}


static s7_pointer big_is_negative(s7_scheme *sc, s7_pointer args)
{
  #define H_is_negative "(negative? num) returns #t if the real number num is negative (less than 0)"
  if (!s7_is_real(car(args)))
    return(s7_wrong_type_arg_error(sc, "negative?", 0, car(args), "a real"));
  
  return(make_boolean(sc, s7_is_negative(car(args))));
}


static s7_pointer big_is_nan(s7_scheme *sc, s7_pointer args) 
{
  #define H_is_nan "(nan? obj) returns #t if obj is a NaN"
  s7_pointer x;

  x = car(args);
  if (s7_is_number(x))
    {
#ifndef _MSC_VER
      return(make_boolean(sc, (is_NaN(s7_real_part(x))) || (is_NaN(s7_imag_part(x)))));
#else
      /* is this actually different from isinf below? */
      if (is_NaN(s7_real_part(x)) || is_NaN(s7_imag_part(x)))
	return(sc->T);
      else return(sc->F);
#endif
    }

  return(sc->F);
}


static s7_pointer big_is_infinite(s7_scheme *sc, s7_pointer args)
{
  #define H_is_infinite "(infinite? obj) returns #t if obj is an infinite real"
  /* (infinite? 1e310) would return #t if left to g_is_infinite
   */
  s7_pointer x;

  x = car(args);
  switch (type(x))
    {
    case T_INTEGER:
    case T_RATIO:
      return(sc->F);

    case T_REAL:
      return(make_boolean(sc, isinf(s7_real_part(x))));

    case T_COMPLEX:
      return(make_boolean(sc, (isinf(s7_real_part(x))) || (isinf(s7_imag_part(x)))));

    case T_BIG_INTEGER:
    case T_BIG_RATIO:
      return(sc->F);

    case T_BIG_REAL:
      return(make_boolean(sc, mpfr_inf_p(big_real(x)) != 0));

    case T_BIG_COMPLEX:
      return(make_boolean(sc, 
			  (mpfr_inf_p(big_real(big_real_part(sc, args))) != 0) ||
			  (mpfr_inf_p(big_real(big_imag_part(sc, args))) != 0)));

    default:
      return(sc->F);
    }
}


static s7_pointer big_lognot(s7_scheme *sc, s7_pointer args)
{
  if (type(car(args)) == T_BIG_INTEGER)
    {
      s7_pointer p;
      mpz_t n;
      mpz_init(n);
      mpz_com(n, big_integer(car(args)));
      p = mpz_to_big_integer(sc, n);
      mpz_clear(n);
      return(p);
    }
  return(g_lognot(sc, args));
}


static s7_pointer big_integer_length(s7_scheme *sc, s7_pointer args)
{
  if (type(car(args)) == T_BIG_INTEGER)
    {
      s7_pointer result;
      mpfr_t n;
      mpfr_init_set_z(n, big_integer(car(args)), GMP_RNDN);
      if (mpfr_cmp_ui(n, 0) < 0)
	mpfr_neg(n, n, GMP_RNDN);
      else mpfr_add_ui(n, n, 1, GMP_RNDN);
      mpfr_log2(n, n, GMP_RNDU);
      result = make_integer(sc, mpfr_get_si(n, GMP_RNDU));
      mpfr_clear(n);
      return(result);
    }
  return(g_integer_length(sc, args));
}


static s7_pointer big_ash(s7_scheme *sc, s7_pointer args)
{
  s7_pointer p0, p1;

  p0 = car(args);
  p1 = cadr(args);
  /* here, as in expt, there are cases like (ash 1 63) which need to be handled as bignums
   *   so there's no way to tell when it's safe to drop into g_ash instead.
   */
  if ((s7_is_integer(p0)) && /* this includes bignum ints... */
      (s7_is_integer(p1)))
    {
      mpz_t n;
      s7_Int shift;
      s7_pointer p;
      bool p0_is_big;
      int p0_compared_to_zero = 0;

      p0_is_big = is_big_number(p0);
      if (p0_is_big)
	p0_compared_to_zero = mpz_cmp_ui(big_integer(p0), 0);
      else 
	{
	  if (s7_integer(p0) > 0)
	    p0_compared_to_zero = 1;
	  else
	    {
	      if (s7_integer(p0) < 0)
		p0_compared_to_zero = -1;
	      else p0_compared_to_zero = 0;
	    }
	}

      if (p0_compared_to_zero == 0)
	return(small_int(0));

      if (is_big_number(p1))
	{
	  if (!mpz_fits_sint_p(big_integer(p1)))
	    {
	      if (mpz_cmp_ui(big_integer(p1), 0) > 0)
		return(s7_out_of_range_error(sc, "ash", 2, p1, "shift is too large"));

	      /* here if p0 is negative, we need to return -1 */
	      if (p0_compared_to_zero == 1)
		return(small_int(0));
	      return(make_integer(sc, -1));
	    }
	  shift = mpz_get_si(big_integer(p1));
	}
      else 
	{
	  shift = s7_integer(p1);
	  if (shift < S7_LONG_MIN)
	    {
	      if (p0_compared_to_zero == 1)
		return(small_int(0));
	      return(make_integer(sc, -1));
	    }
	}

      mpz_init_set(n, big_integer(promote_number(sc, T_BIG_INTEGER, p0)));
      if (shift > 0)     /* left */
	mpz_mul_2exp(n, n, shift);
      else
	{
	  if (shift < 0) /* right */
	    mpz_fdiv_q_2exp(n, n, (unsigned int)(-shift));
	}
      p = mpz_to_big_integer(sc, n);
      mpz_clear(n);
      return(p);
    }
  return(g_ash(sc, args));
}


static s7_pointer big_bits(s7_scheme *sc, s7_pointer args, const char *name, int start, s7_function g_bits, 
			   void (*mpz_bits)(mpz_ptr, mpz_srcptr, mpz_srcptr))
{
  s7_pointer x;
  bool use_bigs = false;
  for (x = args; is_not_null(x); x = cdr(x))
    {
      if (!s7_is_integer(car(x)))
	return(s7_wrong_type_arg_error(sc, name, position_of(x, args), car(x), "an integer"));  
      if (type(car(x)) == T_BIG_INTEGER)
	use_bigs = true;
    }
  if (use_bigs)
    {
      mpz_t n;
      mpz_init_set_si(n, 0);
      if (start == -1)
	mpz_sub_ui(n, n, 1);
      for (x = args; is_not_null(x); x = cdr(x))
	mpz_bits(n, n, big_integer(promote_number(sc, T_BIG_INTEGER, car(x))));
      x = mpz_to_big_integer(sc, n);
      mpz_clear(n);
      return(x);
    }
  return(g_bits(sc, args));
}


static s7_pointer big_logand(s7_scheme *sc, s7_pointer args)
{
  if (is_null(args))
    return(make_integer(sc, -1));
  return(big_bits(sc, args, "logand", -1, g_logand, mpz_and));
}


static s7_pointer big_logior(s7_scheme *sc, s7_pointer args)
{
  if (is_null(args))
    return(small_int(0));
  return(big_bits(sc, args, "logior", 0, g_logior, mpz_ior));
}


static s7_pointer big_logxor(s7_scheme *sc, s7_pointer args)
{
  if (is_null(args))
    return(small_int(0));
  return(big_bits(sc, args, "logxor", 0, g_logxor, mpz_xor));
}


static s7_pointer big_rationalize(s7_scheme *sc, s7_pointer args)
{
  #define H_rationalize "(rationalize x err) returns the ratio with lowest denominator within err of x"
  /* currently (rationalize 1/0 1e18) -> 0
   * remember to pad with many trailing zeros:
   *
   *    : (rationalize 0.1 0)
   *    3602879701896397/36028797018963968
   *    :(rationalize 0.1000000000000000 0)
   *    1/10
   *
   * PERHAPS: gmp number reader used if gmp -- could this be the trailing zeros problem?  (why is the non-gmp case ok?)
   *          also the bignum function is faking it.
   */
  s7_pointer p0, p1 = NULL, p;
  mpfr_t error, ux, x0, x1;
  mpz_t i, i0, i1;
  double xx;

  p0 = car(args);
  if (!s7_is_real(p0))
    return(s7_wrong_type_arg_error(sc, "rationalize", (is_null(cdr(args))) ? 0 : 1, p0, "a real"));

  /* p0 can be exact, but we still have to check it for simplification */
  if (is_not_null(cdr(args)))
    {
      double err_x;
      p1 = cadr(args);
      if (!s7_is_real(p1))              /* (rationalize (expt 2 60) -) */
	return(s7_wrong_type_arg_error(sc, "rationalize error limit,", 2, p1, "a real"));

      if (is_big_number(p1))
	mpfr_init_set(error, big_real(promote_number(sc, T_BIG_REAL, p1)), GMP_RNDN);
      else mpfr_init_set_d(error, s7_number_to_real(p1), GMP_RNDN);

      err_x = mpfr_get_d(error, GMP_RNDN);
      if (is_NaN(err_x))
	{
	  mpfr_clear(error);
	  return(s7_out_of_range_error(sc, "rationalize", 2, cadr(args), "error term is NaN"));
	}
      if (mpfr_inf_p(error) != 0)
	{
	  mpfr_clear(error);
	  return(small_int(0));
	}
      mpfr_abs(error, error, GMP_RNDN);
    }
  else mpfr_init_set_d(error, default_rationalize_error, GMP_RNDN);
  
  if (is_big_number(p0))
    mpfr_init_set(ux, big_real(promote_number(sc, T_BIG_REAL, p0)), GMP_RNDN);
  else mpfr_init_set_d(ux, s7_number_to_real(p0), GMP_RNDN);

  xx = mpfr_get_d(ux, GMP_RNDN);  
  if (is_NaN(xx))
    {
      mpfr_clear(ux);
      mpfr_clear(error);
      return(s7_out_of_range_error(sc, "rationalize", 1, car(args), "argument is NaN"));
    }
  if (mpfr_inf_p(ux) != 0)
    {
      mpfr_clear(ux);
      mpfr_clear(error);
      return(s7_out_of_range_error(sc, "rationalize", 1, car(args), "argument is infinite"));
    }

  mpfr_init_set(x0, ux, GMP_RNDN);        /* x0 = ux - error */
  mpfr_sub(x0, x0, error, GMP_RNDN);
  mpfr_init_set(x1, ux, GMP_RNDN);        /* x1 = ux + error */
  mpfr_add(x1, x1, error, GMP_RNDN);
  mpz_init(i);
  mpfr_get_z(i, x0, GMP_RNDU);            /* i = ceil(x0) */
  
  if (mpfr_cmp_ui(error, 1) >= 0)         /* if (error >= 1.0) */
    {
      mpz_t n;
      
      if (mpfr_cmp_ui(x0, 0) < 0)                /* if (x0 < 0) */
	{
	  if (mpfr_cmp_ui(x1, 0) < 0)            /*   if (x1 < 0) */
	    {
	      mpz_init(n);
	      mpfr_get_z(n, x1, GMP_RNDD);       /*     num = floor(x1) */
	    }
	  else mpz_init_set_ui(n, 0);            /*   else num = 0 */
	}
      else mpz_init_set(n, i);                   /* else num = i */
      
      mpz_clear(i);
      mpfr_clear(ux);
      mpfr_clear(x0);
      mpfr_clear(x1);
      mpfr_clear(error);
      p = mpz_to_big_integer(sc, n);
      mpz_clear(n);
      return(p);
    }
  
  if (mpfr_cmp_z(x1, i) >= 0)                /* if (x1 >= i) */
    {
      mpz_t n;
      
      if (mpz_cmp_ui(i, 0) >= 0)            /* if (i >= 0) */
	mpz_init_set(n, i);                 /*   num = i */
      else
	{
	  mpz_init(n);
	  mpfr_get_z(n, x1, GMP_RNDD);      /* else num = floor(x1) */
	}
      
      mpz_clear(i);
      mpfr_clear(ux);
      mpfr_clear(x0);
      mpfr_clear(x1);
      mpfr_clear(error);
      p = mpz_to_big_integer(sc, n);
      mpz_clear(n);
      return(p);
    }
  
  {
    mpz_t p0, q0, r, r1, p1, q1, old_p1, old_q1;
    mpfr_t val, e0, e1, e0p, e1p, old_e0, old_e1, old_e0p;
    
    mpz_init(i0);
    mpz_init(i1);
    mpfr_get_z(i0, x0, GMP_RNDD);            /* i0 = floor(x0) */
    mpfr_get_z(i1, x1, GMP_RNDU);            /* i1 = ceil(x1) */
    
    mpz_init_set(p0, i0);                    /* p0 = i0 */
    mpz_init_set_ui(q0, 1);                  /* q0 = 1 */
    mpz_init_set(p1, i1);                    /* p1 = i1 */
    mpz_init_set_ui(q1, 1);                  /* q1 = 1 */
    mpfr_init(e0);
    mpfr_init(e1);
    mpfr_init(e0p);
    mpfr_init(e1p);
    mpfr_sub_z(e0, x0, i1, GMP_RNDN);        /* e0 = i1 - x0 */
    mpfr_neg(e0, e0, GMP_RNDN);
    mpfr_sub_z(e1, x0, i0, GMP_RNDN);        /* e1 = x0 - i0 */
    mpfr_sub_z(e0p, x1, i1, GMP_RNDN);       /* e0p = i1 - x1 */
    mpfr_neg(e0p, e0p, GMP_RNDN);
    mpfr_sub_z(e1p, x1, i0, GMP_RNDN);       /* e1p = x1 - i0 */
    
    mpfr_init(val);
    
    mpfr_init(old_e0);
    mpfr_init(old_e1);
    mpfr_init(old_e0p);
    
    mpz_init(r);
    mpz_init(r1);
    mpz_init(old_p1);
    mpz_init(old_q1);
    
    while (true)
      {
	mpfr_set_z(val, p0, GMP_RNDN);
	mpfr_div_z(val, val, q0, GMP_RNDN);  /* val = p0/q0 */
	
	if (((mpfr_cmp(x0, val) <= 0) &&      /* if ((x0 <= val) && (val <= x1)) */
	     (mpfr_cmp(val, x1) <= 0)) ||
	    (mpfr_cmp_ui(e1, 0) == 0) ||
	    (mpfr_cmp_ui(e1p, 0) == 0)) 
	  /* these last 2 are probably not needed -- they protect against running out of bits in the non-gmp case above */
	  {
	    mpq_t q;
	    mpq_init(q);
	    mpq_set_num(q, p0);            /* return(p0/q0) */
	    mpq_set_den(q, q0);
	    
	    mpz_clear(i);
	    mpz_clear(i0);
	    mpz_clear(i1);
	    mpfr_clear(ux);
	    mpfr_clear(x0);
	    mpfr_clear(x1);
	    mpfr_clear(error);
	    
	    mpz_clear(p0);
	    mpz_clear(q0);
	    mpz_clear(r);
	    mpz_clear(r1);
	    mpz_clear(p1);
	    mpz_clear(q1);
	    mpz_clear(old_p1);
	    mpz_clear(old_q1);
	    
	    mpfr_clear(val);
	    mpfr_clear(e0);
	    mpfr_clear(e1);
	    mpfr_clear(e0p);
	    mpfr_clear(e1p);
	    mpfr_clear(old_e0);
	    mpfr_clear(old_e1);
	    mpfr_clear(old_e0p);
	    
	    p = mpq_to_big_ratio(sc, q);
	    mpq_clear(q);
	    return(p);
	  }
	
	mpfr_div(val, e0, e1, GMP_RNDN);
	mpfr_get_z(r, val, GMP_RNDD);           /* r = floor(e0/e1) */
	mpfr_div(val, e0p, e1p, GMP_RNDN);
	mpfr_get_z(r1, val, GMP_RNDU);          /* r1 = ceil(e0p/e1p) */
	if (mpz_cmp(r1, r) < 0)                 /* if (r1 < r) */
	  mpz_set(r, r1);                       /*   r = r1 */
	
	mpz_set(old_p1, p1);                    /* old_p1 = p1 */
	mpz_set(p1, p0);                        /* p1 = p0 */
	mpz_set(old_q1, q1);                    /* old_q1 = q1 */
	mpz_set(q1, q0);                        /* q1 = q0 */
	
	mpfr_set(old_e0, e0, GMP_RNDN);         /* old_e0 = e0 */
	mpfr_set(e0, e1p, GMP_RNDN);            /* e0 = e1p */
	mpfr_set(old_e0p, e0p, GMP_RNDN);       /* old_e0p = e0p */
	mpfr_set(e0p, e1, GMP_RNDN);            /* e0p = e1 */
	mpfr_set(old_e1, e1, GMP_RNDN);         /* old_e1 = e1 */
	
	mpz_mul(p0, p0, r);                     /* p0 = old_p1 + r * p0 */
	mpz_add(p0, p0, old_p1); 
	
	mpz_mul(q0, q0, r);                     /* q0 = old_q1 + r * q0 */
	mpz_add(q0, q0, old_q1); 
	
	mpfr_mul_z(e1, e1p, r, GMP_RNDN);       /* e1 = old_e0p - r * e1p */
	mpfr_sub(e1, old_e0p, e1, GMP_RNDN); 
	
	mpfr_mul_z(e1p, old_e1, r, GMP_RNDN);   /* e1p = old_e0 - r * old_e1 */
	mpfr_sub(e1p, old_e0, e1p, GMP_RNDN); 
      }
  }
}


static s7_pointer big_exact_to_inexact(s7_scheme *sc, s7_pointer args)
{
  #define H_exact_to_inexact "(exact->inexact num) converts num to an inexact number; (exact->inexact 3/2) = 1.5"
  s7_pointer p;

  p = car(args);
  if (!s7_is_number(p))   /* apparently (exact->inexact 1+i) is not an error */
    return(s7_wrong_type_arg_error(sc, "exact->inexact", 0, p, "a number"));

  if (!s7_is_rational(p))
    return(p);

  return(promote_number(sc, T_BIG_REAL, to_big(sc, p)));
}


static s7_pointer big_inexact_to_exact(s7_scheme *sc, s7_pointer args)
{
  #define H_inexact_to_exact "(inexact->exact num) converts num to an exact number; (inexact->exact 1.5) = 3/2"

  s7_pointer p;
  p = car(args);

  if (s7_is_rational(p))
    return(p);

  if (!s7_is_real(p))
    return(s7_wrong_type_arg_error(sc, "inexact->exact", 0, p, "a real number"));

  return(big_rationalize(sc, args));
}


static s7_pointer big_convert_to_int(s7_scheme *sc, s7_pointer args,
				     const char *caller,
				     void (*div_func)(mpz_ptr, mpz_srcptr, mpz_srcptr),
				     mp_rnd_t mode)
{
  /* we can't go to the normal (non-gmp) functions here */
  s7_pointer p;
  mpz_t n;

  p = car(args);
  if (!s7_is_real(p))
    return(s7_wrong_type_arg_error(sc, caller, 0, p, "a real"));

  if (s7_is_integer(p))
    return(p);

  p = to_big(sc, p);
  if (type(p) == T_BIG_RATIO)
    {
      /* apparently we have to do the divide by hand */
      mpz_t d;
      mpz_init_set(n, mpq_numref(big_ratio(p)));
      mpz_init_set(d, mpq_denref(big_ratio(p)));
      div_func(n, n, d);
      mpz_clear(d);
    }
  else
    {
      if ((big_is_nan(sc, args) == sc->T) ||
	  (big_is_infinite(sc, args)) == sc->T)
	return(s7_out_of_range_error(sc, caller, 0, p, "argument is NaN or infinite"));

      mpz_init(n);
      mpfr_get_z(n, big_real(p), mode);
    }
  p = mpz_to_big_integer(sc, n);
  mpz_clear(n);
  return(p);
}


static s7_pointer big_floor(s7_scheme *sc, s7_pointer args)
{
  #define H_floor "(floor x) returns the integer closest to x toward -inf"
  return(big_convert_to_int(sc, args, "floor", mpz_fdiv_q, GMP_RNDD));
}


static s7_pointer big_ceiling(s7_scheme *sc, s7_pointer args)
{
  #define H_ceiling "(ceiling x) returns the integer closest to x toward inf"
  return(big_convert_to_int(sc, args, "ceiling", mpz_cdiv_q, GMP_RNDU));
}


static s7_pointer big_truncate(s7_scheme *sc, s7_pointer args)
{
  #define H_truncate "(truncate x) returns the integer closest to x toward 0"
  return(big_convert_to_int(sc, args, "truncate", mpz_tdiv_q, GMP_RNDZ));
}


static s7_pointer big_round(s7_scheme *sc, s7_pointer args)
{
  #define H_round "(round x) returns the integer closest to x"
  s7_pointer p;
  mpz_t n;

  p = car(args);
  if (!s7_is_real(p))
    return(s7_wrong_type_arg_error(sc, "round", 0, p, "a real"));

  if (s7_is_integer(p))
    return(p);

  p = to_big(sc, p);
  if (type(p) == T_BIG_INTEGER)
    return(p);

  if (type(p) == T_BIG_RATIO)
    {
      int rnd;
      mpz_t rm;
      mpz_init_set(n, mpq_numref(big_ratio(p)));
      mpz_init(rm);
      mpz_fdiv_qr(n, rm, n, mpq_denref(big_ratio(p)));
      mpz_mul_ui(rm, rm, 2);
      rnd = mpz_cmpabs(rm, mpq_denref(big_ratio(p)));
      mpz_fdiv_q(rm, rm, mpq_denref(big_ratio(p)));
      if (rnd > 0)
	mpz_add(n, n, rm);
      else
	{
	  if (rnd == 0)
	    {
	      if (mpz_odd_p(n))
		mpz_add_ui(n, n, 1);
	    }
	}
      mpz_clear(rm);
      p = mpz_to_big_integer(sc, n);
      mpz_clear(n);
      return(p);
    }

  if ((big_is_nan(sc, args) == sc->T) ||
      (big_is_infinite(sc, args)) == sc->T)
    return(s7_out_of_range_error(sc, "round", 0, p, "argument is NaN or infinite"));

  {
    int cmp_res;
    mpz_t fl, ce;
    mpfr_t x, dfl, dce;
    mpfr_init_set(x, big_real(p), GMP_RNDN);
    mpz_init(fl);
    mpfr_get_z(fl, x, GMP_RNDD);           /* fl = floor(x) */
    mpz_init(ce);
    mpfr_get_z(ce, x, GMP_RNDU);           /* ce = ceil(x) */
    mpfr_init(dfl);
    mpfr_sub_z(dfl, x, fl, GMP_RNDN);      /* dfl = x - fl */
    mpfr_init(dce);
    mpfr_sub_z(dce, x, ce, GMP_RNDN);      /* dce = -(ce - x) */
    mpfr_neg(dce, dce, GMP_RNDN);          /*    and reversed */
    cmp_res = mpfr_cmp(dfl, dce);
    if (cmp_res > 0)                       /* if (dfl > dce) return(ce) */
      mpz_init_set(n, ce);
    else
      {
	if (cmp_res < 0)                   /* if (dfl < dce) return(fl) */
	  mpz_init_set(n, fl);
	else
	  {
	    if (mpz_even_p(fl))
	      mpz_init_set(n, fl);         /* if (mod(fl, 2) == 0) return(fl) */
	    else mpz_init_set(n, ce);      /* else return(ce) */
	  }
      }
    mpz_clear(fl);
    mpz_clear(ce);
    mpfr_clear(dfl);
    mpfr_clear(dce);
    mpfr_clear(x);
    p = mpz_to_big_integer(sc, n);
    mpz_clear(n);
    return(p);
  }
}


static s7_pointer big_quotient(s7_scheme *sc, s7_pointer args)
{
  #define H_quotient "(quotient x1 x2) returns the integer quotient of x1 and x2; (quotient 4 3) = 1"

  s7_pointer x, y, p;
  x = car(args);
  y = cadr(args);

  if (!s7_is_real(x))
    return(s7_wrong_type_arg_error(sc, "quotient", 1, x, "a real"));
  if (!s7_is_real(y))
    return(s7_wrong_type_arg_error(sc, "quotient", 2, y, "a real"));

  if ((s7_is_integer(x)) &&
      (s7_is_integer(y)))
    {
      mpz_t n;
      x = to_big(sc, x);
      y = to_big(sc, y);

      if (big_is_zero_1(sc, y) == sc->T)
	return(division_by_zero_error(sc, "quotient", args));
	  
      mpz_init_set(n, big_integer(x));
      mpz_tdiv_q(n, n, big_integer(y));

      p = mpz_to_big_integer(sc, n);
      mpz_clear(n);
      return(p);
    }
  return(big_truncate(sc, list_1(sc, big_divide(sc, args))));
}


static s7_pointer big_remainder(s7_scheme *sc, s7_pointer args)
{
  #define H_remainder "(remainder x1 x2) returns the integer remainder of x1 and x2; (remainder 10 3) = 1"
  s7_pointer x, y, p;
  x = car(args);
  y = cadr(args);

  if (!s7_is_real(x))
    return(s7_wrong_type_arg_error(sc, "remainder", 1, x, "a real"));
  if (!s7_is_real(y))
    return(s7_wrong_type_arg_error(sc, "remainder", 2, y, "a real"));

  if ((s7_is_integer(x)) &&
      (s7_is_integer(y)))
    {
      mpz_t n;
      x = to_big(sc, x);
      y = to_big(sc, y);

      if (big_is_zero_1(sc, y) == sc->T)
	return(division_by_zero_error(sc, "remainder", args));
	  
      mpz_init_set(n, big_integer(x));
      mpz_tdiv_r(n, n, big_integer(y));

      p = mpz_to_big_integer(sc, n);
      mpz_clear(n);
      return(p);
    }
  return(big_subtract(sc, 
          list_2(sc, x, 
           big_multiply(sc, 
            list_2(sc, y, 
             big_quotient(sc, args))))));
}


static s7_pointer big_modulo(s7_scheme *sc, s7_pointer args)
{
  #define H_modulo "(modulo x1 x2) returns x1 mod x2; (modulo 4 3) = 1.  The arguments can be real numbers."
  s7_pointer a, b, p;

  a = car(args);
  if (!s7_is_real(a))
    return(s7_wrong_type_arg_error(sc, "modulo", 1, a, "a real"));

  b = cadr(args);
  if (!s7_is_real(b))
    return(s7_wrong_type_arg_error(sc, "modulo", 2, b, "a real"));

  a = to_big(sc, a);
  b = to_big(sc, b);

  if ((s7_is_integer(a)) &&
      (s7_is_integer(b)))
    {
      s7_pointer x, y;
      int cy, cz;
      mpz_t n;
      
      y = promote_number(sc, T_BIG_INTEGER, b);
      if (mpz_cmp_ui(big_integer(y), 0) == 0)
	return(a);
      
      x = promote_number(sc, T_BIG_INTEGER, a);
      /* mpz_mod is too tricky here */

      mpz_init_set(n, big_integer(x));
      mpz_fdiv_r(n, n, big_integer(y));
      cy = mpz_cmp_ui(big_integer(y), 0);
      cz = mpz_cmp_ui(n, 0);
      if (((cy < 0) && (cz > 0)) ||
	  ((cy > 0) && (cz < 0)))
	mpz_add(n, n, big_integer(y));

      p = mpz_to_big_integer(sc, n);
      mpz_clear(n);
      return(p);
    }
  return(big_subtract(sc, 
          list_2(sc, a, 
           big_multiply(sc, 
            list_2(sc, b, 
    	     big_floor(sc, 
              list_1(sc, 
               big_divide(sc, 
	        list_2(sc, a, b)))))))));
}


static int big_real_scan_args(s7_scheme *sc, s7_pointer args)
{
  int i, result_type = T_INTEGER;
  s7_pointer arg;

  for (i = 1, arg = args; is_not_null(arg); i++, arg = cdr(arg)) 
    {
      s7_pointer p;
      p = car(arg);
      if (!s7_is_real(p))
	return(-i);
      if (is_big_number(p))
	result_type = big_type_to_result_type(result_type, type(p));
      else result_type = normal_type_to_result_type(result_type, type(p));
    }
  return(result_type);
}


static s7_pointer big_max(s7_scheme *sc, s7_pointer args)
{
  int result_type;
  s7_pointer x, result, arg;
  
  result_type = big_real_scan_args(sc, args);
  if (result_type < 0)
    return(s7_wrong_type_arg_error(sc, "max", -result_type, s7_list_ref(sc, args, -1 - result_type), "a real number"));

  if (result_type < T_BIG_INTEGER)
    return(g_max(sc, args));

  result = promote_number(sc, result_type, car(args));

  for (x = cdr(args); is_not_null(x); x = cdr(x)) 
    {
      arg = promote_number(sc, result_type, car(x));
      switch (result_type)
	{
	case T_BIG_INTEGER:
	  if (mpz_cmp(big_integer(result), big_integer(arg)) < 0)
	    result = arg;
	  break;
	  
	case T_BIG_RATIO:
	  if (mpq_cmp(big_ratio(result), big_ratio(arg)) < 0)
	    result = arg;
	  break;

	case T_BIG_REAL:
	  if (mpfr_cmp(big_real(result), big_real(arg)) < 0)
	    result = arg;
	  break;
	}
    }
  if (result_type == T_BIG_RATIO) /* maybe actual result was an int */
    {
      if (mpz_cmp_ui(mpq_denref(big_ratio(result)), 1) == 0)
	{
	  mpz_t n;
	  s7_pointer p;
	  mpz_init_set(n, mpq_numref(big_ratio(result)));
	  p = mpz_to_big_integer(sc, n);
	  mpz_clear(n);
	  return(p);
	}
    }
  return(result);
}


static s7_pointer big_min(s7_scheme *sc, s7_pointer args)
{
  int result_type;
  s7_pointer x, result, arg;
  
  result_type = big_real_scan_args(sc, args);
  if (result_type < 0)
    return(s7_wrong_type_arg_error(sc, "min", -result_type, s7_list_ref(sc, args, -1 - result_type), "a real number"));

  if (result_type < T_BIG_INTEGER)
    return(g_min(sc, args));

  result = promote_number(sc, result_type, car(args));

  for (x = cdr(args); is_not_null(x); x = cdr(x)) 
    {
      arg = promote_number(sc, result_type, car(x));
      switch (result_type)
	{
	case T_BIG_INTEGER:
	  if (mpz_cmp(big_integer(result), big_integer(arg)) > 0)
	    result = arg;
	  break;
	  
	case T_BIG_RATIO:
	  if (mpq_cmp(big_ratio(result), big_ratio(arg)) > 0)
	    result = arg;
	  break;

	case T_BIG_REAL:
	  if (mpfr_cmp(big_real(result), big_real(arg)) > 0)
	    result = arg;
	  break;
	}
    }
  if (result_type == T_BIG_RATIO) /* maybe actual result was an int */
    {
      if (mpz_cmp_ui(mpq_denref(big_ratio(result)), 1) == 0)
	{
	  mpz_t n;
	  s7_pointer p;
	  mpz_init_set(n, mpq_numref(big_ratio(result)));
	  p = mpz_to_big_integer(sc, n);
	  mpz_clear(n);
	  return(p);
	}
    }
  return(result);
}


static s7_pointer big_less(s7_scheme *sc, s7_pointer args)
{
  #define H_less "(< x1 ...) returns #t if its arguments are in increasing order"
  int result_type;
  s7_pointer x, previous, current;
  
  result_type = big_real_scan_args(sc, args);
  if (result_type < 0)
    return(s7_wrong_type_arg_error(sc, "<", -result_type, s7_list_ref(sc, args, -1 - result_type), "a real number"));

  /* don't try to use g_less here */
  if (result_type < T_BIG_INTEGER)
    result_type += 4;
  previous = promote_number(sc, result_type, car(args));

  for (x = cdr(args); is_not_null(x); x = cdr(x)) 
    {
      current = promote_number(sc, result_type, car(x));
      switch (result_type)
	{
	case T_BIG_INTEGER:
	  if (mpz_cmp(big_integer(previous), big_integer(current)) >= 0) return(sc->F);
	  break;
	  
	case T_BIG_RATIO:
	  if (mpq_cmp(big_ratio(previous), big_ratio(current)) >= 0) return(sc->F);
	  break;

	case T_BIG_REAL:
	  if (mpfr_cmp(big_real(previous), big_real(current)) >= 0) return(sc->F);
	  break;
	}
      previous = current;
    }
  return(sc->T);
}


static s7_pointer big_less_or_equal(s7_scheme *sc, s7_pointer args)
{
  #define H_less_or_equal "(<= x1 ...) returns #t if its arguments are in increasing order"
  int result_type;
  s7_pointer x, previous, current;
  
  result_type = big_real_scan_args(sc, args);
  if (result_type < 0)
    return(s7_wrong_type_arg_error(sc, "<=", -result_type, s7_list_ref(sc, args, -1 - result_type), "a real number"));

  if (result_type < T_BIG_INTEGER)
    result_type += 4;
  previous = promote_number(sc, result_type, car(args));

  for (x = cdr(args); is_not_null(x); x = cdr(x)) 
    {
      current = promote_number(sc, result_type, car(x));
      switch (result_type)
	{
	case T_BIG_INTEGER:
	  if (mpz_cmp(big_integer(previous), big_integer(current)) > 0) return(sc->F);
	  break;
	  
	case T_BIG_RATIO:
	  if (mpq_cmp(big_ratio(previous), big_ratio(current)) > 0) return(sc->F);
	  break;

	case T_BIG_REAL:
	  if (mpfr_cmp(big_real(previous), big_real(current)) > 0) return(sc->F);
	  break;
	}
      previous = current;
    }
  return(sc->T);
}


static s7_pointer big_greater(s7_scheme *sc, s7_pointer args)
{
  #define H_greater "(> x1 ...) returns #t if its arguments are in decreasing order"
  int result_type;
  s7_pointer x, previous, current;

  result_type = big_real_scan_args(sc, args);
  if (result_type < 0)
    return(s7_wrong_type_arg_error(sc, ">", -result_type, s7_list_ref(sc, args, -1 - result_type), "a real number"));

  if (result_type < T_BIG_INTEGER)
    result_type += 4;
  previous = promote_number(sc, result_type, car(args));

  for (x = cdr(args); is_not_null(x); x = cdr(x)) 
    {
      current = promote_number(sc, result_type, car(x));
      switch (result_type)
	{
	case T_BIG_INTEGER:
	  if (mpz_cmp(big_integer(previous), big_integer(current)) <= 0) return(sc->F);
	  break;
	  
	case T_BIG_RATIO:
	  if (mpq_cmp(big_ratio(previous), big_ratio(current)) <= 0) return(sc->F);
	  break;

	case T_BIG_REAL:
	  if (mpfr_cmp(big_real(previous), big_real(current)) <= 0) return(sc->F);
	  break;
	}
      previous = current;
    }
  return(sc->T);
}


static s7_pointer big_greater_or_equal(s7_scheme *sc, s7_pointer args)
{
  #define H_greater_or_equal "(>= x1 ...) returns #t if its arguments are in decreasing order"
  int result_type;
  s7_pointer x, previous, current;
  
  result_type = big_real_scan_args(sc, args);
  if (result_type < 0)
    return(s7_wrong_type_arg_error(sc, ">=", -result_type, s7_list_ref(sc, args, -1 - result_type), "a real number"));

  if (result_type < T_BIG_INTEGER)
    result_type += 4;
  previous = promote_number(sc, result_type, car(args));

  for (x = cdr(args); is_not_null(x); x = cdr(x)) 
    {
      current = promote_number(sc, result_type, car(x));
      switch (result_type)
	{
	case T_BIG_INTEGER:
	  if (mpz_cmp(big_integer(previous), big_integer(current)) < 0) return(sc->F);
	  break;
	  
	case T_BIG_RATIO:
	  if (mpq_cmp(big_ratio(previous), big_ratio(current)) < 0) return(sc->F);
	  break;

	case T_BIG_REAL:
	  if (mpfr_cmp(big_real(previous), big_real(current)) < 0) return(sc->F);
	  break;
	}
      previous = current;
    }
  return(sc->T);
}


static s7_pointer big_equal(s7_scheme *sc, s7_pointer args)
{
  int result_type = T_INTEGER;
  s7_pointer x, y, result;

  for (x = args; is_not_null(x); x = cdr(x)) 
    {
      s7_pointer p;
      p = car(x);
      if (!is_number(p))
	{
	  if (!is_big_number(p))
	    return(s7_wrong_type_arg_error(sc, "=", position_of(x, args), p, "a number"));
	  else result_type = big_type_to_result_type(result_type, type(p));
	}
      else 
	{
	  switch (type(p))
	    {
	    case T_INTEGER:
	    case T_RATIO:
	      break;

	    case T_REAL:
	      if (is_NaN(real(p)))  /* (= (bignum "3") 1/0) */
		return(sc->F);

	    default:
	      if ((is_NaN(real_part(p))) ||
		  (is_NaN(imag_part(p))))
		return(sc->F);
	    }
	  result_type = normal_type_to_result_type(result_type, type(p));
	}
    }

  if (result_type < T_BIG_INTEGER)
    return(g_equal(sc, args));

  result = promote_number(sc, result_type, car(args));

  for (y = cdr(args); is_not_null(y); y = cdr(y)) 
    {
      s7_pointer arg;
      arg = promote_number(sc, result_type, car(y));
      switch (result_type)
	{
	case T_BIG_INTEGER:
	  if (mpz_cmp(big_integer(result), big_integer(arg)) != 0) return(sc->F);
	  break;
	  
	case T_BIG_RATIO:
	  if (mpq_cmp(big_ratio(result), big_ratio(arg)) != 0) return(sc->F);
	  break;
	  
	case T_BIG_REAL:
	  if (mpfr_cmp(big_real(result), big_real(arg)) != 0) return(sc->F);
	  break;
	  
	case T_BIG_COMPLEX:
	  if (mpc_cmp(big_complex(result), big_complex(arg)) != 0) return(sc->F);
	  break;
	}
    }
  return(sc->T);
}


static s7_pointer big_gcd(s7_scheme *sc, s7_pointer args)
{
  #define H_gcd "(gcd ...) returns the greatest common divisor of its rational arguments"
  bool rats = false;
  s7_pointer x;

  for (x = args; is_not_null(x); x = cdr(x)) 
    if (!s7_is_rational(car(x)))
      return(s7_wrong_type_arg_error(sc, "gcd", position_of(x, args), car(x), "an integer or ratio"));
    else rats = ((rats) || (!s7_is_integer(car(x))));

  if (is_null(cdr(args)))    /* (gcd -2305843009213693951/4611686018427387903) */
    return(big_abs(sc, args)); 

  if (!rats)
    {
      mpz_t n;
      mpz_init(n);
      for (x = args; is_not_null(x); x = cdr(x)) 
	{
	  mpz_gcd(n, n, big_integer(promote_number(sc, T_BIG_INTEGER, car(x))));
	  if (mpz_cmp_ui(n, 1) == 0)
	    {
	      mpz_clear(n);
	      return(small_int(1));
	    }
	}
      x = mpz_to_big_integer(sc, n);
      mpz_clear(n);
      return(x);
    }

  {
    s7_pointer rat;
    mpq_t q;
    mpz_t n, d;
    
    rat = promote_number(sc, T_BIG_RATIO, car(args));
    mpz_init_set(n, mpq_numref(big_ratio(rat)));
    mpz_init_set(d, mpq_denref(big_ratio(rat)));
    for (x = cdr(args); is_not_null(x); x = cdr(x))
      {
	rat = promote_number(sc, T_BIG_RATIO, car(x));
	mpz_gcd(n, n, mpq_numref(big_ratio(rat)));
	mpz_lcm(d, d, mpq_denref(big_ratio(rat)));
      }
    if (mpz_cmp_ui(d, 1) == 0)
      {
	rat = mpz_to_big_integer(sc, n);
	mpz_clear(n);
	mpz_clear(d);
	return(rat);
      }

    mpq_init(q);
    mpq_set_num(q, n);
    mpq_set_den(q, d);
    mpz_clear(n);
    mpz_clear(d);

    x = mpq_to_big_ratio(sc, q);
    mpq_clear(q);
    return(x);
  }
}


static s7_pointer big_lcm(s7_scheme *sc, s7_pointer args)
{
  #define H_lcm "(lcm ...) returns the least common multiple of its rational arguments"
  s7_pointer x;
  bool rats = false;

  for (x = args; is_not_null(x); x = cdr(x)) 
    if (!s7_is_rational(car(x)))
      return(s7_wrong_type_arg_error(sc, "lcm", position_of(x, args), car(x), "an integer or ratio"));
    else rats = ((rats) || (!s7_is_integer(car(x))));
  
   if (is_null(cdr(args)))    /* (lcm -2305843009213693951/4611686018427387903) */
      return(big_abs(sc, args)); 
    
   if (!rats)
    {
      mpz_t n;
      mpz_init(n);
      mpz_set_ui(n, 1);
      for (x = args; is_not_null(x); x = cdr(x)) 
	{
	  mpz_lcm(n, n, big_integer(promote_number(sc, T_BIG_INTEGER, car(x))));
	  if (mpz_cmp_ui(n, 0) == 0)
	    {
	      mpz_clear(n);
	      return(small_int(0));
	    }
	}
      x = mpz_to_big_integer(sc, n);
      mpz_clear(n);
      return(x);
    }

  {
    s7_pointer rat;
    mpq_t q;
    mpz_t n, d;
    
    rat = promote_number(sc, T_BIG_RATIO, car(args));
    mpz_init_set(n, mpq_numref(big_ratio(rat)));
    if (mpz_cmp_ui(n, 0) == 0)
      {
	mpz_clear(n);
	return(small_int(0));
      }

    mpz_init_set(d, mpq_denref(big_ratio(rat)));
    for (x = cdr(args); is_not_null(x); x = cdr(x))
      {
	rat = promote_number(sc, T_BIG_RATIO, car(x));
	mpz_lcm(n, n, mpq_numref(big_ratio(rat)));
	if (mpz_cmp_ui(n, 0) == 0)
	  {
	    mpz_clear(n);
	    mpz_clear(d);
	    return(small_int(0));
	  }
	mpz_gcd(d, d, mpq_denref(big_ratio(rat)));
      }

    if (mpz_cmp_ui(d, 1) == 0)
      {
	rat = mpz_to_big_integer(sc, n);
	mpz_clear(n);
	mpz_clear(d);
	return(rat);
      }

    mpq_init(q);
    mpq_set_num(q, n);
    mpq_set_den(q, d);
    mpz_clear(n);
    mpz_clear(d);
    x = mpq_to_big_ratio(sc, q);
    mpq_clear(q);
    return(x);
  }
}


static s7_pointer g_bignum_precision(s7_scheme *sc, s7_pointer args)
{
  #define H_bignum_precision "(bignum-precision) returns the number of bits used for floats and complex numbers. It can be set."
  return(make_integer(sc, (int)mpfr_get_default_prec()));
}


static s7_pointer g_set_bignum_precision(s7_scheme *sc, s7_pointer args)
{
  mp_prec_t bits;
  s7_Int precision;

  if (!s7_is_integer(car(args)))
    return(s7_wrong_type_arg_error(sc, "set! bignum-precision", 0, car(args), "an integer"));

  precision = s7_integer(car(args));
  if (precision <= 1)                   /* (set! (bignum-precision) 1) causes mpfr to segfault! (also 0 and -1) */
    return(s7_out_of_range_error(sc, "set! bignum-precision", 0, car(args), "has to be greater than 1"));    

  bits = (mp_prec_t)precision;
  mpfr_set_default_prec(bits);
  mpc_set_default_precision(bits);
  s7_symbol_set_value(sc, make_symbol(sc, "pi"), big_pi(sc));
  return(car(args));
}


typedef struct {
  gmp_randstate_t state;
} s7_big_rng_t;


static char *print_big_rng(s7_scheme *sc, void *val)
{
  char *buf;
  s7_big_rng_t *r = (s7_big_rng_t *)val;
  buf = (char *)malloc(64 * sizeof(char));
  snprintf(buf, 64, "#<big-rng %p>", r);
  return(buf);
}


static void free_big_rng(void *val)
{
  s7_big_rng_t *r = (s7_big_rng_t *)val;
  gmp_randclear(r->state);
  free(r);
}


static bool equal_big_rng(void *val1, void *val2)
{
  return(val1 == val2);
}

  
static s7_pointer make_big_random_state(s7_scheme *sc, s7_pointer args)
{
  #define H_make_random_state "(make-random-state seed) returns a new random number state initialized with 'seed'. \
Pass this as the second argument to 'random' to get a repeatable random number sequence:\n\
    (let ((seed (make-random-state 1234))) (random 1.0 seed))"

  s7_big_rng_t *r;
  s7_pointer seed;

  seed = car(args);
  if (!s7_is_integer(seed))
    return(s7_wrong_type_arg_error(sc, "make-random-state,", 0, seed, "an integer"));

  if (is_big_number(seed))
    {
      r = (s7_big_rng_t *)calloc(1, sizeof(s7_big_rng_t));
      gmp_randinit_default(r->state);
      gmp_randseed(r->state, big_integer(seed));
      return(s7_make_object(sc, big_rng_tag, (void *)r));
    }

  return(s7_make_random_state(sc, args));
}


static s7_pointer big_random(s7_scheme *sc, s7_pointer args)
{
  s7_pointer num, state, x;
  state = sc->NIL;

  num = car(args); 
  if (!s7_is_number(num))
    return(s7_wrong_type_arg_error(sc, "random", (is_null(cdr(args))) ? 0 : 1, num, "a number"));

  if (is_not_null(cdr(args)))
    {
      state = cadr(args);
      if ((!is_c_object(state)) ||
	  ((object_type(state) != big_rng_tag) &&
	   (object_type(state) != rng_tag)))
	return(s7_wrong_type_arg_error(sc, "random state,", 2, state, "a random-state object"));
    }

  if (big_is_zero(sc, args) == sc->T)
    return(num);

  if ((is_big_number(num)) ||
      ((is_c_object(state)) &&
       (object_type(state) == big_rng_tag)))
    {
      /* bignum case -- provide a state if none was passed,
       *   promote num if bignum state was passed but num is not a bignum
       *   if num==0, just return 0 (above) since gmp otherwise throws an arithmetic exception
       */
      s7_big_rng_t *r = NULL;

      if (is_null(state))
	{
	  /* no state passed, so make one */

	  if (!sc->default_big_rng)
	    {
	      mpz_t seed;
	      r = (s7_big_rng_t *)calloc(1, sizeof(s7_big_rng_t));
	      mpz_init_set_ui(seed, (unsigned int)time(NULL));
	      gmp_randinit_default(r->state);
	      gmp_randseed(r->state, seed);
	      mpz_clear(seed);
	      sc->default_big_rng = (void *)r;
	    }
	  else r = (s7_big_rng_t *)(sc->default_big_rng);
	}
      else
	{
	  /* state was passed, check its type */

	  if (object_type(state) == rng_tag)
	    {
	      /* here "num" is a bignum, the state was passed, but it is intended for non-bignums */
	      if (type(num) == T_BIG_REAL)
		num = s7_make_real(sc, (s7_Double)mpfr_get_d(big_real(num), GMP_RNDN));
	      else
		{
		  if (type(num) == T_BIG_INTEGER)
		    num = make_integer(sc, big_integer_to_s7_Int(big_integer(num)));
		  else
		    {
		      if (type(num) == T_BIG_RATIO)
			num = s7_make_ratio(sc, 
					    big_integer_to_s7_Int(mpq_numref(big_ratio(num))), 
					    big_integer_to_s7_Int(mpq_denref(big_ratio(num))));
		    }
		}
	      return(g_random(sc, list_2(sc, num, state)));
	    }
	  r = (s7_big_rng_t *)s7_object_value(state);
	}

      if (!is_big_number(num))
	{
	  switch (type(num))
	    {
	    case T_INTEGER:
	      num = promote_number(sc, T_BIG_INTEGER, num);
	      break;

	    case T_RATIO:
	      num = promote_number(sc, T_BIG_RATIO, num);
	      break;

	    case T_REAL:
	      num = promote_number(sc, T_BIG_REAL, num);
	      break;

	    default:
	      num = promote_number(sc, T_BIG_COMPLEX, num);
	      break;
	    }
	}

      /* finally both the state and the number are big */
	    
      switch (type(num))
	{
	case T_BIG_INTEGER:
	  {
	    mpz_t n;
	    mpz_init(n);
	    mpz_urandomm(n, r->state, big_integer(num));
	    
	    /* this does not work if num is a negative number -- you get positive results.
	     *   so check num for sign, and negate result if necessary.
	     */
	    
	    if (mpz_cmp_ui(big_integer(num), 0) < 0)
	      mpz_neg(n, n);
	    
	    x = mpz_to_big_integer(sc, n);
	    mpz_clear(n);
	    return(x);
	  }
      
	case T_BIG_RATIO:
	  {
	    mpfr_t n, e;
	    mpfr_t rat;
	    
	    mpfr_init_set_ui(n, 1, GMP_RNDN);
	    mpfr_urandomb(n, r->state);
	    mpfr_init_set_q(rat, big_ratio(num), GMP_RNDN);
	    mpfr_mul(n, n, rat, GMP_RNDN);
	    
	    mpfr_init_set_str(e, "0.0000001", 10, GMP_RNDN);
	    mpfr_mul(e, e, rat, GMP_RNDN);
	    mpfr_clear(rat);
	    /* as in g_random, small ratios are a problem because the error term (default_rationalize_error = 1e-12 here)
	     *   clobbers everything to 0.
	     */
	    x = big_rationalize(sc, list_2(sc, mpfr_to_big_real(sc, n), mpfr_to_big_real(sc, e)));
	    mpfr_clear(n);
	    mpfr_clear(e);
	    return(x);
	  }
      
	case T_BIG_REAL:
	  {
	    mpfr_t n;
	    mpfr_init_set_ui(n, 1, GMP_RNDN);
	    mpfr_urandomb(n, r->state);
	    mpfr_mul(n, n, big_real(num), GMP_RNDN);
	    x = mpfr_to_big_real(sc, n);
	    mpfr_clear(n);
	    return(x);
	  }
	  
	case T_BIG_COMPLEX:
	  {
	    mpc_t n;
	    mpc_init(n);
	    mpc_urandom(n, r->state);
	    mpfr_mul(mpc_realref(n), mpc_realref(n), mpc_realref(big_complex(num)), GMP_RNDN);
	    mpfr_mul(mpc_imagref(n), mpc_imagref(n), mpc_imagref(big_complex(num)), GMP_RNDN);
	    x = mpc_to_big_complex(sc, n);
	    mpc_clear(n);
	    return(x);
	  }
	}
    }

  return(g_random(sc, args));
}


static void s7_gmp_init(s7_scheme *sc)
{
  s7_define_function(sc, "+",                   big_add,              0, 0, true,  H_add);
  s7_define_function(sc, "-",                   big_subtract,         1, 0, true,  H_subtract);
  s7_define_function(sc, "*",                   big_multiply,         0, 0, true,  H_multiply);
  s7_define_function(sc, "/",                   big_divide,           1, 0, true,  H_divide);

  s7_define_function(sc, "max",                 big_max,              1, 0, true,  H_max);
  s7_define_function(sc, "min",                 big_min,              1, 0, true,  H_min);
  s7_define_function(sc, "<",                   big_less,             2, 0, true,  H_less);
  s7_define_function(sc, "<=",                  big_less_or_equal,    2, 0, true,  H_less_or_equal);
  s7_define_function(sc, ">",                   big_greater,          2, 0, true,  H_greater);
  s7_define_function(sc, ">=",                  big_greater_or_equal, 2, 0, true,  H_greater_or_equal);
  s7_define_function(sc, "=",                   big_equal,            2, 0, true,  H_equal);

  s7_define_function(sc, "numerator",           big_numerator,        1, 0, false, H_numerator);
  s7_define_function(sc, "denominator",         big_denominator,      1, 0, false, H_denominator);
  s7_define_function(sc, "rationalize",         big_rationalize,      1, 1, false, H_rationalize);
  s7_define_function(sc, "exact->inexact",      big_exact_to_inexact, 1, 0, false, H_exact_to_inexact);
  s7_define_function(sc, "inexact->exact",      big_inexact_to_exact, 1, 0, false, H_inexact_to_exact);
  s7_define_function(sc, "floor",               big_floor,            1, 0, false, H_floor);
  s7_define_function(sc, "ceiling",             big_ceiling,          1, 0, false, H_ceiling);
  s7_define_function(sc, "truncate",            big_truncate,         1, 0, false, H_truncate);
  s7_define_function(sc, "round",               big_round,            1, 0, false, H_round);
  s7_define_function(sc, "quotient",            big_quotient,         2, 0, false, H_quotient);
  s7_define_function(sc, "remainder",           big_remainder,        2, 0, false, H_remainder);
  s7_define_function(sc, "modulo",              big_modulo,           2, 0, false, H_modulo);
  s7_define_function(sc, "gcd",                 big_gcd,              0, 0, true,  H_gcd);
  s7_define_function(sc, "lcm",                 big_lcm,              0, 0, true,  H_lcm);

  s7_define_function(sc, "make-rectangular",    big_make_rectangular, 2, 0, false, H_make_rectangular);
  s7_define_function(sc, "make-polar",          big_make_polar,       2, 0, false, H_make_polar);
  s7_define_function(sc, "real-part",           big_real_part,        1, 0, false, H_real_part);
  s7_define_function(sc, "imag-part",           big_imag_part,        1, 0, false, H_imag_part);
  s7_define_function(sc, "angle",               big_angle,            1, 0, false, H_angle);
  s7_define_function(sc, "magnitude",           big_magnitude,        1, 0, false, H_magnitude);
  s7_define_function(sc, "abs",                 big_abs,              1, 0, false, H_abs);

  s7_define_function(sc, "lognot",              big_lognot,           1, 0, false, H_lognot);
  s7_define_function(sc, "logior",              big_logior,           0, 0, true,  H_logior);
  s7_define_function(sc, "logxor",              big_logxor,           0, 0, true,  H_logxor);
  s7_define_function(sc, "logand",              big_logand,           0, 0, true,  H_logand);
  s7_define_function(sc, "ash",                 big_ash,              2, 0, false, H_ash);
  s7_define_function(sc, "integer-length",      big_integer_length,   1, 0, false, H_integer_length);
  
  s7_define_function(sc, "even?",               big_is_even,          1, 0, false, H_is_even);
  s7_define_function(sc, "odd?",                big_is_odd,           1, 0, false, H_is_odd);
  s7_define_function(sc, "zero?",               big_is_zero,          1, 0, false, H_is_zero);
  s7_define_function(sc, "positive?",           big_is_positive,      1, 0, false, H_is_positive);
  s7_define_function(sc, "negative?",           big_is_negative,      1, 0, false, H_is_negative);
  s7_define_function(sc, "infinite?",           big_is_infinite,      1, 0, false, H_is_infinite);
  s7_define_function(sc, "nan?",                big_is_nan,           1, 0, false, H_is_nan);

  s7_define_function(sc, "exp",                 big_exp,              1, 0, false, H_exp);
  s7_define_function(sc, "expt",                big_expt,             2, 0, false, H_expt);
  s7_define_function(sc, "log",                 big_log,              1, 1, false, H_log);
  s7_define_function(sc, "sqrt",                big_sqrt,             1, 0, false, H_sqrt);
  s7_define_function(sc, "sin",                 big_sin,              1, 0, false, H_sin);
  s7_define_function(sc, "cos",                 big_cos,              1, 0, false, H_cos);
  s7_define_function(sc, "tan",                 big_tan,              1, 0, false, H_tan);
  s7_define_function(sc, "asin",                big_asin,             1, 0, false, H_asin);
  s7_define_function(sc, "acos",                big_acos,             1, 0, false, H_acos);
  s7_define_function(sc, "atan",                big_atan,             1, 1, false, H_atan);
  s7_define_function(sc, "sinh",                big_sinh,             1, 0, false, H_sinh);
  s7_define_function(sc, "cosh",                big_cosh,             1, 0, false, H_cosh);
  s7_define_function(sc, "tanh",                big_tanh,             1, 0, false, H_tanh);
  s7_define_function(sc, "asinh",               big_asinh,            1, 0, false, H_asinh);
  s7_define_function(sc, "acosh",               big_acosh,            1, 0, false, H_acosh);
  s7_define_function(sc, "atanh",               big_atanh,            1, 0, false, H_atanh);

  big_rng_tag = s7_new_type("<big-random-number-generator>", print_big_rng, free_big_rng, equal_big_rng, NULL, NULL, NULL);
  s7_define_function(sc, "random",              big_random,           1, 1, false, H_random);
  s7_define_function(sc, "make-random-state",   make_big_random_state,1, 1, false, H_make_random_state);

  s7_define_function(sc, "bignum",              g_bignum,             1, 1, false, H_bignum);
  s7_define_function(sc, "bignum?",             g_is_bignum,          1, 0, false, H_is_bignum);
  s7_define_function_with_setter(sc, "bignum-precision", g_bignum_precision, g_set_bignum_precision, 0, 0, H_bignum_precision);

  add_max = (1 << (s7_int_bits - 1));
  mpfr_set_default_prec((mp_prec_t)DEFAULT_BIGNUM_PRECISION); 
  mpc_set_default_precision((mp_prec_t)DEFAULT_BIGNUM_PRECISION);

  s7_symbol_set_value(sc, make_symbol(sc, "pi"), big_pi(sc));

  /* if these fixnum limits were read as strings, they'd be bignums in the gmp case, 
   *   so for consistency make the symbolic versions bignums as well.
   */
  s7_symbol_set_value(sc, make_symbol(sc, "most-positive-fixnum"), s7_Int_to_big_integer(sc, s7_integer(s7_name_to_value(sc, "most-positive-fixnum"))));
  s7_symbol_set_value(sc, make_symbol(sc, "most-negative-fixnum"), s7_Int_to_big_integer(sc, s7_integer(s7_name_to_value(sc, "most-negative-fixnum"))));

  s7_provide(sc, "gmp");
}

#endif
/* WITH_GMP */




/* -------------------------------- initialization -------------------------------- */

s7_scheme *s7_init(void) 
{
  int i;
  s7_scheme *sc;

#if WITH_OPTIMIZATION
  finder = find_symbol_or_bust;
#endif
  init_types();
  init_ctables();
  init_mark_functions();
  init_pows();
  init_uppers();

#if WITH_COUNTS
  init_counts();
#endif

  sc = (s7_scheme *)calloc(1, sizeof(s7_scheme)); /* malloc is not recommended here */
  
  sc->gc_off = true;                         /* sc->args and so on are not set yet, so a gc during init -> segfault */
  sc->gc_stats = false;
  init_gc_caches(sc);

  sc->longjmp_ok = false;
  sc->symbol_table_is_locked = false;

  sc->strbuf_size = INITIAL_STRBUF_SIZE;
  sc->strbuf = (char *)calloc(sc->strbuf_size, sizeof(char));
  sc->print_width = MAX_STRING_LENGTH;
  
  sc->read_line_buf = NULL;
  sc->read_line_buf_size = 0;

  sc->NIL =         &sc->_NIL;
  sc->GC_NIL =      &sc->_GC_NIL;
  sc->T =           &sc->_T;
  sc->F =           &sc->_F;
  sc->EOF_OBJECT =  &sc->_EOF_OBJECT;
  sc->UNSPECIFIED = &sc->_UNSPECIFIED;  
  sc->UNDEFINED =   &sc->_UNDEFINED;
  sc->NO_VALUE =    &sc->_NO_VALUE;  
  sc->ELSE =        &sc->_ELSE;
  sc->TEMP_CELL =   &sc->_TEMP_CELL;
  sc->TEMP_CELL_1 = &sc->_TEMP_CELL_1;
  sc->TEMP_CELL_2 = &sc->_TEMP_CELL_2;
  sc->TEMP_CELL_3 = &sc->_TEMP_CELL_3;

  set_type(sc->NIL, T_NIL | T_GC_MARK | T_IMMUTABLE);
  car(sc->NIL) = cdr(sc->NIL) = sc->UNSPECIFIED;
  frame_id(sc->NIL) = -1;
  sc->NIL->hloc = NOT_IN_HEAP;
  
  set_type(sc->GC_NIL, T_UNTYPED | T_GC_MARK | T_IMMUTABLE);
  car(sc->GC_NIL) = cdr(sc->GC_NIL) = sc->UNSPECIFIED;
  sc->GC_NIL->hloc = NOT_IN_HEAP;
  
  set_type(sc->T, T_BOOLEAN | T_GC_MARK | T_IMMUTABLE);
  car(sc->T) = cdr(sc->T) = sc->UNSPECIFIED;
  sc->T->hloc = NOT_IN_HEAP;
  
  set_type(sc->F, T_BOOLEAN | T_GC_MARK | T_IMMUTABLE);
  car(sc->F) = cdr(sc->F) = sc->UNSPECIFIED;
  sc->F->hloc = NOT_IN_HEAP;
  
  set_type(sc->EOF_OBJECT, T_UNTYPED | T_GC_MARK | T_IMMUTABLE);
  car(sc->EOF_OBJECT) = cdr(sc->EOF_OBJECT) = sc->UNSPECIFIED;
  sc->EOF_OBJECT->hloc = NOT_IN_HEAP;  

  set_type(sc->UNSPECIFIED, T_UNTYPED | T_GC_MARK | T_IMMUTABLE);
  car(sc->UNSPECIFIED) = cdr(sc->UNSPECIFIED) = sc->UNSPECIFIED;
  sc->UNSPECIFIED->hloc = NOT_IN_HEAP;
  
  set_type(sc->UNDEFINED, T_UNTYPED | T_GC_MARK | T_IMMUTABLE);
  car(sc->UNDEFINED) = cdr(sc->UNDEFINED) = sc->UNSPECIFIED;
  sc->UNDEFINED->hloc = NOT_IN_HEAP;
  
  set_type(sc->NO_VALUE, T_UNTYPED | T_GC_MARK | T_IMMUTABLE);
  car(sc->NO_VALUE) = cdr(sc->NO_VALUE) = sc->UNSPECIFIED;
  sc->NO_VALUE->hloc = NOT_IN_HEAP;

  set_type(sc->ELSE, T_UNTYPED | T_GC_MARK | T_IMMUTABLE);
  car(sc->ELSE) = cdr(sc->ELSE) = sc->UNSPECIFIED;
  sc->ELSE->hloc = NOT_IN_HEAP;
  /* "else" is added to the global environment below -- can't do it here
   *    because the symbol table and environment don't exist yet.
   */

  set_type(sc->TEMP_CELL, T_PAIR | T_GC_MARK | T_IMMUTABLE);
  cdr(sc->TEMP_CELL) = sc->NIL;
  sc->TEMP_CELL->hloc = NOT_IN_HEAP;

  set_type(sc->TEMP_CELL_1, T_PAIR | T_GC_MARK | T_IMMUTABLE);
  car(sc->TEMP_CELL_1) = cdr(sc->TEMP_CELL_1) = sc->NIL;
  car(sc->TEMP_CELL) = sc->TEMP_CELL_1;
  sc->TEMP_CELL_1->hloc = NOT_IN_HEAP;
  
  set_type(sc->TEMP_CELL_2, T_PAIR | T_GC_MARK | T_IMMUTABLE);
  car(sc->TEMP_CELL_2) = cdr(sc->TEMP_CELL_2) = sc->NIL;
  sc->TEMP_CELL_2->hloc = NOT_IN_HEAP;

  set_type(sc->TEMP_CELL_3, T_PAIR | T_GC_MARK | T_IMMUTABLE);
  car(sc->TEMP_CELL_3) = cdr(sc->TEMP_CELL_3) = sc->NIL;
  sc->TEMP_CELL_3->hloc = NOT_IN_HEAP;

  sc->T1_1 = &sc->_T1_1;
  set_type(sc->T1_1, T_PAIR | T_GC_MARK | T_IMMUTABLE);
  car(sc->T1_1) = sc->NIL;
  cdr(sc->T1_1) = sc->NIL;
  sc->T1_1->hloc = NOT_IN_HEAP;

  sc->T2_1 = &sc->_T2_1;
  sc->T2_2 = &sc->_T2_2;

  set_type(sc->T2_1, T_PAIR | T_GC_MARK | T_IMMUTABLE);
  car(sc->T2_1) = sc->NIL;
  cdr(sc->T2_1) = sc->T2_2;
  sc->T2_1->hloc = NOT_IN_HEAP;
  set_type(sc->T2_2, T_PAIR | T_GC_MARK | T_IMMUTABLE);
  car(sc->T2_2) = sc->NIL;
  cdr(sc->T2_2) = sc->NIL;
  sc->T2_2->hloc = NOT_IN_HEAP;

  sc->T3_1 = &sc->_T3_1;
  sc->T3_2 = &sc->_T3_2;
  sc->T3_3 = &sc->_T3_3;

  set_type(sc->T3_1, T_PAIR | T_GC_MARK | T_IMMUTABLE);
  car(sc->T3_1) = sc->NIL;
  cdr(sc->T3_1) = sc->T3_2;
  sc->T3_1->hloc = NOT_IN_HEAP;
  set_type(sc->T3_2, T_PAIR | T_GC_MARK | T_IMMUTABLE);
  car(sc->T3_2) = sc->NIL;
  cdr(sc->T3_2) = sc->T3_3;
  sc->T3_2->hloc = NOT_IN_HEAP;
  set_type(sc->T3_3, T_PAIR | T_GC_MARK | T_IMMUTABLE);
  car(sc->T3_3) = sc->NIL;
  cdr(sc->T3_3) = sc->NIL;
  sc->T3_3->hloc = NOT_IN_HEAP;

  sc->input_port_stack = sc->NIL;
  sc->code = sc->NIL;
  sc->cur_code = ERROR_INFO_DEFAULT;
  sc->args = sc->NIL;
  sc->value = sc->NIL;
  sc->w = sc->NIL;
  sc->x = sc->NIL;
  sc->y = sc->NIL;
  sc->z = sc->NIL;

  sc->temp1 = sc->NIL;
  sc->temp2 = sc->NIL;
  sc->temp3 = sc->NIL;
  sc->temp4 = sc->NIL;

  sc->error_exiter = NULL;
  sc->begin_hook = NULL;
  sc->default_rng = NULL;
  
  sc->heap_size = INITIAL_HEAP_SIZE;
  if ((sc->heap_size % 32) != 0)
    sc->heap_size = 32 * (int)ceil((double)(sc->heap_size) / 32.0);
  sc->heap = (s7_pointer *)malloc(sc->heap_size * sizeof(s7_pointer));
  
  sc->free_heap = (s7_cell **)malloc(sc->heap_size * sizeof(s7_cell *));
  sc->free_heap_top = (s7_cell **)(sc->free_heap + INITIAL_HEAP_SIZE);
  sc->free_heap_trigger = (s7_cell **)(sc->free_heap + GC_TRIGGER_SIZE);
  sc->previous_free_heap_top = sc->free_heap_top;

  {
    s7_cell *cells;
    cells = (s7_cell *)calloc(INITIAL_HEAP_SIZE, sizeof(s7_cell));
    for (i = 0; i < INITIAL_HEAP_SIZE; i++)
      {
	sc->heap[i] = &cells[i];
 	sc->free_heap[i] = sc->heap[i];
 	sc->heap[i]->hloc = i;
      }
  }

  permanent_heap = (unsigned char *)calloc(PERMANENT_HEAP_SIZE, sizeof(unsigned char));
  permanent_heap_top = (unsigned char *)(permanent_heap + PERMANENT_HEAP_SIZE);
  
  /* this has to precede s7_make_* allocations */
  sc->protected_objects_size = INITIAL_PROTECTED_OBJECTS_SIZE;
  sc->protected_objects_loc = 0;
  sc->protected_objects = s7_make_vector(sc, INITIAL_PROTECTED_OBJECTS_SIZE);
  for (i = 0; i < INITIAL_PROTECTED_OBJECTS_SIZE; i++)
    vector_element(sc->protected_objects, i) = sc->GC_NIL;
  set_immutable(sc->protected_objects);
  
  sc->stack = s7_make_vector(sc, INITIAL_STACK_SIZE);
  sc->stack_start = vector_elements(sc->stack);
  sc->stack_end = sc->stack_start;
  sc->stack_size = INITIAL_STACK_SIZE;
  sc->stack_resize_trigger = (s7_pointer *)(sc->stack_start + sc->stack_size / 2);
  set_type(sc->stack, T_STACK);

  {
    s7_cell *cells;
    sc->all_g_args = (s7_pointer *)malloc((GC_TRIGGER_SIZE + 1) * sizeof(s7_pointer));
    cells = (s7_cell *)calloc(GC_TRIGGER_SIZE + 1, sizeof(s7_cell));
    for (i = 0; i <= GC_TRIGGER_SIZE; i++)
      {
	sc->all_g_args[i] = &cells[i];
	set_type(sc->all_g_args[i], T_PAIR | T_IMMUTABLE);
	car(sc->all_g_args[i]) = sc->NIL;
	cdr(sc->all_g_args[i]) = sc->NIL;
      }
  }

  initialize_op_stack(sc);
  
  /* keep the symbol table out of the heap */
  sc->symbol_table = (s7_pointer)calloc(1, sizeof(s7_cell));
  set_type(sc->symbol_table, T_VECTOR);
  vector_length(sc->symbol_table) = SYMBOL_TABLE_SIZE;
  vector_elements(sc->symbol_table) = (s7_pointer *)malloc(SYMBOL_TABLE_SIZE * sizeof(s7_pointer));
  s7_vector_fill(sc, sc->symbol_table, sc->NIL);
  sc->symbol_table->hloc = NOT_IN_HEAP;
  
  sc->gensym_counter = 0;
  tracing = false;
  trace_all = false;

  sc->trace_list = (s7_pointer *)calloc(INITIAL_TRACE_LIST_SIZE, sizeof(s7_pointer));
  sc->trace_list_size = INITIAL_TRACE_LIST_SIZE;
  sc->trace_top = 0;
  sc->trace_depth = 0;
  sc->no_values = 0;
  sc->s7_call_line = 0;
  sc->s7_call_file = NULL;
  sc->s7_call_name = NULL;
  sc->safety = 0;
  sc->circle_info = NULL;
  sc->fdats = (format_data **)calloc(8, sizeof(format_data *));
  sc->num_fdats = 8;

  sc->global_env = s7_make_vector(sc, 512);
  type(sc->global_env) = T_ENVIRONMENT;
  vector_fill_pointer(sc->global_env) = 0;
  for (i = 0; i < 512; i++)
    vector_element(sc->global_env, i) = sc->NIL;
  sc->envir = sc->NIL;
  
  /* keep the small_ints out of the heap */
  small_ints = (s7_pointer *)malloc((NUM_SMALL_INTS + 1) * sizeof(s7_pointer));
  {
    s7_cell *cells;
    cells = (s7_cell *)calloc((NUM_SMALL_INTS + 1), sizeof(s7_cell));
    for (i = 0; i <= NUM_SMALL_INTS; i++) 
      {
	s7_pointer p;
	small_ints[i] = &cells[i];
	p = small_ints[i];
	typeflag(p) = T_IMMUTABLE | T_INTEGER;
	p->hloc = NOT_IN_HEAP;
	integer(p) = i;
      }
  }

  real_zero = make_permanent_real(0.0);
  real_one = make_permanent_real(1.0);
  real_NaN = make_permanent_real(NAN);
  real_pi = make_permanent_real(3.1415926535897932384626433832795029L); /* M_PI is not good enough for s7_Double = long double */

  /* keep the characters out of the heap */
  chars = (s7_pointer *)malloc((NUM_CHARS + 1) * sizeof(s7_pointer));
  chars[0] = sc->EOF_OBJECT;
  chars++;
  {
    s7_cell *cells;
    cells = (s7_cell *)calloc(NUM_CHARS, sizeof(s7_cell));
    for (i = 0; i < NUM_CHARS; i++) 
      {
	s7_pointer p;
	p = &cells[i];
	typeflag(p) = T_IMMUTABLE | T_CHARACTER;
	p->hloc = NOT_IN_HEAP;
	character(p) = (unsigned char)i;
	chars[i] = p;
      }
  }

  make_standard_ports(sc);

  assign_syntax(sc, "quote",             OP_QUOTE);
  assign_syntax(sc, "if",                OP_IF);
  assign_syntax(sc, "begin",             OP_BEGIN);
  assign_syntax(sc, "set!",              OP_SET);
  assign_syntax(sc, "let",               OP_LET);
  assign_syntax(sc, "let*",              OP_LET_STAR);
  assign_syntax(sc, "letrec",            OP_LETREC);
  assign_syntax(sc, "cond",              OP_COND);
  assign_syntax(sc, "and",               OP_AND);
  assign_syntax(sc, "or",                OP_OR);
  assign_syntax(sc, "case",              OP_CASE);
  assign_syntax(sc, "do",                OP_DO);

  set_immutable(assign_syntax(sc, "with-environment",  OP_WITH_ENV));

  assign_syntax(sc, "lambda",            OP_LAMBDA);
  assign_syntax(sc, "lambda*",           OP_LAMBDA_STAR);      /* optional, key, rest args */
  assign_syntax(sc, "define",            OP_DEFINE);
  assign_syntax(sc, "define*",           OP_DEFINE_STAR);
  assign_syntax(sc, "define-constant",   OP_DEFINE_CONSTANT);  /* unsetabble and unrebindable bindings */

  assign_syntax(sc, "defmacro",          OP_DEFMACRO);         /* CL-style macro syntax */
  assign_syntax(sc, "defmacro*",         OP_DEFMACRO_STAR);
  assign_syntax(sc, "define-macro",      OP_DEFINE_MACRO);     /* Scheme-style macro syntax */
  assign_syntax(sc, "define-macro*",     OP_DEFINE_MACRO_STAR); 
  assign_syntax(sc, "define-expansion",  OP_DEFINE_EXPANSION); /* read-time (immediate) macro expansion */
  assign_syntax(sc, "define-bacro",      OP_DEFINE_BACRO);     /* macro expansion in calling environment */
  assign_syntax(sc, "define-bacro*",     OP_DEFINE_BACRO_STAR);

  sc->QUOTE_UNCHECKED =       assign_internal_syntax(sc, "quote",   OP_QUOTE_UNCHECKED);  
  sc->LET_UNCHECKED =         assign_internal_syntax(sc, "let",     OP_LET_UNCHECKED);  
  sc->LET_STAR_UNCHECKED =    assign_internal_syntax(sc, "let*",    OP_LET_STAR_UNCHECKED);  
  sc->LETREC_UNCHECKED =      assign_internal_syntax(sc, "letrec",  OP_LETREC_UNCHECKED);  
  sc->LET_NO_VARS =           assign_internal_syntax(sc, "let",     OP_LET_NO_VARS);  
  sc->LET_C =                 assign_internal_syntax(sc, "let",     OP_LET_C);  
  sc->LET_C_P =               assign_internal_syntax(sc, "let",     OP_LET_C_P);  
  sc->LET_S =                 assign_internal_syntax(sc, "let",     OP_LET_S);  
  sc->LET_S_P =               assign_internal_syntax(sc, "let",     OP_LET_S_P);  
  sc->LET_Q =                 assign_internal_syntax(sc, "let",     OP_LET_Q);  
  sc->LET_ALL_C =             assign_internal_syntax(sc, "let",     OP_LET_ALL_C);  
  sc->LET_ALL_S =             assign_internal_syntax(sc, "let",     OP_LET_ALL_S);  
  sc->LET_ALL_Q =             assign_internal_syntax(sc, "let",     OP_LET_ALL_Q);  
  sc->LET_ALL_G =             assign_internal_syntax(sc, "let",     OP_LET_ALL_G);  
  sc->NAMED_LET =             assign_internal_syntax(sc, "let",     OP_NAMED_LET);  
  sc->NAMED_LET_NO_VARS =     assign_internal_syntax(sc, "let",     OP_NAMED_LET_NO_VARS);  
  sc->SET_UNCHECKED =         assign_internal_syntax(sc, "set!",    OP_SET_UNCHECKED);  
  sc->SET_SYMBOL_C =          assign_internal_syntax(sc, "set!",    OP_SET_SYMBOL_C);  
  sc->SET_SYMBOL_S =          assign_internal_syntax(sc, "set!",    OP_SET_SYMBOL_S);  
  sc->SET_SYMBOL_Q =          assign_internal_syntax(sc, "set!",    OP_SET_SYMBOL_Q);  
  sc->SET_SYMBOL_R =          assign_internal_syntax(sc, "set!",    OP_SET_SYMBOL_R);  
  sc->SET_SYMBOL_P =          assign_internal_syntax(sc, "set!",    OP_SET_SYMBOL_P);  
  sc->CASE_UNCHECKED =        assign_internal_syntax(sc, "case",    OP_CASE_UNCHECKED);  
  sc->CASE_INT =              assign_internal_syntax(sc, "case",    OP_CASE_INT);  
  sc->CASE_SIMPLE =           assign_internal_syntax(sc, "case",    OP_CASE_SIMPLE);  
  sc->CASE_SIMPLER =          assign_internal_syntax(sc, "case",    OP_CASE_SIMPLER);  
  sc->CASE_SIMPLEST =         assign_internal_syntax(sc, "case",    OP_CASE_SIMPLEST);  
  sc->COND_UNCHECKED =        assign_internal_syntax(sc, "cond",    OP_COND_UNCHECKED);  
  sc->COND_SIMPLE =           assign_internal_syntax(sc, "cond",    OP_COND_SIMPLE);  
  sc->DO_UNCHECKED =          assign_internal_syntax(sc, "do",      OP_DO_UNCHECKED);  
  sc->LAMBDA_UNCHECKED =      assign_internal_syntax(sc, "lambda",  OP_LAMBDA_UNCHECKED);  
  sc->LAMBDA_STAR_UNCHECKED = assign_internal_syntax(sc, "lambda*", OP_LAMBDA_STAR_UNCHECKED);  
  sc->DEFINE_UNCHECKED =      assign_internal_syntax(sc, "define",  OP_DEFINE_UNCHECKED);  
  sc->DEFINE_STAR_UNCHECKED = assign_internal_syntax(sc, "define*", OP_DEFINE_STAR_UNCHECKED);  
  sc->SET_NORMAL =            assign_internal_syntax(sc, "set!",    OP_SET_NORMAL);
  sc->SET_PAIR =              assign_internal_syntax(sc, "set!",    OP_SET_PAIR);
  sc->AND_UNCHECKED =         assign_internal_syntax(sc, "and",     OP_AND_UNCHECKED);
  sc->AND_P =                 assign_internal_syntax(sc, "and",     OP_AND_P);
  sc->OR_UNCHECKED =          assign_internal_syntax(sc, "or",      OP_OR_UNCHECKED);
  sc->OR_P =                  assign_internal_syntax(sc, "or",      OP_OR_P);

  sc->IF_P_P_P =              assign_internal_syntax(sc, "if",      OP_IF_P_P_P);
  sc->IF_P_P =                assign_internal_syntax(sc, "if",      OP_IF_P_P);
  sc->IF_ANDP_P =             assign_internal_syntax(sc, "if",      OP_IF_ANDP_P);
  sc->IF_ANDP_P_P =           assign_internal_syntax(sc, "if",      OP_IF_ANDP_P_P);
  sc->IF_ORP_P =              assign_internal_syntax(sc, "if",      OP_IF_ORP_P);
  sc->IF_ORP_P_P =            assign_internal_syntax(sc, "if",      OP_IF_ORP_P_P);
  sc->IF_B_P =                assign_internal_syntax(sc, "if",      OP_IF_B_P);
  sc->IF_B_P_P =              assign_internal_syntax(sc, "if",      OP_IF_B_P_P);
  sc->IF_P_P_X =              assign_internal_syntax(sc, "if",      OP_IF_P_P_X);
  sc->IF_P_X_P =              assign_internal_syntax(sc, "if",      OP_IF_P_X_P);
  sc->IF_P_X =                assign_internal_syntax(sc, "if",      OP_IF_P_X);
  sc->IF_P_X_X =              assign_internal_syntax(sc, "if",      OP_IF_P_X_X);
  sc->IF_X_P_P =              assign_internal_syntax(sc, "if",      OP_IF_X_P_P);
  sc->IF_X_P =                assign_internal_syntax(sc, "if",      OP_IF_X_P);
  sc->IF_X_P_X =              assign_internal_syntax(sc, "if",      OP_IF_X_P_X);
  sc->IF_X_X_P =              assign_internal_syntax(sc, "if",      OP_IF_X_X_P);
  sc->IF_X_X =                assign_internal_syntax(sc, "if",      OP_IF_X_X);
  sc->IF_X_X_X =              assign_internal_syntax(sc, "if",      OP_IF_X_X_X);

#if WITH_OPTIMIZATION
  sc->COND_SIMPLER =          assign_internal_syntax(sc, "cond",    OP_COND_SIMPLER);  
  sc->SAFER_AND_X =           assign_internal_syntax(sc, "and",     OP_SAFER_AND_X);  
  sc->SAFE_AND_S =            assign_internal_syntax(sc, "and",     OP_SAFE_AND_S);  
  sc->SAFE_OR_S =             assign_internal_syntax(sc, "or",      OP_SAFE_OR_S);  
  sc->SAFER_OR_S =            assign_internal_syntax(sc, "or",      OP_SAFER_OR_S);  
  sc->SAFER_OR_C =            assign_internal_syntax(sc, "or",      OP_SAFER_OR_C);  
  sc->SAFER_OR_CEQ =          assign_internal_syntax(sc, "or",      OP_SAFER_OR_CEQ);  
  sc->SAFER_OR_X =            assign_internal_syntax(sc, "or",      OP_SAFER_OR_X);  
  sc->SAFE_CASE_S =           assign_internal_syntax(sc, "case",    OP_SAFE_CASE_S);  
  sc->SAFE_IF1 =              assign_internal_syntax(sc, "if",      OP_SAFE_IF1);  
  sc->SAFE_IF2 =              assign_internal_syntax(sc, "if",      OP_SAFE_IF2);  
  sc->IF_O_P_P =              assign_internal_syntax(sc, "if",      OP_IF_O_P_P);
  sc->IF_O_P =                assign_internal_syntax(sc, "if",      OP_IF_O_P);
  sc->SAFE_IF_ANDX_P =        assign_internal_syntax(sc, "if",      OP_SAFE_IF_ANDX_P);
  sc->SAFE_IF_ORX_P =         assign_internal_syntax(sc, "if",      OP_SAFE_IF_ORX_P);
  sc->SAFE_IF_CC_X_P =        assign_internal_syntax(sc, "if",      OP_SAFE_IF_CC_X_P);  
  sc->SAFE_IF_CC_P =          assign_internal_syntax(sc, "if",      OP_SAFE_IF_CC_P);  
  sc->IF_ORCEQ_P =            assign_internal_syntax(sc, "if",      OP_IF_ORCEQ_P);
  sc->IF_ORCEQ_P_P =          assign_internal_syntax(sc, "if",      OP_IF_ORCEQ_P_P);
  sc->SAFE_IF_CC_P_P =        assign_internal_syntax(sc, "if",      OP_SAFE_IF_CC_P_P);  
  sc->SAFE_IF_CEQ_P_P =       assign_internal_syntax(sc, "if",      OP_SAFE_IF_CEQ_P_P);  
  sc->SAFE_IF_CS_P =          assign_internal_syntax(sc, "if",      OP_SAFE_IF_CS_P);  
  sc->SAFE_IF_CS_P_P =        assign_internal_syntax(sc, "if",      OP_SAFE_IF_CS_P_P);  
  sc->SAFE_IF_CS_Q_P =        assign_internal_syntax(sc, "if",      OP_SAFE_IF_CS_Q_P);  
  sc->SAFE_IF_CS_P_X =        assign_internal_syntax(sc, "if",      OP_SAFE_IF_CS_P_X);  
  sc->SAFE_IF_CS_X_P =        assign_internal_syntax(sc, "if",      OP_SAFE_IF_CS_X_P);
  sc->SAFE_IF_CSS_X_P =       assign_internal_syntax(sc, "if",      OP_SAFE_IF_CSS_X_P);
  sc->SAFE_IF_CSC_X_P_A =     assign_internal_syntax(sc, "if",      OP_SAFE_IF_CSC_X_P_A);
  sc->SAFE_IF_CSC_X_P =       assign_internal_syntax(sc, "if",      OP_SAFE_IF_CSC_X_P);
  sc->SAFE_IF_CSQ_P =         assign_internal_syntax(sc, "if",      OP_SAFE_IF_CSQ_P);    
  sc->SAFE_IF_CSQ_P_P =       assign_internal_syntax(sc, "if",      OP_SAFE_IF_CSQ_P_P);    
  sc->SAFE_IF_CSS_P =         assign_internal_syntax(sc, "if",      OP_SAFE_IF_CSS_P);    
  sc->SAFE_IF_CSS_P_P =       assign_internal_syntax(sc, "if",      OP_SAFE_IF_CSS_P_P);    
  sc->SAFE_IF_CSC_P =         assign_internal_syntax(sc, "if",      OP_SAFE_IF_CSC_P);    
  sc->SAFE_IF_CSC_P_P =       assign_internal_syntax(sc, "if",      OP_SAFE_IF_CSC_P_P);    
  sc->SAFE_IF_C_SS_P =        assign_internal_syntax(sc, "if",      OP_SAFE_IF_C_SS_P);    
  sc->SAFE_IF_IS_PAIR_P =     assign_internal_syntax(sc, "if",      OP_SAFE_IF_IS_PAIR_P);  
  sc->SAFE_IF_IS_PAIR_P_X =   assign_internal_syntax(sc, "if",      OP_SAFE_IF_IS_PAIR_P_X);  
  sc->SAFE_IF_IS_PAIR_P_P =   assign_internal_syntax(sc, "if",      OP_SAFE_IF_IS_PAIR_P_P);  
  sc->SAFE_IF_IS_SYMBOL_P =   assign_internal_syntax(sc, "if",      OP_SAFE_IF_IS_SYMBOL_P);  
  sc->SAFE_IF_IS_SYMBOL_P_X = assign_internal_syntax(sc, "if",      OP_SAFE_IF_IS_SYMBOL_P_X);  
  sc->SAFE_IF_IS_SYMBOL_P_P = assign_internal_syntax(sc, "if",      OP_SAFE_IF_IS_SYMBOL_P_P);  
  sc->SAFE_IF_IS_EOF_P_P =    assign_internal_syntax(sc, "if",      OP_SAFE_IF_IS_EOF_P_P);  
  sc->LET_R =                 assign_internal_syntax(sc, "let",     OP_LET_R);  
  sc->LET_R_P =               assign_internal_syntax(sc, "let",     OP_LET_R_P);  
  sc->LET_CAR_P =             assign_internal_syntax(sc, "let",     OP_LET_CAR_P);  
  sc->LET_READ_CHAR_P =       assign_internal_syntax(sc, "let",     OP_LET_READ_CHAR_P);  
  sc->LET_ALL_R =             assign_internal_syntax(sc, "let",     OP_LET_ALL_R);  
  sc->LET_O =                 assign_internal_syntax(sc, "let",     OP_LET_O);  
  sc->LET_O_P =               assign_internal_syntax(sc, "let",     OP_LET_O_P);  
  sc->LET_OX_P =              assign_internal_syntax(sc, "let",     OP_LET_OX_P);  
  sc->LET_UNKNOWN_S_P =       assign_internal_syntax(sc, "let",     OP_LET_UNKNOWN_S_P);  
  sc->LET_C_D =               assign_internal_syntax(sc, "let",     OP_LET_C_D);  
  sc->INCREMENT_1 =           assign_internal_syntax(sc, "set!",    OP_INCREMENT_1);  
  sc->DECREMENT_1 =           assign_internal_syntax(sc, "set!",    OP_DECREMENT_1);  
  sc->SET_CDR =               assign_internal_syntax(sc, "set!",    OP_SET_CDR);
  sc->SET_CONS =              assign_internal_syntax(sc, "set!",    OP_SET_CONS);
  sc->SIMPLE_DO =             assign_internal_syntax(sc, "do",      OP_SIMPLE_DO);
  sc->SIMPLE_DO_P =           assign_internal_syntax(sc, "do",      OP_SIMPLE_DO_P);
  sc->SIMPLE_DO_OPT =         assign_internal_syntax(sc, "do",      OP_SIMPLE_DO_OPT);
  sc->SAFE_SIMPLE_DO =        assign_internal_syntax(sc, "do",      OP_SAFE_SIMPLE_DO);
  sc->DOTIMES =               assign_internal_syntax(sc, "do",      OP_DOTIMES);
  sc->DOTIMES_C_C =           assign_internal_syntax(sc, "do",      OP_DOTIMES_C_C);
  sc->SIMPLE_DOTIMES =        assign_internal_syntax(sc, "do",      OP_SIMPLE_DOTIMES);
  sc->SAFE_DO =               assign_internal_syntax(sc, "do",      OP_SAFE_DO);
  sc->DOX =                   assign_internal_syntax(sc, "do",      OP_DOX);
  sc->safe_do_level = 0;
  sc->safe_do_ids_size = 8;
  sc->safe_do_ids = (long long int *)calloc(sc->safe_do_ids_size, sizeof(long long int));
#endif

  sc->LAMBDA =      make_symbol(sc, "lambda");
  sc->LAMBDA_STAR = make_symbol(sc, "lambda*");
  sc->QUOTE =       make_symbol(sc, "quote");
  sc->UNQUOTE =     make_symbol(sc, "unquote");
  sc->MACROEXPAND = make_symbol(sc, "macroexpand");
  sc->FEED_TO =     make_symbol(sc, "=>");
  sc->APPLY =       make_symbol(sc, "apply");
  sc->CDR =         make_symbol(sc, "cdr");
  sc->SET =         make_symbol(sc, "set!");

  #define s_is_type_name "[?]"                         /* these were "(?)" etc, but the procedure-source needs to be usable */
  sc->S_IS_TYPE = make_symbol(sc, s_is_type_name);

  #define s_type_make_name "[make]"
  sc->S_TYPE_MAKE = make_symbol(sc, s_type_make_name);

  #define s_type_ref_name "[ref]"
  sc->S_TYPE_REF = make_symbol(sc, s_type_ref_name);

  #define s_type_arg_name "[arg]"
  sc->S_TYPE_ARG = make_symbol(sc, s_type_arg_name);

  sc->QQ_VALUES =       make_symbol(sc, "{values}");
  sc->QQ_LIST =         make_symbol(sc, "{list}");
  sc->QQ_APPLY_VALUES = make_symbol(sc, "{apply} {values}");
  sc->QQ_APPEND =       make_symbol(sc, "{append}");
  sc->MULTIVECTOR =     make_symbol(sc, "{multivector}");

  s7_make_slot(sc, sc->NIL, make_symbol(sc, "else"), sc->ELSE);

  sc->VECTOR = make_symbol(sc, "vector");
  typeflag(sc->VECTOR) |= T_SETTER; 
  
  sc->ERROR =          make_symbol(sc, "error");
  sc->READ_ERROR =     make_symbol(sc, "read-error");
  sc->SYNTAX_ERROR =   make_symbol(sc, "syntax-error");
  sc->WRONG_TYPE_ARG = make_symbol(sc, "wrong-type-arg");

  sc->WRONG_TYPE_ARG_INFO = sc->NIL;
  for (i = 0; i < 7; i++)
    sc->WRONG_TYPE_ARG_INFO = permanent_cons(sc->F, sc->WRONG_TYPE_ARG_INFO, T_PAIR);
  car(sc->WRONG_TYPE_ARG_INFO) = s7_make_permanent_string("~A argument ~D, ~S, is a~A ~A but should be ~A");

  sc->SIMPLE_WRONG_TYPE_ARG_INFO = sc->NIL;
  for (i = 0; i < 6; i++)
    sc->SIMPLE_WRONG_TYPE_ARG_INFO = permanent_cons(sc->F, sc->SIMPLE_WRONG_TYPE_ARG_INFO, T_PAIR);
  car(sc->SIMPLE_WRONG_TYPE_ARG_INFO) = s7_make_permanent_string("~A argument, ~S, is a~A ~A but should be ~A");

  sc->WRONG_NUMBER_OF_ARGS = make_symbol(sc, "wrong-number-of-args");
  sc->FORMAT_ERROR = make_symbol(sc, "format-error");
  sc->OUT_OF_RANGE = make_symbol(sc, "out-of-range");
  sc->OUT_OF_RANGE_INFO = sc->NIL;
  for (i = 0; i < 5; i++)
    sc->OUT_OF_RANGE_INFO = permanent_cons(sc->F, sc->OUT_OF_RANGE_INFO, T_PAIR);
  car(sc->OUT_OF_RANGE_INFO) = s7_make_permanent_string("~A argument ~D, ~S, is out of range (~A)");

  sc->SIMPLE_OUT_OF_RANGE_INFO = sc->NIL;
  for (i = 0; i < 4; i++)
    sc->SIMPLE_OUT_OF_RANGE_INFO = permanent_cons(sc->F, sc->SIMPLE_OUT_OF_RANGE_INFO, T_PAIR);
  car(sc->SIMPLE_OUT_OF_RANGE_INFO) = s7_make_permanent_string("~A argument, ~S, is out of range (~A)");

  sc->KEY_KEY =              s7_make_keyword(sc, "key");
  sc->KEY_OPTIONAL =         s7_make_keyword(sc, "optional");
  sc->KEY_ALLOW_OTHER_KEYS = s7_make_keyword(sc, "allow-other-keys");
  sc->KEY_REST =             s7_make_keyword(sc, "rest");

  sc->__FUNC__ = make_symbol(sc, "__func__");

  sc->s_function_args = permanent_cons(sc->F, sc->NIL, T_PAIR);

  sc->TOO_MANY_ARGUMENTS = s7_make_permanent_string("~A: too many arguments: ~A");
  sc->NOT_ENOUGH_ARGUMENTS = s7_make_permanent_string("~A: not enough arguments: ~A");

  sc->gc_off= false;

  /* pws first so that make-procedure-with-setter has a type tag */
  s7_define_safe_function(sc, "make-procedure-with-setter",   g_make_procedure_with_setter,   2, 0, false, H_make_procedure_with_setter);
  s7_define_safe_function(sc, "procedure-with-setter?",       g_is_procedure_with_setter,     1, 0, false, H_is_procedure_with_setter);
  pws_tag = s7_new_type("<procedure-with-setter>", pws_print, pws_free,	pws_equal, pws_mark, pws_apply,	pws_set);
  

  s7_define_safe_function(sc, "gensym",                    g_gensym,                   0, 1, false, H_gensym);
  s7_define_safe_function(sc, "symbol-table",              g_symbol_table,             0, 0, false, H_symbol_table);
  s7_define_safe_function(sc, "symbol?",                   g_is_symbol,                1, 0, false, H_is_symbol);
  s7_define_safe_function(sc, "symbol->string",            g_symbol_to_string,         1, 0, false, H_symbol_to_string);
  s7_define_safe_function(sc, "string->symbol",            g_string_to_symbol,         1, 0, false, H_string_to_symbol);
  s7_define_safe_function(sc, "symbol",                    g_symbol,                   1, 0, false, H_symbol);
  s7_define_safe_function(sc, "symbol->value",             g_symbol_to_value,          1, 1, false, H_symbol_to_value);
  s7_define_function_with_setter(sc, "symbol-access",      g_symbol_get_access, g_symbol_set_access, 1, 0, H_symbol_access);
  
  s7_define_safe_function(sc, "outer-environment",         g_outer_environment,        1, 0, false, H_outer_environment);
  s7_define_safe_function(sc, "global-environment",        g_global_environment,       0, 0, false, H_global_environment);
  s7_define_safe_function(sc, "current-environment",       g_current_environment,      0, 0, false, H_current_environment);
  s7_define_constant_function(sc, "initial-environment",   g_initial_environment,      0, 0, false, H_initial_environment);
  s7_define_function(sc, "augment-environment",            g_augment_environment,      1, 0, true,  H_augment_environment);
  s7_define_function(sc, "augment-environment!",           g_augment_environment_direct, 1, 0, true,  H_augment_environment_direct);
  s7_define_safe_function(sc, "environment?",              g_is_environment,           1, 0, false, H_is_environment);
  s7_define_safe_function(sc, "environment->list",         g_environment_to_list,      1, 0, false, H_environment_to_list);
  s7_define_safe_function(sc, "provided?",                 g_is_provided,              1, 0, false, H_is_provided);
  s7_define_safe_function(sc, "provide",                   g_provide,                  1, 0, false, H_provide);
  s7_define_safe_function(sc, "defined?",                  g_is_defined,               1, 1, false, H_is_defined);
  s7_define_safe_function(sc, "constant?",                 g_is_constant,              1, 0, false, H_is_constant);
  s7_define_safe_function(sc, "macro?",                    g_is_macro,                 1, 0, false, H_is_macro);


  s7_define_safe_function(sc, "keyword?",                  g_is_keyword,               1, 0, false, H_is_keyword);
  s7_define_safe_function(sc, "make-keyword",              g_make_keyword,             1, 0, false, H_make_keyword);
  s7_define_safe_function(sc, "symbol->keyword",           g_symbol_to_keyword,        1, 0, false, H_symbol_to_keyword);
  s7_define_safe_function(sc, "keyword->symbol",           g_keyword_to_symbol,        1, 0, false, H_keyword_to_symbol);
  

  s7_define_safe_function(sc, "port-line-number",          g_port_line_number,         0, 1, false, H_port_line_number);
  s7_define_safe_function(sc, "port-filename",             g_port_filename,            0, 1, false, H_port_filename);
  s7_define_safe_function(sc, "pair-line-number",          g_pair_line_number,         1, 0, false, H_pair_line_number);
  
  s7_define_safe_function(sc, "input-port?",               g_is_input_port,            1, 0, false, H_is_input_port);
  s7_define_safe_function(sc, "output-port?",              g_is_output_port,           1, 0, false, H_is_output_port);
  s7_define_safe_function(sc, "port-closed?",              g_is_port_closed,           1, 0, false, H_is_port_closed);
  s7_define_safe_function(sc, "char-ready?",               g_is_char_ready,            0, 1, false, H_is_char_ready);
  s7_define_safe_function(sc, "eof-object?",               g_is_eof_object,            1, 0, false, H_is_eof_object);
  /* this should be named eof? (what isn't an object?) */
  
  s7_define_safe_function(sc, "current-input-port",        g_current_input_port,       0, 0, false, H_current_input_port);
  s7_define_safe_function(sc, "set-current-input-port",    g_set_current_input_port,   1, 0, false, H_set_current_input_port);
  s7_define_safe_function(sc, "current-output-port",       g_current_output_port,      0, 0, false, H_current_output_port);
  s7_define_safe_function(sc, "set-current-output-port",   g_set_current_output_port,  1, 0, false, H_set_current_output_port);
  s7_define_safe_function(sc, "current-error-port",        g_current_error_port,       0, 0, false, H_current_error_port);
  s7_define_safe_function(sc, "set-current-error-port",    g_set_current_error_port,   1, 0, false, H_set_current_error_port);
  s7_define_safe_function(sc, "close-input-port",          g_close_input_port,         1, 0, false, H_close_input_port);
  s7_define_safe_function(sc, "close-output-port",         g_close_output_port,        1, 0, false, H_close_output_port);
  s7_define_safe_function(sc, "open-input-file",           g_open_input_file,          1, 1, false, H_open_input_file);
  s7_define_safe_function(sc, "open-output-file",          g_open_output_file,         1, 1, false, H_open_output_file);
  s7_define_safe_function(sc, "open-input-string",         g_open_input_string,        1, 0, false, H_open_input_string);
  s7_define_safe_function(sc, "open-output-string",        g_open_output_string,       0, 0, false, H_open_output_string);
  s7_define_safe_function(sc, "get-output-string",         g_get_output_string,        1, 0, false, H_get_output_string);
  
  s7_define_safe_function(sc, "read-char",                 g_read_char,                0, 1, false, H_read_char);
  s7_define_safe_function(sc, "peek-char",                 g_peek_char,                0, 1, false, H_peek_char);
  s7_define_function(sc, "read",                           g_read,                     0, 1, false, H_read);
  /* read can't be safe because it messes with the stack, expecting to be all by itself in the call sequence (not embedded in OP_SAFE_C_opSq for example)
   */
  s7_define_safe_function(sc, "newline",                   g_newline,                  0, 1, false, H_newline);
  s7_define_safe_function(sc, "write-char",                g_write_char,               1, 1, false, H_write_char);
  s7_define_safe_function(sc, "write",                     g_write,                    1, 1, false, H_write);
  s7_define_safe_function(sc, "display",                   g_display,                  1, 1, false, H_display);
  s7_define_safe_function(sc, "read-byte",                 g_read_byte,                0, 1, false, H_read_byte);
  s7_define_safe_function(sc, "write-byte",                g_write_byte,               1, 1, false, H_write_byte);
  s7_define_safe_function(sc, "read-line",                 g_read_line,                0, 2, false, H_read_line);
  
  s7_define_function(sc, "call-with-input-string",         g_call_with_input_string,   2, 0, false, H_call_with_input_string);
  s7_define_function(sc, "call-with-input-file",           g_call_with_input_file,     2, 0, false, H_call_with_input_file);
  s7_define_function(sc, "with-input-from-string",         g_with_input_from_string,   2, 0, false, H_with_input_from_string);
  s7_define_function(sc, "with-input-from-file",           g_with_input_from_file,     2, 0, false, H_with_input_from_file);
  
  s7_define_function(sc, "call-with-output-string",        g_call_with_output_string,  1, 0, false, H_call_with_output_string);
  s7_define_function(sc, "call-with-output-file",          g_call_with_output_file,    2, 0, false, H_call_with_output_file);
  s7_define_function(sc, "with-output-to-string",          g_with_output_to_string,    1, 0, false, H_with_output_to_string);
  s7_define_function(sc, "with-output-to-file",            g_with_output_to_file,      2, 0, false, H_with_output_to_file);
  
  
#if (!WITH_GMP)
  s7_define_safe_function(sc, "make-polar",                g_make_polar,               2, 0, false, H_make_polar);
  s7_define_safe_function(sc, "make-rectangular",          g_make_rectangular,         2, 0, false, H_make_rectangular);
  s7_define_safe_function(sc, "magnitude",                 g_magnitude,                1, 0, false, H_magnitude);
  s7_define_safe_function(sc, "angle",                     g_angle,                    1, 0, false, H_angle);
  s7_define_safe_function(sc, "real-part",                 g_real_part,                1, 0, false, H_real_part);
  s7_define_safe_function(sc, "imag-part",                 g_imag_part,                1, 0, false, H_imag_part);
  s7_define_safe_function(sc, "numerator",                 g_numerator,                1, 0, false, H_numerator);
  s7_define_safe_function(sc, "denominator",               g_denominator,              1, 0, false, H_denominator);
  s7_define_safe_function(sc, "rationalize",               g_rationalize,              1, 1, false, H_rationalize);
  s7_define_safe_function(sc, "abs",                       g_abs,                      1, 0, false, H_abs);
  s7_define_safe_function(sc, "exp",                       g_exp,                      1, 0, false, H_exp);
  s7_define_safe_function(sc, "log",                       g_log,                      1, 1, false, H_log);
  s7_define_safe_function(sc, "sin",                       g_sin,                      1, 0, false, H_sin);
  s7_define_safe_function(sc, "cos",                       g_cos,                      1, 0, false, H_cos);
  s7_define_safe_function(sc, "tan",                       g_tan,                      1, 0, false, H_tan);
  s7_define_safe_function(sc, "asin",                      g_asin,                     1, 0, false, H_asin);
  s7_define_safe_function(sc, "acos",                      g_acos,                     1, 0, false, H_acos);
  s7_define_safe_function(sc, "atan",                      g_atan,                     1, 1, false, H_atan);
  s7_define_safe_function(sc, "sinh",                      g_sinh,                     1, 0, false, H_sinh);
  s7_define_safe_function(sc, "cosh",                      g_cosh,                     1, 0, false, H_cosh);
  s7_define_safe_function(sc, "tanh",                      g_tanh,                     1, 0, false, H_tanh);
  s7_define_safe_function(sc, "asinh",                     g_asinh,                    1, 0, false, H_asinh);
  s7_define_safe_function(sc, "acosh",                     g_acosh,                    1, 0, false, H_acosh);
  s7_define_safe_function(sc, "atanh",                     g_atanh,                    1, 0, false, H_atanh);
  s7_define_safe_function(sc, "sqrt",                      g_sqrt,                     1, 0, false, H_sqrt);
  s7_define_safe_function(sc, "expt",                      g_expt,                     2, 0, false, H_expt);
  s7_define_safe_function(sc, "floor",                     g_floor,                    1, 0, false, H_floor);
  s7_define_safe_function(sc, "ceiling",                   g_ceiling,                  1, 0, false, H_ceiling);
  s7_define_safe_function(sc, "truncate",                  g_truncate,                 1, 0, false, H_truncate);
  s7_define_safe_function(sc, "round",                     g_round,                    1, 0, false, H_round);
  s7_define_safe_function(sc, "lcm",                       g_lcm,                      0, 0, true,  H_lcm);
  s7_define_safe_function(sc, "gcd",                       g_gcd,                      0, 0, true,  H_gcd);
  s7_define_safe_function(sc, "+",                         g_add,                      0, 0, true,  H_add);
  s7_define_safe_function(sc, "-",                         g_subtract,                 1, 0, true,  H_subtract);
  s7_define_safe_function(sc, "*",                         g_multiply,                 0, 0, true,  H_multiply);
  s7_define_safe_function(sc, "/",                         g_divide,                   1, 0, true,  H_divide);
  s7_define_safe_function(sc, "max",                       g_max,                      1, 0, true,  H_max);
  s7_define_safe_function(sc, "min",                       g_min,                      1, 0, true,  H_min);
  s7_define_safe_function(sc, "quotient",                  g_quotient,                 2, 0, false, H_quotient);
  s7_define_safe_function(sc, "remainder",                 g_remainder,                2, 0, false, H_remainder);
  s7_define_safe_function(sc, "modulo",                    g_modulo,                   2, 0, false, H_modulo);
  s7_define_safe_function(sc, "=",                         g_equal,                    2, 0, true,  H_equal);
  s7_define_safe_function(sc, "<",                         g_less,                     2, 0, true,  H_less);
  s7_define_safe_function(sc, ">",                         g_greater,                  2, 0, true,  H_greater);
  s7_define_safe_function(sc, "<=",                        g_less_or_equal,            2, 0, true,  H_less_or_equal);
  s7_define_safe_function(sc, ">=",                        g_greater_or_equal,         2, 0, true,  H_greater_or_equal);
  s7_define_safe_function(sc, "even?",                     g_is_even,                  1, 0, false, H_is_even);
  s7_define_safe_function(sc, "odd?",                      g_is_odd,                   1, 0, false, H_is_odd);
  s7_define_safe_function(sc, "zero?",                     g_is_zero,                  1, 0, false, H_is_zero);
  s7_define_safe_function(sc, "positive?",                 g_is_positive,              1, 0, false, H_is_positive);
  s7_define_safe_function(sc, "negative?",                 g_is_negative,              1, 0, false, H_is_negative);
  s7_define_safe_function(sc, "infinite?",                 g_is_infinite,              1, 0, false, H_is_infinite);
  s7_define_safe_function(sc, "nan?",                      g_is_nan,                   1, 0, false, H_is_nan);

  s7_define_safe_function(sc, "inexact->exact",            g_inexact_to_exact,         1, 0, false, H_inexact_to_exact);
  s7_define_safe_function(sc, "exact->inexact",            g_exact_to_inexact,         1, 0, false, H_exact_to_inexact);

  s7_define_safe_function(sc, "random",                    g_random,                   1, 1, false, H_random);
  s7_define_safe_function(sc, "make-random-state",         s7_make_random_state,       1, 1, false, H_make_random_state);

  s7_define_safe_function(sc, "integer-length",            g_integer_length,           1, 0, false, H_integer_length);
  s7_define_safe_function(sc, "logior",                    g_logior,                   0, 0, true,  H_logior);
  s7_define_safe_function(sc, "logxor",                    g_logxor,                   0, 0, true,  H_logxor);
  s7_define_safe_function(sc, "logand",                    g_logand,                   0, 0, true,  H_logand);
  s7_define_safe_function(sc, "lognot",                    g_lognot,                   1, 0, false, H_lognot);
  s7_define_safe_function(sc, "ash",                       g_ash,                      2, 0, false, H_ash);
#endif
  s7_define_safe_function(sc, "logbit?",                   g_logbit,                   2, 0, false, H_logbit);

  s7_define_safe_function(sc, "random-state->list",        s7_random_state_to_list,    0, 1, false, H_random_state_to_list);
  s7_define_safe_function(sc, "integer-decode-float",      g_integer_decode_float,     1, 0, false, H_integer_decode_float);
  s7_define_safe_function(sc, "exact?",                    g_is_exact,                 1, 0, false, H_is_exact);
  s7_define_safe_function(sc, "inexact?",                  g_is_inexact,               1, 0, false, H_is_inexact);

  rng_tag = s7_new_type_x("<random-number-generator>", print_rng, free_rng, equal_rng, NULL, NULL, NULL, NULL, copy_random_state, NULL, NULL);

  s7_define_safe_function(sc, "number?",                   g_is_number,                1, 0, false, H_is_number);
  s7_define_safe_function(sc, "integer?",                  g_is_integer,               1, 0, false, H_is_integer);
  s7_define_safe_function(sc, "real?",                     g_is_real,                  1, 0, false, H_is_real);
  s7_define_safe_function(sc, "complex?",                  g_is_complex,               1, 0, false, H_is_complex);
  s7_define_safe_function(sc, "rational?",                 g_is_rational,              1, 0, false, H_is_rational);

  s7_define_safe_function(sc, "number->string",            g_number_to_string,         1, 1, false, H_number_to_string);
  s7_define_safe_function(sc, "string->number",            g_string_to_number,         1, 1, false, H_string_to_number);

  
  s7_define_safe_function(sc, "char-upcase",               g_char_upcase,              1, 0, false, H_char_upcase);
  s7_define_safe_function(sc, "char-downcase",             g_char_downcase,            1, 0, false, H_char_downcase);
  s7_define_safe_function(sc, "char->integer",             g_char_to_integer,          1, 0, false, H_char_to_integer);
  s7_define_safe_function(sc, "integer->char",             g_integer_to_char,          1, 0, false, H_integer_to_char);
  
  s7_define_safe_function(sc, "char-upper-case?",          g_is_char_upper_case,       1, 0, false, H_is_char_upper_case);
  s7_define_safe_function(sc, "char-lower-case?",          g_is_char_lower_case,       1, 0, false, H_is_char_lower_case);
  s7_define_safe_function(sc, "char-alphabetic?",          g_is_char_alphabetic,       1, 0, false, H_is_char_alphabetic);
  s7_define_safe_function(sc, "char-numeric?",             g_is_char_numeric,          1, 0, false, H_is_char_numeric);
  s7_define_safe_function(sc, "char-whitespace?",          g_is_char_whitespace,       1, 0, false, H_is_char_whitespace);
  s7_define_safe_function(sc, "char?",                     g_is_char,                  1, 0, false, H_is_char);
  
  s7_define_safe_function(sc, "char=?",                    g_chars_are_equal,          2, 0, true,  H_chars_are_equal);
  s7_define_safe_function(sc, "char<?",                    g_chars_are_less,           2, 0, true,  H_chars_are_less);
  s7_define_safe_function(sc, "char>?",                    g_chars_are_greater,        2, 0, true,  H_chars_are_greater);
  s7_define_safe_function(sc, "char<=?",                   g_chars_are_leq,            2, 0, true,  H_chars_are_leq);
  s7_define_safe_function(sc, "char>=?",                   g_chars_are_geq,            2, 0, true,  H_chars_are_geq);
  s7_define_safe_function(sc, "char-ci=?",                 g_chars_are_ci_equal,       2, 0, true,  H_chars_are_ci_equal);
  s7_define_safe_function(sc, "char-ci<?",                 g_chars_are_ci_less,        2, 0, true,  H_chars_are_ci_less);
  s7_define_safe_function(sc, "char-ci>?",                 g_chars_are_ci_greater,     2, 0, true,  H_chars_are_ci_greater);
  s7_define_safe_function(sc, "char-ci<=?",                g_chars_are_ci_leq,         2, 0, true,  H_chars_are_ci_leq);
  s7_define_safe_function(sc, "char-ci>=?",                g_chars_are_ci_geq,         2, 0, true,  H_chars_are_ci_geq);
  
  
  s7_define_safe_function(sc, "string?",                   g_is_string,                1, 0, false, H_is_string);
  s7_define_safe_function(sc, "make-string",               g_make_string,              1, 1, false, H_make_string);
  s7_define_safe_function(sc, "string-length",             g_string_length,            1, 0, false, H_string_length);
  s7_define_safe_function(sc, "string-ref",                g_string_ref,               2, 0, false, H_string_ref);
  s7_define_safe_function(sc, "string-set!",               g_string_set,               3, 0, false, H_string_set);
  s7_define_safe_function(sc, "string=?",                  g_strings_are_equal,        2, 0, true,  H_strings_are_equal);
  s7_define_safe_function(sc, "string<?",                  g_strings_are_less,         2, 0, true,  H_strings_are_less);
  s7_define_safe_function(sc, "string>?",                  g_strings_are_greater,      2, 0, true,  H_strings_are_greater);
  s7_define_safe_function(sc, "string<=?",                 g_strings_are_leq,          2, 0, true,  H_strings_are_leq);
  s7_define_safe_function(sc, "string>=?",                 g_strings_are_geq,          2, 0, true,  H_strings_are_geq);
  s7_define_safe_function(sc, "string-ci=?",               g_strings_are_ci_equal,     2, 0, true,  H_strings_are_ci_equal);
  s7_define_safe_function(sc, "string-ci<?",               g_strings_are_ci_less,      2, 0, true,  H_strings_are_ci_less);
  s7_define_safe_function(sc, "string-ci>?",               g_strings_are_ci_greater,   2, 0, true,  H_strings_are_ci_greater);
  s7_define_safe_function(sc, "string-ci<=?",              g_strings_are_ci_leq,       2, 0, true,  H_strings_are_ci_leq);
  s7_define_safe_function(sc, "string-ci>=?",              g_strings_are_ci_geq,       2, 0, true,  H_strings_are_ci_geq);
  
  s7_define_safe_function(sc, "string-append",             g_string_append,            0, 0, true,  H_string_append);
  s7_define_safe_function(sc, "string-fill!",              g_string_fill,              2, 0, false, H_string_fill);
  s7_define_safe_function(sc, "string-copy",               g_string_copy,              1, 0, false, H_string_copy);
  s7_define_safe_function(sc, "substring",                 g_substring,                2, 1, false, H_substring);
  s7_define_safe_function(sc, "string",                    g_string,                   0, 0, true,  H_string);
  s7_define_safe_function(sc, "list->string",              g_list_to_string,           1, 0, false, H_list_to_string);
  s7_define_safe_function(sc, "string->list",              g_string_to_list,           1, 0, false, H_string_to_list);
  s7_define_safe_function(sc, "object->string",            g_object_to_string,         1, 1, false, H_object_to_string);
  s7_define_safe_function(sc, "format",                    g_format,                   1, 0, true,  H_format);


  s7_define_safe_function(sc, "null?",                     g_is_null,                  1, 0, false, H_is_null);
  s7_define_safe_function(sc, "list?",                     g_is_list,                  1, 0, false, H_is_list);
  s7_define_safe_function(sc, "pair?",                     g_is_pair,                  1, 0, false, H_is_pair);
  s7_define_safe_function(sc, "cons",                      g_cons,                     2, 0, false, H_cons);
  s7_define_safe_function(sc, "car",                       g_car,                      1, 0, false, H_car);
  s7_define_safe_function(sc, "cdr",                       g_cdr,                      1, 0, false, H_cdr);
  /* possibly a bad idea -- macro expander has built-in cdr */

  s7_define_function(sc, "set-car!",                       g_set_car,                  2, 0, false, H_set_car);
  s7_define_function(sc, "set-cdr!",                       g_set_cdr,                  2, 0, false, H_set_cdr);
  s7_define_safe_function(sc, "caar",                      g_caar,                     1, 0, false, H_caar);
  s7_define_safe_function(sc, "cadr",                      g_cadr,                     1, 0, false, H_cadr);
  s7_define_safe_function(sc, "cdar",                      g_cdar,                     1, 0, false, H_cdar);
  s7_define_safe_function(sc, "cddr",                      g_cddr,                     1, 0, false, H_cddr);
  s7_define_safe_function(sc, "caaar",                     g_caaar,                    1, 0, false, H_caaar);
  s7_define_safe_function(sc, "caadr",                     g_caadr,                    1, 0, false, H_caadr);
  s7_define_safe_function(sc, "cadar",                     g_cadar,                    1, 0, false, H_cadar);
  s7_define_safe_function(sc, "cdaar",                     g_cdaar,                    1, 0, false, H_cdaar);
  s7_define_safe_function(sc, "caddr",                     g_caddr,                    1, 0, false, H_caddr);
  s7_define_safe_function(sc, "cdddr",                     g_cdddr,                    1, 0, false, H_cdddr);
  s7_define_safe_function(sc, "cdadr",                     g_cdadr,                    1, 0, false, H_cdadr);
  s7_define_safe_function(sc, "cddar",                     g_cddar,                    1, 0, false, H_cddar);
  s7_define_safe_function(sc, "caaaar",                    g_caaaar,                   1, 0, false, H_caaaar);
  s7_define_safe_function(sc, "caaadr",                    g_caaadr,                   1, 0, false, H_caaadr);
  s7_define_safe_function(sc, "caadar",                    g_caadar,                   1, 0, false, H_caadar);
  s7_define_safe_function(sc, "cadaar",                    g_cadaar,                   1, 0, false, H_cadaar);
  s7_define_safe_function(sc, "caaddr",                    g_caaddr,                   1, 0, false, H_caaddr);
  s7_define_safe_function(sc, "cadddr",                    g_cadddr,                   1, 0, false, H_cadddr);
  s7_define_safe_function(sc, "cadadr",                    g_cadadr,                   1, 0, false, H_cadadr);
  s7_define_safe_function(sc, "caddar",                    g_caddar,                   1, 0, false, H_caddar);
  s7_define_safe_function(sc, "cdaaar",                    g_cdaaar,                   1, 0, false, H_cdaaar);
  s7_define_safe_function(sc, "cdaadr",                    g_cdaadr,                   1, 0, false, H_cdaadr);
  s7_define_safe_function(sc, "cdadar",                    g_cdadar,                   1, 0, false, H_cdadar);
  s7_define_safe_function(sc, "cddaar",                    g_cddaar,                   1, 0, false, H_cddaar);
  s7_define_safe_function(sc, "cdaddr",                    g_cdaddr,                   1, 0, false, H_cdaddr);
  s7_define_safe_function(sc, "cddddr",                    g_cddddr,                   1, 0, false, H_cddddr);
  s7_define_safe_function(sc, "cddadr",                    g_cddadr,                   1, 0, false, H_cddadr);
  s7_define_safe_function(sc, "cdddar",                    g_cdddar,                   1, 0, false, H_cdddar);
  s7_define_safe_function(sc, "assq",                      g_assq,                     2, 0, false, H_assq);
  s7_define_safe_function(sc, "assv",                      g_assv,                     2, 0, false, H_assv);
  s7_define_function(sc, "assoc",                          g_assoc,                    2, 1, false, H_assoc);
  s7_define_safe_function(sc, "memq",                      g_memq,                     2, 0, false, H_memq);
  s7_define_safe_function(sc, "memv",                      g_memv,                     2, 0, false, H_memv);
  s7_define_function(sc, "member",                         g_member,                   2, 1, false, H_member);
  s7_define_safe_function(sc, "append",                    g_append,                   0, 0, true,  H_append);
  s7_define_function(sc, "list",                           g_list,                     0, 0, true,  H_list);
  {
    s7_pointer p;
    p = s7_symbol_value(sc, make_symbol(sc, "list"));
    set_copy_args(p);
  }
  s7_define_safe_function(sc, "list-ref",                  g_list_ref,                 2, 0, true,  H_list_ref);
  s7_define_safe_function(sc, "list-set!",                 g_list_set,                 3, 0, true,  H_list_set);
  s7_define_function(sc, "list-tail",                      g_list_tail,                2, 0, false, H_list_tail);
  /* perhaps with setter? */
  s7_define_safe_function(sc, "make-list",                 g_make_list,                1, 1, false, H_make_list);

  s7_define_safe_function(sc, "length",                    g_length,                   1, 0, false, H_length);
  s7_define_safe_function(sc, "copy",                      g_copy,                     1, 0, false, H_copy);
  s7_define_function(sc, "fill!",                          g_fill,                     2, 0, false, H_fill);
  s7_define_safe_function(sc, "reverse",                   g_reverse,                  1, 0, false, H_reverse);
  s7_define_function(sc, "reverse!",                       g_reverse_in_place,         1, 0, false, H_reverse_in_place); /* used by Snd code */
  

  s7_define_safe_function(sc, "vector?",                   g_is_vector,                1, 0, false, H_is_vector);
  s7_define_safe_function(sc, "vector->list",              g_vector_to_list,           1, 0, false, H_vector_to_list);
  s7_define_safe_function(sc, "list->vector",              g_list_to_vector,           1, 0, false, H_list_to_vector);
  s7_define_safe_function(sc, "vector-fill!",              g_vector_fill,              2, 0, false, H_vector_fill);
  s7_define_safe_function(sc, "vector",                    g_vector,                   0, 0, true,  H_vector);
#if WITH_OPTIMIZATION
  {
    s7_pointer f;
    f = s7_symbol_value(sc, sc->VECTOR);
    set_setter(f); /* for optimized DO's benefit */
  }
#endif
  s7_define_safe_function(sc, "vector-length",             g_vector_length,            1, 0, false, H_vector_length);
  s7_define_safe_function(sc, "vector-ref",                g_vector_ref,               2, 0, true,  H_vector_ref);
  s7_define_safe_function(sc, "vector-set!",               g_vector_set,               3, 0, true,  H_vector_set);
  s7_define_safe_function(sc, "make-vector",               g_make_vector,              1, 1, false, H_make_vector);
  s7_define_safe_function(sc, "vector-dimensions",         g_vector_dimensions,        1, 0, false, H_vector_dimensions);
  s7_define_function(sc, "sort!",                          g_sort,                     2, 0, false, H_sort);


  s7_define_safe_function(sc, "hash-table",                g_hash_table,               0, 0, true,  H_hash_table);
  s7_define_safe_function(sc, "hash-table?",               g_is_hash_table,            1, 0, false, H_is_hash_table);
  s7_define_safe_function(sc, "make-hash-table",           g_make_hash_table,          0, 1, false, H_make_hash_table);
  s7_define_safe_function(sc, "hash-table-ref",            g_hash_table_ref,           2, 0, true,  H_hash_table_ref);
  s7_define_safe_function(sc, "hash-table-set!",           g_hash_table_set,           3, 0, false, H_hash_table_set);
  s7_define_safe_function(sc, "hash-table-size",           g_hash_table_size,          1, 0, false, H_hash_table_size);
  s7_define_safe_function(sc, "make-hash-table-iterator",  g_make_hash_table_iterator, 1, 0, false, H_make_hash_table_iterator);


  s7_define_safe_function(sc, "hook?",                     g_is_hook,                  1, 0, false, H_is_hook);
  s7_define_safe_function(sc, "make-hook",                 g_make_hook,                0, 2, false, H_make_hook);
  s7_define_function(sc, "hook",                           g_hook,                     0, 0, true,  H_hook);
  s7_define_function(sc, "hook-apply",                     g_hook_apply,               1, 0, true,  H_hook_apply);
  s7_define_safe_function(sc, "hook-arity",                g_hook_arity,               1, 0, false, H_hook_arity);
  s7_define_safe_function(sc, "hook-documentation",        g_hook_documentation,       1, 0, false, H_hook_documentation);
  s7_define_function_with_setter(sc, "hook-functions",     g_hook_functions, g_hook_set_functions, 1, 0, H_hook_functions);

  s7_define_function(sc, "call/cc",                        g_call_cc,                  1, 0, false, H_call_cc);
  s7_define_function(sc, "call-with-current-continuation", g_call_cc,                  1, 0, false, H_call_cc);
  s7_define_safe_function(sc, "continuation?",             g_is_continuation,          1, 0, false, H_is_continuation);
  s7_define_function(sc, "call-with-exit",                 g_call_with_exit,           1, 0, false, H_call_with_exit);

  s7_define_function(sc, "load",                           g_load,                     1, 1, false, H_load);
  s7_define_function(sc, "eval",                           g_eval,                     1, 1, false, H_eval);
  s7_define_function(sc, "eval-string",                    g_eval_string,              1, 1, false, H_eval_string);
  s7_define_function(sc, "apply",                          g_apply,                    1, 0, true,  H_apply);

  s7_define_function(sc, "for-each",                       g_for_each,                 2, 0, true,  H_for_each);
  s7_define_function(sc, "map",                            g_map,                      2, 0, true,  H_map);

  s7_define_function(sc, "values",                         g_values,                   0, 0, true,  H_values);
  s7_define_function(sc, "dynamic-wind",                   g_dynamic_wind,             3, 0, false, H_dynamic_wind);
  s7_define_function(sc, "catch",                          g_catch,                    3, 0, false, H_catch);
  s7_define_function(sc, "error",                          g_error,                    0, 0, true,  H_error);

  /* these are internal for quasiquote's use */
  s7_define_constant_function(sc, "{values}",              g_qq_values,                0, 0, true,  H_qq_values);
  s7_define_constant_function(sc, "{apply}",               g_apply,                    1, 0, true,  H_apply);
  s7_define_constant_function(sc, "{apply} {values}",      g_apply_values,             0, 0, true,  H_apply_values);
  s7_define_constant_function(sc, "{append}",              g_append,                   0, 0, true,  H_append);
  s7_define_constant_function(sc, "{multivector}",         g_qq_multivector,           1, 0, true,  H_qq_multivector);
  s7_define_constant_function(sc, "{list}",                g_qq_list,                  0, 0, true,  H_qq_list);
  {
    s7_pointer p;
    p = s7_symbol_value(sc, make_symbol(sc, "{list}"));
    set_type(p, (T_C_LST_ARGS_FUNCTION | T_PROCEDURE | T_COPY_ARGS));
  }

  s7_define_safe_function(sc, "stacktrace",                g_stacktrace,               0, 2, false, H_stacktrace);
  s7_define_safe_function(sc, "trace",                     g_trace,                    0, 0, true,  H_trace);
  s7_define_safe_function(sc, "untrace",                   g_untrace,                  0, 0, true,  H_untrace);
  s7_define_safe_function(sc, "gc",                        g_gc,                       0, 1, false, H_gc);

  s7_define_safe_function(sc, "procedure?",                g_is_procedure,             1, 0, false, H_is_procedure);
  s7_define_safe_function(sc, "procedure-documentation",   g_procedure_documentation,  1, 0, false, H_procedure_documentation);
  s7_define_safe_function(sc, "help",                      g_help,                     1, 0, false, H_help);
  s7_define_safe_function(sc, "procedure-arity",           g_procedure_arity,          1, 0, false, H_procedure_arity);
  s7_define_safe_function(sc, "procedure-source",          g_procedure_source,         1, 0, false, H_procedure_source);
  s7_define_safe_function(sc, "procedure-environment",     g_procedure_environment,    1, 0, false, H_procedure_environment);
  s7_define_safe_function(sc, "unoptimize",                g_unoptimize,               1, 0, false, H_unoptimize);
  s7_define_safe_function(sc, "procedure-setter",          g_procedure_setter,         1, 0, false, H_procedure_setter);
  s7_define_safe_function(sc, "procedure-name",            g_procedure_name,           1, 0, false, H_procedure_name);
  
  s7_define_safe_function(sc, "not",                       g_not,                      1, 0, false, H_not);
  s7_define_safe_function(sc, "boolean?",                  g_is_boolean,               1, 0, false, H_is_boolean);
  s7_define_safe_function(sc, "eq?",                       g_is_eq,                    2, 0, false, H_is_eq);
  s7_define_safe_function(sc, "eqv?",                      g_is_eqv,                   2, 0, false, H_is_eqv);
  s7_define_safe_function(sc, "equal?",                    g_is_equal,                 2, 0, false, H_is_equal);
  
  s7_define_safe_function(sc, "s7-version",                g_s7_version,               0, 0, false, H_s7_version);

  #define object_set_name "(generalized set!)"
  s7_define_function(sc, object_set_name,                  g_internal_object_set,      1, 0, true,  "internal object setter redirection");
  sc->OBJECT_SET = s7_symbol_value(sc, make_symbol(sc, object_set_name));

  s7_define_function(sc, s_is_type_name,                   s_is_type,                  2, 0, false, "internal object type check");
  s7_define_function(sc, s_type_make_name,                 s_type_make,                2, 0, false, "internal object creation");
  s7_define_function(sc, s_type_ref_name,                  s_type_ref,                 2, 0, false, "internal object value");

  s7_define_function_star(sc, "make-type", g_make_type, "print equal getter setter length name copy reverse fill", H_make_type);
  set_global(make_symbol(sc, "length"));
  set_global(make_symbol(sc, "copy"));

  /* -------- *gc-stats* -------- */
  s7_define_variable(sc, "*gc-stats*", sc->F);
  s7_symbol_set_access(sc, s7_make_symbol(sc, "*gc-stats*"), 
		       list_3(sc, 
			      sc->F, 
			      s7_make_function(sc, "(set *gc-stats*)", g_gc_stats_set, 2, 0, false, "called if *gc-stats* is set"), 
			      s7_make_function(sc, "(bind *gc-stats*)", g_gc_stats_set, 2, 0, false, "called if *gc-stats* is bound")));

  /* -------- *features* -------- */
  s7_define_variable(sc, "*features*", sc->NIL);
  s7_symbol_set_access(sc, s7_make_symbol(sc, "*features*"), 
		       list_3(sc, 
			      sc->F, 
			      s7_make_function(sc, "(set *features*)", g_features_set, 2, 0, false, "called if *features* is set"), 
			      s7_make_function(sc, "(bind *features*)", g_features_set, 2, 0, false, "called if *features* is bound")));


  /* these hook variables should use s7_define_constant, but we need to be backwards compatible */

  /* -------- *load-hook* -------- */
  sc->load_hook = s7_make_hook(sc, 1, 0, false, "*load-hook* is called when a file is loaded.  Its functions take one argument, the filename.");
  s7_define_variable(sc, "*load-hook*", sc->load_hook);
  s7_symbol_set_access(sc, s7_make_symbol(sc, "*load-hook*"), 
		       list_3(sc, 
			      sc->F, 
			      s7_make_function(sc, "(set *load-hook*)", g_load_hook_set, 2, 0, false, 
					       "called if *load-hook* is set"), 
			      sc->F));

  /* -------- *trace-hook* -------- */
  sc->trace_hook = s7_make_hook(sc, 2, 0, false, 
				"*trace-hook* customizes tracing.  Its functions take 2 arguments, the function being traced, and its current arguments.");
  s7_define_variable(sc, "*trace-hook*", sc->trace_hook); 
  s7_symbol_set_access(sc, s7_make_symbol(sc, "*trace-hook*"), 
		       list_3(sc, 
			      sc->F, 
			      s7_make_function(sc, "(set *trace-hook*)", g_trace_hook_set, 2, 0, false, 
					       "called if *trace-hook* is set"), 
			      sc->F));
  
  /* -------- *unbound-variable-hook* -------- */
  sc->unbound_variable_hook = s7_make_hook(sc, 1, 0, false, "*unbound-variable-hook* is called when an unbound variable is encountered.  Its functions \
take 1 argument, the unbound symbol.");
  s7_define_variable(sc, "*unbound-variable-hook*", sc->unbound_variable_hook); 
  s7_symbol_set_access(sc, s7_make_symbol(sc, "*unbound-variable-hook*"), 
		       list_3(sc, 
			      sc->F, 
			      s7_make_function(sc, "(set *unbound-variable-hook*)", g_unbound_variable_hook_set, 2, 0, false, 
					       "called if *unbound-variable-hook* is set"), 
			      sc->F));
  
  /* -------- *error-hook* -------- */
  sc->error_hook = s7_make_hook(sc, 2, 0, false, "*error-hook* is called when an error is not caught.  Its functions take two arguments, \
the error type and the info passed to the error handler.");
  s7_define_variable(sc, "*error-hook*", sc->error_hook);
  s7_symbol_set_access(sc, s7_make_symbol(sc, "*error-hook*"), 
		       list_3(sc, 
			      sc->F, 
			      s7_make_function(sc, "(set *error-hook*)", g_error_hook_set, 2, 0, false, 
					       "called if *error-hook* is set"), 
			      sc->F));

  /* -------- *vector-print-length* -------- */
  s7_define_variable(sc, "*vector-print-length*", small_int(8));
  sc->vector_print_length = global_slot(make_symbol(sc, "*vector-print-length*"));
  s7_symbol_set_access(sc, s7_make_symbol(sc, "*vector-print-length*"), 
		       list_3(sc, 
			      sc->F, 
			      s7_make_function(sc, "(set *vector-print-length*)", g_vector_print_length_set, 2, 0, false, 
					       "called if *vector-print-length* is set"), 
			      sc->F));

  /* -------- *safety* -------- */
  s7_define_variable(sc, "*safety*", small_int(sc->safety));
  s7_symbol_set_access(sc, s7_make_symbol(sc, "*safety*"), 
		       list_3(sc, 
			      sc->F, 
			      s7_make_function(sc, "(set *safety*)", g_safety_set, 2, 0, false, "called if *safety* is set"), 
			      s7_make_function(sc, "(bind *safety*)", g_safety_bind, 2, 0, false, "called if *safety* is bound")));


  /* -------- *load-path* -------- */
  s7_define_variable(sc, "*load-path*", sc->NIL);
  s7_symbol_set_access(sc, s7_make_symbol(sc, "*load-path*"), 
		       list_3(sc, 
			      sc->F, 
			      s7_make_function(sc, "(set *load-path*)", g_load_path_set, 2, 0, false, "called if *load-path* is set"), 
			      s7_make_function(sc, "(bind *load-path*)", g_load_path_set, 2, 0, false, "called if *load-path* is bound")));

  /* -------- *#readers* -------- */
  s7_define_variable(sc, "*#readers*", sc->NIL);
  sc->sharp_readers = global_slot(make_symbol(sc, "*#readers*"));
  s7_symbol_set_access(sc, s7_make_symbol(sc, "*#readers*"), 
		       list_3(sc, 
			      sc->F, 
			      s7_make_function(sc, "(set *#readers*)", g_sharp_readers_set, 2, 0, false, "called if *#readers* is set"), 
			      s7_make_function(sc, "(bind *#readers*)", g_sharp_readers_set, 2, 0, false, "called if *#readers* is bound")));


  /* -------- *error-info* -------- */
  sc->error_info = s7_make_and_fill_vector(sc, ERROR_INFO_SIZE, ERROR_INFO_DEFAULT);
  s7_define_constant(sc, "*error-info*", sc->error_info);


  /* the next two are for the test suite */
  s7_make_procedure_with_setter(sc, "-s7-symbol-table-locked?", 
				g_symbol_table_is_locked, 0, 0, 
				g_set_symbol_table_is_locked, 1, 0, 
				H_symbol_table_is_locked);
  s7_define_safe_function(sc, "-s7-stack-size", g_stack_size, 0, 0, false, "current stack size");

  s7_provide(sc, "s7");

#if WITH_EXTRA_EXPONENT_MARKERS
  s7_provide(sc, "dfls-exponents");
#endif

  {
    s7_pointer sym;
    sc->VECTOR_SET = s7_symbol_value(sc, sym = make_symbol(sc, "vector-set!"));
    typeflag(sc->VECTOR_SET) |= T_SETTER; 
    set_setter(sym);
    
    sc->LIST_SET = s7_symbol_value(sc, sym = make_symbol(sc, "list-set!"));
    typeflag(sc->LIST_SET) |= T_SETTER; 
    set_setter(sym);
    
    sc->HASH_TABLE_SET = s7_symbol_value(sc, sym = make_symbol(sc, "hash-table-set!"));
    typeflag(sc->HASH_TABLE_SET) |= T_SETTER; 
    set_setter(sym);

    sc->ENVIRONMENT_SET = s7_make_function(sc, "(environment-set!)", g_environment_set, 3, 0, false, "internal setter for environments");
    typeflag(sc->ENVIRONMENT_SET) |= T_SETTER; 
    set_setter(sym);

    sym = make_symbol(sc, "cons");
    set_setter(sym);
    sym = s7_symbol_value(sc, sym);
    set_setter(sym);
    
    sc->HASH_TABLE_ITERATE = s7_make_function(sc, "(hash-table iterate)", g_hash_table_iterate, 1, 0, false, "internal hash-table iteration function");
    sc->STRING_SET = s7_symbol_value(sc, make_symbol(sc, "string-set!"));
  }

  s7_function_set_setter(sc, "car",                 "set-car!");
  s7_function_set_setter(sc, "cdr",                 "set-cdr!");
  s7_function_set_setter(sc, "hash-table-ref",      "hash-table-set!");
  s7_function_set_setter(sc, "vector-ref",          "vector-set!");
  s7_function_set_setter(sc, "list-ref",            "list-set!");
  s7_function_set_setter(sc, "string-ref",          "string-set!");
  s7_function_set_setter(sc, "current-input-port",  "set-current-input-port");
  s7_function_set_setter(sc, "current-output-port", "set-current-output-port");
  s7_function_set_setter(sc, "current-error-port",  "set-current-error-port");

#if WITH_OPTIMIZATION
  init_choosers(sc);
#endif

  {
    int i, top;
#if WITH_GMP
    #define S7_LOG_LLONG_MAX 36.736800
    #define S7_LOG_LONG_MAX  16.6355322
#else
    /* actually not safe = (log (- (expt 2 63) 1)) and (log (- (expt 2 31) 1)) 
     *   (using 63 and 31 bits)
     */
    #define S7_LOG_LLONG_MAX 43.668274
    #define S7_LOG_LONG_MAX  21.487562
#endif

    top = sizeof(s7_Int);
    s7_int_max = (top == 8) ? S7_LONG_MAX : S7_SHORT_MAX;
    s7_int_min = (top == 8) ? S7_LONG_MIN : S7_SHORT_MIN;
    s7_int_bits = (top == 8) ? 63 : 31;
    s7_int_digits = (top == 8) ? 18 : 8;

    s7_Int_max = (top == 8) ? S7_LLONG_MAX : S7_LONG_MAX;
    s7_Int_min = (top == 8) ? S7_LLONG_MIN : S7_LONG_MIN;

    s7_int_digits_by_radix[0] = 0;
    s7_int_digits_by_radix[1] = 0;

    for (i = 2; i < 17; i++)
      s7_int_digits_by_radix[i] = (int)(floor(((top == 8) ? S7_LOG_LLONG_MAX : S7_LOG_LONG_MAX) / log((double)i)));

    s7_define_constant(sc, "most-positive-fixnum", make_permanent_integer((top == 8) ? S7_LLONG_MAX : ((top == 4) ? S7_LONG_MAX : S7_SHORT_MAX)));
    s7_define_constant(sc, "most-negative-fixnum", make_permanent_integer((top == 8) ? S7_LLONG_MIN : ((top == 4) ? S7_LONG_MIN : S7_SHORT_MIN)));

    if (top == 4) default_rationalize_error = 1.0e-6;
    s7_define_constant(sc, "pi", real_pi);

    /* for s7_Double, float gives about 9 digits, double 18, long Double claims 28 but I don't see more than about 22? */
  }

#if WITH_GMP
  s7_gmp_init(sc);
#endif

#if WITH_DUMP_HEAP  
  s7_define_function(sc, "dump-heap", g_dump_heap, 0, 0, false, "write heap contents to heap.data"); 
#endif

  /* macroexpand 
   *   needs to be a bacro so locally defined macros are expanded:
   *   (let () (define-macro (hi a) `(+ ,a 1)) (macroexpand (hi 2)))
   */
  s7_eval_c_string(sc, "(define-bacro (macroexpand __mac__) `(,(procedure-source (car __mac__)) ',__mac__))");
  /*
   * it's possible to clobber this definition via (for example):
   *   (set-car! (cddr (procedure-source macroexpand)) '(0))
   */
  
  /* quasiquote
   */
  s7_define_macro(sc, "quasiquote", g_quasiquote, 1, 0, false, H_quasiquote);

  /* letrec* -- the less said the better...
   */
  s7_eval_c_string(sc, "(define-macro (letrec* bindings . body)                            \n\
                          (if (null? body)                                                 \n\
                              (error 'syntax-error \"letrec* has no body\")                \n\
                              (if (not (list? bindings))                                   \n\
                                  (error 'syntax-error \"letrec* bindings are messed up\") \n\
                                  `(let (,@(map (lambda (var&init)                         \n\
                                                  (list (car var&init) #<undefined>))      \n\
                                                bindings))                                 \n\
                                     ,@(map (lambda (var&init)                             \n\
                                              (if (not (null? (cddr var&init)))            \n\
                                                  (error 'syntax-error \"letrec* variable has more than one value\")) \n\
                                              (list 'set! (car var&init) (cadr var&init))) \n\
                                             bindings)                                     \n\
                                     ,@body))))");


  /* call-with-values is almost a no-op in this context */
  s7_eval_c_string(sc, "(define-macro (call-with-values producer consumer) `(,consumer (,producer)))"); 
  /* (call-with-values (lambda () (values 1 2 3)) +) */

  s7_eval_c_string(sc, "(define-macro (multiple-value-bind vars expression . body) `((lambda ,vars ,@body) ,expression))");
  /* (multiple-value-bind (a b) (values 1 2) (+ a b)) */
  /*   named "receive" in srfi-8 which strikes me as perverse */

  s7_eval_c_string(sc, "(define-macro (multiple-value-set! vars expr . body)   \n\
                          (let ((local-vars (map (lambda (n) (gensym)) vars))) \n\
                            `((lambda ,local-vars ,@(map (lambda (n ln) `(set! ,n ,ln)) vars local-vars) ,@body) ,expr)))");


  /* cond-expand (uses *features*) */
  s7_eval_c_string(sc, "(define-macro (cond-expand . clauses)              \n\
                          ;; taken from MIT scheme?                        \n\
                          (define (got? fr)                                \n\
                            (cond                                          \n\
                             ((memq fr *features*) #t)                     \n\
                             ((not (pair? fr)) #f)                         \n\
                             ((eq? 'and (car fr))                          \n\
                              (let loop ((clauses (cdr fr)))               \n\
                        	(or (null? clauses)                        \n\
                        	    (and (got? (car clauses))              \n\
                        		 (loop (cdr clauses))))))          \n\
                             ((eq? 'or (car fr))                           \n\
                              (let loop ((clauses (cdr fr)))               \n\
                        	(and (pair? clauses)                       \n\
                        	     (or (got? (car clauses))              \n\
                        		 (loop (cdr clauses))))))          \n\
                             ((eq? 'not (car fr))                          \n\
                              (not (got? (and (pair? (cdr fr))             \n\
                        		      (cadr fr)))))                \n\
                             (else #f)))                                   \n\
                          (let loop ((clauses clauses))                    \n\
                            (if (pair? clauses)                            \n\
                        	(let* ((feature (if (pair? (car clauses))  \n\
                        			    (caar clauses)         \n\
                        			    (error 'wrong-type-arg \"cond-expand clause ~A is not a list\" (car clauses))))\n\
                        	       (code (cons 'begin (cdar clauses))))\n\
                        	  (cond                                    \n\
                        	   ((and (eq? 'else feature)               \n\
                        		 (null? (cdr clauses)))            \n\
                        	    code)                                  \n\
                        	   ((got? (unoptimize feature))            \n\
                        	    code)                                  \n\
                        	   (else (loop (cdr clauses))))))))");

  /* fprintf(stderr, "size: %d %d, max op: %d\n", (int)sizeof(s7_cell), (int)sizeof(s7_extended_cell), OP_MAX_DEFINED); */
  /* 64 bit machine: size: 48 72, max op: 263 */
#if WITH_GMP
  /* fprintf(stderr, "sizes: %d %d %d %d\n", sizeof(mpz_t), sizeof(mpq_t), sizeof(mpfr_t), sizeof(mpc_t)); */
  /* sizes: 16 32 32 64 
   *   which means if we embed mpc_t in the s7_cell number union, it will grow by 24 bytes? 72 from 48
   */
#endif
  save_initial_environment(sc);

  return(sc);
}


/* should we fix this?
 *    :(let ((lst '("12" "34"))) (lst 0 1))
 *    ;list-ref argument 1, "12", is string but should be a pair
 *    ;    (lst 0 1)
 *    :(let ((lst '("12" "34"))) ((lst 0) 1))
 *    #\2
 *    (let ((lst (list #(1 2) #(3 4)))) (lst 0 1))
 *    ;list-ref argument 1, #(1 2), is vector but should be a pair
 *    ;    (lst 0 1)
 */

/* TODO: the symbol-access stuff is not fully implemented
 *       t342.scm for tests, augment env, closure arg names, do step?
 *       what about recursion during this process (i.e. ref to accessed var in accessor)? -- infinite loop possible here!
 *       block recursive call on accessor? -- how? (no dynamic vars, so no clean way to disable)
 *       are optimized calls ok in this regard?
 *       all call_symbol_bind uses really should be embedded
 *       call gc in the symbol access stuff and unbound variable to flush out bugs [or eval-string?]
 *
 * (set! (procedure-setter abs) ...)?
 *
 * other uses of s7_call: all the object stuff [see note in that section], readers, unbound_variable
 *
 * these are currently scarcely ever used: SAFE_C_opQSq C_XDX
 *
 * lint     13424 -> 1231 [1237]
 * bench    52019 -> 7875 [8268]
 * index    44300 -> 4988 [4992]
 */
